Package: fasterRaster
Type: Package
Title: Faster Raster and Spatial Vector Processing Using 'GRASS'
Version: 8.4.1.2
Date: 2026-04-17
Authors@R: 
	c(
		person(
			given = c('Adam', 'B.'),
			family = 'Smith',
			role = c('cre', 'aut'),
			email = 'adam.smith@mobot.org',
			comment = c(ORCID = '0000-0002-6420-1659')
		)
	)
Maintainer: Adam B. Smith <adam.smith@mobot.org>
Description: Processing of large-in-memory/large-on disk rasters and spatial
	vectors using 'GRASS' <https://grass.osgeo.org/>. Most functions in the
	'terra' package are	recreated. Processing of medium-sized and smaller
	spatial objects will nearly always be faster using 'terra' or 'sf', but
	for large-in-memory/large-on-disk objects, 'fasterRaster' may be faster.
	To use most of the functions, you must have the stand-alone version (not
	the 'OSGeoW4' installer version) of 'GRASS' 8.0 or higher.
Depends: R (>= 4.0.0)
Imports: data.table (>= 1.14.8), DT, graphics, grDevices, methods,
        omnibus (>= 1.2.15), rgrass (>= 0.3-9), sf, shiny, terra (>=
        1.7), utils
Suggests: knitr, rmarkdown
License: GPL (>= 3)
SystemRequirements: GRASS (>= 8)
URL: https://github.com/adamlilith/fasterRaster,
        https://adamlilith.github.io/fasterRaster/
BugReports: https://github.com/adamlilith/fasterRaster/issues
VignetteBuilder: knitr
Encoding: UTF-8
LazyData: true
LazyLoad: yes
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-04-17 21:08:09 UTC; adame
Author: Adam B. Smith [cre, aut] (ORCID:
    <https://orcid.org/0000-0002-6420-1659>)
Repository: CRAN
Date/Publication: 2026-04-17 21:50:02 UTC
Built: R 4.6.0; ; 2026-04-29 17:20:19 UTC; unix

# Generated by roxygen2: do not edit by hand

export(.g.proj)
export(.g.region)
export(addons)
export(appFuns)
export(fastData)
export(faster)
export(grassHelp)
export(grassInfo)
export(grassStarted)
export(installAddon)
export(is.polygons)
export(makeGRaster)
export(makeGVector)
export(mow)
export(pcs)
export(print)
export(removeAddon)
export(seqToSQL)
export(st_coordinates)
export(st_crs)
export(sun)
exportClasses(GLocation)
exportClasses(GRaster)
exportClasses(GRegion)
exportClasses(GSpatial)
exportClasses(GVector)
exportMethods("$")
exportMethods("$<-")
exportMethods("%in%")
exportMethods("%notin%")
exportMethods("[")
exportMethods("[<-")
exportMethods("[[")
exportMethods("[[<-")
exportMethods("activeCat<-")
exportMethods("add<-")
exportMethods("addCats<-")
exportMethods("addTable<-")
exportMethods("levels<-")
exportMethods("names<-")
exportMethods(.workDir)
exportMethods(Arith)
exportMethods(Compare)
exportMethods(E)
exportMethods(Logic)
exportMethods(N)
exportMethods(S)
exportMethods(W)
exportMethods(abs)
exportMethods(acos)
exportMethods(activeCat)
exportMethods(activeCats)
exportMethods(addCats)
exportMethods(aggregate)
exportMethods(allNA)
exportMethods(anyNA)
exportMethods(app)
exportMethods(appCheck)
exportMethods(as.contour)
exportMethods(as.data.frame)
exportMethods(as.data.table)
exportMethods(as.doub)
exportMethods(as.float)
exportMethods(as.int)
exportMethods(as.lines)
exportMethods(as.points)
exportMethods(as.polygons)
exportMethods(asin)
exportMethods(atan)
exportMethods(atan2)
exportMethods(bioclims)
exportMethods(bottom)
exportMethods(breakPolys)
exportMethods(buffer)
exportMethods(c)
exportMethods(catNames)
exportMethods(categories)
exportMethods(cats)
exportMethods(ceiling)
exportMethods(cellSize)
exportMethods(centroids)
exportMethods(classify)
exportMethods(clump)
exportMethods(clusterPoints)
exportMethods(colbind)
exportMethods(combineLevels)
exportMethods(compareGeom)
exportMethods(complete.cases)
exportMethods(compositeRGB)
exportMethods(concats)
exportMethods(connectors)
exportMethods(convHull)
exportMethods(coordRef)
exportMethods(cos)
exportMethods(count)
exportMethods(crds)
exportMethods(crop)
exportMethods(crs)
exportMethods(datatype)
exportMethods(delaunay)
exportMethods(denoise)
exportMethods(dim)
exportMethods(dim3d)
exportMethods(disagg)
exportMethods(distance)
exportMethods(dropRows)
exportMethods(dropTable)
exportMethods(droplevels)
exportMethods(erase)
exportMethods(exp)
exportMethods(expanse)
exportMethods(ext)
exportMethods(extend)
exportMethods(extract)
exportMethods(fast)
exportMethods(fillHoles)
exportMethods(fillNAs)
exportMethods(fixBridges)
exportMethods(fixDangles)
exportMethods(fixLines)
exportMethods(floor)
exportMethods(flow)
exportMethods(flowPath)
exportMethods(focal)
exportMethods(fractalRast)
exportMethods(fragmentation)
exportMethods(freq)
exportMethods(geomorphons)
exportMethods(geomtype)
exportMethods(global)
exportMethods(grassGUI)
exportMethods(grid)
exportMethods(head)
exportMethods(hexagons)
exportMethods(hillshade)
exportMethods(hist)
exportMethods(horizonHeight)
exportMethods(init)
exportMethods(interpIDW)
exportMethods(interpSplines)
exportMethods(intersect)
exportMethods(is.2d)
exportMethods(is.3d)
exportMethods(is.cell)
exportMethods(is.doub)
exportMethods(is.factor)
exportMethods(is.float)
exportMethods(is.int)
exportMethods(is.lines)
exportMethods(is.lonlat)
exportMethods(is.na)
exportMethods(is.points)
exportMethods(kernel)
exportMethods(kurtosis)
exportMethods(layerCor)
exportMethods(levels)
exportMethods(ln)
exportMethods(log)
exportMethods(log10)
exportMethods(log10p)
exportMethods(log1p)
exportMethods(log2)
exportMethods(logic)
exportMethods(longlat)
exportMethods(mask)
exportMethods(maskNA)
exportMethods(match)
exportMethods(max)
exportMethods(mean)
exportMethods(median)
exportMethods(merge)
exportMethods(min)
exportMethods(minmax)
exportMethods(missing.cases)
exportMethods(missingCats)
exportMethods(mmode)
exportMethods(multivarEnvSim)
exportMethods(nacell)
exportMethods(names)
exportMethods(ncell)
exportMethods(ncell3d)
exportMethods(ncol)
exportMethods(ndepth)
exportMethods(neighborhoodMatrix)
exportMethods(neighbourhoodMatrix)
exportMethods(ngeom)
exportMethods(nlevels)
exportMethods(nlyr)
exportMethods(noise)
exportMethods(nonnacell)
exportMethods(not.na)
exportMethods(nrow)
exportMethods(nsubgeom)
exportMethods(nunique)
exportMethods(pairs)
exportMethods(plot)
exportMethods(plotRGB)
exportMethods(predict)
exportMethods(princomp)
exportMethods(project)
exportMethods(quantile)
exportMethods(rNormRast)
exportMethods(rSpatialDepRast)
exportMethods(rUnifRast)
exportMethods(rWalkRast)
exportMethods(range)
exportMethods(rast)
exportMethods(rasterize)
exportMethods(rbind)
exportMethods(regress)
exportMethods(remove0)
exportMethods(removeAngles)
exportMethods(removeBridges)
exportMethods(removeDangles)
exportMethods(removeDupCentroids)
exportMethods(removeDups)
exportMethods(removeSmallPolys)
exportMethods(reorient)
exportMethods(replaceNAs)
exportMethods(res)
exportMethods(res3d)
exportMethods(resample)
exportMethods(round)
exportMethods(ruggedness)
exportMethods(rvoronoi)
exportMethods(sampleRast)
exportMethods(scale)
exportMethods(scalepop)
exportMethods(sdpop)
exportMethods(segregate)
exportMethods(selectRange)
exportMethods(show)
exportMethods(simplifyGeom)
exportMethods(sin)
exportMethods(sineRast)
exportMethods(skewness)
exportMethods(smoothGeom)
exportMethods(snap)
exportMethods(sources)
exportMethods(spatSample)
exportMethods(sqrt)
exportMethods(st_as_sf)
exportMethods(st_buffer)
exportMethods(st_crs)
exportMethods(stdev)
exportMethods(streams)
exportMethods(stretch)
exportMethods(subset)
exportMethods(subst)
exportMethods(sum)
exportMethods(summary)
exportMethods(tail)
exportMethods(tan)
exportMethods(terrain)
exportMethods(thinLines)
exportMethods(thinPoints)
exportMethods(tiles)
exportMethods(top)
exportMethods(topology)
exportMethods(trim)
exportMethods(trunc)
exportMethods(union)
exportMethods(unscale)
exportMethods(update)
exportMethods(var)
exportMethods(varpop)
exportMethods(vect)
exportMethods(vegIndex)
exportMethods(voronoi)
exportMethods(wetness)
exportMethods(which.max)
exportMethods(which.min)
exportMethods(writeRaster)
exportMethods(writeVector)
exportMethods(xor)
exportMethods(xres)
exportMethods(yres)
exportMethods(zext)
exportMethods(zonal)
exportMethods(zonalGeog)
exportMethods(zres)
import(data.table, except = shift)
import(sf)
import(terra, except = shift)
importFrom(data.table,":=")
importFrom(data.table,as.data.table)
importFrom(graphics,par)
importFrom(terra,median)

# fasterRaster
<!-- badges: start -->

[![cran version](https://www.r-pkg.org/badges/version/fasterRaster)](https://cran.r-project.org/package=fasterRaster) [![R-CMD-check](https://github.com/adamlilith/fasterRaster/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/adamlilith/fasterRaster/actions/workflows/R-CMD-check.yaml) [![GPLv3 license](https://img.shields.io/badge/License-GPLv3-blue.svg)](https://perso.crans.org/besson/LICENSE.html)

<!-- badges: end -->
Faster raster processing in `R` using `GRASS`

<a href="https://adamlilith.github.io/fasterRaster/index.html"><img src="man/figures/logo.png" align="right" height="200" alt="fasterRaster website" /></a>

`fasterRaster` is an **R** package designed specifically to handle large-in-memory/large-on-disk spatial rasters and vectors. `fasterRaster` does this using Open Source Geospatial's <a href="https://grass.osgeo.org/">`GRASS`</a>

`fasterRaster` was created with five design principles:

* **Value added**: `fasterRaster` complements `terra` and `sf`, and is highly dependent on them! It is useful for analyzing large-in-memory/large-on-disk rasters and vectors that those packages struggle to handle. For medium- and small-size objects, `terra` and `sf` will almost always be faster.  
* **Familiarity**: If you know how to use `terra`, you basically know how to use `fasterRaster`! That's because most of the functions have the same name and almost the same arguments as `terra` functions.  
* **Comparability**: To the degree possible, outputs from `fasterRaster` are the same as those from functions in `terra` with the same name.  
* **Simplicity**: `GRASS` requires users to track things like "locations" or "projects", "mapsets", and "regions" for which there is no comparable analog in the `terra` or `sf` packages. `fasterRaster` handles these behind the scenes so you don't need to.  
* **It's R**: The `rgrass` package provides a powerful conduit through which you can run `GRASS` tools from `R`. As such, it provides much more flexibility than `fasterRaster`. However, to use `rgrass`, you need to know what `GRASS` tools you want to use and be familiar with `GRASS` syntax. `fasterRaster` obviates this step but uses `rgrass` as a backend, allowing you to focus on `R` syntax and look up help for functions the normal way you do in `R`. You don't need to know `GRASS`!

`fasterRaster` makes heavy use of the <a href="https://cran.r-project.org/package=rgrass">`rgrass`</a> package by Roger Bivand and others, the <a href="https://cran.r-project.org/package=rgrass">`terra`</a> package by Robert Hijmans, the <a href="https://cran.r-project.org/package=sf">`sf`</a> package by Edzer Pebesma, Roger Bivand, and others, and of course <a href="https://grass.osgeo.org/">`GRASS`</a>, so is greatly indebted to all of these creators!

# Vignettes & documentation
**fasterRaster** comes with four user-oriented vignettes, plus a `pkgdown` <a href="https://adamlilith.github.io/fasterRaster/index.html">site</a> with full documentation:

o <a href="https://adamlilith.github.io/fasterRaster/articles/fasterRaster.html">Getting started</a> (also reproduced below)  
o <a href="https://adamlilith.github.io/fasterRaster/articles/GRasters.html">Types of `GRaster`s</a>  
o <a href="https://adamlilith.github.io/fasterRaster/articles/faster_fasterRaster.html">Making **fasterRaster** faster</a>  
o <a href="https://adamlilith.github.io/fasterRaster/articles/addons.html">Addons</a>  
o <a href="https://adamlilith.github.io/fasterRaster/">Documentation</a>  

# Installation
To install `fasterRaster`, please use:

`install_packages('fasterRaster', dependencies = TRUE)`  

You can get the latest stable release using:

`remotes::install_github('adamlilith/fasterRaster', dependencies = TRUE)`  

...and the development version from:

`remotes::install_github('adamlilith/fasterRaster@intuitive_fasterRaster', dependencies = TRUE)`  

To use `fasterRaster` you must install [GRASS version 8.3+](https://grass.osgeo.org/) on your operating system. **You will need to use the stand-alone installer, not the Open Source Geospatial (OS Geo) installer.**

**Optional**: A few functions in **fasterRaster** require **GRASS** "addon" tools, which do not come bundled with **GRASS**. You do not need to install these addons if you do not use functions that call them. A list of functions that require addons can be seen in the "addons" vignette (in **R**, use `vignette("addons", package = "fasterRaster")`). This vignette also explains how to install addons.

# An example

The example presented here is the same as that presented in the the ["getting started"](https://adamlilith.github.io/fasterRaster/articles/fasterRaster.html) vignette.

We'll do a simple operation in which we:

1. Add a buffer to lines representing rivers, then

2. Calculate the distance to from each cell to the closest buffer and burn the distance values into a raster.

To do this, we'll be using maps representing the middle of the eastern coast of Madagascar. We will also use the `terra` and `sf` packages.

```
library(terra) # GIS for rasters and vectors
library(sf) # GIS for vectors
library(fasterRaster)

# Get example elevation raster and rivers vector:
madElev <- fastData('madElev') # SpatRaster with elevation
madRivers <- fastData('madRivers') # sp vector with rivers

# Plot inputs:
plot(madElev)
plot(st_geometry(madRivers), col = "lightblue", add = TRUE)
```
<img src="man/figures/elev_rivers.png"/>  

Before you use nearly any function in the package, you need to tell `fasterRaster` where `GRASS` is installed on your system. The installation folder will vary by operating system and maybe `GRASS` version, but will look something like this:  

```
# Choose the appropriate one, and modify as needed:
grassDir <- "C:/Program Files/GRASS GIS 8.4" # Windows
grassDir <- "/Applications/GRASS-8.4.app/Contents/Resources" # Mac OS
grassDir <- "/usr/local/grass" # Linux
```

Now, use the `faster()` function to tell `fasterRaster` where `GRASS` is installed:
```
faster(grassDir = grassDir)
```

The `fast()` function is the key function for loading a raster or vector into `fasterRaster` format. Rasters in this package are called `GRaster`s and vectors `GVector`s (the "G" stands for `GRASS`). We will now convert the `madElev` raster, which is a `SpatRaster` from the `terra` package, into a `GRaster`.
```
elev <- fast(madElev)
elev
```
You will see the `GRaster`s metadata:
```
class       : GRaster
topology    : 2D 
dimensions  : 1024, 626, NA, 1 (nrow, ncol, ndepth, nlyr)
resolution  : 59.85157, 59.85157, NA (x, y, z)
extent      : 731581.552, 769048.635, 1024437.272, 1085725.279 (xmin, xmax, ymin, ymax)
coord ref.  : Tananarive (Paris) / Laborde Grid 
name(s)     : madElev 
datatype    : integer 
min. value  :       1 
max. value  :     570
```

Next, we'll do the same for the rivers vector. In this case, the vector is an `sf` object from the `sf` package, but we could also use a `SpatVector` from the `terra` package.
```
rivers <- fast(madRivers)
rivers
```
```
class       : GVector
geometry    : 2D lines 
dimensions  : 11, 11, 5 (geometries, sub-geometries, columns)
extent      : 731627.1, 762990.132, 1024541.235, 1085580.454 (xmin, xmax, ymin, ymax)
coord ref.  : Tananarive (Paris) / Laborde Grid 
names       :   F_CODE_DES          HYC_DESCRI      NAM   ISO     NAME_0 
type        :        <chr>               <chr>    <chr> <chr>      <chr> 
values      : River/Stream Perennial/Permanent MANANARA   MDG Madagascar 
              River/Stream Perennial/Permanent MANANARA   MDG Madagascar 
              River/Stream Perennial/Permanent      UNK   MDG Madagascar 
             ...and  8  more rows

```

Now, let's add a 1000-m buffer to the rivers using `buffer()`. As much as possible, `fasterRaster` functions have the same names and same arguments as their counterparts in the `terra` package to help users who are familiar with that package.

Note, though, that the output from `fasterRaster` is not necessarily guaranteed to be the same as output from the respective functions `terra`. This is because there are different methods to do the same thing, and the developers of `GRASS` may have chosen different methods than the developers of other GIS packages.
```
# width in meters because CRS is projected
river_buffers <- buffer(rivers, width = 1000)
```

Now, let's calculate the distances between the buffered areas and all cells on the raster map using `distance()`.
```
dist_to_rivers_meters <- distance(elev, river_buffers)
```

Finally, let's plot the output.
```
plot(dist_to_rivers_meters)
plot(river_buffers, border = 'white', add = TRUE)
plot(rivers, col = "lightblue", add = TRUE)
```

<img src="man/figures/dist_to_rivers.png"/>  

And that's how it's done!  You can do almost anything in `fasterRaster`  you can do with `terra`. The examples above do not show the advantage of `fasterRaster` because the they do not use in large-in-memory/large-on-disk spatial datasets. For very large datasets, `fasterRaster` can be much faster! For example, for a large raster (many cells), the `distance()` function in `terra` can take many days to run and even crash `R`, whereas in `fasterRaster`, it could take just a few minutes or hours.

## Exporting `GRaster`s and `GVector`s from a `GRASS` session

You can convert a `GRaster` to a `SpatRaster` raster using `rast()`:

`terra_elev <- rast(elev)`  

To convert a `GVector` to the `terra` package's `SpatVector`, use `vect()`:

```
terra_rivers <- vect(rivers)
```

You can use `writeRaster()` and `writeVector()` to save `fasterRaster` rasters and vectors directly to disk. This will *always* be faster than using `rast()` or `vect()` and then saving.
```
elev_temp_file <- tempfile(fileext = ".tif") # save as GeoTIFF
writeRaster(elev, elev_temp_file)

vect_temp_shp <- tempfile(fileext = ".shp") # save as shapefile
vect_temp_gpkg <- tempfile(fileext = ".gpkg") # save as GeoPackage
writeVector(rivers, vect_temp_shp)
writeVector(rivers, vect_temp_gpkg)
```

# Versioning

`fasterRaster` versions will look something like `8.3.1.2`, or more generally, `M1.M2.S1.S2`. Here, `M1.M2` will mirror the version of `GRASS` for which `fasterRaster` was built and tested. For example, `fasterRaster` version 8.4.x.x will work using `GRASS` 8.4 (and version 8.3). The values in `S1.S2` refer to "major" and "minor" versions of `fasterRaster`.  That is, a change in the value of `S1` (e.g., from `x.x.1.0` to `x.x.2.0`) indicates changes that potentially break older code developed with a prior version of `fasterRaster`.  A change in `S2` refers to a bug fix, additional functionality in an existing function, or the addition of an entirely new function.

Note that the `M1.M2` and `S1.S2` increment independently. For example, if the version changes from `8.3.1.5` to `8.4.1.5`, then the new version has been tested on `GRASS` 8.4, but code developed with version `8.3.1.x` of `fasterRaster` should still work.

**NOTE**: While `fasterRaster` is still in beta/alpha release, the version will look something like `8.3.0.7XXX`, following Hadley Wickham's guidelines for versioning under development.

# Further reading

* Robert Hijman's [`terra`](https://cran.r-project.org/package=terra) package and Edzer Pebesma's [`sf`](https://cran.r-project.org/package=sf) package are good places to start if you are not familiar with doing GIS in `R`.  
* Roger Bivand's [`rgrass`](https://cran.r-project.org/package=rgrass) package allows users to call any `GRASS` function with all of its functionality, which in some cases is far beyond what is allowed by `fasterRaster`.  
* The [GRASS](https://grass.osgeo.org/) website is authoritative and contains the manual on all the `GRASS` functions used in this package and more.  
* The Wiki on [how to run `GRASS` in `R` or `R` in `GRASS`](https://grasswiki.osgeo.org/wiki/R_statistics/rgrass) will help you to become a power-user of `GRASS` in `R`.

# Citation
Please see and cite:

Smith, A.B. 2025. `fasterRaster`: GIS in `R` using `GRASS` for large vectors and rasters. *Transactions in GIS* 30:e70238. DOI: [10.1111/tgis.70238](https://doi.org/10.1111/tgis.70238), Open-access preprint: [10.31223/X52R0M](https://doi.org/10.31223/X52R0M)


`~` Adam

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
fig.path = 'man/figures/'

## ----install, eval = FALSE----------------------------------------------------
# install.packages("fasterRaster")

## ----install_dev, eval = FALSE------------------------------------------------
# remotes::install_github("adamlilith/fasterRaster", dependencies = TRUE)

## ----packages-----------------------------------------------------------------
library(terra)
library(sf)
library(data.table)
library(fasterRaster)

## ----grassDir_examples, eval = FALSE------------------------------------------
# grassDir <- "C:/Program Files/GRASS GIS 8.4" # Windows
# grassDir <- "/Applications/GRASS-8.4.app/Contents/Resources" # Mac OS
# grassDir <- "/usr/local/grass" # Linux

## ----grassDir, echo = FALSE---------------------------------------------------
grassDir <- "C:/Program Files/GRASS GIS 8.4" # Windows

## ----faster_grassDir, eval = FALSE--------------------------------------------
# faster(grassDir = grassDir)

## ----madElev, eval = FALSE----------------------------------------------------
# madElev <- fastData("madElev") # example SpatRaster
# madElev
# 
# class       : SpatRaster
# dimensions  : 1024, 626, 1  (nrow, ncol, nlyr)
# resolution  : 59.85157, 59.85157  (x, y)
# extent      : 731581.6, 769048.6, 1024437, 1085725  (xmin, xmax, ymin, ymax)
# coord. ref. : Tananarive (Paris) / Laborde Grid
# source      : madElev.tif
# name        : madElev
# min value   :       1
# max value   :     570

## ----elev, eval = FALSE-------------------------------------------------------
# elev <- fast(madElev)
# elev
# 
# class       : GRaster
# topology    : 2D
# dimensions  : 1024, 626, NA, 1 (nrow, ncol, ndepth, nlyr)
# resolution  : 59.85157, 59.85157, NA (x, y, z)
# extent      : 731581.552, 769048.635, 1024437.272, 1085725.279 (xmin, xmax, ymin, ymax)
# coord ref.  : Tananarive (Paris) / Laborde Grid
# name(s)     : madElev
# datatype    : integer
# min. value  :       1
# max. value  :     570

## ----elev_from_file, eval = FALSE---------------------------------------------
# rastFile <- system.file("extdata", "madElev.tif", package = "fasterRaster")
# elev2 <- fast(rastFile)

## ----madRivers, eval = FALSE--------------------------------------------------
# madRivers <- fastData("madRivers") # sf vector
# madRivers
# 
# Simple feature collection with 11 features and 5 fields
# Geometry type: LINESTRING
# Dimension:     XY
# Bounding box:  xmin: 731627.1 ymin: 1024541 xmax: 762990.1 ymax: 1085580
# Projected CRS: Tananarive (Paris) / Laborde Grid
# First 10 features:
#        F_CODE_DES          HYC_DESCRI      NAM ISO     NAME_0                       geometry
# 1180 River/Stream Perennial/Permanent MANANARA MDG Madagascar LINESTRING (739818.2 108005...
# 1185 River/Stream Perennial/Permanent MANANARA MDG Madagascar LINESTRING (739818.2 108005...
# 1197 River/Stream Perennial/Permanent      UNK MDG Madagascar LINESTRING (747857.8 108558...
# 1216 River/Stream Perennial/Permanent      UNK MDG Madagascar LINESTRING (739818.2 108005...
# 1248 River/Stream Perennial/Permanent      UNK MDG Madagascar LINESTRING (762990.1 105737...
# 1256 River/Stream Perennial/Permanent      UNK MDG Madagascar LINESTRING (742334.2 106858...
# 1257 River/Stream Perennial/Permanent      UNK MDG Madagascar LINESTRING (731803.7 105391...
# 1264 River/Stream Perennial/Permanent      UNK MDG Madagascar LINESTRING (755911.6 104957...
# 1300 River/Stream Perennial/Permanent      UNK MDG Madagascar LINESTRING (731871 1044531,...
# 1312 River/Stream Perennial/Permanent      UNK MDG Madagascar LINESTRING (750186.1 103441...

## ----rivers, eval = FALSE-----------------------------------------------------
# rivers <- fast(madRivers)
# rivers
# 
# class       : GVector
# geometry    : 2D lines
# dimensions  : 11, 11, 5 (geometries, sub-geometries, columns)
# extent      : 731627.0998, 762990.1321, 1024541.23477, 1085580.45359 (xmin, xmax, ymin, ymax)
# coord ref.  : Tananarive (Paris) / Laborde Grid
# names       :   F_CODE_DES      HYC_DESCRI      NAM   ISO     NAME_0
# type        :        <chr>           <chr>    <chr> <chr>      <chr>
# values      : River/Stream Perennial/Perm~ MANANARA   MDG Madagascar
#               River/Stream Perennial/Perm~ MANANARA   MDG Madagascar
#               River/Stream Perennial/Perm~      UNK   MDG Madagascar
#               (and 8 more rows)

## ----how_to_plot, eval = FALSE------------------------------------------------
# plot(elev)
# plot(rivers, col = 'lightblue', add = TRUE)

## ----multiplication, eval = FALSE---------------------------------------------
# elev_feet <- elev * 3.28084
# elev_feet
# 
# class       : GRaster
# topology    : 2D
# dimensions  : 1024, 626, NA, 1 (nrow, ncol, ndepth, nlyr)
# resolution  : 59.85157, 59.85157, NA (x, y, z)
# extent      : 731581.552, 769048.635, 1024437.272, 1085725.279 (xmin, xmax, ymin, ymax)
# coord ref.  : Tananarive (Paris) / Laborde Grid
# name(s)     :    layer
# datatype    :   double
# min. value  :   3.2808
# max. value  : 1870.056

## ----log, eval = FALSE--------------------------------------------------------
# log10_elev <- log10(elev)
# log10_elev
# 
# class       : GRaster
# topology    : 2D
# dimensions  : 1024, 626, NA, 1 (nrow, ncol, ndepth, nlyr)
# resolution  : 59.85157, 59.85157, NA (x, y, z)
# extent      : 731581.552, 769048.635, 1024437.272, 1085725.279 (xmin, xmax, ymin, ymax)
# coord ref.  : Tananarive (Paris) / Laborde Grid
# name(s)     :              log
# datatype    :           double
# min. value  :                0
# max. value  : 2.75587485567249

## ----distance_buffers, eval = FALSE-------------------------------------------
# dist <- distance(elev, rivers)
# dist
# 
# class       : GRaster
# topology    : 2D
# dimensions  : 1024, 626, NA, 1 (nrow, ncol, ndepth, nlyr)
# resolution  : 59.85157, 59.85157, NA (x, y, z)
# extent      : 731581.552, 769048.635, 1024437.272, 1085725.279 (xmin, xmax, ymin, ymax)
# coord ref.  : Tananarive (Paris) / Laborde Grid
# name(s)     :         distance
# datatype    :           double
# min. value  :                0
# max. value  : 21310.9411762729

## ----buffer, eval = FALSE-----------------------------------------------------
# 
# river_buff <- buffer(rivers, 2000)
# river_buff
# 
# class       : GVector
# geometry    : 2D polygons
# dimensions  : 1, 5, 0 (geometries, sub-geometries, columns)
# extent      : 729629.19151, 764989.97343, 1022544.92079, 1087580.24979 (xmin, xmax, ymin, ymax)
# coord ref.  : Tananarive (Paris) / Laborde Grid
# 

## ----plot_rivers_buffer, eval = FALSE-----------------------------------------
# plot(dist)
# plot(rivers, col = 'lightblue', add = TRUE)
# plot(river_buff, border = 'white', add = TRUE)

## ----rast, eval = FALSE-------------------------------------------------------
# terra_elev <- rast(elev)

## ----vect, eval = FALSE-------------------------------------------------------
# terra_rivers <- vect(rivers)
# sf_rivers <- st_as_sf(rivers)

## ----write, eval = FALSE------------------------------------------------------
# elev_temp_file <- tempfile(fileext = ".tif") # save as GeoTIFF
# writeRaster(elev, elev_temp_file)
# 
# vect_temp_file <- tempfile(fileext = ".shp") # save as shapefile
# writeVector(rivers, vect_temp_file)

#' Reduce linear features on a raster so linear features are 1 cell wide
#'
#' @description The `thinLines()` function attempts to reduce linear features on a raster to just 1 cell wide. You may need to run `thinLines()` multiple times on the same raster (or experiment with the `iter` argument) to get acceptable output. `thinLines()` can be helpful to run on a raster before using [as.lines()].
#' 
#' @param x A `GRaster`.
#' 
#' @param iter Numeric integer: Number of iterations (default is 200).
#' 
#' @returns A `GRaster`.
#'
#' @seealso [as.lines()], **GRASS** manual page for tool `r.thin` (see `grassHelp("r.thin")`)
#' 
#' @example man/examples/ex_asLines.r
#' 
#' @aliases thinLines
#' @rdname thinLines
#' @exportMethod thinLines
methods::setMethod(
    f = "thinLines",
    signature = c(x = "GRaster"),
    function(x, iter = 200) {
    .locationRestore(x)
    .region(x)

    nLayer <- nlyr(x)
    srcs <- .makeSourceName("thin", "raster", nLayer)
    for (i in seq_len(nLayer)) {
        args <- list(
            cmd = "r.thin",
            input = sources(x)[i],
            output = srcs[i],
            iterations = iter,
            flags = c(.quiet(), "overwrite"),
            intern = TRUE
        )
        do.call(rgrass::execGRASS, args = args)
    }
    
    makeGRaster(srcs, names(x))
    
    } # EOF
)

#' Get "GRASS" projection of raster or vector
#'
#' @description "Getter" for the @projection slot.
#'
#' @param x A `GRaster` or `GVector`.
#'
#' @returns Character.
#'
#' @keywords internal
.projection <- function(x) x@projection

#' Create a grid GVector
#'
#' @description This function creates a `GVector` of "wall-to-wall" cells (like a lattice). The input can be a `GVector` or `GRaster`, which provides the extent of the output.
#'
#' @param x A `GRaster` or `GVector`.
#'
#' @param nx,ny Integer or numeric:
#' * If `use` is `"number"`, then these values represent the number of rows and columns in the grid.
#'  * If `use` is `size`, then these values represent the size of the cells in the x- and y-dimensions.
# In either case, either `nx` or `ny` can be left `NULL` if the other value is defined. In this case, then the size of the cells in the opposing direction will be the same as the size in the direction that is defined. For example, if `nx` is 10 and `use` is `number`, then the size of the cells in the y-dimension will be the same as the size in the x-dimension. The extent of the output will be expanded in the eastward and southward directions to accommodate an integer number of cells in the desired direction.
#'
#' @param use Character: How to generate the grid. If this is `number` (default), then `nx` and `ny` are taken to be the number of grid cells. If `size`, then `nx` and `ny` are taken to be the size of the grid cells.
#'
#' @param angle Numeric: Degrees by which to rotate grid (from north, clockwise).
#'
#' @returns A `GVector`.
#'
#' @example man/examples/ex_grid_hexagons.r
#'
#' @seealso [hexagons()], tool `v.mkgrid` in **GRASS**
#'
#' @aliases grid
#' @rdname grid
#' @exportMethod grid
methods::setMethod(
	f = "grid",
	signature = c(x = "GRaster"),
	function(x, nx = NULL, ny = NULL, use = "number", angle = 0) .grid(x, nx = nx, ny = ny, use = use, angle = angle)
)

#' @aliases grid
#' @rdname grid
#' @exportMethod grid
methods::setMethod(
	f = "grid",
	signature = c(x = "GVector"),
	function(x, nx = NULL, ny = NULL, use = "number", angle = 0) .grid(x, nx = nx, ny = ny, use = use, angle = angle)
)

#' @noRd
.grid <- function(x, nx, ny, use, angle) {

	.locationRestore(x)
	.region(x)

	use <- omnibus::pmatchSafe(use, c("number", "size"), nmax = 1L)

	extent <- ext(x, vector = TRUE)

	# calculate cell number and re-calibrate GRASS region
	if (is.null(nx) & is.null(ny)) {
		stop("At least one of `nx` or `ny` must be defined.")
	} else if (!is.null(nx) & is.null(ny)) {

		if (use == "number") {
		
			dim <- extent[2L] - extent[1L]
			xsize <- ysize <- dim / nx

		} else if (use == "size") {
			xsize <- ysize <- nx
		}
		
		ycellsFract <- (extent[4L] - extent[3L]) / ysize
		ycells <- ceiling(ycellsFract)
		delta <- 0.5 * (ycells - ycellsFract)
		newNorth <- extent[4L] + delta * ysize
		newSouth <- extent[3L] - delta * ysize

		.regionExt(c(extent[1L], extent[2L], newSouth, newNorth), respect = "dimensions")

	} else if (is.null(nx) & !is.null(ny)) {

		if (use == "number") {
		
			dim <- extent[4L] - extent[3L]
			xsize <- ysize <- dim / ny

		} else if (use == "size") {
			xsize <- ysize <- ny
		}

		xcellsFract <- (extent[2L] - extent[1L]) / xsize
		xcells <- ceiling(xcellsFract)
		delta <- 0.5 * (xcells - xcellsFract)
		newWest <- extent[1L] - delta * xsize
		newEast <- extent[2L] + delta * xsize
	
		.regionExt(c(newWest, newEast, extent[3L], extent[4L]), respect = "dimensions")

	} else {

		if (use == "number") {

			xsize <- (extent[2L] - extent[1L]) / nx
			ysize <- (extent[4L] - extent[3L]) / ny

		} else if (use == "size") {
		
			xsize <- nx
			ysize <- ny

			xcellsFract <- (extent[2L] - extent[1L]) / xsize
			ycellsFract <- (extent[4L] - extent[3L]) / ysize

			xcells <- ceiling(xcellsFract)
			ycells <- ceiling(ycellsFract)

			xdelta <- 0.5 * (xcells - xcellsFract)
			ydelta <- 0.5 * (ycells - ycellsFract)

			newWest <- extent[1L] - xdelta * xsize
			newEast <- extent[2L] + xdelta * xsize
		
			newNorth <- extent[4L] + ydelta * ysize
			newSouth <- extent[3L] - ydelta * ysize

			.regionExt(c(newWest, newEast, newSouth, newNorth), respect = "dimensions")
		
		}

	}

	src <- .makeSourceName("v_mkgrid", "vector")
	args <- list(
		cmd = "v.mkgrid",
		map = src,
		position = "region",
		box = c(xsize, ysize),
		flags = c(.quiet(), "overwrite")
	)

	angle <- 360 - angle
	if (!(angle %in% c(0, 360))) args$angle <- angle

	do.call(rgrass::execGRASS, args = args)
	makeGVector(src)

}

if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Elevation raster
madElev <- fastData("madElev")

# Plant specimen points vector
madDypsis <- fastData("madDypsis")

# Rivers lines vector
madRivers <- fastData("madRivers")

# Polygons vector
madCoast4 <- fastData("madCoast4")
madAnt <- madCoast4[madCoast4$NAME_4 == "Antanambe", ]

# Convert to fasterRaster format:
elev <- fast(madElev)
dypsis <- fast(madDypsis)
rivers <- fast(madRivers)
coast <- fast(madCoast4)
ant <- fast(madAnt)

### Crop raster by vector:
rastByVect <- crop(elev, ant)
plot(elev, col = "gray", legend = FALSE)
plot(rastByVect, add = TRUE)
plot(ant, add = TRUE)

### Crop raster by raster:

# For this example, make the SpatRaster smaller, then crop by this.
templateRast <- crop(madElev, madAnt)

template <- fast(templateRast)
rastByRast <- crop(elev, template)

plot(elev, col = "gray", legend = FALSE)
plot(rastByRast, add = TRUE)

### Crop vector by raster:

# For this example, make the SpatRaster smaller, then crop by this.
templateRast <- crop(madElev, madAnt)

template <- fast(templateRast)
ptsByRast <- crop(dypsis, template)

plot(elev, col = "gray", legend = FALSE)
plot(templateRast, add = TRUE)
plot(dypsis, add = TRUE)
plot(ptsByRast, pch = 21, bg = "red", add = TRUE)

### Crop vector by vector:
ptsSubset <- dypsis[1:10] # use first 10 points as cropping template

# Crop points vector by convex hull around points:
ptsByPts <- crop(dypsis, ptsSubset)
plot(dypsis)
plot(convHull(ptsSubset), lty = "dotted", border = "blue", add = TRUE)
plot(ptsByPts, col = "red", add = TRUE)
plot(ptsSubset, col = "blue", pch = 3, cex = 1.6, add = TRUE)
legend("topleft",
    legend = c("Dypsis", "Selected", "Crop template", "Convex hull"),
    pch = c(16, 16, 3, NA),
    lwd = c(NA, NA, NA, 1),
    col = c("black", "red", "blue", "blue"),
    lty = c(NA, NA, NA, "dotted"),
    xpd = NA,
    bg = "white"
)

# Crop points vector by extent of points:
ptsByPts <- crop(dypsis, ptsSubset, ext = TRUE)
plot(dypsis)
plot(ptsByPts, col = "red", add = TRUE)
plot(ptsSubset, col = "blue", pch = 3, cex = 1.6, add = TRUE)
legend("topleft",
    legend = c("Dypsis", "Selected", "Crop template"),
    pch = c(16, 16, 3),
    lwd = c(NA, NA, NA),
    col = c("black", "red", "blue"),
    lty = c(NA, NA, NA),
    xpd = NA,
    bg = "white"
)

# Crop points vector by convex hull around lines:
ptsByPts <- crop(dypsis, rivers)
plot(rivers, col = "blue", pch = 3, cex = 1.6)
plot(dypsis, add = TRUE)
plot(convHull(rivers), lty = "dotted", border = "blue", add = TRUE)
plot(ptsByPts, col = "red", add = TRUE)
legend("topleft",
    legend = c("Dypsis", "Selected", "Rivers", "Convex hull"),
    pch = c(16, 16, NA, NA),
    lwd = c(NA, NA, 1, 1),
    col = c("black", "red", "blue", "blue"),
    lty = c(NA, NA, "solid", "dotted"),
    xpd = NA,
    bg = "white"
)

# Crop points vector by extent of lines:
ptsByPts <- crop(dypsis, rivers, ext = TRUE)
plot(rivers, col = "blue", pch = 3, cex = 1.6)
plot(dypsis, add = TRUE)
plot(convHull(rivers), lty = "dotted", border = "blue", add = TRUE)
plot(ptsByPts, col = "red", add = TRUE)
legend("topleft",
    legend = c("Dypsis", "Selected", "Rivers"),
    pch = c(16, 16, NA),
    lwd = c(NA, NA, 1),
    col = c("black", "red", "blue"),
    lty = c(NA, NA, "solid"),
    xpd = NA,
    bg = "white"
)

# Crop points vector by polygon:
ptsByPts <- crop(dypsis, ant)
plot(dypsis)
plot(ant, border = "blue", pch = 3, cex = 1.6, add = TRUE)
plot(ptsByPts, col = "red", add = TRUE)
legend("topleft",
    legend = c("Dypsis", "Selected", "Antanambe"),
    pch = c(16, 16, NA),
    lwd = c(NA, NA, 1),
    col = c("black", "red", "blue"),
    lty = c(NA, NA, "solid"),
    xpd = NA,
    bg = "white"
)

# Crop lines vector by polygon:
linesByPoly <- crop(rivers, ant)
plot(rivers)
plot(ant, border = "blue", pch = 3, cex = 1.6, add = TRUE)
plot(linesByPoly, col = "red", add = TRUE)
legend("topleft",
    legend = c("Rivers", "Selected", "Antanambe"),
    col = c("black", "red", "blue"),
    lwd = 1,
    xpd = NA,
    bg = "white"
)

# Crop polygon vector by convex hull around points:
polyByPoints <- crop(ant, dypsis)
plot(dypsis, col = "red")
plot(ant, border = "blue", add = TRUE)
plot(polyByPoints, border = "red", add = TRUE)
legend("topleft",
    legend = c("Dypsis", "Antanambe", "Selected"),
    col = c("red", "blue", "red"),
    pch = c(16, NA, NA),
    lwd = c(NA, 1, 1),
    xpd = NA,
    bg = "white"
)

}

#' Terrain ruggedness index
#'
#' @description For a given focal grid cell, the terrain ruggedness index (TRI) is calculated by taking the square root of the average of the squared difference between the focal cell's elevation and the elevations of the 8 surrounding cells, or \deqn{\sqrt(\sum_{i = 1}^{8}(m_i - m_0)^2 / 8)} where \eqn{m_0} is the elevation of the focal cell and \eqn{m_i} is the elevation of the *i*th grid cell. Areas that are entirely flat will have a value of `NA` assigned to them.
#'
#' By default, TRI is calculated for a 3x3 moving window around each focal cell. However, you can use a larger-sized window. In this case, the window size must be an odd number >= 3, and you must have the `r.tri` **GRASS** addon installed. If it is not installed, the function will try to install it.
#'
#' @param x A `GRaster`.
#' @param size Integer (default is 3): Size of the moving window. Must be an odd integer >= 3.
#' @param exponent Numeric >= 0 and <= 4. Used to reduce the influence of cells farther from the focal cell (larger areas can yield noisier results if the exponent is small). All cells are weighted equally when `exponent = 0`.
#'
#' @returns A `GRaster`.
#'
#' @seealso [terrain()], [wetness()], [geomorphons()]
#'
#' @references Riley, S.J., DeGloria, S.D., and Elliot, R. 1999. A terrain ruggedness index that quantifies topographic heterogeneity. *Intermountain Journal of Sciences* 5:23-27.
#'
#' @example man/examples/ex_ruggedness_wetness.r
#'
#' @aliases ruggedness
#' @rdname ruggedness
#' @exportMethod ruggedness
methods::setMethod(
	f = "ruggedness",
	signature = c(x = "GRaster"),
	function(x, size = 3, exponent = 0) {

	if (size %% 2 != 1) stop("Argument `size` must be an odd integer >= 3.")
	if (exponent < 0 | exponent > 4) stop("Argument `exponent` must in the range [0, 4].")
	if (size > 3) .addons("r.tri")

	.locationRestore(x)
	.region(x)

	nLayers <- nlyr(x)
	srcs <- .makeSourceName("terrainRuggednessIndex_r_mapcalc", "raster", n = nLayers)
	for (i in seq_len(nLayers)) {
	
		if (size == 3) {

			ex <- paste0(srcs[i], " = sqrt((",
				"(", sources(x)[i], "[-1, -1] - ", sources(x), ")^2 + ",
				"(", sources(x)[i], "[-1, 0] - ", sources(x), ")^2 + ",
				"(", sources(x)[i], "[-1, 1] - ", sources(x), ")^2 + ",
				"(", sources(x)[i], "[0, -1] - ", sources(x), ")^2 + ",
				"(", sources(x)[i], "[0, 1] - ", sources(x), ")^2 + ",
				"(", sources(x)[i], "[1, -1] - ", sources(x), ")^2 + ",
				"(", sources(x)[i], "[1, 0] - ", sources(x), ")^2 + ",
				"(", sources(x)[i], "[1, 1] - ", sources(x), ")^2) / 8)")
			
			rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		} else {
		
			rgrass::execGRASS(
				cmd = "r.tri",
				input = sources(x)[i],
				output = srcs[i],
				size = size,
				exponent = exponent,
				# processes = faster("cores"), # does not create raster
				flags = c(.quiet(), "overwrite")
			)
		
		}

	} # next layer
	makeGRaster(srcs, names = paste0(names(x), "_TRI"))

	} # EOF
)

if (grassStarted()) {

# Setup
library(terra)

# Elevation raster
madElev <- fastData("madElev")

# Convert to GRaster:
elev <- fast(madElev)

# Topographic wetness index:
twi <- wetness(elev)
names(twi) <- 'TWI'
plot(c(elev, twi))

# Terrain ruggedness index:
tri <- ruggedness(elev)
tri7 <- ruggedness(elev, size = 7)
triSmooth7 <- ruggedness(elev, size = 7, exponent = 4)

tris <- c(elev, tri, tri7, triSmooth7)
names(tris) <- c("elevation", "TRI in 3x3", "TRI in 7x7", "Smoothed TRIin 7x7")
plot(tris)

}

if (grassStarted()) {

# Setup
library(terra)

# Plant specimens (points), elevation (raster)
madDypsis <- fastData("madDypsis")
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster, and sf to a GVector
dypsis <- fast(madDypsis)
elev <- fast(madElev)

### Get coordinates:
dypsisPoints <- crds(dypsis)
elevPoints <- crds(elev)

head(dypsisPoints)
head(elevPoints)

}

#' Make unique GRASS name for rasters, vectors, etc.
#'
#' @param x Character or `NULL`: Descriptive string. **Developers, please note**: To assist with debugging, **GRASS** objects created by a **GRASS** tool have the tool named in this argument (with underscores). Example: "v_in_ogr" or "r_resample".
#' 
#' @param type Character: `raster`, `raster3D`, `vector`, or `table`.
#' 
#' @param n Numeric integer: Number of names to make
#'
#' @param name `NULL` (default) or `character`: Name of the output, attached as an attribute.
#' 
#' @returns Character vector.
#' 
#' @keywords internal
.makeSourceName <- function(x = NULL, type = NULL, n = 1L, name = NULL) {

	if (is.null(x) & is.null(type)) stop("Both `x` and `type` cannot be NULL at the same time.")

 	type <- tolower(type)

	names <- ""
	if (!is.null(x)) {
		if (inherits(x, "SpatRaster")) {
			
			type <- "raster"
			names <- names(x)
			names <- .fixNames(names)
			n <- terra::nlyr(x)
			
		} else if (inherits(x, "GRaster")) {

			type <- "raster"
			names <- names(x)
			names <- .fixNames(names)
			n <- nlyr(x)

		} else if (inherits(x, c("SpatVector", "sf"))) {
			type <- "vector"
		} else if (inherits(x, "character")) {
			type <- x
		}
	} else {
		type <- omnibus::pmatchSafe(type, c("raster", "raster3d", "vector", "group", "region", "table"))
	}

	type[type == "raster3d"] <- "rast3d"
	type[type %in% c("GRaster", "raster")] <- "rast"
	type[type %in% c("GVector", "vector")] <- "vect"
 	type[type == "group"] <- "group"
	type[type == "region"] <- "region"
	type[type == "table"] <- "table"

	src <- omnibus::rstring(1L)
	if (n > 1L) src <- paste0(src, "_", 1L:n)
	src <- if (names[1L] != "") {
  		paste0(type, "_", names, "_", src)
	} else {
  		paste0(type, "_", src)
	}
	if (!is.null(name)) names(src) <- name
	src

}

.fixNames <- function(names) {

	names <- gsub(names, pattern = " ", replacement="_")
	names <- gsub(names, pattern = "\\.", replacement="_")
	names <- gsub(names, pattern = "\\+", replacement="_")
	names <- gsub(names, pattern = "-", replacement="_")
	names <- gsub(names, pattern = "\\?", replacement="_")
	names <- gsub(names, pattern = "!", replacement="_")
	names <- gsub(names, pattern = "\\*", replacement="_")
	names <- gsub(names, pattern = "\\(", replacement="_")
	names <- gsub(names, pattern = "\\)", replacement="_")
	names <- gsub(names, pattern = "\\[", replacement="_")
	names <- gsub(names, pattern = "\\]", replacement="_")
	names <- gsub(names, pattern = "\\|", replacement="_")
	names <- gsub(names, pattern = "@", replacement="_")
	names <- gsub(names, pattern = "\\$", replacement="_")
	names <- gsub(names, pattern = "#", replacement="_")
	names <- gsub(names, pattern = "%", replacement="_")
	names <- gsub(names, pattern = "\\^", replacement="_")
	names <- gsub(names, pattern = "&", replacement="_")
	names <- gsub(names, pattern = "=", replacement="_")
	names <- gsub(names, pattern = "\\'", replacement="_")
	names <- gsub(names, pattern = "\\{", replacement="_")
	names <- gsub(names, pattern = "}", replacement="_")
	names <- gsub(names, pattern = ";", replacement="_")
	names <- gsub(names, pattern = ":", replacement="_")
	names <- gsub(names, pattern = "<", replacement="_")
	names <- gsub(names, pattern = ">", replacement="_")
	names <- gsub(names, pattern = "/", replacement="_")
	names <- gsub(names, pattern = "`", replacement="_")
	names <- gsub(names, pattern = "~", replacement="_")
	names

}

#' Select values from rasters in a stack based on values in another raster
#'
#' @description `selectRange()` selects values from `GRaster`s in a "stack" based on the values in another "selection" raster. For example, if the stack has three layers (call them A, B, and C), the "selection" raster could have values of 1, 2, or 3 in each cell. The raster that is returned will have values from A wherever the selection raster is 1, B from where it is 2, and C from where it is 3.
#'
#' @param x A `GRaster`, typically with more than one layer.
#'
#' @param y A `GRaster` with integer values. The raster will be rounded if it does not. The values are typically between 1 and the number of layers in `x`, but wherever they are outside this range, the returned raster will have `NA` values.
#'
#' @returns A `GRaster`.
#'
#' @example man/examples/ex_selectRange.r
#'
#' @aliases selectRange
#' @rdname selectRange
#' @exportMethod selectRange
methods::setMethod(
	f = "selectRange",
	signature = c(x = "GRaster"),
	function(x, y) {
	
	.locationRestore(x)
	.region(x)

	if (nlyr(y) > 1L) {
		warning("The`y` raster has more than one layer. On the first will be used.")
		y <- y[[1L]]
	}
	if (datatype(y, "GRASS") != "CELL") y <- round(y)

	src <- .makeSourceName("selectRange", "raster")
	xGnames <- sources(x)
	yGname <- sources(y)
	nLayers <- nlyr(x)

	ex <- paste0(src, " = ")
	for (i in seq_len(nLayers)) {
		ex <- paste0(ex, " if(", yGname, " == ", i, ", ", xGnames[i], ",")
	}
	ex <- paste0(ex, " null()", paste(rep(")", nLayers), collapse = ""))

	args <- list(
		cmd = "r.mapcalc",
		expression = ex,
		flags = c(.quiet(), "overwrite"),
		intern = TRUE
	)
	do.call(rgrass::execGRASS, args = args)

	makeGRaster(src, "selectRange")

	} # EOF
)

#' Combine values/categories of multiple GRasters into a single GRaster
#'
#' @description This function takes from 2 to 10 integer or categorical (factor) `GRaster`s and creates a single `GRaster` that has one value per combination of values in the inputs. For example, say that there were two input rasters, with values 1 and 2 in the one raster, and 3 and 4 in the other. If the following combinations of values occurred between the two rasters, then the output raster would be re-coded with the new values:
#'
#' | `input_raster1` | `input_raster2` | `output_raster` |
#' | --------------- | --------------- | --------------- |
#' | 1               | 3               | 0               |
#' | 1               | 4               | 1               |
#' | 2               | 3               | 2               |
#' | 2               | 4               | 3               |
#'
#' If the argument `na.rm` is set to `TRUE` (which it is, by default), then whenever at least one cell has an `NA` value, then the output will also have an `NA` (i.e., a new category number is not created). However, if `na.rm` is `FALSE`, then combinations that include an `NA` are assigned a new category number, unless all values are `NA` (in which case the output will be `NA`).
#'
#' The difference between this function and [combineLevels()] is that this one creates a "combined" `GRaster` with a combined levels table, whereas `combineLevels()` just merges the levels tables.
#'
#' If the inputs are all categorical rasters, then a [levels()] table will also be returned with the new levels.
#'
#' @param x A `GRaster` with one or more layers, each of which must be have cells that represent integers or categories (factors).
#'
#' @param ... Either missing or integer/categorical (factor) `GRaster`s.
#'
#' @param na.rm Logical: If `TRUE` (default), then any combinations that include an `NA` cell will result in an `NA` cell in the output.
#'
#' @returns A `GRaster`. If the inputs are all categorical (factor) rasters, then a levels table will also be returned with the new combined levels.
#'
#' @example man/examples/ex_GRaster_categorical.r
#'
#' @seealso [combineLevels()], [terra::concats()], `vignette("GRasters", package = "fasterRaster")`, **GRASS** manual page for tool `r.cross` (see `grassHelp("r.cross")`)
#'
#' @aliases concats
#' @rdname concats
#' @exportMethod concats
methods::setMethod(
	f = "concats",
	signature = c(x = "GRaster"),
	function(x, ..., na.rm = TRUE) {

	.locationRestore(x)
	.region(x)

	dt <- datatype(x)
	if (any(!(dt %in% c("integer", "factor")))) stop("All rasters must be of type integer or factor.")
	
	allFactors <- all(is.factor(x))
	inSrcs <- sources(x)

	dots <- list(...)
	if (length(dots) > 0L) {

		dots <- omnibus::unlistRecursive(dots)
		xFirst <- x[[1L]]
	
		for (i in seq_along(dots)) {

			compareGeom(xFirst, dots[[i]])
			
			dt <- datatype(dots[[i]])
			if (any(!(dt %in% c("integer", "factor")))) stop("All rasters must be of type integer or factor.")
			allFactors <- allFactors & all(is.factor(dots[[i]]))
			
			inSrcs <- c(inSrcs, sources(dots[[i]]))
		
		}
		nLayersDots <- sapply(dots, nlyr)
	} else {
  		nLayersDots <- 0L
	}

	nl <- nlyr(x) + nLayersDots
	if (nl < 2L | nl > 10L) stop("From 2 to 10 raster layers can be combined. Too few/many layers.")

	# combine
	src <- .makeSourceName("r_cross", "raster")
	args <- list(
		cmd = "r.cross",
		input = inSrcs,
		output = src,
		flags = c(.quiet(), "overwrite")
	)

	if (na.rm) args$flags <- c(args$flags, "z")
	do.call(rgrass::execGRASS, args = args)

	# collate levels tables
	if (allFactors) {
	
		# get levels tables from rasters
		levs <- levels(x)
		if (length(dots) > 0L) {
			for (i in seq_along(dots)) levs <- c(levs, levels(dots[[i]]))
		}

		# get levels table from output
		glevs <- rgrass::execGRASS(
			cmd = "r.category",
			map = src,
			separator = "pipe",
			intern = TRUE
		)

		glevs <- sapply(glevs, strsplit, split = "\\|")
		n <- length(glevs)
	
		# populate new levels table
		table <- data.table::data.table(value = rep(NA_integer_, n), category = rep(NA_character_, n))

		for (i in seq_along(glevs))	{

			val <- as.integer(glevs[[i]][1L])
			table$value[i] <- val
			
			gcats <- strsplit(glevs[[i]][2L], split = "; ")[[1L]]
			gcats <- gsub(gcats, pattern = "category ", replacement = "")
			gcats[gcats == "NULL"] <- NA_integer_
			gcats <- as.integer(gcats)

			comboCat <- rep(NA_character_, nl)
			for (j in seq_along(levs)) {

				if (is.na(gcats[j])) {
					comboCat[j] <- "NA"
				} else {
					comboCat[j] <- unlist(levs[[j]][levs[[j]][[1L]] == gcats[j], 2L])
				}

			}

			comboCat <- paste(comboCat, collapse = " | ")
			table$category[i] <- comboCat
		
		}

	} else {
		table <- NULL
	}

	makeGRaster(src, names = "combineCats", levels = table)
	
	} # EOF
)

#' BIOCLIM rasters
#'
#' @description The BIOCLIM set of bioclimatic variables were created for modeling species' geographic distributions (Booth et al. 2014). This function can create the "standard" 19 set of variables, plus several more from an "extended" set.
#'
#' "Classic" set of BIOCLIM variables (Booth et al. 2014):
#' The units reported below assume that input rasters are in mm (precipitation) and deg C (temperature), and that each raster represents a month (but other time units are allowed, with corresponding changes to the temporal units assumed below).
#'
#' * BIO1: Mean annual temperature, calculated using monthly means (deg C)
#' * BIO2: Mean diurnal range across months (average of monthly difference between maximum and minimum temperature) (deg C)
#' * BIO3: Isothermality (100 * BIO02 / BIO07; unit-less)
#' * BIO4: Temperature seasonality (standard deviation across months of average monthly temperature * 100; deg C)
#' * BIO5: Maximum temperature of the warmest month (based on maximum temperature; deg C)
#' * BIO6: Minimum temperature of the coldest month (based on minimum temperature; deg C)
#' * BIO7: Range of annual temperature (BIO05 - BIO06; deg C)
#' * BIO8: Temperature of the wettest quarter (based on mean temperature; deg C)
#' * BIO9: Temperature of the driest quarter (based on mean temperature; deg C)
#' * BIO10: Temperature of the warmest quarter (based on mean temperature; deg C)
#' * BIO11: Temperature of the coldest quarter (based on mean temperature; deg C)
#'
#' * BIO12: Total annual precipitation (mm)
#' * BIO13: Precipitation of the wettest month (mm)
#' * BIO14: Precipitation of the driest month (mm)
#' * BIO15: Precipitation seasonality (100 * coefficient of variation; unit-less)
#' * BIO16: Precipitation of the wettest quarter (mm)
#' * BIO17: Precipitation of the driest quarter (mm)
#' * BIO18: Precipitation of the warmest quarter (based on mean temperature; mm)
#' * BIO19: Precipitation of the coldest quarter (based on mean temperature; mm)
#'
#' "Extended" set of BIOCLIM variables (starts at 41 to avoid conflicts with Kriticos et al. 2014):
#' * BIO41: Temperature of the quarter following the coldest quarter (based on mean temperature; deg C)
#' * BIO42: Temperature of the quarter following the warmest quarter (based on mean temperature; deg C)
#' * BIO43: Precipitation of the quarter following the coldest quarter (based on mean temperature; mm)
#' * BIO44: Precipitation of the quarter following the warmest quarter (based on mean temperature; mm)  
#'
#' * BIO45: Temperature of the quarter following the driest quarter (based on mean temperature; deg C)
#' * BIO46: Temperature of the quarter following the wettest quarter (based on mean temperature; deg C)
#' * BIO47: Precipitation of the quarter following the driest quarter (based on mean temperature; mm)
#' * BIO48: Precipitation of the quarter following the wettest quarter (based on mean temperature; mm)  
#'
#' * BIO49: Hottest month (based on maximum temperature)
#' * BIO50: Coldest month (based on minimum temperature)
#' * BIO51: Wettest month
#' * BIO52: Driest month  
#'
#' * BIO53: First month of the warmest quarter (based on mean temperature)
#' * BIO54: First month of the coldest quarter (based on mean temperature)
#' * BIO55: First month of the wettest quarter
#' * BIO56: First month of the driest quarter  
#'
#' * BIO57: The greatest decrease in temperature from one month to the next (deg C; always >= 0)
#' * BIO58: The greatest increase in temperature from one month to the next (deg C; always >= 0)
#' * BIO59: The greatest decrease in precipitation from one month to the next (mm; always >= 0)
#' * BIO60: The greatest increase in precipitation from one month to the next (mm; always >= 0)
#' 
#' By default, "quarter" refers to any consecutive run of three months, not a financial quarter. A quarter can thus include November-December-January, or December-January-February, for example. However, the length of a quarter can be changed using the argument `quarter`.
#'
#' The variables are defined assuming that the input rasters represent monthly values (12 rasters for min/max temperature and precipitation), but you can also use sets of 52 rasters, representing one per week, in which case "quarter" would be a successive run of 3 weeks. You could also attempt 365 rasters, in which case a "quarter" would be a run of 3 successive days.
#'
#' BIOCLIMs 41 through 44 are added here to capture the "shoulder" seasons (spring and autumn) important in temperature regions. BIOCLIMs 45 through 48 are also included for consistency.
#'
#' BIOCLIMs 49 through 60 are not bioclimatic variables per se, but useful for assessing the properties of the variables that are defined based on the "-est" month or quarter.
#'
#' @param ppt A multi-layered `GRaster` or `SpatRaster`, representing monthly/weekly/daily precipitation.
#'
#' @param tmin,tmax A multi-layered `GRaster` or `SpatRaster`, representing monthly/weekly/daily minimum and maximum temperature.
#'
#' @param tmean Either `NULL` (default), or a multi-layered `GRaster` or `SpatRaster`, representing monthly/weekly/daily average temperature. If `NULL`, `tmean` will be calculated internally from `tmin` and `tmax`. Providing these rasters thus saves time if you already have them on hand.
#'
#' @param bios Any of:
#' * Numeric values: Calculate these BIOCLIM variables. For example, `bios = c(1, 12)` calculates BIOCLIMs 1 and 12.
#' * `NULL` (default): Calculate BIOCLIMs 1 through 19
#' * `"*"`: Calculate all BIOCLIMs this function can calculate.
#' * `"+"`: Calculate BIOCLIMs 41 onward.
#' * Any combination of the above except `NULL` (e.g., `c(1, 12, "+")`).
#'
#' @param sample Logical: If `TRUE` (default), BIO4 and 15 are calculated with the sample standard deviation. If `FALSE`, then the population standard deviation is used.
#'
#' @param quarter Numeric: Length of a "quarter". BIOCLIM variables are typically calculated using monthly-averaged rasters (e.g., precipitation and temperature of January, February, etc.), in which case a "quarter" is 3 months (so the default for `quarter` is 3). However, this function can accommodate any set of rasters representing a time series (e.g., 365 for daily rasters), in which case the user can decide what constitutes a "quarter" for calculation of the any BIOCLIMs that use "quarters" in their definitions.
#'
#' @param pptDelta Numeric: Value to add to precipitation for calculation of BIO15 (coefficient of variation of precipitation, times 100). Adding a small value avoids division by 0. The default is 1.
#'
#' @param verbose Logical: If `TRUE` (default), display progress.
#'
#' @returns A `GRaster` with one or more layers.
#'
#' @references Booth, T.H., Nix, H.A., Busby, J.R., and Hutchinson, M.F.  2014.  BIOCLIM: The first species distribution modeling package, its early applications and relevance to most current MaxEnt studies. *Diversity and Distributions* 20:1-9 \doi{10.1111/ddi.12144}.
#'
#' Kriticos, D.J., Jarošik, V., and Otam N.  2014.  Extending the suite of BIOCLIM variables: A proposed registry system and case study using principal components analysis. *Methods in Ecology and Evolution* 5:956-960 \doi{10.1111/2041-210X.12244}.
#'
#' @example man/examples/ex_bioclims.r
#'
#' @aliases bioclims
#' @rdname bioclims
#' @exportMethod bioclims
methods::setMethod(
	f = "bioclims",
	signature = c(ppt = "GRaster"),
	function(
		ppt,
		tmin,
		tmax,
		tmean = NULL,
		bios = NULL,
		sample = TRUE,
		quarter = 3,
		pptDelta = 1,
		verbose = TRUE
	) {

	# BIOCLIM numbers
	classicBios <- 1L:19L
	extendedBios <- 41L:60L
	allBios <- c(classicBios, extendedBios)

	meanTempBios <- c(1, 4, 8:11, 18:19, 41:46, 53:54, 57:60) # BIOCLIMs that require mean temp
	quarterTempBios <- c(10, 11, 18, 19, 41:46, 53:54) # BIOCLIMs that require quarter temperatures
	quarterPrecipBios <- c(8, 9, 16, 17, 16:19, 45:48, 55:56)
	wettestQuarterBios <- c(8, 16, 46, 48, 55)
	driestQuarterBios <- c(9, 17, 45, 47, 56)
	warmestQuarterBios <- c(10, 18, 42, 44, 53)
	coldestQuarterBios <- c(11, 19, 41, 43, 54)
	
	### number of tasks (used for progress bar
	nTasks <- 2 + length(bios) +
		(any(meanTempBios %in% bios) & is.null(tmean)) +
		any(quarterTempBios %in% bios) +
		any(quarterPrecipBios %in% bios) +
		any(wettestQuarterBios %in% bios) +
		any(driestQuarterBios %in% bios) +
		any(warmestQuarterBios %in% bios) +
		any(coldestQuarterBios %in% bios) +
		any(c(57, 58) %in% bios) +
		any(c(59, 60) %in% bios) +
		(sample & 4 %in% bios) +
		(sample & 15 %in% bios)

	if (verbose | faster("verbose")) {
		pb <- utils::txtProgressBar(min = 0, max = nTasks, initial = 0, style = 3, width = 30)
	}
	tasks <- 0

	if (is.null(bios)) {
		bios <- classicBios
	} else if (any(bios == "*")) {
		bios <- allBios
	} else if (any(bios == "+")) {
		bios <- bios[bios != "+"]
		bios <- c(bios, extendedBios)
	}

	bios <- unique(bios)
	bios <- sort(bios)
	bios <- as.integer(bios)

	if (quarter <= 1) stop("Argument `quarter` must be > 1.")

	nLayers <- nlyr(ppt)

	if (any(omnibus::notIn(bios, allBios))) stop("Argument `bios` has at least one invalid value. Valid values include:\n  ", paste(allBios, collapse = " "))

	if (any(omnibus::notIn(bios, allBios))) stop("Argument `bios` has at least one invalid value. Valid values include:\n  ", paste(allBios, collapse = ", "))

	if (quarter > nLayers) stop("Argument `quarter` cannot be more than the number of raster layers in any of the input GRasters.")

	compareGeom(ppt, tmin, lyrs = TRUE)
	compareGeom(ppt, tmax, lyrs = TRUE)
	if (!is.null(tmean)) compareGeom(ppt, tmean, lyrs = TRUE)
	
	.locationRestore(ppt)
	.region(ppt)

	ppt <- sources(ppt)
	tmin <- sources(tmin)
	tmax <- sources(tmax)
	if (!is.null(tmean)) tmean <- sources(tmean)

	### calculate tmean if needed
	if (is.null(tmean) & any(bios %in% meanTempBios)) {

		tmean <- .makeSourceName("r_mapcalc", "raster", nLayers)
		for (i in seq_len(nLayers)) {

			ex <- paste0(tmean[i], " = (", tmin[i], " + ", tmax[i], ") / 2")
			rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		}
		on.exit(.rm(tmean, type = "raster", warn = FALSE, verify = FALSE), add = TRUE)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	### quarter temperature rasters
	if (any(bios %in% quarterTempBios)) {


		tmeanByQuarter <- .calcQuarterGR(tmean, fun = "mean", quarter = quarter)
		on.exit(.rm(tmeanByQuarter, type = "raster", warn = FALSE, verify = FALSE), add = TRUE)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	### quarter precipitation rasters
	if (any(bios %in% quarterPrecipBios)) {

		pptByQuarter <- .calcQuarterGR(ppt, fun = "sum", quarter = quarter)
		on.exit(.rm(pptByQuarter, type = "raster", warn = FALSE, verify = FALSE), add = TRUE)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	### MAIN
	########

	out <- NULL

	### BIO53 (warmest quarter)
	if (any(bios %in% warmestQuarterBios)) {

		bio53Minus1 <- .makeSourceName("r_mapcalc", "raster")
		bio53 <- .makeSourceName("r_mapcalc", "raster", name = "bio53")
		
		rgrass::execGRASS(
			cmd = "r.series",
			input = tmeanByQuarter,
			output = bio53Minus1,
			method = "max_raster",
			nprocs = faster("cores"),
			memory = faster("memory"),
			flags = c(.quiet(), "overwrite", "n")
		)

		on.exit(.rm(bio53Minus1, type = "raster", warn = FALSE, verify = FALSE), add = TRUE)

		ex <- paste0(bio53, " = ", bio53Minus1, " + 1")
		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		# vector of warmest quarters
		warmestQuarters <- .freq(bio53, dtype = "CELL", digits = NA, bins = NA)
		warmestQuarters <- warmestQuarters[["value"]]

		if (53 %in% bios) {
			out <- c(out, bio53)
			names(out)[length(out)] <- "bio53"
		}

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	### BIOCLIM 54 (coldest quarter)
	if (any(bios %in% coldestQuarterBios)) {

		bio54Plus1 <- .makeSourceName("r_mapcalc", "raster")
		bio54 <- .makeSourceName("r_mapcalc", "raster", name = "bio54")
		
		rgrass::execGRASS(
			cmd = "r.series",
			input = tmeanByQuarter,
			output = bio54Plus1,
			method = "min_raster",
			nprocs = faster("cores"),
			memory = faster("memory"),
			flags = c(.quiet(), "overwrite", "n")
		)

		on.exit(.rm(bio54Plus1, type = "raster", warn = FALSE, verify = FALSE), add = TRUE)

		ex <- paste0(bio54, " = ", bio54Plus1, " + 1")
		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		# get vector of all coldest quarters
		coldestQuarters <- .freq(bio54, dtype = "CELL", digits = NA, bins = NA)
		coldestQuarters <- coldestQuarters[["value"]]

		if (54 %in% bios) {
			out <- c(out, bio54)
			names(out)[length(out)] <- "bio54"
		}

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	### BIOCLIM 55 (wettest quarter)
	if (any(bios %in% wettestQuarterBios)) {
		
		bio55Plus1 <- .makeSourceName("r_mapcalc", "raster")
		bio55 <- .makeSourceName("r_mapcalc", "raster", name = "bio55")

		rgrass::execGRASS(
			cmd = "r.series",
			input = pptByQuarter,
			output = bio55Plus1,
			method = "max_raster",
			nprocs = faster("cores"),
			memory = faster("memory"),
			flags = c(.quiet(), "overwrite", "n")
		)

		on.exit(.rm(bio55Plus1, type = "raster", warn = FALSE, verify = FALSE), add = TRUE)

		# increment index (first is 0)
		ex <- paste0(bio55, " = ", bio55Plus1, " + 1")
		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		# get vector of all wettest quarters
		wettestQuarters <- .freq(bio55, dtype = "CELL", digits = NA, bins = NA)
		wettestQuarters <- wettestQuarters[["value"]]

		if (55 %in% bios) {
			out <- c(out, bio55)
			names(out)[length(out)] <- "bio55"
		}

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	### BIOCLIM 56 (driest quarter)
	if (any(bios %in% driestQuarterBios)) {

		bio56Plus1 <- .makeSourceName("r_mapcalc", "raster")
		bio56 <- .makeSourceName("r_mapcalc", "raster", name = "bio56")
		
		rgrass::execGRASS(
			cmd = "r.series",
			input = pptByQuarter,
			output = bio56Plus1,
			method = "min_raster",
			nprocs = faster("cores"),
			memory = faster("memory"),
			flags = c(.quiet(), "overwrite", "n")
		)

		on.exit(.rm(bio56Plus1, type = "raster", warn = FALSE, verify = FALSE), add = TRUE)

		ex <- paste0(bio56, " = ", bio56Plus1, " + 1")
		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		# get vector of all driest quarters
		driestQuarters <- .freq(bio56, dtype = "CELL", digits = NA, bins = NA)
		driestQuarters <- driestQuarters[["value"]]

		if (56 %in% bios) {
			out <- c(out, bio56)
			names(out)[length(out)] <- "bio56"
		}

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	### BIOCLIM 1
	if (1 %in% bios) {

		bio1 <- .makeSourceName("r_series", "raster", n = 1L, name = "bio1")
		rgrass::execGRASS(
			"r.series",
			input = paste(tmean, collapse = ","),
			output = bio1,
			method = "average",
			nprocs = faster("cores"),
			memory = faster("memory"),
			flags = c(.quiet(), "overwrite", "n")
		)

		out <- c(out, bio1)
		names(out)[length(out)] <- "bio1"

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)
	
	}

	### BIOCLIM 2
	if (any(c(2, 3) %in% bios)) {
		
		meanMaxSrc <- .makeSourceName("r_series", "raster", n = 1L)
		meanMinSrc <- .makeSourceName("r_series", "raster", n = 1L)
		bio2 <- .makeSourceName("r_series", "raster", n = 1L, name = "bio2")

		# mean of maximums
		rgrass::execGRASS(
			"r.series",
			input = paste(tmax, collapse = ","),
			output = meanMaxSrc,
			method = "average",
			nprocs = faster("cores"),
			memory = faster("memory"),
			flags = c(.quiet(), "overwrite", "n")
		)

		# mean of minimums
		rgrass::execGRASS(
			"r.series",
			input = paste(tmin, collapse = ","),
			output = meanMinSrc,
			method = "average",
			nprocs = faster("cores"),
			memory = faster("memory"),
			flags = c(.quiet(), "overwrite", "n")
		)

		# bio2
		ex <- paste0(bio2, " = ", meanMaxSrc, " - ", meanMinSrc)

		on.exit(.rm(c(meanMaxSrc, meanMinSrc), type = "raster", warn = FALSE, verify = FALSE), add = TRUE)

		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		if (2 %in% bios) {
			out <- c(out, bio2)
			names(out)[length(out)] <- "bio2"
		}
	
		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	### BIOCLIM 5
	if (any(c(3, 5, 7) %in% bios)) {
		
		bio5 <- .makeSourceName("r_series", "raster", n = 1L, name = "bio5")

		rgrass::execGRASS(
			"r.series",
			input = paste(tmax, collapse = ","),
			output = bio5,
			method = "maximum",
			nprocs = faster("cores"),
			memory = faster("memory"),
			flags = c(.quiet(), "overwrite", "n")
		)

		if (5 %in% bios) {
			out <- c(out, bio5)
			names(out)[length(out)] <- "bio5"
		}

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)
	
	}

	### BIOCLIM 6
	if (any(c(3, 6, 7) %in% bios)) {
		
		bio6 <- .makeSourceName("r_series", "raster", n = 1L, name = "bio6")

		rgrass::execGRASS(
			"r.series",
			input = paste(tmin, collapse = ","),
			output = bio6,
			method = "minimum",
			nprocs = faster("cores"),
			memory = faster("memory"),
			flags = c(.quiet(), "overwrite", "n")
		)

		if (6 %in% bios) {
			out <- c(out, bio6)
			names(out)[length(out)] <- "bio6"
		}

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	### BIOCLIM 7 (annual temp range)
	if (any(c(3, 7) %in% bios)) {
		
		bio7 <- .makeSourceName("r_mapcalc", "raster", n = 1L, name = "bio7")
		ex <- paste0(bio7, " = ", bio5, " - ", bio6)
		
		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		if (7 %in% bios) {
			out <- c(out, bio7)
			names(out)[length(out)] <- "bio7"
		}

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)
	
	}

	### BIOCLIM 3 (100 * bio2 / bio7)
	if (3 %in% bios) {
		
		bio3 <- .makeSourceName("r_mapcalc", "raster", n = 1L, name = "bio3")

		bio2 <- out[names(out) == "bio2"]
		bio7 <- out[names(out) == "bio7"]
		
		ex <- paste0(bio3, " = 100 * ", bio2, " / ", bio7)
		
		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		out <- c(out, bio3)
		names(out)[length(out)] <- "bio3"

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)
	
	}

	### BIOCLIM 4 (100 * sd of temp)
	if (4 %in% bios) {
		
		bio4 <- .makeSourceName("r_mapcalc", "raster", n = 1L, name = "bio4")

		rgrass::execGRASS(
			"r.series",
			input = paste(tmean, collapse = ","),
			output = bio4,
			method = "stddev",
			nprocs = faster("cores"),
			memory = faster("memory"),
			flags = c(.quiet(), "overwrite", "n")
		)

		# convert to sample SD as per R
		if (sample) {
		
			tasks <- tasks + 1
			if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

			srcIn <- bio4
			bio4 <- .makeSourceName("r_mapcalc", "raster", n = 1L, name = "bio4")

			ex <- paste0(bio4, " = 100 * sqrt((", nLayers, " * (", srcIn, "^2)) / (", nLayers, " - 1))")
			
			rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))
			on.exit(.rm(c(srcIn), type = "raster", warn = FALSE, verify = FALSE), add = TRUE)
		
		}

		out <- c(out, bio4)
		names(out)[length(out)] <- "bio4"

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)
	
	}

	# BIOCLIM 8 (mean temp of wettest quarter)
	if (8 %in% bios) {
		
		# burn temperature of wettest quarter to cells
		bio8 <- .makeSourceName("r_mapcalc", "raster", n = 1L, name = "bio8")

		ex <- paste0(bio8, " = ")
		for (i in wettestQuarters) {
			ex <- paste0(ex, "if(", bio55, "==", i, ",", tmeanByQuarter[i], ",")
		}
		ex <- paste0(ex, "null()", paste(rep(")", length(wettestQuarters)), collapse = ""))

		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		out <- c(out, bio8)
		names(out)[length(out)] <- "bio8"

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIOCLIM 9 (mean temp of driest quarter)
	if (9 %in% bios) {

		# burn temperature of driest quarter to cells
		bio9 <- .makeSourceName("r_mapcalc", "raster", n = 1L, name = "bio9")

		ex <- paste0(bio9, " = ")
		for (i in driestQuarters) {
			ex <- paste0(ex, "if(", bio56, "==", i, ",", tmeanByQuarter[i], ",")
		}
		ex <- paste0(ex, "null()", paste(rep(")", length(driestQuarters)), collapse = ""))

		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		out <- c(out, bio9)
		names(out)[length(out)] <- "bio9"

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIOCLIM 10 (mean temp of warmest quarter)
	if (10 %in% bios) {

		# burn temperature of warmest quarter to cells
		bio10 <- .makeSourceName("r_mapcalc", "raster", n = 1L, name = "bio10")

		ex <- paste0(bio10, " = ")
		for (i in warmestQuarters) {
			ex <- paste0(ex, "if(", bio53, "==", i, ",", tmeanByQuarter[i], ",")
		}
		ex <- paste0(ex, "null()", paste(rep(")", length(warmestQuarters)), collapse = ""))

		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		out <- c(out, bio10)
		names(out)[length(out)] <- "bio10"

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIOCLIM 11 (mean temp of coldest quarter)
	if (11 %in% bios) {

		# burn temperature of coldest quarter to cells
		bio11 <- .makeSourceName("r_mapcalc", "raster", n = 1L, name = "bio11")

		ex <- paste0(bio11, " = ")
		for (i in coldestQuarters) {
			ex <- paste0(ex, "if(", bio54, "==", i, ",", tmeanByQuarter[i], ",")
		}
		ex <- paste0(ex, "null()", paste(rep(")", length(coldestQuarters)), collapse = ""))

		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		out <- c(out, bio11)
		names(out)[length(out)] <- "bio11"

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIOCLIM 12 (total precipitation)
	if (12 %in% bios) {

		bio12 <- .makeSourceName("r_mapcalc", "raster", n = 1L, name = "bio12")

		input <- paste0(ppt, collapse = ",")

		rgrass::execGRASS(
			cmd = "r.series",
			input = input,
			output = bio12,
			method = "sum",
			nprocs = faster("cores"),
			memory = faster("memory"),
			flags = c(.quiet(), "overwrite", "n")
		)

		out <- c(out, bio12)
		names(out)[length(out)] <- "bio12"

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIOCLIM 13 (wettest month precipitation)
	if (13 %in% bios) {

		bio13 <- .makeSourceName("r_mapcalc", "raster", n = 1L, name = "bio13")

		input <- paste0(ppt, collapse = ",")

		rgrass::execGRASS(
			cmd = "r.series",
			input = input,
			output = bio13,
			method = "maximum",
			nprocs = faster("cores"),
			memory = faster("memory"),
			flags = c(.quiet(), "overwrite", "n")
		)

		out <- c(out, bio13)
		names(out)[length(out)] <- "bio13"

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIOCLIM 14 (driest month precipitation)
	if (14 %in% bios) {

		bio14 <- .makeSourceName("r_mapcalc", "raster", n = 1L, name = "bio14")

		input <- paste0(ppt, collapse = ",")

		rgrass::execGRASS(
			cmd = "r.series",
			input = input,
			output = bio14,
			method = "minimum",
			nprocs = faster("cores"),
			memory = faster("memory"),
			flags = c(.quiet(), "overwrite", "n")
		)

		out <- c(out, bio14)
		names(out)[length(out)] <- "bio14"

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	### BIOCLIM 15 (CV of precipitation)
	if (15 %in% bios) {
		
		# as per dismo::biovars and WorldClim, add 1 to precipitation to avoid division by 0
		if (pptDelta != 0) {

			pptDeltaSrc <- .makeSourceName("bioclims", "raster", nLayers)
			for (i in seq_along(ppt)) {
				ex <- paste0(pptDeltaSrc[i], " = ", ppt[i], " + ", pptDelta)
				rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))
			}

			tasks <- tasks + 1
			if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

		} else {
			pptDeltaSrc <- ppt
		}

		sdSrc <- .makeSourceName("r_mapcalc", "raster", n = 1L, name = "bio15")

		rgrass::execGRASS(
			"r.series",
			input = paste(pptDeltaSrc, collapse = ","),
			output = sdSrc,
			method = "stddev",
			nprocs = faster("cores"),
			memory = faster("memory"),
			flags = c(.quiet(), "overwrite", "n")
		)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

		# convert to sample SD as per R
		if (sample) {
		
			sdSrcIn <- sdSrc
			sdSrc <- .makeSourceName("r_mapcalc", "raster", n = 1L)

			ex <- paste0(sdSrc, " = sqrt((", nLayers, " * (", sdSrcIn, "^2)) / (", nLayers, " - 1))")
			rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))
		
			on.exit(.rm(c(sdSrcIn), type = "raster", warn = FALSE, verify = FALSE), add = TRUE)

			tasks <- tasks + 1
			if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

		}

		meanSrc <- .makeSourceName("bioclims_r_series", "ppt_mean")
		rgrass::execGRASS(
			"r.series",
			input = paste(pptDeltaSrc, collapse = ","),
			output = meanSrc,
			method = "average",
			nprocs = faster("cores"),
			memory = faster("memory"),
			flags = c(.quiet(), "overwrite", "n")
		)

		bio15 <- .makeSourceName("r_mapcalc", "raster", name = "bio15")
		
		ex <- paste0(bio15, " = 100 * ", sdSrc, " / ", meanSrc)
		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		out <- c(out, bio15)
		names(out)[length(out)] <- "bio15"

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)
	
	}

	# BIOCLIM 16 (precipitation of the wettest quarter)
	if (16 %in% bios) {

		# burn precipitation of wettest quarter to cells
		bio16 <- .makeSourceName("r_mapcalc", "raster", n = 1L, name = "bio16")

		ex <- paste0(bio16, " = ")
		for (i in wettestQuarters) {
			ex <- paste0(ex, "if(", bio55, "==", i, ",", pptByQuarter[i], ",")
		}
		ex <- paste0(ex, "null()", paste(rep(")", length(wettestQuarters)), collapse = ""))

		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		out <- c(out, bio16)
		names(out)[length(out)] <- "bio16"

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO17: Precipitation of the driest quarter
	if (17 %in% bios) {

		# burn precipitation of driest quarter to cells
		bio17 <- .makeSourceName("r_mapcalc", "raster", n = 1L, name = "bio17")

		ex <- paste0(bio17, " = ")
		for (i in driestQuarters) {
			ex <- paste0(ex, "if(", bio56, "==", i, ",", pptByQuarter[i], ",")
		}
		ex <- paste0(ex, "null()", paste0(rep(")", length(driestQuarters)), collapse = ""))

		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		out <- c(out, bio17)
		names(out)[length(out)] <- "bio17"

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO18: Precipitation of the warmest quarter (based on mean temperature)
	if (18 %in% bios) {

		# burn precipitation of warmest quarter to cells
		bio18 <- .makeSourceName("r_mapcalc", "raster", n = 1L, name = "bio18")

		ex <- paste0(bio18, " = ")
		for (i in warmestQuarters) {
			ex <- paste0(ex, "if(", bio53, "==", i, ",", pptByQuarter[i], ",")
		}
		ex <- paste0(ex, "null()", paste(rep(")", length(warmestQuarters) - 1L), collapse = ""), ")")

		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		out <- c(out, bio18)
		names(out)[length(out)] <- "bio18"

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO19: Precipitation of the coldest quarter (based on mean temperature)
	if (19 %in% bios) {

		# burn precipitation of warmest quarter to cells
		bio19 <- .makeSourceName("r_mapcalc", "raster", n = 1L, name = "bio19")

		ex <- paste0(bio19, " = ")
		for (i in coldestQuarters) {
			ex <- paste0(ex, "if(", bio54, "==", i, ",", pptByQuarter[i], ",")
		}
		ex <- paste0(ex, "null()", paste(rep(")", length(coldestQuarters)), collapse = ""))

		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		out <- c(out, bio19)
		names(out)[length(out)] <- "bio19"

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO41: Temperature of the quarter following the coldest quarter (based on mean temperature)
	if (41 %in% bios) {

		# burn precipitation of warmest quarter to cells
		bio41 <- .makeSourceName("r_mapcalc", "raster", n = 1L, name = "bio41")
		n <- length(coldestQuarters)

		ex <- paste0(bio41, " = ")
		for (i in coldestQuarters) {

			if (i + quarter <= nLayers) {
				target <- i + quarter
			} else {
				target <- i + quarter - n
			}
			ex <- paste0(bio41, " = if(", bio54, "==", i, ",", tmeanByQuarter[target], ",")
	
		}
		ex <- paste0(ex, "null()", paste(rep(")", length(coldestQuarters)), collapse = ""))

		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		out <- c(out, bio41)
		names(out)[length(out)] <- "bio41"

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO42: Temperature of the quarter following the warmest quarter (based on mean temperature; deg C)
	if (42 %in% bios) {

		# burn precipitation of warmest quarter to cells
		bio42 <- .makeSourceName("bioclims_bio42", "raster")

		ex <- paste0(bio42, " = ")
		for (i in warmestQuarters) {

			if (i + quarter <= nLayers) {
				target <- i + quarter
			} else {
				target <- i + quarter - nLayers
			}
			ex <- paste0(bio42, " = if(", bio53, "==", i, ",", tmeanByQuarter[target], ",")
	
		}
		ex <- paste0(ex, "null()", paste(rep(")", length(warmestQuarters)), collapse = ""))

		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		out <- c(out, bio42)
		names(out)[length(out)] <- "bio42"

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO43: Precipitation of the quarter following the coldest quarter (based on mean temperature; mm)
	if (43 %in% bios) {

		# burn precipitation of warmest quarter to cells
		bio43 <- .makeSourceName("bioclims_bio43", "raster")

		ex <- paste0(bio43, " = ")
		for (i in coldestQuarters) {

			if (i + quarter <= nLayers) {
				target <- i + quarter
			} else {
				target <- i + quarter - nLayers
			}
			ex <- paste0(bio43, " = if(", bio54, "==", i, ",", pptByQuarter[target], ",")
	
		}
		ex <- paste0(ex, "null()", paste(rep(")", length(coldestQuarters)), collapse = ""))

		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		out <- c(out, bio43)
		names(out)[length(out)] <- "bio43"

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO44: Precipitation of the quarter following the warmest quarter (based on mean temperature; mm)
	if (44 %in% bios) {

		bio44 <- .makeSourceName("bioclims_bio44", "raster")

		ex <- paste0(bio44, " = ")
		for (i in warmestQuarters) {

			if (i + quarter <= nLayers) {
				target <- i + quarter
			} else {
				target <- i + quarter - nLayers
			}
			ex <- paste0(bio44, " = if(", bio53, "==", i, ",", pptByQuarter[target], ",")
	
		}
		ex <- paste0(ex, "null()", paste(rep(")", length(warmestQuarters)), collapse = ""))

		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		out <- c(out, bio44)
		names(out)[length(out)] <- "bio44"

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO45: Temperature of the quarter following the driest quarter (based on mean temperature; deg C)
	if (45 %in% bios) {

		bio45 <- .makeSourceName("bioclims_bio45", "raster")

		ex <- paste0(bio45, " = ")
		for (i in driestQuarters) {

			if (i + quarter <= nLayers) {
				target <- i + quarter
			} else {
				target <- i + quarter - nLayers
			}
			ex <- paste0(bio45, " = if(", bio56, "==", i, ",", tmeanByQuarter[target], ",")
	
		}
		ex <- paste0(ex, "null()", paste(rep(")", length(warmestQuarters)), collapse = ""))

		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		out <- c(out, bio45)
		names(out)[length(out)] <- "bio45"

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO46: Temperature of the quarter following the wettest quarter (based on mean temperature; deg C)
	if (46 %in% bios) {

		bio46 <- .makeSourceName("bioclims_bio46", "raster")

		ex <- paste0(bio46, " = ")
		for (i in wettestQuarters) {

			if (i + quarter <= nLayers) {
				target <- i + quarter
			} else {
				target <- i + quarter - nLayers
			}
			ex <- paste0(bio46, " = if(", bio55, "==", i, ",", tmeanByQuarter[target], ",")
	
		}
		ex <- paste0(ex, "null()", paste(rep(")", length(warmestQuarters)), collapse = ""))

		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		out <- c(out, bio46)
		names(out)[length(out)] <- "bio46"

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO47: Precipitation of the quarter following the driest quarter (based on mean temperature; mm)
	if (47 %in% bios) {

		bio47 <- .makeSourceName("bioclims_bio47", "raster")

		ex <- paste0(bio47, " = ")
		for (i in driestQuarters) {

			if (i + quarter <= nLayers) {
				target <- i + quarter
			} else {
				target <- i + quarter - nLayers
			}
			ex <- paste0(bio47, " = if(", bio56, "==", i, ",", pptByQuarter[target], ",")
	
		}
		ex <- paste0(ex, "null()", paste(rep(")", length(warmestQuarters)), collapse = ""))

		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		out <- c(out, bio47)
		names(out)[length(out)] <- "bio47"

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO48: Precipitation of the quarter following the wettest quarter (based on mean temperature; mm)
	if (48 %in% bios) {

		bio48 <- .makeSourceName("bioclims_bio48", "raster")

		ex <- paste0(bio48, " = ")
		for (i in wettestQuarters) {

			if (i + quarter <= nLayers) {
				target <- i + quarter
			} else {
				target <- i + quarter - nLayers
			}
			ex <- paste0(bio48, " = if(", bio55, "==", i, ",", pptByQuarter[target], ",")
	
		}
		ex <- paste0(ex, "null()", paste(rep(")", length(warmestQuarters)), collapse = ""))

		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		out <- c(out, bio48)
		names(out)[length(out)] <- "bio48"

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO49: Hottest month (based on maximum temperature)
	if (49 %in% bios) {

		bio49 <- .makeSourceName("bioclims_bio49", "raster")
		bio49Minus1 <- .makeSourceName("bioclims_bio49Minus1", "raster")
		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		rgrass::execGRASS(
			"r.series",
			input = tmax,
			output = bio49Minus1,
			method = "max_raster",
			nprocs = faster("cores"),
			memory = faster("memory"),
			flags = c(.quiet(), "overwrite", "n")
		)

		on.exit(.rm(bio49Minus1, type = "raster", warn = FALSE, verify = FALSE), add = TRUE)

		ex <- paste0(bio49, " = ", bio49Minus1, " + 1")
		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		out <- c(out, bio49)
		names(out)[length(out)] <- "bio49"

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO50: Coldest month (based on minimum temperature)
	if (50 %in% bios) {

		bio50 <- .makeSourceName("bioclims_bio50", "raster")
		bio50Minus1 <- .makeSourceName("bioclims_bio50Minus1", "raster")
		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		rgrass::execGRASS(
			"r.series",
			input = tmin,
			output = bio50Minus1,
			method = "min_raster",
			nprocs = faster("cores"),
			memory = faster("memory"),
			flags = c(.quiet(), "overwrite", "n")
		)

		on.exit(.rm(bio50Minus1, type = "raster", warn = FALSE, verify = FALSE), add = TRUE)

		ex <- paste0(bio50, " = ", bio50Minus1, " + 1")
		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		out <- c(out, bio50)
		names(out)[length(out)] <- "bio50"

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO51: Wettest month.
	if (51 %in% bios) {

		bio51 <- .makeSourceName("bioclims_bio51", "raster")
		bio51Minus1 <- .makeSourceName("bioclims_bio51Minus1", "raster")
		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		rgrass::execGRASS(
			"r.series",
			input = ppt,
			output = bio51Minus1,
			method = "max_raster",
			nprocs = faster("cores"),
			memory = faster("memory"),
			flags = c(.quiet(), "overwrite", "n")
		)

		on.exit(.rm(bio51Minus1, type = "raster", warn = FALSE, verify = FALSE), add = TRUE)

		ex <- paste0(bio51, " = ", bio51Minus1, " + 1")
		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		out <- c(out, bio51)
		names(out)[length(out)] <- "bio51"

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO52: Driest month.
	if (52 %in% bios) {

		bio52 <- .makeSourceName("bioclims_bio52", "raster")
		bio52Minus1 <- .makeSourceName("bioclims_bio52Minus1", "raster")
		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		rgrass::execGRASS(
			"r.series",
			input = ppt,
			output = bio52Minus1,
			method = "min_raster",
			nprocs = faster("cores"),
			memory = faster("memory"),
			flags = c(.quiet(), "overwrite", "n")
		)

		on.exit(.rm(bio52Minus1, type = "raster", warn = FALSE, verify = FALSE), add = TRUE)

		ex <- paste0(bio52, " = ", bio52Minus1, " + 1")
		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		out <- c(out, bio52)
		names(out)[length(out)] <- "bio52"

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO57 and BIO58
	if (any(c(57, 58) %in% bios)) {

		# difference between temperature of each month and the next
		deltaTmeanSrcs <- .makeSourceName("bioclims_delta_temp", "raster", nLayers)
		for (i in seq_len(nLayers)) {

			iNext <- if (i == nLayers) { 1L } else { i + 1L }
			ex <- paste0(deltaTmeanSrcs[i], " = ", tmean[i], " - ", tmean[iNext])
			rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		}
		on.exit(.rm(deltaTmeanSrcs, type = "raster", warn = FALSE, verify = FALSE), add = TRUE)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

		# BIO57: The largest *decrease* in temperature from one month to the next
		if (57 %in% bios) {
			
			bio57 <- .makeSourceName("bioclims_bio57", "raster")
			rgrass::execGRASS(
				"r.series",
				input = deltaTmeanSrcs,
				output = bio57,
				method = "maximum",
				nprocs = faster("cores"),
				memory = faster("memory"),
				flags = c(.quiet(), "overwrite", "n")
			)

			out <- c(out, bio57)
			names(out)[length(out)] <- "bio57"

			tasks <- tasks + 1
			if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

		}

		# BIO58: The largest *increase* in temperature from one month to the next
		if (58 %in% bios) {
			
			src <- .makeSourceName("bioclims_bio58_delta", "raster")
			rgrass::execGRASS(
				"r.series",
				input = deltaTmeanSrcs,
				output = src,
				method = "minimum",
				nprocs = faster("cores"),
				memory = faster("memory"),
				flags = c(.quiet(), "overwrite", "n")
			)

			bio58 <- .makeSourceName("bioclims_bio58", "raster")
			ex <- paste0(bio58, " = -1 * ", src)
			rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

			out <- c(out, bio58)
			names(out)[length(out)] <- "bio58"

			tasks <- tasks + 1
			if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

		}

	}

	# BIO59 and BIO60: The largest decrease/increase in precipitation from one month to the next
	if (any(c(59, 60) %in% bios)) {

		# difference between precip of each month and the next
		deltaPptSrcs <- .makeSourceName("bioclims_delta_ppt", "raster", nLayers)
		for (i in seq_len(nLayers)) {

			iNext <- if (i == nLayers) { 1L } else { i + 1L }
			ex <- paste0(deltaPptSrcs[i], " = ", ppt[i], " - ", ppt[iNext])
			rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))
		}
		on.exit(.rm(deltaPptSrcs, type = "raster", warn = FALSE, verify = FALSE), add = TRUE)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

		# BIO59: The largest decrease in precipitation from one month to the next
		if (59 %in% bios) {
			
			bio59 <- .makeSourceName("bioclims_bio59", "raster")
			rgrass::execGRASS(
				"r.series",
				input = deltaPptSrcs,
				output = bio59,
				method = "maximum",
				nprocs = faster("cores"),
				memory = faster("memory"),
				flags = c(.quiet(), "overwrite", "n")
			)

			out <- c(out, bio59)
			names(out)[length(out)] <- "bio59"

			tasks <- tasks + 1
			if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

		}

		# BIO60: The largest increase in precipitation from one month to the next
		if (60 %in% bios) {
			
			src <- .makeSourceName("bioclims_bio60_delta", "raster")
			rgrass::execGRASS(
				"r.series",
				input = deltaPptSrcs,
				output = src,
				method = "minimum",
				nprocs = faster("cores"),
				memory = faster("memory"),
				flags = c(.quiet(), "overwrite", "n")
			)

			bio60 <- .makeSourceName("bioclims_bio60", "raster")
			ex <- paste0(bio60, " = -1 * ", src)
			rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

			out <- c(out, bio60)
			names(out)[length(out)] <- "bio60"

			tasks <- tasks + 1
			if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

		}

	}

	# sort BIOCLIMs
	outNames <- names(out)
	nc <- nchar(outNames)
	shorts <- out[nc == 4L]
	longs <- out[nc == 5L]

	shorts <- shorts[order(names(shorts))]
	longs <- longs[order(names(longs))]

	out <- c(shorts, longs)
	
	# return
	names <- names(out)
	out <- makeGRaster(out, names = names)

	if (verbose | faster("verbose")) {
		tasks <- tasks + 1
		utils::setTxtProgressBar(pb, tasks)
		close(pb)
	}
	out


	} # EOF
)


#' Calculate a series of "quarter" rasters and return their [sources()] names. A quarter can be summarized using 
#'
#' x Character vector of raster [sources()] names (usually 12 or 52)
#' fun Character: Name of function to summarize across a quarter. Allowed values are "sum", "min", "max", or "mean".
#' quarter Numeric: Length of a quarter
#'
#' @noRd
.calcQuarterGR <- function(x, fun, quarter) {

	qs <- c(x, x[1L:(quarter - 1L)])
	n <- length(x)
	srcs <- .makeSourceName("r_mapcalc", "raster", n = n)
	for (i in seq_len(n)) {

		select <- qs[i:(i + quarter - 1L)]

		method <- if (fun == "sum") {
			"sum"
		} else if (fun == "min") {
			"minimum"
		} else if (fun == "max") {
			"maximum"
		} else if (fun == "mean") {
			"average"
		}

		rgrass::execGRASS(
			"r.series",
			input = select,
			output = srcs[i],
			method = method,
			nprocs = faster("cores"),
			memory = faster("memory"),
			flags = c(.quiet(), "overwrite", "n")
		)

	} # next quarter
	srcs

}

#' Replace a raster layer or a column from a vector's data table
#'
#' @description The `$<-` notation can be used to replace a specific layer in a multi-layer `GRaster`, or a to replace a specific column from a `GVector`'s data table.
#'
#' @param x A `GRaster` or `GVector`.
#' @param name Character: Name of the `GRaster` layer to replace, or name of the `GVector` column to replace.
#' @param value Character: The name of a `GRaster` layer or the name of a column in a `GVector`'s data table. Names of rasters and vector tables' columns cab be obtained using [names()].
#'
#' @returns A `GRaster` or the column of a `GVector`.
#'
#' @example man/examples/ex_GRaster_GVector_subset_assign.r
#'
#' @seealso [$], \code{\link[fasterRaster]{[[<-}}, and \code{\link[fasterRaster]{add<-}}
#'
#' @name $<-
#' @aliases $<-,GRaster-method
#' @docType methods
#' @rdname replace_dollar
#' @exportMethod $<-
methods::setMethod(
    "$<-",
    signature = c(x = "GRaster"),
    function(x, name, value) {

	if (is.logical(name)) {
		if (length(name) < nlyr(x)) name <- rep(name, length.out = nlyr(x))
		i <- which(name)
	} else if (is.character(name)) {
		i <- match(name, names(x))
	}

	x[[name]] <- value
	names(x)[i] <- names(value)
	x

	} # EOF
)

#' @name $<-
#' @aliases $<-,GVector-method
#' @docType methods
#' @rdname replace_dollar
#' @exportMethod $<-
methods::setMethod(
	"$<-",
	signature = c(x = "GVector"),
	function(x, name, value) {

	ng <- ngeom(x)
	if (!is.null(value)) {
		lvalue <- length(value)
		if (lvalue < ng) value <- rep(value, length.out = ng) 
	}
	x@table <- data.table::set(x@table, i = NULL, j = name, value = value)
	x

	} # EOF
)

#' Statistics on cells of a GRaster stratified by cells of another raster
#'
#' @description Function `zonal()` calculates statistics (mean, sum, etc.) on cells of a `GRaster` by "zones" created by cells of another `GRaster` or `GVector`.
#'
#' @param x A `GRaster` for which to calculate summary statistics.
#'
#' @param z A `GRaster` or `GVector` used to define zones:
#' * If `z` is a `GRaster`, then it must be of type `integer` or `factor` (see `vignette("GRasters", package = "fasterRaster")`). Zones will be established based on cells that have the same value in this raster.
#' * If `z` is a `GVector`, zones will be created for each geometry. If geometries overlap, then the zonal statistics will be calculated for the ones on top. Thus statistics for the zones defined by geometries below these may not represent all the cells covered by that geometry.
#'
#' @param fun Character vector: Name of the function(s) to summarize `x` with. These can include:
#' * `"*"`: All of the functions below.
#' * `"cv"`: Sample coefficient of variation (expressed as a proportion of the mean).
#' * `"cvpop"`: Population coefficient of variation (expressed as a proportion of the mean).
#' * `"max"` and `"min"`: Highest and lowest values across non-`NA` cells.
#' * `"mean"` (default): Average.
#' * `"meanAbs"`: Mean of absolute values.
#' * `"median"`: Median.
#' * `"quantile"`: Quantile (see also argument `probs`).
#' * `"range"`: Range.
#' * `"sd"`: Sample standard deviation.
#' * `"sdpop"`: Population standard deviation.
#' * `"sum"`: Sum.
#' * `"var"`: Sample variance.
#' * `"varpop"`: Population variance.
#'
#' @param probs Numeric: Quantile at which to calculate `quantile`. Only a single value between 0 and 1 is allowed.
#'
#' @returns A `data.frame` or `data.table`.
#'
#' @example man/examples/ex_zonal.r
#'
#' @aliases zonal
#' @rdname zonal
#' @exportMethod zonal
methods::setMethod(
	f = "zonal",
	signature = c(x = "GRaster"),
	function(x, z, fun = "mean", probs = 0.5) {
	
	if (inherits(z, "GRaster")) {
		.zonalByRaster(x = x, z = z, fun = fun, probs = probs)
	} else if (inherits(z, "GVector")) {
		gtype <- geomtype(z, grass = TRUE)
		.zonalByVector(x = x, z = z, fun = fun, probs = probs, gtype = gtype)
	} else {
		stop("Argument `z` must be a GRaster or GVector.")
	}

	} # EOF
)

#' @noRd
.zonalByRaster <- function(x, z, fun, probs) {

	if (!is.cell(z)[1L] & !is.factor(z)[1L]) stop("GRaster `z` must be an integer or factor raster, or a GVector.")

	if (nlyr(z) > 1L) {
	
		warning("The zone raster in `z` has more than one layer. Only the first will be used.")
		z <- z[[1L]]
	
	}

	compareGeom(x, z)
	.locationRestore(x)
	.region(x)

	### get zone labels
	if (is.factor(z)) z <- droplevels(z)
	freqs <- freq(z)
	zones <- freqs[["value"]]
	
	.zonal(
		x = sources(x),
		z = sources(z),
		fun = fun,
		probs = probs,
		zones = zones,
		xnames = names(x)
	)

}

#' Internal function for zonal() when y is a GVector
#'
#' @param x	GRaster [sources()] name.
#' @param z	GVector [sources()] name.
#' @param fun	Character
#' @param gtype `geomtype(z, grass = TRUE)` ("area", "line", or "point")
#'
#' @keywords internal
.zonalByVector <- function(x, z, fun, probs, gtype) {

	compareGeom(x, z)
	.locationRestore(x)
	.region(x)

	# rasterize by field in data table
	cats <- .vCats(z)
	uniCats <- unique(cats)
	db <- data.table::data.table(cat = uniCats, recats = uniCats)
	.vAttachDatabase(z, db, cats = cats)

	zonalSrc <- .makeSourceName("zonal_v_to_rast", "raster")
	args <- list(
		cmd = "v.to.rast",
		input = sources(z),
		output = zonalSrc,
		use = "attr",
		attribute_column = "recats",
		type = gtype,
		memory = faster("memory"),
		flags = c(.quiet(), "overwrite")
	)

	if (gtype == "line") args$flags <- c(args$flags, "d")
	do.call(rgrass::execGRASS, args = args)
	on.exit(.rm(zonalSrc, type = "raster", warn = FALSE), add = TRUE)
	
	zones <- .freq(zonalSrc, dtype = "CELL")
	zones <- zones[["value"]]

	.zonal(
		x = sources(x),
		z = zonalSrc,
		fun = fun,
		probs = probs,
		zones = zones,
		xnames = names(x)
	)

}

#' Internal function for zonal()
#' 
#' @param x [sources()] name of `GRaster`.
#' @param z [sources()] name of "zones" `GRaster`.
#' @param fun Character: Name of function(s).
#' @param probs Numeric in the range 0 to 1, inclusive.
#' @param zones Vector of zone values (integers).
#' @param xnames Character: Names of `x`.
#' @keywords internal
.zonal <- function(x, z, fun, probs, zones, xnames) {

	### get zone labels
	numZones <- length(zones)

	### calculate statistics
	nLayers <- length(x)
	out <- data.table()
	for (i in seq_len(nLayers)) {
	
		srcs <- .makeSourceName("zonal_r_mapcalc", "raster", numZones)
	
		for (j in seq_along(zones)) {
		
			ex <- paste0(srcs[j], " = if (", z, " == ", zones[j], ", ", x[i], ", null())")
		
			rgrass::execGRASS(
				cmd = "r.mapcalc",
				expression = ex,
				flags = c(.quiet(), "overwrite")
			)

			thisOut <- .global(x = srcs[j], fun = fun, probs = probs)

			prepend <- data.table::data.table(layer = i, name = xnames[i], zone = zones[j])

			thisOut <- cbind(prepend, thisOut)
			if (nrow(out) == 0L) {
				out <- thisOut
			} else {
				out <- rbind(out, thisOut)
			}

		} # next zone in z
	
	} # next x layer

	if (!faster("useDataTable")) out <- as.data.frame(out)
	out

}

if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster
elev <- fast(madElev)

# Calculate all topographic metrics
topos <- terrain(elev, v = "*")
topos

plot(topos) # NB Aspect has values of NA when it cannot be defined

# Calculate a hillshade raster
hs <- hillshade(elev)
plot(hs)

}

#' Replace a specific value(s) in a GRaster
#'
#' @description This function replaces one or more user-specified values in a raster with other values. See [classify()] for replacing ranges of values.
#'
#' @param x A `GRaster`.
#'
#' @param from,to Vectors of numeric or character values. The value(s) in `from` will be replaced with the value(s) in `to`. They must be the same length, or, if you supply a single value for `to`, then all values in `from` will be converted to the same value of `to`. Numeric/integer or character vectors can be used:
#' * `from` and `to` are numeric or integer vectors: Values in `from` will be replaced by their corresponding value in `to`.
#' * `from` and `to` are character vectors: You can use a character vector for `from` and `to`. In this case, the input raster must be a factor (categorical) raster or an integer raster. If `from` is a character vector, these levels(categories) will be replaced by the levels in `to.` This can add levels to a `GRaster` `to` has labels that do not match any existing labels in `from`.
#' * `from` is a character vector and `to` is an integer vector: Cells in `x` that correspond to the given label will have their values replaced by the corresponding value in `to`, and be matched the the corresponding label. If no label corresponds to the new value, a new level will be created. The input must be a categorical raster.
#' * `from` is an integer vector and `to` is a character vector: Cells in `x` that have a value in `from` will be replaced by values that match the labels in `to`. If the input raster does not have a value that corresponds to the given label in `y`, then a new value will be created.
# * You can include `NA` in either `from` or `to` to replace `NA` values or to assign `NA` values.
#' 
#' @param others `NULL` (default), `NA`, numeric, or a character:
#' * `NULL` (default): Values that do not appear in `from` will be unchanged.
#' * `NA`: Values that do not appear in `from` will be set to `NA`.
#' * Character: Cells in `x` that do not appear in `to` will be assigned to this level. In this case, `x` must be a categorical (factor) raster.
#'
#' @param warn Logical: If `TRUE` (default), display a warning when new levels are created.
#'
#' @returns A `GRaster`.
#'
#' @seealso [terra::subst()], [classify()]
#'
#' @example man/examples/ex_subst.r
#'
#' @aliases subst
#' @rdname subst
#' @exportMethod subst
methods::setMethod(
	f = "subst",
	signature = c(x = "GRaster"),
	function(x, from, to, others = NULL, warn = TRUE) {
	
	if (length(from) != length(to) & length(to) != 1L) stop("Arguments `from` and `to` must have the same length.")
	if (is.character(from) & !all(is.factor(x))) stop("You can only use a character vector for `from` if all\n  layers in `x` are categorical rasters.")
	if (is.character(to) & any(datatype(x, "GRASS") != "CELL")) stop("You can only use a character vector for `to` if all\n  layers in `x` are categorical or integer.")
	if (is.character(others) & !all(is.factor(x))) stop("You can only use a character for `others` if all\n  layers in `x` are categorical.")

	if (length(from) > 1L & length(to) == 1L) to <- rep(to, length(from))

	.locationRestore(x)
	.region(x)

	levs <- cats(x)
	nLayers <- nlyr(x)
	srcs <- .makeSourceName("subst_r_mapcalc", "raster", n = nLayers)
	for (i in seq_len(nLayers)) {
	
		if (is.factor(x)[i]) {
			
			thisLev <- levs[[i]]
			ac <- activeCat(x, layer = i)

			# `from` is a character vector
			if (is.character(from)) {

				if (faster("useDataTable")) {
					thisFrom <- thisLev[[1L]][match(from, thisLev[[ac + 1L]])]
				} else {
					thisFrom <- thisLev[ , 1L][match(from, thisLev[  , ac + 1L, drop = TRUE])]
				}
				if (any(is.na(thisFrom) & !is.na(from))) {
					stop("Argument `from` contains one or more values that do not appear in the GRaster's levels table.\n  These will be added ")
				}

			# `from` is numeric/logical
			} else {
				thisFrom <- from
				thisFrom <- as.integer(thisFrom)
			}
		
			if (any(is.character(to))) {

				if (faster("useDataTable")) {
					thisTo <- thisLev[[1L]][match(to, thisLev[[ac + 1L]])]
				} else {
					thisTo <- thisLev[ , 1L][match(to, thisLev[  , ac + 1L, drop = TRUE])]
				}

				# create new levels
				if (any(is.na(thisTo) & !is.na(to))) {
				
					if (warn) warning("Not all labels in `to` are found in `x`. New levels will be created.")
					
					topVal <- max(.maxVal(x)[i], thisLev[[ac]])
					nNew <- sum(is.na(unique(thisTo)) & !is.na(unique(to)))
					newVals <- 1L:nNew + topVal
					thisTo[is.na(thisTo) & !is.na(to)] <- newVals
					thisTo <- as.integer(thisTo)
					
					# create table to `rbind` to levels table
					newLev <- .addLevels(ac = ac, thisLev = thisLev, thisTo = thisTo, labels = to)

					# add new levels to table
					thisLev <- rbind(thisLev, newLev)
			
				}

			# `to` is integer/numeric
			} else if (!is.character(to)) {

				thisTo <- to
				if (any(omnibus::countDecDigits(thisTo) > 0L)) stop("When using `subst()` on a categorical rasters, the\n  values in `to` must be integers or character values.")
				thisTo <- as.integer(thisTo)

				# create table to `rbind` to levels table
				newLev <- .addLevels(ac = ac, thisLev = thisLev, thisTo = thisTo, labels = NULL)

				# add new levels to table
				thisLev <- rbind(thisLev, newLev)
			
			}
	
			# # remove levels that were re-assigned
			# removes <- thisFrom[!(thisFrom %in% thisTo)]
			# removes <- removes[!is.na(removes)]
			# if (length(removes) > 0L) levs[[i]] <- levs[[i]][!(levs[[i]][[1L]] %in% removes)]

			# remove duplicates
			if (faster("useDataTable")) {
				levs[[i]] <- thisLev[!duplicated(thisLev[ , c(1L, ac + 1L), with = FALSE])]
			} else {
				levs[[i]] <- thisLev[!duplicated(thisLev[ , c(1L, ac + 1L), drop = FALSE]), , drop = FALSE]
			}

		} else {

			# raster is not a factor raster
			thisFrom <- from
			thisTo <- to

		}

		ex <- paste0(srcs[i], " = ")
		for (j in seq_along(thisFrom)) {
			if (is.na(thisFrom[j]) & is.na(thisTo[j])) {
				ex <- paste0(ex, "if(isnull(", sources(x)[i], "),null(),")
			} else if (is.na(thisFrom[j]) & !is.na(thisTo[j])) {
				ex <- paste0(ex, "if(isnull(", sources(x)[i], "),", thisTo[j], ",")
			} else if (!is.na(thisFrom[j]) & is.na(thisTo[j])) {
				ex <- paste0(ex, "if(", sources(x)[i], "==", thisFrom[j], ",null(),")
			} else {
				ex <- paste0(ex, "if(", sources(x)[i], "==", thisFrom[j], ",", thisTo[j], ",")
			}
		}

		if (is.null(others)) {
			ex <- paste0(ex, sources(x)[i])
		} else if (is.na(others)) {
			ex <- "null()"
		} else if (is.character(others) & is.factor(x)[i]) {

			# get new value for `others`
			if (faster("useDataTable")) {
				thisOthers <- levs[[i]][levs[[i]][[ac + 1]] == others, 1L]
				thisOthers <- thisOthers[[1L]]
			} else {
				# thisOthers <- levs[[i]][ , ac + 1L, drop = TRUE] == others, 1L]
				thisOthers <- levs[[i]][levs[[i]][ , ac + 1L, drop = TRUE] == others, 1L]
			}
	
			if (length(thisOthers) == 0L) {
				
				# add new level
				if (warn) warning("the value in `others` is not found among existing level labels. A new level will be created.")
				newVal <- max(0L, .maxVal(x)[i], max(thisTo, na.rm = TRUE)) + 1L
				newVal <- as.integer(newVal)
				newLabel <- others
				newLev <- .addLevels(ac = ac, thisLev = thisLev, thisTo = newVal, labels = others)
				levs[[i]] <- rbind(levs[[i]], newLev)
				ex <- paste0(ex, newVal)

			} else {
				ex <- paste0(ex, thisOthers)
			}

		
		} else if (is.character(others) & !is.factor(x)[i]) {
			stop("Cannot add a level to a non-factor GRaster. To make a GRaster categorical, use `levels<-` or `categories`.")
		} else {
			ex <- paste0(ex, others)
		}

		ex <- paste0(ex, paste(rep(")", length(thisFrom)), collapse = ""))
		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))
	
	}
	makeGRaster(srcs, names = names(x), levels = levs)
	
	}
)


#' Creates a new data.table with same structure as the existing levels table
#'
#' @param ac Integer: Active category index (increment by 1 to get the actual column).
#' @param thisLev `data.table`: The existing levels table, to be used as a template.
#' @param thisTo Integer(s): The new values to add (integers).
#' @param `NULL` or character: The new label(s) to add (`NULL` if they already exist in `thisLev`, or character if not).
#'
#' @returns A `data.table`.
#'
#' @noRd
.addLevels <- function(ac, thisLev, thisTo, labels = NULL) {

	# create table to `rbind` to levels table
	for (j in 1L:ncol(thisLev)) {
		if (j == 1L) {
			newLev <- data.table::data.table(DUMMY__ = thisTo)
		} else {
			if (j == ac + 1L) {

				if (!is.null(labels)) {
					thisLabels <- labels
				} else {
					thisLabels <- thisLev[[j]][match(thisTo, thisLev[[1L]])]
				}
				add <- data.table::data.table(DUMMY__ = thisLabels)
			
			} else {
				add <- data.table::data.table(DUMMY__ = NA)
			}
			newLev <- cbind(newLev, add)
		}
		names(newLev)[j] <- names(thisLev)[j]
	}
	newLev <- newLev[!duplicated(newLev)]
	newLev <- newLev[!is.na(newLev[[1L]])]

	newLev

}

#' File extension
#'
#' @description Returns the file extension. Borrowed from **tools** package.
#'
#' @param x Character.
#'
#' @returns Character.
#'
#' @aliases fileExt
#' @noRd
.fileExt <- function(x) {

	pos <- regexpr("\\.([[:alnum:]]+)$", x)
    ifelse(pos > -1L, substring(x, pos + 1L), "")

}

if (grassStarted()) {

# Setup
library(terra)

# Example data:
madForest <- fastData("madForest2000") # raster

### Fragmentation classes from a SpatRaster

fragTerra <- fragmentation(madForest)
plot(fragTerra)
levels(fragTerra)
freq(fragTerra)

### Fragmentation classes from a GRaster

# Convert to GRaster:
forest <- fast(madForest)

# Fragmentation class:
frag <- fragmentation(forest)
plot(frag)
levels(frag)
freq(frag)

}

if (grassStarted()) {

# Setup
library(sf)

# Rivers vector
madRivers <- fastData("madRivers")

# Convert sf to a GVector
rivers <- fast(madRivers)

# Convert GVector to data.frame or data.table
as.data.frame(rivers)
as.data.table(rivers)

# Subset rivers vector
rivers1 <- rivers[1:2]
rivers2 <- rivers[10:11]

# Concatenate rivers
riversCombo <- rbind(rivers1, rivers2)
riversCombo

# Add columns
newCol <- data.frame(new = 1:11)
riversCol <- colbind(rivers, newCol)
riversCol

# Remove table
riversCopy <- rivers
riversCopy # has data table
riversCopy <- dropTable(riversCopy)
riversCopy # no data table

# Add a new table
newTable <- data.frame(num = 1:11, letters = letters[1:11])
addTable(riversCopy) <- newTable
riversCopy

}

#' Replace NAs in a data.table or data.frame column, or in a vector
#'
#' @description This function replaces `NA`s in one or more `data.table`, `data.frame`, or `matrix` columns, or in vectors, with a user-defined value.
#'
#' @param x A `data.table` or `data.frame` or `matrix`, or a vector of numeric, integer, logical, or character values.
#'
#' @param replace A value of any atomic class (numeric, integer, character, Date, etc.): Value to to replace `NA`s.
#'
#' @param cols `NULL`, character, numeric, integer, or logical vector: Indicates columns for which to replace `NA`s. If `NULL`, then all columns will have `NA`s replaced. If a character, this is the column name(s). If numeric or integer, this is the columns' indices. If logical, columns with `TRUE` have `NA`s replaced. If a logical vector has fewer than the total number of columns, it will be recycled.
#'
#' @returns A `data.table`, `data.frame`, `matrix`, or vector.
#'
#' @example man/examples/ex_data_table.r
#'
#' @aliases replaceNAs
#' @rdname replaceNAs
#' @exportMethod replaceNAs
methods::setMethod(
	f = "replaceNAs",
	signature = c(x = "data.frame"),
	function(x, replace, cols = NULL) {
	
	if (length(replace) != 1L) stop("Argument `replace` must be a single value.")

	if (is.null(cols)) cols <- seq_len(ncol(x))

	if (is.logical(cols)) {
		if (length(cols) < ncol(x)) cols <- rep(cols, length.out = ncol(x))
		cols <- which(cols)
	}

	for (j in cols) {
	
		if (anyNA(x[ , j, drop = TRUE])) {
			x[is.na(x[ , j, drop = TRUE]), j] <- replace
		}
	
	}
	x

	} # EOF
)

#' @aliases replaceNAs
#' @rdname replaceNAs
#' @exportMethod replaceNAs
methods::setMethod(
	f = "replaceNAs",
	signature = c(x = "matrix"),
	function(x, replace, cols = NULL) {
	
	if (length(replace) != 1L) stop("Argument `replace` must be a single value.")

	if (is.null(cols)) cols <- seq_len(ncol(x))

	if (is.logical(cols)) {
		if (length(cols) < ncol(x)) cols <- rep(cols, length.out = ncol(x))
		cols <- which(cols)
	}

	for (j in cols) {
	
		if (anyNA(x[ , j, drop = TRUE])) {
			x[is.na(x[ , j, drop = TRUE]), j] <- replace
		}
	
	}
	x

	} # EOF
)

#' @aliases replaceNAs
#' @rdname replaceNAs
#' @exportMethod replaceNAs
methods::setMethod(
    f = "replaceNAs",
    signature = c(x = "data.table"),
    function(x, replace, cols = NULL) {

		if (length(replace) != 1L) stop("Argument `replace` must be a single value.")

		if (is.null(cols)) cols <- seq_len(ncol(x))
        
		if (is.logical(cols)) {
            if (length(cols) < ncol(x)) cols <- rep(cols, length.out = ncol(x))
            cols <- which(cols)
        }

		# replace!
		if ((is.numeric(replace) | is.integer(replace)) && all(sapply(x, "class") == "numerfic")) {
		
   			data.table::setnafill(x, type = "const", fill = replace, cols = cols)
		
		} else {

			for (j in cols) {

				data.table::set(x, which(is.na(x[[j]])), j, replace)

			}
		}
		x

    } # EOF
)

#' @aliases replaceNAs
#' @rdname replaceNAs
#' @exportMethod replaceNAs
methods::setMethod(
    f = "replaceNAs",
    signature = c(x = "numeric"),
    function(x, replace) .replaceNAsAtomic(x, replace)
)

#' @aliases replaceNAs
#' @rdname replaceNAs
#' @exportMethod replaceNAs
methods::setMethod(
    f = "replaceNAs",
    signature = c(x = "integer"),
    function(x, replace) .replaceNAsAtomic(x, replace)
)

#' @aliases replaceNAs
#' @rdname replaceNAs
#' @exportMethod replaceNAs
methods::setMethod(
    f = "replaceNAs",
    signature = c(x = "logical"),
    function(x, replace) .replaceNAsAtomic(x, replace)
)

#' @aliases replaceNAs
#' @rdname replaceNAs
#' @exportMethod replaceNAs
methods::setMethod(
    f = "replaceNAs",
    signature = c(x = "character"),
    function(x, replace) .replaceNAsAtomic(x, replace)
)

#' Replace NAs in an atomic vector
#'
#' @param x A vector of numeric, integer, logical, or character values.
#' @param replace What to replace `NA` with.
#'
#' @returns A vector of (usually) the same class as the input.
#'
#' @noRd
.replaceNAsAtomic <- function(x, replace) {

	if (length(replace) != 1L) stop("Argument `replace` must be a single value.")
	x[is.na(x)] <- replace
	x

}

#' Area of GRaster cells
#'
#' @description `cellArea()` returns a raster will cell values equal to their area. To get the area of all cells of a raster, see [expanse()].
#'
#' @param x A `GRaster`.
#'
#' @param mask Logical: If `TRUE`, then cells that are `NA` in `x` are also `NA` in the output.
#'
#' @param lyrs Logical:
#' * If `lyrs` is `FALSE` (default), then the output has a single layer. In this case, if `mask` is `TRUE`, then cells that are `NA` in the *first* layer are also `NA` in the output.
#' * If `lyrs` is `TRUE`, then the output has the same number of layers as `x` if `mask` is also `TRUE`. In this case, cells that area `NA` in the input will also be `NA` in the output in the respective layer.
#'
#' @param unit Character: Units of area. Partial matching is used, and case is ignored. Can be any of:
#' * `"meters2"` (default), `"metres2"`, or `"m2"`
#' * `"km2"` or `"kilometers2"`
#' * `"ha"` or `"hectares"`
#' * `"ac"` or `"acres"`
#' * `"mi2"` or `"miles2"`
#' * `"ft2"` or `"feet2"`
#'
#' @returns A `GRaster`.
#'
#' @example man/examples/ex_cellSize.r
#'
#' @seealso [terra::cellSize()], [expanse()], [zonalGeog()], [omnibus::convertUnits()]
#'
#' @aliases cellSize
#' @rdname cellSize
#' @exportMethod cellSize
methods::setMethod(
	f = "cellSize",
	signature = c(x = "GRaster"),
	function(x, mask = FALSE, lyrs = FALSE, unit = "meters2") {
	
	units <- c("m2", "meters2", "metres2", "km2", "kilometers2", "ha", "hectares", "ac", "acres", "mi2", "miles2", "ft2", "feet2")
	unit <- omnibus::pmatchSafe(unit, units, useFirst = TRUE, nmax = 1L)

	if (unit == "metres") unit <- "meters"
	unit <- omnibus::expandUnits(unit)

	f <- omnibus::convertUnits(from = "meters2", to = unit)
	f <- format(f, scientific = FALSE)

	.region(x)

	if (lyrs & mask) {
		nLayers <- nlyr(x)
	} else {
		nLayers <- 1L
	}

	### mask
	if (mask) {
		
		# remove mask on exit
		on.exit(.removeMask(), add = TRUE)

		# make mask from first layer
		rgrass::execGRASS(
			cmd = "r.mask",
			raster = sources(x)[1L],
			flags = c(.quiet(), "overwrite")
		)

	}

	srcs <- .makeSourceName("r_mapcalc", "raster", n = nLayers)

	for (i in seq_len(nLayers)) {

		# mask
		if (i > 1L & mask & lyrs) {
		
			rgrass::execGRASS(
				cmd = "r.mask",
				raster = sources(x)[i],
				flags = c(.quiet(), "overwrite")
			)
		
		}

		ex <- if (mask) {
			paste0(srcs[i], " = area() * MASK * ", f)
		} else {
			paste0(srcs[i], " = area() * ", f)
		}

		rgrass::execGRASS(
			cmd = "r.mapcalc",
			expression = ex,
			flags = c(.quiet(), "overwrite")
		)

	} # next raster layer

	makeGRaster(srcs, paste0("area_", unit, "_", seq_len(nLayers)))
		
	} # EOF
)

#' Classify GRaster cell values
#'
#' @description This function classifies a `GRaster`` so that cells that have values within a given range are assigned a new value. The [subst()] function is a simpler method for replacing specific values or category levels.
#'
#' @param x A `GRaster`.
#' @param rcl Reclassification system:
#' * A single integer: Number of "bins" into which to divide values. Arguments `include.lowest` and `right` apply.
#' * A vector of numeric values: Breakpoints of bins into which to divide values. These will be sorted from lowest to highest before classification. Arguments `include.lowest` and `right` apply.
#' * A 2-column `matrix`, `data.frame`, or `data.table`: The first column provides specific values in `x` to be replaced, and the second provides the values they are replaced with. This method is only useful for classifying `integer` `GRaster`s. Arguments `include.lowest` and `right` are ignored. Cells will be classified in the order in which values are listed in the first column.
#' * A 3-column `matrix`, `data.frame`, or `data.table`: The first column provides the lower value of a bin, the second the upper value, and the third the value to assign to the cells in the bin. Arguments `include.lowest` and `right` apply. Cells will be classified in the order of how intervals are listed (intervals will not be sorted).
#'
#' @param include.lowest,right Logical: These arguments determine how cells that have values exactly equal to the lower or upper ends of an interval are classified.
# * `include.lowest = FALSE` (default) and `right = TRUE` (also default): Intervals will be "left-open, right-closed". Cells with values equal to the lowest lower boundary will not be reclassified.
#' * `include.lowest = TRUE` and `right = TRUE`: All intervals will be "left-open, right-closed" except for the lowest interval, which will be "left-closed/right-closed".
#' * `include.lowest = FALSE` and `right = FALSE`: Intervals will be "left-closed/right-open". Cells with values equal to the highest higher boundary will not be reclassified.
#' * `include.lowest = TRUE` and `right = FALSE`: All intervals will be "left-closed/right-open", except for the highest interval, which will be "right-closed/left-closed".
#' * `right = NA`: Only useful for classifying `integer` `GRaster`s. All intervals are "left-closed/right-closed". This is easier than accounting for "open" intervals when dealing with integers. Argument `include.lowest` is ignored.
#'
#' @param others Integer or `NULL` (default), or `NA`: Value to assign to cells that do not fall into the set intervals. Cells with `NA` values are not reclassified. Setting `others` equal to `NULL` or `NA` replaces all other values with `NA`. The value will be coerced to an integer value.
#' 
#' @returns A `GRaster`. The raster will be a categorical `GRaster` if the original values were continuous (i.e., a single- or double-precision raster), or of type "integer" if the input was an integer. See `vignette("GRasters", package = "fasterRaster")`.
#'
#' @seealso [terra::classify()], [subst()]
#'
#' @example man/examples/ex_classify.r
#'
#' @aliases classify
#' @rdname classify
#' @exportMethod classify
methods::setMethod(
	f = "classify",
	signature = c(x = "GRaster"),
	function(x, rcl, include.lowest = FALSE, right = TRUE, others = NULL) {

	if (is.na(right) & !all(is.int(x))) stop("You can only use `right = NA` for GRasters of type `integer`.")

	.locationRestore(x)
	.region(x)
	
	nLayers <- nlyr(x)

	if (!is.null(others) && !is.na(others) && !is.integer(others) && omnibus::compareFloat(others, as.integer(others), "!=")) {
		warning("Converting value of `others` to an integer.")
		others <- as.integer(others)
	}

	if (inherits(rcl, c("matrix", "data.frame", "data.table")) && ncol(rcl) == 2L) {
		if (!all(is.int(x))) stop("You can only use a 2-column matrix for `rcl` if all layers of a GRaster are of type `integer`.")
		right <- NA
	}

	### classify
	srcs <- .makeSourceName("reclass_r_mapcalc", "raster", nLayers)
	levs <- list()
	for (i in seq_len(nLayers)) {

		makeLevs <- !is.na(right) # make levels table?

		# rcl is a scalar or vector
		if (!inherits(rcl, c("matrix", "data.frame", "data.table"))) {

			# rcl is a scalar			
			if (length(rcl) == 1L) {

				if (rcl <= 1) stop("Argument `rcl` must be a value >= 2, or a vector, matrix, or data.frame.")
				mm <- minmax(x[[i]])
				fromTos <- seq(mm["min", 1L], mm["max", 1L], length.out = rcl + 1L)
				
				fromTos <- data.table::data.table(
					from = fromTos[1L:(length(fromTos) - 1L)],
					to = fromTos[2L:length(fromTos)],
					new = 1L:(length(fromTos) - 1L)
				)

			# rcl is a vector
			} else {

				rcl <- sort(rcl)
				len <- length(rcl)
				fromTos <- data.table::data.table(from = rcl[1L:(len - 1L)], to = rcl[2L:len], value = 1L:(len - 1L))
			
			}

		# rcl is a 2-column matrix
		} else if (ncol(rcl) == 2L) {
			
			if (!inherits(rcl, "data.table")) rcl <- data.table::as.data.table(rcl)
			fromTos <- data.table::data.table(from = rcl[[1L]], to = rcl[[1L]], value = rcl[[2L]])

			right <- NA
			makeLevs <- FALSE
		
		# rcl is a 2-column matrix
		} else if (ncol(rcl) == 3L) {

			if (!inherits(rcl, "data.table")) rcl <- data.table::as.data.table(rcl)
			fromTos <- rcl
			if (!is.integer(fromTos[[3L]]) && !all(omnibus::compareFloat(fromTos[[3L]], as.integer(fromTos[[3L]]), "=="))) makeLevs <- FALSE

		} else {
			stop("Argument `rcl` must be a single integer, a numeric vector, or a 2- or 3-column matrix or data.frame.")
		}

		### create r.mapcalc formula
		ex <- paste0(srcs[i], " = ")

		labels <- rep(NA_character_, nrow(fromTos))
		for (j in 1L:nrow(fromTos)) {

			if (is.na(right)) {
				lower <- ">="
				upper <- "<="
			} else if (right & include.lowest) {
				if (j == which.min(fromTos[[1L]])) {
					lower <- ">="
				} else {
					lower <- ">"
				}
				upper <- "<="
			} else if (right & !include.lowest) {
				lower <- ">"
				upper <- "<="
			} else if (!right & include.lowest) {
				lower <- ">="
				if (j == which.max(fromTos[[2L]])) {
					upper <- "<="
				} else {
					upper <- "<"
				}
			} else if (!right & !include.lowest) {
				lower <- ">="
				upper <- "<"
			}

			ex <- paste0(ex, "if(", sources(x)[i], lower, fromTos[j, 1L], " & ", sources(x)[i], upper, fromTos[j, 2L], ", ", fromTos[j, 3L], ", ")

			# labels for levels table
			if (!is.na(right)) {

				leftBrack <- if (lower == ">") { "(" } else if (lower == ">=") { "["}
				rightBrack <- if (upper == "<") { ")" } else if (upper == "<=") { "]"}
				labels[j] <- paste0(leftBrack, fromTos[j, 1L], "-", fromTos[j, 2L], rightBrack)

			}

		}

		if (is.null(others) || is.na(others)) {
			ex <- paste0(ex, "null()")
		} else {
			ex <- paste0(ex, others)
		}
		ex <- paste0(ex, paste(rep(")", j), collapse = ""))
		
		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))
		
		# create levels table
		if (!is.na(right)) {

			levs[[i]] <- data.table::data.table(ID = as.integer(fromTos[[3L]]), DUMMY = labels)

			if (!is.null(others) && !is.na(others)) levs[[i]] <- rbind(levs[[i]], data.table::data.table(ID = as.integer(others), DUMMY = 'others'))

			names(levs[[i]])[2L] <- names(x)[i]

		} else {
			levs <- NULL
		}

	}

	out <- makeGRaster(srcs, names = names(x), levels = levs, ac = 1L)
	out
	
	} # EOF
)

#' Test if addons directory exists and if an addon is installed
#'
#' @description These functions handle **GRASS** addons, which are optional tools that can be installed. Most functions in **fasterRaster** rely on "base" **GRASS** tools (not addons), but a few do.
#'
#' * `addons()`: Either lists all installed addons or verifies if one or more specific addons are installed.
#' * `installAddon()`: Installs a **GRASS** addon. An addon typically only needs installed once. You can install an addon, quit and restart **R**, attach **fasterRaster**, and any installed addons can be used without using this function again. 
#' * `removeAddon()`: Delete an installed addon from your system.
#'
#' @param x Either `NULL` or a character specifying the name of a **GRASS** addons tool. If `NULL`, a vector of installed addons is returned. If a character vector is provided, a logical vector is returned, one per value in `x` that indicates if the respective addon is installed.
#'
#' @param check Logical: If `TRUE`, check to see if the addon is available (`installAddon()`)or if it is installed (`removeAddon()`).
#'
#' @returns `addons()`: Logical. The other functions invisibly return a logical value indicating if the operation succeeded or not.
#'
#' @example man/examples/ex_addons.r
#'
#' @aliases addons
#' @rdname addons
#' @export
addons <- function(x = NULL) {

	out <- rgrass::execGRASS("g.extension", flags = "a", intern = TRUE)
	if (is.null(x)) {
		out <- sort(out)
	} else if (!is.null(x)) {
		out <- x %in% out
		names(out) <- x
	}
	out
	
}

#' @aliases installAddon
#' @rdname addons
#' @export
installAddon <- function(x, check = TRUE) {

	if (check) {

		# is it available?
		avails <- rgrass::execGRASS("g.extension", flags = c("l", .quiet()), intern = TRUE)
		if (!(x %in% avails)) {

			warning("The addon is not available on the official GRASS addon repository.")
			return(invisible(FALSE))

		}
	}

	# remove addon before re-installing
	if (addons(x)) removeAddon(x, check = FALSE)

	rgrass::execGRASS("g.extension", operation = "add", extension = x, flags = .quiet())
	success <- addons(x)
	if (!success) warning("Addon could not be installed.")
	invisible(success)

}

#' @aliases removeAddon
#' @rdname addons
#' @export
removeAddon <- function(x, check = TRUE) {

	if (check) {
		avails <- rgrass::execGRASS("g.extension", flags = c("a", .quiet()))
		if (!(x %in% avails)) warning("The addon is not installed.")
		return(invisible(FALSE))
	}

	rgrass::execGRASS("g.extension", operation = "removeadd", extension = x, flags = c(.quiet(), "f"))
	invisible(TRUE)

}

#' Test to see if a given addon is installed, and if not, installs it.
#'
#' @aliases .addons
#' @rdname addons
#' @keywords internal
.addons <- function(x) {

	installed <- addons(x)
	if (!installed) {
		installAddon(x, check = FALSE)
	}

}



# # Tests if addons folder is installed
# addons <- function(xx = NULL, fail = TRUE, verbose = TRUE) {

# 	out <- TRUE

# 	ao <- faster("addonsDir")
# 	if (is.null(ao) || !file.exists(ao)) {
		
# 		msg <- paste0("The `addons` folder is incorrect. See `faster()` and `vignette(", dQuote("addons", q = FALSE), ", package = ", dQuote("fasterRaster", q = FALSE), ").")

# 		if (fail) {
# 			stop(msg)
# 		} else if (!fail & verbose) {
# 			warning(msg)
# 		}

# 		out <- FALSE

# 	}
	
# 	if (!is.null(x)) {

# 		extensions <- list.files(paste0(faster("addonsDir"), "/bin"))
# 		exts <- .fileExt(extensions)
# 		extensions <- substr(extensions, 1L, nchar(extensions) - nchar(exts) - 1L)

# 		if (!(x %in% extensions)) {

# 			msg <- paste0("The addon extension `", x, "` cannot be found. See `vignette(", dQuote("addons", q = FALSE), ", package = ", dQuote("fasterRaster", q = FALSE), ").")
# 			if (fail) {
# 				stop(msg)
# 			} else if (!fail & verbose) {
# 				warning(msg)
# 			}
		
# 			out <- FALSE

# 		}
# 	}
	
# 	if (verbose) {
# 		if (is.null(x)) {
# 			omnibus::say("Addons directory exists.")
# 		} else {
# 			if (out) {
# 				omnibus::say("Addon `", x, "` is installed.")
# 			} else {
# 				omnibus::say("Addon `", x, "` cannot be found.")
# 			}
# 		}
# 	}
# 	invisible(out)

# }

if (grassStarted()) {

# Setup
library(terra)

# Elevation and land cover raster
madElev <- fastData("madElev") # integer raster
madCover <- fastData("madCover") # categorical raster

# Convert to GRasters
elev <- fast(madElev)
cover <- fast(madCover)

# Subset elevation raster to just a few values to make example faster:
elevSubset <- elev[elev <= 3]
segregate(elevSubset)
segregate(elevSubset, keep = TRUE, other = -1)

# Segregate the factor raster
segregate(cover)

classes <- c("Grassland with mosaic forest", "Mosaic cropland/vegetation")
seg <- segregate(cover, classes = classes)
plot(seg)

}

#' Identify clusters of points
#'
#' @description `clusterPoints()` partitions points in a "points" `GVector` into clusters.
#'
#' @param x A "points" `GVector`.
#'
#' @param method Character: Method used to identify clusters. Explanations of methods are provided in the help page for the **GRASS** tool `v.cluster`, available using `grassHelp("v.cluster")`.
#' * `"DBSCAN"` (default): Density-Based Spatial Clustering of Applications with Noise. 
#' * `"DBSCAN2"`: A modification of DBSCAN.
#' * `"density"`: Cluster points by relative density.
#' * `"OPTICS"`: Ordering Points to Identify the Clustering Structure
#' * `"OPTICS2"`: A modification of OPTICS.
#'
#' Case is ignored, but partial matching is not used.
#'
#' @param minIn Integer, numeric integer, or `NULL` (default): Minimum number of points in a cluster. If `NULL`, then `minIn` is set to 3 for a 2-dimensional vector and 4 for a 3-dimensional vector.
#'
#' @param maxDist Numeric or `NULL` (default): Maximum distance between neighboring points in a cluster for DBSCAN, DBSCAN2, and OPTICS. If `NULL`, the maximum distance will be set to the 99th quantile of observed pairwise distances between points.
#'
#' @returns A vector of integers indicating the cluster to which each point belongs.
#'
#' @seealso **GRASS** manual page for tool `v.cluster` (see `grassHelp("v.cluster")`)
#'
#' @aliases clusterPoints
#' @rdname clusterPoints
#' @exportMethod clusterPoints
methods::setMethod(
	f = "clusterPoints",
	signature = c(x = "GVector"),
	function(x, method = "DBSCAN", minIn = NULL, maxDist = NULL) {
	
	if (geomtype(x) != "points") stop("Only points GVectors can be clustered.")
	
	if (!is.null(maxDist)) {
		if (is.infinite(maxDist)) stop("Argument `maxDist` cannot be infinite.")
	}

	method <- tolower(method)

	.locationRestore(x)

	src <- .makeSourceName("v_cluster", "vector")
	args <- list(
		cmd = "v.cluster",
		input = sources(x),
		output = src,
		method = method,
		layer = "2",
		flags = c(.quiet(), "overwrite")
	)
	if (!is.null(minIn)) args$min <- minIn
	if (!is.null(maxDist)) args$distance <- maxDist
	do.call(rgrass::execGRASS, args = args)

	.vCats(src, layer = 2)

	} # EOF
)

if (grassStarted()) {

### Setup for all examples

library(sf)
library(terra)

# Climate raster, elevation raster, rivers vector
madElev <- fastData("madElev")
madRivers <- fastData("madRivers")
madChelsa <- fastData("madChelsa")

# Convert objects into fasterRaster formats
chelsa <- fast(madChelsa)
elev <- fast(madElev)
rivers <- fast(madRivers)

### Project raster without resampling
elevWGS84 <- project(elev, crs(chelsa))
elevWGS84

### Project raster and resample to resolution of another raster
elevWGS84Resamp <- project(elev, chelsa)
elevWGS84Resamp

res(elevWGS84)
res(elevWGS84Resamp)
res(chelsa)

### Project vector
riversWGS84 <- project(rivers, chelsa)
riversWGS84
cat(crs(rivers)) # using "cat()" to make it look nice
cat(crs(riversWGS84))

}

#' Increase/decrease the size of a vector or around non-NA cells of a raster
#'
#' Buffers can be constructed for `GRaster`s or `GVector`s. For rasters, the `buffer()` function creates a buffer around non-`NA` cells. The output will be a raster. For vectors, the `buffer()` and `st_buffer()` functions create a vector polygon larger or smaller than the focal vector.
#'
#' Note that in some cases, topologically incorrect vectors can be created when buffering. This can arise when buffers intersect to create intersections that technically belong to two or more geometries. This issue can be resolved by dissolving borders between buffered geometries using `dissolve = TRUE`, but as of now, there is no fix if you do not want to dissolve geometries. A workaround would be to create a different `GVector` for each geometry, and then buffer each individually :(.
#'
#' @param x A `GRaster` or `GVector`.
#' 
#' @param width Numeric: For rasters -- Maximum distance cells must be from focal cells to be within the buffer. For rasters, if the buffering unit is `"cells`", then to get `n` cell widths, use `n + epsilon`, where `epsilon` is a small number (e.g., 0.001). The larger the buffer, this smaller this must be to ensure just `n` cells are included.
#' 
#' For vectors, distance from the object to place the buffer. Negative values create "inside" buffers. Units are in the same units as the current coordinate reference system (e.g., degrees for WGS84 or NAD83, often meters for projected systems).
#' 
#' @param dist Vectors -- Same as `width`.
#' 
#' @param background Numeric: Rasters -- Value to assign to cells that are not `NA` and not part of the buffer (default is 0).
#' 
#' @param unit Character: Rasters -- Indicates the units of \code{width}. Can be one of:
#'
#' * `"cells"`: Units are numbers of cells.
#' * `"meters"` (default), `"metres"`, or `"m"`
#' * `"kilometers"` or `"km"`
#' * `"feet"` or `"ft"`
#' * `"miles"` or `"mi"`
#' * `"nautical miles"` or `"nmi"`
#'
#' Partial matching is used and case is ignored.
#'
#' @param method Character: Rasters -- Only used if `units` is `"cells"`. Indicates the manner in which distances are calculated for adding of cells:
#' * `"Euclidean"`: Euclidean distance (default)
#' * `"Manhattan"`: "taxi-cab" distance
#' * `"maximum"`: Maximum of the north-south and east-west distances between points.
#'
#' Partial matching is used and case is ignored.
#'
#' @param lowMemory Logical: Rasters -- Only used if buffering a raster and `units` is not `"meters"`. If `FALSE` (default) use faster, memory-intensive procedure. If `TRUE` then use the slower, low-memory version. To help decide which to use, consider using the low-memory version on a system with 1 GB of RAM for a raster larger than about 32000 x 32000 cells, or for a system with  with 8 GB of RAM a raster larger than about 90000 x 90000 cells.
#'
#' @param capstyle,endCapStyle Character: Vectors -- Style for ending the "cap" of buffers around lines. Valid options include `"rounded"`, `"square"`, and "`flat`".
#'
#' @param dissolve Logical (`GVector`s): If `TRUE` (default), dissolve all buffers after creation. If `FALSE`, construct a buffer for each geometry. Note that overlapping buffers can cause this function to fail because it creates a topologically ambiguous polygon. Thus, using `dissolve = TRUE` is recommended.
#'
#' @seealso [terra::buffer()], [sf::st_buffer()]
#'
#' @returns A `GRaster` or a `GVector`.
#'
#' @example man/examples/ex_buffer.R
#'
#' @aliases buffer
#' @rdname buffer
#' @exportMethod buffer
methods::setMethod(
	"buffer",
	signature(x = "GRaster"),
	function(
		x,
		width,
		unit = "meters",
		method = "Euclidean",
		background = 0,
		lowMemory = FALSE
	) {

	.locationRestore(x)
	.region(x)

	units <- c("cells", "m", "meters", "metres", "km", "kilometers", "ft", "feet", "mi", "miles", "nmi", "nautical miles")
	unit <- omnibus::pmatchSafe(unit, units, useFirst = TRUE, nmax = 1L)
	if (unit != "cells") unit <- omnibus::expandUnits(unit)
	if (unit == "nautical miles") unit <- "nautmiles"

	if (nlyr(x) > 1L) out <- list()

	# for each layer
	for (i in 1L:nlyr(x)) {

		srcBuffer <- .makeSourceName("r_grow_r_buffer", "raster")

		### buffering by cells
		if (unit == "cells") {
		
			methods <- c("euclidean", "manhattan", "maximum")
			method <- tolower(method)
			method <- omnibus::pmatchSafe(method, method, nmax = 1L)

			args <- list(
				cmd = "r.grow",
				input = sources(x)[i],
				output = srcBuffer,
				radius = width,
				metric = method,
				old = 1,
				new = 1,
				flags = c(.quiet(), "overwrite")
			)
		
		### buffering by distance
		} else {
		
			args <- list(
				cmd = ifelse(lowMemory, "r.buffer.lowmem", "r.buffer"),
				input = sources(x)[i],
				output = srcBuffer,
				distances = width,
				units = unit,
				flags = c(.quiet(), "overwrite")
			)
			
		}

		### buffer
		do.call(rgrass::execGRASS, args)

		### reclass
		src <- .makeSourceName("buffer", "raster")
		ex <- if (!is.na(background)) {
			paste0(src, " = if(isnull(", srcBuffer, "), ", background, ", if(", srcBuffer, " == 2, 1, 1))")
		} else {
			paste0(src, " = if(", srcBuffer, " == 2, 1, 1)")
		}
		rgrass::execGRASS("r.mapcalc", expression = ex, flags=c(.quiet(), "overwrite"))
		on.exit(.rm(srcBuffer, type = "raster", warn = FALSE), add = TRUE)
		
		if (nlyr(x) > 1L) {
			group[[i]] <- makeGRaster(src, names(x[[i]]))
		} else {
			out <- makeGRaster(src, names(x[[i]]))
		}
		
	} # next layer
	
	if (nlyr(x) > 1L) out <- c(out)
	out

	} # EOF
)

#' @aliases buffer
#' @rdname buffer
#' @exportMethod buffer
methods::setMethod(
	"buffer",
	signature(x = "GVector"),
	function(
		x,
		width,
		capstyle = "round",
		dissolve = TRUE
	) {

	# .message(msg = "buffer", message = "As of GRASS 8.3, terra's buffer() function is much faster, even for very large vectors.")
	
	.locationRestore(x)

	### buffer
	src <- .makeSourceName("v_buffer", "vector")
	args <- list(
		cmd = "v.buffer",
		input = sources(x),
		output = src,
		distance = width,
		flags = c(.quiet(), "overwrite")
	)

	args$type <- geomtype(x, grass = TRUE)
	if (!dissolve) args$flags <- c(args$flags, "t")

	capstyle <- tolower(capstyle)
	capstyle <- omnibus::pmatchSafe(capstyle, c("round", "square", "flat"))
	if (capstyle == "square") {
		args$flags <- c(args$flags, "s")
	} else if (capstyle == "flat") {
		args$flags <- c(args$flags, "c")
	}

	do.call(rgrass::execGRASS, args)

	# # # # dissolve by category
	# # # if (dissolve) {

	# # # 	# dissolve
	# # # 	srcBuff <- src
	# # # 	src <- .makeSourceName("v_dissolve", "vector")
	# # # 	# catCol <- .vNames(srcBuff)
	# # # 	args <- list(
	# # # 		cmd = "v.dissolve",
	# # # 		input = srcBuff,
	# # # 		output = src,
	# # # 		# column = "cat",
	# # # 		flags = c(.quiet(), "overwrite")
	# # # 	)
		
	# # # 	if (.vHasDatabase(srcBuff)) args$column <- .vNames(srcBuff)[1L]
	# # # 	do.call(rgrass::execGRASS, args = args)

	# # # }

	makeGVector(src)
	
	} # EOF
)

#' @aliases st_buffer
#' @rdname buffer
#' @exportMethod st_buffer
methods::setMethod(
	"st_buffer",
	signature(x = "GVector"),
	function(x, dist, endCapStyle = "round", dissolve = FALSE) {
		buffer(x, width = dist, capstyle = endCapStyle, dissolve = dissolve)
	} # EOF
)

st_buffer <- function(x) UseMethod("st_buffer", x)

if (grassStarted()) {

# Setup
library(terra)

# Elevation raster:
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster:
elev <- fast(madElev)

### Fill NAs:
bilinear <- fillNAs(elev)
bicubic <- fillNAs(elev, method = "bicubic")
rst <- fillNAs(elev, method = "rst")

maps <- c(elev, bilinear, bicubic, rst)
names(maps) <- c("original", "bilinear", "bicubic", "RST")
plot(maps)

### Constrain interpolated values to > 0
constrained <- fillNAs(elev, min = 0)

# Compare unconstrained and constrained:
minmax(bilinear)
minmax(constrained)

### Interpolate to only first 10 cells away from non-NA cells:
restrained <- fillNAs(elev, cells = 10)

maps <- c(elev, restrained)
names(maps) <- c("Original", "within 10 cells")
plot(maps)

}

#' Number of categories in a categorical raster
#'
#' @description This function reports the number of categories (levels) in a categorical `GRaster`.
#'
#' @param x A `GRaster`.
#'
#' @return A named, numeric vector of integers. The values represent the number of categories (rows) that appear in the raster's levels table.
#'
#' @seealso [levels()], [terra::levels()], [droplevels()], `vignette("GRasters", package = "fasterRaster")`
#'
#' @example man/examples/ex_GRaster_categorical.r
#'
#' @aliases nlevels
#' @rdname nlevels
#' @exportMethod nlevels
setMethod(
    f = "nlevels",
    signature = "GRaster",
    definition = function(x) {

	out <- sapply(x@levels, nrow)
	names(out) <- names(x)
	out

	} # EOF
)

#' Count number of levels from a data.frame/table, list, or SpatRaster
#'
#' @description Counts number of levels in a character string (specifically, the empty string `""`), a `data.frame`, `data.table`, or list of `data.frame`s or `data.table`s or empty strings.
#'
#' @param x A `GRaster`, `SpatRaster`, *or* a `data.frame`, `data.table`, an empty string, or a list thereof.
#'
#' @keywords internal
.nlevels <- function(x) {

	if (inherits(x, c("SpatRaster", "GRaster"))) x <- levels(x)
	
	if (!is.list(x)) x <- list(x)
	n <- rep(NA_integer_, length(list))

	for (i in seq_along(x)) {
	
		if (inherits(x[[i]], "character")) {
			if (is.null(x[[i]]) || x[[i]] == "") {
				n[i] <- 0L
			} else {
				stop("Argument x must be a data frame, data table, an empty string or a list of any of these.")
			}
		} else if (inherits(x[[i]], c("data.frame", "data.table"))) {
			n[i] <- nrow(x[[i]])
		}
	}
	n

}

if (grassStarted()) {

# Setup
library(terra)

# Elevation raster, rivers vector
madElev <- fastData("madElev")

# Convert to a GRaster
elev <- fast(madElev)

# Cell coordinates
init(elev, "x")
init(elev, "y")

# Cell row or column
init(elev, "row")
init(elev, "col")

# Chess
elevAgg <- aggregate(elev, 32) # make cells bigger so we can see them

chessOdd <- init(elevAgg, "chess")
chessEven <- init(elevAgg, "chess", odd = FALSE)

chess <- c(chessOdd, chessEven)
names(chess) <- c("odd", "even")
plot(chess)

# Chess with user-defined values
elevAgg <- aggregate(elev, 32) # make cells bigger so we can see

chessOdd13 <- init(elevAgg, "chess", vals = c(0, 13))
chessEven13 <- init(elevAgg, "chess", odd = FALSE, vals = c(0, 13))

chess13 <- c(chessOdd13, chessEven13)
names(chess13) <- c("odd", "even")
plot(chess13)

# Regular
elevAgg <- aggregate(elev, 32) # make cells bigger so we can see

regOdd <- init(elevAgg, "regular")
regEven <- init(elevAgg, "regular", odd = FALSE)

reg <- c(regOdd, regEven)
names(reg) <- c("odd", "even")
plot(reg)

}

#' Voronoi tessellation
#'
#' @description This function creates a Voronoi tessellation from a set of spatial points or polygons.
#'
#' @param x A `GVector` "points" object.
#' @param buffer Numeric: By default, this function creates a vector that has an extent exactly the same as the input data. However, the apparent extent can be changed by setting this value to a value different from 0. Negative values reduce the size of the extent, and positive extend it.  Units are in map units.
#'
#' @returns A `GVector`.
#'
#' @seealso [terra::voronoi()], [sf::st_voronoi()], tool `v.voronoi` in **GRASS**
#'
#' @example man/examples/ex_delaunay_voronoi.r
#'
#' @aliases voronoi
#' @rdname voronoi
#' @exportMethod voronoi
methods::setMethod(
	f = "voronoi",
	signature = c(x = "GVector"),
	definition = function(x, buffer = 0) {

	if (!(geomtype(x) %in% c("points", "polygons"))) stop("The vector must represent spatial points or polygons.")
	
	.locationRestore(x)

	# do not expand region beyond x
	.regionExt(x, respect = "dimensions")
	if (buffer != 0) {
		
		# set region extent to buffered vector
		extent <- .regionExt()
		w <- extent["xmin"]
		e <- extent["xmax"]
		s <- extent["ymin"]
		n <- extent["ymax"]
		
		w <- w - buffer
		e <- e + buffer
		s <- s - buffer
		n <- n + buffer
		
		w <- as.character(w)
		e <- as.character(e)
		s <- as.character(s)
		n <- as.character(n)
		
		rgrass::execGRASS("g.region", n = n, s = s, e = e, w = w, flags = c("o", .quiet(), "overwrite"))
		
	}

	src <- .makeSourceName("v_voronoi", "vect")
	args <- list(
		cmd = "v.voronoi",
		input = sources(x),
		output = src,
		flags = c(.quiet(), "overwrite")
	)

	if (geomtype(x) == "polygons") args$flags <- c(args$flags, "a")

	do.call(rgrass::execGRASS, args = args)
	makeGVector(src)
	
	} # EOF
)

if (grassStarted()) {

# Example data
madElev <- fastData("madElev") # elevation raster
madLANDSAT <- fastData("madLANDSAT") # multi-layer raster
madRivers <- fastData("madRivers") # lines vector

# Convert SpatRaster to GRaster and SpatVector to GVector
elev <- fast(madElev)
rivers <- fast(madRivers)
landsat <- fast(madLANDSAT)

# Plot:
plot(elev)
plot(rivers, add = TRUE)

# Histograms:
hist(elev)
hist(landsat)

# Plot surface reflectance in RGB:
plotRGB(landsat, 3, 2, 1) # "natural" color
plotRGB(landsat, 4, 1, 2, stretch = "lin") # emphasize near-infrared (vegetation)

# Make composite map from RGB layers and plot in grayscale:
comp <- compositeRGB(r = landsat[[3]], g = landsat[[2]], b = landsat[[1]])
grays <- paste0("gray", 0:100)
plot(comp, col = grays)

}

if (grassStarted()) {

# Setup
library(terra)

# Elevation raster
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster:
elev <- fast(madElev)

# Create spatially exclusive tiles:
exclusive <- tiles(elev, n = 2, verbose = TRUE)

startpar <- par(mfrow = c(2, 3))
plot(elev, main = "Original")

for (i in seq_along(exclusive)) {
	plot(exclusive[[i]], ext = elev, main = paste("Tile", i))
}
par(startpar)

# Create tiles that overlap:
overlaps <- tiles(elev, n = 2, overlap = 200, verbose = TRUE)

startpar <- par(mfrow = c(2, 3))
plot(elev, main = "Original")

for (i in seq_along(overlaps)) {
	plot(overlaps[[i]], ext = elev, main = paste("Tile", i))
}
par(startpar)

}

if (grassStarted()) {

# Setup
library(terra)

# Elevation raster
madElev <- fastData("madElev")

# Convert SpatRaster to a GRaster:
elev <- fast(madElev)

# Trim NA rows/columns:
trimmedElev <- trim(elev)
dim(elev)
dim(trimmedElev)

# Trim a "stack" of rasters. We will artificially add NA rows and columns to
# one raster to demonstrate how the trim() function removes only rows/columns
# that are NA for *all* rasters.
elevNAs <- elev

fun <- " = if(col() > ncols() - 200, null(), madElev)"
elevNAs <- app(elevNAs, fun)

# Notice raster is "narrower" because we added NA columns
plot(elevNAs)

elevs <- c(elev, elevNAs)
trimmedElevs <- trim(elevs)
trimmedElevNAs <- trim(elevNAs)

dim(elevs)
dim(trimmedElevNAs)
dim(trimmedElevs)

}

if (grassStarted()) {

# Setup
library(terra)

# Example data: Elevation and land cover
madElev <- fastData("madElev")
madForest2000 <- fastData("madForest2000")
madCover <- fastData("madCover")

# Convert to GRasters:
elev <- fast(madElev)
forest2000 <- fast(madForest2000)
cover <- fast(madCover)

# Rename
names(elev) <- "elev"
names(forest2000) <- "forest"

# Geometric statistics for an integer raster zoned by elevation:
fun <-
  "= if (elev <400 & forest == 1, 0, if (elev >=400 & forest == 1, 1, null()))"
forestByElev <- app(c(elev, forest2000), fun = fun)
plot(forestByElev, main = "forest < 400 m & >= 400 m")
zonalGeog(forestByElev)

# Geometric statistics for a categorical raster:
zonalGeog(cover)

}

#' @title Create a GRaster or GVector
#'
#' @description `fast()` creates a `GRaster` or `GVector` from 1) a file; 2) from a `SpatRaster`, `SpatVector`, or `sf` vector; or 3) from a numeric vector, `matrix`, `data.frame`, or `data.table`. Behind the scenes, this function will also create a connection to **GRASS** if none has yet been made yet.
#'
#' **GRASS** supports loading from disk a variety of raster formats (see the **GRASS** manual page for `r.in.gdal` (see `grassHelp("r.in.gdal")`) and vector formats `v.in.ogr` (see grassHelp("v.in.ogr")`), though not all of them will work with this function.
#'
#' Note that `GVectors` may fail to be created if they contain issues that do not coincide with the topological data model used by **GRASS**. The most common of these is overlapping polygons. See *Details* on how to fix these kinds of issues.
#'
#' Note also that **GRASS** (and thus, **fasterRaster**) is *not* very fast when loading vectors. So, if the vector is large and you only want a portion of it, consider using the `extent` argument to load the spatial subset you need.
#'
#' @param x Any one of:
#' * A `SpatRaster` raster. Rasters can have one or more layers.
#' * A `SpatVector` or `sf` spatial vector. See especially arguments `correct`, `area`, `snap`, `steps`, and `verbose`.
#' * A character string or a vector of strings with the path(s) and filename(s) of one or more rasters or one vector to be loaded directly into **GRASS**. The function will attempt to ascertain the type of object from the file extension (raster or vector), but it can help to indicate which it is using the `rastOrVect` argument if it is unclear. For rasters, see especially argument `levels`. For vectors, see especially arguments `correct`, `resolve`, `area`, `snap`, `steps`, and `verbose`.
#' * A vector with an even number of numeric values representing longitude/latitude pairs. See arguments `geom`, `keepgeom`, and `crs`.
#' * A `data.frame`, `data.table`, or `matrix`: Create a `points` `GVector`. Two of the columns must represent longitude and latitude. See arguments `geom`, `keepgeom`, and `crs`.
#' * Missing: Creates a generic `GRaster` or `GVector`. You must specify `rastOrVect`; for example, `fast(rastOrVect = "raster")`. Also see argument `crs`.
#'
#' @param rastOrVect Either `NULL` (default), or `"raster"` or `"vector"`: If `x` is a filename, then the function will try to ascertain whether it represents a raster or a vector, but sometimes this will fail. In that case, it can help to specify if the file holds a raster or vector. Partial matching is used.
#'
#' @param levels (`GRaster`s only): Any of:
#' * Logical: If `TRUE` (default) and at least one layer of a raster is of type `integer`, search for a "levels" file, load it, and attach levels. A levels file will have the same name as the raster file, but end with any of "rdata", "rdat", "rda", "rds", "csv", or "tab" (case will generally not matter). If such a file is not found, no levels will be assigned. The levels file must contain either a `data.frame`, `data.table`, or `list` of `data.frame`s or `data.table`s, or `NULL`.
#' * A `data.frame`, `data.table`, or list of `data.frame`s or `data.table`s with categories for categorical rasters. The first column of a table corresponds to raster values and must be of type `integer`. A subsequent column corresponds to category labels. By default, the second column is assumed to represent labels, but this can be changed with \code{\link[fasterRaster]{activeCat<-}}. Level tables can also be `NULL` (e.g., `data.fame(NULL)`). You can also assign levels after loading a raster using \code{\link[fasterRaster]{levels<-}}.
#' * `NULL`: Do not attach a levels table.
#'#'
#' @param geom Character or integer vector: If `x` is a `data.frame`, `data.table`, or `matrix`, this specifies which columns of `x` represent longitude and latitude. Columns can be given by name (a character vector) or index (a numeric or integer vector). The default is to use the first two columns of `x`.
#'
#' @param crs String: Coordinate reference system (CRS) WKT2 string. This argument is used for creating a `GVector` from a `numeric` vector or a `data.frame` or similar, or from `fast(rastOrVect = "vector")` or `fast(rastOrVect = "raster")`. By default, the function will use the value of [crs()] (no arguments), which is the CRS of the current **GRASS** "project/location" (see `vignette("projects_mapsets", package = "fasterRaster")`).
#'
#' @param keepgeom Logical: If `x` is a set of `numeric` coordinates, or a `data.frame` or similar, then they can be coerced into a `points` `GVector`. If `keepgeom` is `TRUE`, then the coordinates will be included in the data table of the `GVector`. The default is `FALSE`.
#'
#' @param extent (`GVector`s only): Either a `NULL` (default), or a `GVector`, a `SpatVector`, a `SpatExtent` object, an `sf` vector, a `bbox` object, or a numeric vector of 4 values providing a bounding box. If provided, only vector features within this bounding box are imported. If `extent` is a numeric vector, the values *must* be in the order west, east, south, north. If `NULL`, the entire vector is imported.
#'
#' @param correct Logical (`GVector`s only): Correct topological issues. See *Details* for more details! By default, this is `TRUE`.
#'
#' @param snap `GVector`s only: Numeric or `NULL` (default). The value of `snap` indicates how close vertices need to be for them to be shifted to to the same location. Units of `snap` are map units (usually meters), or degrees for unprojected CRSs. For lines and polygons vectors, a value of `NULL` will invoke an iterative procedure to find an optimal, smallest value of `snap`. To turn snapping off, set `snap = 0`. See *Details* for more details!
#'
#' @param area Polygon `GVector`s only: Either a positive numeric value or `NULL` (default). Remove polygons with an area smaller than this value. Units of `area` are in square meters (regardless of the CRS). If `NULL`, then an iterative procedure is used to identify a value of `area` that results in a topologically correct polygon vector. For point and lines vectors, this argument is ignored. To turn area removal off, set `area = 0`. See *Details* for more details! 
#'
#' @param steps `GVector`s only: A positive integer > 1 (default is 10). When using automatic vector correction (i.e., either `snap = NULL` and/or `area = NULL`), this is the number of values of `snap` and/or `area` to try to generate a correct topology, including no snapping or polygon removal (i.e, `snap = 0` and `area = 0`).
#'
#' @param dropTable `GVector`s only: Logical. If `TRUE`, then drop the data table associated with a vector. By default, this is `FALSE`. See *Details* for more details!
#'
#' @param resolve `GVector`s only: Character. If a `GVector` would be topologically invalid after a first attempt at creating it, then this method will be used to resolve the issue and create a valid `GVector`. Partial matching is used.
#' * `"disaggregate"`: Coerce each area of overlap between polygons into its own geometry. The output will not have a data table associated with it.
#' * `"aggregate"`: Coerce all geometries into a "multipart" geometry so it acts like a single geometry. The output will not have a data table associated with it.
#' * `NA` (default): Do neither of the above and if either `snap` or `area` is `NULL`, keep trying to create the `GVector`. Upon success, the `GVector` will retain any data table associated with it unless `dropTable` is `FALSE`.
#'
#' @param verbose `GVector`s only: Logical. Displays progress when using automatic topology correction.
#'
#' @param ... Other arguments::
#' * `table` (`GVector`s--useful mainly to developers, not most users): A `data.frame` or `data.table` with one row per geometry in a `GVector`. Serves as an attribute table.
#' * `xVect` (`GVector`s--useful mainly to developers, not most users): The `SpatVector` that corresponds to the file named by `x`.
#'
#' @details **GRASS** uses a "topological" model for vectors. Topological issues generally arise only with polygon vectors, not point or line vectors. Sometimes, polygons created in other software are topologically incorrect--the borders of adjacent polygons may cross one another, or there may be small gaps between them. These errors can be corrected by slightly shifting vertices and/or removing small polygons that result from intersections of larger ones that border one another. A topological system also recognizes that boundaries to adjacent polygons are shared by the areas, so should not be ascribed attributes that belong to both areas (e.g., the shared border between two countries "belongs" to both countries).
#'
#' By default, `fast()` will try to correct topological errors in vectors. There are three levels of correction, and they are not necessarily mutually exclusive:
#' 1. **Automatic correction**: By default, `fast()` will apply automatic topology correction. You can turn this off using the `correct = FALSE` argument, though in most cases this is not recommended.
#' 2. ***Manual* snapping and/or area removal**: In addition to correction from step 1, you can cause vertices of polygons close to one another to be "snapped" to same place and/or polygons that are smaller than some threshold to be removed. Problems with mis-aligned vertices arise when adjacent polygons are meant to share borders, but slight differences in the locations of the vertices cause them to  mis-align. This mis-alignment can also produce small "slivers" of polygons that are the areas where they overlap. You can snap vertices within a given distance of one another using the `snap` argument followed by a numeric value, like `snap = 0.000001`. Units of `snap` are in map units (usually meters) for projected coordinate reference systems and degrees for unprojected systems (e.g., WGS84, NAD83, NAD27). You can also remove polygons that are smaller than a particular area using the `area` argument followed by a numeric value (e.g., `area = 1`). The units of `area` are in meters-squared, regardless of the coordinate reference system. Note that using `snap` and `area` entails some risk, as it is possible for nearby vertices to actually be distinct and for small areas to be legitimate.
#' 3. **Automatic snapping and/or area removal**: In addition to the correction from step 1, you can use automatic `snap` and/or `area` correction on polygons vectors by setting `snap` and/or `area` to `NULL` (i.e., their default values). If just `snap` is `NULL`, only automatic snapping will be performed, and if just `area` is `NULL`, then only automatic area removal will be performed. Regardless, you will also need to set an integer value for `steps`, which is the number of steps to take between the smallest value of `snap` and/or `area` and the maximum value attempted. The function will then proceed by first attempting `snap = 0` and/or `area = 0` (i.e., no snapping or area removal). If this does not produce a topologically correct vector, **GRASS** will (internally) suggest a range for `snap`. The `fast()` function then creates `steps` values from the lowest to the highest values of this range evenly-spaced along the log values of this range, then proceed to repeat the importing process until either the vector is imported correctly or the maximum value of `snap` is reached and results in a failed topology. Smaller values of `step` will result in more fine-grained attempts so are less likely to yield overcorrection, but can also take more time. The value of `area` in automatic correction is set to `snap^2`. **NB**: Automated snapping and area removal are only performed on polygons vectors, even if `snap` or `area` is `NULL`. To snap lines or points, you must set `snap` equal to a numeric value. The `area` correction is ignored for points and lines.
#'
#' Issues can also arise due to:
#'
#' * **Data table-vector mismatching**: If your vector has a data table ("attribute table") associated with it, errors can occur if there are more/fewer geometries (or multi-geometries) per row in the table. If you do not really need the data table to do your analysis, you can remove it (and thus obviate this error) using `dropTable = TRUE`.
#' * **Dissolving or aggregating "invalid" geometries**: Using the `resolve` argument, you can create a topologically valid vector by either coercing all overlapping portions of polygons into their own geometries (`resolve = "disaggregate"`), or by coercing them into a single, combined geometry (`resolve = "aggregate"`). Aggregation/disaggregation will be implemented after loading the vector into **GRASS** using the settings given by `snap` and `area`. Aggregation/disaggregation will cause any associated data table to be dropped (it forces `dropTable` to be `TRUE`). The default action is to do neither aggregation nor disaggregation (`resolve = NA`). You can also do this outside **fasterRaster** using [terra::aggregate()] or [terra::disagg()].
#'
#' If none of these fixes work, you can try:
#'
#' * **Correction outside of *fasterRaster***: Before you convert the vector into **fasterRaster**'s `GVector` format, you can also try using the [terra::makeValid()] or [sf::st_make_valid()] tools to fix issues, then use `fast()`. You can also use [terra::aggregate()] or [terra::disagg()] to combine/split problematic geometries.
#' * **Post-load correction**: If you do get a vector loaded into `GVector` format, you can also use a set of **fasterRaster** vector-manipulation [tools][breakPolys] or [fillHoles()] to fix issues.
#'
#' @seealso \code{\link[rgrass]{read_RAST}} and \code{\link[rgrass]{read_VECT}}, [vector cleaning][breakPolys], [fillHoles()], plus **GRASS** modules `v.in.ogr` (see `grassHelp("v.in.ogr")`) and `r.import` (`grassHelp("r.import")`)
#'
#' @returns A `GRaster` or `GVector`.
#'
#' @example man/examples/ex_fast.r
#'
#' @aliases fast,character-method
#' @rdname fast
#' @name fast
#' @exportMethod fast
methods::setMethod(
	"fast",
	signature(x = "character"),
	function(
		x,
		rastOrVect = NULL,
		levels = TRUE,
		extent = NULL,
		correct = TRUE,
		snap = NULL,
		area = NULL,
		steps = 10,
		dropTable = FALSE,
		resolve = NA,
		verbose = TRUE,
		...
	) {

	### debugging
	if (FALSE) {
	
		rastOrVect <- NULL
		levels <- TRUE
		extent <- NULL
		correct <- TRUE
		snap <- NULL
		area <- NULL
		steps <- 10
		dropTable <- FALSE
		resolve <- NA
		verbose <- TRUE
		dots <- list()
	
	}

	if (is.na(faster("grassDir"))) stop("You must specify the folder in which GRASS is installed using faster().")

	# Error for cases where there are fewer geometries than rows in the data table. Topology correction using `snap` or `area` can only the number of geometries, so if the number of geometries is larger than the number of table rows, then iteratively trying larger values of `snap` or `area` will continue to create a mismatch between the number of geometries and table rows. In this case, just throw an error to obviate wasting time.
	nGeometriesVsTableRowsError <- "Since the data table has more rows than there are vector geometries,\n  increasing `snap` and/or `area` can only decrease the number of geometries.\n  Try smaller values of `snap` and/or `area`, or remove the table using\n  `dropTable = TRUE`."

	### dots
	########
	
	dots <- list(...)
	dotNames <- names(dots)

	x <- normalizePath(x, mustWork = FALSE)

	### raster or vector?
	#####################
	if (is.null(rastOrVect)) {

		### attempt to get type from extension
		# 3-letter extensions
		rastExtensions <- c("asc", "asci", "ascii", "grd", "hgt", "img", "mem", "tif", "tifg", "saga")
		vectExtensions <- c("shp", "gpkg", "kml")

		ext <- .fileExt(x[1L])
		ext <- tolower(ext)

		if (any(ext %in% rastExtensions)) {
			rastOrVect <- "raster"
		} else if (any(ext %in% vectExtensions)) {
			rastOrVect <- "vector"
		} else {
			stop("Cannot determine data if raster or vector from file name. Please use argument `rastOrVect`.")
		}

	} else {
		### user supplied rastOrVect
		rastOrVect <- omnibus::pmatchSafe(rastOrVect, c("raster", "vector"))
	}

	if (rastOrVect == "vector" & length(x) > 1L) stop("Cannot load more than one vector at a time.")

	### raster from disk
	####################
	
	if (rastOrVect == "raster") {
		
		# multiple rasters
		if (length(x) > 1L) {

			for (i in seq_along(x)) {
			
				thisOut <- fast(x[i], rastOrVect = "raster", ...)
			
				if (i == 1L) {
					out <- thisOut
				} else {
					out <- c(out, thisOut)
				}
			
			}

		# load just one raster
		} else {

			xRast <- terra::rast(x)
			nLayers <- terra::nlyr(xRast)
			
			xNames <- names(xRast)
			fileExtension <- .fileExt(x)
			nc <- nchar(fileExtension) + 1
			namesFile <- substr(x, 1, nchar(x) - nc)
			namesFile <- paste0(namesFile, '_names.csv')
			if (file.exists(namesFile)) {
				xNames <- utils::read.csv(namesFile)
				xNames <- xNames$name
			}
			
			location <- .locationFind(xRast, match = "crs")

			if (is.null(location) | !grassStarted()) {

				.locationCreate(x = xRast)
				location <- .location()

			}

			.locationRestore(x = location)

			.region(xRast)
			src <- .makeSourceName("fast_r_external", type = "raster", n = 1L)
			rgrass::execGRASS(
				cmd = "r.external",
				input = x,
				output = src,
				flags = c(.quiet(), "overwrite", "o") # overriding projection check!
			)
			
			if (!.exists(src)) stop("Raster not loaded. Check the file name. You may need to use an absolute (not relative) file path.")
			if (nLayers > 1L) src <- paste0(src, ".", seq_len(nLayers))

			### raster levels
			# if (is.list(levels)) {
			# 	if (length(levels) == 1L & length(xNames) > 1L & is.null(levels[[1L]])) levels <- NULL
			# } else if (inherits(levels, c("data.frame", "data.table"))) {
				# levels <- list(levels)
			if (is.logical(levels) && levels) {
				
				info <- .rastInfo(src)
				
				# search for and add levels
				if (any(info$grassDataType == "CELL")) {

					levelsFileName <- x
					levelsFileExt <- .fileExt(x)
					levelsFileName <- substr(levelsFileName, 1L, nchar(x) - nchar(levelsFileExt))

					# extensions <- c("rds", "RDS", "rdata", "RData", "rda", "RDa", "Rda", "Rdat", "rdat", "RDat", "csv", "CSV", "tab", "TAB")
					extensions <- c("rds", "rdata", "rda", "rdat", "csv", "tab")
					levelsFileNames <- paste0(levelsFileName, extensions)

					fileExists <- file.exists(levelsFileNames)
					if (any(fileExists)) {
					
						# if (sum(fileExists) > 1L) warning("More than one `levels` file found. Only the first will be used.")
						levelsFileName <- levelsFileNames[fileExists]
						extension <- extensions[fileExists]

						if (tolower(extension) == "rds") {
							levels <- readRDS(levelsFileName)
						} else if (tolower(extension) %in% c("rdata", "rda", "rdat")) {
							load(levelsFileName)
						} else if (tolower(extension) == "tab") {
							levels <- utils::read.delim(levelsFileName)
						} else if (tolower(extension) == "csv") {
							levels <- utils::read.csv(levelsFileName)
						}

					} else {
						levels <- NULL
					}

				} else {
					levels <- NULL
				}
			
			}

			if (exists("info", inherits = FALSE)) {
				out <- makeGRaster(info, names = xNames, levels = levels)
			} else {
				out <- makeGRaster(src, names = xNames, levels = levels)
			}
			
		}

	### vector from disk
	####################
	
	} else if (rastOrVect == "vector") {

		# x is a filename and xVect is missing: we have NOT come through methods for SpatVectors or sf objects
		if (!any(dotNames == "xVect")) {
			
			x <- terra::vect(x)
			out <- fast(x, extent = extent, correct = correct, snap = snap, area = area, steps = steps, resolve = resolve, dropTable = dropTable, verbose = verbose, ...)

		# x is a filename and xVect is present: we HAVE come here through a method for SpatVectors or sf objects
		} else {

			if (!is.na(resolve)) resolve <- omnibus::pmatchSafe(resolve, c("aggregate", "disaggregate"), n = 1L)

			xVect <- dots$xVect
			if (any(dotNames == "table")) {
				table <- dots$table
			} else {
				table <- NULL
			}

			gtype <- terra::geomtype(xVect)
			if (gtype == "points") {
				gtype <- "point"
			} else if (gtype == "lines") {
				gtype <- "line"
			} else if (gtype == "polygons") {
				gtype <- "area"
			}

			# location, location, location...
			location <- .locationFind(xVect, return = "name", match = "crs")

			if (is.null(location) | !grassStarted()) {

				.locationCreate(x = xVect)
				location <- .location()

			}
			.locationRestore(x = location)

			# correct topology?
			if (correct) {
				correctTopoFlag <- NULL
			} else {
				correctTopoFlag <- "c" # no correction
			}

			# no snapping/area removal for particular geometry types
			if (is.null(snap) & gtype == "point") snap <- -1
			if (is.null(area) & gtype %in% c("line", "point")) area <- 0

			### first try (no snapping or area removal)
			if (is.null(snap) || snap <= 0) {
				thisSnap <- -1
				thisSnapNice <- "no"
			} else {
				thisSnap <- snap
				thisSnapNice <- paste0(thisSnap, " map-units")
			}

			if (is.null(area) || area == 0) {
				thisArea <- 0
				thisAreaNice <- "no polygon removal"
			} else {
				thisArea <- area
				thisAreaNice <- paste0("removal of polygons of <", thisArea, " m2")
			}

			if ((verbose | faster("verbose")) & gtype == "area") {
				omnibus::say("Creating GVector with ", thisSnapNice, " snapping and ", thisAreaNice, "...")
			}

			src <- .makeSourceName("fast_v_in_ogr", "vector")
			if (is.null(snap) & (is.null(area) || area == 0)) {

				# slower if we need to record messages
				suppressMessages(
					assessment <- rgrass::execGRASS(
						cmd = "v.in.ogr",
						input = x,
						output = src,
						snap = thisSnap,
						min_area = thisArea,
						flags = c("verbose", "overwrite", "t", "o", correctTopoFlag),
						ignore.stderr = FALSE,
						Sys_show.output.on.console = FALSE,
						echoCmd = FALSE, # displays GRASS command
						intern = TRUE
					)
				)

				valid <- !any(c(
					grepl(assessment, pattern = "WARNING: The output contains topological errors"),
					grepl(assessment, pattern = "Invalid argument"),
					assessment == "1"
				))

				if (valid) { # more thorough test... slower

					info <- .vectInfo(src)
					valid <- .validVector(info, table)
				
					if (!valid & !is.na(resolve)) {

						table <- NULL
						src <- .aggDisaggVect(src, resolve = resolve, verbose = verbose)

						# info <- .vectInfo(src)
						# valid <- .validVector(info, table)
						valid <- TRUE

					}

					if (!valid & verbose & !dropTable) {
						omnibus::say("   Vector has ", info$nGeometries, " valid geometries, ", sum(is.na(info$cats)), " invalid geometries, and ", nrow(table), " rows in its data table.")
						if (info$nGeometries < nrow(table)) {
							stop(nGeometriesVsTableRowsError)
						}
					} else if (!valid & verbose) {
						omnibus::say("   Vector has ", info$nGeometries, " valid geometries and ", sum(is.na(info$cats)), " invalid geometries.")
					}

				}

			} else {

				# faster if we remove fluff
				rgrass::execGRASS(
					cmd = "v.in.ogr",
					input = x,
					output = src,
					snap = thisSnap,
					min_area = thisArea,
					flags = c(.quiet(), "overwrite", "t", correctTopoFlag)
				)

				# if (!exists("table", inherits = TRUE, mode = "numeric")) table <- .vAsDataTable(src)
				info <- .vectInfo(src)
				valid <- .validVector(info, table)

				if (!valid & !is.na(resolve)) {

					table <- NULL
					src <- .aggDisaggVect(src, resolve = resolve, verbose = verbose)

					# info <- .vectInfo(src)
					# valid <- .validVector(info, table)
					valid <- TRUE

				}

				if (!valid & verbose & !dropTable) {
					omnibus::say("   Vector has ", info$nGeometries, " valid geometries, ", sum(is.na(info$cats)), " invalid geometries, and ", nrow(table), " rows in its data table.")
						if (info$nGeometries < nrow(table)) {
							stop(nGeometriesVsTableRowsError)
						}
				} else if (!valid & verbose) {
					omnibus::say("   Vector has ", info$nGeometries, " valid geometries and ", sum(is.na(info$cats)), " invalid geometries.")
				}

			}
			
			### automated vector topology correction
			if (!valid & (is.null(snap) | is.null(area))) {
			
				if (!exists("assessment", inherits = FALSE)) {
				
					suppressMessages(
						assessment <- rgrass::execGRASS(
							cmd = "v.in.ogr",
							input = x,
							output = src,
							snap = thisSnap,
							min_area = thisArea,
							flags = c("verbose", "overwrite", "t", "o", correctTopoFlag),
							ignore.stderr = FALSE,
							Sys_show.output.on.console = FALSE,
							echoCmd = FALSE, # displays GRASS command
							intern = TRUE
						)
					)
				
				}

				stepsMinus1 <- steps - 1L
				snapRange <- assessment[grepl(assessment, pattern = "Estimated range of snapping threshold:")]

				# generic snap range
				if (length(snapRange) == 0L) {

					snapRange <- c(1E-08, 1)

				# GRASS-suggested snap range
				} else {

					snapRange <- substr(snapRange, 41L, nchar(snapRange))
					snapRange <- sub(snapRange, pattern = "\\]", replacement = "")
					snapRange <- strsplit(snapRange, split = ", ")[[1L]]
					# snapRange <- c(snapRange[[1L]], snapRange[[2L]])
					snapRange <- as.numeric(snapRange)

				}

				# create sequence of snap values
				# evenly-spaced in log space bc suggested min/max values are usually several OOMs apart
				snapRange <- log(snapRange)
				snaps <- seq(snapRange[1L], snapRange[2L], length.out = steps - 1L)
				snaps <- exp(snaps)

				digits <- abs(floor(log10(c(snaps[1L]^2))))

				# snap AUTO and area AUTO
				if (is.null(snap) & is.null(area)) {
				
					step <- 1L
					while (!valid & step <= stepsMinus1) {
					
						thisSnap <- snaps[step]
						thisArea <- snaps[step]^2

						if ((verbose | faster("verbose")) & gtype == "area") {
							
							thisSnapNice <- round(thisSnap, digits)
							thisAreaNice <- round(thisArea, digits)

							omnibus::say("Iteration ", step, ": Snapping at ", thisSnapNice, " map-units and removing polygons of <", thisAreaNice, " m2...")

						}

						src <- .makeSourceName("fast_v_in_ogr", "vector")
						rgrass::execGRASS(
							cmd = "v.in.ogr",
							input = x,
							output = src,
							snap = thisSnap,
							min_area = thisArea,
							flags = c(.quiet(), "overwrite", "t", correctTopoFlag)
						)
						
						made <- .exists(src)
						if (made) {
							info <- .vectInfo(src)
							valid <- .validVector(info, table)
						} else {
							valid <- FALSE
						}

						if (made & !valid & !is.na(resolve)) {

							table <- NULL
							src <- .aggDisaggVect(src, resolve = resolve, verbose = verbose)

							# info <- .vectInfo(src)
							# valid <- .validVector(info, table)
							valid <- TRUE

						}

						if (made & !valid & verbose & !dropTable) {
							omnibus::say("   Vector has ", info$nGeometries, " valid geometries, ", sum(is.na(info$cats)), " invalid geometries, and ", nrow(table), " rows in its data table.")
							if (info$nGeometries < nrow(table)) {
								stop(nGeometriesVsTableRowsError)
							}
						} else if (made & !valid & verbose) {
							omnibus::say("   Vector has ", info$nGeometries, " valid geometries and ", sum(is.na(info$cats)), " invalid geometries.")
						} else if (!made) {
							omnibus::say("   Failed to create vector.")
						}

						step <- step + 1L

					}
					
				# snap AUTO and area NUMERIC
				} else if (is.null(snap) & is.numeric(area)) {
					
					step <- 1L
					while (!valid & step <= stepsMinus1) {
					
						thisSnap <- snaps[step]
						thisArea <- area

						if ((verbose | faster("verbose")) & gtype == "area") {
							
							thisSnapNice <- round(thisSnap, digits)
							omnibus::say("Iteration ", step, ": Snapping at ", thisSnapNice, " map-units...")

						}

						src <- .makeSourceName("fast_v_in_ogr", "vector")
						rgrass::execGRASS(
							cmd = "v.in.ogr",
							input = x,
							output = src,
							snap = thisSnap,
							min_area = thisArea,
							flags = c(.quiet(), "overwrite", "t", correctTopoFlag)
						)
						
						made <- .exists(src)

						if (made) {
							info <- .vectInfo(src)
							valid <- .validVector(info, table)
						} else {
							valid <- FALSE
						}

						if (made & !valid & !is.na(resolve)) {

							table <- NULL
							src <- .aggDisaggVect(src, resolve = resolve, verbose = verbose)

							# info <- .vectInfo(src)
							# valid <- .validVector(info, table)
							valid <- TRUE

						}

						if (made & !valid & verbose & !dropTable) {
							omnibus::say("   Vector has ", info$nGeometries, " valid geometries, ", sum(is.na(info$cats)), " invalid geometries, and ", nrow(table), " rows in its data table.")
							if (info$nGeometries < nrow(table)) {
								stop(nGeometriesVsTableRowsError)
							}
						} else if (made & !valid & verbose) {
							omnibus::say("   Vector has ", info$nGeometries, " valid geometries and ", sum(is.na(info$cats)), " invalid geometries.")
						} else if (!made) {
							omnibus::say("   Failed to create vector.")
						}
						step <- step + 1L

					}
					
				# snap NUMERIC and area AUTO
				} else if (is.numeric(snap) & is.null(area)) {
					
					step <- 1L
					while (!valid & step <= stepsMinus1) {
					
						if (snap <= 0) {
							thisSnap <- -1
						} else {
							thisSnap <- snap
						}
						thisArea <- snaps[step]^2
						
						if ((verbose | faster("verbose")) & gtype == "area") {
						
							thisAreaNice <- round(thisArea, digits)
							omnibus::say("Iteration ", step, ": Removing polygons of <", thisAreaNice, " m2...")

						}

						src <- .makeSourceName("fast_v_in_ogr", "vector")
						rgrass::execGRASS(
							cmd = "v.in.ogr",
							input = x,
							output = src,
							snap = thisSnap,
							min_area = thisArea,
							flags = c(.quiet(), "overwrite", "t", correctTopoFlag)
						)
						
						made <- .exists(src)
						if (made) {
							info <- .vectInfo(src)
							valid <- .validVector(info, table)
							if (!exists("table", inherits = TRUE)) table <- .vAsDataTable(src)
						} else {
							valid <- FALSE
						}

						if (made & !valid & !is.na(resolve)) {

							table <- NULL
							src <- .aggDisaggVect(src, resolve = resolve, verbose = verbose)

							# info <- .vectInfo(src)
							# valid <- .validVector(info, table)
							valid <- TRUE

						}

						if (made & !valid & verbose & !dropTable) {
							omnibus::say("   Vector has ", info$nGeometries, " valid geometries, ", sum(is.na(info$cats)), " invalid geometries, and ", nrow(table), " rows in its data table.")
							if (info$nGeometries < nrow(table)) {
								stop(nGeometriesVsTableRowsError)
							}
						} else if (made & !valid & verbose) {
							omnibus::say("   Vector has ", info$nGeometries, " valid geometries and ", sum(is.na(info$cats)), " invalid geometries.")
						} else if (!made) {
							omnibus::say("   Failed to create vector.")
						}
						step <- step + 1L

					}
				
				} # next type of topology correction while loading
		
			} # if not valid and doing automated correction

			if (!valid) {

				msg <- paste0("Vector has an invalid topology. Try:\n  * Setting the `correct` argument to `TRUE`;\n  * Increasing the value(s) of the `snap` and/or `area` arguments;\n  * Using automated `snap` and/or `area` correction (set `snap` and/or `area` to NULL);\n  * Dropping the data table associated with the vector using `dropTable = TRUE`;\n  * Aggregating or disaggregating polygons with the `resolve` argument; or\n  * Attempting correction of the vector outside of fasterRaster with `terra::makeValid()` or `sf::st_make_valid()` before using fast().")

				stop(msg)

			}

			out <- makeGVector(src = src, table = table)
			
			# # save/reload vector... seems to fix errors with subset_single_bracket on polygon vectors later on
			# if (reload & geomtype(out) == "polygons") {

			# 	if (verbose | faster("verbose")) omnibus::say('Reloading GVector...')
			# 	tempVect <- tempfile(fileext = ".gpkg")

			# 	args <- list(
			# 		cmd = "v.out.ogr",
			# 		input = sources(out),
			# 		output = tempVect,
			# 		format = "GPKG",
			# 		flags = c(.quiet(), "s", "overwrite")
			# 		# flags = c(.quiet(), "s", "c") # "c" ==> save geometries lacking a cat number
			# 	)
			# 	do.call(rgrass::execGRASS, args)

			# 	Sys.sleep(0.5)
			# 	src <- .makeSourceName("v_in_ogr", "vector")
			# 	rgrass::execGRASS(
			# 		cmd = "v.in.ogr",
			# 		input = tempVect,
			# 		output = src,
			# 		flags = c(.quiet(), "overwrite", "t", "c")
			# 	)
			# 	out <- makeGVector(src, table = table)

			# 	# writeVector(out, filename = tempVect, overwrite = TRUE)
			# 	# out <- fast(tempVect, reload = FALSE, table = table) # cannot use vect() on this... weird!
			
			# }

			if ((verbose | faster("verbose")) & geomtype(out) == "polygons") omnibus::say("Topologically valid vector created.")

		} # x is a filename and xVect supplied

	}
	out

	} # EOF
)

#' @aliases fast,SpatRaster-method
#' @rdname fast
#' @exportMethod fast
methods::setMethod(
	"fast",
	signature(x = "SpatRaster"),
	function(x, ...) {

	if (is.na(faster("grassDir"))) stop("Before using fasterRaster functions, you must specify the folder in which GRASS is installed using faster().")

	rastFile <- terra::sources(x)
	levels <- .getLevels(x)

	if (any(rastFile == "")) {

		tempFile <- tempfile(fileext = ".tif")
		terra::writeRaster(x, filename = tempFile, overwrite = TRUE)
		rastFile <- tempFile

	}

	# if this SpatRaster is a subset of rasters in a file, save the subset to disk... otherwise, fast(signature = 'character') will load the entire set of layers and fail to match the levels() from the subset
	thisRast <- terra::rast(rastFile)
	nlay <- terra::nlyr(thisRast)
	nlev <- length(levels)

	if (nlay != nlev) {

		tempFile <- tempfile(fileext = ".tif")
		terra::writeRaster(x, filename = tempFile, overwrite = TRUE)
		rastFile <- tempFile

	}

	fast(x = rastFile, rastOrVect = "raster", levels = levels)

	} # EOF
)

#' @aliases fast,SpatVector-method
#' @rdname fast
#' @exportMethod fast
methods::setMethod(
	"fast",
	signature(x = "SpatVector"),
	function(x, extent = NULL, correct = TRUE, snap = NULL, area = NULL, steps = 10, dropTable = FALSE, resolve = NA, verbose = TRUE) .fastVector(x, extent = extent, correct = correct, snap = snap, area = area, steps = steps, dropTable = dropTable, resolve = resolve, verbose = verbose)
)

#' @aliases fast,sf-method
#' @rdname fast
#' @exportMethod fast
methods::setMethod(
	"fast",
	signature(x = "sf"),
	function(x, extent = NULL, correct = TRUE, snap = NULL, area = NULL, steps = 10, resolve = NA, dropTable = FALSE, verbose = TRUE) .fastVector(x, correct = correct, snap = snap, area = area, steps = steps, extent = extent, dropTable = dropTable, resolve = resolve, verbose = verbose)
)

#' @rdname fast
#' @aliases fast,missing-method
#' @exportMethod fast
methods::setMethod(
	"fast",
	signature(x = "missing"),
	function(x, rastOrVect, crs = "") {
	
	if (crs == "") { 
	
		crs <- tryCatch(crs(), error = function(cond) FALSE)

		if (is.logical(crs)) stop("You must provide a coordinate reference system with `crs` or\n  have created or loaded at least one GRaster or GVector.")

	}

	rastOrVect <- omnibus::pmatchSafe(rastOrVect, c("raster", "vector"), nmax = 1L)
	if (rastOrVect == "raster") {

		x <- terra::rast()
		x <- terra::project(x, crs)
		x[] <- 1L
		tf <- tempfile(fileext = ".tif")
		terra::writeRaster(x, tf, overwrite = TRUE)
		out <- fast(tf)

	} else {

		x <- c(-180, 180, -90, 90)
		x <- terra::ext(x)
		x <- terra::as.polygons(x, crs = "epsg:4326")
		x <- terra::project(x, crs)
		out <- .fastVector(x, correct = TRUE, snap = FALSE, area = FALSE, steps = 10, extent = NULL, dropTable = FALSE, resolve = NA, verbose = FALSE)

	}
	out

	} # EOF
)

#' @rdname fast
#' @aliases fast,numeric-method
#' @exportMethod fast
methods::setMethod(
	"fast",
	signature(x = "numeric"),
	function(x, crs = "", keepgeom = FALSE) {
	
	if (length(x) %% 2L != 0L) stop("You must supply an even number of numeric values, each representing longitude/latitude pairs.")
	x <- matrix(x, ncol = 2L, byrow = TRUE)
	.fastDF(x = x, geom = 1:2, crs = crs, keepgeom = keepgeom)

	} # EOF
)

#' @rdname fast
#' @aliases fast,data.frame-method
#' @exportMethod fast
methods::setMethod(
	"fast",
	signature(x = "data.frame"),
	function(x, geom = 1:2, crs = "", keepgeom = FALSE) .fastDF(x = x, geom = geom, crs = crs, keepgeom = keepgeom)
)

#' @rdname fast
#' @aliases fast,data.table-method
#' @exportMethod fast
methods::setMethod(
	"fast",
	signature(x = "data.table"),
	function(x, geom = 1:2, crs = "", keepgeom = FALSE) .fastDF(x = x, geom = geom, crs = crs, keepgeom = keepgeom)
)

#' @rdname fast
#' @aliases fast,matrix-method
#' @exportMethod fast
methods::setMethod(
	"fast",
	signature(x = "matrix"),
	function(x, geom = 1:2, crs = "", keepgeom = FALSE) .fastDF(x = x, geom = geom, crs = crs, keepgeom = keepgeom)
)

#' Create a points GVector from a data.frame or similar
#'
#' @param x Two numeric values, or a `data.frame`, `data.table`, or `matrix`.
#' @param geom 2-element 
#' @param crs WKT2 string
#' @param keepgeom Logical
#' @noRd
.fastDF <- function(x, geom, crs, keepgeom) {

	if (crs == "") {
	
		crs <- tryCatch(crs(), error=function(cond) FALSE)

		if (is.logical(crs)) stop("You must provide a coordinate reference system with `crs` or\n  have created or loaded at least one GRaster or GVector.")

	}

	if (!is.data.frame(x)) x <- as.data.frame(x)

	if (is.numeric(geom) | is.integer(geom)) {
		xcol <- names(x)[geom[1L]]
		ycol <- names(x)[geom[2L]]
	} else {
		xcol <- geom[1L]
		ycol <- geom[2L]
	}

	xVect <- terra::vect(x, geom = c(xcol, ycol), crs = crs, keepgeom = keepgeom)
	fast(xVect, correct = TRUE)

}


#' Test if a vector is valid
#'
#' @param x Either the [sources()] name of a **GRASS** vector or a `vectInfo` object created by [.vectInfo()].
#' @param table Either `NULL` (no table) or a `data.frame`, `data.table`, or `matrix`
#'
#' @returns Logical.
#'
#' @noRd
.validVector <- function(x, table) {

	# does table have same number of rows as vector geometries?
	if (!is.null(table) && nrow(table) > 0L) {

		if (!inherits(x, "vectInfo")) x <- .vectInfo(x)
		nGeoms <- x$nGeometries
		tableValid <- nGeoms == nrow(table)

	} else {
		tableValid <- TRUE
	}

	# if the table is valid, are its category numbers valid?
	if (tableValid) {
		if (!inherits(x, "vectInfo")) x <- .vectInfo(x)
		catsValid <- x$catsValid
	} else {
		catsValid <- FALSE
	}

	tableValid & catsValid

}

# 1. Write vector to disk (if needed)
# 2. Send to fast(signature = "character")
#' @noRd
.fastVector <- function(
	x,
	extent,
	correct,
	snap,
	area,
	steps,
	dropTable,
	resolve,
	verbose,
	...
) {

	if (is.na(faster("grassDir"))) stop("You must specify the folder in which GRASS is installed using faster().")

	dots <- list(...)
	dotNames <- names(dots)
	
	if (!inherits(x, "SpatVector")) x <- terra::vect(x)
	
	# crop
	if (!is.null(extent)) {
		
		if (inherits(extent, "GSpatial")) {
			extent <- ext(extent)
		} else if (inherits(extent, c("SpatVector", "SpatRaster", "SpatExtent"))) {
			extent <- terra::ext(extent)
		} else if (inherits(extent, "sf")) {
			extent <- sf::st_bbox
			extent <- as.vector(extent)
			extent <- extent[c(1L, 3L, 2L, 4L)]
			extent <- terra::ext(extent)
		} else if (is.numeric(extent)) {
			if (length(extent) != 4L) stop("The `extent` argument must be a spatial object or a vector of four numeric values.")
			extent <- terra::ext(extent)
		} else {
			stop("The `extent` argument must be a spatial object or a vector of four numeric values.")
		}

		x <- terra::crop(x, extent)

		if (nrow(x) == 0L) {
			warning("No geometries occur within the given extent. Returning `NULL`.")
			return(NULL)
		}
		
	}
	
	xVect <- x

	# remove data frame
	if (dropTable) {
		table <- NULL
	} else {
		table <- data.table::as.data.table(xVect)
	}

	if (terra::sources(x) == "") {

		# remove all but one column of data table
		# NB we ***need** a table with the GVector--otherwise, subset_single_bracket does not work as expected
		if (!is.null(table) && ncol(table) > 1L) {
			nc <- 2L:ncol(xVect)
			xVect[ , nc] <- NULL 
		}
		
		# NB we ***need** a table with the GVector--otherwise, subset_single_bracket does not work as expected
		if (is.null(table)) xVect$DUMMYDUMMY_ <- 1L:nrow(xVect)

		vectFile <- paste0(faster("workDir"), "/", omnibus::rstring(1L), ".gpkg")
		terra::writeVector(xVect, filename = vectFile, filetype = "GPKG", overwrite = TRUE)
	
	} else {
		vectFile <- terra::sources(x)
	}

	# sometimes, terra::sources() adds the layer name after "::" to the filename if the vector was saved by v.out.ogr
	if (grepl(vectFile, pattern = "::")) {

		colons <- regexpr(vectFile, pattern = "::", fixed = TRUE)
		vectFile <- substr(vectFile, 1L, colons - 1L)

	}

	# NB not passing extent bc already cropped if we wanted to do that
	args <- list(
		x = vectFile,
		rastOrVect = "vector",
		extent = NULL,
		correct = correct,
		snap = snap,
		area = area,
		steps = steps,
		dropTable = dropTable,
		resolve = resolve,
		table = table,
		xVect = xVect,
		verbose = verbose
	)
	args <- c(args, list(...))
	do.call(fast, args = args)
	
}

#' Set and get categories for categorical rasters
#'
#' @description `GRaster`s can represent categorical data. Cell values are actually integers, each corresponding to a category, such as "desert" or "wetland." A categorical raster is associated with a table that matches each value to a category name. The table must be `NULL` (i.e., no categories--so not a categorical raster), or have at least two columns. The first column must have integers and represent raster values. One or more subsequent columns must have category labels. The column with these labels is the "active category".
#'
#' * `levels()`: Displays the "levels" table of a raster (just the value and active category columns).
#' * `cats()`: Displays the entire "levels" table of a raster.
#' * `levels()<-`: (Re)assigns the "levels" table to each layer of a raster. Assigning a "levels" table to an integer raster makes it a categorical raster.
#' * `categories()`: (Re)assigns the "levels" table to specific layer(s) of a raster.
#' * For a complete list of functions relevant to categorical rasters, see `vignette("GRasters", package = "fasterRaster")).
#'
#' @param x A `GRaster`.
#'
#' @param value A `data.frame`, `data.table`, a list of `data.frames` or `data.tables` with one per raster layer, or a categorical `SpatRaster`. The table's first column is the "value" column and must contain numeric values (of class `numeric` or `character`). If a `SpatRaster` is supplied, then its categories will be transferred to the `GRaster`.
#'
#' @param layer Numeric integers, logical vector, or character: For `cats()` and `categories()`, this specifies the layer(s)for which to obtain level(s).
#'
#' @param active An integer or a character: The index or column name of the column used for category labels (the "active column"). Following [terra::activeCat()], the first column of the "levels" table is ignored, so a value of 1 means to use the second column of the table for labels. A value of 2 means to use the third column, and so on.
#'
#' @returns Values returned are:
#' * `levels()` and `cats()`: A list of `data.frame`s or `data.table`s, one per raster layer.
#' * `levels()<-` and `categories()`: A `GRaster`.
#'
#' @seealso [terra::levels()], \code{\link[fasterRaster]{levels<-}}, [terra::cats()], [terra::categories()], see `vignette("GRasters", package = "fasterRaster")`
#'
#' @example man/examples/ex_GRaster_categorical.r
#'
#' @aliases levels
#' @rdname levels
#' @exportMethod levels
methods::setMethod(
    f = "levels",
    signature = c(x = "GRaster"),
    function(x) {

	out <- cats(x)
	numLevels <- nlevels(x)

	for (i in seq_along(out)) {

		if (numLevels[i] > 0L) {
		
			names <- catNames(x, i)[[1L]]
			active <- activeCat(x, layer = i, name = TRUE)
			value <- names[1L]

			cols <- c(value, active)

			if (faster("useDataTable")) {
				out[[i]] <- out[[i]][ , ..cols]
			} else {
				out[[i]] <- out[[i]][ , cols, drop = FALSE]
			}

		}
	}

	names(out) <- names(x)
	out

    } # EOF
)

#' @aliases cats
#' @rdname levels
#' @exportMethod cats
methods::setMethod(
    f = "cats",
    signature = c(x = "GRaster"),
    function(x, layer = 1:nlyr(x)) {
	
	layer <- .layerIndex(layer, x, recycle = TRUE)
	
	out <- x@levels[layer]
	numLevels <- .nlevels(out)
	for (i in seq_along(out)) {
		if (numLevels[i] == 0L) out[[i]] <- data.table::data.table(NULL)
	}
	
	# out <- lapply(out, .replace0LevelsWithNull)

	if (!faster("useDataTable")) {
		for (i in seq_along(out)) {
			if (numLevels[i] > 0L) out[[i]] <- as.data.frame(out[[i]])
		}
	}
	names(out) <- names(x)[layer]
	out

    } # EOF
)

# #' Replaces data.table/frames in a list with 0 rows with NULL
# #'
# #' Have to do this because defining a list element as 'NULL' removes that element entirely. Fix from https://stackoverflow.com/questions/7944809/assigning-null-to-a-list-element-in-r.
# #'
# #' @param y A data.table/frame
# #'
# #' @returns A data.table/frame or 'NULL'.
# #'
# #' @noRd
# .replace0LevelsWithNull <- function(y) {

	# if (nrow(y) == 0L) {
		# NULL
	# } else {
		# y
	# }

# }

#' @aliases categories
#' @rdname levels
#' @exportMethod categories
methods::setMethod(
	f = "categories",
	signature = c(x = "GRaster"),
	function(x, layer = 1, value, active = 1) {

	# layer <- NULL # obviates check(): "no visible binding for global variable `layer`"
	layer <- .layerIndex(layer = layer, x, recycle = TRUE)

	if (!inherits(value, "list")) value <- list(value)

	if (length(list) != length(layer)) stop("The number of level tables is not the same as the number of raster layers.")

	for (i in layer) {

		if (is.character(value[[i]]) && value[[i]] == "") {
			value[[i]] <- data.table::data.table(NULL)
		} else if (!inherits(value[[i]], "data.table")) {
   			value[[i]] <- data.table::as.data.table(value[[i]])
		}

		# set value to integer and sort by value
		if (!is.character(value[[i]])) {

			names <- names(value[[i]])
			valueCol <- names[1L]

			# convert first column to integer
			value[[i]][ , (valueCol) := lapply(.SD, as.integer), .SDcols = valueCol]

			# detect non-unique values
			unis <- unique(value[[i]][, 1L])
			numUnis <- nrow(unis)
			if (numUnis < nrow(value[[i]])) stop("The value column (the first column) must have unique values.")

			# sort by first column
			data.table::setorderv(value[[i]], col = valueCol)
		
		}

	}

	for (i in layer) {
		x@levels[[i]] <- value[[i]]
		x@activeCat[i] <- as.integer(active + 1L)
	}
	methods::validObject(x)
	x

	} # EOF
)

#' @aliases levels<-
#' @rdname levels
#' @exportMethod levels<-
methods::setMethod(
	f = "levels<-",
	signature = c(x = "GRaster", value = "data.frame"),
	function(x, value) {
	
	value <- list(value)
	levels(x) <- value
	x
	
	} # EOF
)

#' @aliases levels<-
#' @rdname levels
#' @exportMethod levels<-
methods::setMethod(
	f = "levels<-",
	signature = c(x = "GRaster", value = "data.table"),
	function(x, value) {
	
	value <- list(value)
	levels(x) <- value
	x
	
	} # EOF
)

# #' @aliases levels<-
# #' @rdname levels
# #' @exportMethod levels<-
# methods::setMethod(
# 	f = "levels<-",
# 	signature = c(x = "GRaster", value = "SpatRaster"),
# 	function(x, value) {

# 	value <- cats(value)
# 	levels(x) <- value
# 	x@activeCat <- terra::activeCat(value, layer = 1L)
# 	x

# 	} # EOF
# )

#' @aliases levels<-
#' @rdname levels
#' @exportMethod levels<-
methods::setMethod(
	f = "levels<-",
	signature = c(x = "GRaster", value = "GRaster"),
	function(x, value) {

	value <- cats(value)
	levels(x) <- value
	x@activeCat <- activeCats(value)
	x

	} # EOF
)

#' @aliases levels<-
#' @rdname levels
#' @exportMethod levels<-
methods::setMethod(
	f = "levels<-",
	signature = c(x = "GRaster", value = "list"),
	function(x, value) {
	
	if (length(value) != nlyr(x)) stop("The number of level tables is not the same as the number of raster layers.\n  If you want to change the level table of specifc layers, use categories().")

	for (i in seq_along(value)) {

		# convert empty strings to NULL data.tables
		if (is.character(value[[i]]) && value[[i]] == "") {

			value[[i]] <- data.table::data.table(NULL)

		# convert to data.table
		} else if (!inherits(value[[i]], "data.table")) {
			
			value[[i]] <- data.table::as.data.table(value[[i]])

		}

		names <- names(value[[i]])
		valueCol <- names[1L]

		# convert first column to integer
		value[[i]][ , (valueCol) := lapply(.SD, as.integer), .SDcols = valueCol]

		# detect non-unique values
		unis <- unique(value[[i]][ , 1L])
		numUnis <- nrow(unis)
		if (numUnis < nrow(value[[i]])) stop("The value column (the first column) must have unique values.")

		# sort by first column
		data.table::setorderv(value[[i]], col = valueCol)

	}

	x@activeCat <- rep(2L, nlyr(x))
	x@levels <- value
 	a <- methods::validObject(x)
	x

	} # EOF
)

#' Combine levels table from multiple categorical GRasters
#'
#' @description This function creates a single "levels" table from the levels tables of one or more categorical `GRaster`s.
#'
#' The difference between this function and [concats()] is that `concats()` creates a "combined" `GRaster` with a combined levels table, whereas this one just merges the levels tables.
#'
#' @param x A `GRaster` or a `list` of `GRaster`s.
#' @param ... Arguments to pass to [data.table::merge()].
#'
#' @returns A `list` with a "levels" table (a `data.frame` or `data.table`), and the active category number for the new table. Following [terra::activeCat()], the number is offset by 1, so a value of 1 indicates that the second column in the table should be used for the category labels, a value of 2 indicates the third column should be used, and so on.
#'
#' @example man/examples/ex_GRaster_categorical.r
#'
#' @seealso [concats()], [terra::concats], `vignette("GRasters", package = "fasterRaster")`
#'
#' @aliases combineLevels
#' @rdname combineLevels
#' @exportMethod combineLevels
methods::setMethod(
	f = "combineLevels",
	signature = c(x = "GRaster"),
	function(x, ...) {

	if (!any(is.factor(x))) {
	
		out <- data.table::data.table(NULL)
		if (!faster("useDataTable")) out <- as.data.frame(out)
	
	} else {
		x <- cats(x)
		out <- .combineLevels(x = x, ...)
	}
	out
	
	}
)

#' @aliases combineLevels
#' @rdname combineLevels
#' @exportMethod combineLevels
methods::setMethod(
	f = "combineLevels",
	signature = c(x = "list"),
	function(x, ...) {
	
	if (!all(sapply(x, inherits, what = 'GRaster'))) stop("`x` must be a `GRaster` with one or more layers, or a `list` of `GRaster`s.")
	xx <- x
	x <- list()
	for (i in seq_along(xx)) x <- c(x, cats(xx[[i]]))
	.combineLevels(x = x, ...)
	
	}
)

#' Combine "levels" tables
#'
#' @param x A list of `data.frame`s or `data.table`s.
#'
#' @noRd
.combineLevels <- function(x, ...) {


	for (i in seq_along(x)) {
		if (!inherits(x[[i]], "data.table")) x[[i]] <- data.table::as.data.table(x[[i]])
	}

	# get first levels table
	i <- 1L
	out <- data.table::data.table(NULL)
	while (nrow(out) == 0L & i <= length(x)) {
		out <- x[[i]]
		i <- i + 1L
	}

	if (i <= length(x)) {
		
		by.x <- names(out)[1L]
		for (j in i:length(x)) {
			cats2 <- x[[j]]
			if (nrow(cats2) > 0L) {
				by.y <- names(cats2)[1L]
				# out <- merge(out, cats2, by.x = by.x, by.y = by.y)
				out <- merge(out, cats2, all = TRUE, ...)
			}
		}
		
	}
	
	if (!faster("useDataTable")) out <- as.data.frame(out)
	out
	
}

if (grassStarted()) {

# Setup
library(sf)
library(terra)

### This example creates a simple model of Dypsis distribution using
# elevation, distance to forest, land cover class, and nearness to rivers.

# Elevation raster, forest cover in year 2000, land cover class, and
# points where Dypsis plants have been collected
madElev <- fastData("madElev")
madForest2000 <- fastData("madForest2000")
madCover <- fastData("madCover")
madRivers <- fastData("madRivers")
madDypsis <- fastData("madDypsis")

# Convert SpatRasters to GRasters and sf vectors to GVectors:
elev <- fast(madElev)
forest <- fast(madForest2000)
cover <- fast(madCover)
rivers <- fast(madRivers)
dypsis <- fast(madDypsis)

# Distance to forest
distToForest <- distance(forest, unit = "m")
distToForest <- log1p(distToForest) # log(x + 1) of distance
names(distToForest) <- "distToForest"

# "Stack" elevation and forest cover
continuous <- c(elev, distToForest)

# Scale continuous predictors to mean of 0 and sd of 1
continuousScaled <- scale(continuous)
names(continuousScaled) <- c("elevation", "distToForest")

# Project land cover raster
coverProj <- project(cover, continuousScaled)

# Near a river?
riverBuffer <- buffer(rivers, 5000)
nearRiver <- rasterize(riverBuffer, elev, background = 0)
names(nearRiver) <- "nearRiver"
levels(nearRiver) <- data.frame(value = 0:1, label = c("far", "near"))

# Combine continuous/categorical data
covariateRasters <- c(continuousScaled, coverProj, nearRiver)
plot(covariateRasters)

# Extract environmental values at Dypsis locations:
presEnv <- extract(covariateRasters, dypsis, cats = TRUE)
presEnv$presBg <- 1
head(presEnv)

# Extract elevation and forest cover at 2000 background sites:
bgEnv <- spatSample(covariateRasters, size = 3000, values = TRUE, cats = TRUE)
bgEnv <- bgEnv[stats::complete.cases(bgEnv), ]
bgEnv <- bgEnv[1:2000, ]
bgEnv$presBg <- 0
head(bgEnv)

# Combine presence and background data:
env <- rbind(presEnv, bgEnv)

# Calibrate model:
form <- presBg ~ elevation + distToForest +
  I(distToForest^2) + elevation * distToForest +
  madCover + nearRiver

model <- stats::glm(form, data = env, family = stats::binomial)
summary(model)

# Make predictions and map:
prediction <- predict(covariateRasters, model, type = "response")
prediction

# Not a great model!
plot(prediction, main = "Predicted")
plot(dypsis, pch = 1, add = TRUE)

}

.onUnload <- function(lib, pkg) {
	
	# remove GRASS files from cache
	mow("*", verbose = faster("verbose"), ask = FALSE)

	# remove special environment
	rm(.fasterRaster)

	# remove GRASS files
	rgrass::unset.GIS_LOCK()
	rgrass::remove_GISRC()
	rgrass::unlink_.gislock()

}

.onUnattach <- function(lib, pkg) {

	# remove GRASS files from cache
	mow("*", verbose = faster("verbose"), ask = FALSE)

	# remove special environment
	rm(.fasterRaster)

	# remove GRASS files
	rgrass::unset.GIS_LOCK()
	rgrass::remove_GISRC()
	rgrass::unlink_.gislock()

}

#' Test if a GRaster or GVector is 2- or 3-dimensional
#'
#' @description Test whether a `GRaster` or `GVector` is 2- or 3-dimensional.
#'
#' @param x An object that inherits from the `GSpatial` class (i.e., a `GRaster` or `GVector`).
#'
#' @returns Logical.
#'
#' @seealso [topology()]
#'
#' @example man/examples/ex_GRaster_GVector.r
#'
#' @aliases is.2d
#' @rdname is.2d
#' @exportMethod is.2d
methods::setMethod(
	f = "is.2d",
	signature = "GSpatial",
	definition = function(x) topology(x) == "2D"
)

#' @aliases is.3d
#' @rdname is.2d
#' @exportMethod is.3d
methods::setMethod(
	f = "is.3d",
	signature = "GSpatial",
	definition = function(x) topology(x) == "3D"
)

if (grassStarted()) {

# Setup
library(terra)

# Example data: Elevation and land cover rasters
madElev <- fastData("madElev")
madCover <- fastData("madCover")

### match() with an integer raster:

elev <- fast(madElev)

# Cells in elevation raster replaced with index in which they appear
# in the table:
table <- c(10, 20, 30, 40, 50)
elevIndex <- match(elev, table)
elevIndexNeg <- match(elev, table, nomatch = -100)

plot(c(elevIndex, elevIndexNeg))

### Using %in% and %notin% on an integer GRaster:

elev <- fast(madElev)
table <- c(10, 20, 30, 40, 50)

ins <- elev %in% table
notins <- elev %notin% table

plot(c(ins, notins))

### match() with a categorical raster:

cover <- fast(madCover)
cover <- droplevels(cover)
levels(cover)

forestLabels <- c(
   "Sparse broadleaved evergreen/semi-deciduous forest",
   "Broadleaved deciduous forest",
   "Grassland with mosaic forest",
   "Flooded forest"
)

forestClasses <- match(cover, forestLabels)
plot(forestClasses)
levels(forestClasses)

forestNoMatch <- match(cover, forestLabels, nomatch = -100)
plot(forestNoMatch)
levels(forestNoMatch)

### Using %in% and %notin% on a categorical GRaster:

cover <- fast(madCover)
cover <- droplevels(cover)
levels(cover)

forestLabels <- c(
   "Sparse broadleaved evergreen/semi-deciduous forest",
   "Broadleaved deciduous forest",
   "Grassland with mosaic forest",
   "Flooded forest"
)

forest <- cover %in% forestLabels
plot(forest)

notForest <- cover %notin% forestLabels
plot(notForest)

}

#' Create longitude/latitude rasters
#' 
#' @description `longlat()` creates two rasters, one with cell values equal to the longitude of the cell centers, and one with cell values equal to the latitude of the cell centers.
#' 
#' @param x A `GRaster`.
#'
#' @param degrees Logical: If `TRUE` (default), coordinate values of cells will be in degrees. If `FALSE`, and `x` is in a projected coordinate reference system, values will represent coordinates in map units (usually meters). Values will always be in degrees when the coordinate reference system is unprojected (e.g., WGS84, NAD83, etc.).
#' 
#' @returns A `GRaster` stack.
#'
#' @seealso [init()]
#' 
#' @example man/examples/ex_longlat.r
#' 
#' @aliases longlat
#' @rdname longlat
#' @exportMethod longlat
methods::setMethod(
    f = "longlat",
    signature(x = "GRaster"),
    function(x, degrees = TRUE) {

    .locationRestore(x)
    .region(x)
    srcs <- .longlat(x, degrees = degrees)
   makeGRaster(srcs, c("longitude", "latitude"))

    } # EOF
)

#' @noRd
.longlat <- function(x, degrees) {

    if (inherits(x, "GRaster")) {
        src <- sources(x)
    } else {
        src <- x
    }

    srcs <- .makeSourceName(c("longlat_long", "longlat_lat"), "raster")
    if (degrees) {

        rgrass::execGRASS(
            cmd = "r.latlong",
            input = src,
            output = srcs[1L],
            flags = c("l", .quiet(), "overwrite")
        )

        rgrass::execGRASS(
            cmd = "r.latlong",
            input = src,
            output = srcs[2L],
            flags = c(.quiet(), "overwrite")
        )

    } else {
    
        # longitude
        ex <- paste0(srcs[1L], " = col(", src, ")")
        rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))
    
        # latitude
        ex <- paste0(srcs[2L], " = row(", src, ")")
        rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))
    
    }
    srcs

}

#' Re-make vector "category" (cat) values
#' 
#' @description Each geometry in a **GRASS** vector has a "category" number (abbreviated "cat" in output and modules). Geometries can have the same or different numbers, but for functions to work as intended, they often need to have sequential category values, starting at 1, with no skips between integers. This function reconstitutes the category values of a vector in **GRASS** so they being with 1 and have no skips. **This function is mostly of use to developers.**
#'
#' @param x A `GVector` or the [sources()] name of a vector in **GRASS**.
#' 
#' @param gtype Character: Type of vector features in **GRASS** format (i.e., either `point`, `line`, or `area`). See [geomtype()].
#'
#' @param cats `NULL` (default) or `integer` or `character` vector: Category values of the **GRASS** vector. Supplying these can speed the re-assignment of categories. The values of `cats` is *not* used as the new category values. Rather, they will be used to indicate which geometries belong to the same multi-part feature.
#'
#' @returns The [sources()] name of a vector.
#'
#' @seealso [.vIncrementCats()], [.vCats()]
#' 
#' @aliases .vRecat
#' @rdname vRecat
#' @keywords internal
.vRecat <- function(x, gtype, cats = NULL) {

	if (inherits(x, "GVector")) {
		.locationRestore(x)
		src <- sources(x)
	} else {
		src <- x
	}

	if (is.null(cats)) cats <- .vCats(src)

	# srcIn <- src
	# src <- .makeSourceName("v_category", "vector")

	# rgrass::execGRASS(
	# 	cmd = "v.category",
	# 	input = srcIn,
	# 	output = src,
	# 	option = "del",
	# 	cat = -1,
	# 	type = gtype,
	# 	flags = c(.quiet(), "overwrite", "t")
	# )

	# rgrass::execGRASS(
	# 	cmd = "v.category",
	# 	input = srcIn,
	# 	output = src,
	# 	option = "add",
	# 	type = gtype,
	# 	flags = c(.quiet(), "overwrite", "t")
	# )

	srcIn <- src
	src <- .makeSourceName("vRecat_v_reclass", "vector")

	reCats <- as.character(cats)
	.vAttachDatabase(srcIn, table = data.frame(reCat = reCats), cats = cats)

	rgrass::execGRASS(
		cmd = "v.reclass",
		input = srcIn,
		output = src,
		column = "reCat",
		# type = gtype,
		flags = c(.quiet(), "overwrite")
	)

	# # # if (is.null(cats) & is.null(start)) {

	# # # 	cats <- .vCats(src, db = TRUE)
	# # # 	cats <- omnibus::renumSeq(cats)
	# # # 	cats <- data.frame(newcat = cats)
	
	# # # } else if (!is.null(cats)) {
	
	# # # 	cats <- data.frame(newcat = as.integer(cats))
	
	# # # } else {
	
	# # # 	cats <- .vCats(src)
	# # # 	cats <- cats + (start - min(cats))
	
	# # # }

	# # # tf <- tempfile(fileext = ".csv")
	# # # tft <- paste0(tf, "t")
	# # # utils::write.csv(cats, tf, row.names = FALSE)
	# # # keyTableType <- '"Integer"'
	# # # write(keyTableType, tft)

	# # # # import table
	# # # srcTable <- .makeSourceName("db_in_ogr_table", "table")
	# # # rgrass::execGRASS(
	# # # 	cmd = "db.in.ogr",
	# # # 	input = tf,
	# # # 	output = srcTable,
	# # # 	flags = c(.quiet(), "overwrite")
	# # # )

	# # # # detach existing table
	# # # .vDetachDatabase(src)

	# # # # attach table
	# # # rgrass::execGRASS(
	# # # 	cmd = "v.db.connect",
	# # # 	map = src,
	# # # 	table = srcTable,
	# # # 	key = "newcat",
	# # # 	flags = c(.quiet(), "o")
	# # # )

	# # # # reclass
	# # # srcIn <- src
	# # # src <- .makeSourceName("v_reclass", "vector")
	# # # rgrass::execGRASS(
	# # # 	cmd = "v.reclass",
	# # # 	input = srcIn,
	# # # 	output = src,
	# # # 	layer = srcTable,
	# # # 	column = "newcat",
	# # # 	type = "centroid",
	# # # 	flags = c(.quiet(), "overwrite")
	# # # )

	src

}

#' Create lines connecting nearest features of two GVectors
#'
#' @description `connectors()` creates a lines `GVector` which represent the shortest (Great Circle) paths between each feature of one `GVector` and the nearest feature of another `GVector`.
#'
#' @param x,y `GVector`s.
#' @param minDist,maxDist Either `NULL` (default) or numeric values: Ignore features separated by less than or greater than these distances.
#'
#' @returns A `GVector` with a data table that has the length of each connecting line in meters.
#' 
#' @seealso **GRASS** manual for tool `v.distance` (see `grassHelp("v.distance")`).
#'
#' @example man/examples/ex_connectors.r
#'
#' @aliases connectors
#' @rdname connectors
#' @exportMethod connectors
methods::setMethod(
	"connectors",
	signature(x = "GVector", y = "GVector"),
	function(x, y, minDist = NULL, maxDist = NULL) {

	if (!is.null(minDist) && minDist < 0) stop("Argument `minDist` must be positive or NULL.")
	if (!is.null(maxDist) && maxDist < 0) stop("Argument `maxDist` must be positive or NULL.")
	if ((!is.null(maxDist) & !is.null(maxDist)) && (minDist > maxDist)) stop("Argument `minDist` is greater than `maxDist`.")

	compareGeom(x, y)
	.locationRestore(x)

	if (is.null(minDist)) minDist <- -1
	if (is.null(maxDist)) maxDist <- -1

	if (!.vHasDatabase(x)) {

		cats <- .vCats(x)
		# db <- data.frame(TEMPTEMP_minDist_meters = -1)
		db <- data.frame(cat = cats, TEMPTEMP_minDist_meters = -1)
		.vAttachDatabase(x, table = db, replace = TRUE, cats = cats)
		
	}

	# create connectors
	src <- .makeSourceName("connectors", "vector")

	rgrass::execGRASS(
		cmd = "v.distance",
		from = sources(x),
		to = sources(y),
		output = src,
		upload = "dist",
		dmin = minDist,
		dmax = maxDist,
		column = "TEMPTEMP_minDist_meters",
		flags = c(.quiet(), "overwrite")
	)
	
	# # remove column that was added to x
	# rgrass::execGRASS("v.db.dropcolumn", map = sources(x), columns = "TEMPTEMP_minDist_meters", flags = .quiet())

	table <- .vAsDataTable(x)
	table <- table[ , "TEMPTEMP_minDist_meters"]
	names(table) <- "length_meters"
	makeGVector(src, table = table)

	} # EOF
)

#' Data type of a raster
#'
#' @description In **fasterRaster**, rasters can have three data types: "factor" (categorical rasters), "integer" (integers), "float" (floating point values, accurate to the 6th to 9th decimal places), and "double" (double-precision values, accurate to the 15th to 17th decimal places). The type of raster can be checked with:
#'
#' * `is.factor()`: The raster will have integer values and categories matched to the integers (see levels()).
#' * `is.int()`: Are values integers? Note that `is.int()` will return `FALSE` for categorical rasters, even though cell values are technically integers.
#' * `is.cell()`: Are values integers (`TRUE` for `integer` and categorical rasters).
#' * `is.float()`: Are values floating-point precision?
#' * `is.doub()`: Are values double-floating point precision?
#'
#' @param x A `GRaster`.
#' 
#' @returns Logical.
#' 
#' @seealso [datatype()], [terra::datatype()], [as.int()], [as.float()], [as.doub()], [is.factor()], `vignette("GRasters", package = "fasterRaster")`
#'
#' @example man/examples/ex_GRaster.r
#' 
#' @aliases is.int
#' @rdname is.int
#' @exportMethod is.int
methods::setMethod(
    f = "is.int",
    signature = c(x = "GRaster"),
    function(x) datatype(x, "GRASS") == "CELL" & nlevels(x) == 0L
)

is.int <- function(x) UseMethod("is.int", x)

#' @aliases is.cell
#' @rdname is.int
#' @exportMethod is.cell
methods::setMethod(
    f = "is.cell",
    signature = c(x = "GRaster"),
    function(x) datatype(x, "GRASS") == "CELL"
)

#' @aliases is.float
#' @rdname is.int
#' @exportMethod is.float
methods::setMethod(
    f = "is.float",
    signature = c(x = "GRaster"),
    function(x) datatype(x, "GRASS") == "FCELL"
)

#' @aliases is.doub
#' @rdname is.int
#' @exportMethod is.doub
methods::setMethod(
    f = "is.doub",
    signature = c(x = "GRaster"),
    function(x) datatype(x, "GRASS") == "DCELL"
)

#' @aliases is.factor
#' @rdname is.int
#' @exportMethod is.factor
methods::setMethod(
    f = "is.factor",
    signature = c(x = "GRaster"),
    function(x) nlevels(x) > 0L & datatype(x, "GRASS") == "CELL"
)

#' Minimum convex hull around a spatial vector
#'
#' Create a minimum convex hull around a spatial vector.
#'
#' @param x A `GVector`.
#' @param by Character: If `""` (default), then a convex hull is created for all geometries together. Otherwise, this is the name of a field in the vector. Hulls will be created for each set of geometries with the same value in this column.
#'
#' @return A `GVector`.
#'
#' @seealso \code{link[terra]{convHull}}, \code{link[sf]{st_convex_hull}}, tool `v.hull` in **GRASS** (see `grassHelp("v.hull")`)
#'
#' @example man/examples/ex_convHull.r
#'
#' @aliases convHull
#' @rdname convHull
#' @exportMethod convHull
methods::setMethod(
	f = "convHull",
	signature = c(x = "GVector"),
	definition = function(x, by = "") {

	if (by == "") {
		
		src <- .makeSourceName("v_hull", "vector")
	
		rgrass::execGRASS(
			cmd = "v.hull",
			input = sources(x),
			output = src,
			flags = c(.quiet(), "overwrite")
		)

		out <- makeGVector(src)
		
	} else {
	
		byCol <- which(names(x) == by)
		bys <- as.data.table(x)[[by]]
		uniques <- unique(bys)
	
		n <- length(uniques)

		srcs <- .makeSourceName("v_hull", "vector", n)
		vects <- list()
		for (i in seq_len(n)) {

			uniq <- uniques[i]
			selected <- which(bys == uniq)

			xx <- x[selected]

			if (ngeom(xx) < 3L) {

				warnings("Skipping ", uniq, " in column ", by, " because it has <3 locations.")

			} else {
					
				# select
				args <- list(
					cmd = "v.hull",
					input = sources(xx),
					output = srcs[i],
					flags = c(.quiet(), "overwrite"),
					intern = TRUE
				)
				
				do.call(rgrass::execGRASS, args=args)
				vects[[i]] <- makeGVector(srcs[i])

			}
				
		} # next set
		
		# remove any NULL polygons
		vects <- Filter(Negate(is.null), vects)
		
		if (length(vects) == 0L) {
			out <- NULL
		} else {

			# concatenate
			args <- list(x = vects[[1L]])
			if (length(vects) > 1L) args <- c(args, list(vects[2:length(vects)]))
			out <- do.call("c", args = args)

		}
	
	}
	out
		
	} # EOF
)

#' Create raster representing one or more random walks
#'
#' This function creates a raster where the cell values represent the number of times one or more random "walkers" traverse the cell. If you simulate multiple random walkers, you can do computation in parallel, which can be controlled by allowing **fasterRaster** to use multiple cores and more memory using the "cores" and "memory" arguments in the [faster()] function.
#'
#' This function needs the **GRASS** addon `r.random.walk`. If it is not installed, it will try to install it.#'
#' @param x A `GRaster` to serve as a template.
#'
#' @param n Numeric: Number of walkers. Default  is 1.
#'
#' @param steps Numeric: Number of steps taken by each walker. Default is 100000.
#'
#' @param directions Either 4 or 8: Directions in which a walker can turn at any point. If 4, then walks are confined to north/south/east/west directions (Rook's case). If 8, then the cardinal and subcardinal directions are allowed (Queen's case).
#'
#' @param avoid Logical: If `FALSE` (default), then walkers can traverse their own walks. If `TRUE`, walkers avoid their own trails. A self-avoiding random walk can take much longer to compute.
#'
#' @param sameStart Logical: If `FALSE` (default), walkers can begin anywhere. If `TRUE`, then walkers start from the same place.
#'
#' @param seed Integer or `NULL` (default): If `NULL`, then a random seed is generated by the function. If numeric, results will be deterministic. In either case, the value will be rounded to the nearest integer.
#'
#' @param check Logical: If `TRUE` (default), function will check to see if the addon `r.random.walk` has been installed. If it has not, it will attempt to install it.
#'
#' @returns A `GRaster` with cell values representing the number of times one or more walkers traversed the cell.
#'
#' @seealso [rNormRast()], [rUnifRast()], [rSpatialDepRast()], [fractalRast()]

#' @example man/examples/ex_randRast.r
#'
#' @aliases rWalkRast
#' @rdname rWalkRast
#' @exportMethod rWalkRast
methods::setMethod(
	f = "rWalkRast",
	signature = c(x = "GRaster"),
	function(x, n = 1, steps = 100000, directions = 8, avoid = FALSE, sameStart = FALSE, seed = NULL, check = TRUE) {

	if (!(directions %in% c(4, 8))) stop("The `directions` argument must be 4 or 8.")
	directions <- as.character(directions)
	
	if (!is.null(seed)) seed <- round(seed)

	if (check) .addons("r.random.walk")

    .locationRestore(x)
    .region(x)

	src <- .makeSourceName("rRandWalk_r_random_walk", "raster")
	args <- list(
		cmd = "r.random.walk",
		output = src,
		steps = steps,
		directions = directions,
		memory = faster("memory"),
		nprocs = faster("cores"),
		nwalkers = n,
		flags = c(.quiet(), "overwrite")
	)

	if (!is.null(seed)) {
		args <- c(args, seed = seed)
	} else if (is.null(seed)) {
		args$flags <- c(args$flags, "s")
	}
	if (avoid) args$flags <- c(args$flags, "a")
	if (sameStart) args$flags <- c(args$flags, "t")

	do.call(rgrass::execGRASS, args = args)

	# srcIn <- src
	# src <- .makeSourceName("rRandWalk_r_mapcalc", "raster")
	# ex <- paste0(src, " = int(", srcIn, ")")
	# rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

	makeGRaster(src, "randomWalk")

	} # EOF
)

if (grassStarted()) {

# Setup
library(terra)

# Elevation and points
madElev <- fastData("madElev")
madDypsis <- fastData("madDypsis")

# Convert to fasterRaster formats:
elev <- fast(madElev)
dypsis <- fast(madDypsis)

# Aggregate cells of the raster so they are bigger
elevAgg <- aggregate(elev, 32)

# Remove all but one or two points per cell
thin1 <- thinPoints(dypsis, elevAgg, n = 1)
thin2 <- thinPoints(dypsis, elevAgg, n = 2)

# Plot
plot(elevAgg)
plot(dypsis, add = TRUE)
plot(thin2, col = "yellow", add = TRUE)
plot(thin1, col = "red", add = TRUE)
legend(
   "bottomright",
   legend = c("In original & thin 1 & 2",
     "In just thin 1 & 2", "In just thin 1"),
   pch = 16,
   col = c("black", "yellow", "red"),
   bg = "white",
   xpd = NA
)

}

#' Randomly sample cells from a GRaster
#'
#' @description `sampleRast()` randomly samples cells from non-`NA` cells of a raster. The output will be a raster with selected non-`NA` cells, and all other cells set to `NA`. To generate random points, see [spatSample()].
#'
#' @param x A `GRaster`.
#'
#' @param size Numeric: Number of cells or proportion of cells to select.
#' 
#' @param prop Logical: If `TRUE`, the value of `size` will be interpreted as a proportion of cells. The default is `FALSE` (`size` is interpreted as the number of cells to select).
#'
#' @param maskvalues Numeric vector, including `NA`, or `NULL` (default): Values in the raster to select from. All others will be ignored. If this is `NULL`, then only non-`NA` cells will be selected for retention.
#'
#' @param updatevalue Numeric or `NULL` (default): Value to assign to masked cells. If `NULL`, then the values in the input raster are retained.
#' 
#' @param test Logical: If `TRUE`, and `size` is greater than the number of non-`NA` cells in `x`, then fail. Testing this can take a long time for large rasters. The default is `FALSE`.
#'
#' @param seed `NULL` (default) or numeric: If `NULL`, then a random seed will be generated for the random number generator. Otherwise a seed can be provided.
#' 
#' @returns A `GRaster`.
#' 
#' @seealso [spatSample()]; [terra::spatSample()], tool `r.random` in **GRASS**
#' 
#' @example man/examples/ex_sampleRast_spatSample.r
#' 
#' @aliases sampleRast
#' @rdname sampleRast
#' @exportMethod sampleRast
methods::setMethod(
    f = "sampleRast",
    signature = c(x = "GRaster"),
    function(
        x,
        size,
        prop = FALSE,
        maskvalues = NA,
        updatevalue = NULL,
        test = FALSE,
        seed = NULL
    ) {

    if (size <= 0) stop("Cannot select this number of cells.")
    if (prop && size > 1) stop("Cannot select more than 100% of cells.")
    if (!prop && size > ncell(x)) stop("Cannot select more than the total number of cells.")
    if (test) {
        if (any(size > nacell(x))) stop("At least one raster has too few non-NA cells.")
    }

    .locationRestore(x)
    .region(x)

    if (!prop) {
        # npoints <- as.character(size)
        npoints <- formatC(size, format = "f", drop0trailing = TRUE) # overcome scientific notation

    } else {
        npoints <- paste0(100 * npoints, "%")
    }

    nLayers <- nlyr(x)
    srcs <- .makeSourceName("sampleRast_r_random", "raster", nLayers)
    for (i in seq_len(nLayers)) {
    
        args <- list(
            cmd = "r.random",
            input = sources(x)[i],
            cover = sources(x)[i],
            npoints = npoints,
            raster = srcs[i],
            flags = c(.quiet(), "overwrite")
        )

        if (is.null(seed)) {
            args$flags <- c(args$flags, "s")
        } else {
            args$seed <- seed
        }
        # if (selectNA) args$flags <- c(args$flags, "size")

        # custom mask values
        if (length(maskvalues) > 1L || !is.na(maskvalues)) {

            maskSrc <- .makeSourceName("mask", "raster")

            if (anyNA(maskvalues)) {
                maskValuesHaveNAs <- TRUE
                maskvalues <- maskvalues[!is.na(maskvalues)]
            } else {
                maskValuesHaveNAs <- FALSE
            }

            maskvalues <- as.character(maskvalues)
            maskvalues <- paste(paste0(sources(x)[i], " == "), maskvalues)
            if (maskValuesHaveNAs) maskvalues <- c(maskvalues, paste0("isnull(", sources(mask), ")"))

            maskvalues <- paste(maskvalues, collapse = " | ")

            ex <- paste0(maskSrc, " = if(", maskvalues, ", 1, null())")
            rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"), intern = TRUE)

        } else {
            maskSrc <- sources(x)[i]
        }
        args$input <- maskSrc

        do.call(rgrass::execGRASS, args = args)

    } # next raster

    ### change masked values
    if (!is.null(updatevalue)) {

        srcsUpdate <- .makeSourceName("random", "raster", nLayers)
        for (i in seq_len(nLayers)) {

            ex <- paste0(srcsUpdate[i], " = if(!isnull(", srcs[i], "), ", updatevalue, ", null())")
            rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"), intern = TRUE)

        } # next raster
        out <- makeGRaster(srcsUpdate, names(x))

    } else {
        levels <- cats(x)
        ac <- activeCats(x)
        out <- makeGRaster(srcs, names(x), levels = levels, ac = ac)
    }
    out

    } # EOF
)

if (grassStarted()) {

# Setup
library(terra)

# Climate rasters:
madChelsa <- fastData("madChelsa")

# Convert to GRasters:
chelsa <- fast(madChelsa)

### Center and scale rasters
# Scale with using sample SD:
chScaled <- scale(chelsa)
chScaled

# Scale with using population SD:
chScaledPop <- scalepop(chelsa)
chScaledPop

# Means are very close to 0 and SDs to 1:
global(chScaled, c("mean", "sd", "min", "max"))
global(chScaledPop, c("mean", "sd", "min", "max"))

# Get original means and sd's:
centers <- attributes(chScaled)$center
scales <- attributes(chScaled)$scale
centers
scales

### Unscale rasters:
chUnscaled <- unscale(chScaled, center = centers, scale = scales)

# Means and SD are returned to original values:
global(chUnscaled, c("mean", "sd", "min", "max")) # unscaled
global(chelsa, c("mean", "sd", "min", "max")) # original

}

if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Elevation raster, outline of a part of Madagascar, and rivers vector:
madElev <- fastData("madElev")
madCoast0 <- fastData("madCoast0")
madRivers <- fastData("madRivers")

# Convert to GRaster and GVectors:
elev <- fast(madElev)
coast <- fast(madCoast0)
rivers <- fast(madRivers)

# For this example, we will first crop to a small extent.
river <- rivers[1]
elevCrop <- crop(elev, river)
elevPoints <- as.points(elevCrop)
elevPoints

plot(elevCrop)
plot(elevPoints, pch = '.', add = TRUE)

# Extract points from vectors:
coastPoints <- as.points(coast)
riversPoints <- as.points(rivers)

plot(coast)
plot(coastPoints, add = TRUE)

plot(rivers, col = "blue", add = TRUE)
plot(riversPoints, col = "blue", add = TRUE)

}

#' @name madRivers
#'
#' @title Major rivers in a selected portion of Madagascar
#'
#' @description Spatial lines object of major rivers in a portion of Madagascar.
#'
#' @docType data
#'
#' @usage data(madRivers)
#'
#' @format An object of class `sf`.
#'
#' @keywords Madagascar
#'
#' @source [DIVA-GIS](https://diva-gis.org/)
#'
#' @example man/examples/ex_fastData.r
#'
NULL

if (grassStarted()) {

# Setup
library(sf)
library(terra)

### Points, line, and polygon centroids

# Point centroids:
madDypsis <- fastData("madDypsis")
dypsis <- fast(madDypsis)

dypMean <- centroids(dypsis)
dypMedian <- centroids(dypsis, method = "median")
dypPMedian <- centroids(dypsis, method = "pmedian")

plot(dypsis)
plot(dypMean, col = "red", add = TRUE)
plot(dypMedian, col = "green", pch = 2, add = TRUE)
plot(dypPMedian, col = "blue", pch = 3, add = TRUE)
legend("bottomright",
   legend = c("mean", "median", "pmedian"),
   col = c("red", "green", "blue"),
   pch = c(16, 2, 3),
   xpd = NA
)

# Line centroids:
madRivers <- fastData("madRivers")
rivers <- fast(madRivers)

riversMid <- centroids(rivers)
riversMean <- centroids(rivers, method = "mean")
riversMedian <- centroids(rivers, method = "median")

plot(rivers)
plot(riversMid, col = "red", add = TRUE)
plot(riversMean, col = "green", pch = 2, add = TRUE)
plot(riversMedian, col = "blue", pch = 3, add = TRUE)
legend("bottomright",
   legend = c("mid", "mean", "median"),
   col = c("red", "green", "blue"),
   pch = c(16, 2, 3),
   xpd = NA
)

# Polygon centroids:
madCoast4 <- fastData("madCoast4")
coast4 <- fast(madCoast4)

coastMean <- centroids(coast4)
coastMedian <- centroids(coast4, method = "median")
coastBMedian <- centroids(coast4, method = "bmedian")

plot(coast4)
plot(coastMean, col = "red", add = TRUE)
plot(coastMedian, col = "green", pch = 2, add = TRUE)
plot(coastBMedian, col = "blue", pch = 3, add = TRUE)
legend("bottomright",
   legend = c("mean", "median", "bmedian"),
   col = c("red", "green", "blue"),
   pch = c(16, 2, 1),
   xpd = NA
)

### Centroids of integer GRaster "clumps"

# Load elevation raster
madElev <- fastData("madElev")
elev <- fast(madElev)

# Create clumps of similarly-valued cells
clumps <- clump(elev, minDiff = 0.01, minClumpSize = 1000)

# Centroids:
clumpCents <- centroids(clumps)
clumpCents

plot(clumps)
plot(clumpCents, add = TRUE)


}

#' Fill NA cells in a raster using interpolation
#'
#' @description This function uses splines to fill `NA` cells in a raster based on the values of nearby cells. Depending on the method used, not all `NA` cells can be filled.
#'
#' @param x A `GRaster`.
#'
#' @param lambda Either `NULL` (default), or a numeric value > 0: If `NULL`, then the function will use leave-one-out crossvalidation to find the optimal value.
#'
#' @param method Character: Type of spline, either "`bilinear`" (default), "`bicubic`", or "`RST`" (regularized splines with tension). Partial matching is used and case is ignored.
#'
#' **Note**: The RST method will often display warnings, but these can be ignored.
#'
#' @param min,max Numeric: Lowest and highest values allowed in the interpolated values. Values outside these bounds will be truncated to the minimum/maximum value(s) allowed. The default imposes no constraints. For multi-layered rasters, you can supply a single value for `min` and/or `max`, or multiple values (one per layer). Values will be recycled if there are fewer than one or them per layer in the raster.
#'
#' @param cells Integer or numeric integer: Number of cells away from the non-`NA` cells to fill. For example, if `cells = 2`, then only cells within a 2-cell buffer of non-`NA` cells will be filled. The default is `Inf` (fill all possible cells--some methods may not be able to do this, depending on the configuration of the raster).
#'
#' @returns A `GRaster`.
#'
#' @example man/examples/ex_fillNAs.r
#'
#' @seealso [terra::interpNear()], **GRASS** tool `r.fillnulls` (see `grassHelp("r.fillnulls")`)
#'
#' @aliases fillNAs
#' @rdname fillNAs
#' @exportMethod fillNAs
methods::setMethod(
	f = "fillNAs",
	signature = c(x = "GRaster"),
	function(x, lambda = NULL, method = "bilinear", min = -Inf, max = Inf, cells = Inf) {
	
	method <- tolower(method)
	method <- omnibus::pmatchSafe(method, c("bilinear", "bicubic", "rst"), n = 1L)

	.locationRestore(x)
	.region(x)

	nLayers <- nlyr(x)

	if (length(min) < nLayers) min <- rep(min, length.out = nLayers)
	if (length(max) < nLayers) max <- rep(max, length.out = nLayers)

	mm <- minmax(x)
	if (any(min > mm["min", ])) warning("The `min` value is greater than the actual minimum value in at least one raster layer.\n  Observed values will be truncated.")

	if (any(max > mm["max", ])) warning("The `max` value is less than the actual maximum value in at least one raster layer.\n  Observed values will be truncated.")

	srcs <- .makeSourceName("r_fillnulls", "raster", nLayers)

	for (i in seq_len(nLayers)) {
	
		# mask
		if (!is.infinite(cells)) {
		
			maskSrc <- .makeSourceName("r_grow", "raster")

			rgrass::execGRASS(
				cmd = "r.grow",
				input = sources(x)[i],
				output = maskSrc,
				radius = cells + 0.01,
				metric = "euclidean",
				old = 1,
				new = 1,
				flags = c(.quiet(), "overwrite")
			)

			rgrass::execGRASS(
				cmd = "r.mask",
				raster = maskSrc,
				flags = c(.quiet(), "overwrite")
			)
		
		}

		args <- list(
			cmd = "r.fillnulls",
			input = sources(x)[i],
			output = srcs[i],
			method = method,
			memory = faster("memory"),
			flags = c(.quiet(), "overwrite"),
			Sys_show.output.on.console = FALSE
		)
	
		do.call(rgrass::execGRASS, args = args)

		# unmask
		if (!is.infinite(cells)) .removeMask()

		# constrain interpolation
		if (!is.infinite(min[i]) | !is.infinite(max[i])) {

			srcIn <- srcs[i]
			srcs[i] <- .makeSourceName("r_mapcalc", "raster")

			if (!is.infinite(min[i]) & is.infinite(max[i])) {
		
				ex <- paste0(srcs[i], " = if(", srcIn, " < ", min[i], ", ", min[i], ", ", srcIn, ")")

			} else if (is.infinite(min[i]) & !is.infinite(max[i])) {

				ex <- paste0(srcs[i], " = if(", srcIn, " > ", max[i], ", ", max[i], ", ", srcIn, ")")

			} else if (!is.infinite(min[i]) & !is.infinite(max[i])) {
			
				ex <- paste0(srcs[i], " = if(", srcIn, " < ", min[i], ", ", min[i], ", if (", srcIn, " > ", max[i], ", ", max[i], ", ", srcIn, "))")

			}

			rgrass::execGRASS(
				cmd = "r.mapcalc",
				expression = ex,
				flags = c(.quiet(), "overwrite")
			)

			.rm(srcIn, type = "raster", warn = FALSE)
		
		}

	}
	makeGRaster(srcs, names(x))

	} # EOF
)

#' Create fractal raster
#'
#' @description `fractalRast()` creates a raster with a fractal pattern.
#'
#' @param x A `GRaster`. The output will have the same extent and dimensions as this raster.
#'
#' @param n A numeric integer: Number of rasters to generate. 
#'
#' @param mu,sigma Numeric: Mean and sample standard deviation of output.
#'
#' @param dimension Numeric: Fractal dimension. Must be between 2 and 3.
#'
#' @returns A `GRaster`.
#'
#' @example man/examples/ex_randRast.r
#' 
#' @seealso [rSpatialDepRast()], [rNormRast()], [rUnifRast()], [rWalkRast()], **GRASS** manual page for tool `r.surf.fractal` (see `grassHelp("r.surf.fractal")`)
#'
#' @aliases fractalRast
#' @rdname fractalRast
#' @exportMethod fractalRast
methods::setMethod(
    f = "fractalRast",
    signature = c(x = "GRaster"),
    function(x,
        n = 1,
        mu = 0,
        sigma = 1,
        dimension = 2.05) {

    if (any(dimension <= 2) | any(dimension >= 3)) stop("Argument `dimension` must be in the range (2, 3).")

    mu <- rep(mu, length.out = n)
    sigma <- rep(sigma, length.out = n)
    dimension <- rep(dimension, length.out = n)

    .locationRestore(x)
    .region(x)

    srcFracts <- .makeSourceName("fractal", "raster", n) # base fractal rasters
    srcs <- .makeSourceName("fractalScaled", "raster", n) # rescaled fractal rasters

    for (i in seq_len(n)) {

        rgrass::execGRASS(
            cmd = "r.surf.fractal",
            output = srcFracts[i],
            dimension = dimension[i],
            flags = c(.quiet(), "overwrite")
        )

        xmu <- .global(srcFracts[i], fun = "mean")
        xsigma <- .global(srcFracts[i], fun = "sdpop")

        ex <- paste0(srcs[i], " = ", mu[i], " + ", sigma[i], " * (", srcFracts[i], " - ", xmu, ") / ", xsigma)

        rgrass::execGRASS(
            "r.mapcalc",
            expression = ex,
            flags = c(.quiet(), "overwrite")
        )

    } # next raster
    
    on.exit(.rm(srcFracts, type = "raster", warn = FALSE), add = TRUE)
    makeGRaster(srcs, "fractal")

    } # EOF
)

#' Subset a GRaster layer, or return values from a column of a GVector's table
#'
#' @description The `dollar` notation can be used to get a single layer of a multi-layer `GRaster` or the values of a column from a `GVector`'s data table.
#'
#' @param x A `GRaster` or `GVector`.
#' @param name Character: The name of a `GRaster` or of a column of a `GVector`'s data table. Names of rasters and vectors can be found using [names()].
#' @returns A `GRaster` or vector of the same type as the `GVector`'s column.
#'
#' @example man/examples/ex_GRaster_GVector_subset_assign.r
#' 
#' @seealso [subset()], \code{\link[fasterRaster]{[}}, \code{\link[fasterRaster]{[[}}
#' 
#' @name $
#' @aliases $,GRaster-method
#' @docType methods
#' @rdname subset_dollar
#' @exportMethod $
methods::setMethod(
    f = "$",
    signature = c(x = "GRaster"),
    function(x, name) {
	
	# test indices
	i <- .layerIndex(name, x, recycle = TRUE)
	mm <- minmax(x)
	
	new(
		"GRaster",
		location = .location(x),
		mapset = .mapset(x),
		workDir = faster("workDir"),
		crs = crs(x),
		projection = .projection(x),
		nLayers = length(i),
		dimensions = dim3d(x),
		topology = topology(x),
		extent = as.vector(ext(x)),
		zextent = zext(x),
		sources = sources(x)[i],
		names = names(x)[i],
		datatypeGRASS = datatype(x, "GRASS")[i],
		resolution = res3d(x),
		minVal = mm["min", i],
		maxVal = mm["max", i],
		activeCat = x@activeCat[i],
		levels = cats(x)[i]
	)
	
	} # EOF
)

#' @name $
#' @aliases $,GVector-method
#' @docType methods
#' @rdname subset_dollar
#' @exportMethod $
methods::setMethod(
	"$",
	signature = c(x = "GVector"),
	function(x, name) x@table[[name]]
)

# NB SQL WHERE queries can be up to 30000 character long... longer fails in GRASS

.restore <- NULL

# bad form:
# require(data.table)
# require(sf)
# require(terra)

# creates conflicts with terra::shift and data.table::shift
# #' @import terra
#' @importFrom data.table as.data.table
# #' @import terra

#' @import sf
#' @rawNamespace import(data.table, except = shift)
#' @rawNamespace import(terra, except = shift)
# #' @importFrom data.table as.data.table
# #' @importFrom methods new
# #' @importFrom methods setMethod
# #' @importFrom methods setValidity
# #' @importFrom methods validObject

### data.table
.datatable.aware <- TRUE

if (getRversion() >= "2.15.1") {
    utils::globalVariables(c(".", ".I", ".N", ".SD"), utils::packageName())
    utils::globalVariables(c("..ac", "..cols", "..i", "..j", "j", "i", "..fun"), utils::packageName())
}

### GENERICS
#' @importFrom data.table :=
# if (!isGeneric("[")) methods::setGeneric(name = "[", def = function(x, ...) standardGeneric("["))
# if (!isGeneric("[<-")) methods::setGeneric(name = "[<-", def = function(x, ...) standardGeneric("[<-"))

# $ is already generic
# $<- is already generic
# [ is already generic
# [<- is already generic
# [[ is already generic
# [[<- is already generic

methods::setGeneric(name = "%in%", package = "terra")
methods::setGeneric(name = "%notin%", def = function(x, ...) package = "omnibus")
methods::setGeneric(name = ".copyGSpatial", def = function(x, ...) standardGeneric(".copyGSpatial"))
methods::setGeneric(name = ".exists", def = function(x, ...) standardGeneric(".exists"))
methods::setGeneric(name = ".location", def = function(x, ...) standardGeneric(".location"))
methods::setGeneric(name = ".locationCreate", def = function(x, ...) standardGeneric(".locationCreate"))
methods::setGeneric(name = ".locationFind", def = function(x, ...) standardGeneric(".locationFind"))
methods::setGeneric(name = ".locationRestore", def = function(x, ...) standardGeneric(".locationRestore"))
methods::setGeneric(name = ".mapset", def = function(x, ...) standardGeneric(".mapset"))
methods::setGeneric(name = ".region", def = function(x, ...) standardGeneric(".region"))
methods::setGeneric(name = ".regionDim", def = function(x, ...) standardGeneric(".regionDim"))
methods::setGeneric(name = ".regionExt", def = function(x, ...) standardGeneric(".regionExt"))
methods::setGeneric(name = ".regionRes", def = function(x, ...) standardGeneric(".regionRes"))

# methods::setGeneric(name = "activeCat", package = "terra")
methods::setGeneric(name = "activeCat", def = function(x, ...) standardGeneric("activeCat"))
methods::setGeneric(name = "activeCats", def = function(x, ...) standardGeneric("activeCats"))
methods::setGeneric(name = "activeCat<-", package = "terra")
methods::setGeneric(name = "add<-", package = "terra")
methods::setGeneric(name = "addCats", package = "terra")
methods::setGeneric(name = "addCats<-", def = function(x, ..., value) standardGeneric("addCats<-"))
methods::setGeneric(name = "addTable<-", def = function(x, ..., value) standardGeneric("addTable<-"))
# methods::setGeneric(name = "andAndAnd", def = function(e1, e2) standardGeneric("andAndAnd"))
methods::setGeneric(name = "allNA", package = "terra")
# anyNA: generic is implicit
methods::setGeneric(name = "app", package = "terra")
# methods::setGeneric(name = "appFuns", def = function(x, ...) standardGeneric("appFuns"))
methods::setGeneric(name = "appCheck", def = function(x, fun, ...) standardGeneric("appCheck"))
methods::setGeneric(name = "as.contour", package = "terra")
methods::setGeneric("as.data.frame")
methods::setGeneric(name = "as.data.table", package = "data.table")
# methods::setGeneric(name = "as.data.table", def = function(x, ...) standardGeneric("as.data.table"))
methods::setGeneric(name = "as.doub", def = function(x, ...) standardGeneric("as.doub"))
methods::setGeneric(name = "as.float", def = function(x, ...) standardGeneric("as.float"))
# as.integer() is primitive
methods::setGeneric(name = "as.int", package = "terra")
setGeneric("as.lines", function(x,...) standardGeneric("as.lines"))
setGeneric("as.points", function(x,...) standardGeneric("as.points"))
setGeneric("as.polygons", function(x,...) standardGeneric("as.polygons"))
setGeneric("aggregate", function(x, ...) standardGeneric("aggregate"), package = "stats")

methods::setGeneric(name = "barplot", package = "terra")
methods::setGeneric(name = "bioclims", def = function(ppt, ...) standardGeneric("bioclims"))
methods::setGeneric(name = "bottom", def = function(x, ...) standardGeneric("bottom"))
methods::setGeneric(name = "breakPolys", def = function(x, ...) standardGeneric("breakPolys"))
methods::setGeneric(name = "buffer", package = "terra")

# c() is primitive
methods::setGeneric(name = "categories", package = "terra")
methods::setGeneric(name = "catNames", def = function(x, ...) standardGeneric("catNames"))
methods::setGeneric(name = "cats", package = "terra")
# methods::setGeneric(name = "cbind")
# from https://stackoverflow.com/questions/27886535/proper-way-to-use-cbind-rbind-with-s4-classes-in-package
# methods::getGeneric("cbind")
# methods::setGeneric(name = "cbind", signature = "...")
methods::setGeneric(name = "colbind", def = function(x, ...) standardGeneric("colbind"))
methods::setGeneric(name = "cellSize", package = "terra")
methods::setGeneric(name = "centroids", package = "terra")
methods::setGeneric(name = "classify", package = "terra")
methods::setGeneric(name = "clump", def = function(x, ...) standardGeneric("clump"))
methods::setGeneric(name = "clusterPoints", def = function(x, ...) standardGeneric("clusterPoints"))
methods::setGeneric(name = "compareGeom", package = "terra")
methods::setGeneric("complete.cases", function(...) standardGeneric("complete.cases"), package = "stats")
methods::setGeneric(name = "compositeRGB", def = function(r, ...) standardGeneric("compositeRGB"))
methods::setGeneric(name = "combineLevels", def = function(x, ...) standardGeneric("combineLevels"))
methods::setGeneric(name = "concats", def = function(x, ...) standardGeneric("concats"))
methods::setGeneric(name = "connectors", def = function(x, y, ...) standardGeneric("connectors"))
methods::setGeneric(name = "coordRef", def = function(x, ...) standardGeneric("coordRef"))
methods::setGeneric(name = "convHull", package = "terra")
methods::setGeneric(name = "count", def = function(x, ...) standardGeneric("count"))
methods::setGeneric(name = "crds", package = "terra")
methods::setGeneric(name = "crs", package = "terra")
methods::setGeneric(name = "crop", package = "terra")

# dim() is primitive
methods::setGeneric(name = "datatype", def = function(x, ...) standardGeneric("datatype"))
methods::setGeneric(name = "delaunay", package = "terra")
methods::setGeneric(name = "denoise", def = function(x, ...) standardGeneric("denoise"))
methods::setGeneric(name = "dim3d", def = function(x, ...) standardGeneric("dim3d"))
methods::setGeneric(name = "disagg", package = "terra")
methods::setGeneric(name = "distance", def = function(x, y, ...) standardGeneric("distance"))
methods::setGeneric(name = "droplevels", package = "terra")
methods::setGeneric(name = "dropRows", def = function(x, ...) standardGeneric("dropRows"))
methods::setGeneric(name = "dropTable", def = function(x, ...) standardGeneric("dropTable"))

methods::setGeneric(name = "E", def = function(x, ...) standardGeneric("E"))
# methods::setGeneric(name = "elide", def = function(x, ...) standardGeneric("elide"))
methods::setGeneric(name = "erase", package = "terra")
methods::setGeneric(name = "ext", package = "terra")
methods::setGeneric(name = "extend", package = "terra")
methods::setGeneric(name = "expanse", package = "terra")
methods::setGeneric(name = "extract", package = "terra")

methods::setGeneric(name = "fast", def = function(x, ...) standardGeneric("fast"))
methods::setGeneric(name = "fillHoles", package = "terra")
methods::setGeneric(name = "fillNAs", def = function(x, ...) standardGeneric("fillNAs"))
methods::setGeneric(name = "fixBridges", def = function(x, ...) standardGeneric("fixBridges"))
methods::setGeneric(name = "fixDangles", def = function(x, ...) standardGeneric("fixDangles"))
methods::setGeneric(name = "fixLines", def = function(x, ...) standardGeneric("fixLines"))
methods::setGeneric(name = "flow", def = function(x, ...) standardGeneric("flow"))
methods::setGeneric(name = "flowPath", def = function(x, ...) standardGeneric("flowPath"))
methods::setGeneric(name = "focal", package = "terra")
methods::setGeneric(name = "fractalRast", def = function(x, ...) standardGeneric("fractalRast"))
methods::setGeneric(name = "freq", package = "terra")
methods::setGeneric(name = "fragmentation", def = function(x, ...) standardGeneric("fragmentation"))

methods::setGeneric(name = "geomtype", package = "terra")
methods::setGeneric(name = "geomorphons", def = function(x, ...) standardGeneric("geomorphons"))
methods::setGeneric(name = "global", package = "terra")
methods::setGeneric(name = "grassGUI", def = function(x, ...) standardGeneric("grassGUI"))
methods::setGeneric(name = "grid", def = function(x, ...) standardGeneric("grid"))

methods::setGeneric(name = "head", package = "utils")
methods::setGeneric(name = "hexagons", def = function(x, ...) standardGeneric("hexagons"))
methods::setGeneric(name = "hillshade", def = function(x, ...) standardGeneric("hillshade"))
methods::setGeneric(name = "hist", package = "terra")
methods::setGeneric(name = "horizonHeight", def = function(x, ...) standardGeneric("horizonHeight"))

methods::setGeneric(name = "init", package = "terra")
methods::setGeneric(name = "intersect", package = "terra")
methods::setGeneric(name = "interpIDW", package = "terra")
methods::setGeneric(name = "interpSplines", def = function(x, y, ...) standardGeneric("interpSplines"))
methods::setGeneric(name = "is.2d", def = function(x) standardGeneric("is.2d"))
methods::setGeneric(name = "is.3d", def = function(x) standardGeneric("is.3d"))
methods::setGeneric(name = "is.factor", def = function(x) standardGeneric("is.factor"))
# is.double() is primitive and cannot be made generic
methods::setGeneric(name = "is.cell", def = function(x) standardGeneric("is.cell"))
methods::setGeneric(name = "is.doub", def = function(x) standardGeneric("is.doub"))
methods::setGeneric(name = "is.float", def = function(x) standardGeneric("is.float"))
# methods::setGeneric(name = "is.int", def = function(x) standardGeneric("is.int"))
methods::setGeneric(name = "is.int", package = "terra")
methods::setGeneric(name = "is.lines", package = "terra")
methods::setGeneric(name = "is.lonlat", package = "terra")
if (!isGeneric("is.na")) { methods::setGeneric(name = "is.na", def = function(x) standardGeneric("is.na")) }
methods::setGeneric(name = "is.points", package = "terra")
methods::setGeneric(name = "is.polygons", package = "terra")

# methods::setGeneric(name = "keepRows", def = function(x, ...) standardGeneric("keepRows"))
methods::setGeneric(name = "kernel", def = function(x, ...) standardGeneric("kernel"))
methods::setGeneric(name = "kurtosis", def = function(x, ...) standardGeneric("kurtosis"))

# levels (in base) is generic
# levels<- (in base) is primitive
methods::setGeneric(name = "layerCor", package = "terra")
methods::setGeneric(name = "ln", def = function(x, ...) standardGeneric("ln"))
methods::setGeneric(name = "longlat", def = function(x, ...) standardGeneric("longlat"))
methods::setGeneric(name = "log10p", def = function(x) standardGeneric("log10p"))

methods::setGeneric(name = "mask", package = "terra")
methods::setGeneric(name = "maskNA", def = function(x, ...) standardGeneric("maskNA"))
methods::setGeneric(name = "match", package = "terra")
methods::setGeneric(name = "minmax", package = "terra")
methods::setGeneric(name = "missing.cases", def = function(...) standardGeneric("missing.cases"))
methods::setGeneric(name = "missingCats", def = function(x, ...) standardGeneric("missingCats"))
methods::setGeneric(name = "merge", package = "terra")
#' @importFrom terra median
methods::setGeneric(name = "median", def = function(x, na.rm) standardGeneric("median"))
methods::setGeneric(name = "mmode", def = function(x, ...) package = "omnibus")
methods::setGeneric(name = "multivarEnvSim", def = function(ref, proj, ...) standardGeneric("multivarEnvSim"))

# "names" (in base) is primitive
methods::setGeneric(name = "N", def = function(x, ...) standardGeneric("N"))
methods::setGeneric(name = "ncell", package = "terra")
methods::setGeneric(name = "ncell3d", def = function(x) standardGeneric("ncell3d"))
methods::setGeneric(name = "ncol", def = function(x) standardGeneric("ncol")) # in base
methods::setGeneric(name = "ndepth", def = function(x) standardGeneric("ndepth"))
methods::setGeneric(name = "neighborhoodMatrix", def = function(x, ...) standardGeneric("neighborhoodMatrix"))
methods::setGeneric(name = "neighbourhoodMatrix", def = function(x, ...) standardGeneric("neighbourhoodMatrix"))
methods::setGeneric(name = "new", package = "methods")
methods::setGeneric(name = "nlyr", package = "terra")
methods::setGeneric(name = "nacell", def = function(x, ...) standardGeneric("nacell"))
methods::setGeneric(name = "ngeom", def = function(x, ...) standardGeneric("ngeom"))
methods::setGeneric(name = "nlevels", package = "terra")
methods::setGeneric(name = "noise", def = function(x, ...) standardGeneric("noise"))
methods::setGeneric(name = "nonnacell", def = function(x, ...) standardGeneric("nonnacell"))
methods::setGeneric(name = "not.na", def = function(x, ...) standardGeneric("not.na")) # in base
methods::setGeneric(name = "nrow", def = function(x) standardGeneric("nrow")) # in base
methods::setGeneric(name = "nsubgeom", def = function(x, ...) standardGeneric("nsubgeom"))
methods::setGeneric(name = "nunique", def = function(x, ...) standardGeneric("nunique"))

# methods::setGeneric(name = "orOrOr", def = function(e1, e2) standardGeneric("orOrOr"))
methods::setGeneric(name = "reorient", def = function(x, ...) standardGeneric("reorient"))

methods::setGeneric(name = "pairs", package = "terra")
methods::setGeneric(name = "plot", package = "terra")
methods::setGeneric(name = "plotRGB", package = "terra")
methods::setGeneric(name = "predict", package = "terra")
methods::setGeneric(name = "princomp", package = "terra")
methods::setGeneric(name = "print") # base
methods::setGeneric(name = "project", package = "terra")

methods::setGeneric(name = "quantile", package = "terra")

methods::setGeneric(name = "rast", package = "terra")
methods::setGeneric(name = "rasterize", package = "terra")
# methods::setGeneric(name = "rbind")

# from https://stackoverflow.com/questions/27886535/proper-way-to-use-cbind-rbind-with-s4-classes-in-package
methods::getGeneric("rbind")
methods::setGeneric(name = "rbind", signature = "...")
methods::setGeneric(name = "regress", package = "terra")
methods::setGeneric(name = "remove0", def = function(x, ...) standardGeneric("remove0"))
methods::setGeneric(name = "removeAngles", def = function(x, ...) standardGeneric("removeAngles"))
methods::setGeneric(name = "removeBridges", def = function(x, ...) standardGeneric("removeBridges"))
methods::setGeneric(name = "removeDangles", def = function(x, ...) standardGeneric("removeDangles"))
methods::setGeneric(name = "removeDupCentroids", def = function(x, ...) standardGeneric("removeDupCentroids"))
methods::setGeneric(name = "removeDups", def = function(x, ...) standardGeneric("removeDups"))
methods::setGeneric(name = "removeSmallPolys", def = function(x, ...) standardGeneric("removeSmallPolys"))
methods::setGeneric(name = "replaceNAs", def = function(x, ...) standardGeneric("replaceNAs"))
methods::setGeneric(name = "res", package = "terra")
methods::setGeneric(name = "res3d", def = function(x) standardGeneric("res3d"))
methods::setGeneric(name = "resample", package = "terra")
methods::setGeneric(name = "rNormRast", def = function(x, ...) standardGeneric("rNormRast"))
methods::setGeneric(name = "ruggedness", def = function(x, ...) standardGeneric("ruggedness"))
methods::setGeneric(name = "rUnifRast", def = function(x, ...) standardGeneric("rUnifRast"))
methods::setGeneric(name = "rvoronoi", def = function(x, ...) standardGeneric("rvoronoi"))
methods::setGeneric(name = "rWalkRast", def = function(x, ...) standardGeneric("rWalkRast"))

methods::setGeneric(name = "S", def = function(x, ...) standardGeneric("S"))
methods::setGeneric(name = "sampleRast", def = function(x, ...) standardGeneric("sampleRast"))
methods::setGeneric(name = "scale", package = "terra")
methods::setGeneric(name = "scalepop", def = function(x, ...) standardGeneric("scalepop"))
methods::setGeneric(name = "sdpop", def = function(x, ...) standardGeneric("sdpop"))
methods::setGeneric(name = "selectRange", def = function(x, ...) standardGeneric("selectRange"))
methods::setGeneric(name = "segregate", def = function(x, ...) standardGeneric("segregate"))
methods::setGeneric(name = "show", package = "methods")
methods::setGeneric(name = "simplifyGeom", package = "terra")
methods::setGeneric(name = "sineRast", def = function(x, ...) standardGeneric("sineRast"))
methods::setGeneric(name = "smoothGeom", def = function(x, ...) standardGeneric("smoothGeom"))
methods::setGeneric(name = "skewness", def = function(x, ...) standardGeneric("skewness"))
methods::setGeneric(name = "snap", package = "terra")
methods::setGeneric(name = "sources", package = "terra")
methods::setGeneric(name = "spatSample", package = "terra")
methods::setGeneric(name = "rSpatialDepRast", def = function(x, ...) standardGeneric("rSpatialDepRast"))

methods::setGeneric(name = "streams", def = function(x, ...) standardGeneric("streams"))
methods::setGeneric(name = "stretch", package = "terra")
methods::setGeneric(name = "st_as_sf", package = "sf")
# methods::setGeneric(name = "st_bbox", package = "sf")
methods::setGeneric(name = "st_buffer", package = "sf")
methods::setGeneric(name = "st_crs", package = "sf")
# methods::setGeneric(name = "st_distance", package = "sf")
methods::setGeneric(name = "stdev", package = "terra")
methods::setGeneric(name = "subset", package = "terra")
methods::setGeneric(name = "subst", package = "terra")
methods::setGeneric(name = "summary", def = function(object, ...) standardGeneric("summary"))

methods::setGeneric(name = "terrain", package = "terra")
methods::setGeneric(name = "thinLines", def = function(x, ...) standardGeneric("thinLines"))
methods::setGeneric(name = "thinPoints", def = function(x, y, ...) standardGeneric("thinPoints"))
methods::setGeneric(name = "tiles", def = function(x, ...) standardGeneric("tiles"))
methods::setGeneric(name = "top", def = function(x, ...) standardGeneric("top"))
methods::setGeneric(name = "topology", def = function(x, ...) standardGeneric("topology"))
methods::setGeneric(name = "trim", package = "terra")

methods::setGeneric(name = "update", package = "terra")
methods::setGeneric(name = "union", package = "terra")
methods::setGeneric(name = "unscale", def = function(x, ...) standardGeneric("unscale"))

methods::setGeneric(name = "varpop", def = function(x, ...) standardGeneric("varpop"))
methods::setGeneric(name = "vegIndex", def = function(x, ...) standardGeneric("vegIndex"))
methods::setGeneric(name = "vect", package = "terra")
methods::setGeneric(name = "voronoi", package = "terra")

methods::setGeneric(name = ".workDir", def = function(x, ...) standardGeneric(".workDir"))
methods::setGeneric(name = "W", def = function(x, ...) standardGeneric("W"))
methods::setGeneric(name = "wetness", def = function(x, ...) standardGeneric("wetness"))
methods::setGeneric(name = "writeRaster", package = "terra")
methods::setGeneric(name = "writeVector", package = "terra")

methods::setGeneric(name = "xres", package = "terra")
methods::setGeneric(name = "xor", def = function(x, y, ...) standardGeneric("xor"))

methods::setGeneric(name = "yres", package = "terra")

methods::setGeneric(name = "zext", def = function(x) standardGeneric("zext"))
methods::setGeneric(name = "zonal", package = "terra")
methods::setGeneric(name = "zonalGeog", def = function(x, ...) standardGeneric("zonalGeog"))
methods::setGeneric(name = "zres", def = function(x) standardGeneric("zres"))

#' Rows of a GRaster or GVector's table that have no NAs or that have NAs
#'
#' @description When applied to a categorical `GRaster`, `compete.cases()` returns `TRUE` for each row of the "levels" table that has no `NA`s in it. In contrast, `missing.cases()` returns `TRUE` for each row that has at least one `NA` in it. If the raster is not categorical, then `NA` is returned.
#'
#' When applied to a `GVector` with a data table, `complete.cases()` returns `TRUE` for each row where there are no `NA`s. if the `GVector` has no data table, then a vector of `TRUE` values the same length as the total number of geometries will be returned. In contrast, `missing.cases()` returns `TRUE` for every row that has at least one `NA` in it. If the `GVector` has no data table, then a vector of `FALSE` values is returned.
#'
#' @param ... A `GRaster` or `GVector`.
#'
#' @param levels Logical (`GRaster`s only): If `TRUE` (default), then assess only the "value" and [activeCat()] columns of the levels table (see [levels()]). If `FALSE`, then assess all columns (see [cats()]).
#'
#' @returns Both `complete.cases()` and `missing.cases()` return the same type of object. The output depends on the input:
#' * A categorical `GRaster` with just one layer: A logical vector.
#' * An integer, float, or double `GRaster` with just one layer: `NA`.
#' * A `GRaster` with multiple layers: A list with one element per layer with either logical vectors or `NA`s, as per above.
#' * A `GVector` with a data table: A logical vector.
#' * A `GVector` without a data table: `NA`.
#'
#' @seealso [missingCats()] 
#'
#' @example man/examples/ex_complete.cases.r
#'
#' @aliases complete.cases
#' @rdname complete.cases
#' @exportMethod complete.cases
methods::setMethod(
	f = "complete.cases",
	signature = "GRaster",
	function(..., levels = TRUE) {
	
	dots <- list(...)
	if (length(dots) != 1L) stop("This function only assess complete cases for a single GRaster at a time (though it can be multi-layered).")
	x <- dots[[1L]]
	
	if (levels) {
		categs <- levels(x)
	} else {
		categs <- cats(x)
	}

	out <- list()
	for (i in seq_len(nlyr(x))) {
	
		if (is.factor(x)[i]) {
			out[[i]] <- stats::complete.cases(categs[[i]])
		} else {
			out[[i]] <- NA
		}
	
	}
	
	if (nlyr(x) == 1L) out <- out[[1L]]
	out
	
	} # EOF
)

#' @aliases complete.cases
#' @rdname complete.cases
#' @exportMethod complete.cases
methods::setMethod(
	f = "complete.cases",
	signature = "GVector",
	function(...) {
	
	dots <- list(...)
	if (length(dots) != 1L) stop("Tis function can only assess complete cases for a single GVector at a time.")
	x <- dots[[1L]]
	
	table <- x@table
	if (nrow(table) == 0L) {
		out <- rep(TRUE, ngeom)
	} else {
		out <- stats::complete.cases(table)
	}
	out
	
	} # EOF
)

#' @aliases missing.cases
#' @rdname complete.cases
#' @exportMethod missing.cases
methods::setMethod(
	f = "missing.cases",
	signature = "GRaster",
	function(..., levels = TRUE) {

	dots <- list(...)
	if (length(dots) != 1L) stop("This function can only assess complete cases for a single GRaster at a time (though it can be multi-layered).")
	x <- dots[[1L]]
	
	if (levels) {
		categs <- levels(x)
	} else {
		categs <- cats(x)
	}
	
	out <- list()
	for (i in seq_len(nlyr(x))) {

		if (is.factor(x)[i]) {
			out[[i]] <- !stats::complete.cases(categs[[i]])
		} else {
			out[[i]] <- NA
		}
	
	}
	
	if (nlyr(x) == 1L) out <- out[[1L]]
	out
	
	} # EOF
)

#' @aliases missing.cases
#' @rdname complete.cases
#' @exportMethod missing.cases
methods::setMethod(
	f = "missing.cases",
	signature = "GVector",
	function(...) {

	dots <- list(...)
	if (length(dots) != 1L) stop("This function can only assess missing cases for a single GVector at a time.")
	x <- dots[[1L]]
	
	table <- x@table
	if (nrow(table) == 0L) {
		rep(FALSE, ngeom(x))
	} else {
		out <- !stats::complete.cases(table)
	}
	out
	
	} # EOF
)

#' Kernel density estimator of points
#'
#' @description `kernel()` creates a raster using a kernel density estimator of the density of points in a "points" `GVector`.
#'
#' @param x A "points" `GVector`.
#'
#' @param y A `GRaster`: The extent and resolution of this raster will be used to create the density raster. Otherwise, values in this raster are ignored.
#'
#' @param kernel Character: Name of the kernel function to use. Possible values include:
#' * `"Epanechnikov"` (default)
#' * `"Gaussian"`
#' * `"uniform"`
#' * `"triangular"`
#' * `"quartic"`
#' * `"triweight"`
#' * `"cosine"`
#'
#' Partial matching is used, and case is ignored.
#'
#' @param optimize Logical: If `TRUE` (default), then attempt to find the optimal radius less than or equal to the `radius` value using the "Gaussian" kernel. If `FALSE`, use the `radius` value as-is.
#'
#' @param h Numeric or `NULL` (default): Smoothing bandwidth of kernel estimator.
#'
#' If this is `NULL`, the Epanechnikov kernel is used, and `optimize` is `TRUE`, then Silverman's rule-of-thumb is used to estimate the optimal value of `h`:
#' \deqn{h = 0.9 * min(\sigma_x / n^1/6, \sigma_y / n^1/6)}
#'
#' If the Gaussian kernel is used, and `optimize` is `TRUE`, then the **GRASS** `v.kernel` function will attempt to identify the optimal bandwidth, up to the value of `h`, if `h` is defined.
#'
#' Otherwise, if `h` is `NULL`, then the value will be arbitrarily set at 1/5th of the shorter of the distance of the x- and y-extent of the points.
#'
#' @seealso **GRASS** manual page for tool `v.kernel` (see `grassHelp("v.kernel")`)
#'
#' @returns A `GRaster`.
#'
#' @example man/examples/ex_kernel.r
#'
#' @aliases kernel
#' @rdname kernel
#' @exportMethod kernel
methods::setMethod(
	f = "kernel",
	signature = c(x = "GVector"),
	function(
		x,
		y,
  		kernel = "Epanechnikov",
		optimize = TRUE,
		h = NULL
	) {
	
 	kernFx <- omnibus::pmatchSafe(kernel, c("gaussian", "uniform", "triangular", "epanechnikov", "quartic", "triweight", "cosine"), nmax = 1L)

	.locationRestore(y)
	compareGeom(x, y)
	.region(y)

	# assign value to radius
	if (is.null(h) & optimize & kernFx == "epanechinokov") {
	
		coords <- crds(x)
		sds <- apply(coords, 2L, "sd")
		n_1_6 <- nrow(x) ^ (1 / 6)
		h <- 0.9 * min(sds[1L] / n_1_6, sds[2L] / n_1_6)

	} else if (is.null(h)) {

		extent <- ext(x, vector = TRUE)
		xext <- extent[2L] - extent[1L]
		yext <- extent[4L] - extent[3L]
	
		extent <- min(xext, yext)
		h <- 0.2 * extent

	}

	src <- .makeSourceName("v_kernel", "raster")
	args <- list(
		cmd = "v.kernel",
		input = sources(x),
		output = src,
		radius = h,
		kernel = kernFx,
		flags = c(.quiet(), "overwrite")
	)

	if (optimize & kernFx == "gaussian") args$flags <- c(args$flags, "o")
	do.call(rgrass::execGRASS, args = args)

	makeGRaster(src, "kernelDensity")
	
	} # EOF
)

#' Create a raster with random values drawn from a normal distribution
#'
#' @description `rNormRast()` creates a raster with values drawn from a normal distribution.
#'
#' @param x A `GRaster`: The output will have the same extent and dimensions as this raster.
#'
#' @param n An integer: Number of rasters to generate. 
#'
#' @param mu,sigma Numeric: Mean and sample standard deviation of output. If creating more than one raster, you can provide one value per raster. If there are fewer, they will be recycled.
#'
#' @param seed Numeric integer or `NULL`: Random seed. If `NULL`, then a random seed is used for each raster. If provided, there should be one seed value per raster.
#'
#' @returns A `GRaster`.
#'
#' @example man/examples/ex_randRast.r
#' 
#' @seealso [rSpatialDepRast()], [fractalRast()], [rUnifRast()], [rWalkRast()], **GRASS** manual page for tool `r.random.surface` (see `grassHelp("r.random.surface")`)
#'
#' @aliases rNormRast
#' @rdname rNormRast
#' @exportMethod rNormRast
methods::setMethod(
    f = "rNormRast",
    signature = c(x = "GRaster"),
    function(x,
        n = 1,
        mu = 0,
        sigma = 1,
        seed = NULL) {

    if (!is.null(seed)) if (length(seed) != n) stop("You must provide one value of `seed` per raster, or set it to NULL.")

    mu <- rep(mu, length.out = n)
    sigma <- rep(sigma, length.out = n)

    .locationRestore(x)
    .region(x)

    srcs <- .makeSourceName("rnormScaled", "raster", n)

    for (i in seq_len(n)) {

        args <- list(
            cmd = "r.surf.gauss",
            output = srcs[i],
            mean = mu[i],
            sigma = sigma[i],
            flags = c(.quiet(), "overwrite"),
            intern = TRUE
        )
        if (!is.null(seed)) args$seed <- seed[i]
        do.call(rgrass::execGRASS, args = args)

    } # next raster
    makeGRaster(srcs, "rnorm")

    } # EOF
)

#' Revert to a previously-created "GRASS" "location"
#'
#' @description This function resets the connection to a previously-created **GRASS** "location". The session must have been already created using [fast()] in the current **R** session. This function is typically only of use to developers.
#'
#' @param x Either:
#' * A character: Name of the "location" in **GRASS**.
#' * An integer: Index of the "location" in `.fasterRaster$locations`.
#' * A `GSpatial` object (usually a `GRaster` or `GVector`).
#'
#' Any of these can be found using `.locationFind()`.
#'
#' @return An object of class `GLocation` (invisibly) if successful. An error will likely result if not.
#'
#' @aliases .locationRestore
#' @rdname locationRestore
#' @keywords internal
methods::setMethod(
	f = ".locationRestore",
	signature = c(x = "character"),
	function(x) ..locationRestore(x)
)

#' @aliases .locationRestore
#' @rdname locationRestore
#' @keywords internal
methods::setMethod(
	f = ".locationRestore",
	signature = c(x = "integer"),
	function(x) ..locationRestore(x)
)

#' @aliases .locationRestore
#' @rdname locationRestore
#' @keywords internal
methods::setMethod(
	f = ".locationRestore",
	signature = c(x = "numeric"),
	function(x) ..locationRestore(x)
)

#' @aliases .locationRestore
#' @rdname locationRestore
#' @keywords internal
methods::setMethod(
	f = ".locationRestore",
	signature = c(x = "GSpatial"),
	function(x) ..locationRestore(x)
)

#' @keywords internal
..locationRestore <- function(x) {

	if (inherits(x, "character")) {
		location <- x
	} else if (inherits(x, c("numeric", "integer"))) {
		location <- names(.fasterRaster$locations)[x]
	} else if (inherits(x, "GSpatial")) {
		location <- x@location
		mapset <- x@mapset
	}

	index <- .locationFind(location, return = "index")

	if (is.null(index)) {
		stop("Location has not been created. You must use faster() to set the installation directory where GRASS is installed, then use fast() at least once to start GRASS.")
	}

	opts <- faster()
	grassDir <- opts$grassDir
	addonsDir <- opts$addonsDir
	workDir <- 	opts$workDir
	mapset <- "PERMANENT"

	coordRef <- .fasterRaster$locations[[index]]
	if (location != .fasterRaster$activeLocation) {

		### reconnect to location
		emptyRast <- terra::rast(matrix(1L), crs = coordRef)

		# suppressWarnings(
		# 	session <- rgrass::initGRASS(
		# 		gisBase = grassDir,
		# # 		addon_base = addonsDir,
		# 		home = workDir,
		# 		gisDbase = workDir, # ?
		# 		SG = emptyRast,
		# 		location = location,
		# 		mapset = mapset,
		# 		override = TRUE, # must be TRUE to restart, even in different location/mapset
		# 		remove_GISRC = FALSE, # ???
		# 		ignore.stderr = TRUE
		# 	)
		# )

		if (grassInfo("versionNumber") >= 8.4) {
			
			rgrass::execGRASS(
				cmd = "g.mapset",
				mapset = mapset,
				project = location,
				flags = c(.quiet(), "overwrite")#,
				# dbase = workDir
			)
		
		} else {
			
			rgrass::execGRASS(
				cmd = "g.mapset",
				mapset = mapset,
				location = location,
				flags = c(.quiet(), "overwrite")#,
				# dbase = workDir
			)
		
		}

		# .fasterRaster$activeLocation <<- location
		assign(x = "activeLocation", value = location, pos = .fasterRaster)
		Sys.sleep(0.1) # obviates failures to switch locations(?)

	}

	session <- GLocation(
		location = location,
		mapset = mapset,
		crs = coordRef,
		workDir = workDir
	)

	invisible(session)

}

#' Match CRS of a GSpatial object and an existing "GRASS" location
#'
#' @description The function searches the set of available **GRASS** "projects" (previously known as "locations") for one that has a coordinate reference system matching a `GSpatial` object. If none are found, or if no connection with **GRASS** has yet been made, then it returns `NULL`. Otherwise, it returns either the index or the name of the matching location.
#'
#' @param x Either:
#' * Missing: Returns names and coordinate reference system strings of all "locations".
#' * A character representing a coordinate reference system in WKT format
#' * A `SpatRaster`, `SpatVector`, or `sf` vector with a coordinate reference system
#' * A `GSpatial` object (usually a `GRaster` or `GVector`)
#'
#' @param return Either:
#' * `"name"` (default): Returns the name of the "location" with a coordinate reference system the same as `x`.
#' * `"index"`: Returns the index of this "location" in `.fasterRaster$locations` of the `.fasterRaster` environment.
#' * `"crs"`: Returns the coordinate reference system of this "project/location".
#'
#' @param match Character: Method used to find the location. If `match` is "`name`"" (default), then the name of the location is used. If `match` is "`crs`", then the coordinate reference system of each location is checked for a match.
#'
#' @returns Character, integer, or `NULL` (if no match is found).
#'
#' @aliases .locationFind,missing-method
#' @rdname locationFind
#' @keywords internal
methods::setMethod(
	f = ".locationFind",
	signature = c(x = "missing"),
	function(x, return = "name") {

	if (!grassStarted()) {
		out <- NULL
	} else {
	
		return <- omnibus::pmatchSafe(return, c("name", "index", "crs"), nmax = 1L)
		if (return == "name") {
			out <- names(.fasterRaster$locations)
		} else if (return == "index") {
			out <- seq_along(.fasterRaster$locations)
		} else if (return == "crs") {
			out <- .fasterRaster$locations
		}
	
	}
	out
	
	} # EOF
)

#' @aliases .locationFind,GLocation-method
#' @rdname locationFind
#' @keywords internal
methods::setMethod(
	f = ".locationFind",
	signature = c(x = "GLocation"),
	function(x, return = "name", match = "name") ..locationFind(x = x, return = return, match = match)
)

#' @aliases .locationFind,SpatRaster-method
#' @rdname locationFind
#' @keywords internal
methods::setMethod(
	f = ".locationFind",
	signature = c(x = "SpatRaster"),
	function(x, return = "name", match = "name") ..locationFind(x = x, return = return, match = match)
)

#' @aliases .locationFind,SpatVector-method
#' @rdname locationFind
#' @keywords internal
methods::setMethod(
	f = ".locationFind",
	signature = c(x = "SpatVector"),
	function(x, return = "name", match = "name") ..locationFind(x = x, return = return, match = match)
)

#' @aliases .locationFind,sf-method
#' @rdname locationFind
#' @keywords internal
methods::setMethod(
	f = ".locationFind",
	signature = c(x = "sf"),
	function(x, return = "name", match = "name") ..locationFind(x = x, return = return, match = match)
)

#' @aliases .locationFind,character-method
#' @rdname locationFind
#' @keywords internal
methods::setMethod(
	f = ".locationFind",
	signature = c(x = "character"),
	function(x, return = "name", match = "name") ..locationFind(x = x, return = return, match = match)
)

#' @noRd
..locationFind <- function(x, return, match) {

	if (!grassStarted()) {
		out <- NULL
	} else {

		return <- omnibus::pmatchSafe(return, c("name", "index", "crs"), nmax = 1L)
		match <- omnibus::pmatchSafe(match, c("name", "crs"), nmax = 1L)

		if (inherits(x, "character") & match == "name") {
			matches <- names(.fasterRaster$locations) == x
		} else if (inherits(x, c("character", "CRS")) & match == "crs") {
			matches <- sapply(.fasterRaster$locations, terra::same.crs, y = x)
		} else if (inherits(x, c("GLocation")) & match == "name") {
			matches <- names(.fasterRaster$locations) == .location(x)
		} else if (inherits(x, c("GLocation", "SpatRaster", "SpatVector", "sf")) & match == "crs") {
			matches <- sapply(.fasterRaster$locations, terra::same.crs, y = x)
		} else {
			stop("Cannot match the input to a searchable aspect of .fasterRaster$locations.")
		}
		
		index <- which(matches)
		
		if (length(index) == 0L) {
			out <- NULL
		} else {
			
			if (return == "name") {
				out <- names(.fasterRaster$locations)[index]
			} else if (return == "index") {
				out <- index
			} else if (return == "crs") {
				out <- .fasterRaster$locations[[index]]
			}
		}
		
	}
	out

}

#' Determine if GRasters and/or GVectors are geographically comparable
#'
#' `compareGeom()` compares geographic metadata between two or more `GRaster`s and/or `GVector`s. In many cases, spatial objects must be comparable for them to "interact" (e.g., conducting arithmetic operations, masking, etc.).
#'
#' @param x,y,... `GRaster`s or `GVector`s. If `y` is `GRaster`, then the `...` must also be `GRaster`s (or missing). If `y` is `GVector`, then the `...` must also be `GVector`s (or missing).
#'
#' @param location,mapset Logical: Compare **GRASS** "project/location" and "mapsets" (see `vignette("projects_mapsets", package = "fasterRaster")`). Default is `TRUE`.
#' 
#' @param crs Logical: Compare coordinate reference systems. Default is `TRUE`.
#' 
#' @param lyrs Logical (rasters only): Compare number of layers of "stacked" rasters. Note this is different from number of vertical "depths" of a raster. Default is `FALSE`.
#' 
#' @param topo Logical: Test for same topology (2D or 3D). By default, this is `TRUE` for raster-raster comparisons, and `FALSE` for all others.
#'
#' @param ext Logical: If `TRUE`, test for same extent. By default, is `TRUE` for raster-raster comparison and `FALSE` for all others.
#'
#' @param zext Logical: Test for same vertical extents (3D only). By default, this is `TRUE` for raster-raster comparisons, and `FALSE` for all others.
#' 
#' @param rowcol Logical (rasters only): Test for same number of rows and columns. Default is `TRUE`.
#' 
#' @param depths Logical (rasters only): Test for same number of depths. Default is `TRUE`.
#' 
#' @param res Logical (rasters only): Test for same resolution in x- and y-dimensions. Default is `TRUE`.
#'
#' @param zres Logical (rasters only): Test for same resolution in z dimension. Default is `TRUE`.
#' 
#' @param geometry Logical (vector-vector comparison only): Compare geometry. Default is `FALSE`.
#'
#' @param stopOnError Logical: If `TRUE` (default), throw an error with an explanation if the objects are not comparable. If `FALSE` (default), return `TRUE` or `FALSE`.
#'
#' @param messages Logical: If `TRUE` (default), display a warning if a condition is not met. This only comes into effect if `stopOnError` is `FALSE`.
#'
#' @returns Logical (invisibly): `TRUE` for no mismatches detected, `FALSE` for incompatibility), or side effect of throwing an error.
#'
#' @aliases compareGeom
#' @rdname compareGeom
#' @exportMethod compareGeom
methods::setMethod(
	"compareGeom",
	signature = c(x = "GRaster", y = "GRaster"),
	definition = function(
		x,
		y,
		...,
		location = TRUE,
		mapset = TRUE,
		topo = TRUE,
		lyrs = FALSE,
		crs = TRUE,
		ext = TRUE,
		zext = TRUE,
		rowcol = TRUE,
		depths = TRUE,
		res = TRUE,
		zres = TRUE,
		stopOnError = TRUE,
		messages = TRUE
	) {

	out <- TRUE

	y <- list(y, ...)
	for (i in seq_along(y)) {

		if (location) out <- .locationCompare(out=out, x=x, y=y[[i]], stopOnError = stopOnError, messages = messages)
		if (mapset) out <- .mapsetCompare(out = out, x=x, y=y[[i]], stopOnError = stopOnError, messages = messages)
		if (crs) out <- .crsCompare(out = out, x = x, y = y[[i]], stopOnError = stopOnError, messages = messages)

		if (lyrs) {
			if (nlyr(x) != nlyr(y[[i]])) {
				msg <- "The rasters have a different number of layers."
				if (stopOnError) stop(msg)
				if (messages & !stopOnError) warning(msg)
				out <- FALSE
			}
		}

		if (topo) out <- .topoCompare(out=out, x=x, y=y[[i]], stopOnError=stopOnError, messages=messages)
		if (ext) out <- .extentCompare(out=out, x=x, y=y[[i]], stopOnError=stopOnError, messages=messages)
		if (zext) out <- .zextentCompare(out=out, x=x, y=y[[i]], stopOnError=stopOnError, messages=messages)

		if (rowcol) {
			xdim <- dim(x)[1L:2L]
			ydim <- dim(y[[i]])[1L:2L]

			if (!all(xdim == ydim)) {
				msg <- "The rasters have different numbers of rows and/or columns."
				if (stopOnError) stop(msg)
				if (messages & !stopOnError) warning(msg)
				out <- FALSE
			}
		}

		if (depths) {

			msg <- "The rasters have a different numbers of depths."

			xd <- ndepth(x)
			yd <- ndepth(y[[i]])
			xna <- is.na(xd)
			yna <- is.na(yd)
			if (any(xna) | any(yna)) {
				if (!all(xna == yna)) {
					if (stopOnError) stop(msg)
					if (messages & !stopOnError) warning(msg)
					out <- FALSE
				}
			}

			if (any(!xna) | any(!yna)) {
				if (!all(xd[!xna] == yd[!yna])) {
					if (stopOnError) stop(msg)
					if (messages & !stopOnError) warning(msg)
					out <- FALSE
				}
			}

		}

		if (res) {
			xres <- res(x)[1L:2L]
			yres <- res(y[[i]])[1L:2L]
			
			if (!all(omnibus::compareFloat(xres, yres, "=="))) {
				msg <- "The rasters have a different horizontal resolutions."
				if (stopOnError) stop(msg)
				if (messages & !stopOnError) warning(msg)
				out <- FALSE
			}
		}

		if (zres) {
			
			xzres <- zres(x)
			yzres <- zres(y[[i]])

			if (!is.na(xzres) & !is.na(yzres)) {

				msg <- "The rasters have a different vertical resolutions."

				if (
					(is.na(xzres) & !is.na(yzres)) |
					(!is.na(xzres) & is.na(yzres))
				) {
					if (stopOnError) stop(msg)
					if (messages & !stopOnError) warning(msg)
					out <- FALSE
				} else if (omnibus::compareFloat(xzres, yzres, "!=")) {
					if (stopOnError) stop(msg)
					if (messages & !stopOnError) warning(msg)
					out <- FALSE
				}

			}

		}

	} # next object to compare to

	invisible(out)

	} # EOF
)

#' @aliases compareGeom
#' @rdname compareGeom
#' @exportMethod compareGeom
methods::setMethod(f = "compareGeom",
	signature = c(x = "GVector", y = "GVector"),
	definition = function(
		x,
		y,
		...,
		location = TRUE,
		mapset = TRUE,
		topo = FALSE,
		crs = TRUE,
		ext = FALSE,
		zext = FALSE,
		geometry = FALSE,
		stopOnError = TRUE,
		messages = TRUE
	) {

	.compareGeneric(
		x = x,
		y = y,
		...,
		location = location,
		mapset = mapset,
		crs = crs,
		topo = topo,
		ext = ext,
		zext = zext,
		geometry = geometry,
		stopOnError = stopOnError,
		messages = messages
	)

	} # EOF
)

#' @aliases compareGeom
#' @rdname compareGeom
#' @exportMethod compareGeom
methods::setMethod(f = "compareGeom",
	signature = c(x = "GRaster", y = "GVector"),
	definition = function(
		x,
		y,
		...,
		location = TRUE,
		mapset = TRUE,
		topo = FALSE,
		crs = TRUE,
		ext = FALSE,
		zext = FALSE,
		stopOnError = TRUE,
		messages = TRUE
	) {

	.compareGeneric(
		x = x,
		y = y,
		...,
		location = location,
		mapset = mapset,
		crs = crs,
		topo = topo,
		ext = ext,
		zext = zext,
		geometry = FALSE,
		stopOnError = stopOnError,
		messages = messages
	)

	} # EOF
)

#' @aliases compareGeom
#' @rdname compareGeom
#' @exportMethod compareGeom
methods::setMethod(f = "compareGeom",
	signature = c(x = "GVector", y = "GRaster"),
	definition = function(
		x,
		y,
		...,
		location = TRUE,
		mapset = TRUE,
		topo = FALSE,
		crs = TRUE,
		ext = FALSE,
		zext = FALSE,
		stopOnError = TRUE,
		messages = TRUE
	) {

	.compareGeneric(
		x = x,
		y = y,
		...,
		location = location,
		mapset = mapset,
		crs = crs,
		topo = topo,
		ext = ext,
		zext = zext,
		geometry = FALSE,
		stopOnError = stopOnError,
		messages = messages
	)

	} # EOF
)

# x is a GRaster and y and ... GVectors OR
# x is a GVector and y and ... GRasters
.compareGeneric <- function(
	x,
	y,
	...,
	location,
	mapset,
	crs,
	topo,
	ext,
	zext,
	geometry,
	stopOnError,
	messages
) {

	out <- TRUE
	y <- list(y, ...)
	
	allGVector <- sapply(y, inherits, "GVector")
	allGRaster <- sapply(y, inherits, "GRaster")
	allGVector <- all(allGVector)
	allGRaster <- all(allGRaster)
	allSameClass <- allGRaster | allGVector

	if (!allSameClass) stop("The objects supplied in ... must be of the same class as the y object.")

	for (i in seq_along(y)) {

		if (location) out <- .locationCompare(out=out, x=x, y=y[[i]], stopOnError=stopOnError, messages=messages)
		if (mapset) out <- .mapsetCompare(out=out, x=x, y=y[[i]], stopOnError=stopOnError, messages=messages)
		if (crs) out <- .crsCompare(out=out, x=x, y=y[[i]], stopOnError=stopOnError, messages=messages)
		if (topo) out <- .topoCompare(out=out, x=x, y=y[[i]], stopOnError=stopOnError, messages=messages)
		if (ext) out <- .extentCompare(out=out, x=x, y=y[[i]], stopOnError=stopOnError, messages=messages)
		if (zext) out <- .zextentCompare(out=out, x=x, y=y[[i]], stopOnError=stopOnError, messages=messages)

		if (geometry) {
			if (geomtype(x) != geomtype(y[[i]])) {
				msg <- "The vectors have a different geometries."
				if (stopOnError) stop(msg)
				if (messages & !stopOnError) warning(msg)
				out <- FALSE
			}
		}

	} # next object
	out

}

### compare locations
.locationCompare <- function(out, x, y, stopOnError, messages) {

	if (.location(x) != .location(y)) {
		msg <- paste0("The objects have different coordinate reference systems. You can use\n  project() to warp one of them to the reference system of the other.")
		if (stopOnError) stop(msg)
		if (messages & !stopOnError) warning(msg)
		out <- FALSE
	}
	out
}

### compare mapsets
.mapsetCompare <- function(out, x, y, stopOnError, messages) {

	if (.mapset(x) != .mapset(y)) {
		msg <- "The objects are in different GRASS mapsets."
		if (stopOnError) stop(msg)
		if (messages & !stopOnError) warning(msg)
		out <- FALSE
	}
	out
}

### compare CRS
.crsCompare <- function(out, x, y, stopOnError, messages) {

	if (st_crs(x) != st_crs(y)) {
		msg <- "The objects have different coordinate reference systems."
		if (stopOnError) stop(msg)
		if (messages & !stopOnError) warning(msg)
		out <- FALSE
	}
	out
}

### compare extents
.extentCompare <- function(out, x, y, stopOnError, messages) {
	
	xx <- ext(x, vector=TRUE)
	yy <- ext(y, vector=TRUE)
	if (any(omnibus::compareFloat(xx, yy, "!="))) {
		msg <- "The objects have different extents."
		if (stopOnError) stop(msg)
		if (messages & !stopOnError) warning(msg)
		out <- FALSE
	}
	out
	
}

### compare z-extents
.zextentCompare <- function(out, x, y, stopOnError, messages) {

	if (topology(x) == "3D" & topology(y) == "3D") {
	
		xzext <- zext(x)
		xbottom <- xzext[1L]
		xtop <- xzext[2L]

		yzext <- zext(y)
		ybottom <- yzext[1L]
		ytop <- yzext[2L]
		
		if (anyNA(c(xzext, yzext))) {
			msg <- "At least one raster has a missing z-ext."
			if (stopOnError) stop(msg)
			if (messages & !stopOnError) warning(msg)
			out <- FALSE
		}
		if (omnibus::compareFloat(xbottom, ybottom, "!=")) {
			msg <- "Objects have different vertical extents."
			if (stopOnError) stop(msg)
			if (messages & !stopOnError) warning(msg)
			out <- FALSE
		}
		if (omnibus::compareFloat(xtop, ytop, "!=")) {
			stop("Objects have different vertical extents.")
			if (stopOnError) stop(msg)
			if (messages & !stopOnError) warning(msg)
			out <- FALSE
		}
	}
	out
	
} # EOF

### compare topology
.topoCompare <- function(out, x, y, stopOnError, messages) {

	if (topology(x) != topology(y)) {
		msg <- "The rasters have different topologies."
		if (stopOnError) stop(msg)
		if (messages & !stopOnError) warning(msg)
		out <- FALSE
	}
	out
	
}

#' Apply a principal component analysis (PCA) to layers of a GRaster
#'
#' @description This function applies a principal component analysis to layers of a `GRaster`.
#' 
#' @param x A `GRaster` with two or more layers.
#' 
#' @param scale Logical: If `TRUE` (default), input layers will be rescaled by dividing each layer by its overall population standard deviation. Rasters will always be centered (have their mean subtracted from values). Centering and scaling is recommended when rasters values are in different units.
#'
#' @param scores Logical: If `TRUE`, the `prcomp` object will have the scores attached to it. This can greatly increase the size of the object in memory if the input raster has many cells. It will also take more time. If `FALSE` (default), then skip returning scores.
#' 
#' @returns A multi-layer `GRaster` with one layer per principal component axis. The [pcs()] function can be used on the output raster to retrieve a `prcomp` object from the raster, which includes rotations (loadings) and proportions of variance explained.
#' 
#' @seealso [terra::princomp()], [terra::prcomp()]
#' 
#' @example man/examples/ex_princomp.r
#' 
#' @aliases princomp
#' @rdname princomp
#' @exportMethod princomp
methods::setMethod(
	f = "princomp",
	signature = c(x = "GRaster"),
	function(x, scale = TRUE, scores = FALSE) {

	if (nlyr(x) < 2L) stop("Input raster must have 2 or more layers.")

	.locationRestore(x)
	.region(x)

	input <- paste(sources(x), collapse = ",")
	src <- .makeSourceName("i_pca", "raster")

	args <- list(
		cmd = "i.pca",
		input = input,
		output = src,
		rescale = c(0, 0),
		flags = c(.quiet(), "overwrite"),
		intern = TRUE
	)

	if (scale) args$flags <- c(args$flags, "n")

	# execute PCA and get rasters
	info <- do.call(rgrass::execGRASS, args = args)
	layers <- 1L:nlyr(x)
	srcs <- paste0(src, ".", layers)
	out <- makeGRaster(srcs, paste0("PC", layers))

	### get PCA information
	#######################

	n <- nlyr(x)

	# standard deviations
	sdev <- rep(NA_real_, n)

	# rotations
	rotation <- matrix(NA_real_, nrow = n, ncol = n, dimnames = list(names(x), names(out)))

	for (pc in layers) {
		
		this <- info[pc]

		# standard deviation
		start <- regexpr(this, pattern = " ")[1L] + 1L
		stop <- regexpr(this, pattern = "\\(")[1L] - 1L
		var <- substr(this, start, stop)
		var <- as.numeric(var)
		sd <- sqrt(var)
		sdev[pc] <- sd

		# rotations
		start <- regexpr(this, pattern = "\\(")[1L] + 1L
		end <- regexpr(this, pattern = ")")[1L] - 1L
		rots <- substr(this, start, end)
		rots <- strsplit(rots, ",")[[1L]]
		rots <- as.numeric(rots)

		rotation[ , pc] <- rots

	}

	# center
	center <- global(x, "mean")
	center <- c(center)[[1L]]
	names(center) <- names(x)

	# scale
	if (!scale) {
		scale <- FALSE
	} else {
		scale <- global(x, "sd")
		scale <- c(scale)[[1L]]
		names(scale) <- names(x)
	}

	pca <- list(
		sdev = sdev,
		rotation = rotation,
		center = center,
		scale = scale
	)

	# return scores
	if (scores) {

		sc <- as.points(out, values = TRUE)
		sc <- as.data.table(sc)
		sc <- as.matrix(sc)
		pca$scores <- sc

	}

	class(pca) <- "prcomp"
	attr(out, "pca") <- pca
	out

	} # EOF
)

#' Retrieve a principal components model from a PCA GRaster
#' 
#' @param x A `GRaster` created by [princomp()] 
#' 
#' @returns An object of class `prcomp`.
#' 
#' @seealso [princomp()], [terra::princomp()], tool `i.pca` in **GRASS**
#' 
#' @example man/examples/ex_princomp.r
#'
#' @aliases pcs
#' @rdname pcs
#' @export
pcs <- function(x) attr(x, "pca")

#' Calculate cell values based on values of nearby cells
#' 
#' @description This function calculates statistics on a moving "neighborhood" of cells of a raster. The neighborhood can be a square, circle, or a user-defined set of cells (with or without weights).
#' 
#' @param x A `GRaster`.
#' 
#' @param w Numeric integer or a square matrix with an odd number of rows and columns: The size and nature of the neighborhood:
#' * "Square" neighborhoods (when `circle = FALSE`): An odd integer >= 3, indicating indicates the size of a "square" neighborhood (number of cells wide and number or cells tall).
#' * "Circular" neighborhoods (when `circle = TRUE`): An odd integer >=3, indicating the diameter of the circle.
#' * A matrix of cell weights: The matrix must be square and have an odd number of rows and columns (example: `matrix(c(0.5, 1, 0.5, 1, 2, 1, 0.5, 1, 0.5), nrow=3)`). You cannot use a weights matrix when `circle = TRUE`. Cells with `NA` as a weight will be ignored. Note that weighted matrices should not be used for function `min`, `max`, `count`, `nunique`, or `interspersion`.
#' 
#' @param fun Character: Name of the function to apply to the neighborhood:
#' * "`mean`" (default)
#' * "`median`"
#' * "`mode`"
#' * "`min`" or "`max`": Minimum or maximum. Should not use a weights matrix.
#' * "`range`": Difference between the maximum and minimum.  Should not use a weights matrix.
#' * "`sd`": Sample standard deviation.  NB: This is the same as the [stats::sd()] function.
#' * "`sdpop`": Population standard deviation. NB: This is the same as the function "stddev" in the **GRASS** tool `r.neighbors`.
#' * "`sum`": Sum of non-`NA`` cells.
#' * "`count`": Number of non-`NA cells. Should not use a weights matrix.
#' * "`var`": Sample variance.  NB: This is the same as the [stats::var()] function.
#' * "`varpop`": Population variance. NB: This is the same as the function "variance" in the **GRASS** tool `r.neighbors`.
#' * "`nunique`": Number of unique values. Should not use a weights matrix.
#' * "`interspersion`": Proportion of cells with values different from focal cell (e.g., if 6 of 8 cells have different values, then the interspersion is 6/8 = 0.75). NB: This is slightly different from how it is defined in the **GRASS** tool `r.neighbors`. Should not use a weights matrix.
#' * "`quantile`": Quantile of values. The value in argument `quantile` is used to specify the quantile.
#' 
#' The center cell value is always included in the calculations, and all calculations ignore `NA` cells (i.e., they do not count as cells in the focal neighborhood).
#'
#' @param circle Logical: If `FALSE` (default), use a square neighborhood. If `TRUE`, use a circular neighborhood. When this is `TRUE`, argument `w` cannot be a matrix.
#'
#' @param quantile Numeric between 0 and 1, inclusive: Quantile to calculate when `fun = "quantile"`. The default value is 0.5 (median), and valid values must be in the range between 0 and 1, inclusive.
#'
#' @returns A `GRaster`.
#' 
#' @seealso [terra::focal()], **GRASS** manual page for tool `r.neighbors` (see `grassHelp("r.neighbors")`)
#' 
#' @example man/examples/ex_focal.r
#' 
#' @aliases focal
#' @rdname focal
#' @exportMethod focal
methods::setMethod(
    f = "focal",
    signature = c(x = "GRaster"),
    function(x, w = 3, fun = "sum", circle = FALSE, quantile = 0.5) {

	if (inherits(w, "numeric") && (w < 3 | omnibus::compareFloat(w %% 2, 0, "=="))) stop("Argument `w` must be an odd integer >= 3.")
	
	if (inherits(w, "matrix") && (nrow(w) != ncol(w) | omnibus::compareFloat(nrow(w) %% 2, 0, "==") | omnibus::compareFloat(ncol(w) %% 2, 0, "=="))) stop("Matrix `w` must have the same number of rows and columns, and it must have an odd number of each.")
	
	if (is.matrix(w) & circle) ("Cannot use a circular neighborhood and a weights matrix at the same time.")
	
    funs <- c("mean", "sum", "sd", "sdpop", "var", "varpop", "median", "mode", "max", "min", "count", "range", "nunique", "interspersion", "quantile")

    if (!is.character(fun)) stop("Argument `fun` must be a character.")

    fun <- omnibus::pmatchSafe(fun, funs, useFirst = TRUE)
    if (fun == "mean") fun <- "average"
    if (fun == "min") fun <- "minimum"
    if (fun == "max") fun <- "maximum"
    if (fun == "nunique") fun <- "diversity"
	
    if (fun == "quantile") {
        if (quantile < 0 | quantile > 1) stop("Argument `quantile` must be in the range [0, 1].")
    }

    .locationRestore(x)
    .region(x)

	# size of neighborhood
	size <- if (inherits(w, "matrix")) {
		nrow(w)
	} else {
		w
	}

    args <- list(
        cmd = "r.neighbors",
        input = NA_character_,
        output = NA_character_,
        method = fun,
        size = size,
        nprocs = faster("cores"),
        memory = faster("memory"),
        flags = c(.quiet(), "overwrite")
    )

    if (circle) args$flags <- c(args$flags, "c")
    if (fun == "quantile") args$quantile = quantile

    ### weights matrix
    if (inherits(w, "matrix")) {
        
		if (anyNA(w)) w[is.na(w)] <- 0
		
		weight <- as.data.frame(w)
		colnames(weight) <- NULL
		
        # GRASS wants backslashes
		rand <- omnibus::rstring(1L)
		weightFile <- paste0(tempdir(), "/TEMPTEMPweights", rand, ".txt")
		weightFileBack <- paste0(tempdir(), "\\TEMPTEMPweights", rand, ".txt")
		
		sink(weightFile)
		print(weight, row.names = FALSE)
		sink()
		
		args$weight <- weightFileBack
        args$weighting_function <- "file"
        
    }

    ### process each layer of raster
    nLayers <- nlyr(x)
    for (i in seq_len(nLayers)) {

        # must convert to double in some cases
        if (fun == "average") {

			# convert to double to obviate issues with imprecision
			gnDouble <- .makeSourceName("double", "raster")
			ex <- paste0(gnDouble, " = double(", sources(x)[i], ")")
			rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"), intern = TRUE)

			# focal sums (numerator)
			gnSum <- .makeSourceName("sum", "raster")
			thisArgs <- args
			thisArgs$method <- "sum"
			thisArgs$input <- gnDouble
			thisArgs$output <- gnSum
			do.call(rgrass::execGRASS, args = thisArgs)

			# ones mask (for denominator)
			gnOnes <- .makeSourceName("ones", "raster")
			ex <- paste0(gnOnes, " = if(isnull(", gnDouble, "), null(), double(1))")
			rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"), intern = TRUE)

			# count raster (sum of ones--denominator)
			gnCount <- .makeSourceName("count", "raster")
			thisArgs <- args
			thisArgs$method <- "sum"
			thisArgs$input <- gnOnes
			thisArgs$output <- gnCount
			do.call(rgrass::execGRASS, args = thisArgs)

			# calculate mean
			src <- .makeSourceName(fun, "raster")
			ex <- paste0(src, " = ", gnSum, " / ", gnCount)
			rgrass::execGRASS("r.mapcalc", expression = ex, flags=c(.quiet(), "overwrite"), intern = TRUE)

        } else if (fun %in% c("sdpop", "sd", "var", "varpop")) {
	
			# convert to double to obviate issues with imprecision
			gnDouble <- .makeSourceName("double", "raster")
			ex <- paste0(gnDouble, " = double(", sources(x)[i], ")")
			rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"), intern = TRUE)
		
			### numerator LHS
			
				# square values
				gnSquared <- .makeSourceName("squared", "raster")
				ex <- paste0(gnSquared, " = ", gnDouble, "^2")
				rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"), intern = TRUE)
				
				# focal sums of squares
				gnSoS <- .makeSourceName("sumOfSquares", "raster")
				thisArgs <- args
				thisArgs$method <- "sum"
				thisArgs$input <- gnSquared
				thisArgs$output <- gnSoS
				do.call(rgrass::execGRASS, args = thisArgs)

			### numerator RHS
				
				# focal sums
				gnSums <- .makeSourceName("sum", "raster")
				thisArgs <- args
				thisArgs$method <- "sum"
				thisArgs$input <- gnDouble
				thisArgs$output <- gnSums
				do.call(rgrass::execGRASS, args = thisArgs)
				
				# squared sums raster
				gnSquaredSums <- .makeSourceName("squaredSums", "raster")
				ex <- paste0(gnSquaredSums, " = ", gnSums, "^2")
				rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"), intern = TRUE)

			### counts
				
				# ones mask... used to obviate problems where NULL cells are counted as 0
				gnOnes <- .makeSourceName("ones", "raster")
				ex <- paste0(gnOnes, " = if(isnull(", gnDouble, "), null(), double(1))")
				rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"), intern = TRUE)
				
				# count raster (sum of ones)
				gnCount <- .makeSourceName("count", "raster")
				thisArgs <- args
				thisArgs$method <- "sum"
				thisArgs$input <- gnOnes
				thisArgs$output <- gnCount
				do.call(rgrass::execGRASS, args = thisArgs)

			### final
				
				# numerator
				gnNumer <- .makeSourceName("numerator", "raster")
				ex <- paste0(gnNumer, " = ", gnSoS, " - (", gnSquaredSums, " / ", gnCount, ")")
				rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"), intern = TRUE)
				
				# sd or variance
				src <- .makeSourceName(fun, "raster")
				ex <- if (fun == "sd") {
					paste0(src, " = sqrt(", gnNumer, " / (", gnCount, " - 1))")
				} else if (fun == "sdpop") {
					paste0(src, " = sqrt(", gnNumer, " / ", gnCount, ")")
				} else if (tolower(fun) == "var") {
					paste0(src, " = ", gnNumer, " / (", gnCount, " - 1)")
				} else if (tolower(fun) == "varpop") {
					paste0(src, " = ", gnNumer, " / (", gnCount, ")")
				}

				rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"), intern = TRUE)

        } else if (fun == "interspersion") {
		
			gnInter <- .makeSourceName("interspersion", "raster")
			args$output <- gnInter
			args$input <- sources(x)[i]
			do.call(rgrass::execGRASS, args = args)

			src <- .makeSourceName("focal", "raster")
			ex <- paste0(src, " = (double(", gnInter, ") - 1) / 100")
			rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"), intern = TRUE)
			
        } else {

			src <- .makeSourceName("focal", "raster")
			args$output <- src
			args$input <- sources(x)[i]
			do.call(rgrass::execGRASS, args = args)

        }
	
		if (is.factor(x)[i] & fun %in% c("mode", "min", "max")) {
			levels <- cats(x)[[i]]
		} else {
			levels <- NULL
		}
	
        this <- makeGRaster(src, names(x)[i], levels = levels)
        if (i == 1L) {
            out <- this
        } else {
            out <- c(out, this)
        }

    } # next raster
    out

    } # EOF
) 

#' Find which cells of a GRaster match certain values
#'
#' @description The `match()` function takes a `GRaster` and a numeric, integer or character vector as inputs and returns a `GRaster` with cell values that correspond to the index of each element in the vector that matched the original cell value. For example, if a 4-cell raster had values 3, `NA`, 5, 4, and the vector was `c(3, 4)`, then the output would be a 4-cell raster with values 1, `NA`, `NA`, 2 because the first value in the vector was 3 (so the cell with 3 is assigned 1), and because the second value in the vector was 4 (so the cell with 4 was assigned 2). The other two values had no matches.
#'
#' If the `GRaster` is categorical, then the vector can be category labels instead of numeric values.
#'
#' The `%in%` operator returns a `GRaster` with cell values that are 1 if their original values appeared in the vector, and 0 if not (or `NA` if the original value was `NA`). If the `GRaster` is categorical, then the vector can be category labels instead of numeric values.
#'
#' The `%notin%` operator returns 1 for cells with values that are *not* found in the vector, and 0 otherwise. If the `GRaster` is categorical, then the vector can be category labels instead of numeric values.
#'
#' @param x A `GRaster`: Note that any kind of `GRaster` is acceptable (integer, float, double, or categorical), but matching may not work as intended for float and double rasters because of problems with comparing floating-point values.
#'
#' @param table A numeric, integer, or character vector.
#'
#' @param nomatch Numeric or integer: Value to return when no match is found.
#'
#' @returns A `GRaster`.
#'
#' @seealso [terra::match()], [match()], [omnibus::notIn()]
#'
#' @example man/examples/ex_match.r
#'
#' @aliases match
#' @rdname match
#' @exportMethod match
methods::setMethod(
	f = "match",
	signature = c(x = "GRaster"),
	function(x, table, nomatch = NA) {
	
	if (is.character(table) & !all(is.factor(x))) {
		stop("Can only match category labels to categorical rasters.")
	}

	.locationRestore(x)
	.region(x)

	nLayers <- nlyr(x)
	srcs <- .makeSourceName("r_mapcalc", "raster", nLayers)

	for (i in seq_len(nLayers)) {
	
		if (is.character(table)) {

			levs <- cats(x)[[i]]
			ac <- activeCat(x)[i]

			if (faster("useDataTable")) {
				labels <- levs[[ac + 1L]]
				matches <- match(table, labels)
				vals <- levs[[1L]][matches]
			} else {
				labels <- levs[ , ac + 1L, drop = TRUE]
				matches <- match(table, labels)
				vals <- levs[ , 1L, drop = TRUE][matches]
			}

		} else {
			vals <- table
		}
	
		ex <- paste0(srcs[i], " = if(isnull(", sources(x)[i],"), null(), if(", sources(x)[i], "==", vals[1L], ",1,")

		if (length(vals) > 1L) {
			for (count in 2L:length(vals)) {
				ex <- paste0(ex, "if(", sources(x)[i], "==", vals[count], ",", count, ",")
			}
			
			ex <- if (is.na(nomatch)) {
				paste0(ex, "null())")
			} else {
				paste0(ex, nomatch, ")")
			}
			
			for (count in 2L:length(vals)) {
				ex <- paste0(ex, ")")
			}
		}

		ex <- paste0(ex, ")")

		rgrass::execGRASS(
			cmd = "r.mapcalc",
			expression = ex,
			flags = c(.quiet(), "overwrite")
		)

	} # next raster

	# if (!is.na(nomatch)) {
	
	# 	srcsIn <- srcs
	# 	srcs <- .makeSourceName("r_mapcalc", "raster", nLayers)
	# 	for (i in seq_len(nLayers)) {

	# 		ex <- paste0(srcs[i], " = if(isnull(", sources(x)[i], "), null(), ", srcsIn[i], ")")
	# 		rgrass::execGRASS(
	# 			cmd = "r.mapcalc",
	# 			expression = ex,
	# 			flags = c(.quiet(), "overwrite")
	# 		)

	# 	}
	
	# }

	# add category levels
	if (is.character(table)) {
	
		levs <- .freq(srcs, dtype = rep("CELL", length(srcs)))

		acs <- activeCat(x, names = TRUE)

  		if (!inherits(levs, "list")) levs <- list(levs)
		if (!faster("useDataTable")) {
			for (i in seq_along(levs)) levs[[i]] <- data.table::as.data.table(levs[[i]])
		}

		for (i in seq_along(acs)) {

			if (nrow(levs[[i]]) == 0L) {
				levs[[i]] <- ""
			} else {

				levs[[i]] <- levs[[i]][ , "count" := NULL]
				values <- levs[[i]][["value"]]
				labels <- table[match(values, seq_along(table))]
				labels[is.na(labels)] <- "nomatch"

				levs[[i]][ , acs[i] := labels]

				levs[[i]][ , "value" := lapply(.SD, as.integer), .SDcols = "value"]
				data.table::setorderv(levs[[i]], col = "value")

			}
				
		}
		
	} else {
		levs <- NULL
	}
	makeGRaster(srcs, names(x), levels = levs)
	
	} # EOF
)

#' @aliases %in%
#' @rdname match
#' @exportMethod %in%
methods::setMethod(
	f = "%in%",
	signature = c(x = "GRaster"),
	function(x, table) {
	
	.locationRestore(x)
	.region(x)

	if (is.character(table) & !all(is.factor(x))) {
		stop("Can only match category labels to categorical rasters.")
	}

	nLayers <- nlyr(x)
	srcs <- .makeSourceName("r_mapcalc", "raster", nLayers)

	for (i in seq_len(nLayers)) {
	
		if (is.character(table)) {

			levs <- cats(x)[[i]]
			ac <- activeCat(x)[i]

			# labels <- levs[[ac + 1L]]
			# matches <- match(table, labels)
			# vals <- levs[[1L]][matches]

			if (faster("useDataTable")) {
				labels <- levs[[ac + 1L]]
				matches <- match(table, labels)
				vals <- levs[[1L]][matches]
			} else {
				labels <- levs[ , ac + 1L, drop = TRUE]
				matches <- match(table, labels)
				vals <- levs[ , 1L, drop = TRUE][matches]
			}

		} else {
			vals <- table
		}

		if (is.na(vals[1L])) {

			ex <- paste0(srcs[i], " = 0 * ", sources(x))

		} else {

			ex <- paste0(srcs[i], " = if(isnull(", sources(x)[i], "), null(), if(", sources(x)[i], "==", vals[1L])

			if (length(vals) > 1L) {
				for (count in 2L:length(vals)) {
					ex <- paste0(ex, "|", sources(x)[i], "==", vals[count])
				}
			}

			ex <- paste0(ex, ",1,0))")

		}
		
		rgrass::execGRASS(
			cmd = "r.mapcalc",
			expression = ex,
			flags = c(.quiet(), "overwrite")
		)
		
	} # next raster

	makeGRaster(srcs, names(x))

	} # EOF
)

#' @aliases %notin%
#' @rdname match
#' @exportMethod %notin%
methods::setMethod(
	f = "%notin%",
	signature = c(x = "GRaster"),
	function(x, table) {
	
	.locationRestore(x)
	.region(x)

	if (is.character(table) & !all(is.factor(x))) {
		stop("Can only match category labels to categorical rasters.")
	}

	nLayers <- nlyr(x)
	srcs <- .makeSourceName("r_mapcalc", "raster", nLayers)

	for (i in seq_len(nLayers)) {
	
		if (is.character(table)) {

			levs <- levels(x)[[i]]
			ac <- activeCat(x)[i]
			labels <- levs[[ac + 1L]]
			matches <- match(table, labels)
			vals <- levs[[1L]][matches]

		} else {
			vals <- table
		}
	
		ex <- paste0(srcs[i], " = if(isnull(", sources(x)[i], "), null(), if(", sources(x)[i], "!=", vals[1L])

		if (length(vals) > 1L) {
			for (count in 2L:length(vals)) {
				ex <- paste0(ex, "&", sources(x)[i], "!=", vals[count])
			}
		}

		ex <- paste0(ex, ",1,0))")

		rgrass::execGRASS(
			cmd = "r.mapcalc",
			expression = ex,
			flags = c(.quiet(), "overwrite")
		)
	
	} # next raster

	makeGRaster(srcs, names(x))
	
	} # EOF
)

if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Elevation raster, rivers vector, locations of Dypsis plants
madElev <- fastData("madElev")
madRivers <- fastData("madRivers")
madDypsis <- fastData("madDypsis")

# Convert a SpatRaster to a GRaster, and sf to a GVector
elev <- fast(madElev)
rivers <- fast(madRivers)
dypsis <- fast(madDypsis)

### case 1: GRaster by itself

# Distance between NA cells and nearest non-NA cells
naDist <- distance(elev)
names(naDist) <- "NA Distance"
plot(naDist)

# Distance between non-NA cells and nearest NA cells
nonNaDist <- distance(elev, fillNA = FALSE)
names(nonNaDist) <- "non-NA Distance"
plot(nonNaDist)

# Distance between cells with an elevation of 3 and any other cell that != 3
distFocal3 <- distance(elev, target = 3)
names(distFocal3) <- "Distance from 3"
plot(distFocal3)

# Distance between any cell and cells with a value of 3
distTo3 <- distance(elev, fillNA = FALSE, target = 3)
names(distTo3) <- "Distance to 3"
plot(distTo3)

### Case 2: GRaster and GVector
distToVect <- distance(elev, rivers)

plot(distToVect)
plot(rivers, add = TRUE)

### Case 3: GVector vs GVector
plot(rivers)
plot(dypsis, add = TRUE)

distToRivers <- distance(dypsis, rivers, unit = "yd")
distToPlants <- distance(rivers, dypsis)
distToRivers
distToPlants

### Case 4: GVector vs itself
distToItself <- distance(dypsis)
distToItself

}

if (grassStarted()) {

# Setup
library(terra)

# Elevation raster
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster:
elev <- fast(madElev)

# Convert non-NA to 1, NA cells remain NA
elevMask <- maskNA(elev)
elevMask
plot(c(elev, elevMask))

# Convert NA to 1, non-NA cells become NA
elevInvertMask <- maskNA(elev, invert = TRUE)
elevInvertMask
plot(c(elev, elevInvertMask))

# Convert NA to 200, non-NA cells keep their values
elevInvertRetain <- maskNA(elev, value = 200, invert = TRUE, retain = TRUE)
elevInvertRetain
plot(c(elev, elevInvertRetain))

}

#' Sample random points from a GRaster or GVector
#'
#' @description `spatSample()` randomly locates points across a `GRaster` or `GVector`. It can return a `GVector`, the coordinates, values associated with the points, or all of these. If you want to generate a raster with randomly-sampled cells, see [sampleRast()].
#' 
#' @param x A `GRaster` or `GVector`.
#' 
#' @param size Numeric value > 0: Number of points to create.
#' 
#' @param as.points Logical: If `FALSE` (default), the output is a `data.frame` or `data.table`. If `TRUE`, the output is a "points" `GVector`.
#' 
#' @param values Logical: If `TRUE` (default), values of the `GRaster` at points are returned.
#' 
#' @param cats Logical: If `TRUE` (default) and the `GRaster` is categorical, then return the category label of each cell. If `values` is also `TRUE`, then the cell value will also be returned.
#' 
#' @param xy Logical: If `TRUE`, return the longitude and latitude of each point. Default is `FALSE`.
#'
#' @param strata Either `NULL` (default), or a `GVector` defining strata. If supplied, the `size` argument will be interpreted as number of points to place per geometry in `strata`. Note that using strata can dramatically slow the process.
#'
#' @param byStratum Logical: If `FALSE` (default), then `size` number of points will be placed within the entire area delineated by `strata`. If `TRUE`, then `size` points will be placed within each subgeometry of `strata`.
#'
#' @param zlim Either `NULL` (default), or a vector of two numbers defining the lower and upper altitudinal bounds of coordinates. This cannot be combined with `values = TRUE` or `cats = TRUE`.
#'
#' @param seed Either `NULL` (default) or an integer: Random number seed. If this is `NULL`, the a seed will be set randomly. Values will be rounded to the nearest integer.
#'
#' @param verbose Logical: If `TRUE`, display progress. Default is `FALSE`.
#'
#' @returns A `data.frame`, `data.table`, or `GVector`.
#' 
#' @seealso [sampleRast()], [terra::spatSample()], tool `v.random` in **GRASS** (see `grassHelp("v.random")`)
#'
#' @example man/examples/ex_sampleRast_spatSample.r
#'
#' @aliases spatSample
#' @rdname spatSample
#' @exportMethod spatSample
methods::setMethod(
	f = "spatSample",
	signature = "GRaster",
	function(
		x,
		size,
		as.points = FALSE,
		values = TRUE,
		cats = TRUE,
		xy = FALSE,
		strata = NULL,
		byStratum = FALSE,
		zlim = NULL,
		seed = NULL,
		verbose = FALSE
	) {

	# for debugging
	if (FALSE) {
	
		as.points <- FALSE
		values <- TRUE
		cats <- TRUE
		xy <- FALSE
		strata <- NULL
		byStratum <- FALSE
		zlim <- NULL
		seed <- NULL
		verbose <- TRUE
	
	}

	if (!is.null(zlim) & (values | cats)) stop("You cannot at present extract values or categories using 3D points.")
	if (!xy & !as.points & !values & !cats) stop("At least one of `xy`, `as.points`, `values`, or `cats` must be TRUE.")

	.locationRestore(x)
	.region(x)
	
	# ### fast point location... use R
	# ### v.random is REALLY slow for even moderately large sample sizes
	# if (is.null(strata)) {

	# 	# for unprojected, we want to adjust for smaller cells near poles, so we oversample, then subsample
	# 	extent <- ext(x, vector = TRUE)
	# 	if (is.lonlat(x)) {

	# 		# sampling scheme for latitude:
	# 		# divide y extent into large number of bins
	# 		# weight each bin by cos(latitude)
	# 		# sample with replacement from bin y values
	# 		# add a uniform random number +- half bin width

	# 		yres <- yres(x)
	# 		yLower <- extent[3L] + 0.5 * yres
	# 		yUpper <- extent[4L] - 0.5 * yres
	# 		ydim <- dim(x)[2L]
	# 		nMarks <- 100 * ydim
	# 		halfDelta <- 0.5 * (yUpper - yLower) / (nMarks - 1)
	# 		yMarks <- seq(yLower, yUpper, length.out = nMarks)
	# 		w <- abs(cos(pi * yMarks / 180))
	# 		ys <- sample(yMarks, size, prob = w, replace = TRUE)
	# 		ys <- ys + stats::runif(size, -halfDelta, halfDelta)
		
	# 	} else {
	# 		# if not long/lat, just sample uniformly
	# 		ys <- stats::runif(size, extent[3L], extent[4L])
	# 	}

	# 	xs <- stats::runif(size, extent[1L], extent[2L])
	# 	if (!is.null(zlim)) zs <- stats::runif(size, zlim[1L], zlim[2L])

	# 	if (xy) {
	# 		if (is.null(zlim)) {
	# 			out <- data.table::data.table(x = xs, y = ys)
	# 		} else {
	# 			out <- data.table::data.table(x = xs, y = ys, z = zs)
	# 		}
	# 	}

	# 	### NB The script below ingests points in chunks bc that has proven faster than doing all at once for larger numbers of points.

	# 	xs <- round(xs, 7L)
	# 	ys <- round(ys, 7L)
	# 	if (!is.null(zlim)) zs <- round(zs, 7L)

	# 	xs[xs < extent[1L]] <- extent[1L]
	# 	xs[xs > extent[2L]] <- extent[2L]
	# 	ys[ys < extent[3L]] <- extent[3L]
	# 	ys[ys > extent[4L]] <- extent[4L]

	# 	### use v.in.ascii to ingest each subset of points
	# 	##################################################

	# 	nAtATime <- 2E5 # optimal-ish size based on manual checks
	# 	sets <- ceiling(size / nAtATime)
	# 	srcs <- .makeSourceName("spatSample_v_in_ascii", "vector", n = sets)

	# 	if ((verbose | faster("verbose")) & sets > 1)  {
	# 		omnibus::say("Ingesting points...")
	# 		pb <- utils::txtProgressBar(min = 0, max = sets, initial = 0, style = 3, width = 30)
	# 	}

	# 	for (set in seq_len(sets)) {

	# 		if ((verbose | faster("verbose")) & sets > 1) utils::setTxtProgressBar(pb, set)

	# 		index <- (nAtATime * (set - 1) + 1):(min(nAtATime * set, size))
	# 		thisXs <- xs[index]
	# 		thisYs <- ys[index]
	# 		if (!is.null(zlim)) thisZs <- zs[index]

	# 		if (is.null(zlim)) {
	# 			coords <- data.table::data.table(cat = index, x = thisXs, y = thisYs)
	# 		} else {
	# 			coords <- data.table::data.table(cat = index, x = thisXs, y = thisYs, z = thisZs)
	# 		}
	# 		coords[ , coords := do.call(paste, c(.SD, sep = "|"))]
	# 		coords <- coords[ , "coords", drop = FALSE]

	# 		tf <- tempfile(fileext = ".txt")
	# 		data.table::fwrite(coords, tf, col.names = FALSE, quote = FALSE, scipen = 20L)

	# 		args <- list(
	# 			"v.in.ascii",
	# 			input = tf,
	# 			output = srcs[set],
	# 			format = "point",
	# 			separator = "pipe",
	# 			cat = 1, x = 2, y = 3,
	# 			flags = c(.quiet(), "overwrite", "t", "n", "b") # "b" --> no topology
	# 			# flags = c(.quiet(), "overwrite", "t", "n") # "b" --> no topology
	# 		)
	# 		if (!is.null(zlim)) {
	# 			args$flags <- c(args$flags, "z")	
	# 			args$z <- 4
	# 		}
	# 		do.call(rgrass::execGRASS, args = args)
	# 		unlink(tf)

	# 		# if (set > 1) {
				
	# 		# 	topCat <- min(nAtATime * set, size)
	# 		# 	srcs[set] <- .vIncrementCats(srcs[set], add = topCat)

	# 		# }

	# 	} # next set

	# 	if ((verbose | faster("verbose")) & sets > 1)  close(pb)

	# 	### use v.patch to combine subsets of points
	# 	############################################

	# 	# seems like we can combine at least 11 vectors at a time, but not too many more
	# 	srcsAtATime <- 10L # number of sources to combine at a time (plus the running `x` source)

	# 	nSrcs <- length(srcs)
	# 	sets <- ceiling(nSrcs / srcsAtATime)

	# 	if ((verbose | faster("verbose")) & sets > 1)  {
	# 		omnibus::say("Collapsing points...")
	# 		pb <- utils::txtProgressBar(min = 0, max = sets, initial = 0, style = 3, width = 30)
	# 	}

	# 	if (nSrcs == 1L) {
	# 		src <- srcs
	# 	} else {
			
	# 		for (set in seq_len(sets)) {

	# 			if ((verbose | faster("verbose")) & sets > 1L)  utils::setTxtProgressBar(pb, set)

	# 			index <- (1 + srcsAtATime * (set - 1)) : min(nSrcs, set * srcsAtATime)
	# 			srcIn <- srcs[index]
	# 			input <- paste(srcIn, collapse = ",")
	# 			if (set > 1) input <- paste0(src, ",", input)
			
	# 			src <- .makeSourceName("spatSample_v_patch", "vector")

	# 			rgrass::execGRASS(
	# 				cmd = "v.patch",
	# 				input = input,
	# 				output = src,
	# 				flags = c(.quiet(), "overwrite", "b", "n") # "n" ==> no topology
	# 				# flags = c(.quiet(), "overwrite") # "n" ==> input has topology, "b" ==> don't build topology
	# 			)
				
	# 			.rm(srcIn, type = "vector", warn = FALSE)

	# 		}

	# 		if ((verbose | faster("verbose")) & sets > 1)  close(pb)

	# 	} # if > 1 set

	# # if strata is not NULL
	# } else {
		
	### locating by stratum
	if (!is.null(strata)) {

		if (size > 250000) .message("spatSample_strata", "Using `strata` when selecting a large number of points can take a long time.")

		src <- .makeSourceName("spatSample_v_random", "vector")

		args <- list(
			cmd = "v.random",
			output = src,
			npoints = size,
			restrict = sources(strata),
			flags = c(.quiet(), "overwrite")
		)

		if (byStratum) args$flags <- c(args$flags, "a")

		if (!is.null(seed)) {
			seed <- round(seed)
			args$seed <- seed
		}

		args$restrict <- sources(strata)

		if (!is.null(zlim)) {
			args$zmin <- zlim[1L]
			args$zmax <- zlim[2L]
			args$flags <- c(args$flags, "z")
		}

		# build topology... needed for GVector or for extraction/coordinates
		if (!as.points & !xy & !(values | cats)) args$flags <- c(args$flags, "b")

		# args$flags <- c(args$flags, "b") ### do not create topology... problems? YES!
		do.call(rgrass::execGRASS, args = args)

	### not loacting by stratum
	} else {
		
		if (verbose | faster("verbose")) omnibus::say("Placing points...")
		src <- .makeSourceName("spatSample_v_random", "vector")
		args <- list(
			cmd = "v.random",
			output = src,
			npoints = size,
			flags = c(.quiet(), "overwrite")
		)

		# build topology... needed for GVector or for extraction/coordinates
		if (!xy & !(values | cats)) args$flags <- c(args$flags, "b")

		if (!is.null(seed)) args$seed <- round(seed)
		do.call(rgrass::execGRASS, args = args)

	}

	if (xy) out <- .crdsVect(src, z = !is.null(zlim), gtype = "points")

	# extract values from raster
	if (values | cats) {

		# rgrass::execGRASS(
		# 	cmd = "v.build",
		# 	map = src,
		# 	flags = c(.quiet(), "overwrite")
		# )

		vals <- .extractFromRasterAtPoints(x = x, y = src, xNames = names(x), dtype = datatype(x), levels = levels(x), cats = cats, verbose = verbose)

		if (exists("out", inherits = FALSE)) {
			out <- cbind(out, vals)
		} else {
			out <- vals
		}

	}

	if (as.points) {

		# # build topology
		# rgrass::execGRASS(
		# 	"v.build",
		# 	map = src,
		# 	option = "build",
		# 	flags = c(.quiet(), "overwrite")
		# )

		thisCats <- if (!byStratum) { 1:size } else { NULL }
		.vAttachDatabase(src, cats = thisCats)
		if (exists("out", inherits = FALSE)) {
			# info <- .vectInfo(src)
			# nGeometries <- info$nGeometries
			# n <- nGeometries / size
			# out <- makeGVector(src, table = out, cats = rep(1:size, each = n))
			out <- makeGVector(src, table = out, cats = thisCats)
		} else {
			out <- makeGVector(src, cats = thisCats)
		}
	} else {
		if (!faster("useDataTable")) out <- as.data.frame(out)
	}
	out

	} # EOF

)

#' @aliases spatSample
#' @rdname spatSample
#' @exportMethod spatSample
methods::setMethod(
	f = "spatSample",
	signature = "GVector",
	function(
		x,
		size,
		as.points = FALSE,
		values = TRUE,
		xy = FALSE,
		byStratum = FALSE,
		zlim = NULL,
		seed = NULL
	) {

	if (!is.null(seed)) {
		if (!omnibus::is.wholeNumber(seed)) stop("Argument `seed` must be an integer or NULL.")
		seed <- round(seed)
	}
	
	.locationRestore(x)
	.region(x)

	if (geomtype(x) != "polygons") x <- convHull(x)


	# # if sampling by strata, keep polygons as-is
	# if (!is.null(strata)) {

	# 	srcRestrict <- sources(x)

	# # if not sampling by strata, dissolve all polygons
	# } else {

	# 	# make copy of vector and force all categories to 1
	# 	gtype <- geomtype(x)
	# 	srcRestrict <- .aggregate(sources(x), dissolve = TRUE, gtype = gtype, copy = TRUE)

	# }
	
	src <- .makeSourceName("spatSample_v_random", "vector")
	args <- list(
		cmd = "v.random",
		output = src,
		npoints = size,
		restrict = sources(x),
		flags = c(.quiet(), "overwrite")
	)

	if (!is.null(seed)) {
		seed <- round(seed)
		args$seed <- seed
	}

	# if (!is.null(strata) & byStratum) args$flags <- c(args$flags, "a")
	if (byStratum) args$flags <- c(args$flags, "a")

	if (!is.null(zlim)) {
		args$zmin <- zlim[1L]
		args$zmax <- zlim[2L]
		args$flags <- c(args$flags, "z")
	}

	do.call(rgrass::execGRASS, args = args)

	# return coordinates
	if (xy) coords <- .crdsVect(src, z = is.3d(x), gtype = "points")

	# extract values from vector
	if (values) {

		if (!xy) coords <- .crdsVect(x = src, z = !is.null(zlim), gtype = "points")
		vals <- .extractFromVect(x, y = coords, xy = xy)
		vals$id.y <- NULL

	} # if wanting values

	out <- NULL
	if (xy & values) {
		out <- cbind(xy, vals)
	} else if (!xy & values) {
		out <- vals
	} else if (xy & !values) {
		out <- coords
	}

	if (as.points) {
		out <- makeGVector(src, table = out)
	} else {
		if (!faster("useDataTable")) out <- as.data.frame(out)
	}
	out

	} # EOF

)

#' Make predictions from a linear or generalized linear model to a GRaster
#'
#' @description This version of the `predict()` function make predictions to a set of `GRaster`s from a model object.
#'
#' The model must be either a linear model, which is of class `lm` and typically created using the [stats::lm()] function or a generalized linear model (GLM), which is class `glm` and typically created using [stats::glm()]. Other packages can also create `lm` or `glm` objects, but they may not work in this function. For example, generalized additive models, which can be created using the `gam()` function in the **mgcv** package, inherit the `glm` class, but cannot be used in this function. However, `glm` objects created with the **speedglm** package should work with this function.
#'
#' This `predict()` function can handle:
#' * Linear predictors and intercepts like `1 + x`;
#' * Quadratic terms like `x^2` (or, in **R** formula notation, `I(x^2)`);
#' * Two-way interaction terms between scalars like `x1:x2` and `x1 * x2`;
#' * Categorical predictors (i.e., categorical `GRaster`s; see `vignette("GRasters", package = "fasterRaster")`);
#' * Two-way interactions between a categorical predictor and a scalar predictor; and
#' * Two-way interactions between categorical predictors.
#'
#' @param object A `GRaster` with one or more layers.
#'
#' @param model An `lm` or `glm` model object.
#'
#' @param type Character: Type of prediction to make. This can be either `link` (default; predictions are made on the scale of the link function) or `response` (predictions are made on the scale of the response variable). This function can only make predictions on the scale of the response for the identity, logit, log, or cloglog (complementary log-log) link functions.
#'
#' @returns A `GRaster`.
#'
#' @seealso [terra::predict()]; [stats::predict()]
#'
#' @example man/examples/ex_predict.r
#'
#' @aliases predict
#' @rdname predict
#' @exportMethod predict
methods::setMethod(
	f = "predict",
	signature = c(object = "GRaster"),
	function(object, model, type = "response") {
	
	type <- omnibus::pmatchSafe(type, c("link", "response"), nmax = 1L)

	.locationRestore(object)
	.region(object)

	### parse model to GRASS r.mapcalc format
	#########################################

	coeffs <- stats::coef(model)

	terms <- data.table(
		term = names(coeffs),
		value = coeffs,
		termClean = NA_character_
	)

	terms <- terms[!is.na(terms$value)]

	terms$term <- gsub(terms$term, pattern = "I\\(", replacement = "")
	terms$term <- gsub(terms$term, pattern = "\\(", replacement = "")
	terms$term <- gsub(terms$term, pattern = "\\)", replacement = "")

	# intercept
	i <- which(terms$term == "Intercept")
	if (length(i) > 0L) {
		terms$termClean[i] <- as.character(terms$value[i])
	}

	# single factors... assuming we just use levels of factors as intercepts, not in interaction terms
	if (any(is.factor(object))) {

		facts <- which(is.factor(object))

		# use if() statement to obtain offsets due to each level of each factor
		for (fact in facts)	{
		
			levs <- levels(object)[[fact]]
			levs$TEMPTEMP__ <- paste0(names(object)[fact], levs[[2L]])
			levs <- levs[levs[[3L]] %in% names(coeffs), ]
			nLevs <- nrow(levs)

			if (nLevs > 0L) {

				for (j in seq_len(nLevs)) {

					thisTerm <- which(terms$term == levs$TEMPTEMP__[j])
					coeff <- coeffs[names(coeffs) == levs$TEMPTEMP__[j]]

					terms$termClean[thisTerm] <- paste0("if(", sources(object)[fact], "==", levs[[1L]][j], ",", coeff, ",0)")

				}

			}
		
		}
	
	}

	# interactions between factors
	i <- which(is.na(terms$termClean) & grepl(terms$term, pattern = ":"))
	if (sum(is.factor(object)) >= 2L & length(i) > 0L) {

		# list contains three items of information:
		# each element named after the layer's name
		# each element contains factor values
		# names of values are the names as they'd appear in coefficients
		facts <- which(is.factor(object))
		possibleNames <- list()
		for (fact in facts) {
			
			levs <- levels(object)[[fact]]
			index <- length(possibleNames) + 1L
			possibleNames[[index]] <- levs[[1L]]
			names(possibleNames[[index]]) <- paste0(names(object)[fact], levs[[2L]])
			names(possibleNames)[index] <- sources(object)[fact]

		}

		# replace "nice" formula term with if() statement testing equality for each factor level combination in model
		for (ii in i) { # each row in terms table
			for (jj in 1L:(length(possibleNames) - 1L)) { # each element in factor list
				for (jCount in seq_along(possibleNames[[jj]])) { # each item in the element
					jItem <- possibleNames[[jj]][jCount]
					jName <- names(jItem)
					for (kk in (jj + 1L):length(possibleNames)) { # remaining elements in factor list
						for (kCount in seq_along(possibleNames[[kk]])) { # each item in element
							kItem <- possibleNames[[kk]][kCount]
							kName <- names(kItem)
							if (terms$term[ii] == paste0(jName, ":", kName) | terms$term[ii] == paste0(kName, ":", jName)) {
								
								jSrc <- names(possibleNames)[jj]
								kSrc <- names(possibleNames)[kk]

								# value of coefficient... at least one of these will be empty
								# both will be empty if there were no cases of crosses between the two factor levels
								thisCoeff1 <- coeffs[paste0(jName, ":", kName)]
								thisCoeff2 <- coeffs[paste0(kName, ":", jName)]
								thisCoeff <- c(thisCoeff1, thisCoeff2)
								thisCoeff <- thisCoeff[!is.na(thisCoeff)]

								if (length(thisCoeff) > 0L) {

									terms$termClean[[ii]] <- paste0("if(", jSrc, "==", jItem, " & ", kSrc, "==", kItem, ",", thisCoeff, ",0)")
								} else {
									terms$termClean[[ii]] <- NA
								}
								
							}
						}
					}
				}
			}
		}

	}

	# interactions between factors and scalars
	i <- which(is.na(terms$termClean) & grepl(terms$term, pattern = ":"))
	if (any(is.factor(object)) & any(!is.factor(object)) & length(i) > 0L) {
		
		# list contains three items of information:
		# each element named after the layer's name
		# each element contains factor values
		# names of values are the names as they'd appear in coefficients
		facts <- which(is.factor(object))
		possibleNames <- list()
		for (fact in facts) {
			
			levs <- levels(object)[[fact]]
			index <- length(possibleNames) + 1L
			possibleNames[[index]] <- levs[[1L]]
			names(possibleNames[[index]]) <- paste0(names(object)[fact], levs[[2L]])
			names(possibleNames)[index] <- sources(object)[fact]

		}

		scalarNames <- names(object)[!is.factor(object)]
		scalarNames <- scalarNames[order(nchar(scalarNames), decreasing = TRUE)]
		for (ii in i) {
		
			term <- terms$term[ii]
			matchScalar <- FALSE
			scalar <- 1L
			while (!matchScalar & scalar <= length(scalarNames)) {

				scalarName <- scalarNames[scalar]
				if (grepl(term, pattern = scalarName)) {

					matchScalar <- TRUE
					matchFact <- FALSE
					fact <- 1L
					while (!matchFact & fact <= length(possibleNames)) {

						for (j in seq_along(possibleNames[[fact]])) {
							
							factNameLevel <- names(possibleNames[[fact]][j])
							if (grepl(term, pattern = factNameLevel)) {
								
								matchFact <- TRUE

								thisCoeff1 <- coeffs[paste0(factNameLevel, ":", scalarName)]
								thisCoeff2 <- coeffs[paste0(scalarName, ":", factNameLevel)]
								thisCoeff <- c(thisCoeff1, thisCoeff2)
								thisCoeff <- thisCoeff[!is.na(thisCoeff)]

								scalarSrc <- sources(object)[[which(names(object) == scalarName)]]
								factSrc <- names(possibleNames)[fact]
								levelValue <- possibleNames[[fact]][j]

								terms$termClean[ii] <- paste0(thisCoeff, " * if(", factSrc, "==", levelValue, ",1,0) * ", scalarSrc)

							} # if factor is in term
						
						} # next level of factor

						fact <- fact + 1L

					} # next factor
				
				} # if scalar is in this term
			
				scalar <- scalar + 1L

			} # next scalar

		} # next term

	} # scalar * factor

	# scalar linear terms
	i <- which(is.na(terms$termClean) & !grepl(terms$term, pattern = ":"))
	if (length(i) > 0L) {
	
		for (j in i) {
		
			term <- .matchRasterNamesToFormula(object, terms$term[j])
			term <- paste(terms$value[j], "*", term)
			terms$termClean[j] <- term

		}
	
	}
	
	# numeric interaction terms
	i <- which(is.na(terms$termClean) & grepl(terms$term, pattern = ":"))
	if (length(i) > 0L) {
	
		for (j in i) {
		
			term <- terms$term[j]
			term <- strsplit(term, split = ":")[[1L]]
			term <- .matchRasterNamesToFormula(object, term)
			term <- paste(term, collapse = "*")
			term <- paste(terms$value[j], "*", term)
			terms$termClean[j] <- term
		
		}
	
	}

	# there is some suggestion that long "expression" terms won't work, so we do it in chunks
	nTermsAtATime <- 20L
	sets <- ceiling(nrow(terms) / nTermsAtATime)
	for (set in seq_len(sets)) {

		index <- (nTermsAtATime * (set - 1L) + 1L):(min(nTermsAtATime * set, nrow(terms)))
		# ex <- paste(terms$termClean[index], collapse = " + ")
		ex <- paste(terms$termClean, collapse = " + ")
		if (set == 1L) {
			src <- .makeSourceName(paste0("predict_r_mapcalc_set", set), "raster")
			ex <- paste0(src, " = ", ex)
		} else {
			srcIn <- src
			src <- .makeSourceName(paste0("predict_r_mapcalc_set", set), "raster")
			ex <- paste0(src, " = ", srcIn, " + ", ex)
		}

		rgrass::execGRASS(
			cmd = "r.mapcalc",
			expression = ex,
			flags = c(.quiet(), "overwrite")
		)

	}

	# return prediction on response scale
	fam <- stats::family(model)
	link <- fam$link

	if (type == "response" & link != "identity") {
	
		srcIn <- src
		src <- .makeSourceName("predict_r_mapcalc", "raster")

		if (link == "log") {
			ex <- paste0(src, " = exp(", srcIn, ")")
		} else if (link == "logit") {
			ex <- paste0(src, " = exp(", srcIn, ") / (1 + exp(", srcIn, "))")
		} else if (link == "cloglog") {
			ex <- paste0(src, " = 1 - exp(-1 * exp(", srcIn, "))")
		} else {
			stop("The ", link, " link function is not supported.")
		}
	
		rgrass::execGRASS(cmd = "r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

	}

	makeGRaster(src, names = "prediction")

	} # EOF
)

#' @param object `GRaster` stack.
#' @param term Character vector to which to match names.
#'
#' @returns A character vector same length as `object` with [sources()] names.
#'
#' @noRd
.matchRasterNamesToFormula <- function(object, term) {

	# find names of rasters and replace from longest to shortest in length to help obviate duplicate names
	names <- names(object)
	srcs <- sources(object)
	nc <- nchar(names)
	ncOrder <- order(nc)
	names <- names[ncOrder]
	srcs <- srcs[ncOrder]

	for (i in seq_along(names)) {
		term <- gsub(term, pattern = names[i], replacement = srcs[i], fixed = TRUE)
	}
	term

}

#' Fix issues with geometries of a vector
#'
#' @description These functions are intended to help fix geometric issues with a `GVector`. Note that the functionality of the `snap()` and `removeAreas()` functions can also be implemented when using [fast()] to create a `GVector`.
#'
#' * `breakPolys()`: Break topologically clean areas. This is similar to `fixLines()`, except that it does not break loops. Topologically clean vectors may occur if the vector was imported from a format that does not enforce topology, such as a shapefile. Duplicate geometries are automatically removed after breaking.
#' * `fixBridges()`: Change "bridges" to "islands" (which are topologically incorrect) within geometries to lines.
#' * `fixDangles()`: Change "dangles" hanging off boundaries to lines if shorter than `tolerance` distance. If `tolerance` is <0, all dangles will be changed to lines.  Units of `tolerance` are in map units, or in degrees for unprojected CRSs. If `tolerance` <0, all dangles are removed, and the function will retain only closed loops and lines connecting loops. Dangles will be removed from longest to shortest.
#' * `fixLines()`: Break lines at intersections and lines that form closed loops.
#' * `remove0()`: Remove all boundaries and lines with a length of 0.
#' * `removeAngles()`: Collapse lines that diverge at an angle that is computationally equivalent to 0. This tool often needs to be followed with the `break()` and `removeDups()` methods.
#' * `removeBridges()`: Remove "bridges" to "islands" (which are topologically incorrect) within geometries.
#' * `removeDangles()`: Remove "dangling" lines if shorter than `tolerance` distance. If `tolerance` is <0, all dangles will be removed. Units of `tolerance` are in map units, or in degrees for unprojected CRSs. If `tolerance` <0, all dangles are removed, and the function will retain only closed loops and lines connecting loops. Dangles will be removed from longest to shortest.
#' * `removeDupCentroids()`: Remove duplicated area centroids. In **GRASS**, closed polygons have their attributes mapped to a (hidden) centroid of the polygon.
#' * `removeDups()`: Remove duplicated features and area centroids. 
#' * `removeSmallPolys()`: Remove polygons smaller than `tolerance`. Units of `tolerance` are in square meters (regardless of the CRS).
#' * `snap()`: Snap lines/boundaries to each other if they are less than `tolerance` apart. Subsequent removal of dangles may be needed. Units of `tolerance` are map units, or degrees for unprojected CRSs.
#'
#' @param x A `GVector`.
#'
#' @param tolerance Numeric or `NULL` (default): Minimum distance in map units (degrees for unprojected, usually meters for projected) or minimum area (in meters-squared, regardless of projection).
#'
#' @seealso [terra::topology()], [fillHoles()], [terra::removeDupNodes()], *Details* section in [fast()], [simplifyGeom()], [smoothGeom()], **GRASS** manual page for tool `v.clean` (see `grassHelp("v.clean")`)
#'
#' @returns A `GVector`.
#'
#' @example man/examples/ex_simplify_smooth_clean_GVector.r
#'
#' @aliases breakPolys
#' @rdname breakPolys
#' @exportMethod breakPolys
methods::setMethod(
	f = "breakPolys",
	signature = c(x = "GVector"),
	function(x) {
	
	if (!is.polygons(x)) stop("This tool can only be applied to polygon GVectors.")
	.cleanGeom(x, method = "bpol", tolerance = NULL, topo = topology(x))

	} # EOF
)

#' @aliases fixBridges
#' @rdname breakPolys
#' @exportMethod fixBridges
methods::setMethod(
	f = "fixBridges",
	signature = c(x = "GVector"),
	function(x) .cleanGeom(x, method = "chbridge", tolerance = NULL, topo = topology(x))
)

#' @aliases fixDangles
#' @rdname breakPolys
#' @exportMethod fixDangles
methods::setMethod(
	f = "fixDangles",
	signature = c(x = "GVector"),
	function(x, tolerance = -1) .cleanGeom(x, method = "chdangle", tolerance = tolerance, topo = topology(x))
)

#' @aliases fixLines
#' @rdname breakPolys
#' @exportMethod fixLines
methods::setMethod(
	f = "fixLines",
	signature = c(x = "GVector"),
	function(x) {
	
	if (is.points(x)) stop("This tool can only be applied to line or polygon GVectors.")
	.cleanGeom(x, method = "break", tolerance = NULL, topo = topology(x))

	} #EOF
)

#' @aliases remove0
#' @rdname breakPolys
#' @exportMethod remove0
methods::setMethod(
	f = "remove0",
	signature = c(x = "GVector"),
	function(x) {
	
	if (is.points(x)) stop("This tool can only be applied to line or polygon GVectors.")
	.cleanGeom(x, method = "rmline", tolerance = NULL, topo = topology(x))

	} # EOF
)

#' @aliases removeAngles
#' @rdname breakPolys
#' @exportMethod removeAngles
methods::setMethod(
	f = "removeAngles",
	signature = c(x = "GVector"),
	function(x) .cleanGeom(x, method = "rmsa", tolerance = NULL, topo = topology(x))
)

#' @aliases removeBridges
#' @rdname breakPolys
#' @exportMethod removeBridges
methods::setMethod(
	f = "removeBridges",
	signature = c(x = "GVector"),
	function(x) .cleanGeom(x, method = "rmbridge", tolerance = NULL, topo = topology(x))
)

#' @aliases removeDangles
#' @rdname breakPolys
#' @exportMethod removeDangles
methods::setMethod(
	f = "removeDangles",
	signature = c(x = "GVector"),
	function(x, tolerance = -1) .cleanGeom(x, method = "rmdangle", tolerance = tolerance, topo = topology(x))
)

#' @aliases removeDupCentroids
#' @rdname breakPolys
#' @exportMethod removeDupCentroids
methods::setMethod(
	f = "removeDupCentroids",
	signature = c(x = "GVector"),
	function(x) {
	
	if (!is.polygons(x)) stop("This tool can only be applied to polygon GVectors.")
	.cleanGeom(x, method = "rmdac", tolerance = NULL, topo = topology(x))
	
	} # EOF
)

#' @aliases removeDups
#' @rdname breakPolys
#' @exportMethod removeDups
methods::setMethod(
	f = "removeDups",
	signature = c(x = "GVector"),
	function(x) .cleanGeom(x, method = "rmdupl", tolerance = NULL, topo = topology(x))
)

#' @aliases removeSmallPolys
#' @rdname breakPolys
#' @exportMethod removeSmallPolys
methods::setMethod(
	f = "removeSmallPolys",
	signature = c(x = "GVector"),
	function(x , tolerance) {
	
	if (!is.polygons(x)) stop("This tool can only be applied to polygon GVectors.")
	.cleanGeom(x, method = "rmarea", tolerance = tolerance, topo = topology(x))

	} # EOF
)

#' @aliases snap
#' @rdname breakPolys
#' @exportMethod snap
methods::setMethod(
	f = "snap",
	signature = c(x = "GVector"),
	function(x , tolerance) .cleanGeom(x, method = "snap", tolerance = tolerance, topo = topology(x))
)

#' Implement vector cleaning using `sources()` name of a vector 
#'
#' @param x `GVector` or [sources()] name of a `GVector`.
#' @param method Character.
#' @param tolerance `NULL` or numeric.
#' @param topo "2D" or "3D"
#'
#' @noRd
.cleanGeom <- function(x, method, tolerance, topo) {

	if (inherits(x, "GVector")) {
		.locationRestore(x)
		src <- sources(x)
		table <- x@table
	} else {
		src <- x
		table <- NULL
	}

	src <- .copyGSpatial(src, type = "vector", topo = topo, reshapeRegion = FALSE)
	
	src <- .makeSourceName("v_clean", "vector")
	args <- list(
		cmd = "v.clean",
		input = sources(x),
		output = src,
		tool = method,
		flags = c("c", .quiet(), "overwrite")
	)

	if (!is.null(tolerance)) args$threshold <- tolerance
	do.call(rgrass::execGRASS, args = args)

	out <- makeGVector(src)
	if (!is.null(table)) {
	
		if (ngeom(out) == nrow(table)) out@table <- table
		methods::validObject(out)
	
	}
	out

}

#' @title GRASS "mapset" of an object or the active session
#'
#' @description **GRASS** "locations" are sets of one or more rasters and/or vectors with the same coordinate reference systems, and may or may not represent the same actual location on Earth. **GRASS** "mapsets" are like subfolders of locations, and are collections of rasters and/or vectors typically related to the same general project. This function returns the mapset of an object or the current mapset. This said, **fasterRaster** always uses the "PERMANENT" mapset, so there is very little reason to use this function as-is. See `vignette("projects_mapsets", package = "fasterRaster")`.
#'
#' @param x A Either:
#' * Missing: Reports mapset of currently active **GRASS** "projects"/"locations".
#'  * A `GLocation` object or an object that contains the `GLocation` class (i.e., a `GSpatial` object: a `GRaster` or `GVector`): Reports the CRS of the object.
#'
#' @return A character string.
#'
#' @seealso **GRASS** [locations and mapsets](https://grass.osgeo.org/grass82/manuals/grass_database.html)
#'
#' @aliases .mapset
#' @rdname mapset
#' @keywords internal
methods::setMethod(
	f = ".mapset",
	signature = "GLocation",
	definition = function(x) unname(x@mapset)
)

#' @aliases .mapset
#' @rdname mapset
#' @keywords internal
methods::setMethod(
	f = ".mapset",
	signature = "missing",
	definition = function(x) {
	
	rgrass::execGRASS("g.mapset", flags = c(.quiet(), 'p'), intern = TRUE)
	
	} # EOF
)

#' Meta-data on active GRASS locations
#'
#' @description **GRASS** uses "locations"to store sets of rasters and vectors with the same coordinate reference system (CRS). These rasters and vectors may or may not be in the same actual location on Earth--they just have the same CRS. This function returns information on all of the **GRASS** "locations" that have been initialized. It is mainly useful for developers.
#'
#' NB: **fasterRaster** always uses the "PERMANENT" mapset.
#'
#' @param warn Logical: If `TRUE` (default), display a warning if no **GRASS** "locations" have been created.
#'
#' @returns A named `list`. The names are the "location's" names and the values are the coordinate reference strings.
#'
#' @aliases .locations
#' @rdname locations
#' @keywords internal
.locations <- function(warn = TRUE) {

	if (!grassStarted()) {
	
		if (warn) warning("No GRASS locations have been created.")
		out <- NULL
		
	} else {
		out <- .fasterRaster$locations
	}
	out

}

#' Metadata on rasters and vectors in GRASS
#'
#' @description `.rastInfo()` and `.vectInfo()` fetch metadata on rasters and vectors in **GRASS**. The `print()`, `show()`, and `summary()` functions can be used to display this metadata.
#'
#' @param x A `GRaster`, `GVector`, or `sources`.
#'
#' @returns Metadata on the extent, dimensions, resolution, bottom/top, etc. of rasters and vectors in **GRASS**.
#'
#' @keywords internal
.rastInfo <- function(x) {

	src <- if (!inherits(x, "character")) {
		sources(x)
	} else {
		x
	}

	### more than one raster
	#######################
	if (length(src) > 1L) {
		
		for (i in seq_along(src)) {
		
			this <- .rastInfo(src[i])
			out <- if (i == 1L) {
				this
			} else {
    			omnibus::appendLists(out, this)
			}
		
		}
	
	} else {
	### a single source
	##################

		rasters <- .ls(c("rasters", "rasters3d"))
		type <- names(rasters[rasters == src])

		### 2D raster
		if (type == "raster") {

			suppressMessages(
				niceInfo <- rgrass::execGRASS(
					"r.info",
					flags = .quiet(),
					map = src,
					intern = TRUE,
					Sys_show.output.on.console = FALSE,
					echoCmd = FALSE
				)
			)

			suppressMessages(
				extentInfo <- rgrass::execGRASS(
					"r.info",
					flags = c("g", .quiet()),
					map = src,
					intern = TRUE,
					Sys_show.output.on.console = FALSE,
					echoCmd = FALSE
				)
			)

			suppressMessages(
				rangeInfo <- rgrass::execGRASS(
					"r.info",
					flags = c("r", .quiet()),
					map = src,
					intern = TRUE,
					Sys_show.output.on.console = FALSE,
					echoCmd = FALSE
				)
			)
			
			topology <- "2D"

			ztop <- NA_real_
			zbottom <- NA_real_
			
		### 3D raster
		} else if (type == "raster3d") {

			suppressMessages(
				niceInfo <- rgrass::execGRASS(
					"r3.info",
					flags = c(.quiet()),
					map = src,
					intern = TRUE,
					Sys_show.output.on.console = FALSE,
					echoCmd = FALSE
				)
			)

			suppressMessages(
				extentInfo <- rgrass::execGRASS(
					"r3.info",
					flags = c("g", .quiet()),
					map = src,
					intern = TRUE,
					Sys_show.output.on.console = FALSE,
					echoCmd = FALSE
				)
			)
		
			suppressMessages(
				rangeInfo <- rgrass::execGRASS(
					"r3.info",
					flags = c("r", .quiet()),
					map = src,
					intern = TRUE,
					Sys_show.output.on.console = FALSE,
					echoCmd = FALSE
				)
			)

			topology <- "3D"
		
			ztop <- extentInfo[grepl(extentInfo, pattern = "top=")]
   			zbottom <- extentInfo[grepl(extentInfo, pattern = "bottom=")]
			
			ztop <- sub(ztop, pattern="top=", replacement="")
			zbottom <- sub(zbottom, pattern="bottom=", replacement="")
			
			ztop <- as.numeric(ztop)
			zbottom <- as.numeric(zbottom)
			
		}

		# projection
		index <- grepl(niceInfo, pattern = " \\|        Projection: ")
		projection <- niceInfo[index]
		projection <- sub(projection, pattern = " \\|        Projection: ", replacement = "")
		projection <- substr(projection, 1L, nchar(projection) - 1L)
		projection <- trimws(projection)
		
		# extent
		west <- extentInfo[grepl(extentInfo, pattern = "west=")]
		east <- extentInfo[grepl(extentInfo, pattern = "east=")]
		south <- extentInfo[grepl(extentInfo, pattern = "south=")]
		north <- extentInfo[grepl(extentInfo, pattern = "north=")]

		west <- sub(west, pattern="west=", replacement="")
		east <- sub(east, pattern="east=", replacement="")
		south <- sub(south, pattern="south=", replacement="")
		north <- sub(north, pattern="north=", replacement="")
		
		west <- as.numeric(west)
		east <- as.numeric(east)
		south <- as.numeric(south)
		north <- as.numeric(north)

		# dimensions
		rows <- extentInfo[grepl(extentInfo, pattern="rows=")]
		cols <- extentInfo[grepl(extentInfo, pattern="cols=")]
		depths <- extentInfo[grepl(extentInfo, pattern="depths=")]

		rows <- sub(rows, pattern="rows=", replacement="")
		cols <- sub(cols, pattern="cols=", replacement="")
		depths <- sub(depths, pattern="depths=", replacement="")
		
		rows <- as.integer(rows)
		cols <- as.integer(cols)
		depths <- as.integer(depths)
		if (length(depths) == 0L) depths <- NA_integer_
		
		# resolution
		ewres <- extentInfo[grepl(extentInfo, pattern="ewres=")]
		nsres <- extentInfo[grepl(extentInfo, pattern="nsres=")]
		tbres <- extentInfo[grepl(extentInfo, pattern="tbres=")]

		ewres <- sub(ewres, pattern="ewres=", replacement="")
		nsres <- sub(nsres, pattern="nsres=", replacement="")
		tbres <- sub(tbres, pattern="tbres=", replacement="")
		
		ewres <- as.numeric(ewres)
		nsres <- as.numeric(nsres)
		tbres <- as.numeric(tbres)
		if (length(tbres) == 0L) tbres <- NA_real_

		# data type
		grassDataType <- extentInfo[grepl(extentInfo, pattern="datatype=")]
		grassDataType <- sub(grassDataType, pattern="datatype=", replacement="")
		grassDataType <- gsub(grassDataType, pattern="\"", replacement="")

		# ### Any categories? Note that "r.info" incorrectly reports number of categories, so we will use r.info just to determine if there are any categories.
		# anyCats <- niceInfo[grepl(niceInfo, pattern = "Number of Categories:")]
		# start <- regexpr(anyCats, pattern = "Number of Categories: ")
    	# start <- start[1L] + nchar("Number of Categories: ")
		# anyCats <- substr(anyCats, 63, nchar(anyCats))
		# anyCats <- gsub(anyCats, pattern = "\\|", replacement = "")
		# anyCats <- as.integer(anyCats)

		# if (anyCats == 0L) {
		# 	numCategories <- 0L
		# } else {
		
		# 	catinfo <- rgrass::execGRASS(
		# 		"r.category",
		# 		map = src,
		# 		flags = .quiet(),
		# 		separator = "pipe",
		# 		intern = TRUE
		# 	)

		# 	numCategories <- length(catinfo)
		# 	numCategories <- as.integer(numCategories)
		
		# }

		# mininum/maximum value
		minVal <- rangeInfo[grepl(rangeInfo, pattern="min=")]
		maxVal <- rangeInfo[grepl(rangeInfo, pattern="max=")]
		
		minVal <- sub(minVal, pattern="min=", replacement="")
		maxVal <- sub(maxVal, pattern="max=", replacement="")
		
		minVal <- if (minVal == "NULL") { NA_real_ } else { as.numeric(minVal) }
		maxVal <- if (maxVal == "NULL") { NA_real_ } else { as.numeric(maxVal) }

		out <- list(
			sources = src,
			type = type,
			topology = topology,
			projection = projection,
			
			west = west,
			east = east,
			south = south,
			north = north,
			
			ztop = ztop,
			zbottom = zbottom,
		
			rows = rows,
			cols = cols,
			depths = depths,
		
			ewres = ewres,
			nsres = nsres,
			tbres = tbres,
			
			grassDataType = grassDataType,
			
			# numCategories = numCategories,
			minVal = minVal,
			maxVal = maxVal
			
		)
		
	} # if just one raster

	class(out) <- "rastInfo"
	out

}

#' @method print rastInfo
#' @aliases print
#' @rdname show
print.rastInfo <- function(x, ...) {

	cat("Source(s)     :", x$sources, "\n")
	cat("Type(s)       :", x$type, "\n")
	cat("Topology      :", x$topology, "\n")
	cat("Projection    :", x$projection, "\n")
	cat("Resolution    :", x$ewres, x$nsres, x$tbres, "(x, y, z)\n")
	cat("Dimensions    :", x$rows, x$cols, x$depths, "(rows, cols, depths)\n")
	cat("Extent (WESN) :", paste(x$west, x$east, x$south, x$north, collapse = ", "), "\n")
	cat("Z extent (BT) :", paste(x$zbottom, x$ztop, collapse = ", "), "\n")
	cat("Datatype      :", x$grassDataType, "\n")
	cat("Min/max       :", x$minVal, x$maxVal, "\n")

}

#' @rdname show
#' @keywords internal
show.rastInfo <- function(x) print.rastInfo(x)

# #' @rdname show
# #' @keywords internal
# summary.rastInfo <- function(x) print.rastInfo(x)

#' Metadata on a vector in GRASS
#'
#' @description This function queries **GRASS** to obtain metadata on a vector.
#'
#' @param x A `GVector` or the [sources()] name of one.
#' @param integer Logical. If `TRUE` (default), the "cats" of a vector will be returned as type integer. If `FALSE`, they will be returned as a character vector.
#' @param cats `NULL` (default) or an integer vector of category numbers, one per geometry.
#'
#' @returns A `vectInfo` object (a list).
#' @keywords internal
.vectInfo <- function(x, integer = TRUE, cats = NULL) {

	src <- if (inherits(x, "GVector")) {
		sources(x)
	} else {
		x
	}
	
	# rgrass::execGRASS("g.region", vector = x, flags=c("o", .quiet()))

	### extent/topology
	suppressMessages(
		extentInfo <- rgrass::execGRASS(
			"v.info",
			flags = c("g", .quiet()),
			map = src,
			intern = TRUE,
			Sys_show.output.on.console = FALSE,
			echoCmd = FALSE
		)
	)
	
	### geometry
	suppressMessages(
		geomInfo <- rgrass::execGRASS(
			"v.info",
			map = src,
			flags = c("t", .quiet()),
			intern = TRUE,
			Sys_show.output.on.console = FALSE,
			echoCmd = FALSE
		)
	)

	# extent
	west <- extentInfo[grepl(extentInfo, pattern = "west=")]
	east <- extentInfo[grepl(extentInfo, pattern = "east=")]
	south <- extentInfo[grepl(extentInfo, pattern = "south=")]
	north <- extentInfo[grepl(extentInfo, pattern = "north=")]

	west <- sub(west, pattern = "west=", replacement = "")
	east <- sub(east, pattern = "east=", replacement = "")
	south <- sub(south, pattern = "south=", replacement = "")
	north <- sub(north, pattern = "north=", replacement = "")

	west <- as.numeric(west)
	east <- as.numeric(east)
	south <- as.numeric(south)
	north <- as.numeric(north)

	# topology (2/3D)
	topology <- geomInfo[grepl(geomInfo, pattern = "map3d=")]
	topology <- sub(topology, pattern = "map3d=", replacement = "")
	topology <- if (topology == "0") {
		"2D"
	} else if (topology == "1") {
		"3D"
	}

	# top/bottom
	if (topology == "2D") {
		ztop <- zbottom <- NA_real_
	} else {
		zbottom <- extentInfo[grepl(extentInfo, pattern = "bottom=")]
		ztop <- extentInfo[grepl(extentInfo, pattern = "top=")]

		zbottom <- sub(zbottom, pattern = "bottom=", replacement = "")
		ztop <- sub(ztop, pattern = "top=", replacement = "")

		zbottom <- as.numeric(zbottom)
		ztop <- as.numeric(ztop)
	}
 
	# geometry (points/lines/polygons)
	points <- geomInfo[grepl(geomInfo, pattern="points=")]
	lines <- geomInfo[grepl(geomInfo, pattern="lines=")]
	boundaries <- geomInfo[grepl(geomInfo, pattern="boundaries=")]
	centroids <- geomInfo[grepl(geomInfo, pattern="centroids=")]
	areas <- geomInfo[grepl(geomInfo, pattern="areas=")]

	points <- sub(points, pattern="points=", replacement="")
	lines <- sub(lines, pattern="lines=", replacement="")
	boundaries <- sub(boundaries, pattern="boundaries=", replacement="")
	centroids <- sub(centroids, pattern="centroids=", replacement="")
	areas <- sub(areas, pattern="areas=", replacement="")

	points <- as.integer(points)
	lines <- as.integer(lines)
	boundaries <- as.integer(boundaries)
	centroids <- as.integer(centroids)
	areas <- as.integer(areas)

	if (points > 0 & lines == 0 & boundaries == 0 & centroids == 0 & areas == 0) {
		geometry <- "points"
		# nGeometries <- points
	} else if (points == 0 & lines > 0 & boundaries == 0 & centroids == 0 & areas == 0) {
		geometry <- "lines"
		# nGeometries <- lines
	} else if (points == 0 & lines == 0 & boundaries > 0 & centroids > 0 & areas > 0) {
		geometry <- "polygons"
		# nGeometries <- centroids
	} else {
		geometry <- NA_character_
		# nGeometries <- NA_integer_
	}
	
	# nGeometries <- as.integer(nGeometries)
	if (is.null(cats)) cats <- .vCats(src, integer = integer)
	nGeometries <- length(unique(cats))
	catsValid <- !any(grepl(cats, pattern = "[/]")) & !anyNA(cats)

	# ### fields
	# suppressWarnings(
	# 	fieldinfo <- rgrass::execGRASS(
	# 		"v.info",
	# 		flags = c("c", .quiet()),
	# 		map = src,
	# 		intern = TRUE,
	# 		Sys_show.output.on.console = FALSE,
	# 		echoCmd = FALSE
	# 	)
	# )
	
	# fieldData <- rgrass::execGRASS(
	# 	"v.db.connect",
	# 	map = src,
	# 	flags = c("g", .quiet()),
	# 	intern = TRUE
	# )

	suppressMessages(
		attribMetaInfo <- rgrass::execGRASS(
			"v.info",
			flags = "e",
			map = src,
			intern = TRUE,
			Sys_show.output.on.console = FALSE,
			echoCmd = FALSE
		)
	)

	# hasNoFields <- any(grepl(fieldData, pattern="is not connected to a database")) | any(grepl(fieldinfo, pattern="ERROR: Database connection not defined for layer"))
	
	# if (hasNoFields) {

	# 	nFields <- as.integer(0)
	# 	dbLayer <- NA_character_
	# 	fields <- NA_character_
	# 	classes <- NA_character_

	# # vector has a data frame
	# } else {

	# 	# fields and types
	# 	fieldinfo <- fieldinfo[!grepl(fieldinfo, pattern ="Displaying column types/names for database connection of layer")]
		
	# 	fieldinfo <- strsplit(fieldinfo, "\\|")
	# 	fields <- classes <- rep(NA, length(fieldinfo))
		
	# 	for (i in seq_along(fieldinfo)) {
		
	# 		type <- fieldinfo[[i]][1L]
	# 		type <- if (type == "CHARACTER") {
	# 			"character"
	# 		} else if (type == "INTEGER") {
	# 			"integer"
	# 		} else if (type == "DOUBLE PRECISION") {
	# 			"numeric"
	# 		} else {
	# 			NA_character_
	# 		}
		
	# 		classes[i] <- type
	# 		fields[i] <- fieldinfo[[i]][2L]
	# 	}

	# 	cat <- which(fields == "cat")
	# 	fields <- fields[-cat]
	# 	classes <- classes[-cat]

	# 	nFields <- length(fields)
	# 	nFields <- as.integer(nFields)

  	# 	dbLayer <- attribMetaInfo[grepl(attribMetaInfo, pattern = "attribute_layer_name")]
	# 	dbLayer <- if (length(dbLayer) == 0) {
	# 		NA_character_
	# 	} else {
	# 		gsub(dbLayer, pattern="attribute_layer_name=", replacement="")
	# 	}

	# }

  	projection <- attribMetaInfo[grepl(attribMetaInfo, pattern = "projection=")]
 	projection <- sub(projection, pattern = "projection=", replacement = "")
		
	out <- list(
		sources = src,
		type = "vector",
		projection = projection,
		topology = topology,
		geometry = geometry,
		
		cats = cats,
		catsValid = catsValid,
		nGeometries = nGeometries,
		
		west = west,
		east = east,
		south = south,
		north = north,
	
		zbottom = zbottom,
		ztop = ztop
		
		# nFields = nFields,
		# dbLayer = dbLayer,
		# fields = fields,
		# classes = classes
	)
	
	class(out) <- "vectInfo"
	out

}

#' @method print vectInfo
#' @aliases print
#' @rdname show
#' @keywords internal
print.vectInfo <- function(x, ...) {

	cats <- x$cats
	if (length(cats) > 6L) {
		cats <- paste0(paste(head(x$cats), collapse = ", "), ", ...\n")
	} else {
		cats <- paste0(paste(head(x$cats), collapse = ", "), "\n")
	}

	cat("Source        :", x$sources, "\n")
	cat("Geometry      :", x$geometry, "\n")
	cat("Projection    :", x$projection, "\n")
	cat("Topology      :", x$topology, "\n")
	cat("Extent (WESN) :", paste(x$west, x$east, x$south, x$north, collapse = ", "), "\n")
	cat("Z extent (BT) :", paste(x$zbottom, x$ztop, collapse = ", "), "\n")
	cat("Geometries    :", x$nGeometries, "\n")
	cat("Subgeometries :", length(x$cats), "\n")
	cat("Cats          :", cats)
	cat("Cats valid    :", x$catsValid, "\n")

}

# #' @rdname show
# #' @keywords internal
# summary.vectInfo <- function(x) print(x)

#' @method show vectInfo
#' @aliases show
#' @rdname show
#' @keywords internal
show.vectInfo <- function(x) print(x)

#' Rescale values in a GRaster
#' 
#' @description `stretch()` rescales the values in a `GRaster`. All values can be rescaled, or just values in a user-defined range. This range can be given by specifying either the lower and upper bounds of the range using `smin` and `smax`, and/or by the quantiles (across all cells of the raster) using `minq` and `maxq`.
#' 
#' @param x A `GRaster`.
#' 
#' @param minv,maxv Numeric: Minimum and maximum values to which to rescale values.
#' 
#' @param minq,maxq Numeric: Specifies range of values to rescale, given by their quantiles. The default is to stretch all values (the 0th and 100th quantiles). One or both are  ignored if `smin` and/or `smax` are provided.
#' 
#' @param smin,smax Numeric or `NA`: Specifies range of values to rescale. If `NA` (default), then all values are rescaled.
#' 
#' @returns A `GRaster`.
#' 
#' @seealso [terra::stretch()] and tool `r.rescale` in **GRASS** (not used on this function)
#' 
#' @example man/examples/ex_stretch.r
#' 
#' @aliases stretch
#' @rdname stretch
#' @exportMethod stretch
methods::setMethod(
    f = "stretch",
    signature = c(x = "GRaster"),
    definition = function(
        x,
        minv = 0, maxv = 255,
        minq = 0, maxq = 1,
        smin = NA, smax = NA
    ) {

    if (is.na(minv)) stop("Invalid lower bound to which to stretch.")
    if (is.na(maxv)) stop("Invalid upper bound to which to stretch.")

    if (is.na(minq) & is.na(smin)) stop("Invalid lower bound of range to stretch.")
    if (!is.na(minq) && minq < 0) stop("Invalid lower bound of range to stretch.")

    if (is.na(maxq) & is.na(smax)) stop("Invalid upper bound of range to stretch.")
    if (!is.na(maxq) && maxq > 1) stop("Invalid upper bound of range to stretch.")

    .locationRestore(x)
    .region(x)
	
	versionNumber <- grassInfo("versionNumber")

    mm <- minmax(x)
	
    for (i in 1L:nlyr(x)) {
        
        ### lower "from" bound
        if (!is.na(smin)) {
            lowerFrom <- smin
        } else {
            
            if (minq == 0) {
                lowerFrom <- mm[1L, i]
            } else {

                # get lower quantile
                args <- list(
                    cmd = "r.univar",
                    flags = c(.quiet(), "r", "e"),
                    map = sources(x)[i],
                    Sys_show.output.on.console = FALSE,
                    echoCmd = FALSE,
                    intern = TRUE,
                    percentile = minq * 100
                )

                # nprocs became available with GRASS eight point three
                if (versionNumber >= 8.3) args$nprocs <- faster("cores")

                info <- do.call(rgrass::execGRASS, args)
                pattern <- "percentile: "
                lowerFrom <- info[grepl(info, pattern=pattern)]
                lowerFrom <- strsplit(lowerFrom, split=":")[[1L]][2L]
                lowerFrom <- as.numeric(lowerFrom)

            } # if calculating quantile
        } # if calculating lower quantile

        ### upper "from" bound
        if (!is.na(smax)) {
            upperFrom <- smax
        } else {
            
            if (maxq == 1) {
                upperFrom <- mm[2L, i]
            } else {

                # get lower quantile
                args <- list(
                    cmd = "r.univar",
                    flags = c(.quiet(), "r", "e"),
                    map = sources(x)[i],
                    Sys_show.output.on.console = FALSE,
                    echoCmd = FALSE,
                    intern = TRUE,
                    percentile = maxq * 100
                )

                # nprocs became available with GRASS eight point three
                if (versionNumber >= 8.3) args$nprocs <- faster("cores")

                info <- do.call(rgrass::execGRASS, args)
                pattern <- "percentile: "
                upperFrom <- info[grepl(info, pattern=pattern)]
                upperFrom <- strsplit(upperFrom, split=":")[[1L]][2L]
                upperFrom <- as.numeric(upperFrom)

            } # if calculating quantile
        } # if calculating upper quantile


        ### truncate values at min/max
        if (mm[1L, i] < lowerFrom | mm[2L, i] > upperFrom) {

            gnTrunc <- .makeSourceName("truncated", "rast")
            ex <- paste0(gnTrunc, " = if(", sources(x)[i], " < ", lowerFrom, ", ", lowerFrom, ", if(", sources(x)[i], " > ", upperFrom, ", ", upperFrom, ", ", sources(x)[i], "))")

            args <- list(
                cmd = "r.mapcalc",
                expression = ex,
                flags = c(.quiet(), "overwrite"),
                intern = TRUE
            )

            do.call(rgrass::execGRASS, args=args)

        } else {
            gnTrunc <- sources(x)[i]
        }

        src <- .makeSourceName("stretch", "rast")
        scale <- (maxv - minv) / (upperFrom - lowerFrom)
        ex <- paste0(src, " = ", minv, " + (", scale, " * (", gnTrunc, " - ", lowerFrom, "))")
        
        args <- list(
            cmd = "r.mapcalc",
            expression = ex,
            flags = c(.quiet(), "overwrite"),
            intern = TRUE
        )

        do.call(rgrass::execGRASS, args=args)

        this <- makeGRaster(src, names(x)[i])
        if (i == 1L) {
            out <- this
        } else {
            out <- c(out, this)
        }

    } # next layer
    out

    } # EOF
)

if (grassStarted()) {

# Setup
library(terra)

# Example data
madChelsa <- fastData("madChelsa")

# Convert a SpatRaster to a GRaster:
chelsa <- fast(madChelsa)

# Correlation
layerCor(chelsa, "cor")

# Covariance
layerCor(chelsa, "cov")

}

#' Remove rows in a data.table, data.frame, or matrix.
#'
#' @description As of September of 2023, the **data.table** package does not have a function for removing rows by index. This function does this job.
#'
#' @param x A `data.table` or `data.frame`.
#' @param drops Numeric, integer, or logical vector: Indices or indicators of rows to remove.
#'
#' If a logical vector is supplied, rows that correspond to `TRUE` will be removed. If the vector is shorter than the number of rows, values of `drops` will be recycled.
#'
#' @returns A `data.table` or `data.frame`.
#'
#' @example man/examples/ex_data_table.r
#'
#' @aliases dropRows
#' @rdname dropRows
#' @exportMethod dropRows
methods::setMethod(
	f = "dropRows",
	signature = c(x = "data.table"),
	function(x, drops) {

	if (nrow(x) > 0L) {

  		drops <- .rowIndices(x, select = drops)

		# select row indexes to keep
		keeps <- setdiff(x[ , .I], drops)

		cols <- names(x)
		
		# this is the subsetted table
		xSubset <- data.table::data.table(x[[1L]][keeps]) 
  		# x[ , 1L := NULL] # delete
		
		data.table::setnames(xSubset, cols[1L])

		if (length(cols) > 1L) {
			for (col in cols[2L:length(cols)]) {
    			these <- x[[col]][keeps]
				xSubset[ , (col) := these]
				# x[ , (col) := NULL] # delete
			}
		}
		out <- xSubset
	
	} else {
		out <- x # has no rows
	}
	out

	} # EOF
)

#' @aliases dropRows
#' @rdname dropRows
#' @exportMethod dropRows
methods::setMethod(
	f = "dropRows",
	signature = c(x = "data.frame"),
	function(x, drops) {
	
	if (nrow(x) > 0L) {

  		drops <- .rowIndices(x, select = drops)
		out <- x[-drops, , drop = FALSE]

	} else {
		out <- x # has no rows
	}
	out
		
	} # EOF
)

#' @aliases dropRows
#' @rdname dropRows
#' @exportMethod dropRows
methods::setMethod(
	f = "dropRows",
	signature = c(x = "matrix"),
	function(x, drops) {
	
	if (nrow(x) > 0L) {

  		drops <- .rowIndices(x, select = drops)
		out <- x[-drops, , drop = FALSE]

	} else {
		out <- x # has no rows
	}
	out
		
	} # EOF
)

#' Get indices of rows to remove
#'
#' @param x A `data.frame` or `data.table` or `matrix`.
#' @param select Numeric, integer, logical, or character.
#'
#' @noRd
 .rowIndices <- function(x, select) {
     if (is.logical(select)) {
         if (length(select) != nrow(x)) {
             select <- rep(select, length.out = nrow(x))
         }
         select <- which(select)
     } else if (is.character(select)) {
         if (inherits(x, "data.table")) stop("Row names can only be used with a data.frame or matrix, not with a data.table.")
		 rns <- rownames(x)
         select <- match(select, rns)
     }
     select
 }

if (grassStarted()) {

# Setup
library(terra)

# Climate raster:
madChelsa <- fastData("madChelsa")

# Convert a SpatRaster to a GRaster:
chelsa <- fast(madChelsa)

### Denoise:
quiet <- denoise(chelsa, scale = TRUE)

compare1 <- c(chelsa[["bio1"]], quiet[["bio1"]])
plot(compare1)

compare2 <- c(chelsa[["bio7"]], quiet[["bio7"]])
plot(compare2)

### Noise:
loud <- noise(chelsa, scale = TRUE)

compare1 <- c(chelsa[["bio1"]], loud[["bio1"]])
plot(compare1)

compare2 <- c(chelsa[["bio7"]], loud[["bio7"]])
plot(compare2)

}

#' Format a numeric series into an SQL value call
#'
#' @description This function takes as its argument a vector of integers or numeric values, and converts sequential runs to a range while keeping non-sequential values as-is. For example, `c(1, 5, 6, 7, 8, 9, 15, 16, 20)` becomes `"1,5-9,15-16,20"`. This reduces the number of characters necessary to supply to a SQL condition. This function is mainly of use to developers.
#'
#' @param x A vector of numerical values. The vector should be sorted from  lowers to highest for the most efficient "compression" of sequential ranges. Values will be coerced to class `integer`.
#'
#' @param maxChar Integer or numeric: Maximum number of characters to include in the output. If the output has more than this number of characters, the remainder is dropped, and the `trim` attribute of the output is set to `TRUE`. The default is 29900, which is the maximum length of an SQL statement that **GRASS** seems to be able to handle (minus a safety margin).
#'
#' @param sort Logical: If `TRUE` (default), sort `x` before converting to SQL. This can reduce the length of the output.
#'
#' @returns A character string. The string has three attributes. The `trim` attribute is `TRUE` or `FALSE`, depending on whether `maxChar` was reached or not (and subsequent numbers dropped from the string). The `lastIndex` attribute is the last index of `x` that was processed (i.e., the index of the last value in the output), and the number of values represented by the output.
#'
#' @examples
#'
#' x <- 1:5
#' seqToSQL(x)
#'
#' x <- c(1:5, 7)
#' seqToSQL(x)
#'
#' x <- c(1:5, 7, 15:16)
#' y <- c(1:5, 7, 15:16, 20)
#' seqToSQL(x)
#' seqToSQL(y)
#' 
#' seqToSQL(x, maxChar = 5)
#' seqToSQL(y, maxChar = 8)
#'
#' seqToSQL(10:1, sort = FALSE)
#' seqToSQL(10:1, sort = TRUE)
#'
#' @export
seqToSQL <- function(x, maxChar = 29900, sort = TRUE) {

	if (sort) x <- sort(x)
	x <- as.integer(x)
	x <- unique(x)
	n <- length(x)
	if (n == 0L) {
		out <- NULL
		trim <- NA		
		lastIndex <- NA_integer_
	} else if (n == 1L) {
		out <- as.character(x)
		trim <- FALSE
		lastIndex <- 1L
	} else {

		nMinus1 <- n - 1L

		out <- character()
		inseq <- FALSE
		xchar <- as.character(x)
		
		i <- 1L
		nc <- 0L
		while (nc <= maxChar & i <= nMinus1) {
		
			nextX <- x[i + 1L]
			nextInSeries <- x[i] + 1L
		
			# if sequential
			if (nextX == nextInSeries) {
			
				# if starting new series
				if (!inseq) {
					from <- xchar[i]
					inseq <- TRUE
				}
			
			} else {
			
				if (inseq) {
					to <- xchar[i]
					out <- paste(c(out, paste0(from, "-", to)), collapse = ",") 
					inseq <- FALSE
				} else {
					out <- paste(c(out, xchar[i]), collapse = ",") 
				}
			
			}
			
   			# hacky
			nc <- max(length(out), nchar(out) + inseq + 1L)
			i <- i + 1L
		
		} # next x
		
		## ended prematurely because string was too long
		if (nc > maxChar) {
		
		   	trim <- TRUE
			lastIndex <- i - 1L
		
		## got through the entire set
		} else {
			
			if (inseq && x[n - 1L] + 1L == x[n]) {
			
				to <- xchar[n]
				out <- paste(c(out, paste0(from, "-", to)), collapse = ",") 
				
			} else {
				# out <- c(out, as.character(c(n - 1L, n)))
				out <- paste(c(out, xchar[n]), collapse = ",") 
			}
			
			trim <- FALSE
			lastIndex <- n
			
		}
		
	}
	
	attr(out, "trim") <- trim
	attr(out, "lastIndex") <- lastIndex
	out

}

#' Save a GVector to disk
#'
#' @description This function saves a `GVector` to disk directly from a **GRASS** session.
#'
#' By default, files will be of OGC GeoPackage format (extension "`.gpkg`"), but this can be changed with the `format` argument. You can see a list of supported formats by simply using this function with no arguments, as in `writeVector()`, or by consulting the online help page for **GRASS** tool `v.out.ogr` (see `grassHelp("v.out.ogr")`).
#'
#' Note that if the vector has a data table attached and at least one numeric or integer column has an `NA` or `NaN` value, the function will yield a warning like:
#' ```
#' Warning 1: Invalid value type found in record 2 for field column_with_NA_or_NaN. This warning will no longer be emitted.
#' ```
#' Also note that despite the promise, this warning will be displayed again.
#'
#' @param x A `GVector`.
#'
#' @param filename Character: Path and file name.
#'
#' @param overwrite Logical: If `FALSE` (default), do not save over existing files.
#'
#' @param format Character or `NULL`: File format. If `NULL` (default), then the function will attempt to get the format from the file name extension. Partial matching is used and case is ignored. You can see a list of formats using `writeVector()` (no arguments). Some common formats include:
#'	* `"GPKG"`: OGC GeoPackage (extension `.gpkg`).
#'	* `"CSV"`: Comma-separated value... saves the data table only, not the geometries (extension `.csv`).
#'	* `"ESRI Shapefile"`: ESRI shapefile (extension `.shp`).
#'	* `"GeoJSON"`: GeoJSON (extension `GeoJSON`)
#'	* `"KML"`: Keyhole Markup Language (extension `.kml`)
#'	* `"netCDF"`: NetCDF (extension `.ncdf`)
#'	* `"XLSX"`: MS Office Open XML spreadsheet (extension `.xlsx`).
#'
#' @param attachTable Logical: If `TRUE` (default), attach the attribute to table to the vector before saving it. If `FALSE`, the attribute table will not be attached.
#'
#' @param ... Additional arguments to send to **GRASS** tool `v.out.ogr` (see `grassHelp("v.out.ogr")`).
#'
#' @returns Invisibly returns a `GRaster` (the input, `x`). Also saves the vector to disk.
#'
#' @seealso [terra::writeVector()], [sf::st_write()], **GRASS** tool `v.out.ogr` (see `grassHelp("v.out.ogr")`)
#'
#' @example man/examples/ex_writeVector.r
#'
#' @seealso [terra::writeVector()], the **GRASS** tool manual page for `v.out.ogr` (see `grassHelp("v.out.ogr")`)
#'
#' @aliases writeVector
#' @rdname writeVector
#' @exportMethod writeVector
setMethod(
	"writeVector",
	signature(x = "GVector", filename = "character"),
	function(
		x,
		filename,
		overwrite = FALSE,
		format = NULL,
		attachTable = TRUE,
		...
	) {

	### going to overwrite anything?
	if (!overwrite && file.exists(filename)) stop(paste0("File already exists and `overwrite` is FALSE:\n ", filename))

	.locationRestore(x)

	filename <- normalizePath(filename, mustWork = FALSE)

	if (attachTable & nrow(x) > 0L) {

		table <- as.data.frame(x)
		.vAttachDatabase(x, table = table, replace = TRUE)

	}

	### general arguments
	args <- list(
		cmd = "v.out.ogr",
		input = sources(x),
		output = filename,
		flags = c(.quiet(), "s")
		# flags = c(.quiet(), "s", "c") # "c" ==> save geometries lacking a cat number
	)
	if (overwrite) args$flags <- c(args$flags, "overwrite")

	if (!is.null(format)) {

		formats <- writeVector()
		formats <- formats$extension

		format <- gsub(format, pattern = " ", replacement = "_")
		formats <- gsub(formats, pattern = " ", replacement = "_")

		format <- omnibus::pmatchSafe(format, formats, nmax = 1L)

	} else if (is.null(format)) {
	
		fileExt <- .fileExt(filename)
		fileExt <- tolower(fileExt)
		
		if (fileExt == "shp") {
			format <- "ESRI_Shapefile"
		} else if (fileExt == "gml") {
			format <- "GML"
		} else if (fileExt == "kml") {
			format <- "KML"
			args$type <- "face"
		} else if (fileExt == "ncdf") {
			format <- "netCDF"
		} else if (fileExt == "gpkg") {
			format <- "GPKG"
		} else if (fileExt == "csv") {
			format <- "CSV"
		} else if (fileExt == "pdf") {
			format <- "PDF"
		}

	}
	args$format <- format

	if (is.3d(x) & !is.null(format)) {

		if (format == "ESRI_Shapefile") {

			gtype <- geomtype(x)
			if (gtype == "points") {
				args$lco <- "SHPT=POINTZ"
			} else if (gtype == "lines") {
				args$lco <- "SHPT=ARCZ"
			} else if (gtype == "polygons") {
				args$lco <- "SHPT=POLYGONZ"
			}
		
		} else if (format == "KML") {
			args$dsco = "AltitudeMode=absolute"
		}
	
	}

	do.call(rgrass::execGRASS, args)
	
	# out <- suppressWarnings(terra::vect(filename))
	# if (nrow(x) > 0L && any(names(out) == "cat")) out$cat <- NULL
	# if (nrow(x) > 0L && any(names(out) == "cat_")) out$cat_ <- NULL

	# invisible(out)
	invisible(x)

	} # EOF
)

#' @aliases writeVector
#' @rdname writeVector
#' @exportMethod writeVector
setMethod(
	"writeVector",
	signature(x = "missing", filename = "missing"),
	function(x, filename) {

		forms <- rgrass::execGRASS("v.out.ogr", flags = "l", intern = TRUE, ignore.stderr = TRUE)

		forms <- forms[forms != "Supported formats:"]
		forms <- trimws(forms)
		forms <- sort(forms)
		forms <- strsplit(forms, ": ")
		forms <- do.call(rbind, forms)
		forms <- as.data.frame(forms)
		names(forms) <- c("extension", "filetype")
		forms$extension <- gsub(forms$extension, pattern = " \\(rw\\+\\)", replacement = "")
		forms$extension <- gsub(forms$extension, pattern = " \\(rw\\)", replacement = "")
		
		forms
		
	} # EOF
)

#' Subset geometries of a GVector
#'
#' @description The `[` operator returns a subset or remove specific geometries of a `GVector`. You can get the number of geometries using [ngeom()]. Note that you cannot use this function to change the "order" in which geometries or their associated records in a data table appear. For example, `vector[1:3]` and `vector[3:1]` will yield the exact same results.
#'
#' Note that subsetting can take a very long time if you are retaining only a small number of geometries from a vector with many geometries. The routine selects geometries by removing those that are not in `i`. So if you can write code to remove fewer geometries (i.e., an "inverse" selection), it may go faster.
#'
#' @param x A `GVector`.
#'
#' @param i Numeric integer, integer, or logical vector: Indicates which geometry(ies) to obtain. Negative numeric or integer values will remove the given geometries from the output. If a logical vector is supplied and it is not the same length as the number of geometries, it will be recycled.
#'
#' @param j Numeric integer, integer, logical, or character: Indices or name(s) of the column(s) to obtain. You can see column names using [names()]. Negative numeric or integer values will remove the given columns from the output. If a logical vector is supplied and it is not the same length as the number of columns, it will be recycled.
#'
#' @returns A `GVector`.
#'
#' @example man/examples/ex_GRaster_GVector_subset_assign.r
#'
#' @seealso [subset()], [$], \code{\link[fasterRaster]{[[}}
#'
#' @name [
#' @aliases [,GVector,ANY,ANY-method
#' @docType methods
#' @rdname subset_single_bracket
#' @exportMethod [
methods::setMethod(
	"[",
	signature = c(x = "GVector", i = "ANY", j = "ANY"),
	function(x, i, j) {

	.locationRestore(x)
	nGeoms <- ngeom(x)

	if (missing(i) & missing(j)) {
		out <- x
	} else if (missing(i) & !missing(j)) {
		out <- x[[i = j]]
	} else if (all(1L:nGeoms %in% i) & !missing(j)) {
		out <- x[[i = j]]
	} else if (all(1L:nGeoms %in% i) & missing(j)) {
		out <- x
	} else if (!missing(i)) {

		if (nrow(x) > 0 & ((is.numeric(i) | is.integer(i)) & any(i != order(i)))) {

			.message(
				msg = "subset_square_bracket",
				message =
"The GVector returned by `[` (subset_single_bracket) may have the order of items in\n  its data table changed from the order they appear in the index used to select features.\n  For example, if you use `vector[3:1]`,  `vector[1:3]`. If this\n  message appeared after you called a function other than `[`, you can probably ignore it."
			)

		}

		if (is.logical(i)) {
			if (length(i) < nGeoms) i <- rep(i, length.out = nGeoms)
			i <- which(i)
		}

		if (any(i > 0L) & any(i < 0L)) stop("Cannot mix positive and negative indices.")

		# negative indices... removing features
		if (all(i < 0L)) {

			reverseRowSelect <- TRUE
			iRev <- i
			iSeq <- seq_len(nGeoms) # all possible i's
			i <-  iSeq[i] # retained i's
			removeAll <- length(i) == 0L

		} else {
			# positive indices... keeping features
			reverseRowSelect <- removeAll <- FALSE
		}
		
		if (any(i > nGeoms) | any(i == 0L)) stop("Index out of bounds.")
		i <- sort(i) # NB MUST have this, or data table will be mis-matched to features!!!!!

		if (removeAll) {
			out <- NULL # removed all
		} else if (all(1L:nGeoms %in% i) & missing(j)) {
			return(x)
		} else {

			# # # # will be deleting geometries from this copy
			# # # src <- .copyGSpatial(x)
			# # # cats <- .vCats(src)
			# # # cats <- sort(unique(cats))
			# # # iNot <- seq_len(nGeoms)
			# # # iNot <- iNot[!(iNot %in% i)]
			# # # cats <- cats[iNot] # will be removed

			# # # # delete geometries in subsets bc removing large numbers of geometries at a time is inefficient
			# # # done <- FALSE
			# # # n <- 200000L # maximum number of cats to delete at a time (seqToSQL() usually stops far short)
			# # # # n <- 10L # maximum number of cats to delete at a time (seqToSQL() usually stops far short)
			# # # startFrom <- 1L
			# # # numRemoved <- 0L
			# # # stopAt <- min(length(cats), startFrom + n - 1L)
			# # # while (!done) {
				
			# # 	# thisCats <- cats[startFrom:stopAt]
			# # 	# thisCats <- seqToSQL(thisCats)

			# # # # <<< ORIGINAL -- selecting by deleting unwanted... VERY slow for large vectors >>>
			# # 	# rgrass::execGRASS(
			# # 		# cmd = "v.edit",
			# # 		# map = src,
			# # 		# # type = gtype, # breaks
			# # 		# tool = "delete",
			# # 		# cats = thisCats,
			# # 		# flags = c(.quiet(), "overwrite")
			# # 	# )
			# # # # <<< end ORIGINAL >>>

			# # 	# # args <- list(
			# # 	# # 	cmd = "v.edit",
			# # 	# # 	map = src,
			# # 	# # 	# type = gtype, # breaks
			# # 	# # 	tool = "delete",
			# # 	# # 	cats = thisCats,
			# # 	# # 	flags = c(.quiet(), "overwrite")
			# # 	# # )
			# # 	# # # if (geomtype(x) == "polygons") args$type <- "area"
			# # 	# # # if (geomtype(x) == "points") args$flags <- c(args$flags, "b")
			# # 	# # do.call(rgrass::execGRASS, args = args)

			# # 	# # rgrass::execGRASS("v.build", map = src, option = "build", flags = c(.quiet(), "overwrite"))

			# # 	# numRemoved <- numRemoved + attr(thisCats, "lastIndex")
			# # 	# if (numRemoved == length(cats)) {
			# # 		# done <- TRUE
			# # 	# } else {
			# # 		# startFrom <- startFrom + attr(thisCats, "lastIndex") # what was the index of the last used cat?
			# # 		# stopAt <- min(length(cats), startFrom + n)
			# # 	# }

			# # # } # next removal

			# # # # check to see if all cats have been removed
			# # # # a hack... for some reason, seqToSQL may not list all cats
			# # # newCats <- .vCats(src)
			# # # badCats <- cats[cats %in% newCats]
			# # # if (length(badCats) > 0L) {
			
			# # 	# badCats <- cats[cats %in% newCats]
			# # 	# thisCats <- seqToSQL(badCats)

			# # 	# rgrass::execGRASS(
			# # 		# cmd = "v.edit",
			# # 		# map = src,
			# # 		# # type = gtype, # breaks
			# # 		# tool = "delete",
			# # 		# cats = thisCats,
			# # 		# flags = c(.quiet(), "overwrite")
			# # 	# )

			# # 	# newCats <- newCats[!(newCats %in% badCats)]
			
			# # # }

			# ### WORKS!!!!! Keeping copy.
			# src <- .makeSourceName("subset_single_bracket_v_extract", "vector")
			# cats <- .vCats(sources(x))
			# cats <- unique(cats)
			# cats <- cats[cats %in% i] # selects same as GVector but wrong geometry
			# sqlCats <- seqToSQL(cats)
			# sqlCats <- as.character(sqlCats)
			# gtype <- geomtype(x, grass = TRUE)
			# ### ^^^ WORKS!!!!! Keeping copy.
			
			gtype <- geomtype(x, grass = TRUE)

			src <- .makeSourceName("subset_single_bracket_v_extract", "vector")
			cats <- .vCats(sources(x))
			cats <- unique(cats)
			keepCats <- cats[cats %in% i] # selects same as GVector but wrong geometry

			args <- list(
				cmd = "v.extract",
				input = sources(x),
				output = src,
				type = gtype,
				# flags = c(.quiet(), "overwrite")
				flags = c(.quiet(), "overwrite", "t") # "t" ==> Do not copy attributes
			)

			if (gtype == "point") {

				# selecting geometries using database method
				qid <- rep(0L, nGeoms)
				qid[keepCats] <- 1L

				db <- data.table::data.table(cat = cats, qid = qid)
				.vAttachDatabase(x, db, cats = cats)

				args$where <- "qid = 1" # will not work with inequalities!

			} else if (gtype %in% c("area", "line")) {
			
				sqlCats <- seqToSQL(keepCats)
				sqlCats <- as.character(sqlCats)
				args$cats <- sqlCats
			
			}

			do.call(rgrass::execGRASS, args = args)
			# src <- .vRecat(src, gtype = gtype, cats = keepCats) # using cats argument removes some point geometries!!!
			src <- .vRecat(src, gtype = gtype) # retains all point geometries

# 				srcs <- character()
# 				trim <- TRUE
# 				j <- 1L
# 				while (trim) {
				
# 					if (j > 1L) {
# 						thisKeepCats <- seqToSQL(keepCats[(stoppedAt + 1):length(i)])
# 						stoppedAt <- attr(thisKeepCats, "lastIndex")
# 						trim <- attr(thisKeepCats, "trim")
# 					}

# 					thisKeepCats <- as.character(thisKeepCats)

# 					srcs[j] <- .makeSourceName("single_bracket_select_v_extract", "vector")
# 					rgrass::execGRASS(
# 						cmd = "v.extract",
# 						input = sources(x),
# 						output = srcs[j],
# 						cats = thisKeepCats,
# 						type = gtype,
# 						flags = c(.quiet(), "overwrite")
# 					)

# 					j <- j + 1L
				
# 				}
			
			
			# }

			# renumber cats so they start from 1 and are sequential
			
		} # not removing all and not removing nothing

		### select data table rows and columns
		if (nrow(x) == 0L) {
			table <- NULL
		} else if (!removeAll) {

			table <- x@table

			# select columns
			if (missing(j)) {
				# not removing all columns
				removeAllCols <- FALSE
			} else {
				
				nc <- ncol(table)
				if (is.character(j)) {
					j <- match(j, names(table))
				} else if (is.logical(j)) {

					if (length(j) < nc) j <- rep(j, length.out = nc)
					j <- which(j)
				}
				
				if (all(j < 0L)) {
					reverseColSelect <- TRUE
					j <- j * -1L
					removeAllCols <-  (all(sort(j) == seq_len(ncol(x))))
				} else {
					reverseColSelect <- removeAllCols <- FALSE
				}

				if (any(j > nc)) stop("Index out of bounds.")

				if (removeAllCols) {
					table <- data.table::data.table(NULL)
				} else {
				
					# ..j <- NULL

					if (reverseColSelect) {
						j <- setdiff(names(x), names(x)[j])
					} else {
						j <- names(x)[j]
					}
					
					table <- table[ , ..j, with = FALSE]

				}
			
			}

			# select rows
			if (!removeAllCols) {
				if (reverseRowSelect) {
					table <- table[iRev]
				} else {
					table <- table[i]
				}
			}

		} # vector has table
		
		if (removeAll) {
			out <- NULL
		} else {
			# out <- makeGVector(src, table = table, cats = newCats)
			# out <- makeGVector(src, table = table)
			out <- makeGVector(src, table = table, cats = keepCats)
		}

	}
	out

	} # EOF
)

#' @name [
#' @aliases [,GRaster,GRaster,ANY-method
#' @docType methods
#' @rdname subset_single_bracket
#' @exportMethod [
methods::setMethod(
	"[",
	signature = c(x = "GRaster", i = "GRaster", j = "ANY"),
	function(x, i, j) {

	.locationRestore(x)
	.region(x)

	if (missing(j)) j <- NULL

	nLayers <- nlyr(x)
	nLayersI <- nlyr(i)

	if (nLayersI != nLayers) {

		if (nlyr(i) != 1L) warning("The number of layers in the subset raster (inside the `[]`) does not have\n  the same number of layers as the raster to be subset. The layers inside\n  the square brackets will be recycled.")

		iIndex <- rep(seq_len(nLayersI), length.out = nLayers)
		i <- i[[iIndex]]

	}

	srcs <- .makeSourceName("subset_single_bracket", "raster", nLayers)
	for (count in seq_len(nLayers)) {
			
		if (!(.minVal(i)[count] %in% c(NA, 0, 1)) & !(.maxVal(i)[count] %in% c(NA, 0, 1))) stop("The GRaster in `i` must 0, 1, or NA values.")

		ex <- paste0(srcs[count], " = if(", sources(i)[count], " == 1, ", sources(x)[count], ", 0)")
		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))
	
	}
	makeGRaster(srcs, names(x), levels = cats(x), ac = activeCats(x))

	} # EOF
)

#' Topographic wetness index
#'
#' @description This function creates a raster map with values equal to the topographic wetness index (TWI), which is a measure of how much overland water flow tends to accumulate in or flow away from a location.
#'
#' @param x A `GRaster` (typically representing elevation). The raster must be projected (i.e., not in WGS84, NAD83, et cetera).
#'
#' @returns A `GRaster`.
#'
#' @seealso [terrain()], [ruggedness()], [geomorphons()], **GRASS** manual for tool `r.topidx` (see `grassHelp("r.topidx")`)
#'
#' @example man/examples/ex_ruggedness_wetness.r
#'
#' @aliases wetness
#' @rdname wetness
#' @exportMethod wetness
methods::setMethod(
	f = "wetness",
	signature = c(x = "GRaster"),
	function(x) {
	
	if (.projection(x) == "Latitude-Longitude") stop("Raster must be projected before calculating TWI.")

	.locationRestore(x)
	.region(x)

	nLayers <- nlyr(x)
	srcs <- .makeSourceName("r_topidx", "raster", n = nLayers)

	for (i in seq_len(nLayers)) {
	
		rgrass::execGRASS(
			cmd = "r.topidx",
			input = sources(x)[i],
			output = srcs[i],
			flags = c(.quiet(), "overwrite")
		)
	
	}
	makeGRaster(srcs, paste0(names(x), "_twi"))

	} # EOF
)

#' Names of columns of a GRASS vector's attribute table
#'
#' @description This function returns the column names of a **GRASS** vector's attribute table.
#'
#' @param x A `GVector` or the name of a vector in **GRASS**.
#' 
#' @returns Character vector.
#'
#' @aliases .vNames
#' @rdname vNames
#' @keywords internal
.vNames <- function(x) {

	if (inherits(x, "GVector")) {
		.locationRestore(x)
		src <- sources(x)
	} else {
		src <- x
	}

	if (.vHasDatabase(src)) {

		out <- rgrass::execGRASS(
			cmd = "db.columns",
			table = src,
			flags = .quiet(),
			intern = TRUE
		)
		
	} else {
		out <- NULL
	}
	out

}

#' Coordinate reference system of a GRaster or GVector
#' 
#' @description Get the coordinate reference system (CRS) of a `GRaster` or `GVector`s, or from the currently active **GRASS** "project/location" (see `vignette("projects_mapsets", package = "fasterRaster")`):
#' * `crs()`: Return a WKT string (an object of class `character`).
#' * `st_crs()`: Return a `crs` object (from the **sf** package).
#' * `coordRef()`: Return a `list`.
#'
#' @param x An object that inherits from a `GLocation` (i.e., a `GRaster` or `GVector`) or missing. If missing, the coordinate reference system of the currently active **GRASS** "project/location" is reported.
#' 
#' @param ... Other arguments (generally unused).
#'
#' @returns Function `crs()` returns a `character` object, `st_crs()` returns `crs` object, and `coordRef()` a `list`.
#' 
#' @seealso [terra::crs()], [sf::st_crs()]
#' 
#' @example man/examples/ex_GRaster.r
#'
#' @aliases crs
#' @rdname crs
#' @exportMethod crs
methods::setMethod(
    f = "crs",
    signature = c(x = "missing"),
    definition = function(x) {

	locs <- .locations()
	al <- .fasterRaster$activeLocation
	locs[[al]]

    } # EOF
)

#' @aliases crs
#' @rdname crs
#' @exportMethod crs
methods::setMethod(
	f = "crs",
	signature = "GLocation",
	definition = function(x) x@crs
)

#' @rdname crs
#' @aliases st_crs
#' @exportMethod st_crs
methods::setMethod(
	f = "st_crs",
	signature = "missing",
	definition = function(x, ...) {
		
	out <- crs()
	sf::st_crs(out)

	} # EOF
)

#' @rdname crs
#' @aliases st_crs
#' @exportMethod st_crs
methods::setMethod(
	f = "st_crs",
	signature = "GLocation",
	definition = function(x, ...) {

	out <- x@crs
	out <- sf::st_crs(out)
	out

	} # EOF
)

#' @aliases st_crs
#' @rdname crs
#' @export
st_crs <- function(x, ...) UseMethod("st_crs", x)

#' @rdname crs
#' @aliases coordRef
#' @exportMethod coordRef
methods::setMethod(
	f = "coordRef",
	signature = "missing",
	function(x) {
	
	info <- rgrass::execGRASS("g.proj", flags = "p", intern = TRUE)
	out <- list()

	if (any(grepl(info, pattern = "name       :"))) {

		y <- info[grepl(info, pattern = "name       :")]
		y <- strsplit(y, split = ":")[[1L]][2L]
		y <- trimws(y)

		out <- c(out, list(name = y))

	}

	if (any(grepl(info, pattern = "ellps      :"))) {

		y <- info[grepl(info, pattern = "ellps      :")]
		y <- strsplit(y, split = ":")[[1L]][2L]
		y <- trimws(y)

		out <- c(out, list(ellipse = y))

	}

	if (any(grepl(info, pattern = "datum"))) {

		y <- info[grepl(info, pattern = "datum")]
		y <- strsplit(y, split = ":")[[1L]][2L]
		y <- trimws(y)

		out <- c(out, list(datum = y))

	}

	if (any(grepl(info, pattern = "EPSG:"))) {

		y <- info[grepl(info, pattern = "EPSG:")]
		y <- strsplit(y, split = "EPSG:")[[1L]][2L]
		y <- trimws(y)

		out <- c(out, list(epsg = y))

	}

	if (any(grepl(info, pattern = "proj       :"))) {

		y <- info[grepl(info, pattern = "proj       :")]
		y <- strsplit(y, split = ":")[[1L]][2L]
		y <- trimws(y)

		out <- c(out, list(proj = y))

	}

	if (any(grepl(info, pattern = "lat_0      :"))) {

		y <- info[grepl(info, pattern = "lat_0      :")]
		y <- strsplit(y, split = ":")[[1L]][2L]
		y <- trimws(y)

		y <- as.numeric(y)

		out <- c(out, list(lat_0 = y))

	}

	if (any(grepl(info, pattern = "lon_0      :"))) {

		y <- info[grepl(info, pattern = "lon_0      :")]
		y <- strsplit(y, split = ":")[[1L]][2L]
		y <- trimws(y)

		y <- as.numeric(y)

		out <- c(out, list(lon_0 = y))

	}

	if (any(grepl(info, pattern = "azi        :"))) {

		y <- info[grepl(info, pattern = "azi        :")]
		y <- strsplit(y, split = ":")[[1L]][2L]
		y <- trimws(y)

		y <- as.numeric(y)

		out <- c(out, list(azimuth = y))

	}

	if (any(grepl(info, pattern = "x_0        :"))) {

		y <- info[grepl(info, pattern = "x_0        :")]
		y <- strsplit(y, split = ":")[[1L]][2L]
		y <- trimws(y)

		y <- as.numeric(y)

		out <- c(out, list(x_0 = y))

	}

	if (any(grepl(info, pattern = "y_0        :"))) {

		y <- info[grepl(info, pattern = "y_0        :")]
		y <- strsplit(y, split = ":")[[1L]][2L]
		y <- trimws(y)

		y <- as.numeric(y)

		out <- c(out, list(y_0 = y))

	}

	if (any(grepl(info, pattern = "k          :"))) {

		y <- info[grepl(info, pattern = "k          :")]
		y <- strsplit(y, split = ":")[[1L]][2L]
		y <- trimws(y)

		y <- as.numeric(y)

		out <- c(out, list(k = y))

	}

	if (any(grepl(info, pattern = "meters     :"))) {

		y <- info[grepl(info, pattern = "meters     :")]
		y <- strsplit(y, split = ":")[[1L]][2L]
		y <- trimws(y)

		y <- as.numeric(y)

		out <- c(out, list(meters = y))

	}

	if (any(grepl(info, pattern = "pm         :"))) {

		y <- info[grepl(info, pattern = "pm         :")]
		y <- strsplit(y, split = ":")[[1L]][2L]
		y <- trimws(y)

		out <- c(out, list(prime_meridian = y))

	}

	if (any(grepl(info, pattern = "unit       :"))) {

		y <- info[grepl(info, pattern = "unit       :")]
		y <- strsplit(y, split = ":")[[1L]][2L]
		y <- trimws(y)

		out <- c(out, list(unit = y))

	}

	if (any(grepl(info, pattern = "units      :"))) {

		y <- info[grepl(info, pattern = "units      :")]
		y <- strsplit(y, split = ":")[[1L]][2L]
		y <- trimws(y)

		out <- c(out, list(proj_unit = y))

	}
	out$crs <- crs()
	out

	} # EOF
)

#' @rdname crs
#' @aliases coordRef
#' @exportMethod coordRef
methods::setMethod(
	f = "coordRef",
	signature = c(x = "GRaster"),
	function(x) {
	
	.locationRestore(x)
	coordRef()

	} # EOF
)

#' @rdname crs
#' @aliases coordRef
#' @exportMethod coordRef
methods::setMethod(
	f = "coordRef",
	signature = c(x = "GVector"),
	function(x) {
	
	.locationRestore(x)
	coordRef()

	} # EOF
)

if (grassStarted()) {

# Setup
library(terra)

# Elevation raster
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster:
elev <- fast(madElev)

# Cell size, no masking, single layer
cs1 <- cellSize(elev)
plot(cs1)

# Cell size, with masking, single layer
cs2 <- cellSize(elev, mask = TRUE)
plot(cs2)

# Cell size, no masking, multilayer
elev2 <- c(elev, log(elev - 200))
cs3 <- cellSize(elev2)
plot(cs3)

# Cell size, masking by 1st layer, multilayer (ignores subsequent layers)
cs4 <- cellSize(elev2, mask = TRUE)
plot(cs4)

# Cell size, masking by each layer, multilayer
cs5 <- cellSize(elev2, mask = TRUE, lyrs = TRUE)
plot(cs5)

}

#' Has "GRASS" been started or not?
#'
#' @description Returns `TRUE` or `FALSE`, depending on whether a **GRASS** connection has been made or not within the current **R** session. Usually used only by developers. **GRASS** is started the first time [fast()] is used.
#'
#' @returns Logical.
#'
#' @examples
#'
#' grassStarted()
#'
#' @aliases grassStarted
#' @rdname grassStarted
#' @export
grassStarted <- function() {
	exists(".fasterRaster", inherits = TRUE) && .fasterRaster$grassStarted
}

#' Names of columns of the levels table of a categorical raster
#'
#' @description This function returns the column names of each "levels" table of a categorical raster (see `vignette("GRasters", package = "fasterRaster")`).
#'
#' @param x A `GRaster`.
#'
#' @param layer `NULL`, numeric integer, or character: The index (indices) or name(s) of one or more raster layers. The default is `NULL`, in which case all names for all layers are returned.
#'
#' @returns A list of character vectors.
#'
#' @example man/examples/ex_GRaster_categorical.r
#'
#' @seealso [cats()], `vignette("GRasters", package = "fasterRaster")`
#'
#' @aliases catNames
#' @rdname catNames
#' @exportMethod catNames
methods::setMethod(
	f = "catNames",
	signature = c(x = "GRaster"),
	function (x, layer = NULL) {
	
	if (is.null(layer)) layer <- seq_len(nlyr(x))
	if (is.character(layer)) layer <- match(layer, names(x))
	if (is.logical(layer)) {
		if (length(layer) != nlyr(x)) {
			layer <- rep(layer, length.out = nlyr(x))
		}
		layer <- which(layer)
	}

	categs <- x@levels[layer]
	lapply(categs, names)
	
	} # EOF
)

#' @aliases catNames
#' @rdname catNames
#' @exportMethod catNames
methods::setMethod(
	f = "catNames",
	signature = c(x = "SpatRaster"),
	function (x, layer = NULL) {
	
	if (is.null(layer)) layer <- seq_len(terra::nlyr(x))
	if (is.character(layer)) layer <- match(layer, names(x))
	if (is.logical(layer)) {
		if (length(layer) != nlyr(x)) {
			layer <- rep(layer, length.out = terra::nlyr(x))
		}
		layer <- which(layer)
	}

	categs <- terra::cats(x[[layer]])
	lapply(categs, names)
	
	} # EOF
)

#' Get index of raster layers
#'
#' @param layer Integer, numeric, logical, or character: Refers to one or more layers.
#'
#' @param x A `GRaster`.
#'
#' @param recycle Logical: If `TRUE` (default), and `layer` is logical and smaller in number than the number of layers, then recycle the vector of `layer`.
#'
#' @param negate Logical: If `TRUE`, return indices of all layers *not* identified in `layer`.
#'
#' @returns An integer vector.
#'
#' @keywords internal
.layerIndex <- function(layer, x, recycle = TRUE, negate = FALSE) {

	nx <- nlyr(x)

	if (is.character(layer)) {
		layer <- match(layer, names(x))
		if (anyNA(layer)) stop("Layer not found.")
	} else if (is.logical(layer)) {
		if (recycle && length(layer) < nx) layer <- rep(layer, length.out = nx)
		layer <- which(layer)
	} else if (inherits(layer, c("numeric", "integer"))) {
		if (any(layer > nx) | any(layer < 1L)) stop("Index out of bounds.")
	}
	if (any(!(layer %in% seq_len(nx)))) stop("Raster only contains ", nlyr(x), " layer(s).")
	
	if (negate) layer <- seq_len(nx)[-layer]

	layer

}

#' "Stack" a GRaster
#'
#' @description This function "stacks" one `GRaster` with another. It has the same functionality as [c()].
#'
#' @param x,value A `GRaster`.
#'
#' @returns A `GRaster`.
#'
#' @seealso [c()], [terra::add<-], [terra::c()]
#'
#' @example man/examples/ex_GRaster_GVector_subset_assign.r
#'
#' @name add<-
#' @aliases add<-,GRaster,GRaster-method
#' @docType methods
#' @rdname add
#' @exportMethod add<-
methods::setMethod(
	f = "add<-",
	signature = c(x = "GRaster", value = "GRaster"),
	function(x, value) c(x, value)
)

#' Sine wave rasters
#'
#' @description This function creates one or more rasters with sine waves in the north-south and east-west directions.
#'
#' @param x A `GRaster`.
#'
#' @param ns,ew Numeric: Number of complete sine waves (i.e., wavelengths) in the north-south and east-west directions. A wavelength of 1 creates a "full" sine wave (e.g., starting at 0 at one end and ending at 0 at the other). A wavelength of 2 would create two such waves, and so on. A value of 0 creates no waves in the given direction (i.e., each row or column has constant values). The default value is 1.
#'
#' @param nsOffset,ewOffset Numeric: Offset of the sine waves from the edges of the raster, expressed as a proportion of the length of the raster. The default is 0, so the values of the outermost cells will be close to 0 (but not exactly 0 because centers of cells at the raster edges are not on the actual edge). If an offset value is 0.2, for example, then it will be pushed "inward" toward the middle of the raster by 20% of the raster's extent.
#'
#' @param nsAmp,ewAmp Numeric: Amplitude (minimum and maximum of the sine wave) in the north-south and east-west directions. The default is 1. Note that when north-south and east-west waves are created (i.e., `ns` and `ew` are both > 0), the effective amplitude is halved so that the sum is equal to `nsAmp + ewAmp`.
#'
#' @param combos Logical: If `TRUE` (default), create sine rasters using all possible combinations of values of `ns`, `ew`, `nsOffset`, `ewOffset`, and `amp`. If `FALSE`, you can only supply either one value per parameter, or the same number of values per parameter. In this latter case, one raster will be created per pairwise set of the unique parameters. For example, you could specify 3 values for `ns` and either one or three values for any of the other parameters, and three rasters would be created.
#'
#' @param mask Either `NULL` (default), or a `GRaster` or `GVector`: Used as a mask for the output. If this is a `GVector`, then only cells that are not `NA` in the mask have values. If this is a `GVector`, then only cells that overlap the vector have values assigned. All other values will be `NA`.
#'
#' @param verbose Logical: If `TRUE`, display progress.
#'
#' @returns A `GRaster`.
#'
#' @example man/examples/ex_sineRast.r
#'
#' @aliases sineRast
#' @rdname sineRast
#' @exportMethod sineRast
methods::setMethod(
	f = "sineRast",
	signature = c(x = "GRaster"),
	function(
		x,
		ns = 1,
		ew = 1,
		nsOffset = 0,
		ewOffset = 0,
		nsAmp = 1,
		ewAmp = 1,
		combos = FALSE,
		mask = NULL,
		verbose = FALSE
	) {

	if (any(ns < 0 | ew < 0)) stop("Arguments `ns` and `ew` must be >= 0.")

	if (combos) {

		settings <- expand.grid(ns = ns, ew = ew, nsOffset = nsOffset, ewOffset = ewOffset, nsAmp = nsAmp, ewAmp = ewAmp)
		settings <- data.table::as.data.table(settings)

	} else {

		lengths <- c(length(ns), length(ew), length(nsOffset), length(ewOffset), length(nsAmp), length(ewAmp))
		maxLength <- max(lengths)
	
		bad <- any(lengths != 1 & lengths != maxLength)
		if (bad) stop("If `combo` is FALSE, then arguments `ns`, `ew`, `nsOffset`,  `ewOffset`,\n  `nsAmp`, and `ewAmp` must be a single value or the same length. You can mix\n  and match. For example, `ns` can have a single value, and `ew` can have\n  3 values. But in this case, the other parameters must have 1 or 3 values.")

		settings <- data.table::data.table(ns = ns, ew = ew, nsOffset = nsOffset, ewOffset = ewOffset, nsAmp = nsAmp, ewAmp = ewAmp)

	}

	ns <- settings$ns
	ew <- settings$ew
	nsOffset <- settings$nsOffset
	ewOffset <- settings$ewOffset
	nsAmp <- settings$nsAmp
	ewAmp <- settings$ewAmp
	nRastsToMake <- nrow(settings)

	.locationRestore(x)
	.region(x)

	if (nlyr(x) > 1) x <- x[[1L]]
	extent <- ext(x, vector = TRUE)

	if ((verbose | faster("verbose")) & nRastsToMake > 1)  {
		pb <- utils::txtProgressBar(min = 0, max = nRastsToMake, initial = 0, style = 3, width = 30)
	}

	srcs <- .makeSourceName("sineRast", "raster", nrow(settings))
	for (i in seq_len(nRastsToMake)) {

		if ((verbose | faster("verbose")) & nRastsToMake > 1) utils::setTxtProgressBar(pb, i)

		### create expression
		# no waves
		if (omnibus::compareFloat(ns[i], 0, "==") & omnibus::compareFloat(ew[i], 0, "==")) {
			ex <- paste0(srcs[i], " = 0")
		# waves in at least on direction
		} else {

			# scale wavelength
			# set cell values to [0, 2* pi] along extent
			ymin <- extent[3L] + 0.5 * res(x)[2L]
			ymax <- extent[4L] - 0.5 * res(x)[2L]
			yext <- ymax - ymin

			xmin <- extent[1L] + 0.5 * res(x)[1L]
			xmax <- extent[2L] - 0.5 * res(x)[1L]
			xext <- xmax - xmin

			ewOffset <- 360 * ewOffset
			nsOffset <- 360 * nsOffset

			if (ns[i] > 0 & ew[i] > 0) {
				halfNsAmp <- nsAmp[i] / 2
				halfEwAmp <- ewAmp[i] / 2
				ex <- paste0(srcs[i], " = ", halfNsAmp, " * sin(", ewOffset[i] + ew[i] * 360, " * ((x() - ", xmin, ") / ", xext, ")) + ", halfEwAmp, " * sin(", nsOffset[i] + ns[i] * 360, " * ((y() - ", ymin, ") / ", yext, "))")
			} else if (ns[i] > 0 & omnibus::compareFloat(ew[i], 0, "==")) {
				ex <- paste0(srcs[i], " = ", nsAmp[i], " * sin(", nsOffset[i] + ns[i] * 360, " * ((y() - ", ymin, ") / ", yext, "))")
			} else if (omnibus::compareFloat(ns[i], 0, "==") & ew[i] > 0) {
				ex <- paste0(srcs[i], " = ", ewAmp[i], " * sin(", ewOffset[i] + ew[i] * 360, " * ((x() - ", xmin, ") / ", xext, "))")
			}

		}
		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		# mask
		if (!is.null(mask)) {

			if (inherits(mask, "GRaster")) {
				maskType <- "raster"
			} else if (inherits(mask, "GVector")) {
				maskType <- "vector"
			} else {
				stop("Argument `mask` must be a `GRaster` or `GVector`.")
			}

			srcs[i] <- .mask(x = srcs[i], mask = sources(mask), maskType = maskType, inverse = FALSE, maskvalues = NA, updatevalue = NA)
		
		}

	}
	if ((verbose | faster("verbose")) & nRastsToMake > 1)  close(pb)

	names <- paste0("ns", ns, "_ew", ew, "_nsOffset", nsOffset, "_ewOffset", ewOffset, "_nsAmp", nsAmp, "_ewAmp", ewAmp)
	makeGRaster(srcs, names)

	} # EOF
)

#' Add rows or columns to the "levels" table of a categorical raster
#'
#' @description `addCats()` and `addCats()<-` add information to a categorical`GRaster's "levels" table.
#' * addCats()` uses [data.table::merge()] or [cbind()] to do this--it does not add new rows, but rather new columns.
#' * `addCats()<-` uses [rbind()] to add new categories (rows) to the "levels" table.
#'
#' GRaster`s can represent categorical data (see `vignette("GRasters", package = "fasterRaster")`). Cell values are actually integers, each corresponding to a category, such as "desert" or "wetland." A categorical raster is associated with a "levels" table that matches each value to a category name. The table must be `NULL` (i.e., no categories--so not a categorical raster), or have at least two columns. The first column must have integers and represent raster values. One or more subsequent columns must have category labels. The column with these labels is the "active category".
#'
#' @param x A `GRaster`.
#'
#' @param value A `data.frame`, `data.table`, a list of `data.frames` or `data.tables` with one per raster layer, or a categorical `SpatRaster`. The table's first column is the "value" column and must contain numeric values (of class `numeric` or `character`). If a `SpatRaster` is supplied, then its categories will be transferred to the `GRaster`.
#'
#' @param merge Logical (function `addCats()`): If `FALSE` (default), columns will be combined with the existing "levels" table using [cbind()]. If `TRUE`, they will be combined using [data.table::merge()].
#'
#' @param layer Numeric integers, logical vector, or character: Layer(s) to which to add or from which to drop levels.
#'
#' @returns A `GRaster`. The "levels" table of the raster is modified.
#'
#' @seealso [terra::addCats()], [concats()], [combineLevels()], [droplevels()], `vignette("GRasters", package = "fasterRaster"`
#'
#' @example man/examples/ex_GRaster_categorical.r
#'
#' @aliases addCats
#' @rdname addCats
#' @exportMethod addCats
methods::setMethod(
	f = "addCats",
	signature = c(x = "GRaster"),
	function(x, value, merge = FALSE, layer = 1) {

	layer <- .layerIndex(layer, x, recycle = TRUE)

	if (!inherits(value, "data.table")) {
		value <- data.table::as.data.table(value)
	}

	for (i in layer) {

		if (merge) {

			levelValCol <- names(x@levels[[i]])[1L]
			valueValCol <- names(value)[1L]
			data.table::setkeyv(value, valueValCol)

			x@levels[[i]] <- merge(x@levels[[i]], value, by.x = levelValCol, by.y = valueValCol, all.x = TRUE, allow.cartesian = FALSE)

			data.table::setkeyv(x@levels[[i]], levelValCol)

		} else {
		
			x@levels[[i]] <- cbind(x@levels[[i]], value)
		
		}
	}

 	methods::validObject(x)
	x

	} # EOF
)

#' @aliases addCats<-
#' @rdname addCats
#' @exportMethod addCats<-
methods::setMethod(
	f = "addCats<-",
	signature = c(x = "GRaster"),
	function(x, layer = 1, value) {

	if (missing(value)) value <- 1L

	layer <- .layerIndex(layer, x, recycle = TRUE)

	if (!inherits(value, "data.table")) value <- data.table::as.data.table(value)

	for (i in layer) {

		val <- names(value)[1L]
		value[ , (val) := lapply(.SD, as.integer), .SDcols = val]
  		x@levels[[i]] <- rbind(x@levels[[i]], value)
		data.table::setkeyv(x@levels[[i]], val)

		if (length(unique(x@levels[[i]][[val]])) < nrow(x@levels[[i]])) stop("More than one label assigned to the same value in the levels table.")
	
	}

 	methods::validObject(x)
	x

	} # EOF
)

#' "fasterRaster": Faster raster and spatial vector processing using "GRASS"
#'
#' @description **fasterRaster**: Processing of large-in-memory/-on disk rasters and spatial vectors in using **GRASS**. Most functions in the **terra** and **sf** packages are recreated. Processing of medium-sized and smaller spatial objects will nearly always be faster using **terra** or **sf**. To use most of the functions you must have the stand-alone version of **GRASS** version 8.3 or higher (not the **OSGeoW4** installer version). Note that due to differences in how **GRASS**, **terra**, and **sf** were implemented, results will not always be strictly comparable between functions for the same operation.
#'
#' ## Most useful tutorials and functions:
#' * The quick-start guide to getting started with **fasterRaster**: `vignette("fasterRaster", package = "fasterRaster")`: 
#' * Types of `GRaster`s: `vignette("GRasters", package = "fasterRaster")`
#' * How to speed up **fasterRaster**: `vignette("faster_fasterRaster", package = "fasterRaster")`
#' * [faster()]: Set the directory where **GRASS** is installed on your system, and set or get other package-wide options. This function must be run once before using most **fasterRaster** functions.
#' * [fast()]: Convert a `SpatRaster`, `SpatVector`, or `sf` vector to **fasterRaster**'s raster format (`GRaster`s) or vector format (`GVector`s), or load one from a file
#' * [rast()], [vect()], and [st_as_sf()]: Convert `GRaster`s and `GVector`s to `SpatRaster`s, `SpatVector`s, or `sf` vectors
#' * [writeRaster()] and [writeVector()]: Save `GRaster`s or `GVector`s to disk
#'
#' ## Properties of `GRasters`
#' * [crs()]: Coordinate reference system
#' * [coordRef()]: Coordinate reference system
#' * [datatype()]: Data type
#' * [dim()] and [dim3d()]: Number of rows, columns, and depths
#' * [ext()], [N()], [S()], [E()], [W()], [top()], and [bottom()]: Spatial extent
#' * [freq()]: Frequencies of cell values in a raster
#' * [is.2d()] and [is.3d()]: Is an object 2- or 3-dimensional?
#' * [is.int()], [is.cell()], [is.float()], [is.doub()]: `GRaster` data type (integer/float/double)
#' * [is.factor()]: Does a raster represent categorical data?
#' * [is.lonlat()]: Is an object projected (e.g., in WGS84)?
#' * [levels()]: Names of levels in a categorical `GRaster`
#' * [minmax()]: Minimum and maximum values across all non-`NA` cells
#' * [names()]: `GRaster` names
#' * [ncol()]: Number of columns
#' * [nacell()]: Number of `NA` cells
#' * [ncell()]: Number of cells
#' * [ncell3d()]: Number of cells of a 3D `GRaster`
#' * [ndepth()]: Number of depths of a 3D `GRaster`
#' * [nlyr()]: Number of layers
#' * [nonnacell()]: Number of non-`NA` cells
#' * [nrow()]: Number of rows
#' * [nlevels()]: Number of categories
#' * [res()], [res3d()], [xres()], [yres()], and [zres()]: Spatial resolution
#' * [sources()]: Name of the raster file in the **GRASS** cache
#' * [`topology()`][topology]: Dimensionally (2D or 3D)
#' * [zext()]: Vertical extent
#' * [zres()]: Vertical resolution
#' 
#' ## Functions that operate on or create `GRasters`
#' * [Arithmetic]: Mathematical operations on `GRaster`s: `+`, `-`, `*`, `/`, `^`, `%%` (modulus), `%/%` (integer division)
#' * [Logical comparisons][Compare-methods]: `<`, `<=`, `==`, `!=`, `>=`, and `>`, plus \code{\link[fasterRaster]{%in%}} and \code{\link[fasterRaster]{%notin%}} (for categorical rasters only)
#' * [Logical operators][Logic-methods]: `|`and `&`
#'
#' Mathematical functions that are applied to each layer of a `GRaster`:
#' * Working with `NA`s: [is.na()], [not.na()], and [maskNA()] 
#' * Trigonometry: [sin()], [cos()], [tan()], [asin()], [acos()], [atan()], [atan2()] 
#' * Logarithms and powers: [exp()], [log()], [ln()], [log1p()], [log2()], [log10()], [sqrt()] 
#' * Rounding: [round()], [floor()], [ceiling()], [trunc()] 
#' * Signs: [abs()] 
#'
#' Mathematical functions that are applied across layers of multi-layered `GRaster`s:
#' * Numeration: [sum()], [count()] 
#' * Central tendency: [mean()], [mmode()], [median()] 
#' * Dispersion: [stdev()], [var()], [varpop()], [nunique()], [range()], [quantile()], [skewness()], [kurtosis()]
#' * Extremes: [min()], [max()], [which.min()], [which.max()] 
#' * `NA`s: [allNA()], [anyNA()] 
#' 
#' Subsetting, assigning, and replacing `GRaster` layers
#' * [$], \code{\link[fasterRaster]{[[}}, or [subset()]: Subset or remove specific layers of a `GRaster`
#' * \code{\link[fasterRaster]{[<-}}: Replace values of cells of a `GRaster`
#' * \code{\link[fasterRaster]{[[<-}}: Replace specific layers of a `GRaster`
#' * \code{\link[fasterRaster]{add<-}}: Replace specific layers of a `GRaster`
#'
#' Operations on `GRaster`s
#' * [as.int()], [as.float()], [as.doub()]: Change data type (integer/float/double)
#' * [as.lines()]: Convert a `GRaster` to a "lines" vector
#' * [as.points()]: Convert a `GRaster` to a "points" vector
#' * [as.polygons()]: Convert a `GRaster` to a "polygons" vector
#' * [aggregate()]: Aggregate values of `GRaster` cells into larger cells
#' * [bioclims()]: BIOCLIM rasters (classic set and extended set)
#' * [buffer()]: Create a buffer around non-`NA` cells
#' * [app()]: Apply a user-defined function to multiple layers of a `GRaster` (with helper functions [appFuns()] and [appCheck()])
#' * [c()]: "Stack" two or more rasters
#' * [cellSize()]: Cell area
#' * [classify()]: Partition cell values into strata
#' * [clump()]: Group adjacent cells with similar values
#' * [combineLevels()]: Combine the "levels" tables of two or more categorical `GRaster`s
#' * [concats()]: Combine values from two or more categorical and/or integer rasters by concatenating them
#' * [crop()]: Remove parts of a `GRaster`
#' * [denoise()]: Remove "noise" from a `GRaster` using a principal components analysis (PCA)
#' * [distance()]: Distance to non-`NA` cells, or vice versa
#' * [extend()]: Add rows and columns to a `GRaster`
#' * [extract()]: Extract values from a `GRaster` at locations of a `GVector`
#' * [fillNAs()]: Fill `NA` cells
#' * [focal()]: Calculate cell values based on values of nearby cells
#' * [fragmentation()]: Landscape fragmentation class from Riitters et al. (2020)
#' * [global()]: Summary statistics across cells of each `GRaster` layer
#' * [hist()]: Histogram of `GRaster` values
#' * [interpIDW()]: Interpolate values at points to a `GRaster`
#' * [kernel()]: Kernel density estimator of points
#' * [layerCor()]: Correlation or covariance between two or more `GRaster` layers
#' * [mask()]: Remove values in a `GRaster` based on values in another `GRaster` or vector
#' * [maskNA()]: Mask all non-NA cells or all NA cells
#' * [match()], \code{\link[fasterRaster]{%in%}}, and \code{\link[fasterRaster]{%notin%}}: Find which cells of a `GRaster` match or do not match certain values
#' * [merge()]: Combine two or more rasters with different extents and fill in `NA`s
#' * [multivarEnvSim()]: Multivariate environmental similarity surface (MESS)
#' * \code{\link[fasterRaster]{names<-}}: Assign names to a `GRaster`
#' * [noise()]: Remove coarse-scale trends from a `GRaster`, leaving just fine-scale "noise"
#' * [pairs()]: Plot correlations between `GRaster` layers
#' * [pcs()]: Retrieve a principal components model from a PCA `GRaster` generated using `princomp()`
#' * [plot()]: Display a `GRaster`
#' * [project()]: Change coordinate reference system and cell size
#' * [predict()]: Make predictions to a `GRaster` from a linear model or generalized linear model
#' * [princomp()]: Apply a principal components analysis (PCA) to a `GRaster`
#' * [regress()]: Regression intercept, slope, r2, and t-value across each set of cells
#' * [resample()]: Change cell size
#' * [reorient()]: Convert degrees between 'north-orientation' and 'east orientation'
#' * [sampleRast()]: Randomly sample cells from a `GRaster`
#' * [scale()], [scalepop()], and [unscale()]: Subtract means and divide by standard deviations, or inverse of that
#' * [selectRange()]: Select values from rasters in a stack based on values in another `GRaster`
#' * [spatSample()]: Randomly points from a `GRaster`
#' * [stretch()]: Rescale values in a GRaster
#' * [subst()]: Re-assign cell values
#' * [thinLines()]: Reduce linear features on a `GRaster` so linear features are 1 cell wide
#' * [tiles()]: Divide a `GRaster` into spatially exclusive subsets (though with possible overlap)
#' * [trim()]: Remove rows and columns from a `GRaster` that are all `NA`
#' * [zonal()]: Statistics (mean, sum, etc.) on areas of a `GRaster` defined by sets of cells with the same values in another `GRaster`, or by geometries in a `GVector`
#' * [zonalGeog()]: Geographic statistics (area, perimeter, fractal dimension, etc.) for sets of cells with the same values
#'
#' ## Creating `GRaster`s *de novo*
#' * [fractalRast()]: Create a fractal `GRaster`
#' * [init()]: GRaster with values equal to row, column, coordinate, regular, or "chess"
#' * [longlat()]: Create longitude/latitude rasters
#' * [makeGRaster()]: Create a `GRaster` from a **GRASS** raster file
#' * [makeGVector()]: Create a `GVector` from a **GRASS** vector file
#' * [rNormRast()]: A random `GRaster` with values drawn from a normal distribution
#' * [rSpatialDepRast()]: Create a random `GRaster` with or without spatial dependence
#' * [rUnifRast()]: A random `GRaster` with values drawn from a uniform distribution
#' * [rWalkRast()]: Paths of random walkers
#' * [sineRast()]: Sine wave rasters
#' 
#' ## Analysis of terrain and hydrology
#' * [as.contour()]: Contour lines from a `GRaster`
#' * [flow()]: Identify watershed basins and direction and accumulation of flow
#' * [flowPath()]: Path of water flow across a landscape
#' * [geomorphons()]: Identify terrain feature types
#' * [`hillshade()`][hillshade]: Create a hillshade `GRaster`
#' * [horizonHeight()]: Horizon height
#' * [sun()]: Solar radiance and irradiance
#' * [ruggedness()]: Terrain Ruggedness Index
#' * [streams()]: Create stream network
#' * [terrain()]: Slope, aspect, curvature, and partial slopes
#' * [wetness()]: Topographic wetness index
#' 
#' ## Operations on categorical (factor) `GRaster`s
#' * \code{\link[fasterRaster]{%in%}}, and \code{\link[fasterRaster]{%notin%}}: Mask cells that match or do not match a given category
#' * [activeCat()] and [activeCats()]: Column(s) that defines category labels
#' \code{\link[fasterRaster]{activeCat<-}}: Set column that defines category labels
#' * [addCats()]: Add new columns to a "levels" table
#' \code{\link[fasterRaster]{addCats<-}}: Add new rows (levels) to a "levels" table
#' * [categories()]: Set "levels" table for specific layers of a categorical raster
#' * [catNames()]: Column names of each "levels" table
#' * [cats()]: "Levels" table of a categorical raster
#' * [combineLevels()]: Combine the "levels" tables of two or more categorical `GRaster`s
#' * [complete.cases()]: Find rows of a categorical `GRaster`'s "levels" table that have no `NA`s in them
#' * [concats()]: Combine categories from two or more categorical rasters by concatenating them
#' * [droplevels()]: Remove one or more levels
#' * [freq()]: Frequency of each category across cells of a raster
#' * [is.factor()]: Is a raster categorical?
#' * [levels()]: "Levels" table of a categorical raster  
#' * \code{\link[fasterRaster]{levels<-}}: Set "levels" table of a categorical raster
#' * [match()], \code{\link[fasterRaster]{%in%}}, and \code{\link[fasterRaster]{%notin%}}: Find which cells of a `GRaster` match or do not match certain category labels
#' * [minmax()]: "Lowest" and "highest" category values of categorical rasters (when argument `levels = TRUE`)
#' * [missing.cases()]: Find rows of a categorical `GRaster`'s "levels" table that have at least one `NA` in them
#' * [missingCats()]: Values that have no category assigned to them
#' * [nlevels()]: Number of levels
#' * [segregate()]: Create one GRaster layer per unique value in a GRaster
#' * [subst()]: Re-assign category levels
#' * [zonalGeog()]: Geographic statistics (area, perimeter, fractal dimension, etc.) for sets of cells with the same values
#' 
#' ## Analysis of remote sensing rasters
#' * [compositeRGB()]: Combine red, green, and blue color bands to make a composite `GRaster`
#' * [plotRGB()]: Display a multispectral `GRaster` using red, blue, green, and alpha channels
#' * [vegIndex()]: Vegetation indices from surface reflectance
#'
#' ## Functions that operate on **terra** `SpatRaster`s
#' * [bioclims()]: BIOCLIM rasters (classic set and extended set)
#' * [fragmentation()]: Landscape fragmentation class from Riitters et al. (2020)
#'
#' ## Properties of `GVector`s
#' * [crs()]: Coordinate reference system
#' * [coordRef()]: Coordinate reference system
#' * [datatype()]: Data type of fields
#' * [dim()]: Number of geometries and columns
#' * [expanse()]: Area of polygons or length of lines
#' * [ext()], [N()], [S()], [E()], [W()], [top()], and [bottom()]: Spatial extent
#' * [geomtype()]: Type of vector (points, lines, polygons)
#' * [is.2d()] and [is.3d()]: Is an object 2- or 3-dimensional?
#' * [is.lonlat()]: Is an object projected (e.g., in WGS84)?
#' * [is.points()], [is.lines()], [is.polygons()]: Does a `GVector` represent points, lines, or polygons?
#' * [names()]: Names of `GVector` fields
#' * [ncol()]: Number of fields
#' * [ngeom()]: Number of geometries (points, lines, polygons)
#' * [nrow()]: Number of rows in a vector data table
#' * [nsubgeom()]: Number of sub-geometries (points, lines, polygons that make up single- and multipart geometries)
#' * [`sources()`][sources]: Name of the vector file in the **GRASS** cache
#' * [`topology()`][topology]: Dimensionally (2D or 3D)
#' * [zext()]: Vertical extent
#'
#' ## Subsetting and assigning geometries or rows and columns of `GVector`s
#' * [$] or \code{\link[fasterRaster]{[[}}: Subset columns of a `GVector`'s data table
#' * \code{\link[fasterRaster]{[}} or [subset()]: Subset geometries of a `GVector`
#' * \code{\link[fasterRaster]{$<-}}: Replace specific columns of a `GVector`'s data table or add columns
#' * \code{\link[fasterRaster]{addTable<-}}: Add a data table to a `GVector`
#' * [dropTable()]: Remove a `GVector`s data table
#'
#' ## Operations on `GVector`s
#' * [aggregate()]: Combine `GVector` geometries
#' * [as.data.frame()]: Convert a `GVector`'s attribute table to a `data.frame`
#' * [as.data.table()]: Convert a `GVector`'s attribute table to a `data.table`
#' * [as.points()]: Extract vertex coordinates from a "lines" or "polygons" `GVector`
#' * [buffer()]: Create a polygon around/inside a `GVector`
#' * [clusterPoints()]: Identify clusters of points
#' * [centroids()]: Centroid(s) of a `GVector`
#' * [colbind()]: Add columns to the data table of a `GVector`
#' * [complete.cases()]: Find rows of a `GVector`'s data table that have no `NA`s in them
#' * [connectors()]: Create lines connecting nearest features of two `GVector`s
#' * [convHull()]: Minimum convex hull
#' * [crds()]: Extract coordinates of a `GVector`
#' * [crop()]: Remove parts of a `GVector`
#' * [delaunay()]: Delaunay triangulation
#' * [disagg()]: Separate multipart geometries into singlepart geometries
#' * [distance()]: Distance between geometries in two `GVector`, or from a `GVector` to cells of a `GRaster`
#' * [erase()] or `-`: Remove part of a `GVector` that overlaps with another
#' * [expanse()]: Area of polygons or length of lines
#' * [extract()]: Extract values from a `GVector` at specific points
#' * [grid()]: Create a grid `GVector`
#' * [head()]: First rows of a `GVector`'s data table
#' * [hexagons()]: Create a hexagonal grid
#' * [interpIDW()]: Interpolate values at points to a `GRaster` using inverse-distance weighting
#' * [interpSplines()]: Interpolate values at points to a `GRaster` using splines
#' * [intersect()] or `*`: Intersection of two `GVectors`
#' * [kernel()]: Kernel density estimator of points
#' * [missing.cases()]: Find rows of a `GVector`'s data table that have at least `NA` in them
#' * \code{\link[fasterRaster]{names<-}}: Assign names to columns of a `GVector`s data table
#' * [neighborhoodMatrix()] and [neighbourhoodMatrix()]: Neighborhood matrix of a polygons `GVector`
#' * [project()]: Change coordinate reference system
#' * [rasterize()]: Convert a `GVector` to a `GRaster`
#' * [rbind()]: Combine `GVectors`
#' * [simplifyGeom()]: Remove vertices
#' * [smoothGeom()]: Remove "angular" aspects of features
#' * [st_as_sf()]: Convert a `GVector` to a `sf` vector
#' * [st_buffer()]: Create a polygon around/inside a `GVector`
#' * [tail()]: Last rows of a `GVector`'s data table
#' * [thinPoints()]: Reduce number of points in same raster cell
#' * [union()] or `+`: Combine two `GVector`s
#' * [voronoi()]: Voronoi tessellation
#' * [xor()] or `/`: Select parts of polygons not shared by two `GVector`s
#'
#' ## Creating `GVector`s *de novo*
#' * [rvoronoi()]: Random Voronoi tesselation
#'
#' ## Fixing issues with `GVector`s
#' (See also *Details* [fast()].)
#' * [breakPolys()]: Break topologically clean areas
#' * [fillHoles()]: Fill "holes" of a `GVector`
#' * [fixBridges()]: Change "bridges" to "islands"
#' * [fixDangles()]: Change "dangles" hanging off boundaries to lines
#' * [fixLines()]: Break lines at intersections and lines that form closed loops
#' * [remove0()]: Remove all boundaries and lines with a length of 0
#' * [removeAngles()]: Collapse lines that diverge at an angle that is computationally equivalent to 0
#' * [removeBridges()]: Remove "bridges" to "islands"
#' * [removeDangles()]: Remove "dangling" lines
#' * [removeDupCentroids()]: Remove duplicated area centroids
#' * [removeDups()]: Remove duplicated features and area centroids
#' * [removeSmallPolys()]: Remove small polygons
#' * [snap()]: Snap lines/boundaries to each other
#' 
#' ## Converting between data types
#' * [as.contour()]: Convert a `GRaster` to a `GVector` representing contour lines
#' * [as.doub()]: Convert a `GRaster` to a double-floating point raster (**GRASS** data type `DCELL`)
#' * [as.data.frame()]: Convert `GVector` to a `data.frame`
#' * [as.data.table()]: Convert `GVector` to a `data.table`
#' * [as.float()]: Convert a `GRaster` to a floating-point raster (**GRASS** data type `FCELL`)
#' * [as.int()]: Convert a `GRaster` to an integer raster (**GRASS** data type `CELL`)
#' * [as.points()], [as.lines()], and [as.polygons()]: Convert a `GRaster` to a `GVector`
#' * [categories()] and \code{\link[fasterRaster]{levels<-}}: Convert an integer raster to a categorical ("factor") raster.
#' * [fast()]: Convert a `SpatRaster` to a `GRaster`; a `SpatVector`, `sf` vector, numeric vector, `matrix`, `data.frame`, or `data.table` to a `GVector`; or load a vector or raster from a file
#' * [rast()]: Convert a `GRaster` to a `SpatRaster`
#' * [rasterize()]: Convert a `GVector` to a `GRaster`
#' * [st_as_sf()]: Convert a `GVector` to a `sf` vector
#' * [vect()]: Convert a `GVector` to a `SpatVector`
#'
#' ## General purpose functions
#' * [addons()]: Show installed **GRASS** addons
#' * [compareGeom()]: Determine if geographic metadata is same between `GRaster`s and/or `GVector`s
#' * [dropRows()]: Remove rows from a `data.frame` or `data.table`
#' * [grassGUI()]: Start the **GRASS** GUI (not recommended for most users!!!)
#' * [grassHelp()]: Open the help page for a **GRASS** tool.
#' * [grassInfo()]: **GRASS** version and citation
#' * [grassStarted()]: Has a connection **GRASS** been made within the current **R** session?
#' * [installAddon()]: Install a **GRASS** addon
#' * [mow()]: Remove unused rasters and vectors from the **GRASS** cache
#' * [reorient()]: Convert degrees between 'north-orientation' and 'east orientation'
#' * [replaceNAs()]: Replace `NA`s in columns of a `data.table` or `data.frame`, or in a vector
#' * [removeAddon()]: Delete **GRASS** addon from your system
#' * [seqToSQL()]: Format a numeric series into an SQL value call
#' * [update()]: Refresh metadata in a `GRaster` or `GVector` object
#'
#' ## Data objects
#' * [fastData()]: Helper function to quickly obtain example rasters and vectors
#' * [appFunsTable][appFunsTable] (see also [appFuns()]): Functions usable by the [app()] function
#' * [madChelsa][madChelsa]: Climate rasters for of a portion of eastern Madagascar
#' * [madCoast0][madCoast0], [madCoast4][madCoast4], and [madCoast][madCoast]: Borders of an eastern portion of Madagascar
#' * [madCover]: Land cover raster
#' * [madCoverCats][madCoverCats]: Table of land cover classes
#' * [madDypsis][madDypsis]: Specimens records of species in the genus *Dypsis*
#' * [madElev][madElev]: Elevation raster
#' * [madForest2000][madForest2000] and [madForest2014][madForest2014]: Forest cover in 2000 and 2014
#' * [madLANDSAT][madLANDSAT]: Surface reflectance in 2023
#' * [madPpt][madPpt], [madTmin][madTmin], [madTmax][madTmax]: Rasters of mean monthly precipitation, and minimum and maximum temperature
#' * [madRivers][madRivers]: Rivers vector
#' * [vegIndices][vegIndices]: Vegetation indices that can be calculated using [vegIndex()]
#' 
#' ## Esoteric tutorials and arcane notes
#' * Comparisons between `GRegion`s can be performed using the `==` and `!=` operators.
#' * Vignette on **GRASS** "projects/locations" and "mapsets": `vignette("projects_mapsets", package = "fasterRaster")`
#' * Vignette on **GRASS** "regions": `vignette("regions", package = "fasterRaster")`
#' * Vignette on **GRASS** 3-dimensional `GRaster`s and `GVector`s: `vignette("three_d_objects", package = "fasterRaster")`
#' * Vignette on **fasterRaster** hidden functions: `vignette("hidden_functions", package = "fasterRaster")`
#'
#' ## Classes
#' * [GLocation][GLocation]: Fundamental class; points to a "location/project" in **GRASS**
#' * [GSpatial][GSpatial]: Basic class of any spatial object
#' * [GRegion][GRegion]: Points to a "region" of a "location/project" in **GRASS**
#' * [GRaster][GRaster]: Raster class
#' * [GVector][GVector]: Spatial vector class
#'
#' @author Adam B. Smith
#' @name fasterRaster
#' @keywords internal
"_PACKAGE"

#' Create a hexagonal grid
#'
#' @description This function creates a `GVector` of "wall-to-wall" hexagons. The input can be a `GVector` or `GRaster`, which provides the extent of the output.
#'
#' @param x A `GRaster` or `GVector`.
#'
#' @param ny Integer or numeric integer: Number of rows of hexagons that span the extent of object `x`.
#'
#' @param expand One or two numeric values: Expand the region by this proportion in both directions (a single value) or in the x- and y-dimensions separately. Expanding the region can be helpful to ensure the entire area of interest is covered by polygons, which can otherwise leave gaps at the edges. The number of rows and columns will be increased, but the number of hexagons that span `x` will still be `ny`.
#'
#' @param angle Numeric: Degrees by which to rotate grid (from north, clockwise).
#'
#' @returns A `GVector`.
#'
#' @example man/examples/ex_grid_hexagons.r
#'
#' @seealso [grid()], tool `v.mkgrid` in **GRASS**
#'
#' @aliases hexagons
#' @rdname hexagons
#' @exportMethod hexagons
methods::setMethod(
	f = "hexagons",
	signature = c(x = "GRaster"),
	function(x, ny = 10, expand = 0, angle = 0) .hexagons(x, ny = ny, expand = expand, angle = angle)
)

#' @aliases hexagons
#' @rdname hexagons
#' @exportMethod hexagons
methods::setMethod(
	f = "hexagons",
	signature = c(x = "GVector"),
	function(x, ny = 10, expand = 0, angle = 0) .hexagons(x, ny = ny, expand = expand, angle = angle)
)

#' @noRd
.hexagons <- function(x, ny, expand, angle) {

	.locationRestore(x)
	.region(x)

	extent <- ext(vector = TRUE)
	ysize <- (extent[4L] - extent[3L]) / ny

	# expand region?
	xp <- expand[1L]
	xp <- omnibus::compareFloat(xp, 0, "!=")

	if (length(expand) == 1L) {
		yp <- xp
	} else {
		yp <- expand[2L]
		yp <- omnibus::compareFloat(yp, 0, "!=")
	}

	if (xp | yp) {
	
		if (length(expand) == 1L) expand[2L] <- expand[1L]

		extent <- ext(vector = TRUE)
		xdim <- extent[2L] - extent[1L]
		ydim <- extent[4L] - extent[3L]
		
		xdelta <- 0.5 * expand[1L] * xdim
		ydelta <- 0.5 * expand[2L] * ydim

		extent <- c(
			extent[1L] - xdelta,
			extent[2L] + xdelta,
			extent[3L] - ydelta,
			extent[4L] + ydelta
		)

		.regionExt(extent, respect = "dimensions")

	}

	src <- .makeSourceName("v_mkgrid", "vector")
	args <- list(
		cmd = "v.mkgrid",
		map = src,
		position = "region",
		# grid = c(ycells, 0), # order is reversed
		box = c(ysize, ysize),
		flags = c(.quiet(), "overwrite", "h")
	)

	angle <- 360 - angle
	if (!(angle %in% c(0, 360))) args$angle <- angle

	do.call(rgrass::execGRASS, args = args)
	makeGVector(src)

}

#' Create a random raster with or without spatial dependence
#'
#' @description `rSpatialDepRast()` creates a raster with random values in cells. Across the raster, values are approximately normally distributed, though a raster with a "true" normal distribution can be made with [rNormRast()]. Spatial dependence can be introduced, though all together the values will still be approximately normally distributed.
#'
#' @param x A `GRaster`: The output will have the same extent and dimensions as this raster.
#'
#' @param n An integer: Number of rasters to generate. 
#'
#' @param mu,sigma Numeric: Mean and sample standard deviation of output. If creating more than one raster, you can provide one value per raster. If there are fewer, they will be recycled.
#'
#' @param dist Numeric: Maximum distance of spatial autocorrelation (in map units--typically meters). Default is 0 (no spatial autocorrelation). If creating more than one raster, you can provide one value per raster. If there are fewer, values will be recycled.
#'
#' @param exponent Numeric > 0: Distance decay exponent. If creating more than one raster, you can provide one value per raster. If there are fewer, values will be recycled.
#'
#' @param delay Numeric >= 0: Values >0 force the distance decay of similarity to remain constant up to this distance. Beyond this distance, the decay exponent takes effect. Default is 0. If creating more than one raster, you can provide one value per raster. If there are fewer, values will be recycled.
#' 
#' @param seed Numeric integer or `NULL`: Random seed. If `NULL`, then a random seed is used for each raster. If provided, there should be one seed value per raster.
#'
#' @returns A `GRaster`.
#'
#' @example man/examples/ex_randRast.r
#' 
#' @seealso [rNormRast()], [fractalRast()], [rUnifRast()], [rWalkRast()], **GRASS** manual page for tool `r.random.surface` (see `grassHelp("r.random.surface")`)
#'
#' @aliases rSpatialDepRast
#' @rdname rSpatialDepRast
#' @exportMethod rSpatialDepRast
methods::setMethod(
    f = "rSpatialDepRast",
    signature = c(x = "GRaster"),
    function(
        x,
        n = 1,
        mu = 0,
        sigma = 1,
        dist = 0,
        exponent = 1,
        delay = 0,
        seed = NULL
    ) {

    if (any(dist < 0)) stop("Argument `dist` must be >= 0.")
    if (any(exponent <= 0)) stop("Argument `exponent` must be > 0.")
    if (any(delay < 0)) stop("Argument `delay` must be >= 0.")
    if (!is.null(seed)) if (length(seed) != n) stop("You must provide one value of `seed` per raster, or set it to NULL.")

    mu <- rep(mu, length.out = n)
    sigma <- rep(sigma, length.out = n)
    dist <- rep(dist, length.out = n)
    exponent <- rep(exponent, length.out = n)
    delay <- rep(delay, length.out = n)

    .locationRestore(x)
    .region(x)

    srcRands <- .makeSourceName("rand", "raster", n)
    srcs <- .makeSourceName("randScaled", "raster", n)

    for (i in seq_len(n)) {

        args <- list(
            cmd = "r.random.surface",
            output = srcRands[i],
            distance = dist[i],
            exponent = exponent[i],
            flat = delay[i],
            flags = c(.quiet(), "overwrite")
        )
        if (!is.null(seed)) args$seed <- seed[i]
        do.call(rgrass::execGRASS, args = args)

        xmu <- .global(srcRands[i], fun = "mean")
        xsigma <- .global(srcRands[i], fun = "sdpop")

        ex <- paste0(srcs[i], " = ", mu[i], " + ", sigma[i], " * (", srcRands[i], " - ", xmu, ") / ", xsigma)

        rgrass::execGRASS(
            cmd = "r.mapcalc",
            expression = ex,
            flags = c(.quiet(), "overwrite")
        )
        
    } # next raster
    makeGRaster(srcs, "rnorm")

    } # EOF
)

#' Spatial resolution
#'
#' @description Spatial resolution of a `GRaster`:
#' * `res()`: 2-dimensional resolution (x and y).\cr\cr
#' * `res3d()`: 3-dimensional resolution (z, y, and z).\cr\cr
#' * `xres()`, `yres()`, and `zres()`: East-west resolution, north-south resolution, and top-bottom resolution.
#'
#' @param x A `GRaster`, `GRegion`, or missing. If missing, the resolution of the currently active "region" is returned (see `vignette("regions", package = "fasterRaster")`).
#'
#' @return A numeric vector. For both `res()` and `res3d()`, the first value is the length of cells in the x-direction and the second the length of cells in the y-direction. For `res3d()` the third value is height of a voxel (the z-direction). `xres()`, `yres()`, and `zres()` each return a single value.
#'
#' @seealso [terra::res()]
#'
#' @example man/examples/ex_GRaster.r
#'
#' @aliases res
#' @rdname res
#' @exportMethod res
#' @keywords res
setMethod(
	f = "res",
	signature = "missing",
 	definition = function(x) res(.region())
)

#' @aliases res
#' @rdname res
#' @exportMethod res
setMethod(
	f = "res",
	signature = "GRegion",
	definition = function(x) x@resolution[1L:2L]
)

#' @aliases xres
#' @rdname res
#' @exportMethod xres
setMethod(
	f = "xres",
	signature = "missing",
 	definition = function(x) xres(.region())
)

#' @aliases xres
#' @rdname res
#' @exportMethod xres
setMethod(
	f = "xres",
	signature = "GRegion",
	definition = function(x) x@resolution[1L]
)

#' @aliases yres
#' @rdname res
#' @exportMethod yres
setMethod(
	f = "yres",
	signature = "missing",
 	definition = function(x) yres(.region())
)

#' @aliases yres
#' @rdname res
#' @exportMethod yres
setMethod(
	f = "yres",
	signature = "GRegion",
	definition = function(x) x@resolution[2L]
)

#' @aliases zres
#' @rdname res
#' @exportMethod zres
setMethod(
	f = "zres",
	signature = "missing",
 	definition = function(x) zres(.region())
)

#' @aliases zres
#' @rdname res
#' @exportMethod zres
setMethod(
	f = "zres",
	signature = "GRegion",
	definition = function(x) x@resolution[3L]
)

#' @aliases res3d
#' @rdname res
#' @exportMethod res3d
setMethod(
	f = "res3d",
	signature = "missing",
 	definition = function(x) res3d(.region())
)

#' @aliases res3d
#' @rdname res
#' @exportMethod res3d
setMethod(
	f = "res3d",
	signature = "GRegion",
	definition = function(x) c(xres=x@resolution[1L], yres=x@resolution[2L], zres=x@resolution[3L])
)

#' Hillshading
#' 
#' @description Hillshade rasters are often used for display purposes because they make topographical relief look "real" to the eye.
#'
#' @param x A `GRaster` (typically representing elevation).
#'
#' @param angle Numeric: The altitude of the sun above the horizon in degrees. Valid values are in the range \[0, 90\], and the default value is 45 (half way from the horizon to overhead).
#'
#' @param direction The direction (azimuth) in which the sun is shining in degrees. Valid values are in the range 0 to 360. The default is 0, meaning the sun is at due south (180 degrees) and shining due north (0 degrees). Note that in this function, 0 corresponds to north and 180 to south, but in the **GRASS** tool `r.relief`, "east orientation" is used (0 is east, 90 is north, etc.).
#'
#' @param zscale Numeric: Value by which to exaggerate terrain. The default is 1.  Numbers greater than this will increase apparent relief, and less than this (even negative) will diminish it.
#'
#' @returns A `GRaster`.
#'
#' @example man/examples/ex_terrain.r
#' 
#' @aliases hillshade
#' @rdname hillshade
#' @exportMethod hillshade
methods::setMethod(
    f = "hillshade",
    signature = c(x = "GRaster"),
    definition = function(x, angle = 45, direction = 0, zscale = 1) {

    if (angle < 0 | angle > 90) stop("Argument `angle` must be in the range [0, 90].")
    if (direction < 0 | direction > 360) stop("Argument `angle` must be in the range [0, 360].")

    .locationRestore(x)
    .region(x)

    # convert from east to north orientation
    direction <- ((360 - direction) %% 360 + 90) %% 360

    nLayers <- nlyr(x)
    srcs <- .makeSourceName("shade", "raster", n = nLayers)
    for (i in seq_len(nLayers)) {

        args <- list(
            cmd = "r.relief",
            input = sources(x)[i],
            output = srcs[i],
            altitude = angle,
            azimuth = direction,
            zscale = zscale,
            flags = c(.quiet(), "overwrite")
        )

        do.call(rgrass::execGRASS, args = args)

    }
    makeGRaster(srcs, paste0(names(x), "_shade"))

    } # EOF
)

#' Category column values of a GRASS vector
#'
#' @description Returns values in the `cat` column of a vector in **GRASS**.
#'
#' @param x A `GVector` or the name of a vector in **GRASS**.
#'
#' @param layer Integer, numeric integer, or character: Layer from which to obtain category values.
#'
#' @param db Logical: If `TRUE`, return category numbers from the database table associated with the vector. If `FALSE` (default), return category numbers from the actual vector.
#'
#' @param integer Logical: If `TRUE` (default), return category values as integers. In some cases, a geometry can have multiple categories, in which case `NA` is returned. If `FALSE`, return category values as strings (and thus, if a geometry has more than one category, does not convert to `NA`).
#'
#' @returns A vector.
#'
#' @aliases .vCats
#' @rdname vCats
#' @keywords internal
.vCats <- function(x, layer = 1, db = FALSE, integer = TRUE) {

	if (inherits(x, "GVector")) {
		.locationRestore(x)
		src <- sources(x)
	} else {
		src <- x
	}

	layer <- as.character(layer)

	if (db) {
		dbase <- .vAsDataTable(src)
		if (any(names(dbase) == "frid")) {
			out <- dbase[["frid"]]
		} else if (any(names(dbase) == "cat")) {
			out <- dbase[["cat"]]
		} else {
			out <- NULL
		}
	} else {

		suppressMessages(out <- rgrass::execGRASS(
			cmd = "v.category",
			input = src,
			layer = layer,
			option = "print",
			flags = .quiet(),
			intern = TRUE,
			Sys_show.output.on.console = FALSE,
			echoCmd = FALSE
		))
	
	}

	if (integer) {
		
		# # ensure categories are ordered
		# out <- strsplit(out, split = "/")
		# out <- lapply(out, as.integer)
		# out <- lapply(out, sort)
		# out <- sapply(out, paste, collapse = "/")
		
		# # assign each category combination a new category value
		# uniq <- unique(out)
		# nuniq <- length(uniq)
		
		# out <- factor(out, seq_len(nuniq))
		# out <- levels(out)
		
		out <- suppressWarnings(as.integer(out))
		
	}
	out

}

if (grassStarted()) {

# Setup
library(terra)

# Elevation raster, rivers vector
madLANDSAT <- fastData("madLANDSAT")

# Convert a SpatRaster to a GRaster:
landsat <- fast(madLANDSAT)

# See available vegetation indices:
vegIndices

# Normalized Difference Vegetation Index and Enhanced Vegetation Index:
indices <- c("ndvi", "evi")
vi <- vegIndex(landsat, index = indices, r = 1, b = 3, nir = 4, bits = 8)
plot(vi)

# All indices using R and NIR:
rnir <- vegIndex(landsat, index = "rnir", r = 1, nir = 4, bits = 8)

# Note: Some values are highly skewed
plot(rnir)

}

if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev") # raster
madCover <- fastData("madCover") # categorical raster

# Convert to GRasters
elev <- fast(madElev) # raster
cover <- fast(madCover) # categorical raster

# Frequencies of integer raster values
f <- freq(elev)
print(f) # have to do this sometimes if output is a data table

# Frequencies of categorical raster values
f <- freq(cover)
print(f) # have to do this sometimes if output is a data table

# Frequencies of given values
f <- freq(elev, value = 1)
print(f) # have to do this sometimes if output is a data table

# When a GRaster has non-integer values, they will be binned:
f <- freq(elev + 0.1, bins = 10)
print(f)

}

#' Create red-green-blue plot from a raster with RGB layers
#'
#' @description This function takes as its main argument a `GRaster` with at least three layers typically representing red, green, and blue components (plus possibly an "alpha", or transparency layer). As with [plot()], this function is somewhat of a hack in that it downsamples the layers to a coarser resolution using [aggregate()], saves the raster to disk, then uses [terra::plotRGB()] to do the actual plotting.
#'
#' @param x A `GRaster`. Values must be in the range from 0 to 255.
#'
#' @param r,g,b Either a numeric integer or the [names()] of layers representing red, green, and blue components.
#'
#' @param a Either `NULL` (default), or a numeric integer or the [names()] of a layer representing transparency.
#'
#' @param simplify Logical: If `TRUE` (default), then downsample the `GRaster` before plotting. This can save time for very dense rasters.
#'
#' @param ... Arguments to pass to [terra::plotRGB()].
#'
#' @returns Nothing (makes a plot).
#'
#' @seealso [terra::plotRGB()], [plot()], [compositeRGB()]
#'
#' @example man/examples/ex_plot.r
#'
#' @aliases plotRGB
#' @rdname plotRGB
#' @exportMethod plotRGB
methods::setMethod(
	f = "plotRGB",
	signature = c(x = "GRaster"),
	function(x, r = 1, g = 2, b = 3, a = NULL, simplify = TRUE, ...) {
	
	.locationRestore(x)
	.region(x)
	
	# simplify
	# simplify
	if (simplify & !is.3d(x)) {
		
		# screenRes <- rpanel::rp.screenresolution()
		screenRes <- omnibus::screenRes()
		screenWidth <- screenRes[["width"]]
		screenHeight <- screenRes[["width"]]

		rows <- nrow(x)
		cols <- ncol(x)
		# pow <- 5 # power by which to reduce dimensions, set at 5 semi-empirically
		if (screenWidth < cols | screenHeight < rows) {
		
			ewScale <- max(1, floor(0.5 * cols / screenWidth))
			nsScale <- max(1, floor(0.5 * rows / screenHeight))

			if (ewScale >= 2 | nsScale >= 2) {

				if (any(datatype(x) %in% c("integer", "factor"))) {
					fun <- "mode"
				} else {
					fun <- "mean"
				}

				fact <- c(ewScale, nsScale)
				rgb <- c(r, g, b)
				if (!is.null(a)) rgb <- c(rgb, a)
				xx <- xx[[rgb]]
				x <- aggregate(xx, fact = fact, fun = fun)

			}
			
		}

	}
	
	tf <- tempfile(fileext = ".tif")
	writeRaster(x, tf, overwrite = TRUE)
	xx <- terra::rast(tf)
	
	terra::plotRGB(xx, r = r, g = g, b = b, a = a, ...)
	
	} # EOF
)

#' Smooth the geometry of a vector
#'
#' @description `smoothGeom()` makes line segments of a vector appear less angular.
#'
#' @param x A `GVector`.
#'
#' @param method Character: Method used to smooth line segments. Partial matching is used, and case does not matter:
#'
#' * `"Hermite"`: Hermite interpolation (default): Guarantees that the output vector always passes through the original points. This method adds points (possibly many) by constructing cubic splines with points approximately `dist` apart. The number of points can be reduced by specifying a smaller value of `angle`, which specifies the minimum angle between two successive line segments.
#' * `"Chaiken"`: Chaiken's algorithm: Guarantees that the new vector always touches the midpoint of each original line segment. The points on the new line are at least `dist` apart.
#'
#' @param dist Numeric > 0 of `NULL` (default): Minimum distance (see `method`). Units are in map units. If `NULL`, then 2% of the minimum of the x-, y-, and z-extent will be used.
#'
#' @param angle Numeric > 0: Maximum angle for the Hermite algorithm. Default is 3.
#'
#' @seealso [simplifyGeom()], [terra::simplifyGeom()], [geometry cleaning][breakPolys], **GRASS** manual page for tool `v.generalize` (see `grassHelp("v.generalize")`)
#'
#' @returns A `GVector`.
#'
#' @example man/examples/ex_simplify_smooth_clean_GVector.r
#'
#' @aliases smoothGeom
#' @rdname smoothGeom
#' @exportMethod smoothGeom
methods::setMethod(
	f = "smoothGeom",
	signature = c(x = "GVector"),
	function(x, method = "Hermite", dist = NULL, angle = 3) {

	# automatic distance
	if (is.null(dist)) {
		extent <- ext(x, vector=TRUE)
		xext <- extent[2L] - extent[1L]
		yext <- extent[4L] - extent[3L]
		zext <- diff(zext)
		dist <- 0.02 * min(xext, yext, zext, na.rm=TRUE)
	} else {
		if (dist < 0) stop("Argument `dist` must be > 0.")
	}

	.locationRestore(x)
	
	method <- tolower(method)
	method <- omnibus::pmatchSafe(method, c("hermite", "chaiken"))
	
	if (method == "hermite" && angle < 0) stop("Argument `angle` must be >0.")
	
	src <- .makeSourceName("generalized", "vect")
	args <- list(
		cmd = "v.generalize",
		input = sources(x),
		output = src,
		method = method,
		threshold = dist,
		flags = c(.quiet(), "overwrite"),
		intern = TRUE
	)

	if (method == "hermite") args$angle_thresh <- angle
	
	do.call(rgrass::execGRASS, args=args)
	makeGVector(src)

	} # EOF
)

if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Elevation raster and points at which Dypsis have been collected:
madElev <- fastData("madElev")
madDypsis <- fastData("madDypsis")

# Convert a SpatRaster to a GRaster and sf to GVector:
elev <- fast(madElev)
dypsis <- fast(madDypsis)

# We have an elevation raster but will pretend we don't and interpolate
# the value of elevation associated with each point.

# Extract elevation values from elevation raster:
dypsisElev <- extract(elev, dypsis)

# Add column to dypsis:
dypsis <- colbind(dypsis, dypsisElev)

# For expediency, we will also increase the size of spline steps in the
# x- and y-dimensions:
xlength <- 10 * xres(elev)
ylength <- 10 * yres(elev)

# Interpolate to a raster (need to increase size of spline length first):
interpElev <- interpSplines(dypsis, y = elev, field = "madElev",
lambda = 0.01, xlength = xlength, ylength = ylength)

# Plot:
oldpar <- par(mfrow = c(1, 2))
plot(elev, main = "Observed")
plot(dypsis, pch = 1, add = TRUE)

plot(interpElev, main = "Interpolated")
plot(dypsis, pch = 1, add = TRUE)

par(oldpar)

### Find optimal lambda using cross-validation (takes a while):
lambdas <- interpSplines(dypsis, y = elev, field = "madElev",
xlength = xlength, ylength = ylength, interpolate = FALSE)

lambdas

}

if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster
elev <- fast(madElev)

### Stretch based on user-defined range

#  fasterRaster
fr <- stretch(elev, smin=1, smax=100)
fr

# terra
tr <- stretch(madElev, smin = 1, smax = 100)
tr

# Compare fasterRaster to terra output
fr <- rast(fr)
fr <- extend(fr, tr)
fr - tr

### Stretch values in a certain quantile range

#  fasterRaster
fr <- stretch(elev, minq = 0.25, maxq = 0.75)
fr

# terra
tr <- stretch(madElev, minq = 0.25, maxq = 0.75)
tr

# Compare fasterRaster to terra output
fr <- rast(fr)
fr <- extend(fr, tr)
fr - tr

}

if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster:
elev <- fast(madElev)

# Focal calculations:
sums <- focal(elev, fun = "sum")
means <- focal(elev, fun = "mean")

# Focal calculations on a circular window:
sds <- focal(elev, fun = "sd") # square
sdsCircle <- focal(elev, fun = "sd", circle = TRUE) # circle

sds
sdsCircle

plot(sds - sdsCircle)

# Focal calculations with user-defined weights:
w <- matrix(c(1, 0, 1, 0, 1, 0, 1, 0, 1), ncol = 3)
w
sumsWeighted <- focal(elev, fun = "sum", w = w)

s <- c(sums, sumsWeighted)
minmax(s)

}

#' Combine two or more rasters with different extents and fill in NAs
#'
#' @description `merge()` combines two or more `GRaster`s, possibly with different extents, into a single larger `GRaster`. Where the same cell has different values in each raster, the value of the first raster's cell is used. If this is `NA`, then the value of the second raster's cell is used, and so on.
#'
#' @param x,y,... `GRaster`s.
#'
#' @returns A `GRaster`.
#' 
#' @seealso [terra::merge()], [terra::mosaic()], and **GRASS** tool `r.patch`
#'
#' @example man/examples/ex_merge.r
#'
#' @aliases merge
#' @rdname merge
#' @exportMethod merge
methods::setMethod(
	f = "merge",
	signature = c(x = "GRaster", y = "GRaster"),
	definition = function(x, y, ...) {

	.locationRestore(x)
	compareGeom(x, y, lyrs=TRUE, ext=FALSE, rowcol=FALSE, depths=TRUE, res=TRUE, zres=TRUE)
	
	x <- list(x, y, ...)

	# set region to combined extent
	rasts <- paste(sapply(x, sources), collapse=",")
	rgrass::execGRASS("g.region", raster = rasts, flags=c("o", .quiet()))

	# combine
	src <- .makeSourceName("merge_r_patch", "raster")
	rgrass::execGRASS(
		cmd = "r.patch",
		input = rasts,
		output = src,
		nprocs = faster("cores"),
		memory = faster("memory"),
		flags = c(.quiet(), "overwrite")
	)

	# combine levels
	levels <- combineLevels(x)
	makeGRaster(src, levels = levels)
	
	} # EOF
)

#' Contour lines from a "GRaster"
#'
#' @description
#' Create a `GVector` of contour lines from a `GRaster`.
#'
#' @param x A `GRaster`.
#'
#' @param nlevels Numeric: A positive integer or missing (default). Number of levels at which to calculate contours. Levels will be calculated in equal-sized steps from the smallest to the largest value of `x`. Either `nlevels` or `levels` must be specified.
#'
#' @param levels Numeric vector: A numeric vector of values at which to calculate contour lines. Either `nlevels` or `levels` must be specified.
#'
#' @seealso [terra::as.contour()], **GRASS** manual page for tool `r.contour` (see `grassHelp("r.contour")`)
#'
#' @returns A `GVector` representing contour lines.
#'
#' @example man/examples/ex_asContour.r
#'
#' @aliases as.contour
#' @rdname as.contour
#' @exportMethod as.contour
setMethod(
	"as.contour",
	signature(x = "GRaster"),
	function(x, nlevels, levels) {
	
	.locationRestore(x)
	.region(x)

	if (!missing(nlevels) & !missing(levels)) stop("Please specify either `nlevels` or `levels`, but not both.")
	if (missing(nlevels) & missing(levels)) stop("Please specify either `nlevels` or `levels`.")

	if (!missing(nlevels)) {
		mm <- minmax(x)
		levels <- seq(mm[1L, 1L], mm[2L, 1L], length.out=nlevels + 2L) # adding 2 bc creates no contours for min/max values
	}

	input <- sources(x)
	src <- .makeSourceName("r_contour", "vector")

	### execute
	args <- list(
		cmd = "r.contour",
		input = input,
		output = src,
		levels = levels,
		cut = 2,
		flags = .quiet()
	)
	
	do.call(rgrass::execGRASS, args=args)
	makeGVector(src)
	
	} # EOF
)

#' Area of polygons or length of lines
#'
#' @description This function calculates the area of each polygon in a "polygons" `GVector` or the length of lines in a "lines" `GVector`.
#'
#' @param x A "polygons" or "lines" `GVector`.
#'
#' @param unit Character: Units in which to report values. Areal units are squared, linear are not. Can be any of:
#' * `"meters"`(default), `"metres"`, or `"m"`
#' * `"km"` or `"kilometers"`
#' * `"ha"` or `"hectares"`
#' * `"ft"` or `"feet"`
#' * `"ac"` or `"acres"`
#' * `"percent"`
#'
#' Partial matching is used and case is ignored.
#'
#' @returns Numeric values, one per geometry in `x`.
#'
#' @example man/examples/ex_GVector.r
#'
#' @aliases expanse
#' @rdname expanse
#' @exportMethod expanse
methods::setMethod(
	f = "expanse",
	signature = c(x = "GVector"),
	function(x, unit = "m") {
	
	gtype <- geomtype(x, grass = TRUE)
	if (!(gtype %in% c("area", "line"))) stop("GVector must represent lines or polygons.")

	.locationRestore(x)
	
	units <- c("m", "m2", "meters2", "metres2", "km2", "kilometers2", "acres", "ha", "hectares", "ft2", "feet2", "mi2", "miles2", "%", "percent")
	unit <- omnibus::pmatchSafe(unit, units, useFirst = TRUE, nmax = 1L)
	if (unit == "m2") unit <- "meters"
	if (unit == "meters2") unit <- "meters"
	if (unit == "metres2") unit <- "meters"
	if (unit == "km2") unit <- "kilometers"
	if (unit == "kilometers2") unit <- "kilometers"
	if (unit == "mi2") unit <- "miles"
	if (unit == "miles2") unit <- "miles"
	if (unit == "ft2") unit <- "feet"
	if (unit == "feet2") unit <- "feet"
	if (unit == "ha") unit <- "hectares"
	if (unit == "%") unit <- "percent"
	unit <- omnibus::expandUnits(unit)

	option <- if (gtype == "area") {
		"area"
	} else if (gtype == "line") {
		"length"
	}

	info <- rgrass::execGRASS(
		cmd = "v.report",
		map = sources(x),
		option = option,
		units = unit,
		flags = c(.quiet(), "d"),
		intern = TRUE
	)

	info <- strsplit(info, split = "\\|")
	header <- info[[1L]]
	
	if (gtype == "line") {
		column <- which(header == "length")
	} else {
		column <- which(header == "area")
	}

	out <- rep(NA_character_, ngeom(x))
	for (i in seq_along(out)) {
	
		val <- info[[i + 1L]][column]
		out[i] <- val
	
	}
	out <- as.numeric(out)
	out

	} # EOF
)

if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")

# convert a SpatRaster to a GRaster
elev <- fast(madElev)

# calculate horizon height in north and east directions
hhNorth <- horizonHeight(elev)
hhNorth
plot(hhNorth)

# calculate horizon height in east and north directions
hhEast <- horizonHeight(elev, northIs0 = FALSE)
hhEast
plot(hhEast)

}

#' Multivariate environmental similarity surface (MESS)
#'
#' @description The multivariate environmental similarity surface (MESS) indicates the degree to which a set of "projection" environmental conditions fall inside or outside a set of "reference" conditions. Values of 1 indicate a location falls at the exact median of all variables. Values of 0 indicate that the location has at least one environmental covariate that is at the upper or lower end of the range of reference conditions, and values <1 indicate that at least one variable falls above or below the reference conditions. MESS can be used, for example, to indicate the degree to which a model constructed in one time period and/or location must extrapolate when projected to another time period and/or location.
#'
#' @param ref A `data.frame`, `data.table`, a points `GVector`, or "stack" of `GRasters`: This represents the set of "reference" environmental conditions:
#' * `data.frame` or `data.table`: There must be one column per layer in `proj`, and the columns must have the same names as the layers in `proj`.
#' * `GRaster` with one or more layers: Must have the same [names()] as the `GRaster`s in `proj`. Values are assumed to be continuous (not categorical/factors).
#'
#' @param proj A `GRaster` or missing. If a `GRaster`, it must have the same layers as can have with one or more layers as `ref`. Values are assumed to be continuous (not categorical/factors). If missing, then `ref` is used, in which case the output represents the relative difference of each cell from the overall layer median.
#'
#' @param na.rm Logical: If `FALSE` (default), and `ref` is a `data.frame`, `data.table`, or `matrix`, and has `NA`s in it, then the function will likely fail.
#'
#' @returns A `GRaster` "stack". There will be one layer per layer in `ref`, indicating the MESS score for that variable. There will also be a layer named "MESS" which represents the MESS value across all variables (the minimum value of each of the individual MESS rasters). A final layer represents the layer which is most different (has the lowest MESS value).
#'
#' @references Elith, J, Kearney, M, and Phillips, S. 2010. The art of modelling range-shifting species. *Methods in Ecology and Evolution* 1:330-342. \doi{10.1111/j.2041-210X.2010.00036.x} (see especially the Supplement)
#' 
#' @example man/examples/ex_multivarEnvSim.r
#'
#' @aliases multivarEnvSim
#' @rdname multivarEnvSim
#' @exportMethod multivarEnvSim
methods::setMethod(
	f = "multivarEnvSim",
	signature = c(ref = "GRaster", proj = "GRaster"),
	definition = function(
		ref,
		proj
	) {

	lyrs <- nlyr(ref) == nlyr(proj)
	names1 <- names(ref) %in% names(proj)
	names2 <- names(proj) %in% names(ref)

	if (!all(lyrs, names1, names2)) stop("The `ref` and `proj` set of GRasters must have the same number of layers with the same names.")

	ref <- ref[[names(proj)]]

	medians <- global(ref, "median")
	mm <- minmax(ref)

	.locationRestore(proj)
	.region(proj)

	.multivarEnvSim(ref = ref, proj = proj, medians = medians, mm = mm)

	} # EOF
)

#' @aliases multivarEnvSim
#' @rdname multivarEnvSim
#' @exportMethod multivarEnvSim
methods::setMethod(
	f = "multivarEnvSim",
	signature = c(ref = "GRaster", proj = "missing"),
	definition = function(ref, proj) multivarEnvSim(ref = ref, proj = ref)
)

#' @aliases multivarEnvSim
#' @rdname multivarEnvSim
#' @exportMethod multivarEnvSim
methods::setMethod(
	f = "multivarEnvSim",
	signature = c(ref = "data.frame", proj = "GRaster"),
	definition = function(ref, proj, na.rm = FALSE) {
	
	names2 <- names(proj) %in% names(ref)
	if (!all(names2)) stop("Not all layer names in `proj` appear in `ref.")

	.locationRestore(proj)
	.region(proj)

	ref <- ref[ , names(proj), drop = FALSE]

	medians <- matrix(apply(ref, 2, "median", na.rm = na.rm), ncol = 1)
	rownames(medians) <- colnames(ref)
	colnames(medians) <- "median"

	mins <- apply(ref, 2, "min", na.rm = na.rm)
	maxs <- apply(ref, 2, "max", na.rm = na.rm)

	mm <- matrix(c(mins, maxs), nrow = 2, byrow = TRUE)
	rownames(mm) <- c("min", "max")
	colnames(mm) <- names(ref)

	.multivarEnvSim(ref = ref, proj = proj, medians = medians, mm = mm)
	
	} # EOF
)

#' @aliases multivarEnvSim
#' @rdname multivarEnvSim
#' @exportMethod multivarEnvSim
methods::setMethod(
	f = "multivarEnvSim",
	signature = c(ref = "data.table", proj = "GRaster"),
	definition = function(ref, proj, na.rm = FALSE) {
	
	names2 <- names(proj) %in% names(ref)
	if (!all(names2)) stop("Not all layer names in `proj` appear in `ref.")

	.locationRestore(proj)
	.region(proj)

	cols <- names(proj)
	ref <- ref[ , ..cols]

	medians <- matrix(apply(ref, 2, "median", na.rm = na.rm), ncol = 1)
	rownames(medians) <- colnames(ref)
	colnames(medians) <- "median"

	mins <- apply(ref, 2, "min", na.rm = na.rm)
	maxs <- apply(ref, 2, "max", na.rm = na.rm)

	mm <- matrix(c(mins, maxs), nrow = 2, byrow = TRUE)
	rownames(mm) <- c("min", "max")
	colnames(mm) <- names(ref)

	.multivarEnvSim(ref = ref, proj = proj, medians = medians, mm = mm)
	
	} # EOF
)

#' @aliases multivarEnvSim
#' @rdname multivarEnvSim
#' @exportMethod multivarEnvSim
methods::setMethod(
	f = "multivarEnvSim",
	signature = c(ref = "matrix", proj = "GRaster"),
	definition = function(ref, proj, na.rm = FALSE) {

	ref <- data.table::as.data.table(ref)
	multivarEnvSim(ref = ref, proj = proj, na.rm = na.rm)
	
	} # EOF
)

# Hidden function for multivariate environmental similarity surface (MESS)
#
# ref `GRaster`, `data.frame`, `data.table`, or `matrix`
# proj `GRaster`
# medians Numeric: Vector of median values of `ref`
# mm Numeric matrix of minium/maximum values of `ref`. Top rows is miniums and bottom is maximums. Columns must have same names as `ref`.
.multivarEnvSim <- function(ref, proj, medians, mm) {

	### MESS values for each variable
	nLayers <- nlyr(proj)
	srcs <- .makeSourceName("mess_r_mapcalc", "raster", nLayers)
	for (i in seq_len(nLayers)) {
	
		layer <- names(proj)[i]
		ySrc <- sources(proj)[i]

		thisMedian <- medians[layer, "median"]
		thisMin <- mm["min", layer]
		thisMax <- mm["max", layer]

		medianToMax <- thisMax - thisMedian
		medianToMin <- thisMedian - thisMin

		ex <- paste0(srcs[i], " = if(", ySrc," >= ", thisMedian, ", 1 - ((", ySrc, " - ", thisMedian, ") / ", medianToMax, "), 1 - ((", thisMedian, " - ", ySrc, ") / ", medianToMin, "))")

		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))
	
	}

	### MESS values across all variables
	srcOverall <- .makeSourceName("mess_r_mapcalc", "raster")
	ex <- paste0(srcOverall, " = min(", paste(srcs, collapse = ","),")")
	rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

	### most different variable
	srcMostDiffFrom0 <- .makeSourceName("mess_r_series", "raster")
	rgrass::execGRASS(
		cmd = "r.series",
		input = paste(srcs, collapse=","),
		output = srcMostDiffFrom0,
		method = "min_raster",
		nprocs = faster("cores"),
		memory = faster("memory"),
		flags = c(.quiet(), "overwrite")
	)

	# add 1 bc "r.series::min_raster" returns 0 for the first raster
	srcMostDiffFrom1 <- .makeSourceName("mess_r_mapcalc", "raster")
	ex <- paste0(srcMostDiffFrom1, " = int(", srcMostDiffFrom0, " + 1)")
	rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

	mostDiffLevels <- data.table::data.table(value = 1L:nLayers, layer = names(proj))
	levs <- vector(mode = "list", length = nLayers + 2)
	levs[[length(levs)]] <- mostDiffLevels

	srcsAll <- c(srcs, srcOverall, srcMostDiffFrom1)
	outNames <- c(names(proj), "MESS", "mostDifferent")
	makeGRaster(srcsAll, names = outNames, levels = levs)


}

#' Convert a GRaster, or lines or polygons GVector to a points vector
#'
#' @description `as.points()` converts a `GRaster`, or a lines or polygons `GVector` to a points `GVector`.
#'
#' For `GRasters`, the points have the coordinates of cell centers and are assigned the cells' values. Only non-`NA` cells will be converted to points.
#'
#' For `GVectors`, each point will have the attributes of the line or polygon to which it belonged. Points are extracted from each vertex.
#' 
#' @param x A `GRaster`, `GVector`.
#' 
#' @param values Logical: If `TRUE` (default), create an attribute table with raster cell values, with one row per point.
#' 
#' @returns A `points` `GVector`.
#' 
#' @seealso [crds()], [as.lines()], [as.polygons()], [terra::as.points()], and modules `v.to.points` and `r.to.vect` in **GRASS**
#' 
#' @example man/examples/ex_asPoints.r
#' 
#' @aliases as.points
#' @rdname as.points
#' @exportMethod as.points
methods::setMethod(
    f = "as.points",
    signature = c(x = "GRaster"),
    function(x, values = TRUE) {
        
        if (nlyr(x) > 1L) warning("Only the first layer of `x` will be converted to points.")

        .locationRestore(x)
        .region(x)

        src <- .makeSourceName("as_points_r_to_vect", "vector")

        rgrass::execGRASS(
            cmd = "r.to.vect",
            input = sources(x)[1L],
            output = src,
            type = "point",
            column = names(x)[1L],
            flags = c(.quiet(), "overwrite")
        )

        if (values) {
            table <- .vAsDataTable(src)
            table$cat <- NULL
            table$label <- NULL
        } else {
            table <- NULL
        }

        # nLayers <- nlyr(x)

        # if (nLayers > 1L & values) {
        
        #     for (i in 2L:nLayers) {

        #         thisSrc <- .makeSourceName("as_points_r_to_vect", "vector")

        #         rgrass::execGRASS(
        #             cmd = "r.to.vect",
        #             input = sources(x)[i],
        #             output = thisSrc,
        #             type = "point",
        #             column = names(x)[i],
        #             flags = c(.quiet(), "overwrite")
        #         )

        #         thisTable <- .vAsDataTable(thisSrc)
        #         cols <- names(x)[i]
        #         thisTable <- thisTable[ , ..cols]

        #         table <- cbind(table, thisTable)

        #     }
        
        # } # next layer

        makeGVector(src, table = table)

    } # EOF
)

#' @aliases as.points
#' @rdname as.points
#' @exportMethod as.points
methods::setMethod(
    f = "as.points",
    signature = c(x = "GVector"),
    function(x) {

    gtype <- geomtype(x, grass = TRUE)
    if (gtype == "point") {
        warning("Vector object is already a points object.")
        return(x)
    }

    .locationRestore(x)

    src <- .makeSourceName("as_points_v_to_points", "vector")
    args <- list(
        cmd = "v.to.points",
        input = sources(x),
        output = src,
        use = "vertex",
        type = gtype,
        flags = c(.quiet(), "overwrite")
    )
    do.call(rgrass::execGRASS, args = args)
    makeGVector(src, table = NULL)

    } # E#OF
)

#' Solar radiance and irradiance
#'
#' The `sun()` function calculates beam (direct), diffuse and ground reflected solar irradiation for a given day and set of topographic and atmospheric conditions. The function relies on the **GRASS** tool `r.sun`, the manual page for which contains a detailed explanation (see `grassHelp("r.sun")`)
#'
#' @param elevation A `GRaster` with values representing elevation (typically in meters).
#'
#' @param coeff_bh A `GRaster`: A raster with values of the real-sky beam radiation coefficient. Valid values are between 0 and 1.
#'
#' @param coeff_dh A `GRaster`: A raster with values of the real-sky diffuse radiation coefficient. Valid values are between 0 and 1.
#'
#' @param slope A `GRaster`: This is a raster representing topographic slope in degrees. It can be generated using [terrain()].
#'
#' @param aspect A `GRaster`: This is a raster representing topographic aspect in degrees. It can be generated using [terrain()]. If generated with that function, "east orientation" *must* be used (i.e., argument `northIs0` must be `FALSE`).
#'
#' @param hh A "stack" of `GRaster`s: This represents height of the horizon in radians in particular directions. Horizon height can be calculated using [horizonHeight()]. The directions *must* be in "east orientation" (i.e., argument `northIs0` in `horzionHeight()` must be `FALSE`). The directions must correspond with the sequence given by `horizon_step` (see next argument). For example, if `horizon_step` is 90, then `hh` must contain rasters representing horizon height at 0 (east), 90 (north), 180 (west), and 270 (south) aspects.
#'
#' @param horizon_step Numeric >0: Difference between angular steps in which horizon height is measured. One horizon height raster will be made per value from 0 to (360 - `horizon_step`) degrees.
#'
#' @param albedo A `GRaster` or a numeric value: This is either a raster with values of ground albedo or a numeric value (in which case albedo is assumed to be the same everywhere). Albedo is unit-less, and the default value is 0.2.
#'
#' @param linke A `GRaster` or a numeric value: This is either a raster with values of the Linke atmospheric turbidity coefficient or a numeric value (in which case the same value is assumed for all locations). The Linke coefficient is unit-less. The default value is 3, but see also the **GRASS** manual page for tool `r.sun` (`grassHelp("r.sun")`).
#'
#' @param day Positive integer between 1 to 365, inclusive: Day of year for which to calculate ir/radiation. Default is 1 (January 1st).
#'
#' @param step Positive integer between 0 and 24, inclusive. Time step in hours for all-day radiation sums. Decimal values are OK.
#'
#' @param declination Numeric or `NULL` (default). Declination value. If `NULL`, this is calculated automatically.
#'
#' @param solar_constant Positive numeric: The solar constant (solar energy hitting the top of the atmosphere). Default is 1367. Units are W / m^2.
#'
#' @param distance_step Positive numeric between 0.5 and 1.5, inclusive: Sampling distance coefficient. Default is 1.
#'
#' @param npartitions Positive numeric. Number of chunks in which to read input files. Default is 1.
#'
#' @param beam_rad Logical: If `TRUE` (default), generate a raster with beam irradiation with units of Wh / m^2 / day ("mode 2" of the `r.sun` **GRASS** tool).
#'
#' @param diff_rad Logical: If `TRUE` (default), generate a raster representing irradiation in Wh / m^2 /day
#'
#' @param refl_rad Logical: If `TRUE` (default), generate a raster with ground-reflected irradiation with units of Wh / m^2 / day ("mode 2" of the `r.sun` **GRASS** tool).
#'
#' @param glob_rad Logical:. If `TRUE` (default), generate a raster with total irradiance/irradiation with units of Wh / m^2 / day ("mode 2" of the `r.sun` **GRASS** tool).
#'
#' @param insol_time Logical: If `TRUE` (default), generate a raster with total insolation time in hours ("mode 2" of the `r.sun` **GRASS** tool).
#'
#' @param lowMemory Logical: If `TRUE`, use the low-memory version of the `r.sun` **GRASS** tool. The default is `FALSE`.
#'
#' @returns A raster or raster stack stack with the same extent, resolution, and coordinate reference system as `elevation`. Assuming all possible rasters are generated they represent:
#' * `beam_rad`: Beam radiation (Watt-hours/m2/day)
#' * `diff_rad`: Diffuse radiation (Watt-hours/m2/day)
#' * `refl_rad`: Reflected radiation (Watt-hours/m2/day)
#' * `glob_rad`: Global radiation (Watt-hours/m2/day)
#' * `insol_time`: Insolation duration (hours)
#' 
#' @seealso [terrain()], [horizonHeight()], **GRASS** manual page for tool `r.sun` (see `grassHelp("r.sun")`)
#'
#' @example man/examples/ex_sun.r
#'
#' @export
sun <- function(
		elevation,
		coeff_bh,
		coeff_dh,

		slope,
		aspect,

		hh,
		horizon_step = 90,
		
		albedo = 0.2,
		linke = 3,

		day = 1,
		step = 0.5,
		declination = NULL,
		solar_constant = 1367,
		
		distance_step = 1,
		npartitions = 1,

		beam_rad = TRUE,
		diff_rad = TRUE,
		refl_rad = TRUE,
		glob_rad = TRUE,
		insol_time = TRUE,

		lowMemory = FALSE
		
	) {

	if (!beam_rad & !diff_rad & !refl_rad & !glob_rad & !insol_time) stop("No output requested.")

	compareGeom(elevation, coeff_bh, coeff_dh, slope, aspect, hh)
	
	.locationRestore(elevation)
	.region(elevation)

	hhSrc <- sources(hh)[1L]
	directions <- seq(0, 359.9999999, by = horizon_step)
	hhSrc <- gsub(hhSrc, pattern = paste0("_", sprintf("%03.0f", directions[1L])), replacement = "")
		
	args <- list(
		cmd = "r.sun",
		flags = c(.quiet(), "overwrite"),
		day = day,
		step = step,
		nprocs = faster("cores"),
		distance_step = distance_step,
		npartitions = npartitions,
		elevation = sources(elevation),
		coeff_bh = sources(coeff_bh),
		coeff_dh = sources(coeff_dh),
		slope = sources(slope),
		aspect = sources(aspect),
		horizon_basename = hhSrc,
		horizon_step = horizon_step,
		solar_constant = solar_constant
	)

	if (lowMemory) args$flags <- c(args$flags, "m")
	if (!is.null(declination)) args <- c(args, declination = declination)

	# albedo
	if (inherits(albedo, "GRaster")) { # raster
		compareGeom(elevation, albedo)
		args <- c(args, albedo = sources(albedo))
	} else {
		args <- c(args, albedo_value = albedo)
	}
	
	# linke
	if (inherits(linke, "GRaster")) { # raster
		compareGeom(elevation, linke)
		args <- c(args, linke = sources(linke))
	} else {
		args <- c(args, linke_value = linke)
	}

	# output names
	if (beam_rad) args <- c(args, beam_rad = .makeSourceName("sun_beam_rad", "raster"))
	if (diff_rad) args <- c(args, diff_rad = .makeSourceName("sun_diff_rad", "raster"))
	if (refl_rad) args <- c(args, refl_rad = .makeSourceName("sun_refl_rad", "raster"))
	if (glob_rad) args <- c(args, glob_rad = .makeSourceName("sun_glob_rad", "raster"))
	if (insol_time) args <- c(args, insol_time = .makeSourceName("sun_insol_time", "raster"))

	do.call(rgrass::execGRASS, args=args)

	out <- elevation
	
	if (beam_rad) {
		this <- makeGRaster(args$beam_rad, "beam_rad")
		out <- c(out, this)
	}
	
	if (diff_rad) {
		this <- makeGRaster(args$diff_rad, "diff_rad")
		out <- c(out, this)
	}
	
	if (refl_rad) {
		this <- makeGRaster(args$refl_rad, "refl_rad")
		out <- c(out, this)
	}
	
	if (glob_rad) {
		this <- makeGRaster(args$glob_rad, "glob_rad")
		out <- c(out, this)
	}
	
	if (insol_time) {
		this <- makeGRaster(args$insol_time, "insol_time")
		out <- c(out, this)
	}
	
	n <- nlyr(out)
	out <- out[[2L:n]]
	out

} # EOF

#' @title Classes for fasterRaster locations, rasters, and vectors
#'
#' @aliases GVector
#' @rdname GLocation
#' @exportClass GVector
GVector <- methods::setClass(
	"GVector",
	contains = "GSpatial",
	slots = list(
		projection = "character",
		nGeometries = "integer",
		nSubgeometries = "integer",
		geometry = "character",
		table = "data.table"
	),
	prototype = prototype(
		projection = NA_character_,
		geometry = NA_character_,
		nGeometries = NA_integer_,
		nSubgeometries = NA_integer_,
		table = data.table::data.table(NULL)
	)
)


methods::setValidity("GVector",
	function(object) {

		info <- .vectInfo(object)

		if (!info$catsValid) {
			"Vector has invalid topology. See the *Details* section in `fast()` on how to correct topology."
		} else if (!all(object@geometry %in% c(NA_character_, "points", "lines", "polygons"))) {
			paste0("@geometry can only be NA, `points`, `lines`, or `polygons`.")
		# } else if (length(unique(.vCats(object)) != object@nGeometries)) {
			# "The number of @nGeometries is not the same as the number of unique `cat` values in the vector attribute table in GRASS."
		# } else if (object@nGeometries > object@nSubgeometries) {
			# "The number of sub-geometries in @nSubgeometries must be <= the number of geometries in @nGeometries."
		} else if (object@nGeometries == 0L) {
			"GVector has no geometries."
		} else if (nrow(object@table) > 0L && nrow(object@table) != object@nGeometries) {
			"The data.table in @table must be a NULL table (data.table(NULL)), or\n  it must have the same number of rows as @nGeometries. See the *Details* section in `fast()` on how to correct topology."
		} else {
			TRUE
		}
	} # EOF
)

#' Create a GVector from a GRASS vector
#'
#' @description Create a `GVector` from a vector existing in the current **GRASS** session.
#'
#' @param src Character: The name of the vector in **GRASS**.
#'
#' @param table A `data.table`, `data.frame`, `GVector` with a table, or character. This can be `data.table(NULL)` or `data.frame(NULL)` if there is no table associated with the vector. If a character, this is interpreted as the name of the table in **GRASS**.
#'
#' @param build Logical: If `TRUE` (default), build topology using **GRASS** tool `v.build`.
#'
#' @param extensive Logical: If `TRUE`, do extensive topological checks using `v.build`. The default is `FALSE`.
#'
#' @param cats `NULL` (default) or an integer vector: Values of the "cats" (categories) of the vector in **GRASS**. This is useful *only* for speeding up the `GVector` creation process when the "cats" have already been ascertained. For most users, this will be better left as `NULL`.
#'
#' @param fail Logical: If `TRUE` (default), and the vector either has a 0 east-west or north-south extent, then exit the function with an error. If `fail` is `FALSE`, then display a warning and return `NULL`.
#'
#' @returns A `GVector` (or `NULL` if `fail` is `TRUE` and the `GVector` would be invalid).
#'
#' @seealso [makeGRaster()]
#'
#' @example man/examples/ex_fast.r
#'
#' @aliases makeGVector
#' @rdname makeGVector
#' @export makeGVector
makeGVector <- function(src, table = NULL, build = TRUE, extensive = FALSE, cats = NULL, fail = TRUE) {

	if (inherits(table, "GVector")) table <- table@table
	if (is.null(table)) table <- data.table::data.table(NULL)
	if (!inherits(table, "data.table")) table <- data.table::as.data.table(table)

	if (is.null(cats)) cats <- .vCats(src, db = FALSE)
	nGeoms <- length(unique(cats))
	nSubgeoms <- length(cats)

	info <- .vectInfo(src, cats = cats)

	# NB test for zero extent... note that lines and points can have zero extent
	if (is.na(info$geometry) || (info$geometry == "polygons" & ((info$west - info$east) == 0 | (info$north - info$south) == 0))) {
	
		msg <- "Vector has no geometries."
		if (fail) {
			stop(msg)
		} else {
			warning(msg)
			return(NULL)
		}
	
	}

	# build topology
	if (build) {

		args <- list(
			cmd = "v.build",
			map = src,
			option = "build",
			flags = c(.quiet(), "overwrite"),
			echoCmd = FALSE
		)
		if (extensive) args$flags <- c(args$flags, "e")
		do.call(rgrass::execGRASS, args = args)

	}
	
	methods::new(
		"GVector",
		location = .location(),
		mapset = "PERMANENT",
		workDir = faster("workDir"),
		crs = crs(),
  		projection = info[["projection"]][1L],
		topology = info[["topology"]][1L],
		sources = src,
		geometry = info[["geometry"]][1L],
		nGeometries = nGeoms,
		nSubgeometries = nSubgeoms,
		extent = c(info[["west"]][1L], info[["east"]][1L], info[["south"]][1L], info[["north"]][1L]),
		zextent = c(info[["zbottom"]], info[["ztop"]]),
		table = table
	)

}

if (grassStarted()) {

# Setup
library(terra)

# Climatic conditions with 4 variables
madChelsa <- fastData("madChelsa")
chelsa <- fast(madChelsa)

# Simulate new conditions by multiplying values by (1 + small number)
proj <- 1.05 * chelsa
names(proj) <- names(chelsa) # proj must have same names as ref

messes <- multivarEnvSim(ref = chelsa, proj = proj)
plot(messes)

# Where is at least one variable outside the reference range?
extrap <- messes[["MESS"]] < 0
levs <- data.frame(value = 0:1, labels = c('no extrapolation', 'extrapolation'))
levels(extrap) <- levs
plot(extrap)

# Using a data frame as "reference" conditions:
madDypsis <- fastData("madDypsis") # Dypsis occurrences
dypsis <- fast(madDypsis)

dypEnv <- extract(chelsa, dypsis)
dypMess <- multivarEnvSim(ref = dypEnv, proj = proj)
plot(dypMess)

}

#' @name madChelsa
#'
#' @title Rasters of bioclimatic variables for an eastern portion of Madagascar
#'
#' @description Rasters of bioclimatic variables for an eastern portion of Madagascar from CHELSA version 2.1 in unprojected (WGS84) coordinates. Values represent averages across 1980-2010. Only these BIOCLIM variables are included:
#' 	* `bio1`: Mean annual temperature (deg C)
#' 	* `bio7`: Temperature annual range (hottest - coldest month temperature; deg C)
#' 	* `bio12`: Total annual precipitation (mm)
#' 	* `bio15`: Precipitation seasonality (unit-less)
#'
#' @docType data
#'
#' @format An object of class `SpatRaster` in unprojected (WGS84) coordinates.
#'
#' @keywords climate Madagascar
#'
#' @references Karger, D.N., Conrad, O., Bohner, J., Kawohl, T., Kreft, H., Soria-Auza, R.W., Zimmermann, N.E., Linder, H.P., and Kessler, M.  2017.  Climatologies at high resolution for the earth's land surface areas. *Scientific Data* 4:170122. \doi{10.1038/sdata.2017.122}
#'
#' @source \doi{https://doi.org/10.1038/sdata.2017.122}
#'
#' @example man/examples/ex_fastData.r
#'
NULL

#' Set or get options shared across fasterRaster functions
#'
#' @description `faster()` either sets or gets options used across **fasterRaster** functions. Its use can vary:
#' * Get current values of a particular option: Use `faster("option_name")`. Values will remain unchanged.
#' * Get current values of all options: Use `faster()` (no arguments). Values will remain unchanged.
#' * Get default values of a particular option: Use `faster("option_name", default = TRUE)`. Values will remain unchanged.
#' * Get default values of all options: Use `faster(default = TRUE)`. Values will remain unchanged.
#' * Set values of particular options: Use the form `faster(option 1 = value1, option2 = value2)`.
#' * Set all options to their defaults: Use `faster(restore = TRUE)`.
#'
#' You cannot simultaneously get and set options.
#'
#' To run most **fasterRaster** functions, you must set the `grassDir` option.
#'
#' @param ... Either:
#' * A character vector: Name(s) of option(s) to get values of;
#' * An option and the value of the option using an `option = value` pattern; or
#' * A names `list` with names that match the options you wish to change, and with values to assign to each option.
#'
#' Options include:
#'
#' * `grassDir` (character): The folder in which **GRASS** is installed on your computer. You must set this option to run most **fasterRaster** functions. Depending on your operating system, your install directory will look something like this:
#'     * Windows: `"C:/Program Files/GRASS GIS 8.4"`
#'     * Mac OS: `"/Applications/GRASS-8.4.app/Contents/Resources"`
#'     * Linux: `"/usr/local/grass"`
#'
#' * `addonsDir` (character): Folder in which **GRASS** addons are stored. If `NA` and `grassDir` is not `NA`, this will be assumed to be `file.path(grassDir, "addons")`. The default values is `NA`.
#'
#' * `cores` (integer/numeric integer): Number of processor cores to use on a task. The default is 2. Some **GRASS** modules are parallelized.
#'
#' * `memory` (integer/numeric): The amount of memory to allocate to a task, in GB, for **GRASS**. The default is 2048 MB (i.e., 2 GB). Some **GRASS** modules can take advantage of more memory.
#'
#' * `useDataTable` (logical): If `FALSE` (default), functions that return tabular output produce `data.frame`s. If `TRUE`, output will be `data.table`s from the **data.table** package. This can be much faster, but it might require you to know how to use `data.table`s if you want to manipulate them in **R**. You can always convert them to `data.frame`s using [base::as.data.frame()].
#' 
#' * `verbose` (logical): If `TRUE`, show progress during function operations and other messages. Default is `FALSE`. This overrides the value of any `verbose` argument in a function.
#'
#' * `debug` (logical): If `TRUE`, show **GRASS** messages and otherwise hidden slots in classes. This is mainly used for debugging, so most users will want to keep this at its default, `FALSE`.
#'
#' * `workDir` (character): The folder in which **GRASS** rasters, vectors, and other objects are created and manipulated. By default, this is given by [tempdir()]. Note that on some systems, changing the default folder to somewhere else can cause problems with **fasterRaster** being able to find rasters in **GRASS** that have been created.
#'
#' @param restore Logical: If `TRUE`, the all options will be reset to their default values. The default is `FALSE`.
#'
#' @param default Logical: Return the default value(s) of the option(s). The default value of `default` is `FALSE`.
#'
#' @return If options are changed, then a named list of option values *before* they were changed is returned invisibly.
#'
#' If option values are requested, a named list with option values is returned (not invisibly).
#'
#' @example man/examples/ex_options.r
#'
#' @aliases faster
#' @rdname faster
#' @export
faster <- function(
	...,
	default = FALSE,
	restore = FALSE
) {

	opts <- list(...)
	if (length(opts) == 1L && is.list(opts[[1L]])) opts <- opts[[1L]]
	# if (length(opts) == 1L && inherits(opts[[1L]], "list")) opts <- opts[[1L]]
	# # if (length(opts) == 1L && is.null(names(opts))) names(opts) <- "grassDir"
	# # if (length(opts) > 1L && any(names(opts) == "")) names(opts)[names(opts) == ""] <- "grassDir"
	nOpts <- length(opts)
	
	if (default & restore) {
		stop("Cannot request default values and restore to default values at the same time.")
	} else if (nOpts > 0L & restore) {
		stop("Cannot simultaneously set options and restore options to default values.")
	}

	# getting options if ... has no names
	gettingOpts <- is.null(names(opts)) & !restore

	# retrieve options
	if (gettingOpts) {
		out <- .getFastOptions(..., default = default)
		out
	} else {
		# out <- .setFastOptions(..., restore = restore)
		out <- .setFastOptions(opts, restore = restore)
		invisible(out)
	}

}

#' Set arguments shared across functions
#' @noRd
.setFastOptions <- function(
	opts,
	restore = FALSE
) {

	# opts <- list(...)

	namesOfOpts <- .namesOfOptions()
	if (!is.null(opts) && any(!(names(opts) %in% namesOfOpts))) stop("Invalid option(s): ", paste(names(opts[!(opts %in% namesOfOpts)]), collapse=", "))
	
	# retrieve in case we want to reset them
	out <- faster()

	### check for validity
	error <- paste0("Option `grassDir` must be NULL (which is likely to fail)\n  or a single character string. The default is `", .grassDirDefault(), "`.")
	if (any(names(opts) %in% "grassDir")) {
		if (!is.na(opts$grassDir)) {
   			if (!is.character(opts$grassDir) || length(opts$grassDir) != 1L) stop(error)
		}
		if (!file.exists(opts$grassDir)) {
			opts$grassDir <- NA_character_
			warning("`grassDir` is invalid. This directory does not exist. Value has been set to NA.")
		}
	}

	error <- paste0("Option `addonsDir` must be NULL or a single character string. The default is `", .addonsDirDefault(), "`.")
	if (any(names(opts) %in% "addonsDir")) {
		if (!is.na(opts$addonsDir)) {
   			if (!is.character(opts$addonsDir) || length(opts$addonsDir) != 1L) stop(error)
		}

	} else if ((any(names(opts) %in% "grassDir") && !is.na(opts$grassDir)) & (is.null(.fasterRaster$options$addonsDir) || is.na(.fasterRaster$options$addonsDir))) {
	
		opts$addonsDir <- file.path(opts$grassDir, "addons")
	
	}

	error <- paste0("Option `workDir` must be a single character string. The default is\n  ", .workDirDefault(), ".")
	if (any(names(opts) %in% "workDir")) {

  		if (!is.character(opts$workDir) || length(opts$workDir) != 1L) stop(error)

		workDir <- omnibus::forwardSlash(opts$workDir)

	}

	# if (any(names(opts) %in% "nAtATime")) {
	# 	if (!is.numeric(opts$nAtATime) | (opts$nAtATime <= 0 & !omnibus::is.wholeNumber(opts$nAtATime))) stop("Option `nAtATime` must be an integer >= 1. The default is ", .nAtATimeDefault(), ".")
	# }

	if (any(names(opts) %in% "cores")) {
		if (!inherits(opts$cores, c("integer", "numeric")) || opts$cores <= 0 || !omnibus::is.wholeNumber(opts$cores)) stop("Option `cores` must be an integer >= 1. The default is ", .coresDefault(), ".")
	}

	# if (any(names(opts) %in% "clean")) {
  	# 	if (is.na(opts$clean) || !is.logical(opts$clean)) stop("Option `clean` must be a logical. The default is ", .cleanDefault(), ".")
	# }

	if (any(names(opts) %in% "debug")) {
  		if (is.na(opts$debug) || !is.logical(opts$debug)) stop("Option `debug` must be a logical. The default is ", .debugDefault(), ".")
	}

	if (any(names(opts) %in% "verbose")) {
  		if (is.na(opts$verbose) || !is.logical(opts$verbose)) stop("Option `verbose` must be a logical. The default is ", .verboseDefault(), ".")
	}

	if (any(names(opts) %in% "memory")) {
		if (!is.numeric(opts$memory) || opts$memory <= 0) stop("Option `memory` must be a positive number. The default is ", .memoryDefault(), " (GB).")
	}

	if (any(names(opts) %in% "useDataTable")) {

		if (is.na(opts$useDataTable) || !is.logical(opts$useDataTable)) stop("Option `useDataTable` must be a logical. The default is ", .useDataTableDefault(), ".")
	
	}

	### set the options
	if (length(opts) == 0L) {
		opts <- as.list(namesOfOpts)
		names(opts) <- namesOfOpts
	}

	for (opt in names(opts)) {

		# default
		if (restore) {
			val <- paste0(".", opt, "Default()")
			val <- eval(str2expression(val))
		} else {
			val <- opts[[opt]]
		}
		
		.fasterRaster$options[[opt]] <- val

	}

	if (any(names(opts) %in% "debug")) {
		info <- rgrass::set.echoCmdOption(.fasterRaster$options$debug)
		info <- rgrass::set.echoCmdOption(value = .fasterRaster$options$debug)
	}

	invisible(out)

}

#' Report arguments shared across functions
#' @noRd
.getFastOptions <- function(..., default = FALSE) {

	namesOfOpts <- .namesOfOptions()
	
	opts <- unlist(list(...))
	
	if (!is.null(opts) && any(!(opts %in% namesOfOpts))) stop("Invalid option(s): ", paste(opts[!(opts %in% namesOfOpts)], collapse=", "))

	### return default values
	if (default) {

		if (length(opts) == 0L) opts <- namesOfOpts

		out <- list()
		for (opt in opts) {

			ex <- paste0(".", opt, "Default()")
			out[[length(out) + 1L]] <- eval(str2expression(ex))

		}
		names(out) <- opts

	### return options as they currently are
	} else {

		out <- list()
		# we have no options :(
		if (length(opts) == 0L) {
			# out <- .fasterRaster$options
			out <- get("options", .fasterRaster)
		# we have options, people!
		} else {
			out <- list()
			for (opt in opts) {
				
				# out[[length(out) + 1L]] <- .fasterRaster$options[[opt]]
				out[[length(out) + 1L]] <- get(opt, .fasterRaster$options)
			
			}
			names(out) <- opts
		}

	}

	if (length(out) == 1L) out <- unlist(out)
	out

}

#' Remove or retain "noise" in a raster using PCA
#'
#' @description `denoise()` applies a principal component analysis (PCA) to layers of a `GRaster`, then uses the PCA to predict values back to a raster. This retains only coarse-scale trends, thereby removing "noise" (locally extreme values that fall far from a PC axis).
#' 
#' `noise()` does the opposite by first constructing the PCA, predicting values back to the raster, then subtracting these values from the original, removing coarse-scale trends and thereby leaving "noise".
#' 
#' @param x A `GRaster` with two or more layers.
#' 
#' @param scale Logical: If `TRUE`, input layers will be rescaled by dividing each layer by its overall population standard deviation. Note that rasters will always be centered (have their mean subtracted from values). Centering and scaling is recommended when rasters values are in different units. The default is `FALSE` (do not scale).
#' 
#' @param percent Numeric integer or integer in the range 50 to 99 (default is 80): Minimum total variation explained in the retained PC axes. Higher values will cause `denoise()` to remove less noise, and `noise()` to return less noise. If this value to too low to retain even one axis, the function will fail with a message to that effect.
#' 
#' @returns A multi-layer `GRaster` with one layer per input.
#' 
#' @seealso [princomp()], [stats::prcomp()], **GRASS** manual page for tool `i.pca` (see `grassHelp("i.pca")`)
#' 
#' @example man/examples/ex_denoise_noise.r
#' 
#' @aliases denoise
#' @rdname denoise
#' @exportMethod denoise
methods::setMethod(
	f = "denoise",
	signature = c(x = "GRaster"),
	function(x, scale = FALSE, percent = 80) {

	if (nlyr(x) < 2L) stop("Input raster must have 2 or more layers.")

	# if (percent %% 1 != 0) {
	if (!omnibus::is.wholeNumber(percent)) {
		warning("Argument `percent` must be an integer. Value will be rounded.")
		percent <- round(percent)
	}

	if (percent < 50 | percent > 99) stop("Argument `percent` must be an integer in the range [50, 99].")

	.locationRestore(x)
	.region(x)

	input <- paste(sources(x), collapse = ",")
	src <- .makeSourceName("i_pca", "raster")

	args <- list(
		cmd = "i.pca",
		input = input,
		output = src,
		rescale = c(0, 0),
		percent = percent,
		flags = c(.quiet(), "overwrite", "f"),
		intern = TRUE
	)

	if (scale) args$flags <- c(args$flags, "n")

	# execute PCA and get rasters
	info <- do.call(rgrass::execGRASS, args = args)
	layers <- 1L:nlyr(x)
	srcs <- paste0(src, ".", layers)
	makeGRaster(srcs, names(x))

	} # EOF
)

#' @aliases noise
#' @rdname denoise
#' @exportMethod noise
methods::setMethod(
	f = "noise",
	signature = c(x = "GRaster"),
	function(x, scale = FALSE, percent = 80) {

	denoised <- denoise(x, scale = scale, percent = 80)
	out <- x - denoised
	names(out) <- names(x)
	out

	} # EOF
)

#' @title Classes for "fasterRaster" locations, rasters, and vectors
#'
#' @aliases GRaster
#' @rdname GLocation
#' @exportClass GRaster
GRaster <- methods::setClass(
	"GRaster",
	contains = "GRegion",
	slots = list(
		projection = "character",
		datatypeGRASS = "character",
		nLayers = "integer",
		names = "character",
		minVal = "numeric",
		maxVal = "numeric",
		activeCat = "integer",
		levels = "list"
	),
	prototype = prototype(
		projection = NA_character_,
		datatypeGRASS = NA_character_,
		nLayers = NA_integer_,
		names = NA_character_,
		minVal = NA_real_,
		maxVal = NA_real_,
		activeCat = NA_integer_,
		levels = list(data.table::data.table(NULL))
	)
)

#' Evaluate the @levels column slot of a `GRaster`
#'
#' @description Test whether a list of "levels" tables is valid. First column must be integer. Must have >= 2 columns.
#' 
#' @param object A `GRaster`.
#' @returns `TRUE` if invalid, `FALSE` otherwise.
#'
#' @noRd
.validLevelTable <- function(object) {

	levels <- object@levels
	nlevs <- .nlevels(levels)

	for (i in seq_along(levels)) {
	
		nr <- nlevs[i]
	
		if (nr != 0) {
	
			if (object@datatypeGRASS[i] != "CELL") {
				return(TRUE)
			# } else if (nr == 1L) {
				# return(TRUE)
			} else if (!inherits(levels[[i]][[1L]], "integer")) {
				return(TRUE)
			}
		
		}
	
	}
	FALSE

}

#' Evaluate the @activeCat slot of a GRaster
#'
#' @description Test whether `@activeCat` values are valid.
#' 
#' @param object A `GRaster`.
#' @returns `TRUE` if invalid, `FALSE` otherwise.
#'
#' @noRd
.validActiveCat <- function(object) {

	bad <- FALSE
	numCols <- sapply(object@levels, ncol)
	
	isFact <- is.factor(object)

	if (any(isFact & is.na(object@activeCat))) {
		bad <- TRUE
	} else if (any(isFact & (object@activeCat < 1L | object@activeCat > numCols))) {
		bad <- TRUE
	}
	bad

}

#' Evaluate whether a raster's datatype can accommodate levels
#'
#' @description A `GRaster` must have the `CELL` datatype to have levels.
#' 
#' @param object A `GRaster`
#' @returns `TRUE` if invalid, `FALSE` otherwise.
#'
#' @noRd
.canMakeRasterCategorical <- function(object) {

	dt <- object@datatypeGRASS
	numLevels <- nlevels(object)
	if (any(numLevels > 0L)) {
		out <- any(dt[numLevels > 0L] != "CELL")
	} else {
		out <- FALSE
	}
	out

}

#' Test if GRaster is valid
#'
#' @noRd
methods::setValidity("GRaster",
	function(object) {
		if (!all(object@datatypeGRASS %in% c("CELL", "FCELL", "DCELL"))) {
			paste0("@datatypeGRASS can only be NA, ``CELL`, `FCELL`, or `DCELL`.")
		} else if (!is.na(object@dimensions[3L]) && object@dimensions[3L] <= 0L) {
			"Third value in @dimensions must be NA or a positive integer."
		} else if (!is.na(object@resolution[3L]) && object@resolution[3L] <= 0) {
			"Third value in @resolution must be NA or a positive real value."
		} else if (object@nLayers < 1) {
			"@nLayers must be a positive integer."
		} else if (object@nLayers != length(object@sources)) {
			"@nLayers is different from the number of @sources."
		} else if (object@nLayers != length(object@names)) {
			"@names must be @nLayers in length."
		} else if (object@nLayers != length(object@minVal)) {
			"@minVal must be @nLayers in length."
		} else if (object@nLayers != length(object@maxVal)) {
			"@maxVal must be @nLayers in length."
		} else if (length(object@activeCat) != object@nLayers) {
			"The number of @activeCat values must be the same as the number of @nLayers."
		} else if (length(object@levels) != object@nLayers) {
			"The number of tables in @levels must be the same as the number of @nLayers."
		} else if (.canMakeRasterCategorical(object)) {
			"Only rasters of datatype `CELL` can be categorical."
		} else if (.validLevelTable(object)) {
			"Each table must be a NULL `data.table`, or if not, the first column\n  must be an integer, and there must be >1 columns."
		} else if (.validActiveCat(object)) {
			"@activeCat must be `NA_integer_`, or an integer between 1 and the number of columns in each `data.table` in @levels, minus 1."
		} else {
			TRUE
		}
	} # EOF
)

#' Create a GRaster from a GRASS raster
#'
#' @description Create a `GRaster` from a raster existing in the current **GRASS** session.
#'
#' @param src Character: The name of the raster in **GRASS**.
#' @param names Character: Name of the raster (cf. [names()]).
#' @param levels `NULL` (default), a `data.frame`, `data.table`, an empty string (`""`), or a list of `data.frame`s, `data.table`s, and/or empty strings: These become the raster's [levels()]. If `""`, then no levels are defined.
#' @param ac Vector of numeric/integer values >=1, or `NULL` (default): Active category column (offset by 1, so 1 really means the second column, 2 means the third, etc.). A value of `NULL` uses an automated procedure to figure it out.
#' @param fail Logical: If `TRUE` (default), and the raster either has a 0 east-west or north-south extent, then exit the function with an error. If `fail` is `FALSE`, then display a warning and return `NULL`.
#'
#' @returns A `GRaster`.
#'
#' @seealso [makeGVector()]
#'
#' @example man/examples/ex_fast.r
#'
#' @aliases makeGRaster
#' @rdname makeGRaster
#' @export makeGRaster
makeGRaster <- function(src, names = "raster", levels = "", ac = NULL, fail = TRUE) {

	if (inherits(src, "rastInfo")) {
		info <- src
		src <- info$sources
	} else {
		info <- .rastInfo(src)
	}

	if (!all(.exists(src))) stop("No raster was created in GRASS. The error is likely in the calling function.")

	# test for zero extent
	if (any(abs(info$west - info$east) < omnibus::eps()) | any(abs(info$north - info$south) < omnibus::eps())) {
		msg <- "Raster has 0 east-west extent and/or north-south extent."
		if (fail) {
			stop(msg)
		} else {
			warning(msg)
			return(NULL)
		}
	}

	# levels: convert empty strings to NULL data.tables and data.frames to data.tables
	if (is.null(levels) || all(levels == "")) levels <- data.table::data.table(NULL)
	if (!inherits(levels, "list")) levels <- list(levels)
	levels <- lapply(levels, data.table::as.data.table)
	if (length(levels) == 1L & length(src) > 1L) for (i in 2L:length(src)) levels[[i]] <- levels[[1L]]

	# active category and levels
	if (is.null(ac)) {
		
		ac <- rep(NA_integer_, length(levels))
		for (i in seq_along(levels)) {
			if (is.null(levels[[i]]) || (is.character(levels[[i]]) && levels[[i]] == "")) {
				ac[i] <- NA_integer_
				levels[[i]] <- data.table::data.table(NULL)
			} else if (!inherits(levels[[i]], "data.table")) {
				ac[i] <- 2L
				levels[[i]] <- data.table::data.table(levels[[i]])
			} else if (inherits(levels[[i]], "data.table")) {
				ac[i] <- 2L
			}
		}
	
	} else {
		ac <- as.integer(ac)
		ac <- ac + 1L
	}

	# nLayers <- length(info$sources)
	nLayers <- length(src)
	if (length(names) < nLayers) names <- rep(names, length.out = nLayers)

	out <- methods::new(
		"GRaster",
		location = .location(),
		mapset = "PERMANENT",
		workDir = faster("workDir"),
		crs = crs(),
		projection = info[["projection"]][1L],
		topology = info[["topology"]][1L],
		extent = c(info[["west"]][1L], info[["east"]][1L], info[["south"]][1L], info[["north"]][1L]),
		zextent = c(info[["zbottom"]][1L], info[["ztop"]][1L]),
		nLayers = nLayers,
		dimensions = c(info[["rows"]][1L], info[["cols"]][1L], info[["depths"]][1L]),
		resolution = c(info[["ewres"]][1L], info[["nsres"]][1L], info[["tbres"]][1L]),
		sources = src,
		names = names,
		datatypeGRASS = info[["grassDataType"]],
		minVal = info[["minVal"]],
		maxVal = info[["maxVal"]],
		activeCat = ac,
		levels = levels
	)
	out <- .makeUniqueNames(out)
	# out <- droplevels(out, layer = 1L:nLayers)
	out

}

#' Remove rows from the "levels" table of a categorical raster
#'
#' @description `droplevels()` removes levels (category values) from the "levels" table of a categorical `GRaster`.
#'
#' @param x A `GRaster`.
#'
#' @param level `NULL`, character, numeric, integer, or logical: Level(s) to drop. If `NULL` (default), then all levels without values in the raster are dropped (this may remove the "levels" table entirely if all levels are dropped, converting the raster to an `integer`-type raster).  If a character, this is the category label(s) to drop. If numeric or integer, this is the category value(s) to drop. If logical, values that are `TRUE` are dropped.
#'
#' @param layer Numeric integers, logical vector, or character: Layer(s) to which to add or from which to drop levels.
#'
#' @returns A `GRaster`. The "levels" table of the raster is modified.
#'
#' @seealso [missingCats()], [missing.cases()], [terra::droplevels()], `vignette("GRasters", package = "fasterRaster")`
#'
#' @example man/examples/ex_GRaster_categorical.r
#'
#' @aliases droplevels
#' @rdname droplevels
#' @exportMethod droplevels
methods::setMethod(
	f = "droplevels",
	signature = c(x = "GRaster"),
	function(x, level = NULL, layer = 1) {
	
	layer <- .layerIndex(layer, x, recycle = TRUE)
	
	levs <- cats(x)
	isFact <- is.factor(x)
	for (i in layer) {

		if (isFact[i]) {

			# remove all non-extant levels
			if (is.null(level)) {

			    freqs <- freq(x[[i]])
				
				if (!faster("useDataTable")) {
					freqs <- data.table::as.data.table(freqs)
					levs[[i]] <- data.table::as.data.table(levs[[i]])
				}

				data.table::setkeyv(freqs, names(freqs)[1L])
				data.table::setkeyv(levs[[i]], names(levs[[i]])[1L])

				cols <- names(levs[[i]])

    			# x@levels[[i]] <- levs[[i]][unique(levs[[i]][freqs$value != 0L, which = TRUE]), ]
    			x@levels[[i]] <- levs[[i]][freqs$count != 0L]

			} else if (is.character(level)) {
			
				ac <- activeCat(x, layer = i, names = TRUE)

				x@levels[[i]] <- levs[[i]][levs[[i]][ , !(get(ac) %in% levels)]]
			
			} else if (is.logical(level)) {
			
				if (length(level) < nrow(levs[[i]])) {

					level <- rep(level, length.out = nrow(levs[[i]]))
				}

    			x@levels[[i]] <- levs[[i]][!level]
			
			} else if (is.numeric(levels) | is.integer(levels)) {
			
				drops <- which(x@levels[[i]][[1L]] %in% levels)
				x@levels[[i]] <- dropRows(x@levels[[i]], drops = drops)
			
			}

			if (nrow(x@levels[[i]]) == 0L) x@activeCat[i] <- NA_integer_

		} # if this layer has levels

	} # next raster
	methods::validObject(x)
	x
	
	} # EOF
)

droplevels <- function(x) UseMethod("droplevels", x)

#' Get a GLocation's working directory
#'
#' @description This function returns the working directory of a `GLocation` object.
#'
#' @param x A `GLocation` object or missing. If an object, returns the working folder in which the object is saved by **GRASS**. If missing, then just returns the working folder (same as `faster("workDir")`).
#'
#' @returns Character.
#'
#' @aliases .workDir
#' @rdname workDir
#' @exportMethod .workDir
#' @keywords internal
methods::setMethod(
	f = ".workDir",
	signature = c(x = "GLocation"),
	function(x)	x@workDir
)

#' @aliases .workDir
#' @rdname workDir
#' @exportMethod .workDir
#' @keywords internal
methods::setMethod(
	f = ".workDir",
	signature = c(x = "missing"),
	function(x)	faster("workDir")
)

#' Frequencies of cell values in a raster
#'
#' @description `freq()` tabulates the frequency of cell values in a raster. For rasters where [datatype()] is `integer` or `factor`, the frequency of each value or level is reported. For other rasters, the range of values is divided into bins, and the number of cells with values in each bin is reported.
#'
#' @param x A `GRaster`.
#'
#' @param digits Numeric integer: Number of digits by which to round raster values. Ignored for integer and categorical rasters.
#'
#' @param bins Positive numeric integer: Number of bins in which to divide values of `numeric` rasters. The default is 100. For `integer` and categorical rasters, each value is tallied (i.e., this is ignored).
#'
#' @param value Numeric or `NULL` (default): If numeric, only cells with this value will be counted. If `NULL`, all values will be counted.
#'
#' @returns A `data.frame` or a named `list` of `data.frame`s, one per layer in `x`.
#'
#' @seealso [terra::freq()], tool `r.stats` in **GRASS**
#'
#' @example man/examples/ex_freq.r
#'
#' @aliases freq
#' @rdname freq
#' @exportMethod freq
methods::setMethod(
	f = "freq",
	signature = c(x = "GRaster"),
	definition = function(
		x,
		digits = 3,
		bins = 100,
		value = NULL
	) {

	dtype <- datatype(x, type = "GRASS")
	.freq(x = x, dtype = dtype, digits = digits, bins = bins, value = value)
	
	} # EOF
)

#' @param x `GRaster` or [sources()] name(s)
#' @param dtype [datatype()] using `GRASS`-style output
#' @param levels The [levels()] table from a `GRaster`. Used if `x` is the [sources()] name of a `GRaster` and the raster is categorical. Note this is *not* a list of `levels` tables, but rather the actual table.
#' @param digits Integer
#' @param bins Integer > 0
#' @param value NULL or numeric
#'
#' @noRd
.freq <- function(x, dtype, levels = NULL, digits = 3, bins = 100, value = NULL) {

	if (inherits(x, "GRaster")) {
		.locationRestore(x)
		.region(x)
		src <- sources(x)
	} else {
		src <- x
	}

	nLayers <- length(src)

	if (any(dtype != "CELL") & bins <= 0) stop("Argument `bins` must be a positive integer.")
	
	# get values for each raster
	out <- list()
	for (i in seq_len(nLayers)) {

		thisSrc <- src[i]

		if (!is.null(value)) {

			thisSrc <- .makeSourceName("r_mapcalc", "raster")
			ex <- paste0(thisSrc, " = if(", src[i], " == ", value, ", ", value, ", null())")
			rgrass::execGRASS(cmd = "r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))
		
		}

		args <- list(
			cmd = "r.stats",
			input = thisSrc,
   			separator = "pipe",
			flags = c("c", "n", .quiet()),
			intern = TRUE
		)
		
		if (dtype[i] != "CELL") args$nsteps = bins
		
		data <- do.call(rgrass::execGRASS, args = args)
		
		bads <- which(grepl(data, pattern = "\b"))
		if (length(bads) > 0L) data <- data[-bads]
	
		data <- strsplit(data, split = "\\|")

		# categorical/integer data
		if (dtype[i] == "CELL") {
			
			freqs <- do.call(rbind, data)
			freqs <- data.table::as.data.table(freqs)
			names(freqs) <- c("value", "count")

			freqs[ , c("value", "count") := lapply(.SD, as.integer), .SDcols = c("value", "count")]

			data.table::setkeyv(freqs, "value")

			# add level labels
			if ((inherits(x, "GRaster") & is.factor(x)[i]) | !is.null(levels)) {

				if (inherits(x, "GRaster")) {
          			levs <- levels(x)[[i]]
				} else {
					levs <- levels
				}
				xValCol <- names(freqs)[1L]
				yValCol <- names(levs)[1L]
				freqs <- merge(freqs, levs, by.x = xValCol, by.y = yValCol, all.y = TRUE)
				
				freqs <- replaceNAs(freqs, cols = "count", replace = 0)

			}
		
		# continuous data
		} else {
	
			n <- length(data)
			freqs <- data.table::data.table(from = rep(NA_real_, n), to = rep(NA_real_, n), count = rep(NA_integer_, n))

			for (j in seq_along(data)) {
			
				count <- data[[j]][2L]
				count <- as.numeric(count)

				fromTo <- strsplit(data[[j]][1L], split="-")[[1L]]
				fromTo <- as.numeric(fromTo)

				if (length(fromTo) == 2L) {
					from <- fromTo[1L]
					to <- fromTo[2L]
				} else if (length(fromTo) == 3L) {
					if (is.na(fromTo[1L])) {
						from <- -1 * fromTo[2L]
						to <- fromTo[3L]
					} else {
						from <- fromTo[1L]
						to <- -1 * fromTo[3L]
					}
				} else {
					from <- -1 * fromTo[2L]
					to <- -1 * fromTo[4L]
				}
					
				freqs$from[j] <- from
				freqs$to[j] <- to
				freqs$count[j] <- count

			} # next datum
			
			freqs$from <- round(freqs$from, digits)
			freqs$to <- round(freqs$to, digits)
		
		}

		# if (!faster("useDataTable")) freqs <- as.data.frame(freqs)
		out[[i]] <- freqs
		names(out)[i] <- names(x)[i]

	} # next layer
	
	# # add factor labels
	# facts <- is.factor(x)
	# if (any(facts)) {
	
	# 	for (i in which(facts)) {
		
	# 		xx <- x[[i]]
	# 		ac <- activeCat(xx)
	# 		catName <- names(levels(xx)[[1L]])[ac + 1L]
			
	# 		labels <- levels(xx)[[1L]][[ac + 1L]]
	# 		vals <- levels(xx)[[1L]][[1L]]
			
	# 		out[[i]] <- cbind(
	# 			out[[i]][ , 1L],
	# 			data.table::data.table(
	# 				DUMMY = labels[match(out[[i]][["value"]], vals)]
	# 			),
	# 			out[[i]][ , 2L]
	# 		)
			
	# 		names(out[[i]])[2L] <- catName
		
	# 	}
	
	# }
	
	if (length(out) == 1L) out <- out[[1L]]
	out

}

#' Create stream network
#'
#' @description This function estimates the course of streams and rivers from an elevation raster. It is based on the **GRASS** tool `r.stream.extract`, where more details can be found (see `grassHelp("r.stream.extract")`)
#'
#' @param x A `GRaster` representing elevation.
#'
#' @param accumulation Either `NULL` (default) or a raster representing flow accumulation. If not supplied, an accumulation will created internally. You can generate an accumulation raster using [flow()].
#'
#' @param depression Either `NULL` (default) or a `GRaster` representing depressions (areas from which streams will not flow out of).
#'
#' @param flowThreshold Numeric > 0: Minimum threshold for a stream to be generated. The default is 1, which is not necessarily a reasonable value.
#'
#' @param dirThreshold Numeric (default is `Inf`): When flow exceeds this threshold, its direction is estimated using a single-flow direction algorithm. Below this threshold, a multi-direction flow model is used. This is the `d8cut` parameter in `r.stream.extract`, and it is only used if `accumulation` is `NULL`.  The default is 1, which is not necessarily a reasonable value.
#'
#' @param montgomery Numeric: The "Montgomery" exponent for slope, multiplied by accumulation as per `accumulation * slope^montgomery`. This value is then compared to the threshold to determine if it is sufficient. The default is 0 (i.e., no slope scaling).
#'
#' @param minLength Numeric: First-order streams less than this length are removed (units in cells). Default is 0 (no removal).
#'
#' @returns A `GRaster`.
#'
#' @example man/examples/ex_streams.r
#'
#' @seealso [flow()], [flowPath()], **GRASS** manual for tool `r.stream.extract` (see `grassHelp("r.stream.extract")`)
#'
#' @aliases streams
#' @rdname streams
#' @exportMethod streams
methods::setMethod(
	f = "streams",
	signature = "GRaster",
	function(
		x,
		accumulation = NULL,
		depression = NULL,
		flowThreshold = 1,
		dirThreshold = 1,
		montgomery = 0,
		minLength = 1
	) {

	if (nlyr(x) > 1L) stop("Only a single-layer GRaster can be supplied to `x`.")
	if (!is.null(accumulation) && nlyr(accumulation) > 1L) stop("Only a single-layer GRaster can be supplied to `accumulation`.")

	.locationRestore(x)
	.region(x)
	if (!is.null(accumulation)) compareGeom(x, accumulation)

	src <- .makeSourceName("streams_r_stream_extract", "raster")
	args <- list(
		cmd = "r.stream.extract",
		stream_raster = src,
		elevation = sources(x),
		threshold = flowThreshold,
		d8cut = dirThreshold,
		mexp = montgomery,
		stream_length = minLength,
		memory = faster("memory"),
		flags = c(.quiet(), "overwrite")
	)
	
	do.call(rgrass::execGRASS, args = args)
	makeGRaster(src, names = "streams")
	
	} # EOF
)

#' Minimum and maximum values or categories of a GRaster
#'
#' @description `minmax()` reports the minimum and maximum values across all non-NA cells of a `GRaster`. When the `levels` argument is `TRUE` and the raster is categorical, the function reports the "lowest" and "highest" category values in a categorical (factor) `GRaster`.
#'
#' @param x A `GRaster`.
#'
#' @param levels Logical: If `TRUE` and the raster is a categorical raster, return the "lowest" and "highest" categories. The default is `FALSE`.
#'
#' @return `minmax()` returns a numeric matrix, and `minmax(..., levels = TRUE)` returns a `data.frame` with category names. In the latter case, non-categorical rasters will have `NA` values.
#' 
#' @seealso [terra::minmax()]
#' 
#' @example man/examples/ex_GRaster.r
#'
#' @aliases minmax
#' @rdname minmax
#' @exportMethod minmax
setMethod(
	f = "minmax",
	signature = "GRaster",
	definition = function(x, levels = FALSE) {
	
	if (!levels) {
	
		out <- matrix(c(x@minVal, x@maxVal), nrow = 2, byrow = TRUE, dimnames = list(c("min", "max"), names(x)))

	} else {

		numLevels <- nlevels(x)
		for (i in seq_len(nlyr(x))) {

			if (numLevels[i] > 0L) {
				
				ac <- x@activeCat[[i]]
    			cats <- x@levels[[i]][ , ..ac]
				cats <- cats[[1L]]
				thisMin <- cats[1L]
				thisMax <- tail(cats, 1L)

				this <- data.frame(TEMPTEMP_ = c(thisMin, thisMax), row.names = c("min", "max"))
				
				names(this) <- names(x)[i]

			} else {

				this <- data.frame(TEMPTEMP_ = c(NA_character_, NA_character_), row.names = c("min", "max"))
				
				names(this) <- names(x)[i]

			}

			if (i == 1L) {
				out <- this
			} else {
				out <- cbind(out, this)
			}

		} # next raster

	}
	out

	} # EOF
)


#' Get minimum value from GRaster metadata
#' @param x A `GRaster`.
#' @keywords internal
.minVal <- function(x) x@minVal

#' Get maximum value from GRaster metadata
#' @param x A `GRaster`.
#' @keywords internal
.maxVal <- function(x) x@maxVal

#' Report or change the extent, dimensions, and/or resolution of a region GRASS
#'
#' @description These functions either change the extent, dimensions, and/or resolution of a **GRASS** "region" or report the current region's extent, dimensions, and/or resolution (see `vignette("regions", package = "fasterRaster")`). These functions are mostly used internally and rarely of interest to most users.
#' * `.region()`: All 2D and 3D aspects of a region.
#' * `.regionDim()`: x- and y-dimensions.
#' * `.regionExt()`: x- and y-extent.
#' * `.regionRes()`: x- and y-resolution.
#'
#' @param x Any of:
#'	* Missing (default): Reports the extent, resolution, and dimensions of the current region. All other arguments will be ignored. You can also use [ext()], [dim()], and [res()] and related functions with missing arguments.
#'	* A `GSpatial`, `GRegion`, `GRaster`, `GVector` object: Sets the region"s extent, dimensions, and/or resolution to those of the object.
#'	* A `numeric` vector. This will resize the region's extent, resample the region's resolution/dimensions, or both, to ensure the desired dimensions or resolution are retained:
#'       * 2 values for `.regionDim()`: Number of rows and columns
#'       * 4 values for `.regionExt()`: Westernmost and easternmost easting (longitude), and southernmost and northernmost northing (latitude)
#'       * 2 values for `.regionRes()`: Size of cells in the x- and y-dimensions
#'
#' @param trim A `GRaster` or `NULL` (default). If a `GRaster`, then the region will be trimmed to the non-`NA` cells in this raster. `trim` can only be non-`NULL` if `x` is a `GRaster`. Ignored if `NULL`.
#' 
#' @param respect Character or `GRaster`: Indicates what aspect(s) of the current region to retain. Different functions allow for a different aspect to be retained. Partial matching is used.
#' * `.regionDim()`: `"extent"` (extent unchanged, resolution may be changed) or `"resolution"` (resolution unchanged, extent may be changed).
#' * `.regionExt()`: `"dimensions"` (dimensions unchanged, extent may be changed) or `"resolution"` (resolution unchanged, extent may be changed).
#' * `.regionRes()`: `"extent"` (extent may be changed and/or dimensions may be changed to accommodate desired cell size) or `"dimensions"` (extent may be changed, dimensions unchanged).
#' Alternatively, a `GRaster` can be supplied:
#' * `.regionDim()`: New region will have same extent and resolution.
#' * `.regionExt()`: New region will have same dimensions and resolution.
#' * `.regionRes()`: New region will have same extent and dimensions.
#' 
#' In this case, the new region"s registration will be the same as this raster, and cell resolution will be the same
#' 
#' Note: In most cases extent cannot be retained exactly if the resolution is changed. When resolution is changed, the actual extent will be the user-supplied extent expanded by zero to one rows or zero to one columns to accommodate an integer number of cells of the desired size. The western and northern limits of the extent will be retained, while the eastern and southern limits of the extent will be moved to accommodate an integer number of columns and rows.
#'
#' @returns The value returned depends on how the function is used:
#' * If used with no arguments, `.region()` returns a `GRegion` object.
#' * If used with no arguments, `.regionDim()`, `.regionExt()`, and `.regionRes()` return numeric or integer vectors.
#' * If the function is used to change reshape/resample the region, it returns a `GRegion` object reflecting the region *before* it was changed. This allows users to revert to the original region if desired.
#' 
#' @details When resizing extent, **terra** keeps the `xmin` (west) and `ymax` (north) the fixed and shifts `xmax` (east) and `ymin` (south) as needed. To retain as much fidelity between **fasterRaster** and **terra** as possible, these functions do the same to the region.
#'
#' @aliases .region
#' @rdname region
#' @keywords internal
methods::setMethod(
	f = ".region",
	signature = "missing",
	definition = function(x) {
	
	info <- rgrass::execGRASS(cmd = "g.region", flags=c("p", "3", "u"), intern = TRUE)
	
	projection <- info[grepl(info, pattern = "projection:")]
	longLatProj <- grepl(projection, pattern = "(Latitude-Longitude)")
	
	# horizontal extent
	n <- info[grepl(info, pattern="north:")]
	s <- info[grepl(info, pattern="south:")]
	e <- info[grepl(info, pattern="east:")]
	w <- info[grepl(info, pattern="west:")]
	
	n <- gsub(n, pattern="north:", replacement="")
	s <- gsub(s, pattern="south:", replacement="")
	e <- gsub(e, pattern="east:", replacement="")
	w <- gsub(w, pattern="west:", replacement="")
	
	n <- trimws(n)
	s <- trimws(s)
	e <- trimws(e)
	w <- trimws(w)
	
	if (longLatProj) {
		
		n <- strsplit(n, ":")[[1L]]
		s <- strsplit(s, ":")[[1L]]
		e <- strsplit(e, ":")[[1L]]
		w <- strsplit(w, ":")[[1L]]
	
		nIsSouth <- any(grepl(n, pattern="S"))
		sIsSouth <- any(grepl(s, pattern="S"))
		eIsWest <- any(grepl(e, pattern="W"))
		wIsWest <- any(grepl(w, pattern="W"))

		n <- gsub(n, pattern="N", replacement="")
		n <- gsub(n, pattern="S", replacement="")
	
		s <- gsub(s, pattern="N", replacement="")
		s <- gsub(s, pattern="S", replacement="")
	
		e <- gsub(e, pattern="E", replacement="")
		e <- gsub(e, pattern="W", replacement="")

		w <- gsub(w, pattern="E", replacement="")
		w <- gsub(w, pattern="W", replacement="")
	
		n <- as.numeric(n)
		s <- as.numeric(s)
		e <- as.numeric(e)
		w <- as.numeric(w)
	
		if (length(n) == 2L) n <- n[1L] + n[2L] / 60
		if (length(s) == 2L) s <- s[1L] + s[2L] / 60
		if (length(e) == 2L) e <- e[1L] + e[2L] / 60
		if (length(w) == 2L) w <- w[1L] + w[2L] / 60
		
		if (length(n) == 3L) n <- n[1L] + n[2L] / 60 + n[3L] / 3600
		if (length(s) == 3L) s <- s[1L] + s[2L] / 60 + s[3L] / 3600
		if (length(e) == 3L) e <- e[1L] + e[2L] / 60 + e[3L] / 3600
		if (length(w) == 3L) w <- w[1L] + w[2L] / 60 + w[3L] / 3600
		
		if (nIsSouth) n <- -1 * n
		if (sIsSouth) s <- -1 * s
		if (eIsWest) e <- -1 * e
		if (wIsWest) w <- -1 * w
	
	} else {
		n <- as.numeric(n)
		s <- as.numeric(s)
		e <- as.numeric(e)
		w <- as.numeric(w)
	}
		
	extent <- c(w, e, s, n)
	
	# vertical extent
	top <- info[grepl(info, pattern="top:")]
	bottom <- info[grepl(info, pattern="bottom:")]
	
	top <- gsub(top, pattern="top:", replacement="")
	bottom <- gsub(bottom, pattern="bottom:", replacement="")
	
	top <- trimws(top)
	bottom <- trimws(bottom)
	
	top <- as.numeric(top)
	bottom <- as.numeric(bottom)
	
	zextent <- c(bottom=bottom, top=top)
	
	# dimensions
	rows <- info[grepl(info, pattern="rows:")]
	cols <- info[grepl(info, pattern="cols:")]
	depths <- info[grepl(info, pattern="depths:")]

	rows <- gsub(rows, pattern="rows:", replacement="")
	cols <- gsub(cols, pattern="cols:", replacement="")
	depths <- gsub(depths, pattern="depths:", replacement="")
	
	rows <- trimws(rows)
	cols <- trimws(cols)
	depths <- trimws(depths)
	
	rows <- as.numeric(rows)
	cols <- as.numeric(cols)
	depths <- as.numeric(depths)
	
	dims <- c(rows=rows, cols=cols, depths=depths)
	dims <- as.integer(dims)
	
	# resolution
	ewres <- info[grepl(info, pattern="ewres:")]
	nsres <- info[grepl(info, pattern="nsres:")]
	tbres <- info[grepl(info, pattern="tbres:")]

	ewres <- gsub(ewres, pattern="ewres:", replacement="")
	nsres <- gsub(nsres, pattern="nsres:", replacement="")
	tbres <- gsub(tbres, pattern="tbres:", replacement="")
	
	ewres <- trimws(ewres)
	nsres <- trimws(nsres)
	tbres <- trimws(tbres)
	
	tbres <- as.numeric(tbres)
	
	if (longLatProj) {
	
		ewres <- strsplit(ewres, ":")[[1L]]
		nsres <- strsplit(nsres, ":")[[1L]]
	
		ewres <- as.numeric(ewres)
		nsres <- as.numeric(nsres)
	
		len <- length(ewres)
		if (len == 2L) {
			ewres <- ewres[1L] + ewres[2L] / 60
		} else if (len == 3L) {
			ewres <- ewres[1L] + ewres[2L] / 60 + ewres[3L] / 3600
		}
		
		len <- length(nsres)
		if (length(nsres) == 2L) {
			nsres <- nsres[1L] + nsres[2L] / 60
		} else if (len == 3L) {
			nsres <- nsres[1L] + nsres[2L] / 60 + nsres[3L] / 3600
		}
	
	} else {
		
		ewres <- as.numeric(ewres)
		nsres <- as.numeric(nsres)
		tbres <- as.numeric(tbres)
		
	}
		
	resol <- c(xres=ewres, yres=nsres, zres=tbres)
	
	GRegion(
		location = .location(),
		mapset = .mapset(),
		workDir = faster("workDir"),
		crs = crs(),
		topology = "3D",
		extent = extent,
		zextent = zextent,
		sources = NA_character_,
		dimensions = dims,
		resolution = resol
	)

	} # EOF
)

#' @rdname region
#' @aliases .region
#' @keywords internal
methods::setMethod(
	f = ".region",
	signature = "SpatRaster",
	definition = function(x) {
	
	initials <- .region()

	dims <- dim(x)
	rows <- dims[1L]
	cols <- dims[2L]

	extent <- as.vector(terra::ext(x))
	w <- as.character(extent[1L])
	e <- as.character(extent[2L])
	s <- as.character(extent[3L])
	n <- as.character(extent[4L])

	rgrass::execGRASS(
		cmd = "g.region",
		n = n,
		s = s,
		e = e,
		w = w,
		rows = rows,
		cols = cols,
		flags=c("o", .quiet())
	)

	invisible(initials)

	} # EOF
)

#' @rdname region
#' @aliases .region
#' @keywords internal
methods::setMethod(
	f = ".region",
	signature = c(x = "GRegion"),
	definition = function(x) {

	.locationRestore(x)
	initials <- .region()

	w <- W(x, TRUE)
	e <- E(x, TRUE)
	s <- S(x, TRUE)
	n <- N(x, TRUE)
	t <- top(x, TRUE)
	b <- bottom(x, TRUE)

	rows <- nrow(x)
	cols <- ncol(x)

	tbres <- zres(x)
	tbres <- as.character(tbres)

	args <- list(
		cmd = "g.region",
		n = n, s = s, e = e, w = w,
		rows = rows, cols = cols,
		flags = c("o", .quiet())
	)

	if (!is.na(t)) args <- c(args, t = t)
	if (!is.na(b)) args <- c(args, b = b)
	if (!is.na(tbres)) args <- c(args, tbres = tbres)

	do.call(rgrass::execGRASS, args=args)
	invisible(initials)
		
	} # EOF
)

#' @rdname region
#' @aliases .region
#' @keywords internal
methods::setMethod(
	f = ".region",
	signature = c(x = "GRaster"),
	definition = function(x, trim = NULL) {
	
	.locationRestore(x)
	initials <- .region()

	topo <- topology(x)

	if (any(topo %in% "2D") & any(topo %in% "3D")) stop("Cannot mix 2D- and 3D-rasters when defining region.")

	# ensure validity of trim raster
	if (!is.null(trim)) {
	
		trimToTopo <- topology(trim)
		if (any(!(topo %in% trimToTopo))) stop("Topology of `trim` does not match topology of `x`.")
		
		trim <- sources(trim)
		if (length(trim) != 1L) stop("Argument `trim` can have only one layer.")
	
	}

	src <- sources(x)[1L]

	if (all(topo == "2D") & is.null(trim)) {
		rgrass::execGRASS("g.region", raster=src, flags=c("o", .quiet()), intern=TRUE)
	} else if (all(topo == "2D") & !is.null(trim)) {
		rgrass::execGRASS("g.region", raster=src, zoom=trim, flags=c("o", .quiet()), intern=TRUE)
	} else if (all(topo == "3D") & is.null(trim)) {
		rgrass::execGRASS("g.region", raster_3d=src, flags=c("o", .quiet()), intern=TRUE)
	} else if (all(topo == "3D") & !is.null(trim)) {
		rgrass::execGRASS("g.region", raster_3d=src, zoom=trim, flags=c("o", .quiet()), intern=TRUE)
	} else {
		stop("Could not reshape region.")
	}

	invisible(initials)
		
	} # EOF
)

#' @rdname region
#' @aliases .region
#' @keywords internal
methods::setMethod(
	f = ".region",
	signature = c(x = "GVector"),
	definition = function(x) {
	
	.locationRestore(x)
	initials <- .region()

	rgrass::execGRASS(
		cmd = "g.region",
		vector = sources(x),
		flags=c("o", .quiet())
	)
	invisible(initials)
		
	}
)

#################
### regionExt ###
#################

#' @rdname region
#' @aliases .regionExt
#' @keywords internal
methods::setMethod(
	f = ".regionExt",
	signature = "missing",
	definition = function(x) .region()@extent
)

#' @rdname region
#' @aliases .regionExt
#' @keywords internal
methods::setMethod(
	f = ".regionExt",
	signature = "numeric",
	definition = function(x, respect) ..regionExt(x, respect)
)

#' @rdname region
#' @aliases .regionExt
#' @keywords internal
methods::setMethod(
	f = ".regionExt",
	signature = "GSpatial",
	definition = function(x, respect) ..regionExt(x, respect)
)

..regionExt <- function(x, respect) {

	respect <- omnibus::pmatchSafe(respect, c("resolution", "dimensions"))

	if (inherits(x, "GSpatial")) {
		.locationRestore(x)
		x <- ext(x, vector=TRUE)
	} else if (inherits(x, c("numeric", "integer"))) {
		if (length(x) != 4L) stop("Please supply a numeric vector of four values.")
		if (any(is.na(x))) stop("Extent cannot have an NA value.")
	}

	initials <- .region()
	
	w <- x[1L]
	e <- x[2L]
	s <- x[3L]
	n <- x[4L]
	
	ewres <- res(initials)[1L]
	nsres <- res(initials)[2L]

	if (respect == "resolution") {

		cols <- ceiling((w - e) / ewres)
		rows <- ceiling((n - s) / nsres)

		e <- w + cols * ewres
		s <- n - rows * nsres

	} else if (respect == "dimensions") {

		cols <- ncol(initials)
		rows <- nrow(initials)

		ewres <- (e - w) / cols
		nsres <- (n - s) / rows

	}

	w <- as.character(x[1L])
	e <- as.character(x[2L])
	s <- as.character(x[3L])
	n <- as.character(x[4L])

	ewres <- as.character(ewres)
	nsres <- as.character(nsres)

	rgrass::execGRASS(
		cmd = "g.region",
		w = w, e = e, s = s, n = n,
		ewres = ewres, nsres = nsres,
		flags = c("o", .quiet())
	)
	invisible(initials)

}

#################
### regionDim ###
#################

#' @rdname region
#' @aliases .regionDim
#' @keywords internal
methods::setMethod(f = ".regionDim",
	signature = c(x = "missing"),
	definition = function(x) .region()@dimensions
)

#' @rdname region
#' @aliases .regionDim
#' @keywords internal
methods::setMethod(
	f = ".regionDim",
	signature = "numeric",
	definition = function(x, respect) ..regionDim(x, respect)
)

#' @rdname region
#' @aliases .regionDim
#' @keywords internal
methods::setMethod(
	f = ".regionDim",
	signature = "GRegion",
	definition = function(x, respect) ..regionDim(x, respect)
)

..regionDim <- function(x, respect) {

	respect <- omnibus::pmatchSafe(respect, c("extent", "resolution"))

	if (inherits(x, "GRegion")) {
		x <- dim(x)[1L:2L]
	} else if (inherits(x, "SpatRaster")) {
		x <- dim(x)[1L:2L]
	} else if (inherits(x, c("numeric", "integer"))) {
		if (length(x) != 2L | any(x <= 0) | any(!omnibus::is.wholeNumber(x))) {
			stop("Please supply a numeric vector of two positive integer values.")
		}
	}

	initials <- .region()

	rows <- x[1L]
	cols <- x[2L]

	extent <- ext(initials, vector=TRUE)
	w <- extent[["xmin"]]
	e <- extent[["xmax"]]
	s <- extent[["ymin"]]
	n <- extent[["ymax"]]

	resol <- initials@resolution
	ewres <- resol[1L]
	nsres <- resol[2L]

	if (respect == "extent") {

		ewres <- (e - w) / cols
		nsres <- (n - s) / rows

	} else if (respect == "resolution") {

		e <- w + cols * ewres
		s <- n - rows * nsres

	}

	w <- as.character(w)
	e <- as.character(e)
	s <- as.character(s)
	n <- as.character(n)

	ewres <- as.character(ewres)
	nsres <- as.character(nsres)

	args <- list(
		cmd = "g.region",
		w = w, e = e, s = s, n = n,
		rows = rows, cols = cols,
		ewres = ewres, nsres = nsres,
		flags = c("o", .quiet())
	)

	do.call(rgrass::execGRASS, args=args)
	invisible(initials)

}

#################
### regionRes ###
#################

#' @rdname region
#' @aliases .regionRes
#' @keywords internal
methods::setMethod(f = ".regionRes",
	signature = c(x = "missing"),
	definition = function(x) .region()@resolution[1L:2L]
)

#' @rdname region
#' @aliases .regionRes
#' @keywords internal
methods::setMethod(
	f = ".regionRes",
	signature = "numeric",
	definition = function(x, respect) ..regionRes(x, respect)
)

#' @rdname region
#' @aliases .regionRes
#' @keywords internal
methods::setMethod(
	f = ".regionRes",
	signature = "GRegion",
	definition = function(x, respect) ..regionRes(x, respect)
)

# set resolution of region
..regionRes <- function(x, respect) {

	respect <- omnibus::pmatchSafe(respect, c("extent", "dimensions"))

	if (inherits(x, "GRegion")) {
		x <- res(x)[1L:2L]
	} else if (inherits(x, c("numeric", "integer"))) {
		if (length(x) == 1L) x <- rep(x, 2L)
		if (length(x) != 2L | any(x <= 0)) {
			stop("Please supply a numeric vector of one or two positive values.")
		}
	}

	initials <- .region()

	ewres <- x[1L]
	nsres <- x[2L]

	w <- W(initials)
	e <- E(initials)
	s <- S(initials)
	n <- N(initials)

	if (respect == "extent") {

		cols <- ceiling((e - w) / ewres)
		rows <- ceiling((n - s) / nsres)

		e <- w + cols * ewres
		s <- n - rows * nsres

	} else if (respect == "dimensions") {

		e <- w + ncol(initials) * ewres
		s <- n - nrow(initials) * nsres

	}

	w <- as.character(w)
	e <- as.character(e)
	s <- as.character(s)
	n <- as.character(n)

	ewres <- as.character(ewres)
	nsres <- as.character(nsres)

	rgrass::execGRASS(
		cmd = "g.region",
		w = w, e = e, s = s, n = n,
		ewres = ewres, nsres=nsres,
		flags = c("o", .quiet())
	)
	invisible(initials)

}

#' @name madCoast
#'
#' @title Shapefile of a portion of the coastline of Madagascar
#'
#' @description Borders of a selected portion of Madagascar
#'
#' @docType data
#'
#' @format ESRI Shapefile.
#'
#' @keywords Madagascar
#'
#' @source [Database of Global Administrative Areas Version 2.8 (GADM)](https://gadm.org)
#'
#' @example man/examples/ex_fastData.r
#'
NULL

#' Slope, aspect, curvature, and partial slopes
#'
#' `terrain()` calculates topographic indices, including slope, aspect, curvature, and partial slopes (slopes in the east-west or north-south directions).
#'
#' @param x A `GRaster` (typically representing elevation).
#' @param v Name of the topographic metric(s) to calculate. Valid values include one or more of:
#'
#' * `"slope"`: Slope. Units are given by argument `units`.
#' * `"aspect"`: Aspect. When argument `northIs0` is `TRUE` (default), then aspect is given in degrees from north going clockwise (0 = north, 90 = east, 180 = south, 270 = west).  Units are given by argument `units`.
#' * `"profileCurve"`: Profile curvature.
#' * `"tanCurve"`: Tangential curvature.
#' * `"dx"`: Slope in east-west direction.
#' * `"dy"`: Slope in north-south direction.
#' * `"dxx"`: Second partial derivative in east-west direction.
#' * `"dyy"`: Second partial derivative in north-south direction.
#' * `"dxy"`: Second partial derivative along east-west and north-south direction.
#' * `"*"`: All of the above.
#'
#' @param units Character: "Units" in which to calculate slope and aspect: either `"degrees"` for degrees (default), `"radians"`, or `"percent"`. Partial matching is used.
#'
#' @param undefinedAspect Numeric or `NA` (default): Value to assign to flat areas for which aspect cannot be calculated.
#'
#' @param northIs0 Logical: If `TRUE` (default), aspect will be reported in "north orientation," such that 0 is north, and degrees run clockwise (90 is east, 180 south, 270 west). If `FALSE`, then aspect will be reported in "east orientation," such that 0 is east, and degrees run counterclockwise (90 is north, 180 west, 270 south). The latter is the default in **GRASS**, but the former is the default in [terra::terrain()] function, so is used here as the default. **Note:** The [sun()] function requires aspect to be in east orientation.
#'
#' @returns A `GRaster` with one or more layers.
#' 
#' @seealso [terra::terrain()], [ruggedness()], [wetness()], [geomorphons()], tool `r.slope.aspect` in **GRASS**
#'
#' @example man/examples/ex_terrain.r
#'
#' @aliases terrain
#' @rdname terrain
#' @exportMethod terrain
methods::setMethod(
	f = "terrain",
	signature = c(x = "GRaster"),
	definition = function(
		x,
		v = "slope",
		units = "degrees",
		undefinedAspect = NA,
		northIs0 = TRUE
	) {

	metrics <- c("slope", "aspect", "profileCurve", "tanCurve", "dx", "dy", "dxx", "dyy", "dxy")
	if (any(v == "*")) v <- metrics
	v <- sort(v)
	v <- omnibus::pmatchSafe(v, metrics, useFirst = TRUE)
	
	units <- omnibus::pmatchSafe(units, c("degrees", "percent", "radians"), n = 1L)
	if (units == "radians") {
		unitRadians <- TRUE
		units <- "degrees"
	} else {
		unitRadians <- FALSE
	}

	if (nlyr(x) > 1) {
		warning("The input raster has > 1 layer. Only the first will be used.")
		x <- x[[1L]]
	}

	.locationRestore(x)
	.region(x)
	
	args <- list(
		cmd = "r.slope.aspect",
		elevation = sources(x),
		nprocs = faster("cores"),
		memory = faster("memory"),
		flags = c(.quiet(), "overwrite", "e")
	)
	
	if ("slope" %in% v) {
		args$slope <- .makeSourceName("slope", "rast")
		units <- omnibus::pmatchSafe(units, c("degrees", "percent"))
		args$format <- units
	}

	if ("aspect" %in% v) {
		args$aspect <- .makeSourceName("aspect", "rast")
		units <- omnibus::pmatchSafe(units, c("degrees", "percent"))
		args$format <- units
		if (northIs0) args$flags <- c(args$flags, "n")
	}

	if ("profileCurve" %in% v) args$pcurvature = .makeSourceName("profileCurve", "rast")
	if ("tanCurve" %in% v) args$tcurvature = .makeSourceName("tanCurve", "rast")
	if ("dx" %in% v) args$dx = .makeSourceName("dx", "rast")
	if ("dy" %in% v) args$dy = .makeSourceName("dy", "rast")
	if ("dxx" %in% v) args$dxx = .makeSourceName("dxx", "rast")
	if ("dyy" %in% v) args$dyy = .makeSourceName("dyy", "rast")
	if ("dxy" %in% v) args$dxy = .makeSourceName("dxy", "rast")
	
	do.call(rgrass::execGRASS, args)
	
	if ("slope" %in% v) {

		# convert to radians
		if (unitRadians) {

			oldSlopeSrc <- args$slope
			slopeSrc <- .makeSourceName("terrain_r_mapcalc_slope", "raster")
			ex <- paste0(slopeSrc, " = ", pi, " * ", oldSlopeSrc, " / 180")
			rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))
			on.exit(.rm(oldSlopeSrc, type = "raster", warn = FALSE), add = TRUE)
		
		} else {
			slopeSrc <- args$slope
		}

		out <- makeGRaster(slopeSrc, "slope")
	}
	
	if ("aspect" %in% v) {

		# force undefined aspect to a new value (r.terrain assigns undefined aspect to -9999)
		oldAspectSrc1 <- args$aspect
		aspectSrc <- .makeSourceName("terrain_r_mapcalc_undefined", "raster")
		if (is.na(undefinedAspect)) {
			ex <- paste0(aspectSrc, " = if(", oldAspectSrc1, " < 0, null(), ", oldAspectSrc1, ")")
		} else {
			ex <- paste0(aspectSrc, " = if(", oldAspectSrc1, " < 0, ", undefinedAspect, ", ", oldAspectSrc1, ")")
		}
		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))
		on.exit(.rm(oldAspectSrc1, type = "raster", warn = FALSE), add = TRUE)
	
		# convert to radians
		if (unitRadians) {

			oldAspectSrc2 <- aspectSrc
			aspectSrc <- .makeSourceName("terrain_r_mapcalc_radians", "raster")
			ex <- paste0(aspectSrc, " = ", pi, " * ", oldAspectSrc2, " / 180")
			rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))
			on.exit(.rm(oldAspectSrc2, type = "raster", warn = FALSE), add = TRUE)
		
		}

		this <- makeGRaster(aspectSrc, "aspect")
		if (exists("out", inherits = FALSE)) {
			out <- c(out, this)
		} else {
			out <- this
		}
	}

	if ("profileCurve" %in% v) {
		this <- makeGRaster(args$pcurvature, "profileCurve")
		if (exists("out", inherits = FALSE)) {
			out <- c(out, this)
		} else {
			out <- this
		}
	}

	if ("tanCurve" %in% v) {
		this <- makeGRaster(args$tcurvature, "tanCurve")
		if (exists("out", inherits = FALSE)) {
			out <- c(out, this)
		} else {
			out <- this
		}
	}

	if ("dx" %in% v) {
		this <- makeGRaster(args$dx, "dx")
		if (exists("out", inherits = FALSE)) {
			out <- c(out, this)
		} else {
			out <- this
		}
	}

	if ("dy" %in% v) {
		this <- makeGRaster(args$dy, "dy")
		if (exists("out", inherits = FALSE)) {
			out <- c(out, this)
		} else {
			out <- this
		}
	}

	if ("dxx" %in% v) {
		this <- makeGRaster(args$dxx, "dxx")
		if (exists("out", inherits = FALSE)) {
			out <- c(out, this)
		} else {
			out <- this
		}
	}

	if ("dyy" %in% v) {
		this <- makeGRaster(args$dyy, "dyy")
		if (exists("out", inherits = FALSE)) {
			out <- c(out, this)
		} else {
			out <- this
		}
	}

	if ("dxy" %in% v) {
		this <- makeGRaster(args$dxy, "dxy")
		if (exists("out", inherits = FALSE)) {
			out <- c(out, this)
		} else {
			out <- this
		}
	}
	
	out

	} # EOF
)

#' Connect to "GRASS"
#'
#' @description This function initializes a **GRASS** "project" (previously known in **GRASS** as a "location"; see `vignette("projects_mapsets", package = "fasterRaster")`). You need to run this function (often just once) before you use most functions in **fasterRaster**. This function is of use to developers, not most users.
#'
#' @param x Any object from which a coordinate reference system (CRS) can be acquired. Ergo, any of:
#' * A `SpatRaster`, `SpatVector`, `SpatExtent`, `stars`, or `sf` object
#' * A `crs` object (i.e., from [sf::st_crs()]).
#' * A CRS (coordinate reference system) WKT string. Some PROJ4 strings *might* work, too.
#'
#' @param location Character or `NULL` (default): Name of the location.
#'
#' @param overwrite Logical: If `FALSE` (default), and a **GRASS** "coordinate reference frame" with the given name has already been created, then the function will fail. If `TRUE`, then the existing **GRASS** "coordinate reference frame" of the same name will be overwritten. *NOTE*: This will **not** remove any **R** objects associated with rasters or vectors in the "location", but they will no longer work because the objects they point to will be overwritten.
#
#' @param warn Logical: If `TRUE` (default) and `overwrite` is `TRUE`, then display a warning.
#'
#' @return A [GLocation] object (invisibly).
#'
#' @aliases .locationCreate
#' @rdname locationCreate
#' @keywords internal
methods::setMethod(
	f = ".locationCreate",
	signature = c(x = "character"),
	function(x, location = NULL, overwrite = FALSE, warn = TRUE) ..locationCreate(x = x, location = location, overwrite = overwrite, warn = warn)
)

#' @aliases .locationCreate
#' @rdname locationCreate
#' @keywords internal
methods::setMethod(
	f = ".locationCreate",
	signature = c(x = "SpatRaster"),
	function(x, location = NULL, overwrite = FALSE, warn = TRUE) ..locationCreate(x = x, location = location, overwrite = overwrite, warn = warn)
)

#' @aliases .locationCreate
#' @rdname locationCreate
#' @keywords internal
methods::setMethod(
	f = ".locationCreate",
	signature = c(x = "SpatVector"),
	function(x, location = NULL, overwrite = FALSE, warn = TRUE) ..locationCreate(x = x, location = location, overwrite = overwrite, warn = warn)
)

#' @aliases .locationCreate
#' @rdname locationCreate
#' @keywords internal
methods::setMethod(
	f = ".locationCreate",
	signature = c(x = "sf"),
	function(x, location = NULL, overwrite = FALSE, warn = TRUE) ..locationCreate(x = x, location = location, overwrite = overwrite, warn = warn)
)

#' @noRd
..locationCreate <- function(
	x,
	location = NULL,
	overwrite = FALSE,
	warn = TRUE
) {

	# for debugging
	if (FALSE) {

		x <- madElev
		location <- NULL
		overwrite <- TRUE
		warn <- TRUE
	
	}

	# rgrass::set.ignore.stderrOption(!faster("debug"))

	### CRS
	if (inherits(x, c("SpatRaster", "SpatVector", "SpatExtent"))) {
		coordRef <- terra::crs(x)
	} else if (inherits(x, c("sf", "stars"))) {
		coordRef <- sf::st_crs(x)
	} else if (inherits(x, "crs")) {
		coordRef <- unclass(x)$wkt
	} else if (inherits(x, "character")) {
		coordRef <- x
	}

	if (inherits(coordRef, "crs")) coordRef <- coordRef$wkt

	if (is.null(location)) {
	
		string <- sf::st_crs(x)$input
		niceCRS <- gsub(string, pattern = "\\/", replacement = "")
		niceCRS <- gsub(niceCRS, pattern = "\\(", replacement = "")
		niceCRS <- gsub(niceCRS, pattern = "\\)", replacement = "")
		niceCRS <- gsub(niceCRS, pattern = " ", replacement = "_")
		niceCRS <- gsub(niceCRS, pattern = "__", replacement = "_")
		
		# sf::st_crs(x)$input sometimes returns entire thing
		if (nchar(niceCRS) > 50) {
			between <- gregexpr(string, pattern = "\"")[[1L]]
			niceCRS <- substr(string, between[1L] + 1L, between[2L] - 1L)
		}

		rand <- omnibus::rstring(1L)
		location <- paste0(niceCRS, "_", rand)

	}

	locs <- .locations(warn = FALSE)
	locationExists <- !is.null(locs) && any(location == locs)

	opts <- faster()
	grassDir <- opts$grassDir
	addonsDir <- opts$addonsDir
	workDir <- opts$workDir

	if (is.na(grassDir)) stop("You must specify the folder in which GRASS is installed using faster().")
	
	### do we need a new GRASS session or to switch the location/working directory?
	if (!overwrite & locationExists) {

		stop("Location with same name already exists. If you want to overwrite it, use `overwrite = TRUE`.")

	} else if (overwrite & locationExists) {
	
		if (warn) warning(paste0("The GRASS session with these properties has been overwritten:\n  * location: ", location, "\n  * workDir: ", workDir, ".\n  All previously existing files have been removed."), immediate. = TRUE)
		
		rgrass::unset.GIS_LOCK()
		rgrass::remove_GISRC()
		rgrass::unlink_.gislock()

		fp <- file.path(workDir, location)
		files <- list.files(fp, include.dirs = TRUE, full.names = TRUE, recursive = TRUE)
		files <- rev(files)
		unlink(files, recursive = TRUE)
		unlink(fp, recursive = TRUE)
	
	}
		
	### start new GRASS session
	emptyRast <- terra::rast(matrix(1L), crs = coordRef)

	dir.create(workDir, showWarnings = FALSE, recursive = TRUE)

	### start the GRASS session
	suppressWarnings(
		session <- rgrass::initGRASS(
			gisBase = grassDir,
			# addon_base = addonsDir,
			home = workDir,
			gisDbase = workDir, # ?
			SG = emptyRast,
			location = location,
			mapset = "PERMANENT",
			override = TRUE, # must be TRUE to restart, even in different location/mapset
			remove_GISRC = overwrite, # ???
			ignore.stderr = TRUE
		)
	)

	rgrass::set.ignore.stderrOption(!faster("debug"))
	.fasterRaster$grassStarted <- TRUE
	.fasterRaster$grassVersion <- grassInfo("versionNumber")
	
	n <- length(.fasterRaster$locations)
	.fasterRaster$locations[[n + 1L]] <- coordRef
	names(.fasterRaster$locations)[n + 1L] <- location

	.fasterRaster$activeLocation <- location

	session <- GLocation(
		location = location,
		mapset = "PERMANENT",
		crs = coordRef,
		workDir = workDir
	)

	invisible(session)

}

if (grassStarted()) {

# Setup
library(terra)

# Climate raster:
madChelsa <- fastData("madChelsa")

# Convert a SpatRaster to a GRaster:
chelsa <- fast(madChelsa)

# Generate raster with layers representing principal component predictions:
pcRast <- princomp(chelsa, scale = TRUE)
plot(pcRast)

# Get information on the PCA:
prinComp <- pcs(pcRast)

prinComp
summary(prinComp)
plot(prinComp)

}

#' @title Mathematical operations on each layer of a GRasters
#'
#' @description You can apply mathematical functions to each layer of a `GRaster`. The output is a `GRaster` with the same number or layers as the input. Available functions include:\cr
#'
#' * `NA`s:
#'      * `is.na()`
#'      * `not.na()`
#' * Absolute value: `abs()`
#' * Trigonometric functions (assumes values are in radians):
#'      * `cos()`
#'      * `sin()`
#'      * `tan()`
#'      * `acos()`
#'      * `asin()`
#'      * `atan()`
#'      * `atan2()`
#' * Exponential and logarithmic functions:
#'      * `exp()`
#'      * `log()` (natural log)
#'      * `ln()` (also natural log)
#'      * `log2()` (log, base 2)
#'      * `log10()` (log, base 10)
#'      * `log1p()` (same as `log(x + 1)`)
#'      * `log10p()` (same as `log(x + 1, base = 10)`)
#' * Power functions:
#'      * `sqrt()`
#'      * `x^y`
#' * Rounding:
#'      * `round()`
#'      * `floor()` (round down)
#'      * `ceiling()` (round up)
#'      * `trunc()` (remove decimal portion)
#'
#' @param x,y `GRaster`s.
#'
#' @param falseNA Logical (function `not.na()`): If `FALSE` (default), non-`NA` cells will be converted to 1, and `NA` cells to 0. If `TRUE`, non-`NA` cells will be converted to  and `NA` cells will stay as `NA`.
#'
#' @param base Numeric: Base of the logarithm.
#'
#' @param digits Numeric: Number of digits to round to. If negative, then rounding is to the nearest positive power of 10. For example, if `digits = -2`, then the `GRaster` values are rounded to the nearest 100.
#'
#' @returns A `GRaster`.
#'
#' @example man/examples/ex_GRaster_arithmetic_by_layer.r
#'
#' @aliases is.na
#' @rdname math
#' @exportMethod is.na
setMethod(
	"is.na",
	signature(x = "GRaster"),
	function(x) {
	
	.locationRestore(x)
	.region(x)

	srcs <- .makeSourceName("isNA", "raster", nlyr(x))
	for (i in seq_len(nlyr(x))) {
	
		ex <- paste0(srcs[i], " = int(if(isnull(", sources(x)[i], "), 1, 0))")
		rgrass::execGRASS(
			cmd = "r.mapcalc",
			expression = ex,
			flags = c(.quiet(), "overwrite")
		)
	
	}
	makeGRaster(srcs, names(x))

	} # EOF
)

#' @aliases not.na
#' @rdname math
#' @exportMethod not.na
methods::setMethod(
	f = "not.na",
	signature(x = "GRaster"),
	function(x, falseNA = FALSE) {
	
	.locationRestore(x)
	.region(x)

	srcs <- .makeSourceName("not_na_r_mapcalc", "raster", nlyr(x))
	for (i in seq_len(nlyr(x))) {
	
		ex <- if (falseNA) {
			paste0(srcs[i], " = int(if(isnull(", sources(x)[i], "), 1, 0))")
		} else {
   			paste0(srcs[i], " = int(if(isnull(", sources(x)[i], "), 0, 1))")
		}
		# ex <- if (falseNA) {
		# 	paste0(srcs[i], " = int(if(isnull(", sources(x)[i], "), 1, null()))")
		# } else {
   		# 	paste0(srcs[i], " = int(if(isnull(", sources(x)[i], "), 1, 0))")
		# }
		rgrass::execGRASS(
			cmd = "r.mapcalc",
			expression = ex,
			flags = c(.quiet(), "overwrite")
		)
	
	}
	makeGRaster(srcs, names(x))

	}
)

#' @aliases abs
#' @rdname math
#' @exportMethod abs
setMethod(
	"abs",
	signature(x = "GRaster"),
	function(x) .genericRastFx("abs", x)
)

#' @aliases sin
#' @rdname math
#' @export
#' @exportMethod sin
setMethod(
	"sin",
	signature(x = "GRaster"),
	function(x) .genericTrig("sin", x)
)

#' @aliases cos
#' @rdname math
#' @export
#' @exportMethod cos
setMethod(
	"cos",
	signature(x = "GRaster"),
	function(x) .genericTrig("cos", x)
)

#' @aliases tan
#' @rdname math
#' @export
#' @exportMethod tan
setMethod(
	"tan",
	signature(x = "GRaster"),
	function(x) .genericTrig("tan", x)
)

#' @aliases asin
#' @rdname math
#' @export
#' @exportMethod asin
setMethod(
	"asin",
	signature(x = "GRaster"),
	function(x) .genericArcTrig("asin", x)
)

#' @aliases acos
#' @rdname math
#' @export
#' @exportMethod acos
setMethod(
	"acos",
	signature(x = "GRaster"),
	function(x) .genericArcTrig("acos", x)
)

#' @aliases atan
#' @rdname math
#' @export
#' @exportMethod atan
setMethod(
	"atan",
	signature(x = "GRaster"),
	function(x) .genericArcTrig("atan", x)
)

#' @aliases atan2
#' @rdname math
#' @export
#' @exportMethod atan2
setMethod(
	"atan2",
	signature(y = "GRaster", x = "GRaster"),
	function(y, x) {

	compareGeom(y, x)
	.locationRestore(x)
	.region(x)
	
	ny <- nlyr(y)
	nx <- nlyr(x)
	
	if (ny == 1L & nx > 1L) {
		y <- y[[rep(1L, nx)]]
	} else if (ny > 1L & nx == 1L) {
		x <- x[[rep(1L, ny)]]
	} else if (ny != nx) {
		stop("Rasters must have the same number of layers, or at least one raster must have one layer.")
	}
	
	n <- max(ny, nx)
	
	prec <- .getPrec(c(x, y), NULL)

	for (i in seq_len(n)) {
		
		src <- .makeSourceName("arith_by_layer", "rast")
		ex <- paste0(src, " = atan(", prec, "(", sources(x)[i], ") , ", prec, "(", sources(y)[i], "))  * (", pi, " / 180)")
		this <- .genericArithRast(name = "atan2", src = src, ex = ex)
		if (i == 1L) {
			out <- this
		} else {
			out <- c(out, this)
		}

	}
	out
		
	} # EOF
)

#' @aliases exp
#' @rdname math
#' @exportMethod exp
setMethod(
	"exp",
	signature = "GRaster",
	function(x) .genericRastFx("exp", x)
)

#' @aliases log1p
#' @rdname math
#' @exportMethod log1p
setMethod(
	"log1p",
	signature = "GRaster",
	function(x) {

	.locationRestore(x)
	.region(x)

	srcs <- .makeSourceName("arith_by_layer", "rast", nlyr(x))
	for (i in seq_len(nlyr(x))) {

		ex <- paste0(srcs[i], " = log(", sources(x)[i], " + 1)")
		rgrass::execGRASS(
			cmd = "r.mapcalc",
			expression = ex,
			flags = c(.quiet(), "overwrite")
		)

	}
	makeGRaster(srcs, names(x))
		
	} # EOF
)

#' @aliases log10p
#' @rdname math
#' @exportMethod log10p
setMethod(
	"log10p",
	signature = "GRaster",
	function(x) {

	.locationRestore(x)
	.region(x)

	srcs <- .makeSourceName("arith_by_layer", "rast", nlyr(x))
	for (i in seq_len(nlyr(x))) {

		ex <- paste0(srcs[i], " = log(", sources(x)[i], " + 1, 10)")
		rgrass::execGRASS(
			cmd = "r.mapcalc",
			expression = ex,
			flags = c(.quiet(), "overwrite")
		)

	}
	makeGRaster(srcs, names(x))
		
	} # EOF
)

#' @aliases log
#' @rdname math
#' @export
#' @exportMethod log
setMethod(
	"log",
	signature = "GRaster",
	function(x, base = exp(1)) .genericFx2("log", x, base)
)

#' @aliases log ln
#' @rdname math
#' @exportMethod ln
setMethod(
	"ln",
	signature = "GRaster",
 	function(x) .genericFx2("log", x, y = exp(1))
)

#' @aliases log2
#' @rdname math
#' @export
#' @exportMethod log2
setMethod(
	"log2",
	signature = "GRaster",
	function(x) .genericFx2("log", x, 2)
)

#' @aliases log10
#' @rdname math
#' @exportMethod log10
setMethod(
	"log10",
	signature = "GRaster",
	function(x) .genericFx2("log", x, 10)
)

#' @aliases sqrt
#' @rdname math
#' @exportMethod sqrt
setMethod(
	"sqrt",
	signature = "GRaster",
	function(x) .genericRastFx("sqrt", x)
)

#' @aliases round
#' @rdname math
#' @export
#' @exportMethod round
setMethod(
	"round",
	signature = "GRaster",
	function(x, digits = 0) {

		roundto <- if (digits == 0) {
			1
		} else if (digits > 0) {
			as.numeric(paste0("0.", paste(rep(0, digits - 1), collapse=""), "1"))
		} else if (digits < 0) {
			as.numeric(paste0("1", paste(rep(0, abs(digits)), collapse="")))
		}

		.genericFx2("round", x, roundto)
		
	} # EOF
)

#' @aliases floor
#' @rdname math
#' @export
#' @exportMethod floor
setMethod(
	"floor",
	signature = "GRaster",
	function(x) .genericRastFx("floor", x)
)

#' @aliases ceiling
#' @rdname math
#' @export
#' @exportMethod ceiling
setMethod(
	"ceiling",
	signature = "GRaster",
	function(x) .genericRastFx("ceil", x)
)

#' @aliases trunc
#' @rdname math
#' @export
#' @exportMethod trunc
setMethod(
	"trunc",
	signature = "GRaster",
	function(x) .genericRastFx("int", x)
)

#' Generic trigonometry function
#' @param fx	Character: Name of the function in **GRASS** tool `r.series`.
#' @param x		A `GRaster`.
#' @noRd
.genericTrig <- function(fx, x) {

	.locationRestore(x)
	.region(x)
	
	prec <- .getPrec(x, NULL)

	srcs <- .makeSourceName("r.mapcalc", "rast", nlyr(x))
	for (i in 1L:nlyr(x)) {
	
  		ex <- paste0(srcs[i], " = ", fx, "(", prec, "(", sources(x)[i], ") * 180 / ", pi, ")")
		rgrass::execGRASS(
			"r.mapcalc",
			expression = ex,
			flags = c(.quiet(), "overwrite")
		)
		
	}
	makeGRaster(srcs, names(x))

}

#' Generic "arc"-trigonometry function
#' @param fx	Character: Name of the function in **GRASS** tool `r.series`.
#' @param x		A `GRaster`.
#' @noRd
.genericArcTrig <- function(fx, x) {

	.locationRestore(x)
	.region(x)

 	prec <- .getPrec(x, "/")

	srcs <- .makeSourceName("arith_by_layer", "rast", nlyr(x))
	for (i in 1L:nlyr(x)) {
	
		ex <- paste0(srcs[i], " = ", fx, "(", prec, "(", sources(x)[i], ") * ", pi, " / 180)")
		rgrass::execGRASS(
			"r.mapcalc",
			expression = ex,
			flags = c(.quiet(), "overwrite")
		)
	}

	makeGRaster(srcs, names(x))

}

#' Generic function with one input (the GRaster)
#' @param fx	Character: Name of the function in **GRASS** tool `r.series`.
#' @param x		A `GRaster`.
#' @noRd
.genericRastFx <- function(fx, x) {

	.locationRestore(x)
	.region(x)
	
	nLayers <- nlyr(x)
	srcs <- .makeSourceName("arith_by_layer", "raster", nLayers)
	
	prec <- .getPrec(x, NULL)

	for (i in seq_len(nLayers)) {
	
		ex <- paste0(srcs[i], " = ", fx, "(", prec, "(", sources(x)[i], "))")

		rgrass::execGRASS(
			cmd = "r.mapcalc",
			expression = ex,
			flags = c(.quiet(), "overwrite")
		)
		
	}
	makeGRaster(srcs, names(x))

}

#' Generic function with two inputs (GRaster and a numeric)
#' @param fx	Character: Name of the function in **GRASS** tool `r.series`.
#' @param x		A `GRaster`.
#' @param y		A numeric.
#' @noRd
.genericFx2 <- function(fx, x, y) {

	.locationRestore(x)
	.region(x)

 	prec <- .getPrec(x, NULL)

	nLayers <- nlyr(x)
	srcs <- .makeSourceName("arith_by_layer", "raster", nLayers)
	
	for (i in seq_len(nLayers)) {
	
		ex <- paste0(srcs[i], " = ", fx, "(", prec, "(", sources(x)[i], "), ", y, ")")

		rgrass::execGRASS(
			cmd = "r.mapcalc",
			expression = ex,
			flags = c(.quiet(), "overwrite")
		)
	}
	makeGRaster(srcs, fx)

}

#' Convert degrees between 'north-orientation' and 'east orientation'
#'
#' @description This function converts facing between "north orientation" and "east orientation".
#'
#' In "north orientation" systems, a 0-degree facing is north, and the angle of facing proceeds clockwise. For example, a 90 degree facing faces east, 180 south, and 270 west. In "east orientation", a 0-degree facing is east, and the facing angle proceeds counter-clockwise. For example, 90 is north, 180 is west, and 270 south.
#'
#' @param x A numeric vector or a `GRaster` with cell values equal to facing (in degrees).
#'
#' @param units Character: "Units" of values in `x`: either `"degrees"` for degrees (default) or `"radians"`. Partial matching is used.
#'
#' @returns A `GRaster` or numeric vector. Values will be in the range between 0 and 360 and represents facing in the system "opposing" the input's system. For example, if the input is north orientation, the output will be east orientation. If the input is in east orientation, the output will be in north orientation.
#'
#' @example man/examples/ex_reorient.r
#'
#' @aliases reorient
#' @rdname reorient
#' @exportMethod reorient
methods::setMethod(
	f = "reorient",
	signature = c(x = "GRaster"),
	definition = function(x, units = "degrees") {

	units <- omnibus::pmatchSafe(units, c("degrees", "radians"), n = 1L)

	.locationRestore(x)
	.region(x)

	nLayers <- nlyr(x)
	srcs <- .makeSourceName("reorient", "raster", n = nLayers)

	for (i in seq_len(nLayers)) {

		if (units == "degrees") {
			ex <- paste0(srcs[i], " = ((360 - (", sources(x)[i], " % 360)) + 90) % 360")
		} else if (units == "radians") {
			ex <- paste0(srcs[i], " = ((2 * ", pi, " - (", sources(x)[i], " % (2 * ", pi, "))) + (", pi, " / 4)) % (2 * ", pi, ")")
		}
		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

	}
	makeGRaster(srcs, names(x))

	} # EOF
)

#' @aliases reorient
#' @rdname reorient
#' @exportMethod reorient
methods::setMethod(
	f = "reorient",
	signature = c(x = "numeric"),
	definition = function(x, units = "degrees") {
	
	units <- omnibus::pmatchSafe(units, c("degrees", "radians"), n = 1L)

	if (units == "degrees") {
		((360 - (x %% 360)) + 90) %% 360
	} else if (units == "radians") {
		((2 * pi - (x %% (2 * pi))) + (pi / 4)) %% (2 * pi)
	}
	
	} # EOF
)

library(data.table)

dt <- data.table(
   x = 1:10,
   y = letters[1:10],
   z = rnorm(10)
)

# make some values NA
dt[x == 4 | x == 8, y := NA_character_]
dt

# Replace NAs:
replaceNAs(dt, replace = -99, cols = "y")
dt

# Drop rows:
dropped <- dropRows(dt, 8:10)
dropped

# NB May not print... in that case, use:
print(dropped)

# We can also use replaceNAs() on vectors:
y <- 1:10
y[c(2, 10)] <- NA
replaceNAs(y, -99)

# Same as:
y <- 1:10
y[c(2, 10)] <- NA
y[is.na(y)] <- -99

if (grassStarted()) {

# Setup
library(terra)

# Elevation raster
madElev <- fastData("madElev")

# Convert SpatRaster to a GRaster:
elev <- fast(madElev)

# Create a "stack" of rasters for us to operate on:
x <- c(elev, elev^2, sqrt(elev))

# Demonstrate check for badly-named rasters:
names(x) <- c("cos", "asin", "exp")
fun <- "= cos / asin + exp"
appCheck(x, fun, failOnBad = FALSE)

# Rename rasters acceptable names and run the function:
names(x) <- c("x1", "x2", "x3")
fun <- "= (x1 / x2) + x3"
appCheck(x, fun, failOnBad = FALSE)
app(x, fun = fun)

# This is the same as:
(x[[1]] / x[[2]]) + x[[3]]

# We can view a table of app() functions using appFuns():
appFuns()

# We can also get the same table using:
data(appFunsTable)

# Apply other functions:
fun <- "= median(x1 / x2, x3, x1 * 2, cos(x2))"
app(x, fun = fun)

fun <- "= round(x1) * tan(x2) + log(x3, 10)"
app(x, fun = fun)

# Demonstrate effects of data type:
fun <- "= x1 + x3"
app(x, fun = fun, datatype = "float") # output is floating-point
app(x, fun = fun, datatype = "integer") # output is integer

# Some functions override the "datatype" argument:
fun <- "= sin(x2)"
app(x, fun = fun, datatype = "integer")

# Make a raster with random values [1:4], with equal probability of each:
fun <- "= round(rand(0.5, 4.5))"
rand <- app(elev, fun = fun)
rand

freqs <- freq(rand) # cell frequencies
print(freqs)

}

if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Example data
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster
elev <- fast(madElev)
elevs <- c(elev, elev, log10(elev) - 1, sqrt(elev))
names(elevs) <- c("elev1", "elev2", "log_elev", "sqrt_elev")

elev
elevs

# do some math
elev + 100
elev - 100
elev * 100
elev / 100
elev ^ 2
elev %/% 100 # divide then round down
elev %% 100 # modulus

100 + elev
100 %/% elev
100 %% elev

elevs + 100
100 + elevs

# math with logicals
elev + TRUE
elev - TRUE
elev * TRUE
elev / TRUE
elev ^ TRUE
elev %/% TRUE # divide then round down
elev %% TRUE # modulus

elevs + TRUE
TRUE + elevs

# Raster interacting with raster(s):
elev + elev
elev - elev
elev * elev
elev / elev
elev ^ log(elev)
elev %/% sqrt(elev) # divide then round down
elev %% sqrt(elev) # modulus

elevs + elev
elev * elevs

# sign
abs(-1 * elev)
abs(elevs)

# powers
sqrt(elevs)

# trigonometry
sin(elev)
cos(elev)
tan(elev)

asin(elev)
acos(elev)
atan(elev)

atan(elevs)
atan2(elev, elev^1.2)
atan2(elevs, elev^1.2)
atan2(elev, elevs^1.2)
atan2(elevs, elevs^1.2)

# logarithms
exp(elev)
log(elev)
ln(elev)
log2(elev)
log1p(elev)
log10(elev)
log10p(elev)
log(elev, 3)

log(elevs)

# rounding
round(elev + 0.5)
floor(elev + 0.5)
ceiling(elev + 0.5)
trunc(elev + 0.5)

}

#' @title Make a copy of an object in GRASS
#'
#' @description Create a copy of a `GRaster` or `GVector` in **GRASS**.  This function is used internally and is of little use to most users.  This only creates a copy of the object in the **GRASS** session--to make a `GRaster` or `GVector`, [makeGRaster()] or [makeGVector()] need to be called after making the copy. Note that if the object is multi-layered, then a copy is made of each layer.
#'
#' @param x `GRaster`, `GVector`, or character: The object or the [sources()] name(s) of the object(s) to be copied. Can take multi-layered objects or multiple `sources`.
#'
#' @param type Character or `NULL` (default): Either "raster" or "vector". If a character, there must be one per value in `x`. If `NULL`, will attempt to auto-detect (takes longer).
#'
#' @param topo `NULL` (default) or `"2D"` or `"3D"`.
#' 
#' @param reshapeRegion Logical: If `TRUE`, reshape the region to match `x` (`GRaster`s only).
#'
#' @returns Character vector representing the [sources()] names of the copied object(s), plus makes a copy of the given object(s) in **GRASS**.
#'
#' @aliases .copyGSpatial
#' @rdname copyGSpatial
#' @keywords internal
methods::setMethod(
	f = ".copyGSpatial",
	signature = c(x = "GRaster"),
	function(x, reshapeRegion = TRUE) .copyGRaster(x, topo = topology(x), reshapeRegion = reshapeRegion)
)

#' @aliases .copyGSpatial
#' @rdname copyGSpatial
#' @keywords internal
methods::setMethod(
	f = ".copyGSpatial",
	signature = c(x = "GVector"),
	function(x) .copyGVector(x)
)

#' @aliases .copyGSpatial
#' @rdname copyGSpatial
#' @keywords internal
methods::setMethod(
	f = ".copyGSpatial",
	signature = c(x = "character"),
	function(x, type = NULL, topo = NULL, reshapeRegion = TRUE) {
	
	if (is.null(type)) {

		srcs <- .ls()
		types <- names(srcs)
		type <- types[match(x, srcs)]

	} else {
		type <- omnibus::pmatchSafe(type, c("raster", "vector", "raster3d", "vector3d"), nmax = 2L)
		if (length(type == 2L)) type <- type[1L]
	}

	n <- length(x)
	srcs <- rep(NA_character_, n)

	for (i in seq_len(n)) {
		
		if (type[i] %in% c("raster", "raster3d")) {
			srcs[i] <- .copyGRaster(x[i], topo = topo, reshapeRegion = reshapeRegion)
		} else if (type[i] %in% c("vector", "vector3d")) {
			srcs[i] <- .copyGVector(x[i])
		}
	
	}
	srcs
		
	} # EOF

)

#' @param x A `GRaster` or [sources()] name of one.
#' @param topo "2D" or "3D"
#' @param reshapeRegion Logical. If `TRUE`, `x` must be a `GRaster`.
#'
#' @returns [sources()] names of copied rasters.
#'
#' @aliases .copyGRaster
#' @rdname copyGSpatial
#' @keywords internal
.copyGRaster <- function(x, topo = "2D", reshapeRegion = TRUE) {

	# NB This function could use `g.copy`, but in some cases it does not have the desired effect. For example, when a MASK raster is present, it correctly copies cells that are not masked, but when the MASK is removed, the masked cells re-appear. Similarly, it seems to ignore the region when copying.

	if (inherits(x, "GRaster")) {

		.locationRestore(x)
		if (reshapeRegion) .region(x)
		srcs <- sources(x)

	} else {

		srcs <- x
		if (reshapeRegion) {

			args <- list(
				cmd = "g.region",
				flags = .quiet()
			)

			if (is.null(topo)) {

				topo <- "2D" # guessing!
				
				if (!is.null(.quiet)) warning("Assuming raster is 2D.")

				args$raster <- srcs[1L]

			} else if (topo == "2D") {
				args$raster <- srcs[1L]
			} else if (topo == "3D") {
				args$raster_3d <- srcs[1L]
			}

			do.call(rgrass::execGRASS, args = args)

		}

	}

	nLayers <- length(srcs)
	out <- .makeSourceName("copyGSpatial_r_mapcalc", type = "raster", nLayers)

	for (i in seq_len(nLayers)) {

		ex <- paste0(out[i], " = ", srcs[i])

		rgrass::execGRASS(
			cmd = "r.mapcalc",
			expression = ex,
			flags = c(.quiet(), "overwrite")
		)

	}

	out
	
}

#' @param x A `GVector` or the [sources()] name of one.
#'
#' @aliases .copyGRaster
#' @rdname copyGSpatial
#' @keywords internal
.copyGVector <- function(x) {

	if (inherits(x, "GVector")) {
		.locationRestore(x)
		srcs <- sources(x)
	} else {
		srcs <- x
	}

	n <- length(srcs)

	out <- .makeSourceName("copyGSpatial", type = "vector", n = n)

	for (i in seq_len(n)) {
		
		fromTo <- paste0(srcs[i], ",", out[i])
		rgrass::execGRASS(
			cmd = "g.copy",
			vector = fromTo,
			flags = c(.quiet(), "overwrite")
		)

	}

	out

}

#' Tests if a GRASS vector is linked to an attribute table
#'
#' @description **GRASS** vectors can be lined to one or more attribute tables, or "databases.". This function tests to see if the vector does indeed have a database. This function is typically used by developers.
#'
#' @param x A `GVector` or the [sources()] name of a vector in **GRASS**.
#'
#' @returns Logical.
#'
#' @aliases .vHasDatabase
#' @rdname vHasDatabase
#' @keywords internal
.vHasDatabase <- function(x) {

	if (inherits(x, "GVector")) {
		.locationRestore(x)
		src <- sources(x)
	} else {
		src <- x
	}

	info <- rgrass::execGRASS(
		"v.db.connect",
		map = src,
		flags = c(.quiet(), "p"),
		intern = TRUE
	)

	if (length(info) < 1L) {
		FALSE
	} else if (length(info) > 1L) {
		any(grepl(info, pattern = "is connected by"))
	} else {
		!grepl(info, pattern = "is not connected to a database")
	}

}

if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")
madChelsa <- fastData("madChelsa")

### What raster formats can we attempt to write?
writeRaster()

### Save GRaster to disk (using temporary file)
elev <- fast(madElev)
filename <- tempfile(fileext = ".tif")
writeRaster(elev, filename)

# Load raster from disk
elev2 <- fast(filename)
elev2

### Save multi-layer GRaster to disk in one file (using temporary file)
chelsa <- fast(madChelsa)
filename <- tempfile(fileext = ".tif")
writeRaster(chelsa, filename)

# Load raster from disk
chelsa2 <- fast(filename)
chelsa2

### Save multi-layer GRaster to disk layer-by-layer (using temporary file)
chelsa <- fast(madChelsa)
filename <- tempfile(fileext = ".tif")
writeRaster(chelsa, filename, byLayer = TRUE)

# Load one of the rasters from disk
filename2 <- sub(filename, pattern = ".tif", replacement = "_bio1.tif")
chelsaBio1 <- fast(filename2)
chelsaBio1

}

#' Combine one or more GVectors
#'
#' @description `rbind()` combines two or more `GVector`s of the same type (points, lines, or polygons) and same coordinate reference system. You can speed operations by putting the vector that is largest in memory first in `rbind(...)`. If the `GVector`s have data tables, these will also be combined using `rbind()` if their column names and data types match.
#'
#' @param ... One or more `GVector`s.
#' @param deparse.level See [rbind()].
#'
#' @returns A `GVector`.
#'
#' @example man/examples/ex_cbind_rbind_addTable_dropTable.r
#'
#' @seealso [colbind()], \code{\link[fasterRaster]{addTable<-}}, [dropTable()]
#'
#' @aliases rbind
#' @rdname rbind
#' @exportMethod rbind
methods::setMethod(
	f = "rbind",
	signature = "GVector",
	function(...) {
	
	dots <- list(...)
	.locationRestore(dots[[1L]])

	# unlist any lists
	dots <- omnibus::unlistRecursive(dots)

	# comparable?
	nDots <- length(dots)
	if (nDots == 1L) {
		return(dots[[1L]])
	} else if (nDots > 1L) {
		for (i in 2L:nDots) compareGeom(dots[[1L]], dots[[i]], geometry = TRUE, topo = TRUE)
	}

	### recat everything
	# v.patch does not increment cat numbers, so different geometries will have the same cat.
	# This step increases cat numbers in each subsequent vector so conflicts do not happen.

	srcs <- sapply(dots, sources)
	src <- .rbind(srcs)

	# x <- dots[[1L]]
	# src <- sources(x)
	# srcs <- sapply(dots, sources)

	# cats <- .vCats(src, db = FALSE)
	# topCat <- max(cats)
	
	# for (i in 2L:nDots) {

	# 	srcs[i] <- .vIncrementCats(srcs[i], add = topCat)
	# 	cats <- .vCats(srcs[i], db = FALSE)
	# 	topCat <- max(cats)
	
	# }

	# ### combine vectors
	# # seems like we can combine at least 11 vectors at a time, but not a lot at a time
	# srcsAtATime <- 10L # number of sources to combine at a time (plus the running `x` source)

	# nSrcs <- length(srcs)
	# sets <- ceiling(nSrcs / srcsAtATime)
	
	# for (set in seq_len(sets)) {

	# 	index <- (1L + srcsAtATime * (set - 1L)) : min(nSrcs, set * srcsAtATime)
	# 	srcIn <- srcs[index]
	# 	input <- paste(srcIn, collapse = ",")
	# 	input <- paste0(src, ",", input)
	
	# 	src <- .makeSourceName("v_patch", "vector")

	# 	rgrass::execGRASS(
	# 		cmd = "v.patch",
	# 		input = input,
	# 		output = src,
	# 		flags = c(.quiet(), "overwrite")
	# 	)
		
	# }

	### combine data tables if possible
	tables <- lapply(dots, as.data.frame)
	tables <- lapply(tables, data.table::as.data.table)
	
	nRows <- sapply(tables, nrow)
	nGeoms <- sapply(dots, ngeom)
	if (all(nRows == nGeoms)) {
		table <- tryCatch(
			do.call(rbind, tables),
			error = function(cond) FALSE
		)
	} else {
		table <- FALSE
	}

	if (is.logical(table)) {

		warning("Data tables cannot be combined so will be dropped from the output vector.")
		table <- NULL

	}

	makeGVector(src, table = table)
	
	} # EOF
)

#' Merge vectors in GRASS
#'
#' This function combines vectors with the same geometry in **GRASS**.
#'
#' @param srcs Character vector: The [sources()] names of vectors to combine. They must have the same [geomtype()].
#' @returns Character: [sources()] name of the combined vector in **GRASS**.
#'
#' @noRd
.rbind <- function(srcs) {

	thisCats <- .vCats(srcs[1L], db = FALSE)
	topCat <- max(thisCats)
	
	nSrcs <- length(srcs)
	for (i in 2L:nSrcs) {

		srcs[i] <- .vIncrementCats(srcs[i], add = topCat)
		thisCats <- .vCats(srcs[i], db = FALSE)
		topCat <- max(thisCats)
	
	}

	### combine vectors
	# seems like we can combine at least 11 vectors at a time, but not a lot at a time
	srcsAtATime <- 10L # number of sources to combine at a time (plus the running `x` source)

	sets <- ceiling(nSrcs / srcsAtATime)
	
	src <- srcs[1L]
	for (set in seq_len(sets)) {

		index <- (1L + srcsAtATime * (set - 1L)) : min(nSrcs, set * srcsAtATime)
		srcIn <- srcs[index]
		input <- paste(srcIn, collapse = ",")
		input <- paste0(src, ",", input)
	
		src <- .makeSourceName("v_patch", "vector")

		rgrass::execGRASS(
			cmd = "v.patch",
			input = input,
			output = src,
			flags = c(.quiet(), "overwrite")
		)
		
	}
	src

}

# rbind <- function(...) UseMethod("rbind")

if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Plant specimens (points) and land cover
madDypsis <- fastData("madDypsis")
madCover <- fastData("madCover")

# Convert to GVector and GRaster
dypsis <- fast(madDypsis)
cover <- fast(madCover)

### GVector

# Look at the data table:
as.data.table(dypsis)

# Which rows have no NAs?
complete.cases(dypsis)

# Which rows have at least one NA (opposite of above)?
missing.cases(dypsis)

### GRaster

# Look at the levels table:
levels(cover)

# Which rows of levels table have no NAs?
complete.cases(cover)

# Which rows have at least one NA (opposite of above)?
missing.cases(cover)

}

#' Remove rows and columns from a raster that are all NA
#'
#' @description This function removes any rows and columns from  a `GRaster` that are all `NA`.
#'
#' If the `GRaster` is a stack of rasters, then the rasters will all be trimmed to the same extent such that none have any rows or columns that are all `NA`. In other words, if at least one raster in the stack has a non-`NA` cell in a row or column, all rasters will retain that row or column.
#'
#' @param x A `GRaster`.
#'
#' @param pad Numeric integer: Number of `NA` rows and columns to retain. The default is 0.
#'
#' @returns A `GRaster`.
#'
#' @seealso [terra::trim()], [extend()], and **GRASS** tool `g.region`
#'
#' @example man/examples/ex_trim.r
#'
#' @aliases trim
#' @rdname trim
#' @exportMethod trim
methods::setMethod(
	f = "trim",
	signature = c(x = "GRaster"),
	function(x, pad = 0) {
	
	if (pad < 0) stop("Argument `pad` must be >= 0.")

	.locationRestore(x)
	.region(x)

	nLayers <- nlyr(x)

	# for each raster, find smallest extent necessary to encompass all non-NA rows/columns
	n <- s <- e <- w <- rep(NA_real_, nLayers)
	for (i in seq_len(nLayers)) {
	
		rgrass::execGRASS(
			cmd = "g.region",
			raster = sources(x)[i],
			zoom = sources(x)[i],
			flags = c(.quiet(), "overwrite")
		)

		reg <- .region()
		w[i] <- W(reg)
		e[i] <- E(reg)
		s[i] <- S(reg)
		n[i] <- N(reg)
	
	}

	wDelta <- w - W(x)
	eDelta <- e - E(x)
	sDelta <- s - S(x)
	nDelta <- n - N(x)

	w <- w[which.min(wDelta)]
	e <- e[which.min(eDelta)]
	s <- s[which.min(sDelta)]
	n <- n[which.min(nDelta)]

	ewres <- xres(x)
	nsres <- yres(x)

	if (pad > 0) {
	
		w <- max(w - pad * ewres, W(x))
		e <- min(e + pad * ewres, E(x))
  		s <- max(s - pad * nsres, S(x))
		n <- min(n + pad * nsres, N(x))
	
	}

	w <- as.character(w)
	e <- as.character(e)
	s <- as.character(s)
	n <- as.character(n)

	ewres <- as.character(ewres)
	nsres <- as.character(nsres)

	rgrass::execGRASS(
		cmd = "g.region",
		w = w, e = e, s = s, n = n,
		ewres = ewres, nsres = nsres,
		flags = c(.quiet(), "overwrite")
	)

	src <- .copyGSpatial(x, reshapeRegion = FALSE)
	makeGRaster(src, names(x), levels = cats(x), ac = activeCats(x))
	
	} # EOF
)

#' Regression intercept, slope, r2, and t-value across each set of cells
#'
#' @description This function performs a regression on each set of cells in a multi-layered `GRaster`. The output is a `GRaster` with the intercept, slope, r^2 value, and Student's t value. The regression formula is as `y ~ 1 + x`, where `x` is the layer number of each layer (e.g., 1 for the first or top layer in the input `GRaster`, 2 for the second or second-to-top layer, etc.). Note that this is restricted version of the functionality in [terra::regress()].
#'
#' @param y A multi-layer `GRaster`.
#'
#' @param x Ignored.
#'
#' @param na.rm Logical: If `FALSE`, any series of cells with `NA` in at least one cell results in an `NA` in the output.
#'
#' @returns A multi-layer `GRaster`.
#'
#' @seealso [terra::regress()]
#'
#' @example man/examples/ex_GRaster_arithmetic_across_layers.r
#'
#' @aliases regress
#' @rdname regress
#' @exportMethod regress
methods::setMethod(
	f = "regress",
	signature = c(y = "GRaster", x = "missing"),
	function(y, x, na.rm = FALSE) {
	
	if (nlyr(y) < 2) stop("The `GRaster` must have >1 layer.")

	.locationRestore(y)
	.region(y)

	fx <- "offset"
	fxName <- "intercept"
	intercept <- .genericMultiLayer(fx = fx, fxName = fxName, x = y, na.rm = na.rm)

	fx <- "slope"
	fxName <- "slope"
	slope <- .genericMultiLayer(fx = fx, fxName = fxName, x = y, na.rm = na.rm)

	fx <- "detcoeff"
	fxName <- "r2"
	r2 <- .genericMultiLayer(fx = fx, fxName = fxName, x = y, na.rm = na.rm)
		
	fx <- "tvalue"
	fxName <- "tvalue"
	tvalue <- .genericMultiLayer(fx = fx, fxName = fxName, x = y, na.rm = na.rm)
		
	c(intercept, slope, r2, tvalue)

	} # EOF
)

#' Extract values from a GRaster at locations in a points GVector
#'
#' @description `extract()` obtains the values of a `GRaster` or `GVector` associated with the locations of a set of points. The output depends on the input:
#' * **Case #1: `x` is a numeric or integer `GRaster` and `y` is a points `GVector`**: Returns values of cells that have points. If `xy` is `TRUE`, also returns the coordinates of the points.
#' * **Case #2: `x` is a categorical (factor) `GRaster` and `y` is a points `GVector`**: Same as case #1, but if `cats` is `TRUE`, returns category labels of cells that have points. If `xy` is `TRUE`, also returns the coordinates of the points.
#' * **Case #3: `x` is a categorical `GRaster` and `y` is a lines or polygons `GVector`**: Returns a summary (e.g., mean, standard deviation, etc.) of all cells that overlap the line(s) or polygon(s).
#' * **Case #4: `x` is a `GVector` and `y` is a points `GVector`**: Returns the data table row associated each point. If `xy` is `TRUE`, also returns the coordinates of the points.
#' Note that whenever a points `GVector` is allowed for `y`, a `data.frame`, `data.table`, `matrix`, or `numeric` values representing points can be used instead.
#'
#' @param x A `GRaster` or `GVector`.
#'
#' @param y A `GVector`, *or* a `data.frame` or `matrix` where the first two columns represent longitude and latitude (in that order), *or* a two-element numeric vector where the first column represents longitude and the second latitude. Values of `x` will be extracted from the points in `y`. `GVector`s can be of types points, lines, or polygons. 
#'
#' @param fun Character vector: Name(s) of function(s) to apply to values. This is used when `x` is a `GRaster` and `y` is a lines or polygons `GVector`. The method(s) specified by `fun` will be applied to all cell values that overlap with each geometry (i.e., individual cell values will not be returned). Valid functions include:
#' * `"countNonNA"`: Number of overlapping cells.
#' * `"countNA"`: Number of overlapping `NA` cells.
#' * `"mean"`: Average.
#' * `"min"`: Minimum.
#' * `"max"`: Minimum.
#' * `"sum"`: Sum.
#' * `"range"`: Maximum - minimum.
#' * `"sd"`: Sample standard deviation (same as [stats::sd()]).
#' * `"sdpop"`: Population standard deviation.
#' * `"var"`: Sample variance (same as [stats::var()]).
#' * `"varpop"`: Population variance.
#' * `"cv"`: Coefficient of variation.
#' * `"cvpop"`: Population coefficient of variation.
#' * `"median"`: Median.
#' * `"quantile"`: Quantile; you can specify the quantile using the `prob` argument.
#'
#' @param prob Numeric in the range from 0 to 1: Quantile which to calculate. The value of `prob` will be rounded to the nearest hundredth.
#'
#' @param overlap Logical: If `TRUE` (default), and `y` is a lines or polygons `GVector`, then account for potential overlap of geometries when extracting. This can be slow, so if you are sure geometries do not overlap, you can change this to `FALSE`. This argument is ignored if `y` is a points `GVector`.
#'
#' @param xy Logical: If `TRUE` and `y` represents points, also return the coordinates of each point. Default is `FALSE.`
#'
#' @param cats Logical (extracting from a raster): If `TRUE` (default) and `x` is a categorical `GRaster`, then return the category labels instead of the values.
#'
#' @param verbose Logical: If `TRUE`, display progress (will only function when extracting from points on a `GRaster` when the number of `GRaster`s is large, or when extracting using a "points" `GVector` with lots of points).
#'
#' @param ... Arguments to pass to [project()]. This is used only if extracting from a `GRaster` at locations specified by a `GVector`, and they have a different coordinate reference system. In this case, users should specify the `wrap` argument to [project()].
#'
#' @returns A `data.frame` or `data.table`.
#'
#' @example man/examples/ex_extract.r
#'
#' @seealso [terra::extract()], and modules `r.what` and `v.what` in **GRASS**
#'
#' @aliases extract
#' @rdname extract
#' @exportMethod extract
methods::setMethod(
    f = "extract",
    signature = c(x = "GRaster", y = "GVector"),
    function(
        x,
        y,
        fun = "mean",
        prob = 0.5,
        overlap = TRUE,
        xy = FALSE,
        cats = TRUE,
        verbose = FALSE,
        ...
    ) {

    .locationRestore(x)

    sameCRS <- compareGeom(x, y, stopOnError = FALSE, messages = FALSE)
    if (!sameCRS) {
    
        warning("The GVector has a different coordinate reference system than the GRaster.\n  The GVector was projected to the CRS of the GRaster. If the GVector spans\n  the international date line, you may need to specify `wrap = TRUE` when\n  using this function.", immediate. = TRUE)

        dots <- list(...)
        if (any(names(dots) == "wrap")) {
            wrap <- dots$wrap
        } else {
            wrap = FALSE
        }
        y <- project(y, x, wrap = wrap)

    }

    compareGeom(x, y)

    nLayers <- nlyr(x)
    gtype <- geomtype(y)

    ### if GVector is lines/polygons
    if (gtype %in% c("lines", "polygons")) {

        ### assuming may be overlap in geometries
        if (overlap) {
        
            numGeoms <- ngeom(y)

            for (i in seq_len(numGeoms)) {
            
                yy <- y[i]
                yy <- .aggregate(yy, gtype = "polygons", dissolve = FALSE)
                yy <- makeGVector(yy)
                
                thisOut <- extract(
                    x = x, y = yy,
                    fun = fun,
                    prob = prob,
                    overlap = FALSE
                )

                if (i == 1L) {
                    out <- thisOut
                } else {
                    out <- rbind(out, thisOut)
                }
            
            } # next geometry
        
        ### assuming no overlap in geometries
        } else {

            funs <- c(
                "countNonNA",
                "countNA",
                "min",
                "max",
                "range",
                "mean",
                "sd",
                "sdpop",
                "var",
                "varpop",
                "cv",
                "cvpop",
                "sum",
                "median",
                "quantile"
            )

            fun <- omnibus::pmatchSafe(fun, funs)

            # get sample versions
            sampleSd <- any(fun == "sd")
            sampleVar <- any(fun == "var")
            sampleCv <- any(fun == "cv")

            if (any(fun == "quantile")) {
                
                if (length(prob) > 1L) stop("Argument `prob` can only have one value.")
                if (prob > 1 | prob < 0) stop("Argument `prob` must be in the range [0, 1].")
                prob <- round(100 * prob)
            
            }

            method <- fun
            method <- method[!(method %in% c("sd", "var", "cv"))]
            method[method == "countNonNA"] <- "number"
            method[method == "countNA"] <- "null_cells"
            method[method == "min"] <- "minimum"
            method[method == "max"] <- "maximum"
            method[method == "mean"] <- "average"
            method[method == "sdpop"] <- "stddev"
            method[method == "varpop"] <- "variance"
            method[method == "cvpop"] <- "coeff_var"
            method[method == "quantile"] <- "percentile"
        
            method <- unique(method)

            for (i in seq_len(nLayers)) {

                prefix <- omnibus::rstring(1L)

                if (length(method) > 0L) {
                    
                    if (!.vHasDatabase(y)) .vAttachDatabase(y)

                    args <- list(
                        cmd = "v.rast.stats",
                        raster = sources(x)[i],
                        map = sources(y),
                        column_prefix = prefix,
                        method = method,
                        flags = c(.quiet())
                    )
                    if (gtype == "lines") args$flags <- c(args$flags, "d")
                    if (any(method == "percentile")) args$percentile <- prob
                    do.call(rgrass::execGRASS, args = args)

                    extracted <- .vAsDataTable(y)

                } else {
                    extracted <- data.table::data.table(NULL)
                }

                # sample SD, var, CV
                if (sampleSd | sampleVar | sampleCv) {
                    
                    sampleStats <- .sampleStatsFromRast(i = i, x = x, y = y, fun = fun, extracted = extracted, prefix = prefix)
                    extracted <- cbind(extracted, sampleStats)
                    
                }

                extracted <- extracted[!duplicated(extracted)]

                if (any(names(extracted) == "cat")) extracted$cat <- NULL
                if (any(names(extracted) == "cat_")) extracted$cat_ <- NULL
                nms <- names(extracted)
                nms <- gsub(nms, pattern = paste0(prefix, "_"), replacement = "")

                nms[nms == "number"] <- "countNonNA"
                nms[nms == "null_cells"] <- "countNA"
                nms[nms == "minimum"] <- "min"
                nms[nms == "maximum"] <- "max"
                nms[nms == "average"] <- "mean"
                nms[nms == "stddev"] <- "sdpop"
                nms[nms == "variance"] <- "varpop"
                nms[nms == "coeff_var"] <- "cvpop"
                nms[grepl(nms, pattern = "percentile")] <- "quantile"

                names(extracted) <- nms

                if (any(fun == "cvpop")) {
                    cvpop <- NULL
                    extracted[ , cvpop := cvpop / 100]
                }

                extracted <- extracted[ , ..fun]
                names(extracted) <- paste0(names(x)[i], "_", names(extracted))

                if (i == 1L) {
                    out <- extracted
                } else {
                    out <- cbind(out, extracted)
                }

            } # next raster layer

        } # no overlap/overlap in geometries

    ### if GVector is points
    } else {

        if (xy) {

            coords <- crds(y, z = is.3d(y))
            out <- data.table::as.data.table(coords)

        }

        vals <- .extractFromRasterAtPoints(x = x, y = y, cats = cats, verbose = verbose)

        if (exists("out", inherits = FALSE)) {
            out <- cbind(out, vals)
        } else {
            out <- vals
        }
        

    } # if lines/polygons or points
    if (!faster("useDataTable")) out <- as.data.frame(out)
    out

    } # EOF
)

#' Function to calculate sample sd, variance, and CV from raster
#'
#' @noRd
.sampleStatsFromRast <- function(i, x, y, fun, extracted, prefix) {

    ### calculate variance
    x2 <- x[[i]]^2

    # n * SUM(x^2)
    if (any(fun == "countNonNA")) {
        
        count <- extracted[[paste0(prefix, "_number")]]

        numer1 <- extract(
            x = x2,
            y = y,
            fun = "sum"
        )

        numer1 <- numer1[[1L]]
        numer1 <- numer1 * count

    
    } else {

        numer1 <- extract(
            x = x2,
            y = y,
            fun = c("countNonNA", "sum")
        )

        count <- numer1[[paste0(names(x2), "_countNonNA")]]
        count <- as.numeric(count) # avoid integer overflow
        sums <- numer1[[paste0(names(x2), "_sum")]]
        sums <- as.numeric(sums)
        numer1 <- count * sums

    }

    # (SUM(x))^2
    if (!any(fun == "sum")) {

        numer2 <- extract(
            x = x[[i]],
            y = y,
            fun = "sum"
        )
        numer2 <- numer2[[paste0(names(x), "_sum")]]

    } else {
        numer2 <- extracted[[paste0(prefix, "_sum")]]
    }

    numer2 <- as.numeric(numer2)
    numer2 <- numer2^2

    denom <- count * (count - 1)
    variance <- (numer1 - numer2) / denom

    ### calculate mean for CV
    if (any(fun == "cv")) {
    
        if (any(fun == "mean")) {
            means <- extracted[[paste0(prefix, "_average")]]
        } else {
        
            means <- extract(
                x = x[[i]],
                y = y,
                fun = "mean"
            )

            means <- means[[paste0(names(x), "_mean")]]
        
        }
    
    }

    thisExtracted <- data.table::data.table(NULL)

    if (any(fun == "var")) {
        var <- NULL
        thisExtracted[ , var := variance]
    }

    if (any(fun == "sd")) {
        sd <- NULL
        thisExtracted[ , sd := sqrt(variance)]
    }

    if (any(fun == "cv")) {
        cv <- NULL
        thisExtracted[ , cv := sqrt(variance) / means]
    }

    thisExtracted

}

#' @aliases extract
#' @rdname extract
#' @exportMethod extract
methods::setMethod(
    f = "extract",
    signature = c(x = "GRaster", y = "data.frame"),
    function(
        x,
        y,
        xy = FALSE,
        cats = TRUE
    ) {

    if (ncol(y) < 2L) stop("Argument `y` must have at least two columns. The first must represent longitude and the second latitude.")
    if (ncol(y) > 2L) warning("Argument `y` has more than two columns. The first will be assumed to represent longitude and the second latitude.")

    y <- terra::vect(y, geom = colnames(y)[1L:2L], crs = crs(x), keepgeom = FALSE)
    y <- fast(y)
    extract(x = x, y = y, xy = xy, cats = cats)

    } # EOF
)

#' @aliases extract
#' @rdname extract
#' @exportMethod extract
methods::setMethod(
    f = "extract",
    signature = c(x = "GRaster", y = "data.table"),
    function(
        x,
        y,
        xy = FALSE,
        cats = TRUE
    ) {

    y <- as.data.frame(y)
    extract(x = x, y = y, xy = xy, cats = cats)

    } # EOF
)

#' @aliases extract
#' @rdname extract
#' @exportMethod extract
methods::setMethod(
    f = "extract",
    signature = c(x = "GRaster", y = "matrix"),
    function(
        x,
        y,
        xy = FALSE,
        cats = TRUE
    ) {

    y <- as.data.frame(y)
    extract(x = x, y = y, xy = xy, cats = cats)

    } # EOF
)

#' @aliases extract
#' @rdname extract
#' @exportMethod extract
methods::setMethod(
    f = "extract",
    signature = c(x = "GRaster", y = "numeric"),
    function(
        x,
        y,
        xy = FALSE,
        cats = TRUE
    ) {

    if (length(y) != 2L) stop("Argument `y` must have two values, longitude and latitude.")
    y <- cbind(y)
    colnames(y) <- c("x", "y")
    extract(x = x, y = y, xy = xy, cats = cats)

    } # EOF
)

#' @aliases extract
#' @rdname extract
#' @exportMethod extract
methods::setMethod(
    f = "extract",
    signature = c(x = "GVector", y = "GVector"),
    function(x, y, xy = FALSE, verbose = TRUE) {

    if (geomtype(y) != "points") stop("Argument`y` must be a points vector.")
    if (is.3d(y)) warning("Coordinates in the z-dimension will be ignored.")

    .locationRestore(x)
    coords <- crds(y, z = FALSE)
    .extractFromVect(x, coords, xy, verbose = verbose)

    } # EOF

)

#' @aliases extract
#' @rdname extract
#' @exportMethod extract
methods::setMethod(
    f = "extract",
    signature = c(x = "GVector", y = "data.frame"),
    function(x, y, xy = FALSE, verbose = TRUE) .extractFromVect(x, y[ , 1L:2L], xy, verbose = verbose)
)

#' @aliases extract
#' @rdname extract
#' @exportMethod extract
methods::setMethod(
    f = "extract",
    signature = c(x = "GVector", y = "data.table"),
    function(x, y, xy = FALSE, verbose = TRUE) {

    y <- as.data.frame(y)
    .extractFromVect(x, y[ , 1L:2L], xy, verbose = verbose)

    } # EOF
)

#' @aliases extract
#' @rdname extract
#' @exportMethod extract
methods::setMethod(
    f = "extract",
    signature = c(x = "GVector", y = "matrix"),
    function(x, y, xy = FALSE, verbose = TRUE) .extractFromVect(x, y[ , 1L:2L], xy, verbose = verbose)
)

#' @aliases extract
#' @rdname extract
#' @exportMethod extract
methods::setMethod(
    f = "extract",
    signature = c(x = "GVector", y = "numeric"),
    function(x, y, xy = FALSE) {
        
    y <- cbind(y)
    .extractFromVect(x, y, xy, verbose = FALSE)

    } # EOF
)

# #' For internal use by other functions
# #' 
# #' @param y Character: [sources()] of points vector
# #' 
# #' @noRd
# methods::setMethod(
#     f = "extract",
#     signature = c(x = "GVector", y = "character"),
#     function(x, y, xy = FALSE, verbose = TRUE) {

#     y <- .crdsVect(y, z = FALSE, gm = "points", verbose = verbose)
#     .extractFromVect(x, y, xy)

#     } # EOF
# )

#' @param x GVector
#' @param coords 2-column matrix of coordinates
#' @param xy T/F: Return coordinates
#' @param nperSet Number of points to extract... too large throws an error. 1000 seems to break, so going with 900.
#' @param data.table return as `data.table`
#' @param verbose Logical.
#' 
#' @noRd
.extractFromVect <- function(x, y, xy, nPerSet = 900L, data.table = TRUE, verbose = FALSE) {

    .locationRestore(x)

    if (!inherits(y, "data.table")) y <- data.table::as.data.table(y)

    colnames(y) <- c("x", "y")
    n <- nrow(y)

	if (n > nPerSet) {


		sets <- ceiling(n / nPerSet)
        if (verbose | faster("verbose")) {
            pb <- utils::txtProgressBar(min = 0, max = sets, initial = 0, style = 3, width = 30)
        }

		out <- data.table::data.table()
		for (set in seq_len(sets)) {

            if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, set)
			
			yy <- y[(1L + (set - 1L) * nPerSet):min(set * nPerSet, n)]
			thisOut <- .extractFromVect(x, y = yy, xy = xy, data.table = TRUE, verbose = FALSE)
			out <- rbind(out, thisOut)
		
		}

        if (verbose | faster("verbose")) close(pb)
		
	} else {

		coords <- rep(NA_real_, 2L * n)
		coords[seq(1L, 2 * n, by = 2L)] <- y[["x"]]
		coords[seq(2L, 2 * n, by = 2L)] <- y[["y"]]

		gtype <- geomtype(x, grass = TRUE)

		info <- rgrass::execGRASS(
			cmd = "v.what",
			map = sources(x),
			coordinates = coords,
			type = gtype,
			flags = c(.quiet(), "verbose"), # could use "g" to make output easier to parse
			intern = TRUE
		)

		nonnas <- which(grepl(info, pattern = "Category: "))
		nas <- which(grepl(info, pattern = "Nothing found."))

		lenNonnas <- length(nonnas)
		lenNas <- length(nas)

		# get index of non-NAs
		if (lenNonnas > 0L & lenNas >= 0L) {
			
			names(nonnas) <- rep("notna", lenNonnas)
			names(nas) <- rep("na", lenNas)
			together <- c(nonnas, nas)
			together <- sort(together)

			nonnasIndex <- which(names(together) == "notna")

		} else if (lenNonnas > 0L & lenNas == 0) {
			nonnasIndex <- 1L:n
		}

		### at least one point was on the vector
		if (lenNonnas > 0L) {

			nons <- info[nonnas]
			nons <- sub(nons, pattern = "Category: ", replacement = "")
			nons <- as.integer(nons)

		}

		# vector has no data table (return IDs)
		if (nrow(x) == 0L) {

			id.x <- NULL
			out <- data.table::data.table(id.y = 1L:n, id.x = NA_integer_)
			if (lenNonnas > 0L) out[nonnasIndex, id.x := nons]

		# vector has data table
		} else {

			# make template of data table (1 row, all NAs)
			out <- as.data.table(x)[1L]

			classes <- sapply(out, class)
			cols <- names(x)
			
			for (i in seq_along(classes)) {

				if (classes[i] == "integer") {
					assign <- NA_integer_
				} else if (classes[i] == "numeric") {
					assign <- NA_real_
				} else if (classes[i] == "character") {
					assign <- NA_character_
				} else {
					assign <- NA
				}

				col <- cols[i]
				out[1L, (col) := assign]

			}

			# create template data table with all NAs
			out <- out[rep(1L, n)]
			id.y <- data.table::data.table(id.y = 1L:n)
			out <- cbind(id.y, out)

			# assign values to rows
			if (lenNonnas > 0L) {
			
				for (i in seq_along(cols)) {
					
					col <- cols[i]
					out[nonnasIndex, (col) := x@table[nons, get(col)]]
				
				}

			}

		} # vector has a data.table

		if (xy) out <- cbind(y, out)
		
	}
	
	if (!data.table & !faster("useDataTable")) out <- as.data.frame(out)
	out
    
} # EOF

#' Extract values from a set of rasters at points
#'
#' @param x `GRaster` or [sources()] name of one or more `GRaster`s.
#' @param y A points `GVector` or [sources()] name of one
#' @param dtype **GRASS** format `datatype` of each raster in `x`. Will be obtained from `x` if `x` is a `GRaster`.
#' @param xNames [names()] of `x`.  Will be obtained from `x` if `x` is a `GRaster`.
#' @param levels [levels()] of a `GRaster`. Will be taken from `x` if `NULL` and `x` is a `GRaster`.
#' @param cats Logical: Extract category labels instead of values
#' @param verbose Logical.
#'
#' @returns A `data.table`.
#'
#' @noRd
.extractFromRasterAtPoints <- function(
    x,
    y,
    xNames = NULL,
    dtype = NULL,
    levels = NULL,
    cats = TRUE,
    verbose = FALSE
) {

    if (inherits(x, "GRaster")) {
        
        nLayers <- nlyr(x)
        dtype <- datatype(x, "GRASS")
        xNames <- names(x)
        xSrc <- sources(x)
        levels <- levels(x)

    } else {

        if (is.null(xNames)) stop("If `x` is the sources() name of a GRaster, `xNames` must be supplied (i.e., not NULL).")
        if (is.null(levels)) stop("If `x` is the sources() name of a GRaster, `levels` must be supplied (i.e., not NULL).")
        if (is.null(dtype)) stop("If `x` is the sources() name of a GRaster, `dtype` must be supplied (i.e., not NULL).")
        nLayers <- length(x)
        xSrc <- x
    }

    if (inherits(y, "GVector")) {
        ySrc <- sources(y)
    } else {
        ySrc <- y
    }

    ### extracting sets of rasters at a time to obviate issues with too-long of a GRASS command string if there are too many rasters
    layersAtATime <- 30L
    sets <- ceiling(nLayers / layersAtATime)
    if (verbose | faster("verbose")) {
        omnibus::say("Extracting values...")
        if (sets > 1) pb <- utils::txtProgressBar(min = 0, max = sets, initial = 0, style = 3, width = 30)
    }

    # # don't know why, but having this here obviates extraction of all NAs on first extract attempt
    # if (inherits(x, "GRaster")) x[[1L]] <- x[[1L]] + 0

    for (set in seq_len(sets)) {

        if ((verbose | faster("verbose")) & sets > 1)  utils::setTxtProgressBar(pb, set)

        index <- (layersAtATime * (set - 1) + 1):min(layersAtATime * set, nLayers)
        xxSrc <- xSrc[index]

        vals <- rgrass::execGRASS(
            cmd = "r.what",
            map = xxSrc,
            points = ySrc,
            null_value = "NA",
            separator = "pipe",
            flags = c(.quiet(), "overwrite"),
            intern = TRUE
        )

        vals <- strsplit(vals, split = "\\|")
        vals <- do.call(rbind, vals)
        vals <- vals[ , 4L:ncol(vals)]
        vals <- data.table::as.data.table(vals)
        names(vals) <- xNames[index]

        suppressWarnings(
            for (i in 1:ncol(vals)) {
                dt <- dtype
                col <- names(vals)[i]
                if (dt[i] == "CELL") {
                    vals[ , (col) := as.integer(get(col)), by = .SD]
                } else {
                    vals[ , (col) := as.numeric(get(col)), by = .SD]
                }
            }
        )

        if (exists("out", inherits = FALSE)) {
            out <- cbind(out, vals)
        } else {
            out <- vals
        }

    } # next set of rasters
    if ((verbose | faster("verbose")) & sets > 1) close(pb)

    # category label instead of value
    levelRows <- sapply(levels, nrow)
    if (cats & any(levelRows > 0)) {

        facts <- which(levelRows > 0)
        for (fact in facts) {

            factName <- xNames[fact]

            levs <- levels[[fact]]
            if (!inherits(levs, "data.table")) levs <- data.table::as.data.table(levs)
            this <- levs[match(out[[fact]], levs[[1L]]), 2L]
            # names(this) <- names(xx)[fact]
            # this <- this[ , lapply(.SD, as.factor)] # convert to factor... do we want this?
            # out[ , (factName) := this[[1L]]]
            out[ , (factName) := this[[1]]]

        }

    } # extracting category labels

    out

}

#' Get levels from a SpatRaster or a file
#'
#' @param x A `SpatRaster`, `data.frame`, `data.table`, empty string, `NULL`, or a file name. CSV, TAB, RDS, and RDA are allowed file types.
#'
#' @returns A list.
#'
#' @noRd
.getLevels <- function(x) {

	if (is.null(x)) {
		levels <- NULL
	} else if (inherits(x, "SpatRaster")) {
		levels <- terra::cats(x)
	} else if (is.character(x)) {
		if (x == "") {
			levels <- list(x)
		} else {
		
			# load from file
		    nc <- nchar(x)
			suffix <- substr(x, nc - 3L, nc)
			
			suffix <- tolower(suffix)
			
			if (suffix %in% c(".csv", ".tab")) {
				levels <- data.table::fread(x, nThread = faster("cores"), na.strings = "<not defined>", showProgress = FALSE)
			} else if (suffix == ".rds") {
				levels <- readRDS(x)
			} else if (suffix == ".rda") {
				levelsObj <- load(x)
				levels <- get(levelsObj)
			} else {

				suffix <- substr(x, nc - 5L, nc)

				suffix <- tolower(suffix)

				if (suffix == ".rdata") {
					levelsObj <- load(levels)
					levels <- get(levelsObj)
				} else {
					stop("Cannot open a file with level data of this type.\n  Supported types include .csv, .tab, .rds, .rda/.rdata (case is ignored).")
				}

			}
		}
	
	} else if (!inherits(x, c("list", "data.frame", "data.table"))) {
		stop("Argument `levels` must be a data.frame, data.table, an empty string, a list of these, OR NULL or a file name.")
	}
	levels

}

#' @name vegIndices
#'
#' @title Table of vegetation indices that can be calculated from remote sensing surface reflectance data using [vegIndex()]. A near-comprehensive table of indices can be found on the [Index Database: A Database for Remote Sensing Indices](https://www.indexdatabase.de).
#'
#' @description A table of vegetation indices that ca be calculated using [vegIndex()]. Columns include:
#' * `index``: Abbreviation of the index.
#' * `definition`: Index name
#' * `R`, `G`, `B`, `NIR`, `channel5`, `channel7`: Whether or not the index uses the red, green, blue, or near-infrared channels, and channels 5 and 7.
#' * `soilLineslope`, `soilIntercept`, `soilNR`: Whether or not the index requires soil line slope, soil intercept, and a soil noise reduction factor.
#'
#' @docType data
#'
#' @format An object of class `data.frame`.
#'
#' @keywords Remote sensing
#'
#' @seealso [vegIndex()]
#'
#' @example man/examples/ex_fastData.r
NULL

#' Replace values of a GRaster
#'
#' @description The `[<-` operator can be used to replace all of the values of a `GRaster`, or specific values depending on the expression in `i`. For example, you could use `rast[] <- 10` to assign 10 to all cells, or `rast[rast > 0] <- 10` to assign all cells with values >0 to 10. You can also use one raster to set values in another, as in `rast1[rast2 > 0] <- 10`.
#'
#' @param x A `GRaster`.
#' @param i Either missing or a conditional statement that resolves to a `GRaster`.
#' @param j Not used
#' @param value A numeric, integer, or logical value, or `NA`. Only a single value can be used.
#'
#' @returns A `GRaster`.
#'
#' @example man/examples/ex_GRaster_GVector_subset_assign.r
#'
#' @name [<-
#' @aliases [<-,GRaster,missing,ANY-method
#' @docType methods
#' @rdname replace_single_square_bracket
#' @exportMethod [<-
setMethod(
	f = "[<-",
	signature = c(x = "GRaster", i = "missing", j = "ANY"),
	definition = function(x, i, j, value) {
	
	i <- NULL
	if (missing(j)) {
		j <- NULL
	} else {
		warning("Argument j will be ignored.")
	}
	
	if (!inherits(value, c("numeric", "integer", "logical"))) stop("Can only assign numeric, integer, or logical values to a raster.")
	if (length(value) != 1L) stop("Cannot assign multiple values to a raster.")
	
	.locationRestore(x)
	.region(x)
	
	if (is.na(value)) {
		value <- "null()"
	} else if (is.logical(value)) {
		value <- as.integer(value)
	}
	
	nLayers <- nlyr(x)
	srcs <- .makeSourceName(x, "raster", nLayers)
	for (count in seq_len(nLayers)) {

		ex <- paste0(srcs[count], " = ", value)

		rgrass::execGRASS(
			cmd = "r.mapcalc",
			expression = ex,
			flags = c(.quiet(), "overwrite")
		)

	} # next raster
	makeGRaster(srcs, names(x))
	
	} # EOF
)

#' @name [<-
#' @aliases [<-,GRaster,GRaster,ANY-method
#' @docType methods
#' @rdname replace_single_square_bracket
#' @exportMethod [<-
setMethod(
	f = "[<-",
	signature = c(x = "GRaster", i = "GRaster", j = "ANY"),
	definition = function(x, i, j, value) {
	
	if (missing(j)) {
		j <- NULL
	} else {
		warning("Argument j will be ignored.")
	}
	
	if (!inherits(value, c("numeric", "integer", "logical"))) stop("Can only assign numeric, integer, or logical values to a raster.")
	if (length(value) != 1L) stop("Cannot assign multiple values to a raster.")
	
	.locationRestore(x)
	.region(x)
	
	if (is.na(value)) {
		value <- "null()"
	} else if (is.logical(value)) {
		value <- as.integer(value)
	}
	
	nLayers <- nlyr(x)
	nLayersI <- nlyr(i)

	if (nLayersI != nLayers) {

		if (nlyr(i) != 1L) warning("The number of layers in the subset raster (inside the `[]`) does not have\n  the same number of layers as the raster to be subset. The layers inside\n  the square brackets will be recycled.")

		iIndex <- rep(seq_len(nLayersI), length.out = nLayers)
		i <- i[[iIndex]]

	}

	srcs <- .makeSourceName(x, "raster", nLayers)
	for (count in seq_len(nLayers)) {

		if (inherits(i, "GRaster")) {
			if (!(.minVal(i)[count] %in% c(NA, 0, 1)) & !(.maxVal(i)[count] %in% c(NA, 0, 1))) stop("The GRaster in `i` must 0, 1, or NA values.")
			ex <- paste0(srcs[count], " = if(", sources(i)[count], " == 1, ", value, ", ", sources(x)[count], ")")
		} else {
			ex <- paste0(srcs[count], " = ", value)
		}

		rgrass::execGRASS(
			cmd = "r.mapcalc",
			expression = ex,
			flags = c(.quiet(), "overwrite")
		)

	} # next raster
	makeGRaster(srcs, names(x))
	
	} # EOF
)

#' Create a raster with random values drawn from a uniform distribution
#'
#' @description `rUnifRast()` creates a raster with values drawn from a uniform (flat) distribution.
#'
#' @param x A `GRaster`. The output will have the same extent and dimensions as this raster.
#'
#' @param n A numeric integer: Number of rasters to generate. 
#'
#' @param low,high Numeric: Minimum and maximum values from which to select.
#'
#' @param seed Numeric integer vector or `NULL`: Random seed. If `NULL`, then a different seed will be generated by **GRASS**. Defining this is useful if you want to recreate rasters.  If provided, there should be one seed value per raster.
#'
#' @returns A `GRaster`.
#'
#' @example man/examples/ex_randRast.r
#' 
#' @seealso [rNormRast()], [rSpatialDepRast()], [fractalRast()], [rWalkRast()], **GRASS** manual page for tool `r.random.surface` (see `grassHelp("r.random.surface")`)
#'
#' @aliases rUnifRast
#' @rdname rUnifRast
#' @exportMethod rUnifRast
methods::setMethod(
    f = "rUnifRast",
    signature = c(x = "GRaster"),
    function(
		x,
        n = 1,
        low = 0,
        high = 1,
        seed = NULL
	) {

    if (!is.null(seed)) if (length(seed) != n) stop("You must provide one value of `seed` per raster, or set it to NULL.")

    .locationRestore(x)
    .region(x)

    srcs <- .makeSourceName("runif", "raster", n)

    for (i in seq_len(n)) {

		# # have to add/subtract a fractional number to avoid producing a rounded (CELL) raster
		# num <- 1 / sqrt(pi * exp(1))

		# ex <- paste0(srcs[i], " = float(rand(", low, " + ", num, ", ", high, " + ", num, ") - ", num, ")")
		ex <- paste0(srcs[i], " = float(rand(", low, " + 0.1, ", high, " + 0.1) - 0.1)")
        args <- list(
            cmd = "r.mapcalc",
            expression = ex,
            flags = c(.quiet(), "overwrite"),
            intern = TRUE
        )

        if (!is.null(seed)) {
			args$seed <- seed[i]
		} else {
			args$flags <- c(args$flags, "s")
		}
        do.call(rgrass::execGRASS, args = args)

    } # next raster
    makeGRaster(srcs, rep("runif", n))

    } # EOF
)

#' Save a GRaster to disk
#'
#' @description
#' This function saves a `GRaster` to disk directly from a **GRASS** session. It is faster than using [rast()], then saving the output of that to disk (because `rast()` actually save the raster to disk, anyway).
#'
#' The function will attempt to ascertain the file type to be ascertained from the file extension, but you can specify the format using the `format` argument (see entry for `...`). You can see a list of supported formats by simply using this function with no arguments, as in `writeRaster()`, or by consulting the online help page for the **GRASS** tool `r.out.gdal` (see `grassHelp("r.out.gdal")`). Only the `GeoTIFF` file format is guaranteed to work for multi-layered rasters.
#'
#' The function will attempt to optimize the `datatype` argument, but this can take a long time. You can speed this up by setting `datatype` manually. Note that if you are saving a "stack" of `GRaster`s with different `datatype`s, the one with the highest information density will be used (e.g., low-bit integer < high-bit integer < floating-point < double-floating point). This can make rasters with lower datatypes much larger on disk. In these cases, it make be best to save rasters with similar `datatype`s together.
#'
#' @param x A `GRaster` or missing: If missing, a table of supported file types is reported.
#' @param filename Character: Path and file name.
#' @param overwrite Logical: If `FALSE` (default), do not save over existing file(s).
#' @param datatype `NULL` (default) or character: The datatype of the values stored in non-ASCII rasters. If `NULL`, this will be ascertained from the raster, and the function usually does a good job at it. However, you can force it manually, but note that in some cases, trying to save a `GRaster` using an inappropriate `datatype` for its values can result in an error or in the function exiting without an error but also without having written the raster to disk. The argument can take any of those shown below under the first four columns, but whatever is used, it will be converted to the **GDAL** version.
#'
#'    | **fasterRaster**   | **terra**   | **GRASS**   | **GDAL**   | **Values** |
#'    | ------------------ | ----------- | ----------- | ---------- | ------ |
#'    | `integer`          | `INT1U`     | `CELL`      | `Byte`     | Integer values from 0 to 255 |
#'    | `integer`          | `INT2U`     | `CELL`      | `UInt16`   | Integer values from 0 to 65,534 |
#'    | `integer`          | `INT2S`     | `CELL`      | `Int16`    | Integer values from -32,767 to -32,767 |
#'    | `integer`          | `INT4S`     | `CELL`      | `Int32`    | Integer values from -2,147,483,647 to 2,147,483,647 |
#'    | `float`            | `FLT4S`     | `FCELL`     | `Float32`  | Values from -3.4E+38 to 3.4E+38, including decimal values |
#'    | `double`           | `FLT8S`     | `DCELL`     | `Float64`  | Values from -1.79E+308 to 1.79E+308, including decimal values |
#'    | `factor`           | `INT`*      | `CELL`      | `INT*`     | Integer values corresponding to categories
#'
#' `*` Depends on the integers (signed/unsigned, range of values). Categorical rasters will have an associated file saved with them that has category values and labels. The file name will be the same as the raster's file name, but end with the extension given by `levelsExt` (`.csv` by default).
#'
#' @param byLayer Logical: If `FALSE` (default), multi-layer rasters will be saved in one file. If `TRUE`, the each layer will be saved in a separate file. The filename from `filename` will be amended so that it ends with `_<name>` (then the file extension), where `<name>` is give by [names()]. Note that if any characters in raster names will not work in a file name, then the function will fail (e.g., a backslash or question mark).
#'
#' @param names Logical: If `TRUE` (default), save a file with raster layer names. The file will have the same name as the raster file but end with "`_names.csv`". Currently, the [names()] attribute of rasters cannot be saved in the raster, which can create confusion when multi-layered rasters are saved. Turning on this option will save the ancillary file with layer names. If it exists, this file will be read by [fast()] so layer names are assigned when the raster is read by that function. The absence of a "names" file will not create any issues with this function or [fast()], other than not having the metadata on layer names.
#' 
#' @param levelsExt Character, logical, or `NULL` (default): Name of the file extension for the "levels" file that accompanies a categorical `GRaster`. When saving categorical rasters, the raster file is accompanied with a "levels" file that contain information on the levels of the raster. This file is the same as `filename`, except it has a different extension. Valid values depend on how many raster layers are saved at a time (case is ignored):
#' * DefaultOne raster layer: `".csv"`
#' * Two or more layers, with at least one categorical raster: `".rds"`, `".rda"`, `".rdat"`, `".rdata"`
#' * Any: `NULL` or `TRUE` automatically selects either `".csv"` (one raster layer) or `".rds` (two or more)
#' * Any: `FALSE` disables saving of a levels file.
#'
#' @param compress Character: Type of compression to use for GeoTIFF files:
#'    * `"LZW"` (default)
#'    * `"DEFLATE"`
#'    * `"PACKBITS"`
#'    * `"LZMA"`
#'    * `NULL`: No compression is used, but the file can still be reduced in size by using zip, gzip, or other compressions.
#'
#' @param warn Logical: If `TRUE` (default), display a warning if the `datatype` argument does not match the value given by `datatype(x, "GDAL")`, or if the `fileExt` argument will not work with the given raster and so has been automatically changed.
#'
#' @param ... Additional arguments. These can include:
#' * `bigTiff`: Logical: If `TRUE`, and the file format is a GeoTIFF and would be larger than 4 GB (regardless of compression), then the file will be saved in BIGTIFF format.
#' * `format`: Character, indicating file format. This is usually ascertained from the file extension, but in case this fails, it can be stated explicitly. When using other formats, you may have to specify the `createopts` argument, too (see help page for **GRASS** tool `r.out.gdal`). Two common formats include:
#'    * `"GTiff"` (default): GeoTIFF `filename` ends in `.tif`.
#'    * `"ASC"`: ASCII `filename` ends in `.asc`
#' * Additional arguments to send to **GRASS** modules `r.out.gdal` and `r.out.ascii`.
#' * `precision`: Numeric: For ASCII files, you may need to state the number of significant digits. 32-bit values have 7 digits and 64-bit values have 16. So in these cases the argument would be `precision=7` or `precision=16`.
#'
#' @returns A `GRaster` (invisibly). A raster is also saved to disk.
#'
#' @seealso [terra::writeRaster()], **GRASS** tool `r.out.gdal` (see `grassHelp("r.out.gdal")`)
#'
#' @example man/examples/ex_writeRaster.r
#'
#' @aliases writeRaster
#' @rdname writeRaster
#' @exportMethod writeRaster
setMethod(
	"writeRaster",
	signature(x = "GRaster", filename = "character"),
	function(
		x,
		filename,
		overwrite = FALSE,
		datatype = NULL,
		byLayer = FALSE,
		names = TRUE,
		levelsExt = NULL,
		compress = "LZW",
		warn = TRUE,
		...
	) {
	
	if (!is.null(levelsExt) & !is.logical(levelsExt)) {

		levelsExts <- c(".csv", ".rds", ".rda", ".rdata", "csv", "rds", "rda", "rdata", "rdat")
		levelsExt <- omnibus::pmatchSafe(levelsExt, levelsExts, nmax = 1L)
		if (substr(levelsExt, 1L, 1L) != ".") levelsExt <- paste0(".", levelsExt)
		
	}

	.locationRestore(x)
	.region(x)

	nLayers <- nlyr(x)

	# save each layer separately
	if (byLayer) {
	
		for (i in seq_len(nLayers)) {
		
			xx <- x[[i]]
			extension <- .fileExt(filename)
			fn <- substr(filename, 1L, nchar(filename) - nchar(extension) - 1L)
			fn <- paste0(fn, "_", names(xx), ".", extension)
			writeRaster(xx, filename = fn, overwrite = overwrite, datatype = datatype, byLayer = FALSE, names = names, levelsExt = levelsExt, compress = compress, warn = warn, ...)
		
		}
	
	# save all layers in a single file
	} else {

		dots <- list(...)
		filename <- trimws(filename)

		flags <- c(.quiet())
		if (overwrite) flags <- c(flags, "overwrite")

		filename <- normalizePath(filename, mustWork = FALSE)

		### going to overwrite anything?
		if (!overwrite) {
			if (file.exists(filename)) stop("File already exists and `overwrite` is FALSE:\n  ", filename)
		}

		### format
		extension <- .fileExt(filename)
		extension <- tolower(extension)
		
		ascii <- if ("format" %in% names(dots)) {
			if (("format" %in% names(dots) && tolower(dots$format) %in% c("asc", "asci", "ascii")) || extension %in% c("asc", "asci", "ascii")) { TRUE } else { FALSE }
		} else {
			FALSE
		}

		geotiff <- ("format" %in% names(dots) && tolower(dots$format) == "gtiff") | extension == "tif" | extension == "tiff"

		### save
		if (ascii) {
			
			if (nlyr(x) > 1L) stop("Cannot save multi-layer GRaster as a single ASCII file. Save each layer individually.")
			
			rgrass::execGRASS(
				dmc = "r.out.ascii",
				input = x,
				output = filename,
				flags = flags,
				...
			)

			if (names) .saveNames(x = x, filename = filename)

		} else {

			thisFlags <- c(flags, "c")

			## if multi-layered raster stack, then group first... only guaranteed to work with GeoTIFFs
			if (nLayers > 1L) {

				srcGroup <- .makeSourceName("i_group", type = "group")
				input <- sources(x)
				
				rgrass::execGRASS(
					cmd = "i.group",
					group = srcGroup,
					input = input,
					flags = .quiet()
				)

				src <- srcGroup

			} else {
				src <- sources(x)
			}
			
			# data type
			if (is.null(datatype)) datatype <- datatype(x, "GRASS")

			bounds <- minmax(x)
			mins <- bounds["min", ]
			maxs <- bounds["max", ]

			if (any(datatype %in% c("double", "FLT8S", "DCELL"))) {

				datatype <- "Float64"

			} else if (any(datatype %in% c("float", "FLT4S", "FCELL"))) {

				datatype <- "Float32"

			} else if (any(datatype %in% c("factor", "integer", "logical", "CELL"))) {

				if (all(mins >= 0L & maxs <= 255L)) {
					datatype <- "Byte"
				} else if (all(mins >= 0L & maxs <= 65534L)) {
					datatype <- "UInt16"
				} else if (any(mins < 0L) & all(maxs >= -32767L) & all(maxs <= 32767L)) {
					datatype <- "Int16"
				} else if (any(mins < 0L) & all(maxs >= -2147483647L) & all(maxs <= -2147483647)) {
					datatype <- "Int32"
				# } else if (all(mins >= -3.4E+38) & all(maxs <= 3.4E+38)) {
					# datatype <- "Float32" # causes r.out.gdal not to write some DCELL rasters
				} else {
					datatype <- "Float64"
				}

			} else {
			
				# if (all(mins >= -3.4E+38) & all(maxs <= 3.4E+38)) {
					# datatype <- "Float32"
				# } else {
					datatype <- "Float64"
				# }
			
			}
			
			if (any(datatype == "INT1U")) datatype[datatype == "INT1U"] <- "Byte" # will not write, for some reason
			# if (any(datatype == "INT1U")) datatype[datatype == "INT1U"] <- "UInt16"
			if (any(datatype == "INT2U")) datatype[datatype == "INT2U"] <- "UInt16"
			if (any(datatype == "INT2S")) datatype[datatype == "INT2S"] <- "Int32"
			if (any(datatype == "FLT4S")) datatype[datatype == "FLT4S"] <- "Float32"
			if (any(datatype == "FLT8S")) datatype[datatype == "FLT8S"] <- "Float64"
			
			if (length(datatype) > 1L) {
				datatype <- if (any(datatype == "Float64")) {
					"Float64"
				} else if (any(datatype == "Float32")) {
					"Float32"
				} else if (any(datatype == "Int32")) {
					"Int32"
				} else if (any(datatype == "UInt16")) {
					"UInt16"
				} else if (any(datatype == "Byte")) {
					"Byte" # Will not write, for some reason
					# "UInt16"
				}
				
			}

			if (any(datatype(x, "fasterRaster") %in% c("float", "double")) & datatype %in% c("Byte", "UInt16", "Int32")) {

				stop("Trying to save non-integer rasters with an integer datatype. You can:\n  * Change the datatype of the raster to integer using as.int(), trunc(), round(), floor(), or ceiling();\n  * Change the raster to a non-integer type using as.float() or as.doub();\n  * When using writeRaster(), set argument `datatype` to `FLT4S` or `FLT8S`, depending on their minimum and maximum values.")
				
			}

			# if (any(!(datatype(x, "GDAL") %in% datatype))) {
			# 	flags <- c(flags, "f")
			# 	if (warn) warning("Argument `datatype` does not match the data type of the raster. Data may be lost.")
			# }

			if (!("createopt" %in% names(dots))) createopt <- NULL

			# GeoTIFF options
			if (geotiff) {

				createopt <- c(createopt, "PROFILE=GeoTIFF")
				if (!is.null(compress)) {
					compress <- omnibus::pmatchSafe(compress, c("LZW", "DEFLATE", "PACKBITS", "LZMA"), nmax = 1L)
					createopt <- c(createopt, paste0("COMPRESS=", toupper(compress)))
				}
				if ("bigTiff" %in% names(dots) && dots$bigTiff) {
					createopt <- c(createopt, "BIGTIFF=YES")
				} else {
					createopt <- c(createopt, "BIGTIFF=IF_NEEDED")
				}
				if (datatype %in% c("Byte", "UInt16", "Int32")) createopt <- c(createopt, "PREDICTOR=2")
				if (datatype %in% c("Float32", "Float64")) createopt <- c(createopt, "PREDICTOR=3")
				
				createopt <- unique(createopt)
				createopt <- paste(createopt, collapse=",")

				# mm <- minmax(x)
				# metaopt <- paste0("TIFFTAG_MINSAMPLEVALUE=", paste(mm[1L, ], collapse=","))
				# metaopt <- c(metaopt, paste0("TIFFTAG_MAXSAMPLEVALUE=", paste(mm[2L, ], collapse=",")))
				# metaopt <- paste0("STATISTICS_MINIMUM=", paste(mm[1L, ], collapse=","))
				# metaopt <- c(metaopt, paste0("STATISTICS_MAXIMUM=", paste(mm[2L, ], collapse=",")))
				# metaopt <- unique(metaopt)
				# metaopt <- paste(metaopt, collapse=",")
				
			}

			# save
			rgrass::execGRASS(
				cmd = "r.out.gdal",
				input = src,
				output = filename,
				type = datatype,
				format = "GTiff",
				createopt = createopt,
				# metaopt = 'TIFFTAG_IMAGEDESCRIPTION="TEST',
				flags = thisFlags
			)

			if (names) .saveNames(x = x, filename = filename)

		}
		
		isFact <- is.factor(x)
		if (any(isFact)) {

			if (is.logical(levelsExt)) {
				if (levelsExt) {
					levelsExt <- NULL
					saveLevels <- TRUE
				} else {
					saveLevels <- FALSE
				}
			} else {
				saveLevels <- TRUE
			}

			if (saveLevels) {

				categs <- cats(x)
				if (is.null(levelsExt)) {
					if (length(categs) > 1L) {
						levelsExt <- ".rds"
					} else {
						levelsExt <- ".csv"
					}
				}
				
				if (length(categs) > 1L & tolower(levelsExt) == ".csv") {

					if (warn) warning("You cannot save levels files of multi-layered rasters using a `levelsExt` value of `.csv`.\n  The file extension has been changed to `.rds`, which saves files that can be read using `readRDS()`.")

					levelsExt <- ".rds"

				}

				levelFileName <- substr(filename, 1L, nchar(filename) - nchar(levelsExt))
				levelFileName <- paste0(levelFileName, levelsExt)
				
				# save
				if (tolower(levelsExt) %in% ".rds") {
					saveRDS(categs, levelFileName)
				} else if (tolower(levelsExt) %in% c(".rds", ".rda", ".rdat", ".rdata")) {
					save(categs, levelFileName)
				} else if (tolower(levelsExt) %in% ".csv") {
					utils::write.csv(categs, levelFileName, row.names = FALSE)
				}

			} # "yes" to saving levels

		} # save levels

		# out <- terra::rast(filename)
		# names(out) <- names(x)
		
		# if (any(isFact)) {
		# 	for (i in which(isFact)) {
		# 		cats <- as.data.frame(cats(x)[[i]])
		# 		out <- terra::categories(out, layer = i, value = cats, active = activeCat(x, layer = i))
		# 	}
		# }
		
		# if (mm) out <- terra::setMinMax(out)
		# invisible(out)

	} # save all layers in a single file
	
	invisible(x)
	
	} # EOF
	
)

#' @aliases writeRaster
#' @rdname writeRaster
#' @export
#' @exportMethod writeRaster
setMethod(
	"writeRaster",
	signature(x = "missing", filename = "missing"),
	function(x, filename) {
	
	forms <- rgrass::execGRASS(
		cmd = "r.out.gdal",
		flags = "l",
		intern = TRUE
	)
	
	forms <- forms[forms != "Supported formats:"]
	forms <- trimws(forms)
	forms <- sort(forms)
	formsHeader <- c("Supported raster file formats:", forms)
	cat(paste(formsHeader, collapse = "\n"))
	cat("\n")
	utils::flush.console()
	invisible(forms)
	
	} # EOF
)

### save "names" file for each layer of a raster
.saveNames <- function(x, filename) {

	namesFile <- filename
	n <- nchar(filename)
	ne <- nchar(.fileExt(filename))
	namesFile <- substr(namesFile, 1L, n - ne - 1L)
	namesFile <- paste0(namesFile, '_names.csv')

	namesDf <- data.table::data.table(
		layer = 1:nlyr(x),
		name = names(x)
	)

	data.table::fwrite(namesDf, namesFile, row.names = FALSE)

}

#' Returns .quiet() or NULL for "flags" argument to GRASS modules
#'
#' @description A function for developers used for setting the "quiet' argument in `flags` arguments passed to [rgrass::execGRASS()]. If `faster("debug")` is `TRUE`, the string "quiet" is returned. If `FALSE`, then `NULL` is returned.
#'
#' @returns A string (.quiet()) or `NULL`.
#'
#' @aliases .quiet
#' @rdname quiet
#' @keywords internal
.quiet <- function() {

	if (faster("debug")) {
		out <- 'verbose'
	} else {
		out <- "quiet"
	}
	out

}

#' Return first or last part of the data frame of a GVector
#'
#' @description Return the first or last part of a `GVector`'s data table.
#'
#' @param x A `GVector`.
#' @param n Integer: Number of rows to display.
#' @param keepnums Logical: If no `rownames` are present, create them. Default is `TRUE`.
#' @param ... Other arguments.
#'
#' @returns A `data.table` or `data.frame`.
#' 
#' @seealso [terra::head()], [terra::tail()]
#'
#' @example man/examples/ex_GVector.r
#'
#' @aliases head
#' @rdname head
#' @exportMethod head
methods::setMethod(
	f = "head",
	signature = c(x = "GVector"),
	definition = function(x, n = 6L, keepnums = TRUE, ...) {

		nr <- nrow(x)
		nr <- min(n, nr)
		out <- x@table[1L:nr]
		if (!faster("useDataTable")) out <- as.data.frame(out)
		out

	} # EOF
)

#' @aliases tail
#' @rdname head
#' @exportMethod tail
methods::setMethod(
    f = "tail",
    signature = c(x = "GVector"),
    definition = function(x, n = 6L, keepnums = TRUE, ...) {

	nr <- nrow(x)
	nfrom <- max(1L, nr - n + 1L)
	out <- x@table[nfrom:nr]
	if (!faster("useDataTable")) out <- as.data.frame(out)
	out

    } # EOF
)

#' Open the help page for a GRASS tool
#'
#' @description This function opens the manual page for a **GRASS** tool (function) in your browser.
#'
#' @param x Character: Any of:
#' * The name of a **GRASS** tool (e.g., `"r.mapcalc"`).
#' * `"toc"`: **GRASS** manual table of contents.
#' * `"index"`: Display an index of topics.
#'
#' @param online Logical: If `FALSE` (default), show the manual page that was included with your installation of **GRASS** on your computer.  If `FALSE`, show the manual page online (requires an Internet connection). In either case, the manual page will display for the version of **GRASS** you have installed.
#'
#' @returns Nothing (opens a web page).
#'
#' @example man/examples/ex_grassHelp.r
#'
#' @aliases grassHelp
#' @rdname grassHelp
#' @export
grassHelp <- function(x, online = FALSE) {

	if (interactive()) {

		if (!grassStarted()) {
			omnibus::say("You must start GRASS first by using fast() at least once before opening a manual page.")
			return()
		}

		x <- tolower(x)
		if (length(x) != 1L) stop("Only one help page can be opened at a time.")
		
		args <- list(
			cmd = "g.manual",
			entry = x,
			flags = .quiet()
		)
		
		if (x == "index") {
			args$flags <- c(args$flags, "t")
		} else if (x == "toc") {
			args$flags <- c(args$flags, "i")
		}

		do.call(rgrass::execGRASS, args = args)

	} else {
		warning("You can only open GRASS manual pages in an interactive R session.")
	}

}

#' Path of water flow across a landscape
#'
#' @description This function finds the least-cost pathway from a set of starting points to the lowest cells accessible from them while, in each step, traversing "down" slope gradients. It is intended to depict the path a drop of water would take when flowing across a landscape. For a single starting point, the defaults settings will produce a raster with cells with values of 1 along the path. All other cells will be set to `NA`.
#'
#' @param x A `GRaster` with a single layer, typically representing elevation.
#'
#' @param y A "points" `GVector`. The `GVector` must have <= 1024 points.
#'
#' @param return Character: Indicates the type of values "burned" into the cells of the output raster. Case is ignored and partial matching is used, but only one option can be selected.
#' * "`ID`" (default): Cells in each path are labeled with the index of the starting point. A cell in the flow path of the first point will have a value of 1, a cell in the flow path of the second point will have a value of 2, and so on.
#' * "`sequence`": The output raster's cells will start with 1 at the source point(s), then accumulate so that the next cell in the flow path is 2, the one after that 3, and so on.
#' * "`copy`": The cells in the flow path will have the elevation raster's values in the cells along the flow path(s).
#' * "`accumulation`": Cells in the flow path will accumulate the elevation raster's cell values. For example, if the starting cell has an elevation of 700 and the next cell in the drainage path has a value of 600 and the one after that 500, then the first cell in the path will have a value of 700, the next 1300 (= 700 + 600), and the third 1800 (= 700 + 600 + 500).
#' * A numeric value: All cells in flow paths will be assigned this value.
#'
#' @returns A `GRaster`.
#'
#' @example man/examples/ex_flowPath.r
#'
#' @seealso [flow()], [streams()], the **GRASS** tool `r.drain` (see `grassHelp("r.drain")`)
#'
#' @aliases flowPath
#' @rdname flowPath
#' @exportMethod flowPath
methods::setMethod(
	f = "flowPath",
	signature = c(x = "GRaster"),
	function(x, y, return = "ID") {
	
	if (nlyr(x) > 1L) stop("This function can only use a single-layered GRaster as input.")
	if (geomtype(y) != "points") stop("The GVector must represent points.")
	if (ngeom(y) > 1024L) stop("The GVector can only have up to 1024 points in it.")

	if (!is.numeric(return)) {
		returns <- c("ID", "sequence", "copy", "accumulation")
		return <- omnibus::pmatchSafe(return, returns, nmax = 1L)
	}

	.locationRestore(x)
	.region(x)

	ySrc <- .copyGSpatial(y)
	ySrc <- .vRecat(ySrc, gtype = "point")

	src <- .makeSourceName("flowPath_r_drain", type = "raster")
	args <- list(
		cmd = "r.drain",
		input = sources(x),
		output = src,
		start_points = ySrc,
		flags = c(.quiet(), "overwrite")
	)

	if (return == "copy") {
		args$flags <- c(args$flags, "c")
	} else if (return == "accumulation") {
		args$flags <- c(args$flags, "a")
	} else if (return == "sequence") {
		args$flags <- c(args$flags, "n")
	}

	do.call(rgrass::execGRASS, args = args)

	# force output to a numeric value
	if (is.numeric(return)) {

		srcIn <- src
		src <- .makeSourceName("flowPath_r_mapcalc", "raster")
		ex <- paste0(src , " = ", return, " + (", srcIn, " * 0)")
		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

	}

	makeGRaster(src, names = "flowPath")

	} # EOF
)

#' Apply a function to a set of rasters
#'
#' @description `app()` applies a function to a set of "stacked" rasters. It is similar to the [terra::app()] and [terra::lapp()] functions.
#'
#' `appFuns()` provides a table of **GRASS** functions that can be used by `app()` and their equivalents in **R**.
#'
#' `appCheck()` tests whether a formula supplied to `app()` has any "forbidden" function calls.
#' 
#' The `app()` function operates in a manner somewhat different from [terra::app()]. The function to be applied *must* be written as a character string. For example, if the `GRaster` had layer names "`x1`" and "`x2`", then the function might be like `"= max(sqrt(x1), log(x2))"`. Rasters **cannot** have the same names as functions used in the formula. In this example, the rasters could not be named "max", "sqrt", or "log". Note that the name of a `GRaster` is given by [names()]--this can be different from the name of the object in **R**.
#' 
#' The `app()` function will automatically check for `GRaster` names that appear also to be functions that appear in the formula. However, you can check a formula before running `app()` by using the `appCheck()` function. You can obtain a list of `app()` functions using `appFuns()`. Note that these are sometimes different from how they are applied in **R**.
#'
#' Tips:
#' * Make sure your `GRaster`s have `names()`. The function matches on these, not the name of the variable you use in **R** for the `GRaster`.
#' * Use `null()` instead of `NA`, and use `isnull()` instead of `is.na()`.
#' * If you want to calculate values using while ignoring `NA` (or `null`) values, see the functions that begin with `n` (like `nmean`).
#' * Be mindful of the data type that a function returns. In **GRASS**, these are `CELL` (integer), `FCELL` (floating point values--precise to about the 7th decimal place), and `DCELL` (double-floating point values--precise to about the 15th decimal place; commensurate with the **R** `numeric` type). In cases where you want a `GRaster` to be treated like a float or double type raster, wrap the name of the `GRaster` in the `float()` or `double()` functions. This is especially useful if the `GRaster` might be assumed to be the `CELL` type because it only contains integer values. You can get the data type of a raster using [datatype()] with the `type` argument set to `GRASS`. You can change the data type of a `GRaster` using [as.int()], [as.float()], and [as.doub()]. Note that categorical rasters are really `CELL` (integer) rasters with an associated "levels" table. You can also change a `CELL` raster to a `FCELL` raster by adding then subtracting a decimal value, as in `x - 0.1 + 0.1`. See `vignette("GRasters", package = "fasterRaster")`.
#' * The `rand()` function returns integer values by default. If you want non-integer values, use the tricks mentioned above to datatype non-integer values. For example, if you want uniform random values in the range between 0 and 1, use something like `= float(rand(0 + 0.1, 1 + 0.1) - 0.1)`.
#' * `GRaster`s with short names can cause the function to mis-specify the formula in **GRASS**. This happens when, for example, the `fun` argument contains an "if" statement, and the raster name is "i" or "f", which would incorrectly replace those letters in the "if" with the **GRASS** name of the raster. To avoid this, avoid using short raster names that could be part of longer operators using by `app()` (see [appFuns()] for a list of these operators).
#'
#' @param x A `GRaster` with one or more named layers.
#'
#' @param fun Character: The function to apply. This must be written as a character string that follows these rules:
#' 
#' * It must use typical arithmetic operators like `+`, `-`, `*`, `/` and/or functions that can be seen using `appFuns(TRUE)`.
#' * The [names()] of the rasters do not match any of the functions in the `appFuns(TRUE)` table. Note that `x` and `y` are forbidden names :(
#'
#' The help page for **GRASS** tool `r.mapcalc` will be especially helpful. You can see this page using `grassHelp("r.mapcalc")`.
#' 
#' @param datatype Character: This ensures that rasters are treated as a certain type before they are operated on. This is useful when using rasters that have all integer values, which **GRASS** can assume represent integers, even if they are not supposed to. In this case, the output of operations on this raster might be an integer if otherwise not corrected. Partial matching is used, and options include:
#' * `"integer"`: Force all rasters to integers by truncating their values. The output may still be of type `float` if the operation creates non-integer values.
#' * `"float"`: Force rasters to be considered floating-point values.
#' * `"double"`: Force rasters to be considered double-floating point values.
#' * `"auto"` (default): Ensure that rasters are represented by their native [datatype()] (i.e., "CELL" rasters as integers, "FCELL" rasters as floating-point, and "DCELL" as double-floating point).
#'
#' @param seed Numeric integer vector or `NULL` (default): A number for the random seed. Used only for `app()` function `rand()`, that generates a random number. If `NULL`, a seed will be generated. Defining the seed is useful for replicating a raster made with `rand()`. This must be an integer!
#' 
#' @param warn Logical (function `appFuns()`): If `TRUE` (default), display a warning when `allFuns()` is not called interactively.
#'
#' @param msgOnGood Logical (function `appCheck()`): If `TRUE` (default), display a message if no overt problems with the raster names and formula are detected.
#' 
#' @param failOnBad Logical (function `appCheck()`): If `TRUE` (default), fail if overt problems with raster names and the formula are detected.
#'
#' @returns A `GRaster`.
#'
#' @seealso [terra::app()], [terra::lapp()], [subst()], [classify()], and especially the **GRASS** manual page for tool `r.mapcalc` (see `grassHelp("r.mapcalc")`)
#'
#' @example man/examples/ex_app.r
#'
#' @aliases app
#' @rdname app
#' @exportMethod app
methods::setMethod(
    f = "app",
    signature = c(x = "GRaster"),
    function(x, fun, datatype = "auto", seed = NULL) {

    fun <- trimws(fun)
    if (substr(fun, 1L, 2L) != "=") fun <- paste0("= ", substr(fun, 2L, nchar(fun)))
    if (substr(fun, 2L, 2L) != " ") fun <- paste0("= ", substr(fun, 2L, nchar(fun)))
    if (substr(fun, 1L, 2L) == "==") stop("Argument `fun` cannot begin with a double equals sign.")

    appCheck(x, fun, msgOnGood = FALSE, failOnBad = TRUE)

    .locationRestore(x)
    .region(x)

    # replace raster names with sources
    # replacing from longest to shortest to avoid issues with nestedness

    datatype <- omnibus::pmatchSafe(datatype, c("integer", "float", "double", "auto"))
    if (datatype == "integer") datatype = "int"

    # table with GRASS name, R name, and the "stub" string that will temporarily represent it in "fun"
    # later, we replace all values of "rname" with the corresponding values of "datatypeSrc"
    xwalk <- data.table(
        src = sources(x),
        datatypeSrc = NA_character_,
        rname = names(x),
        stub = NA_character_
    )
    xwalk <- xwalk[rev(order(nchar(xwalk$rname)))]

    # datatype
    dt <- datatype(x, "GRASS")
    if (datatype == "auto") {
        srcs <- if ("DCELL" %in% dt) {
            paste0("double(", xwalk$src, ")")
        } else if ("FCELL" %in% dt) {
            paste0("float(", xwalk$src, ")")
        } else if ("CELL" %in% dt) {
            paste0("int(", xwalk$src, ")")
        } else {
            stop('Unknown datatype.')
        }
    } else {
        if (!any("FCELL" %in% dt)) {
            srcs <- paste0(datatype, "(", xwalk$src, ")")
        }
    }

    xwalk$datatypeSrc <- srcs

    for (i in 1:nrow(xwalk)) {

        rname <- xwalk$rname[i]
        stub <- paste0('#', i, '#')
        xwalk$stub[i] <- stub

        fun <- gsub(fun, pattern = rname, replacement = stub, fixed = TRUE)

    }

    for (i in 1:nrow(xwalk)) {
        fun <- gsub(x = fun, pattern = xwalk$stub[i], replacement = xwalk$datatypeSrc[i], fixed = TRUE)
    }

    src <- .makeSourceName("app", "raster")
    fun <- paste(src, fun)

    args <- list(
        cmd = "r.mapcalc",
        expression = fun,
        flags = c(.quiet(), "overwrite"),
        intern = TRUE
    )
    if (is.null(seed)) {
		args$flags <- c(args$flags, "s")
	} else {
		args$seed <- seed
	}
    do.call(rgrass::execGRASS, args = args)
    makeGRaster(src, "layer")

    } # EOF
)

#' @aliases appFuns
#' @rdname app
#' @export
appFuns <- function(warn = TRUE) {
	
	appFunsTable <- NULL
    utils::data("appFunsTable", envir = environment(), package = "fasterRaster")

	if (interactive()) {
	
    	showableCols <- c("Type", "GRASS_Function", "R_Function", "Definition", "Returns")

		shiny::shinyApp(
			ui = shiny::fluidPage(DT::DTOutput("tbl")),
			server = function(input, output) {
				output$tbl <- DT::renderDT(
					# appFunsTable[ , showableCols],
					appFunsTable[ , c("Type", "GRASS_Function", "R_Function", "Definition", "Returns")],
					caption = shiny::HTML("Functions that can be used in the <b>fasterRaster</b> app() function and their equivalents in <b>R</b>.<br/>Note that in GRASS, 'null()' is the same as 'NA'."),
					options = list(
						pageLength = nrow(appFunsTable),
						width = "100%",
						scrollX = TRUE
					)
					# options = list(lengthChange = FALSE)
				)
			}
		)
        
    } else if (warn) {
        warning("You must be running R interactively to view the table using appFuns().")
    }

}

#' @aliases appCheck
#' @rdname app
#' @exportMethod appCheck
methods::setMethod(
    f = "appCheck",
    signature = c(x = "GRaster", fun = "character"),
    function(x, fun, msgOnGood = TRUE, failOnBad = TRUE) {

    # any forbidden names in rasters?
    ns <- names(x)
    
	appFunsTable <- NULL
    utils::data("appFunsTable", envir = environment(), package = "fasterRaster")

    if (inherits(appFunsTable, "data.table")) {
        funs <- appFunsTable[["GRASS_Function"]]
    } else {
        funs <- appFunsTable$GRASS_Function
    }

    parens <- regexec(funs, pattern = "\\(")
    parens <- unlist(parens)
    parens <- parens - 1L
    funs <- substr(funs, 1L, parens)
    
    bads <- funs[funs %in% ns]

    if (length(bads) == 0L) {
        if (msgOnGood) {
            cat("Raster name(s) are likely acceptable.\n")
            utils::flush.console()
        }
    } else {

        realBads <- character()
        for (bad in bads) {
            if (grepl(fun, pattern = bad)) realBads <- c(realBads, bad)
        }

        if (length(realBads) > 0L) {
            
            msg <- paste0("At least one raster has a forbidden name that seems to appear in the equation:\n  ", fun)
            if (failOnBad) {
                stop(msg, "\n", paste(realBads, collapse = " "))
            } else {
                warning(msg)
                return(realBads)
            }
        } else if (msgOnGood) {
            if (any(nchar(ns) <= 3L)) {
                msg <-
                "The GRasters have one or more forbidden names, but the forbidden names do not seem to appear in\n  the equation. Use the equation with caution, or rename the GRasters. Also, consider renaming rasters\n  with very names with <=3 characters (e.g., 'a', 'a1', 'f_t', 'egg') to avoid potential issues."
            } else {
                msg <- "The GRasters have one or more forbidden names, but they do not seem to appear in\n  the equation. Use the equation with caution, or rename the GRasters."
            }
            warning(msg)
        }
    }

    invisible(bads)

    } # EOF
)

if (grassStarted()) {

# Setup
library(terra)

# Load rasters with precipitation and min/max temperature
madPpt <- fastData("madPpt")
madTmin <- fastData("madTmin")
madTmax <- fastData("madTmax")

### Classic and extended BIOCLIMs from SpatRasters
bcSR <- bioclims(madPpt, madTmin, madTmax, bios = "*")
bcSR

### BIOCLIMs from GRasters
ppt <- fast(madPpt)
tmin <- fast(madTmin)
tmax <- fast(madTmax)

# For small rasters, takes longer to run compared to SpatRaster version:
bc <- bioclims(ppt, tmin, tmax, bios = c(1, 5, 12))
bc
plot(bc)

}

if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Elevation raster, rivers vector
madElev <- fastData("madElev")
madRivers <- fastData("madRivers")

# Convert a SpatRaster to a GRaster, and sf to a GVector
elev <- fast(madElev)
rivers <- fast(madRivers)

### Buffer a raster by a given distance:
buffByDist <- buffer(elev, width = 2000) # 2000-m buffer
plot(buffByDist, legend = FALSE)
plot(madElev, add = TRUE)

### Buffer a raster by a given number of cells:
buffByCells <- buffer(elev, width = 20.01, unit = "cells") # 20-cell buffer
plot(buffByCells, legend = FALSE)
plot(madElev, add = TRUE)

### Buffer a vector:
buffRivers <- buffer(rivers, width = 2000, dissolve = TRUE) # 2000-m buffer
plot(buffRivers)
plot(st_geometry(madRivers), col = "blue", add = TRUE)

}

#' Geometry of a GVector (points, lines, or polygons)
#'
#' @description `geomtype()` reports whether a `GVector` represents points, lines, or polygons. The "`is.*`" functions test whether the `GVector` represents points, lines, or polygons.
#'
#' @param x A `GVector`.
#' 
#' @param grass Logical: If `FALSE` (default), return **terra**-like geometry types ("points", "lines", or "polygons"). If `TRUE`, return **GRASS**-like geometry types ("point", "line", "area"--note that these are a subset of the available types and may not be the "true" **GRASS** type).
#'
#' @return `geomtype()` returns either "points", "lines", or "polygons" if the `grass` arguments is `FALSE`, or "point", "line", "area" if `grass` is `TRUE`. The "`is.*`" functions return `TRUE` or `FALSE`.
#' 
#' @seealso [terra::geomtype()]
#'
#' @example man/examples/ex_GVector.r
#'
#' @aliases geomtype
#' @rdname geomtype
#' @exportMethod geomtype
methods::setMethod(
	f = "geomtype",
	signature = "GVector",
	definition = function(x, grass = FALSE) {

  	out <- x@geometry
	if (grass) {
		if (out == "points") {
			out <- "point"
		} else if (out == "lines") {
			out <- "line"
		} else if (out == "polygons") {
			out <- "area"
		}
	}
	out

  } # EOF
)

#' @aliases is.points
#' @rdname geomtype
#' @exportMethod is.points
methods::setMethod(
	f = "is.points",
	signature = "GVector",
	definition = function(x) {
		gt <- geomtype(x)
		gt == "points"
	} # EOF
)

#' @aliases is.lines
#' @rdname geomtype
#' @exportMethod is.lines
methods::setMethod(
	f = "is.lines",
	signature = "GVector",
	definition = function(x) {
		gt <- geomtype(x)
		gt == "lines"
	} # EOF
)

#' @aliases is.polygons
#' @rdname geomtype
#' @export is.polygons
methods::setMethod(
	f = "is.polygons",
	signature = "GVector",
	definition = function(x) {
		gt <- geomtype(x)
		gt == "polygons"
	} # EOF
)

#' Get geometry type from the sources() name of a vector
#'
#' @param x A `GVector` or the [sources()] name of one.
#'
#' @returns Character ("point", "line", or "area"; i.e., in **GRASS** nomenclature).
#'
#' @keywords internal
.geomtype <- function(x) {

	if (inherits(x, "GVector")) x <- sources(x)
	info <- .vectInfo(x)
	out <- info$geometry

	if (out == "points") {
		out <- "point"
	} else if (out == "lines") {
		out <- "line"
	} else if (out == "polygons") {
		out <- "area"
	}
	out

}

#' Reduce number of points in same raster cell
#'
#' @description This function thins a "points" `GVector` so that it has no more than `n` points per grid cell in a raster.
#'
#' @param x A "points" `GVector`.
#'
#' @param y A `GRaster`.
#'
#' @param n Integer or numeric integer: Maximum number of points to remain in a cell. The default is 1.
#'
#' @returns A "points" `GVector`.
#'
#' @example man/examples/ex_thinPoints.r
#'
#' @aliases thinPoints
#' @rdname thinPoints
#' @exportMethod thinPoints
methods::setMethod(
	f = "thinPoints",
	signature = c(x = "GVector", y = "GRaster"),
	function(x, y, n = 1) {
	
	.locationRestore(x)
	compareGeom(x, y)
	.region(y)

	src <- .makeSourceName("v_decimate", "vector")
	rgrass::execGRASS(
		cmd = "v.decimate",
		input = sources(x),
		output = src,
		cell_limit = n,
		flags = c(.quiet(), "overwrite", "g", "t")
	)

	if (nrow(x) > 0L) {
	
		table <- as.data.table(x)
		xCats <- .vCats(x)
		outCats <- .vCats(src)
		keeps <- which(outCats %in% xCats)
		table <- table[keeps]
	
	} else {
		table <- NULL
	}

	makeGVector(src, table = table)
	
	} # EOF
)

#' Display a fasterRaster object
#'
#' @description Display a `GLocation`, `GSpatial`, `GRegion`, `GRaster`, or `GVector` object.
#'
#' @param object,x An object of class `GLocation`, `GSpatial`, `GRegion`, `GRaster`, or `GVector`.
#'
#' @param ... Other arguments.
#'
#' @returns Nothing (side effect is to display metadata on the given object).
#'
#' @example man/examples/ex_GRaster_GVector.r
#'
#' @aliases show
#' @rdname show
#' @exportMethod show
methods::setMethod(
	f = "show", 
	signature = "GLocation",
	definition = function(object) {
	
	debug <- faster("debug")
	
	cat("class       :", paste(class(object), collapse=", "), "\n")
	cat("location    :", object@location, "\n")
	cat("mapset      :", object@mapset, "\n")
	cat("coord ref.  :", .showCRS(object), "\n")
	cat("workDir     :", object@workDir, "\n")

	}
)

#' @method print GLocation
#' @aliases print
#' @rdname show
#' @export print
print.GLocation <- function(x, ...) show(x)

#' @aliases summary
#' @rdname show
#' @exportMethod summary
methods::setMethod(
	f = "summary",
	signature = "GLocation",
	definition = function(object) show(object)
)


#' @aliases show
#' @rdname show
#' @exportMethod show
methods::setMethod(
	f = "show",
	signature = "GSpatial",
	definition = function(object) {

	debug <- faster("debug")

	digs <- min(3, getOption("digits"))
	extent <- round(object@extent, digs)
	
	cat("class       :", paste(class(object), collapse=", "), "\n")
	if (debug) {
		cat("location    :", object@location, "\n")
		cat("mapset      :", object@mapset, "\n")
		cat("workDir     :", object@workDir, "\n")
	}
	cat("topology    :", object@topology, "\n")
	cat("extent      :", paste(extent, collapse=", "), "(xmin, xmax, ymin, ymax)\n")
	if (debug & object@topology == "3D") cat("z extent    :", paste(object@zextent, collapse=", "), " (bottom, top)\n")
	cat("coord ref.  :", format(st_crs(object)), "\n")
	
	}
)

#' @method print GSpatial
#' @aliases print
#' @rdname show
#' @export print
print.GSpatial <- function(x, ...) show(x)

#' @aliases summary
#' @rdname show
#' @exportMethod summary
methods::setMethod(
	f = "summary",
	signature = "GSpatial",
	definition = function(object) show(object)
)

#' @aliases show
#' @rdname show
#' @exportMethod show
methods::setMethod(
	f = "show",
	signature = "GRegion",
	definition = function(object) {

	debug <- faster("debug")

	digs <- min(5, getOption("digits"))
	resol <- round(object@resolution, digs)

	extent <- round(object@extent, max(round(digs / 2), 2))
	zextent <- round(object@zextent, max(round(digs / 2), 2))
	
	cat("class       :", paste(class(object), collapse=", "), "\n")
	if (debug) {
		cat("location    :", object@location, "\n")
		cat("mapset      :", object@mapset, "\n")
		cat("workDir     :", object@workDir, "\n")
	}
	cat("topology    :", object@topology, "\n")
	cat("coord ref.  :", format(st_crs(object)), "\n")
	cat("dimensions  :", paste(object@dimensions, collapse=", "), "(nrow, ncol, ndepth)\n")
	cat("resolution  :", paste(resol, collapse=", "), "(x, y, z)\n")
	cat("extent      :", paste(extent, collapse=", "), "(xmin, xmax, ymin, ymax)\n")
	cat("z extent    :", paste(object@zextent, collapse=", "), " (bottom, top)\n")
	
	}
)

#' @method print GRegion
#' @aliases print
#' @rdname show
#' @export print
print.GRegion <- function(x, ...) show(x)

#' @aliases summary
#' @rdname show
#' @exportMethod summary
methods::setMethod(
	f = "summary",
	signature = "GRegion",
	definition = function(object) show(object)
)

#' @aliases show
#' @rdname show
#' @exportMethod show
methods::setMethod(
	f = "show",
	signature = "GRaster",
	definition = function(object) {

	numLevels <- nlevels(object)
	if (any(numLevels > 0L)) {
  		minColWidth <- 14L
	} else {
  		minColWidth <- 4L
	}
	maxColWidth <- 17L
	debug <- faster("debug")

	digs <- min(5, getOption("digits"))
	resol <- round(object@resolution, digs)
	if (length(resol) == 2L) resol <- c(resol, NA_real_)
	
	extent <- round(object@extent, min(digs, 3))

	### pad everything by same amount so display of data for each later appears in neat columns
	minVal <- object@minVal
	maxVal <- object@maxVal
	sigs <- omnibus::roundedSigDigits(c(minVal, maxVal))
	sigs <- -1 * min(sigs) + 3
	minVal <- round(minVal, sigs)
	maxVal <- round(maxVal, sigs)
	
	minValLength <- nchar(minVal)
	maxValLength <- nchar(maxVal)
	
	if (anyNA(minValLength)) minValLength[is.na(minValLength)] <- 2L
	if (anyNA(maxValLength)) maxValLength[is.na(maxValLength)] <- 2L
	
	# minimum and maximum category values
	mmCats <- minmax(object, levels = TRUE)
	minCat <- mmCats["min", , drop = TRUE]
	maxCat <- mmCats["max", , drop = TRUE]
	minCat[is.na(minCat)] <- "NA"
	maxCat[is.na(maxCat)] <- "NA"
	nCats <- nlevels(object)
	
	if (debug) {
		activeCat <- object@activeCat - 1L
		if (anyNA(activeCat)) activeCat[is.na(activeCat)] <- "NA"
	}

	# truncate long names
	namesNice <- object@names
	ncharNames <- nchar(namesNice)
	if (any(ncharNames > maxColWidth)) {
    	namesNice[ncharNames > maxColWidth] <-
   			paste0(substr(namesNice[ncharNames > maxColWidth], 1L, maxColWidth - 1L), "~")
	}

	# truncate long sources
	if (debug) {
	
		sources <- object@sources
		ncharNames <- nchar(sources)
		if (any(ncharNames > maxColWidth)) {
			sources[ncharNames > maxColWidth] <-
				paste0(substr(sources[ncharNames > maxColWidth], 1L, maxColWidth - 1L), "~")
		}
		
	}

	# datatype
	datatypeFR <- datatype(object, "fasterRaster")
	if (debug) datatypeGRASS <- datatype(object, "GRASS")

	# number of characters of longest display item
	nc <- pmax(
		rep(3, object@nLayers),
		nchar(namesNice),
		nchar(datatypeFR),
		nchar(nCats),
		minValLength, maxValLength
	)

 	if (debug) {
		nc <- pmax(
			nc,
			nchar(activeCat),
			nchar(object@datatypeGRASS),
			nchar(sources)
		)
	}
	
	if (any(nc < minColWidth)) nc[nc < minColWidth] <- minColWidth

	# truncate category names
	if (any(nchar(minCat) > nc)) {
  		minCat[nchar(minCat) > nc] <- paste0(substr(minCat[nchar(minCat) > nc], 1L, nc - 1L), "~")
	}
	if (any(nchar(maxCat) > nc)) {
  		maxCat[nchar(maxCat) > nc] <- paste0(substr(maxCat[nchar(maxCat) > nc], 1L, nc - 1L), "~")
	}

 	# pad with extra spaces
 	sourcesNice <- names <- datatypeNiceFR <- datatypeNiceGRASS <- nCatsNice <- minValNice <- maxValNice <- minCatNice <- maxCatNice <- rep(NA, object@nLayers)

	for (i in seq_len(object@nLayers)) {

		fmt <- paste0("%", nc[i], "s")
		if (debug) {
			sourcesNice[i] <- sprintf(fmt, sources[i])
    		activeCat[i] <- sprintf(fmt, activeCat[i])
   			datatypeNiceGRASS[i] <- sprintf(fmt, datatypeGRASS[i])
		}
		# names[i] <- sprintf(fmt, object@names[i])
		names[i] <- sprintf(fmt, namesNice[i])
		datatypeNiceFR[i] <- sprintf(fmt, datatypeFR[i])
		nCatsNice[i] <- sprintf(fmt, nCats[i])
		minCatNice[i] <- sprintf(fmt, minCat[i])
		maxCatNice[i] <- sprintf(fmt, maxCat[i])
		minValNice[i] <- sprintf(fmt, minVal[i])
		maxValNice[i] <- sprintf(fmt, maxVal[i])

	}
	
	# concatenate each line into single strings
	names <- paste(names, collapse=" ")
	datatypeNiceFR <- paste(datatypeNiceFR, collapse=" ")
	minValNice <- paste(minValNice, collapse=" ")
	maxValNice <- paste(maxValNice, collapse=" ")
 	nCatsNice <- paste(nCatsNice, collapse = " ")
	minCatNice <- paste(minCatNice, collapse=" ")
	maxCatNice <- paste(maxCatNice, collapse=" ")

	if (debug) {
		sources <- paste(sourcesNice, collapse=" ")
  		datatypeNiceGRASS <- paste(datatypeNiceGRASS, collapse = " ")
		activeCat <- paste(activeCat, collapse = "")
	}

	cat("class       : GRaster\n")
	if (debug) {
		cat("location    :", object@location, "\n")
		cat("mapset      :", object@mapset, "\n")
		cat("workDir     :", object@workDir, "\n")
	}
	cat("topology    :", object@topology, "\n")
	cat("dimensions  :", paste(c(object@dimensions, object@nLayers), collapse=", "), "(nrow, ncol, ndepth, nlyr)\n")
	cat("resolution  :", paste(resol, collapse=", "), "(x, y, z)\n")
	cat("extent      :", paste(extent, collapse=", "), "(xmin, xmax, ymin, ymax)\n")
	if (debug | object@topology == "3D") {
		cat("z extent    :", paste(object@zextent, collapse=", "), "(bottom, top)\n")
	}
	cat("coord ref.  :", format(st_crs(object)), "\n")
	if (debug) cat("projection  :", object@projection, "\n")
	cat("name(s)     :", names, "\n")
	if (debug) cat("sources     :", sources, "\n")
	if (!debug) {
		cat("datatype    :", datatypeNiceFR, "\n")
	} else {
		cat("type (fR)   :", datatypeNiceFR, "\n")
		cat("type (GR)   :", datatypeNiceGRASS, "\n")
	}
	cat("min. value  :", minValNice, "\n")
	cat("max. value  :", maxValNice, "\n")
	if (any(numLevels > 0L)) {
  		cat("categories  :", nCatsNice, "\n")
  		if (debug) cat("active cat. :", activeCat, "\n")
		cat("min. categ. :", minCatNice, "\n")
		cat("max. categ. :", maxCatNice, "\n")
	}

	} # EOF
)

#' @method print GRaster
#' @aliases print
#' @rdname show
#' @export print
print.GRaster <- function(x, ...) show(x)

#' @aliases summary
#' @rdname show
#' @exportMethod summary
methods::setMethod(
	f = "summary",
	signature = "GRaster",
	definition = function(object) show(object)
)

#' @aliases show
#' @rdname show
#' @exportMethod show
methods::setMethod(
	f = "show",
	signature = "GVector",
	definition = function(object) {

	# maximum number of fields to show
	maxFieldsToShow <- 8L
  	maxColWidth <- 17L
	maxDigits <- 5L

	debug <- faster("debug")

	digs <- min(maxDigits, getOption("digits"))
	extent <- round(object@extent, digs)
	zextent <- round(object@zextent, digs)

	# concatenate fields across vectors for display
	# making list, each element becomes a line in the display
	# each element contains field n from first vector, field n from second vector, etc.
	# will then print each element on one line
	nFields <- ncol(object@table)
	nRows <- nrow(object@table)
	table <- head(object@table, 3)
	if (nRows > 0L) {

		fields <- colnames(table)
		classes <- sapply(table, class)
		maxFieldsToShow <- min(nFields, maxFieldsToShow)
		
  		row1 <- table[1L, 1L:maxFieldsToShow]
		row1 <- as.data.frame(row1)
		row1 <- unlist(row1)
		row1 <- as.character(row1)
		if (anyNA(row1)) row1[is.na(row1)] <- "NA"
		
		if (nRows >= 2L) {
			row2 <- table[2L, 1L:maxFieldsToShow]
			row2 <- as.data.frame(row2)
			row2 <- unlist(row2)
			row2 <- as.character(row2)
			if (anyNA(row2)) row2[is.na(row2)] <- "NA"
		} else {
			row2 <- NULL
		}
		
		if (nRows >= 3L) {
			row3 <- table[3L, 1L:maxFieldsToShow]
   			row3 <- as.data.frame(row3)
			row3 <- unlist(row3)
			row3 <- as.character(row3)
			if (anyNA(row3)) row3[is.na(row3)] <- "NA"
		} else {
			row3 <- NULL
		}

		fields <- fields[1L:maxFieldsToShow]
		classes <- classes[1L:maxFieldsToShow]
		
		classes <- gsub(classes, pattern = "integer", replacement = "<int>")
		classes <- gsub(classes, pattern = "numeric", replacement = "<num>")
		classes <- gsub(classes, pattern = "complex", replacement = "<cpx>")
		classes <- gsub(classes, pattern = "character", replacement = "<chr>")
		classes <- gsub(classes, pattern = "logical", replacement = "<log>")
		classes <- gsub(classes, pattern = "Date", replacement = "<Dte>")

		# column widths
		ncFields <- nchar(fields)
		ncFieldClasses <- nchar(classes)
		ncRow1 <- nchar(row1)
		nc <- pmax(ncFields, ncFieldClasses, ncRow1)
		
		if (!is.null(row2)) ncTableCells <- nchar(row2)
		if (!is.null(row3)) ncTableCells <- pmax(ncTableCells, nchar(row3))
		if (!is.null(row2))	nc <- pmax(nc, ncTableCells)
		if (any(nc > maxColWidth)) nc[nc > maxColWidth] <- maxColWidth

		# format table values
		tooLongIndex <- which(nchar(row1) > maxColWidth)
		if (length(tooLongIndex) > 0L) {
			row1[tooLongIndex] <- paste0(substr(row1[tooLongIndex], 1L, nc[tooLongIndex] - 1L), "~")
		}

		if (!is.null(row2)) {		

			tooLongIndex <- which(nchar(row2) > nc)
			if (length(tooLongIndex) > 0L) {
				row2[tooLongIndex] <- paste0(substr(row2[tooLongIndex], 1L, nc[tooLongIndex] - 1L), "~")
			}
		
		}
		
		if (!is.null(row3)) {		

   			tooLongIndex <- which(nchar(row3) > nc)
			if (length(tooLongIndex) > 0L) {
				row3[tooLongIndex] <- paste0(substr(row3[tooLongIndex], 1L, nc[tooLongIndex] - 1L), "~")
			}
		
		}
		
		for (i in seq_along(fields)) {
		
			fmt	<- paste0("%", nc[i], "s")
			fields[i] <- sprintf(fmt, fields[i])
			classes[i] <- sprintf(fmt, classes[i])
   			row1[i] <- sprintf(fmt, row1[i])
   			if (!is.null(row2)) row2[i] <- sprintf(fmt, row2[i])
   			if (!is.null(row3)) row3[i] <- sprintf(fmt, row3[i])
		
		}
		
		if (nFields > maxFieldsToShow) {
			classes <- c(classes, paste("(and", nFields - maxFieldsToShow, "more column(s))"))
		}
		
	}
	
	cat("class       : GVector\n")
	if (debug) {
		cat("location    :", object@location, "\n")
		cat("mapset      :", object@mapset, "\n")
		cat("source      :", object@sources, "\n")
		cat("workDir     :", object@workDir, "\n")
	}
 	cat("geometry    :", object@topology, object@geometry, "\n")
	cat("dimensions  :", paste0(object@nGeometries, ", ", object@nSubgeometries, ", ", nFields), "(geometries, sub-geometries, columns)\n")
	# cat("dimensions  :", paste0(object@nGeometries, ", ", nFields), "(geometries, columns)\n")
	cat("extent      :", paste(extent, collapse=", "), "(xmin, xmax, ymin, ymax)\n")
	if (debug | object@topology == "3D") cat("z extent    :", paste(zextent, collapse=", "), "(bottom, top)\n")
	cat("coord ref.  :", format(st_crs(object)), "\n")
	if (debug) cat("projection  :", object@projection, "\n")

	if (nFields > 0L) {
		cat("names       :", fields, "\n")
		cat("type        :", classes, "\n")
		cat("values      :", row1, "\n")
		if (!is.null(row2)) cat("             ", row2, "\n")
		if (!is.null(row3)) cat("             ", row3, "\n")
		if (nRows > 3L) cat(paste0("              (and ", nRows - 3L, " more rows)\n"))
	}
	
	} # EOF

)

#' @method print GVector
#' @aliases print
#' @rdname show
#' @export print
print.GVector <- function(x, ...) show(x)

#' @aliases summary
#' @rdname show
#' @exportMethod summary
methods::setMethod(
    f = "summary",
    signature = "GVector",
    definition = function(object) show(object)
)

#' Show the CRS for an object, and make it pretty, even if it doesn't want to
#' @noRd
.showCRS <- function(x) {

	# format(st_crs(x))

	x <- crs(x)
	x <- sf::st_crs(x)

	if (!is.na(x$input) && x$input != x$wkt) {
		out <- x$input # pre-made
	} else {
  		start <- regexec(x$wkt, pattern = "CONVERSION")[[1L]]
		if (start == -1L) {
			out <- NA_character_
		} else {
   			start <- start + nchar("CONVERSION") + 1L
			x <- substr(x$wkt, start = start, stop = nchar(x$wkt))

   			stop <- regexec(x, pattern = ",")[[1L]] - 2L
			x <- substr(x, start = 2L, stop)
			if (grepl(x, pattern = "METHOD")) {
				out <- NA_character_
			} else {
			   out <- x
			}
		}
	}
	out

}

#' Logic-methods operations on GRasters
#'
#' @description You can do logical operations on `GRaster`s. A cell with a value of 1 is interpreted as `TRUE`, and a value of 0 is interpreted as `FALSE`. You can compare:
#' * A `GRaster` to another `GRaster`
#' * A `GRaster` to a logical value (`TRUE` or `FALSE`, but not `NA`--see [not.na()])
#' * A `GRaster` to a numeric or integer value that is 0 or 1
#' 
#' Operators include:
#' * `|`: `TRUE` if either condition is `TRUE` (or 1), but returns `NA` if either condition is `NA`.
#' * `&`: `TRUE` if both conditions are `TRUE` (or 1), but `NA` if either is `NA`.
#'
#' @param e1,e2 Two `GRaster`s, or a `GRaster` and a logical value (`TRUE` or `FALSE`, but not `NA`), a numeric value that is 0 or 1 (but not `NA_real_`), or an integer value that is 0 or 1 (but not `NA_integer_`).
#'
#' @returns A binary `GRaster` (1 ==> `TRUE`, 0 ==> `FALSE`, plus `NA` when comparison results in `NA`).
#'
#' @example man/examples/ex_GRaster_comparison_logic.r
#'
#' @aliases Logic-methods
#' @rdname Logic-methods
#' @exportMethod Logic
methods::setMethod(
	f = "Logic",
	signature(e1 = "GRaster", e2 = "GRaster"),
    function(e1, e2) {
	
		compareGeom(e1, e2)
		.locationRestore(e1)
		.region(e1)

		if (nlyr(e1) > 1L & nlyr(e2) == 1L) {
			e2 <- e2[[rep(1L, nlyr(e1))]]
		} else if (nlyr(e1) == 1L & nlyr(e2) > 1L) {
			e1 <- e1[[rep(1L, nlyr(e2))]]
		} else if (nlyr(e1) != nlyr(e2)) {
			stop("Rasters must have the same number of layers, or at least one raster must have a single layer.")
		}

		oper <- as.vector(.Generic)[1L]
		for (i in 1L:nlyr(e1)) {

			name <- paste0(names(e1)[i], "_", names(e2)[i])
			src <- .makeSourceName("logic", "raster")

			ex <- paste0(src, "= int(if(", sources(e1)[i], " ", oper, " ", sources(e2)[i], "))")
			this <- .genericArithRast(name = name, src = src, ex = ex)
			
			if (i == 1L) {
				out <- this
			} else {
				out <- c(out, this)
			}
			
		}
		out
		
	} # EOF
)

# logical raster
#' @aliases Logic-methods
#' @rdname Logic-methods
#' @exportMethod Logic
methods::setMethod(
	f = "Logic",
	signature(e1 = "logical", e2 = "GRaster"),
    function(e1, e2) {
	
		.locationRestore(e2)
		.region(e2)

		e1 <- as.integer(e1)
		
		oper <- as.vector(.Generic)[1L]
		for (i in 1L:nlyr(e2)) {
			
			name <- names(e2)[i]
			src <- .makeSourceName("logic", "rast")

			ex <- paste(src, "= int(if(", e1, " ", oper, " ", sources(e2)[i], "))")
			this <- .genericArithRast(name = name, src = src, ex = ex)
			
			if (i == 1L) {
				out <- this
			} else {
				out <- c(out, this)
			}
			
		}
		out
			
	} # EOF
)

# raster logical
#' @aliases Logic-methods
#' @rdname Logic-methods
#' @exportMethod logic
methods::setMethod(
	f = "Logic",
	signature(e1 = "GRaster", e2 = "logical"),
    function(e1, e2) {

		.locationRestore(e1)
		.region(e1)

		e2 <- as.integer(e2)

		oper <- as.vector(.Generic)[1L]
		for (i in 1L:nlyr(e1)) {

			name <- names(e1)[i]
			src <- .makeSourceName("logic", "rast")
			
			ex <- paste(src, "= int(if(", sources(e1)[i], " ", oper, " ", e2, "))")
			this <- .genericArithRast(name = name, src = src, ex = ex)
			
			if (i == 1L) {
				out <- this
			} else {
				out <- c(out, this)
			}
			
		}
		out
			
	} # EOF
)

# raster numeric
#' @aliases Logic-methods
#' @rdname Logic-methods
#' @exportMethod logic
methods::setMethod(
	f = "Logic",
	signature(e1 = "GRaster", e2 = "numeric"),
    function(e1, e2) {

		.locationRestore(e1)
		.region(e1)

		if (!omnibus::is.wholeNumber(e2) | !(omnibus::compareFloat(e2, 0, "==") | omnibus::compareFloat(e2, 1, "=="))) stop("Can only perform logical operations using GRasters and\n  either another GRaster, TRUE, FALSE, 1, or 0.")
		e2 <- as.integer(e2)

		oper <- as.vector(.Generic)[1L]
		for (i in 1L:nlyr(e1)) {

			name <- names(e1)[i]
			src <- .makeSourceName("logic", "rast")
			
			ex <- paste(src, "= int(if(", sources(e1)[i], " ", oper, " ", e2, "))")
			this <- .genericArithRast(name = name, src = src, ex = ex)
			
			if (i == 1L) {
				out <- this
			} else {
				out <- c(out, this)
			}
			
		}
		out
			
	} # EOF
)

# numeric raster
#' @aliases Logic-methods
#' @rdname Logic-methods
#' @exportMethod logic
methods::setMethod(
	f = "Logic",
	signature(e1 = "numeric", e2 = "GRaster"),
    function(e1, e2) {

		.locationRestore(e1)
		.region(e1)

		if (!omnibus::is.wholeNumber(e1) | !(omnibus::compareFloat(e1, 0, "==") | omnibus::compareFloat(e1, 1, "=="))) stop("Can only perform logical operations using GRasters and\n  either another GRaster, TRUE, FALSE, 1, or 0.")
		e1 <- as.integer(e1)

		oper <- as.vector(.Generic)[1L]
		for (i in 1L:nlyr(e1)) {

			name <- names(e2)[i]
			src <- .makeSourceName("logic", "rast")
			
			ex <- paste(src, "= int(if(", e1, " ", oper, " ", sources(e2)[i], "))")
			this <- .genericArithRast(name = name, src = src, ex = ex)
			
			if (i == 1L) {
				out <- this
			} else {
				out <- c(out, this)
			}
			
		}
		out
			
	} # EOF
)

# raster integer
#' @aliases Logic-methods
#' @rdname Logic-methods
#' @exportMethod logic
methods::setMethod(
	f = "Logic",
	signature(e1 = "GRaster", e2 = "integer"),
    function(e1, e2) {

		.locationRestore(e1)
		.region(e1)

		if (!(e2 == 0L | e2 == 1L)) stop("Can only perform logical operations using GRasters and\n  either another GRaster, TRUE, FALSE, 1, or 0.")
		e2 <- as.integer(e2)

		oper <- as.vector(.Generic)[1L]
		for (i in 1L:nlyr(e1)) {

			name <- names(e1)[i]
			src <- .makeSourceName("logic", "rast")
			
			ex <- paste(src, "= int(if(", sources(e1)[i], " ", oper, " ", e2, "))")
			this <- .genericArithRast(name = name, src = src, ex = ex)
			
			if (i == 1L) {
				out <- this
			} else {
				out <- c(out, this)
			}
			
		}
		out
			
	} # EOF
)

# integer raster
#' @aliases Logic-methods
#' @rdname Logic-methods
#' @exportMethod logic
methods::setMethod(
	f = "Logic",
	signature(e1 = "integer", e2 = "GRaster"),
    function(e1, e2) {

		.locationRestore(e1)
		.region(e1)

		if (!(e1 == 0L | e1 == 1L)) stop("Can only perform logical operations using GRasters and\n  either another GRaster, TRUE, FALSE, 1, or 0.")
		e1 <- as.integer(e1)

		oper <- as.vector(.Generic)[1L]
		for (i in 1L:nlyr(e1)) {

			name <- names(e2)[i]
			src <- .makeSourceName("logic", "rast")
			
			ex <- paste(src, "= int(if(", e1, " ", oper, " ", sources(e2)[i], "))")
			this <- .genericArithRast(name = name, src = src, ex = ex)
			
			if (i == 1L) {
				out <- this
			} else {
				out <- c(out, this)
			}
			
		}
		out
			
	} # EOF
)

# # raster raster
# #' @aliases Logic-methods
# #' @rdname Logic-methods
# #' @exportMethod orOrOr
# methods::setMethod(
# 	f = "orOrOr",
# 	signature = c(e1 = "GRaster", e2 = "GRaster"),
# 	function(e1, e2) {
	
# 	compareGeom(e1, e2)
# 	.locationRestore(e1)
# 	.region(e1)

# 	if (nlyr(e1) > 1L & nlyr(e2) == 1L) {
# 		e2 <- e2[[rep(1L, nlyr(e1))]]
# 	} else if (nlyr(e1) == 1L & nlyr(e2) > 1L) {
# 		e1 <- e1[[rep(1L, nlyr(e2))]]
# 	} else if (nlyr(e1) != nlyr(e2)) {
# 		stop("Rasters must have the same number of layers, or at least one raster must have a single layer.")
# 	}

# 	for (i in 1L:nlyr(e1)) {

# 		name <- paste0(names(e1)[i], "_", names(e2)[i])
# 		src <- .makeSourceName("logic", "raster")

# 		ex <- paste0(src, "= int(if(", sources(e1)[i], " ||| ", sources(e2)[i], "))")
# 		this <- .genericArithRast(name = name, src = src, ex = ex)
		
# 		if (i == 1L) {
# 			out <- this
# 		} else {
# 			out <- c(out, this)
# 		}
		
# 	}
# 	out
		
# 	} # EOF
# )

# # raster raster
# #' @aliases Logic-methods
# #' @rdname Logic-methods
# #' @exportMethod orOrOr
# methods::setMethod(
# 	f = "andAndAnd",
# 	signature = c(e1 = "GRaster", e2 = "GRaster"),
# 	function(e1, e2) {
	
# 	compareGeom(e1, e2)
# 	.locationRestore(e1)
# 	.region(e1)

# 	if (nlyr(e1) > 1L & nlyr(e2) == 1L) {
# 		e2 <- e2[[rep(1L, nlyr(e1))]]
# 	} else if (nlyr(e1) == 1L & nlyr(e2) > 1L) {
# 		e1 <- e1[[rep(1L, nlyr(e2))]]
# 	} else if (nlyr(e1) != nlyr(e2)) {
# 		stop("Rasters must have the same number of layers, or at least one raster must have a single layer.")
# 	}

# 	for (i in 1L:nlyr(e1)) {

# 		name <- paste0(names(e1)[i], "_", names(e2)[i])
# 		src <- .makeSourceName("logic", "raster")

# 		ex <- paste0(src, "= int(if(", sources(e1)[i], " &&& ", sources(e2)[i], "))")
# 		this <- .genericArithRast(name = name, src = src, ex = ex)
		
# 		if (i == 1L) {
# 			out <- this
# 		} else {
# 			out <- c(out, this)
# 		}
		
# 	}
# 	out
		
# 	} # EOF
# )

# #' @name `|||`
# #' @title Logic-methods operations on GRasters
# #' @rdname Logic-methods
# #' @export
# `|||` <- function(e1, e2) orOrOr(e1, e2)

# #' @name `&&&`
# #' @title Logic-methods operations on GRasters
# #' @rdname Logic-methods
# #' @export
# `&&&` <- function(e1, e2) andAndAnd(e1, e2)

if (grassStarted()) {

# Setup
library(sf)

# Polygon of coastal Madagascar and Dypsis specimens
madCoast4 <- fastData("madCoast4") # polygons
madDypsis <- fastData("madDypsis") # points

# Convert vectors:
coast4 <- fast(madCoast4)
dypsis <- fast(madDypsis)

# Create another polygons vector from a convex hull around Dypsis points
hull <- convHull(dypsis)

### union()

unioned <- union(coast4, hull)
plot(unioned)

plus <- coast4 + hull # same as union()

### intersect

inter <- intersect(coast4, hull)
plot(coast4)
plot(hull, border = "red", add = TRUE)
plot(inter, border = "blue", add = TRUE)

### xor

xr <- xor(coast4, hull)
plot(coast4)
plot(xr, border = "blue", add = TRUE)

### erase

erased <- erase(coast4, hull)
plot(coast4)
plot(erased, border = "blue", add = TRUE)

minus <- coast4 - hull # same as erase()

}

if (grassStarted()) {

# Setup
library(terra)

# Elevation
madElev <- fastData("madElev")

# Convert to GRaster:
elev <- fast(madElev)

# Thin elevation raster:
thinned <- thinLines(elev, iter = 300)
plot(thinned)

# Convert to lines:
rastToLines <- as.lines(thinned)
plot(rastToLines)

# We can clean this:
cleanLines <- fixDangles(x = rastToLines)
plot(rastToLines, col = "red")
plot(cleanLines, add = TRUE)

}

# DEFAULTS FOR ALL GLOBAL SETTINGS

# global PUBLIC options
.grassDirDefault <- function() NA_character_
.addonsDirDefault <- function() NA_character_
# .workDirDefault <- function() file.path(omnibus::forwardSlash(tempdir()), omnibus::rstring(1))
.workDirDefault <- function() omnibus::forwardSlash(tempdir())

.locationDefault <- function() "default"
# .mapsetDefault <- function() "PERMANENT"

.coresDefault <- function() 2
.verboseDefault <- function() FALSE
.debugDefault <- function() FALSE
# .cleanDefault <- function() TRUE
.memoryDefault <- function() 2048 # in MB
# .nAtATimeDefault <- function() 1000000 # number of indices to select at a time
# .rasterPrecisionDefault <- function() "double" # "FCELL"/"float", or "DCELL"/"double"
.useDataTableDefault <- function() FALSE # logical

#' Names or table of PUBLIC options
#' @param type NULL (show all names), OR "numeric", "logical", or "character", OR "table" (show table)
#' @noRd
.namesOfOptions <- function(type = NULL) {
	
	opts <- data.frame(
		name = c(
			"grassDir",
			"addonsDir",
			# "clean",
			"verbose",
			"debug",
			"cores",
			"memory",
			"location",
			# "mapset",
			"workDir",
			"useDataTable"
		),
		type = c(
			"character",
			"character",
			# "logical",
			"logical",
			"logical",
			"numeric",
			"numeric",
			"character",
			# "character",
			"character",
			"logical"
		)
	)
		
	if (is.null(type)) {
		opts$name
	} else if (type == "table") {
		opts
	} else {
		opts$name[opts$type == type]
	}
	
}

#' Fill holes in a GVector
#'
#' @description `fillHoles()` removes holes in a `GVector`.
#'
#' @param x A `GVector`.
#'
#' @param fail Logical: If `TRUE` (default), and **GRASS 8.3** or higher is not installed, cause an error. If `FALSE`, a warning will be displayed and a `NULL` value will be returned. This function requires **GRASS 8.3** or higher to be installed.
#'
#' @returns A `GVector`.
#'
#' @example man/examples/ex_GVector.r
#'
#' @seealso [terra::fillHoles()], **GRASS** manual page for tool `v.fill.holes` (see `grassHelp("v.fill.holes")`)
#'
#' @aliases fillHoles
#' @rdname fillHoles
#' @exportMethod fillHoles
methods::setMethod(
	f = "fillHoles",
	signature(x = "GVector"),
	function(x, fail = TRUE) {
	
	if (grassInfo("versionNumber") < 8.4) {
	
		if (fail) {
			msg <- "This function requires GRASS 8.4 or above."
			stop(msg)
		} else {
			msg <- "This function requires GRASS 8.4 or above. A NULL value has been returned."
			warning(msg)
			out <- NULL
		}

	} else {

		.locationRestore(x)
		src <- .makeSourceName("v_fill_holes", "vector")
		rgrass::execGRASS(
			cmd = "v.fill.holes",
			input = sources(x),
			output = src,
			flags = c(.quiet(), "overwrite")
		)

		out <- makeGVector(src, table = x)

	}

	out

	} # EOF
)

if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster
elev <- fast(madElev)

# Calculate stream channels
streams <- streams(elev)
plot(streams)

}

#' Coordinates of a vector"s features or a raster"s cell centers
#'
#' @description Returns the coordinates of the center of cells of a `GRaster` or coordinates of a `GVector`'s vertices. The output will be a `matrix`, `data.frame`, or `list`. If you want the output to be a "points" `GVector`, use [as.points()].
#'
#' @param x A `GVector` or a `GRaster`.
#' @param z If `TRUE` (default), return x-, y-, and z-coordinates. If `FALSE`, just return x- and y-coordinates. For 2-dimensional objects, z-coordinates will all be 0.
#' @param na.rm Logical: If `TRUE`, remove cells that are `NA` (`GRaster`s only).
#'
#' @returns A `matrix`, `data.frame`, or `list`.
#'
#' @seealso [terra::crds()]
#'
#' @example man/examples/ex_crds.r
#'
#' @aliases crds
#' @rdname crds
#' @exportMethod crds
methods::setMethod(
    f = "crds",
    signature = c(x = "GRaster"),
    function(x, z = is.3d(x), na.rm = TRUE) {

	..char <- NULL

	.locationRestore(x)
	.region(x)

	flags <- c(.quiet(), "overwrite")
	if (!na.rm) flags <- c(flags, "i")

	temp <- paste0(tempfile(), ".csv")
	rgrass::execGRASS("r.out.xyz", input = sources(x), output = temp, separator = "comma", flags = flags, intern = TRUE)
	# see https://grass.osgeo.org/grass82/manuals/r.out.xyz.html
	out <- data.table::fread(temp)

	if (is.3d(x)) {
		names(out) <- c("x", "y", "z", names(x))
	} else {
		names(out) <- c("x", "y", names(x))
	}

	# convert to numerics
	classes <- sapply(out, class)
	if (any(classes %in% "character")) {
		chars <- which(classes == "character")
		for (char in chars) out[ , ..char] <- as.numeric(out[ , ..char])
	}

	if (!faster("useDataTable")) out <- as.data.frame(out)
	out
	
    } # EOF
)

#' @aliases crds
#' @rdname crds
#' @exportMethod crds
methods::setMethod(
	f = "crds",
	signature = c(x = "GVector"),
	function(x, z = is.3d(x)) {

	cats <- if (geomtype(x) == "points") { 1:ngeom(x) } else { NULL }
	.crdsVect(x, z = z, cats = cats)
	
	} # EOF
)

#' @rdname crds
#' @export
st_coordinates <- function(x) {
	if (inherits(x, "GVector")) {
		cats <- if (geomtype(x) == "points") { 1:ngeom(x) } else { NULL }
		.crdsVect(x, z = FALSE, cats = cats)
	} else {
		sf::st_coordinates(x)
	}
}

#' Extract coordinates for vector
#'
#' @param x A GVector or [sources()] name of one.
#' @param z T/F Extract z-coordinate?
#' @param gtype Character: [geomtype()] of the vector.
#' @param cats Either `NULL` (default), or a numeric/integer vector with one unique value per point in `x`. Used to speed attachment of a database to a points vector, which is needed to extract coordinates using `v.to.db`. If `NULL`, category values are obtained by querying the points vector (which can be slow).
#' 
#' @noRd
.crdsVect <- function(x, z, gtype = NULL, cats = NULL) {

	if (inherits(x, "GVector")) {
		.locationRestore(x)
		gtype <- geomtype(x)
		src <- sources(x)
		cats <- 1:nrow(x)
	} else {
		src <- x
	}

	# if lines or polygons, convert to points first
	if (gtype %in% c("lines", "polygons")) {

		stop("crds() will only work on points vectors :(.")
		
		# ####### NB seems to work on lines but disagrees with st_coordinates() and crds()

		# if (z && is.3d(x)) warning("z coordinates ignored.")
	
		# src <- .makeSourceName("points", "vect")
		# rgrass::execGRASS("v.to.points", input=sources(x), output=src, use="vertex", flags=c(.quiet(), "overwrite"), intern=TRUE)
		# pts <- makeGVector(src)
		# pts <- vect(pts)

		# out <- terra::crds(pts)

	} else if (gtype == "points") {

		if (!.vHasDatabase(src)) .vAttachDatabase(src, cats = cats)

		info <- rgrass::execGRASS(
			cmd = "v.to.db",
			map = src,
			flags = c(.quiet(), "p"),
			option = "coor",
			type = "point",
			intern = TRUE
		)

		info <- info[!grepl(info, pattern = "cat|x|y|z")]
		info <- info[grepl(info, pattern = "\\|")]
		# cutAt <- which(info == "Reading features...")
		# if (length(cutAt) > 0L) info <- info[1L:(cutAt - 1L)]
		
		info <- strsplit(info, split = "\\|")
		info <- lapply(info, as.numeric)
		out <- do.call(rbind, info)
		
		if (z) {
			out <- out[ , 2L:4L, drop = FALSE]
			colnames(out) <- c("x", "y", "z")
		} else {
			out <- out[ , 2L:3L, drop = FALSE]
			colnames(out) <- c("x", "y")
		}
		
	}	
		
	if (faster("useDataTable")) out <- data.table::as.data.table(out)
	out
	
}

if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")
madCover <- fastData("madCover")

### Substitution within an integer/numeric raster

# Convert SpatRaster to GRaster
elev <- fast(madElev)

# Simple substitution of one value, keeping all other values
newElev <- elev
newElev[newElev == 100] <- -100
newElev[newElev > 500] <- 500
hist(newElev)

# Simple substitution of one value, keeping all other values
substituted <- subst(elev, from = 300, to = -300)
substituteds <- c(elev, substituted)
names(substituteds) <- c("original", "substituted")
plot(substituteds)

# Simple substitution of three values, keeping all other values
substituted <- subst(elev, from = c(299, 300, 301), to = c(-699, -600, -601))
substituteds <- c(elev, substituted)
names(substituteds) <- c("original", "substituted")
plot(substituteds)

# Simple substitution of three values to one other value, retaining remainder
substituted <- subst(elev, from = c(299, 300, 301), to = -1000)
substituteds <- c(elev, substituted)
names(substituteds) <- c("original", "substituted")
plot(substituteds)

# Simple substitution of one value, setting all other values to 100
substituted <- subst(elev, from = 300, to = -300, others = 100)
substituteds <- c(elev, substituted)
names(substituteds) <- c("original", "substituted")
plot(substituteds)

### Substitution within a factor/categorical raster

# Convert a SpatRaster to a GRaster:
cover <- fast(madCover)

cover <- droplevels(cover) # remove unused levels
levels(cover) # levels of "cover"

# Substitute using level name, replace with EXISTING level label
from <- "Mosaic cropland/vegetation"
to <- "Mosaic crops"
categ <- subst(cover, from = from, to = to)
freq(cover) # original frequencies of each land cover class
freq(categ) # note change in frequency of "from" and "to" categories
plot(c(cover, categ))

# Substitute using level name, replace with NEW level label
from <- c("Mosaic crops", "Mosaic cropland/vegetation")
to <- c("Mixed cropland")
categ <- subst(cover, from = from, to = to)
freq(cover) # original frequencies of each land cover class
freq(categ) # note change in frequency of "from" and "to" categories
plot(c(cover, categ))

# Substitute using level name, replace with NEW level label
from <- c("Mosaic crops", "Mosaic cropland/vegetation")
to <- c("Mixed cropland", "Mixed cropland/vegetation")
categ <- subst(cover, from = from, to = to)
freq(cover) # original frequencies of each land cover class
freq(categ) # note change in frequency of "from" and "to" categories
plot(c(cover, categ))

# Substitute using level name, replace with VALUE of an existing label
from <- c("Mosaic crops", "Mosaic cropland/vegetation")
to <- 120
categ <- subst(cover, from = from, to = to)
freq(cover) # original frequencies of each land cover class
freq(categ) # note change in frequency of "from" and "to" categories
plot(c(cover, categ))

# Substitute using level name, replace with new level name, replace all others
from <- c("Mosaic crops", "Mosaic cropland/vegetation")
to <- "Crops"
categ <- subst(cover, from = from, to = to, others = "Other")
freq(cover) # original frequencies of each land cover class
freq(categ) # note change in frequency of "from" and "to" categories
plot(c(cover, categ))

}

if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Example data
madChelsa <- fastData("madChelsa")

# Convert a SpatRaster to a GRaster
chelsa <- fast(madChelsa)
chelsa # 4 layers

# Central tendency
mean(chelsa)
mmode(chelsa)
median(chelsa)

# Statistics
nunique(chelsa)
sum(chelsa)
count(chelsa)
min(chelsa)
max(chelsa)
range(chelsa)
skewness(chelsa)
kurtosis(chelsa)

stdev(chelsa)
stdev(chelsa, pop = FALSE)
var(chelsa)
varpop(chelsa)

# Which layers have maximum/minimum?
which.min(chelsa)
which.max(chelsa)

# Regression

# Note the intercept is different for fasterRaster::regress().
regress(chelsa)
regress(madChelsa, 1:nlyr(madChelsa))

# Note: To get quantiles for each layer, use global().
quantile(chelsa, 0.1)

# NAs
madForest2000 <- fastData("madForest2000")
forest2000 <- fast(madForest2000)
forest2000 <- project(forest2000, chelsa, method = "near")

chelsaForest <- c(chelsa, forest2000)

nas <- anyNA(chelsaForest)
plot(nas)

allNas <- allNA(chelsaForest)
plot(allNas)

}

#' Plot a histogram of raster values
#'
#' @description This function creates a histogram of values in `GRaster`. The function is modeled after [graphics::hist()], but actually uses [graphics::barplot()].
#'
#' @param x A `GRaster`.
#'
#' @param layer Character, numeric, or integer: Indicates which layer of a multi-layer `GRaster` for which to plot a histogram. The layer can be identified using its [name()] (character) or index (numeric or integer). If this is missing, then up to `maxnl` layers are plotted.
#'
#' @param maxnl Maximum number of layers for which to create histograms. This is 16 by default, but ignored if `layer` is defined.
#'
#' @param bins Positive numeric integer: Number of bins in which to divide values of a raster with continuous values. For `integer` and categorical rasters, each value is tallied.
#'
#' @param freq Logical: If `TRUE` (default), plot the frequency of values. If `FALSE`, plot the density of values (i.e., the number in each bin divided by the total number of cells with non-`NA` values).
#'
#' @param ... Arguments to pass to [graphics::barplot()].
#'
#' @returns A named list of `data.frame`s (invisibly), one per layer plotted, and creates a graph.
#'
#' @example man/examples/ex_plot.r
#'
#' @importFrom graphics par
#'
#' @aliases hist
#' @rdname hist
#' @exportMethod hist
methods::setMethod(
	f = "hist",
	signature = c(x = "GRaster"),
	function(x, layer, maxnl = 16, bins = 30, freq = TRUE, ...) {

	if (!missing(layer)) {
		layer <- .layerIndex(layer = layer, x = x)
	} else {
		layer <- seq(1L, min(maxnl, nlyr(x)))
	}
	numLayers <- length(layer)
	
	x <- x[[layer]]

	freqs <- freq(x, bins = bins)
	if (length(layer) == 1L) freqs <- list(freqs)

	# plot
	ncols <- ceiling(sqrt(numLayers))
	nrows <- ceiling(numLayers / ncols)
	pars <- graphics::par()
	on.exit(graphics::par(pars), add = TRUE)

	graphics::par(mfrow = c(nrows, ncols))
	for (i in seq_along(layer)) {
	
		if (!inherits(freqs[[i]], "data.table")) freqs[[i]] <- as.data.table(freqs[[i]])

		if (is.factor(x)[i]) {
			xs <- levels(x)[[i]][[2L]]
			las <- 3
			xlab <- ""
		} else if (is.cell(x[[i]])) {
   			xs <- freqs[[i]][["value"]]
			las <- 0
			xlab <- "Value"
		} else {

			TEMPTEMP_mean_ <- NULL
   			data.table::setDT(freqs[[i]])[ , TEMPTEMP_mean_ := rowMeans(.SD, na.rm = TRUE), .SDcols = c("from", "to")]

			xs <- freqs[[i]][["TEMPTEMP_mean_"]]
			las <- 0
			xlab <- "Mean bin value"
		
		}
		
		ys <- freqs[[i]][["count"]]
		if (!freq) ys <- ys / sum(ys)
		
		ylab <- if (freq) { "Frequency"} else { "Density" }
		
		graphics::barplot(ys, main = names(x)[i], xlab = xlab, ylab = ylab, names.arg = xs, las = las, ...)
	
	}
	
	invisible(freqs)
	
	} # EOF
)

if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")
elev <- fast(madElev)

# Calculate flow accumulation and watershed basins
water <- flow(elev, return = c("accum", "basins"))
water

elevWater <- c(elev, water)
plot(elevWater)

}

#' Classes for fasterRaster sessions, regions, rasters, and vectors
#'
#' @description The `G`-suite of S4 classes contain pointers to **GRASS** objects or metadata about the current **GRASS** session. Most users will manipulate objects using these classes, but do not need to know the details.
#'
#' * The `GLocation` class stores information about the **GRASS** "project"/"location"(see `vignette("projects_mapsets", package = "fasterRaster")`), and coordinate reference system. Contained by all the rest.
#'
#' * The `GSpatial` class contains the `GLocation` class and stores information about spatial objects (extent, topology) plus the name of the file representing it in **GRASS** (its `source`). Contained by `GRegion`, `GRaster`, and `GVector`.
#'
#' * The `GRegion` class contains the `GSpatial` class and stores information about grids (dimensions and resolution). They do have `sources`, but these are not used (they're always `NA`). Contained by `GRaster`. The `GRegion` corresponds to **GRASS** "regions", though `GRegion` objects are not actually pointers to **GRASS** "region" files (see `vignette("regions", package = "fasterRaster")`).
#'
#' * The `GRaster` class contains the `GRegion` class and represents rasters. It stores information on number of layers, categories, min/max values, and user-friendly names. Categorical `GRaster`s are associated with a "levels" table for representing categorical data (e.g., wetlands, forest, etc.).
#'
#' * The `GVector` class contains the `GSpatial` class and represents spatial vectors. It may or may not have an associated `data.table` (i.e., a `data.frame`), which contains metadata about each geometry in the vector.
#'
#' @slot location	Character (all classes): The **GRASS** "project"/"location" of the object. The default value is `default`. Can be obtained using the hidden function `.location()`. See `vignette("projects_mapsets", package = "fasterRaster")`.
#'
#' @slot mapset		Character (all classes): The **GRASS** "mapset". Default value is `PERMANENT`. Typically hidden to users. Can be obtained using the hidden function `.mapset()`. See `vignette("projects_mapsets", package = "fasterRaster")`.
#'
#' @slot workDir	Character (all classes): Directory in which **GRASS** stores files.
#'
#' @slot topology	Character (`GSpatial` objects, including `GRegion`s, `GRaster`s, and `GVector`s): Valid values are `2D` (2-dimensional--most rasters and vectors) or `3D` (3-dimensional--e.g., LIDAR data). Can be obtained using [topology()].
#'
#' @slot sources		Character (`GRaster`s and `GVector`s): Name of the object in **GRASS**. These are typically made on-the-fly and provide the pointer to the object from **R** to **GRASS**. Changing them manually will break the connection. Can be obtained using [sources()].
#'
#' @slot names		Character  (`GRaster`s only): Name of a raster or each raster layer in. Can be obtained using [names()].
#'
#' @slot crs		Character (all classes): Coordinate reference systems string (preferably in WKT2 format). Can be obtained using [crs()] or [st_crs()].
#'
#' @slot projection Character: The **GRASS** "projection" for a `GRaster` or `GVector`. Can be obtained using `.projection()`.
#'
#' @slot dimensions	Dimensions:
#' * `GRegion`s and `GRaster`s: Vector of three integers indicating number of rows, columns, and depths (for 3D objects). Can be obtained using [dim()], plus [nrow()], [ncol()], and [ndepth()].
#' * `GVectors`s: Vector of two integers indicating number of geometries and number of fields. Can be obtained using [dim()], plus [nrow()] and [ncol()].
#'
#' @slot extent		Numeric vector with four values (`GSpatial` objects, including `GRegion`s, `GRaster`s, and `GVector`s): Extent of the object listed in order from westernmost longitude, easternmost longitude, southernmost latitude, northernmost latitude. Can be obtained using [ext()].
#'
#' @slot zextent	Numeric (`GSpatial` objects, including `GRegion`s, `GRaster`s, and `GVector`s): Bottom- and top-most extents of 3D `GRaster`s and `GVector`s. Can be obtained using [zext()].
#'
#' @slot geometry	Character (`GVectors`s): Either `points`, `lines`, or `polygons`. Can be obtained using [geomtype()].
#'
#' @slot nLayers Integer (`GRaster`s): Number of layers ("stacked" rasters--different from number of depths of 3D rasters). Can be obtained using [nlyr()].
#'
#' @slot nGeometries Integer (`GVector`s): Number of features (points, lines, or polygons). Can be obtained using [nrow()].
#'
#' @slot datatypeGRASS Character (`GRaster`s): Type of data stored in a raster, as interpreted by `GRASS`. This is either `CELL` (integers), `FCELL` (floating-point values), or `DCELL` (double-values). Can be obtained using [datatype()].
#'
#' @slot resolution	Vector of two numeric values (`GRegion`s, including `GRaster`s): Size of a raster cell in the east-west direction and in the north-south direction. Can be obtained using [res()] and [res3d()].
#'
#' @slot minVal,maxVal Numeric (`GRaster`s): Minimum and maximum value across all cells. Can be obtained using [minmax()].
#'
#' @slot activeCat Integer (`GRaster`s): Column index of the category labels. Must be >0. Note that from the user's standpoint, 1 is subtracted from this number. So a value if `@activeCat` is `2`, then the user would see "1" when printed. Can be obtained using [activeCat()].
#'
#' @slot levels List of `data.table`s (`GRaster`s): Tables for categorical rasters. If a raster is not categorical, the `data.table` is `NULL`, as in `data.table(NULL)`. Can be obtained using `levels()` or `cats()`.
#'
#' @slot table `data.table` (`GVector`s): Table with metadata, one row per geometry (point, line, or plane). If no table is associated with the vector, this must be `data.table(NULL)`. The column with the category value is given in `@catName`.
#'
#' @slot catName Character (`GVector`s): Name of the column in the vector's database that contains category values (integers).
#'
#' @return An object of class `GLocation`, `GSpatial`, `GRegion`, `GRaster`, or `GVector`.
#'
#' @aliases GLocation
#' @rdname GLocation
#' @exportClass GLocation
GLocation <- methods::setClass(
	Class = "GLocation",
	slots = list(
		location = "character",			# GRASS location
		mapset = "character",			# GRASS MAPSET
		crs = "character",				# coordinate reference system
		workDir = "character" 			# working directory
	),
	prototype = prototype(
		location = NA_character_,
		mapset = NA_character_,
		crs = NA_character_,
		workDir = NA_character_
	)
)

setValidity("GLocation",
	function(object) {
		if (length(object@location) != 1L) {
			"@location can only be a single character string."
		} else if (length(object@mapset) != 1L) {
			"@mapset can only be a single character string."
		} else if (length(object@crs) != 1L) {
			"@crs can only be a single character string."
		} else if (length(object@workDir) != 1L) {
			"@workDir can only be a single character string."
		} else {
			TRUE
		}
	} # EOF
)

# GLocation <- function(
	# location = NA_character_,
	# mapset = NA_character_,
	# crs = NA_character_
# ) {
	# methods::new("GLocation", location = location, mapset = mapset, crs = crs)
# }

#' Values in a categorical raster with no assigned category
#'
#' @description This function reports the values in a categorical `GRaster` that have no matching category label in its "levels" table.
#'
#' `GRaster`s can represent categorical data. Cell values are actually integers, each corresponding to a category, such as "desert" or "wetland." A categorical raster is associated with a table that matches each value to a category name.
#'
#' @param x A `GRaster`.
#'
#' @param layer Numeric integers, logical vector, or character: Layer(s) for which to obtain missing categories.
#' 
#' @returns A numeric vector (if `x`is just one layer), or a named list of numeric vectors, one per layer in `x`.
#'
#' @seealso [missingCats()], [missing.cases()], [droplevels()], `vignette("GRasters", package = "fasterRaster")`
#'
#' @example man/examples/ex_GRaster_categorical.r
#'
#' @aliases missingCats
#' @rdname missingCats
#' @exportMethod missingCats
methods::setMethod(
    f = "missingCats",
    signature = c(x = "GRaster"),
    function(x, layer = 1:nlyr(x)) {
    
    layer <- .layerIndex(layer, x, recycle = TRUE)

    levs <- levels(x)

    out <- list()
    for (i in layer) {

        if (!is.factor(x)[i]) {
            out[[i]] <- numeric()
        } else {
            
            this <- .freq(x[[i]], dtype = "CELL")
            this <- this[count == 0]
            this <- this[[3L]]
            out[[i]] <- this

        } # if this layer has levels

    } # next raster
    if (length(out) == 1L) {
        out <- out[[1L]]
    } else {
        names(out) <- names(x)[layer]
    }
    out

    } # EOF
)

if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Elevation raster, outline of a part of Madagascar, and rivers vector:
madElev <- fastData("madElev") # raster
madDypsis <- fastData("madDypsis") # points vector
madRivers <- fastData("madRivers") # lines vector
madCoast4 <- fastData("madCoast4") # polygons vector

# Convert to GRaster and GVectors:
elev <- fast(madElev)
dypsis <- fast(madDypsis)
coast4 <- fast(madCoast4)
rivers <- fast(madRivers)

# Convert points, line, and polygons vectors to rasters:
points <- rasterize(dypsis, elev)
plot(points)

lines <- rasterize(rivers, elev)
plot(lines)

polys <- rasterize(coast4, elev)
plot(polys)

communes <- rasterize(coast4, elev, field = "NAME_4")
plot(communes)

# Change background value:
polysNeg1 <- rasterize(coast4, elev, background = -1)
plot(polysNeg1)

# Make one layer per river:
byRiver <- rasterize(rivers, elev, field = "NAM", by = "NAM")
plot(byRiver)

}

#' Delete a "GRASS" "location"
#'
#' @description This function deletes a **GRASS** "project"/"location", rasters, and vectors therein. This function should be used cautiously and is mainly of use to developers. See `vignette("projects_mapsets", package = "fasterRaster")`.
#'
#' @param location Character: Name of the **GRASS** "location".
#' @param mapset Character or `NULL` (default): Name of the mapset to delete. If `NULL`, then all mapsets in the given "location" will be deleted.
#' @param workDir Character: Either `NULL` (default) or a character string of the directory in which the location to be removed resides. If `NULL`, then the working directory will be obtained from `faster("workDir")`.
#'
#' @returns If successful, `TRUE` (invisibly). If not, then `FALSE` (also invisibly), plus a warning.
#'
#' @rdname location
#' @keywords internal
.locationDelete <- function(
	location,
	mapset = NULL,
	workDir = NULL
) {

	if (is.null(workDir)) workDir <- faster("workDir")

	path <- if (is.null(mapset)) {
		file.path(workDir, location)
	} else {
		file.path(workDir, location, mapset)
	}

	if (!file.exists(path)) {
		warning("The specified <location>, <mapset>, and/or <workDir> do not exist.")
		out <- FALSE
	} else {

		rgrass::unset.GIS_LOCK()
		rgrass::remove_GISRC()
		rgrass::unlink_.gislock()

		files <- list.files(path, include.dirs = TRUE, full.names = TRUE, recursive = TRUE)
		files <- rev(files)
		unlink(files, recursive = TRUE)
		unlink(path, recursive = TRUE)

		out <- TRUE

		# remove location from set of available ones
		if (is.null(mapset)) {
		
			index <- .locationFind(location, match = "name")
	
			.fasterRaster[[index]] <- NULL
			
			if (.location() == location) .fasterRaster$activeLocation <- NA_character_
	
		}

	}

	invisible(out)

}

#' Center and scale a GRaster, or the opposite
#'
#' @description `scale()` and `scalepop()` center and scale layers in a `GRaster` by subtracting from each raster its mean value (centering), then dividing by its standard deviation (scaling). This is useful for using the raster in a linear model, for example, because unscaled predictors can lead to numerical instability. The `scale()` function uses the sample standard deviation, and the `scalepop()` function uses the population standard deviation. For even moderately-sized rasters, the difference between these two is negligible, but the `scalepop()` function can be much faster than the `scale()` function.
#'
#' The `unscale()` function does the opposite of `scale()` and `scalepop()`: it multiples each layer by a value (presumably, its standard deviation), and adds another value (presumably, its mean).
#'
#' @param x A `GRaster`.
#'
#' @param center Value depends on the function:
#' * `scale()`:
#'     * Logical: If `TRUE` (default), subtract from each raster layer its mean. If `FALSE`, do not.
#'     * Numeric: A single value, in which case the same value will be used across all layers of `x`, or one value per layer in `x`.
#' * `unscale()`: Numeric vector or `NULL` (default): This can be a single value, which will be recycled if there is more than one layer in the raster, or one value per raster layer. If a value is `NA`, then no un-centering will be performed on the relevant raster layer. If `NULL`, then no un-centering is done.
#'
#' @param scale Value depends on the function:
#' * `scale()`:
#'     * Logical: If `TRUE` (default), divide each raster layer by its standard deviation. If `FALSE`, do not.
#'     * Numeric: A single value, in which case the same value will be used across all layers of `x`, or one value per layer in `x`.
#' * `unscale()`: Numeric vector or `NULL` (default): This can be a single value, which will be recycled if there is more than one layer in the raster, or one value per raster layer. If a value is `NA`, then no unscaling will be done on the relevant raster layer. If `NULL`, then no un-scaling is done.
#'
#' @returns All functions return a `GRaster`. The output of `scale()` and `scalepop()` will have two attributes, "center" and "scale", which have the means and standard deviations of the original rasters (if `center` and `scale` are `TRUE`, otherwise, they will be `NA`). These can be obtained using `attributes(output_raster)$center` and `attributes(output_raster)$scale`.
#'
#' @example man/examples/ex_scale_unscale.r
#'
#' @aliases scale
#' @rdname scale
#' @exportMethod scale
methods::setMethod(
	f = "scale",
	signature = c(x = "GRaster"),
	function(x, center = TRUE, scale = TRUE) {

	sample <- TRUE
	.scale(x, center = center, scale = scale, sample = sample)

	} # EOF
)

#' @aliases scalepop
#' @rdname scale
#' @exportMethod scalepop
methods::setMethod(
	f = "scalepop",
	signature = c(x = "GRaster"),
	function(x, center = TRUE, scale = TRUE) {

	sample <- FALSE
	.scale(x, center = center, scale = scale, sample = sample)

	} # EOF
)

#' @param x `GRaster`
#' @param center,scale Logical or numeric
#' @param sample Logical
#'
#' @noRd
.scale <- function(x, center, scale, sample) {

	.locationRestore(x)
	.region(x)

	nLayers <- nlyr(x)

	#### calculate centers and scales then center and scale
	#######################################################
	if (is.logical(center) & is.logical(scale)) {

		if (!center & !scale) {
			warning("No scaling performed because neither `center` nor `scale` are TRUE.")
			return(x)
		}
		
		if (center) {
			fx <- "mean"
		} else {
			fx <- NULL
		}

		if (scale) {
			if (sample) {
				sdfx <- "sd"
			} else{
				sdfx <- "sdpop"
			}
		} else {
			sdfx <- NULL
		}
		fx <- c(fx, sdfx)

		stats <- global(x, fx)

		srcs <- .makeSourceName("scale_r_mapcalc", "raster", n = nLayers)
		for (i in seq_len(nLayers)) {
		
			if (center) mu <- stats[i, "mean"]
			if (scale) sigma <- stats[i, sdfx]

			if (center & scale) {
				ex <- paste0(srcs[i], " = (", sources(x)[i], " - ", mu, ") / ", sigma)
			} else if (center & !scale) {
				ex <- paste0(srcs[i], " = ", sources(x)[i], " - ", mu)
			} else if (!center & scale) {
				ex <- paste0(srcs[i], " = ", sources(x)[i], " / ", sigma)
			}

			rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))
		
		} # next layer
		out <- makeGRaster(srcs, names(x))

		if (center) {
			vals <- stats[ , "mean"]
			names(vals) <- names(x)
			attr(out, "center") <- vals
		} else {
			attr(out, "center") <- NA_real_
		}

		if (scale) {
			vals <- stats[ , sdfx]
			names(vals) <- names(x)
			attr(out, "scale") <- vals
		} else {
			attr(out, "scale") <- NA_real_
		}

	### user provides centers but not scales
	########################################
	} else if (!is.logical(center) & is.logical(scale)) {
	
		len <- length(center)
		if (len != nLayers) {
			if (len == 1) {
				center <- rep(center, nLayers)
				warning("Using the same center for all rasters.")
			} else {
				stop("Argument `center` must be TRUE, FALSE, a single numeric value, or have the same number of numeric values as `x` has layers.")
			}
		}
		
		if (scale) {
			if (sample) {
				sdfx <- "sd"
			} else{
				sdfx <- "sdpop"
			}
			stats <- global(x, sdfx)
		}

		nLayers <- nlyr(x)
		srcs <- .makeSourceName("scale_r_mapcalc", "raster", n = nLayers)
		for (i in seq_len(nLayers)) {
		
			mu <- center[i]
			if (scale) sigma <- stats[i, sdfx]

			if (scale) {
				ex <- paste0(srcs[i], " = (", sources(x)[i], " - ", mu, ") / ", sigma)
			} else if (!scale) {
				ex <- paste0(srcs[i], " = ", sources(x)[i], " - ", mu)
			}

			rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))
		
		} # next layer
		out <- makeGRaster(srcs, names(x))

		names(center) <- names(x)
		attr(out, "center") <- center

		if (scale) {
			vals <- stats[ , sdfx]
			names(vals) <- names(x)
			attr(out, "scale") <- vals
		} else {
			attr(out, "scale") <- NA_real_
		}
	
	### calculate centers but user supplied scales
	##############################################
	} else if (is.logical(center) & !is.logical(scale)) {
	
		len <- length(scale)
		if (len != nLayers) {
			if (len == 1) {
				scale <- rep(scale, len)
				warning("Using the same scale for all rasters.")
			} else {
				stop("Argument `scale` must be TRUE, FALSE, a single numeric value, or have the same number of numeric values as `x` has layers.")
			}
		}

		if (center) {
			fx <- "mean"
			stats <- global(x, fx)
		}

		srcs <- .makeSourceName("scale_r_mapcalc", "raster", n = nLayers)
		for (i in seq_len(nLayers)) {
		
			if (center) mu <- stats[i, fx]
			sigma <- scale[i]

			if (center) {
				ex <- paste0(srcs[i], " = (", sources(x)[i], " - ", mu, ") / ", sigma)
			} else if (!center) {
				ex <- paste0(srcs[i], " = ", sources(x)[i], " / ", sigma)
			}

			rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		} # next layer

		if (center) {
			vals <- stats[ , "mean"]
			names(vals) <- names(x)
			attr(out, "center") <- vals
		} else {
			attr(out, "center") <- NA_real_
		}

		names(scale) <- names(x)
		attr(out, "scale") <- scale

	### user supplied centers and scales
	####################################
	} else if (!is.logical(center) & !is.logical(scale)) {
	
		len <- length(center)
		if (len != nLayers) {
			if (len == 1) {
				center <- rep(center, nLayers)
				warning("Using the same center for all rasters.")
			} else {
				stop("Argument `center` must be TRUE, FALSE, a single numeric value, or have the same number of numeric values as `x` has layers.")
			}
		}

		len <- length(scale)
		if (len != nLayers) {
			if (len == 1) {
				scale <- rep(scale, nLayers)
				warning("Using the same scale for all rasters.")
			} else {
				stop("Argument `scale` must be TRUE, FALSE, a single numeric value, or have the same number of numeric values as `x` has layers.")
			}
		}

		srcs <- .makeSourceName("scale_r_mapcalc", "raster", n = nLayers)
		for (i in seq_len(nLayers)) {
		
			mu <- center[i]
			sigma <- scale[i]

			ex <- paste0(srcs[i], " = (", sources(x)[i], " - ", mu, ") / ", sigma)
			rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))
		
		} # next layer
		out <- makeGRaster(srcs, names(x))

		names(center) <- names(x)
		names(scale) <- names(x)
		attr(out, "center") <- center
		attr(out, "scale") <- scale

	}
	out

} # EOF

#' @aliases unscale
#' @rdname scale
#' @exportMethod unscale
methods::setMethod(
	f = "unscale",
	signature = c(x = "GRaster"),
	function(x, center = NULL, scale = NULL) {

	if (is.null(center) & is.null(scale)) {
		warning("No unscaling performed because neither `center` `scale` are NULL.")
		return(x)

	}
	
	.locationRestore(x)
	.region(x)

	nLayers <- nlyr(x)

	if (!is.null(center)) {

		if (length(center) == 1L) center <- rep(center, nLayers)
		if (length(center) != nLayers) stop("The `center` argument must be a single value, one value per layer in the GRaster, or NULL.")

	}

	if (!is.null(scale)) {

		if (length(scale) == 1L) scale <- rep(scale, nLayers)
		if (length(scale) != nLayers) stop("The `scale` argument must be a single value, one value per layer in the GRaster, or NULL.")

	}

	srcs <- .makeSourceName("r_mapcalc", "raster", n = nLayers)
	for (i in seq_len(nLayers)) {
	
		if (!is.null(center)) mu <- center[i]
		if (!is.null(scale)) sigma <- scale[i]

		if (!is.null(center) & !is.null(scale)) {

			if (!is.na(mu) & !is.na(sigma)) {
				ex <- paste0(srcs[i], " = (", sources(x)[i], " * ", sigma, ") + ", mu)
			} else if (is.na(mu) & !is.na(sigma)) {
				ex <- paste0(srcs[i], " = ", sources(x)[i], " * ", sigma)
			} else if (!is.na(mu) & is.na(sigma)) {
				ex <- paste0(srcs[i], " = ", sources(x)[i], " + ", mu)
			} else if (is.na(mu) & is.na(sigma)) {
				ex <- NULL
			}

		} else if (is.null(center) & !is.null(scale)) {
		
			if (!is.na(sigma)) {
				ex <- paste0(srcs[i], " = ", sources(x)[i], " * ", sigma)
			} else {
				ex <- NULL
			}

		} else if (!is.null(center) & is.null(scale)) {

			if (!is.na(mu)) {
				ex <- paste0(srcs[i], " = ", sources(x)[i], " + ", mu)
			} else {
				ex <- NULL
			}

		}

		if (!is.null(ex)) {

			rgrass::execGRASS(
				cmd = "r.mapcalc",
				expression = ex,
				flags = c(.quiet(), "overwrite")
			)

		} else {
			srcs[i] <- sources(x)[i]
		}
	
	} # next layer
	makeGRaster(srcs, names(x))

	} # EOF
)

#' Test if a coordinate reference system is unprojected
#'
#' @description `is.lonlat()` attempts to determine if a coordinate reference system is unprojected (e.g., WGS84, NAD83, NAD27, etc.). For `GRaster`s and `GVector`s, the function should always be correct. For WKT character strings and `sf` vectors, it does this by looking for the "CONVERSION[" tag in the WKT string (or the object's WKT string), and if it finds one, returns `FALSE`. This may not be truthful in all cases.
#'
#' @param x A WKT coordinate reference string or an object from which on can be obtained (e.g., a `GRaster`, `GVector`, `GRegion`, `GLocation`, `SpatRaster`, `SpatVector`, or `sf` object).
#'
#' @returns Logical (`TRUE` if unprojected, `FALSE` otherwise).
#'
#' @seealso [terra::is.lonlat()]
#'
#' @aliases is.lonlat
#' @rdname is.lonlat
#' @exportMethod is.lonlat
methods::setMethod(
    f = "is.lonlat",
    signature = c(x = "character"),
    function(x) grepl(x, pattern = "CONVERSION[")[[1L]]
)

#' @aliases is.lonlat
#' @rdname is.lonlat
#' @exportMethod is.lonlat
methods::setMethod(
    f = "is.lonlat",
    signature = c(x = "GLocation"),
    function(x) .projection(x) == "Latitude-Longitude"
)

#' @aliases is.lonlat
#' @rdname is.lonlat
#' @exportMethod is.lonlat
methods::setMethod(
    f = "is.lonlat",
    signature = c(x = "sf"),
    function(x) is.lonlat(crs(x))
)

#' Combine red, green, and blue color bands to make a composite GRaster
#'
#' @description This function takes as arguments three rasters typically representing red, green, and blue color bands, and returns a single raster with values based on their combination. Typically, this raster should be plotted in grayscale.
#'
#' @param r,g,b Either:
#' * One `GRaster` with one band each for `r`, `g`, or `b` representing red, green, and blue color bands; or
#' * `r` is  single `GRaster` with 3 bands (R, G, and B bands), and `g` and `b` are `NULL`.
#'
#' @param levels Either a single value that is an integer, or a vector of integers: Number of levels of red, green, and blue intensities represented in `r`, `g`, and `b`. If a single value is supplied, it is assumed that all three have the same number of levels. If three values are supplied, then they correspond to the R, G, and B bands. The default is 256 (assume that R, G, and B rasters have values between 0 and 255).
#'
#' @param dither Logical: If `TRUE`, apply Floyd-Steinberg dithering. Default is `FALSE`.
#'
#' @returns A `GRaster`.
#'
#' @example man/examples/ex_plot.r
#'
#' @seealso [plotRGB()], [terra::plotRGB()], **GRASS** manual page for tool `r.composite` (see `grassHelp("r.composite")`)
#'
#' @aliases compositeRGB
#' @rdname compositeRGB
#' @exportMethod compositeRGB
methods::setMethod(
	f = "compositeRGB",
	signature(r = "GRaster"),
	function(r, g = NULL, b = NULL, levels = 256, dither = FALSE) {

	msg <- paste0("Argument `r` must have 1 band (in which case arguments `g` and `b` must also be single-layer GRasters),\n  or `r` must have 3 bands (and `g` and `b` must be NULL).")

	if (!(nlyr(r) %in% c(1L, 3L))) stop(msg)

	if (nlyr(r) == 3L) {
 		if (!is.null(g) | !is.null(b)) stop(msg)
		g <- r[[2L]]
		b <- r[[3L]]
		r <- r[[1L]]
	} else if (nlyr(r) == 1L) {
		if (nlyr(g) != 1L | nlyr(b) != 1L) stop(msg)
	}

	compareGeom(r, g)
	compareGeom(r, b)
	.locationRestore(r)
	.region(r)

	if (!(length(levels %in% c(1L, 3L)))) stop("Argument `levels` must have 1 or 3 values.")
	if (length(levels) == 1L) levels <- rep(levels, 3L)

 	src <- .makeSourceName("r_composite", "raster")
	args <- list(
		cmd = "r.composite",
		red = sources(r),
		green = sources(g),
		blue = sources(b),
		level_red = levels[1L],
		level_green = levels[2L],
		level_blue = levels[3L],
		output = src,
		flags = c(.quiet(), "overwrite")
	)

	if (dither) args$flags <- c(args$flags, "d")

	do.call(rgrass::execGRASS, args = args)
	makeGRaster(src, "compositeRGB")
	
	} # EOF
)

#' Create one GRaster layer per unique value in a GRaster
#'
#' @description This function creates a multi-layered `GRaster` for every unique values in an input `GRaster`. By default, the output will have a value of 1 wherever the input has the given value, and 0 elsewhere. This is useful for creating dummy variable `GRaster` layers for use with models that have factors, especially if the input `GRaster` is categorical. Note that the [predict()] function in **fasterRaster** usually does not need this treatment of `GRaster`s since it can handle categorical rasters already.
#'
#' @param x A `GRaster`.
#'
#' @param classes Either `NULL` (default) or a character vector with category labels for which to create outputs. If the input is not a categorical/factor or `integer` `GRaster`, this is ignored.
#'
#' @param keep Logical: If `FALSE` (default), then the original value in the input `GRaster` will be retained in the each of the output `GRaster` layers wherever the input had the respective value. Other cells will be assigned a value of `other`.
#'
#' @param other Numeric or `NA`: Value to assign to cells that do not have the target value.
#'
#' @param bins Numeric: Number of bins in which to put values. This is only used for `GRaster`s that are not categorical/factor rasters or `integer` rasters.
#'
#' @param digits Numeric: Number of digits to which to round input if it is a `numeric` or `double` `GRaster` (see `vignettes("GRasters", package = "fasterRaster")`).
#'
#' @returns If the input `x` is a single-layered `GRaster`, the output will be a multi-layered `GRaster` with one layer per value in the input, or one layer per values in `classes`. If the input is a multi-layered `GRaster`, the output will be a `list` of multi-layered `GRaster`s.
#'
#' @example man/examples/ex_segregate.r
#'
#' @seealso [terra::segregate()]
#'
#' @aliases segregate
#' @rdname segregate
#' @exportMethod segregate
methods::setMethod(
	f = "segregate",
	signature = c(x = "GRaster"),
	function(x, classes = NULL, keep = FALSE, other = 0, bins = 100, digits = 3) {

	.locationRestore(x)
	.region(x)

	if (!is.null(classes) & (all(is.factor(x)) | all(is.integer(x)))) x <- x[x %in% classes]
	if (is.na(other)) other <- "null()"

	out <- list()
	nLayers <- nlyr(x)
	for (i in seq_len(nLayers)) {
	
		# unique values
		dtype <- datatype(x, type = "GRASS")[i]
		freqs <- freq(x[[i]], digits = digits, bins = bins)
		freqs <- freqs[freqs$count > 0, ]
		vals <- freqs$value

		nVals <- length(vals)
		if (nVals == 0) {
			thisOut <- NA
		} else {
			
			thisSrcs <- .makeSourceName("segregate", "raster", nVals)
			for (j in seq_len(nVals)) {

				if (keep) {
					thisKeep <- vals[j]
				} else {
					thisKeep <- 1
				}
				ex <- paste0(thisSrcs[j], " = if(", sources(x)[i], " == ", vals[j], ", ", thisKeep, ", ", other, ")")
				rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

			}

			names <- if (is.factor(x)[i]) {
				freqs[[3]]
			} else {
				as.character(vals)
			}
			thisOut <- makeGRaster(thisSrcs, names = names)

		}
	
		out[[i]] <- thisOut
	
	}

	if (nLayers == 1L) {
		out <- out[[1L]]
	} else {
		names(out) <- names(x)
	}
	out
	
	} # EOF
)

if (grassStarted()) {

# Citation
grassInfo()

# Version number
grassInfo("version")

# Version number
grassInfo("versionNumber")

# Version number
grassInfo("versionNumber")

# Copyright
grassInfo("copyright")

}

if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Example data
madElev <- fastData("madElev") # raster

# Convert to GRasters and GVectors
elev <- fast(madElev)

### spatSample()

# Random points as data.frame or data.table:
randVals <- spatSample(elev, size = 20, values = TRUE)
randVals

# Random points as a points GVector:
randPoints <- spatSample(elev, size = 20, as.points = TRUE)
randPoints
plot(elev)
plot(randPoints, add = TRUE)

# Random points in a select area:
madCoast <- fastData("madCoast4") # vector
coast <- fast(madCoast)
ant <- coast[coast$NAME_4 == "Antanambe"] # subset

restrictedPoints <- spatSample(elev, size = 20, as.points = TRUE,
   strata = ant)

plot(elev)
plot(ant, add = TRUE)
plot(restrictedPoints, add = TRUE) # note 20 points for entire geometry

# Random points, one set per subgeometry:
stratifiedPoints <- spatSample(elev, size = 20, as.points = TRUE,
   strata = ant, byStratum = TRUE)

plot(elev)
plot(ant, add = TRUE)
plot(stratifiedPoints, pch = 21, bg = "red", add = TRUE) # note 20 points per subgeometry

# Random categories:
madCover <- fastData("madCover") # raster
cover <- fast(madCover)

randCover <- spatSample(cover, size = 20, values = TRUE,
     cat = TRUE, xy = TRUE)
randCover

### sampleRast()

# Random cells in non-NA cells:
rand <- sampleRast(elev, 10000)
plot(rand)
nonnacell(rand)

# Use custom values for the mask:
randCustomMask <- sampleRast(elev, 10000, maskvalues = 1:20)
plot(randCustomMask)

# Force selected values to a custom value:
randCustomUpdate <- sampleRast(elev, 10000, updatevalue = 7)
plot(randCustomUpdate)

# Custom values for mask and set selected cells to custom value:
randAll <- sampleRast(elev, 10000, maskvalues = 1:20, updatevalue = 7)
plot(randAll)

}

#' Interpolate values at points to a GRaster using inverse-distance weighting
#'
#' @description This function interpolates values from a set of points to a raster using inverse distance weighting (IDW).
#'
#' @param x A "points" `GVector`.
#'
#' @param y A `GRaster` to serve as a template for interpolation: Only points in `x` that fall inside the extent of the raster will be used for interpolation. You can increase the extent of a `GRaster` using [extend()].
#'
#' @param field Character, integer, or numeric integer: Name or index of the column in `x` with values to interpolate. If `NULL` and if `x` is a 3-dimensional "points" `GVector`, then the interpolation will act on the z-coordinate of each point.
#'
#' @param nPoints Integer or numeric integer: Number of nearest points to use for interpolation. The default is to use all points (`Inf`).
#'
#' @param power Numeric value > 0: Power to which to take distance when interpolating. The default value is two, so the value of each point used for interpolation is \eqn{1 / d^2} where *d* is distance.
#'
#' @returns A `GRaster`.
#'
#' @seealso [terra::interpIDW()], [interpSplines()], [fillNAs()], **GRASS** tool `v.surf.idw` (se `grassHelp("v.surf.idw")`)
#'
#' @aliases interpIDW
#' @rdname interpIDW
#' @exportMethod interpIDW
methods::setMethod(
	f = "interpIDW",
	signature = c(x = "GVector", y = "GRaster"),
	function(x, y, field, nPoints = Inf, power = 2) {

	if (geomtype(x, grass = TRUE) != "point") stop("Argument `x` must be a points GVector.")

	compareGeom(x, y)
	.locationRestore(x)
	.region(y)

	if (is.infinite(nPoints)) nPoints <- nsubgeom(x)

	# copy values field to x
	if (is.null(field) & is.2d(x)) {
		stop("Argument `field` cannot be NULL if `x` is a 2-dimensional points GVector.")
	} else if (is.numeric(field) | is.integer(field)) {
		field <- names(x)[field]
	}

	if (!is.null(field)) {

		if (anyNA(x@table[[field]])) stop("The column `field` has at least one NA in it. NAs are not permissible.")

		cats <- .vCats(x, db = FALSE)
		db <- data.table::data.table(frid = cats, TEMPTEMP_ = x@table[[field]])
		names(db)[2L] <- field
		.vAttachDatabase(x, table = db, replace = TRUE, cats = cats)

	}

	src <- .makeSourceName("interpIDW_v_surf_idw", "raster")

	rgrass::execGRASS(
		cmd = "v.surf.idw",
		input = sources(x),
		output = src,
		npoints = nPoints,
		column = field,
		power = power,
		flags = c(.quiet(), "overwrite")
	)
	makeGRaster(src, "interpolation")

	} # EOF
)

#' Aggregates or disaggregates a vector in GRASS
#'
#' @param src [sources()] name.
#' @param resolve "aggregate" or "disaggregate"
#' @param verbose T/F
#'
#' @returns [sources()] name.
#' @noRd
.aggDisaggVect <- function(src, resolve, verbose) {

	if (resolve == "disaggregate") {

		if (verbose) omnibus::say("Fixing invalid vector by disaggregating polygons. This will remove any data table.")

		# delete categories
		srcIn <- src
		src <- .makeSourceName("fast_v_category_del")
		rgrass::execGRASS(
			cmd = "v.category",
			input = srcIn,
			output = src,
			option = "del",
			cat = -1,
			flags = c(.quiet(), "overwrite")
		)

		# assign each subgeometry its own category
		srcIn <- src
		src <- .makeSourceName("fast_v_category_add")
		rgrass::execGRASS(
			cmd = "v.category",
			input = srcIn,
			output = src,
			option = "add",
			flags = c(.quiet(), "overwrite")
		)

	} else if (resolve == "aggregate") {

		if (verbose | faster("verbose")) omnibus::say("Fixing invalid vector by aggregating polygons. This will remove any data table.")

		srcIn <- src
		src <- .makeSourceName("fast_v_extract")
		rgrass::execGRASS(
			cmd = "v.extract",
			input = srcIn,
			output = src,
			new = 1L,
			flags = c(.quiet(), "overwrite", "d")
		)
	
	}

	src

}

#' Add a database table to a GRASS attribute table
#'
#' @description `.vAttachDatabase()` adds a table to a **GRASS** vector. This table is meant to be "invisible" to most users--they should use interact with attribute tables using the `GVector` slot `@table`. Some functions require tables (e.g., [extract()] and [spatSample()]). **This function is mostly of use to developers.**
#'
#' @param x A `GVector` or the name of a vector in **GRASS**.
#' 
#' @param table Either `NULL` (default), or a `data.frame` or `data.table`, or a numeric or integer vector:
#' * If `NULL`, then a bare minimal table will be created with a column named `cat`, holding sequential integer values.
#' * If a `data.frame` or `data.table` and no column is named `cat`, one will be created with sequential integer values. If the table does have a column named `cat`, then it should have integer (not just numeric) values.
#' * If a `vector`, then these are coerced to type `integer` and used to define the `cat` column.
#'
#' There should be one row/value per geometry in `x`.
#'
#' @param replace Logical: If `TRUE`, replace the existing database connection.
#'
#' @param cats Either `NULL` (default), or an integer vector: This is provided as a means to save time by passing `cats` to this function if it has already been generated by a calling function.
#' 
#' @returns Invisibly returns the [sources()] name of a vector in **GRASS**.
#' 
#' @aliases .vAttachDatabase
#' @rdname vAttachDatabase
#' @keywords internal
.vAttachDatabase <- function(x, table = NULL, replace = TRUE, cats = NULL) {

	if (inherits(x, "GVector")) {
		.locationRestore(x)
		src <- sources(x)
	} else {
		src <- x
	}

	if (replace || !.vHasDatabase(src)) {

		# if no table
		if (is.null(table)) {
			if (is.null(cats)) cats <- .vCats(src, db = FALSE, integer = TRUE)
			table <- data.frame(cat = cats)
		}

		# if table is a vector
		if (inherits(table, c("numeric", "integer"))) {
			table <- as.integer(table)
			table <- data.frame(cat = table)
		}

		# if table does not have a "cat" column
		if (!any("cat" %in% names(table))) {
		
			if (is.null(cats)) cats <- .vCats(src, db = FALSE, integer = TRUE)
			catsRenum <- omnibus::renumSeq(cats)
			table <- table[catsRenum, , drop = FALSE]
			cats <- data.frame(cat = cats)
			table <- cbind(cats, table)
		
		}

		# columns <- names(table)
		# classes <- sapply(table, "class")
		# for (i in seq_len(ncol(table))) {

		# 	if (classes[i] == "integer") {
		# 		columns[i] <- paste0(columns[i], " INTEGER")
		# 	} else if (classes[i] == "numeric") {
		# 		columns[i] <- paste0(columns[i], " DOUBLE PRECISION")
		# 	} else {
		# 		nc <- nchar(table[ , i])
		# 		nc <- max(nc)
		# 		columns[i] <- paste0(columns[i], " VARCHAR(", nc, ")")
		# 	}

		# }

		# save table to disk
		tf <- tempfile(fileext = ".csv")
		tft <- paste0(tf, "t")
		utils::write.csv(table, tf, row.names = FALSE)
		
		classes <- sapply(table, class)
		classes[!(classes %in% c("numeric", "integer", "character", "Date"))] <- '"String"'
		classes[classes == "numeric"] <- '"Real"'
		classes[classes == "integer"] <- '"Integer"'
		classes[classes == "character"] <- '"String"'
		classes[classes == "Date"] <- '"Date"'
		classes <- paste(classes, collapse = ",")
		
		write(classes, tft)

		# import table as database
		srcTable <- .makeSourceName("db_in_ogr_table", NULL)
		rgrass::execGRASS(
			cmd = "db.in.ogr",
			input = tf,
			output = srcTable,
			# key = "cat", # error
			flags = c(.quiet(), "overwrite")
		)

		# disconnect existing table
		if (.vHasDatabase(src)) {
		
			rgrass::execGRASS(
				cmd = "v.db.droptable",
				map = src,
				flags = c(.quiet(), "f")
			)
		
		}

		# connect database to vector
		args <- list(
			cmd = "v.db.connect",
			map = sources(x),
			table = srcTable,
			layer = "1",
			# key = "frid",
			key = "cat_", # adds an underscore, for some reason
			# flags = c(.quiet(), "overwrite", "o")
			flags = c(.quiet(), "overwrite")
		)

		if (grassInfo("versionNumber") <= 8.3) args$flags <- c(args$flags, "o")

		do.call(rgrass::execGRASS, args = args)

		# args <- list(
		# 	cmd = "v.db.addtable",
		# 	map = src,
		# 	columns = columns,
		# 	flags = .quiet(),
		# 	intern = TRUE
		# )

		# info <- do.call(rgrass::execGRASS, args = args)

	}
	invisible(x)

}

.onLoad <- function(lib, pkg) {

	.fasterRaster <<- new.env(parent = emptyenv())
	.fasterRaster$grassStarted <- FALSE
	.fasterRaster$grassVersion <- NA_real_
	.fasterRaster$locations <- list()
	.fasterRaster$activeLocation <- NA_character_
	.fasterRaster$messages <- list()
	.fasterRaster$options <- list()
	# .fasterRaster$maxSqlLength <- 29900L # maximum number of characters (with safety margin) of an SQL query GRASS can handle
	faster(restore = TRUE)
	
}

.onAttach <- function(lib, pkg) {

	.fasterRaster <<- new.env(parent = emptyenv())
	.fasterRaster$grassStarted <- FALSE
	.fasterRaster$grassVersion <- NA_real_
	.fasterRaster$locations <- list()
	.fasterRaster$activeLocation <- NA_character_
	.fasterRaster$messages <- list()
	.fasterRaster$options <- list()
	faster(restore = TRUE)
	
	ver <- read.dcf(file = system.file("DESCRIPTION", package = pkg), fields = "Version")
	packageStartupMessage(paste(pkg, ver))
	packageStartupMessage("To avoid conflicts between functions, please attach the `terra`, `sf`,")
	packageStartupMessage("and `data.table` packages before attaching `fasterRaster` using, for")
	packageStartupMessage("example, `library(terra)`.")
	packageStartupMessage("For guides and table of contents, see `?fasterRaster`.")
	
}

#' GRASS citation, version, and copyright information
#'
#' Report the **GRASS** citation, version/release year, version number, or copyright information.
#'
#' @param x Character: What to return. Any of:
#' * `"citation"` (default)
#' * `"copyright"`: Copyright information
#' * `"version"`: Version number and release year
#' * `"versionNumber"`: Version number as numeric, major and minor only (e.g., 8.4)
#'
#' Partial matching is used and case is ignored.
#' 
#' @return Character.
#'
#' @example man/examples/ex_grassInfo.r
#' 
#' @aliases grassInfo
#' @rdname grassInfo
#' @export grassInfo
grassInfo <- function(x = "citation") {
	
	match <- omnibus::pmatchSafe(x, table = c("citation", "version", "versionNumber", "copyright"), useFirst = TRUE, nmax = 1L)

	if (match == "citation") {
		out <- rgrass::execGRASS("g.version", flags = "x", intern = TRUE)
	} else if (match == "copyright") {
		out <- rgrass::execGRASS("g.version", flags = "c", intern = TRUE)
	} else if (match == "version") {
		out <- rgrass::execGRASS("g.version", intern = TRUE)
	} else if (match == "versionNumber") {

		out <- .fasterRaster$grassVersion
		if (is.na(out)) {

			out <- rgrass::execGRASS("g.version", intern = TRUE)
			out <- strsplit(out, split = " ")[[1L]]
			out <- out[2L]
			out <- substr(out, 1L, 3L)
			out <- as.numeric(out)

		}
	}
	out
}

if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Example data
madElev <- fastData("madElev") # raster
madForest <- fastData("madForest2000") # raster
madCoast <- fastData("madCoast4") # vector

# Convert to GRasters and GVectors
elev <- fast(madElev)
forest <- fast(madForest)
coast <- fast(madCoast)

ant <- coast[coast$NAME_4 == "Antanambe"]

# Mask by a raster or  vector:
maskByRast <- mask(elev, forest)
plot(c(forest, maskByRast))

maskByVect <- mask(elev, ant)
plot(maskByVect)
plot(ant, add = TRUE)

# Mask by a raster or vector, but invert mask:
maskByRastInvert <- mask(elev, forest, inverse = TRUE)
plot(c(forest, maskByRastInvert))

maskByVectInvert <- mask(elev, ant, inverse = TRUE)
plot(maskByVectInvert)
plot(ant, add = TRUE)

# Mask by a raster, but use custom values for the mask:
maskByRastCustomMask <- mask(elev, elev, maskvalues = 1:20)
plot(c(elev <= 20, maskByRastCustomMask))

# Mask by a raster or vector, but force masked values to a custom value:
byRastCustomUpdate <- mask(elev, forest, updatevalue = 7)
plot(byRastCustomUpdate)

byVectCustomUpdate <- mask(elev, ant, updatevalue = 7)
plot(byVectCustomUpdate)

# Mask by a raster, inverse, custom values, and custom update:
byRastAll <-
   mask(elev, elev, inverse = TRUE, maskvalues = 1:20, updatevalue = 7)

plot(c(elev, byRastAll))

}

if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Elevation raster
madElev <- fastData("madElev")

# Convert to GRaster:
elev <- fast(madElev)

### Simple sine waves:
waves <- sineRast(elev, ns = 2, ew = 1)
plot(waves)

### Sine waves with different amplitudes:
amps <- sineRast(elev, nsAmp = c(1, 5), ewAmp = c(1, 5))
amps

### Sine waves with and without north-south offset:
noOffsets <- sineRast(elev, ns = 1, ew = 1)
offsets <- sineRast(elev, ns = 1, ew = 1, nsOffset = 0.25)
offs <- c(noOffsets, offsets)
names(offs) <- c("no offset", "offset")
plot(offs)

### Masking:
madCoast4 <- fastData("madCoast4")
coast4 <- fast(madCoast4, verbose = FALSE)

masked <- sineRast(elev, mask = coast4)
plot(masked)

### Multiple sine waves (multiple rasters):
mults <- sineRast(elev, ns = 1:2, ew = 1:2)
combos <- sineRast(elev, ns = 1:2, ew = 1:2, combos = TRUE)
plot(mults)
plot(combos)

}

#' Does the "GRASS" representation of a GRaster or GVector exist?
#'
#' @description `GRaster`s and `GVector`s are **R** objects that contain a pointer to a raster or vector in **GRASS**. Thus, for a `GRaster` or `GVector` to be functional, the **GRASS** file must exist. This function indicates if that is so.
#'
#' @param x A `GRaster`, `GVector`, or the [sources()] name of one.
#'
#' @returns Logical.
#'
#' @aliases .exists
#' @rdname exists
#' @keywords internal
methods::setMethod(
	f = ".exists",
	signature = c(x = "GRaster"),
	function(x) {
	
	.locationRestore(x)
	src <- sources(x)
	src %in% .ls(type = "raster")
	
	} # EOF
)

#' @aliases .exists
#' @rdname exists
#' @keywords internal
methods::setMethod(
	f = ".exists",
	signature = c(x = "GVector"),
	function(x) {
	
	.locationRestore(x)
	src <- sources(x)
	src %in% .ls(type = "vector")
	
	} # EOF
)

#' @aliases .exists
#' @rdname exists
#' @keywords internal
methods::setMethod(
	f = ".exists",
	signature = c(x = "character"),
	function(x) x %in% .ls()
)

#' Delete objects in the active GRASS session
#'
#' @description Delete the names of all rasters and/or vectors that have been exported to or created in the active **GRASS** session's location and mapset.
#'
#' @param x Character, a `GSpatial` object, or missing (default):
#' * Character:
#'      * Any of `"rasters"` (all rasters), `"vectors"` (all spatial vectors), `"rasters3d"` (3D-rasters), and/or `"groups"` (groups), and all of these types will be deleted. Partial matching is supported.
#'      * The `sources` of the object to be deleted. Argument `type` must be specified.
#' * `GSpatial` object (i.e., a `GRaster` or `GVector`): Delete this object.
#' * Missing: Delete everything in the active **GRASS** session.
#'
#' @param type The type of spatial objects to delete. This can include `"rasters"` (all rasters), `"vectors"` (all spatial vectors), `"rasters3d"` (3D-rasters), and/or `"groups"` (groups). Partial matching is supported. If missing, all objects are candidates for deletion if they match `x`.
#'
#' @param warn Logical: If `TRUE` (default), display warning if no matches or if everything in **GRASS** is to be deleted.
#'
#' @param verify Logical: If `TRUE` (default), the function will search for the item(s) to be deleted first to verify they exist. If this is `FALSE`, then `x` MUST be specified and `type` must be '`raster`' or `'vector'` (one value per value in `x`). This has no effect if `x` is a `GSpatial` object. It's main use is to save a bit of time.
#'
#' @return `TRUE` (invisibly).
#'
#' @seealso [rm()]
#'
#' @aliases .rm
#' @rdname rm
#' @keywords internal
.rm <- function(x, type = NULL, warn = TRUE, verify = TRUE) {

	if (missing(x)) {
		if (warn) {
			ans <- readline(prompt = "Delete everything in the current GRASS session (Y/n)? ")
			if (ans != "Y") {
				message("Nothing deleted.")
				return(invisible(FALSE))
			}
		}
		dels <- .ls()
		type <- names(dels)
	} else if (inherits(x, c("GSpatial"))) {
		dels <- sources(x)
		type <- if (inherits(x, "GRaster")) {
			"raster"
		} else if (inherits(x, "GVector")) {
			"vector"
		}
	} else if (inherits(x, "character")) {
		
		if (verify) {
			
			dels <- .ls()
			dels <- dels[dels %in% x]
			if (length(dels) == 0L) {
				if (warn) warning("Object(s) are not in GRASS. Nothing deleted.")
				return(invisible(FALSE))
			}
			type <- names(dels)
		
		} else {
			dels <- x
			type <- rep(type, length(dels))
		}
	}

	for (i in seq_along(dels)) {
		
		del <- dels[i]
		typ <- type[i]
		
		rgrass::execGRASS(
			"g.remove",
			name = del,
			type = typ,
			flags = c(.quiet(), "f")
		)
	
	}
	
	invisible(TRUE)

}

#' "Stack" GRasters
#'
#' @description `GRaster`s can be "stacked" using this function, effectively creating a multi-layered raster. This is different from creating a 3-dimensional raster, though such an effect can be emulated using stacking. `GVector`s can be combined into a single vector.  Stacks can only be created when:
#' * All objects are the same class (either all `GRaster`s or all `GVector`s).
#' * All objects have the same coordinate reference system (see crs()).
#' * Horizontal extents are the same (see [ext()]).
#' * Horizontal dimensions are the same (see [res()]).
#' * The topology (2- or 3-dimensional) must be the same. If 3D, then all rasters must have the same number of depths and vertical extents (see [topology()]).
#'
#' Data tables associated with `GVector`s will be combined if each vector has a table and if each table has the same columns and data types. Otherwise, the data table will be combined using [merge()].
#'
#' @param x A `GRaster` or a `GVector`.
#'
#' @param ... One or more `GRaster`s, one or more `GVector`s, a list of `GRaster`s, or a list of `GVector`s. You can use a mix of lists and individual rasters or vectors.
#'
#' @return A `GRaster`.
#' 
#' @seealso [add<-], [terra::c()], \code{\link[terra]{add<-}}
#'
#' @example man/examples/ex_c.r
#'
#' @aliases c
#' @rdname c
#' @exportMethod c
setMethod(
	f = "c",
	signature = "GRaster",
	definition = function(x, ...) {

	.locationRestore(x)
	dots <- list(...)

	# unlist any lists
	if (length(dots) > 0L) dots <- omnibus::unlistRecursive(dots)

	### concatenate
	out <- x
	z <- zext(out)
	dims <- dim(out)[1L:3L]

	if (length(dots) >= 1L) {

		for (i in seq_along(dots)) {

			if (!inherits(dots[[i]], "GRaster")) stop("Can only combine GRasters with GRasters.")
			
			compareGeom(out, dots[[i]])
			mmx <- minmax(out)
			mmdots <- minmax(dots[[i]])

			mn <- c(.minVal(out), .minVal(dots[[i]]))
			mx <- c(.maxVal(out), .maxVal(dots[[i]]))
			
			out <- GRaster(
				location = .location(out),
				mapset = .mapset(out),
				workDir = faster("workDir"),
				crs = crs(out),
				projection = .projection(out),
				topology = topology(out),
				sources = c(sources(out), sources(dots[[i]])),
				names = c(names(out), names(dots[[i]])),
				extent = as.vector(ext(out)),
				zextent = z,
				datatypeGRASS = c(out@datatypeGRASS, dots[[i]]@datatypeGRASS),
				dimensions = dims,
				nLayers = nlyr(out) + nlyr(dots[[i]]),
				resolution = res3d(out),
				minVal = mn,
				maxVal = mx,
				activeCat = c(out@activeCat, dots[[i]]@activeCat),
				levels = c(out@levels, dots[[i]]@levels)
			)
		
		} # next GRaster to combine
		
	} # next item
	out <- .makeUniqueNames(out)		
	out
	
	} # EOF
)

#' Neighborhood matrix from a polygons GVector
#'
#' @description This function returns a neighborhood matrix from a polygons `GVector`, which represents which geometries touch one another. It is useful for implementing geostatistical analyses that require indicators about which area features are next to one another.
#'
#' Polygons must share more than one point for them to be considered a neighbors (i.e., same as `spdep::poly2nb(x, queen = FALSE)`).
#'
#' This function needs the **GRASS** addon `v.neighborhoodmatrix`. If it is not installed, it will try to install it.
#'
#' @param x A polygons `GVector.
#'
#' @returns A `list`. Each element represents a polygon. If an element is empty, it has no neighbors. Otherwise, it is a vector of integers, which represent the indices of the polygon(s) to which it is a neighbor.
#'
#' @example man/examples/ex_neighborhoodMatrix.r
#'
#' @aliases neighborhoodMatrix
#' @rdname neighborhoodMatrix
#' @exportMethod neighborhoodMatrix
methods::setMethod(
	f = "neighborhoodMatrix",
	signature = c(x = "GVector"),
	definition = function(x) {

	if (geomtype(x) != "polygons") stop("The input must be a GVector representing polygons.")

	.addons('v.neighborhoodmatrix')

	neighs <- rgrass::execGRASS(
		cmd = "v.neighborhoodmatrix",
		input = sources(x),
		separator = 'pipe',
		flags = c(.quiet(), "b"),
		intern = TRUE
	)

	out <- vector(mode = "list", length = ngeom(x))
	for (i in seq_along(out)) out[[i]] <- numeric()

	neighs <- strsplit(neighs, split = "\\|")
	neighs <- lapply(neighs, as.integer)

	for (i in seq_along(neighs)) {

		first <- neighs[[i]][1L]
		remainder <- neighs[[i]][2L:length(neighs[[i]])]

		remainder <- remainder[remainder != first]

		out[[first]] <- c(out[[first]], remainder)

	}
	out


	} # EOF
)

#' @aliases neighbourhoodMatrix
#' @rdname neighborhoodMatrix
#' @exportMethod neighbourhoodMatrix
methods::setMethod(
	f = "neighbourhoodMatrix",
	signature = c(x = "GVector"),
	definition = function(x) neighborhoodMatrix(x)
)

#' Start the GRASS GUI (potentially dangerous!)
#'
#' @description This function starts the **GRASS** GUI. It is provided merely as a utility... in most cases, it should *not* be used if you are doing any kind of analysis of rasters or vectors using **fasterRaster**. The reason for this prohibition is that **fasterRaster** objects, like `GRaster`s and `GVector`s, are really "pointers" to objects in **GRASS**. If **fasterRaster** points to a **GRASS** object that is changed in **GRASS** but not **R**, then **fasterRaster** will not "know" about it, so changed won't be reflected in the **fasterRaster** object.
#'
#' One aspect of the GUI that is useful but will not change objects is to use it to plot rasters and vectors. However, the a **fasterRaster** object in **R** will have a different name in **GRASS**. The name in **GRASS** of a `GVector` or `GRaster` is given by [sources()].
#'
#' @returns Nothing (starts the **GRASS** GUI).
#'
#' @seealso [mow()]
#'
#' @example man/examples/ex_grassGUI.r
#'
#' @aliases grassGUI
#' @rdname grassGUI
#' @exportMethod grassGUI
methods::setMethod(
	f = "grassGUI",
	signature = c(x = "missing"),
	function() {
		if (grassStarted()) {
			rgrass::execGRASS("g.gui", ui = "wxpython", flags = c(.quiet(), "f"))
		} else {
			warning("GRASS needs to be started by using `fast()` at least once before starting the GUI.")
		}
	}
)

#' Scatterplot of values in each GRaster layer against the others
#'
#' @description `pairs()` generates a scatterplot between values of cells in each layer of a `GRaster` against all the other layers.
#' 
#' @param x A `GRaster` with two or more layers.
#' 
#' @param n A numeric integer, integer, or `NULL` (default): Number of cells to sample. If `NULL`, 50% of the total number of cells will be used.
#' 
#' @param ... Arguments to send to [graphics::pairs()] (which typically sends them to [graphics::plot()]). Special arguments for affecting certain panels include:
#' * Correlation panels -- `cor.cex`: Size of the values of the correlation coefficients.
#' * Histogram panels -- `hist.col`: Histogram color.
#' * Dot-plot panels -- `colramp`: Function taking an integer `n` as input and returning `n` names of colors. The default is: `colorRampPalette(c("white", blues9))`.
#' 
#' @returns Nothing (creates a plot).
#' 
#' @example man/examples/ex_GRaster.r
#' 
#' @aliases pairs
#' @rdname pairs
#' @exportMethod pairs
methods::setMethod(
	f = "pairs",
	signature = c(x = "GRaster"),
	function(x, n = NULL, ...) {

	.locationRestore(x)
	.region(x)

 	if (is.null(n)) n <- round(0.5 * ncell(x))

	samps <- spatSample(x, size = n, values = TRUE, xy = FALSE, as.points = FALSE)

	dots <- list(...)
	if (!("cex" %in% names(dots))) dots$cex <- 0.5
	if (!("pch" %in% names(dots))) dots$pch <- "."

	args <- list(
		x = samps,
		lower.panel = .panelSS,
		upper.panel = .panelCor,
		diag.panel = .panelHist
	)
	args <- c(args, dots)

	do.call(graphics::pairs, args = args)

	} # EOF
)

#' From [graphics::pairs()] help
#' @noRd
.panelCor <- function(x, y, cor.cex = 2, digits = 2, ...) {

	usr <- par(usr = c(0, 1, 0, 1))
	on.exit(graphics::par(usr = usr), add = TRUE)
	r <- abs(stats::cor(x, y))
	txt <- format(c(r, 0.123456789), digits = digits)[1L]
	text(0.5, 0.5, txt, cex = max(1, r * cor.cex))

}

#' From [graphics::pairs()] help
#' @noRd
.panelHist <- function(x, breaks = 20, hist.col = "firebrick3", ...) {

    usr <- par("usr")
    par(usr = c(usr[1L:2L], 0, 1.5))

    h <- graphics::hist(x, breaks = breaks, plot = FALSE)
    breaks <- h$breaks
    nB <- length(breaks)
    y <- h$counts
    y <- y / max(y)
    graphics::rect(breaks[-nB], 0, breaks[-1L], y, col = hist.col)

}

#' @noRd
.panelSS <- function(x, y, colramp = grDevices::colorRampPalette(c("white", grDevices::blues9)), ...) {

	usr <- par(usr = c(0, 1, 0, 1), new = TRUE)
	on.exit(graphics::par(usr = usr), add = TRUE)
	graphics::smoothScatter(x, y, colramp = colramp, nrpoints = Inf, ...)

}

#' @title Topology (2- or 3-dimensions) of a GRaster or GVector
#' @aliases topology
#' @description `GRaster`s and `GVector`s can have 2-dimensional or 3-dimensional coordinates. This function returns the dimensions of the object.
#'
#' @param x A `GSpatial` object (i.e., a `GRaster` or `GVector`).
#'
#' @returns Either "2D" or "3D".
#'
#' @seealso [is.2d()], [is.3d()]
#'
#' @example man/examples/ex_GRaster.r
#'
#' @exportMethod topology
methods::setMethod(
	f = "topology",
	signature = "GSpatial",
	definition = function(x) x@topology
)

if (grassStarted()) {

# See current values for options:
faster("grassDir")
faster("cores")
faster("memory")
faster("useDataTable")
faster() # all options

# See default values for options:
faster("cores", default = TRUE)
faster(default = TRUE) # all options

# Set options (change accordingly for your system!!!)
if (FALSE) {

   opts. <- faster() # remember starting values of options

   faster(grassDir = "C:/Program Files/GRASS GIS 8.4")
   faster(verbose = TRUE, memory = 1024, cores = 1)

   faster(c("grassDir", "verbose", "memory", "cores"))

   faster(opts.) # reset options to starting values

}

}

#' Number of NA or non-NA cells in a raster
#'
#' @description The `nacell()` function counts the number of `NA` cells in a `GRaster`, and the `nonnacell()` reports the number of non-`NA` cells. If the raster is 3D, then all cells in all layers are counted.
#'
#' @param x A `GRaster`.
#'
#' @param warn Logical: If `TRUE` (default), display a warning about how much time the computation could take.
#'
#' @returns A numeric value, one per raster layer in the input.
#'
#' @example man/examples/ex_GRaster.r
#'
#' @seealso [ncell()], [ncell3d()], [terra::ncell()], [dim()], [terra::dim()]
#' 
#' @aliases nacell
#' @rdname nacell
#' @exportMethod nacell
methods::setMethod(
    f = "nacell",
    signature = c(x = "GRaster"),
    function(x, warn = TRUE) {

    .locationRestore(x)
    .region(x)
    ncell(x) - .nonnacell(x, warn = warn)

    } # EOF
)

#' @aliases nonnacell
#' @rdname nacell
#' @exportMethod nonnacell
methods::setMethod(
    f = "nonnacell",
    signature = c(x = "GRaster"),
    function(x, warn = TRUE) .nonnacell(x, warn = warn)
)

#' Number of non-NA cells in a GRaster
#'
#' @param x The [sources()] name of a `GRaster` or a `GRaster`. Location are region will be assumed to be appropriate to the raster.
#' @param warn Logical.
#' @param nLayers `NULL` or number of layers in `x`.
#'
#' @noRd
.nonnacell <- function(x, warn, nLayers = NULL) {

    if (warn) .message("nacell_nonnacell", "Calculating the number of NA or non-NA cells can take a while for large rasters.")

    # # NB `r.univar` can report incorrect numbers of cells sometimes (e.g., # NA cells > total number of cells), so we use `r.mapcalc` to create a 1s raster and sum the cell values.

    if (inherits(x, "GRaster")) {
        srcIn <- sources(x)
        nLayers <- nlyr(x)
    } else {
        srcIn <- x
        if (is.null(nLayers)) stop("Must specifify `nLayers` when supplying a sources() name to this function.")
    }

    out <- rep(NA_real_, nLayers)

    srcs <- .makeSourceName("nacell_r_mapcalc", n = nLayers)
    for (i in seq_len(nLayers)) {
    
        ex <- paste0(srcs[i], " = ", srcIn[i], " * 0 + 1")
        rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

        args <- list(
            cmd = "r.univar",
            flags = c("r", .quiet()),
            map = srcs[i],
            nprocs = faster("cores"),
            Sys_show.output.on.console = FALSE,
            echoCmd = FALSE,
            intern = TRUE
        )

        # Should be GRASS eight point three!
        if (grassInfo("versionNumber") >= 8.3) args$nprocs <- faster("cores")
        info <- do.call(rgrass::execGRASS, args = args)

        # NB need to convert to numeric to obviate overflow
        pattern <- "sum: "
        n <- info[grepl(info, pattern = pattern)]
        n <- sub(n, pattern = pattern, replacement = "")
        out[i] <- as.numeric(n)

    } # next layer
    out

}

#' @title Mathematical operations on two or more GRasters
#'
#' @description These functions can be applied to a "stack" of `GRaster`s with two or more layers. They return a single-layered `GRaster`.  If you want to summarize across cells in a raster (e.g., calculate the mean value of all cells on a raster), use [global()]. Options include:
#' * Numeration: `count()` (number of non-`NA` cells), `sum()`.
#' * Central tendency: `mean()`, `mmode()` (mode), `median()`.
#' * Extremes: `min()`, `max()`, `which.min()` (index of raster with the minimum value), `which.max()` (index of the raster with the maximum value)
#' * Dispersion: `range()`, `stdev()` (standard deviation), `var()` (sample variance), `varpop()` (population variance), `nunique()` (number of unique values), `quantile()` (use argument `probs`), `skewness()`, and `kurtosis()`.
#' * `NA`s: `anyNA()` (any cells are `NA`?), `allNA()` (are all cells `NA`?)
#'
#' @param x A `GRaster`. Typically, this raster will have two or more layers. Values will be calculated within cells across rasters.
#'
#' @param prob Numeric: Quantile to calculate. Used for `quantile()`.
#'
#' @param pop Logical (for `stdev()`): If `TRUE` (default), calculate the population standard deviation across layers. If `FALSE`, calculate the sample standard deviation.
#'
#' @param na.rm Logical: If `FALSE` (default), of one cell value has an `NA`, the result will be `NA`. If `TRUE`, `NA`s are ignored.
#'
#' @returns A `GRaster`.
#'
#' @example man/examples/ex_GRaster_arithmetic_across_layers.r
#'
#' @aliases mean
#' @rdname functions
#' @exportMethod mean
setMethod(
	"mean",
	signature(x = "GRaster"),
	function(x, na.rm = FALSE) {

		fx <- "average"
		fxName <- "mean"
		.genericMultiLayer(fx = fx, fxName = fxName, x = x, na.rm = na.rm)
		
	} # EOF
)

#' @aliases mmode
#' @rdname functions
#' @exportMethod mmode
setMethod(
	"mmode",
	signature(x = "GRaster"),
	function(x, na.rm = FALSE) {

	fx <- "mode"
	fxName <- "mode"
	.genericMultiLayer(fx = fx, fxName = fxName, x = x, na.rm = na.rm)
	
	} # EOF
)

#' @aliases median
#' @rdname functions
#' @exportMethod median
setMethod(
	"median",
	signature(x = "GRaster"),
	function(x, na.rm = FALSE) {

		fx <- "median"
		fxName <- "median"
		.genericMultiLayer(fx = fx, fxName = fxName, x = x, na.rm = na.rm)
		
	} # EOF
)

#' @aliases count
#' @rdname functions
#' @exportMethod count
setMethod(
	"count",
	signature(x = "GRaster"),
	function(x) {

		fx <- "count"
		fxName <- "count"
		.genericMultiLayer(fx = fx, fxName = fxName, x = x, na.rm = TRUE)
		
	} # EOF
)

#' @aliases sum
#' @rdname functions
#' @exportMethod sum
setMethod(
	"sum",
	signature(x = "GRaster"),
	function(x, na.rm = FALSE) {

		fx <- "sum"
		fxName <- "sum"
		.genericMultiLayer(fx = fx, fxName = fxName, x = x, na.rm = na.rm)
		
	} # EOF
)

#' @aliases min
#' @rdname functions
#' @exportMethod min
setMethod(
	"min",
	signature(x = "GRaster"),
	function(x, na.rm = FALSE) {

		fx <- "minimum"
		fxName <- "min"
		.genericMultiLayer(fx = fx, fxName = fxName, x = x, na.rm = na.rm)
		
	} # EOF
)

#' @aliases max
#' @rdname functions
#' @exportMethod max
setMethod(
	"max",
	signature(x = "GRaster"),
	function(x, na.rm = FALSE) {

		.locationRestore(x)
		
		fx <- "maximum"
		fxName <- "max"
		.genericMultiLayer(fx = fx, fxName = fxName, x = x, na.rm = na.rm)
		
	} # EOF
)

#' @aliases which.min
#' @rdname functions
#' @exportMethod which.min
setMethod(
	"which.min",
	signature(x = "GRaster"),
	function(x) {

		fx <- "min_raster"
		fxName <- "whichmin"
		out <- .genericMultiLayer(fx = fx, fxName = fxName, x = x, na.rm = TRUE)
		out + 1
		
	} # EOF
)

#' @aliases which.max
#' @rdname functions
#' @exportMethod which.max
setMethod(
	"which.max",
	signature(x = "GRaster"),
	function(x) {

		fx <- "max_raster"
		fxName <- "whichmax"
		out <- .genericMultiLayer(fx = fx, fxName = fxName, x = x, na.rm = TRUE)
		out + 1
		
	} # EOF
)

#' @aliases sdpop
#' @rdname functions
#' @exportMethod sdpop
setMethod(
	"sdpop",
	signature(x = "numeric"),
	function(x, na.rm = FALSE) {
	sqrt(varpop(x, na.rm = na.rm))
	} # EOF
)

#' @aliases varpop
#' @rdname functions
#' @exportMethod varpop
setMethod(
	"varpop",
	signature(x = "GRaster"),
	function(x, na.rm = FALSE) {

		fx <- "variance"
		fxName <- "varpop"
		.genericMultiLayer(fx = fx, fxName = fxName, x = x, na.rm = na.rm)
		
	} # EOF
)

#' @aliases varpop
#' @rdname functions
#' @exportMethod varpop
setMethod(
	"varpop",
	signature(x = "numeric"),
	function(x, na.rm = FALSE) {

	if (na.rm & anyNA(x)) x <- x[!is.na(x)]
	ss <- sum((x - mean(x))^2)
	n <- length(x)
	ss / n
		
	} # EOF
)

#' @aliases stdev
#' @rdname functions
#' @exportMethod stdev
setMethod(
	"stdev",
	signature(x = "GRaster"),
	function(x, pop = TRUE, na.rm = FALSE) {

	.locationRestore(x)
	.region(x)

	# sample standard deviation
	if (!pop) {

		# sum of squares
		srcSS <- .makeSourceName("ss", "raster")
		srcMean <- .genericMultiLayer(fx = "average", fxName = "mean", x = x, na.rm = na.rm, return = "source")
		ex <- paste0(srcSS, " = ", paste0("(", sources(x), " - ", srcMean, ")^2", collapse=" + "))
		rgrass::execGRASS(
			cmd = "r.mapcalc",
			expression = ex,
			flags = c(.quiet(), "overwrite")
		)

		# N
		srcCount <- .genericMultiLayer(fx = "count", fxName = "count", x = x, na.rm = na.rm, return = "source")
		srcCountMinus1 <- .makeSourceName("nMinus1", "rast")
		ex <- paste0(srcCountMinus1, " = ", srcCount, " - 1")
		rgrass::execGRASS(
			cmd = "r.mapcalc",
			expression = ex,
			flags = c(.quiet(), "overwrite")
		)

		# SD
		prec <- .getPrec(x, NULL)

		src <- .makeSourceName("sd", "rast")
		ex <- paste0(src, " = ", prec, "(sqrt(", srcSS, " / ", srcCountMinus1, "))")
		
		rgrass::execGRASS(
			cmd = "r.mapcalc",
			expression = ex,
			flags = c(.quiet(), "overwrite")
		)

		.rm(c(srcSS, srcCount), type = "raster", warn = FALSE)

		out <- makeGRaster(src, "sd")

	# population standard deviation
	} else {
	
		fx <- "stddev"
		fxName <- "sdpop"
		out <- .genericMultiLayer(fx = fx, fxName = fxName, x = x, na.rm = na.rm)
		
	}
	out
	
	} # EOF
)

#' @aliases var
#' @rdname functions
#' @exportMethod var
setMethod(
	"var",
	signature(x = "GRaster"),
	function(x, na.rm = FALSE) {

	.locationRestore(x)
	.region(x)

	# sum of squares
	srcSS <- .makeSourceName("ss", "rast")
	srcMean <- .genericMultiLayer(fx = "average", fxName = "mean", x = x, na.rm = na.rm, return = "source")
	ex <- paste0(srcSS, " = ", paste0("(", sources(x), " - ", srcMean, ")^2", collapse=" + "))
	rgrass::execGRASS(
		cmd = "r.mapcalc",
		expression = ex,
		flags = c(.quiet(), "overwrite")
	)

	# N
	srcCount <- .genericMultiLayer(fx = "count", fxName = "count", x = x, na.rm = na.rm, return = "source")
	srcCountMinus1 <- .makeSourceName("nMinus1", "rast")
	ex <- paste0(srcCountMinus1, " = ", srcCount, " - 1")
	rgrass::execGRASS(
		cmd = "r.mapcalc",
		expression = ex,
		flags = c(.quiet(), "overwrite")
	)

	# variance
	prec <- .getPrec(x, NULL)

	src <- .makeSourceName("var", "rast")
	ex <- paste0(src, " = ", prec, "(", srcSS, " / ", srcCountMinus1, ")")
	
	rgrass::execGRASS(
		cmd = "r.mapcalc",
		expression = ex,
		flags = c(.quiet(), "overwrite")
	)
	
	.rm(c(srcSS, srcCount), type = "raster", warn = FALSE)
	makeGRaster(src, "var")
		
	} # EOF
)

#' @aliases nunique
#' @rdname functions
#' @exportMethod nunique
setMethod(
	"nunique",
	signature(x = "GRaster"),
	function(x, na.rm = FALSE) {

		fx <- "diversity"
		fxName <- "nunique"
		.genericMultiLayer(fx = fx, fxName = fxName, x = x, na.rm = na.rm)
		
	} # EOF
)

#' @aliases skewness
#' @rdname functions
#' @exportMethod skewness
setMethod(
	"skewness",
	signature(x = "GRaster"),
	function(x, na.rm = FALSE) {

		fx <- "skewness"
		fxName <- "skewness"
		.genericMultiLayer(fx = fx, fxName = fxName, x = x, na.rm = na.rm)
		
	} # EOF
)

#' @aliases kurtosis
#' @rdname functions
#' @exportMethod kurtosis
setMethod(
	"kurtosis",
	signature(x = "GRaster"),
	function(x, na.rm = FALSE) {

		fx <- "kurtosis"
		fxName <- "kurtosis"
		.genericMultiLayer(fx = fx, fxName = fxName, x = x, na.rm = na.rm)
		
	} # EOF
)

#' @aliases range
#' @rdname functions
#' @exportMethod range
setMethod(
	"range",
	signature(x = "GRaster"),
	function(x, na.rm = FALSE) {

		fx <- "range"
		fxName <- "range"
		.genericMultiLayer(fx = fx, fxName = fxName, x = x, na.rm = na.rm)
		
	} # EOF
)

#' @aliases quantile
#' @rdname functions
#' @exportMethod quantile
setMethod(
	"quantile",
	signature(x = "GRaster"),
	function(x, prob, na.rm = FALSE) {

    .locationRestore(x)
    .region(x)

	fx <- "quantile"
	fxName <- "quantile"
	
	src <- .makeSourceName(fxName, "rast")
	args <- list(
		cmd = "r.series",
		input = paste(sources(x), collapse=","),
		output = src,
		method = fx,
		quantile = prob,
		nprocs = faster("cores"),
		memory = faster("memory"),
		flags = c(.quiet(), "overwrite")
	)
	
	if (na.rm) args$flags <- c(args$flags, "n")
	
	do.call(rgrass::execGRASS, args = args)
	makeGRaster(src, fxName)
		
	} # EOF
)

#' @aliases anyNA
#' @rdname functions
#' @exportMethod anyNA
setMethod(
	"anyNA",
	signature(x = "GRaster"),
	function(x) {

    .locationRestore(x)
    .region(x)

	src <- .makeSourceName("anyNA", "raster")

	nl <- nlyr(x)
	ex <- paste0("if(isnull(", sources(x), "))")
	ex <- paste(ex, collapse = " | ")
	ex <- paste0(src, " = ", ex)
	rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))
	makeGRaster(src, "layer")

	} # EOF
)


#' @aliases allNA
#' @rdname functions
#' @exportMethod allNA
setMethod(
	"allNA",
	signature(x = "GRaster"),
	function(x) {

    .locationRestore(x)
    .region(x)

	src <- .makeSourceName("anyNA", "raster")

	nl <- nlyr(x)
	ex <- paste0("if(isnull(", sources(x), "))")
	ex <- paste(ex, collapse = " & ")
	ex <- paste0(src, " = ", ex)
	rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))
	makeGRaster(src, "layer")

	} # EOF
)



# generic function for multi-layer functions
# fx: name of GRASS function
# fxName: stub name of output raster
# x: GRaster
# na.rm: T/F
# return: "GRaster" or "source"
#' @noRd
.genericMultiLayer <- function(fx, fxName, x, na.rm, return = "GRaster") {

	.locationRestore(x)
	.region(x)

	src <- .makeSourceName(fx, "rast")
	args <- list(
		cmd = "r.series",
		input = paste(sources(x), collapse=","),
		output = src,
		method = fx,
		nprocs = faster("cores"),
		memory = faster("memory"),
		flags = c(.quiet(), "overwrite")
	)
	
	if (!na.rm) args$flags <- c(args$flags, "n")
	do.call(rgrass::execGRASS, args = args)

	if (return == "GRaster") {
		if (fxName %in% c("mode", "min", "max")) {
			levels <- combineLevels(x)
		} else {
			levels <- NULL
		}
		out <- makeGRaster(src, fxName, levels = levels)
	} else if (return == "source") {
		out <- src
	} else {
		stop("Invalid value for `return`.")
	}
	out

}

#' Aggregate raster cells into larger cells or combine geometries of a vector
#'
#' @description When applied to a `GRaster`, `aggregate()` creates a new raster with cells that are a multiple of the size of the cells of the original raster. The new cells can be larger or smaller than the original cells (this function thus emulates both the `terra::aggregate()` and [terra::disagg()] functions.)
#' 
#' When applied to a `GVector`, all geometries are combined into a "multipart" geometry, in which geometries are treated as if they were a single unit. Borders between aggregated geometries can be dissolved if the `dissolve` argument is `TRUE`. If the `GVector` has a data table associated with it, the output will also have a data table as long as there is at least one column with values that are all the same. Values of columns that do not have duplicated values will be converted to `NA`.
#'
#' @param x A `GRaster` or `GVector`.
#'
#' @param fact Numeric vector (rasters only): One, two, or three positive values. These reflect the size of the new cells as multiples of the size of the old cells. If just one value is supplied, this is used for all two or three dimensions. If two values are supplied, the first is multiplied by the east-west size of cells, and the second north-south size of cells (the raster must be 2D). If three values are supplied, the third value is used as the multiplier of the vertical dimension of cells. Values are calculated using all cells that have their centers contained by the target cell.
#'
#' Note that unlike `terra::aggregate()` and [terra::disagg()], these values need not be integers.
#'
#' @param fun Character (rasters only): Name of the function used to aggregate. For `GRaster`s, this is the function that summarizes across cells. For `GVector`s, this function will be used to calculate new values of numeric or integer cells.
#' * `mean`: Average (default)
#' * `median`: Median
#' * `mode`: Most common value
#' * `min`: Minimum
#' * `max`: Maximum
#' * `range`: Difference between maximum and minimum
#' * `sum`: Sum
#' * `varpop`: Population variance
#' * `sdpop`: Population standard deviation
#' * `quantile`: Quantile (see argument `prob`)
#' * `count`: Number of non-`NA` cell
#' * `diversity`: Number of unique values
#'
#' @param prob Numeric (rasters only): Quantile at which to calculate `quantile`.
#'
#' @param na.rm Logical (rasters only): If `FALSE` (default), propagate `NA` cells or `NA` values.
#'
#' @param weight Logical (rasters only): If `FALSE`, each source cell that has its center in the destination cell will be counted equally. If `TRUE`, the value of each source will be weighted the proportion of the destination cell the source cell covers.
#' 
#' @returns A `GRaster` or `GVector`.
#' 
#' @seealso [stats::aggregate()], [terra::aggregate()], [disagg()], [terra::disagg()]
#'
#' @example man/examples/ex_aggregate_disagg.r
#'
#' @aliases aggregate
#' @rdname aggregate
#' @exportMethod aggregate
methods::setMethod(
	f = "aggregate",
	signature = c(x = "GRaster"),
	definition = function(
		x,
		fact = 2,
		fun = "mean",
		weight = FALSE,
		prob = NULL,
		na.rm = FALSE
	) {

	if (any(fact <= 0)) stop("Values of `fact` must be > 0.")

	funs <- c("mean", "median", "mode", "min", "maximum", "range", "quantile", "sum", "varpop", "sdpop", "count", "diversity")
	fun <- omnibus::pmatchSafe(tolower(fun), funs)
	
	if (fun == "mean") {
		fun <- "average"
	} else if (fun == "min") {
		fun <- "minimum"
	} else if (fun == "max") {
		fun <- "maximum"
	} else if (fun == "varpop") {
		fun <- "variance"
	} else if (fun == "sdpop") {
		fun <- "stdev"
	} else if (fun == "quantile") {
		
		if (is.null(prob)) stop("A value must be specified for `prob` if the aggregating function is `quantile`.")
		
		if (prob < 0 | prob > 1) stop("Argument `prob` must be in the range [0, 1].")
	
	}
	
	.locationRestore(x)
	.region(x)

	if (is.2d(x)) {
		
		if (length(fact) == 1L) fact <- rep(fact, 2L)
		if (length(fact) == 3L) stop("This is a 2D raster. Only 1 or 2 values are allowed for `resol`.")
	
		resol <- res(x)
		resol <- resol * fact
		.regionRes(resol, respect = "extent")
	
	} else if (is.3d(x)) {
	
		if (length(fact) == 1L) fact <- rep(fact, 3L)
		if (length(fact) == 2L) {
			warning("This is a 3D raster, but `resol` has only 2 values.\n  Assuming third dimension will not be aggregated.")
			fact[3L] <- 1
		}
	
		resol <- res3d()
		resol <- resol * fact
		.regionRes(resol)

	}

	args <- list(
		cmd = "r.resamp.stats",
		input = NA_character_,
		output = NA_character_,
		method = fun,
		flags = c(.quiet(), "overwrite")
	)
	
	if (weight) args$flags <- c(args$flags, "w")
	if (!na.rm) args$flags <- c(args$flags, "n")
	if (fun == "quantile") args <- c(args, quantile = prob)
	
	nLayers <- nlyr(x)
	srcs <- .makeSourceName("aggregate", "raster", n = nLayers)
	for (i in seq_len(nLayers)) {
		
		args$input <- sources(x)[i]
		args$output <- srcs[i]
		
		do.call(rgrass::execGRASS, args = args)
		
		if (datatype(x, type = "GRASS")[i] == "CELL" & fun %in% c("mode", "minimum", "maximum")) {
			
			srcIn <- srcs[i]
			srcs[i] <- .makeSourceName("aggregate_int", "raster")
			ex <- paste0(srcs[i], " = int(", srcIn, ")")
			rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))
			
		}

		if (fun %in% c("median", "mode", "min", "max")) {
			levs <- cats(x)[[i]]
		} else {
			levs <- NULL
		}

		this <- makeGRaster(srcs[i], names = names(x)[i], levels = levs)
		if (i == 1L) {
			out <- this
		} else {
			out <- c(out, this)
		}
		
	}
	out

	} # EOF
)

#' @aliases aggregate
#' @rdname aggregate
#' @exportMethod aggregate
methods::setMethod(
	f = "aggregate",
	signature = c(x = "GVector"),
	function(x) {

	.locationRestore(x)
	gtype <- geomtype(x)
	# src <- .aggregate(x, dissolve = dissolve, gtype = gtype, copy = TRUE)
	src <- .aggregate(x, dissolve = TRUE, gtype = gtype, copy = TRUE)

	# aggregate data table
	if (nrow(x) == 0L) {
		aggTable <- NULL
	} else {

		table <- as.data.table(x)
			
		if (nrow(x) == 1L) {
			aggTable <- table
		} else {

			dups <- rep(FALSE, ncol(table))
			rowsMinus1 <- nrow(table) - 1L
			
			for (i in 1L:ncol(table)) {
				if (sum(duplicated(table[ , ..i])) == rowsMinus1) dups[i] <- TRUE
			}

			if (!any(dups)) {
				aggTable <- NULL
			} else {

				aggTable <- table[1L, ]
				classes <- sapply(table, class)
				
				for (i in 1L:ncol(table)) {

					if (!dups[i]) {
						
						if (classes[i] == "character") {
							assign <- NA_character_
						} else if (classes[i] == "integer") {
							assign <- NA_integer_
						} else if (classes[i] == "numeric") {
							assign <- NA_real_
						} else {
							assign <- NA
						}
						aggTable[1L, i] <- assign
					}
				} # for each data table column

			} # if there is at least one entirely-duplicated row

		} # if table has >= 1 row

	} # if data table has >= 2 rows

	makeGVector(src, table = aggTable)

	} # EOF
)

#' @param x `GVector` or [sources()] name
#' @param gtype geomtype (fasterRaster)
#' @param dissolve Logical
#' @param copy Logical: If TRUE, make copy of vector before operations
#' @noRd
.aggregate <- function(x, dissolve, gtype, copy) {

	if (inherits(x, "GVector")) {
		.locationRestore(x)
		src <- sources(x)
	} else {
		src <- x
	}
	
	if (copy) src <- .copyGVector(src)
	
	oldcats <- .vCats(src, db = FALSE)
	if (length(oldcats) > 1L) {

		table <- data.table::data.table(frid = rep(1L, length(oldcats)))
		.vAttachDatabase(src, table = table, replace = TRUE, cats = oldcats)

		# newcats <- data.frame(oldfr = oldcats, fr = rep(1L, length(oldcats)))
		
		# tf <- tempfile(fileext = ".csv")
		# tft <- paste0(tf, "t")
		# utils::write.csv(newcats, tf, row.names = FALSE)
		# # tableType <- '"Integer","Integer"'
		# tableType <- '"Integer"'
		# write(tableType, tft)

		# # import table with new categories
		# srcTable <- .makeSourceName("db_in_ogr", "table")
		
		# rgrass::execGRASS(
			# cmd = "db.in.ogr",
			# input = tf,
			# output = srcTable,
			# flags = c(.quiet(), "overwrite")
		# )

		# # connect table to copy of vector
		# srcIn <- .copyGVector(src)

		# rgrass::execGRASS(
			# cmd = "v.db.join",
			# map = srcIn,
			# other_table = srcTable,
			# other_column = "oldfr",
			# column = "fr",
			# flags = .quiet()
		# )

		srcIn <- src
		src <- .makeSourceName("aggregate_v_reclass", "vector")
		
		rgrass::execGRASS(
			"v.reclass",
			input = srcIn,
			output = src,
			column = "fr",
			flags = c(.quiet(), "overwrite")
		)

		if (dissolve & gtype == "polygons") {
					
			table <- data.table::data.table(fr = rep(1L, length(oldcats)))

			src <- .copyGSpatial(x)
			.vAttachDatabase(src, table = table, replace = TRUE)

			srcIn <- src
			src <- .makeSourceName("v_extract", "vector")

			rgrass::execGRASS(
				cmd = "v.extract",
				input = srcIn,
				output = src,
				new = 1L,
				flags = c(.quiet(), "overwrite", "d", "t")
			)
		
		}

	} # if >1 geometry

	src

} # EOF

#' Convert a GRaster to a SpatRaster
#'
#' @description The **fasterRaster** version of the `rast()` function converts a `GRaster` to a `SpatRaster` (from the **terra** package).
#'
#' @param x A `GRaster`.
#' @param mm Logical: If `TRUE`, call [terra::setMinMax()] on the raster to ensure it has metadata on the minimum and maximum values. For large rasters, this can take a long time, so the default value of `mm` is `FALSE`.
#' @param ... Additional arguments to send to [writeRaster()]. These are typically unneeded, though `bigTiff` may be of use if the raster is large, and supplying `datatype` can speed conversion of large rasters. See [writeRaster()].
#'
#' @return A `SpatRaster` (**terra** package).
#' 
#' @seealso [terra::rast()]
#'
#' @example man/examples/ex_GRaster.r
#'
#' @aliases rast
#' @rdname rast
#' @exportMethod rast
setMethod(
	"rast",
	signature(x = "GRaster"),
	function(x, mm = FALSE, ...) {

	filename <- tempfile(fileext = ".tif")
	writeRaster(x, filename = filename, format = "GeoTIFF", ...)
	out <- terra::rast(filename)
	if (mm) out <- terra::setMinMax(out)
	names(out) <- names(x)
	
	if (any(is.factor(x))) {
		facts <- which(is.factor(x))
		for (fact in facts) {
			levels(out[[fact]]) <- cats(x)[[i]]
			activeCat(out[[fact]]) <- activeCat(x)[i]
		}
	}

	out

	} # EOF
)

#' Remove rasters and vectors from the GRASS cache
#'
#' @description **fasterRaster** functions attempt to delete rasters and vectors in the **GRASS** cache, but not all intermediate files can be removed. This function can be used to clear the cache of extraneous rasters and vectors.
#'
#' Calling this function inside another function's environment and defining `x` as `"*"` can be very **dangerous**, as it will detect objects outside of that environment, and thus delete any rasters/vectors outside that environment. Here is a guide:
#' * To delete files associated with a single `GRaster` or `GVector`, use `mow(GRaster_to_unlink)` or `mow(GVector_to_unlink)`.
#' * To delete files associated with more than one `GRaster`s and/or `GVector`s, provide them as a list. For example, use `mow(list(GRaster_to_unlink, GVector_to_unlink))`.
#' * To remove all rasters, all vectors, or all rasters and vectors in the **GRASS** cache that are not linked to a `GRaster` or `GVector`, use `mow("*")`.
#' * To remove all rasters or all vectors in the **GRASS** cache, use `mow("*", type = "rasters")` or `mow("*", type = "vectors")`.
#' * To remove all rasters or all vectors in the **GRASS** cache *except* for certain ones, use `mow("*", keep = list(GRaster_to_keep, GVector_to_keep))`. You can combine this with the `keep` argument to retain specific rasters or vectors. For example, you can use `mow("*", type = "rasters", keep = list(GRaster_to_keep))`.
#'
#' @param x Any of:
#' * `"unlinked"` (default): Delete **GRASS** rasters and/or vectors that are unlinked to `GRaster`s or `GVector`s in the environment in which the function was called, or the environment named in `pos`.
#' * A `GRaster` or `GVector`: Delete the **GRASS** raster or vector pointed to by this object.
#' * A `list` of `GRaster`s and/or `GVector`s: Delete the **GRASS** raster(s) and/or vector(s) pointed to by these objects.
#' * `"*"`: Delete *all* **GRASS** rasters and/or vectors pointed to by objects in the environment named in `pos`. Only objects in `keep` will not be deleted.
#' 
#' @param pos Either `NULL` (default), or an environment. This is used only if `x` is `"unlinked"` or `"*"`. In that case, if `pos` is `NULL`, the environment in which this function was called will be searched for `GRaster`s and `GVector`s for removal of their associated **GRASS** rasters and vectors. Otherwise, the named environment will be searched.
#'
#' @param type Either `NULL` or a character vector. This is used only if `x` is `"unlinked"` or `"*"`. If `NULL`, all rasters and vectors in the **GRASS** cache are candidates for deletion. Otherwise, this can be either `"rasters"`, `"vectors"`, or both.
#'
#' @param keep Either `NULL` (default) or a `list()` of `GRaster`s and/or `GVector`s that you want to retain. This is used only if `x` is `"unlinked"` or `"*"`. The rasters and vectors in **GRASS** pointed to by these objects will not be deleted.
#'
#' @param verbose Logical: If `TRUE` (default), report progress.
#'
#' @param ask Logical: If `TRUE` (default), prompt for reassurance. This is used only if `x` is `"unlinked"` or `"*"`.
#'
#' @returns Invisibly returns a named vector with the number of rasters and vectors deleted.
#'
#' @seealso [terra::tmpFiles()]
#'
#' @example man/examples/ex_mow.r
#'
#' @aliases mow
#' @rdname mow
#' @export mow
mow <- function(
	x = "unlinked",
	pos = NULL,
	type = NULL,
	keep = NULL,
	verbose = TRUE,
	ask = TRUE
) {
	
	### delete a single GRaster or GVector
	######################################
	if (inherits(x, "GRaster") | inherits(x, "GVector")) {
		
		.locationRestore(x)
		.rm(x)
		return(invisible(c(rasters = 0, vectors = 0)))
	
	}

	### delete a list of GRasters/GVectors
	######################################
	if (is.list(x)) {

		# make data table of objects
		GSpatials <- data.table::data.table()
		for (i in seq_along(x)) {

			thisClass <- if (inherits(x[[i]], "GRaster")) {
				"GRaster"
			} else if (inherits(x[[i]], "GVector")) {
				"GVector"
			} else {
				"other"
			}

			GSpatials <- rbind(
				GSpatials,
				data.table::data.table(
					sources = sources(x[[i]]),
					project = .location(x[[i]]),
					class = thisClass
				)
			)

		}

		out <- c(rasters = 0L, vectors = 0L)

		# delete rasters
		if (any(GSpatials$class == "GRaster")) {

			projs <- unique(GSpatials$project[GSpatials$class == "GRaster"])
			if (verbose | faster("verbose")) omnibus::say("Deleting ", sum(GSpatials$class == "GRaster"), " raster(s) from the GRASS cache...")
			
			for (proj in projs) {
			
				.locationRestore(proj)

				srcs <- GSpatials$sources[GSpatials$class == "GRaster" & GSpatials$project == proj]
				out[["rasters"]]  <- out[["rasters"]] + length(srcs)
				.rm(srcs, type = "rasters", verify = FALSE)

			}
			
		}

		# delete vectors
		if (any(GSpatials$class == "GVector")) {

			projs <- unique(GSpatials$project[GSpatials$class == "GVector"])
			if (verbose | faster("verbose")) omnibus::say("Deleting ", sum(GSpatials$class == "GVector"), " vector(s) from the GRASS cache...")
			
			for (proj in projs) {
			
				.locationRestore(proj)

				srcs <- GSpatials$sources[GSpatials$class == "GVector" & GSpatials$project == proj]
				out[["vectors"]]  <- out[["vectors"]] + length(srcs)
				.rm(srcs, type = "vectors", verify = FALSE)

			}
			
		}

		return(invisible(out))

	}

	### delete either everything except objects in `keep` or just unlinked rasters/vectors
	######################################################################################
	if (is.character(x)) {

		x <- omnibus::pmatchSafe(x, c("*", "unlinked"), nmax = 1L)

		if (is.null(type)) type <- c("rasters", "vectors")
		type <- omnibus::pmatchSafe(type, c("rasters", "vectors"), nmax = 2L)

		if (ask & x %in% c("*", "unlinked")) {
		
			response <- readline("This action will break any existing GRasters and GVectors.\nAre you sure you want to empty the entire GRASS cache? (Y/n) ")
			if (response != "Y") {
				if (verbose | faster("verbose")) omnibus::say("Nothing deleted.")
				return(invisible(c(rasters = 0, vectors = 0)))
			}
		
		}

		if (x == "*" && !is.null(keep) && !is.list(keep)) stop("Argument `keep` must be a list or NULL.")

		### get GRASS file names in each project/location
		projs <- names(.fasterRaster$locations)
		grassObjs <- list()
		for (proj in projs) {

			.locationRestore(proj)
			grassObjs[[length(grassObjs) + 1L]] <- .ls(type = type)
			names(grassObjs)[length(grassObjs)] <- proj
		
		}

		if (all(sapply(grassObjs, length) == 0L)) {
			if (verbose | faster("verbose")) omnibus::say("Nothing deleted.")
			return(invisible(c(rasters = 0, vectors = 0)))
		}

		### get just UNLINKED rasters/vectors
		if (x == "unlinked") {

			# get all R objects
			if (is.null(pos)) {
				rObjs <- ls(envir = parent.frame(n = 1L))
			} else {
				rObjs <- ls(pos)
			}

			# get sources of R objects
			srcs <- character()
			for (i in seq_along(rObjs)) {

				thisObj <- get(rObjs[[i]])
				if (inherits(thisObj, c("GRaster", "GVector"))) {
					srcs <- c(srcs, sources(thisObj))
				}
			
			}

			# remove linked ones
			for (proj in projs) {
				grassObjs[[proj]] <- grassObjs[[proj]][!(grassObjs[[proj]] %in% srcs)]
			}

		} else if (x == "*") {
		
			# remove keep from list to be deleted
			if (is.list(keep)) {
			
				srcs <- character()
				for (i in seq_along(keep)) srcs <- c(srcs, sources(keep[[i]]))

				for (proj in projs) {
					grassObjs[[proj]] <- grassObjs[[proj]][!(grassObjs[[proj]] %in% srcs)]
				}

			}
		
		}

		out <- c(rasters = 0, vectors = 0)
		for (proj in projs) {
			out[["rasters"]] <- out[["rasters"]] + sum(names(grassObjs[[proj]]) == "raster")
			out[["vectors"]] <- out[["vectors"]] + sum(names(grassObjs[[proj]]) == "vector")
		}

		if (all(sapply(grassObjs, length) == 0)) {
			if (verbose | faster("verbose")) omnibus::say("Nothing deleted.")
			return(invisible(out))
		} else {

			if (verbose | faster("verbose")) omnibus::say("Deleting ", out[["rasters"]], " raster(s) and ", out[["vectors"]], " vector(s)...")
					
			for (proj in projs) {
				if (length(grassObjs[[proj]]) > 0L) {

					if (any(names(grassObjs[[proj]]) == "raster")) {

						.locationRestore(proj)
						.rm(grassObjs[[proj]], type = "raster", warn = FALSE, verify = FALSE)
					
					}
					if (any(names(grassObjs[[proj]]) == "vector")) {

						.locationRestore(proj)
						.rm(grassObjs[[proj]], type = "vector", warn = FALSE, verify = FALSE)
					
					}
				}
			}
		}
		invisible(return(invisible(out)))

	}

} # EOF

#' Refresh metadata in a GRaster or GVector
#'
#' `GRaster`s and `GVector`s are really pointers to objects in **GRASS**. The values displayed when you use `show()` or `print()` for a `GRaster` or `GVector` are stored in **R**. If, on the odd chance that you make a change to a `GRaster` or `GVector` (e.g., using commands in the **rgrass** package), the changes will not be automatically reflected in the `GRaster` or `GVector`. This function can be used to update the objects in **R** to reflect their proper values.
#'
#' @param object A `GRaster` or `GVector`.
#'
#' @returns A `GRaster` or `GVector`.
#'
#' @example man/examples/ex_GRaster.r 
#'
#' @aliases update
#' @rdname update
#' @exportMethod update
methods::setMethod(
    f = "update",
    signature = c(object = "GRaster"),
    definition = function(object) makeGRaster(sources(object), levels = cats(object), ac = activeCats(object))
)

#' @aliases update
#' @rdname update
#' @exportMethod update
methods::setMethod(
    f = "update",
    signature = c(object = "GVector"),
    definition = function(object) makeGVector(sources(object), table = as.data.table(object))
)

if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster
elev <- fast(madElev)

### Calculate input rasters

# Values below are just a guess
coeff_bh <- coeff_dh <- elev
coeff_bh[] <- 0.4
coeff_dh[] <- 0.6

slope <- terrain(elev, "slope")
aspect <- terrain(elev, "aspect", northIs0 = FALSE)

horizon_step <- 90
hh <- horizonHeight(elev, step = horizon_step, northIs0 = FALSE)

### calculate solar ir/radiance

solar <- sun(
	elevation = elev,
	coeff_bh = coeff_bh,
	coeff_dh = coeff_dh,
	slope = slope,
	aspect = aspect,
	hh = hh,
	horizon_step = horizon_step,
	albedo = 0.2,
	linke = 1.5,
	day = 1,
	step = 0.5,
	declination = NULL,
	solar_constant = 1367,
	
	distance_step = 1,
	npartitions = 1,

	beam_rad = TRUE,
	diff_rad = TRUE,
	refl_rad = TRUE,
	glob_rad = TRUE,
	insol_time = TRUE,

	lowMemory = FALSE
)

solar

}

if (grassStarted()) {

# Setup
library(terra)

# Elevation SpatRaster:
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster:
elev <- fast(madElev)

### Calculate zonal statistics using a GRaster as zones

# Generate a "zones" GRaster by dividing raster into areas based on
# high/low elevation.
names(elev) # Use this name in app() formula.
fun <- "= if (madElev <200, 0, if (madElev <400, 1, 2))"
zones <- app(elev, fun = fun)

# Calculate zonal statistics using a raster as zones
zonal(elev, zones, fun = "mean")
zonal(elev, zones, fun = "*") # all statistics

# Calculate zonal statistics on multi-layered GRaster
elev2 <- c(elev, log10(elev))
zonal(elev2, zones, fun = c("mean", "sum", "sdpop"))

### Calculate zonal statistics using a GVector as zones

madCoast4 <- fastData("madCoast4")
coast <- fast(madCoast4)

zonal(elev, z = coast, fun = "mean")

}

#' Select parts of polygons not shared between two GVectors
#'
#' @description The `xor()` function selects the area that does *not* overlap between two "polygon" `GVector`s. You can also use the `/` operator, as in `vect1 / vect2`.
#'
#' @param x,y `GVector`s.
#'
#' @returns A `GVector`.
#'
#' @seealso [crop()], [intersect()], [union()], [erase()]
#' 
#' @example man/examples/ex_union_intersect_xor_erase.r
#'
#' @aliases xor
#' @rdname xor
#' @exportMethod xor
methods::setMethod(
	f = "xor",
	signature = c(x = "GVector", y = "GVector"),
	function(x, y) {

	compareGeom(x, y, geometry = TRUE)
	if (geomtype(x) != "polygons") stop("Only polygon GVectors can be xored.")
	.locationRestore(x)
		
	src <- .makeSourceName("xor_v_overlay", "vector")
	rgrass::execGRASS(
		cmd = "v.overlay",
		ainput = sources(x),
		binput = sources(y),
		output = src,
		operator = "xor",
		snap = -1,
		flags = c(.quiet(), "overwrite")
	)
	makeGVector(src)
	
	} # EOF
)

#' Get or set the column with category labels in a categorical raster
#'
#' @description These functions return or set the column of the labels to be matched to each value in the raster of a categorical `GRaster` (see `vignette("GRasters", package = "fasterRaster")`). *Important*: Following [terra::activeCat()], the first column in the "levels" table is ignored, so an "active category" value of 1 means the second column is used as labels, a value of 2 means the third is used, and so on.
#' * `activeCat()` returns the column of the labels to be matched to each value in the raster for a single raster layer.
#' * `activeCats()` does the same, but for all layers of a `GRaster`.
#' * `activeCat()<-` sets the column to be used as category labels.
#'
#' @param x A categorical `GRaster`.
#' 
#' @param layer Numeric, integer, logical, or character: Indicates for which layer(s) to get or set the active category column. This can be a number (the index of the raster(s)), a logical vector (`TRUE` ==> get/set the active category column, `FALSE` ==> leave as-is), or a character vector (names of layers).
#'
#' @param names Logical: If `TRUE`, display the name(s) of the active column(s). If `FALSE` (default), report the index of the active column. Following [terra::activeCat()], the first column in the levels table is ignored. So, an active column of "1" means the second column is active. "2" means the third column is active, and so on.
#'
#' @param value Numeric, integer, or character. Following [terra::activeCat()], the first column in each levels table is ignored. So, if you want the second column to be the category label, use 1. If you want the third column, use 2, and so on. You can also specify the active column by its column name (though this can't be the first column's name).
#'
#' @returns `activeCat()` returns an integer or character of the active column index or name. `activeCats()` returns a vector of indices or names. `activeCat()<-` returns a `GRaster`.
#'
#' @example man/examples/ex_GRaster_categorical.r
#'
#' @aliases activeCat
#' @rdname activeCat
#' @exportMethod activeCat
methods::setMethod(
	f = "activeCat",
	signature = c(x = "GRaster"),
	function(x, layer = 1, names = FALSE) {
	
	layer <- .layerIndex(layer, x, recycle = TRUE)

	if (names) {

		ac <- x@activeCat
		out <- rep(NA_character_, length(layer))
		
		for (i in seq_along(layer)) {
			if (is.factor(x)[layer[i]]) out[i] <- names(cats(x)[[layer[i]]])[ac[layer[i]]]
		}
	
	} else {
		out <- x@activeCat[layer] - 1L
	}
	out

	} # EOF
)

#' @aliases activeCats
#' @rdname activeCat
#' @exportMethod activeCats
methods::setMethod(
	f = "activeCats",
	signature = c(x = "GRaster"),
	function(x, names = FALSE) {
	
	if (!names) {
		out <- x@activeCat - 1L
	} else if (names) {
		
		out <- rep(NA_character_, nlyr(x))
		facts <- is.factor(x)
		if (any(facts)) {
			for (i in which(facts)) {
				out[i] <- names(cats(x)[[i]])[activeCat(x, layer = i)]
			}
		}
	}
	out

	} # EOF
)

#' @aliases activeCat<-
#' @rdname activeCat
#' @exportMethod activeCat<-
methods::setMethod(
	f = "activeCat<-",
	signature = c(x = "GRaster"),
	function(x, layer = 1, value) {
	
	.activeCatAssign(x = x, layer = layer, value = value)
	
	} # EOF
)

#' @aliases activeCat<-
#' @rdname activeCat
#' @exportMethod activeCat<-
methods::setMethod(
	f = "activeCat<-",
	signature = c(x = "GRaster"),
	function(x, layer = 1, value) {
	
	.activeCatAssign(x = x, layer = layer, value = value)
	
	} # EOF
)

#' @aliases activeCat<-
#' @rdname activeCat
#' @exportMethod activeCat<-
methods::setMethod(
	f = "activeCat<-",
	signature = c(x = "GRaster"),
	function(x, layer = 1, value) {
	
	.activeCatAssign(x = x, layer = layer, value = value)
	
	} # EOF
)

#' Worker function for activeCat<-
#'
#' @param x A `GRaster`.
#' @param layer Numeric, integer, character, logical.
#' @param value Numeric or integer.
#'
#' @returns A `GRaster`.
#'
#' @noRd
.activeCatAssign <- function(x, layer, value) {

	layer <- .layerIndex(layer, x, recycle = TRUE)
	
	facts <- is.factor(x)
	if (any(!(which(facts) %in% layer))) stop("At least one layer is not categorical.\n  The active category cannot be set for this layer.")

	if (is.character(value)) {
		levs <- levels(x)
		for (i in layer) {
			value[i] <- match(value[i], names(levs[[i]]))
		}
	} else {
		value <- value + 1L
	}

	value <- as.integer(value)
	x@activeCat[layer] <- value
	methods::validObject(x)
	x

}

#' @name madDypsis
#'
#' @title Spatial points vector of records of Dypsis in eastern Madagascar
#'
#' @description Spatial points vector of herbarium specimens and observations of plants in the genus *Dypsis* (slender, evergreen palms) from a portion of eastern Madagascar.
#'
#' @docType data
#'
#' @usage data(madDypsis)
#'
#' @format An object of class `sf`.
#'
#' @keywords Dypsis Madagascar
#'
#' @source [Global Biodiversity Information Facility (GBIF)](https://gbif.org)
#'
#' @example man/examples/ex_fastData.r
#'
NULL

if (grassStarted()) {

# Setup
library(sf)

# Points vector of specimens of species in the plant genus Dypsis
madCoast0 <- fastData("madCoast0")

# Convert sf to a GVector:
coast <- fast(madCoast0)

### grid

# grid specified by number of cells in x-dimension
g1 <- grid(coast, nx = 10)
plot(coast, col = "cornflowerblue")
plot(g1, add = TRUE)

# grid specified by number of cells in x- and y-dimension
g2 <- grid(coast, nx = 10, ny = 5)
plot(coast, col = "cornflowerblue")
plot(g2, add = TRUE)

# grid specified by size of cells in both dimensions
g3 <- grid(coast, nx = 1250, ny = 2000, use = "size")
plot(coast, col = "cornflowerblue")
plot(g3, add = TRUE)

### hexagons

hexes <- hexagons(coast, ny = 10)
plot(hexes)
plot(coast, lwd = 2, add = TRUE)

hexes <- hexagons(coast, ny = 10, expand = c(0.3, 0.1))
plot(hexes)
plot(coast, lwd = 2, add = TRUE)

}

#' Increment category values of a "GRASS" vector
#'
#' @description Adds a constant to all category values of a **GRASS** vector. **This function is mostly of use to developers.**
#'
#' @param x A `GVector` or the [sources()] name of one.
#'
#' @param add Integer: Value to add to each category value.
#'
#' @returns The [sources()] name of a **GRASS** vector with category values incremented.
#'
#' @aliases .vIncrementCats
#' @rdname vIncrementCats
#' @keywords internal
.vIncrementCats <- function(x, add) {

	if (inherits(x, "GVector")) {
		.locationRestore(x)
		srcIn <- sources(x)
	} else {
		srcIn <- x
	}

	src <- .makeSourceName("v_category", "vector")
	rgrass::execGRASS(
		cmd = "v.category",
		input = srcIn,
		output = src,
		option = "sum",
		cat = add,
		flags = c(.quiet(), "overwrite", "t")
	)
	src

}

#' Display warning or message
#'
#' @description Display a warning or message if the given warning has not been displayed since **fasterRaster** was attached or if a given number or hours has passed since then.
#'
#' @param msg Character: Name for the message (used internally). Should be able to be assigned to a list (i.e., no spaces, punctuation, etc.).
#' 
#' @param message Text for the message.
#' 
#' @returns `TRUE` (invisibly).
#' 
#' @aliases .message
#' @rdname message
#' @keywords internal
.message <- function(msg, message) {

	# How long do we wait in hours to re-display same message?
	lag_hours <- 8
	lag_sec <- lag_hours * 60 * 60

	message <- paste0(message, "\n  This warning will not be displayed again until ", lag_hours, " hours have passed.")

	now <- Sys.time()
	if (any(names(.fasterRaster$messages) == msg)) {

		if (now - .fasterRaster$messages[[msg]] > lag_sec) {
			warning(message, immediate. = TRUE)
			utils::flush.console()
		    .fasterRaster$messages[[msg]] <- now
		}

	} else {

		.fasterRaster$messages[[msg]] <- now
		warning(message, immediate. = TRUE)
		utils::flush.console()

	}

	invisible(TRUE)
}

if (grassStarted()) {

# Setup
library(sf)

# Rivers vector and locations of Dypsis plants
madRivers <- fastData("madRivers")
madDypsis <- fastData("madDypsis")

# Convert sf's to GVectors:
dypsis <- fast(madDypsis)
rivers <- fast(madRivers)

### Connections from each point to nearest river
consFromDypsis <- connectors(dypsis, rivers)

plot(rivers, col = "blue")
plot(dypsis, add = TRUE)
plot(consFromDypsis, col = "red", add = TRUE)

### Connections from each river to nearest point
consFromRivers <- connectors(rivers, dypsis)

plot(rivers, col = "blue")
plot(dypsis, add = TRUE)
plot(consFromRivers, col = "red", add = TRUE)

}

#' @title Replace layers of a GRaster
#'
#' @description The `[[<-` operator can be used to replace a layer in a multi-layer `GRaster`.
#'
#' @param x A `GRaster`.
#'
#' @param i A numeric integer, integer, logical, or character: Indicates the layer to replace. If a logical vector, then the vector must have the same length as there are layers in `x`.
#'
#' @param value Either a `GRaster` or `NULL`: If `NULL`, then the layer indicated by `i` will be removed.
#'
#' @example man/examples/ex_GRaster_GVector_subset_assign.r
#'
#' @name [[<-
#' @aliases [[<-,GRaster,ANY-method
#' @docType methods
#' @rdname replace_double_square_brackets
#' @exportMethod [[<-
methods::setMethod(
	"[[<-",
	signature = c(x = "GRaster", i = "ANY"),
	function(x, i, value) {

	if (!inherits(value, c("GRaster", "NULL"))) stop("Cannot assign a non-GRaster to a GRaster layer.")
	
	if (is.logical(i)) {
		if (length(i) != nlyr(x)) stop("A logical vector used to select GRaster layers must have the same length as there are number of layers.")
		i <- which(i)
	} else if (is.character(i)) {
		i <- match(i, names(x))
		if (anyNA(i)) stop("At least one name does not appear in this raster layer\'s names.")
	}
	
	if (any(i < 1L)) stop("Cannot use negative indices.")

	nLayers <- nlyr(x)
	out <- x

	# "value" is NULL
	if (is.null(value)) {

		notNulls <- setdiff(seq_len(nLayers), i)

		if (length(notNulls) == 0L) {
			out <- NULL
		} else {

			out <- new(
				"GRaster",
				location = .location(out),
				mapset = .mapset(out),
				workDir = faster("workDir"),
				crs = crs(out),
				projection = .projection(out),
				nLayers = length(notNulls),
				dimensions = dim3d(out),
				topology = topology(out),
				extent = as.vector(ext(out)),
				zextent = zext(out),
				sources = sources(out)[notNulls],
				names = names(out)[notNulls],
				datatypeGRASS = datatype(out, "GRASS")[notNulls],
				resolution = res3d(out),
				minVal = .minVal(out)[notNulls],
				maxVal = .maxVal(out)[notNulls],
				activeCat = activeCat(out)[notNulls],
				levels = out@levels[notNulls]
			)
		
		}
		
	# "value" is a GRaster
	} else {

		if (length(i) != nlyr(value)) stop("The number of rasters to be assigned is different from the number of indices.")

		compareGeom(x, value)
		
		insides <- i[i <= nLayers]
		outsides <- i[i > nLayers]

		lenIns <- length(insides)
		lenOuts <- length(outsides)

		if (lenOuts > 0L) {

			if (any(outsides != seq_len(lenOuts) + nLayers)) stop("Indices are not sequential.")

			valueOutsides <- lenIns + seq_len(lenOuts)
			valueOutsides <- valueOutsides[lenIns + seq_len(lenOuts)]
			out <- c(x, value[[valueOutsides]])

		}

		if (lenIns > 0L) {
		
			valueInsides <- seq_len(lenIns)

			srcs <- sources(out)
			srcs[i] <- sources(value)[valueInsides]

			names <- names(out)
			names[insides] <- names(value)[valueInsides]

			dts <- datatype(out, "GRASS")
			dts[insides] <- datatype(value, "GRASS")[valueInsides]

			minVal <- .minVal(out)
			minVal[insides] <- .minVal(value)[valueInsides]

			maxVal <- .maxVal(out)
			maxVal[insides] <- .maxVal(value)[valueInsides]

			ac <- activeCat(out, layer=seq_len(nlyr(x)))
			ac[insides] <- activeCat(value)[valueInsides]

			levs <- levels(out)
			levs[insides] <- levels(value)[valueInsides]

			out <- new(
				"GRaster",
				location = .location(out),
				mapset = .mapset(out),
				workDir = faster("workDir"),
				crs = crs(out),
				projection = .projection(out),
				nLayers = nlyr(out),
				dimensions = dim(out)[1L:3L],
				topology = topology(out),
				extent = as.vector(ext(out)),
				zextent = zext(out),
				sources = srcs,
				names = names,
				datatypeGRASS = dts,
				resolution = res3d(out),
				minVal = minVal,
				maxVal = maxVal,
				activeCat = ac,
				levels = levs
			)

		}

	}
 	
	if (!is.null(out)) {
		if (length(anyDuplicated(out@names)) > 0L) out@names <- make.unique(out@names)
	}
	out

	} # EOF
)

#' Convert GVector to a data frame
#'
#' @description Convert a `GVector`'s data table to a `data.frame` or `data.table`.
#'
#' @param x A `GVector`.
#'
#' @returns A `data.frame` or `NULL` (if the `GRaster` has no data table).
#' 
#' @seealso [terra::as.data.frame()], [data.table::as.data.table()]
#' 
#' @example man/examples/ex_GRaster_GVector.r
#'
#' @aliases as.data.frame
#' @rdname as.data.frame
#' @exportMethod as.data.frame
methods::setMethod(
	f = "as.data.frame",
	signature = c(x = "GVector"),
	definition = function(x) {
		if (nrow(x) > 0L) {
			as.data.frame(x@table)
		} else {
			NULL
		}

	} # EOF
)

#' @aliases as.data.table
#' @rdname as.data.frame
#' @exportMethod as.data.table
methods::setMethod(
    f = "as.data.table",
    signature = c(x = "GVector"),
    definition = function(x) {
        if (nrow(x) > 0L) {
            x@table
        } else {
            NULL
        }
    } # EOF
)

#' Vegetation indices from surface reflectance
#'
#' @description This function calculates one of many types of vegetation indices from a raster with four bands representing blue (B), green (G), red (R), and near infrared (NIR), plus possibly channels 5 and 7. The function requires rasters that represent surface reflectance, so should have values that fall in the range 0 to 1, unless they are digital number rasters (e.g., integers in the range 0 to 255). If digital number format is used, then the `bits` argument should be defined.
#' 
#' @param x A `GRaster` with one layer per required band. Values should be between 0 and 1.
#' 
#' @param index Character or character vector: The vegetation index or indices to calculate. You can find a list of available indices using [fastData("vegIndices")][fastData] (also see [vegIndices]). The first column, "`index`" provides the name of the index, and these are the values that this argument will accept (e.g., "NDVI", "EVI2"). Partial matching is used, and case is ignored. You can also use these shortcuts:
#' * `"*"`: Calculate *all* indices
#' * `"RNIR"`: Calculate all indices that use R and NIR channels (but not other channels).
#' * `"NotSoil"`: Calculate all indices that use any channels but do not require `soilSlope` or `soilIntercept`.
#'
#' *Note*: A near-comprehensive table of indices can be found on the [Index Database: A Database for Remote Sensing Indices](https://www.indexdatabase.de).
#' 
#' @param r,g,b,nir Numeric or character: Index or [names()] of the layers in `x` that represent the red, green, blue, and near infrared channels. Values must be in the range from 0 to 1 or integers.
#' 
#' @param b5,b7 Numeric or character: Index of names of the layers representing bands 5 and 7. These are used only for GVI and PVI. Values must be in the range from 0 to 1 or integers.
#' 
#' @param bits Either `NULL` (default) or numeric integer or integer with a value of 7, 8, 10, or 16: If the rasters are represented by integers (so do not fall in the range of 0 to 1), then the number of bits can be supplied using `bits`. If this is the case, then they will range from 0 to 2^`n`, where `n` is 7, 8, 10, or 16. If bit rasters are supplied, they must be of [datatype()] "integer". If raster values are in the range from 0 to 1, then `bits` should be `NULL` (default).
#' 
#' @param soilSlope,soilIntercept,soilNR Numeric: Values of the soil slope, intercept, and soil noise reduction factor (0.08, by default). Used only for calculation of MSAVI.
#' 
#' @returns A `GRaster`.
#' 
#' @seealso **GRASS** manual page for tool `i.vi` (see `grassHelp("i.vi")`)
#' 
#' @example man/examples/ex_vegIndex.r
#' 
#' @aliases vegIndex
#' @rdname vegIndex
#' @exportMethod vegIndex
methods::setMethod(
	f = "vegIndex",
	signature = c(x = "GRaster"),
	function(
		x,
		index = "NDVI",
		r = NULL,
		g = NULL,
		b = NULL,
		nir = NULL,
		b5 = NULL,
		b7 = NULL,
		soilSlope = NULL,
		soilIntercept = NULL,
		soilNR = 0.08,
		bits = NULL
	) {

	mm <- minmax(x)
	if (any(mm < 0)) stop("Raster values cannot be < 0.")

	if (is.null(bits)) {

		if (any(mm > 1)) stop("Raster values should be between 0 and 1 if `bits` is NULL.")

	} else {

		if (!all(is.int(x))) stop("If argument `bits` is defined, then rasters must be of type `integer`.")

		maxVal <- 2^bits
		if (any(mm > maxVal)) stop("Maximum value for ", bits, " bits is ", maxVal, ".\n  At least one raster surpasses this.")

	}

	.locationRestore(x)
	.region(x)

	vegIndices <- NULL
	utils::data("vegIndices", envir = environment(), package = "fasterRaster")

	index <- tolower(index)
	if ("*" %in% index) {
		index <- vegIndices$index
	} else if ("notsoil" %in% index) {
    	index <- c(
			index,
			vegIndices[!vegIndices$soilSlope & !vegIndices$soilIntercept, "index", drop = TRUE]
		)
	} else if ("rnir" %in% index) {
    	index <- c(
			index,
			vegIndices[vegIndices$R & !vegIndices$G & !vegIndices$B & vegIndices$NIR & !vegIndices$channel5 & !vegIndices$channel7 & !vegIndices$soilSlope & !vegIndices$soilIntercept, "index", drop = TRUE]
		)
		index <- index[index != "RNIR"]
		index <- index[index != "rnir"]

	}

	# index <- sort(index)
	indexOrder <- order(index)
	index <- sort(index)
	index <- omnibus::pmatchSafe(index, vegIndices$index, useFirst = TRUE)
	index <- index[indexOrder]

	index <- tolower(index)

	if (!is.null(r) && is.character(r)) r <- match(names(x), r)
	if (!is.null(g) && is.character(g)) g <- match(names(x), g)
	if (!is.null(b) && is.character(b)) b <- match(names(x), b)
	if (!is.null(nir) && is.character(nir)) nir <- match(names(x), nir)
	if (!is.null(b5) && is.character(b5)) b5 <- match(names(x), b5)
	if (!is.null(b7) && is.character(b7)) b7 <- match(names(x), b7)

	srcs <- .makeSourceName(index, "raster")
	for (i in seq_along(index)) {

		args <- list(
			cmd = "i.vi",
			viname = index[i],
			output = srcs[i],
			flags = c(.quiet(), "overwrite"),
			intern = TRUE
		)

		if (!is.null(bits)) args$storage_bit <- bits

		if (!is.null(r)) args$red <- sources(x)[r]
		if (!is.null(g)) args$green <- sources(x)[g]
		if (!is.null(b)) args$blue <- sources(x)[b]
		if (!is.null(nir)) args$nir <- sources(x)[nir]
		
		if (!is.null(b5)) args$band5 <- sources(x)[b5]
		if (!is.null(b7)) args$band7 <- sources(x)[b7]
		
  		if (!is.null(soilSlope)) args$soil_line_slope <- soilSlope
  		
		if (!is.null(soilIntercept)) args$soil_line_slope <- soilIntercept

  		if (!is.null(soilNR)) args$soil_noise_reduction <- soilNR

		do.call(rgrass::execGRASS, args = args)

	} # next metric
	makeGRaster(srcs, toupper(index))

	} #OF
)

#' Call GRASS `g.region` tool
#'
#' This function calls the **GRASS** tool `g.region` to display information on the region of the current **GRASS** "project".
#'
#' @returns Displays current region information for the active "project/location" in **GRASS**.
#'
#' @export
#' @keywords internal
.g.region <- function() {
	out <- rgrass::execGRASS("g.region", flags = "p", intern = TRUE)
	out
}

#' Identify terrain feature types
#'
#' @description Geomorphons are idealized terrain types calculated from an elevator raster based on a moving window of a given size. The window is a torus (which can have an inner radius of 0, so can also be a circle), which allows it to identify geomorphons of a given size while ignoring ones larger or smaller. There are 10 basic geomorphons. Consult the the manual for **GRASS** tool `r.geomorphon` using `grassHelp("r.geomorphon")` for more details and diagrams of each type of geomorphon. Geomorphon types include:
#' 1. Flat areas: Focal area has approximately the same elevation as surrounding areas
#' 2. Pits: An area is lower than all other surrounding areas
#' 3. Valley: Focal area has elevation similar to two opposing side of the window but lower than the other two opposing sides
#' 4. Footslope: Focal region is at the "bottom" of a slope
#' 5. Hollow: A small valley/indention in the crest of a hill
#' 6. Slope: Cells in the window form an approximately uniform slope
#' 7. Spur: An extrusion at the foot of a hill (i.e.,, a small hill extending out from the foot of a slope)
#' 8. Shoulder: The crest of a slope
#' 9. Ridge: Opposite of a valley; focal area is higher than two opposing sides but approximately the same elevation as the other two opposing sides
#' 10. Peak: Focal area is higher than any other in the window
#'
#' @param x A single-layer `GRaster`, typically representing elevation.
#'
#' @param inner,outer Integer: Inner and outer radii of the torus used to identify geomorphons, in cells or meters (set by argument `unit`). The inner default value is 0 and the outer default value is 3. The `outer` radius sets the maximum size of a geomorphon that that can be identified, and `inner` sets the smallest size. If `unit` is `"meters"`, the value of `outer` must be larger than the smaller dimension of any cell in the east-west and north-south directions.
#'
#' @param unit Character: Units of `inner` and outer; can be either `"cells"` (default) or `"meters"`. Partial matching is used.
#'
#' @param flat Numeric value >= 0: Minimum difference (in degrees) between the focal area areas around it for a geomorphon to be considered as "flat". Larger cells (i.e., ~1 km resolution or larger) require smaller values (<<1) to correctly identify flat areas. Higher values result in more areas being classified as "flat" geomorphons. The default value is 1.
#'
#' @param flatDist Numeric: Distance (in meters) to correct for the effect of large distances on the diminished capacity to identify "flat" geomorphons. If the distance between the focal area and a surrounding area surpasses this distance, then the effective value of `flat` will be reduced
#'
#' @param mode Character: Method for implementing the zenith/line-of-site search. Partial matching is used:
#' * `"1"` (default): The "original" geomorphon mode (in **GRASS** tool `r.geomorphon`, the "anglev1" method)
#' * `"2"`: Better handling of cases with equal zenith/nadir angles (the "anglev2" method)
#' * `"2d"`: As `"2"`, but takes into account zenith/nadir distance ("anglev2_distance" method)
#'
#' @returns A categorical `GRaster` where each geomorphon is a category (see `vignette("GRasters", package = "fasterRaster")`).
#'
#' @seealso **GRASS** manual for tool `r.geomorphon` (see `grassHelp("r.geomorphon")`)
#'
#' @example man/examples/ex_geomorphons.r
#'
#' @aliases geomorphons
#' @rdname geomorphons
#' @exportMethod geomorphons
methods::setMethod(
	f = "geomorphons",
	signature = c(x = "GRaster"),
	function(
		x,
		inner = 0,
		outer = 3,
		unit = "cells",
		flat = 1,
		flatDist = 0,
		mode = "1"
	) {
	
	unit <- omnibus::pmatchSafe(unit, c("cells", "meters"), nmax = 1L)
	mode <- as.character(mode)
	if (mode == "1") {
		comparison <- "anglev1"
	} else if (mode == "2") {
		comparison <- "anglev2"
	} else if (mode == "2d") {
		 comparison <- "anglev2_distance"
	} else {
		stop("Argument `mode` can only be `1`, `2`, or `2d`.")
	}

	if (nlyr(x) > 1L) stop("This function requires `x` to have just one layer.")

	.locationRestore(x)
	.region(x)

	src <- .makeSourceName("geomorphon_r_geomorphon", "raster")
	args <- list(
		cmd = "r.geomorphon",
		elevation = sources(x),
		forms = src,
		search = outer,
		skip = inner,
		flat = flat,
		dist = flatDist,
		comparison = comparison,
		flags = c(.quiet(), "overwrite")
	)

	if (unit == "meters") args$flags <- c(args$flags, "m")

	do.call(rgrass::execGRASS, args = args)

	levels <- data.table::data.table(
		value = 1L:10L,
		geomorphon = c("flat", "peak", "ridge", "shoulder", "spur", "slope", "hollow", "footslope", "valley", "pit")
	)

	makeGRaster(src, levels = levels)

	} # EOF
)

if (grassStarted()) {

# Setup
library(sf)

# Points vector of specimens of species in the plant genus Dypsis
madDypsis <- fastData("madDypsis")

# Convert sf to a GVector:
dypsis <- fast(madDypsis)

### Convex hull for all plant specimens together:
ch <- convHull(dypsis)

### Convex hull for each species:
head(dypsis) # See the "rightsHolder" column?
chHolder <- convHull(dypsis, by = "rightsHolder")

### Plot:
plot(dypsis)
plot(ch, add = TRUE)
n <- length(chHolder)
for (i in 1:n) {
   plot(chHolder[[i]], border = i, add = TRUE)
}

}

#' @title Classes for "fasterRaster" locations, rasters, and vectors
#'
#' @aliases GSpatial
#' @rdname GLocation
#' @exportClass GSpatial
GSpatial <- methods::setClass(
	Class = "GSpatial",
	contains = "GLocation",
	slots = list(
		topology = "character",			# 2D or 3D
		extent = "numeric",				# horizontal extent (4 numerics)
		zextent = "numeric",			# vertical extent (2 numerics)
		sources = "character"			# name in GRASS
	),
	prototype = prototype(
		topology = NA_character_,
		extent = c(NA_real_, NA_real_, NA_real_, NA_real_),
		zextent = c(NA_real_, NA_real_),
		sources = NA_character_
	)
)

methods::setValidity("GSpatial",
	function(object) {
		if (length(object@location) != 1L) {
			"@location can only be a single character string."
		} else if (length(object@mapset) != 1L) {
			"@mapset can only be a single character string."
		} else if (length(object@crs) != 1L) {
			"@crs can only be a single character string."
		} else if (!all(object@topology %in% c(NA_character_, "2D", "3D"))) {
			paste0("@topology can only be a NA, `2D` or `3D`.")
		} else if (object@topology == "3D" && any(is.na(object@zextent))) {
			paste0("@topology is `3D` but @zextent has at least one NA value.")
		} else if (!anyNA(object@zextent[1L]) & is.na(object@zextent[2L]) | (is.na(object@zextent[1L]) & !is.na(object@zextent[2L]))) {
			"Both values of @zextent must be NA or must be numeric values."
		} else if (!anyNA(object@zextent) && object@zextent[2L] < object@zextent[1L]) {
			"Bottom value of @zextent is greater than the top value of @zextent."
		} else {
			TRUE
		}
	} # EOF
)

# GSpatial <- function(
	# location = NA_character_,
	# mapset = NA_character_,
	# crs = NA_character_,
	# sources = NA_character_,
	# topology = NA_character_,
	# extent = c(NA_real_, NA_real_, NA_real_, NA_real_),
	# ztop = NA_real_,
	# zbottom = NA_real_
# ) {
	# methods::new(
		# "GSpatial",
		# location = location,
		# mapset = mapset,
		# crs = crs,
		# sources = sources,
		# topology = topology,
		# extent = extent,
		# ztop = ztop,
		# zbottom = zbottom
	# )
# }

### Re-orient numeric values:
facings <- c(0, 90, 180, 270, 360)
reorient(facings)

# Re-reorienting returns the same values:
reorient(reorient(facings))

if (grassStarted()) {

### Re-orient a GRaster:

# Setup
library(terra)
madElev <- fastData("madElev")
elev <- fast(madElev)

# Calculate aspect in degrees, using north orientation:
aspectNorth <- terrain(elev, "aspect")

# Re-orient to east-facing:
aspectEast <- reorient(aspectNorth)

# Re-reorienting is the same, to within rounding error:
aspectNorth - reorient(reorient(aspectNorth))

# Plot:
aspects <- c(aspectNorth, aspectEast)
names(aspects) <- c("north_orientation", "east_orientation")
plot(aspects)


}

#' Simplify the geometry of a vector
#'
#' @description `simplifyGeom()` reduces the number of vertices used to represent a vector (i.e., to save memory or disk space). There are several methods available.
#'
#' @param x A `GVector`.
#'
#' @param method Character: Method used to reduce the number of vertices. Partial matching is used, and case does not matter:
#'
#' * `"VR"`: Vertex reduction (default, simplest): If two points p1 and p2 on the same line are closer than the threshold, remove p2. The `tolerance` argument represents this threshold distance.
#' * `"DP"`: Douglas-Peucker (AKA Ramer-Douglas-Peucker) algorithm: Simply stated, for points p1, p2, and p3 on a line, this method constructs a line segment between p1 and p3. If p2 is closer than the threshold to the line segment, it is removed. In this example, the `tolerance` argument refers to the maximum distance between p2 and the line segment.
#' * `"DPR"`: Douglas-Peucker algorithm with reduction: As the Douglas-Pueker method, but each geometry is thinned so that in the end it has only a given proportion of the starting number of points. The `prop` argument refers to this proportion of remaining points.
#' * `"RW`: Reumann-Witkam algorithm: For points p1, p2, p3, and p4 on a line, constructs two line segments parallel to the line segment defined by p1 and p4. These are placed `tolerance` distance one either side of the p1-p4 line segment. If the line segment p1-p2 or p3-p4 falls entirely within the bounds of the two outer parallel segments, p2 and p3 are removed, leaving just p1 and p4.
#'
#' @param tolerance Numeric >= 0: Threshold distance in map units (degrees for unprojected, usually meters for projected). If `NULL`, then 2% of the minimum of the x-, y-, and z-extent will be used.
#'
#' @param prop Positive value between 0 and 1: Proportion of points that will be retained for each geometry when the Douglas-Peucker algorithm with reduction is applied (ignored otherwise). Default is 0.5 (retain 50% of vertices).
#'
#' @seealso [smoothGeom()], [geometry cleaning][breakPolys], [terra::simplifyGeom()], **GRASS** manual page for tool `v.generalize` (see `grassHelp("v.generalize")`)
#'
#' @returns A `GVector`.
#'
#' @example man/examples/ex_simplify_smooth_clean_GVector.r
#'
#' @aliases simplifyGeom
#' @rdname simplifyGeom
#' @exportMethod simplifyGeom
methods::setMethod(
	f = "simplifyGeom",
	signature = c(x = "GVector"),
	function(x, tolerance = NULL, method = "VR", prop = 0.5) {

	# automatic distance
	if (is.null(tolerance)) {
		extent <- ext(x, vector = TRUE)
		xext <- extent[2L] - extent[1L]
		yext <- extent[4L] - extent[3L]
		if (is.3d(x)) {
			zext <- zext(x)
			zext <- diff(zext)
		} else {
			zext <- NULL
		}
		tolerance <- 0.02 * min(xext, yext, zext, na.rm = TRUE)
	} else {
		if (tolerance < 0) stop("Argument `tolerance` must be > 0.")
	}


	.locationRestore(x)
	
	method <- tolower(method)
	method <- omnibus::pmatchSafe(method, c("vr", "dp", "dpr", "rw"), useFirst = TRUE)
	method <- if (method == "vr") {
		"reduction"
	} else if (method == "dp") {
		"douglas"
	} else if (method == "dpr") {
		"douglas_reduction"
	} else if (method == "rw") {
		"reumann"
	}
	
	if (method == "reumann" && (prop < 0 | prop > 1)) stop("Argument `prop` must be in the range [0, 1].")
	
	src <- .makeSourceName("generalized", "vect")
	args <- list(
		cmd = "v.generalize",
		input = sources(x),
		output = src,
		method = method,
		threshold = tolerance,
		flags = c(.quiet(), "overwrite"),
		intern = TRUE
	)

	if (method == "reumann") args$reduction <- 100 * prop
	
	do.call(rgrass::execGRASS, args=args)
	makeGVector(src)

	} # EOF
)

#' Landscape fragmentation class following Riitters et al. (2020)
#'
#' @description Riitters et al. (2020) propose a classification scheme for forest fragmentation (which can be applied to any habitat type). The scheme relies on calculating density (e.g., number of forested cells in a window around a focal cell) and connectivity (number of cases where neighboring cells are both forested). This function calculates these classes from a `GRaster` or `SpatRaster` in which the focal habitat type has cell values of 1, and non-focal habitat type has cell values of 0 or `NA`.
#'
#' Note that by default, the `SpatRaster` and `GRaster` versions will create different results around the border of the raster. The `SpatRaster` version uses the [terra::focal()] function, which will not return an `NA` value when its window overlaps the raster border if the `na.rm` argument is `TRUE`. However, the `GRaster` version uses the **GRASS** tool `r.neighbors`, which does return `NA` values in these cases.
#'
#' The fragmentation classes are:
#' * Value provided by `none`: None (i.e., no forest; default is `NA`).
#' * 1: Patch
#' * 2: Transitional
#' * 3: Perforated
#' * 4: Edge
#' * 5: Undetermined (not possible to obtain when `w = 3`)
#' * 6: Interior
#'
#' @param x A `SpatRaster` or `GRaster` wherein the habitat type has cell values of 1, and non-focal habitat type has cell values of 0 or `NA`
#'
#' @param w An odd integer between 3 and 25: Size of the window across which fragmentation is calculated (in units of "rows" and "columns"). The default is 3, meaning the function uses a 3x3 moving window to calculate fragmentation. For large rasters, compute time is ~*O*(`N`) + *O*(`N * w^2`), where `N` is the number of cells in the raster. So, even a small increase in `w` can increase compute time by a lot. Values >25 will fail. A workaround might be to use [aggregate()] with `fun = 'mode'`, apply `fragmentation()`, then `resample()` to the original resolution.
#'
#' @param undet Character: How to assign the "undetermined" case. Valid values are `"perforated"` (default), `"edge"`, and `"undetermined"`. Partial matching is used. If `Pf` is the proportional density raster cell value and `Pff` the proportional connectivity raster cell value, the undetermined case occurs when `Pf` > 0.6 and `Pf == Pff`.
#'
#' @param none Integer or `NA` (default): Value to assign to a cell with no focal habitat. Riitters et al. use `NA`. This will be forced to an integer if it is not an actual integer.
#'
#' @param na.rm Logical: If `TRUE` (default) and `x` is a `SpatRaster`, then cells near the edge of the raster where the window overlaps the edge can still be assigned a fragmentation class. If `FALSE`, these cells will be assigned a value of `none`.
#' 
#' @param cores Integer: Number of processor cores to use for when processing a `SpatRaster`.
#'
#' @param verbose Logical: If `TRUE` (default), display progress.
#'
#' @returns A categorical `SpatRaster` or `GRaster`. The values assigned to each class can be seen with [levels()].
#'
#' @references Riitters, K., J. Wickham, R. O'Neill, B. Jones, and E. Smith. 2000. Global-scale patterns of forest fragmentation. *Conservation Ecology* 4:3. URL: [http://www.consecol.org/vol4/iss2/art3/](http://www.consecol.org/vol4/iss2/art3/). Also note the [errata](https://www.ecologyandsociety.org/vol4/iss2/art3/errata/january26.2001.html).
#' 
#' @example man/examples/ex_fragmentation.r
#'
#' @aliases fragmentation
#' @rdname fragmentation
#' @exportMethod fragmentation
methods::setMethod(
	f = "fragmentation",
	signature = c(x = "SpatRaster"),
	function(x, w = 3, undet = "undetermined", none = NA, na.rm = TRUE, cores = faster("cores"), verbose = TRUE) {

	# errors?
	if (w > 25 | !omnibus::is.wholeNumber(w) | w < 1L | w %% 2 == 0) stop("Argument `w` must be an odd, positive integer in the range from 3 to 25.")
	undet <- omnibus::pmatchSafe(undet, c("undetermined", "perforated", "edge"), nmax = 1L)

	# setup
	nLayers <- terra::nlyr(x)

	# zeros matrix to store each set of neighbors
	ww <- matrix(0L, w, w)

	# assess if each pair of neighboring cell in cardinal direction has 1 or 2 forested cells or not (0 = no, 1 = 1 cell, 2 = 2 cells)
	# 1st make a list of possible cell neighbors by cell index
	# 2nd assess if neighbors both have target habitat type
	cells <- list()
	w1 <- matrix(1L:(w^2), nrow = w, ncol = w)
	w2 <- w1[1L:(w - 1), 1L:(w - 1)]

	for (i in 1L:(w - 1L)^2) {
			
		len <- length(cells)
		wFrom <- w2[i]
		
		# left-to-right neighbors
		cells[[len + 1L]] <- c(wFrom, wFrom + w)

		# top-to-bottom neighbors
		cells[[len + 2L]] <- c(wFrom, wFrom + 1L)

	}

	# add left-to-right neighbors of bottom row
	for (i in w1[w, 1L:(w - 1L)]) {
	
		len <- length(cells)
		wFrom <- w1[i]
		cells[[len + 1L]] <- c(wFrom, wFrom + w)

	}

	# add top-to-bottom neighbors of right column
	for (i in w1[1L:(w - 1L), w]) {
	
		len <- length(cells)
		wFrom <- w1[i]
		cells[[len + 1L]] <- c(wFrom, wFrom + 1L)

	}

	# for each layer
	nSteps <- nLayers * (1 + length(cells) + 17)
	if (verbose | faster("verbose")) {
		pb <- utils::txtProgressBar(min = 0, max = nSteps, initial = 0, style = 3, width = 30)
	}
	steps <- 0

	for (i in seq_len(nLayers)) {

		xx <- x[[i]]

		# convert NAs to 0s
		steps <- steps + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, steps)
		# xx <- terra::app(xx, fun = function(x) ifelse(is.na(x), 0L, x), cores = cores) # VERY slow!!!
		xx[is.na(x)] <- 0L

		unis <- unique(xx)
		unis <- unis[ , 1L]
		if (length(unis) != 2L) {
			stop("The raster must have values of 0 and 1, NA and 1,\n  or the values must be all NA, all 0, or all 1.")
		}

		connectivities <- list()
		for (j in seq_along(cells)) {

			steps <- steps + 1
			if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, steps)
		
			pair <- cells[[j]]
			www <- ww
			www[pair] <- 1L

			connectivities[[j]] <- focal(xx, w = www, fun = "sum", na.rm = TRUE)

		}

		# stack pff rasters
		steps <- steps + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, steps)
		pffConnected <- do.call(c, connectivities)

		# indicate if at least one cell of pair is forested
		steps <- steps + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, steps)
		pffEither <- pffConnected >= 1L

		# convert to 1 if both cells are forested
		steps <- steps + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, steps)
		pffConnected <- pffConnected == 2L

		# sum forested neighbors across rasters
		steps <- steps + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, steps)
		pffNeighs <- sum(pffConnected)

		# sum either forested across rasters
		steps <- steps + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, steps)
		pffForests <- sum(pffEither)

		# calculate pff
		steps <- steps + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, steps)
		pff <- pffNeighs / pffForests

		# calculate cover
		steps <- steps + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, steps)
		pf <- terra::focal(xx, w = w, "mean", na.rm = na.rm)

		### assign classes
		steps <- steps + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, steps)
		thisOut <- xx * 0L

		# pre-calculate interior mask
		steps <- steps + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, steps)
		interiorMask <- pf == 1 & pff == 1
		notInteriorMask <- 1L - interiorMask

		# pre-calculate difference
		steps <- steps + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, steps)
		delta <- pf - pff

		# patch
		steps <- steps + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, steps)
		mask <- pf > 0 & pf <= 0.4
		thisOut <- terra::mask(thisOut, mask, maskvalues = 1L, updatevalue = 1L)

		# transitional
		steps <- steps + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, steps)
		mask <- pf > 0.4 & pf <= 0.6
		thisOut <- terra::mask(thisOut, mask, maskvalues = 1L, updatevalue = 2L)

		# perforated
		steps <- steps + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, steps)
		if (undet == 'perforated') {
			mask <- pf > 0.6 & delta >= 0
		} else {
			mask <- pf > 0.6 & delta > 0
		}
		thisOut <- terra::mask(thisOut, mask, maskvalues = 1L, updatevalue = 3L)

		# edge
		steps <- steps + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, steps)
		if (undet == 'edge') {
			mask <- pf > 0.6 & delta <= 0
		} else {
			mask <- pf > 0.6 & delta < 0
		}
		thisOut <- terra::mask(thisOut, mask, maskvalues = 1L, updatevalue = 4L)

		# undetermined
		steps <- steps + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, steps)
		if (undet == 'undetermined') {
			mask <- pf > 0.6 & delta == 0
			thisOut <- terra::mask(thisOut, mask, maskvalues = 1L, updatevalue = 5L)
		}

		# interior
		steps <- steps + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, steps)
		thisOut <- terra::mask(thisOut, interiorMask, maskvalues = 1L, updatevalue = 6L)

		# no focal habitat in focal cell
		# Riitters et al.: "If the center pixel was not forest, then a null value was assigned to that location."
		steps <- steps + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, steps)
		thisOut <- terra::mask(thisOut, xx, maskvalues = 0L, updatevalue = none)

		levs <- .fragmentationLevels(undet = undet, none = none)
		levels(thisOut) <- levs

		if (i == 1L) {
			out <- thisOut
		} else {
			out <- c(out, thisOut)
		}

	} # next layer
	if (verbose | faster("verbose")) close(pb)

	out

	} # EOF

)

#' @aliases fragmentation
#' @rdname fragmentation
#' @exportMethod fragmentation
methods::setMethod(
	f = "fragmentation",
	signature = c(x = "GRaster"),
	function(x, w = 3, undet = "undetermined", none = NA, verbose = TRUE) {

	undet <- omnibus::pmatchSafe(undet, c("perforated", "edge", "undetermined"), nmax = 1L)

	mm <- minmax(x)
	msg <- "Raster must be binary (only 1 and/or 0, with NAs allowed."
	if (any(!(mm %in% c(0L, 1L)))) stop(msg)

	# force to integer to obviate issues with floating-point precision
	dtype <- datatype(x)
	if (any(dtype != "integer")) {
		nonIntegers <- which(dtype != "integer")
		for (i in nonIntegers) {
			x[[i]] <- as.int(x[[i]])
		}
	}

	# setup
	.locationRestore(x)
	.region(x)

	nLayers <- terra::nlyr(x)

	# assess if each pair of neighboring cell in cardinal direction has 1 or 2 forested cells or not (0 = no, 1 = 1 cell, 2 = 2 cells)
	ww <- matrix(0L, w, w)

	#### calculate connectivities for Pff

	# list of offset pairs relative to focal cell in cardinal directions
	offsets <- list()
	center <- (w + 1L) / 2L
	
	# get left-right cell pairs referenced by offset from window center
	i <- 1L
	for (row in seq_len(w)) {
		for (col in 1L:(w - 1L)) {

			offsets[[i]] <- c(row - center, col - center, row - center, col - center + 1L)
			i <- i + 1L

		}
	}

	# get up-down cell pairs referenced by offset from window center
	i <- length(offsets) + 1L
	for (row in 1L:(w - 1L)) {
		for (col in seq_len(w)) {

			offsets[[i]] <- c(row - center, col - center, row - center + 1L, col - center)
			i <- i + 1L

		}
	}

	# for each layer
	nSteps <- nLayers * (2 + length(offsets) + 6) + 1
	if (verbose | faster("verbose")) {
		pb <- utils::txtProgressBar(min = 0, max = nSteps, initial = 0, style = 3, width = 30)
	}
	steps <- 0

	# to keep `sources` of temporary files for deletion
	created <- character()

	srcs <- rep(NA_character_, nLayers)
	for (i in seq_len(nLayers)) {

		### remove NAs
		steps <- steps + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, steps)
		srcIn <- sources(x)[i]
		srcXZeros <- .makeSourceName("fragmentation_xSansNAs", "raster")
		ex <- paste0(srcXZeros, " = if(isnull(", srcIn, "), 0, ", srcIn, ")")
		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		created <- c(created, srcXZeros)

		### connectivities (for Pff)
		n <- length(offsets)
		srcConnectivities <- paste0('c', 1:n) # .makeSourceName("frag_connect", "raster", n = length(offsets))
		srcEither <- paste0('e', 1:n) # .makeSourceName("frag_either", "raster", n = length(offsets))
		srcBoth <- paste0('b', 1:n) # .makeSourceName("frag_both", n = length(offsets))
		for (j in seq_along(offsets)) {

			steps <- steps + 1
			if (verbose || faster("verbose")) utils::setTxtProgressBar(pb, steps)

			# tally occurrences of habitat in each pair of cells		
			y1 <- offsets[[j]][1L]
			x1 <- offsets[[j]][2L]
			
			y2 <- offsets[[j]][3L]
			x2 <- offsets[[j]][4L]

			# sum of pairs of cells
			ex <- paste0(srcConnectivities[j], " = ", srcXZeros, "[", y1, ",", x1, "]+", srcXZeros, "[", y2, ",", x2, "]")
			rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

			# does either cell in a pair have habitat?
			ex <- paste0(srcEither[j], " = if(", srcConnectivities[j], ">= 1)")
			rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

			# do both cells have habitat?
			ex <- paste0(srcBoth[j], " = if(", srcConnectivities[j], "==2)")
			rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		}
		created <- c(created, srcConnectivities, srcEither, srcBoth)

		# number of neighbor cell pairs with at least one with habitat
		steps <- steps + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, steps)
		srcNumEithers <- .makeSourceName("fragmentation_sum_eithers", "raster")
		ex <- paste0(srcNumEithers, " = ", paste(srcEither, collapse = "+"))
		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		created <- c(created, srcNumEithers)

		# number of neighbor cell pairs both with habitat
		steps <- steps + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, steps)
		srcNumBoths <- .makeSourceName("fragmentation_sum_both", "raster")
		ex <- paste0(srcNumBoths, " = ", paste(srcBoth, collapse = " + "))
		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		created <- c(created, srcNumBoths)

		# calculate Pff
		steps <- steps + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, steps)
		srcPff <- .makeSourceName("fragmentation_pff", "raster")
		ex <- paste0(srcPff, " = double(", srcNumBoths, ") / double(", srcNumEithers, ")")
		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		created <- c(created, srcPff)

		# calculate cover (Pf)
		steps <- steps + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, steps)
		srcPf <- .makeSourceName("fragmentation_pf", "raster")
		rgrass::execGRASS(
			cmd = "r.neighbors",
			input = srcXZeros,
			output = srcPf,
			size = w,
			method = "average",
			nprocs = faster("cores"),
			memory = faster("memory"),
			flags = c(.quiet(), "overwrite")
		)

		created <- c(created, srcPf)

		# calculate difference between Pf and Pff (used for "perforated", "edge", and "undetermined" cases)
		steps <- steps + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, steps)
		srcDelta <- .makeSourceName("fragmentation_delta", "raster")
		ex <- paste0(srcDelta, " = ", srcPf, " - ", srcPff)
		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		created <- c(created, srcDelta)

		### assign classes
		# NB: The first condition in each forces cases where focal cell does not have focal habitat to NA as per Riitters et al.: "If the center pixel was not forest, then a null value was assigned to that location."
		src <- .makeSourceName("fragmentation_class", "raster")
		
		if (is.na(none)) {
			noneValue <- "null()"
		} else {
			noneValue <- as.int(none)
		}

		if (undet == "undetermined") {
		
			ex <- paste0(src, " = if (", srcXZeros, " == 0, ", noneValue, ", if (", srcPf, " == 0, 0, if (", srcPf, " == 1 & ", srcPff, " == 1, 6, if (", srcPf, " <= 0.4, 1, if (", srcPf, " <= 0.6, 2, if (", srcDelta, " > 0, 3, if (", srcDelta, " < 0, 4, if (", srcDelta, " == 0, 5, null()))))))))")

		} else if (undet == "edge") {
		
			ex <- paste0(src, " = if (", srcXZeros, " == 0, ", noneValue, ", if (", srcPf, " == 0, 0, if (", srcPf, " == 1 & ", srcPff, " == 1, 6, if (", srcPf, " <= 0.4, 1, if (", srcPf, " <= 0.6, 2, if (", srcDelta, " > 0, 3, if (", srcDelta, " <= 0, 4, null())))))))")

		} else if (undet == "perforated") {
		
			ex <- paste0(src, " = if (", srcXZeros, " == 0, ", noneValue, ", if (", srcPf, " == 0, 0, if (", srcPf, " == 1 & ", srcPff, " == 1, 6, if (", srcPf, " <= 0.4, 1, if (", srcPf, " <= 0.6, 2, if (", srcDelta, " >= 0, 3, if (", srcDelta, " < 0, 4, null())))))))")
		
		}
		
		steps <- steps + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, steps)
		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		# ex <- paste0(src, " = ",
		# 	"if (pf == 0, 0,
		# 		if (pf == 1 & pff == 1, 6,
		# 			if (pf <= 0.4, 1,
		# 				if (pf <= 0.6, 2,
		# 					if (pf - pff > 0, 3,
		# 						if (pf - pff < 0, 4,
		# 							if (pf == pff, 5, null())
		# 						)
		# 					)
		# 				)
		# 			)
		# 		)
		# 	)"
		# )

		# # if masking to just focal habitat
		# if (restrict) {

		# 	# multiply by focal habitat (1)/non-focal habitat (0 or NA)
		# 	srcMask <- .makeSourceName("fragmentation_mask", "raster")
		# 	ex <- paste0(srcMask, " = if (", srcIn, " == 1, 1, null())")
		# 	rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		# 	srcIn <- src
		# 	src <- .makeSourceName("fragmentation_class_masked")
		# 	ex <- paste0(src, " = ", srcIn, " * ", srcMask)
		# 	rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))
			
		# 	# make mask, copy with mask in place, remove mask
		# 	# # rgrass::execGRASS(cmd = "r.mask", raster = srcIn, flags = c(.quiet(), "overwrite"))
		# 	# # src <- .copyGRaster(src, topo = topology(x)[i], reshapeRegion = FALSE)
		# 	# # .removeMask()

		# } # else {
		
		# 	# mask to non-NA cells in input
		# 	srcIn <- src
		# 	src <- .makeSourceName("fragmentation_mask_NAs", "raster")
		# 	ex <- paste0(src, " = if(isnull(", sources(x)[i], "), null(), ", srcIn, ")")
		# 	rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		# }

		steps <- steps + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, steps)
		.rm(created, type = "raster", warn = FALSE)
		created <- character()

		srcs[i] <- src

	} # next layer

	steps <- steps + 1
	if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, steps)
	if (verbose | faster("verbose")) close(pb)

	levs <- .fragmentationLevels(undet = undet, none = none)
	levels <- list()
	for (i in seq_len(nLayers)) {
		levels[[i]] <- levs
	}

	makeGRaster(srcs, names = "fragmentation", levels = levels)

	} # EOF
	
)

#' Create levels table for `fragmentation()`.
#'
#' @param undet Value of `undet`.
#' @param none Value of `none`.
#'
#' @returns A `data.frame`.
#'
#' @noRd
.fragmentationLevels <- function(undet, none) {

	if (undet %in% c("perforated", "edge")) {

		if (is.na(none)) {

			levs <- data.frame(value = c(1L:4L, 6L), class = c("patch", "transitional", "perforated", "edge", "interior"))
		
		} else {
		
			levs <- data.frame(value = c(none, 1L:4L, 6L), class = c("none", "patch", "transitional", "perforated", "edge", "interior"))

		}

	} else if (undet == "undetermined") {
	
		if (is.na(none)) {
			
			levs <- data.frame(value = 1L:6L, class = c("patch", "transitional", "perforated", "edge", "undetermined", "interior"))

		} else {
		
			levs <- data.frame(value = c(none, 1L:6L), class = c("none", "patch", "transitional", "perforated", "edge", "undetermined", "interior"))

		}
	
	}
	levs

}

#' Get the datatype of a GRaster or of GVector columns
#'
#' @description For `GRaster`s, `datatype()` returns the data type (see `vignette("GRasters", package = "fasterRaster")`). For `GVector`s, `datatype()` returns the class of each column of the attribute table.
#'
#' @param x A `GRaster` or `GVector`.
#' @param type (`GRaster`s only) `NULL` or character: Type of datatype to report (`GRaster` only):
#' * `"fasterRaster"` (default): Reports the **fasterRaster** type (factor, integer, float, or double)
#' * `"terra"`: Report the (inferred) **terra** data type (e.g., INT2U, FLT4S). Please see the table in the documentation for [writeRaster()` for an explanation of these codes.
#' * `"GRASS"`: Will return "CELL" (integer), "FCELL" (floating-point value), or "DCELL" (double-floating point value)
#' * `"GDAL"`: See [GDAL: Raster Band](https://gdal.org/en/stable/user/raster_data_model.html). Please also see the table in the [writeRaster()] help page.
#'
#' @param forceDouble Logical (`GRaster`s and `SpatRaster`s only): If `TRUE` (default), and the raster appears to represent non-integer values, then the raster will be assumed to represent double-floating point values (**GRASS**: type "DCELL", **terra**: type "FLT8S", **fasterRaster**: type "double", and **GDAL**: type "Float64"). `forceDouble` reports the actual datatype if `type = "fasterRaster"` (i.e., the type is not forced to "double").
#'
#' @returns `datatype()` for a `GRaster` returns a character. `datatype()` for a `GVector` returns a data frame, with one row per field. If the `GVector` has no attribute table, the function returns `NULL`.
#'
#' @seealso [terra::datatype()], `vignette("GRasters", package = "fasterRaster")`
#'
#' @example man/examples/ex_GRaster_GVector.r
#'
#' @aliases datatype
#' @rdname datatype
#' @exportMethod datatype
methods::setMethod(
	f = "datatype",
	signature = c(x = "GRaster"),
	definition = function(x, type = "fasterRaster", forceDouble = TRUE) {

	type <- omnibus::pmatchSafe(type, c("fasterRaster", "GRASS", "terra", "GDAL"))

	out <- x@datatypeGRASS
	
	if (type == "fasterRaster") {

		isFact <- is.factor(x)
		out[out == "CELL" & isFact] <- "factor"
		out[out == "CELL" & !isFact] <- "integer"
		out[out == "FCELL"] <- "float"
		out[out == "DCELL"] <- "double"
	
	} else if (type %in% c("terra", "GDAL")) {

		stats <- .global(x, fun = c("min", "max", "sum"))

		min <- stats[ , "min"]
		max <- stats[ , "max"]
		sum <- stats[ , "sum"]

		integer <- all(omnibus::is.wholeNumber(stats))

	    # remainder <- stats %% 1L
		# integer <- rowSums(abs(remainder)) == 0L

		for (i in seq_along(out)) {

			if (out[i] == "CELL") {
			
				if (min[i] >= 0L & max[i] <= 255L) {
					out[i] <- if (type == "GDAL") { "Byte" } else { "INT1U" }
				} else if (min[i] >= 0L & max[i] <= 65534L) {
					out[i] <- if (type == "GDAL") { "UInt16" } else { "INT2U" }
				} else if (min[i] >= -32767L & max[i] <= -32767L) {
					out[i] <- if (type == "GDAL") { "Int16" } else { "INT2S" }
				} else if (min[i] >= -2147483647L & max[i] <= 2147483647L) {
					out[i] <- if (type == "GDAL") { "Int32" } else { "INT4S" }
				}
				
			} else if (forceDouble) {
				out[i] <- if (type == "GDAL") { "Float64" } else { "FLT8S" }
			} else if (out[i] == "FCELL") {
				out[i] <- if (type == "GDAL") { "Float32" } else { "FLT4S" }
			} else if (out[i] == "DCELL") {
				out[i] <- if (type == "GDAL") { "Float64" } else { "FLT8S" }
			} else {
				warning("Values are too small/large to represent. Assigning `double` type.")
				out[i] <- if (type == "GDAL") { "Float64" } else { "FLT8S" }
			}

		} # next raster

	} # want non-fasterRaster type
	out
	
	} # EOF
)

### NB This overwrites terra::datatype() for signature 'SpatRaster'!
# # # #' @aliases datatype
# # # #' @rdname datatype
# # # #' @exportMethod datatype
# # # methods::setMethod(
# # # 	f = "datatype",
# # # 	signature = c(x = "SpatRaster"),
# # # 	definition = function(x, type = "fasterRaster", forceDouble = TRUE) {
	
# # # 	type <- pmatch(type, c("fasterRaster", "GRASS", "terra", "GDAL"))

# # # 	stats <- terra::global(x, fun = c("min", "max", "sum"), na.rm = TRUE)

# # # 	min <- stats[ , "min"]
# # # 	max <- stats[ , "max"]
# # # 	sum <- stats[ , "sum"]

# # # 	integer <- all(omnibus::is.wholeNumber(stats))

# # # 	# remainder <- stats %% 1
# # # 	# integer <- rowSums(abs(remainder)) == 0

# # # 	nl <- terra::nlyr(x)
	
# # # 	out <- rep(NA_character_, nl)
# # # 	for (i in seq_len(nl)) {
	
# # # 		# integer
# # # 		if (integer[i]) {

# # # 			if (type == "GRASS") {
# # # 				out[i] <- "CELL"
# # # 			} else if (type == "fasterRaster") {
# # # 				out[i] <- "integer"
# # # 			} else {

# # # 				if (min[i] >= 0 & max[i] <= 255) {
# # # 					out[i] <- if (type == "GDAL") { "Byte" } else if (type == "terra") { "INT1U" }
# # # 				} else if (min[i] >= 0 & max[i] <= 65534) {
# # # 					out[i] <- if (type == "GDAL") { "UInt16" } else if (type == "terra") { "INT2U" }
# # # 				} else if (min[i] >= -32767 & max[i] <= -32767) {
# # # 					out[i] <- if (type == "GDAL") { "Int16" } else if (type == "terra") { "INT2S" }
# # # 				} else if (min[i] >= -2147483647 & max[i] <= 2147483647) {
# # # 					out[i] <- if (type == "GDAL") { "Int32" } else if (type == "terra") { "INT4S" }
# # # 				}
				
# # # 			}
			
# # # 		# not integer
# # # 		} else {
			
# # # 			if (forceDouble) {
# # # 				out[i] <- if (type == "GDAL") { "Float64" } else if (type == "terra") { "FLT8S" } else if (type == "GRASS") { "DCELL" } else if (type == "fasterRaster") { "double" }
# # # 			} else if (min[i] > -3.4e+38 & max[i] < 3.4e+38) {
# # # 				out[i] <- if (type == "GDAL") { "Float32" } else if (type == "terra") { "FLT4S" } else if (type == "GRASS") { "FCELL" } else if (type == "fasterRaster") { "float" }
# # # 			} else if (min[i] > -1.79e+308 & max[i] < 1.79e+308) {
# # # 				out[i] <- if (type == "GDAL") { "Float64" } else if (type == "terra") { "FLT8S" } else if (type == "GRASS") { "DCELL" } else if (type == "fasterRaster") { "double" }
# # # 			} else {
# # # 				warning("Values are too small/large to represent. Assigning `double` type.")
# # # 				out[i] <- if (type == "GDAL") { "Float64" } else if (type == "terra") { "FLT8S" } else if (type == "GRASS") { "DCELL"} else if (type == "fasterRaster") { "double" }
# # # 			}
# # # 		}

# # # 	} # next raster

# # # 	if (type == "fasterRaster") {

# # # 		isFact <- is.factor(x)
# # # 		out[out == "CELL" & isFact] <- "factor"
# # # 		out[out == "CELL" & !isFact] <- "integer"
# # # 		out[out == "FCELL"] <- "float"
# # # 		out[out == "DCELL"] <- "double"
	
# # # 	}
# # # 	out
	
# # # 	} # EOF
# # # )

#' @aliases datatype
#' @rdname datatype
#' @exportMethod datatype
setMethod(f = "datatype",
	signature = c(x = "GVector"),
	definition = function(x) {

	if (nrow(x@table) > 0L) {
		sapply(x@table, "class")
	} else {
		NULL
	}

	} # EOF
)

#' Identify watershed basins and direction and accumulation of flow
#'
#' @description The `flow()` function uses a raster representing elevation to compute other rasters representing:
#' * Flow accumulation;
#' * Direction of flow;
#' * Watershed basins;
#' * Flooded areas; and/or
#' * Topographic convergence (log of flow accumulation divided by local slope).
#'
#' More details about the computations can be found at the help page for the **GRASS** tool `r.terraflow`] (see `grassHelp("r.terraflow")`)
#'
#' @param x A `GRaster` with a single layer, typically representing elevation.
#'
#' @param direction Character: Either `"single"` or `"multi"`. This indicates whether a single-direction flow or multi-direction flow model is used. The default is `"multi"`. Partial matching is used and case is ignored.
#'
#' @param dirThreshold Numeric (default is `Inf`): For the multi-direction flow model, this indicates the amount of accumulated flow above which the single-direction flow rule is used to locate the egress of water from a cell. This is the `d8cut` parameter in `r.stream.extract`.
#'
#' @param return Character vector: Indicates what rasters to return. Partial matching is used and case is ignored. Options include:
#' * `"accumulation"` (default): Flow accumulation raster.
#' * `"basins"`: Watershed basins
#' * `"direction"`: Flow direction
#' * `"flooded"`: Flooded areas
#' * `"TCI"`: Topographic convergence index
#' * `"*"`: All of the above
#'
#' @seealso [flowPath()], [streams()], the **GRASS** tool `r.terraflow` (see `grassHelp("r.terraflow")`)
#'
#' @param scratchDir Character or `NULL` (default): Directory in which to store temporary files. The **GRASS** tool `r.terraflow` makes a lot of temporary files. If this is `NULL`, then a temporary folder in the user's working directory will be used (see [getwd()]).
#'
#' @returns A `GRaster`.
#'
#' @example man/examples/ex_flow.r
#'
#' @aliases flow
#' @rdname flow
#' @exportMethod flow
methods::setMethod(
	f = "flow",
	signature = c(x = "GRaster"),
	function(
		x,
		direction = "multi",
		return = "accumulation",
		dirThreshold = Inf,
		scratchDir = NULL
	) {

	if (nlyr(x) > 1L) stop("This function can only use a single-layered GRaster as input.")

	returns <- c("accumulation", "basins", "direction", "flooded", "TCI")
	if (any(return == "*")) return <- returns
	return <- omnibus::pmatchSafe(return, returns)
	return <- unique(return)

	direction <- omnibus::pmatchSafe(direction, c("single", "multi"), nmax = 1L)

	.locationRestore(x)
	.region(x)

	if (is.null(scratchDir)) {
		scratchDir <- paste0(getwd(), "/flow_temporary_files")
		omnibus::dirCreate(scratchDir)
		on.exit(unlink(scratchDir, recursive = TRUE, force = TRUE))
	}

	args <- list(
		cmd = "r.terraflow",
		elevation = sources(x),
		memory = faster("memory"),
		directory = scratchDir,
		flags = c(.quiet(), "overwrite")
	)

	if (direction == "single") args$flags <- c(args$flags, "s")
	if (direction == "multi" & !is.infinite(dirThreshold)) args$d8cut <- dirThreshold

	if (any(return == "accumulation")) args$accumulation <- .makeSourceName("r_flow_accumulation", "raster")
	if (any(return == "direction")) args$direction <- .makeSourceName("r_flow_direction", "raster")
	if (any(return == "basins")) args$swatershed <- .makeSourceName("r_flow_swatershed", "raster")
	if (any(return == "flooded")) args$filled <- .makeSourceName("r_flow_filled", "raster")
	if (any(return == "TCI")) args$tci <- .makeSourceName("r_flow_tci", "raster")

	if (is.null(scratchDir)) {
		scratchDir <- getwd()
		scratchDir <- paste0(scratchDir, '/flow_temp_files')
		omnibus::dirCreate(scratchDir)
		on.exit(unlink(scratchDir, recursive = TRUE, force = TRUE, expand = TRUE), add = TRUE)
	}

	do.call(rgrass::execGRASS, args = args)

	names <- srcs <- character()
	if (any(return == "accumulation")) {
		srcs <- args$accumulation
		names <- "accumulation"
	}
	if (any(return == "direction")) {
		srcs <- c(srcs, args$direction)
		names <- c(names, "direction")
	}
	if (any(return == "basins")) {
		srcs <- c(srcs, args$swatershed)
		names <- c(names, "basins")
	}
	if (any(return == "flooded")) {
		srcs <- c(srcs, args$filled)
		names <- c(names, "flooded")
	}
	if (any(return == "TCI")) {
		srcs <- c(srcs, args$tci)
		names <- c(names, "TCI")
	}
	
	order <- match(return, names)
	srcs <- srcs[order]
	names <- names[order]

	makeGRaster(srcs, names)

	} # EOF
)

#' Coerce as multipart GVector to a singlepart GVector
#'
#' @description `GVectors` can contain a mix of "singlepart" and "multipart" features. A singlepart feature is a single point, set of connected line segments, or a polygon. A multipart feature is a set of lines, sets of connected line segments, or set of polygons that are treated as a single feature. This function converts all multipart features to singlepart features. If the `GVector` has an attribute table, rows will be duplicated so that each of the new `GVector`'s geometries have the rows that correspond to their "parent" geometries.
#'
#' If you want to "split" cells of a `GRaster` into smaller cells, use [aggregate()].
#' 
#' @param x A `GVector`.
#' 
#' @returns A `GVector`.
#' 
#' @seealso [aggregate()]
#' 
#' @example man/examples/ex_aggregate_disagg.r
#' 
#' @aliases disagg
#' @rdname disagg
#' @exportMethod disagg
methods::setMethod(
	f = "disagg",
	signature = c(x = "GVector"),
	function(x) {

	.locationRestore(x)

	srcIn <- .copyGSpatial(x)

	srcDel <- .makeSourceName("disagg_v_category_del", "vector")
	rgrass::execGRASS(
		cmd = "v.category",
		input = srcIn,
		output = srcDel,
		option = "del",
		cat = -1,
		flags = c(.quiet(), "overwrite")
	)

	src <- .makeSourceName("disagg_v_category_add", "vector")
	rgrass::execGRASS(
		cmd = "v.category",
		input = srcDel,
		output = src,
		option = "add",
		flags = c(.quiet(), "overwrite")
	)

	# replicate data table
	if (nrow(x) == 0L) {
		table <- NULL
	} else {
		
		cats <- .vCats(x)
		if (anyNA(cats)) {
			warning("At least one geometry has a combined category. Data table cannot be copied.")
			table <- NULL
		} else {
			table <- x@table
			table <- table[cats]
			.vAttachDatabase(src)
		}

	}

	# # remove existing database
 	# rgrass::execGRASS(
 	# 	cmd = "v.db.connect",
 	# 	map = src,
 	# 	layer = "1",
 	# 	flags = c(.quiet(), "overwrite", "d")
 	# )

	# # attach new database with same number of rows as categories in vector
	# db <- data.frame(cat = .vCats(src))
	# tf <- tempfile(fileext = ".csv")
	# tft <- paste0(tf, "t")
	# utils::write.csv(db, tf, row.names = FALSE)

	# classes <- sapply(table, class)
	# classes[!(classes %in% c("numeric", "integer", "character", "Date"))] <- '"String"'
	# classes[classes == "numeric"] <- '"Real"'
	# classes[classes == "integer"] <- '"Integer"'
	# classes[classes == "character"] <- '"String"'
	# classes[classes == "Date"] <- '"Date"'
	# classes <- paste(classes, collapse = ",")

	# write(classes, tft)

	# srcTable <- .makeSourceName("db_in_ogr_table", NULL)
	# rgrass::execGRASS(
	# 	cmd = "db.in.ogr",
	# 	input = tf,
	# 	output = srcTable,
	# 	# key = "cat",
	# 	flags = c(.quiet(), "overwrite")
	# )

	# # connect database to vector
	# rgrass::execGRASS(
	# 	cmd = "v.db.connect",
	# 	map = sources(x),
	# 	table = srcTable,
	# 	layer = "1",
	# 	key = "cat_",
	# 	flags = c(.quiet(), "overwrite", "o")
	# )

	makeGVector(src, table = table)

	} # EOF
)

if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster:
elev <- fast(madElev)

# Geomorphons:
geos <- geomorphons(elev)
geos
levels(geos) # levels
freq(geos) # frequencies across cells

col <- c("gray90", "red", "orange", "blue", "green", "pink", "firebrick",
  "purple", "gray50", "black")
plot(geos, col = col)

}

#' Remove parts of a GRaster or GVector
#'
#' `crop()` removes parts of a `GRaster` or `GVector` that fall "outside" another raster or vector. You cannot make the `GRaster` or `GVector` larger than it already is (see [extend()]). Rasters may not be cropped to the exact extent, as the extent will be enlarged to encompass an integer number of cells. If you wish to remove certain cells of a raster, see [mask()].
#'
#' @param x A `GRaster` or `GVector` to be cropped.
#' @param y A `GRaster` or `GVector` to serve as a template for cropping.
#' @param extent Logical:
#' * If `y` is a "points" `GVector`: Use the convex hull around `y` to crop `x`.
#' * If `y` is a "lines" or "polygons" `GVector`: If `TRUE`, use the extent of `y` to crop `x`.
#' * if `y` is a `GVector`, `x` will be cropped to the extent of `y` (`extent` is ignored).
#' @param fail Logical: If `TRUE` (default), and the cropped object would have zero extent in at least one dimension, then exit the function with an error. If `FALSE`, then display a warning and return `NULL`.
#'  
#' @details Known differences from [terra::crop()]:
#' * If `x` and `y` are "points" vectors, and `extent` is `TRUE`, **terra** removes points that fall on the extent boundary. **fasterRaster** does not remove points on the extent boundary.
#' * If `x` is a "points" vector and `y` is a "lines" vectors, and `extent` is `FALSE`, **terra** uses the extent of `y` to crop the points.  **fasterRaster** uses the minimum convex hull of the lines vector.
#'
#' @return A `GRaster` or `GVector` (or `NULL` if `fail` is `FALSE` and the output would have a zero east-west and/or north-south extent).
#' 
#' @seealso [terra::crop()], [sf::st_crop()]
#' 
#' @example man/examples/ex_crop.r
#'
#' @aliases crop
#' @rdname crop
#' @exportMethod crop
methods::setMethod(
	f = "crop",
	signature = c(x = "GRaster"),
	definition = function(x, y, fail = TRUE) {
	
	if (inherits(y, "GVector")) {
		compareGeom(x, y)
	} else {
		if (crs(x) != crs(y)) stop("Rasters have different coordinate references systems.")
	}
	.locationRestore(x)

	### change region to match the extent of y but have the same resolution and registration as x
	if (W(y) <= W(x)) {
		w <- W(y)
	} else {
  		cells <- (W(y) - W(x)) / xres(x)
  		if (!omnibus::is.wholeNumber(cells)) {
			cells <- ceiling(cells)
		} else {
			cells <- round(cells)
		}
  		w <- W(x) + cells * xres(x)
	}

	if (E(y) >= E(x)) {
		e <- E(x)
	} else {
  		cells <- (E(x) - E(y)) / xres(x)
		if (inherits(y, "GVector")) {
			cells <- floor(cells)
		} else if (inherits(y, "GRaster")) {
   			if (omnibus::is.wholeNumber(cells)) {
				cells <- round(cells)
			} else {
				cells <- floor(cells)
			}
		}
  		e <- E(x) - cells * xres(x) # !!!???
	}

	if (S(y) <= S(x)) {
		s <- S(x)
	} else {
  		cells <- (S(y) - S(x)) / yres(x)
  		if (inherits(y, "GVector")) {
			cells <- floor(cells)
		} else if (inherits(y, "GRaster")) {
   			if (omnibus::is.wholeNumber(cells)) {
				cells <- round(cells)
			} else {
				cells <- ceiling(cells)
			}
		}
  		s <- S(x) + cells * yres(x) # !!!???
	}

	if (N(y) >= N(x)) {
		n <- N(x)
	} else {
  		cells <- (N(x) - N(y)) / yres(x)
		if (inherits(y, "GVector")) {
			cells <- floor(cells)
		} else if (inherits(y, "GRaster")) {
   			if (omnibus::is.wholeNumber(cells)) {
				cells <- round(cells)
			} else {
				cells <- ceiling(cells)
			}
		}
  		n <- N(x) - cells * yres(x)
	}

	w <- as.character(w)
	e <- as.character(e)
	s <- as.character(s)
	n <- as.character(n)
	ewres <- as.character(xres(x))
	nsres <- as.character(yres(x))

	rgrass::execGRASS(
		cmd = "g.region",
		w = w, e = e, s = s, n = n,
		ewres = ewres, nsres = nsres,
		flags = c(.quiet(), "overwrite")
	)

	### crop by creating copy of focal raster
	srcs <- .copyGSpatial(x, reshapeRegion = FALSE)
	makeGRaster(srcs, names(x), levels = cats(x), ac = activeCats(x), fail = fail)

	} # EOF
)

#' @aliases crop
#' @rdname crop
#' @exportMethod crop
methods::setMethod(
	f = "crop",
	signature = c(x = "GVector"),
	definition = function(x, y, extent = FALSE, fail = TRUE) {

	compareGeom(x, y)
	.locationRestore(x)

	### crop
	src <- .makeSourceName("crop", "vector")
	if (inherits(y, "GVector")) {

		xgeom <- geomtype(x, grass = TRUE)
		ygeom <- geomtype(y, grass = TRUE)

		args <- list(
			cmd = "v.clip",
			input = sources(x),
			output = src,
			flags = c(.quiet(), "overwrite"),
			intern = TRUE
		)
		
		if (extent) {
			.regionExt(y, respect = "dim")
			args$flags <- c(args$flags, "r")
			args$clip <- sources(y)
		} else if (ygeom %in% c("point", "line")) {
			ch <- convHull(y)
			args$clip <- sources(ch)
		} else if (ygeom == "area") {
			args$clip <- sources(y)
		}

	} else if (inherits(y, "GRaster")) {

		.region(y)
		args <- list(
			cmd = "v.clip",
			input = sources(x),
			clip = sources(y),
			output = src,
			flags = c("r", .quiet(), "overwrite")
		)

	}

	do.call(rgrass::execGRASS, args = args)
	makeGVector(src, fail = fail)

	} # EOF
)

#' Are the category values of a vector valid?
#'
#' @description Category values of a **GRASS** vector can be invalid if **GRASS** assigns more than one value to a geometry (e.g., "7/12"). This can occur when the vector was created by software that does not use a topological system (e.g., a shapefile).
#'
#' @param x A `GVector` or the [sources()] name of one.
#'
#' @returns Logical.
#'
#' @keywords internal
.vValidCats <- function(x) {

	if (inherits(x, "GVector")) x <- sources(x)
	meows <- .vCats(x, db = FALSE, integer = TRUE)
	length(meows) > 0L && !anyNA(meows)
	
}

#' Rename a raster or vector in an existing GRASS session
#'
#' Rename a raster or vector in an existing **GRASS** session.
#'
#' @param from,to `sources` of the raster or vector to rename.
#' @param rastOrVect Either `NULL` (default), `"raster"`, or `"vector"`. This specifies the type of object to be renamed. Partial matching is allowed. If left as `NULL` (default), the function will try to identify if the object is a raster or vector, and return an error if there is both a raster and vector of given name. Note that unlike in **R**, **GRASS** can have rasters and vector"s with the same name.
#' 
#' @return The function invisibly returns \code{TRUE} if the desired rasters and/or vectors were named, and \code{FALSE} if raster and/or vector to be renamed did not exist in the \code{GRASS} session. Notably, a raster or vector or both are renamed in an existing \code{GRASS} session.
#'
#' @aliases .rename
#' @keywords internal
.rename <- function(
	from,
	to,
	rastOrVect = NULL
) {

	### function globals
	####################
	
	flags <- c(.quiet(), "overwrite")
	rastOrVect <- omnibus::pmatchSafe(rastOrVect, c("raster", "vector"), error=TRUE)
	
	fromTo <- c(from, to)
	success <- FALSE

	### rename raster
	if (rastOrVect == "raster") {
	
		spatials <- .ls("rasters")

		rgrass::execGRASS("g.rename", raster=fromTo, flags=flags, intern=TRUE)
		success <- TRUE
		
	} else if (rastOrVect == "vector") {
	
		rgrass::execGRASS("g.rename", vector=fromTo, flags=flags, intern=TRUE)
		success <- TRUE
	
	}

	invisible(success)

}

#' Convert a GRASS vector's attribute table to a data.table
#'
#' @description **GRASS** vectors can be linked to an attribute table, which can be exported from **GRASS** to **R** using this function. **This function is mostly of use to developers.**
#'
#' Values in the `cat` column are not necessarily unique--if a value appears more than once, the set of features they index are (in other software) called "multipart" features. The table can have more columns with metadata for each feature.
#'
#' This function is typically used by developers.
#'
#' @param x A `GVector` or the name of a vector in **GRASS**.
#'
#' @returns A `data.table` or `NULL` if the vector is not attached to a database.
#'
#' @aliases .vAsDataTable
#' @rdname vAsDataTable
#' @keywords internal
.vAsDataTable <- function(x) {

	if (inherits(x, "GVector")) {
		.locationRestore(x)
		src <- sources(x)
	} else {
		src <- x
	}

	hasTable <- .vHasDatabase(src)
	if (!hasTable) {
		out <- NULL
	} else {

		tf <- tempfile(fileext = ".csv")
		rgrass::execGRASS(
			"v.db.select",
			map = src,
			separator = "comma",
			null_value = "NA",
			file = tf,
			flags = "overwrite",
			intern = TRUE
		)
		out <- data.table::fread(tf)
	
	}

	out
}

#' Divide a GRaster into spatially exclusive subsets
#'
#' @description This function divides a `GRaster` into "tiles" or spatial subsets which can be used for speeding some raster calculations. Tiles can be mutually exclusive or overlap by a user-defined number of cells.
#'
#' @param x A `GRaster` with one or more layers.
#'
#' @param n Numeric vector: Number of tiles to create. This can be a single number, in which case `x` is divided into `n` × `n` tiles, or two values in which case it is divided into `n[1]` × `n[2]` tiles (rows x columns).
#'
#' @param overlap Numeric vector (default is 0): Number of rows/columns by which to expand the size of tiles so they overlap. This can be a single value or two values. If just one is provided, the tiles will be expanded by `overlap` rows and `overlaps` columns. If two numbers are provided, the tiles will be expanded by `overlap[1]` rows and `overlap[2]` columns.
#'
#' @param verbose Logical: If `TRUE`, display progress. Default is `FALSE`. Progress is only displayed if `x` is a multi-layer `GRaster`.
#'
#' @returns If `x` has just one layer, then the output is a `list` with one element per tile. The [lapply()] and [sapply()] functions can be used to apply functions to each tile in the list. If `x` has more than one layer, then the output will be a `list` of `list`s, with each sub-`list` containing the tiles for one `GRaster` layer.
#'
#' @example man/examples/ex_tiles.r
#'
#' @aliases tiles
#' @rdname tiles
#' @exportMethod tiles
methods::setMethod(
	f = "tiles",
	signature = c("GRaster"),
	function(x, n, overlap = 0, verbose = FALSE) {

	.locationRestore(x)
	.region(x)

	dims <- dim(x)
	extent <- ext(x)
	res <- res(x)

	out <- list()

	### multiple layers
	if (nlyr(x) > 1L) {

		# create tiles
		if (verbose | faster("verbose")) {
			pb <- utils::txtProgressBar(min = 0, max = nlyr(x), initial = 0, style = 3, width = 30)
		}

		for (i in 1:nlyr(x)) {

			if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, i)

			srcs <- .tiles(x = x[[i]], n = n, dims = dims, extent = extent, res = res, overlap = 0)

			out[[i]] <- list()
			for (j in seq_along(srcs)) {

				out[[i]][[j]] <- makeGRaster(srcs[j], names(x)[i], levels = cats(x, layer = i), ac = activeCat(x, layer = i), fail = TRUE)

			} # next tile
		} # next layer

		names(out) <- names(x)
		if (verbose | faster("verbose")) close(pb)

	### single layer
	} else {

		srcs <- .tiles(x = x, n = n, dims = dims, extent = extent, res = res, overlap = 0)
		for (i in seq_along(srcs)) {

			out[[i]] <- makeGRaster(srcs[i], names(x), levels = cats(x), ac = activeCat(x), fail = TRUE)

		} # next tile

	}
	out
	
	} # EOF
)

#' Break a raster into tiles
#' @param x The [sources()] name of a **GRASS** raster.
#' @param n Integer/numeric: Number of tiles. One or two values.
#' @param dims [dim()] of `x`.
#' @param extent A `SpatExtent` object or a vector of four values describing the extent of `x` (WESN).
#' @param res Numeric vector with two values: size of a cell in the x-dimension and the y-dimension in map units.
#' @param overlap Number of rows/columns by which to overlap tiles.
#'
#' @returns Character vector: The [sources()] names of the tiles. Each has a `list` attribute with its extent and dimensions.
#' 
#' @noRd
.tiles <- function(x, n, dims, extent, res, overlap = 0) {

	if (length(n) == 1L) n <- c(n, n)
	if (length(overlap) == 1L) overlap <- c(overlap, overlap)

	if (all(overlap >= dims)) {
		warning("The number of overlap cells is >= than the number of rows and/or columns\n  in the input raster. The output is the same GRaster as the input.")
		return(x)
	}

	# number of columns/rows per tile (without expanding for overlap > 0)
	colsPerTile <- dims[2L] / n[2L]
	rowsPerTile <- dims[1L] / n[1L]
	
	cols <- c(1, seq(colsPerTile, dims[2L], by = colsPerTile))
	rows <- c(1, seq(rowsPerTile, dims[1L], by = rowsPerTile))

	cols <- round(cols)
	rows <- round(rows)

	nTiles <- length(cols) * length(rows)

	if (inherits(extent, "SpatExtent")) extent <- as.vector(extent)
	xmin <- extent[1L]
	xmax <- extent[2L]
	ymin <- extent[3L]
	ymax <- extent[4L]

	xres <- res[1L]
	yres <- res[2L]

	nTiles <- prod(n)

	out <- rep(NA_character_, nTiles) # vector storing `sources` names of raster tiles
	i <- 1L
	spatial <- list() # hold spatial attributes, one element per tile
	for (row in seq_len(n[1L])) {
		for (col in seq_len(n[2L])) {

			# start/end rows/cols
			startRow <- rows[row]
			endRow <- rows[row + 1]

			startCol <- cols[col]
			endCol <- cols[col + 1]

			# extent
			west <- xmin + startCol * xres - ifelse(startCol == 1L, xres, 0)
			east <- xmin + endCol * xres

			north <- ymax - startRow * yres + ifelse(startRow == 1L, yres, 0)
			south <- ymax - endRow * yres

			nrows <- rows[row + 1L] - rows[row] + 1L
			ncols <- cols[col + 1L] - cols[col] + 1L

			# add rows
			if (overlap[1] > 0) {
			
				delta <- overlap[1L] * yres
				if (south > ymin) {

					newBorder <- south - delta
					newBorder <- max(newBorder, ymin)
					newRows <- (south - newBorder) / yres
					nrows <- nrows + newRows
					south <- newBorder

				}
			
				if (north < ymax) {

					newBorder <- north + delta
					newBorder <- min(newBorder, ymax)
					newRows <- (newBorder - north) / yres
					nrows <- nrows + newRows
					north <- newBorder
					
				}

			}

			# add columns
			if (overlap[2] > 0) {
			
				delta <- overlap[2L] * xres
				if (west > xmin) {

					newBorder <- west - delta
					newBorder <- max(newBorder, xmin)
					newCols <- (west - newBorder) / xres
					ncols <- ncols + newCols
					west <- newBorder

				}
			
				if (east < xmax) {

					newBorder <- east + delta
					newBorder <- min(newBorder, xmax)
					newCols <- (newBorder - east) / xres
					ncols <- ncols + newCols
					east <- newBorder
					
				}

			
			}

			# obviating floating-point error
			nrows <- as.integer(nrows)
			ncols <- as.integer(ncols)
			
			westChar <- as.character(west)
			eastChar <- as.character(east)
			southChar <- as.character(south)
			northChar <- as.character(north)

			rgrass::execGRASS(
				cmd = "g.region",
				n = northChar,
				s = southChar,
				e = eastChar,
				w = westChar,
				rows = nrows,
				cols = ncols,
				flags= c("o", .quiet())
			)

			out[i] <- .copyGSpatial(x, reshapeRegion = FALSE)
			spatial[[i]] <- list(
				extent = c(west, east, south, north),
				dim = c(nrows, ncols)
			)
			i <- i + 1L

		} # next column

	} # next row
	attr(out, "spatial") <- spatial
	out

}

#' @name madElev
#'
#' @title Elevation raster for an eastern portion of Madagascar
#'
#' @description Elevation raster for an eastern portion of Madagascar.
#'
#' @docType data
#'
#' @format An object of class `SpatRaster`. Values are mean meters above sea level.
#'
#' @keywords elevation Madagascar
#'
#' @source [WorldClim Version 2.1](https://worldclim.org)
#'
#' @example man/examples/ex_fastData.r
#'
NULL

#' Add a database table to a GRASS attribute table
#'
#' @description `.vDetachDatabase()` detaches the database from a **GRASS** vector and deletes it. This table is meant to be "invisible" to most users--they should use interact with attribute tables using the `GVector` slot `@table`. Some functions do require tables (e.g., [extract()] and [spatSample()]). **This function is mostly of use to developers.**
#'
#' @param x A `GVector` or the name of a vector in **GRASS**.
#' 
#' @returns Invisibly returns the [sources()] name of a vector in **GRASS**.
#' 
#' @aliases .vDetachDatabase
#' @rdname vDetachDatabase
#' @keywords internal
.vDetachDatabase <- function(x) {

	if (inherits(x, "GVector")) {
		.locationRestore(x)
		src <- sources(x)
	} else {
		src <- x
	}

	if (.vHasDatabase(src)) {

		rgrass::execGRASS(
			cmd = "v.db.droptable",
			map = src,
			flags = c(.quiet(), "f")
		)

	}
	invisible(src)

}

if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Elevation raster
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster:
elev <- fast(madElev)

# Find clumps based on exact values. This will appear as a gradient because
# most cells are assigned to a group of 1 cell.
exact <- clump(elev)
plot(exact)

# Clump based on approximate values:
approx <- clump(elev, minDiff = 0.0075)
plot(approx)

# Clump based on approximate values with minimum clump size:
approx20 <- clump(elev, minDiff = 0.005, minClumpSize = 20)
plot(approx20)

approx
approx20

}

if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster:
elev <- fast(madElev)

# What functions can we use with global()?
global()

# Calculate global statistics:
global(elev, fun = c("mean", "var", "varpop"))
global(elev, "quantile", probs = c(0.25, 0.5, 0.75))

global(elev, "*") # calculate all available functions

}

if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Example data: elevation raster and points vector
madElev <- fastData("madElev") # raster
madCover <- fastData("madCover") # categorical raster
madDypsis <- fastData("madDypsis") # points vector
madRivers <- fastData("madRivers") # lines vector
madCoast4 <- fastData("madCoast4") # polygons vector

# Convert to fasterRaster formats:
elev <- fast(madElev) # raster
cover <- fast(madCover) # categorical raster
dypsis <- fast(madDypsis) # points vector
rivers <- fast(madRivers) # lines vector
coast <- fast(madCoast4) # polygons vector

# Get values of elevation at points where Dypsis species are located:
extract(elev, dypsis, xy = TRUE)

# Extract from categorical raster at points:
categories <- extract(cover, dypsis)
categoryValues <- extract(cover, dypsis, cats = FALSE)
categories
categoryValues

# Extract and summarize values on a raster across polygons:
extract(elev, coast, fun = c("sum", "mean", "countNonNA"), overlap = FALSE)

# Extract and summarize values on a raster across lines:
extract(elev, rivers, fun = c("sum", "mean", "countNonNA"), overlap = FALSE)

# Extract from a polygons vector at a points vector:
polysFromPoints <- extract(coast, dypsis, xy = TRUE)
head(polysFromPoints) # first 3 are outside polygons vector, next 3 are inside

}

#' Interpolate values at points to a GRaster using splines
#'
#' @description This function interpolates values in the data table of a "points" `GVector` to a `GRaster` using splines with Tykhonov regularization to avoid overfitting.
#'
#' @param x A "points" `GVector`.
#'
#' @param y A `GRaster`: The output will have the same extent and resolution as this raster.
#'
#' @param field Character or integer or numeric integer: Name or index of the column in `x` with values to interpolate. If `NULL` and if `x` is a 3-dimensional "points" `GVector`, then the interpolation will act on the z-coordinate of each point.
#'
#' @param method Character: The method to use for interpolation can be either `"bilinear"` (default) or `"bicubic"`. Partial matching is used.
#'
#' @param lambda Either `NULL` (default) or numeric value > 0: The Tykhonov regularization parameter. If `NULL`, cross-validation will be used to determine the optimal parameter value. Cross-validation can take quite a while. If you use cross-validation, the output will either be a `GRaster` or a `data.frame`, depending on the value of `interpolate`.
#'
#' @param solver Character: Type of solver to use. Can be either of `"Cholesky"` or `"cg"`. Partial matching is used and case is ignored.
#'
#' @param xlength,ylength Either `NULL` (default), or numeric > 0: Length of the spline step in the x- and y-directions. If `NULL`, these will be set to 4 times the length of the extent in the respective direction.
#'
#' @param interpolate Logical: If `TRUE` (default), then create a `GRaster` with interpolated values. If `FALSE`, return a table with `lambda` values from cross-validation. This argument is ignored if `lambda` is a numeric value.
#'
#' @param verbose Logical: if `TRUE`, display progress.
#'
#' @details If you receive the error, "No data within this subregion. Consider increasing spline step values, try increasing the values of `xlength` and `ylength`.
#'
#' If cross-validation takes too long, or other warnings/errors persist, you can randomly subsample `x` to ~100 points to get an optimum value of `lambda` (using `interpolate = FALSE`), then use this value in the same function again without cross-validation (setting `lambda` equal to this value and `interpolate = TRUE`).
#'
#' @returns Output depends on values of `lambda` and `interpolate`:
#' * `lambda` is `NULL` and `interpolate` is `TRUE`: A `GRaster` with an attribute named `lambdas`. This is a `data.frame` with values of `lambda` that were assessed, plus `mean` (mean residual value) and `rms` (root mean square error). You can see the table using `attr(output_raster, "lambdas", exact = TRUE)`.
#' * `lambda` is `NULL` and `interpolate` is `FALSE`: A `data.frame` with values of `lambdas` that were assessed, plus `mean` (mean residual value) and `rms` (root mean square error). You can see the table using `attr(output_raster, "lambdas", exact = TRUE)`.
#' * `lambda` is a number (`interpolate` is ignored): A `GRaster`.
#'
#' @seealso [interpIDW()], [fillNAs()], **GRASS** tool `v.surf.bspline` (see `grassHelp("v.surf.bspline")`)
#'
#' @aliases interpSplines
#' @rdname interpSplines
#' @exportMethod interpSplines
methods::setMethod(
	f = "interpSplines",
	signature = c(x = "GVector", y = "GRaster"),
	function(
		x,
		y,
		field,
		method = "bilinear",
		lambda = NULL,
		solver = "Cholesky",
		xlength = NULL,
		ylength = NULL,
		interpolate = TRUE,
		verbose = is.null(lambda)
	) {

	.interpSplines(
		x = x,
		y = y,
		field = field,
		method = method,
		lambda = lambda,
		solver = solver,
		xlength = xlength,
		ylength = ylength,
		interpolate = interpolate,
		verbose = verbose
	)

	} # EOF
)

# #' @aliases interpSplines
# #' @rdname interpSplines
# #' @exportMethod interpSplines
# methods::setMethod(
# 	f = "interpSplines",
# 	signature = c(x = "GVector", y = "GVector"),
# 	function(
# 		x,
# 		y,
# 		field,
# 		method = "bilinear",
# 		lambda = NULL,
# 		solver = "Cholesky",
# 		xlength = NULL,
# 		ylength = NULL,
#		interpolate = interpolate,
#		verbose = is.null(lambda)
# 	) {

# 	.interpSplines(
# 		x = x,
# 		y = y,
# 		field = field,
# 		method = method,
# 		lambda = lambda,
# 		solver = solver,
# 		xlength = xlength,
# 		ylength = ylength
#		interpolate = interpolate,
#		verbose = verbose
# 	)

# 	} # EOF
# )


#' @noRd
.interpSplines <- function(
		x,
		y,
		field,
		method,
		lambda,
		solver,
		xlength,
		ylength,
		interpolate,
		verbose
) {

	if (!is.null(lambda)) {
		if (lambda <= 0) stop("Argument `lambda` must be NULL or > 0.")
	}

	if (!is.null(lambda)) interpolate <- TRUE

	if (geomtype(x, grass = TRUE) != "point") stop("Argument `x` must be a points GVector.")

	if (inherits(y, "GVector")) {
	
		if (geomtype(y, grass = TRUE) != "point") stop("Argument `y` must be a points GVector or a GRaster.")
	
	}

	method <- omnibus::pmatchSafe(method, c("bilinear", "bicubic"), nmax = 1L)
	solver <- omnibus::pmatchSafe(solver, c("cholesky", "cg"), nmax = 1L)

	compareGeom(x, y)
	.locationRestore(x)
	if (inherits(y, "GRaster")) .region(y)

	# copy values field to x
	if (is.null(field) & is.2d(x)) {
		stop("Argument `field` cannot be NULL if `x` is a 2-dimensional points GVector.")
	} else if (is.numeric(field) | is.integer(field)) {
		field <- names(x)[field]
	}

	if (!is.null(field)) {

		if (anyNA(x@table[[field]])) stop("The column `field` has at least one NA in it. NAs are not permissible.")

		cats <- .vCats(x, db = FALSE)
		db <- data.table::data.table(frid = cats, TEMPTEMP_ = x@table[[field]])
		names(db)[2L] <- field

		.vAttachDatabase(x, table = db, replace = TRUE, cats = cats)

	}

	if (inherits(y, "GVector")) {
		src <- .makeSourceName("v_surf_idw", "vector")
	} else {
		src <- .makeSourceName("v_surf_idw", "raster")
	}

	if (is.null(lambda)) {
		lambda_i <- 0.01
	} else {
		lambda_i <- lambda
	}

	args <- list(
		cmd = "v.surf.bspline",
		input = sources(x),
		column = field,
		method = method,
		solver = solver,
		lambda_i = lambda_i,
		memory = faster("memory"),
		intern = TRUE,
		flags = c(.quiet(), "overwrite")
	)

	if (inherits(y, "GVector")) {
		args$sparse_input <- sources(y)
		args$output <- src
	} else if (inherits(y, "GRaster")) {
		args$raster_output <- src
	}

	if (!is.null(xlength)) args$ew_step = xlength
	if (!is.null(ylength)) args$ns_step = ylength

	if (is.null(lambda)) args$flags <- c(args$flags, "c")

	if (faster("verbose") | verbose) {
		if (is.null(lambda)) {
			omnibus::say("Optimizing lambda...")
		} else {
			omnibus::say("Interpolating...")
		}
	}

	info <- do.call(rgrass::execGRASS, args = args)
	
	# cross-validation for lambda
	if (is.null(lambda)) {

		info <- strsplit(info, split = "\\|")
		info <- info[-1L]
		info <- lapply(info, trimws)
		info <- lapply(info, as.numeric)
		lambdas <- do.call(rbind, info)
		lambdas <- data.table::as.data.table(lambdas)
		names(lambdas) <- c("lambda", "mean", "rms")
		lambdas$optimum <- FALSE
		lambdas$optimum[which.min(lambdas$rms)] <- TRUE

		index <- which.min(lambdas$rms)
		lambda <- lambdas$lambda[index]

		if (interpolate) {

			out <- interpSplines(
				x = x,
				y = y,
				field = field,
				method = method,
				lambda = lambda,
				solver = solver,
				xlength = xlength,
				ylength = ylength,
				verbose = verbose
			)

			attr(out, "lambdas") <- lambdas

		} else {
			out <- lambdas
			if (!faster("useDataTable")) out <- as.data.frame(out)
		}

	} else if (inherits(y, "GVector")) {
		table <- .vAsDataTable(src)
		out <- makeGVector(src)
	} else if (inherits(y, "GRaster")) {
		out <- makeGRaster(src, "interpolation")
	}
	out

}

if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Example data
madElev <- fastData("madElev")
madCoast4 <- fastData("madCoast4")
madCoast4 <- vect(madCoast4)

# For the example, crop the elevation raster to two communes
madAnt <- madCoast4[madCoast4$NAME_4 == "Antanambe", ]
madMan <- madCoast4[madCoast4$NAME_4 == "Manompana", ]

elevAnt <- crop(madElev, madAnt)
elevMan <- crop(madElev, madMan)

plot(madElev)
plot(elevAnt, col = "red", legend = FALSE, add = TRUE)
plot(elevMan, col = "blue", legend = FALSE, add = TRUE)

# Convert a SpatRaster to a GRaster
ant <- fast(elevAnt)
man <- fast(elevMan)

# merge
antMan <- merge(ant, man)
plot(antMan, main = "Antman!")

}

if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Example data
madRivers <- fastData("madRivers")
rivers <- fast(madRivers)
soam <- rivers[rivers$NAM == "SOAMIANINA"] # select one river for illustration

### Simplify geometry (remove nodes)

vr <- simplifyGeom(soam, tolerance = 2000)
dp <- simplifyGeom(soam, tolerance = 2000, method = "dp")
dpr <- simplifyGeom(soam, tolerance = 2000, method = "dpr", prop = 0.5)
rw <- simplifyGeom(soam, tolerance = 2000, method = "rw")

plot(soam, col = "black", lwd = 3)
plot(vr, col = "blue", add = TRUE)
plot(dp, col = "red", add = TRUE)
plot(dpr, col = "chartreuse", add = TRUE)
plot(rw, col = "orange", add = TRUE)

legend("bottom",
   xpd = NA,
   legend = c(
	  "Original",
      "Vertex reduction",
      "Douglas-Peucker",
      "Douglas-Peucker reduction",
      "Reumann-Witkam"
	),
	col = c("black", "blue", "red", "chartreuse", "orange"),
	lwd = c(3, 1, 1, 1, 1)
)

### Smooth geometry

hermite <- smoothGeom(soam, dist = 2000, angle = 3)
chaiken <- smoothGeom(soam, method = "Chaiken", dist = 2000)

plot(soam, col = "black", lwd = 2)
plot(hermite, col = "blue", add = TRUE)
plot(chaiken, col = "red", add = TRUE)

legend("bottom",
   xpd = NA,
   legend = c(
	  "Original",
      "Hermite",
      "Chaiken"
	),
	col = c("black", "blue", "red"),
	lwd = c(2, 1, 1, 1, 1)
)

### Clean geometry

# Has no effect on this vector!
noDangs <- removeDangles(soam, tolerance = 10000)

plot(soam, col = "black", lwd = 2)
plot(noDangs, col = "red", add = TRUE)

legend("bottom",
   xpd = NA,
   legend = c(
	  "Original",
      "No dangles"
	),
	lwd = c(2, 1),
	col = c("black", "red")
)

}

#' Convert a raster to a polygons vector
#'
#' @description [as.polygons()] converts a `GRaster` to a "polygons" `GVector`. After running this function, [geometry cleaning][breakPolys] may be useful to use to "tidy up" the vector.
#' 
#' @param x A `GRaster`. If more than one layer is in the `GRaster`, only the first will be used (with a warning).
#' 
#' @param round Logical: If `TRUE` (default), values in the raster will be rounded first before conversion to a vector. This causes cells that are adjacent that have the same (rounded) values to be combined into a single polygon. For more control, see [clump()].
#'
#' @param smooth Logical: If `TRUE`, round the corners of square features. Default is `FALSE`.
#' 
#' @returns A `GVector`.
#' 
#' @seealso [as.points()], [as.lines()], [terra::as.polygons()], [geometry cleaning][breakPolys], and **GRASS** tool `r.to.vect`
#' 
#' @example man/examples/ex_asPolygons.r
#' 
#' @aliases as.polygons
#' @rdname as.polygons
#' @exportMethod as.polygons
methods::setMethod(
    f = "as.polygons",
    signature = c(x = "GRaster"),
    function(x, round = TRUE, smooth = FALSE) {

    if (nlyr(x) > 1L) warning("The raster has >1 layer. Only the first will be used.")
    x <- x[[1L]]

    .locationRestore(x)
    .region(x)

    if (round & datatype(x, "GRASS") != "CELL") x <- round(x)

    src <- .makeSourceName("asPolys", "raster")
    args <- list(
        cmd = "r.to.vect",
        input = sources(x),
        output = src,
        type = "area",
        column = names(x),
        flags = c(.quiet(), "overwrite"),
        intern = TRUE
    )
    if (smooth) args$flags <- c(args$flags, "s")
    do.call(rgrass::execGRASS, args = args)
    makeGVector(src)

    } # EOF
)

#' @title Number of rows, columns, depths, cells, and layers
#'
#' @description
#' For `GRegion`s: Number of rows, columns, depths, and cells:
#' * `dim()`: Rows and columns
#' * `dim3d()`: Rows, columns, and depths
#' * `nrow()`: Rows
#' * `ncol()`: Columns
#' * `ndepth()`: Depths (for 3-dimensional rasters only)
#' * `ncell()`: Number of cells (2 dimensions)
#' * `ncell3d()`: Number of cells (3 dimensions)
#'
#' For `GRaster`s: As above, plus number of layers:
#' * `nlyr()`: Layers (number of "stacked" rasters--different from depths of a raster).
#'
#' For `GVector`s: Number of geometries and fields (columns):
#' * `dim()`: Number of geometries and number of columns in data table
#' * `nrow()`: Number of geometries
#' * `ncol()`: Number of columns in data table
#'
#' @param x A `GRegion`, `GRaster`, `GVector`, or missing. If missing, then the dimensions of the currently active "region" are returned (see `vignette("regions", package = "fasterRaster")`).
#'
#' @return A numeric value or vector.
#' 
#' @seealso [ngeom()], [nsubgeom()], [nacell()], [nonnacell()], [terra::dim()]
#'
#' @example man/examples/ex_GRaster_GVector.r
#'
#' @aliases dim
#' @rdname dim
#' @exportMethod dim
methods::setMethod(
	f = "dim",
	signature = "GRegion",
	definition = function(x) {
	c(rows = unname(x@dimensions[1L]), cols = unname(x@dimensions[2L]))
	} # EOF
)


## Error on install: "the method for function 'dim' and signature x="missing" is sealed and cannot be re-defined"
# #' @aliases dim
# #' @rdname dim
# #' @exportMethod dim
# methods::setMethod(
	# f = "dim",
	# signature = "missing",
	# definition = function(x) .region()@dimensions[1L:2L]
# )

#' @aliases dim3d
#' @rdname dim
#' @exportMethod dim3d
methods::setMethod(
	f = "dim3d",
	signature = "missing",
	definition = function(x) .region()@dimensions
)

#' @aliases dim3d
#' @rdname dim
#' @exportMethod dim3d
methods::setMethod(
	f = "dim3d",
	signature = "GRegion",
	definition = function(x) {
	
	c(
		rows = unname(x@dimensions[1L]),
		cols = unname(x@dimensions[2L]),
		depths = unname(x@dimensions[3L])
	)
	
	} # EOF
)

#' @rdname dim
#' @aliases nrow
#' @exportMethod nrow
setMethod(
	f = "nrow",
	signature = "missing",
 	definition = function(x) nrow(.region())
)

#' @rdname dim
#' @aliases nrow
#' @exportMethod nrow
setMethod(
	f = "nrow",
	signature = "GRegion",
	definition = function(x) unname(x@dimensions[1L])
)

#' @rdname dim
#' @aliases ncol
#' @exportMethod ncol
setMethod(
	f = "ncol",
	signature = "missing",
 	definition = function(x) ncol(.region())
)

#' @rdname dim
#' @aliases ncol
#' @exportMethod ncol
setMethod(
	f = "ncol",
	signature = "GRegion",
	definition = function(x) unname(x@dimensions[2L])
)

#' @rdname dim
#' @aliases ndepth
#' @exportMethod ndepth
setMethod(
	f = "ndepth",
	signature = "missing",
	definition = function(x) ndepth(.region())
)

#' @rdname dim
#' @aliases ndepth
#' @exportMethod ndepth
setMethod(
	f = "ndepth",
	signature = "GRegion",
	definition = function(x) unname(x@dimensions[3L])
)

#' @rdname dim
#' @aliases ncell
#' @exportMethod ncell
setMethod(
	f = "ncell",
	signature = "missing",
	definition = function(x) ncell(.region())
)

#' @rdname dim
#' @aliases ncell
#' @exportMethod ncell
setMethod(
	f = "ncell",
	signature = "GRegion",
	definition = function(x) prod(x@dimensions[1L:2L])
)

#' @rdname dim
#' @aliases ncell3d
#' @exportMethod ncell3d
setMethod(
	f = "ncell3d",
	signature = "missing",
 	definition = function(x) ncell3d(.region())
)

#' @rdname dim
#' @aliases ncell3d
#' @exportMethod ncell3d
setMethod(
	f = "ncell3d",
	signature = "GRegion",
	definition = function(x) prod(x@dimensions[1L:3L])
)

#' @aliases dim
#' @rdname dim
#' @exportMethod dim
methods::setMethod(
	f = "dim",
	signature = "GVector",
	definition = function(x) dim(x@table)
)

#' @rdname dim
#' @aliases nrow
#' @exportMethod nrow
setMethod(
	f = "nrow",
	signature = "GVector",
	definition = function(x) nrow(x@table)
)

#' @rdname dim
#' @aliases ncol
#' @exportMethod ncol
setMethod(
	f = "ncol",
	signature = "GVector",
	definition = function(x) ncol(x@table)
)

#' @rdname dim
#' @aliases nlyr
#' @exportMethod nlyr
setMethod(
	f = "nlyr",
	signature = "missing",
 	definition = function(x) nlyr(.region())
)

#' @rdname dim
#' @aliases nlyr
#' @exportMethod nlyr
setMethod(
	f = "nlyr",
	signature = "GRaster",
	definition = function(x) x@nLayers
)

#' Compare-methods operations on GRasters and GRegions
#'
#' @description You can do comparative operations on `GRaster`s using normal operators in **R**: `<`, `<=`, `==`, `!=`, `>=`, and `>`. You can also use \code{\link[fasterRaster]{%in%}} for categorical `GRasters` (see `vignette("GRasters", package = "fasterRaster")`).
#'
#' You can also compare two `GRegion`s using the `==` and `!=` operators. Most users of **fasterRaster** will not have to work much with "regions" (see `vignette("regions", package = "fasterRaster")`), so can ignore this functionality. `GRegion`s are the same if they have the same coordinate reference system, location/project and mapset (see `vignette("projects_mapsets", package = "fasterRaster")`), topology (2D or 3D), extent, and resolution. If both are 3D, then they must also have the same vertical extent and number of depths.
#' 
#' @param e1,e2 Values depend on the type of comparison:
#'
#' * Comparing `GRaster`s to logical, numeric, character values: `e1` and `e2` can be any one of these. Comparison to a character string can be useful when using a categorical raster, in which case you can use something like `raster1 == "Wetlands"` to coerce all "wetland" cells to be 1 (TRUE) and all others 0 (FALSE) or `NA` (if it was originally `NA`).
#' * Comparing a `GRegion` to another `GRegion`: `e1` and `e2` must be `GRegion`s!
#'
#' @returns Comparing `GRaster`s: An "integer" `GRaster` with values of 0 (FALSE), 1 (TRUE), or `NA` (neither).
#'
#' Comparing `GRegion`s: Output is logical.
#'
#' @example man/examples/ex_GRaster_comparison_logic.r
#'
#' @aliases Compare-methods
#' @rdname Compare-methods
#' @exportMethod Compare
# raster raster
methods::setMethod(
	f = "Compare",
	signature(e1 = "GRaster", e2 = "GRaster"),
    function(e1, e2) {
	
		compareGeom(e1, e2)
		.locationRestore(e1)
		.region(e1)

		if (nlyr(e1) > 1L & nlyr(e2) == 1L) {
			e2 <- e2[[rep(1L, nlyr(e1))]]
		} else if (nlyr(e1) == 1L & nlyr(e2) > 1L) {
			e1 <- e1[[rep(1L, nlyr(e2))]]
		} else if (nlyr(e1) != nlyr(e2)) {
			stop("Rasters must have the same number of layers, or at least one raster must have a single layer.")
		}

		oper <- as.vector(.Generic)[1L]
		for (i in 1L:nlyr(e1)) {

			name <- paste0(names(e1)[i], "_", names(e2)[i])
			src <- .makeSourceName("comparison", "raster")

			ex <- paste0(src, "= int(if(", sources(e1)[i], oper, sources(e2)[i], "))")
			this <- .genericArithRast(name = name, src = src, ex = ex)
			
			if (i == 1L) {
				out <- this
			} else {
				out <- c(out, this)
			}
			
		}
		out
		
	} # EOF
)

# logical raster
#' @aliases Compare-methods
#' @rdname Compare-methods
#' @exportMethod Compare
methods::setMethod(
	f = "Compare",
	signature(e1 = "logical", e2 = "GRaster"),
    function(e1, e2) {
	
		.locationRestore(e2)
		.region(e2)

		e1 <- as.integer(e1)
		if (is.na(e1)) e1 <- "null()"
		
		oper <- as.vector(.Generic)[1L]
		for (i in 1L:nlyr(e2)) {
			
			name <- names(e2)[i]
			src <- .makeSourceName("comparison", "rast")

			ex <- paste(src, "= int(if(", e1, " ", oper, " ", sources(e2)[i], "))")
			this <- .genericArithRast(name = name, src = src, ex = ex)
			
			if (i == 1L) {
				out <- this
			} else {
				out <- c(out, this)
			}
			
		}
		out
			
	} # EOF
)

# raster logical
#' @aliases Compare-methods
#' @rdname Compare-methods
#' @exportMethod Compare
methods::setMethod(
	f = "Compare",
	signature(e1 = "GRaster", e2 = "logical"),
    function(e1, e2) {
	
		.locationRestore(e1)
		.region(e1)

		e2 <- as.integer(e2)
		if (is.na(e2)) e2 <- "null()"

		oper <- as.vector(.Generic)[1L]
		for (i in 1L:nlyr(e1)) {

			name <- names(e1)[i]
			src <- .makeSourceName("comparison", "rast")
			
			ex <- paste(src, "= int(if(", sources(e1)[i], " ", oper, " ", e2, "))")
			this <- .genericArithRast(name = name, src = src, ex = ex)
			
			if (i == 1L) {
				out <- this
			} else {
				out <- c(out, this)
			}
			
		}
		out
			
	} # EOF
)

# numeric raster
#' @aliases Compare-methods
#' @rdname Compare-methods
#' @exportMethod Compare
methods::setMethod(
	f = "Compare",
	signature(e1 = "numeric", e2 = "GRaster"),
    function(e1, e2) {
	
		.locationRestore(e2)
		.region(e2)

		if (is.na(e1)) e1 <- "null()"

		oper <- as.vector(.Generic)[1L]
		for (i in 1L:nlyr(e2)) {
		
			name <- names(e2)[i]
			src <- .makeSourceName("comparison", "rast")

			ex <- paste(src, "= int(if(", e1, " ", oper, " ", sources(e2)[i], "))")
			this <- .genericArithRast(name = name, src = src, ex = ex)
			
			if (i == 1L) {
				out <- this
			} else {
				out <- c(out, this)
			}

		}
		out
		
	} # EOF
)

# raster numeric
#' @aliases Compare-methods
#' @rdname Compare-methods
#' @exportMethod Compare
methods::setMethod(
	f = "Compare",
	signature(e1 = "GRaster", e2 = "numeric"),
    function(e1, e2) {
	
		.locationRestore(e1)
		.region(e1)

		if (is.na(e2)) e2 <- "null()"
		oper <- as.vector(.Generic)[1L]

		for (i in 1L:nlyr(e1)) {
		
			name <- names(e1)[i]
			src <- .makeSourceName("comparison", "raster")

			ex <- paste0(src, "= int(if(", sources(e1)[i], " ", oper, " ", e2, "))")
			this <- .genericArithRast(name = name, src = src, ex = ex)
			
			if (i == 1L) {
				out <- this
			} else {
				out <- c(out, this)
			}
			
		}
		out
			
	} # EOF
)

# raster integer
#' @aliases Compare-methods
#' @rdname Compare-methods
#' @exportMethod Compare
methods::setMethod(
	f = "Compare",
	signature(e1 = "GRaster", e2 = "integer"),
    function(e1, e2) {
	
		.locationRestore(e1)
		.region(e1)

		if (is.na(e2)) e2 <- "null()"
		oper <- as.vector(.Generic)[1L]

		for (i in 1L:nlyr(e1)) {
		
			name <- names(e1)[i]
			src <- .makeSourceName("comparison", "rast")

			ex <- paste(src, "= int(if(", sources(e1)[i], " ", oper, " ", e2, "))")
			this <- .genericArithRast(name = name, src = src, ex = ex)
			
			if (i == 1L) {
				out <- this
			} else {
				out <- c(out, this)
			}
			
		}
		out
			
	} # EOF
)

# integer raster
#' @aliases Compare-methods
#' @rdname Compare-methods
#' @exportMethod Compare
methods::setMethod(
    f = "Compare",
    signature(e1 = "integer", e2 = "GRaster"),
    function(e1, e2) {

        .locationRestore(e2)
		.region(e2)

        if (is.na(e1)) e1 <- "null()"

        oper <- as.vector(.Generic)[1L]
        for (i in 1L:nlyr(e2)) {
            name <- names(e2)[i]
            src <- .makeSourceName("comparison", "rast")

            ex <- paste(src, "= int(if(", e1, " ", oper, " ", sources(e2)[i], "))")
            this <- .genericArithRast(name = name, src = src, ex = ex)

            if (i == 1L) {
                out <- this
            } else {
                out <- c(out, this)
            }
        }
        out
    } # EOF
)

# raster character
#' @aliases Compare-methods
#' @rdname Compare-methods
#' @exportMethod Compare
methods::setMethod(
	f = "Compare",
	signature(e1 = "GRaster", e2 = "character"),
    function(e1, e2) {
	
	if (!all(is.factor(e1))) stop("Raster must be categorical for this type of comparison.")
	
	.locationRestore(e1)
	.region(e1)

	oper <- as.vector(.Generic)[1L]
	levs <- levels(e1)
	acs <- activeCat(e1, names = TRUE)

	nLayers <- nlyr(e1)
	srcs <- .makeSourceName("comparison", "raster", nLayers)

	for (i in seq_len(nLayers)) {

		# get value of this category
		ac <- acs[i]
		if (faster("useDataTable")) {
			thisVal <- eval(parse(text = paste0("levs[[", i, "]][", ac, " == e2]")))
			thisVal <- thisVal[[1L]]
		} else {
			thisVal <- eval(parse(text = paste0("levs[[", i, "]][levs[[", i, "]]$", ac, " == e2, 1L]")))
		}
		

		if (length(thisVal) == 0L) {
			return(0L * not.na(e1))
		} else {

			if (oper == "<") {

				thisVal <- min(thisVal)
				ex <- paste0(srcs[i], " = if(", sources(e1)[i], " ", oper, " ", thisVal, ", 1, 0)")

			} else if (oper == "<=") {

				thisVal <- max(thisVal)
				ex <- paste0(srcs[i], " = if(", sources(e1)[i], " ", oper, " ", thisVal, ", 1, 0)")

			} else if (oper == ">") {
				
				thisVal <- max(thisVal)
				ex <- paste0(srcs[i], " = if(", sources(e1)[i], " ", oper, " ", thisVal, ", 1, 0)")

			} else if (oper == ">=") {
				
				thisVal <- min(thisVal)
				ex <- paste0(srcs[i], " = if(", sources(e1)[i], " ", oper, " ", thisVal, ", 1, 0)")

			} else if (oper == "==") {

				ex <- paste0(srcs[i], " = if(", sources(e1)[i], " == ", thisVal[1L])

				if (length(thisVal) > 1L) {
					for (count in 2L:length(thisVal)) {
						ex <- paste0(ex, " | ", sources(e1)[i], " == ", thisVal[count])
					}
				}

				ex <- paste0(ex, ", 1, 0)")

			} else if (oper == "!=") {
			
				ex <- paste0(srcs[i], " = if(", sources(e1)[i], " != ", thisVal[1L])

				if (length(thisVal) > 1L) {
					for (count in 2L:length(thisVal)) {
						ex <- paste0(ex, " & ", sources(e1)[i], " != ", thisVal[count])
					}
				}

				ex <- paste0(ex, ", 1, 0)")
			
			}

			rgrass::execGRASS(cmd = "r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

		}

	} # next layer
	out <- makeGRaster(srcs, "logical", levels = cats(e1), ac = activeCats(e1))
	out
	
	} # EOF
)

# character raster
#' @aliases Compare-methods
#' @rdname Compare-methods
#' @exportMethod Compare
methods::setMethod(
	f = "Compare",
	signature(e1 = "character", e2 = "GRaster"),
    function(e1, e2) {
	
	if (!all(is.factor(e2))) stop("Raster must be categorical for this type of comparison.")
	.locationRestore(e2)
	.region(e2)

	oper <- as.vector(.Generic)[1L]
	levs <- levels(e2)
	acs <- activeCat(e2, names = TRUE)

	nLayers <- nlyr(e2)
	srcs <- .makeSourceName("comparison", "raster", nLayers)

	for (i in seq_len(nLayers)) {

		# get value of this category
		ac <- acs[i]
		thisVal <- eval(parse(text = paste0("levs[[", i, "]][", ac, " == e1]")))
		thisVal <- thisVal[[1L]]

		if (length(thisVal) == 0L) {
			this <- 0L * not.na(e2)
		} else {

			if (oper == "<") {

				thisVal <- min(thisVal)
				ex <- paste0(srcs[i], " = if(", sources(e2)[i], " ", oper, " ", thisVal, ", 1, 0)")

			} else if (oper == "<=") {

				thisVal <- max(thisVal)
				ex <- paste0(srcs[i], " = if(", sources(e2)[i], " ", oper, " ", thisVal, ", 1, 0)")

			} else if (oper == ">") {
				
				thisVal <- max(thisVal)
				ex <- paste0(srcs[i], " = if(", sources(e2)[i], " ", oper, " ", thisVal, ", 1, 0)")

			} else if (oper == ">=") {
				
				thisVal <- min(thisVal)
				ex <- paste0(srcs[i], " = if(", sources(e2)[i], " ", oper, " ", thisVal, ", 1, 0)")

			} else if (oper == "==") {

				ex <- paste0(srcs[i], " = if(", sources(e2)[i], " == ", thisVal[1L])

				if (length(thisVal) > 1L) {
					for (count in 2L:length(thisVal)) {
						ex <- paste0(ex, " | ", sources(e2)[i], " == ", thisVal[count])
					}
				}

				ex <- paste0(ex, ", 1, 0)")

			} else if (oper == "!=") {
			
				ex <- paste0(srcs[i], " = if(", sources(e2)[i], " != ", thisVal[1L])

				if (length(thisVal) > 1L) {
					for (count in 2L:length(thisVal)) {
						ex <- paste0(ex, " & ", sources(e2)[i], " != ", thisVal[count])
					}
				}

				ex <- paste0(ex, ", 1, 0)")
			
			}

			rgrass::execGRASS(
				cmd = "r.mapcalc",
				expression = ex,
				flags = c(.quiet(), "overwrite")
			)

		}

	} # next layer
	out <- makeGRaster(srcs, "logical", levels = cats(e2), ac = activeCats(e2))
	out
		
	} # EOF
)

# GRegion GRegion
#' @aliases Compare-methods
#' @rdname Compare-methods
#' @exportMethod Compare
methods::setMethod(
	f = "Compare",
	signature(e1 = "GRegion", e2 = "GRegion"),
    function(e1, e2) {
	
	oper <- as.vector(.Generic)[1L]
	if (!(oper %in% c("==", "!="))) stop("Can only use == and != operators when performing comparisons involving GRegions.")

	out <- TRUE
	if (.location(e1) != .location(e2)) out <- FALSE
	if (!terra::same.crs(e1, e2)) out <- out & FALSE

	if ((is.3d(e1) & !is.3d(e2)) | (!is.3d(e1) & is.3d(e2))) {
		out <- out & FALSE
	} else if (is.3d(e1) & is.3d(e2)) {
		top1 <- top(e1)
		top2 <- top(e2)
		bottom1 <- bottom(e1)
		bottom2 <- bottom(e2)
		out <- out & omnibus::compareFloat(top1, top2, "==") & omnibus::compareFloat(bottom1, bottom2, "==") & ndepth(e1) == ndepth(e2)
	}

	ext1 <- ext(e1, vector = TRUE)
	ext2 <- ext(e2, vector = TRUE)
	res1 <- res(e1)
	res2 <- res(e2)
	out <- out & all(omnibus::compareFloat(ext1, ext2, "==")) & all(omnibus::compareFloat(res1, res2, "=="))

	if (oper == "!=") out <- !out
	out

	} # EOF
)

### The problem with comparing GRasters and GRegions is that GRegions are always 3D, and GRasters may not be, so the test will always fail.
# # # # GRegion GRaster
# # # #' @aliases Compare-methods
# # # #' @rdname Compare-methods
# # # #' @exportMethod Compare
# # # methods::setMethod(
# # # 	f = "Compare",
# # # 	signature(e1 = "GRegion", e2 = "GRaster"),
# # #     function(e1, e2) {

# # # 	oper <- as.vector(.Generic)[1L]
# # # 	.compareGRegionGRaster(e1 = e1, e2 = e2, oper = oper)

# # # 	} # EOF
# # # )

# # # # GRaster GRegion
# # # #' @aliases Compare-methods
# # # #' @rdname Compare-methods
# # # #' @exportMethod Compare
# # # methods::setMethod(
# # # 	f = "Compare",
# # # 	signature(e1 = "GRaster", e2 = "GRegion"),
# # #     function(e1, e2) {

# # # 	oper <- as.vector(.Generic)[1L]
# # # 	.compareGRegionGRaster(e1 = e1, e2 = e2, oper = oper)

# # # 	} # EOF
# # # )

# # # #' @noRd
# # # .compareGRegionGRaster <- function(e1, e2, oper) {
	
# # # 	if (!(oper %in% c("==", "!="))) stop("Can only use == and != operators when performing comparisons involving GRegions.")

# # # 	out <- TRUE
# # # 	if (.location(e1) != .location(e2)) out <- FALSE
# # # 	if (!terra::same.crs(e1, e2)) out <- out & FALSE

# # # 	if (is.3d(e1) & is.3d(e2)) {
# # # 		top1 <- top(e1)
# # # 		top2 <- top(e2)
# # # 		bottom1 <- bottom(e1)
# # # 		bottom2 <- bottom(e2)
# # # 		out <- out & omnibus::compareFloat(top1, top2, "==") & omnibus::compareFloat(bottom1, bottom2, "==") & ndepth(e1) == ndepth(e2)
# # # 	}

# # # 	ext1 <- ext(e1, vector = TRUE)
# # # 	ext2 <- ext(e2, vector = TRUE)
# # # 	res1 <- res(e1)
# # # 	res2 <- res(e2)
# # # 	out <- out & all(omnibus::compareFloat(ext1, ext2, "==")) & all(omnibus::compareFloat(res1, res2, "=="))

# # # 	if (oper == "!=") out <- !out
# # # 	out

# # # }

# # # # GRegion GVector
# # # #' @aliases Compare-methods
# # # #' @rdname Compare-methods
# # # #' @exportMethod Compare
# # # methods::setMethod(
# # #     f = "Compare",
# # #     signature(e1 = "GRegion", e2 = "GVector"),
# # #     function(e1, e2) {
# # #         oper <- as.vector(.Generic)[1L]
# # #         .compareGRegionGVector(e1 = e1, e2 = e2, oper = oper)
# # #     } # EOF
# # # )

# # # # GVector GRegion
# # # #' @aliases Compare-methods
# # # #' @rdname Compare-methods
# # # #' @exportMethod Compare
# # # methods::setMethod(
# # #     f = "Compare",
# # #     signature(e1 = "GVector", e2 = "GRegion"),
# # #     function(e1, e2) {
# # #         oper <- as.vector(.Generic)[1L]
# # #         .compareGRegionGVector(e1 = e1, e2 = e2, oper = oper)
# # #     } # EOF
# # # )

# # # #' @noRd
# # # .compareGRegionGVector <- function(e1, e2, oper) {
	
# # # 	if (!(oper %in% c("==", "!="))) stop("Can only use == and != operators when performing comparisons involving GRegions.")

# # # 	out <- TRUE
# # # 	if (.location(e1) != .location(e2)) out <- FALSE
# # # 	if (!terra::same.crs(e1, e2)) out <- out & FALSE

# # # 	ext1 <- ext(e1, vector = TRUE)
# # # 	ext2 <- ext(e2, vector = TRUE)
# # # 	out <- out & all(omnibus::compareFloat(ext1, ext2, "==")) & all(omnibus::compareFloat(res1, res2, "=="))

# # # 	if (oper == "!=") out <- !out
# # # 	out

# # # }

#' Summary statistics for GRasters
#'
#' @description `global()` calculates a summary statistic across all the cells of a `GRaster`. It returns a single value for each layer of the raster.
#'
#' @param x A `GRaster` or missing.  If missing, then a vector of all of the accepted function names is returned.
#'
#' @param fun Character vector: The name of the function(s):
#' * `"*"`: All of the functions below.
#' * `"cv"`: Sample coefficient of variation (expressed as a proportion of the mean).
#' * `"cvpop"`: Population coefficient of variation (expressed as a proportion of the mean).
#' * `"max"` and `"min"`: Highest and lowest values across non-`NA` cells. NB: [minmax()] is faster.
#' * `"mean"` (default): Average.
#' * `"meanAbs"`: Mean of absolute values.
#' * `"median"`: Median.
#' * `"quantile"`: Quantile (see also argument `probs`).
#' * `"range"`: Range. Note that following [terra::global()], the minimum and maximum are reported, not the actual range.
#' * `"sd"`: Sample standard deviation (same as [stats::sd()]).
#' * `"sdpop"`: Population standard deviation.
#' * `"sum"`: Sum.
#' * `"var"`: Sample variance (same as [stats::var()]).
#' * `"varpop"`: Population variance.
#'
#' @param probs Numeric within the range from 0 to 1: Quantile(s) at which to calculate `quantile`.
#'
#' @param ... Other arguments (unused).
#'
#' @returns If `x` is missing, the function returns a character vector of all accepted function names. If `x` is a `GRaster`, a data frame with the specified statistics is returned.
#'
#' @seealso [terra::global()] and **GRASS** tool `r.univar`
#'
#' @example man/examples/ex_global.r
#'
#' @aliases global
#' @rdname global
#' @exportMethod global
methods::setMethod(
	f = "global",
	signature = c(x = "GRaster"),
	definition = function(
		x,
		fun = "mean",
		probs = seq(0, 1, 0.25),
		...
	) {

	.global(x = x, fun = fun, probs = probs)

	} # EOF
)

# x 		A `GRaster` or the [sources()] name of one
# fun 		character
# probs		numeric in [0, 1]
#' @noRd
.global <- function(x, fun, probs = seq(0, 1, 0.25)) {

	if (inherits(x, "GRaster")) {
		.locationRestore(x)
		.region(x)
		src <- sources(x)
	} else {
		src <- x
	}

	if (any(fun == "*")) fun <- global()
	funs <- global()
	fun <- omnibus::pmatchSafe(fun, funs, useFirst = TRUE, error = TRUE)

	if (any(fun == "range")) {
		if (length(fun) == 1L) {
			fun <- c("min", "max")
		} else {
			where <- which(fun == "range")
			lf <- length(fun)
			if (where == 1L) {
				fun <- c("min", "max", fun[2L:lf])
			} else if (where == lf) {
				fun <- c(fun[1L:(lf - 1L)], "min", "max")
			} else {
				fun <- c(fun[1L:(where - 1L)], "min", "max", fun[(where + 1L):lf])
			}
		}
	}

	fun <- unique(fun)
	if (any(fun == "quantile") && any(probs < 0) | any(probs > 1)) stop("The value for `probs` must be in the range [0, 1].")
	
	versionNumber <- grassInfo("versionNumber")

	out <- data.frame()
	nLayers <- length(src)

	### for each layer
	for (i in seq_len(nLayers)) {

		if (any(fun == "quantile")) {

			# NB using r.quantile() can be **much** faster than using "r.univar" to get quantiles!!! (minutes vs weeks)

			percents <- 100 * probs
			
			quantInfo <- rgrass::execGRASS(
				cmd = 'r.quantile',
				input = src[i],
				percentiles = percents,
				flags = c(.quiet(), "overwrite"),
				Sys_show.output.on.console = FALSE,
				echoCmd = FALSE,
				intern = TRUE
			)

		}

		if (any(fun != "quantile")) {

			args <- list(
				cmd = "r.univar",
				map = src[i],
				flags = c(.quiet(), "r"),
				Sys_show.output.on.console = FALSE,
				echoCmd = FALSE,
				intern = TRUE
			)

			if (versionNumber >= 8.4) args$nprocs <- faster("cores")
			if (any(fun == "median")) args$flags <- c(args$flags, "e")

			info <- do.call(rgrass::execGRASS, args)

		}
			
		# values for this layer
		nc <- length(fun)
		if (any(fun == "quantile")) nc <- nc + length(probs) - 1
		thisOut <- data.frame(
			matrix(NA_real_,
				ncol = nc,
				nrow = 1L
			),
			row.names = src[i]
		)
		
		if (any(fun == "quantile")) {
			
			colNames <- fun
			isQuant <- which(colNames == "quantile")
			quantNames <- paste0("quantile_", probs)
			nCols <- length(colNames)
			if (length(fun) == 1L & fun[1L] == "quantile") {
				colNames <- quantNames
			} else if (colNames[1L] == "quantile") {
				colNames <- c(quantNames, colNames[2L:nCols])
			} else if (colNames[nCols] == "quantile") {
				colNames <- c(colNames[1L:(nCols - 1L)], quantNames)
			} else {
				colNames <- c(colNames[(1L:(isQuant) - 1L)], quantNames, colNames[(isQuant + 1L):nCols])
			}
			names(thisOut) <- colNames
			
		} else {
			names(thisOut) <- fun
		}
			
		### for each function
		for (countFun in seq_along(fun)) {

			thisFun <- fun[countFun]

			pattern <- if (thisFun == "meanAbs") {
				"mean of absolute values: "
			} else if (thisFun == "mean") {
				"mean: "
			} else if (thisFun == "min") {
				"minimum: "
			# } else if (thisFun == "countNA") {
				# "total null cells: "
			} else if (thisFun == "range") {
				"range: "
			} else if (thisFun == "sdpop") {
				"standard deviation: "
			} else if (thisFun == "varpop") {
				"variance: "
			} else if (thisFun == "sum") {
				"sum: "
			} else { 
				NA
			}
			
			if (!is.na(pattern)) {
			
				this <- info[grepl(info, pattern=pattern)]
				this <- sub(this, pattern = pattern, replacement = "")
				this <- as.numeric(this)

			} else {
			
				# r.stats returns "nan" if all vales on the raster are the same... but the minimum is the given value, so use that
				if (thisFun == "max") {
				
					pattern <- "maximum: "
					this <- info[grepl(info, pattern = pattern)]
					this <- strsplit(this, split=":")[[1L]][2L]
					this <- as.numeric(this)

					if (is.nan(this)) {
						pattern <- "minimum: "
						this <- info[grepl(info, pattern = pattern)]
						this <- strsplit(this, split=":")[[1L]][2L]
						this <- as.numeric(this)
					}
					
				} else if (thisFun == "quantile") {
				
					# pattern <- "percentile: "
					# this <- info[grepl(info, pattern=pattern)]
					# this <- strsplit(this, split=":")[[1L]][2L]
					# this <- as.numeric(this)

					quantInfo <- quantInfo[seq_along(probs)]
					this <- strsplit(quantInfo, split = ":")
					this <- lapply(this, as.numeric)
					this <- do.call(rbind, this)
					this <- this[ , 3L, drop = TRUE]
					
				} else if (thisFun == "median") {
				
					pattern <- "median \\(even number of cells\\): "

					this <- info[grepl(info, pattern = pattern)]

					if (length(this) == 0L) {

						pattern <- "median \\(odd number of cells\\): "

						this <- info[grepl(info, pattern = pattern)]

					}

					this <- sub(this, pattern = pattern, replacement = "")
					this <- as.numeric(this)

				# } else if (thisFun == "countNonNA") {
				
				# 	pattern <- "total null and non-null cells: "
				# 	this1 <- info[grepl(info, pattern=pattern)]
				# 	this1 <- sub(this1, pattern=pattern, replacement="")
				# 	this1 <- as.numeric(this1)
					
				# 	pattern <- "total null cells: "
				# 	this2 <- info[grepl(info, pattern=pattern)]
				# 	this2 <- sub(this2, pattern=pattern, replacement="")
				# 	this2 <- as.numeric(this2)
					
				# 	this <- this1 - this2
					
				} else if (thisFun == "cvpop") {
				
					pattern <- "variation coefficient: "
					this <- info[grepl(info, pattern=pattern)]
					this <- sub(this, pattern=pattern, replacement="")
					this <- sub(this, pattern="%", replacement="")
					this <- as.numeric(this)
					this <- this / 100
					
				} else if (thisFun == "cv") {
				
					pattern <- "mean: "
					mean. <- info[grepl(info, pattern = pattern)]
					mean. <- sub(mean., pattern = pattern, replacement="")
					mean. <- as.numeric(mean.)

					srcSS <- .makeSourceName("r_mapcalc", "rast")
					ex <- paste0(srcSS, " = (", sources(x)[i], " - ", mean., ")^2")
					
					rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))
				
					args <- list(
						cmd = "r.univar",
						flags = c("r", .quiet()),
						map = srcSS,
						nprocs = faster("cores"),
						Sys_show.output.on.console = FALSE,
						echoCmd = FALSE,
						intern = TRUE
					)

					if (versionNumber >= 8.4) args$nprocs <- faster("cores")

					thisInfo <- do.call(rgrass::execGRASS, args = args)
					on.exit(.rm(srcSS, type = "raster", warn = FALSE), add = TRUE)

					pattern <- "sum: "
					ss <- thisInfo[grepl(info, pattern=pattern)]
					ss <- sub(ss, pattern=pattern, replacement="")
					ss <- as.numeric(ss)

					pattern <- "total null and non-null cells: "
					n1 <- thisInfo[grepl(thisInfo, pattern=pattern)]
					n1 <- sub(n1, pattern=pattern, replacement="")
					n1 <- as.numeric(n1)
					
					pattern <- "total null cells: "
					n2 <- thisInfo[grepl(thisInfo, pattern=pattern)]
					n2 <- sub(n2, pattern=pattern, replacement="")
					n2 <- as.numeric(n2)
					
					n <- n1 - n2
					stdev <- sqrt(ss / (n - 1))
					this <- stdev / mean.

				} else if (thisFun %in% c("var", "sd")) {
	
					pattern <- "mean: "
					mean. <- info[grepl(info, pattern = pattern)]
					mean. <- sub(mean., pattern = pattern, replacement="")
					mean. <- as.numeric(mean.)

					srcSS <- .makeSourceName("r_mapcalc", "raster")
					
					ex <- paste0(srcSS, " = (", src[i], " - ", mean., ")^2")

					rgrass::execGRASS(
						cmd = "r.mapcalc",
						expression = ex,
						flags = c(.quiet(), "overwrite")
					)

					args <- list(
						cmd = "r.univar",
						flags = c("r", .quiet()),
						map = srcSS,
						nprocs = faster("cores"),
						Sys_show.output.on.console = FALSE,
						echoCmd = FALSE,
						intern = TRUE
					)

					if (grassInfo("versionNumber") >= 8.4) args$nprocs <- faster("cores")

					thisInfo <- do.call(rgrass::execGRASS, args = args)
					on.exit(.rm(srcSS, type = "raster", warn = FALSE), add = TRUE)

					pattern <- "sum: "
					ss <- thisInfo[grepl(info, pattern=pattern)]
					ss <- sub(ss, pattern=pattern, replacement="")
					ss <- as.numeric(ss)

					# NB need to convert to numeric to obviate overflow
					n <- .nonnacell(x[[i]], warn = FALSE, nLayers = 1)

					this <- ss / (n - 1)
					if (thisFun == "sd") this <- sqrt(this)

				} # custom function
				
			} # match function
			
			if (fun[countFun] != "quantile") {
				thisOut[1L, fun[countFun]] <- this
			} else {
				thisOut[1L, match(paste0("quantile_", probs), colnames(thisOut))] <- this
			}
				
		} # next function

		out <- rbind(out, thisOut)

	} # next layer

	rownames(out) <- names(x)
	out

}

#' @aliases global
#' @rdname global
#' @exportMethod global
methods::setMethod(
	f = "global",
	signature = c(x = "missing"),
	definition = function(x, ...) {
	
	c(
		# "countNA",
		# "countNonNA",
		"cv",
		"cvpop",
		"mean",
		"meanAbs",
		"median",
		"min",
		"max",
		"range",
		"sd",
		"sdpop",
		"sum",
		"var",
		"varpop",
		"quantile"
	)

	} # EOF
)

#' Add columns to the data table of a GVector
#'
#' @description `colbind()` adds columns to the data table of a `GVector`. You can combine multiple a `GVector`'s data table with `data.frame`s, `data.table`s, `matrices`, or the data table(s) from other `GVector`(s). To combine two `GVector`s, see [rbind()].
#'
#' @param x,... The first argument must be a `GVector`. Subsequent arguments can be `data.frame`s, `data.table`s, `matrices`, or `GVector`s. Only the data tables of subsequent `GVector`s are added to the table in `x`; the geometries are ignored.
#'
#' @returns A `GVector`.
#'
#' @seealso [rbind()], \code{\link[fasterRaster]{addTable<-}}, [dropTable()]
#'
#' @example man/examples/ex_cbind_rbind_addTable_dropTable.r
#'
#' @aliases colbind
#' @rdname colbind
#' @exportMethod colbind
methods::setMethod(
	f = "colbind",
	signature = c(x = "GVector"),
	function(x, ...) {

	dots <- list(x, ...)
	for (i in seq_along(dots)) {
		if (inherits(dots[[i]], "GVector")) {
			dots[[i]] <- dots[[i]]@table
		}
	}
	
	table <- do.call(cbind, dots)
	if (!inherits(table, "data.table")) table <- data.table::as.data.table(table)
	x@table <- table
	x
	
	} # EOF
)

if (grassStarted()) {

# Setup
library(sf)

# Example vectors
madDypsis <- fastData("madDypsis") # points
madCoast4 <- fastData("madCoast4") # polygons

# Convert sf vectors to GVectors
dypsis <- fast(madDypsis)
coast4 <- fast(madCoast4)
ant <- coast4[coast4$NAME_4 == "Antanambe"]

# Delaunay triangulation
dypsisDel <- delaunay(dypsis)
plot(dypsisDel)
plot(dypsis, pch = 1, col = "red", add = TRUE)

# Voronoi tessellation
vor <- voronoi(dypsis)
plot(vor)
plot(dypsis, pch = 1, col = "red", add = TRUE)

# Random Voronoi tessellation
rand <- rvoronoi(coast4, size = 100)
plot(rand)

}

#' Display a raster or vector
#'
#' @description `plot()` displays a `GRaster` or `GVector`.
#'
#' This function is essentially a hack, as it it not possible to dependably call the appropriate **GRASS** modules and display a raster or vector without potential confusion on the user side. Instead, this function 1) simplifies the focal `GRaster` or `GVector` to make it smaller when saved to disk; 2) writes the object to disk; 3) (internally) creates a `SpatRaster` or `SpatVector` object; then 4) plots the object using [terra::plot()]. Thus, if you are interested in making maps, it will always be faster to make them directly with **terra** or **sf**.
#'
#' @param x A `GRaster` or `GVector`.
#'
#' @param y Missing--leave as empty.
#'
#' @param simplify Logical: If `TRUE` (default) and the raster has an x- and/or y-resolution that is greater than the screen resolution, then [aggregate()] will be applied to the raster before it is saved to reduce the time it takes to save the raster.
#'
#' @param maxGeoms Positive integer (vectors only): Maximum number of features before vector simplification is applied before saving to disk then creating a `SpatVector` for plotting. The default is 10000.
#' 
#' @param ... Other arguments to send to [terra::plot()].
#'
#' @returns Nothing (displays a raster or vector).
#'
#' @seealso [terra::plot()]
#'
#' @example man/examples/ex_plot.r
#'
#' @aliases plot
#' @rdname plot
#' @exportMethod plot
methods::setMethod(
	f = "plot",
	signature = c(x = "GRaster", y = "missing"),
	function(x, y, simplify = TRUE, ...) {

	# simplify
	if (simplify & !is.3d(x)) {
		
		# screenRes <- rpanel::rp.screenresolution()
		screenRes <- omnibus::screenRes()
		screenWidth <- screenRes[["width"]]
		screenHeight <- screenRes[["width"]]

		rows <- nrow(x)
		cols <- ncol(x)
		# pow <- 5 # power by which to reduce dimensions, set at 5 semi-empirically
		if (screenWidth < cols | screenHeight < rows) {
		
			ewScale <- max(1, floor(0.5 * cols / screenWidth))
			nsScale <- max(1, floor(0.5 * rows / screenHeight))

			if (ewScale >= 2 | nsScale >= 2) {

				if (any(datatype(x) %in% c("integer", "factor"))) {
					fun <- "mode"
				} else {
					fun <- "mean"
				}

				fact <- c(ewScale, nsScale)
				x <- aggregate(x, fact = fact, fun = fun)

			}
			
		}

	}

	# nc <- ncell(x)
	# if (nc > maxcell) {

	# 	rows <- nrow(x)
	# 	cols <- ncol(x)
	# 	rescale <- ceiling(nc / (2 * maxcell))

	# 	if (all(datatype(x) %in% c("integer", "factor"))) {
	# 		fun <- "mode"
	# 	} else {
	# 		fun <- "mean"
	# 	}

	# 	if (rescale != 1L) x <- aggregate(x, fact = rescale, fun = fun)

	# }
	
	# # dtype <- datatype(x, "GRASS")
	# # if (all(dtype %in% "CELL")) {
	
	# # 	if (all(.minVal(x) >= 0L & .maxVal(x) <= 255L)) {
	# # 		dtype <- "Byte"
	# # 	} else if (all(.minVal(x) >= 0L & .maxVal(x) <= 65534L)) {
	# # 		dtype <- "UInt16"
	# # 	} else if (all(.minVal(x) >= -32767L & .maxVal(x) <= 32767L)) {
	# # 		dtype <- "Int16"
	# # 	} else if (all(.minVal(x) >= -2147483647L & .maxVal(x) <= 2147483647L)) {
	# # 		dtype <- "Int32"
	# # 	} else if (all(.minVal(x) >= -3.4E+38 & .maxVal(x) <= 3.4E+38)) {
	# # 		dtype <- "Float32"
	# # 	} else {
	# # 		dtype <- "Float64"
	# # 	}

	# # } else {
	# # 	dtype <- "Float64"
	# # }

	tf <- tempfile(fileext = ".tif")
	writeRaster(x, filename = tf, format = "GeoTIFF", bigTiff = TRUE, overwrite = TRUE, warn = FALSE, ...)
	out <- terra::rast(tf)
	
	facts <- is.factor(x)
	if (any(facts)) {
	
		for (i in which(facts)) {
			levels(out[[i]]) <- levels(x)[[i]]
		}
	
	}
	
	names(out) <- names(x)
	terra::plot(out, ...)
	
	} # EOF
)

#' @aliases plot
#' @rdname plot
#' @exportMethod plot
methods::setMethod(
	f = "plot",
	signature = c(x = "GVector", y = "missing"),
	function(x, y, maxGeoms = 10000, ...) {
	
	# simplify
	if (ngeom(x) > maxGeoms) x <- simplifyGeom(x)
	
	tf <- tempfile(fileext = ".gpkg")
	writeVector(x, filename = tf, format = "GPKG", overwrite = TRUE, attachTable = FALSE)
	v <- terra::vect(tf)
	terra::plot(v, ...)
	
	} # EOF
)

#' Plot using a G-object's [sources()] name
#'
#' @param x The [sources()] name of a `GRaster` or `GVector`
#' @param ... Other arguments to pass to [plot()].
#'
#' @keywords internal
.plot <- function(x, ...) {
	files <- .ls()
	what <- names(files)[files == x]
	if (what == "raster") {
		x <- makeGRaster(x)
	} else if (what == "vector") {
		x <- makeGVector(x)
	}
	plot(x, ...)
}

if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster
elev <- fast(madElev)
elevs <- c(elev, elev, log10(elev) - 1, sqrt(elev))
names(elevs) <- c("elev1", "elev2", "log_elev", "sqrt_elev")

elev
elevs

# Comparisons
elev < 100
elev <= 100
elev == 100
elev != 100
elev > 100
elev >= 100

elev + 100 < 2 * elev

elevs > 10
10 > elevs

# logic
elev < 10 | elev > 200
elev < 10 | cos(elev) > 0.9

elev < 10 | TRUE
TRUE | elev > 200

elev < 10 | FALSE
FALSE | elev > 200

elev < 10 & cos(elev) > 0.9

elev < 10 & TRUE
TRUE & elev > 200

elev < 10 & FALSE
FALSE & elev > 200

}

if (grassStarted()) {

# Setup
library(terra)

# Example data
madRivers <- fastData("madRivers")

# What file formats can we attempt to write?
writeVector()

# Convert SpatVector to GVector
rivers <- fast(madRivers)
rivers

# Save GVector to disk as GeoPackage
filename <- tempfile(fileext = ".gpkg")
writeVector(rivers, filename)

# Save GVector to disk as ESRI Shapefile
filename <- tempfile(fileext = ".shp")
writeVector(rivers, filename)

# Save GVector to disk as Google Earth KML
filename <- tempfile(fileext = ".klm")
writeVector(rivers, filename)

# Save GVector data table to disk as comma-separated file
filename <- tempfile(fileext = ".csv")
writeVector(rivers, filename)

# Save GVector data table to disk as NetCDF
filename <- tempfile(fileext = ".ncdf")
writeVector(rivers, filename)

# Save GVector data table to disk as Excel file
filename <- tempfile(fileext = ".xlsx")
writeVector(rivers, filename)

}

if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Elevation raster and points at which Dypsis have been collected:
madElev <- fastData("madElev")
madDypsis <- fastData("madDypsis")

# Convert a SpatRaster to a GRaster and sf to GVector:
elev <- fast(madElev)
dypsis <- fast(madDypsis)

# We will interpolate the value of collection year in the "year" field of dypsis. But first, we need to remove the NA value.
dypsisNoNA <- dypsis[!is.na(dypsis$year)]

# Interpolate:
idw <- interpIDW(dypsisNoNA, elev, field = "year")

dypsisNoNA$year
plot(idw)
plot(dypsisNoNA, add = TRUE)

}

if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")
madRivers <- fastData("madRivers")

# Send spatial objects to GRASS:
elev <- fast(madElev)
rivers <- fast(madRivers)

# Extend raster by number of rows/columns:
extended1 <- extend(elev, 10, fill = 900)
extended2 <- extend(elev, c(10, 20), fill = 900)
extended3 <- extend(elev, c(10, 80, 0, 100), fill = 900)
dim(elev)
dim(extended1)
dim(extended2)
dim(extended3)

plot(extended3)

# When exporting a raster, NA rows and columns are removed.
extended4 <- extend(elev, 100, fill=1) # default fill is NA
extended4terra <- rast(extended4)

dim(extended4)
dim(extended4terra)

plot(extended4)

# Extend the raster by another object with a wider extent.

# For tis example, first crop the raster, then extend it.
elevCrop <- crop(elev, rivers)
uncrop <- extend(elevCrop, elev, fill = 900)
plot(uncrop)

}

if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")
madCoast4 <- fastData("madCoast4")

# Convert to GRaster and crop to a sub-portion (easier for visualizing)
elev <- fast(madElev)
coast4 <- fast(madCoast4)
ant <- coast4[coast4$NAME_4 == "Antanambe"]
elevAnt <- crop(elev, ant)

# Create a set of random points to serve as starting points:
starts <- spatSample(elevAnt, 10, as.points = TRUE, seed = 2)

# Remove points in water:
starts <- starts[complete.cases(starts)]

# Calculate flow paths and label each by ID:
paths <- flowPath(elevAnt, starts)
paths

plot(elevAnt, legend = FALSE, main = "Flow path for each point")
plot(paths, add = TRUE)
plot(starts, pch = 1, add = TRUE)

# Flow paths with cell values indicating number of cells from each start:
seqs <- flowPath(elevAnt, starts, return = "seq")

plot(elevAnt, legend = FALSE, main = "Sequentially-numbered flow paths")
plot(seqs, add = TRUE)
plot(starts, pch = 1, add = TRUE)

# We can convert flow paths to lines:
seqLines <- as.lines(seqs)
plot(seqLines)
seqLines

}

if (grassStarted() & interactive()) {

# Open help pages for `r.mapcalc` and `r.sun`:
grassHelp("r.mapcalc")
grassHelp("r.sun")

# GRASS table of contents:
grassHelp("toc")

# Index page:
grassHelp("index")

}

#' @name madCoverCats
#'
#' @title Table of land cover classes for an eastern portion of Madagascar
#'
#' @description This data frame corresponds to the [madCover][madCover] raster, which represents land cover for an eastern portion of Madagascar. Note that the land cover classes have been simplified, so this table and raster should *not* be used for "real" analyses.
#'
#' @docType data
#'
#' @format An object of class `data.frame`.
#'
#' @keywords Madagascar
#'
#' @seealso [madCover], `vignette("GRasters", package = "fasterRaster")`
#'
#' @references  Arino O., P. Bicheron, F. Achard, J. Latham, R. Witt and J.-L. Weber. 2008. GlobCover: The most detailed portrait of Earth. European Space Agency Bulletin 136:25-31. [http://due.esrin.esa.int](http://due.esrin.esa.int).
#'
#' @source [http://due.esrin.esa.int](http://due.esrin.esa.int)
#'
#' @example man/examples/ex_fastData.r
#'
NULL

if (grassStarted()) {

# Setup
library(terra)

# Elevation raster
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster:
elev <- fast(madElev)

# Calculate contour lines:
conts <- as.contour(elev, nlevels = 10)

plot(elev)
plot(conts, add = TRUE)

}

if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster:
elev <- fast(madElev)

# Classify using a scalar indicating number of bins
scalar <- classify(elev, 5)
scalar
levels(scalar)

# Classify using a vector, indicating bin break points
vector <- classify(elev, rcl = c(0, 100, 200, 300, 400, 500, 600))
vector
levels(vector)

# Classify using a 2-column matrix (only valid for integer rasters)
rcl <- data.frame(is = c(1:3, 5, 10), becomes = c(100:102, 105, 110))
twoCol <- classify(elev, rcl = rcl)
twoCol

# Classify using a 3-column table
rcl <- data.frame(
   from = c(0, 100, 200, 300, 400, 500),
   to = c(100, 200, 300, 400, 500, 600),
   becomes = c(1, 2, 3, 10, 12, 15)
)
threeCol <- classify(elev, rcl = rcl)
threeCol
levels(threeCol)

# Convert all values outside range to NA (default)
rcl <- c(100, 200, 300)
v1 <- classify(elev, rcl = rcl)
v1
plot(v1)

# Convert all values outside range to -1
rcl <- c(100, 200, 300)
v2 <- classify(elev, rcl = rcl, others = -1)
v2
plot(v2)

### Left-open/right-closed (default)
minmax(elev) # note min/max values
rcl <- c(1, 200, 570)
v3 <- classify(elev, rcl = rcl, others = 10)
levels(v3)
plot(v3)

### Left-closed/right-open
minmax(elev) # note min/max values
rcl <- c(1, 200, 570)
v4 <- classify(elev, rcl = rcl, others = 10, right = FALSE)
levels(v4)

# Left-open except for lowest bin/right-closed
minmax(elev) # note min/max values
rcl <- c(1, 200, 570)
v5 <- classify(elev, rcl = rcl, others = 10, include.lowest = TRUE)
v5 <- droplevels(v5)
levels(v5)

# Left-closed/right-open except for highest bin
minmax(elev) # note min/max values
rcl <- c(1, 200, 570)
v6 <- classify(elev, rcl = rcl, others = 10,
   right = FALSE, include.lowest = TRUE)
v6 <- droplevels(v6)
levels(v6)

}

#' Group adjacent cells with similar values
#'
#' @description `clump()` identifies groups of adjacent cells that have the same value or same approximate value, and assigns them a unique number, creating "clumps" of same- or similar-valued cells.
#' 
#' @param x A `GRaster`.
#' 
#' @param minDiff Numeric in the range [0, 1): Minimum difference between cells in order for them to be assigned to the same clump. This is a proportion of the range across all cells. For example, if `minDiff` is set to 0.01, then the maximum difference between cells in a clump can be up to 1% of the entire range across all cells. Small values can create large clumps. The default is 0, in which case values have to be exactly the same.
#' 
#' @param minClumpSize Numeric integer >= 1. Minimum number of cells in a clump. The default is 1.
#' 
#' @param diagonal Logical: If `TRUE` (default), then cells "connected" at corners will be included as part of the same clump.
#' 
#' @returns A `GRaster`.
#' 
#' @example man/examples/ex_clump.r
#' 
#' @aliases clump
#' @rdname clump
#' @exportMethod clump
methods::setMethod(
    f = "clump",
    signature = c(x = "GRaster"),
    function(x, minDiff = 0, minClumpSize = 1, diagonal = TRUE) {

    if (minDiff < 0 | minDiff >= 1) stop("Argument `minDiff` must be >= 0 and < 1.")

    .locationRestore(x)
    .region(x)

    args <- list(
        cmd = "r.clump",
        input = NA_character_,
        output = NA_character_,
        threshold = minDiff,
        minsize = minClumpSize,
        flags = c(.quiet(), "overwrite"),
        intern = TRUE
    )

    if (diagonal) args$flags <- c(args$flags, "d")

    nLayers <- nlyr(x)
    srcs <- .makeSourceName("clump", "raster", nLayers)
    for (i in seq_len(nLayers)) {

        args$input <- sources(x)[i]
        args$output <- srcs[i]

        do.call(rgrass::execGRASS, args=args)
        this <- makeGRaster(srcs[i])

        if (i == 1L) {
            out <- this
        } else {
            out <- c(out, this)
        }

    } # next raster
    out

    } # EOF
)

#' Mask values in a raster
#'
#' @description The output of `mask()` is a `GRaster` that has the same as values as the input raster. However, if the `mask` argument is a `GRaster`, the output will have `NA` values in the same cells that the `mask` raster has `NA` cells. If the `mask` argument is a `GVector`, then the output raster will have `NA` values in cells the `GVector` does not cover.
#'
#' @param x A `GRaster`.
#' 
#' @param mask A `GRaster` or `GVector`.
#' 
#' @param inverse Logical: If `TRUE`, the effect of the mask is inverted. That is, a copy of the input raster is made, but cells that overlap with an `NA` in the mask raster or are not covered by the mask vector retain their values. Cells that overlap with an `NA` in the mask raster or overlap with the mask vector are forced to `NA`.
#' 
#' @param maskvalues Numeric vector, including `NA` (only for when `mask` is a `GRaster`): The value(s) in the mask raster cells that serve as the mask. The default is `NA`, in which case cells in the input raster that overlap with `NA` cells in the mask are forced to `NA`.
#' 
#' @param updatevalue Numeric, including `NA` (default): The values assigned to masked cells.
#' 
#' @returns A `GRaster`.
#' 
#' @seealso [terra::mask()], **GRASS** tool `r.mask` 
#'
#' @example man/examples/ex_mask.r
#' 
#' @aliases mask
#' @rdname mask
#' @exportMethod mask
methods::setMethod(
    f = "mask",
    signature = c(x = "GRaster", mask = "GRaster"),
    function(x, mask, inverse = FALSE, maskvalues = NA, updatevalue = NA) {
    
    .locationRestore(x)
    .region(x)
    xname <- names(x)

    if (nlyr(mask) > 1L) warning("The mask raster has >1 layer. Only the first layer will be used.")

    cats <- cats(x)
    ac <- activeCats(x)

    x <- sources(x)
    mask <- sources(mask)[1L]

    srcs <- .mask(x = x, mask = mask, maskType = "raster", inverse = inverse, maskvalues = maskvalues, updatevalue = updatevalue)
    makeGRaster(srcs, xname, levels = cats, ac = ac)

    } # EOF
)

#' @aliases mask
#' @rdname mask
#' @exportMethod mask
methods::setMethod(
    f = "mask",
    signature = c(x = "GRaster", mask = "GVector"),
    function(x, mask, inverse = FALSE, updatevalue = NA) {

    .locationRestore(x)
    .region(x)
    xname <- names(x)

    cats <- cats(x)
    ac <- activeCats(x)

    x <- sources(x)
    mask <- sources(mask)

    srcs <- .mask(x = x, mask = mask, maskType = "vector", inverse = inverse, maskvalues = NA, updatevalue = updatevalue)
    makeGRaster(srcs, xname, levels = cats, ac = ac)
    
    } # EOF
)

#' @param x [sources()] name of a `GRaster`.
#' @param mask [sources()] name of the mask `GRaster` or `GVector`.
#' @param maskType "raster" or "vector".
#' @param inverse Logical.
#' @param maskvalues Numeric or `NA`.
#' @param updatevalue Numeric or `NA`
#' @noRd
.mask <- function(x, mask, maskType, inverse = FALSE, maskvalues = NA, updatevalue = NA) {

    # remove mask on exit
    on.exit(.removeMask(), add = TRUE)

    ### mask is a raster
    ####################
    if (maskType == "raster") {

        # custom mask values
        if (length(maskvalues) > 1L || !is.na(maskvalues)) {

            maskSrc <- .makeSourceName("mask_r_mapcalc", "raster")

            if (anyNA(maskvalues)) {
                maskValuesHaveNAs <- TRUE
                maskvalues <- maskvalues[!is.na(maskvalues)]
            } else {
                maskValuesHaveNAs <- FALSE
            }

            maskvalues <- as.character(maskvalues)
            maskvalues <- paste(paste0(sources(mask), "=="), maskvalues)
            if (maskValuesHaveNAs) maskvalues <- c(maskvalues, paste0("isnull(", sources(mask), ")"))

            maskvalues <- paste(maskvalues, collapse = " | ")

            ex <- paste0(maskSrc, " = if(", maskvalues, ",1,null())")
            
            args <- list(
                cmd = "r.mapcalc",
                expression = ex,
                flags = c(.quiet(), "overwrite")
            )

            do.call(rgrass::execGRASS, args = args)

        } else {
            maskSrc <- mask
        }

        ### create a mask
        args <- list(
            cmd = "r.mask",
            flags = c(.quiet(), "overwrite")
        )

        args$raster <- maskSrc
    
    ### mask is a vector
    ####################
    } else if (maskType == "vector") {

        args <- list(
            cmd = "r.mask",
            flags = c(.quiet(), "overwrite")
        )
        args$vector <- mask

    }

    if (inverse) args$flags <- c(args$flags, "i")

    ### create mask
    do.call(rgrass::execGRASS, args = args)

    ### copy rasters with mask operative
    ### NOT CORRECT AS OF 2023/07/19: Using g.copy/.copyGSpatial() will not allow the mask to be retained on exiting the function when the mask is automatically removed.
    n <- length(x)
    srcs <- .copyGRaster(x, topo = topology(x), reshapeRegion = FALSE)
    # srcs <- .makeSourceName("mask", "raster", n)

    # for (i in seq_len(n)) {

        # ex <- paste0(srcs[i], " = ", sources(x))
        # rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"), intern = TRUE)

    # }

    ### change masked values
    if (!is.na(updatevalue)) {

		nLayers <- length(x)
        srcsUpdate <- .makeSourceName("mask_r_mapcalc", "raster", nLayers)

        for (i in seq_len(nLayers)) {

            ex <- paste0(srcsUpdate[i], " = if(!isnull(", srcs[i], "), ", updatevalue, ", null())")
            
            rgrass::execGRASS(
                "r.mapcalc",
                expression = ex,
                flags = c(.quiet(), "overwrite")
            )

        } # next raster
        srcs <- srcsUpdate

    } else {
        out <- srcs
    }
    srcs

} # EOF

### remove mask on exit from within a function
#' @noRd
.removeMask <- function() {

    rgrass::execGRASS(
        cmd = "r.mask",
        flags = c(.quiet(), "overwrite", "r")
    )
    
}

#' Name(s) of a GRaster or columns of a GVector's data table
#'
#' `names()` returns that names(s) of a `GRaster` or of columns of a `GVector`'s data table'.
#'
#' `names(value) <-` assigns a new name to the `GRaster` or to the columns of a `GVector`'s data table.
#'
#' @param x A `GRaster` or `GVector`.
#' @param value Character: Name(s) to assign to the raster(s).
#' 
#' @return Character vector.
#' 
#' @seealso [terra::names()]
#'
#' @example man/examples/ex_GRaster.r
#'
#' @aliases names
#' @rdname names
#' @exportMethod names
setMethod(
	f = "names",
	signature = "GRaster",
	definition = function(x) x@names
)

#' @rdname names
#' @aliases names<-
#' @exportMethod names<-
setMethod(
	f = "names<-",
	signature = "GRaster",
	definition = function(x, value) {
		x@names <- value
		methods::validObject(x)
		x
	}
)

#' @aliases names
#' @rdname names
#' @exportMethod names
setMethod(
	f = "names",
	signature = "GVector",
	definition = function(x) {
	
	if (nrow(x@table) > 0L) {
		names(x@table)
	} else {
		NULL
	}
	
	} # EOF
)

#' @rdname names
#' @aliases names<-
#' @exportMethod names<-
setMethod(
	f = "names<-",
	signature = "GVector",
	definition = function(x, value) {
	
	if (ncol(x) > 0L) names(x@table) <- value
	x
	
	} # EOF
)

#' Make unique names
#'
#' @param x A `GRaster`.
#' @returns A `GRaster`
#'
#' @noRd
.makeUniqueNames <- function(x) {

	if (any(duplicated(names(x)))) {
		names(x) <- make.unique(x@names)
	} else {
		x
	}
	x

}

#' Horizon height
#'
#' @description `horizonHeight()` uses a raster representing elevation to calculate the height of the horizon in a particular direction from each cell on a raster. Height is expressed in radians or degrees from the horizontal.
#'
#' @param x A `GRaster`.
#'
#' @param units Character: Units of the height. Either `radians` (default) or `degrees`. Partial matching is used.
#'
#' @param step Numeric integer between 0 and 360, inclusive: Angle step size (in degrees) for calculating horizon height. The direction in which horizon height is calculated is incremented from 0 to 360, with the last value excluded.
#'
#' @param northIs0 Logical: If `TRUE` (default), horizon height calculated in the 0-degree direction will be facing north,  and proceed clockwise So, under "north orientation", 0 is north, 90 east, 180 south, and 270 west. If `FALSE`, angles are in "east orientation", and proceed counterclockwise from east. So, east is 0, north 90, west 180, and south 270. North orientation is the default for this function in **R**, but east orientation is the default in the **GRASS** tool `r.horizon`. **Note:** The [sun()] function requires aspect to be in east orientation.
#'
#' @param bufferZone Numeric >= 0 (default is 0): A buffer of the specified width will be generated around the raster before calculation of horizon angle. If the coordinate system is in longitude/latitude (e.g., WGS84 or NAD83), then this is specified in degrees. Otherwise units are map units (usually meters).
#'
#' @param maxDist Either `NULL` (default) or numeric >= 0: Maximum distance to consider when finding horizon height in meters. If `NULL`, the maximum distance is the full extent of the raster. Smaller values can decrease run time but also reduce accuracy.
#'
#' @param distance Numeric between 0.5 and 1.5, inclusive (default is 1): This determines the step size when searching for the horizon from a given point. The default value of 1 goes cell-by-cell (i.e., search distance step size is one cell width).
#'
#' @returns A `GRaster` with one or more layers. The layers will be named `height_`*xyz*, where *xyz* is degrees from north or from east, depending on whether north or east orientation is used.
#'
#' @seealso **GRASS** manual page for tool `r.horizon` (see `grassHelp("r.horizon")`)
#'
#' @example man/examples/ex_horizonHeight.r
#' 
#' @aliases horizonHeight
#' @rdname horizonHeight
#' @exportMethod horizonHeight
methods::setMethod(
	f = "horizonHeight",
	signature = c(x = "GRaster"),
	definition = function(
		x,
		units = "radians",
		step = 90,
		northIs0 = TRUE,
		bufferZone = 0,
		distance = 1,
		maxDist = NULL
	) {
	
	units <- omnibus::pmatchSafe(units, c("radians", "degrees"))
	
	if (any(!omnibus::is.wholeNumber(step))) {
		step <- round(step)
		warning("Non-integer value used for argument `step`.\n  Value will be rounded (GRASS will do this anyway).")
	}
	
	if (step < 0 | step >= 360) stop("Argument `step` must be in the range [0, 359].")
	
	src <- .makeSourceName("horizonHeight", "raster")
	args <- list(
		cmd = "r.horizon",
		elevation = sources(x),
		output = src,
		step = step,
		start = 0,
		end = 360,
		bufferzone = bufferZone,
		distance = distance,
		flags = c(.quiet(), "overwrite")
	)
	
	# The "c" flag does not seem to work as described in `r.horizon` when using `start`, `end`, and `step` arguments in the call to `r.horizon`. So, we'll calculate horizon height assuming east orientation and if needed, sort the rasters so they match north orientation.
	# if (northIs0) args$flags <- c(args$flags, "c")
	
	if (units == "degrees") args$flags <- c(args$flags, "d")
	if (!is.null(maxDist)) args$maxdistance <- maxDist

	.locationRestore(x)
	.region(x)

	do.call(rgrass::execGRASS, args = args)

	directions <- seq(0, 359.9999999, by = step)
	if (northIs0) directions <- reorient(directions)

	niceDirections <- sprintf("%03.0f", directions)
	srcs <- paste0(src, "_", niceDirections)
	
	if (northIs0) niceDirections <- sort(niceDirections)

	makeGRaster(srcs, names = paste0("horizonHeight_", niceDirections, "deg"))

	} # EOF
)

#' Geographic statistics for sets of cells with the same values
#'
#' @description This function calculates geographic statistics for each set of cells in an `integer` or `factor` `GRaster`. Statistics include:
#' * Area
#' * Perimeter length
#' * "Compact square" statistic: \eqn{4 \sqrt(area) / perimeter)}
#' * "Compact circle" statistic: \eqn{4 * P / ( 2 \sqrt(\pi * A))} where *P* is the perimeter length and *A* the area.
#' * fractal dimension: \eqn{2 ( log(P) / log(A + 0.001))} where *P* is perimeter length and *A* is area.
#' * The average x- and y-coordinates of each zone.
#'
#' @param x A `GRaster`.
#' 
#' @param unit Character: Units of the output. Any of:
#' * `"meters"` (default)
#' * `"kilometers"` or `"km"`
#' * `"miles"` or `"mi"`
#' * `"yards"` or `"yd"`
#' * `"feet"` or `"ft"`: International foot; 1 foot exactly equal to 0.3048 meters
#' * `"cells"`: Number or cells
#'
#' Partial matching is used and case is ignored.
#'
#' @returns A list of `data.frame`s or a `data.table`s, one per layer in `x`. Only layers that are integers or factors have their geographies calculated. Other layers have `NULL` tables returned.
#'
#' @example man/examples/ex_zonalGeog.r
#'
#' @aliases zonalGeog
#' @rdname zonalGeog
#' @exportMethod zonalGeog
methods::setMethod(
	f = "zonalGeog",
	signature(x = "GRaster"),
	function(x, unit = "meters") {
	
	.locationRestore(x)
	.region(x)

	units <- c("m", "meters", "kilometers", "km", "miles", "feet", "ft", "yards", "yd", "cells")
	unit <- omnibus::pmatchSafe(unit, units, useFirst = TRUE, nmax = 1L)
	unit <- omnibus::expandUnits(unit)

	flags <- if (unit != "cells") {
		c(.quiet(), "m")
	} else {
		.quiet()
	}

	isFact <- is.factor(x)
	isInt <- is.int(x)

	nLayers <- nlyr(x)
	out <- list()
	for (i in seq_len(nLayers)) {

		if (isInt[i] | isFact[i]) {

			args <- list(
				cmd = "r.object.geometry",
				input = sources(x)[i],
				intern = TRUE
			)

			if (!is.null(flags)) args$flags <- flags
			info <- do.call(rgrass::execGRASS, args = args)

			info <- strsplit(info, split = "\\|")
			n <- length(info) - 1L
			if (isFact[i]) {

				xDrop <- x[[i]]
				xDrop <- droplevels(xDrop)
				labels <- levels(xDrop)[[1L]]
				catName <- names(labels)[activeCat(xDrop) + 1L]
				labels <- labels[[catName]]
				thisOut <- data.table::data.table(
					value = rep(NA_integer_, n),
					label = labels,
					area = rep(NA_real_, n),
					perimeter = rep(NA_real_, n),
					compactSquare = rep(NA_real_, n),
					compactCircle = rep(NA_real_, n),
					fractalDimension = rep(NA_real_, n),
					meanX = rep(NA_real_, n),
					meanY = rep(NA_real_, n)
				)

			# integer raster
			} else {
			
				thisOut <- data.table::data.table(
					value = rep(NA_integer_, n),
					area = rep(NA_real_, n),
					perimeter = rep(NA_real_, n),
					compactSquare = rep(NA_real_, n),
					compactCircle = rep(NA_real_, n),
					fractalDimension = rep(NA_real_, n),
					meanX = rep(NA_real_, n),
					meanY = rep(NA_real_, n)
				)

			} # pre-assign data table

			for (j in 2L:length(info)) {

				val <- info[[j]][1L]
				val <- as.integer(val)

				vals <- info[[j]]
				vals <- as.numeric(vals)

				index <- j - 1L
				thisOut$value[index] <- val
				thisOut$area[index] = vals[2L]
				thisOut$perimeter[index] = vals[3L]
				thisOut$compactSquare[index] = vals[4L]
				thisOut$compactCircle[index] = vals[5L]
				thisOut$fractalDimension[index] = vals[6L]
				thisOut$meanX[index] = vals[7L]
				thisOut$meanY[index] = vals[8L]

			}

			area <- perimeter <- NULL
			if (unit == "kilometers") {
				f <- omnibus::convertUnits(from = "meters2", to = "kilometers2")
				thisOut[ , area := area * f]
				f <- omnibus::convertUnits(from = "meters", to = "kilometers")
				thisOut[ , perimeter := perimeter * f]
			} else if (unit == "miles") {
				f <- omnibus::convertUnits(from = "meters2", to = "miles2")
				thisOut[ , area := area * f]
				f <- omnibus::convertUnits(from = "meters", to = "miles")
				thisOut[ , perimeter := perimeter * f]
			} else if (unit == "feet") {
				f <- omnibus::convertUnits(from = "meters2", to = "feet2")
				thisOut[ , area := area * f]
				f <- omnibus::convertUnits(from = "meters", to = "feet")
				thisOut[ , perimeter := perimeter * f]
			} else if (unit == "yards") {
				f <- omnibus::convertUnits(from = "meters2", to = "yards2")
				thisOut[ , area := area * f]
				f <- omnibus::convertUnits(from = "meters", to = "yards")
				thisOut[ , perimeter := perimeter * f]
			}

		# if raster is NOT integer or factor
		} else {
			thisOut <- data.table::data.table(NULL)
		}

		out[[i]] <- thisOut
	
	} # next layer

	names(out) <- names(x)
	if (!faster("useDataTable")) {
		for (i in seq_along(out)) out[[i]] <- as.data.frame(out)
	}
	out

	} # EOF
)

#' Convert a GVector to a GRaster
#'
#' @description The `rasterize()` function converts a `GVector` into a `GRaster`.
#'
#' @param x A `GVector`.
#'
#' @param y A `GRaster`: The new raster will have the same extent and resolution as this raster.
#'
#' @param field Character: Name of a column in the data table of `y` to "burn" into the raster. If not `""` (default), then the output will be a categorical `GRraster`. If `field` is `""`, then all geometries will be "burned" to the raster and have the same value.
#'
#' @param background Numeric or `NA` (default): Value to put in cells that are not covered by the `GVector`. Note that if this is not `NA` and not an integer, then the output cannot be a categorical raster (i.e., there will be no "levels" table associated with it).
#'
#' @param by Either `NULL` (default) or character: If this is not `NULL`, then the `GVector` will be subset by the values in the field named by `by`. The output will be a multi-layer raster, with one layer per unique value in `by`.
#'
#' @param verbose Logical: If `by` is not `NULL`, display progress.
#'
#' @returns A `GRaster`.
#'
#' @seealso [terra::rasterize()], **GRASS** tool `v.to.rast` (see `grassHelp("v.to.rast")`)
#'
#' @example man/examples/ex_rasterize.r
#'
#' @aliases rasterize
#' @rdname rasterize
#' @exportMethod rasterize
methods::setMethod(
	f = "rasterize",
	signature = c(x = "GVector", y = "GRaster"),
	function(
		x,
		y,
		field = "",
		background = NA,
		by = NULL,
		verbose = TRUE
	) {
	
	compareGeom(x, y)
	.locationRestore(x)
	.region(y)

	gtype <- geomtype(x, grass = TRUE)
	out <- .rasterize(x = x, y = y, field = field, background = background, by = by, gtype = gtype, verbose = verbose)
	makeGRaster(out$src, levels = out$levels)

	} # EOF
)

#' @param x [sources()] name of a `GVector`.
#' @param y [sources()] name of a `GRaster`
#' @param field Name of a field in the data table of `x` (can be `""`).
#' @param background Numeric or `NA`.
#' @param by `NULL` or name of a field in the data table of `x`.
#' @param verbose Logical.
#' @param gtype `geomtype(..., grass = TRUE)` of `x` ("area", "line", "point")
#'
#' @returns A `list` with the [sources()] name of the output raster, plus a `levels` table (can be `NULL`).
#'
#' @noRd
.rasterize <- function(x, y, field, background, by, gtype, verbose) {

	### create different raster layer for each geometry
	if (!is.null(by)) {

		bys <- unique(x@table[[by]])
		nBys <- length(bys)
		src <- rep(NA_character_, nBys)
		levels <- list()
		if ((verbose | faster("verbose")) & nBys > 1L) pb <- utils::txtProgressBar(min = 0, max = nBys, initial = 0, style = 3, width = 30)

		for (i in seq_len(nBys)) {

			if (verbose | faster("verbose")) {
				utils::setTxtProgressBar(pb, i)
				utils::flush.console()
			}

			index <- which(x@table[[by]] == bys[i])
			xx <- x[index]

			thisOut <- .rasterize(xx, y, gtype = gtype, field = field, background = background, by = NULL, verbose = FALSE)
			src[i] <- thisOut$src
			levels[[i]] <- thisOut$levels
		
		}
		if (verbose | faster("verbose")) close(pb)

	} else {

		src <- .makeSourceName("rasterize_v_to_rast", "raster")

		# if by geometry but burned to the same raster
		if (field == "") {

			levels <- data.table::data.table(NULL)

			args <- list(
				cmd = "v.to.rast",
				input = sources(x),
				output = src,
				use = "val",
				value = 1,
				type = gtype,
				memory = faster("memory"),
				flags = c(.quiet(), "overwrite")
			)
		
		} else {
			
			# rasterize by field in data table
			cats <- .vCats(x)
			uniCats <- sort(unique(cats))
			fieldVals <- x@table[[field]]
			renums <- omnibus::renumSeq(fieldVals)
			levels <- data.table::data.table(value = unique(renums), DUMMYDUMMY_ = unique(fieldVals))
			names(levels)[2L] <- field
			levels <- levels[order(levels$value)]
			db <- data.table::data.table(cat = uniCats, DUMMYDUMMY_ = renums)
			.vAttachDatabase(x, db, cats = cats)

			args <- list(
				cmd = "v.to.rast",
				input = sources(x),
				output = src,
				use = "attr",
				attribute_column = "DUMMYDUMMY_",
				type = gtype,
				memory = faster("memory"),
				flags = c(.quiet(), "overwrite")
			)

		}

		if (gtype == "line") args$flags <- c(args$flags, "d")
		do.call(rgrass::execGRASS, args = args)

		if (!is.na(background)) {

			srcIn <- src
			src <- .makeSourceName("rasterize_r_mapcalc", "vector")
			ex <- paste0(src, " = if(isnull(", srcIn, "), ", background, ", ", srcIn, ")")

			rgrass::execGRASS("r.mapcalc", expression  = ex, flags = c(.quiet(), "overwrite"))
			.rm(srcIn, type = "raster", warn = FALSE)

			if (nrow(levels) > 0L) {
			
				if (is.integer(background) | omnibus::is.wholeNumber(background)) {

					background <- as.integer(background)

					levAdd <- data.table::data.table(value = background, DUMMYDUMMY_ = "background")
					names(levAdd) <- names(levels)
					levels <- rbind(levels, levAdd)
					levels <- levels[!duplicated(levels)]
					levels <- levels[order(levels$value)]

				} else {
					levels <- data.table::data.table(NULL)
				}
				
			}

		}

	}

	list(src = src, levels = levels)

}

#' Get one of the example rasters or spatial vectors
#'
#' This function is a simple way to get example rasters or spatial vector datasets that come with **fasterRaster**.
#'
#' @param x The name of the raster or spatial vector to get. All of these represent a portion of the eastern coast of Madagascar.
#'
#' Spatial vectors (objects of class `sf` from the **sf** package):
#' * [madCoast0][madCoast0]: Outline of the region (polygon)
#' * [madCoast4][madCoast4]: Outlines of the Fokontanies (Communes) of the region (polygons)
#' * [madDypsis][madDypsis]: Records of plants of the genus *Dypsis* (points)
#' * [madRivers][madRivers]: Major rivers (lines)
#'
#' Rasters (objects of class `SpatRaster` from the **terra** package, saved as GeoTIFF files):
#' * [madChelsa][madChelsa]: Bioclimatic variables
#' * [madCover][madCover]: Land cover
#' * [madElev][madElev]: Elevation
#' * [madForest2000][madForest2000]: Forest cover in year 2000
#' * [madForest2014][madForest2014]: Forest cover in year 2014
#' * [madLANDSAT][madLANDSAT]: Surface reflectance in 2023
#' * [madPpt][madPpt], [madTmin][madTmin], [madTmax][madTmax]: Rasters of mean monthly precipitation, and minimum and maximum temperature.
#'
#' Data frames
#' * [appFunsTable][appFunsTable]: Table of functions usable by [app()]
#' * [madCoverCats][madCoverCats]: Land cover values and categories for `madCover`
#' * [vegIndices][vegIndices]: Vegetation indices that can be calculated with [vegIndex()]
#'
#' @return A `SpatRaster`, `sf` spatial vector, or a `data.frame`.
#'
#' @example man/examples/ex_fastData.r
#'
#' @export
fastData <- function(x) {

	vectors <- c("madCoast0", "madCoast4", "madDypsis", "madRivers")
	tables <- c("appFunsTable", "madCoverCats", "vegIndices")
	rasters <- c("madChelsa", "madCover", "madElev", "madLANDSAT", "madForest2000", "madForest2014", "madPpt", "madTmin", "madTmax")

	if (!inherits(x, "character")) {
		stop("Please supply the name of an example raster or spatial vector in fasterRaster.")
	} else {
		if (x %in% c(vectors, tables)) {

			madCoast0 <- madCoast4 <- madDypsis <- madRivers <- NULL
			madCoverCats <- NULL

			out <- do.call(utils::data, list(x, envir = environment(), package = "fasterRaster"))
			out <- get(out)

		} else if (x %in% rasters) {
		
			rastFile <- system.file("extdata", paste0(x, ".tif"), package = "fasterRaster")
			out <- terra::rast(rastFile)

			if (x == "madCover") {

			    levs <- system.file("extdata", paste0("madCover.csv"), package = "fasterRaster")
				levs <- utils::read.csv(levs)
				levels(out) <- levs

			}

		} else {
			stop("Please supply the name of a data object available in fasterRaster.")
		}
	}

	out

}

#' @name appFunsTable
#'
#' @title Functions that can be used in app()
#'
#' @description This is a table of functions that can be used in the [app()] function, their **R** equivalents, and the [datatype()] they return. You can view this table using `?appFunsTable` or as a searchable, sortable **Shiny** table using [appFuns()].
#'
#' @docType data
#'
#' @format A `data.frame`.
#'
#' @keywords app
#'
#' @source [OSGeo](https://grass.osgeo.org)
#'
#' @example man/examples/ex_app.r
#'
NULL

#' Mask all non-NA cells or all NA cells
#'
#' @description This function converts all non-`NA` cells in a `GRaster` to a single user-defined value, leaving `NA` cells as `NA`. Alternatively, it can convert `NA` cells to a user-defined value, and all non-`NA` cells to `NA.`
#'
#' @param x A `GRaster`.
#' @param value Numeric: Value to which to assign to masked cells. The default is 1.
#' @param invert Logical: If `FALSE` (default), convert non-`NA` cells to `value`, and leave `NA` cells as-is. If `TRUE`, convert all `NA` cells to `value`, and non-`NA` cells to `NA`.
#' @param retain Logical: If `invert` is `TRUE` and `retain` is `FALSE` (default), non-`NA` cells will retain their value. This argument is ignored if `invert` is `FALSE`.
#'
#' @seealso [not.na()], [app()], [mask()]
#'
#' @example man/examples/ex_maskNA.r
#'
#' @returns A `GRaster`.
#'
#' @aliases maskNA
#' @rdname maskNA
#' @exportMethod maskNA
methods::setMethod(
	f = "maskNA",
	signature = c(x = "GRaster"),
	function(x, value = 1, invert = FALSE, retain = FALSE) {
	
	.locationRestore(x)
	.region(x)
	
	nLayers <- nlyr(x)
	srcs <- .makeSourceName("maskNA_r_mapcalc", "raster", n = nLayers)
	
	for (i in seq_len(nLayers)) {

		if (!invert) {
			ex <- paste0(srcs[i], " = if(!isnull(", sources(x)[i], "), ", value, ", null())")
		} else {
			ret <- if (retain) { sources(x)[i] } else { "null()"}
			ex <- paste0(srcs[i], " = if(isnull(", sources(x)[i], "), ", value, ", ", ret, ")")
		}
		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

	}
	makeGRaster(srcs, names = names(x))
	
	} # EOF
)

#' Setup fasterRaster for ABS
#'
#' This is a secret function to be used for faster development of **fasterRaster**. It calls [faster()] to set the install directory for **GRASS**, increases default memory, and number of cores. The function assumes development is on a Windows machine.
#'
#' @param start Logical: If `TRUE`, start the **GRASS** session by creating the `madElev` `GRaster`.
#' @param ver Character: **GRASS**: e.g., "83" or "84".
#'
#' @returns `TRUE` (invisibly).
#'
#' @keywords internal
.backdoor <- function(start = FALSE, ver = "84") {
	
	verNice <- paste0(substr(ver, 1L, 1L), ".", substr(ver, 2L, 2L))
	
	faster(
		grassDir = paste0("C:/Program Files/GRASS GIS ", verNice),
		addonsDir = "C:/Users/adame/AppData/Roaming/GRASS8/addons",
		memory = 1024 * 8,
		cores = 2,
		useDataTable = TRUE,
		verbose = TRUE
	)

	if (start) {
		madElev <- fastData("madElev")
		elev <- fast(madElev)
		out <- elev
	} else {
		out <- TRUE
	}

	invisible(out)
}

#' Convert a raster to a lines vector
#'
#' @description [as.lines()] converts a `GRaster` to a "lines" `GVector`. Before you apply this function, you may need to run [thinLines()] on the raster to reduce linear features to a single-cell width. You may also need to use [clean geometry][breakPolys] (especially the [removeDups()] and [removeDangles()]) afterward to remove duplicated vertices and "dangling" lines.
#' 
#' @param x A `GRaster`. If more than one layer is in the `GRaster`, only the first will be used (with a warning).
#'
#' @returns A `GVector`.
#' 
#' @seealso [as.points()], [as.polygons()], [terra::as.lines()], [thinLines()], [geometry cleaning][breakPolys], and **GRASS** tool `r.to.vect`
#' 
#' @example man/examples/ex_asLines.r
#' 
#' @aliases as.lines
#' @rdname as.lines
#' @exportMethod as.lines
methods::setMethod(
    f = "as.lines",
    signature = c(x = "GRaster"),
    function(x) {

    if (nlyr(x) > 1L) warning("The raster has >1 layer. Only the first will be used.")
    x <- x[[1L]]

    .locationRestore(x)
    .region(x)

    src <- .makeSourceName("asLines", "raster")
    args <- list(
        cmd = "r.to.vect",
        input = sources(x),
        output = src,
        type = "line",
        flags = c(.quiet(), "overwrite"),
        intern = TRUE
    )
    do.call(rgrass::execGRASS, args = args)
    makeGVector(src)

    } # EOF
)

#' @name madLANDSAT
#'
#' @title Rasters of surface reflectance for an eastern portion of Madagascar
#'
#' @description Raster layers of surface reflectance from LANDSAT 9 for an eastern portion of Madagascar taken May 21, 2023. Four bands are represented:
#' * `band2': Blue (450-510 nm)
#' * `band3': Green (530-590 nm)
#' * `band4': Red (640-670 nm)
#' * `band5': Near-infrared (850-880 nm)
#' The rasters have been resampled to 90-m resolution to reduce their size, then rescaled to integers in the range 0 to 255.
#'
#' @docType data
#'
#' @format An object of class `SpatRaster` in Universal Trans-Mercator (UTM), Zone 39 North with a WGS84 coordinate system, at 90 m resolution.
#'
#' @keywords land Madagascar
#'
#' @source United States Geological Survey's [EarthExplorer](https://earthexplorer.usgs.gov). Also see [band definitions](https://www.usgs.gov/faqs/what-are-band-designations-landsat-satellites).
#'
#' @example man/examples/ex_fastData.r
#'
NULL

if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Example data
madElev <- fastData("madElev") # integer raster
madCover <- fastData("madCover") # categorical raster
madCoast4 <- fastData("madCoast4") # polygons vector
madRivers <- fastData("madRivers") # lines vector
madDypsis <- fastData("madDypsis") # points vector

### Create GRasters from SpatRasters

# Create an integer raster:
elev <- fast(madElev)
elev

# Create a categorical raster:
cover <- fast(madCover)
madCover
levels(madCover) # category levels

# Create a GRaster from a file on disk:
rastFile <- system.file("extdata", "madForest2000.tif", package = "fasterRaster")
forest2000 <- fast(rastFile)
forest2000

# Create a 1's raster that spans the world:
ones <- fast(rastOrVect = "raster", crs = "epsg:4326")
ones

### Create GVectors

# Create a GVector from an sf vector:
coast4 <- fast(madCoast4)
coast4

# Create a GVector from a SpatVector:
madRivers <- vect(madRivers)
class(madRivers)
rivers <- fast(madRivers)
rivers

# Create a GVector from a vector on disk:
vectFile <- system.file("extdata/shapes", "madCoast.shp",
   package = "fasterRaster")
coast0 <- fast(vectFile)
coast0

# Import only Dypsis occurrences in a restricted area:
ant <- coast4[coast4$NAME_4 == "Antanambe"]
dypsisRestrict <- fast(madDypsis, extent = ant)
dypsis <- fast(madDypsis)

plot(coast4)
plot(ant, col = "gray80", add = TRUE)
plot(dypsis, add = TRUE)
plot(dypsisRestrict, col = "red", add = TRUE)

# Create a generic GVector that spans the world:
wallToWall <- fast(rastOrVect = "vector", crs = "epsg:4326") # WGS84
wallToWall

# Create a GVector from a numeric vector
pts <- c(-90.2, 38.6, -122.3, 37.9)
pts <- fast(pts, crs = "epsg:4326") # WGS84

# Create a GVector from a matrix (can also use data.frame or data.table):
mat <- matrix(c(-90.2, 38.6, -122.3, 37.9), ncol = 2, byrow = TRUE)
mat <- fast(mat, crs = "epsg:4326", keepgeom = TRUE) # WGS84

}

#' Add rows and columns around a writeRaster
#'
#' @description `extend()` adds cells around a raster, making it larger.
#' 
#' @param x A `GRaster`.
#' 
#' @param y Any of:
#' *  An object from which an extent can be obtained; i.e., a `SpatRaster`, `SpatVector`, `SpatExtent`, `sf` vector, or a `GSpatial` object (any of `GRaster`, `GVector`, or `GRegion`). If the extent of `x` is "outside" the extent of `y` on any side, the side(s) of `x` that are outside will be kept as-is (i.e., the extent of `x` will never be shrunk).
#' * A single positive integer: Number of rows and columns to add to the top, bottom, and sides of the raster.
#' * Two integers >= 0: Number of columns (1st value) to add to the sides, and number of rows (2nd value) to add to the top and bottom of the raster.
#' * Four integers >= 0: Number of rows and columns to add (left column, right column, bottom row, top row).
#' 
#' @param snap Character: Method used to align `y` to `x`. Partial matching is used. This is only used if `y` is not a set of numbers.
#' * `"near"` (default): Round to nearest row/column
#' * `"in"`: Round "inward" toward the extent of `x` to nearest row/column
#' * `"out"`: Round "outward" away from the extent of `x` to the nearest row/column. 
#' 
#' @param fill Numeric: Value to place in the new cells. The default is `NA`.
#' 
#' @returns A `GRaster`.
#' 
#' @details Known issues: When `GRaster`s are saved to disk explicitly using [writeRaster()], or implicitly using [rast()] or [plot()], rows and columns that are entirely `NA` are dropped.
#' 
#' @seealso [terra::extend()]
#'
#' @example man/examples/ex_extend.r
#' 
#' @aliases extend
#' @rdname extend
#' @exportMethod extend
methods::setMethod(
    f = "extend",
    signature = c(x = "GRaster", y = "numeric"),
    function(x, y, fill = NA) .extend(x = x, y = y, snap = NA, fill = fill)
)

#' @aliases extend
#' @rdname extend
#' @exportMethod extend
methods::setMethod(
    f = "extend",
    signature = c(x = "GRaster", y = "SpatRaster"),
    function(x, y, snap = "near", fill = NA) .extend(x = x, y = y, snap = snap, fill = fill)
)

#' @aliases extend
#' @rdname extend
#' @exportMethod extend
methods::setMethod(
    f = "extend",
    signature = c(x = "GRaster", y = "SpatVector"),
    function(x, y, snap = "near", fill = NA) .extend(x = x, y = y, snap = snap, fill = fill)
)

#' @aliases extend
#' @rdname extend
#' @exportMethod extend
methods::setMethod(
    f = "extend",
    signature = c(x = "GRaster", y = "SpatExtent"),
    function(x, y, snap = "near", fill = NA) .extend(x = x, y = y, snap = snap, fill = fill)
)

#' @aliases extend
#' @rdname extend
#' @exportMethod extend
methods::setMethod(
    f = "extend",
    signature = c(x = "GRaster", y = "sf"),
    function(x, y, snap = "near", fill = NA) .extend(x = x, y = y, snap = snap, fill = fill)
)

#' @aliases extend
#' @rdname extend
#' @exportMethod extend
methods::setMethod(
    f = "extend",
    signature = c(x = "GRaster", y = "GSpatial"),
    function(x, y, snap = "near", fill = NA) .extend(x = x, y = y, snap = snap, fill = fill)
)

#' @noRd
.extend <- function(x, y, snap, fill) {

    # resolution of x
    ewres <- xres(x)
    nsres <- yres(x)

    ### find coordinates of new extent
    if (inherits(y, "numeric")) {

        if (!(length(y) %in% c(1L, 2L, 4L))) {
            stop("Argument `y` is invalid.")
        } else if (length(y) == 1L) {
            y <- rep(y, 4L)
        } else if (length(y) == 2L) {
            y <- rep(y, each = 2L)
        }
        
        if (any(!omnibus::is.wholeNumber(y))) stop("Values of `y` must be numeric integers.")
        if (any(y < 0L)) y[y < 0L] <- 0L

    } else {

        if (inherits(y, c("SpatRaster", "SpatVector", "SpatExtent"))) {
            extent <- terra::ext(y)
            extent <- as.vector(extent)
        } else if (inherits(y, "GSpatial")) {
            extent <- ext(y, vector = TRUE)
        } else {
            stop("Argument `y` is invalid.")
        }

        # by how many rows and columns do we grow?
        wgrow <- max(0, W(x) - extent[1L]) / ewres
        egrow <- max(0, extent[2L] - E(x)) / ewres
        sgrow <- max(0, S(x) - extent[3L]) / nsres
        ngrow <- max(0, extent[4L] - N(x)) / nsres

        snap <- omnibus::pmatchSafe(snap, c("near", "in", "out"))
        fx <- if (snap == "near") {
            round
        } else if (snap == "in") {
            floor
        } else if (snap == "out") {
            ceiling
        }

        grows <- c(wgrow, egrow, sgrow, ngrow)
        y <- do.call(fx, list(grows))

    } # if y is an object with an extent

    # new extent coordinates
    w <- W(x) - y[1L] * ewres
    e <- E(x) + y[2L] * ewres
    s <- S(x) - y[3L] * nsres
    n <- N(x) + y[4L] * nsres

    w <- as.character(w)
    e <- as.character(e)
    s <- as.character(s)
    n <- as.character(n)

    ewres <- as.character(ewres)
    nsres <- as.character(nsres)

    # reset region
    rgrass::execGRASS(
        cmd = "g.region",
        w = w, e = e, s = s, n = n,
        ewres = ewres, nsres = nsres,
        flags = c(.quiet(), "overwrite")
    )

    reg <- .region()
    nc <- ncol(reg)
    nr <- nrow(reg)

    ### extent
    nLayers <- nlyr(x)
    srcs <- rep(NA_character_, nLayers)
    for (i in seq_len(nLayers)) {

        if (is.na(fill)) {
            src <- .copyGSpatial(x = x[[i]], reshapeRegion = FALSE)
        } else {
            src <- .makeSourceName("extend", "raster", nLayers)

            ex <- paste0(src, " = if(col() <= ", y[1L], ", ", fill, ", if(col() > ncols() - ", y[2L], ", ", fill, ", if(row() <= ", y[4L], ", ", fill, ", if(row() > nrows() - ", y[3], ", ", fill, ", ", sources(x)[i], "))))")

            args <- list(
                cmd = "r.mapcalc",
                expression = ex,
                region = "current",
                flags = c(.quiet(), "overwrite")
            )
            do.call(rgrass::execGRASS, args = args)
        }

        srcs[i] <- src
        
    } # next raster
    
    makeGRaster(srcs, names = names(x), levels = levels(x), ac = activeCats(x))

} # EOF

if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster
elev <- fast(madElev)

# Make a stack of various versions of "elev" from which to select from:
x <- c(elev, 10 * elev, ln(elev), -1 * elev)
x

# Make a layer with random numbers between 1 and 4:
fun <- "= round(rand(0.5, 4.5))"
y <- app(elev, fun = fun)

selected <- selectRange(x, y)

}

if (grassStarted()) {

# Setup
madForest2000 <- fastData("madForest2000")
madForest2014 <- fastData("madForest2014")

# Convert SpatRasters to GRasters:
forest2000 <- fast(madForest2000)
forest2014 <- fast(madForest2014)

# Combine:
forest <- c(forest2000, forest2014)
forest

nlyr(forest)

}

#' @name madTmin
#'
#' @title Rasters of average monthly minimum temperature for an eastern portion of Madagascar
#'
#' @description Rasters of minimum temperature for an eastern portion of Madagascar from WorldClim 2.1 at ~3.33 arcminute resolution projected to the Tananarive (Paris)/Laborde Grid coordinate reference system. Values represent monthly averages across 1970-2000. Units are in degrees C. These should not be used for formal analysis.
#'
#' @docType data
#'
#' @format An object of class `SpatRaster`.
#'
#' @keywords climate Madagascar
#'
#' @references Fick, S.E. and Hijmans, R.J.  2017.  WorldClim 2: New 1-km spatial resolution climate surfaces for global land areas.  International Journal of Climatology 37:4302-4315. \doi{10.1002/joc.5086}
#'
#' @source \doi{https://dx.doi.org/10.1002/joc.5086}
#'
#' @example man/examples/ex_fastData.r
#'
NULL

#' Create a randomly-positioned tesselation
#'
#' @description This function partitions a region into Voronoi polygons that completely overlap it. Each polygon has a random center. The function is essentially a wrapper for [spatSample()] and [voronoi()].
#'
#' @param x A `GRaster` or `GVector` used to constrain the location of random points used to create the tesselation.
#'
#' @param size Numeric integer or integer: Number of polygons.
#' 
#' @param seed Numeric integer, integer, or `NULL` (default): Value used as a random seed. If `NULL`, a random seed will be generated by **GRASS**.
#'
#' @returns A `GVector`.
#'
#' @example man/examples/ex_delaunay_voronoi.r
#' 
#' @aliases rvoronoi
#' @rdname rvoronoi
#' @exportMethod rvoronoi
methods::setMethod(
	f = "rvoronoi",
	signature = c(x = "GRaster"),
	function(x, size = 100, seed = NULL) .rvoronoi(x = x, size = size, seed = seed)
)

#' @aliases rvoronoi
#' @rdname rvoronoi
#' @exportMethod rvoronoi
methods::setMethod(
	f = "rvoronoi",
	signature = c(x = "GVector"),
	function(x, size = 100, seed = NULL) .rvoronoi(x = x, size = size, seed = seed)
)

#' @noRd
.rvoronoi <- function(x, size, seed) {
	
	rp <- spatSample(x = x, size = size, as.points = TRUE, xy = FALSE, values = FALSE, seed = seed)

	out <- voronoi(rp, buffer = 0)
	out

}

if (grassStarted()) {

# Setup
library(terra)

# Elevation
madElev <- fastData("madElev")

# Convert SpatRaster to GRaster:
elev <- fast(madElev)

# To speed things up, first group cells of similar value:
elevClumps <- clump(elev, minDiff = 0.0115)

# Convert to polygons:
rastToPolys <- as.polygons(elevClumps)
plot(rastToPolys)

}

#' @name madForest2014
#'
#' @title Forest cover in year 2014 for a portion of Madagascar
#'
#' @description Raster of occurrence/non-occurrence of forest cover in a portion of Madagascar. Cells are 30-m in resolution. Values represent forest (1) or non-forest (`NA`).
#'
#' @docType data
#'
#' @format An object of class `SpatRaster`..
#'
#' @keywords Madagascar
#'
#' @references Vielledent, G., Grinand, C., Rakotomala, F.A., Ranaivosoa, \pkg{R}., Rakotoarijaona, J-R., Allnutt, T.F., and Achard, F.  2018.  Combining global tree cover loss data with historical national forest cover maps to look at six decades of deforestation and forest fragmentation in Madagascar. *Biological Conservation* 222:189-197. \doi{10.1016/j.biocon.2018.04.008}.
#'
#' @example man/examples/ex_fastData.r
#'
NULL

#' Delaunay triangulation for points
#'
#' @description This function creates a Delaunay triangulation from a "points" `GVector`.
#'
#' @param x A `GVector` "points" object.
#'
#' @returns A `GVector`.
#'
#' @seealso [terra::delaunay()], tool `v.delaunay` in **GRASS**
#'
#' @example man/examples/ex_delaunay_voronoi.r
#'
#' @aliases delaunay
#' @rdname delaunay
#' @exportMethod delaunay
methods::setMethod(
	f = "delaunay",
	signature = c(x = "GVector"),
	definition = function(x) {

	if (geomtype(x) != "points") stop("The vector must represent spatial points.")
	
	src <- .makeSourceName("v_delaunay", "vect")
	args <- list(
		cmd = "v.delaunay",
		input = sources(x),
		output = src,
		flags = c(.quiet(), "overwrite")
	)

	do.call(rgrass::execGRASS, args = args)
	makeGVector(src)
	
	} # EOF
)

#' Spatial bounds of a GRaster or GVector
#'
#' @description These functions return the extent of a `GSpatial` object (`GRegions`, `GRaster`s, and `GVector`s):
#' 
#' * `ext()`: 2-dimensional spatial extent (i.e., westernmost/easternmost and southernmost/northernmost coordinates of area represented).\cr
#' * `zext()`: Vertical extent (i.e., topmost and bottom-most elevation of the volume represented). The vertical extent is not `NA` only if the object is 3-dimensional.\cr
#' * `W()`, `E()`, `N()`, `S()`: Coordinates of one side of horizontal extent.\cr
#' * `top()` and `bottom()`: Coordinates of top and bottom of vertical extent.\cr
#' 
#' @param x An object that inherits from `GSpatial` (i.e., a `GRaster` or `GVector`) or missing. If missing, then the horizontal or vertical extent of the currently active "region" is returned (see `vignette("regions", package = "fasterRaster")`).
#' 
#' @param vector Logical: If `FALSE` (default), return a `SpatExtent` object. If `TRUE`, return the extent as a named vector.
#' 
#' @param char Logical: If `FALSE` (default), return a numeric value. If `TRUE`, return as a character.
#' 
#' @returns The returned values depend on the function:
#' * `ext()`: A `SpatExtent` object (**terra** package) or a numeric vector.
#' * `zext()`: A numeric vector.
#' * `W()`, `E()`, `N()`, `S()`, `top()`, and `bottom()`: A numeric value or character.
#'
#' @seealso [terra::ext()], [sf::st_bbox()]
#' 
#' @example man/examples/ex_GRaster_GVector.r
#'
#' @aliases ext
#' @rdname ext
#' @exportMethod ext
methods::setMethod(
	f = "ext",
	signature = "missing",
	definition = function(x, vector = FALSE) ext(.region(), vector = vector)
)

#' @aliases ext
#' @rdname ext
#' @exportMethod ext
methods::setMethod(
	f = "ext",
	signature = "GSpatial",
	definition = function(x, vector = FALSE) {

	out <- c(x@extent[1L], x@extent[2L], x@extent[3L], x@extent[4L])
	names(out) <- c("xmin", "xmax", "ymin", "ymax")
	if (!vector) out <- terra::ext(out)
	out

	} # EOF
)

#' @aliases zext
#' @rdname ext
#' @exportMethod zext
methods::setMethod(
	f = "zext",
	signature = "missing",
 	definition = function(x) zext(.region())
)

#' @aliases zext
#' @rdname ext
#' @exportMethod zext
methods::setMethod(
	f = "zext",
	signature = "GSpatial",
	definition = function(x) {
	x@zextent
})

#' @aliases W
#' @rdname ext
#' @exportMethod W
methods::setMethod(
	f = "W",
	signature = "missing",
	definition = function(x, char = FALSE) {
	out <- unname(ext(.region(), vector = TRUE)["xmin"])
	if (char) out <- as.character(out)
	out
})

#' @aliases W
#' @rdname ext
#' @exportMethod W
methods::setMethod(
	f = "W",
	signature = "GSpatial",
	definition = function(x, char = FALSE) {
	out <- unname(ext(x, vector = TRUE)["xmin"])
	if (char) out <- as.character(out)
	out
})

#' @aliases E
#' @rdname ext
#' @exportMethod E
methods::setMethod(
	f = "E",
	signature = "missing",
	definition = function(x, char = FALSE) {
	out <- unname(ext(.region(), vector = TRUE)["xmax"])
	if (char) out <- as.character(out)
	out
})

#' @aliases E
#' @rdname ext
#' @exportMethod E
methods::setMethod(
	f = "E",
	signature = "GSpatial",
	definition = function(x, char = FALSE) {
	out <- unname(ext(x, vector = TRUE)["xmax"])
	if (char) out <- as.character(out)
	out
})

#' @aliases N
#' @rdname ext
#' @exportMethod N
methods::setMethod(
	f = "N",
	signature = "missing",
	definition = function(x, char = FALSE) {
	out <- unname(ext(.region(), vector = TRUE)["ymax"])
	if (char) out <- as.character(out)
	out
})

#' @aliases N
#' @rdname ext
#' @exportMethod N
methods::setMethod(
	f = "N",
	signature = "GSpatial",
	definition = function(x, char = FALSE) {
	out <- unname(ext(x, vector = TRUE)["ymax"])
	if (char) out <- as.character(out)
	out
})

#' @aliases S
#' @rdname ext
#' @exportMethod S
methods::setMethod(
	f = "S",
	signature = "missing",
	definition = function(x, char = FALSE) {
	out <- unname(ext(.region(), vector = TRUE)["ymin"])
	if (char) out <- as.character(out)
	out
})

#' @aliases S
#' @rdname ext
#' @exportMethod S
methods::setMethod(
	f = "S",
	signature = "GSpatial",
	definition = function(x, char = FALSE) {
	out <- unname(ext(x, vector = TRUE)["ymin"])
	if (char) out <- as.character(out)
	out
})

#' @aliases top
#' @rdname ext
#' @exportMethod top
methods::setMethod(
	f = "top",
	signature = c(x = "missing"),
	definition = function(x, char = FALSE) {
	out <- unname(zext(.region())["top"])
	if (char) out <- as.character(out)
	out
})

#' @aliases top
#' @rdname ext
#' @exportMethod top
methods::setMethod(
	f = "top",
	signature = "GSpatial",
	definition = function(x, char = FALSE) {
	out <- unname(zext(x)["top"])
	if (char) out <- as.character(out)
	out
})

#' @aliases bottom
#' @rdname ext
#' @exportMethod bottom
methods::setMethod(
	f = "bottom",
	signature = c(x = "GSpatial"),
	definition = function(x, char = FALSE) {
	out <- unname(zext(.region())["bottom"])
	if (char) out <- as.character(out)
	out
})

#' @aliases bottom
#' @rdname ext
#' @exportMethod bottom
methods::setMethod(
	f = "bottom",
	signature = "GSpatial",
	definition = function(x, char = FALSE) {

	out <- unname(zext(x)["bottom"])
	if (char) out <- as.character(out)
	out

	} # EOF
)

#' Function to get extent from a "sources" name of a raster or vector
#'
#' @param x A `GRaster`, `GSpatial`, or a character ([sources()] name of a `GRaster` or `GVector`).
#'
#' @param rastOrVect Either `NULL` (class taken from `x`, but `x` cannot be a character), or "`raster`" or "`vector`" (partial matching is used).
#'
#' @returns A numeric vector.
#'
#' @keywords internal
.ext <- function(x, rastOrVect = NULL) {

	if (inherits(x, "GRaster")) {
		x <- sources(x)
		rastOrVect <- "raster"
	} else if (inherits(x, "GVector")) {
		x <- sources(x)
		rastOrVect <- "vector"
	}

	if (is.null(rastOrVect)) stop("Cannot determine if x is a raster or vector. Use argument rastOrVect.")
	rastOrVect <- omnibus::pmatchSafe(rastOrVect, c("raster", "vector"), nmax = 1L)

	if (rastOrVect == "raster") {
		info <- .rastInfo(x)
	} else {
		info <- .vectInfo(x)
	}

	c(xmin = info$west, xmax = info$east, ymin = info$south, ymax = info$north)	

}

# #' @aliases st_bbox
# #' @rdname ext
# #' @exportMethod st_bbox
# methods::setMethod(
#     f = "st_bbox",
#     signature = c(obj = "missing"),
#     function(obj, ...) {
#         out <- ext(vector = TRUE)
#         names(out) <- c("xmin", "xmax", "ymin", "ymax")
#         sf::st_bbox(out, crs = st_crs())
#     } # EOF
# )

# #' @aliases st_bbox
# #' @rdname ext
# #' @exportMethod st_bbox
# methods::setMethod(
# 	f = "st_bbox",
# 	signature = c(obj = "GSpatial"),
# 	function(obj, ...) {

# 	out <- obj@extent
# 	names(out) <- c('xmin', 'xmax', 'ymin', 'ymax')
# 	sf::st_bbox(out, crs = st_crs(obj))
	
# 	} # EOF
# )

# #' @rdname ext
# #' @export
# st_bbox <- function(obj, ...) {
# 	dots <- list(...)
# 	if (any(names(dots) == "vector")) {
# 		vector <- dots$vector
# 	} else {
# 		vector <- FALSE
# 	}
#     if (missing(obj)) obj <- .region()
#     if (inherits(obj, "GSpatial")) {
#         out <- obj@extent
#         out <- c(out[1L], out[3L], out[2L], out[4L])
#         names(out) <- c("xmin", "ymin", "xmax", "ymax")
#         if (!vector) out <- sf::st_bbox(out, crs = st_crs(obj))
#     } else {
#         out <- sf::st_bbox(obj, ...)
#         if (vector) out <- as.vector(out)
#     }
#     out
# }

# #' @rdname ext
# #' @exportMethod st_bbox
# setMethod("st_bbox", definition = function(obj, ...) st_bbox(obj, ...))

# st_bbox <- function(obj, ...) UseMethod("st_bbox", obj)

# #' @aliases st_bbox
# #' @rdname ext
# #' @export
# st_bbox <- function(obj, ...) UseMethod("st_crs", obj)

if (grassStarted()) {

# Setup
library(terra)

# Elevation raster
madElev <- fastData("madElev")
elev <- fast(madElev)

### Resample raster to 120 x 120 m
elev120 <- resample(elev, c(120, 120), method="bilinear")
elev
elev120

### Resample using another raster as a template

template <- aggregate(elev, 4)

nearest <- resample(elev, template, method = "nearest")

bilinear <- resample(elev, template, method = "bilinear")
bilinearNoFB <- resample(elev, template, method = "bilinear", fallback = FALSE)

bicubic <- resample(elev, template, method = "bicubic")
bicubicNoFB <- resample(elev, template, method = "bicubic", fallback = FALSE)

lanczos <- resample(elev, template, method = "lanczos")
lanczosNoFB <- resample(elev, template, method = "lanczos", fallback = FALSE)

# rasters resampled without fallback have fewer non-NA cells
resampled <- c(nearest, bilinear, bilinearNoFB, bicubic, bicubicNoFB, lanczos,
    lanczosNoFB)
names(resampled) <- c("nearest", "bilinear", "bilinearNoFB", "bicubic",
    "bicubicNoFB", "lanczos", "lanczosNoFB")
ones <- resampled * 0 + 1
global(ones, "sum") # number of non-NA cells
global(resampled, c("mean", "sd", "min", "max")) # other statistics

# Compare fallback to no fallback
frLanczos <- rast(lanczos)
frLanczosNoFB <- rast(lanczosNoFB)

plot(frLanczos, col = "red",
    main = "Red: Cells in fallback not non-fallback", legend = FALSE)
plot(frLanczosNoFB, add=TRUE)

# Compare fasterRaster with terra
coarserTerra <- aggregate(madElev, 4)
terraLanczos <- resample(madElev, coarserTerra, method = "lanczos")

frLanczos <- extend(frLanczos, terraLanczos)
frLanczosNoFB <- extend(frLanczosNoFB, terraLanczos)

frLanczos - terraLanczos
frLanczosNoFB - terraLanczos

plot(frLanczos - terraLanczos, main = "Difference")
plot(frLanczosNoFB - terraLanczos, main = "Difference")

plot(terraLanczos, col = "red",
    main = "Red: Cells in terra not in FR", legend = FALSE)
plot(frLanczos, add=TRUE)

plot(frLanczos, col = "red",
    main = "Red: Cells in FR not in terra", legend = FALSE)
plot(terraLanczos, add=TRUE)

}

#' Subset layers from a GRaster, or specific rows from a GVector
#'
#' @description `subset()` can be used to subset or remove one or more layers from a `GRaster`. It can also be used to subset or remove rows from a `GVector` with a data table.
#'
#' @param x A `GRaster` or `GVector`.
#'
#' @param subset Numeric integer, integer, logical, or character: Indicates the layer(s) of a `GRaster` to subset, or the rows(s) of a `GVector` to return.
#'
#' @param negate Logical: If `TRUE`, all layers or rows in `subset` will be *removed* from the output. Default is `FALSE`.
#'
#' @returns A `GRaster` or `GVector`.
#'
#' @seealso \code{\link[fasterRaster]{[[}}, \code{\link[fasterRaster]{[}}
#'
#' @example man/examples/ex_GRaster_GVector_subset_assign.r
#'
#' @aliases subset
#' @docType methods
#' @rdname subset
#' @exportMethod subset
methods::setMethod(
	"subset",
	signature = c(x = "GRaster"),
	function(x, subset, negate = FALSE) {

	# test indices
	subset <- .layerIndex(subset, x, recycle = TRUE, negate = negate)
	x[[subset]]

	} # EOF
)

#' @aliases subset
#' @rdname subset
#' @exportMethod subset
methods::setMethod(
	"subset",
	signature = c(x = "GVector"),
	function(x, subset, negate = FALSE) {

	nr <- nrow(x)

	# turn subset into integer
	if (is.logical(subset)) {
		if (length(subset) < nr) subset <- rep(subset, length.out = nr)
		subset <- which(subset)
	}
	if (negate) subset <- seq_len(nr)[-subset]
	x[subset]	
	
	} # EOF
)

if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Elevation raster, plant specimen collections, rivers vector,
# outline of area vector
madElev <- fastData("madElev")
madDypsis <- fastData("madDypsis")

# Convert to fasterRaster format:
elev <- fast(madElev)
dypsis <- fast(madDypsis)

# Kernel density estimation:
kde <- kernel(dypsis, elev)
plot(kde)
plot(dypsis, add = TRUE, pch = 1)

}

#' Subset layers from a GRaster, or specific columns from a GVector
#'
#' @description The `[[` operator can be used to subset or remove one or more layers from a `GRaster`. It can also be used to subset or remove columns from a `GVector` with a data table.
#'
#' @param x A `GRaster` or `GVector`.
#'
#' @param i Numeric integer, integer, logical, or character: Indicates the layer(s) of a `GRaster` to subset, or the column(s) of a `GVector` to return. If negative numeric or integer values are supplied, then these layers or columns will be removed from the output.
#'
#' @param j Ignored for `[[`.
#'
#' @returns A `GRaster` or `GVector`.
#'
#' @seealso [subset()]
#'
#' @example man/examples/ex_GRaster_GVector_subset_assign.r
#'
#' @name [[
#' @aliases [[,GRaster,ANY,ANY-method
#' @docType methods
#' @rdname subset_double_square_brackets
#' @exportMethod [[
methods::setMethod(
	"[[",
	signature = c(x = "GRaster", i = "ANY", j = "ANY"),
	function(x, i, j) {

	# test indices
	i <- .layerIndex(i, x, recycle = TRUE)
	mm <- minmax(x)
	dims <- dim(x)[1L:3L]
	
	out <- new(
		"GRaster",
		location = .location(x),
		mapset = .mapset(x),
		workDir = faster("workDir"),
		crs = crs(x),
		projection = .projection(x),
		nLayers = length(i),
		dimensions = dims,
		topology = topology(x),
		extent = as.vector(ext(x)),
		zextent = zext(x),
		sources = sources(x)[i],
		names = names(x)[i],
		datatypeGRASS = datatype(x, "GRASS")[i],
		resolution = res3d(x),
		minVal = mm["min", i],
		maxVal = mm["max", i],
		activeCat = x@activeCat[i],
		levels = cats(x)[i]
	)
	
	if (length(anyDuplicated(out@names)) > 0L) out@names <- make.unique(out@names)
	out

	} # EOF
)

#' @name [[
#' @aliases [[,GVector,ANY,ANY-method
#' @docType methods
#' @rdname subset_double_square_brackets
#' @exportMethod [[
methods::setMethod(
	"[[",
	signature = c(x = "GVector", i = "ANY", j = "ANY"),
	function(x, i, j) {

	nc <- ncol(x)
	if (nc == 0L) stop("Vector has no data table associated with it.")

	# turn i into integer
	if (is.logical(i)) {
		if (length(i) < nc) i <- rep(i, length.out = nc)
		i <- which(i)
	} else if (is.character(i)) {
		i <- match(i, names(x))
  		if (anyNA(i)) stop("At least one named column does not exist in this vector\'s data table.")
	}

	if (any(i < 0L) & any(i > 0L)) stop("Cannot have both positive and negative indices.")

	if (all(i < 0L)) {
		reverseSelect <- TRUE
		i <- i * -1L
		removeAll <- length(i) == ncol(x) && all(sort(i) == seq_len(ncol(x)))
	} else {
		reverseSelect <- removeAll <- FALSE
	}

	if (any(i > nc)) stop("Index out of bounds.")
	
	# removed all columns
	if (removeAll) {
		x@table <- data.table::data.table(NULL)
	# keep some columns
	} else {

		# ..i <- NULL
		if (reverseSelect) {
			i <- setdiff(names(x), names(x)[i])
		} else {
			i <- names(x)[i]
		}

		x@table <- x@table[ , ..i, with = FALSE]
	}
	x
	
	} # EOF
)

#' @aliases bioclims
#' @rdname bioclims
#' @exportMethod bioclims
methods::setMethod(
	f = "bioclims",
	signature = c(ppt = "SpatRaster"),
	function(
		ppt,
		tmin,
		tmax,
		tmean = NULL,
		bios = NULL,
		sample = TRUE,
		quarter = 3,
		pptDelta = 1,
		verbose = TRUE
	) {

	# BIOCLIM numbers
	classicBios <- 1L:19L
	extendedBios <- 41L:60L
	allBios <- c(classicBios, extendedBios)

	meanTempBios <- c(1, 4, 8:11, 18:19, 41:46, 53:54, 57:60) # BIOCLIMs that require mean temp
	quarterTempBios <- c(10, 11, 18, 19, 41:46, 53:54) # BIOCLIMs that require quarter temperatures
	quarterPrecipBios <- c(8, 9, 16, 17, 16:19, 45:48, 55:56)
	wettestQuarterBios <- c(8, 16, 46, 48, 55)
	driestQuarterBios <- c(9, 17, 45, 47, 56)
	warmestQuarterBios <- c(10, 18, 42, 44, 53)
	coldestQuarterBios <- c(11, 19, 41, 43, 54)
	
	if (is.null(bios)) {
		bios <- classicBios
	} else if (any(bios == "*")) {
		bios <- allBios
	} else if (any(bios == "+")) {
		bios <- bios[bios != "+"]
		bios <- c(bios, extendedBios)
	}

	bios <- unique(bios)
	bios <- sort(bios)
	bios <- as.integer(bios)

	if (quarter <= 1) stop("Argument `quarter` must be > 1.")

	nLayers <- terra::nlyr(ppt)

	if (any(omnibus::notIn(bios, allBios))) stop("Argument `bios` has at least one invalid value. Valid values include:\n  ", paste(allBios, collapse = " "))

	if (nLayers != terra::nlyr(tmin) | nLayers != terra::nlyr(tmax)) stop("The input rasters must have the same number of layers.")

	if (quarter > nLayers) stop("Argument `quarter` cannot be more than the number of raster layers in any input data.")

	### number of tasks (used for progress bar
	nTasks <- length(bios) +
		(any(meanTempBios %in% bios) & is.null(tmean)) +
		any(bios %in% quarterTempBios) +
		any(bios %in% quarterPrecipBios) +
		(any(bios %in% warmestQuarterBios) & !(53 %in% warmestQuarterBios)) +
		(any(bios %in% coldestQuarterBios) & !(54 %in% coldestQuarterBios)) +
		(any(bios %in% wettestQuarterBios) & !(55 %in% wettestQuarterBios)) +
		(any(bios %in% driestQuarterBios) & !(55 %in% driestQuarterBios))
		
	if (verbose | faster("verbose")) {
		pb <- utils::txtProgressBar(min = 0, max = nTasks, initial = 0, style = 3, width = 30)
	}
	tasks <- 0

	### calculate tmean if needed
	if (is.null(tmean) & any(bios %in% meanTempBios)) {
		
		tmean <- (tmin + tmax) / 2

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)
	
	}

	### quarter temperature rasters
	if (any(bios %in% quarterTempBios)) {

		tmeanByQuarter <- .calcQuarterSR(tmean, fun = "mean", quarter = quarter)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)
	
	}

	### quarter precipitation rasters
	if (any(bios %in% quarterPrecipBios)) {

		pptByQuarter <- .calcQuarterSR(ppt, fun = "sum", quarter = quarter)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)
	
	}

	### MAIN
	########

	out <- NULL

	### BIO53 (warmest quarter)
	if (any(bios %in% warmestQuarterBios)) {

		bio53 <- app(tmeanByQuarter, which.max)
		warmestQuarters <- terra::freq(bio53)$value
		names(bio53) <- "bio53"
		if (53 %in% bios) out <- c(out, bio53)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)
		
	}

	### BIOCLIM 54 (coldest quarter)
	if (any(bios %in% coldestQuarterBios)) {

		bio54 <- app(tmeanByQuarter, which.min)
		coldestQuarters <- terra::freq(bio54)$value
		names(bio54) <- "bio54"
		if (54 %in% bios) out <- c(out, bio54)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	### BIOCLIM 55 (wettest quarter)
	if (any(bios %in% wettestQuarterBios)) {

		bio55 <- app(pptByQuarter, which.max)
		wettestQuarters <- terra::freq(bio55)$value
		names(bio55) <- "bio55"
		if (55 %in% bios) out <- c(out, bio55)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)
		
	}

	### BIOCLIM 56 (driest quarter)
	if (any(bios %in% driestQuarterBios)) {

		bio56 <- app(pptByQuarter, which.min)
		driestQuarters <- terra::freq(bio56)$value
		names(bio56) <- "bio56"
		if (56 %in% bios) out <- c(out, bio56)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)
		
	}

	### BIOCLIM 1
	if (any(1 %in% bios)) {
		
		bio1 <- mean(tmean)
		names(bio1) <- "bio1"
		out <- c(out, bio1)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)
	
	}

	### BIOCLIM 2
	if (any(c(2, 3) %in% bios)) {

		delta <- tmax - tmin
		bio2 <- mean(delta)
		names(bio2) <- "bio2"
		if (2 %in% bios) out <- c(out, bio2)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)
	
	}

	### BIOCLIM 5
	if (any(c(3, 5, 7) %in% bios)) {
		
		bio5 <- max(tmax)
		names(bio5) <- "bio5"
		if (5 %in% bios) out <- c(out, bio5)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)
	
	}

	### BIOCLIM 6
	if (any(c(3, 6, 7) %in% bios)) {
		
		bio6 <- min(tmin)
		names(bio6) <- "bio6"
		if (6 %in% bios) out <- c(out, bio6)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	### BIOCLIM 7 (annual temp range)
	if (any(c(3, 7) %in% bios)) {
		
		bio7 <- bio5 - bio6
		names(bio7) <- "bio7"
		if (7 %in% bios) out <- c(out, bio7)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)
	
	}

	### BIOCLIM 3 (100 * bio2 / bio7)
	if (3 %in% bios) {
		
		bio3 <- 100 * bio2 / bio7
		names(bio3) <- "bio3"
		out <- c(out, bio3)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)
	
	}

	### BIOCLIM 4 (100 * sd of temp)
	if (4 %in% bios) {
		
		bio4 <- 100 * terra::stdev(tmean, pop = !sample)
		names(bio4) <- "bio4"
		out <- c(out, bio4)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)
	
	}

	# BIOCLIM 8 (mean temp of wettest quarter)
	if (8 %in% bios) {
		
		mm <- terra::minmax(tmeanByQuarter)
		minVal <- -10 * abs(min(mm[1L, , drop = TRUE]))
		bio8 <- tmean[[1L]] * 0 + minVal
		for (i in wettestQuarters) {
			bio8 <- max(bio8, (bio55 == i) * tmeanByQuarter[[i]])
		}

		names(bio8) <- "bio8"
		out <- c(out, bio8)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIOCLIM 9 (mean temp of driest quarter)
	if (9 %in% bios) {
		
		mm <- terra::minmax(tmeanByQuarter)
		minVal <- -10 * abs(min(mm[1L, , drop = TRUE]))
		bio9 <- tmean[[1L]] * 0 + minVal
		for (i in driestQuarters) {
			bio9 <- max(bio9, (bio56 == i) * tmeanByQuarter[[i]])
		}

		names(bio9) <- "bio9"
		out <- c(out, bio9)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIOCLIM 10 (mean temp of warmest quarter)
	if (10 %in% bios) {

		mm <- terra::minmax(tmeanByQuarter)
		minVal <- -10 * abs(min(mm[1L, , drop = TRUE]))
		bio10 <- tmean[[1L]] * 0 + minVal
		for (i in warmestQuarters) {
			bio10 <- max(bio10, (bio53 == i) * tmeanByQuarter[[i]])
		}

		names(bio10) <- "bio10"
		out <- c(out, bio10)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIOCLIM 11 (mean temp of coldest quarter)
	if (11 %in% bios) {

		mm <- terra::minmax(tmeanByQuarter)
		minVal <- -10 * abs(min(mm[1L, , drop = TRUE]))
		bio11 <- tmean[[1L]] * 0 + minVal
		for (i in coldestQuarters) {
			bio11 <- max(bio11, (bio54 == i) * tmeanByQuarter[[i]])
		}

		bio11 <- max(bio11)
		names(bio11) <- "bio11"
		out <- c(out, bio11)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIOCLIM 12 (total precipitation)
	if (12 %in% bios) {

		bio12 <- sum(ppt)
		names(bio12) <- "bio12"
		if (12 %in% bios) out <- c(out, bio12)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIOCLIM 13 (wettest month precipitation)
	if (13 %in% bios) {

		bio13 <- max(ppt)
		names(bio13) <- "bio13"
		out <- c(out, bio13)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIOCLIM 14 (driest month precipitation)
	if (14 %in% bios) {

		bio14 <- min(ppt)
		names(bio14) <- "bio14"
		out <- c(out, bio14)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	### BIOCLIM 15 (CV of precipitation)
	if (15 %in% bios) {
		
		sd <- terra::stdev(ppt + pptDelta, pop = !sample)
		bio15 <- 100 * sd / mean(ppt + pptDelta)
		names(bio15) <- "bio15"
		out <- c(out, bio15)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)
	
	}

	# BIOCLIM 16 (precipitation of the wettest quarter)
	if (16 %in% bios) {

		bio16 <- ppt[[1L]] * 0
		for (i in wettestQuarters) {
			bio16 <- c(bio16, (bio55 == i) * pptByQuarter[[i]])
		}

		bio16 <- sum(bio16)
		names(bio16) <- "bio16"
		out <- c(out, bio16)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO17: Precipitation of the driest quarter
	if (17 %in% bios) {

		bio17 <- ppt[[1L]] * 0
		for (i in driestQuarters) {
			bio17 <- c(bio17, (bio56 == i) * pptByQuarter[[i]])
		}

		bio17 <- sum(bio17)
		names(bio17) <- "bio17"
		out <- c(out, bio17)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO18: Precipitation of the warmest quarter (based on mean temperature)
	if (18 %in% bios) {

		bio18 <- ppt[[1L]] * 0
		for (i in warmestQuarters) {
			bio18 <- c(bio18, (bio53 == i) * pptByQuarter[[i]])
		}

		bio18 <- sum(bio18)
		names(bio18) <- "bio18"
		out <- c(out, bio18)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO19: Precipitation of the coldest quarter (based on mean temperature)
	if (19 %in% bios) {

		bio19 <- ppt[[1L]] * 0
		for (i in coldestQuarters) {
			bio19 <- c(bio19, (bio54 == i) * pptByQuarter[[i]])
		}

		bio19 <- sum(bio19)
		names(bio19) <- "bio19"
		out <- c(out, bio19)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO41: Temperature of the quarter following the coldest quarter (based on mean temperature)
	if (41 %in% bios) {

		mm <- terra::minmax(tmeanByQuarter)
		minVal <- -10 * abs(min(mm[1L, , drop = TRUE]))
		bio41 <- tmean[[1L]] * 0 + minVal
		for (i in coldestQuarters) {
			if (i + quarter <= nLayers) {
				j <- i + quarter
			} else {
				j <- i + quarter - nLayers
			}
			bio41 <- max(bio41, (bio54 == i) * tmeanByQuarter[[j]])
		}

		names(bio41) <- "bio41"
		out <- c(out, bio41)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO42: Temperature of the quarter following the warmest quarter (based on mean temperature)
	if (42 %in% bios) {

		mm <- terra::minmax(tmeanByQuarter)
		minVal <- -10 * abs(min(mm[1L, , drop = TRUE]))
		bio42 <- tmean[[1L]] * 0 + minVal
		for (i in warmestQuarters) {
			if (i + quarter <= nLayers) {
				j <- i + quarter
			} else {
				j <- i + quarter - nLayers
			}
			bio42 <- max(bio42, (bio53 == i) * tmeanByQuarter[[j]])
		}

		names(bio42) <- "bio42"
		out <- c(out, bio42)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO43: Precipitation of the quarter following the coldest quarter (based on mean temperature; mm)
	if (43 %in% bios) {

		bio43 <- ppt[[1L]] * 0
		for (i in coldestQuarters) {
			if (i + quarter <= nLayers) {
				j <- i + quarter
			} else {
				j <- i + quarter - nLayers
			}
			bio43 <- max(bio43, (bio54 == i) * pptByQuarter[[j]])
		}

		names(bio43) <- "bio43"
		out <- c(out, bio43)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO44: Precipitation of the quarter following the warmest quarter (based on mean temperature; mm)
	if (44 %in% bios) {

		bio44 <- ppt[[1L]] * 0
		for (i in warmestQuarters) {
			if (i + quarter <= nLayers) {
				j <- i + quarter
			} else {
				j <- i + quarter - nLayers
			}
			bio44 <- max(bio44, (bio53 == i) * pptByQuarter[[j]])
		}

		names(bio44) <- "bio44"
		out <- c(out, bio44)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO45: Temperature of the quarter following the driest quarter (based on mean temperature; deg C)
	if (45 %in% bios) {

		mm <- terra::minmax(tmeanByQuarter)
		minVal <- -10 * abs(min(mm[2L, , drop = TRUE]))
		bio45 <- tmean[[1L]] * 0 + minVal
		for (i in driestQuarters) {
			if (i + quarter <= nLayers) {
				j <- i + quarter
			} else {
				j <- i + quarter - nLayers
			}
			bio45 <- max(bio45, (bio56 == i) * tmeanByQuarter[[j]])
		}

		names(bio45) <- "bio45"
		out <- c(out, bio45)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO46: Temperature of the quarter following the wettest quarter (based on mean temperature; deg C)
	if (46 %in% bios) {

		mm <- terra::minmax(tmeanByQuarter)
		minVal <- -10 * abs(min(mm[2L, , drop = TRUE]))
		bio46 <- tmean[[1L]] * 0 + minVal
		for (i in wettestQuarters) {
			if (i + quarter <= nLayers) {
				j <- i + quarter
			} else {
				j <- i + quarter - nLayers
			}
			bio46 <- max(bio46, (bio55 == i) * tmeanByQuarter[[j]])
		}

		names(bio46) <- "bio46"
		out <- c(out, bio46)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO47: Precipitation of the quarter following the driest quarter (based on mean temperature; mm)
	if (47 %in% bios) {

		bio47 <- ppt[[1L]] * 0
		for (i in driestQuarters) {
			if (i + quarter <= nLayers) {
				j <- i + quarter
			} else {
				j <- i + quarter - nLayers
			}
			bio47 <- max(bio47, (bio56 == i) * pptByQuarter[[j]])
		}

		names(bio47) <- "bio47"
		out <- c(out, bio47)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO48: Precipitation of the quarter following the wettest quarter (based on mean temperature; mm)
	if (48 %in% bios) {

		bio48 <- ppt[[1L]] * 0
		for (i in wettestQuarters) {
			if (i + quarter <= nLayers) {
				j <- i + quarter
			} else {
				j <- i + quarter - nLayers
			}
			bio48 <- max(bio48, (bio55 == i) * pptByQuarter[[j]])
		}

		names(bio48) <- "bio48"
		out <- c(out, bio48)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO49: Hottest month
	if (49 %in% bios) {

		bio49 <- terra::app(tmax, which.max)
		names(bio49) <- "bio49"
		out <- c(out, bio49)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO50: Coldest month
	if (50 %in% bios) {

		bio50 <- terra::app(tmin, which.min)
		names(bio50) <- "bio50"
		out <- c(out, bio50)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}
	
	# BIO51: Wettest month
	if (51 %in% bios) {

		bio51 <- terra::app(ppt, which.max)
		names(bio51) <- "bio51"
		out <- c(out, bio51)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO52: Driest month
	if (52 %in% bios) {

		bio52 <- terra::app(ppt, which.min)
		names(bio52) <- "bio52"
		out <- c(out, bio52)

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

	}

	# BIO57: The largest decrease in temperature from one month to the next (deg C).
	# BIO58: The largest increase in temperature from one month to the next (deg C).
	if (any(c(57, 58) %in% bios)) {

		deltas <- tmean
		for (i in 1:nLayers) {
			
			if (i + quarter <= nLayers) {
				target <- i + 1
			} else {
				target <- 1
			}

			deltas[[i]] <- tmean[[i]] - tmean[[target]]

		}

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)
		
		if (57 %in% bios) {

			bio57 <- max(deltas)
			names(bio57) <- "bio57"
			out <- c(out, bio57)

			tasks <- tasks + 1
			if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

		}

		if (58 %in% bios) {

			bio58 <- abs(min(deltas))
			names(bio58) <- "bio58"
			out <- c(out, bio58)

			tasks <- tasks + 1
			if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

		}

	}
	
	# BIO59: The largest decrease in precipitation from one month to the next.
	# BIO60: The largest increase in precipitation from one month to the next.
	if (any(c(59, 60) %in% bios)) {

		deltas <- ppt
		for (i in 1:nLayers) {
			
			if (i + quarter <= nLayers) {
				target <- i + 1
			} else {
				target <- 1
			}

			deltas[[i]] <- ppt[[i]] - ppt[[target]]

		}

		tasks <- tasks + 1
		if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)
		
		if (59 %in% bios) {

			bio59 <- max(deltas)
			names(bio59) <- "bio59"
			out <- c(out, bio59)

			tasks <- tasks + 1
			if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

		}

		if (60 %in% bios) {

			bio60 <- abs(min(deltas))
			names(bio60) <- "bio60"
			out <- c(out, bio60)

			tasks <- tasks + 1
			if (verbose | faster("verbose")) utils::setTxtProgressBar(pb, tasks)

		}

	}

	out <- do.call(c, out)

	outNames <- names(out)
	nc <- nchar(outNames)
	if (any(nc == 4L)) {
		shorts <- out[[nc == 4L]]
		shorts <- shorts[[order(names(shorts))]]
	} else { shorts <- NULL }

	if (any(nc == 5L)) {
		longs <- out[[nc == 5L]]
		longs <- longs[[order(names(longs))]]
	} else { longs <- NULL }

	if (is.null(shorts)) {
		out <- longs
	} else if (is.null(longs)) {
		out <- shorts
	} else {
		out <- c(shorts, longs)
	}

	if (verbose | faster("verbose")) close(pb)
	out

	} # EOF
)


#' Calculate a series of "quarter" rasters and return their [sources()] names. A quarter can be summarized using 
#'
#' x A "stack" of `SpatRaster`s
#' fun Character: Name of function to summarize across a quarter ("sum", "min", "max", or "mean").
#' quarter Numeric: Length of a quarter
#'
#' @noRd
.calcQuarterSR <- function(x, fun, quarter) {

	out <- x
	qs <- c(x, x[[1L:(quarter - 1L)]])

	n <- nlyr(x)
	for (i in seq_len(n)) {

		select <- i:(i + quarter - 1L)
		quart <- qs[[select]]

		args <- list(x = quart)
		out[[i]] <- do.call(fun, args)

	} # next quarter
	names(out) <- paste0("q", 1L:n)
	out

}

#' Coerce raster to integer, float, or double precision
#'
#' @description In **fasterRaster**, rasters can have three data types: "factor" (categorical rasters), "integer" (integers), "float" (floating point values, accurate to 6th to 9th decimal places), and "double" (double-precision values, accurate to the 15th to 17th decimal places). The type of raster can be checked with:
#'
#' * `as.int()`: Coerce values to integers (**GRASS** type `CELL`).
#' * `as.float()`: Coerce values to floating-point precision.
#' * `as.doub()`: Coerce values to double-floating point precision.
#' * Integer rasters can be converted categorical rasters by adding "levels" tables with \code{\link[fasterRaster]{levels<-}} or [categories()].
#' 
#' @param x A `GRaster`.
#' 
#' @returns A `GRaster`.
#' 
#' @seealso [datatype()], [terra::datatype()], [is.int()], [is.float()], [is.doub()], \code{\link[fasterRaster]{levels<-}}, `vignette("GRasters", package = "fasterRaster")`
#'
#' @example man/examples/ex_GRaster.r
#' 
#' @aliases as.int
#' @rdname as.int
#' @exportMethod as.int
methods::setMethod(
    f = "as.int",
    signature = c(x = "GRaster"),
    function(x) .as.type(x, fx = "int")
)

#' @aliases as.float
#' @rdname as.int
#' @exportMethod as.float
methods::setMethod(
    f = "as.float",
    signature = c(x = "GRaster"),
    function(x) .as.type(x, fx = "float")
)

#' @aliases as.doub
#' @rdname as.int
#' @exportMethod as.doub
methods::setMethod(
    f = "as.doub",
    signature = c(x = "GRaster"),
    function(x) .as.type(x, fx = "double")
)

# x: GRaster
# fx: "int", "float", "double"
.as.type <- function(x, fx) {

    .locationRestore(x)
    .region(x)

    for (i in 1L:nlyr(x)) {

        src <- .makeSourceName("as_raster_type", "rast")
        ex <- paste0(src, " = ", fx, "(", sources(x)[i], ")")
        rgrass::execGRASS(cmd = "r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

        this <- makeGRaster(src, names(x)[i])
        if (i == 1L) {
            out <- this
        } else {
            out <- c(out, this)
        }

    }
    out

} # EOF

#' Convert a GVector to a SpatVector or sf vector
#'
#' @description The **fasterRaster** version of the `vect()` function converts a `GVector` to a `SpatVector` (from the **terra** package). The **fasterRaster** version of the `st_as_sf()` function converts a `GVector` to an `sf` object (**sf** package).
#'
#' @param x A `GVector`.
#' @param ... Additional arguments to send to [writeVector()].
#'
#' @returns `vect()` returns a `SpatVector` (**terra** package), and `st_as_sf()` returns an `sf` vector (**sf** package).
#' 
#' @seealso [terra::vect()], [sf::st_as_sf()]
#' 
#' @example man/examples/ex_GVector.r
#'
#' @aliases vect
#' @rdname vect
#' @exportMethod vect
setMethod(
	"vect",
	signature(x = "GVector"),
	function(x, ...) {

	filename <- paste0(omnibus::forwardSlash(tempfile()), ".gpkg")
	writeVector(x, filename = filename, ...)
	out <- terra::vect(filename)
	out
	
	} # EOF
)

#' @aliases st_as_sf
#' @rdname vect
#' @exportMethod st_as_sf
setMethod(
	"st_as_sf",
	signature(x = "GVector"),
	function(x) {

	filename <- paste0(omnibus::forwardSlash(tempfile()), ".gpkg")
	writeVector(x, filename = filename, overwrite = TRUE)
	out <- sf::read_sf(filename)
	out

	} # EOF
)

st_as_sf <- function(x) UseMethod("st_as_sf", x)

#' Name of a raster or vector in a GRASS session
#'
#' @description `sources()` retrieves the name of a raster or vector in **GRASS**. `GRaster`s and `GVector`s are actually pointers to objects stored in a **GRASS** database. When using **fasterRaster** functions on rasters and vectors, the commands are translated into **GRASS** commands and executed on the objects named in the pointers. These objects use a "source" (which is really a filename) to refer to the **GRASS** objects. This function is mostly of use to developers.
#'
#' @param x Either a `GSpatial` object or one that inherits from it (i.e., a `GRaster` or `GVector`), *or* a character. If a character, then the character itself is returned.
#'
#' @returns Character.
#'
#' @example man/examples/ex_GRaster.r
#' 
#' @aliases sources
#' @rdname sources
#' @exportMethod sources
methods::setMethod(
	f = "sources",
	signature = "GRaster",
	definition = function(x) x@sources
)

#' @aliases sources
#' @rdname sources
#' @exportMethod sources
methods::setMethod(
	f = "sources",
	signature = "GVector",
	definition = function(x) x@sources
)

#' @aliases sources
#' @rdname sources
#' @exportMethod sources
methods::setMethod(
	f = "sources",
	signature = "character",
	definition = function(x) x
)

#' List objects in the active GRASS session
#'
#' Display the `sources` (see `sources()`) of all rasters and/or vectors that have been exported to or created in the active **GRASS** session"s location and mapset.
#'
#' @param type The type of spatial objects to display. This can include `"rasters"` (all rasters), `"vectors"` (all spatial vectors), `"rasters3d"` (3D-rasters), and/or `"groups"` (groups). Partial matching is supported. If missing, all objects are displayed.
#' 
#' @return Character vector of names of **GRASS** objects.
#'
#' @seealso [ls()]
#'
#' @aliases .ls
#' @rdname ls
#' @keywords internal
.ls <- function(type = c("rasters", "vectors", "rasters3d", "groups")) {

	rov <- c("rasters", "vectors", "rasters3d", "groups")
	if (is.null(type)) type <- rov
	type <- omnibus::pmatchSafe(type, rov, useFirst = TRUE, n = length(rov))

	# find rasters and vector
	rasts <- vects <- rasts3d <- vects3d <- groups <- character()
	
	# rasters
	if (any(type == "rasters")) {
		rasts <- rgrass::execGRASS("g.list", flags = .quiet(), type = "raster", intern = TRUE, echoCmd = FALSE)
		if (length(rasts) > 0L) names(rasts) <- rep("raster", length(rasts))
		rasts <- sort(rasts)
	}
	
	# 3D rasters
	if (any(type == "rasters3d")) {
		rasts3d <- rgrass::execGRASS("g.list", flags=.quiet(), type="raster_3d", intern=TRUE, echoCmd=FALSE)
		if (length(rasts3d) > 0L) names(rasts3d) <- rep("raster3d", length(rasts3d))
		rasts3d <- sort(rasts3d)
	}
	
	# vectors
	if (any(type == "vectors")) {
		vects <- rgrass::execGRASS("g.list", flags=.quiet(), type="vector", intern=TRUE, echoCmd=FALSE)
		if (length(vects) > 0L) names(vects) <- rep("vector", length(vects))
		vects <- sort(vects)
	}
	
	# groups
	if (any(type == "groups")) {
		groups <- rgrass::execGRASS("g.list", flags=.quiet(), type="group", intern=TRUE, echoCmd=FALSE)
		if (length(groups) > 0L) names(groups) <- rep("group", length(groups))
		groups <- sort(groups)
	}
	
	out <- c(rasts, rasts3d, vects, groups)
	out
	
}

#' Arithmetic operations on GRasters
#'
#' @description **`GRaster`s**: You can do arithmetic operations on `GRaster`s and using normal operators in **R**: `+`, `-`, `*`, `/`, `^`, `%%` (modulus), and `%/%` (integer division). 
#'
#' **`GVector`s**: You can also do arithmetic operations on `GVector`s:\cr\cr
#' `+` operator: Same as [union()]\cr
#' `-` operator: Same as [erase()]\cr
#' `*` operator: Same as [intersect()]\cr
#' `/` operator: Same as [xor()]\cr
#' 
#' @param e1,e2 `GRaster`s, `GVector`s, `numeric`s, `integer`s, or `logical`s.
#'
#' @return A `GRaster`.
#'
#' @example man/examples/ex_GRaster_arithmetic_by_layer.r
#'
#' @aliases Arith
#' @rdname Arithmetic
#' @exportMethod Arith
methods::setMethod(
	f = "Arith",
	signature(e1 = "GRaster", e2 = "logical"),
    function(e1, e2) {
	
		.locationRestore(e1)
  		.region(e1)

		e2 <- as.integer(e2)
		if (is.na(e2)) e2 <- "null()"
		
		for (i in 1L:nlyr(e1)) {
		
			src <- .makeSourceName("math", "rast")
			oper <- as.vector(.Generic)[1L]
			prec <- .getPrec(e1[[i]], oper)
			
			ex <- if (oper == "%/%") {
				paste(src, "= floor(", prec, "(", sources(e1)[i], ") /", e2, ")")
			} else if (oper == "%%") {
				paste0(src, " = ", prec, "(", sources(e1)[i], ") %", e2)
			} else {
				paste0(src, " = ", prec, "(", sources(e1)[i], ")  ", oper, e2)
			}

			name <- "layer"
			if (i == 1L) {
				out <- .genericArithRast(name = name, src = src, ex = ex)
			} else {
				out <- c(out, .genericArithRast(name = name, src = src, ex = ex))
			}
		
		} # next layer
		
		out
		
	} # EOF
)

# logical raster
#' @aliases Arith
#' @rdname Arithmetic
#' @exportMethod Arith
methods::setMethod(
	f = "Arith",
	signature(e1 = "logical", e2 = "GRaster"),
    function(e1, e2) {
	
		.locationRestore(e2)
  		.region(e2)

		e1 <- as.integer(e1)
		if (is.na(e1)) e1 <- "null()"
		
		for (i in 1L:nlyr(e2)) {
		
			src <- .makeSourceName("math", "rast")
			oper <- as.vector(.Generic)[1L]
			prec <- .getPrec(e2[[i]], oper)

			ex <- if (oper == "%/%") {
				paste(src, "= floor(", e1, "/ ", prec, "(", sources(e2)[i], "))")
			} else if (oper == "%%") {
				paste0(src, " = ", e1, "% ", prec, "(", sources(e2)[i], ")")
			} else {
				paste0(src, " = ", e1, oper, " ", prec, "(", sources(e2)[i], ")")
			}

			name <- "layer"
			if (i == 1L) {
				out <- .genericArithRast(name = name, src = src, ex = ex)
			} else {
				out <- c(out, .genericArithRast(name = name, src = src, ex = ex))
			}
		
		} # next layer
		
		out

	} # EOF
)

# raster numeric
#' @aliases Arith
#' @rdname Arithmetic
#' @exportMethod Arith
methods::setMethod(
	f = "Arith",
	signature(e1 = "GRaster", e2 = "numeric"),
    function(e1, e2) {
	
		.locationRestore(e1)
		.region(e1)

		if (is.na(e2)) {
			e2 <- "null()"
		} else {
			e2 <- formatC(e2, format = "f", drop0trailing = TRUE)
		}
		
		for (i in 1L:nlyr(e1)) {
		
			src <- .makeSourceName("math", "rast")
			oper <- as.vector(.Generic)[1L]
			prec <- .getPrec(e1[[i]], oper)

			ex <- if (oper == "%/%") {
				paste(src, "= floor(", prec, "(", sources(e1)[i], ") /", e2, ")")
			} else if (oper == "%%") {
				paste0(src, " = ", prec, "(", sources(e1)[i], ") %", e2)
			} else {
				paste0(src, " = ", prec, "(", sources(e1)[i], ")  ", oper, " ", e2)
			}

			name <- "layer"
			if (i == 1L) {
				out <- .genericArithRast(name = name, src = src, ex = ex)
			} else {
				out <- c(out, .genericArithRast(name = name, src = src, ex = ex))
			}
		
		} # next layer
		
		out
		
	} # EOF
)

# raster integer
#' @aliases Arith
#' @rdname Arithmetic
#' @exportMethod Arith
methods::setMethod(
	f = "Arith",
	signature(e1 = "GRaster", e2 = "integer"),
    function(e1, e2) {
	
		.locationRestore(e1)
  		.region(e1)

		if (is.na(e2)) {
			e2 <- "null()"
		} else {
			e2 <- formatC(e2, format = "f", drop0trailing = TRUE)
		}
		
		for (i in 1L:nlyr(e1)) {
		
			src <- .makeSourceName("math", "rast")
			oper <- as.vector(.Generic)[1L]
			prec <- .getPrec(e1[[i]], oper)
			
			ex <- if (oper == "%/%") {
				paste(src, "= floor(", prec, "(", sources(e1)[i], ") /", e2, ")")
			} else if (oper == "%%") {
				paste0(src, " = ", prec, "(", sources(e1)[i], ") %", e2)
			} else {
				paste0(src, " = ", prec, "(", sources(e1)[i], ")  ", oper, e2)
			}

			name <- "layer"
			if (i == 1L) {
				out <- .genericArithRast(name = name, src = src, ex = ex)
			} else {
				out <- c(out, .genericArithRast(name = name, src = src, ex = ex))
			}
		
		} # next layer
		
		out
		
	} # EOF
)

# numeric raster
#' @aliases Arith
#' @rdname Arithmetic
#' @exportMethod Arith
methods::setMethod(
	f = "Arith",
	signature(e1 = "numeric", e2 = "GRaster"),
    function(e1, e2) {
	
		.locationRestore(e2)
  		.region(e2)

		if (is.na(e1)) {
			e1 <- "null()"
		} else {
			e1 <- formatC(e1, format = "f", drop0trailing = TRUE)
		}
		oper <- as.vector(.Generic)[1L]
		
		for (i in 1L:nlyr(e2)) {
		
			src <- .makeSourceName("math", "rast")
			prec <- .getPrec(e2[[i]], oper)
			ex <- if (oper == "%/%") {
				paste(src, "= floor(", e1, "/ ", prec, "(", sources(e2)[i], "))")
			} else if (oper == "%%") {
				paste0(src, " = ", e1, " % ", prec, "(", sources(e2)[i], ")")
			} else {
				paste0(src, " = ", e1, oper, " ", prec, "(", sources(e2)[i], ")")
			}

			name <- "layer"
			if (i == 1L) {
				out <- .genericArithRast(name = name, src = src, ex = ex)
			} else {
				out <- c(out, .genericArithRast(name = name, src = src, ex = ex))
			}
		
		} # next layer
		
		out
		
	}
)

# integer raster
#' @aliases Arith
#' @rdname Arithmetic
#' @exportMethod Arith
methods::setMethod(
	f = "Arith",
	signature(e1 = "integer", e2 = "GRaster"),
    function(e1, e2) {
	
		.locationRestore(e2)
		.region(e2)

		if (is.na(e1)) {
			e1 <- "null()"
		} else {
			e1 <- formatC(e1, format = "f", drop0trailing = TRUE)
		}
		
		for (i in 1L:nlyr(e2)) {
		
			src <- .makeSourceName("math", "rast")
			oper <- as.vector(.Generic)[1L]
			prec <- .getPrec(e2[[i]], oper)

			ex <- if (oper == "%/%") {
				paste(src, "= floor(", e1, " / ", prec, "(", sources(e2)[i], "))")
			} else if (oper == "%%") {
				paste0(src, " = ", e1, "% ", prec, "(", sources(e2)[i], ")")
			} else {
				paste0(src, " = ", e1, oper, " ", prec, "(", sources(e2)[i], ")")
			}

   			name <- "layer"
			if (i == 1L) {
				out <- .genericArithRast(name = name, src = src, ex = ex)
			} else {
				out <- c(out, .genericArithRast(name = name, src = src, ex = ex))
			}
		
		} # next layer
		
		out
		
	}
)

# # missing raster
# methods::setMethod(
# 	f = "Arith",
# 	signature(e1 = "missing", e2 = "GRaster"),
#     function(e1, e2) {

# 	.locationRestore(e2)
# 	.region(e2)

# 	src <- .makeSourceName("math", "rast")

# 	oper <- as.vector(.Generic)[1L]
# 	print(oper)
# 	if (oper == "-") {
		
# 		ex <- paste0(src, " = -1 * ", sources(e2))
# 		out <- .genericArithRast(x = e2, src = src, ex = ex)

# 	} else {
# 		out <- e2
# 	}

# 	out
		
# 	} # EOF
# )

# raster raster
#' @aliases Arith
#' @rdname Arithmetic
#' @exportMethod Arith
methods::setMethod(
	f = "Arith",
	signature(e1 = "GRaster", e2 = "GRaster"),
    function(e1, e2) {
	
		compareGeom(e1, e2)
		.locationRestore(e1)
		.region(e1)

		oper <- as.vector(.Generic)[1L]
		
		if (nlyr(e1) == nlyr(e2)) {

			for (i in 1L:nlyr(e1)) {

				name <- "layer"
				src <- .makeSourceName("arithmetic", "rast")

				prec1 <- .getPrec(e1[[i]], oper)
				prec2 <- .getPrec(e2[[i]], oper)

				ex <- if (oper == "%/%") {
					paste0(src, " = floor(", prec1, "(", sources(e1)[i], ") / ", prec2, "(", sources(e2)[i], "))")
				} else if (oper == "%%") {
					paste0(src, " = ", prec1, "(", sources(e1)[i], ") % ", prec2, "(", sources(e2)[i], ")")
				} else if (oper == "^") {
					paste0(src, " = exp(", prec1, "(", sources(e1)[i], "), ", prec2, "(", sources(e2)[i], "))")
				} else {
					paste0(src, "= ", prec1, "(", sources(e1)[i], ")  ", oper, " ", prec2, "(", sources(e2)[i], ")")
				}
				
				if (i == 1L) {
					out <- .genericArithRast(name = name, src = src, ex = ex)
				} else {
					this <- .genericArithRast(name = name, src = src, ex = ex)
					out <- c(out, this)
				}
				
			}
			
		} else if (nlyr(e1) == 1L & nlyr(e2) > 1L) {

			prec1 <- .getPrec(e1, oper)
			for (i in 1L:nlyr(e2)) {
			
				name <- "layer"
				src <- .makeSourceName("arithmetic", "rast")

				prec2 <- .getPrec(e2[[i]], oper)

				ex <- if (oper == "%/%") {
					paste0(src, " = floor(", prec1, "(", sources(e1), ") / ", prec2, "(", sources(e2)[i], "))")
				} else if (oper == "%%") {
					paste0(src, " = ", prec1, "(", sources(e1), ") % ", prec2, "(", sources(e2)[i], ")")
				} else if (oper == "^") {
					paste0(src, " = exp(", prec1, "(", sources(e1), "), ", prec2, "(", sources(e2)[i], "))")
				} else {
					paste0(src, "= ", prec1, "(", sources(e1), ")  ", oper, " ", prec2, "(", sources(e2)[i], ")")
				}
				
				if (i == 1L) {
					out <- .genericArithRast(name = name, src = src, ex = ex)
				} else {
					this <- .genericArithRast(name = name, src = src, ex = ex)
					out <- c(out, this)
				}
				
			} # next layer
			
		} else if (nlyr(e1) > 1L & nlyr(e2) == 1L) {

			prec2 <- .getPrec(e2, oper)
			for (i in 1L:nlyr(e1)) {
			
				name <- paste0(names(e1)[i], "_", names(e2))
				src <- .makeSourceName("arithmetic", "rast")

				prec1 <- .getPrec(e1[[i]], oper)

				ex <- if (oper == "%/%") {
					paste0(src, " = floor(", prec1, "(", sources(e1)[i], ") / ", prec2, "(", sources(e2), "))")
				} else if (oper == "%%") {
					paste0(src, " = ", prec1, "(", sources(e1)[i], ") % ", prec2, "(", sources(e2), ")")
				} else if (oper == "^") {
					paste0(src, " = exp(", prec1, "(", sources(e1)[i], "), ", prec2, "(", sources(e2), "))")
				} else {
					paste0(src, "= ", prec1, "(", sources(e1)[i], ") ", oper, " ", prec2, "(", sources(e2), ")")
				}
				
				if (i == 1L) {
					out <- .genericArithRast(name = name, src = src, ex = ex)
				} else {
					this <- .genericArithRast(name = name, src = src, ex = ex)
					out <- c(out, this)
				}
				
			} # next layer
		
		} else {
			stop("Both rasters must have the same number of layers, or at least one raster must have a single layer.")
		}
		out
			
	} # EOF
)

# vector vector
#' @aliases Arith
#' @rdname Arithmetic
#' @exportMethod Arith
methods::setMethod(
	f = "Arith",
	signature(e1 = "GVector", e2 = "GVector"),
    function(e1, e2) {
	
		gtype1 <- geomtype(e1)
		gtype2 <- geomtype(e2)
		if (gtype1 != gtype2) stop("Vectors must be of the same type (polygons, lines, or points).")

		compareGeom(e1, e2)
		.locationRestore(e1)

		oper <- as.vector(.Generic)[1L]

		if (oper == "+") {
			if (gtype1 == "polygons" & gtype2 == "polygons") {
				out <- union(e1, e2)
			} else {
				out <- rbind(e1, e2)
			}
		} else if (oper == "-") {
			out <- erase(e1, e2)
		} else if (oper == "*") {
			out <- intersect(e1, e2)
		} else if (oper == "/") {
			out <- xor(e1, e2)
		} else {
			stop("Only the +, -, *, and / operators are defined for GVectors.")
		}
		
		out
			
	} # EOF
)

#' @noRd
.genericArithRast <- function(name, src, ex) {

	rgrass::execGRASS(cmd = "r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))
	makeGRaster(src, name)
	
}

#' Get function for precision of `GRaster`
#'
#' @param x A `GRaster`.
#' @param oper Character or `NULL`: Operator
#'
#' @returns Either "" (empty) or "double".
#' @noRd
.getPrec <- function(x, oper) {

	if ((is.null(oper) || oper %in% c("/", "%/%")) & any(is.cell(x))) {
		"double"
	} else {
		""
	}

}

if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster
elev <- fast(madElev)

# Create longitude/latitude rasters
ll <- longlat(elev)
ll # note units of cell values!

}

#' Centroid(s) of a vector or clumps in a raster
#'
#' @description This function locates the centroid of each geometry of a `GVector`, or the centroid of each "clump" of same-valued cells in an integer/categorical raster (for information on types of `GRaster`s, see `vignette("GRasters", package = "fasterRaster")`).
#' 
#' To use this function with a `GVector`, you need the **GRASS** `v.centerpoint` addon. To use the function with a `GRaster`, you need the addon `r.centroids`. In either case, the function will try to install the respective addon (i.e., you need to have an internet connection). Once installed, a tool will not need to be installed again.
#'
#' @param x A `GVector` or `GRaster`.
#'
#' @param method `GVector`s: Character or `NULL` (default): Method used for calculating centroids. The method of calculation depends on whether the input is a `points`, `lines`, or `polygons` `GVector`. If the value is `NULL`, then the default method will be chosen, depending on the geometry type of the `GVector`:
#' * `points`:
#'    * `"mean"` (default for `points`): Mean of coordinates.
#'    * `"median"`: Geometric median; more robust to outliers.
#'    * `"pmedian"`: Point in `x` closest to the geometric median.
#' * `lines`:
#'    * `"mid"` (default for `lines`): Mid-point on each line; will fall exactly on the line.
#'    * `"mean"`: Center of gravity of all line segments; may not fall on the line.
#'    * `"median`: Geometric median; may not fall on the line.
#' * `polygons`:
#'    * `"mean"` (default for `polygons`): Center of gravity (area), calculated using area triangulation.
#'    * `"median"`: Geometric mean; may not fall inside the polygon.
#'    * `"bmedian"`: Geometric mean; minimum distance to boundaries; may not fall inside the polygon.
#' 
#' Partial matching is used and case is ignored.
#'
#' @returns If the input is a `GVector`, the output will be a "points" `GVector`. If the input is a `GRaster`, the output will be a "points" `GVector` with a table with statistics on each clump. If the input is a `GRaster` with more than one layer, the output will be a `list` of `GVector`s, with one `GVector` per layer.
#'
#' @example man/examples/ex_centroids.r
#' 
#' @seealso [terra::centroids()]; **GRASS** addon tools `v.centerpoint` and `r.centroids`.
#'
#' @aliases centroids
#' @rdname centroids
#' @exportMethod centroids
methods::setMethod(
	f = "centroids",
	signature = c(x = "GVector"),
	function(x, method = NULL) {
	
	.addons("v.centerpoint")

	gtype <- geomtype(x)
	
	if (is.null(method)) {
	
		if (gtype == "points") {
			method <- "mean"
		} else if (gtype == "lines") {
			method <- "mid"
		} else if (gtype == "polygons") {
			method <- "mean"
		}
	
	}

	if (gtype == "points") {
		methods <- c("mean", "median", "pmedian")
	} else if (gtype == "lines") {
		methods <- c("mid", "mean", "median")
	} else if (gtype == "polygons") {
		methods <- c("mean", "median", "bmedian")
	}

	method <- omnibus::pmatchSafe(method, methods, nmax = 1L)

	src <- .makeSourceName("centroids", "vector")
	args <- list(
		cmd = "v.centerpoint",
		input = sources(x),
		output = src,
		flags = c(.quiet(), "overwrite")
	)
	
	if (gtype == "points") {
		args$pcenter <- method
	} else if (gtype == "lines") {
		args$lcenter <- method
	} else if (gtype == "polygons") {
		args$acenter <- method
	}

	do.call(rgrass::execGRASS, args = args)

	if (gtype == "points") {
		table <- NULL
	} else {
		table <- as.data.table(x)
	}

	makeGVector(src, table = table)
	
	} # EOF
)

#' @aliases centroids
#' @rdname centroids
#' @exportMethod centroids
methods::setMethod(
	f = "centroids",
	signature = c(x = "GRaster"),
	function(x) {
	
	.addons("r.centroids")

	.locationRestore(x)
	.region(x)

	nLayers <- nlyr(x)
	out <- list()
	srcs <- .makeSourceName("r_centroids", "vector", n = nLayers)
	for (i in 1:nLayers) {

		rawTable <- rgrass::execGRASS(
			cmd = "r.centroids",
			input = sources(x)[[i]],
			output = srcs[i],
			flags = c(.quiet(), "overwrite"),
			intern = TRUE
		)

		# parse table
		rawTable <- rawTable[7:(length(rawTable) - 2)]
		lines <- strsplit(rawTable, split = " ")
		table <- data.table::data.table()
		for (j in seq_along(rawTable)) {
		
			line <- lines[[j]]
			line <- line[line != ""]
			line <- as.numeric(line)
			table <- rbind(
				table,
				data.table::data.table(
					clump = line[1L],
					x = line[5L],
					y = line[7L],
					numberOfCells = line[4L],
					meanCellValue = line[2L],
					clumpSum = line[3L]
				)
			)
		
		}

		out[[i]] <- makeGVector(srcs[i], table = table)

	}

	if (nLayers == 1) out <- out[[1]]
	out
	
	} # EOF
)

#' Attach or detach GVector's data table
#'
#' @description `addTable()` adds an entire table to a `GVector`. It will replace any existing table. There must be one row in the table for each geometry (see [ngeom()]). You can also add a table column-by-column using the \code{\link[fasterRaster]{$<-}} operator.
#'
#' `dropTable()` removes a data table associated with a `GVector`.
#'
#' @param x A `GVector`.
#' @param ... Other arguments (ignored).
#' @param value A `data.frame`, `data.table`, or `matrix`.
#'
#' @returns A `GVector`.
#'
#' @seealso \code{\link[fasterRaster]{$<-}}, [colbind()], [rbind()], [as.data.frame()], [as.data.table()]
#'
#' @example man/examples/ex_cbind_rbind_addTable_dropTable.r
#'
#' @aliases addTable<-
#' @rdname addTable
#' @exportMethod addTable<-
methods::setMethod(
	f = "addTable<-",
	signature = c(x = "GVector", value = "data.frame"),
	function(x, ..., value) .addTable(x = x, table = value)
)

#' @aliases addTable<-
#' @rdname addTable
#' @exportMethod addTable<-
methods::setMethod(
	f = "addTable<-",
	signature = c(x = "GVector", value = "data.table"),
	function(x, ..., value) .addTable(x = x, table = value)
)

#' @aliases addTable<-
#' @rdname addTable
#' @exportMethod addTable<-
methods::setMethod(
	f = "addTable<-",
	signature = c(x = "GVector", value = "matrix"),
	function(x, ..., value) .addTable(x = x, table = value)
)

#' Add a data table to a `GVector`
#'
#' @param A `GVector`.
#' @param table A `data.frame`, `data.table`, or `matrix`.
#'
#' @noRd
.addTable <- function(x, table) {

	if (nrow(table) != ngeom(x)) stop("The GVector must have one geometry per row in the table.")
	if (!inherits(table, "data.table")) table <- data.table::as.data.table(table)
	x@table <- table
	methods::validObject(x)
	x

}

#' @aliases dropTable
#' @rdname addTable
#' @exportMethod dropTable
methods::setMethod(
	f = "dropTable",
	signature = c(x = "GVector"),
	function(x) {
	
	out <- x
	out@table <- data.table::data.table(NULL)
	out

	} # EOF
)

#' GRaster with values equal to row, column, coordinate, regular, or "chess"
#'
#' @description This function can be used to make a `GRaster` with cell values equal to the cell center's longitude, latitude, row, or column, or in a "chess"-like or "regular" pattern.
#'
#' @param x A `GRaster` to be used as a template.
#'
#' @param fun Character: Any of:
#' * `"x"` or `"y"`: Cell longitude or latitude
#' * `"row"` or `"col"`: Cell row or column
#' * `"chess"`: Alternating values.
#' * `"regular"`: Evenly-spaced cells with the same value.
#'
#' @param odd Logical: If `TRUE` (default), and `fun` is `"chess"`, then the top left cell in the raster will be a "negative" cell. If `FALSE`, then the top left cell with be "positive".
#'
#' @param vals Vector of two numeric values: If `fun` is `"chess"` or `"regular"`, then assign the first value to "positive" cells and the second value to "negative" cells. The default is `c(1, 0)`
#'
#' @returns A `GRaster` with as many layers as `x`.
#'
#' @seealso [terra::init()], [longlat()]
#'
#' @example man/examples/ex_init.r
#'
#' @aliases init
#' @rdname init
#' @exportMethod init
methods::setMethod(
	f = "init",
	signature = c(x = "GRaster"),
	function(x, fun, odd = TRUE, vals = c(0, 1)) {

	funs <- c("x", "y", "row", "col", "chess", "regular")
	fun <- omnibus::pmatchSafe(fun, funs, n = 1)

	# if (fun %in% c("chess", "regular") && (!omnibus::is.wholeNumber(every) | every < 1)) stop("The value of `every` must be a whole number > 0.")

	.locationRestore(x)
	.region(x)

	vals <- as.character(vals)
	vals[vals == "NA"] <- "null()"

	nLayers<- nlyr(x)
	srcs <- .makeSourceName(paste0("init_", fun), "raster", nLayers)
	for (i in seq_len(nLayers)) {
	
		if (fun == "chess") {
		
			if (odd) {
				ex <- paste0(srcs[i], " = if((row() % 2 == 1 & col() % 2 == 1) | (row() % 2 == 0 & col() % 2 == 0), ", vals[1L], ", ", vals[2L], ")")
			} else {
				ex <- paste0(srcs[i], " = if((row() % 2 == 0 & col() % 2 == 1) | (row() % 2 == 1 & col() % 2 == 0), ", vals[1L], ", ", vals[2L], ")")
			}
		
		} else if (fun == "regular") {
		
			if (odd) {
				ex <- paste0(srcs[i], " = if(row() % 2 == 1 | col() % 2 == 1, ", vals[1L], ", ", vals[2L], ")")
			} else {
				ex <- paste0(srcs[i], " = if(row() % 2 == 0 | col() % 2 == 0, ", vals[1L], ", ", vals[2L], ")")
			}
		
		} else if (fun != "chess") {
		
			ex <- paste0(srcs[i], " = ", fun, "()")

		}
		rgrass::execGRASS("r.mapcalc", expression = ex, flags = c(.quiet(), "overwrite"))

	}
	makeGRaster(srcs, fun)

	} # EOF
)

if (grassStarted()) {

# Setup
library(sf)

# Points vector of specimens of species in the plant genus Dypsis
madDypsis <- fastData("madDypsis")

# Convert sf to a GVector:
dypsis <- fast(madDypsis)

# Cluster:
dbscan <- clusterPoints(dypsis, method = "DBSCAN", maxDist = 10000)
dbscan2 <- clusterPoints(dypsis, method = "DBSCAN2", maxDist = 10000)
density <- clusterPoints(dypsis, method = "density", maxDist = 10000)
optics <- clusterPoints(dypsis, method = "optics", maxDist = 10000)
optics2 <- clusterPoints(dypsis, method = "optics2", maxDist = 10000)

oldpar <- par(mfrow = c(2, 3))
plot(dypsis, pch = 21, cex = 1.4, bg = dbscan, main = "DBSCAN")
plot(dypsis, pch = 21, cex = 1.4, bg = dbscan2, main = "DBSCAN2")
plot(dypsis, pch = 21, cex = 1.4, bg = density, main = "density")
plot(dypsis, pch = 21, cex = 1.4, bg = optics, main = "OPTICS")
plot(dypsis, pch = 21, cex = 1.4, bg = optics2, main = "OPTICS2")
par(oldpar)

}

#' Correlation between GRasters
#'
#' @description This function returns a correlation or covariance matrix between two or more `GRaster` layers. This function returns the sample correlation and covariance (i.e., the denominator is n - 1).
#' 
#' @param x A `GRaster` with two or more layers.
#' @param fun Character: Name of the statistic to calculate; either `"cor"` (default) or `"cov"`. 
#' 
#' @returns A numeric `matrix`.
#' 
#' @example man/examples/ex_layerCor.r
#'
#' @seealso [terra::layerCor()], [stats::cor()], [stats::cov()]
#' 
#' @aliases layerCor
#' @rdname layerCor
#' @exportMethod layerCor
methods::setMethod(
	f = "layerCor",
	signature = c(x = "GRaster"),
	function(x, fun = "cor") {
	
	if (nlyr(x) == 1L) stop("The raster must have >= 2 layers.")
	fun <- omnibus::pmatchSafe(fun, c("cor", "cov"))
	
	.locationRestore(x)
	.region(x)

	args <- list(
		cmd = "r.covar",
		map = paste(sources(x), collapse = ","),
		flags = c(.quiet()),
		intern = TRUE
	)

	if (fun == "cor") args$flags <- c(args$flags, "r")
	info <- do.call(rgrass::execGRASS, args = args)

	n <- substr(info[1L], 5L, nchar(info[1L]))
	n <- as.integer(n)

	nLayers <- nlyr(x)
	out <- matrix(NA_real_, ncol = nLayers, nrow = nLayers, dimnames = list(names(x), names(x)))

	for (i in seq_len(nLayers)) {

		this <- info[i + 1L]
		this <- strsplit(this, " ")[[1]]
		this <- as.numeric(this)
		out[i, ] <- this

	}
	attr(out, "n") <- n
	out	
	
	}
)

#' @title Classes for "fasterRaster" locations, rasters, and vectors
#'
#' @aliases GRegion
#' @rdname GLocation
#' @exportClass GRegion
GRegion <- methods::setClass(
	Class = "GRegion",
	contains = "GSpatial",
	slots = list(
		dimensions = "integer",         # 3 integers
		resolution = "numeric"          # 3 numerics
	),
	prototype = prototype(
		dimensions = c(NA_integer_, NA_integer_, NA_integer_),
		resolution = c(NA_real_, NA_real_, NA_real_)
	)
)

methods::setValidity("GRegion",
	function(object) {
		if (length(object@dimensions) != 3L) {
			"@dimensions must be three elements long."
		} else if (any(object@dimensions[1L:2L] <= 0L)) {
			"First two values in @dimensions must be positive integers."
		} else if (!is.na(object@dimensions[3L]) && object@dimensions[3L] <= 0L) {
			"Third value in @dimensions must be NA_integer_ or a positive integer."
		} else if (length(object@resolution) != 3) {
			"@resolution must be three values long."
		} else if (any(object@resolution[1L:2L] <= 0)) {
			"First two values in @resolution must be positive real values."
		} else if (!is.na(object@resolution[3L]) && object@resolution[3L] <= 0) {
			"Third value in @resolution must be NA_real_ or a positive real value."
		} else {
			TRUE
		}

	} # EOF
)

#' Select parts of a polygon GVector erase shared by another polygon GVector
#'
#' @description The `erase()` function removes from the `x` "polygons" `GVector` parts that overlap with the `y` "polygons" `GVector`. You can also use the `-` operator (e.g., `vect1 - vect2`).
#'
#' @param x,y `GVector`s.
#'
#' @returns A `GVector`.
#'
#' @seealso [c()], [aggregate()], [crop()], [intersect()], [union()], [xor()]
#' 
#' @example man/examples/ex_union_intersect_xor_erase.r
#'
#' @aliases erase
#' @rdname erase
#' @exportMethod erase
methods::setMethod(
	f = "erase",
	signature = c(x = "GVector", y = "GVector"),
	function(x, y) {

	compareGeom(x, y, geometry = TRUE)
	if (geomtype(x) != "polygons") stop("Only polygon GVectors can be notted.")
	.locationRestore(x)
		
	src <- .makeSourceName("v_overlay", "vector")
	rgrass::execGRASS(
		cmd = "v.overlay",
		ainput = sources(x),
		binput = sources(y),
		output = src,
		operator = "not",
		snap = -1,
		flags = c(.quiet(), "overwrite")
	)

	makeGVector(src)
	
	} # EOF
)

#' @title GRASS "location" of an object or the active session
#'
#' @description **GRASS** "projects" or "locations" are sets of one or more rasters and/or vectors with the same coordinate reference systems, and may or may not represent the same actual location on Earth. **GRASS** "mapsets" are like sub-folders of locations, and are collections of rasters and/or vectors typically related to the same general project. Typical users will not need to make changes to the default location (called "location") or mapset (called "PERMANENT"). See `vignette("projects_mapsets", package = "fasterRaster")`.
#'
#' @param x Either:
#' * Missing: Reports location of currently active project/location.
#'  * A `GLocation` object or an object that contains the `GLocation` class (i.e., a `GSpatial` object: a `GRaster` or `GVector`).
#'
#' @return A character string.
#'
#' @seealso [.mapset()]
#'
#' @aliases .location
#' @rdname location
#' @keywords internal
methods::setMethod(
	f = ".location",
	signature = "GLocation",
	definition = function(x) unname(x@location)
)

#' @aliases .location
#' @rdname location
#' @keywords internal
methods::setMethod(
	f = ".location",
	signature = "missing",
	definition = function(x) .fasterRaster$activeLocation
)

---
title: "For developers: Hidden fasterRaster functions"
output: rmarkdown::html_vignette
author: "Adam B. Smith"
date: "`r Sys.Date()`"
vignette: >
  %\VignetteIndexEntry{For developers: Hidden fasterRaster functions}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
fig.path = 'man/figures/'
```

**fasterRaster** contains a set of "private" functions that users can access using `fasterRaster:::functionName`. These functions are useful for power users and developers. Not all hidden functions are listed here. Often, a method will have a hidden function of the same name that starts with a period (e.g., `.plot()`). This "period" function is intended to be supplied the [`sources()`](https://adamlilith.github.io/fasterRaster/reference/sources.html) name of a `GRaster` or `GVector` from other functions so that the calling function does not need to spend the time creating the `GRaster` or `GVector` pointer before calling the function. "Period" functions will, though, often work on `GRaster`s or `GVector`s, though some error-checking and region re-definition is not conducted.

## General
* `.addons()`: Tests if an addon is installed, and if not, attempts to install it.
* `.backdoor()`: Calls [faster()] and sets **GRASS** folder to "`C:/Program Files/GRASS GIS X.Y`", plus other options useful for development.
* `.fileExt()`: Get file extension
* `.ls()`: Lists the `sources` of all objects in the active **GRASS** "project/location"
* `.message()`: Display a warning or message if the given warning has not been displayed since **fasterRaster** was attached or if a given number of hours has passed
* `.quiet()`: Returns "quiet" if `faster("debug")` is `TRUE`
* `.workDir()`: Working directory of a `GLocation` object

## Rasters and vectors
* `.copyGSpatial()`: Make a copy of the **GRASS** file pointed to by a `GRaster` or `GVector`
* `.exists()`: Does the **GRASS** file of a `GRaster` or `GVector` exist?
* `.ext()`: Extent from the [`sources()`](https://adamlilith.github.io/fasterRaster/reference/sources.html) name of a `GRaster` or `GVector`
* `.makeSourceName()`: Makes one or more statistically unique strings that can be used as file names to represent rasters or vectors in **GRASS**
* `.plot()`: Plot using the [`sources()`](https://adamlilith.github.io/fasterRaster/reference/sources.html) name of a `GRaster` or `GVector`
* `.projection()`: Value of the `@projection` slot in a `GRaster` or `GVector`
* `.rastInfo()` and `.vectInfo()`: Metadata for a **GRASS** raster or vector
* `.rename()`: Rename a **GRASS** raster or vector
* `.rm()`: Delete rasters or vectors in **GRASS**

## Vectors
* `.aggDisaggVect()`: Aggregate or disaggregate a vector using its [`sources()`](https://adamlilith.github.io/fasterRaster/reference/sources.html) name.
* `.geomtype()`: Geometry type ("point", "line", or "area") from the [`sources()`](https://adamlilith.github.io/fasterRaster/reference/sources.html) name of a `GVector`
* `.validVector()`: Test if a `GVector` is valid.
* `.vAsDataTable()`: Convert the attribute table linked to a vector in **GRASS** to a `data.table`. This table is distinct from the attribute table attached to a `GVector`
* `.vAttachDatabase()`: Add a database table to the **GRASS** representation of a `GVector`
* `.vCats()`: Get a `data.table` with a single column named `cat`, which corresponds to the **GRASS** attribute table's `cat` column
* `.vDetachDatabase()`: Detach the **GRASS** database from a **GRASS** vector
* `.vHasDatabase()`: Tests if **GRASS** vector has a database
* `.vIncrementCats()`: Increment category values of a `GVector`
* `.vNames()`: "**GRASS**" vector attribute table column names
* `.vRecat()`: Change **GRASS** category indices of a **GRASS** vector
* `.vValidCats()`: Are category values of a vector valid?

## Rasters
* `.layerIndex()`: Gets the index of `GRaster` layers from a numeric, integer, character, or logical vector
* `.minVal()` and `.maxVal()`: Values in the `@minVal` and `@maxVal` slots in a `GRaster`
* `.nlevels()`: Number of levels in a `SpatVector`, `data.frame`, `data.table`, empty string, or a list of `data.frame`s, `data.table`s, and/or empty strings.

## **GRASS** "projects/locations" and "mapsets"
* `.locationCreate()` Make a connection to **GRASS** (i.e., start **GRASS** from within **R**) and create a location
* `.locationDelete()` Deletes all files associated with a **GRASS** "location" and mapset
* `.locationFind()`: Find a specific **GRASS** "location" that already exists
* `.locationRestore()` Reconnect **GRASS** to a previously-created **GRASS** "location"
* `.locations()`: List of all available "locations"
* `.g.proj()`: Runs **GRASS** `g.proj` tool to display projection of current "project"
* `.g.region()`: Runs **GRASS** `g.region` tool to display region of current "project"
* `.mapset()`: **GRASS** "mapset" of an object or the active session

## **GRASS** "regions"
* `.region()`: Change or report the active region's extent and resolution
* `.regionDim()]`: Change or report the active region's resolution (also [`dim()`](https://adamlilith.github.io/fasterRaster/reference/dim.html) and related functions, with no arguments)
* `.regionExt()`: Change or report the active region's extent (also [`ext()`](https://adamlilith.github.io/fasterRaster/reference/ext.html) and related functions, with no arguments)
* `.regionRes()`: Change or report the active region's dimensions (also [`res()`](https://adamlilith.github.io/fasterRaster/reference/res.html) and related functions, with no arguments)

~ FINIS ~

---
title: "Types of GRasters"
output: rmarkdown::html_vignette
author: "Adam B. Smith"
date: "`r Sys.Date()`"
vignette: >
  %\VignetteIndexEntry{Types of GRasters}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
fig.path = 'man/figures/'
```

**fasterRaster** `GRaster`s can represent double-floating point numeric values, integers, or categorical/factor data.

## Double-floating point value
Double-floating point values are accurate to about the 15th to 17th decimal place. These are called "double" rasters in **fasterRaster** and `DCELL` rasters in **GRASS**. These rasters typically take the most memory. All `numeric` values in **R** are double-floating point values.

## Floating point value
Less common that double-floating point rasters, floating point rasters are accurate to about the 7th decimal place. These are called "float" rasters in **fasterRaster** and `FCELL` rasters in **GRASS**. These rasters typically take less memory than double-floating point rasters.

## Integer
Rasters that represent integers are called "integer" rasters in **fasterRaster** and `CELL` rasters in **GRASS**. You can force a raster to be an integer using [`as.int()`](https://adamlilith.github.io/fasterRaster/reference/as.int.html). Some of the functions in [`app()`](https://adamlilith.github.io/fasterRaster/reference/app.html) function will also return integer-type rasters. Integer rasters typically take the least memory.

## Categorical/factor
Categorical rasters (also called "factor" rasters) are actually integer rasters, but have an associated attribute table that maps each integer value to a category label, such as "wetland" or "forest". The table has at least two columns. The first is integer values, and (by default) the second is category names. This second column is the "active" category column that is used for plotting and in some functions. The active column can be changed using [`activeCat<-`](https://adamlilith.github.io/fasterRaster/reference/activeCat.html).

## Functions relevant to raster data types

### Any type of `GRaster`
* [`as.int()`](https://adamlilith.github.io/fasterRaster/reference/as.int.html), [`as.float()`](https://adamlilith.github.io/fasterRaster/reference/as.int.html), and [`as.doub()`](https://adamlilith.github.io/fasterRaster/reference/as.int.html) coerce a raster to an integer, float, or double.
* [`datatype()`](https://adamlilith.github.io/fasterRaster/reference/datatype.html) returns the data type of a `GRaster`.
* [`freq()`](https://adamlilith.github.io/fasterRaster/reference/freq.html): Frequency of each category across cells of a raster
* [`is.factor()`](https://adamlilith.github.io/fasterRaster/reference/is.factor.html) indicates if the raster is a categorical raster.
* [`is.int()`](https://adamlilith.github.io/fasterRaster/reference/is.int.html), [`is.float()`](https://adamlilith.github.io/fasterRaster/reference/is.int.html), and [`is.doub()`](https://adamlilith.github.io/fasterRaster/reference/is.int.html) indicate if values in a a raster are integers, floating-point, or double-floating point precision.

### Categorical `GRaster`s

* [`activeCat()`](https://adamlilith.github.io/fasterRaster/reference/activeCat.html), [`activeCats()`](https://adamlilith.github.io/fasterRaster/reference/activeCat.html), and [`activeCat<-`](https://adamlilith.github.io/fasterRaster/reference/activeCat.html) can be used to see or assign which column in a "levels" table associated with a categorical raster is used as category labels.
* [`addCats()`](https://adamlilith.github.io/fasterRaster/reference/addCats.html) adds information to the "levels" table using `data.table::merge()` (same as `merge()`).
* [`addCats<-`](https://adamlilith.github.io/fasterRaster/reference/addCats.html) add new levels to a "levels" table.
* [`catNames()`](https://adamlilith.github.io/fasterRaster/reference/catNames.html) reports the column names of the "levels" table of each layer of a raster.
* [`cats()`](https://adamlilith.github.io/fasterRaster/reference/cats.html) returns the entire "levels" table of a categorical raster.
* [`combineLevels()`](https://adamlilith.github.io/fasterRaster/reference/combineLevels.html): Combine the "levels" tables of two or more categorical `GRaster`s.
* [`complete.cases()`](https://adamlilith.github.io/fasterRaster/reference/complete.cases.html) finds rows in the levels table that have no `NA`s.
* [`concats()`](https://adamlilith.github.io/fasterRaster/reference/combineCats.html) combines levels of two or more categorical or integer rasters by concatenating them.
* [`droplevels()`](https://adamlilith.github.io/fasterRaster/reference/droplevels.html) removes "unused" levels in a "levels" table.
* [`levels()`](https://adamlilith.github.io/fasterRaster/reference/levels.html) returns the "levels" table of a categorical raster (just the value column and the active column).
* [`levels<-`](https://adamlilith.github.io/fasterRaster/reference/levels.html) and [`categories()`](https://adamlilith.github.io/fasterRaster/reference/categories.html) can be used to assign categories to an integer raster and make it categorical (i.e., a "factor" raster).
* [`match()`](https://adamlilith.github.io/fasterRaster/reference/match.html), [`%in%`](https://adamlilith.github.io/fasterRaster/reference/match.html), and [`$%notin%$`](https://adamlilith.github.io/fasterRaster/reference/match.html): Find which cells of a `GRaster` match or do not match certain category labels
* [`missing.cases()`](https://adamlilith.github.io/fasterRaster/reference/missing.cases.html) finds rows in the levels table that have at least one `NA`.
* [`missingCats()`](https://adamlilith.github.io/fasterRaster/reference/missingCats.html) finds values in categorical rasters that do not have a category assigned to them.
* [`nlevels()`](https://adamlilith.github.io/fasterRaster/reference/nlevels.html) returns the number of levels represented by a categorical raster.
* [`subst()`](https://adamlilith.github.io/fasterRaster/reference/subst.html): Re-assign category levels

## Saving rasters to disk
You can potentially save substantial space on disk by setting the `datatype` argument in [`writeRaster()`](https://adamlilith.github.io/fasterRaster/reference/writeRaster.html) to an appropriate value when saving a raster. This argument allows for finer "divisions" than just integer/float/double-float, so depending on the range of values in your raster, you can optimize file size by selecting the one that best matches the values in the raster. See the documentation for [`writeRaster()`](https://adamlilith.github.io/fasterRaster/reference/writeRaster.html) for more information.


~ FINIS ~

---
title: "Making fasterRaster faster"
output: rmarkdown::html_vignette
author: "Adam B. Smith"
date: '`r format(Sys.Date())`'
vignette: >
  %\VignetteIndexEntry{Making fasterRaster faster}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
fig.path = 'man/figures/'
```

There are several ways to speed up **fasterRaster** functions. These are listed below in order of their most likely gains, with the first few being potentially the largest.

1. **Load rasters and vectors directly from disk**: Use [`fast()`](https://adamlilith.github.io/fasterRaster/reference/fast.html) to load rasters and vectrors directly from disk. Converting `terra` or `sf` objects to `GRaster`s and `GVector`s can be slower. Why? Because if the object does not have a file to which the `R` object points, Use [`fast()`](https://adamlilith.github.io/fasterRaster/reference/fast.html) has to save it to disk first as a GeoTIFF or GeoPackage file, then load it into `GRASS`.

2. **Save `GRaster`s and `GVector`s directly to disk**: Converting `GRaster`s and `GVector`s to `SpatRaster`s or `SpatVector` using [`rast()`](https://adamlilith.github.io/fasterRaster/reference/rast.html) or [`vect()`](https://adamlilith.github.io/fasterRaster/reference/vect.html), then saving them is much slower than just saving them. Why? Because these functions actually save the file to disk then uses the respective function from the respective package to connect to the file.

3. **Increase memory and the number of cores usable by GRASS:** By default, `fasterRaster` use 2 cores and 2048 MB (2 GB) of memory for `GRASS` modules that allow users to specify these values. You can set these to higher values using [`faster()`](https://adamlilith.github.io/fasterRaster/reference/faster.html) and thus potentially speed up some calculations. Functions in newer versions of `GRASS` have more capacity to use these options, so updating `GRASS` to the latest version can help, too.

4. **Do operations on `GRaster`s and `GVector`s in the same coordinate reference system together:** Every time you switch between using a `GRaster` or `GVector` with a different coordinate reference system (CRS), `GRASS` has to spend a few seconds changing to that CRS. You can save some time by doing as much work as possible with objects in one CRS, then switching to work on objects in another CRS.

~ FINIS ~

---
title: "3-dimensional objects"
output: rmarkdown::html_vignette
author: "Adam B. Smith"
date: "`r Sys.Date()`"
vignette: >
  %\VignetteIndexEntry{3-dimensional objects}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
fig.path = 'man/figures/'
```
**fasterRaster** has limited support for 3-dimensional `GRaster`s and `GVector`s. A `GRaster` is like a multi-layered raster, except that layers contain values in "voxels" (3-dimensional pixels), where each layer has a z-resolution, in addition to its normal x- and y-resolutions. These layers are called "depths". A 3-dimensional vector has a z-dimension, so any point (i.e., an actual point, or points that together define a line or polygon) has a third dimension. You can test if an object is 3-dimensional using `topology()`, `is.2d()`, or `is.3d()`, or by looking at its metadata (i.e., just enter the name of the object in **R** and look at what is displayed).

Thus far, *no functions that handle 3-dimensional `GRaster`s or `GVector`s have been tested.* Some functions like `spatSample()` have arguments that allow you to do something with a z-component. Also, some of the "getter" functions can report information about the third dimension of a `GRaster` or `GVector`.

Three-dimensional `GRaster`s and `GVector`s should (?... remember, no testing yet!) work with most functions. In some cases, the third dimension will be ignored or removed. You should get a warning in these cases.

## **fasterRaster** functions that specifically work with 3-dimensional objects

   | **Function**              | **Object type**          |
   | --------------------------|------------------------- |
   | `bottom()`                | `GRaster` or `GVector`   |
   | `dim3d()`                 | `GRaster`                |
   | `is.2d()`                 | `GRaster` or `GVector`   |
   | `is.3d()`                 | `GRaster` or `GVector`   |
   | `ncell3d()`               | `GRaster`                |
   | `ndepth()`                | `GRaster`                |
   | `res3d()`                 | `GRaster`                |
   | `top()`                   | `GRaster` or `GVector`   |
   | `topology()`              | `GRaster` or `GVector`   |
   | `zext()`                  | `GRaster` or `GVector`   |
   | `zres()`                  | `GRaster`                |

~ FINIS ~

---
title: "For developers: Locations/projects and mapsets"
output: rmarkdown::html_vignette
author: "Adam B. Smith"
date: "`r Sys.Date()`"
vignette: >
  %\VignetteIndexEntry{For developers: Locations/projects and mapsets}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
fig.path = 'man/figures/'
```

**GRASS** uses "projects" (which used to be called "locations" before **GRASS** 8.4), and "mapsets" to store files (rasters, vectors, etc.). **fasterRaster** uses projects and mapsets, too, but in a manner that is invisibly to most users. Thus, this tutorial is mostly of interest to developers and other curious people.

## **GRASS** locations/projects
Upon starting, **GRASS** creates (or loads) a project, which corresponds to a folder on the user's system. Importantly, all rasters and vectors in a project must have the same coordinate reference system (CRS). Confusingly, rasters and vectors in the same **GRASS** location do not necessarily have to represent the same place on Earth (this the renaming to "project"). In general, rasters and vectors can only interact with one another if they are in the same project and mapset.  **GRASS** can only have a connection to a single project at a time.

**fasterRaster** handles projects and mapsets automatically, so users typically do not need to manage them. Projects are created on an as-needed basis. Within a given **R** session, if no projects have already been made, the first call to [`fast()`](https://adamlilith.github.io/fasterRaster/reference/fast.html) to create a `GRaster` or `GVector` will 1) make a connection to **GRASS** and 2) create a location with a CRS the same as the raster or vector. The raster or vector is then stored in this location. `fast()` starts the connection and creates the location using the exported by hidden function `.locationCreate()`.

If `fast()` is called and a project already exists that has the same CRS as the raster or vector, one of two things will happen. First, if **GRASS** is already connected to that project, the raster or vector is simply imported. Second, if **GRASS** is not connected to the project that has the appropriate CRS, it will use the hidden function `.locationRestore()` to connect to the proper one, then import the raster or vector.

Other functions might also cause **GRASS** to connect to a pre-existing project Generally, if a function is applied to a `GRaster` or `GVector`, it will first check to see that **GRASS** is connected to the project in which the raster or vector is stored. If not, it will use `.locationRestore()` to do so first.

Creating a new project or switching connections to pre-existing project adds a few seconds to processing time of rasters and vectors. To avoid this, users can work as much as possible on sets of rasters and vectors with the same CRS (i.e., the same location).

## **fasterRaster** mapsets

**GRASS** "mapsets" are sub-folders within a location. Every location *must* have a mapset (and thus, a sub-folder) named "PERMANENT". Users of mapsets are intended to store sub-projects that use rasters and vectors with the same CRS. Users of **GRASS** can switch between mapsets. However, for ease-of-use and development, **fasterRaster** always uses the "PERMANENT" mapset within a given location.

## Functions that manage locations and mapsets

The `crs()` function (wth no arguments) can be used to get the coordinate reference system of the current project/location.

The followoinng functions functions are hidden, but documented. All functions take "`x`" as an argument.

* `.location(x)`: Name of the location of `x` (a **fasterRaster** object), or the current location (if `x` is missing).
* `.locations()`: Names and CRSs of all available locations (takes no arguments).
* `.locationCreate(x)`: Create a location with the same CRS as `x`.
* `.locationRestore(x)`: Connect to a pre-existing location. `x` can be a **fasterRaster** object, a `SpatRaster`, `SpatVector`, or `sf` object, or the name of a location.
* `.locationFind(x, return = <option>)`: Given a `GSpatial` object `x`, find the name, index, or CRS of a location that matches it. If `x` is missing, return a list of all locations. `<option>` can be `"name"` (name of the location), `"index"` (index), or "`crs`" (coordinate reference string in WKT format).
* `.ls()`: List all file names of rasters and/or vectors in the current location.
* `.mapset(x)`: Name of the mapset that contains the object `x`.

## More rabbit-holing

Upon being loaded or attached, **fasterRaster** creates a package-specific environment named `.fasterRaster`, where it keeps a list of projects at `.fasterRaster$locations`. This is a named list of projects, where each element is has the name of the location, and the value of the element is the location's CRS. The current location is tracked at `.fasterRaster$activeLocation`.


~ FINIS ~

---
title: "Getting started with fasterRaster"
output: rmarkdown::html_vignette
author: "Adam B. Smith"
date: "`r Sys.Date()`"
vignette: >
  %\VignetteIndexEntry{Getting started with fasterRaster}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
resource_files:
  - man/figures/elev_rivers.png
  - man/figures/dist_to_rivers.png
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
fig.path = 'man/figures/'
```
**fasterRaster** interfaces with **GRASS** to process rasters and spatial vector data. It is intended as an add-on to the **terra** and **sf** packages, and relies heavily upon them. For most rasters and vectors that are small or medium-sized in memory/disk, those packages will almost always be faster. They may also be faster for very large objects.  But when they aren't, **fasterRaster** can step in.

## Installing **fasterRaster**

You probably already have **fasterRaster** installed on your computer, but if not, you can install the latest release version from CRAN using:
```{r install, eval = FALSE}
install.packages("fasterRaster")
```
or the latest development version using:
```{r install_dev, eval = FALSE}
remotes::install_github("adamlilith/fasterRaster", dependencies = TRUE)
```
(You may need to install the `remotes` package first.)

## Installing **GRASS**

**fasterRaster** uses **GRASS** to do its operations. You will need to install **GRASS** using the "stand-alone" installer, available through the [GRASS](https://grass.osgeo.org/). *Be sure to use the "stand-alone" installer, not the "OSGeo4W" installer!*

Optional: A few functions in **fasterRaster** require **GRASS** "addon" modules, which do not come bundled with **GRASS**. You do not need to install these addons if you do not use functions that call them. A list of functions that require addons can be seen in the "addons" vignette (in **R**, use `vignette("addons", package = "fasterRaster")`). This vignette also explains how to install addons.

## Starting a **fasterRaster** session

I recommend attaching the **data.table**, **terra**, and **sf** packages before attaching **fasterRaster** package to avoid function conflicts. The **data.table** package is not required, but you most surely will use at least one of the other two.
```{r packages}
library(terra)
library(sf)
library(data.table)
library(fasterRaster)
```
To begin, you need to tell **fasterRaster** the full file path of the folder where **GRASS** is installed on your system. Where this is well depend on your operating system and the version of **GRASS** installed.  Three examples below show you what this might look like, but you may need to change the file path to match your case:

```{r grassDir_examples, eval = FALSE}
grassDir <- "C:/Program Files/GRASS GIS 8.4" # Windows
grassDir <- "/Applications/GRASS-8.4.app/Contents/Resources" # Mac OS
grassDir <- "/usr/local/grass" # Linux
```
```{r grassDir, echo = FALSE}
grassDir <- "C:/Program Files/GRASS GIS 8.4" # Windows
```
To tell **fasterRaster** where **GRASS** is installed, use the `faster()` function:

```{r faster_grassDir, eval = FALSE}
faster(grassDir = grassDir)
```
You can also use the [`faster()`](https://adamlilith.github.io/fasterRaster/reference/faster.html) function to set options that affect how **fasterRaster** functions run. This includes setting the amount of maximum memory and number of computer cores allocated to operations.

Importantly, some **fasterRaster** functions require addons (**GRASS** modules that are optional and are not included with teh **GRASS** download). To see how to set up addons so functions that use them can operate, see `vignette("addons", package = "fasterRaster").

## Importing spatial objects into **fasterRaster** `GRaster`s and `GVector`s

In **fasterRaster**, rasters are called `GRaster`s and vectors are called `GVector`s. The easiest (but not always fastest) way to start using a `GRaster` or `GVector` is to convert it from one already in **R**. In the example below, we use a raster that comes with the **fasterRaster** package. The raster represents elevation of a portion of eastern Madagascar. We first load the `SpatRaster` using [`fastData()`](https://adamlilith.github.io/fasterRaster/reference/fastData.html), a helper function for loading example data objects that come with the **fasterRaster** package.
```{r madElev, eval = FALSE}
madElev <- fastData("madElev") # example SpatRaster
madElev

class       : SpatRaster 
dimensions  : 1024, 626, 1  (nrow, ncol, nlyr)
resolution  : 59.85157, 59.85157  (x, y)
extent      : 731581.6, 769048.6, 1024437, 1085725  (xmin, xmax, ymin, ymax)
coord. ref. : Tananarive (Paris) / Laborde Grid 
source      : madElev.tif 
name        : madElev 
min value   :       1 
max value   :     570
```

Now, we do the conversion to a `GRaster` and a `GVector` using [`fast()`](https://adamlilith.github.io/fasterRaster/reference/fast.html). This function can create a `GRaster` or `GVector` from a `SpatRaster` or a file representing a raster.
```{r elev, eval = FALSE}
elev <- fast(madElev)
elev

class       : GRaster
topology    : 2D 
dimensions  : 1024, 626, NA, 1 (nrow, ncol, ndepth, nlyr)
resolution  : 59.85157, 59.85157, NA (x, y, z)
extent      : 731581.552, 769048.635, 1024437.272, 1085725.279 (xmin, xmax, ymin, ymax)
coord ref.  : Tananarive (Paris) / Laborde Grid 
name(s)     : madElev 
datatype    : integer 
min. value  :       1 
max. value  :     570
```

Converting rasters and vectors that are already in **R** to `GRaster`s usually takes more time than loading them directly from disk. To load from disk, simply replace the first argument in `fast()` with a string representing the folder path and file name of the raster you want to load into the session. For example, you can do:

```{r elev_from_file, eval = FALSE}
rastFile <- system.file("extdata", "madElev.tif", package = "fasterRaster")
elev2 <- fast(rastFile)
```

Now, let's create a `GVector`. The `fast()` function can take a `SpatVector` from the **terra** package, an `sf` object from the **sf** package, or a string representing the file path and file name of a vector file (e.g., a GeoPackage file or a shapefile).
```{r madRivers, eval = FALSE}
madRivers <- fastData("madRivers") # sf vector
madRivers

Simple feature collection with 11 features and 5 fields
Geometry type: LINESTRING
Dimension:     XY
Bounding box:  xmin: 731627.1 ymin: 1024541 xmax: 762990.1 ymax: 1085580
Projected CRS: Tananarive (Paris) / Laborde Grid
First 10 features:
       F_CODE_DES          HYC_DESCRI      NAM ISO     NAME_0                       geometry
1180 River/Stream Perennial/Permanent MANANARA MDG Madagascar LINESTRING (739818.2 108005...
1185 River/Stream Perennial/Permanent MANANARA MDG Madagascar LINESTRING (739818.2 108005...
1197 River/Stream Perennial/Permanent      UNK MDG Madagascar LINESTRING (747857.8 108558...
1216 River/Stream Perennial/Permanent      UNK MDG Madagascar LINESTRING (739818.2 108005...
1248 River/Stream Perennial/Permanent      UNK MDG Madagascar LINESTRING (762990.1 105737...
1256 River/Stream Perennial/Permanent      UNK MDG Madagascar LINESTRING (742334.2 106858...
1257 River/Stream Perennial/Permanent      UNK MDG Madagascar LINESTRING (731803.7 105391...
1264 River/Stream Perennial/Permanent      UNK MDG Madagascar LINESTRING (755911.6 104957...
1300 River/Stream Perennial/Permanent      UNK MDG Madagascar LINESTRING (731871 1044531,...
1312 River/Stream Perennial/Permanent      UNK MDG Madagascar LINESTRING (750186.1 103441...
```

```{r rivers, eval = FALSE}
rivers <- fast(madRivers)
rivers

class       : GVector
geometry    : 2D lines 
dimensions  : 11, 11, 5 (geometries, sub-geometries, columns)
extent      : 731627.0998, 762990.1321, 1024541.23477, 1085580.45359 (xmin, xmax, ymin, ymax)
coord ref.  : Tananarive (Paris) / Laborde Grid 
names       :   F_CODE_DES      HYC_DESCRI      NAM   ISO     NAME_0 
type        :        <chr>           <chr>    <chr> <chr>      <chr> 
values      : River/Stream Perennial/Perm~ MANANARA   MDG Madagascar 
              River/Stream Perennial/Perm~ MANANARA   MDG Madagascar 
              River/Stream Perennial/Perm~      UNK   MDG Madagascar 
              (and 8 more rows) 
```

## Operations on `GRaster`s and `GVector`s

You can do operations on `GRaster`s and `GVector`s as if they were `SpatRaster`s, `SpatVector`s, and `sf` objects. For example, you plot them as if the were any other spatial object:
```{r how_to_plot, eval = FALSE}
plot(elev)
plot(rivers, col = 'lightblue', add = TRUE)
```
![Elevation and rivers](./images/elev_rivers.png)

You can use mathematical operators and functions:

```{r multiplication, eval = FALSE}
elev_feet <- elev * 3.28084
elev_feet

class       : GRaster
topology    : 2D 
dimensions  : 1024, 626, NA, 1 (nrow, ncol, ndepth, nlyr)
resolution  : 59.85157, 59.85157, NA (x, y, z)
extent      : 731581.552, 769048.635, 1024437.272, 1085725.279 (xmin, xmax, ymin, ymax)
coord ref.  : Tananarive (Paris) / Laborde Grid 
name(s)     :    layer 
datatype    :   double 
min. value  :   3.2808 
max. value  : 1870.056
```

```{r log, eval = FALSE}
log10_elev <- log10(elev)
log10_elev

class       : GRaster
topology    : 2D 
dimensions  : 1024, 626, NA, 1 (nrow, ncol, ndepth, nlyr)
resolution  : 59.85157, 59.85157, NA (x, y, z)
extent      : 731581.552, 769048.635, 1024437.272, 1085725.279 (xmin, xmax, ymin, ymax)
coord ref.  : Tananarive (Paris) / Laborde Grid 
name(s)     :              log 
datatype    :           double 
min. value  :                0 
max. value  : 2.75587485567249
```

You can also use the many **fasterRaster** functions. In general, these functions have the same names as their **terra** counterparts and often the same arguments. Note that even many **terra** and **fasterRaster** functions have the same name, they do not necessarily produce the exact same output. Much care has been taken to ensure they do, but sometimes there are multiple ways to do the same task, so choices made by the authors of **terra** and **GRASS** can lead to differences.

The following code 1) creates a raster where cell values reflect the distance between them and the nearest river; b) makes a buffer around the rivers; then c) plots the output:
```{r distance_buffers, eval = FALSE}
dist <- distance(elev, rivers)
dist

class       : GRaster
topology    : 2D 
dimensions  : 1024, 626, NA, 1 (nrow, ncol, ndepth, nlyr)
resolution  : 59.85157, 59.85157, NA (x, y, z)
extent      : 731581.552, 769048.635, 1024437.272, 1085725.279 (xmin, xmax, ymin, ymax)
coord ref.  : Tananarive (Paris) / Laborde Grid 
name(s)     :         distance 
datatype    :           double 
min. value  :                0 
max. value  : 21310.9411762729 
```
```{r buffer, eval = FALSE}

river_buff <- buffer(rivers, 2000)
river_buff

class       : GVector
geometry    : 2D polygons 
dimensions  : 1, 5, 0 (geometries, sub-geometries, columns)
extent      : 729629.19151, 764989.97343, 1022544.92079, 1087580.24979 (xmin, xmax, ymin, ymax)
coord ref.  : Tananarive (Paris) / Laborde Grid 

```
```{r plot_rivers_buffer, eval = FALSE}
plot(dist)
plot(rivers, col = 'lightblue', add = TRUE)
plot(river_buff, border = 'white', add = TRUE)
```
![Distance between each cell and nearest major river](./images/dist_to_rivers.png)

And that's how you get started! Now that you have a raster and a vector in your **fasterRaster** "location", you can start doing manipulations and analyses using any of the **fasterRaster** functions!  To see an annotated list of these functions, use `?fasterRaster`.

## Converting and saving `GRaster`s and `GVector`s

You can convert a `GRaster` to a `SpatRaster` raster using [`rast()`](https://adamlilith.github.io/fasterRaster/reference/rast.html):
```{r rast, eval = FALSE}
terra_elev <- rast(elev)
```

To convert a `GVector` to the **terra** package's `SpatVector` format or to an `sf` vector, use [`vect()`](https://adamlilith.github.io/fasterRaster/reference/vect.html) or [`st_as_sf()`](https://adamlilith.github.io/fasterRaster/reference/st_as_sf.html):
```{r vect, eval = FALSE}
terra_rivers <- vect(rivers)
sf_rivers <- st_as_sf(rivers)
```


Finally, you can use [`writeRaster()`](https://adamlilith.github.io/fasterRaster/reference/writeRaster.html) and [`writeVector()`](https://adamlilith.github.io/fasterRaster/reference/writeVector.html) to save `GRaster`s and `GVector`s directly to disk. This will always be faster than using `rast()`, `vect()`, or `st_as_sf()` then saving the result from those functions.
```{r write, eval = FALSE}
elev_temp_file <- tempfile(fileext = ".tif") # save as GeoTIFF
writeRaster(elev, elev_temp_file)

vect_temp_file <- tempfile(fileext = ".shp") # save as shapefile
writeVector(rivers, vect_temp_file)
```

# Known issues

* Comparability between **terra** and **fasterRaster**: As much as possible, **fasterRaster** functions were written to recreate the output that functions in **terra** produce.  However, owing to implementation choices made by the respective developers of **terra** and **GRASS**, outputs are not always the same.

* **fasterRaster** can crash when the temporary folder is cleaned: Some operating systems have automated procedures that clean out the system's temporary folders when they get too large. This can remove files **GRASS** is using and **fasterRaster** is pointing to, rendering them broken. In Windows, this setting can be changed by going to `Settings`, then `Storage`, then `Storage Sense`. Turn off the setting "Keep Windows running smoothly by automatically cleaning up temporary system and app files".

* Disk space fills up: As counter to the previous issue, prolonged use of **fasterRaster** by the same **R** process can create a lot of temporary files in the **GRASS** cache that fills your hard drive. **fasterRaster** does its best to remove these files when they are not needed. However, temporary files can still accumulate. For example, the operation `new_raster <- 2 * old_raster^3` creates a raster file with the `^3` operation, which is then multiplied by 2 to get the desired output. The raster from the `^3` operation is still left in the disk cache, even though it does not have a "name" in **R**. Judicious use of the [`mow()`](https://adamlilith.github.io/fasterRaster/reference/mow.html) function can remove these temporary files.

~ FINIS ~

---
title: "For developers: fasterRaster and GRASS Regions"
output: rmarkdown::html_vignette
author: "Adam B. Smith"
date: "`r Sys.Date()`"
vignette: >
  %\VignetteIndexEntry{For developers: fasterRaster and GRASS Regions}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
fig.path = 'man/figures/'
```
A **GRASS** *region* is a data structure like a raster in that it is composed of a grid, but different in that "cells" in this grid do not contain values. Rather, their resolution and the extent of the region influence how rasters are imported, created, processed, and exported. In most cases, whenever a raster undergoes one of these processes using a **GRASS** tool, the raster will be resampled and/or crop/extend it so that matches the region"s extent and resolution. If ignored, this can cause unintended side effects if the region's geometry doesn't match the raster being processed.

Generally, most users of **fasterRaster** will not need to know how regions work because their management is handled automatically.  This help page is provided to assist power users who may wish to use regions explicitly or develop their own applications based on **fasterRaster**.

Each **GRASS** "location" typically has one active region. The extent and resolution of this region is initially set by the first raster that is imported into it.

### Functions for regions in **fasterRaster**

Properties of the current region can be obtained using the following functions (with no arguments):

* `dim()` and `dim3d()`: Dimensions
* `ext()`: Extent
* `res()` and `res3d()`: Resolution

**fasterRaster** provides several hidden functions for managing regions:

* `.region()`: Extent *and* dimensions/resolution of a region. This function is used by many other **fasterRaster** functions before raster processing to ensure the raster is not changed (by the region).
* `.regionDim()`: Dimensions (number of rows and columns)
* `.regionExt()`: Extent
* `.regionRes()`: Resolution

Most of the hidden functions can be used in three different ways:
* No arguments: The function reports the respective value(s) (e.g., `regionExt()` reports the extent of the region--same as `ext()` with no arguments).
* With a `GRaster`, `GVector`, `SpatRaster`, `SpatVector`, or `sf` object: Resizes and/or resamples the region so it has the same extent and/or resolution as the spatial object. Note that vector objects do not have a resolution, so cannot be used in `.region*()` functions that resample the region's resolution.
* With numeric values (i.e., representing extent, resolution, or dimensions): Resize or resample the region so it has these dimensions. When a region's extent is resized or resolution is changed, it will be extended to have the next-larger number of rows and/or columns so it can accommodate cells of the desired size.

Regions are inherently 3-dimensional, but often have a single depth (i.e., one "layer" of cells).


~ FINIS ~

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/freq.r
\name{freq,GRaster-method}
\alias{freq,GRaster-method}
\alias{freq}
\title{Frequencies of cell values in a raster}
\usage{
\S4method{freq}{GRaster}(x, digits = 3, bins = 100, value = NULL)
}
\arguments{
\item{x}{A \code{GRaster}.}

\item{digits}{Numeric integer: Number of digits by which to round raster values. Ignored for integer and categorical rasters.}

\item{bins}{Positive numeric integer: Number of bins in which to divide values of \code{numeric} rasters. The default is 100. For \code{integer} and categorical rasters, each value is tallied (i.e., this is ignored).}

\item{value}{Numeric or \code{NULL} (default): If numeric, only cells with this value will be counted. If \code{NULL}, all values will be counted.}
}
\value{
A \code{data.frame} or a named \code{list} of \code{data.frame}s, one per layer in \code{x}.
}
\description{
\code{freq()} tabulates the frequency of cell values in a raster. For rasters where \code{\link[=datatype]{datatype()}} is \code{integer} or \code{factor}, the frequency of each value or level is reported. For other rasters, the range of values is divided into bins, and the number of cells with values in each bin is reported.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev") # raster
madCover <- fastData("madCover") # categorical raster

# Convert to GRasters
elev <- fast(madElev) # raster
cover <- fast(madCover) # categorical raster

# Frequencies of integer raster values
f <- freq(elev)
print(f) # have to do this sometimes if output is a data table

# Frequencies of categorical raster values
f <- freq(cover)
print(f) # have to do this sometimes if output is a data table

# Frequencies of given values
f <- freq(elev, value = 1)
print(f) # have to do this sometimes if output is a data table

# When a GRaster has non-integer values, they will be binned:
f <- freq(elev + 0.1, bins = 10)
print(f)

}
}
\seealso{
\code{\link[terra:freq]{terra::freq()}}, tool \code{r.stats} in \strong{GRASS}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/centroids.r
\name{centroids,GVector-method}
\alias{centroids,GVector-method}
\alias{centroids}
\alias{centroids,GRaster-method}
\title{Centroid(s) of a vector or clumps in a raster}
\usage{
\S4method{centroids}{GVector}(x, method = NULL)

\S4method{centroids}{GRaster}(x)
}
\arguments{
\item{x}{A \code{GVector} or \code{GRaster}.}

\item{method}{\code{GVector}s: Character or \code{NULL} (default): Method used for calculating centroids. The method of calculation depends on whether the input is a \code{points}, \code{lines}, or \code{polygons} \code{GVector}. If the value is \code{NULL}, then the default method will be chosen, depending on the geometry type of the \code{GVector}:
\itemize{
\item \code{points}:
\itemize{
\item \code{"mean"} (default for \code{points}): Mean of coordinates.
\item \code{"median"}: Geometric median; more robust to outliers.
\item \code{"pmedian"}: Point in \code{x} closest to the geometric median.
}
\item \code{lines}:
\itemize{
\item \code{"mid"} (default for \code{lines}): Mid-point on each line; will fall exactly on the line.
\item \code{"mean"}: Center of gravity of all line segments; may not fall on the line.
\item \verb{"median}: Geometric median; may not fall on the line.
}
\item \code{polygons}:
\itemize{
\item \code{"mean"} (default for \code{polygons}): Center of gravity (area), calculated using area triangulation.
\item \code{"median"}: Geometric mean; may not fall inside the polygon.
\item \code{"bmedian"}: Geometric mean; minimum distance to boundaries; may not fall inside the polygon.
}
}

Partial matching is used and case is ignored.}
}
\value{
If the input is a \code{GVector}, the output will be a "points" \code{GVector}. If the input is a \code{GRaster}, the output will be a "points" \code{GVector} with a table with statistics on each clump. If the input is a \code{GRaster} with more than one layer, the output will be a \code{list} of \code{GVector}s, with one \code{GVector} per layer.
}
\description{
This function locates the centroid of each geometry of a \code{GVector}, or the centroid of each "clump" of same-valued cells in an integer/categorical raster (for information on types of \code{GRaster}s, see \code{vignette("GRasters", package = "fasterRaster")}).

To use this function with a \code{GVector}, you need the \strong{GRASS} \code{v.centerpoint} addon. To use the function with a \code{GRaster}, you need the addon \code{r.centroids}. In either case, the function will try to install the respective addon (i.e., you need to have an internet connection). Once installed, a tool will not need to be installed again.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)
library(terra)

### Points, line, and polygon centroids

# Point centroids:
madDypsis <- fastData("madDypsis")
dypsis <- fast(madDypsis)

dypMean <- centroids(dypsis)
dypMedian <- centroids(dypsis, method = "median")
dypPMedian <- centroids(dypsis, method = "pmedian")

plot(dypsis)
plot(dypMean, col = "red", add = TRUE)
plot(dypMedian, col = "green", pch = 2, add = TRUE)
plot(dypPMedian, col = "blue", pch = 3, add = TRUE)
legend("bottomright",
   legend = c("mean", "median", "pmedian"),
   col = c("red", "green", "blue"),
   pch = c(16, 2, 3),
   xpd = NA
)

# Line centroids:
madRivers <- fastData("madRivers")
rivers <- fast(madRivers)

riversMid <- centroids(rivers)
riversMean <- centroids(rivers, method = "mean")
riversMedian <- centroids(rivers, method = "median")

plot(rivers)
plot(riversMid, col = "red", add = TRUE)
plot(riversMean, col = "green", pch = 2, add = TRUE)
plot(riversMedian, col = "blue", pch = 3, add = TRUE)
legend("bottomright",
   legend = c("mid", "mean", "median"),
   col = c("red", "green", "blue"),
   pch = c(16, 2, 3),
   xpd = NA
)

# Polygon centroids:
madCoast4 <- fastData("madCoast4")
coast4 <- fast(madCoast4)

coastMean <- centroids(coast4)
coastMedian <- centroids(coast4, method = "median")
coastBMedian <- centroids(coast4, method = "bmedian")

plot(coast4)
plot(coastMean, col = "red", add = TRUE)
plot(coastMedian, col = "green", pch = 2, add = TRUE)
plot(coastBMedian, col = "blue", pch = 3, add = TRUE)
legend("bottomright",
   legend = c("mean", "median", "bmedian"),
   col = c("red", "green", "blue"),
   pch = c(16, 2, 1),
   xpd = NA
)

### Centroids of integer GRaster "clumps"

# Load elevation raster
madElev <- fastData("madElev")
elev <- fast(madElev)

# Create clumps of similarly-valued cells
clumps <- clump(elev, minDiff = 0.01, minClumpSize = 1000)

# Centroids:
clumpCents <- centroids(clumps)
clumpCents

plot(clumps)
plot(clumpCents, add = TRUE)


}
}
\seealso{
\code{\link[terra:centroids]{terra::centroids()}}; \strong{GRASS} addon tools \code{v.centerpoint} and \code{r.centroids}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract.r
\name{extract,GRaster,GVector-method}
\alias{extract,GRaster,GVector-method}
\alias{extract}
\alias{extract,GRaster,data.frame-method}
\alias{extract,GRaster,data.table-method}
\alias{extract,GRaster,matrix-method}
\alias{extract,GRaster,numeric-method}
\alias{extract,GVector,GVector-method}
\alias{extract,GVector,data.frame-method}
\alias{extract,GVector,data.table-method}
\alias{extract,GVector,matrix-method}
\alias{extract,GVector,numeric-method}
\title{Extract values from a GRaster at locations in a points GVector}
\usage{
\S4method{extract}{GRaster,GVector}(
  x,
  y,
  fun = "mean",
  prob = 0.5,
  overlap = TRUE,
  xy = FALSE,
  cats = TRUE,
  verbose = FALSE,
  ...
)

\S4method{extract}{GRaster,data.frame}(x, y, xy = FALSE, cats = TRUE)

\S4method{extract}{GRaster,data.table}(x, y, xy = FALSE, cats = TRUE)

\S4method{extract}{GRaster,matrix}(x, y, xy = FALSE, cats = TRUE)

\S4method{extract}{GRaster,numeric}(x, y, xy = FALSE, cats = TRUE)

\S4method{extract}{GVector,GVector}(x, y, xy = FALSE, verbose = TRUE)

\S4method{extract}{GVector,data.frame}(x, y, xy = FALSE, verbose = TRUE)

\S4method{extract}{GVector,data.table}(x, y, xy = FALSE, verbose = TRUE)

\S4method{extract}{GVector,matrix}(x, y, xy = FALSE, verbose = TRUE)

\S4method{extract}{GVector,numeric}(x, y, xy = FALSE)
}
\arguments{
\item{x}{A \code{GRaster} or \code{GVector}.}

\item{y}{A \code{GVector}, \emph{or} a \code{data.frame} or \code{matrix} where the first two columns represent longitude and latitude (in that order), \emph{or} a two-element numeric vector where the first column represents longitude and the second latitude. Values of \code{x} will be extracted from the points in \code{y}. \code{GVector}s can be of types points, lines, or polygons.}

\item{fun}{Character vector: Name(s) of function(s) to apply to values. This is used when \code{x} is a \code{GRaster} and \code{y} is a lines or polygons \code{GVector}. The method(s) specified by \code{fun} will be applied to all cell values that overlap with each geometry (i.e., individual cell values will not be returned). Valid functions include:
\itemize{
\item \code{"countNonNA"}: Number of overlapping cells.
\item \code{"countNA"}: Number of overlapping \code{NA} cells.
\item \code{"mean"}: Average.
\item \code{"min"}: Minimum.
\item \code{"max"}: Minimum.
\item \code{"sum"}: Sum.
\item \code{"range"}: Maximum - minimum.
\item \code{"sd"}: Sample standard deviation (same as \code{\link[stats:sd]{stats::sd()}}).
\item \code{"sdpop"}: Population standard deviation.
\item \code{"var"}: Sample variance (same as \code{\link[stats:cor]{stats::var()}}).
\item \code{"varpop"}: Population variance.
\item \code{"cv"}: Coefficient of variation.
\item \code{"cvpop"}: Population coefficient of variation.
\item \code{"median"}: Median.
\item \code{"quantile"}: Quantile; you can specify the quantile using the \code{prob} argument.
}}

\item{prob}{Numeric in the range from 0 to 1: Quantile which to calculate. The value of \code{prob} will be rounded to the nearest hundredth.}

\item{overlap}{Logical: If \code{TRUE} (default), and \code{y} is a lines or polygons \code{GVector}, then account for potential overlap of geometries when extracting. This can be slow, so if you are sure geometries do not overlap, you can change this to \code{FALSE}. This argument is ignored if \code{y} is a points \code{GVector}.}

\item{xy}{Logical: If \code{TRUE} and \code{y} represents points, also return the coordinates of each point. Default is \code{FALSE.}}

\item{cats}{Logical (extracting from a raster): If \code{TRUE} (default) and \code{x} is a categorical \code{GRaster}, then return the category labels instead of the values.}

\item{verbose}{Logical: If \code{TRUE}, display progress (will only function when extracting from points on a \code{GRaster} when the number of \code{GRaster}s is large, or when extracting using a "points" \code{GVector} with lots of points).}

\item{...}{Arguments to pass to \code{\link[=project]{project()}}. This is used only if extracting from a \code{GRaster} at locations specified by a \code{GVector}, and they have a different coordinate reference system. In this case, users should specify the \code{wrap} argument to \code{\link[=project]{project()}}.}
}
\value{
A \code{data.frame} or \code{data.table}.
}
\description{
\code{extract()} obtains the values of a \code{GRaster} or \code{GVector} associated with the locations of a set of points. The output depends on the input:
\itemize{
\item \strong{Case #1: \code{x} is a numeric or integer \code{GRaster} and \code{y} is a points \code{GVector}}: Returns values of cells that have points. If \code{xy} is \code{TRUE}, also returns the coordinates of the points.
\item \strong{Case #2: \code{x} is a categorical (factor) \code{GRaster} and \code{y} is a points \code{GVector}}: Same as case #1, but if \code{cats} is \code{TRUE}, returns category labels of cells that have points. If \code{xy} is \code{TRUE}, also returns the coordinates of the points.
\item \strong{Case #3: \code{x} is a categorical \code{GRaster} and \code{y} is a lines or polygons \code{GVector}}: Returns a summary (e.g., mean, standard deviation, etc.) of all cells that overlap the line(s) or polygon(s).
\item \strong{Case #4: \code{x} is a \code{GVector} and \code{y} is a points \code{GVector}}: Returns the data table row associated each point. If \code{xy} is \code{TRUE}, also returns the coordinates of the points.
Note that whenever a points \code{GVector} is allowed for \code{y}, a \code{data.frame}, \code{data.table}, \code{matrix}, or \code{numeric} values representing points can be used instead.
}
}
\examples{
if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Example data: elevation raster and points vector
madElev <- fastData("madElev") # raster
madCover <- fastData("madCover") # categorical raster
madDypsis <- fastData("madDypsis") # points vector
madRivers <- fastData("madRivers") # lines vector
madCoast4 <- fastData("madCoast4") # polygons vector

# Convert to fasterRaster formats:
elev <- fast(madElev) # raster
cover <- fast(madCover) # categorical raster
dypsis <- fast(madDypsis) # points vector
rivers <- fast(madRivers) # lines vector
coast <- fast(madCoast4) # polygons vector

# Get values of elevation at points where Dypsis species are located:
extract(elev, dypsis, xy = TRUE)

# Extract from categorical raster at points:
categories <- extract(cover, dypsis)
categoryValues <- extract(cover, dypsis, cats = FALSE)
categories
categoryValues

# Extract and summarize values on a raster across polygons:
extract(elev, coast, fun = c("sum", "mean", "countNonNA"), overlap = FALSE)

# Extract and summarize values on a raster across lines:
extract(elev, rivers, fun = c("sum", "mean", "countNonNA"), overlap = FALSE)

# Extract from a polygons vector at a points vector:
polysFromPoints <- extract(coast, dypsis, xy = TRUE)
head(polysFromPoints) # first 3 are outside polygons vector, next 3 are inside

}
}
\seealso{
\code{\link[terra:extract]{terra::extract()}}, and modules \code{r.what} and \code{v.what} in \strong{GRASS}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/erase.r
\name{erase,GVector,GVector-method}
\alias{erase,GVector,GVector-method}
\alias{erase}
\title{Select parts of a polygon GVector erase shared by another polygon GVector}
\usage{
\S4method{erase}{GVector,GVector}(x, y)
}
\arguments{
\item{x, y}{\code{GVector}s.}
}
\value{
A \code{GVector}.
}
\description{
The \code{erase()} function removes from the \code{x} "polygons" \code{GVector} parts that overlap with the \code{y} "polygons" \code{GVector}. You can also use the \code{-} operator (e.g., \code{vect1 - vect2}).
}
\examples{
if (grassStarted()) {

# Setup
library(sf)

# Polygon of coastal Madagascar and Dypsis specimens
madCoast4 <- fastData("madCoast4") # polygons
madDypsis <- fastData("madDypsis") # points

# Convert vectors:
coast4 <- fast(madCoast4)
dypsis <- fast(madDypsis)

# Create another polygons vector from a convex hull around Dypsis points
hull <- convHull(dypsis)

### union()

unioned <- union(coast4, hull)
plot(unioned)

plus <- coast4 + hull # same as union()

### intersect

inter <- intersect(coast4, hull)
plot(coast4)
plot(hull, border = "red", add = TRUE)
plot(inter, border = "blue", add = TRUE)

### xor

xr <- xor(coast4, hull)
plot(coast4)
plot(xr, border = "blue", add = TRUE)

### erase

erased <- erase(coast4, hull)
plot(coast4)
plot(erased, border = "blue", add = TRUE)

minus <- coast4 - hull # same as erase()

}
}
\seealso{
\code{\link[=c]{c()}}, \code{\link[=aggregate]{aggregate()}}, \code{\link[=crop]{crop()}}, \code{\link[=intersect]{intersect()}}, \code{\link[=union]{union()}}, \code{\link[=xor]{xor()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/writeVector.r
\name{writeVector,GVector,character-method}
\alias{writeVector,GVector,character-method}
\alias{writeVector}
\alias{writeVector,missing,missing-method}
\title{Save a GVector to disk}
\usage{
\S4method{writeVector}{GVector,character}(
  x,
  filename,
  overwrite = FALSE,
  format = NULL,
  attachTable = TRUE,
  ...
)

\S4method{writeVector}{missing,missing}(x, filename)
}
\arguments{
\item{x}{A \code{GVector}.}

\item{filename}{Character: Path and file name.}

\item{overwrite}{Logical: If \code{FALSE} (default), do not save over existing files.}

\item{format}{Character or \code{NULL}: File format. If \code{NULL} (default), then the function will attempt to get the format from the file name extension. Partial matching is used and case is ignored. You can see a list of formats using \code{writeVector()} (no arguments). Some common formats include:
\itemize{
\item \code{"GPKG"}: OGC GeoPackage (extension \code{.gpkg}).
\item \code{"CSV"}: Comma-separated value... saves the data table only, not the geometries (extension \code{.csv}).
\item \code{"ESRI Shapefile"}: ESRI shapefile (extension \code{.shp}).
\item \code{"GeoJSON"}: GeoJSON (extension \code{GeoJSON})
\item \code{"KML"}: Keyhole Markup Language (extension \code{.kml})
\item \code{"netCDF"}: NetCDF (extension \code{.ncdf})
\item \code{"XLSX"}: MS Office Open XML spreadsheet (extension \code{.xlsx}).
}}

\item{attachTable}{Logical: If \code{TRUE} (default), attach the attribute to table to the vector before saving it. If \code{FALSE}, the attribute table will not be attached.}

\item{...}{Additional arguments to send to \strong{GRASS} tool \code{v.out.ogr} (see \code{grassHelp("v.out.ogr")}).}
}
\value{
Invisibly returns a \code{GRaster} (the input, \code{x}). Also saves the vector to disk.
}
\description{
This function saves a \code{GVector} to disk directly from a \strong{GRASS} session.

By default, files will be of OGC GeoPackage format (extension "\code{.gpkg}"), but this can be changed with the \code{format} argument. You can see a list of supported formats by simply using this function with no arguments, as in \code{writeVector()}, or by consulting the online help page for \strong{GRASS} tool \code{v.out.ogr} (see \code{grassHelp("v.out.ogr")}).

Note that if the vector has a data table attached and at least one numeric or integer column has an \code{NA} or \code{NaN} value, the function will yield a warning like:

\if{html}{\out{<div class="sourceCode">}}\preformatted{Warning 1: Invalid value type found in record 2 for field column_with_NA_or_NaN. This warning will no longer be emitted.
}\if{html}{\out{</div>}}

Also note that despite the promise, this warning will be displayed again.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data
madRivers <- fastData("madRivers")

# What file formats can we attempt to write?
writeVector()

# Convert SpatVector to GVector
rivers <- fast(madRivers)
rivers

# Save GVector to disk as GeoPackage
filename <- tempfile(fileext = ".gpkg")
writeVector(rivers, filename)

# Save GVector to disk as ESRI Shapefile
filename <- tempfile(fileext = ".shp")
writeVector(rivers, filename)

# Save GVector to disk as Google Earth KML
filename <- tempfile(fileext = ".klm")
writeVector(rivers, filename)

# Save GVector data table to disk as comma-separated file
filename <- tempfile(fileext = ".csv")
writeVector(rivers, filename)

# Save GVector data table to disk as NetCDF
filename <- tempfile(fileext = ".ncdf")
writeVector(rivers, filename)

# Save GVector data table to disk as Excel file
filename <- tempfile(fileext = ".xlsx")
writeVector(rivers, filename)

}
}
\seealso{
\code{\link[terra:writeVector]{terra::writeVector()}}, \code{\link[sf:st_write]{sf::st_write()}}, \strong{GRASS} tool \code{v.out.ogr} (see \code{grassHelp("v.out.ogr")})

\code{\link[terra:writeVector]{terra::writeVector()}}, the \strong{GRASS} tool manual page for \code{v.out.ogr} (see \code{grassHelp("v.out.ogr")})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sun.r
\name{sun}
\alias{sun}
\title{Solar radiance and irradiance}
\usage{
sun(
  elevation,
  coeff_bh,
  coeff_dh,
  slope,
  aspect,
  hh,
  horizon_step = 90,
  albedo = 0.2,
  linke = 3,
  day = 1,
  step = 0.5,
  declination = NULL,
  solar_constant = 1367,
  distance_step = 1,
  npartitions = 1,
  beam_rad = TRUE,
  diff_rad = TRUE,
  refl_rad = TRUE,
  glob_rad = TRUE,
  insol_time = TRUE,
  lowMemory = FALSE
)
}
\arguments{
\item{elevation}{A \code{GRaster} with values representing elevation (typically in meters).}

\item{coeff_bh}{A \code{GRaster}: A raster with values of the real-sky beam radiation coefficient. Valid values are between 0 and 1.}

\item{coeff_dh}{A \code{GRaster}: A raster with values of the real-sky diffuse radiation coefficient. Valid values are between 0 and 1.}

\item{slope}{A \code{GRaster}: This is a raster representing topographic slope in degrees. It can be generated using \code{\link[=terrain]{terrain()}}.}

\item{aspect}{A \code{GRaster}: This is a raster representing topographic aspect in degrees. It can be generated using \code{\link[=terrain]{terrain()}}. If generated with that function, "east orientation" \emph{must} be used (i.e., argument \code{northIs0} must be \code{FALSE}).}

\item{hh}{A "stack" of \code{GRaster}s: This represents height of the horizon in radians in particular directions. Horizon height can be calculated using \code{\link[=horizonHeight]{horizonHeight()}}. The directions \emph{must} be in "east orientation" (i.e., argument \code{northIs0} in \code{horzionHeight()} must be \code{FALSE}). The directions must correspond with the sequence given by \code{horizon_step} (see next argument). For example, if \code{horizon_step} is 90, then \code{hh} must contain rasters representing horizon height at 0 (east), 90 (north), 180 (west), and 270 (south) aspects.}

\item{horizon_step}{Numeric >0: Difference between angular steps in which horizon height is measured. One horizon height raster will be made per value from 0 to (360 - \code{horizon_step}) degrees.}

\item{albedo}{A \code{GRaster} or a numeric value: This is either a raster with values of ground albedo or a numeric value (in which case albedo is assumed to be the same everywhere). Albedo is unit-less, and the default value is 0.2.}

\item{linke}{A \code{GRaster} or a numeric value: This is either a raster with values of the Linke atmospheric turbidity coefficient or a numeric value (in which case the same value is assumed for all locations). The Linke coefficient is unit-less. The default value is 3, but see also the \strong{GRASS} manual page for tool \code{r.sun} (\code{grassHelp("r.sun")}).}

\item{day}{Positive integer between 1 to 365, inclusive: Day of year for which to calculate ir/radiation. Default is 1 (January 1st).}

\item{step}{Positive integer between 0 and 24, inclusive. Time step in hours for all-day radiation sums. Decimal values are OK.}

\item{declination}{Numeric or \code{NULL} (default). Declination value. If \code{NULL}, this is calculated automatically.}

\item{solar_constant}{Positive numeric: The solar constant (solar energy hitting the top of the atmosphere). Default is 1367. Units are W / m^2.}

\item{distance_step}{Positive numeric between 0.5 and 1.5, inclusive: Sampling distance coefficient. Default is 1.}

\item{npartitions}{Positive numeric. Number of chunks in which to read input files. Default is 1.}

\item{beam_rad}{Logical: If \code{TRUE} (default), generate a raster with beam irradiation with units of Wh / m^2 / day ("mode 2" of the \code{r.sun} \strong{GRASS} tool).}

\item{diff_rad}{Logical: If \code{TRUE} (default), generate a raster representing irradiation in Wh / m^2 /day}

\item{refl_rad}{Logical: If \code{TRUE} (default), generate a raster with ground-reflected irradiation with units of Wh / m^2 / day ("mode 2" of the \code{r.sun} \strong{GRASS} tool).}

\item{glob_rad}{Logical:. If \code{TRUE} (default), generate a raster with total irradiance/irradiation with units of Wh / m^2 / day ("mode 2" of the \code{r.sun} \strong{GRASS} tool).}

\item{insol_time}{Logical: If \code{TRUE} (default), generate a raster with total insolation time in hours ("mode 2" of the \code{r.sun} \strong{GRASS} tool).}

\item{lowMemory}{Logical: If \code{TRUE}, use the low-memory version of the \code{r.sun} \strong{GRASS} tool. The default is \code{FALSE}.}
}
\value{
A raster or raster stack stack with the same extent, resolution, and coordinate reference system as \code{elevation}. Assuming all possible rasters are generated they represent:
\itemize{
\item \code{beam_rad}: Beam radiation (Watt-hours/m2/day)
\item \code{diff_rad}: Diffuse radiation (Watt-hours/m2/day)
\item \code{refl_rad}: Reflected radiation (Watt-hours/m2/day)
\item \code{glob_rad}: Global radiation (Watt-hours/m2/day)
\item \code{insol_time}: Insolation duration (hours)
}
}
\description{
The \code{sun()} function calculates beam (direct), diffuse and ground reflected solar irradiation for a given day and set of topographic and atmospheric conditions. The function relies on the \strong{GRASS} tool \code{r.sun}, the manual page for which contains a detailed explanation (see \code{grassHelp("r.sun")})
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster
elev <- fast(madElev)

### Calculate input rasters

# Values below are just a guess
coeff_bh <- coeff_dh <- elev
coeff_bh[] <- 0.4
coeff_dh[] <- 0.6

slope <- terrain(elev, "slope")
aspect <- terrain(elev, "aspect", northIs0 = FALSE)

horizon_step <- 90
hh <- horizonHeight(elev, step = horizon_step, northIs0 = FALSE)

### calculate solar ir/radiance

solar <- sun(
	elevation = elev,
	coeff_bh = coeff_bh,
	coeff_dh = coeff_dh,
	slope = slope,
	aspect = aspect,
	hh = hh,
	horizon_step = horizon_step,
	albedo = 0.2,
	linke = 1.5,
	day = 1,
	step = 0.5,
	declination = NULL,
	solar_constant = 1367,
	
	distance_step = 1,
	npartitions = 1,

	beam_rad = TRUE,
	diff_rad = TRUE,
	refl_rad = TRUE,
	glob_rad = TRUE,
	insol_time = TRUE,

	lowMemory = FALSE
)

solar

}
}
\seealso{
\code{\link[=terrain]{terrain()}}, \code{\link[=horizonHeight]{horizonHeight()}}, \strong{GRASS} manual page for tool \code{r.sun} (see \code{grassHelp("r.sun")})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mask.r
\name{mask,GRaster,GRaster-method}
\alias{mask,GRaster,GRaster-method}
\alias{mask}
\alias{mask,GRaster,GVector-method}
\title{Mask values in a raster}
\usage{
\S4method{mask}{GRaster,GRaster}(x, mask, inverse = FALSE, maskvalues = NA, updatevalue = NA)

\S4method{mask}{GRaster,GVector}(x, mask, inverse = FALSE, updatevalue = NA)
}
\arguments{
\item{x}{A \code{GRaster}.}

\item{mask}{A \code{GRaster} or \code{GVector}.}

\item{inverse}{Logical: If \code{TRUE}, the effect of the mask is inverted. That is, a copy of the input raster is made, but cells that overlap with an \code{NA} in the mask raster or are not covered by the mask vector retain their values. Cells that overlap with an \code{NA} in the mask raster or overlap with the mask vector are forced to \code{NA}.}

\item{maskvalues}{Numeric vector, including \code{NA} (only for when \code{mask} is a \code{GRaster}): The value(s) in the mask raster cells that serve as the mask. The default is \code{NA}, in which case cells in the input raster that overlap with \code{NA} cells in the mask are forced to \code{NA}.}

\item{updatevalue}{Numeric, including \code{NA} (default): The values assigned to masked cells.}
}
\value{
A \code{GRaster}.
}
\description{
The output of \code{mask()} is a \code{GRaster} that has the same as values as the input raster. However, if the \code{mask} argument is a \code{GRaster}, the output will have \code{NA} values in the same cells that the \code{mask} raster has \code{NA} cells. If the \code{mask} argument is a \code{GVector}, then the output raster will have \code{NA} values in cells the \code{GVector} does not cover.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Example data
madElev <- fastData("madElev") # raster
madForest <- fastData("madForest2000") # raster
madCoast <- fastData("madCoast4") # vector

# Convert to GRasters and GVectors
elev <- fast(madElev)
forest <- fast(madForest)
coast <- fast(madCoast)

ant <- coast[coast$NAME_4 == "Antanambe"]

# Mask by a raster or  vector:
maskByRast <- mask(elev, forest)
plot(c(forest, maskByRast))

maskByVect <- mask(elev, ant)
plot(maskByVect)
plot(ant, add = TRUE)

# Mask by a raster or vector, but invert mask:
maskByRastInvert <- mask(elev, forest, inverse = TRUE)
plot(c(forest, maskByRastInvert))

maskByVectInvert <- mask(elev, ant, inverse = TRUE)
plot(maskByVectInvert)
plot(ant, add = TRUE)

# Mask by a raster, but use custom values for the mask:
maskByRastCustomMask <- mask(elev, elev, maskvalues = 1:20)
plot(c(elev <= 20, maskByRastCustomMask))

# Mask by a raster or vector, but force masked values to a custom value:
byRastCustomUpdate <- mask(elev, forest, updatevalue = 7)
plot(byRastCustomUpdate)

byVectCustomUpdate <- mask(elev, ant, updatevalue = 7)
plot(byVectCustomUpdate)

# Mask by a raster, inverse, custom values, and custom update:
byRastAll <-
   mask(elev, elev, inverse = TRUE, maskvalues = 1:20, updatevalue = 7)

plot(c(elev, byRastAll))

}
}
\seealso{
\code{\link[terra:mask]{terra::mask()}}, \strong{GRASS} tool \code{r.mask}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/missingCats.r
\name{missingCats,GRaster-method}
\alias{missingCats,GRaster-method}
\alias{missingCats}
\title{Values in a categorical raster with no assigned category}
\usage{
\S4method{missingCats}{GRaster}(x, layer = 1:nlyr(x))
}
\arguments{
\item{x}{A \code{GRaster}.}

\item{layer}{Numeric integers, logical vector, or character: Layer(s) for which to obtain missing categories.}
}
\value{
A numeric vector (if \code{x}is just one layer), or a named list of numeric vectors, one per layer in \code{x}.
}
\description{
This function reports the values in a categorical \code{GRaster} that have no matching category label in its "levels" table.

\code{GRaster}s can represent categorical data. Cell values are actually integers, each corresponding to a category, such as "desert" or "wetland." A categorical raster is associated with a table that matches each value to a category name.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data: Land cover raster
madCover <- fastData("madCover")

# Convert categorical SpatRaster to categorical GRaster:
cover <- fast(madCover)

### Properties of categorical rasters

cover # note categories
is.factor(cover) # Is the raster categorical?
nlevels(cover) # number of levels
levels(cover) # just the value and active column
cats(cover) # all columns
minmax(cover) # min/max values
minmax(cover, levels = TRUE) # min/max categories
catNames(cover) # column names of the levels table
missingCats(cover) # categories in table with no values in raster
freq(cover) # frequency of each category (number of cells)
zonalGeog(cover) # geometric statistics

### Active column

# Which column sets the category labels?
activeCat(cover)
activeCat(cover, names = TRUE)

activeCats(c(cover, cover))

# Choose a different column for category labels:
levels(cover)
activeCat(cover) <- 2
levels(cover)

### Managing levels tables

# Remove unused levels:
nlevels(cover)
cover <- droplevels(cover)
nlevels(cover)

# Re-assign levels:
value <- c(20, 30, 40, 50, 120, 130, 140, 170)
label <- c("Cropland", "Cropland", "Forest", "Forest",
 "Grassland", "Shrubland", "Herbaceous", "Flooded")

newCats <- data.frame(value = value, label = label)

cover <- categories(cover, layer = 1, value = newCats)
cats(cover)

# This is the same as:
levels(cover) <- newCats
cats(cover)

# Are there any values not assigned a category?
missingCats(cover)

# Let's assign a category for value 210 (water):
water <- data.frame(value = 210, label = "Water")
addCats(cover) <- water
levels(cover)

# Add more information to the levels table using merge():
landType <- data.frame(
     Value = c(20, 30, 40, 50, 120),
     Type = c("Irrigated", "Rainfed", "Broadleaf evergreen",
     "Broadleaf deciduous", "Mosaic with forest")
)
cats(cover)
cover <- addCats(cover, landType, merge = TRUE)
cats(cover)

### Logical operations on categorical rasters

cover < "Forest" # 1 for cells with a value < 40, 0 otherwise
cover <= "Forest" # 1 for cells with a value < 120, 0 otherwise
cover == "Forest" # 1 for cells with value of 40-120, 0 otherwise
cover != "Forest" # 1 for cells with value that is not 40-120, 0 otherwise
cover > "Forest" # 1 for cells with a value > 120, 0 otherwise
cover >= "Forest" # 1 for cells with a value >= 120, 0 otherwise

cover \%in\% c("Cropland", "Forest") # 1 for cropland/forest cells, 0 otherwise

### Combine categories from different rasters

# For the example, will create a second categorical raster fromm elevation.

# Divide elevation raster into "low/medium/high" levels:
madElev <- fastData("madElev")
elev <- fast(madElev)
elev <- project(elev, cover, method = "near") # convert to same CRS
fun <- "= if(madElev < 100, 0, if(madElev < 400, 1, 2))"
elevCat <- app(elev, fun)

levs <- data.frame(
     value = c(0, 1, 2),
     elevation = c("low", "medium", "high")
)
levels(elevCat) <- list(levs)

# Combine levels:
combined <- concats(cover, elevCat)
combined
levels(combined)

# Combine levels, treating value/NA combinations as new categories:
combinedNA <- concats(cover, elevCat, na.rm = FALSE)
combinedNA
levels(combinedNA)

}
}
\seealso{
\code{\link[=missingCats]{missingCats()}}, \code{\link[=missing.cases]{missing.cases()}}, \code{\link[=droplevels]{droplevels()}}, \code{vignette("GRasters", package = "fasterRaster")}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fragmentation.r
\name{fragmentation,SpatRaster-method}
\alias{fragmentation,SpatRaster-method}
\alias{fragmentation}
\alias{fragmentation,GRaster-method}
\title{Landscape fragmentation class following Riitters et al. (2020)}
\usage{
\S4method{fragmentation}{SpatRaster}(
  x,
  w = 3,
  undet = "undetermined",
  none = NA,
  na.rm = TRUE,
  cores = faster("cores"),
  verbose = TRUE
)

\S4method{fragmentation}{GRaster}(x, w = 3, undet = "undetermined", none = NA, verbose = TRUE)
}
\arguments{
\item{x}{A \code{SpatRaster} or \code{GRaster} wherein the habitat type has cell values of 1, and non-focal habitat type has cell values of 0 or \code{NA}}

\item{w}{An odd integer between 3 and 25: Size of the window across which fragmentation is calculated (in units of "rows" and "columns"). The default is 3, meaning the function uses a 3x3 moving window to calculate fragmentation. For large rasters, compute time is ~\emph{O}(\code{N}) + \emph{O}(\code{N * w^2}), where \code{N} is the number of cells in the raster. So, even a small increase in \code{w} can increase compute time by a lot. Values >25 will fail. A workaround might be to use \code{\link[=aggregate]{aggregate()}} with \code{fun = 'mode'}, apply \code{fragmentation()}, then \code{resample()} to the original resolution.}

\item{undet}{Character: How to assign the "undetermined" case. Valid values are \code{"perforated"} (default), \code{"edge"}, and \code{"undetermined"}. Partial matching is used. If \code{Pf} is the proportional density raster cell value and \code{Pff} the proportional connectivity raster cell value, the undetermined case occurs when \code{Pf} > 0.6 and \code{Pf == Pff}.}

\item{none}{Integer or \code{NA} (default): Value to assign to a cell with no focal habitat. Riitters et al. use \code{NA}. This will be forced to an integer if it is not an actual integer.}

\item{na.rm}{Logical: If \code{TRUE} (default) and \code{x} is a \code{SpatRaster}, then cells near the edge of the raster where the window overlaps the edge can still be assigned a fragmentation class. If \code{FALSE}, these cells will be assigned a value of \code{none}.}

\item{cores}{Integer: Number of processor cores to use for when processing a \code{SpatRaster}.}

\item{verbose}{Logical: If \code{TRUE} (default), display progress.}
}
\value{
A categorical \code{SpatRaster} or \code{GRaster}. The values assigned to each class can be seen with \code{\link[=levels]{levels()}}.
}
\description{
Riitters et al. (2020) propose a classification scheme for forest fragmentation (which can be applied to any habitat type). The scheme relies on calculating density (e.g., number of forested cells in a window around a focal cell) and connectivity (number of cases where neighboring cells are both forested). This function calculates these classes from a \code{GRaster} or \code{SpatRaster} in which the focal habitat type has cell values of 1, and non-focal habitat type has cell values of 0 or \code{NA}.

Note that by default, the \code{SpatRaster} and \code{GRaster} versions will create different results around the border of the raster. The \code{SpatRaster} version uses the \code{\link[terra:focal]{terra::focal()}} function, which will not return an \code{NA} value when its window overlaps the raster border if the \code{na.rm} argument is \code{TRUE}. However, the \code{GRaster} version uses the \strong{GRASS} tool \code{r.neighbors}, which does return \code{NA} values in these cases.

The fragmentation classes are:
\itemize{
\item Value provided by \code{none}: None (i.e., no forest; default is \code{NA}).
\item 1: Patch
\item 2: Transitional
\item 3: Perforated
\item 4: Edge
\item 5: Undetermined (not possible to obtain when \code{w = 3})
\item 6: Interior
}
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data:
madForest <- fastData("madForest2000") # raster

### Fragmentation classes from a SpatRaster

fragTerra <- fragmentation(madForest)
plot(fragTerra)
levels(fragTerra)
freq(fragTerra)

### Fragmentation classes from a GRaster

# Convert to GRaster:
forest <- fast(madForest)

# Fragmentation class:
frag <- fragmentation(forest)
plot(frag)
levels(frag)
freq(frag)

}
}
\references{
Riitters, K., J. Wickham, R. O'Neill, B. Jones, and E. Smith. 2000. Global-scale patterns of forest fragmentation. \emph{Conservation Ecology} 4:3. URL: \url{http://www.consecol.org/vol4/iss2/art3/}. Also note the \href{https://www.ecologyandsociety.org/vol4/iss2/art3/errata/january26.2001.html}{errata}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/buffer.r
\name{buffer,GRaster-method}
\alias{buffer,GRaster-method}
\alias{buffer}
\alias{buffer,GVector-method}
\alias{st_buffer,GVector-method}
\alias{st_buffer}
\title{Increase/decrease the size of a vector or around non-NA cells of a raster}
\usage{
\S4method{buffer}{GRaster}(
  x,
  width,
  unit = "meters",
  method = "Euclidean",
  background = 0,
  lowMemory = FALSE
)

\S4method{buffer}{GVector}(x, width, capstyle = "round", dissolve = TRUE)

\S4method{st_buffer}{GVector}(x, dist, endCapStyle = "round", dissolve = FALSE)
}
\arguments{
\item{x}{A \code{GRaster} or \code{GVector}.}

\item{width}{Numeric: For rasters -- Maximum distance cells must be from focal cells to be within the buffer. For rasters, if the buffering unit is \verb{"cells}", then to get \code{n} cell widths, use \code{n + epsilon}, where \code{epsilon} is a small number (e.g., 0.001). The larger the buffer, this smaller this must be to ensure just \code{n} cells are included.

For vectors, distance from the object to place the buffer. Negative values create "inside" buffers. Units are in the same units as the current coordinate reference system (e.g., degrees for WGS84 or NAD83, often meters for projected systems).}

\item{unit}{Character: Rasters -- Indicates the units of \code{width}. Can be one of:
\itemize{
\item \code{"cells"}: Units are numbers of cells.
\item \code{"meters"} (default), \code{"metres"}, or \code{"m"}
\item \code{"kilometers"} or \code{"km"}
\item \code{"feet"} or \code{"ft"}
\item \code{"miles"} or \code{"mi"}
\item \code{"nautical miles"} or \code{"nmi"}
}

Partial matching is used and case is ignored.}

\item{method}{Character: Rasters -- Only used if \code{units} is \code{"cells"}. Indicates the manner in which distances are calculated for adding of cells:
\itemize{
\item \code{"Euclidean"}: Euclidean distance (default)
\item \code{"Manhattan"}: "taxi-cab" distance
\item \code{"maximum"}: Maximum of the north-south and east-west distances between points.
}

Partial matching is used and case is ignored.}

\item{background}{Numeric: Rasters -- Value to assign to cells that are not \code{NA} and not part of the buffer (default is 0).}

\item{lowMemory}{Logical: Rasters -- Only used if buffering a raster and \code{units} is not \code{"meters"}. If \code{FALSE} (default) use faster, memory-intensive procedure. If \code{TRUE} then use the slower, low-memory version. To help decide which to use, consider using the low-memory version on a system with 1 GB of RAM for a raster larger than about 32000 x 32000 cells, or for a system with  with 8 GB of RAM a raster larger than about 90000 x 90000 cells.}

\item{capstyle, endCapStyle}{Character: Vectors -- Style for ending the "cap" of buffers around lines. Valid options include \code{"rounded"}, \code{"square"}, and "\code{flat}".}

\item{dissolve}{Logical (\code{GVector}s): If \code{TRUE} (default), dissolve all buffers after creation. If \code{FALSE}, construct a buffer for each geometry. Note that overlapping buffers can cause this function to fail because it creates a topologically ambiguous polygon. Thus, using \code{dissolve = TRUE} is recommended.}

\item{dist}{Vectors -- Same as \code{width}.}
}
\value{
A \code{GRaster} or a \code{GVector}.
}
\description{
Buffers can be constructed for \code{GRaster}s or \code{GVector}s. For rasters, the \code{buffer()} function creates a buffer around non-\code{NA} cells. The output will be a raster. For vectors, the \code{buffer()} and \code{st_buffer()} functions create a vector polygon larger or smaller than the focal vector.
}
\details{
Note that in some cases, topologically incorrect vectors can be created when buffering. This can arise when buffers intersect to create intersections that technically belong to two or more geometries. This issue can be resolved by dissolving borders between buffered geometries using \code{dissolve = TRUE}, but as of now, there is no fix if you do not want to dissolve geometries. A workaround would be to create a different \code{GVector} for each geometry, and then buffer each individually :(.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Elevation raster, rivers vector
madElev <- fastData("madElev")
madRivers <- fastData("madRivers")

# Convert a SpatRaster to a GRaster, and sf to a GVector
elev <- fast(madElev)
rivers <- fast(madRivers)

### Buffer a raster by a given distance:
buffByDist <- buffer(elev, width = 2000) # 2000-m buffer
plot(buffByDist, legend = FALSE)
plot(madElev, add = TRUE)

### Buffer a raster by a given number of cells:
buffByCells <- buffer(elev, width = 20.01, unit = "cells") # 20-cell buffer
plot(buffByCells, legend = FALSE)
plot(madElev, add = TRUE)

### Buffer a vector:
buffRivers <- buffer(rivers, width = 2000, dissolve = TRUE) # 2000-m buffer
plot(buffRivers)
plot(st_geometry(madRivers), col = "blue", add = TRUE)

}
}
\seealso{
\code{\link[terra:buffer]{terra::buffer()}}, \code{\link[sf:geos_unary]{sf::st_buffer()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fasterRaster.r
\docType{package}
\name{fasterRaster}
\alias{fasterRaster-package}
\alias{fasterRaster}
\title{"fasterRaster": Faster raster and spatial vector processing using "GRASS"}
\description{
\strong{fasterRaster}: Processing of large-in-memory/-on disk rasters and spatial vectors in using \strong{GRASS}. Most functions in the \strong{terra} and \strong{sf} packages are recreated. Processing of medium-sized and smaller spatial objects will nearly always be faster using \strong{terra} or \strong{sf}. To use most of the functions you must have the stand-alone version of \strong{GRASS} version 8.3 or higher (not the \strong{OSGeoW4} installer version). Note that due to differences in how \strong{GRASS}, \strong{terra}, and \strong{sf} were implemented, results will not always be strictly comparable between functions for the same operation.
\subsection{Most useful tutorials and functions:}{
\itemize{
\item The quick-start guide to getting started with \strong{fasterRaster}: \code{vignette("fasterRaster", package = "fasterRaster")}:
\item Types of \code{GRaster}s: \code{vignette("GRasters", package = "fasterRaster")}
\item How to speed up \strong{fasterRaster}: \code{vignette("faster_fasterRaster", package = "fasterRaster")}
\item \code{\link[=faster]{faster()}}: Set the directory where \strong{GRASS} is installed on your system, and set or get other package-wide options. This function must be run once before using most \strong{fasterRaster} functions.
\item \code{\link[=fast]{fast()}}: Convert a \code{SpatRaster}, \code{SpatVector}, or \code{sf} vector to \strong{fasterRaster}'s raster format (\code{GRaster}s) or vector format (\code{GVector}s), or load one from a file
\item \code{\link[=rast]{rast()}}, \code{\link[=vect]{vect()}}, and \code{\link[=st_as_sf]{st_as_sf()}}: Convert \code{GRaster}s and \code{GVector}s to \code{SpatRaster}s, \code{SpatVector}s, or \code{sf} vectors
\item \code{\link[=writeRaster]{writeRaster()}} and \code{\link[=writeVector]{writeVector()}}: Save \code{GRaster}s or \code{GVector}s to disk
}
}

\subsection{Properties of \code{GRasters}}{
\itemize{
\item \code{\link[=crs]{crs()}}: Coordinate reference system
\item \code{\link[=coordRef]{coordRef()}}: Coordinate reference system
\item \code{\link[=datatype]{datatype()}}: Data type
\item \code{\link[=dim]{dim()}} and \code{\link[=dim3d]{dim3d()}}: Number of rows, columns, and depths
\item \code{\link[=ext]{ext()}}, \code{\link[=N]{N()}}, \code{\link[=S]{S()}}, \code{\link[=E]{E()}}, \code{\link[=W]{W()}}, \code{\link[=top]{top()}}, and \code{\link[=bottom]{bottom()}}: Spatial extent
\item \code{\link[=freq]{freq()}}: Frequencies of cell values in a raster
\item \code{\link[=is.2d]{is.2d()}} and \code{\link[=is.3d]{is.3d()}}: Is an object 2- or 3-dimensional?
\item \code{\link[=is.int]{is.int()}}, \code{\link[=is.cell]{is.cell()}}, \code{\link[=is.float]{is.float()}}, \code{\link[=is.doub]{is.doub()}}: \code{GRaster} data type (integer/float/double)
\item \code{\link[=is.factor]{is.factor()}}: Does a raster represent categorical data?
\item \code{\link[=is.lonlat]{is.lonlat()}}: Is an object projected (e.g., in WGS84)?
\item \code{\link[=levels]{levels()}}: Names of levels in a categorical \code{GRaster}
\item \code{\link[=minmax]{minmax()}}: Minimum and maximum values across all non-\code{NA} cells
\item \code{\link[=names]{names()}}: \code{GRaster} names
\item \code{\link[=ncol]{ncol()}}: Number of columns
\item \code{\link[=nacell]{nacell()}}: Number of \code{NA} cells
\item \code{\link[=ncell]{ncell()}}: Number of cells
\item \code{\link[=ncell3d]{ncell3d()}}: Number of cells of a 3D \code{GRaster}
\item \code{\link[=ndepth]{ndepth()}}: Number of depths of a 3D \code{GRaster}
\item \code{\link[=nlyr]{nlyr()}}: Number of layers
\item \code{\link[=nonnacell]{nonnacell()}}: Number of non-\code{NA} cells
\item \code{\link[=nrow]{nrow()}}: Number of rows
\item \code{\link[=nlevels]{nlevels()}}: Number of categories
\item \code{\link[=res]{res()}}, \code{\link[=res3d]{res3d()}}, \code{\link[=xres]{xres()}}, \code{\link[=yres]{yres()}}, and \code{\link[=zres]{zres()}}: Spatial resolution
\item \code{\link[=sources]{sources()}}: Name of the raster file in the \strong{GRASS} cache
\item \code{\link[=topology]{topology()}}: Dimensionally (2D or 3D)
\item \code{\link[=zext]{zext()}}: Vertical extent
\item \code{\link[=zres]{zres()}}: Vertical resolution
}
}

\subsection{Functions that operate on or create \code{GRasters}}{
\itemize{
\item \link{Arithmetic}: Mathematical operations on \code{GRaster}s: \code{+}, \code{-}, \code{*}, \code{/}, \code{^}, \code{\%\%} (modulus), \code{\%/\%} (integer division)
\item \link[=Compare-methods]{Logical comparisons}: \code{<}, \code{<=}, \code{==}, \code{!=}, \code{>=}, and \code{>}, plus \code{\link[fasterRaster]{\%in\%}} and \code{\link[fasterRaster]{\%notin\%}} (for categorical rasters only)
\item \link[=Logic-methods]{Logical operators}: \code{|}and \code{&}
}

Mathematical functions that are applied to each layer of a \code{GRaster}:
\itemize{
\item Working with \code{NA}s: \code{\link[=is.na]{is.na()}}, \code{\link[=not.na]{not.na()}}, and \code{\link[=maskNA]{maskNA()}}
\item Trigonometry: \code{\link[=sin]{sin()}}, \code{\link[=cos]{cos()}}, \code{\link[=tan]{tan()}}, \code{\link[=asin]{asin()}}, \code{\link[=acos]{acos()}}, \code{\link[=atan]{atan()}}, \code{\link[=atan2]{atan2()}}
\item Logarithms and powers: \code{\link[=exp]{exp()}}, \code{\link[=log]{log()}}, \code{\link[=ln]{ln()}}, \code{\link[=log1p]{log1p()}}, \code{\link[=log2]{log2()}}, \code{\link[=log10]{log10()}}, \code{\link[=sqrt]{sqrt()}}
\item Rounding: \code{\link[=round]{round()}}, \code{\link[=floor]{floor()}}, \code{\link[=ceiling]{ceiling()}}, \code{\link[=trunc]{trunc()}}
\item Signs: \code{\link[=abs]{abs()}}
}

Mathematical functions that are applied across layers of multi-layered \code{GRaster}s:
\itemize{
\item Numeration: \code{\link[=sum]{sum()}}, \code{\link[=count]{count()}}
\item Central tendency: \code{\link[=mean]{mean()}}, \code{\link[=mmode]{mmode()}}, \code{\link[=median]{median()}}
\item Dispersion: \code{\link[=stdev]{stdev()}}, \code{\link[=var]{var()}}, \code{\link[=varpop]{varpop()}}, \code{\link[=nunique]{nunique()}}, \code{\link[=range]{range()}}, \code{\link[=quantile]{quantile()}}, \code{\link[=skewness]{skewness()}}, \code{\link[=kurtosis]{kurtosis()}}
\item Extremes: \code{\link[=min]{min()}}, \code{\link[=max]{max()}}, \code{\link[=which.min]{which.min()}}, \code{\link[=which.max]{which.max()}}
\item \code{NA}s: \code{\link[=allNA]{allNA()}}, \code{\link[=anyNA]{anyNA()}}
}

Subsetting, assigning, and replacing \code{GRaster} layers
\itemize{
\item \link{$}, \code{\link[fasterRaster]{[[}}, or \code{\link[=subset]{subset()}}: Subset or remove specific layers of a \code{GRaster}
\item \code{\link[fasterRaster]{[<-}}: Replace values of cells of a \code{GRaster}
\item \code{\link[fasterRaster]{[[<-}}: Replace specific layers of a \code{GRaster}
\item \code{\link[fasterRaster]{add<-}}: Replace specific layers of a \code{GRaster}
}

Operations on \code{GRaster}s
\itemize{
\item \code{\link[=as.int]{as.int()}}, \code{\link[=as.float]{as.float()}}, \code{\link[=as.doub]{as.doub()}}: Change data type (integer/float/double)
\item \code{\link[=as.lines]{as.lines()}}: Convert a \code{GRaster} to a "lines" vector
\item \code{\link[=as.points]{as.points()}}: Convert a \code{GRaster} to a "points" vector
\item \code{\link[=as.polygons]{as.polygons()}}: Convert a \code{GRaster} to a "polygons" vector
\item \code{\link[=aggregate]{aggregate()}}: Aggregate values of \code{GRaster} cells into larger cells
\item \code{\link[=bioclims]{bioclims()}}: BIOCLIM rasters (classic set and extended set)
\item \code{\link[=buffer]{buffer()}}: Create a buffer around non-\code{NA} cells
\item \code{\link[=app]{app()}}: Apply a user-defined function to multiple layers of a \code{GRaster} (with helper functions \code{\link[=appFuns]{appFuns()}} and \code{\link[=appCheck]{appCheck()}})
\item \code{\link[=c]{c()}}: "Stack" two or more rasters
\item \code{\link[=cellSize]{cellSize()}}: Cell area
\item \code{\link[=classify]{classify()}}: Partition cell values into strata
\item \code{\link[=clump]{clump()}}: Group adjacent cells with similar values
\item \code{\link[=combineLevels]{combineLevels()}}: Combine the "levels" tables of two or more categorical \code{GRaster}s
\item \code{\link[=concats]{concats()}}: Combine values from two or more categorical and/or integer rasters by concatenating them
\item \code{\link[=crop]{crop()}}: Remove parts of a \code{GRaster}
\item \code{\link[=denoise]{denoise()}}: Remove "noise" from a \code{GRaster} using a principal components analysis (PCA)
\item \code{\link[=distance]{distance()}}: Distance to non-\code{NA} cells, or vice versa
\item \code{\link[=extend]{extend()}}: Add rows and columns to a \code{GRaster}
\item \code{\link[=extract]{extract()}}: Extract values from a \code{GRaster} at locations of a \code{GVector}
\item \code{\link[=fillNAs]{fillNAs()}}: Fill \code{NA} cells
\item \code{\link[=focal]{focal()}}: Calculate cell values based on values of nearby cells
\item \code{\link[=fragmentation]{fragmentation()}}: Landscape fragmentation class from Riitters et al. (2020)
\item \code{\link[=global]{global()}}: Summary statistics across cells of each \code{GRaster} layer
\item \code{\link[=hist]{hist()}}: Histogram of \code{GRaster} values
\item \code{\link[=interpIDW]{interpIDW()}}: Interpolate values at points to a \code{GRaster}
\item \code{\link[=kernel]{kernel()}}: Kernel density estimator of points
\item \code{\link[=layerCor]{layerCor()}}: Correlation or covariance between two or more \code{GRaster} layers
\item \code{\link[=mask]{mask()}}: Remove values in a \code{GRaster} based on values in another \code{GRaster} or vector
\item \code{\link[=maskNA]{maskNA()}}: Mask all non-NA cells or all NA cells
\item \code{\link[=match]{match()}}, \code{\link[fasterRaster]{\%in\%}}, and \code{\link[fasterRaster]{\%notin\%}}: Find which cells of a \code{GRaster} match or do not match certain values
\item \code{\link[=merge]{merge()}}: Combine two or more rasters with different extents and fill in \code{NA}s
\item \code{\link[=multivarEnvSim]{multivarEnvSim()}}: Multivariate environmental similarity surface (MESS)
\item \code{\link[fasterRaster]{names<-}}: Assign names to a \code{GRaster}
\item \code{\link[=noise]{noise()}}: Remove coarse-scale trends from a \code{GRaster}, leaving just fine-scale "noise"
\item \code{\link[=pairs]{pairs()}}: Plot correlations between \code{GRaster} layers
\item \code{\link[=pcs]{pcs()}}: Retrieve a principal components model from a PCA \code{GRaster} generated using \code{princomp()}
\item \code{\link[=plot]{plot()}}: Display a \code{GRaster}
\item \code{\link[=project]{project()}}: Change coordinate reference system and cell size
\item \code{\link[=predict]{predict()}}: Make predictions to a \code{GRaster} from a linear model or generalized linear model
\item \code{\link[=princomp]{princomp()}}: Apply a principal components analysis (PCA) to a \code{GRaster}
\item \code{\link[=regress]{regress()}}: Regression intercept, slope, r2, and t-value across each set of cells
\item \code{\link[=resample]{resample()}}: Change cell size
\item \code{\link[=reorient]{reorient()}}: Convert degrees between 'north-orientation' and 'east orientation'
\item \code{\link[=sampleRast]{sampleRast()}}: Randomly sample cells from a \code{GRaster}
\item \code{\link[=scale]{scale()}}, \code{\link[=scalepop]{scalepop()}}, and \code{\link[=unscale]{unscale()}}: Subtract means and divide by standard deviations, or inverse of that
\item \code{\link[=selectRange]{selectRange()}}: Select values from rasters in a stack based on values in another \code{GRaster}
\item \code{\link[=spatSample]{spatSample()}}: Randomly points from a \code{GRaster}
\item \code{\link[=stretch]{stretch()}}: Rescale values in a GRaster
\item \code{\link[=subst]{subst()}}: Re-assign cell values
\item \code{\link[=thinLines]{thinLines()}}: Reduce linear features on a \code{GRaster} so linear features are 1 cell wide
\item \code{\link[=tiles]{tiles()}}: Divide a \code{GRaster} into spatially exclusive subsets (though with possible overlap)
\item \code{\link[=trim]{trim()}}: Remove rows and columns from a \code{GRaster} that are all \code{NA}
\item \code{\link[=zonal]{zonal()}}: Statistics (mean, sum, etc.) on areas of a \code{GRaster} defined by sets of cells with the same values in another \code{GRaster}, or by geometries in a \code{GVector}
\item \code{\link[=zonalGeog]{zonalGeog()}}: Geographic statistics (area, perimeter, fractal dimension, etc.) for sets of cells with the same values
}
}

\subsection{Creating \code{GRaster}s \emph{de novo}}{
\itemize{
\item \code{\link[=fractalRast]{fractalRast()}}: Create a fractal \code{GRaster}
\item \code{\link[=init]{init()}}: GRaster with values equal to row, column, coordinate, regular, or "chess"
\item \code{\link[=longlat]{longlat()}}: Create longitude/latitude rasters
\item \code{\link[=makeGRaster]{makeGRaster()}}: Create a \code{GRaster} from a \strong{GRASS} raster file
\item \code{\link[=makeGVector]{makeGVector()}}: Create a \code{GVector} from a \strong{GRASS} vector file
\item \code{\link[=rNormRast]{rNormRast()}}: A random \code{GRaster} with values drawn from a normal distribution
\item \code{\link[=rSpatialDepRast]{rSpatialDepRast()}}: Create a random \code{GRaster} with or without spatial dependence
\item \code{\link[=rUnifRast]{rUnifRast()}}: A random \code{GRaster} with values drawn from a uniform distribution
\item \code{\link[=rWalkRast]{rWalkRast()}}: Paths of random walkers
\item \code{\link[=sineRast]{sineRast()}}: Sine wave rasters
}
}

\subsection{Analysis of terrain and hydrology}{
\itemize{
\item \code{\link[=as.contour]{as.contour()}}: Contour lines from a \code{GRaster}
\item \code{\link[=flow]{flow()}}: Identify watershed basins and direction and accumulation of flow
\item \code{\link[=flowPath]{flowPath()}}: Path of water flow across a landscape
\item \code{\link[=geomorphons]{geomorphons()}}: Identify terrain feature types
\item \code{\link[=hillshade]{hillshade()}}: Create a hillshade \code{GRaster}
\item \code{\link[=horizonHeight]{horizonHeight()}}: Horizon height
\item \code{\link[=sun]{sun()}}: Solar radiance and irradiance
\item \code{\link[=ruggedness]{ruggedness()}}: Terrain Ruggedness Index
\item \code{\link[=streams]{streams()}}: Create stream network
\item \code{\link[=terrain]{terrain()}}: Slope, aspect, curvature, and partial slopes
\item \code{\link[=wetness]{wetness()}}: Topographic wetness index
}
}

\subsection{Operations on categorical (factor) \code{GRaster}s}{
\itemize{
\item \code{\link[fasterRaster]{\%in\%}}, and \code{\link[fasterRaster]{\%notin\%}}: Mask cells that match or do not match a given category
\item \code{\link[=activeCat]{activeCat()}} and \code{\link[=activeCats]{activeCats()}}: Column(s) that defines category labels
\code{\link[fasterRaster]{activeCat<-}}: Set column that defines category labels
\item \code{\link[=addCats]{addCats()}}: Add new columns to a "levels" table
\code{\link[fasterRaster]{addCats<-}}: Add new rows (levels) to a "levels" table
\item \code{\link[=categories]{categories()}}: Set "levels" table for specific layers of a categorical raster
\item \code{\link[=catNames]{catNames()}}: Column names of each "levels" table
\item \code{\link[=cats]{cats()}}: "Levels" table of a categorical raster
\item \code{\link[=combineLevels]{combineLevels()}}: Combine the "levels" tables of two or more categorical \code{GRaster}s
\item \code{\link[=complete.cases]{complete.cases()}}: Find rows of a categorical \code{GRaster}'s "levels" table that have no \code{NA}s in them
\item \code{\link[=concats]{concats()}}: Combine categories from two or more categorical rasters by concatenating them
\item \code{\link[=droplevels]{droplevels()}}: Remove one or more levels
\item \code{\link[=freq]{freq()}}: Frequency of each category across cells of a raster
\item \code{\link[=is.factor]{is.factor()}}: Is a raster categorical?
\item \code{\link[=levels]{levels()}}: "Levels" table of a categorical raster
\item \code{\link[fasterRaster]{levels<-}}: Set "levels" table of a categorical raster
\item \code{\link[=match]{match()}}, \code{\link[fasterRaster]{\%in\%}}, and \code{\link[fasterRaster]{\%notin\%}}: Find which cells of a \code{GRaster} match or do not match certain category labels
\item \code{\link[=minmax]{minmax()}}: "Lowest" and "highest" category values of categorical rasters (when argument \code{levels = TRUE})
\item \code{\link[=missing.cases]{missing.cases()}}: Find rows of a categorical \code{GRaster}'s "levels" table that have at least one \code{NA} in them
\item \code{\link[=missingCats]{missingCats()}}: Values that have no category assigned to them
\item \code{\link[=nlevels]{nlevels()}}: Number of levels
\item \code{\link[=segregate]{segregate()}}: Create one GRaster layer per unique value in a GRaster
\item \code{\link[=subst]{subst()}}: Re-assign category levels
\item \code{\link[=zonalGeog]{zonalGeog()}}: Geographic statistics (area, perimeter, fractal dimension, etc.) for sets of cells with the same values
}
}

\subsection{Analysis of remote sensing rasters}{
\itemize{
\item \code{\link[=compositeRGB]{compositeRGB()}}: Combine red, green, and blue color bands to make a composite \code{GRaster}
\item \code{\link[=plotRGB]{plotRGB()}}: Display a multispectral \code{GRaster} using red, blue, green, and alpha channels
\item \code{\link[=vegIndex]{vegIndex()}}: Vegetation indices from surface reflectance
}
}

\subsection{Functions that operate on \strong{terra} \code{SpatRaster}s}{
\itemize{
\item \code{\link[=bioclims]{bioclims()}}: BIOCLIM rasters (classic set and extended set)
\item \code{\link[=fragmentation]{fragmentation()}}: Landscape fragmentation class from Riitters et al. (2020)
}
}

\subsection{Properties of \code{GVector}s}{
\itemize{
\item \code{\link[=crs]{crs()}}: Coordinate reference system
\item \code{\link[=coordRef]{coordRef()}}: Coordinate reference system
\item \code{\link[=datatype]{datatype()}}: Data type of fields
\item \code{\link[=dim]{dim()}}: Number of geometries and columns
\item \code{\link[=expanse]{expanse()}}: Area of polygons or length of lines
\item \code{\link[=ext]{ext()}}, \code{\link[=N]{N()}}, \code{\link[=S]{S()}}, \code{\link[=E]{E()}}, \code{\link[=W]{W()}}, \code{\link[=top]{top()}}, and \code{\link[=bottom]{bottom()}}: Spatial extent
\item \code{\link[=geomtype]{geomtype()}}: Type of vector (points, lines, polygons)
\item \code{\link[=is.2d]{is.2d()}} and \code{\link[=is.3d]{is.3d()}}: Is an object 2- or 3-dimensional?
\item \code{\link[=is.lonlat]{is.lonlat()}}: Is an object projected (e.g., in WGS84)?
\item \code{\link[=is.points]{is.points()}}, \code{\link[=is.lines]{is.lines()}}, \code{\link[=is.polygons]{is.polygons()}}: Does a \code{GVector} represent points, lines, or polygons?
\item \code{\link[=names]{names()}}: Names of \code{GVector} fields
\item \code{\link[=ncol]{ncol()}}: Number of fields
\item \code{\link[=ngeom]{ngeom()}}: Number of geometries (points, lines, polygons)
\item \code{\link[=nrow]{nrow()}}: Number of rows in a vector data table
\item \code{\link[=nsubgeom]{nsubgeom()}}: Number of sub-geometries (points, lines, polygons that make up single- and multipart geometries)
\item \code{\link[=sources]{sources()}}: Name of the vector file in the \strong{GRASS} cache
\item \code{\link[=topology]{topology()}}: Dimensionally (2D or 3D)
\item \code{\link[=zext]{zext()}}: Vertical extent
}
}

\subsection{Subsetting and assigning geometries or rows and columns of \code{GVector}s}{
\itemize{
\item \link{$} or \code{\link[fasterRaster]{[[}}: Subset columns of a \code{GVector}'s data table
\item \code{\link[fasterRaster]{[}} or \code{\link[=subset]{subset()}}: Subset geometries of a \code{GVector}
\item \code{\link[fasterRaster]{$<-}}: Replace specific columns of a \code{GVector}'s data table or add columns
\item \code{\link[fasterRaster]{addTable<-}}: Add a data table to a \code{GVector}
\item \code{\link[=dropTable]{dropTable()}}: Remove a \code{GVector}s data table
}
}

\subsection{Operations on \code{GVector}s}{
\itemize{
\item \code{\link[=aggregate]{aggregate()}}: Combine \code{GVector} geometries
\item \code{\link[=as.data.frame]{as.data.frame()}}: Convert a \code{GVector}'s attribute table to a \code{data.frame}
\item \code{\link[=as.data.table]{as.data.table()}}: Convert a \code{GVector}'s attribute table to a \code{data.table}
\item \code{\link[=as.points]{as.points()}}: Extract vertex coordinates from a "lines" or "polygons" \code{GVector}
\item \code{\link[=buffer]{buffer()}}: Create a polygon around/inside a \code{GVector}
\item \code{\link[=clusterPoints]{clusterPoints()}}: Identify clusters of points
\item \code{\link[=centroids]{centroids()}}: Centroid(s) of a \code{GVector}
\item \code{\link[=colbind]{colbind()}}: Add columns to the data table of a \code{GVector}
\item \code{\link[=complete.cases]{complete.cases()}}: Find rows of a \code{GVector}'s data table that have no \code{NA}s in them
\item \code{\link[=connectors]{connectors()}}: Create lines connecting nearest features of two \code{GVector}s
\item \code{\link[=convHull]{convHull()}}: Minimum convex hull
\item \code{\link[=crds]{crds()}}: Extract coordinates of a \code{GVector}
\item \code{\link[=crop]{crop()}}: Remove parts of a \code{GVector}
\item \code{\link[=delaunay]{delaunay()}}: Delaunay triangulation
\item \code{\link[=disagg]{disagg()}}: Separate multipart geometries into singlepart geometries
\item \code{\link[=distance]{distance()}}: Distance between geometries in two \code{GVector}, or from a \code{GVector} to cells of a \code{GRaster}
\item \code{\link[=erase]{erase()}} or \code{-}: Remove part of a \code{GVector} that overlaps with another
\item \code{\link[=expanse]{expanse()}}: Area of polygons or length of lines
\item \code{\link[=extract]{extract()}}: Extract values from a \code{GVector} at specific points
\item \code{\link[=grid]{grid()}}: Create a grid \code{GVector}
\item \code{\link[=head]{head()}}: First rows of a \code{GVector}'s data table
\item \code{\link[=hexagons]{hexagons()}}: Create a hexagonal grid
\item \code{\link[=interpIDW]{interpIDW()}}: Interpolate values at points to a \code{GRaster} using inverse-distance weighting
\item \code{\link[=interpSplines]{interpSplines()}}: Interpolate values at points to a \code{GRaster} using splines
\item \code{\link[=intersect]{intersect()}} or \code{*}: Intersection of two \code{GVectors}
\item \code{\link[=kernel]{kernel()}}: Kernel density estimator of points
\item \code{\link[=missing.cases]{missing.cases()}}: Find rows of a \code{GVector}'s data table that have at least \code{NA} in them
\item \code{\link[fasterRaster]{names<-}}: Assign names to columns of a \code{GVector}s data table
\item \code{\link[=neighborhoodMatrix]{neighborhoodMatrix()}} and \code{\link[=neighbourhoodMatrix]{neighbourhoodMatrix()}}: Neighborhood matrix of a polygons \code{GVector}
\item \code{\link[=project]{project()}}: Change coordinate reference system
\item \code{\link[=rasterize]{rasterize()}}: Convert a \code{GVector} to a \code{GRaster}
\item \code{\link[=rbind]{rbind()}}: Combine \code{GVectors}
\item \code{\link[=simplifyGeom]{simplifyGeom()}}: Remove vertices
\item \code{\link[=smoothGeom]{smoothGeom()}}: Remove "angular" aspects of features
\item \code{\link[=st_as_sf]{st_as_sf()}}: Convert a \code{GVector} to a \code{sf} vector
\item \code{\link[=st_buffer]{st_buffer()}}: Create a polygon around/inside a \code{GVector}
\item \code{\link[=tail]{tail()}}: Last rows of a \code{GVector}'s data table
\item \code{\link[=thinPoints]{thinPoints()}}: Reduce number of points in same raster cell
\item \code{\link[=union]{union()}} or \code{+}: Combine two \code{GVector}s
\item \code{\link[=voronoi]{voronoi()}}: Voronoi tessellation
\item \code{\link[=xor]{xor()}} or \code{/}: Select parts of polygons not shared by two \code{GVector}s
}
}

\subsection{Creating \code{GVector}s \emph{de novo}}{
\itemize{
\item \code{\link[=rvoronoi]{rvoronoi()}}: Random Voronoi tesselation
}
}

\subsection{Fixing issues with \code{GVector}s}{

(See also \emph{Details} \code{\link[=fast]{fast()}}.)
\itemize{
\item \code{\link[=breakPolys]{breakPolys()}}: Break topologically clean areas
\item \code{\link[=fillHoles]{fillHoles()}}: Fill "holes" of a \code{GVector}
\item \code{\link[=fixBridges]{fixBridges()}}: Change "bridges" to "islands"
\item \code{\link[=fixDangles]{fixDangles()}}: Change "dangles" hanging off boundaries to lines
\item \code{\link[=fixLines]{fixLines()}}: Break lines at intersections and lines that form closed loops
\item \code{\link[=remove0]{remove0()}}: Remove all boundaries and lines with a length of 0
\item \code{\link[=removeAngles]{removeAngles()}}: Collapse lines that diverge at an angle that is computationally equivalent to 0
\item \code{\link[=removeBridges]{removeBridges()}}: Remove "bridges" to "islands"
\item \code{\link[=removeDangles]{removeDangles()}}: Remove "dangling" lines
\item \code{\link[=removeDupCentroids]{removeDupCentroids()}}: Remove duplicated area centroids
\item \code{\link[=removeDups]{removeDups()}}: Remove duplicated features and area centroids
\item \code{\link[=removeSmallPolys]{removeSmallPolys()}}: Remove small polygons
\item \code{\link[=snap]{snap()}}: Snap lines/boundaries to each other
}
}

\subsection{Converting between data types}{
\itemize{
\item \code{\link[=as.contour]{as.contour()}}: Convert a \code{GRaster} to a \code{GVector} representing contour lines
\item \code{\link[=as.doub]{as.doub()}}: Convert a \code{GRaster} to a double-floating point raster (\strong{GRASS} data type \code{DCELL})
\item \code{\link[=as.data.frame]{as.data.frame()}}: Convert \code{GVector} to a \code{data.frame}
\item \code{\link[=as.data.table]{as.data.table()}}: Convert \code{GVector} to a \code{data.table}
\item \code{\link[=as.float]{as.float()}}: Convert a \code{GRaster} to a floating-point raster (\strong{GRASS} data type \code{FCELL})
\item \code{\link[=as.int]{as.int()}}: Convert a \code{GRaster} to an integer raster (\strong{GRASS} data type \code{CELL})
\item \code{\link[=as.points]{as.points()}}, \code{\link[=as.lines]{as.lines()}}, and \code{\link[=as.polygons]{as.polygons()}}: Convert a \code{GRaster} to a \code{GVector}
\item \code{\link[=categories]{categories()}} and \code{\link[fasterRaster]{levels<-}}: Convert an integer raster to a categorical ("factor") raster.
\item \code{\link[=fast]{fast()}}: Convert a \code{SpatRaster} to a \code{GRaster}; a \code{SpatVector}, \code{sf} vector, numeric vector, \code{matrix}, \code{data.frame}, or \code{data.table} to a \code{GVector}; or load a vector or raster from a file
\item \code{\link[=rast]{rast()}}: Convert a \code{GRaster} to a \code{SpatRaster}
\item \code{\link[=rasterize]{rasterize()}}: Convert a \code{GVector} to a \code{GRaster}
\item \code{\link[=st_as_sf]{st_as_sf()}}: Convert a \code{GVector} to a \code{sf} vector
\item \code{\link[=vect]{vect()}}: Convert a \code{GVector} to a \code{SpatVector}
}
}

\subsection{General purpose functions}{
\itemize{
\item \code{\link[=addons]{addons()}}: Show installed \strong{GRASS} addons
\item \code{\link[=compareGeom]{compareGeom()}}: Determine if geographic metadata is same between \code{GRaster}s and/or \code{GVector}s
\item \code{\link[=dropRows]{dropRows()}}: Remove rows from a \code{data.frame} or \code{data.table}
\item \code{\link[=grassGUI]{grassGUI()}}: Start the \strong{GRASS} GUI (not recommended for most users!!!)
\item \code{\link[=grassHelp]{grassHelp()}}: Open the help page for a \strong{GRASS} tool.
\item \code{\link[=grassInfo]{grassInfo()}}: \strong{GRASS} version and citation
\item \code{\link[=grassStarted]{grassStarted()}}: Has a connection \strong{GRASS} been made within the current \strong{R} session?
\item \code{\link[=installAddon]{installAddon()}}: Install a \strong{GRASS} addon
\item \code{\link[=mow]{mow()}}: Remove unused rasters and vectors from the \strong{GRASS} cache
\item \code{\link[=reorient]{reorient()}}: Convert degrees between 'north-orientation' and 'east orientation'
\item \code{\link[=replaceNAs]{replaceNAs()}}: Replace \code{NA}s in columns of a \code{data.table} or \code{data.frame}, or in a vector
\item \code{\link[=removeAddon]{removeAddon()}}: Delete \strong{GRASS} addon from your system
\item \code{\link[=seqToSQL]{seqToSQL()}}: Format a numeric series into an SQL value call
\item \code{\link[=update]{update()}}: Refresh metadata in a \code{GRaster} or \code{GVector} object
}
}

\subsection{Data objects}{
\itemize{
\item \code{\link[=fastData]{fastData()}}: Helper function to quickly obtain example rasters and vectors
\item \link{appFunsTable} (see also \code{\link[=appFuns]{appFuns()}}): Functions usable by the \code{\link[=app]{app()}} function
\item \link{madChelsa}: Climate rasters for of a portion of eastern Madagascar
\item \link{madCoast0}, \link{madCoast4}, and \link{madCoast}: Borders of an eastern portion of Madagascar
\item \link{madCover}: Land cover raster
\item \link{madCoverCats}: Table of land cover classes
\item \link{madDypsis}: Specimens records of species in the genus \emph{Dypsis}
\item \link{madElev}: Elevation raster
\item \link{madForest2000} and \link{madForest2014}: Forest cover in 2000 and 2014
\item \link{madLANDSAT}: Surface reflectance in 2023
\item \link{madPpt}, \link{madTmin}, \link{madTmax}: Rasters of mean monthly precipitation, and minimum and maximum temperature
\item \link{madRivers}: Rivers vector
\item \link{vegIndices}: Vegetation indices that can be calculated using \code{\link[=vegIndex]{vegIndex()}}
}
}

\subsection{Esoteric tutorials and arcane notes}{
\itemize{
\item Comparisons between \code{GRegion}s can be performed using the \code{==} and \code{!=} operators.
\item Vignette on \strong{GRASS} "projects/locations" and "mapsets": \code{vignette("projects_mapsets", package = "fasterRaster")}
\item Vignette on \strong{GRASS} "regions": \code{vignette("regions", package = "fasterRaster")}
\item Vignette on \strong{GRASS} 3-dimensional \code{GRaster}s and \code{GVector}s: \code{vignette("three_d_objects", package = "fasterRaster")}
\item Vignette on \strong{fasterRaster} hidden functions: \code{vignette("hidden_functions", package = "fasterRaster")}
}
}

\subsection{Classes}{
\itemize{
\item \link{GLocation}: Fundamental class; points to a "location/project" in \strong{GRASS}
\item \link{GSpatial}: Basic class of any spatial object
\item \link{GRegion}: Points to a "region" of a "location/project" in \strong{GRASS}
\item \link{GRaster}: Raster class
\item \link{GVector}: Spatial vector class
}
}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/adamlilith/fasterRaster}
  \item \url{https://adamlilith.github.io/fasterRaster/}
  \item Report bugs at \url{https://github.com/adamlilith/fasterRaster/issues}
}

}
\author{
Adam B. Smith
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exists.r
\name{.exists,GRaster-method}
\alias{.exists,GRaster-method}
\alias{.exists}
\alias{.exists,GVector-method}
\alias{.exists,character-method}
\title{Does the "GRASS" representation of a GRaster or GVector exist?}
\usage{
\S4method{.exists}{GRaster}(x)

\S4method{.exists}{GVector}(x)

\S4method{.exists}{character}(x)
}
\arguments{
\item{x}{A \code{GRaster}, \code{GVector}, or the \code{\link[=sources]{sources()}} name of one.}
}
\value{
Logical.
}
\description{
\code{GRaster}s and \code{GVector}s are \strong{R} objects that contain a pointer to a raster or vector in \strong{GRASS}. Thus, for a \code{GRaster} or \code{GVector} to be functional, the \strong{GRASS} file must exist. This function indicates if that is so.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vValidCats.r
\name{.vValidCats}
\alias{.vValidCats}
\title{Are the category values of a vector valid?}
\usage{
.vValidCats(x)
}
\arguments{
\item{x}{A \code{GVector} or the \code{\link[=sources]{sources()}} name of one.}
}
\value{
Logical.
}
\description{
Category values of a \strong{GRASS} vector can be invalid if \strong{GRASS} assigns more than one value to a geometry (e.g., "7/12"). This can occur when the vector was created by software that does not use a topological system (e.g., a shapefile).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vCats.r
\name{.vCats}
\alias{.vCats}
\title{Category column values of a GRASS vector}
\usage{
.vCats(x, layer = 1, db = FALSE, integer = TRUE)
}
\arguments{
\item{x}{A \code{GVector} or the name of a vector in \strong{GRASS}.}

\item{layer}{Integer, numeric integer, or character: Layer from which to obtain category values.}

\item{db}{Logical: If \code{TRUE}, return category numbers from the database table associated with the vector. If \code{FALSE} (default), return category numbers from the actual vector.}

\item{integer}{Logical: If \code{TRUE} (default), return category values as integers. In some cases, a geometry can have multiple categories, in which case \code{NA} is returned. If \code{FALSE}, return category values as strings (and thus, if a geometry has more than one category, does not convert to \code{NA}).}
}
\value{
A vector.
}
\description{
Returns values in the \code{cat} column of a vector in \strong{GRASS}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/backdoor.r
\name{.backdoor}
\alias{.backdoor}
\title{Setup fasterRaster for ABS}
\usage{
.backdoor(start = FALSE, ver = "84")
}
\arguments{
\item{start}{Logical: If \code{TRUE}, start the \strong{GRASS} session by creating the \code{madElev} \code{GRaster}.}

\item{ver}{Character: \strong{GRASS}: e.g., "83" or "84".}
}
\value{
\code{TRUE} (invisibly).
}
\description{
This is a secret function to be used for faster development of \strong{fasterRaster}. It calls \code{\link[=faster]{faster()}} to set the install directory for \strong{GRASS}, increases default memory, and number of cores. The function assumes development is on a Windows machine.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as.polygons.r
\name{as.polygons,GRaster-method}
\alias{as.polygons,GRaster-method}
\alias{as.polygons}
\title{Convert a raster to a polygons vector}
\usage{
\S4method{as.polygons}{GRaster}(x, round = TRUE, smooth = FALSE)
}
\arguments{
\item{x}{A \code{GRaster}. If more than one layer is in the \code{GRaster}, only the first will be used (with a warning).}

\item{round}{Logical: If \code{TRUE} (default), values in the raster will be rounded first before conversion to a vector. This causes cells that are adjacent that have the same (rounded) values to be combined into a single polygon. For more control, see \code{\link[=clump]{clump()}}.}

\item{smooth}{Logical: If \code{TRUE}, round the corners of square features. Default is \code{FALSE}.}
}
\value{
A \code{GVector}.
}
\description{
\code{\link[=as.polygons]{as.polygons()}} converts a \code{GRaster} to a "polygons" \code{GVector}. After running this function, \link[=breakPolys]{geometry cleaning} may be useful to use to "tidy up" the vector.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Elevation
madElev <- fastData("madElev")

# Convert SpatRaster to GRaster:
elev <- fast(madElev)

# To speed things up, first group cells of similar value:
elevClumps <- clump(elev, minDiff = 0.0115)

# Convert to polygons:
rastToPolys <- as.polygons(elevClumps)
plot(rastToPolys)

}
}
\seealso{
\code{\link[=as.points]{as.points()}}, \code{\link[=as.lines]{as.lines()}}, \code{\link[terra:as.polygons]{terra::as.polygons()}}, \link[=breakPolys]{geometry cleaning}, and \strong{GRASS} tool \code{r.to.vect}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rbind.r
\name{rbind,GVector-method}
\alias{rbind,GVector-method}
\alias{rbind}
\title{Combine one or more GVectors}
\usage{
\S4method{rbind}{GVector}(..., deparse.level = 1)
}
\arguments{
\item{...}{One or more \code{GVector}s.}

\item{deparse.level}{See \code{\link[=rbind]{rbind()}}.}
}
\value{
A \code{GVector}.
}
\description{
\code{rbind()} combines two or more \code{GVector}s of the same type (points, lines, or polygons) and same coordinate reference system. You can speed operations by putting the vector that is largest in memory first in \code{rbind(...)}. If the \code{GVector}s have data tables, these will also be combined using \code{rbind()} if their column names and data types match.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)

# Rivers vector
madRivers <- fastData("madRivers")

# Convert sf to a GVector
rivers <- fast(madRivers)

# Convert GVector to data.frame or data.table
as.data.frame(rivers)
as.data.table(rivers)

# Subset rivers vector
rivers1 <- rivers[1:2]
rivers2 <- rivers[10:11]

# Concatenate rivers
riversCombo <- rbind(rivers1, rivers2)
riversCombo

# Add columns
newCol <- data.frame(new = 1:11)
riversCol <- colbind(rivers, newCol)
riversCol

# Remove table
riversCopy <- rivers
riversCopy # has data table
riversCopy <- dropTable(riversCopy)
riversCopy # no data table

# Add a new table
newTable <- data.frame(num = 1:11, letters = letters[1:11])
addTable(riversCopy) <- newTable
riversCopy

}
}
\seealso{
\code{\link[=colbind]{colbind()}}, \code{\link[fasterRaster]{addTable<-}}, \code{\link[=dropTable]{dropTable()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/addons.r
\name{addons}
\alias{addons}
\alias{installAddon}
\alias{removeAddon}
\alias{.addons}
\title{Test if addons directory exists and if an addon is installed}
\usage{
addons(x = NULL)

installAddon(x, check = TRUE)

removeAddon(x, check = TRUE)

.addons(x)
}
\arguments{
\item{x}{Either \code{NULL} or a character specifying the name of a \strong{GRASS} addons tool. If \code{NULL}, a vector of installed addons is returned. If a character vector is provided, a logical vector is returned, one per value in \code{x} that indicates if the respective addon is installed.}

\item{check}{Logical: If \code{TRUE}, check to see if the addon is available (\code{installAddon()})or if it is installed (\code{removeAddon()}).}
}
\value{
\code{addons()}: Logical. The other functions invisibly return a logical value indicating if the operation succeeded or not.
}
\description{
These functions handle \strong{GRASS} addons, which are optional tools that can be installed. Most functions in \strong{fasterRaster} rely on "base" \strong{GRASS} tools (not addons), but a few do.
\itemize{
\item \code{addons()}: Either lists all installed addons or verifies if one or more specific addons are installed.
\item \code{installAddon()}: Installs a \strong{GRASS} addon. An addon typically only needs installed once. You can install an addon, quit and restart \strong{R}, attach \strong{fasterRaster}, and any installed addons can be used without using this function again.
\item \code{removeAddon()}: Delete an installed addon from your system.
}
}
\examples{
if (grassStarted()) {

# What addons are installed?
addons()

# Is a specific addon installed?
addons(c("v.centerpoint", "fake.addon"))

}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vIncrementCats.r
\name{.vIncrementCats}
\alias{.vIncrementCats}
\title{Increment category values of a "GRASS" vector}
\usage{
.vIncrementCats(x, add)
}
\arguments{
\item{x}{A \code{GVector} or the \code{\link[=sources]{sources()}} name of one.}

\item{add}{Integer: Value to add to each category value.}
}
\value{
The \code{\link[=sources]{sources()}} name of a \strong{GRASS} vector with category values incremented.
}
\description{
Adds a constant to all category values of a \strong{GRASS} vector. \strong{This function is mostly of use to developers.}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zonal.r
\name{zonal,GRaster,ANY-method}
\alias{zonal,GRaster,ANY-method}
\alias{zonal}
\title{Statistics on cells of a GRaster stratified by cells of another raster}
\usage{
\S4method{zonal}{GRaster,ANY}(x, z, fun = "mean", probs = 0.5)
}
\arguments{
\item{x}{A \code{GRaster} for which to calculate summary statistics.}

\item{z}{A \code{GRaster} or \code{GVector} used to define zones:
\itemize{
\item If \code{z} is a \code{GRaster}, then it must be of type \code{integer} or \code{factor} (see \code{vignette("GRasters", package = "fasterRaster")}). Zones will be established based on cells that have the same value in this raster.
\item If \code{z} is a \code{GVector}, zones will be created for each geometry. If geometries overlap, then the zonal statistics will be calculated for the ones on top. Thus statistics for the zones defined by geometries below these may not represent all the cells covered by that geometry.
}}

\item{fun}{Character vector: Name of the function(s) to summarize \code{x} with. These can include:
\itemize{
\item \code{"*"}: All of the functions below.
\item \code{"cv"}: Sample coefficient of variation (expressed as a proportion of the mean).
\item \code{"cvpop"}: Population coefficient of variation (expressed as a proportion of the mean).
\item \code{"max"} and \code{"min"}: Highest and lowest values across non-\code{NA} cells.
\item \code{"mean"} (default): Average.
\item \code{"meanAbs"}: Mean of absolute values.
\item \code{"median"}: Median.
\item \code{"quantile"}: Quantile (see also argument \code{probs}).
\item \code{"range"}: Range.
\item \code{"sd"}: Sample standard deviation.
\item \code{"sdpop"}: Population standard deviation.
\item \code{"sum"}: Sum.
\item \code{"var"}: Sample variance.
\item \code{"varpop"}: Population variance.
}}

\item{probs}{Numeric: Quantile at which to calculate \code{quantile}. Only a single value between 0 and 1 is allowed.}
}
\value{
A \code{data.frame} or \code{data.table}.
}
\description{
Function \code{zonal()} calculates statistics (mean, sum, etc.) on cells of a \code{GRaster} by "zones" created by cells of another \code{GRaster} or \code{GVector}.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Elevation SpatRaster:
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster:
elev <- fast(madElev)

### Calculate zonal statistics using a GRaster as zones

# Generate a "zones" GRaster by dividing raster into areas based on
# high/low elevation.
names(elev) # Use this name in app() formula.
fun <- "= if (madElev <200, 0, if (madElev <400, 1, 2))"
zones <- app(elev, fun = fun)

# Calculate zonal statistics using a raster as zones
zonal(elev, zones, fun = "mean")
zonal(elev, zones, fun = "*") # all statistics

# Calculate zonal statistics on multi-layered GRaster
elev2 <- c(elev, log10(elev))
zonal(elev2, zones, fun = c("mean", "sum", "sdpop"))

### Calculate zonal statistics using a GVector as zones

madCoast4 <- fastData("madCoast4")
coast <- fast(madCoast4)

zonal(elev, z = coast, fun = "mean")

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zonal.r
\name{.zonal}
\alias{.zonal}
\title{Internal function for zonal()}
\usage{
.zonal(x, z, fun, probs, zones, xnames)
}
\arguments{
\item{x}{\code{\link[=sources]{sources()}} name of \code{GRaster}.}

\item{z}{\code{\link[=sources]{sources()}} name of "zones" \code{GRaster}.}

\item{fun}{Character: Name of function(s).}

\item{probs}{Numeric in the range 0 to 1, inclusive.}

\item{zones}{Vector of zone values (integers).}

\item{xnames}{Character: Names of \code{x}.}
}
\description{
Internal function for zonal()
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grassStarted.r
\name{grassStarted}
\alias{grassStarted}
\title{Has "GRASS" been started or not?}
\usage{
grassStarted()
}
\value{
Logical.
}
\description{
Returns \code{TRUE} or \code{FALSE}, depending on whether a \strong{GRASS} connection has been made or not within the current \strong{R} session. Usually used only by developers. \strong{GRASS} is started the first time \code{\link[=fast]{fast()}} is used.
}
\examples{

grassStarted()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grassHelp.r
\name{grassHelp}
\alias{grassHelp}
\title{Open the help page for a GRASS tool}
\usage{
grassHelp(x, online = FALSE)
}
\arguments{
\item{x}{Character: Any of:
\itemize{
\item The name of a \strong{GRASS} tool (e.g., \code{"r.mapcalc"}).
\item \code{"toc"}: \strong{GRASS} manual table of contents.
\item \code{"index"}: Display an index of topics.
}}

\item{online}{Logical: If \code{FALSE} (default), show the manual page that was included with your installation of \strong{GRASS} on your computer.  If \code{FALSE}, show the manual page online (requires an Internet connection). In either case, the manual page will display for the version of \strong{GRASS} you have installed.}
}
\value{
Nothing (opens a web page).
}
\description{
This function opens the manual page for a \strong{GRASS} tool (function) in your browser.
}
\examples{
if (grassStarted() & interactive()) {

# Open help pages for `r.mapcalc` and `r.sun`:
grassHelp("r.mapcalc")
grassHelp("r.sun")

# GRASS table of contents:
grassHelp("toc")

# Index page:
grassHelp("index")

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ls.r
\name{.ls}
\alias{.ls}
\title{List objects in the active GRASS session}
\usage{
.ls(type = c("rasters", "vectors", "rasters3d", "groups"))
}
\arguments{
\item{type}{The type of spatial objects to display. This can include \code{"rasters"} (all rasters), \code{"vectors"} (all spatial vectors), \code{"rasters3d"} (3D-rasters), and/or \code{"groups"} (groups). Partial matching is supported. If missing, all objects are displayed.}
}
\value{
Character vector of names of \strong{GRASS} objects.
}
\description{
Display the \code{sources} (see \code{sources()}) of all rasters and/or vectors that have been exported to or created in the active \strong{GRASS} session"s location and mapset.
}
\seealso{
\code{\link[=ls]{ls()}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/princomp.r
\name{princomp,GRaster-method}
\alias{princomp,GRaster-method}
\alias{princomp}
\title{Apply a principal component analysis (PCA) to layers of a GRaster}
\usage{
\S4method{princomp}{GRaster}(x, scale = TRUE, scores = FALSE)
}
\arguments{
\item{x}{A \code{GRaster} with two or more layers.}

\item{scale}{Logical: If \code{TRUE} (default), input layers will be rescaled by dividing each layer by its overall population standard deviation. Rasters will always be centered (have their mean subtracted from values). Centering and scaling is recommended when rasters values are in different units.}

\item{scores}{Logical: If \code{TRUE}, the \code{prcomp} object will have the scores attached to it. This can greatly increase the size of the object in memory if the input raster has many cells. It will also take more time. If \code{FALSE} (default), then skip returning scores.}
}
\value{
A multi-layer \code{GRaster} with one layer per principal component axis. The \code{\link[=pcs]{pcs()}} function can be used on the output raster to retrieve a \code{prcomp} object from the raster, which includes rotations (loadings) and proportions of variance explained.
}
\description{
This function applies a principal component analysis to layers of a \code{GRaster}.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Climate raster:
madChelsa <- fastData("madChelsa")

# Convert a SpatRaster to a GRaster:
chelsa <- fast(madChelsa)

# Generate raster with layers representing principal component predictions:
pcRast <- princomp(chelsa, scale = TRUE)
plot(pcRast)

# Get information on the PCA:
prinComp <- pcs(pcRast)

prinComp
summary(prinComp)
plot(prinComp)

}
}
\seealso{
\code{\link[terra:princomp]{terra::princomp()}}, \code{\link[terra:prcomp]{terra::prcomp()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/project.r
\name{project,GRaster-method}
\alias{project,GRaster-method}
\alias{project}
\alias{project,GVector-method}
\title{Change the coordinate reference system of a GRaster or GVector}
\usage{
\S4method{project}{GRaster}(
  x,
  y,
  align = FALSE,
  method = NULL,
  fallback = TRUE,
  res = "fallback",
  wrap = FALSE,
  verbose = FALSE
)

\S4method{project}{GVector}(x, y, wrap = FALSE)
}
\arguments{
\item{x}{A \code{GRaster} or \code{GVector} to be projected.}

\item{y}{A character or \code{GLocation} object (i.e., typically a \code{GRaster} or \code{GVector}): Used to set the focal \code{GRaster} or \code{GVector}'s new CRS (and resolution and possibly extent, for \code{GRaster}s).}

\item{align}{Logical: If \code{FALSE} (default), and \code{x} and \code{y} are \code{GRaster}s, then the extent of \code{x} will be cropped to the extent of \code{y}. If \code{TRUE}, no cropping is performed.}

\item{method}{Character or \code{NULL} (for \code{GRaster}s only): Method to use to conduct the transformation (rasters only). Partial matching is used.
\itemize{
\item \code{NULL} (default): Automatically choose based on raster properties (\code{near} for categorical data, \code{bilinear} for continuous data).
\item \code{"near"}: Nearest neighbor. Best for categorical data, and often a poor choice for continuous data.  If \code{\link[=datatype]{datatype()}} is \code{integer}, this method will be used by default.
\item \code{"bilinear"}: Bilinear interpolation (default for non-categorical data; uses weighted values from 4 cells).
\item \code{"bicubic"}: Bicubic interpolation (uses weighted values from 16 cells).
\item \code{"lanczos"}: Lanczos interpolation (uses weighted values from 25 cells).
}

\emph{Note #1}: If \code{x} and \code{y} are \code{GRaster}s, and \code{res = "terra"}, then the same \code{method} is used to resample \code{x} to the resolution of \code{y} before projecting \code{x}.

\emph{Note #2}: Methods that use multiple cells will cause the focal cell to become \code{NA} if there is at least one cell with an \code{NA} in the cells it draws from. These \code{NA} cells can often be filled using the \code{fallback} argument.}

\item{fallback}{Logical (for projecting \code{GRaster}s only): If \code{TRUE} (default), then use "lower" resampling methods to fill in \code{NA} cells when a "higher" resampling method is used. For example, if \code{method = "bicubic"}, \code{NA} cells will be filled in using the \code{bilinear} method, except when that results in \code{NA}s, in which case the \code{near} method will be used. Fallback causes fewer cells to revert to \code{NA} values, so may be better at capturing complex "edges" (e.g., coastlines). Fallback does increase processing time because each "lower" method must be applied, then results merged. Fallback is not used if \code{method = "near"}.}

\item{res}{Character (for projecting \code{GRaster}s only): Method used to set the resolution of a \code{GRaster} in the new CRS. This can be one of three options.  Partial matching is used and case ignored:
\itemize{
\item \code{"terra"}: This method creates an output raster that is as close as possible in values and resolution to the one that \code{\link[terra:project]{terra::project()}} would create. However, for large rasters (i.e., many cells), this can fail because \code{terra::project()} encounters memory limits (it is used internally to create a template). This method resamples the focal raster in its starting CRS, then projects it to the destination CRS.
\item \code{"template"}: This method can only be used if \code{y} is a \code{GRaster}. The output will have the same resolution as \code{y} and possibly the same extent (depending on the value of \code{align}). However, unlike the \code{"terra"} method, cell values will not necessarily be as close as possible to what \code{\link[terra:project]{terra::project()}} would generate (unless \code{method = "near"}). Unlike the \code{"terra"} method, this method does not resample the focal raster in its starting CRS before projecting. For large rasters it will be faster than the \code{"terra"} method (especially if \verb{"method = "near"}), and it should be less likely to fail because of memory limits.
\item Two numeric values: Values for the new resolution (x- and y-dimensions).
\item \code{"center"}: This method locates the centroid of the raster to be projected (in the same CRS as the original raster). It then creates four points north, south, east, and west of the centroid, each spaced one cell's width from the centroid. This set of points is then projected to the new CRS. The new cell size in the x-dimension will be the average of the distance between the east and west points from the centroid, and in the y-dimension the average from the centroid to the north and south points.
\item \code{"fallback"} (default): This applies the \code{terra} method first, but if that fails, then tries \code{template}, then \code{center}. This process can take a long time for large rasters.
}}

\item{wrap}{Logical:
\itemize{
\item \code{GRaster}s: When projecting rasters that "wrap around" (i.e., whole-world rasters or rasters that have edges that actually circle around to meet on the globe), \code{wrap} should be \code{TRUE} to avoid removing rows and columns from the "edge" of the map. The default is \code{FALSE}.
\item \code{GVector}s: When projecting vectors that span the international date line at 180E/W, \code{wrap} should be \code{TRUE} to avoid an issue where the coordinates are incorrectly mapped to the range -180 to 180.
}}

\item{verbose}{Logical (for projecting \code{GRaster}s only): If \code{TRUE}, display progress. Default is \code{FALSE}.}
}
\value{
A \code{GRaster} or \code{GVector}.
}
\description{
\code{project()} changes the coordinate reference system (CRS) of a \code{GRaster} or \code{GVector}. It has three use cases:
\itemize{
\item \code{x} is a \code{GRaster} and \code{y} is a \code{GRaster}: \code{x} will be projected to the CRS of \code{y} and resampled to have the same resolution as \code{y}. If argument \code{align} is \code{FALSE}, then it will also be cropped to the extent of \code{y}.
\item \code{x} is a \code{GRaster} and \code{y} is a \code{GVector} or a CRS string (typically in Well-Known Text format): \code{x} will be projected to the CRS specified by \code{y} and resampled but not cropped.
\item \code{x} is a \code{GVector} and \code{y} is a \code{GRaster}, \code{GVector}, or CRS string: The vector will be projected to the CRS of \code{y}.
}
}
\details{
When projecting a raster, the "fallback" methods in \strong{GRASS} tool \code{r.import} are actually used, even though the \code{method} argument takes the strings specifying non-fallback methods. See the manual page for the \code{r.import} \strong{GRASS} tool.
}
\examples{
if (grassStarted()) {

### Setup for all examples

library(sf)
library(terra)

# Climate raster, elevation raster, rivers vector
madElev <- fastData("madElev")
madRivers <- fastData("madRivers")
madChelsa <- fastData("madChelsa")

# Convert objects into fasterRaster formats
chelsa <- fast(madChelsa)
elev <- fast(madElev)
rivers <- fast(madRivers)

### Project raster without resampling
elevWGS84 <- project(elev, crs(chelsa))
elevWGS84

### Project raster and resample to resolution of another raster
elevWGS84Resamp <- project(elev, chelsa)
elevWGS84Resamp

res(elevWGS84)
res(elevWGS84Resamp)
res(chelsa)

### Project vector
riversWGS84 <- project(rivers, chelsa)
riversWGS84
cat(crs(rivers)) # using "cat()" to make it look nice
cat(crs(riversWGS84))

}
}
\seealso{
\code{\link[terra:project]{terra::project()}}, \code{\link[sf:st_transform]{sf::st_transform()}}, \strong{GRASS} manual pages for modules \code{r.proj} and \code{v.proj} (see \code{grassHelp("r.proj")} and \code{grassHelp("v.proj")})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/segregate.r
\name{segregate,GRaster-method}
\alias{segregate,GRaster-method}
\alias{segregate}
\title{Create one GRaster layer per unique value in a GRaster}
\usage{
\S4method{segregate}{GRaster}(x, classes = NULL, keep = FALSE, other = 0, bins = 100, digits = 3)
}
\arguments{
\item{x}{A \code{GRaster}.}

\item{classes}{Either \code{NULL} (default) or a character vector with category labels for which to create outputs. If the input is not a categorical/factor or \code{integer} \code{GRaster}, this is ignored.}

\item{keep}{Logical: If \code{FALSE} (default), then the original value in the input \code{GRaster} will be retained in the each of the output \code{GRaster} layers wherever the input had the respective value. Other cells will be assigned a value of \code{other}.}

\item{other}{Numeric or \code{NA}: Value to assign to cells that do not have the target value.}

\item{bins}{Numeric: Number of bins in which to put values. This is only used for \code{GRaster}s that are not categorical/factor rasters or \code{integer} rasters.}

\item{digits}{Numeric: Number of digits to which to round input if it is a \code{numeric} or \code{double} \code{GRaster} (see \code{vignettes("GRasters", package = "fasterRaster")}).}
}
\value{
If the input \code{x} is a single-layered \code{GRaster}, the output will be a multi-layered \code{GRaster} with one layer per value in the input, or one layer per values in \code{classes}. If the input is a multi-layered \code{GRaster}, the output will be a \code{list} of multi-layered \code{GRaster}s.
}
\description{
This function creates a multi-layered \code{GRaster} for every unique values in an input \code{GRaster}. By default, the output will have a value of 1 wherever the input has the given value, and 0 elsewhere. This is useful for creating dummy variable \code{GRaster} layers for use with models that have factors, especially if the input \code{GRaster} is categorical. Note that the \code{\link[=predict]{predict()}} function in \strong{fasterRaster} usually does not need this treatment of \code{GRaster}s since it can handle categorical rasters already.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Elevation and land cover raster
madElev <- fastData("madElev") # integer raster
madCover <- fastData("madCover") # categorical raster

# Convert to GRasters
elev <- fast(madElev)
cover <- fast(madCover)

# Subset elevation raster to just a few values to make example faster:
elevSubset <- elev[elev <= 3]
segregate(elevSubset)
segregate(elevSubset, keep = TRUE, other = -1)

# Segregate the factor raster
segregate(cover)

classes <- c("Grassland with mosaic forest", "Mosaic cropland/vegetation")
seg <- segregate(cover, classes = classes)
plot(seg)

}
}
\seealso{
\code{\link[terra:segregate]{terra::segregate()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hist.r
\name{hist,GRaster-method}
\alias{hist,GRaster-method}
\alias{hist}
\title{Plot a histogram of raster values}
\usage{
\S4method{hist}{GRaster}(x, layer, maxnl = 16, bins = 30, freq = TRUE, ...)
}
\arguments{
\item{x}{A \code{GRaster}.}

\item{layer}{Character, numeric, or integer: Indicates which layer of a multi-layer \code{GRaster} for which to plot a histogram. The layer can be identified using its \code{\link[terra:names]{terra::name()}} (character) or index (numeric or integer). If this is missing, then up to \code{maxnl} layers are plotted.}

\item{maxnl}{Maximum number of layers for which to create histograms. This is 16 by default, but ignored if \code{layer} is defined.}

\item{bins}{Positive numeric integer: Number of bins in which to divide values of a raster with continuous values. For \code{integer} and categorical rasters, each value is tallied.}

\item{freq}{Logical: If \code{TRUE} (default), plot the frequency of values. If \code{FALSE}, plot the density of values (i.e., the number in each bin divided by the total number of cells with non-\code{NA} values).}

\item{...}{Arguments to pass to \code{\link[graphics:barplot]{graphics::barplot()}}.}
}
\value{
A named list of \code{data.frame}s (invisibly), one per layer plotted, and creates a graph.
}
\description{
This function creates a histogram of values in \code{GRaster}. The function is modeled after \code{\link[graphics:hist]{graphics::hist()}}, but actually uses \code{\link[graphics:barplot]{graphics::barplot()}}.
}
\examples{
if (grassStarted()) {

# Example data
madElev <- fastData("madElev") # elevation raster
madLANDSAT <- fastData("madLANDSAT") # multi-layer raster
madRivers <- fastData("madRivers") # lines vector

# Convert SpatRaster to GRaster and SpatVector to GVector
elev <- fast(madElev)
rivers <- fast(madRivers)
landsat <- fast(madLANDSAT)

# Plot:
plot(elev)
plot(rivers, add = TRUE)

# Histograms:
hist(elev)
hist(landsat)

# Plot surface reflectance in RGB:
plotRGB(landsat, 3, 2, 1) # "natural" color
plotRGB(landsat, 4, 1, 2, stretch = "lin") # emphasize near-infrared (vegetation)

# Make composite map from RGB layers and plot in grayscale:
comp <- compositeRGB(r = landsat[[3]], g = landsat[[2]], b = landsat[[1]])
grays <- paste0("gray", 0:100)
plot(comp, col = grays)

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/app.r
\name{app,GRaster-method}
\alias{app,GRaster-method}
\alias{app}
\alias{appFuns}
\alias{appCheck,GRaster,character-method}
\alias{appCheck}
\title{Apply a function to a set of rasters}
\usage{
\S4method{app}{GRaster}(x, fun, datatype = "auto", seed = NULL)

appFuns(warn = TRUE)

\S4method{appCheck}{GRaster,character}(x, fun, msgOnGood = TRUE, failOnBad = TRUE)
}
\arguments{
\item{x}{A \code{GRaster} with one or more named layers.}

\item{fun}{Character: The function to apply. This must be written as a character string that follows these rules:
\itemize{
\item It must use typical arithmetic operators like \code{+}, \code{-}, \code{*}, \code{/} and/or functions that can be seen using \code{appFuns(TRUE)}.
\item The \code{\link[=names]{names()}} of the rasters do not match any of the functions in the \code{appFuns(TRUE)} table. Note that \code{x} and \code{y} are forbidden names :(
}

The help page for \strong{GRASS} tool \code{r.mapcalc} will be especially helpful. You can see this page using \code{grassHelp("r.mapcalc")}.}

\item{datatype}{Character: This ensures that rasters are treated as a certain type before they are operated on. This is useful when using rasters that have all integer values, which \strong{GRASS} can assume represent integers, even if they are not supposed to. In this case, the output of operations on this raster might be an integer if otherwise not corrected. Partial matching is used, and options include:
\itemize{
\item \code{"integer"}: Force all rasters to integers by truncating their values. The output may still be of type \code{float} if the operation creates non-integer values.
\item \code{"float"}: Force rasters to be considered floating-point values.
\item \code{"double"}: Force rasters to be considered double-floating point values.
\item \code{"auto"} (default): Ensure that rasters are represented by their native \code{\link[=datatype]{datatype()}} (i.e., "CELL" rasters as integers, "FCELL" rasters as floating-point, and "DCELL" as double-floating point).
}}

\item{seed}{Numeric integer vector or \code{NULL} (default): A number for the random seed. Used only for \code{app()} function \code{rand()}, that generates a random number. If \code{NULL}, a seed will be generated. Defining the seed is useful for replicating a raster made with \code{rand()}. This must be an integer!}

\item{warn}{Logical (function \code{appFuns()}): If \code{TRUE} (default), display a warning when \code{allFuns()} is not called interactively.}

\item{msgOnGood}{Logical (function \code{appCheck()}): If \code{TRUE} (default), display a message if no overt problems with the raster names and formula are detected.}

\item{failOnBad}{Logical (function \code{appCheck()}): If \code{TRUE} (default), fail if overt problems with raster names and the formula are detected.}
}
\value{
A \code{GRaster}.
}
\description{
\code{app()} applies a function to a set of "stacked" rasters. It is similar to the \code{\link[terra:app]{terra::app()}} and \code{\link[terra:lapp]{terra::lapp()}} functions.

\code{appFuns()} provides a table of \strong{GRASS} functions that can be used by \code{app()} and their equivalents in \strong{R}.

\code{appCheck()} tests whether a formula supplied to \code{app()} has any "forbidden" function calls.

The \code{app()} function operates in a manner somewhat different from \code{\link[terra:app]{terra::app()}}. The function to be applied \emph{must} be written as a character string. For example, if the \code{GRaster} had layer names "\code{x1}" and "\code{x2}", then the function might be like \code{"= max(sqrt(x1), log(x2))"}. Rasters \strong{cannot} have the same names as functions used in the formula. In this example, the rasters could not be named "max", "sqrt", or "log". Note that the name of a \code{GRaster} is given by \code{\link[=names]{names()}}--this can be different from the name of the object in \strong{R}.

The \code{app()} function will automatically check for \code{GRaster} names that appear also to be functions that appear in the formula. However, you can check a formula before running \code{app()} by using the \code{appCheck()} function. You can obtain a list of \code{app()} functions using \code{appFuns()}. Note that these are sometimes different from how they are applied in \strong{R}.

Tips:
\itemize{
\item Make sure your \code{GRaster}s have \code{names()}. The function matches on these, not the name of the variable you use in \strong{R} for the \code{GRaster}.
\item Use \code{null()} instead of \code{NA}, and use \code{isnull()} instead of \code{is.na()}.
\item If you want to calculate values using while ignoring \code{NA} (or \code{null}) values, see the functions that begin with \code{n} (like \code{nmean}).
\item Be mindful of the data type that a function returns. In \strong{GRASS}, these are \code{CELL} (integer), \code{FCELL} (floating point values--precise to about the 7th decimal place), and \code{DCELL} (double-floating point values--precise to about the 15th decimal place; commensurate with the \strong{R} \code{numeric} type). In cases where you want a \code{GRaster} to be treated like a float or double type raster, wrap the name of the \code{GRaster} in the \code{float()} or \code{double()} functions. This is especially useful if the \code{GRaster} might be assumed to be the \code{CELL} type because it only contains integer values. You can get the data type of a raster using \code{\link[=datatype]{datatype()}} with the \code{type} argument set to \code{GRASS}. You can change the data type of a \code{GRaster} using \code{\link[=as.int]{as.int()}}, \code{\link[=as.float]{as.float()}}, and \code{\link[=as.doub]{as.doub()}}. Note that categorical rasters are really \code{CELL} (integer) rasters with an associated "levels" table. You can also change a \code{CELL} raster to a \code{FCELL} raster by adding then subtracting a decimal value, as in \code{x - 0.1 + 0.1}. See \code{vignette("GRasters", package = "fasterRaster")}.
\item The \code{rand()} function returns integer values by default. If you want non-integer values, use the tricks mentioned above to datatype non-integer values. For example, if you want uniform random values in the range between 0 and 1, use something like \verb{= float(rand(0 + 0.1, 1 + 0.1) - 0.1)}.
\item \code{GRaster}s with short names can cause the function to mis-specify the formula in \strong{GRASS}. This happens when, for example, the \code{fun} argument contains an "if" statement, and the raster name is "i" or "f", which would incorrectly replace those letters in the "if" with the \strong{GRASS} name of the raster. To avoid this, avoid using short raster names that could be part of longer operators using by \code{app()} (see \code{\link[=appFuns]{appFuns()}} for a list of these operators).
}
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Elevation raster
madElev <- fastData("madElev")

# Convert SpatRaster to a GRaster:
elev <- fast(madElev)

# Create a "stack" of rasters for us to operate on:
x <- c(elev, elev^2, sqrt(elev))

# Demonstrate check for badly-named rasters:
names(x) <- c("cos", "asin", "exp")
fun <- "= cos / asin + exp"
appCheck(x, fun, failOnBad = FALSE)

# Rename rasters acceptable names and run the function:
names(x) <- c("x1", "x2", "x3")
fun <- "= (x1 / x2) + x3"
appCheck(x, fun, failOnBad = FALSE)
app(x, fun = fun)

# This is the same as:
(x[[1]] / x[[2]]) + x[[3]]

# We can view a table of app() functions using appFuns():
appFuns()

# We can also get the same table using:
data(appFunsTable)

# Apply other functions:
fun <- "= median(x1 / x2, x3, x1 * 2, cos(x2))"
app(x, fun = fun)

fun <- "= round(x1) * tan(x2) + log(x3, 10)"
app(x, fun = fun)

# Demonstrate effects of data type:
fun <- "= x1 + x3"
app(x, fun = fun, datatype = "float") # output is floating-point
app(x, fun = fun, datatype = "integer") # output is integer

# Some functions override the "datatype" argument:
fun <- "= sin(x2)"
app(x, fun = fun, datatype = "integer")

# Make a raster with random values [1:4], with equal probability of each:
fun <- "= round(rand(0.5, 4.5))"
rand <- app(elev, fun = fun)
rand

freqs <- freq(rand) # cell frequencies
print(freqs)

}
}
\seealso{
\code{\link[terra:app]{terra::app()}}, \code{\link[terra:lapp]{terra::lapp()}}, \code{\link[=subst]{subst()}}, \code{\link[=classify]{classify()}}, and especially the \strong{GRASS} manual page for tool \code{r.mapcalc} (see \code{grassHelp("r.mapcalc")})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simplifyGeom.r
\name{simplifyGeom,GVector-method}
\alias{simplifyGeom,GVector-method}
\alias{simplifyGeom}
\title{Simplify the geometry of a vector}
\usage{
\S4method{simplifyGeom}{GVector}(x, tolerance = NULL, method = "VR", prop = 0.5)
}
\arguments{
\item{x}{A \code{GVector}.}

\item{tolerance}{Numeric >= 0: Threshold distance in map units (degrees for unprojected, usually meters for projected). If \code{NULL}, then 2\% of the minimum of the x-, y-, and z-extent will be used.}

\item{method}{Character: Method used to reduce the number of vertices. Partial matching is used, and case does not matter:
\itemize{
\item \code{"VR"}: Vertex reduction (default, simplest): If two points p1 and p2 on the same line are closer than the threshold, remove p2. The \code{tolerance} argument represents this threshold distance.
\item \code{"DP"}: Douglas-Peucker (AKA Ramer-Douglas-Peucker) algorithm: Simply stated, for points p1, p2, and p3 on a line, this method constructs a line segment between p1 and p3. If p2 is closer than the threshold to the line segment, it is removed. In this example, the \code{tolerance} argument refers to the maximum distance between p2 and the line segment.
\item \code{"DPR"}: Douglas-Peucker algorithm with reduction: As the Douglas-Pueker method, but each geometry is thinned so that in the end it has only a given proportion of the starting number of points. The \code{prop} argument refers to this proportion of remaining points.
\item \verb{"RW}: Reumann-Witkam algorithm: For points p1, p2, p3, and p4 on a line, constructs two line segments parallel to the line segment defined by p1 and p4. These are placed \code{tolerance} distance one either side of the p1-p4 line segment. If the line segment p1-p2 or p3-p4 falls entirely within the bounds of the two outer parallel segments, p2 and p3 are removed, leaving just p1 and p4.
}}

\item{prop}{Positive value between 0 and 1: Proportion of points that will be retained for each geometry when the Douglas-Peucker algorithm with reduction is applied (ignored otherwise). Default is 0.5 (retain 50\% of vertices).}
}
\value{
A \code{GVector}.
}
\description{
\code{simplifyGeom()} reduces the number of vertices used to represent a vector (i.e., to save memory or disk space). There are several methods available.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Example data
madRivers <- fastData("madRivers")
rivers <- fast(madRivers)
soam <- rivers[rivers$NAM == "SOAMIANINA"] # select one river for illustration

### Simplify geometry (remove nodes)

vr <- simplifyGeom(soam, tolerance = 2000)
dp <- simplifyGeom(soam, tolerance = 2000, method = "dp")
dpr <- simplifyGeom(soam, tolerance = 2000, method = "dpr", prop = 0.5)
rw <- simplifyGeom(soam, tolerance = 2000, method = "rw")

plot(soam, col = "black", lwd = 3)
plot(vr, col = "blue", add = TRUE)
plot(dp, col = "red", add = TRUE)
plot(dpr, col = "chartreuse", add = TRUE)
plot(rw, col = "orange", add = TRUE)

legend("bottom",
   xpd = NA,
   legend = c(
	  "Original",
      "Vertex reduction",
      "Douglas-Peucker",
      "Douglas-Peucker reduction",
      "Reumann-Witkam"
	),
	col = c("black", "blue", "red", "chartreuse", "orange"),
	lwd = c(3, 1, 1, 1, 1)
)

### Smooth geometry

hermite <- smoothGeom(soam, dist = 2000, angle = 3)
chaiken <- smoothGeom(soam, method = "Chaiken", dist = 2000)

plot(soam, col = "black", lwd = 2)
plot(hermite, col = "blue", add = TRUE)
plot(chaiken, col = "red", add = TRUE)

legend("bottom",
   xpd = NA,
   legend = c(
	  "Original",
      "Hermite",
      "Chaiken"
	),
	col = c("black", "blue", "red"),
	lwd = c(2, 1, 1, 1, 1)
)

### Clean geometry

# Has no effect on this vector!
noDangs <- removeDangles(soam, tolerance = 10000)

plot(soam, col = "black", lwd = 2)
plot(noDangs, col = "red", add = TRUE)

legend("bottom",
   xpd = NA,
   legend = c(
	  "Original",
      "No dangles"
	),
	lwd = c(2, 1),
	col = c("black", "red")
)

}
}
\seealso{
\code{\link[=smoothGeom]{smoothGeom()}}, \link[=breakPolys]{geometry cleaning}, \code{\link[terra:simplify]{terra::simplifyGeom()}}, \strong{GRASS} manual page for tool \code{v.generalize} (see \code{grassHelp("v.generalize")})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/locationRestore.r
\name{.locationRestore,character-method}
\alias{.locationRestore,character-method}
\alias{.locationRestore}
\alias{.locationRestore,integer-method}
\alias{.locationRestore,numeric-method}
\alias{.locationRestore,GSpatial-method}
\title{Revert to a previously-created "GRASS" "location"}
\usage{
\S4method{.locationRestore}{character}(x)

\S4method{.locationRestore}{integer}(x)

\S4method{.locationRestore}{numeric}(x)

\S4method{.locationRestore}{GSpatial}(x)
}
\arguments{
\item{x}{Either:
\itemize{
\item A character: Name of the "location" in \strong{GRASS}.
\item An integer: Index of the "location" in \code{.fasterRaster$locations}.
\item A \code{GSpatial} object (usually a \code{GRaster} or \code{GVector}).
}

Any of these can be found using \code{.locationFind()}.}
}
\value{
An object of class \code{GLocation} (invisibly) if successful. An error will likely result if not.
}
\description{
This function resets the connection to a previously-created \strong{GRASS} "location". The session must have been already created using \code{\link[=fast]{fast()}} in the current \strong{R} session. This function is typically only of use to developers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scale.r
\name{scale,GRaster-method}
\alias{scale,GRaster-method}
\alias{scale}
\alias{scalepop,GRaster-method}
\alias{scalepop}
\alias{unscale,GRaster-method}
\alias{unscale}
\title{Center and scale a GRaster, or the opposite}
\usage{
\S4method{scale}{GRaster}(x, center = TRUE, scale = TRUE)

\S4method{scalepop}{GRaster}(x, center = TRUE, scale = TRUE)

\S4method{unscale}{GRaster}(x, center = NULL, scale = NULL)
}
\arguments{
\item{x}{A \code{GRaster}.}

\item{center}{Value depends on the function:
\itemize{
\item \code{scale()}:
\itemize{
\item Logical: If \code{TRUE} (default), subtract from each raster layer its mean. If \code{FALSE}, do not.
\item Numeric: A single value, in which case the same value will be used across all layers of \code{x}, or one value per layer in \code{x}.
}
\item \code{unscale()}: Numeric vector or \code{NULL} (default): This can be a single value, which will be recycled if there is more than one layer in the raster, or one value per raster layer. If a value is \code{NA}, then no un-centering will be performed on the relevant raster layer. If \code{NULL}, then no un-centering is done.
}}

\item{scale}{Value depends on the function:
\itemize{
\item \code{scale()}:
\itemize{
\item Logical: If \code{TRUE} (default), divide each raster layer by its standard deviation. If \code{FALSE}, do not.
\item Numeric: A single value, in which case the same value will be used across all layers of \code{x}, or one value per layer in \code{x}.
}
\item \code{unscale()}: Numeric vector or \code{NULL} (default): This can be a single value, which will be recycled if there is more than one layer in the raster, or one value per raster layer. If a value is \code{NA}, then no unscaling will be done on the relevant raster layer. If \code{NULL}, then no un-scaling is done.
}}
}
\value{
All functions return a \code{GRaster}. The output of \code{scale()} and \code{scalepop()} will have two attributes, "center" and "scale", which have the means and standard deviations of the original rasters (if \code{center} and \code{scale} are \code{TRUE}, otherwise, they will be \code{NA}). These can be obtained using \code{attributes(output_raster)$center} and \code{attributes(output_raster)$scale}.
}
\description{
\code{scale()} and \code{scalepop()} center and scale layers in a \code{GRaster} by subtracting from each raster its mean value (centering), then dividing by its standard deviation (scaling). This is useful for using the raster in a linear model, for example, because unscaled predictors can lead to numerical instability. The \code{scale()} function uses the sample standard deviation, and the \code{scalepop()} function uses the population standard deviation. For even moderately-sized rasters, the difference between these two is negligible, but the \code{scalepop()} function can be much faster than the \code{scale()} function.

The \code{unscale()} function does the opposite of \code{scale()} and \code{scalepop()}: it multiples each layer by a value (presumably, its standard deviation), and adds another value (presumably, its mean).
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Climate rasters:
madChelsa <- fastData("madChelsa")

# Convert to GRasters:
chelsa <- fast(madChelsa)

### Center and scale rasters
# Scale with using sample SD:
chScaled <- scale(chelsa)
chScaled

# Scale with using population SD:
chScaledPop <- scalepop(chelsa)
chScaledPop

# Means are very close to 0 and SDs to 1:
global(chScaled, c("mean", "sd", "min", "max"))
global(chScaledPop, c("mean", "sd", "min", "max"))

# Get original means and sd's:
centers <- attributes(chScaled)$center
scales <- attributes(chScaled)$scale
centers
scales

### Unscale rasters:
chUnscaled <- unscale(chScaled, center = centers, scale = scales)

# Means and SD are returned to original values:
global(chUnscaled, c("mean", "sd", "min", "max")) # unscaled
global(chelsa, c("mean", "sd", "min", "max")) # original

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/global.r
\name{global,GRaster-method}
\alias{global,GRaster-method}
\alias{global}
\alias{global,missing-method}
\title{Summary statistics for GRasters}
\usage{
\S4method{global}{GRaster}(x, fun = "mean", probs = seq(0, 1, 0.25), ...)

\S4method{global}{missing}(x, ...)
}
\arguments{
\item{x}{A \code{GRaster} or missing.  If missing, then a vector of all of the accepted function names is returned.}

\item{fun}{Character vector: The name of the function(s):
\itemize{
\item \code{"*"}: All of the functions below.
\item \code{"cv"}: Sample coefficient of variation (expressed as a proportion of the mean).
\item \code{"cvpop"}: Population coefficient of variation (expressed as a proportion of the mean).
\item \code{"max"} and \code{"min"}: Highest and lowest values across non-\code{NA} cells. NB: \code{\link[=minmax]{minmax()}} is faster.
\item \code{"mean"} (default): Average.
\item \code{"meanAbs"}: Mean of absolute values.
\item \code{"median"}: Median.
\item \code{"quantile"}: Quantile (see also argument \code{probs}).
\item \code{"range"}: Range. Note that following \code{\link[terra:global]{terra::global()}}, the minimum and maximum are reported, not the actual range.
\item \code{"sd"}: Sample standard deviation (same as \code{\link[stats:sd]{stats::sd()}}).
\item \code{"sdpop"}: Population standard deviation.
\item \code{"sum"}: Sum.
\item \code{"var"}: Sample variance (same as \code{\link[stats:cor]{stats::var()}}).
\item \code{"varpop"}: Population variance.
}}

\item{probs}{Numeric within the range from 0 to 1: Quantile(s) at which to calculate \code{quantile}.}

\item{...}{Other arguments (unused).}
}
\value{
If \code{x} is missing, the function returns a character vector of all accepted function names. If \code{x} is a \code{GRaster}, a data frame with the specified statistics is returned.
}
\description{
\code{global()} calculates a summary statistic across all the cells of a \code{GRaster}. It returns a single value for each layer of the raster.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster:
elev <- fast(madElev)

# What functions can we use with global()?
global()

# Calculate global statistics:
global(elev, fun = c("mean", "var", "varpop"))
global(elev, "quantile", probs = c(0.25, 0.5, 0.75))

global(elev, "*") # calculate all available functions

}
}
\seealso{
\code{\link[terra:global]{terra::global()}} and \strong{GRASS} tool \code{r.univar}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rasterize.r
\name{rasterize,GVector,GRaster-method}
\alias{rasterize,GVector,GRaster-method}
\alias{rasterize}
\title{Convert a GVector to a GRaster}
\usage{
\S4method{rasterize}{GVector,GRaster}(x, y, field = "", background = NA, by = NULL, verbose = TRUE)
}
\arguments{
\item{x}{A \code{GVector}.}

\item{y}{A \code{GRaster}: The new raster will have the same extent and resolution as this raster.}

\item{field}{Character: Name of a column in the data table of \code{y} to "burn" into the raster. If not \code{""} (default), then the output will be a categorical \code{GRraster}. If \code{field} is \code{""}, then all geometries will be "burned" to the raster and have the same value.}

\item{background}{Numeric or \code{NA} (default): Value to put in cells that are not covered by the \code{GVector}. Note that if this is not \code{NA} and not an integer, then the output cannot be a categorical raster (i.e., there will be no "levels" table associated with it).}

\item{by}{Either \code{NULL} (default) or character: If this is not \code{NULL}, then the \code{GVector} will be subset by the values in the field named by \code{by}. The output will be a multi-layer raster, with one layer per unique value in \code{by}.}

\item{verbose}{Logical: If \code{by} is not \code{NULL}, display progress.}
}
\value{
A \code{GRaster}.
}
\description{
The \code{rasterize()} function converts a \code{GVector} into a \code{GRaster}.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Elevation raster, outline of a part of Madagascar, and rivers vector:
madElev <- fastData("madElev") # raster
madDypsis <- fastData("madDypsis") # points vector
madRivers <- fastData("madRivers") # lines vector
madCoast4 <- fastData("madCoast4") # polygons vector

# Convert to GRaster and GVectors:
elev <- fast(madElev)
dypsis <- fast(madDypsis)
coast4 <- fast(madCoast4)
rivers <- fast(madRivers)

# Convert points, line, and polygons vectors to rasters:
points <- rasterize(dypsis, elev)
plot(points)

lines <- rasterize(rivers, elev)
plot(lines)

polys <- rasterize(coast4, elev)
plot(polys)

communes <- rasterize(coast4, elev, field = "NAME_4")
plot(communes)

# Change background value:
polysNeg1 <- rasterize(coast4, elev, background = -1)
plot(polysNeg1)

# Make one layer per river:
byRiver <- rasterize(rivers, elev, field = "NAM", by = "NAM")
plot(byRiver)

}
}
\seealso{
\code{\link[terra:rasterize]{terra::rasterize()}}, \strong{GRASS} tool \code{v.to.rast} (see \code{grassHelp("v.to.rast")})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compositeRGB.r
\name{compositeRGB,GRaster-method}
\alias{compositeRGB,GRaster-method}
\alias{compositeRGB}
\title{Combine red, green, and blue color bands to make a composite GRaster}
\usage{
\S4method{compositeRGB}{GRaster}(r, g = NULL, b = NULL, levels = 256, dither = FALSE)
}
\arguments{
\item{r, g, b}{Either:
\itemize{
\item One \code{GRaster} with one band each for \code{r}, \code{g}, or \code{b} representing red, green, and blue color bands; or
\item \code{r} is  single \code{GRaster} with 3 bands (R, G, and B bands), and \code{g} and \code{b} are \code{NULL}.
}}

\item{levels}{Either a single value that is an integer, or a vector of integers: Number of levels of red, green, and blue intensities represented in \code{r}, \code{g}, and \code{b}. If a single value is supplied, it is assumed that all three have the same number of levels. If three values are supplied, then they correspond to the R, G, and B bands. The default is 256 (assume that R, G, and B rasters have values between 0 and 255).}

\item{dither}{Logical: If \code{TRUE}, apply Floyd-Steinberg dithering. Default is \code{FALSE}.}
}
\value{
A \code{GRaster}.
}
\description{
This function takes as arguments three rasters typically representing red, green, and blue color bands, and returns a single raster with values based on their combination. Typically, this raster should be plotted in grayscale.
}
\examples{
if (grassStarted()) {

# Example data
madElev <- fastData("madElev") # elevation raster
madLANDSAT <- fastData("madLANDSAT") # multi-layer raster
madRivers <- fastData("madRivers") # lines vector

# Convert SpatRaster to GRaster and SpatVector to GVector
elev <- fast(madElev)
rivers <- fast(madRivers)
landsat <- fast(madLANDSAT)

# Plot:
plot(elev)
plot(rivers, add = TRUE)

# Histograms:
hist(elev)
hist(landsat)

# Plot surface reflectance in RGB:
plotRGB(landsat, 3, 2, 1) # "natural" color
plotRGB(landsat, 4, 1, 2, stretch = "lin") # emphasize near-infrared (vegetation)

# Make composite map from RGB layers and plot in grayscale:
comp <- compositeRGB(r = landsat[[3]], g = landsat[[2]], b = landsat[[1]])
grays <- paste0("gray", 0:100)
plot(comp, col = grays)

}
}
\seealso{
\code{\link[=plotRGB]{plotRGB()}}, \code{\link[terra:plotRGB]{terra::plotRGB()}}, \strong{GRASS} manual page for tool \code{r.composite} (see \code{grassHelp("r.composite")})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aggregate.r
\name{aggregate,GRaster-method}
\alias{aggregate,GRaster-method}
\alias{aggregate}
\alias{aggregate,GVector-method}
\title{Aggregate raster cells into larger cells or combine geometries of a vector}
\usage{
\S4method{aggregate}{GRaster}(
  x,
  fact = 2,
  fun = "mean",
  weight = FALSE,
  prob = NULL,
  na.rm = FALSE
)

\S4method{aggregate}{GVector}(x)
}
\arguments{
\item{x}{A \code{GRaster} or \code{GVector}.}

\item{fact}{Numeric vector (rasters only): One, two, or three positive values. These reflect the size of the new cells as multiples of the size of the old cells. If just one value is supplied, this is used for all two or three dimensions. If two values are supplied, the first is multiplied by the east-west size of cells, and the second north-south size of cells (the raster must be 2D). If three values are supplied, the third value is used as the multiplier of the vertical dimension of cells. Values are calculated using all cells that have their centers contained by the target cell.

Note that unlike \code{terra::aggregate()} and \code{\link[terra:disaggregate]{terra::disagg()}}, these values need not be integers.}

\item{fun}{Character (rasters only): Name of the function used to aggregate. For \code{GRaster}s, this is the function that summarizes across cells. For \code{GVector}s, this function will be used to calculate new values of numeric or integer cells.
\itemize{
\item \code{mean}: Average (default)
\item \code{median}: Median
\item \code{mode}: Most common value
\item \code{min}: Minimum
\item \code{max}: Maximum
\item \code{range}: Difference between maximum and minimum
\item \code{sum}: Sum
\item \code{varpop}: Population variance
\item \code{sdpop}: Population standard deviation
\item \code{quantile}: Quantile (see argument \code{prob})
\item \code{count}: Number of non-\code{NA} cell
\item \code{diversity}: Number of unique values
}}

\item{weight}{Logical (rasters only): If \code{FALSE}, each source cell that has its center in the destination cell will be counted equally. If \code{TRUE}, the value of each source will be weighted the proportion of the destination cell the source cell covers.}

\item{prob}{Numeric (rasters only): Quantile at which to calculate \code{quantile}.}

\item{na.rm}{Logical (rasters only): If \code{FALSE} (default), propagate \code{NA} cells or \code{NA} values.}
}
\value{
A \code{GRaster} or \code{GVector}.
}
\description{
When applied to a \code{GRaster}, \code{aggregate()} creates a new raster with cells that are a multiple of the size of the cells of the original raster. The new cells can be larger or smaller than the original cells (this function thus emulates both the \code{terra::aggregate()} and \code{\link[terra:disaggregate]{terra::disagg()}} functions.)

When applied to a \code{GVector}, all geometries are combined into a "multipart" geometry, in which geometries are treated as if they were a single unit. Borders between aggregated geometries can be dissolved if the \code{dissolve} argument is \code{TRUE}. If the \code{GVector} has a data table associated with it, the output will also have a data table as long as there is at least one column with values that are all the same. Values of columns that do not have duplicated values will be converted to \code{NA}.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")
madCoast4 <- fastData("madCoast4")

### aggregating a GRaster

# Convert:
elev <- fast(madElev)

### Aggregate GRaster by same factor in 2 dimensions
# fasterRaster
agg2 <- aggregate(elev, 2, "mean")
agg2

# Compare rasters aggregated by fasterRaster and terra.
# These should be the same.
agg2terra <- aggregate(madElev, 2)

agg2 <- rast(agg2)
agg2 <- extend(agg2, agg2terra)
agg2 - agg2terra # value is ~0

### Aggregate GRaster by a non-integer factor in 2 dimensions
# fasterRaster
agg2.9 <- aggregate(elev, 2.9, "mean")
agg2.9

# terra
agg2.9terra <- aggregate(madElev, 2.9, "mean")
agg2.9terra

# Compare rasters aggregated by fasterRaster and terra.
# These should be different.
res(agg2.9)
res(agg2.9terra) # terra rounds aggregation factor down
2 * res(madElev) # original resolution multiplied by 2

### Aggregate GRaster by different factor in 2 dimensions
agg2x3 <- aggregate(elev, c(2, 3), "mean")
agg2x3

### aggregating a GVector

madCoast4 <- fastData("madCoast4")

# Convert:
coast4 <- fast(madCoast4)

# Aggregate and disaggregate:
aggCoast <- aggregate(coast4)
disaggCoast <- disagg(coast4)

ngeom(coast4)
ngeom(aggCoast)
ngeom(disaggCoast)

# plot
oldpar <- par(mfrow = c(1, 3))
plot(coast4, main = "Original", col = 1:nrow(coast4))
plot(aggCoast, main = "Aggregated", col = 1:nrow(aggCoast))
plot(disaggCoast, main = "Disaggregated", col = 1:nrow(disaggCoast))
par(oldpar)

}
}
\seealso{
\code{\link[stats:aggregate]{stats::aggregate()}}, \code{\link[terra:aggregate]{terra::aggregate()}}, \code{\link[=disagg]{disagg()}}, \code{\link[terra:disaggregate]{terra::disagg()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as.lines.r
\name{as.lines,GRaster-method}
\alias{as.lines,GRaster-method}
\alias{as.lines}
\title{Convert a raster to a lines vector}
\usage{
\S4method{as.lines}{GRaster}(x)
}
\arguments{
\item{x}{A \code{GRaster}. If more than one layer is in the \code{GRaster}, only the first will be used (with a warning).}
}
\value{
A \code{GVector}.
}
\description{
\code{\link[=as.lines]{as.lines()}} converts a \code{GRaster} to a "lines" \code{GVector}. Before you apply this function, you may need to run \code{\link[=thinLines]{thinLines()}} on the raster to reduce linear features to a single-cell width. You may also need to use \link[=breakPolys]{clean geometry} (especially the \code{\link[=removeDups]{removeDups()}} and \code{\link[=removeDangles]{removeDangles()}}) afterward to remove duplicated vertices and "dangling" lines.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Elevation
madElev <- fastData("madElev")

# Convert to GRaster:
elev <- fast(madElev)

# Thin elevation raster:
thinned <- thinLines(elev, iter = 300)
plot(thinned)

# Convert to lines:
rastToLines <- as.lines(thinned)
plot(rastToLines)

# We can clean this:
cleanLines <- fixDangles(x = rastToLines)
plot(rastToLines, col = "red")
plot(cleanLines, add = TRUE)

}
}
\seealso{
\code{\link[=as.points]{as.points()}}, \code{\link[=as.polygons]{as.polygons()}}, \code{\link[terra:as.lines]{terra::as.lines()}}, \code{\link[=thinLines]{thinLines()}}, \link[=breakPolys]{geometry cleaning}, and \strong{GRASS} tool \code{r.to.vect}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/locations.r
\name{.locations}
\alias{.locations}
\title{Meta-data on active GRASS locations}
\usage{
.locations(warn = TRUE)
}
\arguments{
\item{warn}{Logical: If \code{TRUE} (default), display a warning if no \strong{GRASS} "locations" have been created.}
}
\value{
A named \code{list}. The names are the "location's" names and the values are the coordinate reference strings.
}
\description{
\strong{GRASS} uses "locations"to store sets of rasters and vectors with the same coordinate reference system (CRS). These rasters and vectors may or may not be in the same actual location on Earth--they just have the same CRS. This function returns information on all of the \strong{GRASS} "locations" that have been initialized. It is mainly useful for developers.

NB: \strong{fasterRaster} always uses the "PERMANENT" mapset.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/flow.r
\name{flow,GRaster-method}
\alias{flow,GRaster-method}
\alias{flow}
\title{Identify watershed basins and direction and accumulation of flow}
\usage{
\S4method{flow}{GRaster}(
  x,
  direction = "multi",
  return = "accumulation",
  dirThreshold = Inf,
  scratchDir = NULL
)
}
\arguments{
\item{x}{A \code{GRaster} with a single layer, typically representing elevation.}

\item{direction}{Character: Either \code{"single"} or \code{"multi"}. This indicates whether a single-direction flow or multi-direction flow model is used. The default is \code{"multi"}. Partial matching is used and case is ignored.}

\item{return}{Character vector: Indicates what rasters to return. Partial matching is used and case is ignored. Options include:
\itemize{
\item \code{"accumulation"} (default): Flow accumulation raster.
\item \code{"basins"}: Watershed basins
\item \code{"direction"}: Flow direction
\item \code{"flooded"}: Flooded areas
\item \code{"TCI"}: Topographic convergence index
\item \code{"*"}: All of the above
}}

\item{dirThreshold}{Numeric (default is \code{Inf}): For the multi-direction flow model, this indicates the amount of accumulated flow above which the single-direction flow rule is used to locate the egress of water from a cell. This is the \code{d8cut} parameter in \code{r.stream.extract}.}

\item{scratchDir}{Character or \code{NULL} (default): Directory in which to store temporary files. The \strong{GRASS} tool \code{r.terraflow} makes a lot of temporary files. If this is \code{NULL}, then a temporary folder in the user's working directory will be used (see \code{\link[=getwd]{getwd()}}).}
}
\value{
A \code{GRaster}.
}
\description{
The \code{flow()} function uses a raster representing elevation to compute other rasters representing:
\itemize{
\item Flow accumulation;
\item Direction of flow;
\item Watershed basins;
\item Flooded areas; and/or
\item Topographic convergence (log of flow accumulation divided by local slope).
}

More details about the computations can be found at the help page for the \strong{GRASS} tool \code{r.terraflow}] (see \code{grassHelp("r.terraflow")})
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")
elev <- fast(madElev)

# Calculate flow accumulation and watershed basins
water <- flow(elev, return = c("accum", "basins"))
water

elevWater <- c(elev, water)
plot(elevWater)

}
}
\seealso{
\code{\link[=flowPath]{flowPath()}}, \code{\link[=streams]{streams()}}, the \strong{GRASS} tool \code{r.terraflow} (see \code{grassHelp("r.terraflow")})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rvoronoi.r
\name{rvoronoi,GRaster-method}
\alias{rvoronoi,GRaster-method}
\alias{rvoronoi}
\alias{rvoronoi,GVector-method}
\title{Create a randomly-positioned tesselation}
\usage{
\S4method{rvoronoi}{GRaster}(x, size = 100, seed = NULL)

\S4method{rvoronoi}{GVector}(x, size = 100, seed = NULL)
}
\arguments{
\item{x}{A \code{GRaster} or \code{GVector} used to constrain the location of random points used to create the tesselation.}

\item{size}{Numeric integer or integer: Number of polygons.}

\item{seed}{Numeric integer, integer, or \code{NULL} (default): Value used as a random seed. If \code{NULL}, a random seed will be generated by \strong{GRASS}.}
}
\value{
A \code{GVector}.
}
\description{
This function partitions a region into Voronoi polygons that completely overlap it. Each polygon has a random center. The function is essentially a wrapper for \code{\link[=spatSample]{spatSample()}} and \code{\link[=voronoi]{voronoi()}}.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)

# Example vectors
madDypsis <- fastData("madDypsis") # points
madCoast4 <- fastData("madCoast4") # polygons

# Convert sf vectors to GVectors
dypsis <- fast(madDypsis)
coast4 <- fast(madCoast4)
ant <- coast4[coast4$NAME_4 == "Antanambe"]

# Delaunay triangulation
dypsisDel <- delaunay(dypsis)
plot(dypsisDel)
plot(dypsis, pch = 1, col = "red", add = TRUE)

# Voronoi tessellation
vor <- voronoi(dypsis)
plot(vor)
plot(dypsis, pch = 1, col = "red", add = TRUE)

# Random Voronoi tessellation
rand <- rvoronoi(coast4, size = 100)
plot(rand)

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fastData.r
\name{fastData}
\alias{fastData}
\title{Get one of the example rasters or spatial vectors}
\usage{
fastData(x)
}
\arguments{
\item{x}{The name of the raster or spatial vector to get. All of these represent a portion of the eastern coast of Madagascar.

Spatial vectors (objects of class \code{sf} from the \strong{sf} package):
\itemize{
\item \link{madCoast0}: Outline of the region (polygon)
\item \link{madCoast4}: Outlines of the Fokontanies (Communes) of the region (polygons)
\item \link{madDypsis}: Records of plants of the genus \emph{Dypsis} (points)
\item \link{madRivers}: Major rivers (lines)
}

Rasters (objects of class \code{SpatRaster} from the \strong{terra} package, saved as GeoTIFF files):
\itemize{
\item \link{madChelsa}: Bioclimatic variables
\item \link{madCover}: Land cover
\item \link{madElev}: Elevation
\item \link{madForest2000}: Forest cover in year 2000
\item \link{madForest2014}: Forest cover in year 2014
\item \link{madLANDSAT}: Surface reflectance in 2023
\item \link{madPpt}, \link{madTmin}, \link{madTmax}: Rasters of mean monthly precipitation, and minimum and maximum temperature.
}

Data frames
\itemize{
\item \link{appFunsTable}: Table of functions usable by \code{\link[=app]{app()}}
\item \link{madCoverCats}: Land cover values and categories for \code{madCover}
\item \link{vegIndices}: Vegetation indices that can be calculated with \code{\link[=vegIndex]{vegIndex()}}
}}
}
\value{
A \code{SpatRaster}, \code{sf} spatial vector, or a \code{data.frame}.
}
\description{
This function is a simple way to get example rasters or spatial vector datasets that come with \strong{fasterRaster}.
}
\examples{

### vector data

library(sf)

# For vector data, we can use data(*) or fastData(*):
data(madCoast0) # same as next line
madCoast0 <- fastData("madCoast0") # same as previous
madCoast0
plot(st_geometry(madCoast0))

madCoast4 <- fastData("madCoast4")
madCoast4
plot(st_geometry(madCoast4), add = TRUE)

madRivers <- fastData("madRivers")
madRivers
plot(st_geometry(madRivers), col = "blue", add = TRUE)

madDypsis <- fastData("madDypsis")
madDypsis
plot(st_geometry(madDypsis), col = "red", add = TRUE)

### raster data

library(terra)

# For raster data, we can get the file directly or using fastData(*):
rastFile <- system.file("extdata/madElev.tif", package="fasterRaster")
madElev <- terra::rast(rastFile)

madElev <- fastData("madElev") # same as previous two lines
madElev
plot(madElev)

madForest2000 <- fastData("madForest2000")
madForest2000
plot(madForest2000)

madForest2014 <- fastData("madForest2014")
madForest2014
plot(madForest2014)

# multi-layer rasters
madChelsa <- fastData("madChelsa")
madChelsa
plot(madChelsa)

madPpt <- fastData("madPpt")
madTmin <- fastData("madTmin")
madTmax <- fastData("madTmax")
madPpt
madTmin
madTmax


# RGB raster
madLANDSAT <- fastData("madLANDSAT")
madLANDSAT
plotRGB(madLANDSAT, 4, 1, 2, stretch = "lin")

# categorical raster
madCover <- fastData("madCover")
madCover
madCover <- droplevels(madCover)
levels(madCover) # levels in the raster
nlevels(madCover) # number of categories
catNames(madCover) # names of categories table

plot(madCover)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/match.r
\name{match,GRaster-method}
\alias{match,GRaster-method}
\alias{match}
\alias{\%in\%,GRaster-method}
\alias{\%in\%}
\alias{\%notin\%,GRaster-method}
\alias{\%notin\%}
\title{Find which cells of a GRaster match certain values}
\usage{
\S4method{match}{GRaster}(x, table, nomatch = NA)

\S4method{\%in\%}{GRaster}(x, table)

\S4method{\%notin\%}{GRaster}(x, table)
}
\arguments{
\item{x}{A \code{GRaster}: Note that any kind of \code{GRaster} is acceptable (integer, float, double, or categorical), but matching may not work as intended for float and double rasters because of problems with comparing floating-point values.}

\item{table}{A numeric, integer, or character vector.}

\item{nomatch}{Numeric or integer: Value to return when no match is found.}
}
\value{
A \code{GRaster}.
}
\description{
The \code{match()} function takes a \code{GRaster} and a numeric, integer or character vector as inputs and returns a \code{GRaster} with cell values that correspond to the index of each element in the vector that matched the original cell value. For example, if a 4-cell raster had values 3, \code{NA}, 5, 4, and the vector was \code{c(3, 4)}, then the output would be a 4-cell raster with values 1, \code{NA}, \code{NA}, 2 because the first value in the vector was 3 (so the cell with 3 is assigned 1), and because the second value in the vector was 4 (so the cell with 4 was assigned 2). The other two values had no matches.

If the \code{GRaster} is categorical, then the vector can be category labels instead of numeric values.

The \code{\%in\%} operator returns a \code{GRaster} with cell values that are 1 if their original values appeared in the vector, and 0 if not (or \code{NA} if the original value was \code{NA}). If the \code{GRaster} is categorical, then the vector can be category labels instead of numeric values.

The \verb{\%notin\%} operator returns 1 for cells with values that are \emph{not} found in the vector, and 0 otherwise. If the \code{GRaster} is categorical, then the vector can be category labels instead of numeric values.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data: Elevation and land cover rasters
madElev <- fastData("madElev")
madCover <- fastData("madCover")

### match() with an integer raster:

elev <- fast(madElev)

# Cells in elevation raster replaced with index in which they appear
# in the table:
table <- c(10, 20, 30, 40, 50)
elevIndex <- match(elev, table)
elevIndexNeg <- match(elev, table, nomatch = -100)

plot(c(elevIndex, elevIndexNeg))

### Using \%in\% and \%notin\% on an integer GRaster:

elev <- fast(madElev)
table <- c(10, 20, 30, 40, 50)

ins <- elev \%in\% table
notins <- elev \%notin\% table

plot(c(ins, notins))

### match() with a categorical raster:

cover <- fast(madCover)
cover <- droplevels(cover)
levels(cover)

forestLabels <- c(
   "Sparse broadleaved evergreen/semi-deciduous forest",
   "Broadleaved deciduous forest",
   "Grassland with mosaic forest",
   "Flooded forest"
)

forestClasses <- match(cover, forestLabels)
plot(forestClasses)
levels(forestClasses)

forestNoMatch <- match(cover, forestLabels, nomatch = -100)
plot(forestNoMatch)
levels(forestNoMatch)

### Using \%in\% and \%notin\% on a categorical GRaster:

cover <- fast(madCover)
cover <- droplevels(cover)
levels(cover)

forestLabels <- c(
   "Sparse broadleaved evergreen/semi-deciduous forest",
   "Broadleaved deciduous forest",
   "Grassland with mosaic forest",
   "Flooded forest"
)

forest <- cover \%in\% forestLabels
plot(forest)

notForest <- cover \%notin\% forestLabels
plot(notForest)

}
}
\seealso{
\code{\link[terra:match]{terra::match()}}, \code{\link[=match]{match()}}, \code{\link[omnibus:notIn]{omnibus::notIn()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/disagg.r
\name{disagg,GVector-method}
\alias{disagg,GVector-method}
\alias{disagg}
\title{Coerce as multipart GVector to a singlepart GVector}
\usage{
\S4method{disagg}{GVector}(x)
}
\arguments{
\item{x}{A \code{GVector}.}
}
\value{
A \code{GVector}.
}
\description{
\code{GVectors} can contain a mix of "singlepart" and "multipart" features. A singlepart feature is a single point, set of connected line segments, or a polygon. A multipart feature is a set of lines, sets of connected line segments, or set of polygons that are treated as a single feature. This function converts all multipart features to singlepart features. If the \code{GVector} has an attribute table, rows will be duplicated so that each of the new \code{GVector}'s geometries have the rows that correspond to their "parent" geometries.

If you want to "split" cells of a \code{GRaster} into smaller cells, use \code{\link[=aggregate]{aggregate()}}.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")
madCoast4 <- fastData("madCoast4")

### aggregating a GRaster

# Convert:
elev <- fast(madElev)

### Aggregate GRaster by same factor in 2 dimensions
# fasterRaster
agg2 <- aggregate(elev, 2, "mean")
agg2

# Compare rasters aggregated by fasterRaster and terra.
# These should be the same.
agg2terra <- aggregate(madElev, 2)

agg2 <- rast(agg2)
agg2 <- extend(agg2, agg2terra)
agg2 - agg2terra # value is ~0

### Aggregate GRaster by a non-integer factor in 2 dimensions
# fasterRaster
agg2.9 <- aggregate(elev, 2.9, "mean")
agg2.9

# terra
agg2.9terra <- aggregate(madElev, 2.9, "mean")
agg2.9terra

# Compare rasters aggregated by fasterRaster and terra.
# These should be different.
res(agg2.9)
res(agg2.9terra) # terra rounds aggregation factor down
2 * res(madElev) # original resolution multiplied by 2

### Aggregate GRaster by different factor in 2 dimensions
agg2x3 <- aggregate(elev, c(2, 3), "mean")
agg2x3

### aggregating a GVector

madCoast4 <- fastData("madCoast4")

# Convert:
coast4 <- fast(madCoast4)

# Aggregate and disaggregate:
aggCoast <- aggregate(coast4)
disaggCoast <- disagg(coast4)

ngeom(coast4)
ngeom(aggCoast)
ngeom(disaggCoast)

# plot
oldpar <- par(mfrow = c(1, 3))
plot(coast4, main = "Original", col = 1:nrow(coast4))
plot(aggCoast, main = "Aggregated", col = 1:nrow(aggCoast))
plot(disaggCoast, main = "Disaggregated", col = 1:nrow(disaggCoast))
par(oldpar)

}
}
\seealso{
\code{\link[=aggregate]{aggregate()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.r
\name{plot,GRaster,missing-method}
\alias{plot,GRaster,missing-method}
\alias{plot}
\alias{plot,GVector,missing-method}
\title{Display a raster or vector}
\usage{
\S4method{plot}{GRaster,missing}(x, y, simplify = TRUE, ...)

\S4method{plot}{GVector,missing}(x, y, maxGeoms = 10000, ...)
}
\arguments{
\item{x}{A \code{GRaster} or \code{GVector}.}

\item{y}{Missing--leave as empty.}

\item{simplify}{Logical: If \code{TRUE} (default) and the raster has an x- and/or y-resolution that is greater than the screen resolution, then \code{\link[=aggregate]{aggregate()}} will be applied to the raster before it is saved to reduce the time it takes to save the raster.}

\item{...}{Other arguments to send to \code{\link[terra:plot]{terra::plot()}}.}

\item{maxGeoms}{Positive integer (vectors only): Maximum number of features before vector simplification is applied before saving to disk then creating a \code{SpatVector} for plotting. The default is 10000.}
}
\value{
Nothing (displays a raster or vector).
}
\description{
\code{plot()} displays a \code{GRaster} or \code{GVector}.

This function is essentially a hack, as it it not possible to dependably call the appropriate \strong{GRASS} modules and display a raster or vector without potential confusion on the user side. Instead, this function 1) simplifies the focal \code{GRaster} or \code{GVector} to make it smaller when saved to disk; 2) writes the object to disk; 3) (internally) creates a \code{SpatRaster} or \code{SpatVector} object; then 4) plots the object using \code{\link[terra:plot]{terra::plot()}}. Thus, if you are interested in making maps, it will always be faster to make them directly with \strong{terra} or \strong{sf}.
}
\examples{
if (grassStarted()) {

# Example data
madElev <- fastData("madElev") # elevation raster
madLANDSAT <- fastData("madLANDSAT") # multi-layer raster
madRivers <- fastData("madRivers") # lines vector

# Convert SpatRaster to GRaster and SpatVector to GVector
elev <- fast(madElev)
rivers <- fast(madRivers)
landsat <- fast(madLANDSAT)

# Plot:
plot(elev)
plot(rivers, add = TRUE)

# Histograms:
hist(elev)
hist(landsat)

# Plot surface reflectance in RGB:
plotRGB(landsat, 3, 2, 1) # "natural" color
plotRGB(landsat, 4, 1, 2, stretch = "lin") # emphasize near-infrared (vegetation)

# Make composite map from RGB layers and plot in grayscale:
comp <- compositeRGB(r = landsat[[3]], g = landsat[[2]], b = landsat[[1]])
grays <- paste0("gray", 0:100)
plot(comp, col = grays)

}
}
\seealso{
\code{\link[terra:plot]{terra::plot()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/thinPoints.r
\name{thinPoints,GVector,GRaster-method}
\alias{thinPoints,GVector,GRaster-method}
\alias{thinPoints}
\title{Reduce number of points in same raster cell}
\usage{
\S4method{thinPoints}{GVector,GRaster}(x, y, n = 1)
}
\arguments{
\item{x}{A "points" \code{GVector}.}

\item{y}{A \code{GRaster}.}

\item{n}{Integer or numeric integer: Maximum number of points to remain in a cell. The default is 1.}
}
\value{
A "points" \code{GVector}.
}
\description{
This function thins a "points" \code{GVector} so that it has no more than \code{n} points per grid cell in a raster.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Elevation and points
madElev <- fastData("madElev")
madDypsis <- fastData("madDypsis")

# Convert to fasterRaster formats:
elev <- fast(madElev)
dypsis <- fast(madDypsis)

# Aggregate cells of the raster so they are bigger
elevAgg <- aggregate(elev, 32)

# Remove all but one or two points per cell
thin1 <- thinPoints(dypsis, elevAgg, n = 1)
thin2 <- thinPoints(dypsis, elevAgg, n = 2)

# Plot
plot(elevAgg)
plot(dypsis, add = TRUE)
plot(thin2, col = "yellow", add = TRUE)
plot(thin1, col = "red", add = TRUE)
legend(
   "bottomright",
   legend = c("In original & thin 1 & 2",
     "In just thin 1 & 2", "In just thin 1"),
   pch = 16,
   col = c("black", "yellow", "red"),
   bg = "white",
   xpd = NA
)

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quiet.r
\name{.quiet}
\alias{.quiet}
\title{Returns .quiet() or NULL for "flags" argument to GRASS modules}
\usage{
.quiet()
}
\value{
A string (.quiet()) or \code{NULL}.
}
\description{
A function for developers used for setting the "quiet' argument in \code{flags} arguments passed to \code{\link[rgrass:execGRASS]{rgrass::execGRASS()}}. If \code{faster("debug")} is \code{TRUE}, the string "quiet" is returned. If \code{FALSE}, then \code{NULL} is returned.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge.r
\name{merge,GRaster,GRaster-method}
\alias{merge,GRaster,GRaster-method}
\alias{merge}
\title{Combine two or more rasters with different extents and fill in NAs}
\usage{
\S4method{merge}{GRaster,GRaster}(x, y, ...)
}
\arguments{
\item{x, y, ...}{\code{GRaster}s.}
}
\value{
A \code{GRaster}.
}
\description{
\code{merge()} combines two or more \code{GRaster}s, possibly with different extents, into a single larger \code{GRaster}. Where the same cell has different values in each raster, the value of the first raster's cell is used. If this is \code{NA}, then the value of the second raster's cell is used, and so on.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Example data
madElev <- fastData("madElev")
madCoast4 <- fastData("madCoast4")
madCoast4 <- vect(madCoast4)

# For the example, crop the elevation raster to two communes
madAnt <- madCoast4[madCoast4$NAME_4 == "Antanambe", ]
madMan <- madCoast4[madCoast4$NAME_4 == "Manompana", ]

elevAnt <- crop(madElev, madAnt)
elevMan <- crop(madElev, madMan)

plot(madElev)
plot(elevAnt, col = "red", legend = FALSE, add = TRUE)
plot(elevMan, col = "blue", legend = FALSE, add = TRUE)

# Convert a SpatRaster to a GRaster
ant <- fast(elevAnt)
man <- fast(elevMan)

# merge
antMan <- merge(ant, man)
plot(antMan, main = "Antman!")

}
}
\seealso{
\code{\link[terra:merge]{terra::merge()}}, \code{\link[terra:mosaic]{terra::mosaic()}}, and \strong{GRASS} tool \code{r.patch}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ngeom.r
\name{ngeom,GVector-method}
\alias{ngeom,GVector-method}
\alias{ngeom}
\alias{nsubgeom,GVector-method}
\alias{nsubgeom}
\title{Number of geometries and subgeometries in a vector}
\usage{
\S4method{ngeom}{GVector}(x)

\S4method{nsubgeom}{GVector}(x)
}
\arguments{
\item{x}{A \code{GVector}.}
}
\value{
An integer.
}
\description{
\code{GVector}s represent two types of "geometries". In "singlepart" geometries, each point, set of connected line segments, or polygon is treated like its own feature and has its own row in an attribute table. For example, a province might be composed of islands. In this case, each island would be represented as its own feature and could have its own row in the attribute indicating, say, the name and area of each island.

In "multipart" geometries, features are collected together and thu manipulated as if they were a single feature and have a singe line in an attribute table. Each multipart feature can contain one or more singlepart features. For example, all of the islands comprising  province would be collated together and have a single row in the attribute table indicating the name of the province and the area of the entire province.

\code{ngeom()} returns the number of geometries. Singlepart features are treated as one geometry each, and multipart features are treated as one geometry each.

\code{nsubgeom()} Returns the number of subgeometries. Singlepart geometries each represent a single subgeometry. Multipart geometries represent one or more subgeometries. The number of subgeometries will thus always be the same as or more than the number of geometries.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)

# Example data:
madCoast4 <- fastData("madCoast4")
madRivers <- fastData("madRivers")
madDypsis <- fastData("madDypsis")

# Convert sf vectors to GVectors:
coast <- fast(madCoast4)
rivers <- fast(madRivers)
dypsis <- fast(madDypsis)

# Geographic properties:
ext(rivers) # extent
crs(rivers) # coordinate reference system
st_crs(rivers) # coordinate reference system
coordRef(rivers) # coordinate reference system

# Column names and data types:
names(coast)
datatype(coast)

# Points, lines, or polygons?
geomtype(dypsis)
geomtype(rivers)
geomtype(coast)

is.points(dypsis)
is.points(coast)

is.lines(rivers)
is.lines(dypsis)

is.polygons(coast)
is.polygons(dypsis)

# Number of dimensions:
topology(rivers)
is.2d(rivers) # 2-dimensional?
is.3d(rivers) # 3-dimensional?

# Just the data table:
as.data.frame(rivers)
as.data.table(rivers)

# Top/bottom of the data table:
head(rivers)
tail(rivers)

# Vector or table with just selected columns:
names(rivers)
rivers$NAME
rivers[[c("NAM", "NAME_0")]]
rivers[[c(3, 5)]]

# Select geometries/rows of the vector:
nrow(rivers)
selected <- rivers[2:6]
nrow(selected)

# Plot:
plot(coast)
plot(rivers, col = "blue", add = TRUE)
plot(selected, col = "red", lwd = 2, add = TRUE)

# Vector math:
hull <- convHull(dypsis)

un <- union(coast, hull)
sameAsUnion <- coast + hull
plot(un)
plot(sameAsUnion)

inter <- intersect(coast, hull)
sameAsIntersect <- coast * hull
plot(inter)
plot(sameAsIntersect)

er <- erase(coast, hull)
sameAsErase <- coast - hull
plot(er)
plot(sameAsErase)

xr <- xor(coast, hull)
sameAsXor <- coast / hull
plot(xr)
plot(sameAsXor)

# Vector area and length:
expanse(coast, unit = "km") # polygons areas
expanse(rivers, unit = "km") # river lengths

### Fill holes

# First, we will make some holes by creating buffers around points.
buffs <- buffer(dypsis, 500)

holes <- coast - buffs
plot(holes)

filled <- fillHoles(holes, fail = FALSE)

}
}
\seealso{
\code{\link[=nrow]{nrow()}}, \code{\link[=dim]{dim()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cleanGeom.r
\name{breakPolys,GVector-method}
\alias{breakPolys,GVector-method}
\alias{breakPolys}
\alias{fixBridges,GVector-method}
\alias{fixBridges}
\alias{fixDangles,GVector-method}
\alias{fixDangles}
\alias{fixLines,GVector-method}
\alias{fixLines}
\alias{remove0,GVector-method}
\alias{remove0}
\alias{removeAngles,GVector-method}
\alias{removeAngles}
\alias{removeBridges,GVector-method}
\alias{removeBridges}
\alias{removeDangles,GVector-method}
\alias{removeDangles}
\alias{removeDupCentroids,GVector-method}
\alias{removeDupCentroids}
\alias{removeDups,GVector-method}
\alias{removeDups}
\alias{removeSmallPolys,GVector-method}
\alias{removeSmallPolys}
\alias{snap,GVector-method}
\alias{snap}
\title{Fix issues with geometries of a vector}
\usage{
\S4method{breakPolys}{GVector}(x)

\S4method{fixBridges}{GVector}(x)

\S4method{fixDangles}{GVector}(x, tolerance = -1)

\S4method{fixLines}{GVector}(x)

\S4method{remove0}{GVector}(x)

\S4method{removeAngles}{GVector}(x)

\S4method{removeBridges}{GVector}(x)

\S4method{removeDangles}{GVector}(x, tolerance = -1)

\S4method{removeDupCentroids}{GVector}(x)

\S4method{removeDups}{GVector}(x)

\S4method{removeSmallPolys}{GVector}(x, tolerance)

\S4method{snap}{GVector}(x, tolerance)
}
\arguments{
\item{x}{A \code{GVector}.}

\item{tolerance}{Numeric or \code{NULL} (default): Minimum distance in map units (degrees for unprojected, usually meters for projected) or minimum area (in meters-squared, regardless of projection).}
}
\value{
A \code{GVector}.
}
\description{
These functions are intended to help fix geometric issues with a \code{GVector}. Note that the functionality of the \code{snap()} and \code{removeAreas()} functions can also be implemented when using \code{\link[=fast]{fast()}} to create a \code{GVector}.
\itemize{
\item \code{breakPolys()}: Break topologically clean areas. This is similar to \code{fixLines()}, except that it does not break loops. Topologically clean vectors may occur if the vector was imported from a format that does not enforce topology, such as a shapefile. Duplicate geometries are automatically removed after breaking.
\item \code{fixBridges()}: Change "bridges" to "islands" (which are topologically incorrect) within geometries to lines.
\item \code{fixDangles()}: Change "dangles" hanging off boundaries to lines if shorter than \code{tolerance} distance. If \code{tolerance} is <0, all dangles will be changed to lines.  Units of \code{tolerance} are in map units, or in degrees for unprojected CRSs. If \code{tolerance} <0, all dangles are removed, and the function will retain only closed loops and lines connecting loops. Dangles will be removed from longest to shortest.
\item \code{fixLines()}: Break lines at intersections and lines that form closed loops.
\item \code{remove0()}: Remove all boundaries and lines with a length of 0.
\item \code{removeAngles()}: Collapse lines that diverge at an angle that is computationally equivalent to 0. This tool often needs to be followed with the \code{break()} and \code{removeDups()} methods.
\item \code{removeBridges()}: Remove "bridges" to "islands" (which are topologically incorrect) within geometries.
\item \code{removeDangles()}: Remove "dangling" lines if shorter than \code{tolerance} distance. If \code{tolerance} is <0, all dangles will be removed. Units of \code{tolerance} are in map units, or in degrees for unprojected CRSs. If \code{tolerance} <0, all dangles are removed, and the function will retain only closed loops and lines connecting loops. Dangles will be removed from longest to shortest.
\item \code{removeDupCentroids()}: Remove duplicated area centroids. In \strong{GRASS}, closed polygons have their attributes mapped to a (hidden) centroid of the polygon.
\item \code{removeDups()}: Remove duplicated features and area centroids.
\item \code{removeSmallPolys()}: Remove polygons smaller than \code{tolerance}. Units of \code{tolerance} are in square meters (regardless of the CRS).
\item \code{snap()}: Snap lines/boundaries to each other if they are less than \code{tolerance} apart. Subsequent removal of dangles may be needed. Units of \code{tolerance} are map units, or degrees for unprojected CRSs.
}
}
\examples{
if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Example data
madRivers <- fastData("madRivers")
rivers <- fast(madRivers)
soam <- rivers[rivers$NAM == "SOAMIANINA"] # select one river for illustration

### Simplify geometry (remove nodes)

vr <- simplifyGeom(soam, tolerance = 2000)
dp <- simplifyGeom(soam, tolerance = 2000, method = "dp")
dpr <- simplifyGeom(soam, tolerance = 2000, method = "dpr", prop = 0.5)
rw <- simplifyGeom(soam, tolerance = 2000, method = "rw")

plot(soam, col = "black", lwd = 3)
plot(vr, col = "blue", add = TRUE)
plot(dp, col = "red", add = TRUE)
plot(dpr, col = "chartreuse", add = TRUE)
plot(rw, col = "orange", add = TRUE)

legend("bottom",
   xpd = NA,
   legend = c(
	  "Original",
      "Vertex reduction",
      "Douglas-Peucker",
      "Douglas-Peucker reduction",
      "Reumann-Witkam"
	),
	col = c("black", "blue", "red", "chartreuse", "orange"),
	lwd = c(3, 1, 1, 1, 1)
)

### Smooth geometry

hermite <- smoothGeom(soam, dist = 2000, angle = 3)
chaiken <- smoothGeom(soam, method = "Chaiken", dist = 2000)

plot(soam, col = "black", lwd = 2)
plot(hermite, col = "blue", add = TRUE)
plot(chaiken, col = "red", add = TRUE)

legend("bottom",
   xpd = NA,
   legend = c(
	  "Original",
      "Hermite",
      "Chaiken"
	),
	col = c("black", "blue", "red"),
	lwd = c(2, 1, 1, 1, 1)
)

### Clean geometry

# Has no effect on this vector!
noDangs <- removeDangles(soam, tolerance = 10000)

plot(soam, col = "black", lwd = 2)
plot(noDangs, col = "red", add = TRUE)

legend("bottom",
   xpd = NA,
   legend = c(
	  "Original",
      "No dangles"
	),
	lwd = c(2, 1),
	col = c("black", "red")
)

}
}
\seealso{
\code{\link[terra:topology]{terra::topology()}}, \code{\link[=fillHoles]{fillHoles()}}, \code{\link[terra:topology]{terra::removeDupNodes()}}, \emph{Details} section in \code{\link[=fast]{fast()}}, \code{\link[=simplifyGeom]{simplifyGeom()}}, \code{\link[=smoothGeom]{smoothGeom()}}, \strong{GRASS} manual page for tool \code{v.clean} (see \code{grassHelp("v.clean")})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vNames.r
\name{.vNames}
\alias{.vNames}
\title{Names of columns of a GRASS vector's attribute table}
\usage{
.vNames(x)
}
\arguments{
\item{x}{A \code{GVector} or the name of a vector in \strong{GRASS}.}
}
\value{
Character vector.
}
\description{
This function returns the column names of a \strong{GRASS} vector's attribute table.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/activeCat.r
\name{activeCat,GRaster-method}
\alias{activeCat,GRaster-method}
\alias{activeCat}
\alias{activeCats,GRaster-method}
\alias{activeCats}
\alias{activeCat<-,GRaster-method}
\alias{activeCat<-}
\title{Get or set the column with category labels in a categorical raster}
\usage{
\S4method{activeCat}{GRaster}(x, layer = 1, names = FALSE)

\S4method{activeCats}{GRaster}(x, names = FALSE)

\S4method{activeCat}{GRaster}(x, layer = 1) <- value

\S4method{activeCat}{GRaster}(x, layer = 1) <- value

\S4method{activeCat}{GRaster}(x, layer = 1) <- value
}
\arguments{
\item{x}{A categorical \code{GRaster}.}

\item{layer}{Numeric, integer, logical, or character: Indicates for which layer(s) to get or set the active category column. This can be a number (the index of the raster(s)), a logical vector (\code{TRUE} ==> get/set the active category column, \code{FALSE} ==> leave as-is), or a character vector (names of layers).}

\item{names}{Logical: If \code{TRUE}, display the name(s) of the active column(s). If \code{FALSE} (default), report the index of the active column. Following \code{\link[terra:activeCat]{terra::activeCat()}}, the first column in the levels table is ignored. So, an active column of "1" means the second column is active. "2" means the third column is active, and so on.}

\item{value}{Numeric, integer, or character. Following \code{\link[terra:activeCat]{terra::activeCat()}}, the first column in each levels table is ignored. So, if you want the second column to be the category label, use 1. If you want the third column, use 2, and so on. You can also specify the active column by its column name (though this can't be the first column's name).}
}
\value{
\code{activeCat()} returns an integer or character of the active column index or name. \code{activeCats()} returns a vector of indices or names. \verb{activeCat()<-} returns a \code{GRaster}.
}
\description{
These functions return or set the column of the labels to be matched to each value in the raster of a categorical \code{GRaster} (see \code{vignette("GRasters", package = "fasterRaster")}). \emph{Important}: Following \code{\link[terra:activeCat]{terra::activeCat()}}, the first column in the "levels" table is ignored, so an "active category" value of 1 means the second column is used as labels, a value of 2 means the third is used, and so on.
\itemize{
\item \code{activeCat()} returns the column of the labels to be matched to each value in the raster for a single raster layer.
\item \code{activeCats()} does the same, but for all layers of a \code{GRaster}.
\item \verb{activeCat()<-} sets the column to be used as category labels.
}
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data: Land cover raster
madCover <- fastData("madCover")

# Convert categorical SpatRaster to categorical GRaster:
cover <- fast(madCover)

### Properties of categorical rasters

cover # note categories
is.factor(cover) # Is the raster categorical?
nlevels(cover) # number of levels
levels(cover) # just the value and active column
cats(cover) # all columns
minmax(cover) # min/max values
minmax(cover, levels = TRUE) # min/max categories
catNames(cover) # column names of the levels table
missingCats(cover) # categories in table with no values in raster
freq(cover) # frequency of each category (number of cells)
zonalGeog(cover) # geometric statistics

### Active column

# Which column sets the category labels?
activeCat(cover)
activeCat(cover, names = TRUE)

activeCats(c(cover, cover))

# Choose a different column for category labels:
levels(cover)
activeCat(cover) <- 2
levels(cover)

### Managing levels tables

# Remove unused levels:
nlevels(cover)
cover <- droplevels(cover)
nlevels(cover)

# Re-assign levels:
value <- c(20, 30, 40, 50, 120, 130, 140, 170)
label <- c("Cropland", "Cropland", "Forest", "Forest",
 "Grassland", "Shrubland", "Herbaceous", "Flooded")

newCats <- data.frame(value = value, label = label)

cover <- categories(cover, layer = 1, value = newCats)
cats(cover)

# This is the same as:
levels(cover) <- newCats
cats(cover)

# Are there any values not assigned a category?
missingCats(cover)

# Let's assign a category for value 210 (water):
water <- data.frame(value = 210, label = "Water")
addCats(cover) <- water
levels(cover)

# Add more information to the levels table using merge():
landType <- data.frame(
     Value = c(20, 30, 40, 50, 120),
     Type = c("Irrigated", "Rainfed", "Broadleaf evergreen",
     "Broadleaf deciduous", "Mosaic with forest")
)
cats(cover)
cover <- addCats(cover, landType, merge = TRUE)
cats(cover)

### Logical operations on categorical rasters

cover < "Forest" # 1 for cells with a value < 40, 0 otherwise
cover <= "Forest" # 1 for cells with a value < 120, 0 otherwise
cover == "Forest" # 1 for cells with value of 40-120, 0 otherwise
cover != "Forest" # 1 for cells with value that is not 40-120, 0 otherwise
cover > "Forest" # 1 for cells with a value > 120, 0 otherwise
cover >= "Forest" # 1 for cells with a value >= 120, 0 otherwise

cover \%in\% c("Cropland", "Forest") # 1 for cropland/forest cells, 0 otherwise

### Combine categories from different rasters

# For the example, will create a second categorical raster fromm elevation.

# Divide elevation raster into "low/medium/high" levels:
madElev <- fastData("madElev")
elev <- fast(madElev)
elev <- project(elev, cover, method = "near") # convert to same CRS
fun <- "= if(madElev < 100, 0, if(madElev < 400, 1, 2))"
elevCat <- app(elev, fun)

levs <- data.frame(
     value = c(0, 1, 2),
     elevation = c("low", "medium", "high")
)
levels(elevCat) <- list(levs)

# Combine levels:
combined <- concats(cover, elevCat)
combined
levels(combined)

# Combine levels, treating value/NA combinations as new categories:
combinedNA <- concats(cover, elevCat, na.rm = FALSE)
combinedNA
levels(combinedNA)

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reorient.r
\name{reorient,GRaster-method}
\alias{reorient,GRaster-method}
\alias{reorient}
\alias{reorient,numeric-method}
\title{Convert degrees between 'north-orientation' and 'east orientation'}
\usage{
\S4method{reorient}{GRaster}(x, units = "degrees")

\S4method{reorient}{numeric}(x, units = "degrees")
}
\arguments{
\item{x}{A numeric vector or a \code{GRaster} with cell values equal to facing (in degrees).}

\item{units}{Character: "Units" of values in \code{x}: either \code{"degrees"} for degrees (default) or \code{"radians"}. Partial matching is used.}
}
\value{
A \code{GRaster} or numeric vector. Values will be in the range between 0 and 360 and represents facing in the system "opposing" the input's system. For example, if the input is north orientation, the output will be east orientation. If the input is in east orientation, the output will be in north orientation.
}
\description{
This function converts facing between "north orientation" and "east orientation".

In "north orientation" systems, a 0-degree facing is north, and the angle of facing proceeds clockwise. For example, a 90 degree facing faces east, 180 south, and 270 west. In "east orientation", a 0-degree facing is east, and the facing angle proceeds counter-clockwise. For example, 90 is north, 180 is west, and 270 south.
}
\examples{

### Re-orient numeric values:
facings <- c(0, 90, 180, 270, 360)
reorient(facings)

# Re-reorienting returns the same values:
reorient(reorient(facings))

if (grassStarted()) {

### Re-orient a GRaster:

# Setup
library(terra)
madElev <- fastData("madElev")
elev <- fast(madElev)

# Calculate aspect in degrees, using north orientation:
aspectNorth <- terrain(elev, "aspect")

# Re-orient to east-facing:
aspectEast <- reorient(aspectNorth)

# Re-reorienting is the same, to within rounding error:
aspectNorth - reorient(reorient(aspectNorth))

# Plot:
aspects <- c(aspectNorth, aspectEast)
names(aspects) <- c("north_orientation", "east_orientation")
plot(aspects)


}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/minmax.r
\name{.minVal}
\alias{.minVal}
\title{Get minimum value from GRaster metadata}
\usage{
.minVal(x)
}
\arguments{
\item{x}{A \code{GRaster}.}
}
\description{
Get minimum value from GRaster metadata
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/crds.r
\name{crds,GRaster-method}
\alias{crds,GRaster-method}
\alias{crds}
\alias{crds,GVector-method}
\alias{st_coordinates}
\title{Coordinates of a vector"s features or a raster"s cell centers}
\usage{
\S4method{crds}{GRaster}(x, z = is.3d(x), na.rm = TRUE)

\S4method{crds}{GVector}(x, z = is.3d(x))

st_coordinates(x)
}
\arguments{
\item{x}{A \code{GVector} or a \code{GRaster}.}

\item{z}{If \code{TRUE} (default), return x-, y-, and z-coordinates. If \code{FALSE}, just return x- and y-coordinates. For 2-dimensional objects, z-coordinates will all be 0.}

\item{na.rm}{Logical: If \code{TRUE}, remove cells that are \code{NA} (\code{GRaster}s only).}
}
\value{
A \code{matrix}, \code{data.frame}, or \code{list}.
}
\description{
Returns the coordinates of the center of cells of a \code{GRaster} or coordinates of a \code{GVector}'s vertices. The output will be a \code{matrix}, \code{data.frame}, or \code{list}. If you want the output to be a "points" \code{GVector}, use \code{\link[=as.points]{as.points()}}.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Plant specimens (points), elevation (raster)
madDypsis <- fastData("madDypsis")
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster, and sf to a GVector
dypsis <- fast(madDypsis)
elev <- fast(madElev)

### Get coordinates:
dypsisPoints <- crds(dypsis)
elevPoints <- crds(elev)

head(dypsisPoints)
head(elevPoints)

}
}
\seealso{
\code{\link[terra:crds]{terra::crds()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/voronoi.r
\name{voronoi,GVector-method}
\alias{voronoi,GVector-method}
\alias{voronoi}
\title{Voronoi tessellation}
\usage{
\S4method{voronoi}{GVector}(x, buffer = 0)
}
\arguments{
\item{x}{A \code{GVector} "points" object.}

\item{buffer}{Numeric: By default, this function creates a vector that has an extent exactly the same as the input data. However, the apparent extent can be changed by setting this value to a value different from 0. Negative values reduce the size of the extent, and positive extend it.  Units are in map units.}
}
\value{
A \code{GVector}.
}
\description{
This function creates a Voronoi tessellation from a set of spatial points or polygons.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)

# Example vectors
madDypsis <- fastData("madDypsis") # points
madCoast4 <- fastData("madCoast4") # polygons

# Convert sf vectors to GVectors
dypsis <- fast(madDypsis)
coast4 <- fast(madCoast4)
ant <- coast4[coast4$NAME_4 == "Antanambe"]

# Delaunay triangulation
dypsisDel <- delaunay(dypsis)
plot(dypsisDel)
plot(dypsis, pch = 1, col = "red", add = TRUE)

# Voronoi tessellation
vor <- voronoi(dypsis)
plot(vor)
plot(dypsis, pch = 1, col = "red", add = TRUE)

# Random Voronoi tessellation
rand <- rvoronoi(coast4, size = 100)
plot(rand)

}
}
\seealso{
\code{\link[terra:voronoi]{terra::voronoi()}}, \code{\link[sf:geos_unary]{sf::st_voronoi()}}, tool \code{v.voronoi} in \strong{GRASS}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/resample.r
\name{resample,GRaster,GRaster-method}
\alias{resample,GRaster,GRaster-method}
\alias{resample}
\alias{resample,GRaster,numeric-method}
\title{Change the cell size of a GRaster}
\usage{
\S4method{resample}{GRaster,GRaster}(x, y, method = NULL, fallback = TRUE)

\S4method{resample}{GRaster,numeric}(x, y, method = NULL, fallback = TRUE)
}
\arguments{
\item{x}{The \code{GRaster} to resample.}

\item{y}{Either a \code{GRaster} to serve as a template, or a numeric vector with two or three values. If a numeric vector, the values represent east-west and north-south resolution for 2D rasters, or east-west, north-south, and top-bottom resolution for 3D rasters.}

\item{method}{Character or \code{NULL}: Method to use to assign values to cells. Partial matching is used.
\itemize{
\item \code{NULL} (default): Automatically choose based on raster properties (\code{near} for categorical or integer rasters, \code{bilinear} for continuous data).
\item \code{"near"}: Nearest neighbor. Best for categorical data, and often a poor choice for continuous data.  If \code{\link[=nlevels]{nlevels()}} is >0, this method will be used regardless of the value of \code{method}. If you still want to use a different method, coerce the raster to a different type using \code{\link[=as.int]{as.int()}}, \code{\link[=as.float]{as.float()}}, or \code{\link[=as.doub]{as.doub()}}.
\item \code{"bilinear"}: Bilinear interpolation (default for non-categorical data; uses weighted values from 4 cells).
\item \code{"bicubic"}: Bicubic interpolation (uses weighted values from 16 cells).
\item \code{"lanczos"}: Lanczos interpolation (uses weighted values from 25 cells).
Note that methods that use multiple cells will cause the focal cell to become \code{NA} if there is at least one cell with an \code{NA} in the cells it draws from. These \code{NA} cells can be filled using the \code{fallback} option.
}}

\item{fallback}{Logical: If \code{TRUE} (default), then use "lower" methods to fill in \code{NA} cells when a "higher" method is used. For example, if \code{method = "bicubic"}, \code{NA} cells will be filled in using the \code{bilinear} method, except when that results in \code{NA}s, in which case the \code{near} method will be used. Fallback causes fewer cells to revert to \code{NA} values, so can be better at resampling the edges of rasters. However, fallback does increase processing time because each "lower" method must be applied, then results merged.}
}
\value{
A \code{GRaster}.
}
\description{
\code{resample()} changes the cell size (resolution) of a \code{GRaster} using either another raster as a template or a user-defined resolution. Note that the extent of the output raster may be expanded to accommodate an integer number of cells. The function is not guaranteed to recreate the same output as \code{\link[terra:resample]{terra::resample()}}, even when the same resampling method is used.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Elevation raster
madElev <- fastData("madElev")
elev <- fast(madElev)

### Resample raster to 120 x 120 m
elev120 <- resample(elev, c(120, 120), method="bilinear")
elev
elev120

### Resample using another raster as a template

template <- aggregate(elev, 4)

nearest <- resample(elev, template, method = "nearest")

bilinear <- resample(elev, template, method = "bilinear")
bilinearNoFB <- resample(elev, template, method = "bilinear", fallback = FALSE)

bicubic <- resample(elev, template, method = "bicubic")
bicubicNoFB <- resample(elev, template, method = "bicubic", fallback = FALSE)

lanczos <- resample(elev, template, method = "lanczos")
lanczosNoFB <- resample(elev, template, method = "lanczos", fallback = FALSE)

# rasters resampled without fallback have fewer non-NA cells
resampled <- c(nearest, bilinear, bilinearNoFB, bicubic, bicubicNoFB, lanczos,
    lanczosNoFB)
names(resampled) <- c("nearest", "bilinear", "bilinearNoFB", "bicubic",
    "bicubicNoFB", "lanczos", "lanczosNoFB")
ones <- resampled * 0 + 1
global(ones, "sum") # number of non-NA cells
global(resampled, c("mean", "sd", "min", "max")) # other statistics

# Compare fallback to no fallback
frLanczos <- rast(lanczos)
frLanczosNoFB <- rast(lanczosNoFB)

plot(frLanczos, col = "red",
    main = "Red: Cells in fallback not non-fallback", legend = FALSE)
plot(frLanczosNoFB, add=TRUE)

# Compare fasterRaster with terra
coarserTerra <- aggregate(madElev, 4)
terraLanczos <- resample(madElev, coarserTerra, method = "lanczos")

frLanczos <- extend(frLanczos, terraLanczos)
frLanczosNoFB <- extend(frLanczosNoFB, terraLanczos)

frLanczos - terraLanczos
frLanczosNoFB - terraLanczos

plot(frLanczos - terraLanczos, main = "Difference")
plot(frLanczosNoFB - terraLanczos, main = "Difference")

plot(terraLanczos, col = "red",
    main = "Red: Cells in terra not in FR", legend = FALSE)
plot(frLanczos, add=TRUE)

plot(frLanczos, col = "red",
    main = "Red: Cells in FR not in terra", legend = FALSE)
plot(terraLanczos, add=TRUE)

}
}
\seealso{
\code{\link[terra:resample]{terra::resample()}}, \strong{GRASS} modules \code{r.resample} and \code{r.resamp.interp} (see \verb{grassHelp("}r.resample\verb{") and }grassHelp("\code{r.resamp.interp}")`)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kernel.r
\name{kernel,GVector-method}
\alias{kernel,GVector-method}
\alias{kernel}
\title{Kernel density estimator of points}
\usage{
\S4method{kernel}{GVector}(x, y, kernel = "Epanechnikov", optimize = TRUE, h = NULL)
}
\arguments{
\item{x}{A "points" \code{GVector}.}

\item{y}{A \code{GRaster}: The extent and resolution of this raster will be used to create the density raster. Otherwise, values in this raster are ignored.}

\item{kernel}{Character: Name of the kernel function to use. Possible values include:
\itemize{
\item \code{"Epanechnikov"} (default)
\item \code{"Gaussian"}
\item \code{"uniform"}
\item \code{"triangular"}
\item \code{"quartic"}
\item \code{"triweight"}
\item \code{"cosine"}
}

Partial matching is used, and case is ignored.}

\item{optimize}{Logical: If \code{TRUE} (default), then attempt to find the optimal radius less than or equal to the \code{radius} value using the "Gaussian" kernel. If \code{FALSE}, use the \code{radius} value as-is.}

\item{h}{Numeric or \code{NULL} (default): Smoothing bandwidth of kernel estimator.

If this is \code{NULL}, the Epanechnikov kernel is used, and \code{optimize} is \code{TRUE}, then Silverman's rule-of-thumb is used to estimate the optimal value of \code{h}:
\deqn{h = 0.9 * min(\sigma_x / n^1/6, \sigma_y / n^1/6)}

If the Gaussian kernel is used, and \code{optimize} is \code{TRUE}, then the \strong{GRASS} \code{v.kernel} function will attempt to identify the optimal bandwidth, up to the value of \code{h}, if \code{h} is defined.

Otherwise, if \code{h} is \code{NULL}, then the value will be arbitrarily set at 1/5th of the shorter of the distance of the x- and y-extent of the points.}
}
\value{
A \code{GRaster}.
}
\description{
\code{kernel()} creates a raster using a kernel density estimator of the density of points in a "points" \code{GVector}.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Elevation raster, plant specimen collections, rivers vector,
# outline of area vector
madElev <- fastData("madElev")
madDypsis <- fastData("madDypsis")

# Convert to fasterRaster format:
elev <- fast(madElev)
dypsis <- fast(madDypsis)

# Kernel density estimation:
kde <- kernel(dypsis, elev)
plot(kde)
plot(dypsis, add = TRUE, pch = 1)

}
}
\seealso{
\strong{GRASS} manual page for tool \code{v.kernel} (see \code{grassHelp("v.kernel")})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/princomp.r
\name{pcs}
\alias{pcs}
\title{Retrieve a principal components model from a PCA GRaster}
\usage{
pcs(x)
}
\arguments{
\item{x}{A \code{GRaster} created by \code{\link[=princomp]{princomp()}}}
}
\value{
An object of class \code{prcomp}.
}
\description{
Retrieve a principal components model from a PCA GRaster
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Climate raster:
madChelsa <- fastData("madChelsa")

# Convert a SpatRaster to a GRaster:
chelsa <- fast(madChelsa)

# Generate raster with layers representing principal component predictions:
pcRast <- princomp(chelsa, scale = TRUE)
plot(pcRast)

# Get information on the PCA:
prinComp <- pcs(pcRast)

prinComp
summary(prinComp)
plot(prinComp)

}
}
\seealso{
\code{\link[=princomp]{princomp()}}, \code{\link[terra:princomp]{terra::princomp()}}, tool \code{i.pca} in \strong{GRASS}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/streams.r
\name{streams,GRaster-method}
\alias{streams,GRaster-method}
\alias{streams}
\title{Create stream network}
\usage{
\S4method{streams}{GRaster}(
  x,
  accumulation = NULL,
  depression = NULL,
  flowThreshold = 1,
  dirThreshold = 1,
  montgomery = 0,
  minLength = 1
)
}
\arguments{
\item{x}{A \code{GRaster} representing elevation.}

\item{accumulation}{Either \code{NULL} (default) or a raster representing flow accumulation. If not supplied, an accumulation will created internally. You can generate an accumulation raster using \code{\link[=flow]{flow()}}.}

\item{depression}{Either \code{NULL} (default) or a \code{GRaster} representing depressions (areas from which streams will not flow out of).}

\item{flowThreshold}{Numeric > 0: Minimum threshold for a stream to be generated. The default is 1, which is not necessarily a reasonable value.}

\item{dirThreshold}{Numeric (default is \code{Inf}): When flow exceeds this threshold, its direction is estimated using a single-flow direction algorithm. Below this threshold, a multi-direction flow model is used. This is the \code{d8cut} parameter in \code{r.stream.extract}, and it is only used if \code{accumulation} is \code{NULL}.  The default is 1, which is not necessarily a reasonable value.}

\item{montgomery}{Numeric: The "Montgomery" exponent for slope, multiplied by accumulation as per \code{accumulation * slope^montgomery}. This value is then compared to the threshold to determine if it is sufficient. The default is 0 (i.e., no slope scaling).}

\item{minLength}{Numeric: First-order streams less than this length are removed (units in cells). Default is 0 (no removal).}
}
\value{
A \code{GRaster}.
}
\description{
This function estimates the course of streams and rivers from an elevation raster. It is based on the \strong{GRASS} tool \code{r.stream.extract}, where more details can be found (see \code{grassHelp("r.stream.extract")})
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster
elev <- fast(madElev)

# Calculate stream channels
streams <- streams(elev)
plot(streams)

}
}
\seealso{
\code{\link[=flow]{flow()}}, \code{\link[=flowPath]{flowPath()}}, \strong{GRASS} manual for tool \code{r.stream.extract} (see \code{grassHelp("r.stream.extract")})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/location.r, R/locationDelete.r
\name{.location,GLocation-method}
\alias{.location,GLocation-method}
\alias{.location}
\alias{.location,missing-method}
\alias{.locationDelete}
\title{GRASS "location" of an object or the active session}
\usage{
\S4method{.location}{GLocation}(x)

\S4method{.location}{missing}(x)

.locationDelete(location, mapset = NULL, workDir = NULL)
}
\arguments{
\item{x}{Either:
\itemize{
\item Missing: Reports location of currently active project/location.
\item A \code{GLocation} object or an object that contains the \code{GLocation} class (i.e., a \code{GSpatial} object: a \code{GRaster} or \code{GVector}).
}}

\item{location}{Character: Name of the \strong{GRASS} "location".}

\item{mapset}{Character or \code{NULL} (default): Name of the mapset to delete. If \code{NULL}, then all mapsets in the given "location" will be deleted.}

\item{workDir}{Character: Either \code{NULL} (default) or a character string of the directory in which the location to be removed resides. If \code{NULL}, then the working directory will be obtained from \code{faster("workDir")}.}
}
\value{
A character string.

If successful, \code{TRUE} (invisibly). If not, then \code{FALSE} (also invisibly), plus a warning.
}
\description{
\strong{GRASS} "projects" or "locations" are sets of one or more rasters and/or vectors with the same coordinate reference systems, and may or may not represent the same actual location on Earth. \strong{GRASS} "mapsets" are like sub-folders of locations, and are collections of rasters and/or vectors typically related to the same general project. Typical users will not need to make changes to the default location (called "location") or mapset (called "PERMANENT"). See \code{vignette("projects_mapsets", package = "fasterRaster")}.

This function deletes a \strong{GRASS} "project"/"location", rasters, and vectors therein. This function should be used cautiously and is mainly of use to developers. See \code{vignette("projects_mapsets", package = "fasterRaster")}.
}
\seealso{
\code{\link[=.mapset]{.mapset()}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/message.r
\name{.message}
\alias{.message}
\title{Display warning or message}
\usage{
.message(msg, message)
}
\arguments{
\item{msg}{Character: Name for the message (used internally). Should be able to be assigned to a list (i.e., no spaces, punctuation, etc.).}

\item{message}{Text for the message.}
}
\value{
\code{TRUE} (invisibly).
}
\description{
Display a warning or message if the given warning has not been displayed since \strong{fasterRaster} was attached or if a given number or hours has passed since then.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/addTable_dropTable.r
\name{addTable<-,GVector,data.frame-method}
\alias{addTable<-,GVector,data.frame-method}
\alias{addTable<-}
\alias{addTable<-,GVector,data.table-method}
\alias{addTable<-,GVector,matrix-method}
\alias{dropTable,GVector-method}
\alias{dropTable}
\title{Attach or detach GVector's data table}
\usage{
\S4method{addTable}{GVector,data.frame}(x, ...) <- value

\S4method{addTable}{GVector,data.table}(x, ...) <- value

\S4method{addTable}{GVector,matrix}(x, ...) <- value

\S4method{dropTable}{GVector}(x)
}
\arguments{
\item{x}{A \code{GVector}.}

\item{...}{Other arguments (ignored).}

\item{value}{A \code{data.frame}, \code{data.table}, or \code{matrix}.}
}
\value{
A \code{GVector}.
}
\description{
\code{addTable()} adds an entire table to a \code{GVector}. It will replace any existing table. There must be one row in the table for each geometry (see \code{\link[=ngeom]{ngeom()}}). You can also add a table column-by-column using the \code{\link[fasterRaster]{$<-}} operator.

\code{dropTable()} removes a data table associated with a \code{GVector}.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)

# Rivers vector
madRivers <- fastData("madRivers")

# Convert sf to a GVector
rivers <- fast(madRivers)

# Convert GVector to data.frame or data.table
as.data.frame(rivers)
as.data.table(rivers)

# Subset rivers vector
rivers1 <- rivers[1:2]
rivers2 <- rivers[10:11]

# Concatenate rivers
riversCombo <- rbind(rivers1, rivers2)
riversCombo

# Add columns
newCol <- data.frame(new = 1:11)
riversCol <- colbind(rivers, newCol)
riversCol

# Remove table
riversCopy <- rivers
riversCopy # has data table
riversCopy <- dropTable(riversCopy)
riversCopy # no data table

# Add a new table
newTable <- data.frame(num = 1:11, letters = letters[1:11])
addTable(riversCopy) <- newTable
riversCopy

}
}
\seealso{
\code{\link[fasterRaster]{$<-}}, \code{\link[=colbind]{colbind()}}, \code{\link[=rbind]{rbind()}}, \code{\link[=as.data.frame]{as.data.frame()}}, \code{\link[=as.data.table]{as.data.table()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.r
\name{predict,GRaster-method}
\alias{predict,GRaster-method}
\alias{predict}
\title{Make predictions from a linear or generalized linear model to a GRaster}
\usage{
\S4method{predict}{GRaster}(object, model, type = "response")
}
\arguments{
\item{object}{A \code{GRaster} with one or more layers.}

\item{model}{An \code{lm} or \code{glm} model object.}

\item{type}{Character: Type of prediction to make. This can be either \code{link} (default; predictions are made on the scale of the link function) or \code{response} (predictions are made on the scale of the response variable). This function can only make predictions on the scale of the response for the identity, logit, log, or cloglog (complementary log-log) link functions.}
}
\value{
A \code{GRaster}.
}
\description{
This version of the \code{predict()} function make predictions to a set of \code{GRaster}s from a model object.

The model must be either a linear model, which is of class \code{lm} and typically created using the \code{\link[stats:lm]{stats::lm()}} function or a generalized linear model (GLM), which is class \code{glm} and typically created using \code{\link[stats:glm]{stats::glm()}}. Other packages can also create \code{lm} or \code{glm} objects, but they may not work in this function. For example, generalized additive models, which can be created using the \code{gam()} function in the \strong{mgcv} package, inherit the \code{glm} class, but cannot be used in this function. However, \code{glm} objects created with the \strong{speedglm} package should work with this function.

This \code{predict()} function can handle:
\itemize{
\item Linear predictors and intercepts like \code{1 + x};
\item Quadratic terms like \code{x^2} (or, in \strong{R} formula notation, \code{I(x^2)});
\item Two-way interaction terms between scalars like \code{x1:x2} and \code{x1 * x2};
\item Categorical predictors (i.e., categorical \code{GRaster}s; see \code{vignette("GRasters", package = "fasterRaster")});
\item Two-way interactions between a categorical predictor and a scalar predictor; and
\item Two-way interactions between categorical predictors.
}
}
\examples{
if (grassStarted()) {

# Setup
library(sf)
library(terra)

### This example creates a simple model of Dypsis distribution using
# elevation, distance to forest, land cover class, and nearness to rivers.

# Elevation raster, forest cover in year 2000, land cover class, and
# points where Dypsis plants have been collected
madElev <- fastData("madElev")
madForest2000 <- fastData("madForest2000")
madCover <- fastData("madCover")
madRivers <- fastData("madRivers")
madDypsis <- fastData("madDypsis")

# Convert SpatRasters to GRasters and sf vectors to GVectors:
elev <- fast(madElev)
forest <- fast(madForest2000)
cover <- fast(madCover)
rivers <- fast(madRivers)
dypsis <- fast(madDypsis)

# Distance to forest
distToForest <- distance(forest, unit = "m")
distToForest <- log1p(distToForest) # log(x + 1) of distance
names(distToForest) <- "distToForest"

# "Stack" elevation and forest cover
continuous <- c(elev, distToForest)

# Scale continuous predictors to mean of 0 and sd of 1
continuousScaled <- scale(continuous)
names(continuousScaled) <- c("elevation", "distToForest")

# Project land cover raster
coverProj <- project(cover, continuousScaled)

# Near a river?
riverBuffer <- buffer(rivers, 5000)
nearRiver <- rasterize(riverBuffer, elev, background = 0)
names(nearRiver) <- "nearRiver"
levels(nearRiver) <- data.frame(value = 0:1, label = c("far", "near"))

# Combine continuous/categorical data
covariateRasters <- c(continuousScaled, coverProj, nearRiver)
plot(covariateRasters)

# Extract environmental values at Dypsis locations:
presEnv <- extract(covariateRasters, dypsis, cats = TRUE)
presEnv$presBg <- 1
head(presEnv)

# Extract elevation and forest cover at 2000 background sites:
bgEnv <- spatSample(covariateRasters, size = 3000, values = TRUE, cats = TRUE)
bgEnv <- bgEnv[stats::complete.cases(bgEnv), ]
bgEnv <- bgEnv[1:2000, ]
bgEnv$presBg <- 0
head(bgEnv)

# Combine presence and background data:
env <- rbind(presEnv, bgEnv)

# Calibrate model:
form <- presBg ~ elevation + distToForest +
  I(distToForest^2) + elevation * distToForest +
  madCover + nearRiver

model <- stats::glm(form, data = env, family = stats::binomial)
summary(model)

# Make predictions and map:
prediction <- predict(covariateRasters, model, type = "response")
prediction

# Not a great model!
plot(prediction, main = "Predicted")
plot(dypsis, pch = 1, add = TRUE)

}
}
\seealso{
\code{\link[terra:predict]{terra::predict()}}; \code{\link[stats:predict]{stats::predict()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/07_comparison.r
\name{Compare,GRaster,GRaster-method}
\alias{Compare,GRaster,GRaster-method}
\alias{Compare-methods}
\alias{Compare,logical,GRaster-method}
\alias{Compare,GRaster,logical-method}
\alias{Compare,numeric,GRaster-method}
\alias{Compare,GRaster,numeric-method}
\alias{Compare,GRaster,integer-method}
\alias{Compare,integer,GRaster-method}
\alias{Compare,GRaster,character-method}
\alias{Compare,character,GRaster-method}
\alias{Compare,GRegion,GRegion-method}
\title{Compare-methods operations on GRasters and GRegions}
\usage{
\S4method{Compare}{GRaster,GRaster}(e1, e2)

\S4method{Compare}{logical,GRaster}(e1, e2)

\S4method{Compare}{GRaster,logical}(e1, e2)

\S4method{Compare}{numeric,GRaster}(e1, e2)

\S4method{Compare}{GRaster,numeric}(e1, e2)

\S4method{Compare}{GRaster,integer}(e1, e2)

\S4method{Compare}{integer,GRaster}(e1, e2)

\S4method{Compare}{GRaster,character}(e1, e2)

\S4method{Compare}{character,GRaster}(e1, e2)

\S4method{Compare}{GRegion,GRegion}(e1, e2)
}
\arguments{
\item{e1, e2}{Values depend on the type of comparison:
\itemize{
\item Comparing \code{GRaster}s to logical, numeric, character values: \code{e1} and \code{e2} can be any one of these. Comparison to a character string can be useful when using a categorical raster, in which case you can use something like \code{raster1 == "Wetlands"} to coerce all "wetland" cells to be 1 (TRUE) and all others 0 (FALSE) or \code{NA} (if it was originally \code{NA}).
\item Comparing a \code{GRegion} to another \code{GRegion}: \code{e1} and \code{e2} must be \code{GRegion}s!
}}
}
\value{
Comparing \code{GRaster}s: An "integer" \code{GRaster} with values of 0 (FALSE), 1 (TRUE), or \code{NA} (neither).

Comparing \code{GRegion}s: Output is logical.
}
\description{
You can do comparative operations on \code{GRaster}s using normal operators in \strong{R}: \code{<}, \code{<=}, \code{==}, \code{!=}, \code{>=}, and \code{>}. You can also use \code{\link[fasterRaster]{\%in\%}} for categorical \code{GRasters} (see \code{vignette("GRasters", package = "fasterRaster")}).

You can also compare two \code{GRegion}s using the \code{==} and \code{!=} operators. Most users of \strong{fasterRaster} will not have to work much with "regions" (see \code{vignette("regions", package = "fasterRaster")}), so can ignore this functionality. \code{GRegion}s are the same if they have the same coordinate reference system, location/project and mapset (see \code{vignette("projects_mapsets", package = "fasterRaster")}), topology (2D or 3D), extent, and resolution. If both are 3D, then they must also have the same vertical extent and number of depths.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster
elev <- fast(madElev)
elevs <- c(elev, elev, log10(elev) - 1, sqrt(elev))
names(elevs) <- c("elev1", "elev2", "log_elev", "sqrt_elev")

elev
elevs

# Comparisons
elev < 100
elev <= 100
elev == 100
elev != 100
elev > 100
elev >= 100

elev + 100 < 2 * elev

elevs > 10
10 > elevs

# logic
elev < 10 | elev > 200
elev < 10 | cos(elev) > 0.9

elev < 10 | TRUE
TRUE | elev > 200

elev < 10 | FALSE
FALSE | elev > 200

elev < 10 & cos(elev) > 0.9

elev < 10 & TRUE
TRUE & elev > 200

elev < 10 & FALSE
FALSE & elev > 200

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convHull.r
\name{convHull,GVector-method}
\alias{convHull,GVector-method}
\alias{convHull}
\title{Minimum convex hull around a spatial vector}
\usage{
\S4method{convHull}{GVector}(x, by = "")
}
\arguments{
\item{x}{A \code{GVector}.}

\item{by}{Character: If \code{""} (default), then a convex hull is created for all geometries together. Otherwise, this is the name of a field in the vector. Hulls will be created for each set of geometries with the same value in this column.}
}
\value{
A \code{GVector}.
}
\description{
Create a minimum convex hull around a spatial vector.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)

# Points vector of specimens of species in the plant genus Dypsis
madDypsis <- fastData("madDypsis")

# Convert sf to a GVector:
dypsis <- fast(madDypsis)

### Convex hull for all plant specimens together:
ch <- convHull(dypsis)

### Convex hull for each species:
head(dypsis) # See the "rightsHolder" column?
chHolder <- convHull(dypsis, by = "rightsHolder")

### Plot:
plot(dypsis)
plot(ch, add = TRUE)
n <- length(chHolder)
for (i in 1:n) {
   plot(chHolder[[i]], border = i, add = TRUE)
}

}
}
\seealso{
\code{link[terra]{convHull}}, \code{link[sf]{st_convex_hull}}, tool \code{v.hull} in \strong{GRASS} (see \code{grassHelp("v.hull")})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/subst.r
\name{subst,GRaster-method}
\alias{subst,GRaster-method}
\alias{subst}
\title{Replace a specific value(s) in a GRaster}
\usage{
\S4method{subst}{GRaster}(x, from, to, others = NULL, warn = TRUE)
}
\arguments{
\item{x}{A \code{GRaster}.}

\item{from, to}{Vectors of numeric or character values. The value(s) in \code{from} will be replaced with the value(s) in \code{to}. They must be the same length, or, if you supply a single value for \code{to}, then all values in \code{from} will be converted to the same value of \code{to}. Numeric/integer or character vectors can be used:
\itemize{
\item \code{from} and \code{to} are numeric or integer vectors: Values in \code{from} will be replaced by their corresponding value in \code{to}.
\item \code{from} and \code{to} are character vectors: You can use a character vector for \code{from} and \code{to}. In this case, the input raster must be a factor (categorical) raster or an integer raster. If \code{from} is a character vector, these levels(categories) will be replaced by the levels in \code{to.} This can add levels to a \code{GRaster} \code{to} has labels that do not match any existing labels in \code{from}.
\item \code{from} is a character vector and \code{to} is an integer vector: Cells in \code{x} that correspond to the given label will have their values replaced by the corresponding value in \code{to}, and be matched the the corresponding label. If no label corresponds to the new value, a new level will be created. The input must be a categorical raster.
\item \code{from} is an integer vector and \code{to} is a character vector: Cells in \code{x} that have a value in \code{from} will be replaced by values that match the labels in \code{to}. If the input raster does not have a value that corresponds to the given label in \code{y}, then a new value will be created.
}}

\item{others}{\code{NULL} (default), \code{NA}, numeric, or a character:
\itemize{
\item \code{NULL} (default): Values that do not appear in \code{from} will be unchanged.
\item \code{NA}: Values that do not appear in \code{from} will be set to \code{NA}.
\item Character: Cells in \code{x} that do not appear in \code{to} will be assigned to this level. In this case, \code{x} must be a categorical (factor) raster.
}}

\item{warn}{Logical: If \code{TRUE} (default), display a warning when new levels are created.}
}
\value{
A \code{GRaster}.
}
\description{
This function replaces one or more user-specified values in a raster with other values. See \code{\link[=classify]{classify()}} for replacing ranges of values.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")
madCover <- fastData("madCover")

### Substitution within an integer/numeric raster

# Convert SpatRaster to GRaster
elev <- fast(madElev)

# Simple substitution of one value, keeping all other values
newElev <- elev
newElev[newElev == 100] <- -100
newElev[newElev > 500] <- 500
hist(newElev)

# Simple substitution of one value, keeping all other values
substituted <- subst(elev, from = 300, to = -300)
substituteds <- c(elev, substituted)
names(substituteds) <- c("original", "substituted")
plot(substituteds)

# Simple substitution of three values, keeping all other values
substituted <- subst(elev, from = c(299, 300, 301), to = c(-699, -600, -601))
substituteds <- c(elev, substituted)
names(substituteds) <- c("original", "substituted")
plot(substituteds)

# Simple substitution of three values to one other value, retaining remainder
substituted <- subst(elev, from = c(299, 300, 301), to = -1000)
substituteds <- c(elev, substituted)
names(substituteds) <- c("original", "substituted")
plot(substituteds)

# Simple substitution of one value, setting all other values to 100
substituted <- subst(elev, from = 300, to = -300, others = 100)
substituteds <- c(elev, substituted)
names(substituteds) <- c("original", "substituted")
plot(substituteds)

### Substitution within a factor/categorical raster

# Convert a SpatRaster to a GRaster:
cover <- fast(madCover)

cover <- droplevels(cover) # remove unused levels
levels(cover) # levels of "cover"

# Substitute using level name, replace with EXISTING level label
from <- "Mosaic cropland/vegetation"
to <- "Mosaic crops"
categ <- subst(cover, from = from, to = to)
freq(cover) # original frequencies of each land cover class
freq(categ) # note change in frequency of "from" and "to" categories
plot(c(cover, categ))

# Substitute using level name, replace with NEW level label
from <- c("Mosaic crops", "Mosaic cropland/vegetation")
to <- c("Mixed cropland")
categ <- subst(cover, from = from, to = to)
freq(cover) # original frequencies of each land cover class
freq(categ) # note change in frequency of "from" and "to" categories
plot(c(cover, categ))

# Substitute using level name, replace with NEW level label
from <- c("Mosaic crops", "Mosaic cropland/vegetation")
to <- c("Mixed cropland", "Mixed cropland/vegetation")
categ <- subst(cover, from = from, to = to)
freq(cover) # original frequencies of each land cover class
freq(categ) # note change in frequency of "from" and "to" categories
plot(c(cover, categ))

# Substitute using level name, replace with VALUE of an existing label
from <- c("Mosaic crops", "Mosaic cropland/vegetation")
to <- 120
categ <- subst(cover, from = from, to = to)
freq(cover) # original frequencies of each land cover class
freq(categ) # note change in frequency of "from" and "to" categories
plot(c(cover, categ))

# Substitute using level name, replace with new level name, replace all others
from <- c("Mosaic crops", "Mosaic cropland/vegetation")
to <- "Crops"
categ <- subst(cover, from = from, to = to, others = "Other")
freq(cover) # original frequencies of each land cover class
freq(categ) # note change in frequency of "from" and "to" categories
plot(c(cover, categ))

}
}
\seealso{
\code{\link[terra:subst]{terra::subst()}}, \code{\link[=classify]{classify()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/subset.r
\docType{methods}
\name{subset,GRaster-method}
\alias{subset,GRaster-method}
\alias{subset}
\alias{subset,GVector-method}
\title{Subset layers from a GRaster, or specific rows from a GVector}
\usage{
\S4method{subset}{GRaster}(x, subset, negate = FALSE)

\S4method{subset}{GVector}(x, subset, negate = FALSE)
}
\arguments{
\item{x}{A \code{GRaster} or \code{GVector}.}

\item{subset}{Numeric integer, integer, logical, or character: Indicates the layer(s) of a \code{GRaster} to subset, or the rows(s) of a \code{GVector} to return.}

\item{negate}{Logical: If \code{TRUE}, all layers or rows in \code{subset} will be \emph{removed} from the output. Default is \code{FALSE}.}
}
\value{
A \code{GRaster} or \code{GVector}.
}
\description{
\code{subset()} can be used to subset or remove one or more layers from a \code{GRaster}. It can also be used to subset or remove rows from a \code{GVector} with a data table.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

### GRasters

# Example data
madElev <- fastData("madElev") # elevation raster
madForest2000 <- fastData("madForest2000") # forest raster
madForest2014 <- fastData("madForest2014") # forest raster

# Convert SpatRasters to GRasters
elev <- fast(madElev)
forest2000 <- fast(madForest2000)
forest2014 <- fast(madForest2014)

# Re-assigning values of a GRaster
constant <- elev
constant[] <- pi
names(constant) <- "pi_raster"
constant

# Re-assigning specific values of a raster
replace <- elev
replace[replace == 1] <- -20
replace

# Subsetting specific values of a raster based on another raster
elevInForest <- elev[forest2000 == 1]
plot(c(elev, forest2000, elevInForest), nr = 1)

# Adding and replacing layers of a GRaster
rasts <- c(elev, constant, forest2000)

# Combine with another layer:
add(rasts) <- forest2014 # one way
rasts

rasts <- c(rasts, forest2014) # another way

### Subsetting GRaster layers

# Subset:
rasts <- c(elev, forest2000, forest2014)
rasts[[2:3]]
subset(rasts, 2:3)
subset(rasts, c("madForest2000", "madElev"))
rasts[[c("madForest2000", "madElev")]]
rasts$madForest2000

# Get every other layer:
rasts[[c(FALSE, TRUE)]]

### Replacing layers of a GRaster

# Replace a layer
logElev <- log(elev)
names(logElev) <- "logElev"
rasts$madForest2014 <- logElev
rasts

# Replace a layer:
rasts[[3]] <- forest2000
rasts

### GVectors

# example data
madDypsis <- fastData("madDypsis") # vector of points
madDypsis <- vect(madDypsis)

# Convert SpatVector to GVector
dypsis <- fast(madDypsis)

### Retrieving GVector columns

dypsis$species # Returns the column

dypsis[[c("year", "species")]] # Returns a GVector with these columns
dypsis[ , c("year", "species")] # Same as above

### Subsetting GVector geometries

# Subset three geometries
dypsis[c(1, 4, 10)]

# Subset three geometries and one column. Note order will always be the same
# in the output and may differ in order from terra subsetting.
dypsis[c(1, 4, 10), "species"]
dypsis[c(10, 4, 1), "species"] # fasterRaster: Same order as previous.
madDypsis[c(1, 4, 10), "species"]
madDypsis[c(10, 4, 1), "species"] # terra: different order as previous.

# Get geometries by data table condition
dypsis[dypsis$species == "Dypsis betsimisarakae"]

### (Re)assigning GVector column values

# New column
dypsis$pi <- pi
head(dypsis)

# Re-assign values
dypsis$pi <- "pie"
head(dypsis)

# Re-assign specific values
dypsis$institutionCode[dypsis$institutionCode == "MO"] <-
   "Missouri Botanical Garden"

}
}
\seealso{
\code{\link[fasterRaster]{[[}}, \code{\link[fasterRaster]{[}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rastInfo_vectInfo.r
\name{.rastInfo}
\alias{.rastInfo}
\title{Metadata on rasters and vectors in GRASS}
\usage{
.rastInfo(x)
}
\arguments{
\item{x}{A \code{GRaster}, \code{GVector}, or \code{sources}.}
}
\value{
Metadata on the extent, dimensions, resolution, bottom/top, etc. of rasters and vectors in \strong{GRASS}.
}
\description{
\code{.rastInfo()} and \code{.vectInfo()} fetch metadata on rasters and vectors in \strong{GRASS}. The \code{print()}, \code{show()}, and \code{summary()} functions can be used to display this metadata.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/makeSourceName.r
\name{.makeSourceName}
\alias{.makeSourceName}
\title{Make unique GRASS name for rasters, vectors, etc.}
\usage{
.makeSourceName(x = NULL, type = NULL, n = 1L, name = NULL)
}
\arguments{
\item{x}{Character or \code{NULL}: Descriptive string. \strong{Developers, please note}: To assist with debugging, \strong{GRASS} objects created by a \strong{GRASS} tool have the tool named in this argument (with underscores). Example: "v_in_ogr" or "r_resample".}

\item{type}{Character: \code{raster}, \code{raster3D}, \code{vector}, or \code{table}.}

\item{n}{Numeric integer: Number of names to make}

\item{name}{\code{NULL} (default) or \code{character}: Name of the output, attached as an attribute.}
}
\value{
Character vector.
}
\description{
Make unique GRASS name for rasters, vectors, etc.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grassInfo.r
\name{grassInfo}
\alias{grassInfo}
\title{GRASS citation, version, and copyright information}
\usage{
grassInfo(x = "citation")
}
\arguments{
\item{x}{Character: What to return. Any of:
\itemize{
\item \code{"citation"} (default)
\item \code{"copyright"}: Copyright information
\item \code{"version"}: Version number and release year
\item \code{"versionNumber"}: Version number as numeric, major and minor only (e.g., 8.4)
}

Partial matching is used and case is ignored.}
}
\value{
Character.
}
\description{
Report the \strong{GRASS} citation, version/release year, version number, or copyright information.
}
\examples{
if (grassStarted()) {

# Citation
grassInfo()

# Version number
grassInfo("version")

# Version number
grassInfo("versionNumber")

# Version number
grassInfo("versionNumber")

# Copyright
grassInfo("copyright")

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bioclims_GRaster.r, R/bioclims_SpatRaster.r
\name{bioclims,GRaster-method}
\alias{bioclims,GRaster-method}
\alias{bioclims}
\alias{bioclims,SpatRaster-method}
\title{BIOCLIM rasters}
\usage{
\S4method{bioclims}{GRaster}(
  ppt,
  tmin,
  tmax,
  tmean = NULL,
  bios = NULL,
  sample = TRUE,
  quarter = 3,
  pptDelta = 1,
  verbose = TRUE
)

\S4method{bioclims}{SpatRaster}(
  ppt,
  tmin,
  tmax,
  tmean = NULL,
  bios = NULL,
  sample = TRUE,
  quarter = 3,
  pptDelta = 1,
  verbose = TRUE
)
}
\arguments{
\item{ppt}{A multi-layered \code{GRaster} or \code{SpatRaster}, representing monthly/weekly/daily precipitation.}

\item{tmin, tmax}{A multi-layered \code{GRaster} or \code{SpatRaster}, representing monthly/weekly/daily minimum and maximum temperature.}

\item{tmean}{Either \code{NULL} (default), or a multi-layered \code{GRaster} or \code{SpatRaster}, representing monthly/weekly/daily average temperature. If \code{NULL}, \code{tmean} will be calculated internally from \code{tmin} and \code{tmax}. Providing these rasters thus saves time if you already have them on hand.}

\item{bios}{Any of:
\itemize{
\item Numeric values: Calculate these BIOCLIM variables. For example, \code{bios = c(1, 12)} calculates BIOCLIMs 1 and 12.
\item \code{NULL} (default): Calculate BIOCLIMs 1 through 19
\item \code{"*"}: Calculate all BIOCLIMs this function can calculate.
\item \code{"+"}: Calculate BIOCLIMs 41 onward.
\item Any combination of the above except \code{NULL} (e.g., \code{c(1, 12, "+")}).
}}

\item{sample}{Logical: If \code{TRUE} (default), BIO4 and 15 are calculated with the sample standard deviation. If \code{FALSE}, then the population standard deviation is used.}

\item{quarter}{Numeric: Length of a "quarter". BIOCLIM variables are typically calculated using monthly-averaged rasters (e.g., precipitation and temperature of January, February, etc.), in which case a "quarter" is 3 months (so the default for \code{quarter} is 3). However, this function can accommodate any set of rasters representing a time series (e.g., 365 for daily rasters), in which case the user can decide what constitutes a "quarter" for calculation of the any BIOCLIMs that use "quarters" in their definitions.}

\item{pptDelta}{Numeric: Value to add to precipitation for calculation of BIO15 (coefficient of variation of precipitation, times 100). Adding a small value avoids division by 0. The default is 1.}

\item{verbose}{Logical: If \code{TRUE} (default), display progress.}
}
\value{
A \code{GRaster} with one or more layers.
}
\description{
The BIOCLIM set of bioclimatic variables were created for modeling species' geographic distributions (Booth et al. 2014). This function can create the "standard" 19 set of variables, plus several more from an "extended" set.

"Classic" set of BIOCLIM variables (Booth et al. 2014):
The units reported below assume that input rasters are in mm (precipitation) and deg C (temperature), and that each raster represents a month (but other time units are allowed, with corresponding changes to the temporal units assumed below).
\itemize{
\item BIO1: Mean annual temperature, calculated using monthly means (deg C)
\item BIO2: Mean diurnal range across months (average of monthly difference between maximum and minimum temperature) (deg C)
\item BIO3: Isothermality (100 * BIO02 / BIO07; unit-less)
\item BIO4: Temperature seasonality (standard deviation across months of average monthly temperature * 100; deg C)
\item BIO5: Maximum temperature of the warmest month (based on maximum temperature; deg C)
\item BIO6: Minimum temperature of the coldest month (based on minimum temperature; deg C)
\item BIO7: Range of annual temperature (BIO05 - BIO06; deg C)
\item BIO8: Temperature of the wettest quarter (based on mean temperature; deg C)
\item BIO9: Temperature of the driest quarter (based on mean temperature; deg C)
\item BIO10: Temperature of the warmest quarter (based on mean temperature; deg C)
\item BIO11: Temperature of the coldest quarter (based on mean temperature; deg C)
\item BIO12: Total annual precipitation (mm)
\item BIO13: Precipitation of the wettest month (mm)
\item BIO14: Precipitation of the driest month (mm)
\item BIO15: Precipitation seasonality (100 * coefficient of variation; unit-less)
\item BIO16: Precipitation of the wettest quarter (mm)
\item BIO17: Precipitation of the driest quarter (mm)
\item BIO18: Precipitation of the warmest quarter (based on mean temperature; mm)
\item BIO19: Precipitation of the coldest quarter (based on mean temperature; mm)
}

"Extended" set of BIOCLIM variables (starts at 41 to avoid conflicts with Kriticos et al. 2014):
\itemize{
\item BIO41: Temperature of the quarter following the coldest quarter (based on mean temperature; deg C)
\item BIO42: Temperature of the quarter following the warmest quarter (based on mean temperature; deg C)
\item BIO43: Precipitation of the quarter following the coldest quarter (based on mean temperature; mm)
\item BIO44: Precipitation of the quarter following the warmest quarter (based on mean temperature; mm)
\item BIO45: Temperature of the quarter following the driest quarter (based on mean temperature; deg C)
\item BIO46: Temperature of the quarter following the wettest quarter (based on mean temperature; deg C)
\item BIO47: Precipitation of the quarter following the driest quarter (based on mean temperature; mm)
\item BIO48: Precipitation of the quarter following the wettest quarter (based on mean temperature; mm)
\item BIO49: Hottest month (based on maximum temperature)
\item BIO50: Coldest month (based on minimum temperature)
\item BIO51: Wettest month
\item BIO52: Driest month
\item BIO53: First month of the warmest quarter (based on mean temperature)
\item BIO54: First month of the coldest quarter (based on mean temperature)
\item BIO55: First month of the wettest quarter
\item BIO56: First month of the driest quarter
\item BIO57: The greatest decrease in temperature from one month to the next (deg C; always >= 0)
\item BIO58: The greatest increase in temperature from one month to the next (deg C; always >= 0)
\item BIO59: The greatest decrease in precipitation from one month to the next (mm; always >= 0)
\item BIO60: The greatest increase in precipitation from one month to the next (mm; always >= 0)
}

By default, "quarter" refers to any consecutive run of three months, not a financial quarter. A quarter can thus include November-December-January, or December-January-February, for example. However, the length of a quarter can be changed using the argument \code{quarter}.

The variables are defined assuming that the input rasters represent monthly values (12 rasters for min/max temperature and precipitation), but you can also use sets of 52 rasters, representing one per week, in which case "quarter" would be a successive run of 3 weeks. You could also attempt 365 rasters, in which case a "quarter" would be a run of 3 successive days.

BIOCLIMs 41 through 44 are added here to capture the "shoulder" seasons (spring and autumn) important in temperature regions. BIOCLIMs 45 through 48 are also included for consistency.

BIOCLIMs 49 through 60 are not bioclimatic variables per se, but useful for assessing the properties of the variables that are defined based on the "-est" month or quarter.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Load rasters with precipitation and min/max temperature
madPpt <- fastData("madPpt")
madTmin <- fastData("madTmin")
madTmax <- fastData("madTmax")

### Classic and extended BIOCLIMs from SpatRasters
bcSR <- bioclims(madPpt, madTmin, madTmax, bios = "*")
bcSR

### BIOCLIMs from GRasters
ppt <- fast(madPpt)
tmin <- fast(madTmin)
tmax <- fast(madTmax)

# For small rasters, takes longer to run compared to SpatRaster version:
bc <- bioclims(ppt, tmin, tmax, bios = c(1, 5, 12))
bc
plot(bc)

}
}
\references{
Booth, T.H., Nix, H.A., Busby, J.R., and Hutchinson, M.F.  2014.  BIOCLIM: The first species distribution modeling package, its early applications and relevance to most current MaxEnt studies. \emph{Diversity and Distributions} 20:1-9 \doi{10.1111/ddi.12144}.

Kriticos, D.J., Jarošik, V., and Otam N.  2014.  Extending the suite of BIOCLIM variables: A proposed registry system and case study using principal components analysis. \emph{Methods in Ecology and Evolution} 5:956-960 \doi{10.1111/2041-210X.12244}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/region.r
\name{.region,missing-method}
\alias{.region,missing-method}
\alias{.region}
\alias{.region,SpatRaster-method}
\alias{.region,GRegion-method}
\alias{.region,GRaster-method}
\alias{.region,GVector-method}
\alias{.regionExt,missing-method}
\alias{.regionExt}
\alias{.regionExt,numeric-method}
\alias{.regionExt,GSpatial-method}
\alias{.regionDim,missing-method}
\alias{.regionDim}
\alias{.regionDim,numeric-method}
\alias{.regionDim,GRegion-method}
\alias{.regionRes,missing-method}
\alias{.regionRes}
\alias{.regionRes,numeric-method}
\alias{.regionRes,GRegion-method}
\title{Report or change the extent, dimensions, and/or resolution of a region GRASS}
\usage{
\S4method{.region}{missing}(x)

\S4method{.region}{SpatRaster}(x)

\S4method{.region}{GRegion}(x)

\S4method{.region}{GRaster}(x, trim = NULL)

\S4method{.region}{GVector}(x)

\S4method{.regionExt}{missing}(x)

\S4method{.regionExt}{numeric}(x, respect)

\S4method{.regionExt}{GSpatial}(x, respect)

\S4method{.regionDim}{missing}(x)

\S4method{.regionDim}{numeric}(x, respect)

\S4method{.regionDim}{GRegion}(x, respect)

\S4method{.regionRes}{missing}(x)

\S4method{.regionRes}{numeric}(x, respect)

\S4method{.regionRes}{GRegion}(x, respect)
}
\arguments{
\item{x}{Any of:
\itemize{
\item Missing (default): Reports the extent, resolution, and dimensions of the current region. All other arguments will be ignored. You can also use \code{\link[=ext]{ext()}}, \code{\link[=dim]{dim()}}, and \code{\link[=res]{res()}} and related functions with missing arguments.
\item A \code{GSpatial}, \code{GRegion}, \code{GRaster}, \code{GVector} object: Sets the region"s extent, dimensions, and/or resolution to those of the object.
\item A \code{numeric} vector. This will resize the region's extent, resample the region's resolution/dimensions, or both, to ensure the desired dimensions or resolution are retained:
* 2 values for \code{.regionDim()}: Number of rows and columns
* 4 values for \code{.regionExt()}: Westernmost and easternmost easting (longitude), and southernmost and northernmost northing (latitude)
* 2 values for \code{.regionRes()}: Size of cells in the x- and y-dimensions
}}

\item{trim}{A \code{GRaster} or \code{NULL} (default). If a \code{GRaster}, then the region will be trimmed to the non-\code{NA} cells in this raster. \code{trim} can only be non-\code{NULL} if \code{x} is a \code{GRaster}. Ignored if \code{NULL}.}

\item{respect}{Character or \code{GRaster}: Indicates what aspect(s) of the current region to retain. Different functions allow for a different aspect to be retained. Partial matching is used.
\itemize{
\item \code{.regionDim()}: \code{"extent"} (extent unchanged, resolution may be changed) or \code{"resolution"} (resolution unchanged, extent may be changed).
\item \code{.regionExt()}: \code{"dimensions"} (dimensions unchanged, extent may be changed) or \code{"resolution"} (resolution unchanged, extent may be changed).
\item \code{.regionRes()}: \code{"extent"} (extent may be changed and/or dimensions may be changed to accommodate desired cell size) or \code{"dimensions"} (extent may be changed, dimensions unchanged).
Alternatively, a \code{GRaster} can be supplied:
\item \code{.regionDim()}: New region will have same extent and resolution.
\item \code{.regionExt()}: New region will have same dimensions and resolution.
\item \code{.regionRes()}: New region will have same extent and dimensions.
}

In this case, the new region"s registration will be the same as this raster, and cell resolution will be the same

Note: In most cases extent cannot be retained exactly if the resolution is changed. When resolution is changed, the actual extent will be the user-supplied extent expanded by zero to one rows or zero to one columns to accommodate an integer number of cells of the desired size. The western and northern limits of the extent will be retained, while the eastern and southern limits of the extent will be moved to accommodate an integer number of columns and rows.}
}
\value{
The value returned depends on how the function is used:
\itemize{
\item If used with no arguments, \code{.region()} returns a \code{GRegion} object.
\item If used with no arguments, \code{.regionDim()}, \code{.regionExt()}, and \code{.regionRes()} return numeric or integer vectors.
\item If the function is used to change reshape/resample the region, it returns a \code{GRegion} object reflecting the region \emph{before} it was changed. This allows users to revert to the original region if desired.
}
}
\description{
These functions either change the extent, dimensions, and/or resolution of a \strong{GRASS} "region" or report the current region's extent, dimensions, and/or resolution (see \code{vignette("regions", package = "fasterRaster")}). These functions are mostly used internally and rarely of interest to most users.
\itemize{
\item \code{.region()}: All 2D and 3D aspects of a region.
\item \code{.regionDim()}: x- and y-dimensions.
\item \code{.regionExt()}: x- and y-extent.
\item \code{.regionRes()}: x- and y-resolution.
}
}
\details{
When resizing extent, \strong{terra} keeps the \code{xmin} (west) and \code{ymax} (north) the fixed and shifts \code{xmax} (east) and \code{ymin} (south) as needed. To retain as much fidelity between \strong{fasterRaster} and \strong{terra} as possible, these functions do the same to the region.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vHasDatabase.r
\name{.vHasDatabase}
\alias{.vHasDatabase}
\title{Tests if a GRASS vector is linked to an attribute table}
\usage{
.vHasDatabase(x)
}
\arguments{
\item{x}{A \code{GVector} or the \code{\link[=sources]{sources()}} name of a vector in \strong{GRASS}.}
}
\value{
Logical.
}
\description{
\strong{GRASS} vectors can be lined to one or more attribute tables, or "databases.". This function tests to see if the vector does indeed have a database. This function is typically used by developers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fillNAs.r
\name{fillNAs,GRaster-method}
\alias{fillNAs,GRaster-method}
\alias{fillNAs}
\title{Fill NA cells in a raster using interpolation}
\usage{
\S4method{fillNAs}{GRaster}(
  x,
  lambda = NULL,
  method = "bilinear",
  min = -Inf,
  max = Inf,
  cells = Inf
)
}
\arguments{
\item{x}{A \code{GRaster}.}

\item{lambda}{Either \code{NULL} (default), or a numeric value > 0: If \code{NULL}, then the function will use leave-one-out crossvalidation to find the optimal value.}

\item{method}{Character: Type of spline, either "\code{bilinear}" (default), "\code{bicubic}", or "\code{RST}" (regularized splines with tension). Partial matching is used and case is ignored.

\strong{Note}: The RST method will often display warnings, but these can be ignored.}

\item{min, max}{Numeric: Lowest and highest values allowed in the interpolated values. Values outside these bounds will be truncated to the minimum/maximum value(s) allowed. The default imposes no constraints. For multi-layered rasters, you can supply a single value for \code{min} and/or \code{max}, or multiple values (one per layer). Values will be recycled if there are fewer than one or them per layer in the raster.}

\item{cells}{Integer or numeric integer: Number of cells away from the non-\code{NA} cells to fill. For example, if \code{cells = 2}, then only cells within a 2-cell buffer of non-\code{NA} cells will be filled. The default is \code{Inf} (fill all possible cells--some methods may not be able to do this, depending on the configuration of the raster).}
}
\value{
A \code{GRaster}.
}
\description{
This function uses splines to fill \code{NA} cells in a raster based on the values of nearby cells. Depending on the method used, not all \code{NA} cells can be filled.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Elevation raster:
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster:
elev <- fast(madElev)

### Fill NAs:
bilinear <- fillNAs(elev)
bicubic <- fillNAs(elev, method = "bicubic")
rst <- fillNAs(elev, method = "rst")

maps <- c(elev, bilinear, bicubic, rst)
names(maps) <- c("original", "bilinear", "bicubic", "RST")
plot(maps)

### Constrain interpolated values to > 0
constrained <- fillNAs(elev, min = 0)

# Compare unconstrained and constrained:
minmax(bilinear)
minmax(constrained)

### Interpolate to only first 10 cells away from non-NA cells:
restrained <- fillNAs(elev, cells = 10)

maps <- c(elev, restrained)
names(maps) <- c("Original", "within 10 cells")
plot(maps)

}
}
\seealso{
\code{\link[terra:interpNear]{terra::interpNear()}}, \strong{GRASS} tool \code{r.fillnulls} (see \code{grassHelp("r.fillnulls")})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wetness.r
\name{wetness,GRaster-method}
\alias{wetness,GRaster-method}
\alias{wetness}
\title{Topographic wetness index}
\usage{
\S4method{wetness}{GRaster}(x)
}
\arguments{
\item{x}{A \code{GRaster} (typically representing elevation). The raster must be projected (i.e., not in WGS84, NAD83, et cetera).}
}
\value{
A \code{GRaster}.
}
\description{
This function creates a raster map with values equal to the topographic wetness index (TWI), which is a measure of how much overland water flow tends to accumulate in or flow away from a location.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Elevation raster
madElev <- fastData("madElev")

# Convert to GRaster:
elev <- fast(madElev)

# Topographic wetness index:
twi <- wetness(elev)
names(twi) <- 'TWI'
plot(c(elev, twi))

# Terrain ruggedness index:
tri <- ruggedness(elev)
tri7 <- ruggedness(elev, size = 7)
triSmooth7 <- ruggedness(elev, size = 7, exponent = 4)

tris <- c(elev, tri, tri7, triSmooth7)
names(tris) <- c("elevation", "TRI in 3x3", "TRI in 7x7", "Smoothed TRIin 7x7")
plot(tris)

}
}
\seealso{
\code{\link[=terrain]{terrain()}}, \code{\link[=ruggedness]{ruggedness()}}, \code{\link[=geomorphons]{geomorphons()}}, \strong{GRASS} manual for tool \code{r.topidx} (see \code{grassHelp("r.topidx")})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classify.r
\name{classify,GRaster-method}
\alias{classify,GRaster-method}
\alias{classify}
\title{Classify GRaster cell values}
\usage{
\S4method{classify}{GRaster}(x, rcl, include.lowest = FALSE, right = TRUE, others = NULL)
}
\arguments{
\item{x}{A \code{GRaster}.}

\item{rcl}{Reclassification system:
\itemize{
\item A single integer: Number of "bins" into which to divide values. Arguments \code{include.lowest} and \code{right} apply.
\item A vector of numeric values: Breakpoints of bins into which to divide values. These will be sorted from lowest to highest before classification. Arguments \code{include.lowest} and \code{right} apply.
\item A 2-column \code{matrix}, \code{data.frame}, or \code{data.table}: The first column provides specific values in \code{x} to be replaced, and the second provides the values they are replaced with. This method is only useful for classifying \code{integer} \code{GRaster}s. Arguments \code{include.lowest} and \code{right} are ignored. Cells will be classified in the order in which values are listed in the first column.
\item A 3-column \code{matrix}, \code{data.frame}, or \code{data.table}: The first column provides the lower value of a bin, the second the upper value, and the third the value to assign to the cells in the bin. Arguments \code{include.lowest} and \code{right} apply. Cells will be classified in the order of how intervals are listed (intervals will not be sorted).
}}

\item{include.lowest, right}{Logical: These arguments determine how cells that have values exactly equal to the lower or upper ends of an interval are classified.
\itemize{
\item \code{include.lowest = TRUE} and \code{right = TRUE}: All intervals will be "left-open, right-closed" except for the lowest interval, which will be "left-closed/right-closed".
\item \code{include.lowest = FALSE} and \code{right = FALSE}: Intervals will be "left-closed/right-open". Cells with values equal to the highest higher boundary will not be reclassified.
\item \code{include.lowest = TRUE} and \code{right = FALSE}: All intervals will be "left-closed/right-open", except for the highest interval, which will be "right-closed/left-closed".
\item \code{right = NA}: Only useful for classifying \code{integer} \code{GRaster}s. All intervals are "left-closed/right-closed". This is easier than accounting for "open" intervals when dealing with integers. Argument \code{include.lowest} is ignored.
}}

\item{others}{Integer or \code{NULL} (default), or \code{NA}: Value to assign to cells that do not fall into the set intervals. Cells with \code{NA} values are not reclassified. Setting \code{others} equal to \code{NULL} or \code{NA} replaces all other values with \code{NA}. The value will be coerced to an integer value.}
}
\value{
A \code{GRaster}. The raster will be a categorical \code{GRaster} if the original values were continuous (i.e., a single- or double-precision raster), or of type "integer" if the input was an integer. See \code{vignette("GRasters", package = "fasterRaster")}.
}
\description{
This function classifies a `GRaster`` so that cells that have values within a given range are assigned a new value. The \code{\link[=subst]{subst()}} function is a simpler method for replacing specific values or category levels.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster:
elev <- fast(madElev)

# Classify using a scalar indicating number of bins
scalar <- classify(elev, 5)
scalar
levels(scalar)

# Classify using a vector, indicating bin break points
vector <- classify(elev, rcl = c(0, 100, 200, 300, 400, 500, 600))
vector
levels(vector)

# Classify using a 2-column matrix (only valid for integer rasters)
rcl <- data.frame(is = c(1:3, 5, 10), becomes = c(100:102, 105, 110))
twoCol <- classify(elev, rcl = rcl)
twoCol

# Classify using a 3-column table
rcl <- data.frame(
   from = c(0, 100, 200, 300, 400, 500),
   to = c(100, 200, 300, 400, 500, 600),
   becomes = c(1, 2, 3, 10, 12, 15)
)
threeCol <- classify(elev, rcl = rcl)
threeCol
levels(threeCol)

# Convert all values outside range to NA (default)
rcl <- c(100, 200, 300)
v1 <- classify(elev, rcl = rcl)
v1
plot(v1)

# Convert all values outside range to -1
rcl <- c(100, 200, 300)
v2 <- classify(elev, rcl = rcl, others = -1)
v2
plot(v2)

### Left-open/right-closed (default)
minmax(elev) # note min/max values
rcl <- c(1, 200, 570)
v3 <- classify(elev, rcl = rcl, others = 10)
levels(v3)
plot(v3)

### Left-closed/right-open
minmax(elev) # note min/max values
rcl <- c(1, 200, 570)
v4 <- classify(elev, rcl = rcl, others = 10, right = FALSE)
levels(v4)

# Left-open except for lowest bin/right-closed
minmax(elev) # note min/max values
rcl <- c(1, 200, 570)
v5 <- classify(elev, rcl = rcl, others = 10, include.lowest = TRUE)
v5 <- droplevels(v5)
levels(v5)

# Left-closed/right-open except for highest bin
minmax(elev) # note min/max values
rcl <- c(1, 200, 570)
v6 <- classify(elev, rcl = rcl, others = 10,
   right = FALSE, include.lowest = TRUE)
v6 <- droplevels(v6)
levels(v6)

}
}
\seealso{
\code{\link[terra:classify]{terra::classify()}}, \code{\link[=subst]{subst()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extend.r
\name{extend,GRaster,numeric-method}
\alias{extend,GRaster,numeric-method}
\alias{extend}
\alias{extend,GRaster,SpatRaster-method}
\alias{extend,GRaster,SpatVector-method}
\alias{extend,GRaster,SpatExtent-method}
\alias{extend,GRaster,sf-method}
\alias{extend,GRaster,GSpatial-method}
\title{Add rows and columns around a writeRaster}
\usage{
\S4method{extend}{GRaster,numeric}(x, y, fill = NA)

\S4method{extend}{GRaster,SpatRaster}(x, y, snap = "near", fill = NA)

\S4method{extend}{GRaster,SpatVector}(x, y, snap = "near", fill = NA)

\S4method{extend}{GRaster,SpatExtent}(x, y, snap = "near", fill = NA)

\S4method{extend}{GRaster,sf}(x, y, snap = "near", fill = NA)

\S4method{extend}{GRaster,GSpatial}(x, y, snap = "near", fill = NA)
}
\arguments{
\item{x}{A \code{GRaster}.}

\item{y}{Any of:
\itemize{
\item An object from which an extent can be obtained; i.e., a \code{SpatRaster}, \code{SpatVector}, \code{SpatExtent}, \code{sf} vector, or a \code{GSpatial} object (any of \code{GRaster}, \code{GVector}, or \code{GRegion}). If the extent of \code{x} is "outside" the extent of \code{y} on any side, the side(s) of \code{x} that are outside will be kept as-is (i.e., the extent of \code{x} will never be shrunk).
\item A single positive integer: Number of rows and columns to add to the top, bottom, and sides of the raster.
\item Two integers >= 0: Number of columns (1st value) to add to the sides, and number of rows (2nd value) to add to the top and bottom of the raster.
\item Four integers >= 0: Number of rows and columns to add (left column, right column, bottom row, top row).
}}

\item{fill}{Numeric: Value to place in the new cells. The default is \code{NA}.}

\item{snap}{Character: Method used to align \code{y} to \code{x}. Partial matching is used. This is only used if \code{y} is not a set of numbers.
\itemize{
\item \code{"near"} (default): Round to nearest row/column
\item \code{"in"}: Round "inward" toward the extent of \code{x} to nearest row/column
\item \code{"out"}: Round "outward" away from the extent of \code{x} to the nearest row/column.
}}
}
\value{
A \code{GRaster}.
}
\description{
\code{extend()} adds cells around a raster, making it larger.
}
\details{
Known issues: When \code{GRaster}s are saved to disk explicitly using \code{\link[=writeRaster]{writeRaster()}}, or implicitly using \code{\link[=rast]{rast()}} or \code{\link[=plot]{plot()}}, rows and columns that are entirely \code{NA} are dropped.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")
madRivers <- fastData("madRivers")

# Send spatial objects to GRASS:
elev <- fast(madElev)
rivers <- fast(madRivers)

# Extend raster by number of rows/columns:
extended1 <- extend(elev, 10, fill = 900)
extended2 <- extend(elev, c(10, 20), fill = 900)
extended3 <- extend(elev, c(10, 80, 0, 100), fill = 900)
dim(elev)
dim(extended1)
dim(extended2)
dim(extended3)

plot(extended3)

# When exporting a raster, NA rows and columns are removed.
extended4 <- extend(elev, 100, fill=1) # default fill is NA
extended4terra <- rast(extended4)

dim(extended4)
dim(extended4terra)

plot(extended4)

# Extend the raster by another object with a wider extent.

# For tis example, first crop the raster, then extend it.
elevCrop <- crop(elev, rivers)
uncrop <- extend(elevCrop, elev, fill = 900)
plot(uncrop)

}
}
\seealso{
\code{\link[terra:extend]{terra::extend()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as.contour.r
\name{as.contour,GRaster-method}
\alias{as.contour,GRaster-method}
\alias{as.contour}
\title{Contour lines from a "GRaster"}
\usage{
\S4method{as.contour}{GRaster}(x, nlevels, levels)
}
\arguments{
\item{x}{A \code{GRaster}.}

\item{nlevels}{Numeric: A positive integer or missing (default). Number of levels at which to calculate contours. Levels will be calculated in equal-sized steps from the smallest to the largest value of \code{x}. Either \code{nlevels} or \code{levels} must be specified.}

\item{levels}{Numeric vector: A numeric vector of values at which to calculate contour lines. Either \code{nlevels} or \code{levels} must be specified.}
}
\value{
A \code{GVector} representing contour lines.
}
\description{
Create a \code{GVector} of contour lines from a \code{GRaster}.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Elevation raster
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster:
elev <- fast(madElev)

# Calculate contour lines:
conts <- as.contour(elev, nlevels = 10)

plot(elev)
plot(conts, add = TRUE)

}
}
\seealso{
\code{\link[terra:contour]{terra::as.contour()}}, \strong{GRASS} manual page for tool \code{r.contour} (see \code{grassHelp("r.contour")})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/workDir.r
\name{.workDir,GLocation-method}
\alias{.workDir,GLocation-method}
\alias{.workDir}
\alias{.workDir,missing-method}
\title{Get a GLocation's working directory}
\usage{
\S4method{.workDir}{GLocation}(x)

\S4method{.workDir}{missing}(x)
}
\arguments{
\item{x}{A \code{GLocation} object or missing. If an object, returns the working folder in which the object is saved by \strong{GRASS}. If missing, then just returns the working folder (same as \code{faster("workDir")}).}
}
\value{
Character.
}
\description{
This function returns the working directory of a \code{GLocation} object.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clusterPoints.r
\name{clusterPoints,GVector-method}
\alias{clusterPoints,GVector-method}
\alias{clusterPoints}
\title{Identify clusters of points}
\usage{
\S4method{clusterPoints}{GVector}(x, method = "DBSCAN", minIn = NULL, maxDist = NULL)
}
\arguments{
\item{x}{A "points" \code{GVector}.}

\item{method}{Character: Method used to identify clusters. Explanations of methods are provided in the help page for the \strong{GRASS} tool \code{v.cluster}, available using \code{grassHelp("v.cluster")}.
\itemize{
\item \code{"DBSCAN"} (default): Density-Based Spatial Clustering of Applications with Noise.
\item \code{"DBSCAN2"}: A modification of DBSCAN.
\item \code{"density"}: Cluster points by relative density.
\item \code{"OPTICS"}: Ordering Points to Identify the Clustering Structure
\item \code{"OPTICS2"}: A modification of OPTICS.
}

Case is ignored, but partial matching is not used.}

\item{minIn}{Integer, numeric integer, or \code{NULL} (default): Minimum number of points in a cluster. If \code{NULL}, then \code{minIn} is set to 3 for a 2-dimensional vector and 4 for a 3-dimensional vector.}

\item{maxDist}{Numeric or \code{NULL} (default): Maximum distance between neighboring points in a cluster for DBSCAN, DBSCAN2, and OPTICS. If \code{NULL}, the maximum distance will be set to the 99th quantile of observed pairwise distances between points.}
}
\value{
A vector of integers indicating the cluster to which each point belongs.
}
\description{
\code{clusterPoints()} partitions points in a "points" \code{GVector} into clusters.
}
\seealso{
\strong{GRASS} manual page for tool \code{v.cluster} (see \code{grassHelp("v.cluster")})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vAsDataTable.r
\name{.vAsDataTable}
\alias{.vAsDataTable}
\title{Convert a GRASS vector's attribute table to a data.table}
\usage{
.vAsDataTable(x)
}
\arguments{
\item{x}{A \code{GVector} or the name of a vector in \strong{GRASS}.}
}
\value{
A \code{data.table} or \code{NULL} if the vector is not attached to a database.
}
\description{
\strong{GRASS} vectors can be linked to an attribute table, which can be exported from \strong{GRASS} to \strong{R} using this function. \strong{This function is mostly of use to developers.}

Values in the \code{cat} column are not necessarily unique--if a value appears more than once, the set of features they index are (in other software) called "multipart" features. The table can have more columns with metadata for each feature.

This function is typically used by developers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/terrain.r
\name{terrain,GRaster-method}
\alias{terrain,GRaster-method}
\alias{terrain}
\title{Slope, aspect, curvature, and partial slopes}
\usage{
\S4method{terrain}{GRaster}(
  x,
  v = "slope",
  units = "degrees",
  undefinedAspect = NA,
  northIs0 = TRUE
)
}
\arguments{
\item{x}{A \code{GRaster} (typically representing elevation).}

\item{v}{Name of the topographic metric(s) to calculate. Valid values include one or more of:
\itemize{
\item \code{"slope"}: Slope. Units are given by argument \code{units}.
\item \code{"aspect"}: Aspect. When argument \code{northIs0} is \code{TRUE} (default), then aspect is given in degrees from north going clockwise (0 = north, 90 = east, 180 = south, 270 = west).  Units are given by argument \code{units}.
\item \code{"profileCurve"}: Profile curvature.
\item \code{"tanCurve"}: Tangential curvature.
\item \code{"dx"}: Slope in east-west direction.
\item \code{"dy"}: Slope in north-south direction.
\item \code{"dxx"}: Second partial derivative in east-west direction.
\item \code{"dyy"}: Second partial derivative in north-south direction.
\item \code{"dxy"}: Second partial derivative along east-west and north-south direction.
\item \code{"*"}: All of the above.
}}

\item{units}{Character: "Units" in which to calculate slope and aspect: either \code{"degrees"} for degrees (default), \code{"radians"}, or \code{"percent"}. Partial matching is used.}

\item{undefinedAspect}{Numeric or \code{NA} (default): Value to assign to flat areas for which aspect cannot be calculated.}

\item{northIs0}{Logical: If \code{TRUE} (default), aspect will be reported in "north orientation," such that 0 is north, and degrees run clockwise (90 is east, 180 south, 270 west). If \code{FALSE}, then aspect will be reported in "east orientation," such that 0 is east, and degrees run counterclockwise (90 is north, 180 west, 270 south). The latter is the default in \strong{GRASS}, but the former is the default in \code{\link[terra:terrain]{terra::terrain()}} function, so is used here as the default. \strong{Note:} The \code{\link[=sun]{sun()}} function requires aspect to be in east orientation.}
}
\value{
A \code{GRaster} with one or more layers.
}
\description{
\code{terrain()} calculates topographic indices, including slope, aspect, curvature, and partial slopes (slopes in the east-west or north-south directions).
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster
elev <- fast(madElev)

# Calculate all topographic metrics
topos <- terrain(elev, v = "*")
topos

plot(topos) # NB Aspect has values of NA when it cannot be defined

# Calculate a hillshade raster
hs <- hillshade(elev)
plot(hs)

}
}
\seealso{
\code{\link[terra:terrain]{terra::terrain()}}, \code{\link[=ruggedness]{ruggedness()}}, \code{\link[=wetness]{wetness()}}, \code{\link[=geomorphons]{geomorphons()}}, tool \code{r.slope.aspect} in \strong{GRASS}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runifRast.r
\name{rUnifRast,GRaster-method}
\alias{rUnifRast,GRaster-method}
\alias{rUnifRast}
\title{Create a raster with random values drawn from a uniform distribution}
\usage{
\S4method{rUnifRast}{GRaster}(x, n = 1, low = 0, high = 1, seed = NULL)
}
\arguments{
\item{x}{A \code{GRaster}. The output will have the same extent and dimensions as this raster.}

\item{n}{A numeric integer: Number of rasters to generate.}

\item{low, high}{Numeric: Minimum and maximum values from which to select.}

\item{seed}{Numeric integer vector or \code{NULL}: Random seed. If \code{NULL}, then a different seed will be generated by \strong{GRASS}. Defining this is useful if you want to recreate rasters.  If provided, there should be one seed value per raster.}
}
\value{
A \code{GRaster}.
}
\description{
\code{rUnifRast()} creates a raster with values drawn from a uniform (flat) distribution.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Elevation raster
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster:
elev <- fast(madElev)

### Create a raster with values drawn from a uniform distribution:
unif <- rUnifRast(elev)
plot(unif)

### Create a raster with values drawn from a normal distribution:
norms <- rNormRast(elev, n = 2, mu = c(5, 10), sigma = c(2, 1))
plot(norms)
hist(norms, bins = 100)

# Create a raster with random, seemingly normally-distributed values:
rand <- rSpatialDepRast(elev, dist = 1000)
plot(rand)

# Values appear normal on first inspection:
hist(rand)

# ... but actually are patterned:
hist(rand, bins = 100)

# Create a fractal raster:
fractal <- fractalRast(elev, n = 2, dimension = c(2.1, 2.8))
plot(fractal)
hist(fractal)

### Random walker rasters

# One random walker
walk <- rWalkRast(elev)
plot(walk)

# Random walker with self-avoidance:
walkAvoid <- rWalkRast(elev, steps = 1000, avoid = TRUE, seed = 1)
plot(walkAvoid)

# 10 random walkers:
walk10 <- rWalkRast(elev, n = 10)
plot(walk10)

# 10 random walkers starting in same place:
walkSame10 <- rWalkRast(elev, n = 10, sameStart = TRUE)
plot(walkSame10)


}
}
\seealso{
\code{\link[=rNormRast]{rNormRast()}}, \code{\link[=rSpatialDepRast]{rSpatialDepRast()}}, \code{\link[=fractalRast]{fractalRast()}}, \code{\link[=rWalkRast]{rWalkRast()}}, \strong{GRASS} manual page for tool \code{r.random.surface} (see \code{grassHelp("r.random.surface")})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/seqToSQL.r
\name{seqToSQL}
\alias{seqToSQL}
\title{Format a numeric series into an SQL value call}
\usage{
seqToSQL(x, maxChar = 29900, sort = TRUE)
}
\arguments{
\item{x}{A vector of numerical values. The vector should be sorted from  lowers to highest for the most efficient "compression" of sequential ranges. Values will be coerced to class \code{integer}.}

\item{maxChar}{Integer or numeric: Maximum number of characters to include in the output. If the output has more than this number of characters, the remainder is dropped, and the \code{trim} attribute of the output is set to \code{TRUE}. The default is 29900, which is the maximum length of an SQL statement that \strong{GRASS} seems to be able to handle (minus a safety margin).}

\item{sort}{Logical: If \code{TRUE} (default), sort \code{x} before converting to SQL. This can reduce the length of the output.}
}
\value{
A character string. The string has three attributes. The \code{trim} attribute is \code{TRUE} or \code{FALSE}, depending on whether \code{maxChar} was reached or not (and subsequent numbers dropped from the string). The \code{lastIndex} attribute is the last index of \code{x} that was processed (i.e., the index of the last value in the output), and the number of values represented by the output.
}
\description{
This function takes as its argument a vector of integers or numeric values, and converts sequential runs to a range while keeping non-sequential values as-is. For example, \code{c(1, 5, 6, 7, 8, 9, 15, 16, 20)} becomes \code{"1,5-9,15-16,20"}. This reduces the number of characters necessary to supply to a SQL condition. This function is mainly of use to developers.
}
\examples{

x <- 1:5
seqToSQL(x)

x <- c(1:5, 7)
seqToSQL(x)

x <- c(1:5, 7, 15:16)
y <- c(1:5, 7, 15:16, 20)
seqToSQL(x)
seqToSQL(y)

seqToSQL(x, maxChar = 5)
seqToSQL(y, maxChar = 8)

seqToSQL(10:1, sort = FALSE)
seqToSQL(10:1, sort = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nlevels.r
\name{.nlevels}
\alias{.nlevels}
\title{Count number of levels from a data.frame/table, list, or SpatRaster}
\usage{
.nlevels(x)
}
\arguments{
\item{x}{A \code{GRaster}, \code{SpatRaster}, \emph{or} a \code{data.frame}, \code{data.table}, an empty string, or a list thereof.}
}
\description{
Counts number of levels in a character string (specifically, the empty string \code{""}), a \code{data.frame}, \code{data.table}, or list of \code{data.frame}s or \code{data.table}s or empty strings.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/locationCreate.r
\name{.locationCreate,character-method}
\alias{.locationCreate,character-method}
\alias{.locationCreate}
\alias{.locationCreate,SpatRaster-method}
\alias{.locationCreate,SpatVector-method}
\alias{.locationCreate,sf-method}
\title{Connect to "GRASS"}
\usage{
\S4method{.locationCreate}{character}(x, location = NULL, overwrite = FALSE, warn = TRUE)

\S4method{.locationCreate}{SpatRaster}(x, location = NULL, overwrite = FALSE, warn = TRUE)

\S4method{.locationCreate}{SpatVector}(x, location = NULL, overwrite = FALSE, warn = TRUE)

\S4method{.locationCreate}{sf}(x, location = NULL, overwrite = FALSE, warn = TRUE)
}
\arguments{
\item{x}{Any object from which a coordinate reference system (CRS) can be acquired. Ergo, any of:
\itemize{
\item A \code{SpatRaster}, \code{SpatVector}, \code{SpatExtent}, \code{stars}, or \code{sf} object
\item A \code{crs} object (i.e., from \code{\link[sf:st_crs]{sf::st_crs()}}).
\item A CRS (coordinate reference system) WKT string. Some PROJ4 strings \emph{might} work, too.
}}

\item{location}{Character or \code{NULL} (default): Name of the location.}

\item{overwrite}{Logical: If \code{FALSE} (default), and a \strong{GRASS} "coordinate reference frame" with the given name has already been created, then the function will fail. If \code{TRUE}, then the existing \strong{GRASS} "coordinate reference frame" of the same name will be overwritten. \emph{NOTE}: This will \strong{not} remove any \strong{R} objects associated with rasters or vectors in the "location", but they will no longer work because the objects they point to will be overwritten.}

\item{warn}{Logical: If \code{TRUE} (default) and \code{overwrite} is \code{TRUE}, then display a warning.}
}
\value{
A \link{GLocation} object (invisibly).
}
\description{
This function initializes a \strong{GRASS} "project" (previously known in \strong{GRASS} as a "location"; see \code{vignette("projects_mapsets", package = "fasterRaster")}). You need to run this function (often just once) before you use most functions in \strong{fasterRaster}. This function is of use to developers, not most users.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rm.r
\name{.rm}
\alias{.rm}
\title{Delete objects in the active GRASS session}
\usage{
.rm(x, type = NULL, warn = TRUE, verify = TRUE)
}
\arguments{
\item{x}{Character, a \code{GSpatial} object, or missing (default):
\itemize{
\item Character:
\itemize{
\item Any of \code{"rasters"} (all rasters), \code{"vectors"} (all spatial vectors), \code{"rasters3d"} (3D-rasters), and/or \code{"groups"} (groups), and all of these types will be deleted. Partial matching is supported.
\item The \code{sources} of the object to be deleted. Argument \code{type} must be specified.
}
\item \code{GSpatial} object (i.e., a \code{GRaster} or \code{GVector}): Delete this object.
\item Missing: Delete everything in the active \strong{GRASS} session.
}}

\item{type}{The type of spatial objects to delete. This can include \code{"rasters"} (all rasters), \code{"vectors"} (all spatial vectors), \code{"rasters3d"} (3D-rasters), and/or \code{"groups"} (groups). Partial matching is supported. If missing, all objects are candidates for deletion if they match \code{x}.}

\item{warn}{Logical: If \code{TRUE} (default), display warning if no matches or if everything in \strong{GRASS} is to be deleted.}

\item{verify}{Logical: If \code{TRUE} (default), the function will search for the item(s) to be deleted first to verify they exist. If this is \code{FALSE}, then \code{x} MUST be specified and \code{type} must be '\code{raster}' or \code{'vector'} (one value per value in \code{x}). This has no effect if \code{x} is a \code{GSpatial} object. It's main use is to save a bit of time.}
}
\value{
\code{TRUE} (invisibly).
}
\description{
Delete the names of all rasters and/or vectors that have been exported to or created in the active \strong{GRASS} session's location and mapset.
}
\seealso{
\code{\link[=rm]{rm()}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fast.r
\name{fast}
\alias{fast}
\alias{fast,character-method}
\alias{fast,SpatRaster-method}
\alias{fast,SpatVector-method}
\alias{fast,sf-method}
\alias{fast,missing-method}
\alias{fast,numeric-method}
\alias{fast,data.frame-method}
\alias{fast,data.table-method}
\alias{fast,matrix-method}
\title{Create a GRaster or GVector}
\usage{
\S4method{fast}{character}(
  x,
  rastOrVect = NULL,
  levels = TRUE,
  extent = NULL,
  correct = TRUE,
  snap = NULL,
  area = NULL,
  steps = 10,
  dropTable = FALSE,
  resolve = NA,
  verbose = TRUE,
  ...
)

\S4method{fast}{SpatRaster}(x, ...)

\S4method{fast}{SpatVector}(
  x,
  extent = NULL,
  correct = TRUE,
  snap = NULL,
  area = NULL,
  steps = 10,
  dropTable = FALSE,
  resolve = NA,
  verbose = TRUE
)

\S4method{fast}{sf}(
  x,
  extent = NULL,
  correct = TRUE,
  snap = NULL,
  area = NULL,
  steps = 10,
  resolve = NA,
  dropTable = FALSE,
  verbose = TRUE
)

\S4method{fast}{missing}(x, rastOrVect, crs = "")

\S4method{fast}{numeric}(x, crs = "", keepgeom = FALSE)

\S4method{fast}{data.frame}(x, geom = 1:2, crs = "", keepgeom = FALSE)

\S4method{fast}{data.table}(x, geom = 1:2, crs = "", keepgeom = FALSE)

\S4method{fast}{matrix}(x, geom = 1:2, crs = "", keepgeom = FALSE)
}
\arguments{
\item{x}{Any one of:
\itemize{
\item A \code{SpatRaster} raster. Rasters can have one or more layers.
\item A \code{SpatVector} or \code{sf} spatial vector. See especially arguments \code{correct}, \code{area}, \code{snap}, \code{steps}, and \code{verbose}.
\item A character string or a vector of strings with the path(s) and filename(s) of one or more rasters or one vector to be loaded directly into \strong{GRASS}. The function will attempt to ascertain the type of object from the file extension (raster or vector), but it can help to indicate which it is using the \code{rastOrVect} argument if it is unclear. For rasters, see especially argument \code{levels}. For vectors, see especially arguments \code{correct}, \code{resolve}, \code{area}, \code{snap}, \code{steps}, and \code{verbose}.
\item A vector with an even number of numeric values representing longitude/latitude pairs. See arguments \code{geom}, \code{keepgeom}, and \code{crs}.
\item A \code{data.frame}, \code{data.table}, or \code{matrix}: Create a \code{points} \code{GVector}. Two of the columns must represent longitude and latitude. See arguments \code{geom}, \code{keepgeom}, and \code{crs}.
\item Missing: Creates a generic \code{GRaster} or \code{GVector}. You must specify \code{rastOrVect}; for example, \code{fast(rastOrVect = "raster")}. Also see argument \code{crs}.
}}

\item{rastOrVect}{Either \code{NULL} (default), or \code{"raster"} or \code{"vector"}: If \code{x} is a filename, then the function will try to ascertain whether it represents a raster or a vector, but sometimes this will fail. In that case, it can help to specify if the file holds a raster or vector. Partial matching is used.}

\item{levels}{(\code{GRaster}s only): Any of:
\itemize{
\item Logical: If \code{TRUE} (default) and at least one layer of a raster is of type \code{integer}, search for a "levels" file, load it, and attach levels. A levels file will have the same name as the raster file, but end with any of "rdata", "rdat", "rda", "rds", "csv", or "tab" (case will generally not matter). If such a file is not found, no levels will be assigned. The levels file must contain either a \code{data.frame}, \code{data.table}, or \code{list} of \code{data.frame}s or \code{data.table}s, or \code{NULL}.
\item A \code{data.frame}, \code{data.table}, or list of \code{data.frame}s or \code{data.table}s with categories for categorical rasters. The first column of a table corresponds to raster values and must be of type \code{integer}. A subsequent column corresponds to category labels. By default, the second column is assumed to represent labels, but this can be changed with \code{\link[fasterRaster]{activeCat<-}}. Level tables can also be \code{NULL} (e.g., \code{data.fame(NULL)}). You can also assign levels after loading a raster using \code{\link[fasterRaster]{levels<-}}.
\item \code{NULL}: Do not attach a levels table.
#'
}}

\item{extent}{(\code{GVector}s only): Either a \code{NULL} (default), or a \code{GVector}, a \code{SpatVector}, a \code{SpatExtent} object, an \code{sf} vector, a \code{bbox} object, or a numeric vector of 4 values providing a bounding box. If provided, only vector features within this bounding box are imported. If \code{extent} is a numeric vector, the values \emph{must} be in the order west, east, south, north. If \code{NULL}, the entire vector is imported.}

\item{correct}{Logical (\code{GVector}s only): Correct topological issues. See \emph{Details} for more details! By default, this is \code{TRUE}.}

\item{snap}{\code{GVector}s only: Numeric or \code{NULL} (default). The value of \code{snap} indicates how close vertices need to be for them to be shifted to to the same location. Units of \code{snap} are map units (usually meters), or degrees for unprojected CRSs. For lines and polygons vectors, a value of \code{NULL} will invoke an iterative procedure to find an optimal, smallest value of \code{snap}. To turn snapping off, set \code{snap = 0}. See \emph{Details} for more details!}

\item{area}{Polygon \code{GVector}s only: Either a positive numeric value or \code{NULL} (default). Remove polygons with an area smaller than this value. Units of \code{area} are in square meters (regardless of the CRS). If \code{NULL}, then an iterative procedure is used to identify a value of \code{area} that results in a topologically correct polygon vector. For point and lines vectors, this argument is ignored. To turn area removal off, set \code{area = 0}. See \emph{Details} for more details!}

\item{steps}{\code{GVector}s only: A positive integer > 1 (default is 10). When using automatic vector correction (i.e., either \code{snap = NULL} and/or \code{area = NULL}), this is the number of values of \code{snap} and/or \code{area} to try to generate a correct topology, including no snapping or polygon removal (i.e, \code{snap = 0} and \code{area = 0}).}

\item{dropTable}{\code{GVector}s only: Logical. If \code{TRUE}, then drop the data table associated with a vector. By default, this is \code{FALSE}. See \emph{Details} for more details!}

\item{resolve}{\code{GVector}s only: Character. If a \code{GVector} would be topologically invalid after a first attempt at creating it, then this method will be used to resolve the issue and create a valid \code{GVector}. Partial matching is used.
\itemize{
\item \code{"disaggregate"}: Coerce each area of overlap between polygons into its own geometry. The output will not have a data table associated with it.
\item \code{"aggregate"}: Coerce all geometries into a "multipart" geometry so it acts like a single geometry. The output will not have a data table associated with it.
\item \code{NA} (default): Do neither of the above and if either \code{snap} or \code{area} is \code{NULL}, keep trying to create the \code{GVector}. Upon success, the \code{GVector} will retain any data table associated with it unless \code{dropTable} is \code{FALSE}.
}}

\item{verbose}{\code{GVector}s only: Logical. Displays progress when using automatic topology correction.}

\item{...}{Other arguments::
\itemize{
\item \code{table} (\code{GVector}s--useful mainly to developers, not most users): A \code{data.frame} or \code{data.table} with one row per geometry in a \code{GVector}. Serves as an attribute table.
\item \code{xVect} (\code{GVector}s--useful mainly to developers, not most users): The \code{SpatVector} that corresponds to the file named by \code{x}.
}}

\item{crs}{String: Coordinate reference system (CRS) WKT2 string. This argument is used for creating a \code{GVector} from a \code{numeric} vector or a \code{data.frame} or similar, or from \code{fast(rastOrVect = "vector")} or \code{fast(rastOrVect = "raster")}. By default, the function will use the value of \code{\link[=crs]{crs()}} (no arguments), which is the CRS of the current \strong{GRASS} "project/location" (see \code{vignette("projects_mapsets", package = "fasterRaster")}).}

\item{keepgeom}{Logical: If \code{x} is a set of \code{numeric} coordinates, or a \code{data.frame} or similar, then they can be coerced into a \code{points} \code{GVector}. If \code{keepgeom} is \code{TRUE}, then the coordinates will be included in the data table of the \code{GVector}. The default is \code{FALSE}.}

\item{geom}{Character or integer vector: If \code{x} is a \code{data.frame}, \code{data.table}, or \code{matrix}, this specifies which columns of \code{x} represent longitude and latitude. Columns can be given by name (a character vector) or index (a numeric or integer vector). The default is to use the first two columns of \code{x}.}
}
\value{
A \code{GRaster} or \code{GVector}.
}
\description{
\code{fast()} creates a \code{GRaster} or \code{GVector} from 1) a file; 2) from a \code{SpatRaster}, \code{SpatVector}, or \code{sf} vector; or 3) from a numeric vector, \code{matrix}, \code{data.frame}, or \code{data.table}. Behind the scenes, this function will also create a connection to \strong{GRASS} if none has yet been made yet.

\strong{GRASS} supports loading from disk a variety of raster formats (see the \strong{GRASS} manual page for \code{r.in.gdal} (see \code{grassHelp("r.in.gdal")}) and vector formats \code{v.in.ogr} (see grassHelp("v.in.ogr")`), though not all of them will work with this function.

Note that \code{GVectors} may fail to be created if they contain issues that do not coincide with the topological data model used by \strong{GRASS}. The most common of these is overlapping polygons. See \emph{Details} on how to fix these kinds of issues.

Note also that \strong{GRASS} (and thus, \strong{fasterRaster}) is \emph{not} very fast when loading vectors. So, if the vector is large and you only want a portion of it, consider using the \code{extent} argument to load the spatial subset you need.
}
\details{
\strong{GRASS} uses a "topological" model for vectors. Topological issues generally arise only with polygon vectors, not point or line vectors. Sometimes, polygons created in other software are topologically incorrect--the borders of adjacent polygons may cross one another, or there may be small gaps between them. These errors can be corrected by slightly shifting vertices and/or removing small polygons that result from intersections of larger ones that border one another. A topological system also recognizes that boundaries to adjacent polygons are shared by the areas, so should not be ascribed attributes that belong to both areas (e.g., the shared border between two countries "belongs" to both countries).

By default, \code{fast()} will try to correct topological errors in vectors. There are three levels of correction, and they are not necessarily mutually exclusive:
\enumerate{
\item \strong{Automatic correction}: By default, \code{fast()} will apply automatic topology correction. You can turn this off using the \code{correct = FALSE} argument, though in most cases this is not recommended.
\item \strong{\emph{Manual} snapping and/or area removal}: In addition to correction from step 1, you can cause vertices of polygons close to one another to be "snapped" to same place and/or polygons that are smaller than some threshold to be removed. Problems with mis-aligned vertices arise when adjacent polygons are meant to share borders, but slight differences in the locations of the vertices cause them to  mis-align. This mis-alignment can also produce small "slivers" of polygons that are the areas where they overlap. You can snap vertices within a given distance of one another using the \code{snap} argument followed by a numeric value, like \code{snap = 0.000001}. Units of \code{snap} are in map units (usually meters) for projected coordinate reference systems and degrees for unprojected systems (e.g., WGS84, NAD83, NAD27). You can also remove polygons that are smaller than a particular area using the \code{area} argument followed by a numeric value (e.g., \code{area = 1}). The units of \code{area} are in meters-squared, regardless of the coordinate reference system. Note that using \code{snap} and \code{area} entails some risk, as it is possible for nearby vertices to actually be distinct and for small areas to be legitimate.
\item \strong{Automatic snapping and/or area removal}: In addition to the correction from step 1, you can use automatic \code{snap} and/or \code{area} correction on polygons vectors by setting \code{snap} and/or \code{area} to \code{NULL} (i.e., their default values). If just \code{snap} is \code{NULL}, only automatic snapping will be performed, and if just \code{area} is \code{NULL}, then only automatic area removal will be performed. Regardless, you will also need to set an integer value for \code{steps}, which is the number of steps to take between the smallest value of \code{snap} and/or \code{area} and the maximum value attempted. The function will then proceed by first attempting \code{snap = 0} and/or \code{area = 0} (i.e., no snapping or area removal). If this does not produce a topologically correct vector, \strong{GRASS} will (internally) suggest a range for \code{snap}. The \code{fast()} function then creates \code{steps} values from the lowest to the highest values of this range evenly-spaced along the log values of this range, then proceed to repeat the importing process until either the vector is imported correctly or the maximum value of \code{snap} is reached and results in a failed topology. Smaller values of \code{step} will result in more fine-grained attempts so are less likely to yield overcorrection, but can also take more time. The value of \code{area} in automatic correction is set to \code{snap^2}. \strong{NB}: Automated snapping and area removal are only performed on polygons vectors, even if \code{snap} or \code{area} is \code{NULL}. To snap lines or points, you must set \code{snap} equal to a numeric value. The \code{area} correction is ignored for points and lines.
}

Issues can also arise due to:
\itemize{
\item \strong{Data table-vector mismatching}: If your vector has a data table ("attribute table") associated with it, errors can occur if there are more/fewer geometries (or multi-geometries) per row in the table. If you do not really need the data table to do your analysis, you can remove it (and thus obviate this error) using \code{dropTable = TRUE}.
\item \strong{Dissolving or aggregating "invalid" geometries}: Using the \code{resolve} argument, you can create a topologically valid vector by either coercing all overlapping portions of polygons into their own geometries (\code{resolve = "disaggregate"}), or by coercing them into a single, combined geometry (\code{resolve = "aggregate"}). Aggregation/disaggregation will be implemented after loading the vector into \strong{GRASS} using the settings given by \code{snap} and \code{area}. Aggregation/disaggregation will cause any associated data table to be dropped (it forces \code{dropTable} to be \code{TRUE}). The default action is to do neither aggregation nor disaggregation (\code{resolve = NA}). You can also do this outside \strong{fasterRaster} using \code{\link[terra:aggregate]{terra::aggregate()}} or \code{\link[terra:disaggregate]{terra::disagg()}}.
}

If none of these fixes work, you can try:
\itemize{
\item \strong{Correction outside of \emph{fasterRaster}}: Before you convert the vector into \strong{fasterRaster}'s \code{GVector} format, you can also try using the \code{\link[terra:is.valid]{terra::makeValid()}} or \code{\link[sf:valid]{sf::st_make_valid()}} tools to fix issues, then use \code{fast()}. You can also use \code{\link[terra:aggregate]{terra::aggregate()}} or \code{\link[terra:disaggregate]{terra::disagg()}} to combine/split problematic geometries.
\item \strong{Post-load correction}: If you do get a vector loaded into \code{GVector} format, you can also use a set of \strong{fasterRaster} vector-manipulation \link[=breakPolys]{tools} or \code{\link[=fillHoles]{fillHoles()}} to fix issues.
}
}
\examples{
if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Example data
madElev <- fastData("madElev") # integer raster
madCover <- fastData("madCover") # categorical raster
madCoast4 <- fastData("madCoast4") # polygons vector
madRivers <- fastData("madRivers") # lines vector
madDypsis <- fastData("madDypsis") # points vector

### Create GRasters from SpatRasters

# Create an integer raster:
elev <- fast(madElev)
elev

# Create a categorical raster:
cover <- fast(madCover)
madCover
levels(madCover) # category levels

# Create a GRaster from a file on disk:
rastFile <- system.file("extdata", "madForest2000.tif", package = "fasterRaster")
forest2000 <- fast(rastFile)
forest2000

# Create a 1's raster that spans the world:
ones <- fast(rastOrVect = "raster", crs = "epsg:4326")
ones

### Create GVectors

# Create a GVector from an sf vector:
coast4 <- fast(madCoast4)
coast4

# Create a GVector from a SpatVector:
madRivers <- vect(madRivers)
class(madRivers)
rivers <- fast(madRivers)
rivers

# Create a GVector from a vector on disk:
vectFile <- system.file("extdata/shapes", "madCoast.shp",
   package = "fasterRaster")
coast0 <- fast(vectFile)
coast0

# Import only Dypsis occurrences in a restricted area:
ant <- coast4[coast4$NAME_4 == "Antanambe"]
dypsisRestrict <- fast(madDypsis, extent = ant)
dypsis <- fast(madDypsis)

plot(coast4)
plot(ant, col = "gray80", add = TRUE)
plot(dypsis, add = TRUE)
plot(dypsisRestrict, col = "red", add = TRUE)

# Create a generic GVector that spans the world:
wallToWall <- fast(rastOrVect = "vector", crs = "epsg:4326") # WGS84
wallToWall

# Create a GVector from a numeric vector
pts <- c(-90.2, 38.6, -122.3, 37.9)
pts <- fast(pts, crs = "epsg:4326") # WGS84

# Create a GVector from a matrix (can also use data.frame or data.table):
mat <- matrix(c(-90.2, 38.6, -122.3, 37.9), ncol = 2, byrow = TRUE)
mat <- fast(mat, crs = "epsg:4326", keepgeom = TRUE) # WGS84

}
}
\seealso{
\code{\link[rgrass]{read_RAST}} and \code{\link[rgrass]{read_VECT}}, \link[=breakPolys]{vector cleaning}, \code{\link[=fillHoles]{fillHoles()}}, plus \strong{GRASS} modules \code{v.in.ogr} (see \code{grassHelp("v.in.ogr")}) and \code{r.import} (\code{grassHelp("r.import")})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is.lonlat.r
\name{is.lonlat,character-method}
\alias{is.lonlat,character-method}
\alias{is.lonlat}
\alias{is.lonlat,GLocation-method}
\alias{is.lonlat,sf-method}
\title{Test if a coordinate reference system is unprojected}
\usage{
\S4method{is.lonlat}{character}(x)

\S4method{is.lonlat}{GLocation}(x)

\S4method{is.lonlat}{sf}(x)
}
\arguments{
\item{x}{A WKT coordinate reference string or an object from which on can be obtained (e.g., a \code{GRaster}, \code{GVector}, \code{GRegion}, \code{GLocation}, \code{SpatRaster}, \code{SpatVector}, or \code{sf} object).}
}
\value{
Logical (\code{TRUE} if unprojected, \code{FALSE} otherwise).
}
\description{
\code{is.lonlat()} attempts to determine if a coordinate reference system is unprojected (e.g., WGS84, NAD83, NAD27, etc.). For \code{GRaster}s and \code{GVector}s, the function should always be correct. For WKT character strings and \code{sf} vectors, it does this by looking for the "CONVERSION[" tag in the WKT string (or the object's WKT string), and if it finds one, returns \code{FALSE}. This may not be truthful in all cases.
}
\seealso{
\code{\link[terra:is.lonlat]{terra::is.lonlat()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/denoise_noise.r
\name{denoise,GRaster-method}
\alias{denoise,GRaster-method}
\alias{denoise}
\alias{noise,GRaster-method}
\alias{noise}
\title{Remove or retain "noise" in a raster using PCA}
\usage{
\S4method{denoise}{GRaster}(x, scale = FALSE, percent = 80)

\S4method{noise}{GRaster}(x, scale = FALSE, percent = 80)
}
\arguments{
\item{x}{A \code{GRaster} with two or more layers.}

\item{scale}{Logical: If \code{TRUE}, input layers will be rescaled by dividing each layer by its overall population standard deviation. Note that rasters will always be centered (have their mean subtracted from values). Centering and scaling is recommended when rasters values are in different units. The default is \code{FALSE} (do not scale).}

\item{percent}{Numeric integer or integer in the range 50 to 99 (default is 80): Minimum total variation explained in the retained PC axes. Higher values will cause \code{denoise()} to remove less noise, and \code{noise()} to return less noise. If this value to too low to retain even one axis, the function will fail with a message to that effect.}
}
\value{
A multi-layer \code{GRaster} with one layer per input.
}
\description{
\code{denoise()} applies a principal component analysis (PCA) to layers of a \code{GRaster}, then uses the PCA to predict values back to a raster. This retains only coarse-scale trends, thereby removing "noise" (locally extreme values that fall far from a PC axis).

\code{noise()} does the opposite by first constructing the PCA, predicting values back to the raster, then subtracting these values from the original, removing coarse-scale trends and thereby leaving "noise".
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Climate raster:
madChelsa <- fastData("madChelsa")

# Convert a SpatRaster to a GRaster:
chelsa <- fast(madChelsa)

### Denoise:
quiet <- denoise(chelsa, scale = TRUE)

compare1 <- c(chelsa[["bio1"]], quiet[["bio1"]])
plot(compare1)

compare2 <- c(chelsa[["bio7"]], quiet[["bio7"]])
plot(compare2)

### Noise:
loud <- noise(chelsa, scale = TRUE)

compare1 <- c(chelsa[["bio1"]], loud[["bio1"]])
plot(compare1)

compare2 <- c(chelsa[["bio7"]], loud[["bio7"]])
plot(compare2)

}
}
\seealso{
\code{\link[=princomp]{princomp()}}, \code{\link[stats:prcomp]{stats::prcomp()}}, \strong{GRASS} manual page for tool \code{i.pca} (see \code{grassHelp("i.pca")})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vegIndex.r
\name{vegIndex,GRaster-method}
\alias{vegIndex,GRaster-method}
\alias{vegIndex}
\title{Vegetation indices from surface reflectance}
\usage{
\S4method{vegIndex}{GRaster}(
  x,
  index = "NDVI",
  r = NULL,
  g = NULL,
  b = NULL,
  nir = NULL,
  b5 = NULL,
  b7 = NULL,
  soilSlope = NULL,
  soilIntercept = NULL,
  soilNR = 0.08,
  bits = NULL
)
}
\arguments{
\item{x}{A \code{GRaster} with one layer per required band. Values should be between 0 and 1.}

\item{index}{Character or character vector: The vegetation index or indices to calculate. You can find a list of available indices using \link[=fastData]{fastData("vegIndices")} (also see \link{vegIndices}). The first column, "\code{index}" provides the name of the index, and these are the values that this argument will accept (e.g., "NDVI", "EVI2"). Partial matching is used, and case is ignored. You can also use these shortcuts:
\itemize{
\item \code{"*"}: Calculate \emph{all} indices
\item \code{"RNIR"}: Calculate all indices that use R and NIR channels (but not other channels).
\item \code{"NotSoil"}: Calculate all indices that use any channels but do not require \code{soilSlope} or \code{soilIntercept}.
}

\emph{Note}: A near-comprehensive table of indices can be found on the \href{https://www.indexdatabase.de}{Index Database: A Database for Remote Sensing Indices}.}

\item{r, g, b, nir}{Numeric or character: Index or \code{\link[=names]{names()}} of the layers in \code{x} that represent the red, green, blue, and near infrared channels. Values must be in the range from 0 to 1 or integers.}

\item{b5, b7}{Numeric or character: Index of names of the layers representing bands 5 and 7. These are used only for GVI and PVI. Values must be in the range from 0 to 1 or integers.}

\item{soilSlope, soilIntercept, soilNR}{Numeric: Values of the soil slope, intercept, and soil noise reduction factor (0.08, by default). Used only for calculation of MSAVI.}

\item{bits}{Either \code{NULL} (default) or numeric integer or integer with a value of 7, 8, 10, or 16: If the rasters are represented by integers (so do not fall in the range of 0 to 1), then the number of bits can be supplied using \code{bits}. If this is the case, then they will range from 0 to 2^\code{n}, where \code{n} is 7, 8, 10, or 16. If bit rasters are supplied, they must be of \code{\link[=datatype]{datatype()}} "integer". If raster values are in the range from 0 to 1, then \code{bits} should be \code{NULL} (default).}
}
\value{
A \code{GRaster}.
}
\description{
This function calculates one of many types of vegetation indices from a raster with four bands representing blue (B), green (G), red (R), and near infrared (NIR), plus possibly channels 5 and 7. The function requires rasters that represent surface reflectance, so should have values that fall in the range 0 to 1, unless they are digital number rasters (e.g., integers in the range 0 to 255). If digital number format is used, then the \code{bits} argument should be defined.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Elevation raster, rivers vector
madLANDSAT <- fastData("madLANDSAT")

# Convert a SpatRaster to a GRaster:
landsat <- fast(madLANDSAT)

# See available vegetation indices:
vegIndices

# Normalized Difference Vegetation Index and Enhanced Vegetation Index:
indices <- c("ndvi", "evi")
vi <- vegIndex(landsat, index = indices, r = 1, b = 3, nir = 4, bits = 8)
plot(vi)

# All indices using R and NIR:
rnir <- vegIndex(landsat, index = "rnir", r = 1, nir = 4, bits = 8)

# Note: Some values are highly skewed
plot(rnir)

}
}
\seealso{
\strong{GRASS} manual page for tool \code{i.vi} (see \code{grassHelp("i.vi")})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/writeRaster.r
\name{writeRaster,GRaster,character-method}
\alias{writeRaster,GRaster,character-method}
\alias{writeRaster}
\alias{writeRaster,missing,missing-method}
\title{Save a GRaster to disk}
\usage{
\S4method{writeRaster}{GRaster,character}(
  x,
  filename,
  overwrite = FALSE,
  datatype = NULL,
  byLayer = FALSE,
  names = TRUE,
  levelsExt = NULL,
  compress = "LZW",
  warn = TRUE,
  ...
)

\S4method{writeRaster}{missing,missing}(x, filename)
}
\arguments{
\item{x}{A \code{GRaster} or missing: If missing, a table of supported file types is reported.}

\item{filename}{Character: Path and file name.}

\item{overwrite}{Logical: If \code{FALSE} (default), do not save over existing file(s).}

\item{datatype}{\code{NULL} (default) or character: The datatype of the values stored in non-ASCII rasters. If \code{NULL}, this will be ascertained from the raster, and the function usually does a good job at it. However, you can force it manually, but note that in some cases, trying to save a \code{GRaster} using an inappropriate \code{datatype} for its values can result in an error or in the function exiting without an error but also without having written the raster to disk. The argument can take any of those shown below under the first four columns, but whatever is used, it will be converted to the \strong{GDAL} version.\tabular{lllll}{
   \strong{fasterRaster} \tab \strong{terra} \tab \strong{GRASS} \tab \strong{GDAL} \tab \strong{Values} \cr
   \code{integer} \tab \code{INT1U} \tab \code{CELL} \tab \code{Byte} \tab Integer values from 0 to 255 \cr
   \code{integer} \tab \code{INT2U} \tab \code{CELL} \tab \code{UInt16} \tab Integer values from 0 to 65,534 \cr
   \code{integer} \tab \code{INT2S} \tab \code{CELL} \tab \code{Int16} \tab Integer values from -32,767 to -32,767 \cr
   \code{integer} \tab \code{INT4S} \tab \code{CELL} \tab \code{Int32} \tab Integer values from -2,147,483,647 to 2,147,483,647 \cr
   \code{float} \tab \code{FLT4S} \tab \code{FCELL} \tab \code{Float32} \tab Values from -3.4E+38 to 3.4E+38, including decimal values \cr
   \code{double} \tab \code{FLT8S} \tab \code{DCELL} \tab \code{Float64} \tab Values from -1.79E+308 to 1.79E+308, including decimal values \cr
   \code{factor} \tab \code{INT}* \tab \code{CELL} \tab \verb{INT*} \tab Integer values corresponding to categories \cr
}


\code{*} Depends on the integers (signed/unsigned, range of values). Categorical rasters will have an associated file saved with them that has category values and labels. The file name will be the same as the raster's file name, but end with the extension given by \code{levelsExt} (\code{.csv} by default).}

\item{byLayer}{Logical: If \code{FALSE} (default), multi-layer rasters will be saved in one file. If \code{TRUE}, the each layer will be saved in a separate file. The filename from \code{filename} will be amended so that it ends with \verb{_<name>} (then the file extension), where \verb{<name>} is give by \code{\link[=names]{names()}}. Note that if any characters in raster names will not work in a file name, then the function will fail (e.g., a backslash or question mark).}

\item{names}{Logical: If \code{TRUE} (default), save a file with raster layer names. The file will have the same name as the raster file but end with "\verb{_names.csv}". Currently, the \code{\link[=names]{names()}} attribute of rasters cannot be saved in the raster, which can create confusion when multi-layered rasters are saved. Turning on this option will save the ancillary file with layer names. If it exists, this file will be read by \code{\link[=fast]{fast()}} so layer names are assigned when the raster is read by that function. The absence of a "names" file will not create any issues with this function or \code{\link[=fast]{fast()}}, other than not having the metadata on layer names.}

\item{levelsExt}{Character, logical, or \code{NULL} (default): Name of the file extension for the "levels" file that accompanies a categorical \code{GRaster}. When saving categorical rasters, the raster file is accompanied with a "levels" file that contain information on the levels of the raster. This file is the same as \code{filename}, except it has a different extension. Valid values depend on how many raster layers are saved at a time (case is ignored):
\itemize{
\item DefaultOne raster layer: \code{".csv"}
\item Two or more layers, with at least one categorical raster: \code{".rds"}, \code{".rda"}, \code{".rdat"}, \code{".rdata"}
\item Any: \code{NULL} or \code{TRUE} automatically selects either \code{".csv"} (one raster layer) or \verb{".rds} (two or more)
\item Any: \code{FALSE} disables saving of a levels file.
}}

\item{compress}{Character: Type of compression to use for GeoTIFF files:
\itemize{
\item \code{"LZW"} (default)
\item \code{"DEFLATE"}
\item \code{"PACKBITS"}
\item \code{"LZMA"}
\item \code{NULL}: No compression is used, but the file can still be reduced in size by using zip, gzip, or other compressions.
}}

\item{warn}{Logical: If \code{TRUE} (default), display a warning if the \code{datatype} argument does not match the value given by \code{datatype(x, "GDAL")}, or if the \code{fileExt} argument will not work with the given raster and so has been automatically changed.}

\item{...}{Additional arguments. These can include:
\itemize{
\item \code{bigTiff}: Logical: If \code{TRUE}, and the file format is a GeoTIFF and would be larger than 4 GB (regardless of compression), then the file will be saved in BIGTIFF format.
\item \code{format}: Character, indicating file format. This is usually ascertained from the file extension, but in case this fails, it can be stated explicitly. When using other formats, you may have to specify the \code{createopts} argument, too (see help page for \strong{GRASS} tool \code{r.out.gdal}). Two common formats include:
\itemize{
\item \code{"GTiff"} (default): GeoTIFF \code{filename} ends in \code{.tif}.
\item \code{"ASC"}: ASCII \code{filename} ends in \code{.asc}
}
\item Additional arguments to send to \strong{GRASS} modules \code{r.out.gdal} and \code{r.out.ascii}.
\item \code{precision}: Numeric: For ASCII files, you may need to state the number of significant digits. 32-bit values have 7 digits and 64-bit values have 16. So in these cases the argument would be \code{precision=7} or \code{precision=16}.
}}
}
\value{
A \code{GRaster} (invisibly). A raster is also saved to disk.
}
\description{
This function saves a \code{GRaster} to disk directly from a \strong{GRASS} session. It is faster than using \code{\link[=rast]{rast()}}, then saving the output of that to disk (because \code{rast()} actually save the raster to disk, anyway).

The function will attempt to ascertain the file type to be ascertained from the file extension, but you can specify the format using the \code{format} argument (see entry for \code{...}). You can see a list of supported formats by simply using this function with no arguments, as in \code{writeRaster()}, or by consulting the online help page for the \strong{GRASS} tool \code{r.out.gdal} (see \code{grassHelp("r.out.gdal")}). Only the \code{GeoTIFF} file format is guaranteed to work for multi-layered rasters.

The function will attempt to optimize the \code{datatype} argument, but this can take a long time. You can speed this up by setting \code{datatype} manually. Note that if you are saving a "stack" of \code{GRaster}s with different \code{datatype}s, the one with the highest information density will be used (e.g., low-bit integer < high-bit integer < floating-point < double-floating point). This can make rasters with lower datatypes much larger on disk. In these cases, it make be best to save rasters with similar \code{datatype}s together.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")
madChelsa <- fastData("madChelsa")

### What raster formats can we attempt to write?
writeRaster()

### Save GRaster to disk (using temporary file)
elev <- fast(madElev)
filename <- tempfile(fileext = ".tif")
writeRaster(elev, filename)

# Load raster from disk
elev2 <- fast(filename)
elev2

### Save multi-layer GRaster to disk in one file (using temporary file)
chelsa <- fast(madChelsa)
filename <- tempfile(fileext = ".tif")
writeRaster(chelsa, filename)

# Load raster from disk
chelsa2 <- fast(filename)
chelsa2

### Save multi-layer GRaster to disk layer-by-layer (using temporary file)
chelsa <- fast(madChelsa)
filename <- tempfile(fileext = ".tif")
writeRaster(chelsa, filename, byLayer = TRUE)

# Load one of the rasters from disk
filename2 <- sub(filename, pattern = ".tif", replacement = "_bio1.tif")
chelsaBio1 <- fast(filename2)
chelsaBio1

}
}
\seealso{
\code{\link[terra:writeRaster]{terra::writeRaster()}}, \strong{GRASS} tool \code{r.out.gdal} (see \code{grassHelp("r.out.gdal")})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clump.r
\name{clump,GRaster-method}
\alias{clump,GRaster-method}
\alias{clump}
\title{Group adjacent cells with similar values}
\usage{
\S4method{clump}{GRaster}(x, minDiff = 0, minClumpSize = 1, diagonal = TRUE)
}
\arguments{
\item{x}{A \code{GRaster}.}

\item{minDiff}{Numeric in the range [0, 1): Minimum difference between cells in order for them to be assigned to the same clump. This is a proportion of the range across all cells. For example, if \code{minDiff} is set to 0.01, then the maximum difference between cells in a clump can be up to 1\% of the entire range across all cells. Small values can create large clumps. The default is 0, in which case values have to be exactly the same.}

\item{minClumpSize}{Numeric integer >= 1. Minimum number of cells in a clump. The default is 1.}

\item{diagonal}{Logical: If \code{TRUE} (default), then cells "connected" at corners will be included as part of the same clump.}
}
\value{
A \code{GRaster}.
}
\description{
\code{clump()} identifies groups of adjacent cells that have the same value or same approximate value, and assigns them a unique number, creating "clumps" of same- or similar-valued cells.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Elevation raster
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster:
elev <- fast(madElev)

# Find clumps based on exact values. This will appear as a gradient because
# most cells are assigned to a group of 1 cell.
exact <- clump(elev)
plot(exact)

# Clump based on approximate values:
approx <- clump(elev, minDiff = 0.0075)
plot(approx)

# Clump based on approximate values with minimum clump size:
approx20 <- clump(elev, minDiff = 0.005, minClumpSize = 20)
plot(approx20)

approx
approx20

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.r
\name{.plot}
\alias{.plot}
\title{Plot using a G-object's \code{\link[=sources]{sources()}} name}
\usage{
.plot(x, ...)
}
\arguments{
\item{x}{The \code{\link[=sources]{sources()}} name of a \code{GRaster} or \code{GVector}}

\item{...}{Other arguments to pass to \code{\link[=plot]{plot()}}.}
}
\description{
Plot using a G-object's \code{\link[=sources]{sources()}} name
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/locationFind.r
\name{.locationFind,missing-method}
\alias{.locationFind,missing-method}
\alias{.locationFind,GLocation-method}
\alias{.locationFind,SpatRaster-method}
\alias{.locationFind,SpatVector-method}
\alias{.locationFind,sf-method}
\alias{.locationFind,character-method}
\title{Match CRS of a GSpatial object and an existing "GRASS" location}
\usage{
\S4method{.locationFind}{missing}(x, return = "name")

\S4method{.locationFind}{GLocation}(x, return = "name", match = "name")

\S4method{.locationFind}{SpatRaster}(x, return = "name", match = "name")

\S4method{.locationFind}{SpatVector}(x, return = "name", match = "name")

\S4method{.locationFind}{sf}(x, return = "name", match = "name")

\S4method{.locationFind}{character}(x, return = "name", match = "name")
}
\arguments{
\item{x}{Either:
\itemize{
\item Missing: Returns names and coordinate reference system strings of all "locations".
\item A character representing a coordinate reference system in WKT format
\item A \code{SpatRaster}, \code{SpatVector}, or \code{sf} vector with a coordinate reference system
\item A \code{GSpatial} object (usually a \code{GRaster} or \code{GVector})
}}

\item{return}{Either:
\itemize{
\item \code{"name"} (default): Returns the name of the "location" with a coordinate reference system the same as \code{x}.
\item \code{"index"}: Returns the index of this "location" in \code{.fasterRaster$locations} of the \code{.fasterRaster} environment.
\item \code{"crs"}: Returns the coordinate reference system of this "project/location".
}}

\item{match}{Character: Method used to find the location. If \code{match} is "\code{name}"" (default), then the name of the location is used. If \code{match} is "\code{crs}", then the coordinate reference system of each location is checked for a match.}
}
\value{
Character, integer, or \code{NULL} (if no match is found).
}
\description{
The function searches the set of available \strong{GRASS} "projects" (previously known as "locations") for one that has a coordinate reference system matching a \code{GSpatial} object. If none are found, or if no connection with \strong{GRASS} has yet been made, then it returns \code{NULL}. Otherwise, it returns either the index or the name of the matching location.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rSpatialDepRast.r
\name{rSpatialDepRast,GRaster-method}
\alias{rSpatialDepRast,GRaster-method}
\alias{rSpatialDepRast}
\title{Create a random raster with or without spatial dependence}
\usage{
\S4method{rSpatialDepRast}{GRaster}(
  x,
  n = 1,
  mu = 0,
  sigma = 1,
  dist = 0,
  exponent = 1,
  delay = 0,
  seed = NULL
)
}
\arguments{
\item{x}{A \code{GRaster}: The output will have the same extent and dimensions as this raster.}

\item{n}{An integer: Number of rasters to generate.}

\item{mu, sigma}{Numeric: Mean and sample standard deviation of output. If creating more than one raster, you can provide one value per raster. If there are fewer, they will be recycled.}

\item{dist}{Numeric: Maximum distance of spatial autocorrelation (in map units--typically meters). Default is 0 (no spatial autocorrelation). If creating more than one raster, you can provide one value per raster. If there are fewer, values will be recycled.}

\item{exponent}{Numeric > 0: Distance decay exponent. If creating more than one raster, you can provide one value per raster. If there are fewer, values will be recycled.}

\item{delay}{Numeric >= 0: Values >0 force the distance decay of similarity to remain constant up to this distance. Beyond this distance, the decay exponent takes effect. Default is 0. If creating more than one raster, you can provide one value per raster. If there are fewer, values will be recycled.}

\item{seed}{Numeric integer or \code{NULL}: Random seed. If \code{NULL}, then a random seed is used for each raster. If provided, there should be one seed value per raster.}
}
\value{
A \code{GRaster}.
}
\description{
\code{rSpatialDepRast()} creates a raster with random values in cells. Across the raster, values are approximately normally distributed, though a raster with a "true" normal distribution can be made with \code{\link[=rNormRast]{rNormRast()}}. Spatial dependence can be introduced, though all together the values will still be approximately normally distributed.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Elevation raster
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster:
elev <- fast(madElev)

### Create a raster with values drawn from a uniform distribution:
unif <- rUnifRast(elev)
plot(unif)

### Create a raster with values drawn from a normal distribution:
norms <- rNormRast(elev, n = 2, mu = c(5, 10), sigma = c(2, 1))
plot(norms)
hist(norms, bins = 100)

# Create a raster with random, seemingly normally-distributed values:
rand <- rSpatialDepRast(elev, dist = 1000)
plot(rand)

# Values appear normal on first inspection:
hist(rand)

# ... but actually are patterned:
hist(rand, bins = 100)

# Create a fractal raster:
fractal <- fractalRast(elev, n = 2, dimension = c(2.1, 2.8))
plot(fractal)
hist(fractal)

### Random walker rasters

# One random walker
walk <- rWalkRast(elev)
plot(walk)

# Random walker with self-avoidance:
walkAvoid <- rWalkRast(elev, steps = 1000, avoid = TRUE, seed = 1)
plot(walkAvoid)

# 10 random walkers:
walk10 <- rWalkRast(elev, n = 10)
plot(walk10)

# 10 random walkers starting in same place:
walkSame10 <- rWalkRast(elev, n = 10, sameStart = TRUE)
plot(walkSame10)


}
}
\seealso{
\code{\link[=rNormRast]{rNormRast()}}, \code{\link[=fractalRast]{fractalRast()}}, \code{\link[=rUnifRast]{rUnifRast()}}, \code{\link[=rWalkRast]{rWalkRast()}}, \strong{GRASS} manual page for tool \code{r.random.surface} (see \code{grassHelp("r.random.surface")})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/layerCor.r
\name{layerCor,GRaster-method}
\alias{layerCor,GRaster-method}
\alias{layerCor}
\title{Correlation between GRasters}
\usage{
\S4method{layerCor}{GRaster}(x, fun = "cor")
}
\arguments{
\item{x}{A \code{GRaster} with two or more layers.}

\item{fun}{Character: Name of the statistic to calculate; either \code{"cor"} (default) or \code{"cov"}.}
}
\value{
A numeric \code{matrix}.
}
\description{
This function returns a correlation or covariance matrix between two or more \code{GRaster} layers. This function returns the sample correlation and covariance (i.e., the denominator is n - 1).
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data
madChelsa <- fastData("madChelsa")

# Convert a SpatRaster to a GRaster:
chelsa <- fast(madChelsa)

# Correlation
layerCor(chelsa, "cor")

# Covariance
layerCor(chelsa, "cov")

}
}
\seealso{
\code{\link[terra:layerCor]{terra::layerCor()}}, \code{\link[stats:cor]{stats::cor()}}, \code{\link[stats:cor]{stats::cov()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hillshade.r
\name{hillshade,GRaster-method}
\alias{hillshade,GRaster-method}
\alias{hillshade}
\title{Hillshading}
\usage{
\S4method{hillshade}{GRaster}(x, angle = 45, direction = 0, zscale = 1)
}
\arguments{
\item{x}{A \code{GRaster} (typically representing elevation).}

\item{angle}{Numeric: The altitude of the sun above the horizon in degrees. Valid values are in the range [0, 90], and the default value is 45 (half way from the horizon to overhead).}

\item{direction}{The direction (azimuth) in which the sun is shining in degrees. Valid values are in the range 0 to 360. The default is 0, meaning the sun is at due south (180 degrees) and shining due north (0 degrees). Note that in this function, 0 corresponds to north and 180 to south, but in the \strong{GRASS} tool \code{r.relief}, "east orientation" is used (0 is east, 90 is north, etc.).}

\item{zscale}{Numeric: Value by which to exaggerate terrain. The default is 1.  Numbers greater than this will increase apparent relief, and less than this (even negative) will diminish it.}
}
\value{
A \code{GRaster}.
}
\description{
Hillshade rasters are often used for display purposes because they make topographical relief look "real" to the eye.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster
elev <- fast(madElev)

# Calculate all topographic metrics
topos <- terrain(elev, v = "*")
topos

plot(topos) # NB Aspect has values of NA when it cannot be defined

# Calculate a hillshade raster
hs <- hillshade(elev)
plot(hs)

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sampleRast.r
\name{sampleRast,GRaster-method}
\alias{sampleRast,GRaster-method}
\alias{sampleRast}
\title{Randomly sample cells from a GRaster}
\usage{
\S4method{sampleRast}{GRaster}(
  x,
  size,
  prop = FALSE,
  maskvalues = NA,
  updatevalue = NULL,
  test = FALSE,
  seed = NULL
)
}
\arguments{
\item{x}{A \code{GRaster}.}

\item{size}{Numeric: Number of cells or proportion of cells to select.}

\item{prop}{Logical: If \code{TRUE}, the value of \code{size} will be interpreted as a proportion of cells. The default is \code{FALSE} (\code{size} is interpreted as the number of cells to select).}

\item{maskvalues}{Numeric vector, including \code{NA}, or \code{NULL} (default): Values in the raster to select from. All others will be ignored. If this is \code{NULL}, then only non-\code{NA} cells will be selected for retention.}

\item{updatevalue}{Numeric or \code{NULL} (default): Value to assign to masked cells. If \code{NULL}, then the values in the input raster are retained.}

\item{test}{Logical: If \code{TRUE}, and \code{size} is greater than the number of non-\code{NA} cells in \code{x}, then fail. Testing this can take a long time for large rasters. The default is \code{FALSE}.}

\item{seed}{\code{NULL} (default) or numeric: If \code{NULL}, then a random seed will be generated for the random number generator. Otherwise a seed can be provided.}
}
\value{
A \code{GRaster}.
}
\description{
\code{sampleRast()} randomly samples cells from non-\code{NA} cells of a raster. The output will be a raster with selected non-\code{NA} cells, and all other cells set to \code{NA}. To generate random points, see \code{\link[=spatSample]{spatSample()}}.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Example data
madElev <- fastData("madElev") # raster

# Convert to GRasters and GVectors
elev <- fast(madElev)

### spatSample()

# Random points as data.frame or data.table:
randVals <- spatSample(elev, size = 20, values = TRUE)
randVals

# Random points as a points GVector:
randPoints <- spatSample(elev, size = 20, as.points = TRUE)
randPoints
plot(elev)
plot(randPoints, add = TRUE)

# Random points in a select area:
madCoast <- fastData("madCoast4") # vector
coast <- fast(madCoast)
ant <- coast[coast$NAME_4 == "Antanambe"] # subset

restrictedPoints <- spatSample(elev, size = 20, as.points = TRUE,
   strata = ant)

plot(elev)
plot(ant, add = TRUE)
plot(restrictedPoints, add = TRUE) # note 20 points for entire geometry

# Random points, one set per subgeometry:
stratifiedPoints <- spatSample(elev, size = 20, as.points = TRUE,
   strata = ant, byStratum = TRUE)

plot(elev)
plot(ant, add = TRUE)
plot(stratifiedPoints, pch = 21, bg = "red", add = TRUE) # note 20 points per subgeometry

# Random categories:
madCover <- fastData("madCover") # raster
cover <- fast(madCover)

randCover <- spatSample(cover, size = 20, values = TRUE,
     cat = TRUE, xy = TRUE)
randCover

### sampleRast()

# Random cells in non-NA cells:
rand <- sampleRast(elev, 10000)
plot(rand)
nonnacell(rand)

# Use custom values for the mask:
randCustomMask <- sampleRast(elev, 10000, maskvalues = 1:20)
plot(randCustomMask)

# Force selected values to a custom value:
randCustomUpdate <- sampleRast(elev, 10000, updatevalue = 7)
plot(randCustomUpdate)

# Custom values for mask and set selected cells to custom value:
randAll <- sampleRast(elev, 10000, maskvalues = 1:20, updatevalue = 7)
plot(randAll)

}
}
\seealso{
\code{\link[=spatSample]{spatSample()}}; \code{\link[terra:sample]{terra::spatSample()}}, tool \code{r.random} in \strong{GRASS}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/05_GRaster_functions_by_layer.r
\name{is.na,GRaster-method}
\alias{is.na,GRaster-method}
\alias{is.na}
\alias{not.na,GRaster-method}
\alias{not.na}
\alias{abs,GRaster-method}
\alias{abs}
\alias{sin,GRaster-method}
\alias{sin}
\alias{cos,GRaster-method}
\alias{cos}
\alias{tan,GRaster-method}
\alias{tan}
\alias{asin,GRaster-method}
\alias{asin}
\alias{acos,GRaster-method}
\alias{acos}
\alias{atan,GRaster-method}
\alias{atan}
\alias{atan2,GRaster,GRaster-method}
\alias{atan2}
\alias{exp,GRaster-method}
\alias{exp}
\alias{log1p,GRaster-method}
\alias{log1p}
\alias{log10p,GRaster-method}
\alias{log10p}
\alias{log,GRaster-method}
\alias{log}
\alias{ln,GRaster-method}
\alias{ln}
\alias{log2,GRaster-method}
\alias{log2}
\alias{log10,GRaster-method}
\alias{log10}
\alias{sqrt,GRaster-method}
\alias{sqrt}
\alias{round,GRaster-method}
\alias{round}
\alias{floor,GRaster-method}
\alias{floor}
\alias{ceiling,GRaster-method}
\alias{ceiling}
\alias{trunc,GRaster-method}
\alias{trunc}
\title{Mathematical operations on each layer of a GRasters}
\usage{
\S4method{is.na}{GRaster}(x)

\S4method{not.na}{GRaster}(x, falseNA = FALSE)

\S4method{abs}{GRaster}(x)

\S4method{sin}{GRaster}(x)

\S4method{cos}{GRaster}(x)

\S4method{tan}{GRaster}(x)

\S4method{asin}{GRaster}(x)

\S4method{acos}{GRaster}(x)

\S4method{atan}{GRaster}(x)

\S4method{atan2}{GRaster,GRaster}(y, x)

\S4method{exp}{GRaster}(x)

\S4method{log1p}{GRaster}(x)

\S4method{log10p}{GRaster}(x)

\S4method{log}{GRaster}(x, base = exp(1))

\S4method{ln}{GRaster}(x)

\S4method{log2}{GRaster}(x)

\S4method{log10}{GRaster}(x)

\S4method{sqrt}{GRaster}(x)

\S4method{round}{GRaster}(x, digits = 0)

\S4method{floor}{GRaster}(x)

\S4method{ceiling}{GRaster}(x)

\S4method{trunc}{GRaster}(x)
}
\arguments{
\item{x, y}{\code{GRaster}s.}

\item{falseNA}{Logical (function \code{not.na()}): If \code{FALSE} (default), non-\code{NA} cells will be converted to 1, and \code{NA} cells to 0. If \code{TRUE}, non-\code{NA} cells will be converted to  and \code{NA} cells will stay as \code{NA}.}

\item{base}{Numeric: Base of the logarithm.}

\item{digits}{Numeric: Number of digits to round to. If negative, then rounding is to the nearest positive power of 10. For example, if \code{digits = -2}, then the \code{GRaster} values are rounded to the nearest 100.}
}
\value{
A \code{GRaster}.
}
\description{
You can apply mathematical functions to each layer of a \code{GRaster}. The output is a \code{GRaster} with the same number or layers as the input. Available functions include:\cr
\itemize{
\item \code{NA}s:
\itemize{
\item \code{is.na()}
\item \code{not.na()}
}
\item Absolute value: \code{abs()}
\item Trigonometric functions (assumes values are in radians):
\itemize{
\item \code{cos()}
\item \code{sin()}
\item \code{tan()}
\item \code{acos()}
\item \code{asin()}
\item \code{atan()}
\item \code{atan2()}
}
\item Exponential and logarithmic functions:
\itemize{
\item \code{exp()}
\item \code{log()} (natural log)
\item \code{ln()} (also natural log)
\item \code{log2()} (log, base 2)
\item \code{log10()} (log, base 10)
\item \code{log1p()} (same as \code{log(x + 1)})
\item \code{log10p()} (same as \code{log(x + 1, base = 10)})
}
\item Power functions:
\itemize{
\item \code{sqrt()}
\item \code{x^y}
}
\item Rounding:
\itemize{
\item \code{round()}
\item \code{floor()} (round down)
\item \code{ceiling()} (round up)
\item \code{trunc()} (remove decimal portion)
}
}
}
\examples{
if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Example data
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster
elev <- fast(madElev)
elevs <- c(elev, elev, log10(elev) - 1, sqrt(elev))
names(elevs) <- c("elev1", "elev2", "log_elev", "sqrt_elev")

elev
elevs

# do some math
elev + 100
elev - 100
elev * 100
elev / 100
elev ^ 2
elev \%/\% 100 # divide then round down
elev \%\% 100 # modulus

100 + elev
100 \%/\% elev
100 \%\% elev

elevs + 100
100 + elevs

# math with logicals
elev + TRUE
elev - TRUE
elev * TRUE
elev / TRUE
elev ^ TRUE
elev \%/\% TRUE # divide then round down
elev \%\% TRUE # modulus

elevs + TRUE
TRUE + elevs

# Raster interacting with raster(s):
elev + elev
elev - elev
elev * elev
elev / elev
elev ^ log(elev)
elev \%/\% sqrt(elev) # divide then round down
elev \%\% sqrt(elev) # modulus

elevs + elev
elev * elevs

# sign
abs(-1 * elev)
abs(elevs)

# powers
sqrt(elevs)

# trigonometry
sin(elev)
cos(elev)
tan(elev)

asin(elev)
acos(elev)
atan(elev)

atan(elevs)
atan2(elev, elev^1.2)
atan2(elevs, elev^1.2)
atan2(elev, elevs^1.2)
atan2(elevs, elevs^1.2)

# logarithms
exp(elev)
log(elev)
ln(elev)
log2(elev)
log1p(elev)
log10(elev)
log10p(elev)
log(elev, 3)

log(elevs)

# rounding
round(elev + 0.5)
floor(elev + 0.5)
ceiling(elev + 0.5)
trunc(elev + 0.5)

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vAttachDatabase.r
\name{.vAttachDatabase}
\alias{.vAttachDatabase}
\title{Add a database table to a GRASS attribute table}
\usage{
.vAttachDatabase(x, table = NULL, replace = TRUE, cats = NULL)
}
\arguments{
\item{x}{A \code{GVector} or the name of a vector in \strong{GRASS}.}

\item{table}{Either \code{NULL} (default), or a \code{data.frame} or \code{data.table}, or a numeric or integer vector:
\itemize{
\item If \code{NULL}, then a bare minimal table will be created with a column named \code{cat}, holding sequential integer values.
\item If a \code{data.frame} or \code{data.table} and no column is named \code{cat}, one will be created with sequential integer values. If the table does have a column named \code{cat}, then it should have integer (not just numeric) values.
\item If a \code{vector}, then these are coerced to type \code{integer} and used to define the \code{cat} column.
}

There should be one row/value per geometry in \code{x}.}

\item{replace}{Logical: If \code{TRUE}, replace the existing database connection.}

\item{cats}{Either \code{NULL} (default), or an integer vector: This is provided as a means to save time by passing \code{cats} to this function if it has already been generated by a calling function.}
}
\value{
Invisibly returns the \code{\link[=sources]{sources()}} name of a vector in \strong{GRASS}.
}
\description{
\code{.vAttachDatabase()} adds a table to a \strong{GRASS} vector. This table is meant to be "invisible" to most users--they should use interact with attribute tables using the \code{GVector} slot \verb{@table}. Some functions require tables (e.g., \code{\link[=extract]{extract()}} and \code{\link[=spatSample]{spatSample()}}). \strong{This function is mostly of use to developers.}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/crs.r
\name{crs,missing-method}
\alias{crs,missing-method}
\alias{crs}
\alias{crs,GLocation-method}
\alias{st_crs,missing-method}
\alias{st_crs}
\alias{st_crs,GLocation-method}
\alias{coordRef,missing-method}
\alias{coordRef}
\alias{coordRef,GRaster-method}
\alias{coordRef,GVector-method}
\title{Coordinate reference system of a GRaster or GVector}
\usage{
\S4method{crs}{missing}(x)

\S4method{crs}{GLocation}(x)

\S4method{st_crs}{missing}(x, ...)

\S4method{st_crs}{GLocation}(x, ...)

st_crs(x, ...)

\S4method{coordRef}{missing}(x)

\S4method{coordRef}{GRaster}(x)

\S4method{coordRef}{GVector}(x)
}
\arguments{
\item{x}{An object that inherits from a \code{GLocation} (i.e., a \code{GRaster} or \code{GVector}) or missing. If missing, the coordinate reference system of the currently active \strong{GRASS} "project/location" is reported.}

\item{...}{Other arguments (generally unused).}
}
\value{
Function \code{crs()} returns a \code{character} object, \code{st_crs()} returns \code{crs} object, and \code{coordRef()} a \code{list}.
}
\description{
Get the coordinate reference system (CRS) of a \code{GRaster} or \code{GVector}s, or from the currently active \strong{GRASS} "project/location" (see \code{vignette("projects_mapsets", package = "fasterRaster")}):
\itemize{
\item \code{crs()}: Return a WKT string (an object of class \code{character}).
\item \code{st_crs()}: Return a \code{crs} object (from the \strong{sf} package).
\item \code{coordRef()}: Return a \code{list}.
}
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")
madForest2000 <- fastData("madForest2000")

# Convert SpatRasters to GRasters
elev <- fast(madElev)
forest <- fast(madForest2000)

### GRaster properties

# plotting
plot(elev)

# dimensions
dim(elev) # rows, columns, depths, layers
nrow(elev) # rows
ncol(elev) # columns
ndepth(elev) # depths
nlyr(elev) # layers

res(elev) # resolution (2D)
res3d(elev) # resolution (3D)
zres(elev) # vertical resolution
xres(elev) # vertical resolution
yres(elev) # vertical resolution
zres(elev) # vertical resolution (NA because this is a 2D GRaster)

# cell counts
ncell(elev) # cells
ncell3d(elev) # cells (3D rasters only)

# number of NA and non-NA cells
nacell(elev)
nonnacell(elev)

# topology
topology(elev) # number of dimensions
is.2d(elev) # is it 2-dimensional?
is.3d(elev) # is it 3-dimensional?

minmax(elev) # min/max values

# "names" of the object
names(elev)

# coordinate reference system
crs(elev)
st_crs(elev)
coordRef(elev)

# extent (bounding box)
ext(elev)

# vertical extent (not defined for this raster)
zext(elev)

# data type
datatype(elev) # fasterRaster type
datatype(elev, "GRASS") # GRASS type
datatype(elev, "terra") # terra type
datatype(elev, "GDAL") # GDAL type

is.integer(elev)
is.float(elev)
is.double(elev)
is.factor(elev)

# convert data type
as.int(elev) # integer; note that "elev" is already of type "integer"
as.float(elev) # floating-precision
as.doub(elev) # double-precision

# assigning
pie <- elev
pie[] <- pi # assign all cells to the value of pi
pie

# concatenating multiple GRasters
rasts <- c(elev, forest)
rasts

# subsetting
rasts[[1]]
rasts[["madForest2000"]]

# replacing
rasts[[2]] <- 2 * forest
rasts

# adding layers
rasts[[3]] <- elev > 500 # add a layer
rasts <- c(rasts, sqrt(elev)) # add another
add(rasts) <- ln(elev)
rasts

# names
names(rasts)
names(rasts) <- c("elev_meters", "2_x_forest", "high_elevation", "sqrt_elev", "ln_elev")
rasts

# remove a layer
rasts[["2_x_forest"]] <- NULL
rasts

# number of layers
nlyr(rasts)

# correlation and covariance matrices
madLANDSAT <- fastData("madLANDSAT")
landsat <- fast(madLANDSAT) # projects matrix
layerCor(landsat) # correlation
layerCor(landsat, fun = 'cov') # covariance

}
}
\seealso{
\code{\link[terra:crs]{terra::crs()}}, \code{\link[sf:st_crs]{sf::st_crs()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/madChelsa.r
\docType{data}
\name{madChelsa}
\alias{madChelsa}
\title{Rasters of bioclimatic variables for an eastern portion of Madagascar}
\format{
An object of class \code{SpatRaster} in unprojected (WGS84) coordinates.
}
\source{
\doi{https://doi.org/10.1038/sdata.2017.122}
}
\description{
Rasters of bioclimatic variables for an eastern portion of Madagascar from CHELSA version 2.1 in unprojected (WGS84) coordinates. Values represent averages across 1980-2010. Only these BIOCLIM variables are included:
* \code{bio1}: Mean annual temperature (deg C)
* \code{bio7}: Temperature annual range (hottest - coldest month temperature; deg C)
* \code{bio12}: Total annual precipitation (mm)
* \code{bio15}: Precipitation seasonality (unit-less)
}
\examples{

### vector data

library(sf)

# For vector data, we can use data(*) or fastData(*):
data(madCoast0) # same as next line
madCoast0 <- fastData("madCoast0") # same as previous
madCoast0
plot(st_geometry(madCoast0))

madCoast4 <- fastData("madCoast4")
madCoast4
plot(st_geometry(madCoast4), add = TRUE)

madRivers <- fastData("madRivers")
madRivers
plot(st_geometry(madRivers), col = "blue", add = TRUE)

madDypsis <- fastData("madDypsis")
madDypsis
plot(st_geometry(madDypsis), col = "red", add = TRUE)

### raster data

library(terra)

# For raster data, we can get the file directly or using fastData(*):
rastFile <- system.file("extdata/madElev.tif", package="fasterRaster")
madElev <- terra::rast(rastFile)

madElev <- fastData("madElev") # same as previous two lines
madElev
plot(madElev)

madForest2000 <- fastData("madForest2000")
madForest2000
plot(madForest2000)

madForest2014 <- fastData("madForest2014")
madForest2014
plot(madForest2014)

# multi-layer rasters
madChelsa <- fastData("madChelsa")
madChelsa
plot(madChelsa)

madPpt <- fastData("madPpt")
madTmin <- fastData("madTmin")
madTmax <- fastData("madTmax")
madPpt
madTmin
madTmax


# RGB raster
madLANDSAT <- fastData("madLANDSAT")
madLANDSAT
plotRGB(madLANDSAT, 4, 1, 2, stretch = "lin")

# categorical raster
madCover <- fastData("madCover")
madCover
madCover <- droplevels(madCover)
levels(madCover) # levels in the raster
nlevels(madCover) # number of categories
catNames(madCover) # names of categories table

plot(madCover)
}
\references{
Karger, D.N., Conrad, O., Bohner, J., Kawohl, T., Kreft, H., Soria-Auza, R.W., Zimmermann, N.E., Linder, H.P., and Kessler, M.  2017.  Climatologies at high resolution for the earth's land surface areas. \emph{Scientific Data} 4:170122. \doi{10.1038/sdata.2017.122}
}
\keyword{Madagascar}
\keyword{climate}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/03_options.r
\name{faster}
\alias{faster}
\title{Set or get options shared across fasterRaster functions}
\usage{
faster(..., default = FALSE, restore = FALSE)
}
\arguments{
\item{...}{Either:
\itemize{
\item A character vector: Name(s) of option(s) to get values of;
\item An option and the value of the option using an \code{option = value} pattern; or
\item A names \code{list} with names that match the options you wish to change, and with values to assign to each option.
}

Options include:
\itemize{
\item \code{grassDir} (character): The folder in which \strong{GRASS} is installed on your computer. You must set this option to run most \strong{fasterRaster} functions. Depending on your operating system, your install directory will look something like this:
\itemize{
\item Windows: \code{"C:/Program Files/GRASS GIS 8.4"}
\item Mac OS: \code{"/Applications/GRASS-8.4.app/Contents/Resources"}
\item Linux: \code{"/usr/local/grass"}
}
\item \code{addonsDir} (character): Folder in which \strong{GRASS} addons are stored. If \code{NA} and \code{grassDir} is not \code{NA}, this will be assumed to be \code{file.path(grassDir, "addons")}. The default values is \code{NA}.
\item \code{cores} (integer/numeric integer): Number of processor cores to use on a task. The default is 2. Some \strong{GRASS} modules are parallelized.
\item \code{memory} (integer/numeric): The amount of memory to allocate to a task, in GB, for \strong{GRASS}. The default is 2048 MB (i.e., 2 GB). Some \strong{GRASS} modules can take advantage of more memory.
\item \code{useDataTable} (logical): If \code{FALSE} (default), functions that return tabular output produce \code{data.frame}s. If \code{TRUE}, output will be \code{data.table}s from the \strong{data.table} package. This can be much faster, but it might require you to know how to use \code{data.table}s if you want to manipulate them in \strong{R}. You can always convert them to \code{data.frame}s using \code{\link[base:as.data.frame]{base::as.data.frame()}}.
\item \code{verbose} (logical): If \code{TRUE}, show progress during function operations and other messages. Default is \code{FALSE}. This overrides the value of any \code{verbose} argument in a function.
\item \code{debug} (logical): If \code{TRUE}, show \strong{GRASS} messages and otherwise hidden slots in classes. This is mainly used for debugging, so most users will want to keep this at its default, \code{FALSE}.
\item \code{workDir} (character): The folder in which \strong{GRASS} rasters, vectors, and other objects are created and manipulated. By default, this is given by \code{\link[=tempdir]{tempdir()}}. Note that on some systems, changing the default folder to somewhere else can cause problems with \strong{fasterRaster} being able to find rasters in \strong{GRASS} that have been created.
}}

\item{default}{Logical: Return the default value(s) of the option(s). The default value of \code{default} is \code{FALSE}.}

\item{restore}{Logical: If \code{TRUE}, the all options will be reset to their default values. The default is \code{FALSE}.}
}
\value{
If options are changed, then a named list of option values \emph{before} they were changed is returned invisibly.

If option values are requested, a named list with option values is returned (not invisibly).
}
\description{
\code{faster()} either sets or gets options used across \strong{fasterRaster} functions. Its use can vary:
\itemize{
\item Get current values of a particular option: Use \code{faster("option_name")}. Values will remain unchanged.
\item Get current values of all options: Use \code{faster()} (no arguments). Values will remain unchanged.
\item Get default values of a particular option: Use \code{faster("option_name", default = TRUE)}. Values will remain unchanged.
\item Get default values of all options: Use \code{faster(default = TRUE)}. Values will remain unchanged.
\item Set values of particular options: Use the form \verb{faster(option 1 = value1, option2 = value2)}.
\item Set all options to their defaults: Use \code{faster(restore = TRUE)}.
}

You cannot simultaneously get and set options.

To run most \strong{fasterRaster} functions, you must set the \code{grassDir} option.
}
\examples{
if (grassStarted()) {

# See current values for options:
faster("grassDir")
faster("cores")
faster("memory")
faster("useDataTable")
faster() # all options

# See default values for options:
faster("cores", default = TRUE)
faster(default = TRUE) # all options

# Set options (change accordingly for your system!!!)
if (FALSE) {

   opts. <- faster() # remember starting values of options

   faster(grassDir = "C:/Program Files/GRASS GIS 8.4")
   faster(verbose = TRUE, memory = 1024, cores = 1)

   faster(c("grassDir", "verbose", "memory", "cores"))

   faster(opts.) # reset options to starting values

}

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grassGUI.r
\name{grassGUI,missing-method}
\alias{grassGUI,missing-method}
\alias{grassGUI}
\title{Start the GRASS GUI (potentially dangerous!)}
\usage{
\S4method{grassGUI}{missing}()
}
\value{
Nothing (starts the \strong{GRASS} GUI).
}
\description{
This function starts the \strong{GRASS} GUI. It is provided merely as a utility... in most cases, it should \emph{not} be used if you are doing any kind of analysis of rasters or vectors using \strong{fasterRaster}. The reason for this prohibition is that \strong{fasterRaster} objects, like \code{GRaster}s and \code{GVector}s, are really "pointers" to objects in \strong{GRASS}. If \strong{fasterRaster} points to a \strong{GRASS} object that is changed in \strong{GRASS} but not \strong{R}, then \strong{fasterRaster} will not "know" about it, so changed won't be reflected in the \strong{fasterRaster} object.

One aspect of the GUI that is useful but will not change objects is to use it to plot rasters and vectors. However, the a \strong{fasterRaster} object in \strong{R} will have a different name in \strong{GRASS}. The name in \strong{GRASS} of a \code{GVector} or \code{GRaster} is given by \code{\link[=sources]{sources()}}.
}
\examples{
if (grassStarted()) {

# DANGER: Making changes to rasters/vectors in the GUI can "break" them in R.
if (interactive()) grassGUI()

}
}
\seealso{
\code{\link[=mow]{mow()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/delaunay.r
\name{delaunay,GVector-method}
\alias{delaunay,GVector-method}
\alias{delaunay}
\title{Delaunay triangulation for points}
\usage{
\S4method{delaunay}{GVector}(x)
}
\arguments{
\item{x}{A \code{GVector} "points" object.}
}
\value{
A \code{GVector}.
}
\description{
This function creates a Delaunay triangulation from a "points" \code{GVector}.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)

# Example vectors
madDypsis <- fastData("madDypsis") # points
madCoast4 <- fastData("madCoast4") # polygons

# Convert sf vectors to GVectors
dypsis <- fast(madDypsis)
coast4 <- fast(madCoast4)
ant <- coast4[coast4$NAME_4 == "Antanambe"]

# Delaunay triangulation
dypsisDel <- delaunay(dypsis)
plot(dypsisDel)
plot(dypsis, pch = 1, col = "red", add = TRUE)

# Voronoi tessellation
vor <- voronoi(dypsis)
plot(vor)
plot(dypsis, pch = 1, col = "red", add = TRUE)

# Random Voronoi tessellation
rand <- rvoronoi(coast4, size = 100)
plot(rand)

}
}
\seealso{
\code{\link[terra:voronoi]{terra::delaunay()}}, tool \code{v.delaunay} in \strong{GRASS}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stretch.r
\name{stretch,GRaster-method}
\alias{stretch,GRaster-method}
\alias{stretch}
\title{Rescale values in a GRaster}
\usage{
\S4method{stretch}{GRaster}(x, minv = 0, maxv = 255, minq = 0, maxq = 1, smin = NA, smax = NA)
}
\arguments{
\item{x}{A \code{GRaster}.}

\item{minv, maxv}{Numeric: Minimum and maximum values to which to rescale values.}

\item{minq, maxq}{Numeric: Specifies range of values to rescale, given by their quantiles. The default is to stretch all values (the 0th and 100th quantiles). One or both are  ignored if \code{smin} and/or \code{smax} are provided.}

\item{smin, smax}{Numeric or \code{NA}: Specifies range of values to rescale. If \code{NA} (default), then all values are rescaled.}
}
\value{
A \code{GRaster}.
}
\description{
\code{stretch()} rescales the values in a \code{GRaster}. All values can be rescaled, or just values in a user-defined range. This range can be given by specifying either the lower and upper bounds of the range using \code{smin} and \code{smax}, and/or by the quantiles (across all cells of the raster) using \code{minq} and \code{maxq}.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster
elev <- fast(madElev)

### Stretch based on user-defined range

#  fasterRaster
fr <- stretch(elev, smin=1, smax=100)
fr

# terra
tr <- stretch(madElev, smin = 1, smax = 100)
tr

# Compare fasterRaster to terra output
fr <- rast(fr)
fr <- extend(fr, tr)
fr - tr

### Stretch values in a certain quantile range

#  fasterRaster
fr <- stretch(elev, minq = 0.25, maxq = 0.75)
fr

# terra
tr <- stretch(madElev, minq = 0.25, maxq = 0.75)
tr

# Compare fasterRaster to terra output
fr <- rast(fr)
fr <- extend(fr, tr)
fr - tr

}
}
\seealso{
\code{\link[terra:stretch]{terra::stretch()}} and tool \code{r.rescale} in \strong{GRASS} (not used on this function)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/addCats.r
\name{addCats,GRaster-method}
\alias{addCats,GRaster-method}
\alias{addCats}
\alias{addCats<-,GRaster-method}
\alias{addCats<-}
\title{Add rows or columns to the "levels" table of a categorical raster}
\usage{
\S4method{addCats}{GRaster}(x, value, merge = FALSE, layer = 1)

\S4method{addCats}{GRaster}(x, layer = 1) <- value
}
\arguments{
\item{x}{A \code{GRaster}.}

\item{value}{A \code{data.frame}, \code{data.table}, a list of \code{data.frames} or \code{data.tables} with one per raster layer, or a categorical \code{SpatRaster}. The table's first column is the "value" column and must contain numeric values (of class \code{numeric} or \code{character}). If a \code{SpatRaster} is supplied, then its categories will be transferred to the \code{GRaster}.}

\item{merge}{Logical (function \code{addCats()}): If \code{FALSE} (default), columns will be combined with the existing "levels" table using \code{\link[=cbind]{cbind()}}. If \code{TRUE}, they will be combined using \code{\link[data.table:merge]{data.table::merge()}}.}

\item{layer}{Numeric integers, logical vector, or character: Layer(s) to which to add or from which to drop levels.}
}
\value{
A \code{GRaster}. The "levels" table of the raster is modified.
}
\description{
\code{addCats()} and \verb{addCats()<-} add information to a categorical`GRaster's "levels" table.
\itemize{
\item addCats()` uses \code{\link[data.table:merge]{data.table::merge()}} or \code{\link[=cbind]{cbind()}} to do this--it does not add new rows, but rather new columns.
\item \verb{addCats()<-} uses \code{\link[=rbind]{rbind()}} to add new categories (rows) to the "levels" table.
}

GRaster\verb{s can represent categorical data (see }vignette("GRasters", package = "fasterRaster")\verb{). Cell values are actually integers, each corresponding to a category, such as "desert" or "wetland." A categorical raster is associated with a "levels" table that matches each value to a category name. The table must be }NULL` (i.e., no categories--so not a categorical raster), or have at least two columns. The first column must have integers and represent raster values. One or more subsequent columns must have category labels. The column with these labels is the "active category".
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data: Land cover raster
madCover <- fastData("madCover")

# Convert categorical SpatRaster to categorical GRaster:
cover <- fast(madCover)

### Properties of categorical rasters

cover # note categories
is.factor(cover) # Is the raster categorical?
nlevels(cover) # number of levels
levels(cover) # just the value and active column
cats(cover) # all columns
minmax(cover) # min/max values
minmax(cover, levels = TRUE) # min/max categories
catNames(cover) # column names of the levels table
missingCats(cover) # categories in table with no values in raster
freq(cover) # frequency of each category (number of cells)
zonalGeog(cover) # geometric statistics

### Active column

# Which column sets the category labels?
activeCat(cover)
activeCat(cover, names = TRUE)

activeCats(c(cover, cover))

# Choose a different column for category labels:
levels(cover)
activeCat(cover) <- 2
levels(cover)

### Managing levels tables

# Remove unused levels:
nlevels(cover)
cover <- droplevels(cover)
nlevels(cover)

# Re-assign levels:
value <- c(20, 30, 40, 50, 120, 130, 140, 170)
label <- c("Cropland", "Cropland", "Forest", "Forest",
 "Grassland", "Shrubland", "Herbaceous", "Flooded")

newCats <- data.frame(value = value, label = label)

cover <- categories(cover, layer = 1, value = newCats)
cats(cover)

# This is the same as:
levels(cover) <- newCats
cats(cover)

# Are there any values not assigned a category?
missingCats(cover)

# Let's assign a category for value 210 (water):
water <- data.frame(value = 210, label = "Water")
addCats(cover) <- water
levels(cover)

# Add more information to the levels table using merge():
landType <- data.frame(
     Value = c(20, 30, 40, 50, 120),
     Type = c("Irrigated", "Rainfed", "Broadleaf evergreen",
     "Broadleaf deciduous", "Mosaic with forest")
)
cats(cover)
cover <- addCats(cover, landType, merge = TRUE)
cats(cover)

### Logical operations on categorical rasters

cover < "Forest" # 1 for cells with a value < 40, 0 otherwise
cover <= "Forest" # 1 for cells with a value < 120, 0 otherwise
cover == "Forest" # 1 for cells with value of 40-120, 0 otherwise
cover != "Forest" # 1 for cells with value that is not 40-120, 0 otherwise
cover > "Forest" # 1 for cells with a value > 120, 0 otherwise
cover >= "Forest" # 1 for cells with a value >= 120, 0 otherwise

cover \%in\% c("Cropland", "Forest") # 1 for cropland/forest cells, 0 otherwise

### Combine categories from different rasters

# For the example, will create a second categorical raster fromm elevation.

# Divide elevation raster into "low/medium/high" levels:
madElev <- fastData("madElev")
elev <- fast(madElev)
elev <- project(elev, cover, method = "near") # convert to same CRS
fun <- "= if(madElev < 100, 0, if(madElev < 400, 1, 2))"
elevCat <- app(elev, fun)

levs <- data.frame(
     value = c(0, 1, 2),
     elevation = c("low", "medium", "high")
)
levels(elevCat) <- list(levs)

# Combine levels:
combined <- concats(cover, elevCat)
combined
levels(combined)

# Combine levels, treating value/NA combinations as new categories:
combinedNA <- concats(cover, elevCat, na.rm = FALSE)
combinedNA
levels(combinedNA)

}
}
\seealso{
\code{\link[terra:factors]{terra::addCats()}}, \code{\link[=concats]{concats()}}, \code{\link[=combineLevels]{combineLevels()}}, \code{\link[=droplevels]{droplevels()}}, \verb{vignette("GRasters", package = "fasterRaster"}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/maskNA.r
\name{maskNA,GRaster-method}
\alias{maskNA,GRaster-method}
\alias{maskNA}
\title{Mask all non-NA cells or all NA cells}
\usage{
\S4method{maskNA}{GRaster}(x, value = 1, invert = FALSE, retain = FALSE)
}
\arguments{
\item{x}{A \code{GRaster}.}

\item{value}{Numeric: Value to which to assign to masked cells. The default is 1.}

\item{invert}{Logical: If \code{FALSE} (default), convert non-\code{NA} cells to \code{value}, and leave \code{NA} cells as-is. If \code{TRUE}, convert all \code{NA} cells to \code{value}, and non-\code{NA} cells to \code{NA}.}

\item{retain}{Logical: If \code{invert} is \code{TRUE} and \code{retain} is \code{FALSE} (default), non-\code{NA} cells will retain their value. This argument is ignored if \code{invert} is \code{FALSE}.}
}
\value{
A \code{GRaster}.
}
\description{
This function converts all non-\code{NA} cells in a \code{GRaster} to a single user-defined value, leaving \code{NA} cells as \code{NA}. Alternatively, it can convert \code{NA} cells to a user-defined value, and all non-\code{NA} cells to \code{NA.}
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Elevation raster
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster:
elev <- fast(madElev)

# Convert non-NA to 1, NA cells remain NA
elevMask <- maskNA(elev)
elevMask
plot(c(elev, elevMask))

# Convert NA to 1, non-NA cells become NA
elevInvertMask <- maskNA(elev, invert = TRUE)
elevInvertMask
plot(c(elev, elevInvertMask))

# Convert NA to 200, non-NA cells keep their values
elevInvertRetain <- maskNA(elev, value = 200, invert = TRUE, retain = TRUE)
elevInvertRetain
plot(c(elev, elevInvertRetain))

}
}
\seealso{
\code{\link[=not.na]{not.na()}}, \code{\link[=app]{app()}}, \code{\link[=mask]{mask()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datatype.r
\name{datatype,GRaster-method}
\alias{datatype,GRaster-method}
\alias{datatype}
\alias{datatype,GVector-method}
\title{Get the datatype of a GRaster or of GVector columns}
\usage{
\S4method{datatype}{GRaster}(x, type = "fasterRaster", forceDouble = TRUE)

\S4method{datatype}{GVector}(x)
}
\arguments{
\item{x}{A \code{GRaster} or \code{GVector}.}

\item{type}{(\code{GRaster}s only) \code{NULL} or character: Type of datatype to report (\code{GRaster} only):
\itemize{
\item \code{"fasterRaster"} (default): Reports the \strong{fasterRaster} type (factor, integer, float, or double)
\item \code{"terra"}: Report the (inferred) \strong{terra} data type (e.g., INT2U, FLT4S). Please see the table in the documentation for [writeRaster()` for an explanation of these codes.
\item \code{"GRASS"}: Will return "CELL" (integer), "FCELL" (floating-point value), or "DCELL" (double-floating point value)
\item \code{"GDAL"}: See \href{https://gdal.org/en/stable/user/raster_data_model.html}{GDAL: Raster Band}. Please also see the table in the \code{\link[=writeRaster]{writeRaster()}} help page.
}}

\item{forceDouble}{Logical (\code{GRaster}s and \code{SpatRaster}s only): If \code{TRUE} (default), and the raster appears to represent non-integer values, then the raster will be assumed to represent double-floating point values (\strong{GRASS}: type "DCELL", \strong{terra}: type "FLT8S", \strong{fasterRaster}: type "double", and \strong{GDAL}: type "Float64"). \code{forceDouble} reports the actual datatype if \code{type = "fasterRaster"} (i.e., the type is not forced to "double").}
}
\value{
\code{datatype()} for a \code{GRaster} returns a character. \code{datatype()} for a \code{GVector} returns a data frame, with one row per field. If the \code{GVector} has no attribute table, the function returns \code{NULL}.
}
\description{
For \code{GRaster}s, \code{datatype()} returns the data type (see \code{vignette("GRasters", package = "fasterRaster")}). For \code{GVector}s, \code{datatype()} returns the class of each column of the attribute table.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Example data
madElev <- fastData("madElev")
madForest2000 <- fastData("madForest2000")
madCoast0 <- fastData("madCoast0")
madRivers <- fastData("madRivers")
madDypsis <- fastData("madDypsis")

### GRaster properties

# convert SpatRasters to GRasters
elev <- fast(madElev)
forest <- fast(madForest2000)

# plot
plot(elev)

dim(elev) # rows, columns, depths, layers
nrow(elev) # rows
ncol(elev) # columns
ndepth(elev) # depths
nlyr(elev) # layers

res(elev) # resolution

ncell(elev) # cells
ncell3d(elev) # cells (3D rasters only)

topology(elev) # number of dimensions
is.2d(elev) # is it 2D?
is.3d(elev) # is it 3D?

minmax(elev) # min/max values

# name of object in GRASS
sources(elev)

# "names" of the object
names(elev)

# coordinate reference system
crs(elev)

# extent (bounding box)
ext(elev)

# data type
datatype(elev)

# assigning
copy <- elev
copy[] <- pi # assign all cells to the value of pi
copy

# concatenating multiple GRasters
rasts <- c(elev, forest)
rasts

# adding a raster "in place"
add(rasts) <- ln(elev)
rasts

# subsetting
rasts[[1]]
rasts[["madForest2000"]]

# assigning
rasts[[4]] <- elev > 500

# number of layers
nlyr(rasts)

# names
names(rasts)
names(rasts) <- c("elev_meters", "forest", "ln_elev", "high_elevation")
rasts

### GVector properties

# convert sf vectors to GVectors
coast <- fast(madCoast4)
rivers <- fast(madRivers)
dypsis <- fast(madDypsis)

# extent
ext(rivers)

W(rivers) # western extent
E(rivers) # eastern extent
S(rivers) # southern extent
N(rivers) # northern extent
top(rivers) # top extent (NA for 2D rasters like this one)
bottom(rivers) # bottom extent (NA for 2D rasters like this one)

# coordinate reference system
crs(rivers)
st_crs(rivers)

# column names and data types
names(coast)
datatype(coast)

# name of object in GRASS
sources(rivers)

# points, lines, or polygons?
geomtype(dypsis)
geomtype(rivers)
geomtype(coast)

is.points(dypsis)
is.points(coast)

is.lines(rivers)
is.lines(dypsis)

is.polygons(coast)
is.polygons(dypsis)

# dimensions
nrow(rivers) # how many spatial features
ncol(rivers) # hay many columns in the data frame

# number of geometries and sub-geometries
ngeom(coast)
nsubgeom(coast)

# 2- or 3D
topology(rivers) # dimensionality
is.2d(elev) # is it 2D?
is.3d(elev) # is it 3D?

# Update values from GRASS
# (Reads values from GRASS... will not appear to do anything in this case)
coast <- update(coast)

### operations on GVectors

# convert to data frame
as.data.frame(rivers)
as.data.table(rivers)

# subsetting
rivers[c(1:2, 5)] # select 3 rows/geometries
rivers[-5:-11] # remove rows/geometries 5 through 11
rivers[ , 1] # column 1
rivers[ , "NAM"] # select column
rivers[["NAM"]] # select column
rivers[1, 2:3] # row/geometry 1 and column 2 and 3
rivers[c(TRUE, FALSE)] # select every other geometry (T/F vector is recycled)
rivers[ , c(TRUE, FALSE)] # select every other column (T/F vector is recycled)

# removing data table
noTable <- dropTable(rivers)
noTable
nrow(rivers)
nrow(noTable)

# Refresh values from GRASS
# (Reads values from GRASS... will not appear to do anything in this case
# since the rivers object is up-to-date):
rivers <- update(rivers)

# Concatenating multiple vectors
rivers2 <- rbind(rivers, rivers)
dim(rivers)
dim(rivers2)

}
}
\seealso{
\code{\link[terra:datatype]{terra::datatype()}}, \code{vignette("GRasters", package = "fasterRaster")}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/04_arithmetic.r
\name{Arith,GRaster,logical-method}
\alias{Arith,GRaster,logical-method}
\alias{Arith}
\alias{Arith,logical,GRaster-method}
\alias{Arith,GRaster,numeric-method}
\alias{Arith,GRaster,integer-method}
\alias{Arith,numeric,GRaster-method}
\alias{Arith,integer,GRaster-method}
\alias{Arith,GRaster,GRaster-method}
\alias{Arith,GVector,GVector-method}
\title{Arithmetic operations on GRasters}
\usage{
\S4method{Arith}{GRaster,logical}(e1, e2)

\S4method{Arith}{logical,GRaster}(e1, e2)

\S4method{Arith}{GRaster,numeric}(e1, e2)

\S4method{Arith}{GRaster,integer}(e1, e2)

\S4method{Arith}{numeric,GRaster}(e1, e2)

\S4method{Arith}{integer,GRaster}(e1, e2)

\S4method{Arith}{GRaster,GRaster}(e1, e2)

\S4method{Arith}{GVector,GVector}(e1, e2)
}
\arguments{
\item{e1, e2}{\code{GRaster}s, \code{GVector}s, \code{numeric}s, \code{integer}s, or \code{logical}s.}
}
\value{
A \code{GRaster}.
}
\description{
\strong{\code{GRaster}s}: You can do arithmetic operations on \code{GRaster}s and using normal operators in \strong{R}: \code{+}, \code{-}, \code{*}, \code{/}, \code{^}, \code{\%\%} (modulus), and \code{\%/\%} (integer division).

\strong{\code{GVector}s}: You can also do arithmetic operations on \code{GVector}s:\cr\cr
\code{+} operator: Same as \code{\link[=union]{union()}}\cr
\code{-} operator: Same as \code{\link[=erase]{erase()}}\cr
\code{*} operator: Same as \code{\link[=intersect]{intersect()}}\cr
\code{/} operator: Same as \code{\link[=xor]{xor()}}\cr
}
\examples{
if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Example data
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster
elev <- fast(madElev)
elevs <- c(elev, elev, log10(elev) - 1, sqrt(elev))
names(elevs) <- c("elev1", "elev2", "log_elev", "sqrt_elev")

elev
elevs

# do some math
elev + 100
elev - 100
elev * 100
elev / 100
elev ^ 2
elev \%/\% 100 # divide then round down
elev \%\% 100 # modulus

100 + elev
100 \%/\% elev
100 \%\% elev

elevs + 100
100 + elevs

# math with logicals
elev + TRUE
elev - TRUE
elev * TRUE
elev / TRUE
elev ^ TRUE
elev \%/\% TRUE # divide then round down
elev \%\% TRUE # modulus

elevs + TRUE
TRUE + elevs

# Raster interacting with raster(s):
elev + elev
elev - elev
elev * elev
elev / elev
elev ^ log(elev)
elev \%/\% sqrt(elev) # divide then round down
elev \%\% sqrt(elev) # modulus

elevs + elev
elev * elevs

# sign
abs(-1 * elev)
abs(elevs)

# powers
sqrt(elevs)

# trigonometry
sin(elev)
cos(elev)
tan(elev)

asin(elev)
acos(elev)
atan(elev)

atan(elevs)
atan2(elev, elev^1.2)
atan2(elevs, elev^1.2)
atan2(elev, elevs^1.2)
atan2(elevs, elevs^1.2)

# logarithms
exp(elev)
log(elev)
ln(elev)
log2(elev)
log1p(elev)
log10(elev)
log10p(elev)
log(elev, 3)

log(elevs)

# rounding
round(elev + 0.5)
floor(elev + 0.5)
ceiling(elev + 0.5)
trunc(elev + 0.5)

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/08_logic.r
\name{Logic,GRaster,GRaster-method}
\alias{Logic,GRaster,GRaster-method}
\alias{Logic-methods}
\alias{Logic,logical,GRaster-method}
\alias{Logic,GRaster,logical-method}
\alias{Logic,GRaster,numeric-method}
\alias{Logic,numeric,GRaster-method}
\alias{Logic,GRaster,integer-method}
\alias{Logic,integer,GRaster-method}
\title{Logic-methods operations on GRasters}
\usage{
\S4method{Logic}{GRaster,GRaster}(e1, e2)

\S4method{Logic}{logical,GRaster}(e1, e2)

\S4method{Logic}{GRaster,logical}(e1, e2)

\S4method{Logic}{GRaster,numeric}(e1, e2)

\S4method{Logic}{numeric,GRaster}(e1, e2)

\S4method{Logic}{GRaster,integer}(e1, e2)

\S4method{Logic}{integer,GRaster}(e1, e2)
}
\arguments{
\item{e1, e2}{Two \code{GRaster}s, or a \code{GRaster} and a logical value (\code{TRUE} or \code{FALSE}, but not \code{NA}), a numeric value that is 0 or 1 (but not \code{NA_real_}), or an integer value that is 0 or 1 (but not \code{NA_integer_}).}
}
\value{
A binary \code{GRaster} (1 ==> \code{TRUE}, 0 ==> \code{FALSE}, plus \code{NA} when comparison results in \code{NA}).
}
\description{
You can do logical operations on \code{GRaster}s. A cell with a value of 1 is interpreted as \code{TRUE}, and a value of 0 is interpreted as \code{FALSE}. You can compare:
\itemize{
\item A \code{GRaster} to another \code{GRaster}
\item A \code{GRaster} to a logical value (\code{TRUE} or \code{FALSE}, but not \code{NA}--see \code{\link[=not.na]{not.na()}})
\item A \code{GRaster} to a numeric or integer value that is 0 or 1
}

Operators include:
\itemize{
\item \code{|}: \code{TRUE} if either condition is \code{TRUE} (or 1), but returns \code{NA} if either condition is \code{NA}.
\item \code{&}: \code{TRUE} if both conditions are \code{TRUE} (or 1), but \code{NA} if either is \code{NA}.
}
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster
elev <- fast(madElev)
elevs <- c(elev, elev, log10(elev) - 1, sqrt(elev))
names(elevs) <- c("elev1", "elev2", "log_elev", "sqrt_elev")

elev
elevs

# Comparisons
elev < 100
elev <= 100
elev == 100
elev != 100
elev > 100
elev >= 100

elev + 100 < 2 * elev

elevs > 10
10 > elevs

# logic
elev < 10 | elev > 200
elev < 10 | cos(elev) > 0.9

elev < 10 | TRUE
TRUE | elev > 200

elev < 10 | FALSE
FALSE | elev > 200

elev < 10 & cos(elev) > 0.9

elev < 10 & TRUE
TRUE & elev > 200

elev < 10 & FALSE
FALSE & elev > 200

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/focal.r
\name{focal,GRaster-method}
\alias{focal,GRaster-method}
\alias{focal}
\title{Calculate cell values based on values of nearby cells}
\usage{
\S4method{focal}{GRaster}(x, w = 3, fun = "sum", circle = FALSE, quantile = 0.5)
}
\arguments{
\item{x}{A \code{GRaster}.}

\item{w}{Numeric integer or a square matrix with an odd number of rows and columns: The size and nature of the neighborhood:
\itemize{
\item "Square" neighborhoods (when \code{circle = FALSE}): An odd integer >= 3, indicating indicates the size of a "square" neighborhood (number of cells wide and number or cells tall).
\item "Circular" neighborhoods (when \code{circle = TRUE}): An odd integer >=3, indicating the diameter of the circle.
\item A matrix of cell weights: The matrix must be square and have an odd number of rows and columns (example: \code{matrix(c(0.5, 1, 0.5, 1, 2, 1, 0.5, 1, 0.5), nrow=3)}). You cannot use a weights matrix when \code{circle = TRUE}. Cells with \code{NA} as a weight will be ignored. Note that weighted matrices should not be used for function \code{min}, \code{max}, \code{count}, \code{nunique}, or \code{interspersion}.
}}

\item{fun}{Character: Name of the function to apply to the neighborhood:
\itemize{
\item "\code{mean}" (default)
\item "\code{median}"
\item "\code{mode}"
\item "\code{min}" or "\code{max}": Minimum or maximum. Should not use a weights matrix.
\item "\code{range}": Difference between the maximum and minimum.  Should not use a weights matrix.
\item "\code{sd}": Sample standard deviation.  NB: This is the same as the \code{\link[stats:sd]{stats::sd()}} function.
\item "\code{sdpop}": Population standard deviation. NB: This is the same as the function "stddev" in the \strong{GRASS} tool \code{r.neighbors}.
\item "\code{sum}": Sum of non-`NA`` cells.
\item "\code{count}": Number of non-`NA cells. Should not use a weights matrix.
\item "\code{var}": Sample variance.  NB: This is the same as the \code{\link[stats:cor]{stats::var()}} function.
\item "\code{varpop}": Population variance. NB: This is the same as the function "variance" in the \strong{GRASS} tool \code{r.neighbors}.
\item "\code{nunique}": Number of unique values. Should not use a weights matrix.
\item "\code{interspersion}": Proportion of cells with values different from focal cell (e.g., if 6 of 8 cells have different values, then the interspersion is 6/8 = 0.75). NB: This is slightly different from how it is defined in the \strong{GRASS} tool \code{r.neighbors}. Should not use a weights matrix.
\item "\code{quantile}": Quantile of values. The value in argument \code{quantile} is used to specify the quantile.
}

The center cell value is always included in the calculations, and all calculations ignore \code{NA} cells (i.e., they do not count as cells in the focal neighborhood).}

\item{circle}{Logical: If \code{FALSE} (default), use a square neighborhood. If \code{TRUE}, use a circular neighborhood. When this is \code{TRUE}, argument \code{w} cannot be a matrix.}

\item{quantile}{Numeric between 0 and 1, inclusive: Quantile to calculate when \code{fun = "quantile"}. The default value is 0.5 (median), and valid values must be in the range between 0 and 1, inclusive.}
}
\value{
A \code{GRaster}.
}
\description{
This function calculates statistics on a moving "neighborhood" of cells of a raster. The neighborhood can be a square, circle, or a user-defined set of cells (with or without weights).
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster:
elev <- fast(madElev)

# Focal calculations:
sums <- focal(elev, fun = "sum")
means <- focal(elev, fun = "mean")

# Focal calculations on a circular window:
sds <- focal(elev, fun = "sd") # square
sdsCircle <- focal(elev, fun = "sd", circle = TRUE) # circle

sds
sdsCircle

plot(sds - sdsCircle)

# Focal calculations with user-defined weights:
w <- matrix(c(1, 0, 1, 0, 1, 0, 1, 0, 1), ncol = 3)
w
sumsWeighted <- focal(elev, fun = "sum", w = w)

s <- c(sums, sumsWeighted)
minmax(s)

}
}
\seealso{
\code{\link[terra:focal]{terra::focal()}}, \strong{GRASS} manual page for tool \code{r.neighbors} (see \code{grassHelp("r.neighbors")})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rastInfo_vectInfo.r
\name{.vectInfo}
\alias{.vectInfo}
\title{Metadata on a vector in GRASS}
\usage{
.vectInfo(x, integer = TRUE, cats = NULL)
}
\arguments{
\item{x}{A \code{GVector} or the \code{\link[=sources]{sources()}} name of one.}

\item{integer}{Logical. If \code{TRUE} (default), the "cats" of a vector will be returned as type integer. If \code{FALSE}, they will be returned as a character vector.}

\item{cats}{\code{NULL} (default) or an integer vector of category numbers, one per geometry.}
}
\value{
A \code{vectInfo} object (a list).
}
\description{
This function queries \strong{GRASS} to obtain metadata on a vector.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/06_GRaster_functions_across_layers.r
\name{mean,GRaster-method}
\alias{mean,GRaster-method}
\alias{mean}
\alias{mmode,GRaster-method}
\alias{mmode}
\alias{median,GRaster-method}
\alias{median}
\alias{count,GRaster-method}
\alias{count}
\alias{sum,GRaster-method}
\alias{sum}
\alias{min,GRaster-method}
\alias{min}
\alias{max,GRaster-method}
\alias{max}
\alias{which.min,GRaster-method}
\alias{which.min}
\alias{which.max,GRaster-method}
\alias{which.max}
\alias{sdpop,numeric-method}
\alias{sdpop}
\alias{varpop,GRaster-method}
\alias{varpop}
\alias{varpop,numeric-method}
\alias{stdev,GRaster-method}
\alias{stdev}
\alias{var,GRaster-method}
\alias{var}
\alias{nunique,GRaster-method}
\alias{nunique}
\alias{skewness,GRaster-method}
\alias{skewness}
\alias{kurtosis,GRaster-method}
\alias{kurtosis}
\alias{range,GRaster-method}
\alias{range}
\alias{quantile,GRaster-method}
\alias{quantile}
\alias{anyNA,GRaster-method}
\alias{anyNA}
\alias{allNA,GRaster-method}
\alias{allNA}
\title{Mathematical operations on two or more GRasters}
\usage{
\S4method{mean}{GRaster}(x, na.rm = FALSE)

\S4method{mmode}{GRaster}(x, na.rm = FALSE)

\S4method{median}{GRaster}(x, na.rm = FALSE)

\S4method{count}{GRaster}(x)

\S4method{sum}{GRaster}(x, na.rm = FALSE)

\S4method{min}{GRaster}(x, na.rm = FALSE)

\S4method{max}{GRaster}(x, na.rm = FALSE)

\S4method{which.min}{GRaster}(x)

\S4method{which.max}{GRaster}(x)

\S4method{sdpop}{numeric}(x, na.rm = FALSE)

\S4method{varpop}{GRaster}(x, na.rm = FALSE)

\S4method{varpop}{numeric}(x, na.rm = FALSE)

\S4method{stdev}{GRaster}(x, pop = TRUE, na.rm = FALSE)

\S4method{var}{GRaster}(x, na.rm = FALSE)

\S4method{nunique}{GRaster}(x, na.rm = FALSE)

\S4method{skewness}{GRaster}(x, na.rm = FALSE)

\S4method{kurtosis}{GRaster}(x, na.rm = FALSE)

\S4method{range}{GRaster}(x, na.rm = FALSE)

\S4method{quantile}{GRaster}(x, prob, na.rm = FALSE)

\S4method{anyNA}{GRaster}(x)

\S4method{allNA}{GRaster}(x)
}
\arguments{
\item{x}{A \code{GRaster}. Typically, this raster will have two or more layers. Values will be calculated within cells across rasters.}

\item{na.rm}{Logical: If \code{FALSE} (default), of one cell value has an \code{NA}, the result will be \code{NA}. If \code{TRUE}, \code{NA}s are ignored.}

\item{pop}{Logical (for \code{stdev()}): If \code{TRUE} (default), calculate the population standard deviation across layers. If \code{FALSE}, calculate the sample standard deviation.}

\item{prob}{Numeric: Quantile to calculate. Used for \code{quantile()}.}
}
\value{
A \code{GRaster}.
}
\description{
These functions can be applied to a "stack" of \code{GRaster}s with two or more layers. They return a single-layered \code{GRaster}.  If you want to summarize across cells in a raster (e.g., calculate the mean value of all cells on a raster), use \code{\link[=global]{global()}}. Options include:
\itemize{
\item Numeration: \code{count()} (number of non-\code{NA} cells), \code{sum()}.
\item Central tendency: \code{mean()}, \code{mmode()} (mode), \code{median()}.
\item Extremes: \code{min()}, \code{max()}, \code{which.min()} (index of raster with the minimum value), \code{which.max()} (index of the raster with the maximum value)
\item Dispersion: \code{range()}, \code{stdev()} (standard deviation), \code{var()} (sample variance), \code{varpop()} (population variance), \code{nunique()} (number of unique values), \code{quantile()} (use argument \code{probs}), \code{skewness()}, and \code{kurtosis()}.
\item \code{NA}s: \code{anyNA()} (any cells are \code{NA}?), \code{allNA()} (are all cells \code{NA}?)
}
}
\examples{
if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Example data
madChelsa <- fastData("madChelsa")

# Convert a SpatRaster to a GRaster
chelsa <- fast(madChelsa)
chelsa # 4 layers

# Central tendency
mean(chelsa)
mmode(chelsa)
median(chelsa)

# Statistics
nunique(chelsa)
sum(chelsa)
count(chelsa)
min(chelsa)
max(chelsa)
range(chelsa)
skewness(chelsa)
kurtosis(chelsa)

stdev(chelsa)
stdev(chelsa, pop = FALSE)
var(chelsa)
varpop(chelsa)

# Which layers have maximum/minimum?
which.min(chelsa)
which.max(chelsa)

# Regression

# Note the intercept is different for fasterRaster::regress().
regress(chelsa)
regress(madChelsa, 1:nlyr(madChelsa))

# Note: To get quantiles for each layer, use global().
quantile(chelsa, 0.1)

# NAs
madForest2000 <- fastData("madForest2000")
forest2000 <- fast(madForest2000)
forest2000 <- project(forest2000, chelsa, method = "near")

chelsaForest <- c(chelsa, forest2000)

nas <- anyNA(chelsaForest)
plot(nas)

allNas <- allNA(chelsaForest)
plot(allNas)

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geomorphons.r
\name{geomorphons,GRaster-method}
\alias{geomorphons,GRaster-method}
\alias{geomorphons}
\title{Identify terrain feature types}
\usage{
\S4method{geomorphons}{GRaster}(
  x,
  inner = 0,
  outer = 3,
  unit = "cells",
  flat = 1,
  flatDist = 0,
  mode = "1"
)
}
\arguments{
\item{x}{A single-layer \code{GRaster}, typically representing elevation.}

\item{inner, outer}{Integer: Inner and outer radii of the torus used to identify geomorphons, in cells or meters (set by argument \code{unit}). The inner default value is 0 and the outer default value is 3. The \code{outer} radius sets the maximum size of a geomorphon that that can be identified, and \code{inner} sets the smallest size. If \code{unit} is \code{"meters"}, the value of \code{outer} must be larger than the smaller dimension of any cell in the east-west and north-south directions.}

\item{unit}{Character: Units of \code{inner} and outer; can be either \code{"cells"} (default) or \code{"meters"}. Partial matching is used.}

\item{flat}{Numeric value >= 0: Minimum difference (in degrees) between the focal area areas around it for a geomorphon to be considered as "flat". Larger cells (i.e., ~1 km resolution or larger) require smaller values (<<1) to correctly identify flat areas. Higher values result in more areas being classified as "flat" geomorphons. The default value is 1.}

\item{flatDist}{Numeric: Distance (in meters) to correct for the effect of large distances on the diminished capacity to identify "flat" geomorphons. If the distance between the focal area and a surrounding area surpasses this distance, then the effective value of \code{flat} will be reduced}

\item{mode}{Character: Method for implementing the zenith/line-of-site search. Partial matching is used:
\itemize{
\item \code{"1"} (default): The "original" geomorphon mode (in \strong{GRASS} tool \code{r.geomorphon}, the "anglev1" method)
\item \code{"2"}: Better handling of cases with equal zenith/nadir angles (the "anglev2" method)
\item \code{"2d"}: As \code{"2"}, but takes into account zenith/nadir distance ("anglev2_distance" method)
}}
}
\value{
A categorical \code{GRaster} where each geomorphon is a category (see \code{vignette("GRasters", package = "fasterRaster")}).
}
\description{
Geomorphons are idealized terrain types calculated from an elevator raster based on a moving window of a given size. The window is a torus (which can have an inner radius of 0, so can also be a circle), which allows it to identify geomorphons of a given size while ignoring ones larger or smaller. There are 10 basic geomorphons. Consult the the manual for \strong{GRASS} tool \code{r.geomorphon} using \code{grassHelp("r.geomorphon")} for more details and diagrams of each type of geomorphon. Geomorphon types include:
\enumerate{
\item Flat areas: Focal area has approximately the same elevation as surrounding areas
\item Pits: An area is lower than all other surrounding areas
\item Valley: Focal area has elevation similar to two opposing side of the window but lower than the other two opposing sides
\item Footslope: Focal region is at the "bottom" of a slope
\item Hollow: A small valley/indention in the crest of a hill
\item Slope: Cells in the window form an approximately uniform slope
\item Spur: An extrusion at the foot of a hill (i.e.,, a small hill extending out from the foot of a slope)
\item Shoulder: The crest of a slope
\item Ridge: Opposite of a valley; focal area is higher than two opposing sides but approximately the same elevation as the other two opposing sides
\item Peak: Focal area is higher than any other in the window
}
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster:
elev <- fast(madElev)

# Geomorphons:
geos <- geomorphons(elev)
geos
levels(geos) # levels
freq(geos) # frequencies across cells

col <- c("gray90", "red", "orange", "blue", "green", "pink", "firebrick",
  "purple", "gray50", "black")
plot(geos, col = col)

}
}
\seealso{
\strong{GRASS} manual for tool \code{r.geomorphon} (see \code{grassHelp("r.geomorphon")})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distance.r
\name{distance,GRaster,missing-method}
\alias{distance,GRaster,missing-method}
\alias{distance}
\alias{distance,GRaster,GVector-method}
\alias{distance,GVector,GVector-method}
\alias{distance,GVector,missing-method}
\title{Geographic distance}
\usage{
\S4method{distance}{GRaster,missing}(
  x,
  y,
  target = NA,
  fillNA = TRUE,
  unit = "meters",
  method = ifelse(is.lonlat(x), "geodesic", "Euclidean"),
  minDist = NULL,
  maxDist = NULL
)

\S4method{distance}{GRaster,GVector}(
  x,
  y,
  fillNA = TRUE,
  thick = TRUE,
  unit = "meters",
  method = ifelse(is.lonlat(x), "geodesic", "Euclidean"),
  minDist = NULL,
  maxDist = NULL
)

\S4method{distance}{GVector,GVector}(x, y, unit = "meters", minDist = NULL, maxDist = NULL)

\S4method{distance}{GVector,missing}(x, y, unit = "meters", minDist = NULL, maxDist = NULL)
}
\arguments{
\item{x}{A \code{GRaster} or \code{GVector}.}

\item{y}{Either missing, or a \code{GVector}.}

\item{target}{Numeric: Only applicable for case 1, when \code{x} is a \code{GRaster} and \code{y} is missing.  If this is \code{NA} (default), then cells with \code{NA}s have their values replaced with the distance to the nearest non-\code{NA} cell. If this is another value, then cells with these values have their values replaced with the distance to any other cell (meaning, both \code{NA} and non-\code{NA}, except for cells with this value).}

\item{fillNA}{Logical: Determines which raster cells to fill with distances.
\itemize{
\item Case 1, when \code{x} is a \code{GRaster} and \code{y} is missing: If \code{TRUE} (default), fill values of \code{NA} cells with distances to non-\code{NA} cells. If \code{FALSE}, fill non-\code{NA} cells width distance to \code{NA} cells.
\item Case 2, when \code{x} is a \code{GRaster} and \code{y} is a \code{GVector}: If \code{TRUE} (default), then the returned raster will contain the distance from the cell to the closest feature in the vector. If \code{FALSE}, then cells covered by the vector will have their values replaced with the distance to the nearest cell not covered, and cells that are not covered by the vector will have values of 0.
\item Case 3, when \code{x} is a \code{GVector} and \code{y} is a \code{GVector}: This argument is not used in this case.
}}

\item{unit}{Character: Units of the output. Any of:
\itemize{
\item \code{"meters"}, \code{"metres"}, or \code{"m"} (default)
\item \code{"kilometers"} or \code{"km"}
\item \code{"miles"} or \code{"mi"}
\item \code{"nautical miles"} or \code{"nmi"}
\item \code{"yards"} or \code{"yd"}
\item \code{"feet"} or \code{"ft"} -- international, 1 foot exactly equals 0.3048 meters
}

Partial matching is used and case is ignored.}

\item{method}{Character: The type of distance to calculate. Partial matching is used and capitalization is ignored. Possible values include:
\itemize{
\item \code{Euclidean} (default for projected rasters): Euclidean distance.
\item \code{geodesic} (default for unprojected rasters): Geographic distance. If \code{x} is unprojected (e.g., WGS84 or NAD83), then the \code{method} must be \code{"geodesic"}.
\item \code{squared}: Squared Euclidean distance (faster than just Euclidean distance but same rank--good for cases where only order matters).
\item \code{maximum}: Maximum Euclidean distance.
\item \code{Manhattan}: Manhattan distance (i.e., "taxicab" distance, distance along cells going only north-south and east-west and never along a diagonal).
}}

\item{minDist, maxDist}{Either \code{NULL} (default) or numeric values: Ignore distances less than or greater than these distances.}

\item{thick}{Logical: Only applicable for case 2, when \code{x} is a \code{GRaster} and \code{y} is a \code{GVector}. If \code{TRUE} (default), then the vector will be represented by "thickened" lines (i.e., any cell that the line/boundary touches, not just the ones on the rendering path).}
}
\value{
If \code{x} is a \code{GRaster}, then the output is a \code{GRaster}. If \code{x} is a \code{GVector}, then the output is a numeric vector.
}
\description{
This function produces a raster or a matrix of geographic distances, depending on the input:

\strong{Case 1: Argument \code{x} is a \code{GRaster} and \code{y} is missing:} By default, this function replaces values in all \code{NA} cells with the distance between them and their closest non-\code{NA} cell. Alternatively, all non-\code{NA} cells can have their values replaced by the distance to \code{NA} cells. You can also specify which cells (by value) have their values replaced by distance to other cells.

\strong{Case 2: Argument \code{x} is a \code{GRaster} and \code{y} is a \code{GVector}:} All cells in the raster have their value replaced by the distance to the nearest features in the \code{GVector}. Alternatively, calculate the distance from any cell covered by a vector object and the nearest cell \emph{not} covered by a vector object. Note that the vector is rasterized first.

\strong{Case 3: Argument \code{x} is a \code{GVector} and \code{y} is a \code{GVector}:} A matrix of pairwise distances between all features in one versus the other \code{GVector} is returned.

\strong{Case 4: Argument \code{x} is a \code{GVector} and \code{y} is missing:} A matrix of pairwise distances between all features in the \code{GVector} is returned.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Elevation raster, rivers vector, locations of Dypsis plants
madElev <- fastData("madElev")
madRivers <- fastData("madRivers")
madDypsis <- fastData("madDypsis")

# Convert a SpatRaster to a GRaster, and sf to a GVector
elev <- fast(madElev)
rivers <- fast(madRivers)
dypsis <- fast(madDypsis)

### case 1: GRaster by itself

# Distance between NA cells and nearest non-NA cells
naDist <- distance(elev)
names(naDist) <- "NA Distance"
plot(naDist)

# Distance between non-NA cells and nearest NA cells
nonNaDist <- distance(elev, fillNA = FALSE)
names(nonNaDist) <- "non-NA Distance"
plot(nonNaDist)

# Distance between cells with an elevation of 3 and any other cell that != 3
distFocal3 <- distance(elev, target = 3)
names(distFocal3) <- "Distance from 3"
plot(distFocal3)

# Distance between any cell and cells with a value of 3
distTo3 <- distance(elev, fillNA = FALSE, target = 3)
names(distTo3) <- "Distance to 3"
plot(distTo3)

### Case 2: GRaster and GVector
distToVect <- distance(elev, rivers)

plot(distToVect)
plot(rivers, add = TRUE)

### Case 3: GVector vs GVector
plot(rivers)
plot(dypsis, add = TRUE)

distToRivers <- distance(dypsis, rivers, unit = "yd")
distToPlants <- distance(rivers, dypsis)
distToRivers
distToPlants

### Case 4: GVector vs itself
distToItself <- distance(dypsis)
distToItself

}
}
\seealso{
\code{\link[terra:distance]{terra::distance()}}; \strong{GRASS} modules \code{r.grow.distance} and \code{v.distance}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cellSize.r
\name{cellSize,GRaster-method}
\alias{cellSize,GRaster-method}
\alias{cellSize}
\title{Area of GRaster cells}
\usage{
\S4method{cellSize}{GRaster}(x, mask = FALSE, lyrs = FALSE, unit = "meters2")
}
\arguments{
\item{x}{A \code{GRaster}.}

\item{mask}{Logical: If \code{TRUE}, then cells that are \code{NA} in \code{x} are also \code{NA} in the output.}

\item{lyrs}{Logical:
\itemize{
\item If \code{lyrs} is \code{FALSE} (default), then the output has a single layer. In this case, if \code{mask} is \code{TRUE}, then cells that are \code{NA} in the \emph{first} layer are also \code{NA} in the output.
\item If \code{lyrs} is \code{TRUE}, then the output has the same number of layers as \code{x} if \code{mask} is also \code{TRUE}. In this case, cells that area \code{NA} in the input will also be \code{NA} in the output in the respective layer.
}}

\item{unit}{Character: Units of area. Partial matching is used, and case is ignored. Can be any of:
\itemize{
\item \code{"meters2"} (default), \code{"metres2"}, or \code{"m2"}
\item \code{"km2"} or \code{"kilometers2"}
\item \code{"ha"} or \code{"hectares"}
\item \code{"ac"} or \code{"acres"}
\item \code{"mi2"} or \code{"miles2"}
\item \code{"ft2"} or \code{"feet2"}
}}
}
\value{
A \code{GRaster}.
}
\description{
\code{cellArea()} returns a raster will cell values equal to their area. To get the area of all cells of a raster, see \code{\link[=expanse]{expanse()}}.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Elevation raster
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster:
elev <- fast(madElev)

# Cell size, no masking, single layer
cs1 <- cellSize(elev)
plot(cs1)

# Cell size, with masking, single layer
cs2 <- cellSize(elev, mask = TRUE)
plot(cs2)

# Cell size, no masking, multilayer
elev2 <- c(elev, log(elev - 200))
cs3 <- cellSize(elev2)
plot(cs3)

# Cell size, masking by 1st layer, multilayer (ignores subsequent layers)
cs4 <- cellSize(elev2, mask = TRUE)
plot(cs4)

# Cell size, masking by each layer, multilayer
cs5 <- cellSize(elev2, mask = TRUE, lyrs = TRUE)
plot(cs5)

}
}
\seealso{
\code{\link[terra:cellSize]{terra::cellSize()}}, \code{\link[=expanse]{expanse()}}, \code{\link[=zonalGeog]{zonalGeog()}}, \code{\link[omnibus:convertUnits]{omnibus::convertUnits()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rWalkRast.r
\name{rWalkRast,GRaster-method}
\alias{rWalkRast,GRaster-method}
\alias{rWalkRast}
\title{Create raster representing one or more random walks}
\usage{
\S4method{rWalkRast}{GRaster}(
  x,
  n = 1,
  steps = 1e+05,
  directions = 8,
  avoid = FALSE,
  sameStart = FALSE,
  seed = NULL,
  check = TRUE
)
}
\arguments{
\item{x}{A \code{GRaster} to serve as a template.}

\item{n}{Numeric: Number of walkers. Default  is 1.}

\item{steps}{Numeric: Number of steps taken by each walker. Default is 100000.}

\item{directions}{Either 4 or 8: Directions in which a walker can turn at any point. If 4, then walks are confined to north/south/east/west directions (Rook's case). If 8, then the cardinal and subcardinal directions are allowed (Queen's case).}

\item{avoid}{Logical: If \code{FALSE} (default), then walkers can traverse their own walks. If \code{TRUE}, walkers avoid their own trails. A self-avoiding random walk can take much longer to compute.}

\item{sameStart}{Logical: If \code{FALSE} (default), walkers can begin anywhere. If \code{TRUE}, then walkers start from the same place.}

\item{seed}{Integer or \code{NULL} (default): If \code{NULL}, then a random seed is generated by the function. If numeric, results will be deterministic. In either case, the value will be rounded to the nearest integer.}

\item{check}{Logical: If \code{TRUE} (default), function will check to see if the addon \code{r.random.walk} has been installed. If it has not, it will attempt to install it.}
}
\value{
A \code{GRaster} with cell values representing the number of times one or more walkers traversed the cell.
}
\description{
This function creates a raster where the cell values represent the number of times one or more random "walkers" traverse the cell. If you simulate multiple random walkers, you can do computation in parallel, which can be controlled by allowing \strong{fasterRaster} to use multiple cores and more memory using the "cores" and "memory" arguments in the \code{\link[=faster]{faster()}} function.
}
\details{
This function needs the \strong{GRASS} addon \code{r.random.walk}. If it is not installed, it will try to install it.#'
}
\examples{
if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Elevation raster
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster:
elev <- fast(madElev)

### Create a raster with values drawn from a uniform distribution:
unif <- rUnifRast(elev)
plot(unif)

### Create a raster with values drawn from a normal distribution:
norms <- rNormRast(elev, n = 2, mu = c(5, 10), sigma = c(2, 1))
plot(norms)
hist(norms, bins = 100)

# Create a raster with random, seemingly normally-distributed values:
rand <- rSpatialDepRast(elev, dist = 1000)
plot(rand)

# Values appear normal on first inspection:
hist(rand)

# ... but actually are patterned:
hist(rand, bins = 100)

# Create a fractal raster:
fractal <- fractalRast(elev, n = 2, dimension = c(2.1, 2.8))
plot(fractal)
hist(fractal)

### Random walker rasters

# One random walker
walk <- rWalkRast(elev)
plot(walk)

# Random walker with self-avoidance:
walkAvoid <- rWalkRast(elev, steps = 1000, avoid = TRUE, seed = 1)
plot(walkAvoid)

# 10 random walkers:
walk10 <- rWalkRast(elev, n = 10)
plot(walk10)

# 10 random walkers starting in same place:
walkSame10 <- rWalkRast(elev, n = 10, sameStart = TRUE)
plot(walkSame10)


}
}
\seealso{
\code{\link[=rNormRast]{rNormRast()}}, \code{\link[=rUnifRast]{rUnifRast()}}, \code{\link[=rSpatialDepRast]{rSpatialDepRast()}}, \code{\link[=fractalRast]{fractalRast()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/horizonHeight.r
\name{horizonHeight,GRaster-method}
\alias{horizonHeight,GRaster-method}
\alias{horizonHeight}
\title{Horizon height}
\usage{
\S4method{horizonHeight}{GRaster}(
  x,
  units = "radians",
  step = 90,
  northIs0 = TRUE,
  bufferZone = 0,
  distance = 1,
  maxDist = NULL
)
}
\arguments{
\item{x}{A \code{GRaster}.}

\item{units}{Character: Units of the height. Either \code{radians} (default) or \code{degrees}. Partial matching is used.}

\item{step}{Numeric integer between 0 and 360, inclusive: Angle step size (in degrees) for calculating horizon height. The direction in which horizon height is calculated is incremented from 0 to 360, with the last value excluded.}

\item{northIs0}{Logical: If \code{TRUE} (default), horizon height calculated in the 0-degree direction will be facing north,  and proceed clockwise So, under "north orientation", 0 is north, 90 east, 180 south, and 270 west. If \code{FALSE}, angles are in "east orientation", and proceed counterclockwise from east. So, east is 0, north 90, west 180, and south 270. North orientation is the default for this function in \strong{R}, but east orientation is the default in the \strong{GRASS} tool \code{r.horizon}. \strong{Note:} The \code{\link[=sun]{sun()}} function requires aspect to be in east orientation.}

\item{bufferZone}{Numeric >= 0 (default is 0): A buffer of the specified width will be generated around the raster before calculation of horizon angle. If the coordinate system is in longitude/latitude (e.g., WGS84 or NAD83), then this is specified in degrees. Otherwise units are map units (usually meters).}

\item{distance}{Numeric between 0.5 and 1.5, inclusive (default is 1): This determines the step size when searching for the horizon from a given point. The default value of 1 goes cell-by-cell (i.e., search distance step size is one cell width).}

\item{maxDist}{Either \code{NULL} (default) or numeric >= 0: Maximum distance to consider when finding horizon height in meters. If \code{NULL}, the maximum distance is the full extent of the raster. Smaller values can decrease run time but also reduce accuracy.}
}
\value{
A \code{GRaster} with one or more layers. The layers will be named \code{height_}\emph{xyz}, where \emph{xyz} is degrees from north or from east, depending on whether north or east orientation is used.
}
\description{
\code{horizonHeight()} uses a raster representing elevation to calculate the height of the horizon in a particular direction from each cell on a raster. Height is expressed in radians or degrees from the horizontal.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")

# convert a SpatRaster to a GRaster
elev <- fast(madElev)

# calculate horizon height in north and east directions
hhNorth <- horizonHeight(elev)
hhNorth
plot(hhNorth)

# calculate horizon height in east and north directions
hhEast <- horizonHeight(elev, northIs0 = FALSE)
hhEast
plot(hhEast)

}
}
\seealso{
\strong{GRASS} manual page for tool \code{r.horizon} (see \code{grassHelp("r.horizon")})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/longlat.r
\name{longlat,GRaster-method}
\alias{longlat,GRaster-method}
\alias{longlat}
\title{Create longitude/latitude rasters}
\usage{
\S4method{longlat}{GRaster}(x, degrees = TRUE)
}
\arguments{
\item{x}{A \code{GRaster}.}

\item{degrees}{Logical: If \code{TRUE} (default), coordinate values of cells will be in degrees. If \code{FALSE}, and \code{x} is in a projected coordinate reference system, values will represent coordinates in map units (usually meters). Values will always be in degrees when the coordinate reference system is unprojected (e.g., WGS84, NAD83, etc.).}
}
\value{
A \code{GRaster} stack.
}
\description{
\code{longlat()} creates two rasters, one with cell values equal to the longitude of the cell centers, and one with cell values equal to the latitude of the cell centers.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster
elev <- fast(madElev)

# Create longitude/latitude rasters
ll <- longlat(elev)
ll # note units of cell values!

}
}
\seealso{
\code{\link[=init]{init()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/selectRange.r
\name{selectRange,GRaster-method}
\alias{selectRange,GRaster-method}
\alias{selectRange}
\title{Select values from rasters in a stack based on values in another raster}
\usage{
\S4method{selectRange}{GRaster}(x, y)
}
\arguments{
\item{x}{A \code{GRaster}, typically with more than one layer.}

\item{y}{A \code{GRaster} with integer values. The raster will be rounded if it does not. The values are typically between 1 and the number of layers in \code{x}, but wherever they are outside this range, the returned raster will have \code{NA} values.}
}
\value{
A \code{GRaster}.
}
\description{
\code{selectRange()} selects values from \code{GRaster}s in a "stack" based on the values in another "selection" raster. For example, if the stack has three layers (call them A, B, and C), the "selection" raster could have values of 1, 2, or 3 in each cell. The raster that is returned will have values from A wherever the selection raster is 1, B from where it is 2, and C from where it is 3.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster
elev <- fast(madElev)

# Make a stack of various versions of "elev" from which to select from:
x <- c(elev, 10 * elev, ln(elev), -1 * elev)
x

# Make a layer with random numbers between 1 and 4:
fun <- "= round(rand(0.5, 4.5))"
y <- app(elev, fun = fun)

selected <- selectRange(x, y)

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ext.r
\name{.ext}
\alias{.ext}
\title{Function to get extent from a "sources" name of a raster or vector}
\usage{
.ext(x, rastOrVect = NULL)
}
\arguments{
\item{x}{A \code{GRaster}, \code{GSpatial}, or a character (\code{\link[=sources]{sources()}} name of a \code{GRaster} or \code{GVector}).}

\item{rastOrVect}{Either \code{NULL} (class taken from \code{x}, but \code{x} cannot be a character), or "\code{raster}" or "\code{vector}" (partial matching is used).}
}
\value{
A numeric vector.
}
\description{
Function to get extent from a "sources" name of a raster or vector
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/init.r
\name{init,GRaster-method}
\alias{init,GRaster-method}
\alias{init}
\title{GRaster with values equal to row, column, coordinate, regular, or "chess"}
\usage{
\S4method{init}{GRaster}(x, fun, odd = TRUE, vals = c(0, 1))
}
\arguments{
\item{x}{A \code{GRaster} to be used as a template.}

\item{fun}{Character: Any of:
\itemize{
\item \code{"x"} or \code{"y"}: Cell longitude or latitude
\item \code{"row"} or \code{"col"}: Cell row or column
\item \code{"chess"}: Alternating values.
\item \code{"regular"}: Evenly-spaced cells with the same value.
}}

\item{odd}{Logical: If \code{TRUE} (default), and \code{fun} is \code{"chess"}, then the top left cell in the raster will be a "negative" cell. If \code{FALSE}, then the top left cell with be "positive".}

\item{vals}{Vector of two numeric values: If \code{fun} is \code{"chess"} or \code{"regular"}, then assign the first value to "positive" cells and the second value to "negative" cells. The default is \code{c(1, 0)}}
}
\value{
A \code{GRaster} with as many layers as \code{x}.
}
\description{
This function can be used to make a \code{GRaster} with cell values equal to the cell center's longitude, latitude, row, or column, or in a "chess"-like or "regular" pattern.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Elevation raster, rivers vector
madElev <- fastData("madElev")

# Convert to a GRaster
elev <- fast(madElev)

# Cell coordinates
init(elev, "x")
init(elev, "y")

# Cell row or column
init(elev, "row")
init(elev, "col")

# Chess
elevAgg <- aggregate(elev, 32) # make cells bigger so we can see them

chessOdd <- init(elevAgg, "chess")
chessEven <- init(elevAgg, "chess", odd = FALSE)

chess <- c(chessOdd, chessEven)
names(chess) <- c("odd", "even")
plot(chess)

# Chess with user-defined values
elevAgg <- aggregate(elev, 32) # make cells bigger so we can see

chessOdd13 <- init(elevAgg, "chess", vals = c(0, 13))
chessEven13 <- init(elevAgg, "chess", odd = FALSE, vals = c(0, 13))

chess13 <- c(chessOdd13, chessEven13)
names(chess13) <- c("odd", "even")
plot(chess13)

# Regular
elevAgg <- aggregate(elev, 32) # make cells bigger so we can see

regOdd <- init(elevAgg, "regular")
regEven <- init(elevAgg, "regular", odd = FALSE)

reg <- c(regOdd, regEven)
names(reg) <- c("odd", "even")
plot(reg)

}
}
\seealso{
\code{\link[terra:init]{terra::init()}}, \code{\link[=longlat]{longlat()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/neighborhoodMatrix.r
\name{neighborhoodMatrix,GVector-method}
\alias{neighborhoodMatrix,GVector-method}
\alias{neighborhoodMatrix}
\alias{neighbourhoodMatrix,GVector-method}
\alias{neighbourhoodMatrix}
\title{Neighborhood matrix from a polygons GVector}
\usage{
\S4method{neighborhoodMatrix}{GVector}(x)

\S4method{neighbourhoodMatrix}{GVector}(x)
}
\arguments{
\item{x}{A polygons `GVector.}
}
\value{
A \code{list}. Each element represents a polygon. If an element is empty, it has no neighbors. Otherwise, it is a vector of integers, which represent the indices of the polygon(s) to which it is a neighbor.
}
\description{
This function returns a neighborhood matrix from a polygons \code{GVector}, which represents which geometries touch one another. It is useful for implementing geostatistical analyses that require indicators about which area features are next to one another.

Polygons must share more than one point for them to be considered a neighbors (i.e., same as \code{spdep::poly2nb(x, queen = FALSE)}).

This function needs the \strong{GRASS} addon \code{v.neighborhoodmatrix}. If it is not installed, it will try to install it.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)

# Polygons vector:
madCoast4 <- fastData(madCoast4)
mc4 <- fast(madCoast4)

neighs <- neighborhoodMatrix(mc4)
neighs

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/00a_GLocation_class.r, R/00b_GSpatial_class.r,
%   R/00c_GRegion_class.r, R/00d_GRaster_class.r, R/00e_GVector_class.r
\docType{class}
\name{GLocation-class}
\alias{GLocation-class}
\alias{GLocation}
\alias{GSpatial-class}
\alias{GSpatial}
\alias{GRegion-class}
\alias{GRegion}
\alias{GRaster-class}
\alias{GRaster}
\alias{GVector-class}
\alias{GVector}
\title{Classes for fasterRaster sessions, regions, rasters, and vectors}
\value{
An object of class \code{GLocation}, \code{GSpatial}, \code{GRegion}, \code{GRaster}, or \code{GVector}.
}
\description{
The \code{G}-suite of S4 classes contain pointers to \strong{GRASS} objects or metadata about the current \strong{GRASS} session. Most users will manipulate objects using these classes, but do not need to know the details.
\itemize{
\item The \code{GLocation} class stores information about the \strong{GRASS} "project"/"location"(see \code{vignette("projects_mapsets", package = "fasterRaster")}), and coordinate reference system. Contained by all the rest.
\item The \code{GSpatial} class contains the \code{GLocation} class and stores information about spatial objects (extent, topology) plus the name of the file representing it in \strong{GRASS} (its \code{source}). Contained by \code{GRegion}, \code{GRaster}, and \code{GVector}.
\item The \code{GRegion} class contains the \code{GSpatial} class and stores information about grids (dimensions and resolution). They do have \code{sources}, but these are not used (they're always \code{NA}). Contained by \code{GRaster}. The \code{GRegion} corresponds to \strong{GRASS} "regions", though \code{GRegion} objects are not actually pointers to \strong{GRASS} "region" files (see \code{vignette("regions", package = "fasterRaster")}).
\item The \code{GRaster} class contains the \code{GRegion} class and represents rasters. It stores information on number of layers, categories, min/max values, and user-friendly names. Categorical \code{GRaster}s are associated with a "levels" table for representing categorical data (e.g., wetlands, forest, etc.).
\item The \code{GVector} class contains the \code{GSpatial} class and represents spatial vectors. It may or may not have an associated \code{data.table} (i.e., a \code{data.frame}), which contains metadata about each geometry in the vector.
}
}
\section{Slots}{

\describe{
\item{\code{location}}{Character (all classes): The \strong{GRASS} "project"/"location" of the object. The default value is \code{default}. Can be obtained using the hidden function \code{.location()}. See \code{vignette("projects_mapsets", package = "fasterRaster")}.}

\item{\code{mapset}}{Character (all classes): The \strong{GRASS} "mapset". Default value is \code{PERMANENT}. Typically hidden to users. Can be obtained using the hidden function \code{.mapset()}. See \code{vignette("projects_mapsets", package = "fasterRaster")}.}

\item{\code{workDir}}{Character (all classes): Directory in which \strong{GRASS} stores files.}

\item{\code{topology}}{Character (\code{GSpatial} objects, including \code{GRegion}s, \code{GRaster}s, and \code{GVector}s): Valid values are \verb{2D} (2-dimensional--most rasters and vectors) or \verb{3D} (3-dimensional--e.g., LIDAR data). Can be obtained using \code{\link[=topology]{topology()}}.}

\item{\code{sources}}{Character (\code{GRaster}s and \code{GVector}s): Name of the object in \strong{GRASS}. These are typically made on-the-fly and provide the pointer to the object from \strong{R} to \strong{GRASS}. Changing them manually will break the connection. Can be obtained using \code{\link[=sources]{sources()}}.}

\item{\code{names}}{Character  (\code{GRaster}s only): Name of a raster or each raster layer in. Can be obtained using \code{\link[=names]{names()}}.}

\item{\code{crs}}{Character (all classes): Coordinate reference systems string (preferably in WKT2 format). Can be obtained using \code{\link[=crs]{crs()}} or \code{\link[=st_crs]{st_crs()}}.}

\item{\code{projection}}{Character: The \strong{GRASS} "projection" for a \code{GRaster} or \code{GVector}. Can be obtained using \code{.projection()}.}

\item{\code{dimensions}}{Dimensions:
\itemize{
\item \code{GRegion}s and \code{GRaster}s: Vector of three integers indicating number of rows, columns, and depths (for 3D objects). Can be obtained using \code{\link[=dim]{dim()}}, plus \code{\link[=nrow]{nrow()}}, \code{\link[=ncol]{ncol()}}, and \code{\link[=ndepth]{ndepth()}}.
\item \code{GVectors}s: Vector of two integers indicating number of geometries and number of fields. Can be obtained using \code{\link[=dim]{dim()}}, plus \code{\link[=nrow]{nrow()}} and \code{\link[=ncol]{ncol()}}.
}}

\item{\code{extent}}{Numeric vector with four values (\code{GSpatial} objects, including \code{GRegion}s, \code{GRaster}s, and \code{GVector}s): Extent of the object listed in order from westernmost longitude, easternmost longitude, southernmost latitude, northernmost latitude. Can be obtained using \code{\link[=ext]{ext()}}.}

\item{\code{zextent}}{Numeric (\code{GSpatial} objects, including \code{GRegion}s, \code{GRaster}s, and \code{GVector}s): Bottom- and top-most extents of 3D \code{GRaster}s and \code{GVector}s. Can be obtained using \code{\link[=zext]{zext()}}.}

\item{\code{geometry}}{Character (\code{GVectors}s): Either \code{points}, \code{lines}, or \code{polygons}. Can be obtained using \code{\link[=geomtype]{geomtype()}}.}

\item{\code{nLayers}}{Integer (\code{GRaster}s): Number of layers ("stacked" rasters--different from number of depths of 3D rasters). Can be obtained using \code{\link[=nlyr]{nlyr()}}.}

\item{\code{nGeometries}}{Integer (\code{GVector}s): Number of features (points, lines, or polygons). Can be obtained using \code{\link[=nrow]{nrow()}}.}

\item{\code{datatypeGRASS}}{Character (\code{GRaster}s): Type of data stored in a raster, as interpreted by \code{GRASS}. This is either \code{CELL} (integers), \code{FCELL} (floating-point values), or \code{DCELL} (double-values). Can be obtained using \code{\link[=datatype]{datatype()}}.}

\item{\code{resolution}}{Vector of two numeric values (\code{GRegion}s, including \code{GRaster}s): Size of a raster cell in the east-west direction and in the north-south direction. Can be obtained using \code{\link[=res]{res()}} and \code{\link[=res3d]{res3d()}}.}

\item{\code{minVal,maxVal}}{Numeric (\code{GRaster}s): Minimum and maximum value across all cells. Can be obtained using \code{\link[=minmax]{minmax()}}.}

\item{\code{activeCat}}{Integer (\code{GRaster}s): Column index of the category labels. Must be >0. Note that from the user's standpoint, 1 is subtracted from this number. So a value if \verb{@activeCat} is \code{2}, then the user would see "1" when printed. Can be obtained using \code{\link[=activeCat]{activeCat()}}.}

\item{\code{levels}}{List of \code{data.table}s (\code{GRaster}s): Tables for categorical rasters. If a raster is not categorical, the \code{data.table} is \code{NULL}, as in \code{data.table(NULL)}. Can be obtained using \code{levels()} or \code{cats()}.}

\item{\code{table}}{\code{data.table} (\code{GVector}s): Table with metadata, one row per geometry (point, line, or plane). If no table is associated with the vector, this must be \code{data.table(NULL)}. The column with the category value is given in \verb{@catName}.}

\item{\code{catName}}{Character (\code{GVector}s): Name of the column in the vector's database that contains category values (integers).}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vDetachDatabase.r
\name{.vDetachDatabase}
\alias{.vDetachDatabase}
\title{Add a database table to a GRASS attribute table}
\usage{
.vDetachDatabase(x)
}
\arguments{
\item{x}{A \code{GVector} or the name of a vector in \strong{GRASS}.}
}
\value{
Invisibly returns the \code{\link[=sources]{sources()}} name of a vector in \strong{GRASS}.
}
\description{
\code{.vDetachDatabase()} detaches the database from a \strong{GRASS} vector and deletes it. This table is meant to be "invisible" to most users--they should use interact with attribute tables using the \code{GVector} slot \verb{@table}. Some functions do require tables (e.g., \code{\link[=extract]{extract()}} and \code{\link[=spatSample]{spatSample()}}). \strong{This function is mostly of use to developers.}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/head.r
\name{head,GVector-method}
\alias{head,GVector-method}
\alias{head}
\alias{tail,GVector-method}
\alias{tail}
\title{Return first or last part of the data frame of a GVector}
\usage{
\S4method{head}{GVector}(x, n = 6L, keepnums = TRUE, ...)

\S4method{tail}{GVector}(x, n = 6L, keepnums = TRUE, ...)
}
\arguments{
\item{x}{A \code{GVector}.}

\item{n}{Integer: Number of rows to display.}

\item{keepnums}{Logical: If no \code{rownames} are present, create them. Default is \code{TRUE}.}

\item{...}{Other arguments.}
}
\value{
A \code{data.table} or \code{data.frame}.
}
\description{
Return the first or last part of a \code{GVector}'s data table.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)

# Example data:
madCoast4 <- fastData("madCoast4")
madRivers <- fastData("madRivers")
madDypsis <- fastData("madDypsis")

# Convert sf vectors to GVectors:
coast <- fast(madCoast4)
rivers <- fast(madRivers)
dypsis <- fast(madDypsis)

# Geographic properties:
ext(rivers) # extent
crs(rivers) # coordinate reference system
st_crs(rivers) # coordinate reference system
coordRef(rivers) # coordinate reference system

# Column names and data types:
names(coast)
datatype(coast)

# Points, lines, or polygons?
geomtype(dypsis)
geomtype(rivers)
geomtype(coast)

is.points(dypsis)
is.points(coast)

is.lines(rivers)
is.lines(dypsis)

is.polygons(coast)
is.polygons(dypsis)

# Number of dimensions:
topology(rivers)
is.2d(rivers) # 2-dimensional?
is.3d(rivers) # 3-dimensional?

# Just the data table:
as.data.frame(rivers)
as.data.table(rivers)

# Top/bottom of the data table:
head(rivers)
tail(rivers)

# Vector or table with just selected columns:
names(rivers)
rivers$NAME
rivers[[c("NAM", "NAME_0")]]
rivers[[c(3, 5)]]

# Select geometries/rows of the vector:
nrow(rivers)
selected <- rivers[2:6]
nrow(selected)

# Plot:
plot(coast)
plot(rivers, col = "blue", add = TRUE)
plot(selected, col = "red", lwd = 2, add = TRUE)

# Vector math:
hull <- convHull(dypsis)

un <- union(coast, hull)
sameAsUnion <- coast + hull
plot(un)
plot(sameAsUnion)

inter <- intersect(coast, hull)
sameAsIntersect <- coast * hull
plot(inter)
plot(sameAsIntersect)

er <- erase(coast, hull)
sameAsErase <- coast - hull
plot(er)
plot(sameAsErase)

xr <- xor(coast, hull)
sameAsXor <- coast / hull
plot(xr)
plot(sameAsXor)

# Vector area and length:
expanse(coast, unit = "km") # polygons areas
expanse(rivers, unit = "km") # river lengths

### Fill holes

# First, we will make some holes by creating buffers around points.
buffs <- buffer(dypsis, 500)

holes <- coast - buffs
plot(holes)

filled <- fillHoles(holes, fail = FALSE)

}
}
\seealso{
\code{\link[terra:headtail]{terra::head()}}, \code{\link[terra:headtail]{terra::tail()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/projection.r
\name{.projection}
\alias{.projection}
\title{Get "GRASS" projection of raster or vector}
\usage{
.projection(x)
}
\arguments{
\item{x}{A \code{GRaster} or \code{GVector}.}
}
\value{
Character.
}
\description{
"Getter" for the @projection slot.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as.points.r
\name{as.points,GRaster-method}
\alias{as.points,GRaster-method}
\alias{as.points}
\alias{as.points,GVector-method}
\title{Convert a GRaster, or lines or polygons GVector to a points vector}
\usage{
\S4method{as.points}{GRaster}(x, values = TRUE)

\S4method{as.points}{GVector}(x)
}
\arguments{
\item{x}{A \code{GRaster}, \code{GVector}.}

\item{values}{Logical: If \code{TRUE} (default), create an attribute table with raster cell values, with one row per point.}
}
\value{
A \code{points} \code{GVector}.
}
\description{
\code{as.points()} converts a \code{GRaster}, or a lines or polygons \code{GVector} to a points \code{GVector}.

For \code{GRasters}, the points have the coordinates of cell centers and are assigned the cells' values. Only non-\code{NA} cells will be converted to points.

For \code{GVectors}, each point will have the attributes of the line or polygon to which it belonged. Points are extracted from each vertex.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Elevation raster, outline of a part of Madagascar, and rivers vector:
madElev <- fastData("madElev")
madCoast0 <- fastData("madCoast0")
madRivers <- fastData("madRivers")

# Convert to GRaster and GVectors:
elev <- fast(madElev)
coast <- fast(madCoast0)
rivers <- fast(madRivers)

# For this example, we will first crop to a small extent.
river <- rivers[1]
elevCrop <- crop(elev, river)
elevPoints <- as.points(elevCrop)
elevPoints

plot(elevCrop)
plot(elevPoints, pch = '.', add = TRUE)

# Extract points from vectors:
coastPoints <- as.points(coast)
riversPoints <- as.points(rivers)

plot(coast)
plot(coastPoints, add = TRUE)

plot(rivers, col = "blue", add = TRUE)
plot(riversPoints, col = "blue", add = TRUE)

}
}
\seealso{
\code{\link[=crds]{crds()}}, \code{\link[=as.lines]{as.lines()}}, \code{\link[=as.polygons]{as.polygons()}}, \code{\link[terra:as.points]{terra::as.points()}}, and modules \code{v.to.points} and \code{r.to.vect} in \strong{GRASS}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/regress.r
\name{regress,GRaster,missing-method}
\alias{regress,GRaster,missing-method}
\alias{regress}
\title{Regression intercept, slope, r2, and t-value across each set of cells}
\usage{
\S4method{regress}{GRaster,missing}(y, x, na.rm = FALSE)
}
\arguments{
\item{y}{A multi-layer \code{GRaster}.}

\item{x}{Ignored.}

\item{na.rm}{Logical: If \code{FALSE}, any series of cells with \code{NA} in at least one cell results in an \code{NA} in the output.}
}
\value{
A multi-layer \code{GRaster}.
}
\description{
This function performs a regression on each set of cells in a multi-layered \code{GRaster}. The output is a \code{GRaster} with the intercept, slope, r^2 value, and Student's t value. The regression formula is as \code{y ~ 1 + x}, where \code{x} is the layer number of each layer (e.g., 1 for the first or top layer in the input \code{GRaster}, 2 for the second or second-to-top layer, etc.). Note that this is restricted version of the functionality in \code{\link[terra:regress]{terra::regress()}}.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Example data
madChelsa <- fastData("madChelsa")

# Convert a SpatRaster to a GRaster
chelsa <- fast(madChelsa)
chelsa # 4 layers

# Central tendency
mean(chelsa)
mmode(chelsa)
median(chelsa)

# Statistics
nunique(chelsa)
sum(chelsa)
count(chelsa)
min(chelsa)
max(chelsa)
range(chelsa)
skewness(chelsa)
kurtosis(chelsa)

stdev(chelsa)
stdev(chelsa, pop = FALSE)
var(chelsa)
varpop(chelsa)

# Which layers have maximum/minimum?
which.min(chelsa)
which.max(chelsa)

# Regression

# Note the intercept is different for fasterRaster::regress().
regress(chelsa)
regress(madChelsa, 1:nlyr(madChelsa))

# Note: To get quantiles for each layer, use global().
quantile(chelsa, 0.1)

# NAs
madForest2000 <- fastData("madForest2000")
forest2000 <- fast(madForest2000)
forest2000 <- project(forest2000, chelsa, method = "near")

chelsaForest <- c(chelsa, forest2000)

nas <- anyNA(chelsaForest)
plot(nas)

allNas <- allNA(chelsaForest)
plot(allNas)

}
}
\seealso{
\code{\link[terra:regress]{terra::regress()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/smoothGeom.r
\name{smoothGeom,GVector-method}
\alias{smoothGeom,GVector-method}
\alias{smoothGeom}
\title{Smooth the geometry of a vector}
\usage{
\S4method{smoothGeom}{GVector}(x, method = "Hermite", dist = NULL, angle = 3)
}
\arguments{
\item{x}{A \code{GVector}.}

\item{method}{Character: Method used to smooth line segments. Partial matching is used, and case does not matter:
\itemize{
\item \code{"Hermite"}: Hermite interpolation (default): Guarantees that the output vector always passes through the original points. This method adds points (possibly many) by constructing cubic splines with points approximately \code{dist} apart. The number of points can be reduced by specifying a smaller value of \code{angle}, which specifies the minimum angle between two successive line segments.
\item \code{"Chaiken"}: Chaiken's algorithm: Guarantees that the new vector always touches the midpoint of each original line segment. The points on the new line are at least \code{dist} apart.
}}

\item{dist}{Numeric > 0 of \code{NULL} (default): Minimum distance (see \code{method}). Units are in map units. If \code{NULL}, then 2\% of the minimum of the x-, y-, and z-extent will be used.}

\item{angle}{Numeric > 0: Maximum angle for the Hermite algorithm. Default is 3.}
}
\value{
A \code{GVector}.
}
\description{
\code{smoothGeom()} makes line segments of a vector appear less angular.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Example data
madRivers <- fastData("madRivers")
rivers <- fast(madRivers)
soam <- rivers[rivers$NAM == "SOAMIANINA"] # select one river for illustration

### Simplify geometry (remove nodes)

vr <- simplifyGeom(soam, tolerance = 2000)
dp <- simplifyGeom(soam, tolerance = 2000, method = "dp")
dpr <- simplifyGeom(soam, tolerance = 2000, method = "dpr", prop = 0.5)
rw <- simplifyGeom(soam, tolerance = 2000, method = "rw")

plot(soam, col = "black", lwd = 3)
plot(vr, col = "blue", add = TRUE)
plot(dp, col = "red", add = TRUE)
plot(dpr, col = "chartreuse", add = TRUE)
plot(rw, col = "orange", add = TRUE)

legend("bottom",
   xpd = NA,
   legend = c(
	  "Original",
      "Vertex reduction",
      "Douglas-Peucker",
      "Douglas-Peucker reduction",
      "Reumann-Witkam"
	),
	col = c("black", "blue", "red", "chartreuse", "orange"),
	lwd = c(3, 1, 1, 1, 1)
)

### Smooth geometry

hermite <- smoothGeom(soam, dist = 2000, angle = 3)
chaiken <- smoothGeom(soam, method = "Chaiken", dist = 2000)

plot(soam, col = "black", lwd = 2)
plot(hermite, col = "blue", add = TRUE)
plot(chaiken, col = "red", add = TRUE)

legend("bottom",
   xpd = NA,
   legend = c(
	  "Original",
      "Hermite",
      "Chaiken"
	),
	col = c("black", "blue", "red"),
	lwd = c(2, 1, 1, 1, 1)
)

### Clean geometry

# Has no effect on this vector!
noDangs <- removeDangles(soam, tolerance = 10000)

plot(soam, col = "black", lwd = 2)
plot(noDangs, col = "red", add = TRUE)

legend("bottom",
   xpd = NA,
   legend = c(
	  "Original",
      "No dangles"
	),
	lwd = c(2, 1),
	col = c("black", "red")
)

}
}
\seealso{
\code{\link[=simplifyGeom]{simplifyGeom()}}, \code{\link[terra:simplify]{terra::simplifyGeom()}}, \link[=breakPolys]{geometry cleaning}, \strong{GRASS} manual page for tool \code{v.generalize} (see \code{grassHelp("v.generalize")})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mapset.r
\name{.mapset,GLocation-method}
\alias{.mapset,GLocation-method}
\alias{.mapset}
\alias{.mapset,missing-method}
\title{GRASS "mapset" of an object or the active session}
\usage{
\S4method{.mapset}{GLocation}(x)

\S4method{.mapset}{missing}(x)
}
\arguments{
\item{x}{A Either:
\itemize{
\item Missing: Reports mapset of currently active \strong{GRASS} "projects"/"locations".
\item A \code{GLocation} object or an object that contains the \code{GLocation} class (i.e., a \code{GSpatial} object: a \code{GRaster} or \code{GVector}): Reports the CRS of the object.
}}
}
\value{
A character string.
}
\description{
\strong{GRASS} "locations" are sets of one or more rasters and/or vectors with the same coordinate reference systems, and may or may not represent the same actual location on Earth. \strong{GRASS} "mapsets" are like subfolders of locations, and are collections of rasters and/or vectors typically related to the same general project. This function returns the mapset of an object or the current mapset. This said, \strong{fasterRaster} always uses the "PERMANENT" mapset, so there is very little reason to use this function as-is. See \code{vignette("projects_mapsets", package = "fasterRaster")}.
}
\seealso{
\strong{GRASS} \href{https://grass.osgeo.org/grass82/manuals/grass_database.html}{locations and mapsets}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/g.region.r
\name{.g.region}
\alias{.g.region}
\title{Call GRASS \code{g.region} tool}
\usage{
.g.region()
}
\value{
Displays current region information for the active "project/location" in \strong{GRASS}.
}
\description{
This function calls the \strong{GRASS} tool \code{g.region} to display information on the region of the current \strong{GRASS} "project".
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/thinLines.r
\name{thinLines,GRaster-method}
\alias{thinLines,GRaster-method}
\alias{thinLines}
\title{Reduce linear features on a raster so linear features are 1 cell wide}
\usage{
\S4method{thinLines}{GRaster}(x, iter = 200)
}
\arguments{
\item{x}{A \code{GRaster}.}

\item{iter}{Numeric integer: Number of iterations (default is 200).}
}
\value{
A \code{GRaster}.
}
\description{
The \code{thinLines()} function attempts to reduce linear features on a raster to just 1 cell wide. You may need to run \code{thinLines()} multiple times on the same raster (or experiment with the \code{iter} argument) to get acceptable output. \code{thinLines()} can be helpful to run on a raster before using \code{\link[=as.lines]{as.lines()}}.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Elevation
madElev <- fastData("madElev")

# Convert to GRaster:
elev <- fast(madElev)

# Thin elevation raster:
thinned <- thinLines(elev, iter = 300)
plot(thinned)

# Convert to lines:
rastToLines <- as.lines(thinned)
plot(rastToLines)

# We can clean this:
cleanLines <- fixDangles(x = rastToLines)
plot(rastToLines, col = "red")
plot(cleanLines, add = TRUE)

}
}
\seealso{
\code{\link[=as.lines]{as.lines()}}, \strong{GRASS} manual page for tool \code{r.thin} (see \code{grassHelp("r.thin")})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tiles.r
\name{tiles,GRaster-method}
\alias{tiles,GRaster-method}
\alias{tiles}
\title{Divide a GRaster into spatially exclusive subsets}
\usage{
\S4method{tiles}{GRaster}(x, n, overlap = 0, verbose = FALSE)
}
\arguments{
\item{x}{A \code{GRaster} with one or more layers.}

\item{n}{Numeric vector: Number of tiles to create. This can be a single number, in which case \code{x} is divided into \code{n} × \code{n} tiles, or two values in which case it is divided into \code{n[1]} × \code{n[2]} tiles (rows x columns).}

\item{overlap}{Numeric vector (default is 0): Number of rows/columns by which to expand the size of tiles so they overlap. This can be a single value or two values. If just one is provided, the tiles will be expanded by \code{overlap} rows and \code{overlaps} columns. If two numbers are provided, the tiles will be expanded by \code{overlap[1]} rows and \code{overlap[2]} columns.}

\item{verbose}{Logical: If \code{TRUE}, display progress. Default is \code{FALSE}. Progress is only displayed if \code{x} is a multi-layer \code{GRaster}.}
}
\value{
If \code{x} has just one layer, then the output is a \code{list} with one element per tile. The \code{\link[=lapply]{lapply()}} and \code{\link[=sapply]{sapply()}} functions can be used to apply functions to each tile in the list. If \code{x} has more than one layer, then the output will be a \code{list} of \code{list}s, with each sub-\code{list} containing the tiles for one \code{GRaster} layer.
}
\description{
This function divides a \code{GRaster} into "tiles" or spatial subsets which can be used for speeding some raster calculations. Tiles can be mutually exclusive or overlap by a user-defined number of cells.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Elevation raster
madElev <- fastData("madElev")

# Convert a SpatRaster to a GRaster:
elev <- fast(madElev)

# Create spatially exclusive tiles:
exclusive <- tiles(elev, n = 2, verbose = TRUE)

startpar <- par(mfrow = c(2, 3))
plot(elev, main = "Original")

for (i in seq_along(exclusive)) {
	plot(exclusive[[i]], ext = elev, main = paste("Tile", i))
}
par(startpar)

# Create tiles that overlap:
overlaps <- tiles(elev, n = 2, overlap = 200, verbose = TRUE)

startpar <- par(mfrow = c(2, 3))
plot(elev, main = "Original")

for (i in seq_along(overlaps)) {
	plot(overlaps[[i]], ext = elev, main = paste("Tile", i))
}
par(startpar)

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/madElev.r
\docType{data}
\name{madElev}
\alias{madElev}
\title{Elevation raster for an eastern portion of Madagascar}
\format{
An object of class \code{SpatRaster}. Values are mean meters above sea level.
}
\source{
\href{https://worldclim.org}{WorldClim Version 2.1}
}
\description{
Elevation raster for an eastern portion of Madagascar.
}
\examples{

### vector data

library(sf)

# For vector data, we can use data(*) or fastData(*):
data(madCoast0) # same as next line
madCoast0 <- fastData("madCoast0") # same as previous
madCoast0
plot(st_geometry(madCoast0))

madCoast4 <- fastData("madCoast4")
madCoast4
plot(st_geometry(madCoast4), add = TRUE)

madRivers <- fastData("madRivers")
madRivers
plot(st_geometry(madRivers), col = "blue", add = TRUE)

madDypsis <- fastData("madDypsis")
madDypsis
plot(st_geometry(madDypsis), col = "red", add = TRUE)

### raster data

library(terra)

# For raster data, we can get the file directly or using fastData(*):
rastFile <- system.file("extdata/madElev.tif", package="fasterRaster")
madElev <- terra::rast(rastFile)

madElev <- fastData("madElev") # same as previous two lines
madElev
plot(madElev)

madForest2000 <- fastData("madForest2000")
madForest2000
plot(madForest2000)

madForest2014 <- fastData("madForest2014")
madForest2014
plot(madForest2014)

# multi-layer rasters
madChelsa <- fastData("madChelsa")
madChelsa
plot(madChelsa)

madPpt <- fastData("madPpt")
madTmin <- fastData("madTmin")
madTmax <- fastData("madTmax")
madPpt
madTmin
madTmax


# RGB raster
madLANDSAT <- fastData("madLANDSAT")
madLANDSAT
plotRGB(madLANDSAT, 4, 1, 2, stretch = "lin")

# categorical raster
madCover <- fastData("madCover")
madCover
madCover <- droplevels(madCover)
levels(madCover) # levels in the raster
nlevels(madCover) # number of categories
catNames(madCover) # names of categories table

plot(madCover)
}
\keyword{Madagascar}
\keyword{elevation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interpSplines.r
\name{interpSplines,GVector,GRaster-method}
\alias{interpSplines,GVector,GRaster-method}
\alias{interpSplines}
\title{Interpolate values at points to a GRaster using splines}
\usage{
\S4method{interpSplines}{GVector,GRaster}(
  x,
  y,
  field,
  method = "bilinear",
  lambda = NULL,
  solver = "Cholesky",
  xlength = NULL,
  ylength = NULL,
  interpolate = TRUE,
  verbose = is.null(lambda)
)
}
\arguments{
\item{x}{A "points" \code{GVector}.}

\item{y}{A \code{GRaster}: The output will have the same extent and resolution as this raster.}

\item{field}{Character or integer or numeric integer: Name or index of the column in \code{x} with values to interpolate. If \code{NULL} and if \code{x} is a 3-dimensional "points" \code{GVector}, then the interpolation will act on the z-coordinate of each point.}

\item{method}{Character: The method to use for interpolation can be either \code{"bilinear"} (default) or \code{"bicubic"}. Partial matching is used.}

\item{lambda}{Either \code{NULL} (default) or numeric value > 0: The Tykhonov regularization parameter. If \code{NULL}, cross-validation will be used to determine the optimal parameter value. Cross-validation can take quite a while. If you use cross-validation, the output will either be a \code{GRaster} or a \code{data.frame}, depending on the value of \code{interpolate}.}

\item{solver}{Character: Type of solver to use. Can be either of \code{"Cholesky"} or \code{"cg"}. Partial matching is used and case is ignored.}

\item{xlength, ylength}{Either \code{NULL} (default), or numeric > 0: Length of the spline step in the x- and y-directions. If \code{NULL}, these will be set to 4 times the length of the extent in the respective direction.}

\item{interpolate}{Logical: If \code{TRUE} (default), then create a \code{GRaster} with interpolated values. If \code{FALSE}, return a table with \code{lambda} values from cross-validation. This argument is ignored if \code{lambda} is a numeric value.}

\item{verbose}{Logical: if \code{TRUE}, display progress.}
}
\value{
Output depends on values of \code{lambda} and \code{interpolate}:
\itemize{
\item \code{lambda} is \code{NULL} and \code{interpolate} is \code{TRUE}: A \code{GRaster} with an attribute named \code{lambdas}. This is a \code{data.frame} with values of \code{lambda} that were assessed, plus \code{mean} (mean residual value) and \code{rms} (root mean square error). You can see the table using \code{attr(output_raster, "lambdas", exact = TRUE)}.
\item \code{lambda} is \code{NULL} and \code{interpolate} is \code{FALSE}: A \code{data.frame} with values of \code{lambdas} that were assessed, plus \code{mean} (mean residual value) and \code{rms} (root mean square error). You can see the table using \code{attr(output_raster, "lambdas", exact = TRUE)}.
\item \code{lambda} is a number (\code{interpolate} is ignored): A \code{GRaster}.
}
}
\description{
This function interpolates values in the data table of a "points" \code{GVector} to a \code{GRaster} using splines with Tykhonov regularization to avoid overfitting.
}
\details{
If you receive the error, "No data within this subregion. Consider increasing spline step values, try increasing the values of \code{xlength} and \code{ylength}.

If cross-validation takes too long, or other warnings/errors persist, you can randomly subsample \code{x} to ~100 points to get an optimum value of \code{lambda} (using \code{interpolate = FALSE}), then use this value in the same function again without cross-validation (setting \code{lambda} equal to this value and \code{interpolate = TRUE}).
}
\seealso{
\code{\link[=interpIDW]{interpIDW()}}, \code{\link[=fillNAs]{fillNAs()}}, \strong{GRASS} tool \code{v.surf.bspline} (see \code{grassHelp("v.surf.bspline")})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geomtype.r
\name{.geomtype}
\alias{.geomtype}
\title{Get geometry type from the sources() name of a vector}
\usage{
.geomtype(x)
}
\arguments{
\item{x}{A \code{GVector} or the \code{\link[=sources]{sources()}} name of one.}
}
\value{
Character ("point", "line", or "area"; i.e., in \strong{GRASS} nomenclature).
}
\description{
Get geometry type from the sources() name of a vector
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trim.r
\name{trim,GRaster-method}
\alias{trim,GRaster-method}
\alias{trim}
\title{Remove rows and columns from a raster that are all NA}
\usage{
\S4method{trim}{GRaster}(x, pad = 0)
}
\arguments{
\item{x}{A \code{GRaster}.}

\item{pad}{Numeric integer: Number of \code{NA} rows and columns to retain. The default is 0.}
}
\value{
A \code{GRaster}.
}
\description{
This function removes any rows and columns from  a \code{GRaster} that are all \code{NA}.

If the \code{GRaster} is a stack of rasters, then the rasters will all be trimmed to the same extent such that none have any rows or columns that are all \code{NA}. In other words, if at least one raster in the stack has a non-\code{NA} cell in a row or column, all rasters will retain that row or column.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Elevation raster
madElev <- fastData("madElev")

# Convert SpatRaster to a GRaster:
elev <- fast(madElev)

# Trim NA rows/columns:
trimmedElev <- trim(elev)
dim(elev)
dim(trimmedElev)

# Trim a "stack" of rasters. We will artificially add NA rows and columns to
# one raster to demonstrate how the trim() function removes only rows/columns
# that are NA for *all* rasters.
elevNAs <- elev

fun <- " = if(col() > ncols() - 200, null(), madElev)"
elevNAs <- app(elevNAs, fun)

# Notice raster is "narrower" because we added NA columns
plot(elevNAs)

elevs <- c(elev, elevNAs)
trimmedElevs <- trim(elevs)
trimmedElevNAs <- trim(elevNAs)

dim(elevs)
dim(trimmedElevNAs)
dim(trimmedElevs)

}
}
\seealso{
\code{\link[terra:trim]{terra::trim()}}, \code{\link[=extend]{extend()}}, and \strong{GRASS} tool \code{g.region}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interpIDW.r
\name{interpIDW,GVector,GRaster-method}
\alias{interpIDW,GVector,GRaster-method}
\alias{interpIDW}
\title{Interpolate values at points to a GRaster using inverse-distance weighting}
\usage{
\S4method{interpIDW}{GVector,GRaster}(x, y, field, nPoints = Inf, power = 2)
}
\arguments{
\item{x}{A "points" \code{GVector}.}

\item{y}{A \code{GRaster} to serve as a template for interpolation: Only points in \code{x} that fall inside the extent of the raster will be used for interpolation. You can increase the extent of a \code{GRaster} using \code{\link[=extend]{extend()}}.}

\item{field}{Character, integer, or numeric integer: Name or index of the column in \code{x} with values to interpolate. If \code{NULL} and if \code{x} is a 3-dimensional "points" \code{GVector}, then the interpolation will act on the z-coordinate of each point.}

\item{nPoints}{Integer or numeric integer: Number of nearest points to use for interpolation. The default is to use all points (\code{Inf}).}

\item{power}{Numeric value > 0: Power to which to take distance when interpolating. The default value is two, so the value of each point used for interpolation is \eqn{1 / d^2} where \emph{d} is distance.}
}
\value{
A \code{GRaster}.
}
\description{
This function interpolates values from a set of points to a raster using inverse distance weighting (IDW).
}
\seealso{
\code{\link[terra:interpIDW]{terra::interpIDW()}}, \code{\link[=interpSplines]{interpSplines()}}, \code{\link[=fillNAs]{fillNAs()}}, \strong{GRASS} tool \code{v.surf.idw} (se \code{grassHelp("v.surf.idw")})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multivarEnvSim.r
\name{multivarEnvSim,GRaster,GRaster-method}
\alias{multivarEnvSim,GRaster,GRaster-method}
\alias{multivarEnvSim}
\alias{multivarEnvSim,GRaster,missing-method}
\alias{multivarEnvSim,data.frame,GRaster-method}
\alias{multivarEnvSim,data.table,GRaster-method}
\alias{multivarEnvSim,matrix,GRaster-method}
\title{Multivariate environmental similarity surface (MESS)}
\usage{
\S4method{multivarEnvSim}{GRaster,GRaster}(ref, proj)

\S4method{multivarEnvSim}{GRaster,missing}(ref, proj)

\S4method{multivarEnvSim}{data.frame,GRaster}(ref, proj, na.rm = FALSE)

\S4method{multivarEnvSim}{data.table,GRaster}(ref, proj, na.rm = FALSE)

\S4method{multivarEnvSim}{matrix,GRaster}(ref, proj, na.rm = FALSE)
}
\arguments{
\item{ref}{A \code{data.frame}, \code{data.table}, a points \code{GVector}, or "stack" of \code{GRasters}: This represents the set of "reference" environmental conditions:
\itemize{
\item \code{data.frame} or \code{data.table}: There must be one column per layer in \code{proj}, and the columns must have the same names as the layers in \code{proj}.
\item \code{GRaster} with one or more layers: Must have the same \code{\link[=names]{names()}} as the \code{GRaster}s in \code{proj}. Values are assumed to be continuous (not categorical/factors).
}}

\item{proj}{A \code{GRaster} or missing. If a \code{GRaster}, it must have the same layers as can have with one or more layers as \code{ref}. Values are assumed to be continuous (not categorical/factors). If missing, then \code{ref} is used, in which case the output represents the relative difference of each cell from the overall layer median.}

\item{na.rm}{Logical: If \code{FALSE} (default), and \code{ref} is a \code{data.frame}, \code{data.table}, or \code{matrix}, and has \code{NA}s in it, then the function will likely fail.}
}
\value{
A \code{GRaster} "stack". There will be one layer per layer in \code{ref}, indicating the MESS score for that variable. There will also be a layer named "MESS" which represents the MESS value across all variables (the minimum value of each of the individual MESS rasters). A final layer represents the layer which is most different (has the lowest MESS value).
}
\description{
The multivariate environmental similarity surface (MESS) indicates the degree to which a set of "projection" environmental conditions fall inside or outside a set of "reference" conditions. Values of 1 indicate a location falls at the exact median of all variables. Values of 0 indicate that the location has at least one environmental covariate that is at the upper or lower end of the range of reference conditions, and values <1 indicate that at least one variable falls above or below the reference conditions. MESS can be used, for example, to indicate the degree to which a model constructed in one time period and/or location must extrapolate when projected to another time period and/or location.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Climatic conditions with 4 variables
madChelsa <- fastData("madChelsa")
chelsa <- fast(madChelsa)

# Simulate new conditions by multiplying values by (1 + small number)
proj <- 1.05 * chelsa
names(proj) <- names(chelsa) # proj must have same names as ref

messes <- multivarEnvSim(ref = chelsa, proj = proj)
plot(messes)

# Where is at least one variable outside the reference range?
extrap <- messes[["MESS"]] < 0
levs <- data.frame(value = 0:1, labels = c('no extrapolation', 'extrapolation'))
levels(extrap) <- levs
plot(extrap)

# Using a data frame as "reference" conditions:
madDypsis <- fastData("madDypsis") # Dypsis occurrences
dypsis <- fast(madDypsis)

dypEnv <- extract(chelsa, dypsis)
dypMess <- multivarEnvSim(ref = dypEnv, proj = proj)
plot(dypMess)

}
}
\references{
Elith, J, Kearney, M, and Phillips, S. 2010. The art of modelling range-shifting species. \emph{Methods in Ecology and Evolution} 1:330-342. \doi{10.1111/j.2041-210X.2010.00036.x} (see especially the Supplement)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vRecat.r
\name{.vRecat}
\alias{.vRecat}
\title{Re-make vector "category" (cat) values}
\usage{
.vRecat(x, gtype, cats = NULL)
}
\arguments{
\item{x}{A \code{GVector} or the \code{\link[=sources]{sources()}} name of a vector in \strong{GRASS}.}

\item{gtype}{Character: Type of vector features in \strong{GRASS} format (i.e., either \code{point}, \code{line}, or \code{area}). See \code{\link[=geomtype]{geomtype()}}.}

\item{cats}{\code{NULL} (default) or \code{integer} or \code{character} vector: Category values of the \strong{GRASS} vector. Supplying these can speed the re-assignment of categories. The values of \code{cats} is \emph{not} used as the new category values. Rather, they will be used to indicate which geometries belong to the same multi-part feature.}
}
\value{
The \code{\link[=sources]{sources()}} name of a vector.
}
\description{
Each geometry in a \strong{GRASS} vector has a "category" number (abbreviated "cat" in output and modules). Geometries can have the same or different numbers, but for functions to work as intended, they often need to have sequential category values, starting at 1, with no skips between integers. This function reconstitutes the category values of a vector in \strong{GRASS} so they being with 1 and have no skips. \strong{This function is mostly of use to developers.}
}
\seealso{
\code{\link[=.vIncrementCats]{.vIncrementCats()}}, \code{\link[=.vCats]{.vCats()}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/complete.cases.r
\name{complete.cases,GRaster-method}
\alias{complete.cases,GRaster-method}
\alias{complete.cases}
\alias{complete.cases,GVector-method}
\alias{missing.cases,GRaster-method}
\alias{missing.cases}
\alias{missing.cases,GVector-method}
\title{Rows of a GRaster or GVector's table that have no NAs or that have NAs}
\usage{
\S4method{complete.cases}{GRaster}(..., levels = TRUE)

\S4method{complete.cases}{GVector}(...)

\S4method{missing.cases}{GRaster}(..., levels = TRUE)

\S4method{missing.cases}{GVector}(...)
}
\arguments{
\item{...}{A \code{GRaster} or \code{GVector}.}

\item{levels}{Logical (\code{GRaster}s only): If \code{TRUE} (default), then assess only the "value" and \code{\link[=activeCat]{activeCat()}} columns of the levels table (see \code{\link[=levels]{levels()}}). If \code{FALSE}, then assess all columns (see \code{\link[=cats]{cats()}}).}
}
\value{
Both \code{complete.cases()} and \code{missing.cases()} return the same type of object. The output depends on the input:
\itemize{
\item A categorical \code{GRaster} with just one layer: A logical vector.
\item An integer, float, or double \code{GRaster} with just one layer: \code{NA}.
\item A \code{GRaster} with multiple layers: A list with one element per layer with either logical vectors or \code{NA}s, as per above.
\item A \code{GVector} with a data table: A logical vector.
\item A \code{GVector} without a data table: \code{NA}.
}
}
\description{
When applied to a categorical \code{GRaster}, \code{compete.cases()} returns \code{TRUE} for each row of the "levels" table that has no \code{NA}s in it. In contrast, \code{missing.cases()} returns \code{TRUE} for each row that has at least one \code{NA} in it. If the raster is not categorical, then \code{NA} is returned.

When applied to a \code{GVector} with a data table, \code{complete.cases()} returns \code{TRUE} for each row where there are no \code{NA}s. if the \code{GVector} has no data table, then a vector of \code{TRUE} values the same length as the total number of geometries will be returned. In contrast, \code{missing.cases()} returns \code{TRUE} for every row that has at least one \code{NA} in it. If the \code{GVector} has no data table, then a vector of \code{FALSE} values is returned.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Plant specimens (points) and land cover
madDypsis <- fastData("madDypsis")
madCover <- fastData("madCover")

# Convert to GVector and GRaster
dypsis <- fast(madDypsis)
cover <- fast(madCover)

### GVector

# Look at the data table:
as.data.table(dypsis)

# Which rows have no NAs?
complete.cases(dypsis)

# Which rows have at least one NA (opposite of above)?
missing.cases(dypsis)

### GRaster

# Look at the levels table:
levels(cover)

# Which rows of levels table have no NAs?
complete.cases(cover)

# Which rows have at least one NA (opposite of above)?
missing.cases(cover)

}
}
\seealso{
\code{\link[=missingCats]{missingCats()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/replaceNAs.r
\name{replaceNAs,data.frame-method}
\alias{replaceNAs,data.frame-method}
\alias{replaceNAs}
\alias{replaceNAs,matrix-method}
\alias{replaceNAs,data.table-method}
\alias{replaceNAs,numeric-method}
\alias{replaceNAs,integer-method}
\alias{replaceNAs,logical-method}
\alias{replaceNAs,character-method}
\title{Replace NAs in a data.table or data.frame column, or in a vector}
\usage{
\S4method{replaceNAs}{data.frame}(x, replace, cols = NULL)

\S4method{replaceNAs}{matrix}(x, replace, cols = NULL)

\S4method{replaceNAs}{data.table}(x, replace, cols = NULL)

\S4method{replaceNAs}{numeric}(x, replace)

\S4method{replaceNAs}{integer}(x, replace)

\S4method{replaceNAs}{logical}(x, replace)

\S4method{replaceNAs}{character}(x, replace)
}
\arguments{
\item{x}{A \code{data.table} or \code{data.frame} or \code{matrix}, or a vector of numeric, integer, logical, or character values.}

\item{replace}{A value of any atomic class (numeric, integer, character, Date, etc.): Value to to replace \code{NA}s.}

\item{cols}{\code{NULL}, character, numeric, integer, or logical vector: Indicates columns for which to replace \code{NA}s. If \code{NULL}, then all columns will have \code{NA}s replaced. If a character, this is the column name(s). If numeric or integer, this is the columns' indices. If logical, columns with \code{TRUE} have \code{NA}s replaced. If a logical vector has fewer than the total number of columns, it will be recycled.}
}
\value{
A \code{data.table}, \code{data.frame}, \code{matrix}, or vector.
}
\description{
This function replaces \code{NA}s in one or more \code{data.table}, \code{data.frame}, or \code{matrix} columns, or in vectors, with a user-defined value.
}
\examples{

library(data.table)

dt <- data.table(
   x = 1:10,
   y = letters[1:10],
   z = rnorm(10)
)

# make some values NA
dt[x == 4 | x == 8, y := NA_character_]
dt

# Replace NAs:
replaceNAs(dt, replace = -99, cols = "y")
dt

# Drop rows:
dropped <- dropRows(dt, 8:10)
dropped

# NB May not print... in that case, use:
print(dropped)

# We can also use replaceNAs() on vectors:
y <- 1:10
y[c(2, 10)] <- NA
replaceNAs(y, -99)

# Same as:
y <- 1:10
y[c(2, 10)] <- NA
y[is.na(y)] <- -99
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/connectors.r
\name{connectors,GVector,GVector-method}
\alias{connectors,GVector,GVector-method}
\alias{connectors}
\title{Create lines connecting nearest features of two GVectors}
\usage{
\S4method{connectors}{GVector,GVector}(x, y, minDist = NULL, maxDist = NULL)
}
\arguments{
\item{x, y}{\code{GVector}s.}

\item{minDist, maxDist}{Either \code{NULL} (default) or numeric values: Ignore features separated by less than or greater than these distances.}
}
\value{
A \code{GVector} with a data table that has the length of each connecting line in meters.
}
\description{
\code{connectors()} creates a lines \code{GVector} which represent the shortest (Great Circle) paths between each feature of one \code{GVector} and the nearest feature of another \code{GVector}.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)

# Rivers vector and locations of Dypsis plants
madRivers <- fastData("madRivers")
madDypsis <- fastData("madDypsis")

# Convert sf's to GVectors:
dypsis <- fast(madDypsis)
rivers <- fast(madRivers)

### Connections from each point to nearest river
consFromDypsis <- connectors(dypsis, rivers)

plot(rivers, col = "blue")
plot(dypsis, add = TRUE)
plot(consFromDypsis, col = "red", add = TRUE)

### Connections from each river to nearest point
consFromRivers <- connectors(rivers, dypsis)

plot(rivers, col = "blue")
plot(dypsis, add = TRUE)
plot(consFromRivers, col = "red", add = TRUE)

}
}
\seealso{
\strong{GRASS} manual for tool \code{v.distance} (see \code{grassHelp("v.distance")}).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/crop.r
\name{crop,GRaster-method}
\alias{crop,GRaster-method}
\alias{crop}
\alias{crop,GVector-method}
\title{Remove parts of a GRaster or GVector}
\usage{
\S4method{crop}{GRaster}(x, y, fail = TRUE)

\S4method{crop}{GVector}(x, y, extent = FALSE, fail = TRUE)
}
\arguments{
\item{x}{A \code{GRaster} or \code{GVector} to be cropped.}

\item{y}{A \code{GRaster} or \code{GVector} to serve as a template for cropping.}

\item{fail}{Logical: If \code{TRUE} (default), and the cropped object would have zero extent in at least one dimension, then exit the function with an error. If \code{FALSE}, then display a warning and return \code{NULL}.}

\item{extent}{Logical:
\itemize{
\item If \code{y} is a "points" \code{GVector}: Use the convex hull around \code{y} to crop \code{x}.
\item If \code{y} is a "lines" or "polygons" \code{GVector}: If \code{TRUE}, use the extent of \code{y} to crop \code{x}.
\item if \code{y} is a \code{GVector}, \code{x} will be cropped to the extent of \code{y} (\code{extent} is ignored).
}}
}
\value{
A \code{GRaster} or \code{GVector} (or \code{NULL} if \code{fail} is \code{FALSE} and the output would have a zero east-west and/or north-south extent).
}
\description{
\code{crop()} removes parts of a \code{GRaster} or \code{GVector} that fall "outside" another raster or vector. You cannot make the \code{GRaster} or \code{GVector} larger than it already is (see \code{\link[=extend]{extend()}}). Rasters may not be cropped to the exact extent, as the extent will be enlarged to encompass an integer number of cells. If you wish to remove certain cells of a raster, see \code{\link[=mask]{mask()}}.
}
\details{
Known differences from \code{\link[terra:crop]{terra::crop()}}:
\itemize{
\item If \code{x} and \code{y} are "points" vectors, and \code{extent} is \code{TRUE}, \strong{terra} removes points that fall on the extent boundary. \strong{fasterRaster} does not remove points on the extent boundary.
\item If \code{x} is a "points" vector and \code{y} is a "lines" vectors, and \code{extent} is \code{FALSE}, \strong{terra} uses the extent of \code{y} to crop the points.  \strong{fasterRaster} uses the minimum convex hull of the lines vector.
}
}
\examples{
if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Elevation raster
madElev <- fastData("madElev")

# Plant specimen points vector
madDypsis <- fastData("madDypsis")

# Rivers lines vector
madRivers <- fastData("madRivers")

# Polygons vector
madCoast4 <- fastData("madCoast4")
madAnt <- madCoast4[madCoast4$NAME_4 == "Antanambe", ]

# Convert to fasterRaster format:
elev <- fast(madElev)
dypsis <- fast(madDypsis)
rivers <- fast(madRivers)
coast <- fast(madCoast4)
ant <- fast(madAnt)

### Crop raster by vector:
rastByVect <- crop(elev, ant)
plot(elev, col = "gray", legend = FALSE)
plot(rastByVect, add = TRUE)
plot(ant, add = TRUE)

### Crop raster by raster:

# For this example, make the SpatRaster smaller, then crop by this.
templateRast <- crop(madElev, madAnt)

template <- fast(templateRast)
rastByRast <- crop(elev, template)

plot(elev, col = "gray", legend = FALSE)
plot(rastByRast, add = TRUE)

### Crop vector by raster:

# For this example, make the SpatRaster smaller, then crop by this.
templateRast <- crop(madElev, madAnt)

template <- fast(templateRast)
ptsByRast <- crop(dypsis, template)

plot(elev, col = "gray", legend = FALSE)
plot(templateRast, add = TRUE)
plot(dypsis, add = TRUE)
plot(ptsByRast, pch = 21, bg = "red", add = TRUE)

### Crop vector by vector:
ptsSubset <- dypsis[1:10] # use first 10 points as cropping template

# Crop points vector by convex hull around points:
ptsByPts <- crop(dypsis, ptsSubset)
plot(dypsis)
plot(convHull(ptsSubset), lty = "dotted", border = "blue", add = TRUE)
plot(ptsByPts, col = "red", add = TRUE)
plot(ptsSubset, col = "blue", pch = 3, cex = 1.6, add = TRUE)
legend("topleft",
    legend = c("Dypsis", "Selected", "Crop template", "Convex hull"),
    pch = c(16, 16, 3, NA),
    lwd = c(NA, NA, NA, 1),
    col = c("black", "red", "blue", "blue"),
    lty = c(NA, NA, NA, "dotted"),
    xpd = NA,
    bg = "white"
)

# Crop points vector by extent of points:
ptsByPts <- crop(dypsis, ptsSubset, ext = TRUE)
plot(dypsis)
plot(ptsByPts, col = "red", add = TRUE)
plot(ptsSubset, col = "blue", pch = 3, cex = 1.6, add = TRUE)
legend("topleft",
    legend = c("Dypsis", "Selected", "Crop template"),
    pch = c(16, 16, 3),
    lwd = c(NA, NA, NA),
    col = c("black", "red", "blue"),
    lty = c(NA, NA, NA),
    xpd = NA,
    bg = "white"
)

# Crop points vector by convex hull around lines:
ptsByPts <- crop(dypsis, rivers)
plot(rivers, col = "blue", pch = 3, cex = 1.6)
plot(dypsis, add = TRUE)
plot(convHull(rivers), lty = "dotted", border = "blue", add = TRUE)
plot(ptsByPts, col = "red", add = TRUE)
legend("topleft",
    legend = c("Dypsis", "Selected", "Rivers", "Convex hull"),
    pch = c(16, 16, NA, NA),
    lwd = c(NA, NA, 1, 1),
    col = c("black", "red", "blue", "blue"),
    lty = c(NA, NA, "solid", "dotted"),
    xpd = NA,
    bg = "white"
)

# Crop points vector by extent of lines:
ptsByPts <- crop(dypsis, rivers, ext = TRUE)
plot(rivers, col = "blue", pch = 3, cex = 1.6)
plot(dypsis, add = TRUE)
plot(convHull(rivers), lty = "dotted", border = "blue", add = TRUE)
plot(ptsByPts, col = "red", add = TRUE)
legend("topleft",
    legend = c("Dypsis", "Selected", "Rivers"),
    pch = c(16, 16, NA),
    lwd = c(NA, NA, 1),
    col = c("black", "red", "blue"),
    lty = c(NA, NA, "solid"),
    xpd = NA,
    bg = "white"
)

# Crop points vector by polygon:
ptsByPts <- crop(dypsis, ant)
plot(dypsis)
plot(ant, border = "blue", pch = 3, cex = 1.6, add = TRUE)
plot(ptsByPts, col = "red", add = TRUE)
legend("topleft",
    legend = c("Dypsis", "Selected", "Antanambe"),
    pch = c(16, 16, NA),
    lwd = c(NA, NA, 1),
    col = c("black", "red", "blue"),
    lty = c(NA, NA, "solid"),
    xpd = NA,
    bg = "white"
)

# Crop lines vector by polygon:
linesByPoly <- crop(rivers, ant)
plot(rivers)
plot(ant, border = "blue", pch = 3, cex = 1.6, add = TRUE)
plot(linesByPoly, col = "red", add = TRUE)
legend("topleft",
    legend = c("Rivers", "Selected", "Antanambe"),
    col = c("black", "red", "blue"),
    lwd = 1,
    xpd = NA,
    bg = "white"
)

# Crop polygon vector by convex hull around points:
polyByPoints <- crop(ant, dypsis)
plot(dypsis, col = "red")
plot(ant, border = "blue", add = TRUE)
plot(polyByPoints, border = "red", add = TRUE)
legend("topleft",
    legend = c("Dypsis", "Antanambe", "Selected"),
    col = c("red", "blue", "red"),
    pch = c(16, NA, NA),
    lwd = c(NA, 1, 1),
    xpd = NA,
    bg = "white"
)

}
}
\seealso{
\code{\link[terra:crop]{terra::crop()}}, \code{\link[sf:st_crop]{sf::st_crop()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatSample.r
\name{spatSample,GRaster-method}
\alias{spatSample,GRaster-method}
\alias{spatSample}
\alias{spatSample,GVector-method}
\title{Sample random points from a GRaster or GVector}
\usage{
\S4method{spatSample}{GRaster}(
  x,
  size,
  as.points = FALSE,
  values = TRUE,
  cats = TRUE,
  xy = FALSE,
  strata = NULL,
  byStratum = FALSE,
  zlim = NULL,
  seed = NULL,
  verbose = FALSE
)

\S4method{spatSample}{GVector}(
  x,
  size,
  as.points = FALSE,
  values = TRUE,
  xy = FALSE,
  byStratum = FALSE,
  zlim = NULL,
  seed = NULL
)
}
\arguments{
\item{x}{A \code{GRaster} or \code{GVector}.}

\item{size}{Numeric value > 0: Number of points to create.}

\item{as.points}{Logical: If \code{FALSE} (default), the output is a \code{data.frame} or \code{data.table}. If \code{TRUE}, the output is a "points" \code{GVector}.}

\item{values}{Logical: If \code{TRUE} (default), values of the \code{GRaster} at points are returned.}

\item{cats}{Logical: If \code{TRUE} (default) and the \code{GRaster} is categorical, then return the category label of each cell. If \code{values} is also \code{TRUE}, then the cell value will also be returned.}

\item{xy}{Logical: If \code{TRUE}, return the longitude and latitude of each point. Default is \code{FALSE}.}

\item{strata}{Either \code{NULL} (default), or a \code{GVector} defining strata. If supplied, the \code{size} argument will be interpreted as number of points to place per geometry in \code{strata}. Note that using strata can dramatically slow the process.}

\item{byStratum}{Logical: If \code{FALSE} (default), then \code{size} number of points will be placed within the entire area delineated by \code{strata}. If \code{TRUE}, then \code{size} points will be placed within each subgeometry of \code{strata}.}

\item{zlim}{Either \code{NULL} (default), or a vector of two numbers defining the lower and upper altitudinal bounds of coordinates. This cannot be combined with \code{values = TRUE} or \code{cats = TRUE}.}

\item{seed}{Either \code{NULL} (default) or an integer: Random number seed. If this is \code{NULL}, the a seed will be set randomly. Values will be rounded to the nearest integer.}

\item{verbose}{Logical: If \code{TRUE}, display progress. Default is \code{FALSE}.}
}
\value{
A \code{data.frame}, \code{data.table}, or \code{GVector}.
}
\description{
\code{spatSample()} randomly locates points across a \code{GRaster} or \code{GVector}. It can return a \code{GVector}, the coordinates, values associated with the points, or all of these. If you want to generate a raster with randomly-sampled cells, see \code{\link[=sampleRast]{sampleRast()}}.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Example data
madElev <- fastData("madElev") # raster

# Convert to GRasters and GVectors
elev <- fast(madElev)

### spatSample()

# Random points as data.frame or data.table:
randVals <- spatSample(elev, size = 20, values = TRUE)
randVals

# Random points as a points GVector:
randPoints <- spatSample(elev, size = 20, as.points = TRUE)
randPoints
plot(elev)
plot(randPoints, add = TRUE)

# Random points in a select area:
madCoast <- fastData("madCoast4") # vector
coast <- fast(madCoast)
ant <- coast[coast$NAME_4 == "Antanambe"] # subset

restrictedPoints <- spatSample(elev, size = 20, as.points = TRUE,
   strata = ant)

plot(elev)
plot(ant, add = TRUE)
plot(restrictedPoints, add = TRUE) # note 20 points for entire geometry

# Random points, one set per subgeometry:
stratifiedPoints <- spatSample(elev, size = 20, as.points = TRUE,
   strata = ant, byStratum = TRUE)

plot(elev)
plot(ant, add = TRUE)
plot(stratifiedPoints, pch = 21, bg = "red", add = TRUE) # note 20 points per subgeometry

# Random categories:
madCover <- fastData("madCover") # raster
cover <- fast(madCover)

randCover <- spatSample(cover, size = 20, values = TRUE,
     cat = TRUE, xy = TRUE)
randCover

### sampleRast()

# Random cells in non-NA cells:
rand <- sampleRast(elev, 10000)
plot(rand)
nonnacell(rand)

# Use custom values for the mask:
randCustomMask <- sampleRast(elev, 10000, maskvalues = 1:20)
plot(randCustomMask)

# Force selected values to a custom value:
randCustomUpdate <- sampleRast(elev, 10000, updatevalue = 7)
plot(randCustomUpdate)

# Custom values for mask and set selected cells to custom value:
randAll <- sampleRast(elev, 10000, maskvalues = 1:20, updatevalue = 7)
plot(randAll)

}
}
\seealso{
\code{\link[=sampleRast]{sampleRast()}}, \code{\link[terra:sample]{terra::spatSample()}}, tool \code{v.random} in \strong{GRASS} (see \code{grassHelp("v.random")})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexagons.r
\name{hexagons,GRaster-method}
\alias{hexagons,GRaster-method}
\alias{hexagons}
\alias{hexagons,GVector-method}
\title{Create a hexagonal grid}
\usage{
\S4method{hexagons}{GRaster}(x, ny = 10, expand = 0, angle = 0)

\S4method{hexagons}{GVector}(x, ny = 10, expand = 0, angle = 0)
}
\arguments{
\item{x}{A \code{GRaster} or \code{GVector}.}

\item{ny}{Integer or numeric integer: Number of rows of hexagons that span the extent of object \code{x}.}

\item{expand}{One or two numeric values: Expand the region by this proportion in both directions (a single value) or in the x- and y-dimensions separately. Expanding the region can be helpful to ensure the entire area of interest is covered by polygons, which can otherwise leave gaps at the edges. The number of rows and columns will be increased, but the number of hexagons that span \code{x} will still be \code{ny}.}

\item{angle}{Numeric: Degrees by which to rotate grid (from north, clockwise).}
}
\value{
A \code{GVector}.
}
\description{
This function creates a \code{GVector} of "wall-to-wall" hexagons. The input can be a \code{GVector} or \code{GRaster}, which provides the extent of the output.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)

# Points vector of specimens of species in the plant genus Dypsis
madCoast0 <- fastData("madCoast0")

# Convert sf to a GVector:
coast <- fast(madCoast0)

### grid

# grid specified by number of cells in x-dimension
g1 <- grid(coast, nx = 10)
plot(coast, col = "cornflowerblue")
plot(g1, add = TRUE)

# grid specified by number of cells in x- and y-dimension
g2 <- grid(coast, nx = 10, ny = 5)
plot(coast, col = "cornflowerblue")
plot(g2, add = TRUE)

# grid specified by size of cells in both dimensions
g3 <- grid(coast, nx = 1250, ny = 2000, use = "size")
plot(coast, col = "cornflowerblue")
plot(g3, add = TRUE)

### hexagons

hexes <- hexagons(coast, ny = 10)
plot(hexes)
plot(coast, lwd = 2, add = TRUE)

hexes <- hexagons(coast, ny = 10, expand = c(0.3, 0.1))
plot(hexes)
plot(coast, lwd = 2, add = TRUE)

}
}
\seealso{
\code{\link[=grid]{grid()}}, tool \code{v.mkgrid} in \strong{GRASS}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/00e_GVector_class.r
\name{makeGVector}
\alias{makeGVector}
\title{Create a GVector from a GRASS vector}
\usage{
makeGVector(
  src,
  table = NULL,
  build = TRUE,
  extensive = FALSE,
  cats = NULL,
  fail = TRUE
)
}
\arguments{
\item{src}{Character: The name of the vector in \strong{GRASS}.}

\item{table}{A \code{data.table}, \code{data.frame}, \code{GVector} with a table, or character. This can be \code{data.table(NULL)} or \code{data.frame(NULL)} if there is no table associated with the vector. If a character, this is interpreted as the name of the table in \strong{GRASS}.}

\item{build}{Logical: If \code{TRUE} (default), build topology using \strong{GRASS} tool \code{v.build}.}

\item{extensive}{Logical: If \code{TRUE}, do extensive topological checks using \code{v.build}. The default is \code{FALSE}.}

\item{cats}{\code{NULL} (default) or an integer vector: Values of the "cats" (categories) of the vector in \strong{GRASS}. This is useful \emph{only} for speeding up the \code{GVector} creation process when the "cats" have already been ascertained. For most users, this will be better left as \code{NULL}.}

\item{fail}{Logical: If \code{TRUE} (default), and the vector either has a 0 east-west or north-south extent, then exit the function with an error. If \code{fail} is \code{FALSE}, then display a warning and return \code{NULL}.}
}
\value{
A \code{GVector} (or \code{NULL} if \code{fail} is \code{TRUE} and the \code{GVector} would be invalid).
}
\description{
Create a \code{GVector} from a vector existing in the current \strong{GRASS} session.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Example data
madElev <- fastData("madElev") # integer raster
madCover <- fastData("madCover") # categorical raster
madCoast4 <- fastData("madCoast4") # polygons vector
madRivers <- fastData("madRivers") # lines vector
madDypsis <- fastData("madDypsis") # points vector

### Create GRasters from SpatRasters

# Create an integer raster:
elev <- fast(madElev)
elev

# Create a categorical raster:
cover <- fast(madCover)
madCover
levels(madCover) # category levels

# Create a GRaster from a file on disk:
rastFile <- system.file("extdata", "madForest2000.tif", package = "fasterRaster")
forest2000 <- fast(rastFile)
forest2000

# Create a 1's raster that spans the world:
ones <- fast(rastOrVect = "raster", crs = "epsg:4326")
ones

### Create GVectors

# Create a GVector from an sf vector:
coast4 <- fast(madCoast4)
coast4

# Create a GVector from a SpatVector:
madRivers <- vect(madRivers)
class(madRivers)
rivers <- fast(madRivers)
rivers

# Create a GVector from a vector on disk:
vectFile <- system.file("extdata/shapes", "madCoast.shp",
   package = "fasterRaster")
coast0 <- fast(vectFile)
coast0

# Import only Dypsis occurrences in a restricted area:
ant <- coast4[coast4$NAME_4 == "Antanambe"]
dypsisRestrict <- fast(madDypsis, extent = ant)
dypsis <- fast(madDypsis)

plot(coast4)
plot(ant, col = "gray80", add = TRUE)
plot(dypsis, add = TRUE)
plot(dypsisRestrict, col = "red", add = TRUE)

# Create a generic GVector that spans the world:
wallToWall <- fast(rastOrVect = "vector", crs = "epsg:4326") # WGS84
wallToWall

# Create a GVector from a numeric vector
pts <- c(-90.2, 38.6, -122.3, 37.9)
pts <- fast(pts, crs = "epsg:4326") # WGS84

# Create a GVector from a matrix (can also use data.frame or data.table):
mat <- matrix(c(-90.2, 38.6, -122.3, 37.9), ncol = 2, byrow = TRUE)
mat <- fast(mat, crs = "epsg:4326", keepgeom = TRUE) # WGS84

}
}
\seealso{
\code{\link[=makeGRaster]{makeGRaster()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zonal.r
\name{.zonalByVector}
\alias{.zonalByVector}
\title{Internal function for zonal() when y is a GVector}
\usage{
.zonalByVector(x, z, fun, probs, gtype)
}
\arguments{
\item{x}{GRaster \code{\link[=sources]{sources()}} name.}

\item{z}{GVector \code{\link[=sources]{sources()}} name.}

\item{fun}{Character}

\item{gtype}{\code{geomtype(z, grass = TRUE)} ("area", "line", or "point")}
}
\description{
Internal function for zonal() when y is a GVector
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid.r
\name{grid,GRaster-method}
\alias{grid,GRaster-method}
\alias{grid}
\alias{grid,GVector-method}
\title{Create a grid GVector}
\usage{
\S4method{grid}{GRaster}(x, nx = NULL, ny = NULL, use = "number", angle = 0)

\S4method{grid}{GVector}(x, nx = NULL, ny = NULL, use = "number", angle = 0)
}
\arguments{
\item{x}{A \code{GRaster} or \code{GVector}.}

\item{nx, ny}{Integer or numeric:
\itemize{
\item If \code{use} is \code{"number"}, then these values represent the number of rows and columns in the grid.
\item If \code{use} is \code{size}, then these values represent the size of the cells in the x- and y-dimensions.
}}

\item{use}{Character: How to generate the grid. If this is \code{number} (default), then \code{nx} and \code{ny} are taken to be the number of grid cells. If \code{size}, then \code{nx} and \code{ny} are taken to be the size of the grid cells.}

\item{angle}{Numeric: Degrees by which to rotate grid (from north, clockwise).}
}
\value{
A \code{GVector}.
}
\description{
This function creates a \code{GVector} of "wall-to-wall" cells (like a lattice). The input can be a \code{GVector} or \code{GRaster}, which provides the extent of the output.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)

# Points vector of specimens of species in the plant genus Dypsis
madCoast0 <- fastData("madCoast0")

# Convert sf to a GVector:
coast <- fast(madCoast0)

### grid

# grid specified by number of cells in x-dimension
g1 <- grid(coast, nx = 10)
plot(coast, col = "cornflowerblue")
plot(g1, add = TRUE)

# grid specified by number of cells in x- and y-dimension
g2 <- grid(coast, nx = 10, ny = 5)
plot(coast, col = "cornflowerblue")
plot(g2, add = TRUE)

# grid specified by size of cells in both dimensions
g3 <- grid(coast, nx = 1250, ny = 2000, use = "size")
plot(coast, col = "cornflowerblue")
plot(g3, add = TRUE)

### hexagons

hexes <- hexagons(coast, ny = 10)
plot(hexes)
plot(coast, lwd = 2, add = TRUE)

hexes <- hexagons(coast, ny = 10, expand = c(0.3, 0.1))
plot(hexes)
plot(coast, lwd = 2, add = TRUE)

}
}
\seealso{
\code{\link[=hexagons]{hexagons()}}, tool \code{v.mkgrid} in \strong{GRASS}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/names.r
\name{names,GRaster-method}
\alias{names,GRaster-method}
\alias{names}
\alias{names<-,GRaster-method}
\alias{names<-}
\alias{names,GVector-method}
\alias{names<-,GVector-method}
\title{Name(s) of a GRaster or columns of a GVector's data table}
\usage{
\S4method{names}{GRaster}(x)

\S4method{names}{GRaster}(x) <- value

\S4method{names}{GVector}(x)

\S4method{names}{GVector}(x) <- value
}
\arguments{
\item{x}{A \code{GRaster} or \code{GVector}.}

\item{value}{Character: Name(s) to assign to the raster(s).}
}
\value{
Character vector.
}
\description{
\code{names()} returns that names(s) of a \code{GRaster} or of columns of a \code{GVector}'s data table'.
}
\details{
\verb{names(value) <-} assigns a new name to the \code{GRaster} or to the columns of a \code{GVector}'s data table.
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data
madElev <- fastData("madElev")
madForest2000 <- fastData("madForest2000")

# Convert SpatRasters to GRasters
elev <- fast(madElev)
forest <- fast(madForest2000)

### GRaster properties

# plotting
plot(elev)

# dimensions
dim(elev) # rows, columns, depths, layers
nrow(elev) # rows
ncol(elev) # columns
ndepth(elev) # depths
nlyr(elev) # layers

res(elev) # resolution (2D)
res3d(elev) # resolution (3D)
zres(elev) # vertical resolution
xres(elev) # vertical resolution
yres(elev) # vertical resolution
zres(elev) # vertical resolution (NA because this is a 2D GRaster)

# cell counts
ncell(elev) # cells
ncell3d(elev) # cells (3D rasters only)

# number of NA and non-NA cells
nacell(elev)
nonnacell(elev)

# topology
topology(elev) # number of dimensions
is.2d(elev) # is it 2-dimensional?
is.3d(elev) # is it 3-dimensional?

minmax(elev) # min/max values

# "names" of the object
names(elev)

# coordinate reference system
crs(elev)
st_crs(elev)
coordRef(elev)

# extent (bounding box)
ext(elev)

# vertical extent (not defined for this raster)
zext(elev)

# data type
datatype(elev) # fasterRaster type
datatype(elev, "GRASS") # GRASS type
datatype(elev, "terra") # terra type
datatype(elev, "GDAL") # GDAL type

is.integer(elev)
is.float(elev)
is.double(elev)
is.factor(elev)

# convert data type
as.int(elev) # integer; note that "elev" is already of type "integer"
as.float(elev) # floating-precision
as.doub(elev) # double-precision

# assigning
pie <- elev
pie[] <- pi # assign all cells to the value of pi
pie

# concatenating multiple GRasters
rasts <- c(elev, forest)
rasts

# subsetting
rasts[[1]]
rasts[["madForest2000"]]

# replacing
rasts[[2]] <- 2 * forest
rasts

# adding layers
rasts[[3]] <- elev > 500 # add a layer
rasts <- c(rasts, sqrt(elev)) # add another
add(rasts) <- ln(elev)
rasts

# names
names(rasts)
names(rasts) <- c("elev_meters", "2_x_forest", "high_elevation", "sqrt_elev", "ln_elev")
rasts

# remove a layer
rasts[["2_x_forest"]] <- NULL
rasts

# number of layers
nlyr(rasts)

# correlation and covariance matrices
madLANDSAT <- fastData("madLANDSAT")
landsat <- fast(madLANDSAT) # projects matrix
layerCor(landsat) # correlation
layerCor(landsat, fun = 'cov') # covariance

}
}
\seealso{
\code{\link[terra:names]{terra::names()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/c.r
\name{c,GRaster-method}
\alias{c,GRaster-method}
\alias{c}
\title{"Stack" GRasters}
\usage{
\S4method{c}{GRaster}(x, ...)
}
\arguments{
\item{x}{A \code{GRaster} or a \code{GVector}.}

\item{...}{One or more \code{GRaster}s, one or more \code{GVector}s, a list of \code{GRaster}s, or a list of \code{GVector}s. You can use a mix of lists and individual rasters or vectors.}
}
\value{
A \code{GRaster}.
}
\description{
\code{GRaster}s can be "stacked" using this function, effectively creating a multi-layered raster. This is different from creating a 3-dimensional raster, though such an effect can be emulated using stacking. \code{GVector}s can be combined into a single vector.  Stacks can only be created when:
\itemize{
\item All objects are the same class (either all \code{GRaster}s or all \code{GVector}s).
\item All objects have the same coordinate reference system (see crs()).
\item Horizontal extents are the same (see \code{\link[=ext]{ext()}}).
\item Horizontal dimensions are the same (see \code{\link[=res]{res()}}).
\item The topology (2- or 3-dimensional) must be the same. If 3D, then all rasters must have the same number of depths and vertical extents (see \code{\link[=topology]{topology()}}).
}

Data tables associated with \code{GVector}s will be combined if each vector has a table and if each table has the same columns and data types. Otherwise, the data table will be combined using \code{\link[=merge]{merge()}}.
}
\examples{
if (grassStarted()) {

# Setup
madForest2000 <- fastData("madForest2000")
madForest2014 <- fastData("madForest2014")

# Convert SpatRasters to GRasters:
forest2000 <- fast(madForest2000)
forest2014 <- fast(madForest2014)

# Combine:
forest <- c(forest2000, forest2014)
forest

nlyr(forest)

}
}
\seealso{
\link{add<-}, \code{\link[terra:c]{terra::c()}}, \code{\link[terra]{add<-}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rename.r
\name{.rename}
\alias{.rename}
\title{Rename a raster or vector in an existing GRASS session}
\usage{
.rename(from, to, rastOrVect = NULL)
}
\arguments{
\item{from, to}{\code{sources} of the raster or vector to rename.}

\item{rastOrVect}{Either \code{NULL} (default), \code{"raster"}, or \code{"vector"}. This specifies the type of object to be renamed. Partial matching is allowed. If left as \code{NULL} (default), the function will try to identify if the object is a raster or vector, and return an error if there is both a raster and vector of given name. Note that unlike in \strong{R}, \strong{GRASS} can have rasters and vector"s with the same name.}
}
\value{
The function invisibly returns \code{TRUE} if the desired rasters and/or vectors were named, and \code{FALSE} if raster and/or vector to be renamed did not exist in the \code{GRASS} session. Notably, a raster or vector or both are renamed in an existing \code{GRASS} session.
}
\description{
Rename a raster or vector in an existing \strong{GRASS} session.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/00d_GRaster_class.r
\name{makeGRaster}
\alias{makeGRaster}
\title{Create a GRaster from a GRASS raster}
\usage{
makeGRaster(src, names = "raster", levels = "", ac = NULL, fail = TRUE)
}
\arguments{
\item{src}{Character: The name of the raster in \strong{GRASS}.}

\item{names}{Character: Name of the raster (cf. \code{\link[=names]{names()}}).}

\item{levels}{\code{NULL} (default), a \code{data.frame}, \code{data.table}, an empty string (\code{""}), or a list of \code{data.frame}s, \code{data.table}s, and/or empty strings: These become the raster's \code{\link[=levels]{levels()}}. If \code{""}, then no levels are defined.}

\item{ac}{Vector of numeric/integer values >=1, or \code{NULL} (default): Active category column (offset by 1, so 1 really means the second column, 2 means the third, etc.). A value of \code{NULL} uses an automated procedure to figure it out.}

\item{fail}{Logical: If \code{TRUE} (default), and the raster either has a 0 east-west or north-south extent, then exit the function with an error. If \code{fail} is \code{FALSE}, then display a warning and return \code{NULL}.}
}
\value{
A \code{GRaster}.
}
\description{
Create a \code{GRaster} from a raster existing in the current \strong{GRASS} session.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)
library(terra)

# Example data
madElev <- fastData("madElev") # integer raster
madCover <- fastData("madCover") # categorical raster
madCoast4 <- fastData("madCoast4") # polygons vector
madRivers <- fastData("madRivers") # lines vector
madDypsis <- fastData("madDypsis") # points vector

### Create GRasters from SpatRasters

# Create an integer raster:
elev <- fast(madElev)
elev

# Create a categorical raster:
cover <- fast(madCover)
madCover
levels(madCover) # category levels

# Create a GRaster from a file on disk:
rastFile <- system.file("extdata", "madForest2000.tif", package = "fasterRaster")
forest2000 <- fast(rastFile)
forest2000

# Create a 1's raster that spans the world:
ones <- fast(rastOrVect = "raster", crs = "epsg:4326")
ones

### Create GVectors

# Create a GVector from an sf vector:
coast4 <- fast(madCoast4)
coast4

# Create a GVector from a SpatVector:
madRivers <- vect(madRivers)
class(madRivers)
rivers <- fast(madRivers)
rivers

# Create a GVector from a vector on disk:
vectFile <- system.file("extdata/shapes", "madCoast.shp",
   package = "fasterRaster")
coast0 <- fast(vectFile)
coast0

# Import only Dypsis occurrences in a restricted area:
ant <- coast4[coast4$NAME_4 == "Antanambe"]
dypsisRestrict <- fast(madDypsis, extent = ant)
dypsis <- fast(madDypsis)

plot(coast4)
plot(ant, col = "gray80", add = TRUE)
plot(dypsis, add = TRUE)
plot(dypsisRestrict, col = "red", add = TRUE)

# Create a generic GVector that spans the world:
wallToWall <- fast(rastOrVect = "vector", crs = "epsg:4326") # WGS84
wallToWall

# Create a GVector from a numeric vector
pts <- c(-90.2, 38.6, -122.3, 37.9)
pts <- fast(pts, crs = "epsg:4326") # WGS84

# Create a GVector from a matrix (can also use data.frame or data.table):
mat <- matrix(c(-90.2, 38.6, -122.3, 37.9), ncol = 2, byrow = TRUE)
mat <- fast(mat, crs = "epsg:4326", keepgeom = TRUE) # WGS84

}
}
\seealso{
\code{\link[=makeGVector]{makeGVector()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/g.proj.r
\name{.g.proj}
\alias{.g.proj}
\title{Call GRASS \code{g.proj} tool}
\usage{
.g.proj()
}
\value{
Displays current projection information for the active "project/location" in \strong{GRASS}.
}
\description{
This function calls the \strong{GRASS} tool \code{g.region} to display information on the projection of the current \strong{GRASS} "project".
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mow.r
\name{mow}
\alias{mow}
\title{Remove rasters and vectors from the GRASS cache}
\usage{
mow(
  x = "unlinked",
  pos = NULL,
  type = NULL,
  keep = NULL,
  verbose = TRUE,
  ask = TRUE
)
}
\arguments{
\item{x}{Any of:
\itemize{
\item \code{"unlinked"} (default): Delete \strong{GRASS} rasters and/or vectors that are unlinked to \code{GRaster}s or \code{GVector}s in the environment in which the function was called, or the environment named in \code{pos}.
\item A \code{GRaster} or \code{GVector}: Delete the \strong{GRASS} raster or vector pointed to by this object.
\item A \code{list} of \code{GRaster}s and/or \code{GVector}s: Delete the \strong{GRASS} raster(s) and/or vector(s) pointed to by these objects.
\item \code{"*"}: Delete \emph{all} \strong{GRASS} rasters and/or vectors pointed to by objects in the environment named in \code{pos}. Only objects in \code{keep} will not be deleted.
}}

\item{pos}{Either \code{NULL} (default), or an environment. This is used only if \code{x} is \code{"unlinked"} or \code{"*"}. In that case, if \code{pos} is \code{NULL}, the environment in which this function was called will be searched for \code{GRaster}s and \code{GVector}s for removal of their associated \strong{GRASS} rasters and vectors. Otherwise, the named environment will be searched.}

\item{type}{Either \code{NULL} or a character vector. This is used only if \code{x} is \code{"unlinked"} or \code{"*"}. If \code{NULL}, all rasters and vectors in the \strong{GRASS} cache are candidates for deletion. Otherwise, this can be either \code{"rasters"}, \code{"vectors"}, or both.}

\item{keep}{Either \code{NULL} (default) or a \code{list()} of \code{GRaster}s and/or \code{GVector}s that you want to retain. This is used only if \code{x} is \code{"unlinked"} or \code{"*"}. The rasters and vectors in \strong{GRASS} pointed to by these objects will not be deleted.}

\item{verbose}{Logical: If \code{TRUE} (default), report progress.}

\item{ask}{Logical: If \code{TRUE} (default), prompt for reassurance. This is used only if \code{x} is \code{"unlinked"} or \code{"*"}.}
}
\value{
Invisibly returns a named vector with the number of rasters and vectors deleted.
}
\description{
\strong{fasterRaster} functions attempt to delete rasters and vectors in the \strong{GRASS} cache, but not all intermediate files can be removed. This function can be used to clear the cache of extraneous rasters and vectors.

Calling this function inside another function's environment and defining \code{x} as \code{"*"} can be very \strong{dangerous}, as it will detect objects outside of that environment, and thus delete any rasters/vectors outside that environment. Here is a guide:
\itemize{
\item To delete files associated with a single \code{GRaster} or \code{GVector}, use \code{mow(GRaster_to_unlink)} or \code{mow(GVector_to_unlink)}.
\item To delete files associated with more than one \code{GRaster}s and/or \code{GVector}s, provide them as a list. For example, use \code{mow(list(GRaster_to_unlink, GVector_to_unlink))}.
\item To remove all rasters, all vectors, or all rasters and vectors in the \strong{GRASS} cache that are not linked to a \code{GRaster} or \code{GVector}, use \code{mow("*")}.
\item To remove all rasters or all vectors in the \strong{GRASS} cache, use \code{mow("*", type = "rasters")} or \code{mow("*", type = "vectors")}.
\item To remove all rasters or all vectors in the \strong{GRASS} cache \emph{except} for certain ones, use \code{mow("*", keep = list(GRaster_to_keep, GVector_to_keep))}. You can combine this with the \code{keep} argument to retain specific rasters or vectors. For example, you can use \code{mow("*", type = "rasters", keep = list(GRaster_to_keep))}.
}
}
\examples{
if (grassStarted()) {

# Setup
madElev <- fastData("madElev")
elev <- fast(madElev)

mow(elev, ask = TRUE) # delete GRASS raster attached to `elev`

}
}
\seealso{
\code{\link[terra:tmpFile]{terra::tmpFiles()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/copyGSpatial.r
\name{.copyGSpatial,GRaster-method}
\alias{.copyGSpatial,GRaster-method}
\alias{.copyGSpatial}
\alias{.copyGSpatial,GVector-method}
\alias{.copyGSpatial,character-method}
\alias{.copyGRaster}
\alias{.copyGVector}
\title{Make a copy of an object in GRASS}
\usage{
\S4method{.copyGSpatial}{GRaster}(x, reshapeRegion = TRUE)

\S4method{.copyGSpatial}{GVector}(x)

\S4method{.copyGSpatial}{character}(x, type = NULL, topo = NULL, reshapeRegion = TRUE)

.copyGRaster(x, topo = "2D", reshapeRegion = TRUE)

.copyGVector(x)
}
\arguments{
\item{x}{A \code{GVector} or the \code{\link[=sources]{sources()}} name of one.}

\item{reshapeRegion}{Logical. If \code{TRUE}, \code{x} must be a \code{GRaster}.}

\item{type}{Character or \code{NULL} (default): Either "raster" or "vector". If a character, there must be one per value in \code{x}. If \code{NULL}, will attempt to auto-detect (takes longer).}

\item{topo}{"2D" or "3D"}
}
\value{
Character vector representing the \code{\link[=sources]{sources()}} names of the copied object(s), plus makes a copy of the given object(s) in \strong{GRASS}.

\code{\link[=sources]{sources()}} names of copied rasters.
}
\description{
Create a copy of a \code{GRaster} or \code{GVector} in \strong{GRASS}.  This function is used internally and is of little use to most users.  This only creates a copy of the object in the \strong{GRASS} session--to make a \code{GRaster} or \code{GVector}, \code{\link[=makeGRaster]{makeGRaster()}} or \code{\link[=makeGVector]{makeGVector()}} need to be called after making the copy. Note that if the object is multi-layered, then a copy is made of each layer.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dropRows.r
\name{dropRows,data.table-method}
\alias{dropRows,data.table-method}
\alias{dropRows}
\alias{dropRows,data.frame-method}
\alias{dropRows,matrix-method}
\title{Remove rows in a data.table, data.frame, or matrix.}
\usage{
\S4method{dropRows}{data.table}(x, drops)

\S4method{dropRows}{data.frame}(x, drops)

\S4method{dropRows}{matrix}(x, drops)
}
\arguments{
\item{x}{A \code{data.table} or \code{data.frame}.}

\item{drops}{Numeric, integer, or logical vector: Indices or indicators of rows to remove.

If a logical vector is supplied, rows that correspond to \code{TRUE} will be removed. If the vector is shorter than the number of rows, values of \code{drops} will be recycled.}
}
\value{
A \code{data.table} or \code{data.frame}.
}
\description{
As of September of 2023, the \strong{data.table} package does not have a function for removing rows by index. This function does this job.
}
\examples{

library(data.table)

dt <- data.table(
   x = 1:10,
   y = letters[1:10],
   z = rnorm(10)
)

# make some values NA
dt[x == 4 | x == 8, y := NA_character_]
dt

# Replace NAs:
replaceNAs(dt, replace = -99, cols = "y")
dt

# Drop rows:
dropped <- dropRows(dt, 8:10)
dropped

# NB May not print... in that case, use:
print(dropped)

# We can also use replaceNAs() on vectors:
y <- 1:10
y[c(2, 10)] <- NA
replaceNAs(y, -99)

# Same as:
y <- 1:10
y[c(2, 10)] <- NA
y[is.na(y)] <- -99
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotRGB.r
\name{plotRGB,GRaster-method}
\alias{plotRGB,GRaster-method}
\alias{plotRGB}
\title{Create red-green-blue plot from a raster with RGB layers}
\usage{
\S4method{plotRGB}{GRaster}(x, r = 1, g = 2, b = 3, a = NULL, simplify = TRUE, ...)
}
\arguments{
\item{x}{A \code{GRaster}. Values must be in the range from 0 to 255.}

\item{r, g, b}{Either a numeric integer or the \code{\link[=names]{names()}} of layers representing red, green, and blue components.}

\item{a}{Either \code{NULL} (default), or a numeric integer or the \code{\link[=names]{names()}} of a layer representing transparency.}

\item{simplify}{Logical: If \code{TRUE} (default), then downsample the \code{GRaster} before plotting. This can save time for very dense rasters.}

\item{...}{Arguments to pass to \code{\link[terra:plotRGB]{terra::plotRGB()}}.}
}
\value{
Nothing (makes a plot).
}
\description{
This function takes as its main argument a \code{GRaster} with at least three layers typically representing red, green, and blue components (plus possibly an "alpha", or transparency layer). As with \code{\link[=plot]{plot()}}, this function is somewhat of a hack in that it downsamples the layers to a coarser resolution using \code{\link[=aggregate]{aggregate()}}, saves the raster to disk, then uses \code{\link[terra:plotRGB]{terra::plotRGB()}} to do the actual plotting.
}
\examples{
if (grassStarted()) {

# Example data
madElev <- fastData("madElev") # elevation raster
madLANDSAT <- fastData("madLANDSAT") # multi-layer raster
madRivers <- fastData("madRivers") # lines vector

# Convert SpatRaster to GRaster and SpatVector to GVector
elev <- fast(madElev)
rivers <- fast(madRivers)
landsat <- fast(madLANDSAT)

# Plot:
plot(elev)
plot(rivers, add = TRUE)

# Histograms:
hist(elev)
hist(landsat)

# Plot surface reflectance in RGB:
plotRGB(landsat, 3, 2, 1) # "natural" color
plotRGB(landsat, 4, 1, 2, stretch = "lin") # emphasize near-infrared (vegetation)

# Make composite map from RGB layers and plot in grayscale:
comp <- compositeRGB(r = landsat[[3]], g = landsat[[2]], b = landsat[[1]])
grays <- paste0("gray", 0:100)
plot(comp, col = grays)

}
}
\seealso{
\code{\link[terra:plotRGB]{terra::plotRGB()}}, \code{\link[=plot]{plot()}}, \code{\link[=compositeRGB]{compositeRGB()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/layerIndex.r
\name{.layerIndex}
\alias{.layerIndex}
\title{Get index of raster layers}
\usage{
.layerIndex(layer, x, recycle = TRUE, negate = FALSE)
}
\arguments{
\item{layer}{Integer, numeric, logical, or character: Refers to one or more layers.}

\item{x}{A \code{GRaster}.}

\item{recycle}{Logical: If \code{TRUE} (default), and \code{layer} is logical and smaller in number than the number of layers, then recycle the vector of \code{layer}.}

\item{negate}{Logical: If \code{TRUE}, return indices of all layers \emph{not} identified in \code{layer}.}
}
\value{
An integer vector.
}
\description{
Get index of raster layers
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cbind.r
\name{colbind,GVector-method}
\alias{colbind,GVector-method}
\alias{colbind}
\title{Add columns to the data table of a GVector}
\usage{
\S4method{colbind}{GVector}(x, ...)
}
\arguments{
\item{x, ...}{The first argument must be a \code{GVector}. Subsequent arguments can be \code{data.frame}s, \code{data.table}s, \code{matrices}, or \code{GVector}s. Only the data tables of subsequent \code{GVector}s are added to the table in \code{x}; the geometries are ignored.}
}
\value{
A \code{GVector}.
}
\description{
\code{colbind()} adds columns to the data table of a \code{GVector}. You can combine multiple a \code{GVector}'s data table with \code{data.frame}s, \code{data.table}s, \code{matrices}, or the data table(s) from other \code{GVector}(s). To combine two \code{GVector}s, see \code{\link[=rbind]{rbind()}}.
}
\examples{
if (grassStarted()) {

# Setup
library(sf)

# Rivers vector
madRivers <- fastData("madRivers")

# Convert sf to a GVector
rivers <- fast(madRivers)

# Convert GVector to data.frame or data.table
as.data.frame(rivers)
as.data.table(rivers)

# Subset rivers vector
rivers1 <- rivers[1:2]
rivers2 <- rivers[10:11]

# Concatenate rivers
riversCombo <- rbind(rivers1, rivers2)
riversCombo

# Add columns
newCol <- data.frame(new = 1:11)
riversCol <- colbind(rivers, newCol)
riversCol

# Remove table
riversCopy <- rivers
riversCopy # has data table
riversCopy <- dropTable(riversCopy)
riversCopy # no data table

# Add a new table
newTable <- data.frame(num = 1:11, letters = letters[1:11])
addTable(riversCopy) <- newTable
riversCopy

}
}
\seealso{
\code{\link[=rbind]{rbind()}}, \code{\link[fasterRaster]{addTable<-}}, \code{\link[=dropTable]{dropTable()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compareGeom.r
\name{compareGeom,GRaster,GRaster-method}
\alias{compareGeom,GRaster,GRaster-method}
\alias{compareGeom}
\alias{compareGeom,GVector,GVector-method}
\alias{compareGeom,GRaster,GVector-method}
\alias{compareGeom,GVector,GRaster-method}
\title{Determine if GRasters and/or GVectors are geographically comparable}
\usage{
\S4method{compareGeom}{GRaster,GRaster}(
  x,
  y,
  ...,
  location = TRUE,
  mapset = TRUE,
  topo = TRUE,
  lyrs = FALSE,
  crs = TRUE,
  ext = TRUE,
  zext = TRUE,
  rowcol = TRUE,
  depths = TRUE,
  res = TRUE,
  zres = TRUE,
  stopOnError = TRUE,
  messages = TRUE
)

\S4method{compareGeom}{GVector,GVector}(
  x,
  y,
  ...,
  location = TRUE,
  mapset = TRUE,
  topo = FALSE,
  crs = TRUE,
  ext = FALSE,
  zext = FALSE,
  geometry = FALSE,
  stopOnError = TRUE,
  messages = TRUE
)

\S4method{compareGeom}{GRaster,GVector}(
  x,
  y,
  ...,
  location = TRUE,
  mapset = TRUE,
  topo = FALSE,
  crs = TRUE,
  ext = FALSE,
  zext = FALSE,
  stopOnError = TRUE,
  messages = TRUE
)

\S4method{compareGeom}{GVector,GRaster}(
  x,
  y,
  ...,
  location = TRUE,
  mapset = TRUE,
  topo = FALSE,
  crs = TRUE,
  ext = FALSE,
  zext = FALSE,
  stopOnError = TRUE,
  messages = TRUE
)
}
\arguments{
\item{x, y, ...}{\code{GRaster}s or \code{GVector}s. If \code{y} is \code{GRaster}, then the \code{...} must also be \code{GRaster}s (or missing). If \code{y} is \code{GVector}, then the \code{...} must also be \code{GVector}s (or missing).}

\item{location, mapset}{Logical: Compare \strong{GRASS} "project/location" and "mapsets" (see \code{vignette("projects_mapsets", package = "fasterRaster")}). Default is \code{TRUE}.}

\item{topo}{Logical: Test for same topology (2D or 3D). By default, this is \code{TRUE} for raster-raster comparisons, and \code{FALSE} for all others.}

\item{lyrs}{Logical (rasters only): Compare number of layers of "stacked" rasters. Note this is different from number of vertical "depths" of a raster. Default is \code{FALSE}.}

\item{crs}{Logical: Compare coordinate reference systems. Default is \code{TRUE}.}

\item{ext}{Logical: If \code{TRUE}, test for same extent. By default, is \code{TRUE} for raster-raster comparison and \code{FALSE} for all others.}

\item{zext}{Logical: Test for same vertical extents (3D only). By default, this is \code{TRUE} for raster-raster comparisons, and \code{FALSE} for all others.}

\item{rowcol}{Logical (rasters only): Test for same number of rows and columns. Default is \code{TRUE}.}

\item{depths}{Logical (rasters only): Test for same number of depths. Default is \code{TRUE}.}

\item{res}{Logical (rasters only): Test for same resolution in x- and y-dimensions. Default is \code{TRUE}.}

\item{zres}{Logical (rasters only): Test for same resolution in z dimension. Default is \code{TRUE}.}

\item{stopOnError}{Logical: If \code{TRUE} (default), throw an error with an explanation if the objects are not comparable. If \code{FALSE} (default), return \code{TRUE} or \code{FALSE}.}

\item{messages}{Logical: If \code{TRUE} (default), display a warning if a condition is not met. This only comes into effect if \code{stopOnError} is \code{FALSE}.}

\item{geometry}{Logical (vector-vector comparison only): Compare geometry. Default is \code{FALSE}.}
}
\value{
Logical (invisibly): \code{TRUE} for no mismatches detected, \code{FALSE} for incompatibility), or side effect of throwing an error.
}
\description{
\code{compareGeom()} compares geographic metadata between two or more \code{GRaster}s and/or \code{GVector}s. In many cases, spatial objects must be comparable for them to "interact" (e.g., conducting arithmetic operations, masking, etc.).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zonalGeog.r
\name{zonalGeog,GRaster-method}
\alias{zonalGeog,GRaster-method}
\alias{zonalGeog}
\title{Geographic statistics for sets of cells with the same values}
\usage{
\S4method{zonalGeog}{GRaster}(x, unit = "meters")
}
\arguments{
\item{x}{A \code{GRaster}.}

\item{unit}{Character: Units of the output. Any of:
\itemize{
\item \code{"meters"} (default)
\item \code{"kilometers"} or \code{"km"}
\item \code{"miles"} or \code{"mi"}
\item \code{"yards"} or \code{"yd"}
\item \code{"feet"} or \code{"ft"}: International foot; 1 foot exactly equal to 0.3048 meters
\item \code{"cells"}: Number or cells
}

Partial matching is used and case is ignored.}
}
\value{
A list of \code{data.frame}s or a \code{data.table}s, one per layer in \code{x}. Only layers that are integers or factors have their geographies calculated. Other layers have \code{NULL} tables returned.
}
\description{
This function calculates geographic statistics for each set of cells in an \code{integer} or \code{factor} \code{GRaster}. Statistics include:
\itemize{
\item Area
\item Perimeter length
\item "Compact square" statistic: \eqn{4 \sqrt(area) / perimeter)}
\item "Compact circle" statistic: \eqn{4 * P / ( 2 \sqrt(\pi * A))} where \emph{P} is the perimeter length and \emph{A} the area.
\item fractal dimension: \eqn{2 ( log(P) / log(A + 0.001))} where \emph{P} is perimeter length and \emph{A} is area.
\item The average x- and y-coordinates of each zone.
}
}
\examples{
if (grassStarted()) {

# Setup
library(terra)

# Example data: Elevation and land cover
madElev <- fastData("madElev")
madForest2000 <- fastData("madForest2000")
madCover <- fastData("madCover")

# Convert to GRasters:
elev <- fast(madElev)
forest2000 <- fast(madForest2000)
cover <- fast(madCover)

# Rename
names(elev) <- "elev"
names(forest2000) <- "forest"

# Geometric statistics for an integer raster zoned by elevation:
fun <-
  "= if (elev <400 & forest == 1, 0, if (elev >=400 & forest == 1, 1, null()))"
forestByElev <- app(c(elev, forest2000), fun = fun)
plot(forestByElev, main = "forest < 400 m & >= 400 m")
zonalGeog(forestByElev)

# Geometric statistics for a categorical raster:
zonalGeog(cover)

}
}
